/* MatMult test */

#include <oski/common.h>
#include <oski/matcreate.h>
#include <oski/matmult.h>
#include "util/array_util.h"

/* ------------------------------------------------- */

#define A0_M 4    /* # of rows */
#define A0_N 5    /* # of cols */
#define A0_NNZ 8  /* # of stored non-zeros */
static oski_index_t A0_ptr[A0_M+1] = {1, 4, 7, 8, 9}; /* end row pointers for A0 */
static oski_index_t A0_ind[A0_NNZ] = {1, 2, 5, 3, 4, 5, 5, 5}; /* end column indices pointers for A0 */
static oski_value_t A0_val[A0_NNZ] = {MAKE_VAL_COMPLEX(4, -1), MAKE_VAL_COMPLEX(-10, -1), MAKE_VAL_COMPLEX(10, -2), MAKE_VAL_COMPLEX(-9, 10), MAKE_VAL_COMPLEX(-1, -10), MAKE_VAL_COMPLEX(8, -7), MAKE_VAL_COMPLEX(-3, -4), MAKE_VAL_COMPLEX(6, -2)}; /* end stored non-zero values for A0 */

#define x0_LEN 5   /* vector length */
#define x0_VW  3   /* vector width */
static oski_value_t x0_val[x0_LEN*x0_VW] = {MAKE_VAL_COMPLEX(9, -8), MAKE_VAL_COMPLEX(9, -7), MAKE_VAL_COMPLEX(-2, -8), MAKE_VAL_COMPLEX(-10, 10), MAKE_VAL_COMPLEX(3, 10), MAKE_VAL_COMPLEX(-2, 10), MAKE_VAL_COMPLEX(8, -1), MAKE_VAL_COMPLEX(0, -2), MAKE_VAL_COMPLEX(-4, -7), MAKE_VAL_COMPLEX(4, -4), MAKE_VAL_COMPLEX(10, -8), MAKE_VAL_COMPLEX(-6, -10), MAKE_VAL_COMPLEX(-5, -7), MAKE_VAL_COMPLEX(10, 8), MAKE_VAL_COMPLEX(6, 8)}; /* data */

#define y0_LEN 4   /* vector length */
#define y0_VW  3   /* vector width */
static oski_value_t y0_val[y0_LEN*y0_VW] = {MAKE_VAL_COMPLEX(-9, 6), MAKE_VAL_COMPLEX(1, 10), MAKE_VAL_COMPLEX(9, -5), MAKE_VAL_COMPLEX(1, -8), MAKE_VAL_COMPLEX(1, -10), MAKE_VAL_COMPLEX(10, 3), MAKE_VAL_COMPLEX(1, 1), MAKE_VAL_COMPLEX(-10, 3), MAKE_VAL_COMPLEX(-6, 0), MAKE_VAL_COMPLEX(-3, -3), MAKE_VAL_COMPLEX(-6, 9), MAKE_VAL_COMPLEX(-6, 8)}; /* data */

/*
 *  BEGIN TEST 0
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_0( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	oski_value_t* y_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(182, -185), MAKE_VAL_COMPLEX(177, -166), MAKE_VAL_COMPLEX(-12, -74), MAKE_VAL_COMPLEX(-251, 90), MAKE_VAL_COMPLEX(-2, -6), MAKE_VAL_COMPLEX(54, 31), MAKE_VAL_COMPLEX(-15, 27), MAKE_VAL_COMPLEX(100, -28), MAKE_VAL_COMPLEX(62, -12), MAKE_VAL_COMPLEX(-38, 10), MAKE_VAL_COMPLEX(178, -8), MAKE_VAL_COMPLEX(148, 8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_0, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_0", y_true, y_val, y0_LEN*y0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_0() */
/* END TEST 0 */

/*
 *  BEGIN TEST 1
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	oski_value_t* y_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-128, 227), MAKE_VAL_COMPLEX(-119, 140), MAKE_VAL_COMPLEX(-36, 30), MAKE_VAL_COMPLEX(201, -80), MAKE_VAL_COMPLEX(-60, 20), MAKE_VAL_COMPLEX(-56, -97), MAKE_VAL_COMPLEX(19, -35), MAKE_VAL_COMPLEX(-62, 82), MAKE_VAL_COMPLEX(-50, 48), MAKE_VAL_COMPLEX(26, 14), MAKE_VAL_COMPLEX(-112, 26), MAKE_VAL_COMPLEX(-88, 12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1", y_true, y_val, y0_LEN*y0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1() */
/* END TEST 1 */

/*
 *  BEGIN TEST 2
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_2( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	oski_value_t* y_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-30, 111), MAKE_VAL_COMPLEX(78, 73), MAKE_VAL_COMPLEX(37, -103), MAKE_VAL_COMPLEX(-48, -71), MAKE_VAL_COMPLEX(-62, -87), MAKE_VAL_COMPLEX(101, -46), MAKE_VAL_COMPLEX(15, 1), MAKE_VAL_COMPLEX(-59, 94), MAKE_VAL_COMPLEX(-48, 42), MAKE_VAL_COMPLEX(-45, -3), MAKE_VAL_COMPLEX(15, 114), MAKE_VAL_COMPLEX(8, 106) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_2, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_2", y_true, y_val, y0_LEN*y0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_2() */
/* END TEST 2 */

/*
 *  BEGIN TEST 3
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_3( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	oski_value_t* y_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1138, -843), MAKE_VAL_COMPLEX(-241, -1052), MAKE_VAL_COMPLEX(-284, -808), MAKE_VAL_COMPLEX(-223, 1430), MAKE_VAL_COMPLEX(-499, -619), MAKE_VAL_COMPLEX(1048, -655), MAKE_VAL_COMPLEX(267, 165), MAKE_VAL_COMPLEX(-413, -75), MAKE_VAL_COMPLEX(-302, -152), MAKE_VAL_COMPLEX(-330, 268), MAKE_VAL_COMPLEX(417, -443), MAKE_VAL_COMPLEX(418, -258) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_3, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_3", y_true, y_val, y0_LEN*y0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_3() */
/* END TEST 3 */

/*
 *  BEGIN TEST 4
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_4( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-104, -14), MAKE_VAL_COMPLEX(-53, -10), MAKE_VAL_COMPLEX(-21, 1), MAKE_VAL_COMPLEX(186, -1), MAKE_VAL_COMPLEX(64, -142), MAKE_VAL_COMPLEX(-21, 35), MAKE_VAL_COMPLEX(48, 28), MAKE_VAL_COMPLEX(77, 104), MAKE_VAL_COMPLEX(-161, 115), MAKE_VAL_COMPLEX(-117, -22), MAKE_VAL_COMPLEX(-177, -54), MAKE_VAL_COMPLEX(-38, -41), MAKE_VAL_COMPLEX(-188, 37), MAKE_VAL_COMPLEX(28, 22), MAKE_VAL_COMPLEX(223, -88) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_4, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_4", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_4() */
/* END TEST 4 */

/*
 *  BEGIN TEST 5
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_5( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(49, -66), MAKE_VAL_COMPLEX(6, -69), MAKE_VAL_COMPLEX(-45, 7), MAKE_VAL_COMPLEX(-116, 91), MAKE_VAL_COMPLEX(19, 128), MAKE_VAL_COMPLEX(99, -9), MAKE_VAL_COMPLEX(-48, -93), MAKE_VAL_COMPLEX(-93, -106), MAKE_VAL_COMPLEX(101, -90), MAKE_VAL_COMPLEX(89, -14), MAKE_VAL_COMPLEX(123, -34), MAKE_VAL_COMPLEX(-48, 79), MAKE_VAL_COMPLEX(127, -4), MAKE_VAL_COMPLEX(46, -94), MAKE_VAL_COMPLEX(-153, 48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_5, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_5", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_5() */
/* END TEST 5 */

/*
 *  BEGIN TEST 6
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_6( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-5, -60), MAKE_VAL_COMPLEX(-1, -57), MAKE_VAL_COMPLEX(-38, -16), MAKE_VAL_COMPLEX(10, 70), MAKE_VAL_COMPLEX(49, 18), MAKE_VAL_COMPLEX(34, 38), MAKE_VAL_COMPLEX(20, -35), MAKE_VAL_COMPLEX(-8, -6), MAKE_VAL_COMPLEX(-40, -5), MAKE_VAL_COMPLEX(-4, -28), MAKE_VAL_COMPLEX(-2, -64), MAKE_VAL_COMPLEX(-58, -6), MAKE_VAL_COMPLEX(-43, -1), MAKE_VAL_COMPLEX(62, -16), MAKE_VAL_COMPLEX(50, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_6, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_6", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_6() */
/* END TEST 6 */

/*
 *  BEGIN TEST 7
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_7( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-36, -124), MAKE_VAL_COMPLEX(-36, -75), MAKE_VAL_COMPLEX(-29, 3), MAKE_VAL_COMPLEX(61, 206), MAKE_VAL_COMPLEX(176, 32), MAKE_VAL_COMPLEX(9, -33), MAKE_VAL_COMPLEX(-48, 18), MAKE_VAL_COMPLEX(-112, 81), MAKE_VAL_COMPLEX(-135, -131), MAKE_VAL_COMPLEX(-2, -125), MAKE_VAL_COMPLEX(2, -201), MAKE_VAL_COMPLEX(13, 6), MAKE_VAL_COMPLEX(-55, -154), MAKE_VAL_COMPLEX(-10, -28), MAKE_VAL_COMPLEX(108, 183) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_7, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_7", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_7() */
/* END TEST 7 */

/*
 *  BEGIN TEST 8
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_8( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-135, 1), MAKE_VAL_COMPLEX(-93, 22), MAKE_VAL_COMPLEX(3, 41), MAKE_VAL_COMPLEX(194, -59), MAKE_VAL_COMPLEX(25, -167), MAKE_VAL_COMPLEX(-15, 21), MAKE_VAL_COMPLEX(-135, 19), MAKE_VAL_COMPLEX(-121, 90), MAKE_VAL_COMPLEX(-82, -68), MAKE_VAL_COMPLEX(35, 14), MAKE_VAL_COMPLEX(1, 0), MAKE_VAL_COMPLEX(-70, 183), MAKE_VAL_COMPLEX(-74, 27), MAKE_VAL_COMPLEX(30, -78), MAKE_VAL_COMPLEX(143, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_8, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_8", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_8() */
/* END TEST 8 */

/*
 *  BEGIN TEST 9
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_9( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(76, -13), MAKE_VAL_COMPLEX(38, -37), MAKE_VAL_COMPLEX(-29, -9), MAKE_VAL_COMPLEX(-124, 69), MAKE_VAL_COMPLEX(6, 125), MAKE_VAL_COMPLEX(61, -43), MAKE_VAL_COMPLEX(107, -48), MAKE_VAL_COMPLEX(113, -84), MAKE_VAL_COMPLEX(66, 105), MAKE_VAL_COMPLEX(-63, -18), MAKE_VAL_COMPLEX(-63, -16), MAKE_VAL_COMPLEX(48, -129), MAKE_VAL_COMPLEX(61, 14), MAKE_VAL_COMPLEX(-28, 14), MAKE_VAL_COMPLEX(-129, -46) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_9, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_9", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_9() */
/* END TEST 9 */

/*
 *  BEGIN TEST 10
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_10( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-56, -63), MAKE_VAL_COMPLEX(-49, -63), MAKE_VAL_COMPLEX(-56, 14), MAKE_VAL_COMPLEX(70, 70), MAKE_VAL_COMPLEX(70, -21), MAKE_VAL_COMPLEX(70, 14), MAKE_VAL_COMPLEX(-7, -56), MAKE_VAL_COMPLEX(-14, 0), MAKE_VAL_COMPLEX(-49, 28), MAKE_VAL_COMPLEX(-28, -28), MAKE_VAL_COMPLEX(-56, -70), MAKE_VAL_COMPLEX(-70, 42), MAKE_VAL_COMPLEX(-49, 35), MAKE_VAL_COMPLEX(56, -70), MAKE_VAL_COMPLEX(56, -42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_10, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_10", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_10() */
/* END TEST 10 */

/*
 *  BEGIN TEST 11
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_11( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(490, 38), MAKE_VAL_COMPLEX(272, -273), MAKE_VAL_COMPLEX(-307, -91), MAKE_VAL_COMPLEX(-1009, 270), MAKE_VAL_COMPLEX(-202, 926), MAKE_VAL_COMPLEX(767, -183), MAKE_VAL_COMPLEX(946, -170), MAKE_VAL_COMPLEX(1128, -323), MAKE_VAL_COMPLEX(135, 1129), MAKE_VAL_COMPLEX(-526, -369), MAKE_VAL_COMPLEX(-618, -417), MAKE_VAL_COMPLEX(661, -754), MAKE_VAL_COMPLEX(373, 400), MAKE_VAL_COMPLEX(-204, -152), MAKE_VAL_COMPLEX(-812, -879) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, -3);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_11, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_11", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_11() */
/* END TEST 11 */

/*
 *  BEGIN TEST 12
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_12( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	oski_value_t* y_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(44, -275), MAKE_VAL_COMPLEX(-277, 20), MAKE_VAL_COMPLEX(9, 123), MAKE_VAL_COMPLEX(18, -8), MAKE_VAL_COMPLEX(203, -52), MAKE_VAL_COMPLEX(22, -6), MAKE_VAL_COMPLEX(-8, -58), MAKE_VAL_COMPLEX(72, -58), MAKE_VAL_COMPLEX(102, -108), MAKE_VAL_COMPLEX(146, 97), MAKE_VAL_COMPLEX(-46, -114), MAKE_VAL_COMPLEX(0, -28) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_12, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_12", y_true, y_val, y0_LEN*y0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_12() */
/* END TEST 12 */

/*
 *  BEGIN TEST 13
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_13( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	oski_value_t* y_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-8, 212), MAKE_VAL_COMPLEX(241, -77), MAKE_VAL_COMPLEX(-48, -66), MAKE_VAL_COMPLEX(0, 59), MAKE_VAL_COMPLEX(-179, 115), MAKE_VAL_COMPLEX(-91, 18), MAKE_VAL_COMPLEX(-1, 55), MAKE_VAL_COMPLEX(-21, 10), MAKE_VAL_COMPLEX(-66, 90), MAKE_VAL_COMPLEX(-119, -88), MAKE_VAL_COMPLEX(55, 42), MAKE_VAL_COMPLEX(12, -38) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_13, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_13", y_true, y_val, y0_LEN*y0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_13() */
/* END TEST 13 */

/*
 *  BEGIN TEST 14
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_14( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	oski_value_t* y_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-12, -57), MAKE_VAL_COMPLEX(-52, -15), MAKE_VAL_COMPLEX(7, 55), MAKE_VAL_COMPLEX(38, 21), MAKE_VAL_COMPLEX(48, 25), MAKE_VAL_COMPLEX(-35, 44), MAKE_VAL_COMPLEX(-7, 3), MAKE_VAL_COMPLEX(5, -56), MAKE_VAL_COMPLEX(12, -30), MAKE_VAL_COMPLEX(21, -9), MAKE_VAL_COMPLEX(-33, -48), MAKE_VAL_COMPLEX(-28, -46) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_14, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_14", y_true, y_val, y0_LEN*y0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_14() */
/* END TEST 14 */

/*
 *  BEGIN TEST 15
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_15( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	oski_value_t* y_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-116, 1252), MAKE_VAL_COMPLEX(1048, -556), MAKE_VAL_COMPLEX(44, -228), MAKE_VAL_COMPLEX(378, 94), MAKE_VAL_COMPLEX(-528, 716), MAKE_VAL_COMPLEX(-234, 678), MAKE_VAL_COMPLEX(112, 396), MAKE_VAL_COMPLEX(-550, -70), MAKE_VAL_COMPLEX(-432, 504), MAKE_VAL_COMPLEX(-914, -482), MAKE_VAL_COMPLEX(-54, 298), MAKE_VAL_COMPLEX(-444, -132) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_15, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_15", y_true, y_val, y0_LEN*y0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_15() */
/* END TEST 15 */

/*
 *  BEGIN TEST 16
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_16( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-3, 9), MAKE_VAL_COMPLEX(21, -62), MAKE_VAL_COMPLEX(-30, -18), MAKE_VAL_COMPLEX(66, -21), MAKE_VAL_COMPLEX(-11, 129), MAKE_VAL_COMPLEX(54, 36), MAKE_VAL_COMPLEX(-85, -83), MAKE_VAL_COMPLEX(-120, 67), MAKE_VAL_COMPLEX(45, -24), MAKE_VAL_COMPLEX(111, -32), MAKE_VAL_COMPLEX(50, -127), MAKE_VAL_COMPLEX(-45, 3), MAKE_VAL_COMPLEX(-72, 59), MAKE_VAL_COMPLEX(68, -93), MAKE_VAL_COMPLEX(-53, 90) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_16, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_16", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_16() */
/* END TEST 16 */

/*
 *  BEGIN TEST 17
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_17( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(108, -54), MAKE_VAL_COMPLEX(81, 26), MAKE_VAL_COMPLEX(36, -60), MAKE_VAL_COMPLEX(-186, 81), MAKE_VAL_COMPLEX(8, -30), MAKE_VAL_COMPLEX(-102, 48), MAKE_VAL_COMPLEX(160, 98), MAKE_VAL_COMPLEX(126, -85), MAKE_VAL_COMPLEX(-60, -51), MAKE_VAL_COMPLEX(-63, 8), MAKE_VAL_COMPLEX(64, 85), MAKE_VAL_COMPLEX(21, -111), MAKE_VAL_COMPLEX(48, -137), MAKE_VAL_COMPLEX(-2, 195), MAKE_VAL_COMPLEX(83, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_17, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_17", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_17() */
/* END TEST 17 */

/*
 *  BEGIN TEST 18
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_18( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-7, -26), MAKE_VAL_COMPLEX(-5, -25), MAKE_VAL_COMPLEX(-18, -4), MAKE_VAL_COMPLEX(10, 30), MAKE_VAL_COMPLEX(23, 4), MAKE_VAL_COMPLEX(18, 14), MAKE_VAL_COMPLEX(6, -17), MAKE_VAL_COMPLEX(-4, -2), MAKE_VAL_COMPLEX(-18, 1), MAKE_VAL_COMPLEX(-4, -12), MAKE_VAL_COMPLEX(-6, -28), MAKE_VAL_COMPLEX(-26, 2), MAKE_VAL_COMPLEX(-19, 3), MAKE_VAL_COMPLEX(26, -12), MAKE_VAL_COMPLEX(22, -4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_18, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_18", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_18() */
/* END TEST 18 */

/*
 *  BEGIN TEST 19
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_19( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-40, 380), MAKE_VAL_COMPLEX(-291, -117), MAKE_VAL_COMPLEX(-68, 204), MAKE_VAL_COMPLEX(285, -815), MAKE_VAL_COMPLEX(364, 528), MAKE_VAL_COMPLEX(334, -342), MAKE_VAL_COMPLEX(-1000, 160), MAKE_VAL_COMPLEX(-233, 979), MAKE_VAL_COMPLEX(305, -255), MAKE_VAL_COMPLEX(371, -563), MAKE_VAL_COMPLEX(-477, -549), MAKE_VAL_COMPLEX(82, 364), MAKE_VAL_COMPLEX(157, 729), MAKE_VAL_COMPLEX(-473, -821), MAKE_VAL_COMPLEX(-75, 635) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_19, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_19", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_19() */
/* END TEST 19 */

/*
 *  BEGIN TEST 20
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_20( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1, 43), MAKE_VAL_COMPLEX(51, -10), MAKE_VAL_COMPLEX(6, -22), MAKE_VAL_COMPLEX(34, -99), MAKE_VAL_COMPLEX(-37, 123), MAKE_VAL_COMPLEX(18, -4), MAKE_VAL_COMPLEX(103, -69), MAKE_VAL_COMPLEX(-52, -129), MAKE_VAL_COMPLEX(21, 34), MAKE_VAL_COMPLEX(-81, 12), MAKE_VAL_COMPLEX(2, 129), MAKE_VAL_COMPLEX(67, -61), MAKE_VAL_COMPLEX(-126, 107), MAKE_VAL_COMPLEX(-6, 43), MAKE_VAL_COMPLEX(-159, -40) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_20, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_20", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_20() */
/* END TEST 20 */

/*
 *  BEGIN TEST 21
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_21( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(28, -67), MAKE_VAL_COMPLEX(-24, -11), MAKE_VAL_COMPLEX(-12, -2), MAKE_VAL_COMPLEX(-64, 129), MAKE_VAL_COMPLEX(46, -93), MAKE_VAL_COMPLEX(-24, 34), MAKE_VAL_COMPLEX(-79, 66), MAKE_VAL_COMPLEX(52, 123), MAKE_VAL_COMPLEX(-33, -55), MAKE_VAL_COMPLEX(93, -24), MAKE_VAL_COMPLEX(28, -153), MAKE_VAL_COMPLEX(-85, 31), MAKE_VAL_COMPLEX(111, -128), MAKE_VAL_COMPLEX(36, -19), MAKE_VAL_COMPLEX(177, 64) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_21, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_21", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_21() */
/* END TEST 21 */

/*
 *  BEGIN TEST 22
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_22( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-11, -103), MAKE_VAL_COMPLEX(-4, -98), MAKE_VAL_COMPLEX(-66, -26), MAKE_VAL_COMPLEX(20, 120), MAKE_VAL_COMPLEX(85, 29), MAKE_VAL_COMPLEX(60, 64), MAKE_VAL_COMPLEX(33, -61), MAKE_VAL_COMPLEX(-14, -10), MAKE_VAL_COMPLEX(-69, -7), MAKE_VAL_COMPLEX(-8, -48), MAKE_VAL_COMPLEX(-6, -110), MAKE_VAL_COMPLEX(-100, -8), MAKE_VAL_COMPLEX(-74, 0), MAKE_VAL_COMPLEX(106, -30), MAKE_VAL_COMPLEX(86, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_22, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_22", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_22() */
/* END TEST 22 */

/*
 *  BEGIN TEST 23
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_ROWMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_23( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_ROWMAJ, x0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(77, 290), MAKE_VAL_COMPLEX(-459, -57), MAKE_VAL_COMPLEX(-50, 276), MAKE_VAL_COMPLEX(-570, -524), MAKE_VAL_COMPLEX(867, -296), MAKE_VAL_COMPLEX(-74, -522), MAKE_VAL_COMPLEX(-1046, -585), MAKE_VAL_COMPLEX(-912, 748), MAKE_VAL_COMPLEX(434, -9), MAKE_VAL_COMPLEX(142, 804), MAKE_VAL_COMPLEX(722, 102), MAKE_VAL_COMPLEX(-320, -92), MAKE_VAL_COMPLEX(1349, 1013), MAKE_VAL_COMPLEX(-78, -254), MAKE_VAL_COMPLEX(-288, 1104) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-2, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_23, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_23", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_23() */
/* END TEST 23 */

/*
 *  BEGIN TEST 24
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_24( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	oski_value_t* y_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(5, 33), MAKE_VAL_COMPLEX(-103, -59), MAKE_VAL_COMPLEX(129, 207), MAKE_VAL_COMPLEX(336, 254), MAKE_VAL_COMPLEX(2, 70), MAKE_VAL_COMPLEX(214, -41), MAKE_VAL_COMPLEX(20, -43), MAKE_VAL_COMPLEX(17, -72), MAKE_VAL_COMPLEX(50, -78), MAKE_VAL_COMPLEX(71, 57), MAKE_VAL_COMPLEX(7, -116), MAKE_VAL_COMPLEX(48, -42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_24, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_24", y_true, y_val, y0_LEN*y0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_24() */
/* END TEST 24 */

/*
 *  BEGIN TEST 25
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_25( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	oski_value_t* y_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(70, -96), MAKE_VAL_COMPLEX(84, -30), MAKE_VAL_COMPLEX(-205, -153), MAKE_VAL_COMPLEX(-337, -181), MAKE_VAL_COMPLEX(-1, 21), MAKE_VAL_COMPLEX(-307, 24), MAKE_VAL_COMPLEX(-30, 35), MAKE_VAL_COMPLEX(70, 35), MAKE_VAL_COMPLEX(4, 72), MAKE_VAL_COMPLEX(-41, -33), MAKE_VAL_COMPLEX(38, 29), MAKE_VAL_COMPLEX(-2, -36) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_25, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_25", y_true, y_val, y0_LEN*y0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_25() */
/* END TEST 25 */

/*
 *  BEGIN TEST 26
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_26( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	oski_value_t* y_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-24, 81), MAKE_VAL_COMPLEX(56, 55), MAKE_VAL_COMPLEX(29, -75), MAKE_VAL_COMPLEX(-34, -53), MAKE_VAL_COMPLEX(-44, -65), MAKE_VAL_COMPLEX(75, -32), MAKE_VAL_COMPLEX(11, 1), MAKE_VAL_COMPLEX(-45, 68), MAKE_VAL_COMPLEX(-36, 30), MAKE_VAL_COMPLEX(-33, -3), MAKE_VAL_COMPLEX(9, 84), MAKE_VAL_COMPLEX(4, 78) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_26, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_26", y_true, y_val, y0_LEN*y0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_26() */
/* END TEST 26 */

/*
 *  BEGIN TEST 27
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_27( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	oski_value_t* y_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(193, -786), MAKE_VAL_COMPLEX(365, -16), MAKE_VAL_COMPLEX(-1162, -940), MAKE_VAL_COMPLEX(-2150, -1379), MAKE_VAL_COMPLEX(250, 1), MAKE_VAL_COMPLEX(-2051, 459), MAKE_VAL_COMPLEX(-217, 286), MAKE_VAL_COMPLEX(248, 56), MAKE_VAL_COMPLEX(-74, 360), MAKE_VAL_COMPLEX(-266, -354), MAKE_VAL_COMPLEX(-52, 212), MAKE_VAL_COMPLEX(-308, -260) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_27, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_27", y_true, y_val, y0_LEN*y0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_27() */
/* END TEST 27 */

/*
 *  BEGIN TEST 28
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_28( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-122, 2), MAKE_VAL_COMPLEX(11, -86), MAKE_VAL_COMPLEX(23, 128), MAKE_VAL_COMPLEX(29, 28), MAKE_VAL_COMPLEX(-91, -73), MAKE_VAL_COMPLEX(92, 13), MAKE_VAL_COMPLEX(-39, -153), MAKE_VAL_COMPLEX(77, 108), MAKE_VAL_COMPLEX(-134, 56), MAKE_VAL_COMPLEX(-52, 96), MAKE_VAL_COMPLEX(-65, -67), MAKE_VAL_COMPLEX(-141, 123), MAKE_VAL_COMPLEX(-149, 136), MAKE_VAL_COMPLEX(36, -205), MAKE_VAL_COMPLEX(211, -104) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_28, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_28", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_28() */
/* END TEST 28 */

/*
 *  BEGIN TEST 29
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_29( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-18, -87), MAKE_VAL_COMPLEX(-138, -3), MAKE_VAL_COMPLEX(-119, -70), MAKE_VAL_COMPLEX(141, 62), MAKE_VAL_COMPLEX(209, -6), MAKE_VAL_COMPLEX(46, -27), MAKE_VAL_COMPLEX(-6, 53), MAKE_VAL_COMPLEX(-103, -100), MAKE_VAL_COMPLEX(59, 24), MAKE_VAL_COMPLEX(-16, -132), MAKE_VAL_COMPLEX(-79, -31), MAKE_VAL_COMPLEX(35, -5), MAKE_VAL_COMPLEX(78, -43), MAKE_VAL_COMPLEX(28, 43), MAKE_VAL_COMPLEX(-131, -6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_29, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_29", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_29() */
/* END TEST 29 */

/*
 *  BEGIN TEST 30
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_30( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(129, -18), MAKE_VAL_COMPLEX(123, -9), MAKE_VAL_COMPLEX(30, -84), MAKE_VAL_COMPLEX(-150, 30), MAKE_VAL_COMPLEX(-33, 108), MAKE_VAL_COMPLEX(-78, 78), MAKE_VAL_COMPLEX(78, 39), MAKE_VAL_COMPLEX(12, -18), MAKE_VAL_COMPLEX(6, -87), MAKE_VAL_COMPLEX(60, -12), MAKE_VAL_COMPLEX(138, -12), MAKE_VAL_COMPLEX(6, -126), MAKE_VAL_COMPLEX(-3, -93), MAKE_VAL_COMPLEX(42, 132), MAKE_VAL_COMPLEX(6, 108) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_30, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_30", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_30() */
/* END TEST 30 */

/*
 *  BEGIN TEST 31
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_31( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-376, 120), MAKE_VAL_COMPLEX(763, -93), MAKE_VAL_COMPLEX(217, 821), MAKE_VAL_COMPLEX(-481, -237), MAKE_VAL_COMPLEX(-966, -576), MAKE_VAL_COMPLEX(37, 295), MAKE_VAL_COMPLEX(274, -744), MAKE_VAL_COMPLEX(327, 979), MAKE_VAL_COMPLEX(-730, -262), MAKE_VAL_COMPLEX(-412, 724), MAKE_VAL_COMPLEX(234, -136), MAKE_VAL_COMPLEX(-820, 62), MAKE_VAL_COMPLEX(-1079, 197), MAKE_VAL_COMPLEX(459, -735), MAKE_VAL_COMPLEX(1365, 273) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_31, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_31", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_31() */
/* END TEST 31 */

/*
 *  BEGIN TEST 32
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_32( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-63, -63), MAKE_VAL_COMPLEX(69, -144), MAKE_VAL_COMPLEX(-143, 50), MAKE_VAL_COMPLEX(109, 108), MAKE_VAL_COMPLEX(12, -26), MAKE_VAL_COMPLEX(48, 83), MAKE_VAL_COMPLEX(-2, -150), MAKE_VAL_COMPLEX(-121, 74), MAKE_VAL_COMPLEX(45, 23), MAKE_VAL_COMPLEX(20, -14), MAKE_VAL_COMPLEX(23, -95), MAKE_VAL_COMPLEX(-163, 65), MAKE_VAL_COMPLEX(-117, -118), MAKE_VAL_COMPLEX(38, 61), MAKE_VAL_COMPLEX(191, 62) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_32, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_32", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_32() */
/* END TEST 32 */

/*
 *  BEGIN TEST 33
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_33( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(29, 61), MAKE_VAL_COMPLEX(-101, 140), MAKE_VAL_COMPLEX(131, -30), MAKE_VAL_COMPLEX(-69, -108), MAKE_VAL_COMPLEX(2, 0), MAKE_VAL_COMPLEX(-24, -99), MAKE_VAL_COMPLEX(-16, 136), MAKE_VAL_COMPLEX(117, -70), MAKE_VAL_COMPLEX(-51, -1), MAKE_VAL_COMPLEX(-36, 14), MAKE_VAL_COMPLEX(-59, 91), MAKE_VAL_COMPLEX(155, -33), MAKE_VAL_COMPLEX(113, 142), MAKE_VAL_COMPLEX(-42, -97), MAKE_VAL_COMPLEX(-187, -90) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_33, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_33", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_33() */
/* END TEST 33 */

/*
 *  BEGIN TEST 34
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_34( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-127, -16), MAKE_VAL_COMPLEX(-119, -23), MAKE_VAL_COMPLEX(-50, 72), MAKE_VAL_COMPLEX(150, 10), MAKE_VAL_COMPLEX(59, -94), MAKE_VAL_COMPLEX(94, -54), MAKE_VAL_COMPLEX(-64, -57), MAKE_VAL_COMPLEX(-16, 14), MAKE_VAL_COMPLEX(-28, 81), MAKE_VAL_COMPLEX(-60, -4), MAKE_VAL_COMPLEX(-134, -24), MAKE_VAL_COMPLEX(-38, 118), MAKE_VAL_COMPLEX(-21, 89), MAKE_VAL_COMPLEX(-6, -136), MAKE_VAL_COMPLEX(22, -104) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_34, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_34", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_34() */
/* END TEST 34 */

/*
 *  BEGIN TEST 35
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_35( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_ROWMAJ, y0_VW );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(347, 177), MAKE_VAL_COMPLEX(-1244, -39), MAKE_VAL_COMPLEX(1118, 185), MAKE_VAL_COMPLEX(-396, -707), MAKE_VAL_COMPLEX(222, 628), MAKE_VAL_COMPLEX(397, -336), MAKE_VAL_COMPLEX(-586, 898), MAKE_VAL_COMPLEX(1422, 135), MAKE_VAL_COMPLEX(-663, -791), MAKE_VAL_COMPLEX(-146, -396), MAKE_VAL_COMPLEX(-505, -201), MAKE_VAL_COMPLEX(1079, 213), MAKE_VAL_COMPLEX(-414, 1503), MAKE_VAL_COMPLEX(979, -594), MAKE_VAL_COMPLEX(-478, -1545) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_35, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_35", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_35() */
/* END TEST 35 */

/*
 *  BEGIN TEST 36
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_36( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	oski_value_t* y_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-109, 96), MAKE_VAL_COMPLEX(248, 267), MAKE_VAL_COMPLEX(115, -18), MAKE_VAL_COMPLEX(92, 12), MAKE_VAL_COMPLEX(19, -58), MAKE_VAL_COMPLEX(0, 83), MAKE_VAL_COMPLEX(-28, 8), MAKE_VAL_COMPLEX(-62, -74), MAKE_VAL_COMPLEX(122, 96), MAKE_VAL_COMPLEX(289, -109), MAKE_VAL_COMPLEX(-76, -30), MAKE_VAL_COMPLEX(-32, 48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_36, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_36", y_true, y_val, y0_LEN*y0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_36() */
/* END TEST 36 */

/*
 *  BEGIN TEST 37
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_37( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	oski_value_t* y_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(67, -120), MAKE_VAL_COMPLEX(-286, -243), MAKE_VAL_COMPLEX(-77, 44), MAKE_VAL_COMPLEX(-58, -24), MAKE_VAL_COMPLEX(23, 42), MAKE_VAL_COMPLEX(8, -37), MAKE_VAL_COMPLEX(26, -2), MAKE_VAL_COMPLEX(30, 40), MAKE_VAL_COMPLEX(-134, -120), MAKE_VAL_COMPLEX(-283, 91), MAKE_VAL_COMPLEX(28, 24), MAKE_VAL_COMPLEX(-12, -56) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_37, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_37", y_true, y_val, y0_LEN*y0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_37() */
/* END TEST 37 */

/*
 *  BEGIN TEST 38
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_38( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	oski_value_t* y_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(30, 45), MAKE_VAL_COMPLEX(50, -5), MAKE_VAL_COMPLEX(-25, -45), MAKE_VAL_COMPLEX(-40, -5), MAKE_VAL_COMPLEX(-50, -5), MAKE_VAL_COMPLEX(15, -50), MAKE_VAL_COMPLEX(5, -5), MAKE_VAL_COMPLEX(15, 50), MAKE_VAL_COMPLEX(0, 30), MAKE_VAL_COMPLEX(-15, 15), MAKE_VAL_COMPLEX(45, 30), MAKE_VAL_COMPLEX(40, 30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_38, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_38", y_true, y_val, y0_LEN*y0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_38() */
/* END TEST 38 */

/*
 *  BEGIN TEST 39
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_39( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	oski_value_t* y_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-357, 192), MAKE_VAL_COMPLEX(245, 1434), MAKE_VAL_COMPLEX(172, 52), MAKE_VAL_COMPLEX(-16, 345), MAKE_VAL_COMPLEX(-87, -68), MAKE_VAL_COMPLEX(-236, -85), MAKE_VAL_COMPLEX(-85, -99), MAKE_VAL_COMPLEX(209, -122), MAKE_VAL_COMPLEX(126, 840), MAKE_VAL_COMPLEX(1125, 657), MAKE_VAL_COMPLEX(189, -204), MAKE_VAL_COMPLEX(56, 170) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_39, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_39", y_true, y_val, y0_LEN*y0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_39() */
/* END TEST 39 */

/*
 *  BEGIN TEST 40
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_40( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-147, -8), MAKE_VAL_COMPLEX(-12, -97), MAKE_VAL_COMPLEX(-171, -22), MAKE_VAL_COMPLEX(239, 20), MAKE_VAL_COMPLEX(18, 3), MAKE_VAL_COMPLEX(94, -73), MAKE_VAL_COMPLEX(-69, 32), MAKE_VAL_COMPLEX(-138, 83), MAKE_VAL_COMPLEX(-23, -32), MAKE_VAL_COMPLEX(-18, -133), MAKE_VAL_COMPLEX(-146, -44), MAKE_VAL_COMPLEX(0, 110), MAKE_VAL_COMPLEX(19, 77), MAKE_VAL_COMPLEX(-5, -97), MAKE_VAL_COMPLEX(-29, -28) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_40, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_40", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_40() */
/* END TEST 40 */

/*
 *  BEGIN TEST 41
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_41( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(53, -134), MAKE_VAL_COMPLEX(-68, -43), MAKE_VAL_COMPLEX(55, 34), MAKE_VAL_COMPLEX(-119, 140), MAKE_VAL_COMPLEX(128, -25), MAKE_VAL_COMPLEX(42, 121), MAKE_VAL_COMPLEX(71, -146), MAKE_VAL_COMPLEX(110, -87), MAKE_VAL_COMPLEX(-83, 74), MAKE_VAL_COMPLEX(-30, 69), MAKE_VAL_COMPLEX(54, -112), MAKE_VAL_COMPLEX(-152, -46), MAKE_VAL_COMPLEX(-127, -21), MAKE_VAL_COMPLEX(137, -27), MAKE_VAL_COMPLEX(153, -40) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_41, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_41", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_41() */
/* END TEST 41 */

/*
 *  BEGIN TEST 42
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_42( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-12, -86), MAKE_VAL_COMPLEX(-6, -82), MAKE_VAL_COMPLEX(-56, -20), MAKE_VAL_COMPLEX(20, 100), MAKE_VAL_COMPLEX(72, 22), MAKE_VAL_COMPLEX(52, 52), MAKE_VAL_COMPLEX(26, -52), MAKE_VAL_COMPLEX(-12, -8), MAKE_VAL_COMPLEX(-58, -4), MAKE_VAL_COMPLEX(-8, -40), MAKE_VAL_COMPLEX(-8, -92), MAKE_VAL_COMPLEX(-84, -4), MAKE_VAL_COMPLEX(-62, 2), MAKE_VAL_COMPLEX(88, -28), MAKE_VAL_COMPLEX(72, -4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_42, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_42", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_42() */
/* END TEST 42 */

/*
 *  BEGIN TEST 43
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_43( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-28, -570), MAKE_VAL_COMPLEX(-440, 987), MAKE_VAL_COMPLEX(1362, -131), MAKE_VAL_COMPLEX(-602, 963), MAKE_VAL_COMPLEX(-15, -1033), MAKE_VAL_COMPLEX(409, 332), MAKE_VAL_COMPLEX(-517, -999), MAKE_VAL_COMPLEX(433, -1430), MAKE_VAL_COMPLEX(493, 975), MAKE_VAL_COMPLEX(495, 1158), MAKE_VAL_COMPLEX(116, -320), MAKE_VAL_COMPLEX(-620, 390), MAKE_VAL_COMPLEX(-506, 442), MAKE_VAL_COMPLEX(79, -509), MAKE_VAL_COMPLEX(188, -1049) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_43, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_43", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_43() */
/* END TEST 43 */

/*
 *  BEGIN TEST 44
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_44( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(31, 79), MAKE_VAL_COMPLEX(149, -4), MAKE_VAL_COMPLEX(153, -124), MAKE_VAL_COMPLEX(-191, -70), MAKE_VAL_COMPLEX(-226, 168), MAKE_VAL_COMPLEX(-68, -45), MAKE_VAL_COMPLEX(16, 164), MAKE_VAL_COMPLEX(-44, -129), MAKE_VAL_COMPLEX(12, 58), MAKE_VAL_COMPLEX(52, 23), MAKE_VAL_COMPLEX(50, 66), MAKE_VAL_COMPLEX(134, -64), MAKE_VAL_COMPLEX(48, 10), MAKE_VAL_COMPLEX(-21, 61), MAKE_VAL_COMPLEX(-191, -16) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_44, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_44", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_44() */
/* END TEST 44 */

/*
 *  BEGIN TEST 45
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_45( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(170, -16), MAKE_VAL_COMPLEX(37, 76), MAKE_VAL_COMPLEX(-51, 22), MAKE_VAL_COMPLEX(-49, 10), MAKE_VAL_COMPLEX(103, -33), MAKE_VAL_COMPLEX(-100, 105), MAKE_VAL_COMPLEX(71, -53), MAKE_VAL_COMPLEX(74, 111), MAKE_VAL_COMPLEX(57, -181), MAKE_VAL_COMPLEX(44, 1), MAKE_VAL_COMPLEX(160, 12), MAKE_VAL_COMPLEX(-38, -116), MAKE_VAL_COMPLEX(12, -148), MAKE_VAL_COMPLEX(-9, 161), MAKE_VAL_COMPLEX(125, 178) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_45, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_45", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_45() */
/* END TEST 45 */

/*
 *  BEGIN TEST 46
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_46( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-51, -3), MAKE_VAL_COMPLEX(-48, -6), MAKE_VAL_COMPLEX(-18, 30), MAKE_VAL_COMPLEX(60, 0), MAKE_VAL_COMPLEX(21, -39), MAKE_VAL_COMPLEX(36, -24), MAKE_VAL_COMPLEX(-27, -21), MAKE_VAL_COMPLEX(-6, 6), MAKE_VAL_COMPLEX(-9, 33), MAKE_VAL_COMPLEX(-24, 0), MAKE_VAL_COMPLEX(-54, -6), MAKE_VAL_COMPLEX(-12, 48), MAKE_VAL_COMPLEX(-6, 36), MAKE_VAL_COMPLEX(-6, -54), MAKE_VAL_COMPLEX(6, -42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_46, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_46", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_46() */
/* END TEST 46 */

/*
 *  BEGIN TEST 47
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x0', LAYOUT_COLMAJ>
 *  y : <'y0', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_47( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A0_ptr, A0_ind, A0_val,
		A0_M, A0_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y0_val,
		y0_LEN * y0_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y0_LEN, y0_VW, LAYOUT_COLMAJ, y0_LEN );
	oski_value_t* y_val = array_Duplicate( x0_val,
		x0_LEN * x0_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x0_LEN, x0_VW, LAYOUT_COLMAJ, x0_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(72, 149), MAKE_VAL_COMPLEX(-300, -683), MAKE_VAL_COMPLEX(-573, -698), MAKE_VAL_COMPLEX(-121, 746), MAKE_VAL_COMPLEX(739, 1090), MAKE_VAL_COMPLEX(-176, -31), MAKE_VAL_COMPLEX(664, 45), MAKE_VAL_COMPLEX(-834, 217), MAKE_VAL_COMPLEX(468, 305), MAKE_VAL_COMPLEX(-6, -157), MAKE_VAL_COMPLEX(-28, 14), MAKE_VAL_COMPLEX(-84, -410), MAKE_VAL_COMPLEX(257, 49), MAKE_VAL_COMPLEX(-1, -221), MAKE_VAL_COMPLEX(-469, 754) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_47, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_47", y_true, y_val, x0_LEN*x0_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_47() */
/* END TEST 47 */

#define A48_M 4    /* # of rows */
#define A48_N 5    /* # of cols */
#define A48_NNZ 6  /* # of stored non-zeros */
static oski_index_t A48_ptr[A48_M+1] = {0, 2, 2, 4, 6}; /* end row pointers for A48 */
static oski_index_t A48_ind[A48_NNZ] = {1, 2, 0, 2, 1, 2}; /* end column indices pointers for A48 */
static oski_value_t A48_val[A48_NNZ] = {MAKE_VAL_COMPLEX(3, -8), MAKE_VAL_COMPLEX(6, 10), MAKE_VAL_COMPLEX(10, -1), MAKE_VAL_COMPLEX(7, 1), MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(9, 8)}; /* end stored non-zero values for A48 */

#define x48_LEN 5   /* vector length */
#define x48_VW  3   /* vector width */
static oski_value_t x48_val[x48_LEN*x48_VW] = {MAKE_VAL_COMPLEX(-3, 4), MAKE_VAL_COMPLEX(-1, -10), MAKE_VAL_COMPLEX(10, 5), MAKE_VAL_COMPLEX(6, 2), MAKE_VAL_COMPLEX(6, -9), MAKE_VAL_COMPLEX(-5, 2), MAKE_VAL_COMPLEX(6, 4), MAKE_VAL_COMPLEX(-9, 0), MAKE_VAL_COMPLEX(4, 10), MAKE_VAL_COMPLEX(2, 5), MAKE_VAL_COMPLEX(9, 2), MAKE_VAL_COMPLEX(-4, 5), MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(-8, -4), MAKE_VAL_COMPLEX(-1, -1)}; /* data */

#define y48_LEN 4   /* vector length */
#define y48_VW  3   /* vector width */
static oski_value_t y48_val[y48_LEN*y48_VW] = {MAKE_VAL_COMPLEX(10, 6), MAKE_VAL_COMPLEX(-2, -4), MAKE_VAL_COMPLEX(-8, -3), MAKE_VAL_COMPLEX(-6, -6), MAKE_VAL_COMPLEX(-3, -6), MAKE_VAL_COMPLEX(7, 9), MAKE_VAL_COMPLEX(2, -6), MAKE_VAL_COMPLEX(-7, -1), MAKE_VAL_COMPLEX(2, 2), MAKE_VAL_COMPLEX(7, -10), MAKE_VAL_COMPLEX(-10, -7), MAKE_VAL_COMPLEX(-1, 10)}; /* data */

/*
 *  BEGIN TEST 48
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_48( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	oski_value_t* y_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(120, 96), MAKE_VAL_COMPLEX(-126, -201), MAKE_VAL_COMPLEX(-147, 119), MAKE_VAL_COMPLEX(-54, -54), MAKE_VAL_COMPLEX(-27, -54), MAKE_VAL_COMPLEX(63, 81), MAKE_VAL_COMPLEX(30, 23), MAKE_VAL_COMPLEX(-146, -117), MAKE_VAL_COMPLEX(141, 132), MAKE_VAL_COMPLEX(45, -86), MAKE_VAL_COMPLEX(-321, -105), MAKE_VAL_COMPLEX(17, 242) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_48, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_48", y_true, y_val, y48_LEN*y48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_48() */
/* END TEST 48 */

/*
 *  BEGIN TEST 49
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_49( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	oski_value_t* y_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(40, 0), MAKE_VAL_COMPLEX(94, 137), MAKE_VAL_COMPLEX(19, -167), MAKE_VAL_COMPLEX(-42, -42), MAKE_VAL_COMPLEX(-21, -42), MAKE_VAL_COMPLEX(49, 63), MAKE_VAL_COMPLEX(2, -119), MAKE_VAL_COMPLEX(34, 101), MAKE_VAL_COMPLEX(-109, -100), MAKE_VAL_COMPLEX(67, -74), MAKE_VAL_COMPLEX(161, -7), MAKE_VAL_COMPLEX(-33, -82) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_49, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_49", y_true, y_val, y48_LEN*y48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_49() */
/* END TEST 49 */

/*
 *  BEGIN TEST 50
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_50( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	oski_value_t* y_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-66, -94), MAKE_VAL_COMPLEX(2, 44), MAKE_VAL_COMPLEX(60, 59), MAKE_VAL_COMPLEX(30, 78), MAKE_VAL_COMPLEX(3, 66), MAKE_VAL_COMPLEX(-27, -109), MAKE_VAL_COMPLEX(-42, 46), MAKE_VAL_COMPLEX(59, 37), MAKE_VAL_COMPLEX(-10, -26), MAKE_VAL_COMPLEX(-103, 62), MAKE_VAL_COMPLEX(62, 103), MAKE_VAL_COMPLEX(49, -86) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_50, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_50", y_true, y_val, y48_LEN*y48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_50() */
/* END TEST 50 */

/*
 *  BEGIN TEST 51
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_51( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	oski_value_t* y_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-66, 454), MAKE_VAL_COMPLEX(169, -2189), MAKE_VAL_COMPLEX(-1649, 880), MAKE_VAL_COMPLEX(6, 102), MAKE_VAL_COMPLEX(-21, 78), MAKE_VAL_COMPLEX(9, -137), MAKE_VAL_COMPLEX(-497, 815), MAKE_VAL_COMPLEX(64, -1488), MAKE_VAL_COMPLEX(307, 1853), MAKE_VAL_COMPLEX(-333, -56), MAKE_VAL_COMPLEX(-1751, -1852), MAKE_VAL_COMPLEX(-741, 1468) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(9, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_51, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_51", y_true, y_val, y48_LEN*y48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_51() */
/* END TEST 51 */

/*
 *  BEGIN TEST 52
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_52( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(10, -115), MAKE_VAL_COMPLEX(7, 70), MAKE_VAL_COMPLEX(-93, 48), MAKE_VAL_COMPLEX(-154, -6), MAKE_VAL_COMPLEX(7, 288), MAKE_VAL_COMPLEX(88, -86), MAKE_VAL_COMPLEX(87, 72), MAKE_VAL_COMPLEX(18, -264), MAKE_VAL_COMPLEX(-197, -52), MAKE_VAL_COMPLEX(-51, -26), MAKE_VAL_COMPLEX(-86, 47), MAKE_VAL_COMPLEX(-3, -68), MAKE_VAL_COMPLEX(150, 10), MAKE_VAL_COMPLEX(92, -24), MAKE_VAL_COMPLEX(15, 1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_52, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_52", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_52() */
/* END TEST 52 */

/*
 *  BEGIN TEST 53
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_53( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 35), MAKE_VAL_COMPLEX(87, 62), MAKE_VAL_COMPLEX(-87, -38), MAKE_VAL_COMPLEX(54, 26), MAKE_VAL_COMPLEX(-19, -114), MAKE_VAL_COMPLEX(-34, 18), MAKE_VAL_COMPLEX(-203, -80), MAKE_VAL_COMPLEX(108, 192), MAKE_VAL_COMPLEX(61, -56), MAKE_VAL_COMPLEX(-17, -28), MAKE_VAL_COMPLEX(-56, -3), MAKE_VAL_COMPLEX(19, -34), MAKE_VAL_COMPLEX(70, 50), MAKE_VAL_COMPLEX(52, 16), MAKE_VAL_COMPLEX(7, 5) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_53, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_53", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_53() */
/* END TEST 53 */

/*
 *  BEGIN TEST 54
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_54( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-12, 41), MAKE_VAL_COMPLEX(-38, -77), MAKE_VAL_COMPLEX(95, 10), MAKE_VAL_COMPLEX(54, -2), MAKE_VAL_COMPLEX(21, -90), MAKE_VAL_COMPLEX(-34, 31), MAKE_VAL_COMPLEX(60, 14), MAKE_VAL_COMPLEX(-72, 27), MAKE_VAL_COMPLEX(62, 68), MAKE_VAL_COMPLEX(31, 34), MAKE_VAL_COMPLEX(78, -11), MAKE_VAL_COMPLEX(-17, 52), MAKE_VAL_COMPLEX(-110, -50), MAKE_VAL_COMPLEX(-76, -8), MAKE_VAL_COMPLEX(-11, -5) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_54, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_54", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_54() */
/* END TEST 54 */

/*
 *  BEGIN TEST 55
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_55( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-191, 168), MAKE_VAL_COMPLEX(365, 216), MAKE_VAL_COMPLEX(-162, -71), MAKE_VAL_COMPLEX(246, 346), MAKE_VAL_COMPLEX(410, -569), MAKE_VAL_COMPLEX(-299, -38), MAKE_VAL_COMPLEX(-602, -554), MAKE_VAL_COMPLEX(-123, 840), MAKE_VAL_COMPLEX(272, 152), MAKE_VAL_COMPLEX(-54, -19), MAKE_VAL_COMPLEX(-79, 58), MAKE_VAL_COMPLEX(-12, -67), MAKE_VAL_COMPLEX(150, -10), MAKE_VAL_COMPLEX(88, -36), MAKE_VAL_COMPLEX(15, -1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_55, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_55", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_55() */
/* END TEST 55 */

/*
 *  BEGIN TEST 56
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_56( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(17, -71), MAKE_VAL_COMPLEX(-38, -10), MAKE_VAL_COMPLEX(-7, 47), MAKE_VAL_COMPLEX(0, 284), MAKE_VAL_COMPLEX(217, -31), MAKE_VAL_COMPLEX(-91, -200), MAKE_VAL_COMPLEX(93, -240), MAKE_VAL_COMPLEX(-239, -14), MAKE_VAL_COMPLEX(-25, 174), MAKE_VAL_COMPLEX(-17, 1), MAKE_VAL_COMPLEX(-15, 25), MAKE_VAL_COMPLEX(-11, -17), MAKE_VAL_COMPLEX(40, -20), MAKE_VAL_COMPLEX(20, -20), MAKE_VAL_COMPLEX(4, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_56, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_56", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_56() */
/* END TEST 56 */

/*
 *  BEGIN TEST 57
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_57( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-31, 48), MAKE_VAL_COMPLEX(90, 25), MAKE_VAL_COMPLEX(-38, -7), MAKE_VAL_COMPLEX(-22, -258), MAKE_VAL_COMPLEX(-184, 79), MAKE_VAL_COMPLEX(91, 171), MAKE_VAL_COMPLEX(-125, 262), MAKE_VAL_COMPLEX(257, -31), MAKE_VAL_COMPLEX(-33, -174), MAKE_VAL_COMPLEX(-12, -1), MAKE_VAL_COMPLEX(-13, 16), MAKE_VAL_COMPLEX(-6, -13), MAKE_VAL_COMPLEX(30, -10), MAKE_VAL_COMPLEX(16, -12), MAKE_VAL_COMPLEX(3, -1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_57, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_57", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_57() */
/* END TEST 57 */

/*
 *  BEGIN TEST 58
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_58( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-24, 32), MAKE_VAL_COMPLEX(-8, -80), MAKE_VAL_COMPLEX(80, 40), MAKE_VAL_COMPLEX(48, 16), MAKE_VAL_COMPLEX(48, -72), MAKE_VAL_COMPLEX(-40, 16), MAKE_VAL_COMPLEX(48, 32), MAKE_VAL_COMPLEX(-72, 0), MAKE_VAL_COMPLEX(32, 80), MAKE_VAL_COMPLEX(16, 40), MAKE_VAL_COMPLEX(72, 16), MAKE_VAL_COMPLEX(-32, 40), MAKE_VAL_COMPLEX(-80, -80), MAKE_VAL_COMPLEX(-64, -32), MAKE_VAL_COMPLEX(-8, -8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_58, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_58", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_58() */
/* END TEST 58 */

/*
 *  BEGIN TEST 59
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_59( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-129, -43), MAKE_VAL_COMPLEX(-27, 107), MAKE_VAL_COMPLEX(69, -11), MAKE_VAL_COMPLEX(552, -12), MAKE_VAL_COMPLEX(-89, -413), MAKE_VAL_COMPLEX(-383, 181), MAKE_VAL_COMPLEX(-494, -204), MAKE_VAL_COMPLEX(-1, 487), MAKE_VAL_COMPLEX(346, 16), MAKE_VAL_COMPLEX(1, 17), MAKE_VAL_COMPLEX(25, 15), MAKE_VAL_COMPLEX(-17, 11), MAKE_VAL_COMPLEX(-20, -40), MAKE_VAL_COMPLEX(-20, -20), MAKE_VAL_COMPLEX(-2, -4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_59, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_59", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_59() */
/* END TEST 59 */

/*
 *  BEGIN TEST 60
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_60( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	oski_value_t* y_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(68, -44), MAKE_VAL_COMPLEX(-30, 20), MAKE_VAL_COMPLEX(-4, 144), MAKE_VAL_COMPLEX(-60, 58), MAKE_VAL_COMPLEX(-153, -135), MAKE_VAL_COMPLEX(65, -65), MAKE_VAL_COMPLEX(-133, -118), MAKE_VAL_COMPLEX(-232, 15), MAKE_VAL_COMPLEX(-61, 128), MAKE_VAL_COMPLEX(-87, -46), MAKE_VAL_COMPLEX(77, 201), MAKE_VAL_COMPLEX(107, 150) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_60, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_60", y_true, y_val, y48_LEN*y48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_60() */
/* END TEST 60 */

/*
 *  BEGIN TEST 61
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_61( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	oski_value_t* y_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(62, -82), MAKE_VAL_COMPLEX(-38, -6), MAKE_VAL_COMPLEX(-73, -36), MAKE_VAL_COMPLEX(-54, 8), MAKE_VAL_COMPLEX(51, 156), MAKE_VAL_COMPLEX(98, -4), MAKE_VAL_COMPLEX(51, 64), MAKE_VAL_COMPLEX(189, 86), MAKE_VAL_COMPLEX(99, -150), MAKE_VAL_COMPLEX(-35, -99), MAKE_VAL_COMPLEX(-222, -79), MAKE_VAL_COMPLEX(39, -95) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_61, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_61", y_true, y_val, y48_LEN*y48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_61() */
/* END TEST 61 */

/*
 *  BEGIN TEST 62
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_62( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	oski_value_t* y_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(74, 58), MAKE_VAL_COMPLEX(-12, -34), MAKE_VAL_COMPLEX(-61, -32), MAKE_VAL_COMPLEX(-42, -54), MAKE_VAL_COMPLEX(-18, -51), MAKE_VAL_COMPLEX(47, 79), MAKE_VAL_COMPLEX(22, -46), MAKE_VAL_COMPLEX(-55, -15), MAKE_VAL_COMPLEX(14, 18), MAKE_VAL_COMPLEX(66, -73), MAKE_VAL_COMPLEX(-73, -66), MAKE_VAL_COMPLEX(-18, 79) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_62, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_62", y_true, y_val, y48_LEN*y48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_62() */
/* END TEST 62 */

/*
 *  BEGIN TEST 63
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_63( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	oski_value_t* y_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-142, 234), MAKE_VAL_COMPLEX(38, 16), MAKE_VAL_COMPLEX(-121, 240), MAKE_VAL_COMPLEX(12, -36), MAKE_VAL_COMPLEX(228, -795), MAKE_VAL_COMPLEX(-103, -21), MAKE_VAL_COMPLEX(97, -519), MAKE_VAL_COMPLEX(-512, -854), MAKE_VAL_COMPLEX(-689, 211), MAKE_VAL_COMPLEX(11, 112), MAKE_VAL_COMPLEX(139, 700), MAKE_VAL_COMPLEX(-431, 458) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_63, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_63", y_true, y_val, y48_LEN*y48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_63() */
/* END TEST 63 */

/*
 *  BEGIN TEST 64
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_64( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-67, -60), MAKE_VAL_COMPLEX(42, 16), MAKE_VAL_COMPLEX(-197, -80), MAKE_VAL_COMPLEX(26, 54), MAKE_VAL_COMPLEX(-27, 170), MAKE_VAL_COMPLEX(168, -126), MAKE_VAL_COMPLEX(-115, -15), MAKE_VAL_COMPLEX(79, -189), MAKE_VAL_COMPLEX(-212, -17), MAKE_VAL_COMPLEX(-26, -36), MAKE_VAL_COMPLEX(-76, 2), MAKE_VAL_COMPLEX(22, -48), MAKE_VAL_COMPLEX(100, 60), MAKE_VAL_COMPLEX(72, 16), MAKE_VAL_COMPLEX(10, 6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_64, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_64", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_64() */
/* END TEST 64 */

/*
 *  BEGIN TEST 65
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_65( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(107, 40), MAKE_VAL_COMPLEX(-74, 68), MAKE_VAL_COMPLEX(137, 0), MAKE_VAL_COMPLEX(-66, -94), MAKE_VAL_COMPLEX(-57, -122), MAKE_VAL_COMPLEX(-120, 130), MAKE_VAL_COMPLEX(83, -41), MAKE_VAL_COMPLEX(-7, 225), MAKE_VAL_COMPLEX(220, -79), MAKE_VAL_COMPLEX(30, -12), MAKE_VAL_COMPLEX(12, -54), MAKE_VAL_COMPLEX(30, 24), MAKE_VAL_COMPLEX(-60, 60), MAKE_VAL_COMPLEX(-24, 48), MAKE_VAL_COMPLEX(-6, 6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_65, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_65", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_65() */
/* END TEST 65 */

/*
 *  BEGIN TEST 66
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_66( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-10, -20), MAKE_VAL_COMPLEX(42, 16), MAKE_VAL_COMPLEX(-40, 30), MAKE_VAL_COMPLEX(-20, 20), MAKE_VAL_COMPLEX(24, 42), MAKE_VAL_COMPLEX(2, -24), MAKE_VAL_COMPLEX(-28, 16), MAKE_VAL_COMPLEX(18, -36), MAKE_VAL_COMPLEX(-48, -4), MAKE_VAL_COMPLEX(-24, -2), MAKE_VAL_COMPLEX(-26, 32), MAKE_VAL_COMPLEX(-12, -26), MAKE_VAL_COMPLEX(60, -20), MAKE_VAL_COMPLEX(32, -24), MAKE_VAL_COMPLEX(6, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_66, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_66", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_66() */
/* END TEST 66 */

/*
 *  BEGIN TEST 67
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_67( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-178, -355), MAKE_VAL_COMPLEX(275, -78), MAKE_VAL_COMPLEX(-231, -496), MAKE_VAL_COMPLEX(10, 418), MAKE_VAL_COMPLEX(-255, 354), MAKE_VAL_COMPLEX(723, 62), MAKE_VAL_COMPLEX(-250, -166), MAKE_VAL_COMPLEX(597, -528), MAKE_VAL_COMPLEX(-713, -369), MAKE_VAL_COMPLEX(-34, -27), MAKE_VAL_COMPLEX(-71, 22), MAKE_VAL_COMPLEX(8, -51), MAKE_VAL_COMPLEX(110, 30), MAKE_VAL_COMPLEX(72, -4), MAKE_VAL_COMPLEX(11, 3) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_67, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_67", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_67() */
/* END TEST 67 */

/*
 *  BEGIN TEST 68
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_68( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-115, -54), MAKE_VAL_COMPLEX(104, -86), MAKE_VAL_COMPLEX(-113, 10), MAKE_VAL_COMPLEX(98, 150), MAKE_VAL_COMPLEX(203, -72), MAKE_VAL_COMPLEX(-126, -124), MAKE_VAL_COMPLEX(-61, -27), MAKE_VAL_COMPLEX(-159, -75), MAKE_VAL_COMPLEX(-46, 103), MAKE_VAL_COMPLEX(-36, 26), MAKE_VAL_COMPLEX(2, 76), MAKE_VAL_COMPLEX(-48, -22), MAKE_VAL_COMPLEX(60, -100), MAKE_VAL_COMPLEX(16, -72), MAKE_VAL_COMPLEX(6, -10) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_68, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_68", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_68() */
/* END TEST 68 */

/*
 *  BEGIN TEST 69
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_69( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(124, 42), MAKE_VAL_COMPLEX(-101, 116), MAKE_VAL_COMPLEX(83, -25), MAKE_VAL_COMPLEX(-116, -156), MAKE_VAL_COMPLEX(-221, 99), MAKE_VAL_COMPLEX(141, 118), MAKE_VAL_COMPLEX(43, 15), MAKE_VAL_COMPLEX(186, 75), MAKE_VAL_COMPLEX(34, -133), MAKE_VAL_COMPLEX(30, -41), MAKE_VAL_COMPLEX(-29, -82), MAKE_VAL_COMPLEX(60, 7), MAKE_VAL_COMPLEX(-30, 130), MAKE_VAL_COMPLEX(8, 84), MAKE_VAL_COMPLEX(-3, 13) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_69, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_69", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_69() */
/* END TEST 69 */

/*
 *  BEGIN TEST 70
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_70( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(5, 10), MAKE_VAL_COMPLEX(-21, -8), MAKE_VAL_COMPLEX(20, -15), MAKE_VAL_COMPLEX(10, -10), MAKE_VAL_COMPLEX(-12, -21), MAKE_VAL_COMPLEX(-1, 12), MAKE_VAL_COMPLEX(14, -8), MAKE_VAL_COMPLEX(-9, 18), MAKE_VAL_COMPLEX(24, 2), MAKE_VAL_COMPLEX(12, 1), MAKE_VAL_COMPLEX(13, -16), MAKE_VAL_COMPLEX(6, 13), MAKE_VAL_COMPLEX(-30, 10), MAKE_VAL_COMPLEX(-16, 12), MAKE_VAL_COMPLEX(-3, 1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_70, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_70", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_70() */
/* END TEST 70 */

/*
 *  BEGIN TEST 71
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_ROWMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_71( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_ROWMAJ, x48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-846, 194), MAKE_VAL_COMPLEX(-186, -676), MAKE_VAL_COMPLEX(-1144, -92), MAKE_VAL_COMPLEX(1448, 160), MAKE_VAL_COMPLEX(340, -1608), MAKE_VAL_COMPLEX(-1350, -72), MAKE_VAL_COMPLEX(-774, -418), MAKE_VAL_COMPLEX(-1200, 858), MAKE_VAL_COMPLEX(578, 296), MAKE_VAL_COMPLEX(32, 22), MAKE_VAL_COMPLEX(62, -24), MAKE_VAL_COMPLEX(-4, 46), MAKE_VAL_COMPLEX(-100, -20), MAKE_VAL_COMPLEX(-64, 8), MAKE_VAL_COMPLEX(-10, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_71, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_71", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_71() */
/* END TEST 71 */

/*
 *  BEGIN TEST 72
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_72( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	oski_value_t* y_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-9, 160), MAKE_VAL_COMPLEX(-14, -156), MAKE_VAL_COMPLEX(15, -142), MAKE_VAL_COMPLEX(-36, -48), MAKE_VAL_COMPLEX(-15, -45), MAKE_VAL_COMPLEX(40, 70), MAKE_VAL_COMPLEX(59, 48), MAKE_VAL_COMPLEX(-159, 2), MAKE_VAL_COMPLEX(44, -53), MAKE_VAL_COMPLEX(19, 172), MAKE_VAL_COMPLEX(-164, -231), MAKE_VAL_COMPLEX(63, -111) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_72, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_72", y_true, y_val, y48_LEN*y48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_72() */
/* END TEST 72 */

/*
 *  BEGIN TEST 73
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_73( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	oski_value_t* y_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-7, -156), MAKE_VAL_COMPLEX(20, 158), MAKE_VAL_COMPLEX(-4, 137), MAKE_VAL_COMPLEX(48, 48), MAKE_VAL_COMPLEX(24, 48), MAKE_VAL_COMPLEX(-56, -72), MAKE_VAL_COMPLEX(-55, -40), MAKE_VAL_COMPLEX(167, -8), MAKE_VAL_COMPLEX(-48, 53), MAKE_VAL_COMPLEX(-16, -155), MAKE_VAL_COMPLEX(181, 228), MAKE_VAL_COMPLEX(-72, 100) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_73, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_73", y_true, y_val, y48_LEN*y48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_73() */
/* END TEST 73 */

/*
 *  BEGIN TEST 74
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_74( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	oski_value_t* y_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(66, 94), MAKE_VAL_COMPLEX(-2, -44), MAKE_VAL_COMPLEX(-60, -59), MAKE_VAL_COMPLEX(-30, -78), MAKE_VAL_COMPLEX(-3, -66), MAKE_VAL_COMPLEX(27, 109), MAKE_VAL_COMPLEX(42, -46), MAKE_VAL_COMPLEX(-59, -37), MAKE_VAL_COMPLEX(10, 26), MAKE_VAL_COMPLEX(103, -62), MAKE_VAL_COMPLEX(-62, -103), MAKE_VAL_COMPLEX(-49, 86) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_74, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_74", y_true, y_val, y48_LEN*y48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_74() */
/* END TEST 74 */

/*
 *  BEGIN TEST 75
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_75( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	oski_value_t* y_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-83, -1284), MAKE_VAL_COMPLEX(748, 876), MAKE_VAL_COMPLEX(175, 1271), MAKE_VAL_COMPLEX(-54, 54), MAKE_VAL_COMPLEX(-54, 27), MAKE_VAL_COMPLEX(81, -63), MAKE_VAL_COMPLEX(-855, -400), MAKE_VAL_COMPLEX(672, -715), MAKE_VAL_COMPLEX(208, 657), MAKE_VAL_COMPLEX(-1220, -1948), MAKE_VAL_COMPLEX(1676, 688), MAKE_VAL_COMPLEX(610, 1749) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_75, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_75", y_true, y_val, y48_LEN*y48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_75() */
/* END TEST 75 */

/*
 *  BEGIN TEST 76
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_76( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(61, -58), MAKE_VAL_COMPLEX(-167, 26), MAKE_VAL_COMPLEX(153, -43), MAKE_VAL_COMPLEX(-14, -58), MAKE_VAL_COMPLEX(-102, -3), MAKE_VAL_COMPLEX(-30, 27), MAKE_VAL_COMPLEX(-6, 106), MAKE_VAL_COMPLEX(-9, -129), MAKE_VAL_COMPLEX(60, -82), MAKE_VAL_COMPLEX(30, -41), MAKE_VAL_COMPLEX(-7, -64), MAKE_VAL_COMPLEX(122, -28), MAKE_VAL_COMPLEX(-125, 130), MAKE_VAL_COMPLEX(8, 84), MAKE_VAL_COMPLEX(-3, 13) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_76, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_76", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_76() */
/* END TEST 76 */

/*
 *  BEGIN TEST 77
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_77( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(25, 85), MAKE_VAL_COMPLEX(3, 51), MAKE_VAL_COMPLEX(-128, -157), MAKE_VAL_COMPLEX(12, -56), MAKE_VAL_COMPLEX(-87, -45), MAKE_VAL_COMPLEX(94, 46), MAKE_VAL_COMPLEX(38, -232), MAKE_VAL_COMPLEX(63, 282), MAKE_VAL_COMPLEX(86, -46), MAKE_VAL_COMPLEX(43, -23), MAKE_VAL_COMPLEX(-13, -101), MAKE_VAL_COMPLEX(-13, 66), MAKE_VAL_COMPLEX(15, 100), MAKE_VAL_COMPLEX(-28, 76), MAKE_VAL_COMPLEX(-8, 10) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_77, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_77", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_77() */
/* END TEST 77 */

/*
 *  BEGIN TEST 78
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_78( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-15, -55), MAKE_VAL_COMPLEX(97, 61), MAKE_VAL_COMPLEX(-115, 55), MAKE_VAL_COMPLEX(-60, 40), MAKE_VAL_COMPLEX(39, 117), MAKE_VAL_COMPLEX(17, -59), MAKE_VAL_COMPLEX(-78, 26), MAKE_VAL_COMPLEX(63, -81), MAKE_VAL_COMPLEX(-118, -34), MAKE_VAL_COMPLEX(-59, -17), MAKE_VAL_COMPLEX(-81, 67), MAKE_VAL_COMPLEX(-17, -71), MAKE_VAL_COMPLEX(160, -20), MAKE_VAL_COMPLEX(92, -44), MAKE_VAL_COMPLEX(16, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_78, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_78", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_78() */
/* END TEST 78 */

/*
 *  BEGIN TEST 79
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_79( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(101, 127), MAKE_VAL_COMPLEX(119, -219), MAKE_VAL_COMPLEX(-202, 269), MAKE_VAL_COMPLEX(48, -4), MAKE_VAL_COMPLEX(15, -81), MAKE_VAL_COMPLEX(48, -110), MAKE_VAL_COMPLEX(-286, -180), MAKE_VAL_COMPLEX(393, 120), MAKE_VAL_COMPLEX(58, 58), MAKE_VAL_COMPLEX(29, 29), MAKE_VAL_COMPLEX(11, 13), MAKE_VAL_COMPLEX(-5, 206), MAKE_VAL_COMPLEX(-5, -230), MAKE_VAL_COMPLEX(-68, -4), MAKE_VAL_COMPLEX(-10, -4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_79, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_79", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_79() */
/* END TEST 79 */

/*
 *  BEGIN TEST 80
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_80( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(30, -105), MAKE_VAL_COMPLEX(70, 343), MAKE_VAL_COMPLEX(6, -244), MAKE_VAL_COMPLEX(-58, 14), MAKE_VAL_COMPLEX(-3, 102), MAKE_VAL_COMPLEX(-39, -58), MAKE_VAL_COMPLEX(128, -60), MAKE_VAL_COMPLEX(-176, -32), MAKE_VAL_COMPLEX(-82, -60), MAKE_VAL_COMPLEX(-41, -30), MAKE_VAL_COMPLEX(-64, 51), MAKE_VAL_COMPLEX(-83, -243), MAKE_VAL_COMPLEX(139, 202), MAKE_VAL_COMPLEX(84, -8), MAKE_VAL_COMPLEX(13, 3) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_80, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_80", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_80() */
/* END TEST 80 */

/*
 *  BEGIN TEST 81
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_81( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-12, 81), MAKE_VAL_COMPLEX(-64, -283), MAKE_VAL_COMPLEX(-66, 214), MAKE_VAL_COMPLEX(22, -26), MAKE_VAL_COMPLEX(-33, -48), MAKE_VAL_COMPLEX(69, 46), MAKE_VAL_COMPLEX(-164, 36), MAKE_VAL_COMPLEX(230, 32), MAKE_VAL_COMPLEX(58, 0), MAKE_VAL_COMPLEX(29, 0), MAKE_VAL_COMPLEX(10, -63), MAKE_VAL_COMPLEX(107, 213), MAKE_VAL_COMPLEX(-79, -142), MAKE_VAL_COMPLEX(-36, 32), MAKE_VAL_COMPLEX(-7, 3) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_81, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_81", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_81() */
/* END TEST 81 */

/*
 *  BEGIN TEST 82
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_82( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(22, 4), MAKE_VAL_COMPLEX(-38, 24), MAKE_VAL_COMPLEX(0, -50), MAKE_VAL_COMPLEX(-4, -28), MAKE_VAL_COMPLEX(-48, -6), MAKE_VAL_COMPLEX(18, 16), MAKE_VAL_COMPLEX(4, -32), MAKE_VAL_COMPLEX(18, 36), MAKE_VAL_COMPLEX(32, -36), MAKE_VAL_COMPLEX(16, -18), MAKE_VAL_COMPLEX(-10, -40), MAKE_VAL_COMPLEX(28, 6), MAKE_VAL_COMPLEX(-20, 60), MAKE_VAL_COMPLEX(0, 40), MAKE_VAL_COMPLEX(-2, 6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_82, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_82", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_82() */
/* END TEST 82 */

/*
 *  BEGIN TEST 83
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_83( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_ROWMAJ, y48_VW );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-322, -89), MAKE_VAL_COMPLEX(1360, -311), MAKE_VAL_COMPLEX(-1416, -281), MAKE_VAL_COMPLEX(-18, 14), MAKE_VAL_COMPLEX(15, 36), MAKE_VAL_COMPLEX(-12, 343), MAKE_VAL_COMPLEX(-510, -912), MAKE_VAL_COMPLEX(331, 1200), MAKE_VAL_COMPLEX(-38, -8), MAKE_VAL_COMPLEX(-19, -4), MAKE_VAL_COMPLEX(68, -89), MAKE_VAL_COMPLEX(-832, 611), MAKE_VAL_COMPLEX(901, -227), MAKE_VAL_COMPLEX(28, -16), MAKE_VAL_COMPLEX(5, -1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_83, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_83", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_83() */
/* END TEST 83 */

/*
 *  BEGIN TEST 84
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_84( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	oski_value_t* y_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(5, 182), MAKE_VAL_COMPLEX(-10, -40), MAKE_VAL_COMPLEX(-27, 45), MAKE_VAL_COMPLEX(-82, 169), MAKE_VAL_COMPLEX(-19, -186), MAKE_VAL_COMPLEX(45, 95), MAKE_VAL_COMPLEX(-81, -34), MAKE_VAL_COMPLEX(-162, -195), MAKE_VAL_COMPLEX(82, -91), MAKE_VAL_COMPLEX(83, -76), MAKE_VAL_COMPLEX(-44, -152), MAKE_VAL_COMPLEX(51, -92) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_84, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_84", y_true, y_val, y48_LEN*y48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_84() */
/* END TEST 84 */

/*
 *  BEGIN TEST 85
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_85( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	oski_value_t* y_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(29, -148), MAKE_VAL_COMPLEX(6, 22), MAKE_VAL_COMPLEX(-2, -65), MAKE_VAL_COMPLEX(64, -199), MAKE_VAL_COMPLEX(13, 159), MAKE_VAL_COMPLEX(-26, -52), MAKE_VAL_COMPLEX(95, 12), MAKE_VAL_COMPLEX(135, 184), MAKE_VAL_COMPLEX(-76, 101), MAKE_VAL_COMPLEX(-45, 43), MAKE_VAL_COMPLEX(11, 114), MAKE_VAL_COMPLEX(-65, 131) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_85, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_85", y_true, y_val, y48_LEN*y48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_85() */
/* END TEST 85 */

/*
 *  BEGIN TEST 86
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_86( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	oski_value_t* y_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(60, -32), MAKE_VAL_COMPLEX(-26, -2), MAKE_VAL_COMPLEX(-39, 31), MAKE_VAL_COMPLEX(-48, 12), MAKE_VAL_COMPLEX(-39, -3), MAKE_VAL_COMPLEX(66, -8), MAKE_VAL_COMPLEX(-24, -28), MAKE_VAL_COMPLEX(-26, 32), MAKE_VAL_COMPLEX(16, -4), MAKE_VAL_COMPLEX(-29, -65), MAKE_VAL_COMPLEX(-65, 29), MAKE_VAL_COMPLEX(47, 35) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_86, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_86", y_true, y_val, y48_LEN*y48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_86() */
/* END TEST 86 */

/*
 *  BEGIN TEST 87
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_87( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	oski_value_t* y_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-644, -1100), MAKE_VAL_COMPLEX(6, 42), MAKE_VAL_COMPLEX(-797, 11), MAKE_VAL_COMPLEX(-1684, -1188), MAKE_VAL_COMPLEX(1033, 535), MAKE_VAL_COMPLEX(-36, -102), MAKE_VAL_COMPLEX(280, -904), MAKE_VAL_COMPLEX(1840, -90), MAKE_VAL_COMPLEX(620, 972), MAKE_VAL_COMPLEX(-93, 69), MAKE_VAL_COMPLEX(493, 625), MAKE_VAL_COMPLEX(1159, 1273) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_87, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_87", y_true, y_val, y48_LEN*y48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_87() */
/* END TEST 87 */

/*
 *  BEGIN TEST 88
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_88( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-55, -1), MAKE_VAL_COMPLEX(8, 65), MAKE_VAL_COMPLEX(-24, -65), MAKE_VAL_COMPLEX(14, -42), MAKE_VAL_COMPLEX(-63, -42), MAKE_VAL_COMPLEX(28, -27), MAKE_VAL_COMPLEX(31, 44), MAKE_VAL_COMPLEX(7, -108), MAKE_VAL_COMPLEX(70, -28), MAKE_VAL_COMPLEX(35, -14), MAKE_VAL_COMPLEX(-93, -123), MAKE_VAL_COMPLEX(167, -72), MAKE_VAL_COMPLEX(-230, 125), MAKE_VAL_COMPLEX(-28, 56), MAKE_VAL_COMPLEX(-7, 7) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_88, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_88", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_88() */
/* END TEST 88 */

/*
 *  BEGIN TEST 89
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_89( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(94, 49), MAKE_VAL_COMPLEX(-131, -83), MAKE_VAL_COMPLEX(114, -40), MAKE_VAL_COMPLEX(28, -24), MAKE_VAL_COMPLEX(-27, -57), MAKE_VAL_COMPLEX(-19, 93), MAKE_VAL_COMPLEX(35, -104), MAKE_VAL_COMPLEX(-34, 216), MAKE_VAL_COMPLEX(62, 10), MAKE_VAL_COMPLEX(31, 5), MAKE_VAL_COMPLEX(144, 21), MAKE_VAL_COMPLEX(-119, 135), MAKE_VAL_COMPLEX(80, -35), MAKE_VAL_COMPLEX(-44, 28), MAKE_VAL_COMPLEX(-8, 2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_89, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_89", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_89() */
/* END TEST 89 */

/*
 *  BEGIN TEST 90
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_90( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(40, 5), MAKE_VAL_COMPLEX(-66, 47), MAKE_VAL_COMPLEX(-5, -90), MAKE_VAL_COMPLEX(-10, -50), MAKE_VAL_COMPLEX(-87, -6), MAKE_VAL_COMPLEX(34, 27), MAKE_VAL_COMPLEX(4, -58), MAKE_VAL_COMPLEX(36, 63), MAKE_VAL_COMPLEX(54, -68), MAKE_VAL_COMPLEX(27, -34), MAKE_VAL_COMPLEX(-22, -71), MAKE_VAL_COMPLEX(51, 8), MAKE_VAL_COMPLEX(-30, 110), MAKE_VAL_COMPLEX(4, 72), MAKE_VAL_COMPLEX(-3, 11) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_90, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_90", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_90() */
/* END TEST 90 */

/*
 *  BEGIN TEST 91
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_91( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-45, -267), MAKE_VAL_COMPLEX(-60, 248), MAKE_VAL_COMPLEX(-54, -112), MAKE_VAL_COMPLEX(16, 32), MAKE_VAL_COMPLEX(60, -12), MAKE_VAL_COMPLEX(172, -32), MAKE_VAL_COMPLEX(-247, 135), MAKE_VAL_COMPLEX(484, -186), MAKE_VAL_COMPLEX(-24, 56), MAKE_VAL_COMPLEX(-12, 28), MAKE_VAL_COMPLEX(101, -337), MAKE_VAL_COMPLEX(396, 300), MAKE_VAL_COMPLEX(-325, -505), MAKE_VAL_COMPLEX(-16, -48), MAKE_VAL_COMPLEX(0, -8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_91, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_91", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_91() */
/* END TEST 91 */

/*
 *  BEGIN TEST 92
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_92( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-76, -6), MAKE_VAL_COMPLEX(57, 52), MAKE_VAL_COMPLEX(29, -98), MAKE_VAL_COMPLEX(38, -14), MAKE_VAL_COMPLEX(-6, -69), MAKE_VAL_COMPLEX(9, -28), MAKE_VAL_COMPLEX(165, -106), MAKE_VAL_COMPLEX(-186, 33), MAKE_VAL_COMPLEX(60, 34), MAKE_VAL_COMPLEX(30, 17), MAKE_VAL_COMPLEX(-40, -106), MAKE_VAL_COMPLEX(-100, -47), MAKE_VAL_COMPLEX(-64, 41), MAKE_VAL_COMPLEX(-56, 12), MAKE_VAL_COMPLEX(-9, -1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_92, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_92", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_92() */
/* END TEST 92 */

/*
 *  BEGIN TEST 93
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_93( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(47, 3), MAKE_VAL_COMPLEX(-10, -87), MAKE_VAL_COMPLEX(-24, 163), MAKE_VAL_COMPLEX(-30, 50), MAKE_VAL_COMPLEX(69, 72), MAKE_VAL_COMPLEX(-34, 9), MAKE_VAL_COMPLEX(-167, 148), MAKE_VAL_COMPLEX(159, -78), MAKE_VAL_COMPLEX(-98, 16), MAKE_VAL_COMPLEX(-49, 8), MAKE_VAL_COMPLEX(57, 157), MAKE_VAL_COMPLEX(63, 42), MAKE_VAL_COMPLEX(84, -121), MAKE_VAL_COMPLEX(52, -64), MAKE_VAL_COMPLEX(11, -7) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_93, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_93", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_93() */
/* END TEST 93 */

/*
 *  BEGIN TEST 94
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_94( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(38, -9), MAKE_VAL_COMPLEX(-44, 65), MAKE_VAL_COMPLEX(-35, -80), MAKE_VAL_COMPLEX(-26, -42), MAKE_VAL_COMPLEX(-81, 24), MAKE_VAL_COMPLEX(40, 13), MAKE_VAL_COMPLEX(-16, -54), MAKE_VAL_COMPLEX(54, 45), MAKE_VAL_COMPLEX(26, -80), MAKE_VAL_COMPLEX(13, -40), MAKE_VAL_COMPLEX(-44, -57), MAKE_VAL_COMPLEX(49, -10), MAKE_VAL_COMPLEX(10, 110), MAKE_VAL_COMPLEX(28, 64), MAKE_VAL_COMPLEX(1, 11) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_94, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_94", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_94() */
/* END TEST 94 */

/*
 *  BEGIN TEST 95
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x48', LAYOUT_COLMAJ>
 *  y : <'y48', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_95( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A48_ptr, A48_ind, A48_val,
		A48_M, A48_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y48_val,
		y48_LEN * y48_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y48_LEN, y48_VW, LAYOUT_COLMAJ, y48_LEN );
	oski_value_t* y_val = array_Duplicate( x48_val,
		x48_LEN * x48_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x48_LEN, x48_VW, LAYOUT_COLMAJ, x48_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(325, 776), MAKE_VAL_COMPLEX(-212, -1350), MAKE_VAL_COMPLEX(-111, 777), MAKE_VAL_COMPLEX(52, -36), MAKE_VAL_COMPLEX(-36, -102), MAKE_VAL_COMPLEX(-544, 302), MAKE_VAL_COMPLEX(-1377, -24), MAKE_VAL_COMPLEX(915, 939), MAKE_VAL_COMPLEX(104, 28), MAKE_VAL_COMPLEX(52, 14), MAKE_VAL_COMPLEX(241, 1058), MAKE_VAL_COMPLEX(188, 1278), MAKE_VAL_COMPLEX(-16, -493), MAKE_VAL_COMPLEX(-80, 40), MAKE_VAL_COMPLEX(-14, 2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_95, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_95", y_true, y_val, x48_LEN*x48_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_95() */
/* END TEST 95 */

#define A96_M 4    /* # of rows */
#define A96_N 5    /* # of cols */
#define A96_NNZ 4  /* # of stored non-zeros */
static oski_index_t A96_ptr[A96_M+1] = {1, 2, 4, 4, 5}; /* end row pointers for A96 */
static oski_index_t A96_ind[A96_NNZ] = {5, 1, 5, 2}; /* end column indices pointers for A96 */
static oski_value_t A96_val[A96_NNZ] = {MAKE_VAL_COMPLEX(-10, 6), MAKE_VAL_COMPLEX(9, -6), MAKE_VAL_COMPLEX(10, 2), MAKE_VAL_COMPLEX(9, 10)}; /* end stored non-zero values for A96 */

#define x96_LEN 5   /* vector length */
#define x96_VW  3   /* vector width */
static oski_value_t x96_val[x96_LEN*x96_VW] = {MAKE_VAL_COMPLEX(-6, 3), MAKE_VAL_COMPLEX(1, 8), MAKE_VAL_COMPLEX(7, 8), MAKE_VAL_COMPLEX(10, -5), MAKE_VAL_COMPLEX(10, 3), MAKE_VAL_COMPLEX(8, -10), MAKE_VAL_COMPLEX(-10, 1), MAKE_VAL_COMPLEX(10, 10), MAKE_VAL_COMPLEX(-0, -3), MAKE_VAL_COMPLEX(8, 9), MAKE_VAL_COMPLEX(-9, 1), MAKE_VAL_COMPLEX(-9, 10), MAKE_VAL_COMPLEX(3, -4), MAKE_VAL_COMPLEX(-9, 5), MAKE_VAL_COMPLEX(4, 2)}; /* data */

#define y96_LEN 4   /* vector length */
#define y96_VW  3   /* vector width */
static oski_value_t y96_val[y96_LEN*y96_VW] = {MAKE_VAL_COMPLEX(-1, 9), MAKE_VAL_COMPLEX(-7, 8), MAKE_VAL_COMPLEX(-7, -7), MAKE_VAL_COMPLEX(6, -10), MAKE_VAL_COMPLEX(1, -2), MAKE_VAL_COMPLEX(1, 10), MAKE_VAL_COMPLEX(-3, -2), MAKE_VAL_COMPLEX(-10, 7), MAKE_VAL_COMPLEX(0, 5), MAKE_VAL_COMPLEX(6, -3), MAKE_VAL_COMPLEX(10, -2), MAKE_VAL_COMPLEX(-0, -5)}; /* data */

/*
 *  BEGIN TEST 96
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_96( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	oski_value_t* y_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-59, -8), MAKE_VAL_COMPLEX(62, -194), MAKE_VAL_COMPLEX(46, 19), MAKE_VAL_COMPLEX(30, 130), MAKE_VAL_COMPLEX(-28, 121), MAKE_VAL_COMPLEX(88, -16), MAKE_VAL_COMPLEX(23, -3), MAKE_VAL_COMPLEX(38, -99), MAKE_VAL_COMPLEX(-30, -38), MAKE_VAL_COMPLEX(124, 121), MAKE_VAL_COMPLEX(-7, 202), MAKE_VAL_COMPLEX(193, 35) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_96, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_96", y_true, y_val, y96_LEN*y96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_96() */
/* END TEST 96 */

/*
 *  BEGIN TEST 97
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_97( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	oski_value_t* y_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(78, -81), MAKE_VAL_COMPLEX(16, 121), MAKE_VAL_COMPLEX(17, 58), MAKE_VAL_COMPLEX(-100, -36), MAKE_VAL_COMPLEX(16, -102), MAKE_VAL_COMPLEX(-88, -85), MAKE_VAL_COMPLEX(5, 26), MAKE_VAL_COMPLEX(69, 39), MAKE_VAL_COMPLEX(35, -12), MAKE_VAL_COMPLEX(-187, -97), MAKE_VAL_COMPLEX(-95, -192), MAKE_VAL_COMPLEX(-198, 15) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_97, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_97", y_true, y_val, y96_LEN*y96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_97() */
/* END TEST 97 */

/*
 *  BEGIN TEST 98
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_98( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	oski_value_t* y_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(38, 68), MAKE_VAL_COMPLEX(-9, 91), MAKE_VAL_COMPLEX(-84, -14), MAKE_VAL_COMPLEX(-8, -100), MAKE_VAL_COMPLEX(-3, -19), MAKE_VAL_COMPLEX(57, 65), MAKE_VAL_COMPLEX(-31, 1), MAKE_VAL_COMPLEX(-35, 99), MAKE_VAL_COMPLEX(25, 35), MAKE_VAL_COMPLEX(27, -51), MAKE_VAL_COMPLEX(60, -64), MAKE_VAL_COMPLEX(-25, -35) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_98, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_98", y_true, y_val, y96_LEN*y96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_98() */
/* END TEST 98 */

/*
 *  BEGIN TEST 99
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_99( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	oski_value_t* y_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(350, 65), MAKE_VAL_COMPLEX(-440, -270), MAKE_VAL_COMPLEX(25, 260), MAKE_VAL_COMPLEX(70, -90), MAKE_VAL_COMPLEX(495, 160), MAKE_VAL_COMPLEX(290, -780), MAKE_VAL_COMPLEX(-5, 65), MAKE_VAL_COMPLEX(85, 0), MAKE_VAL_COMPLEX(10, 0), MAKE_VAL_COMPLEX(305, -770), MAKE_VAL_COMPLEX(630, -305), MAKE_VAL_COMPLEX(-25, -815) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_99, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_99", y_true, y_val, y96_LEN*y96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_99() */
/* END TEST 99 */

/*
 *  BEGIN TEST 100
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_100( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(53, -87), MAKE_VAL_COMPLEX(48, -72), MAKE_VAL_COMPLEX(84, -27), MAKE_VAL_COMPLEX(-10, -27), MAKE_VAL_COMPLEX(75, -18), MAKE_VAL_COMPLEX(-77, -39), MAKE_VAL_COMPLEX(65, 72), MAKE_VAL_COMPLEX(10, -133), MAKE_VAL_COMPLEX(-24, 23), MAKE_VAL_COMPLEX(30, -121), MAKE_VAL_COMPLEX(72, 64), MAKE_VAL_COMPLEX(134, 7), MAKE_VAL_COMPLEX(-14, -184), MAKE_VAL_COMPLEX(130, -98), MAKE_VAL_COMPLEX(94, 86) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_100, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_100", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_100() */
/* END TEST 100 */

/*
 *  BEGIN TEST 101
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_101( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-20, 63), MAKE_VAL_COMPLEX(-62, 25), MAKE_VAL_COMPLEX(-134, -14), MAKE_VAL_COMPLEX(-45, 67), MAKE_VAL_COMPLEX(-138, 10), MAKE_VAL_COMPLEX(39, 107), MAKE_VAL_COMPLEX(-6, -88), MAKE_VAL_COMPLEX(-80, 83), MAKE_VAL_COMPLEX(27, -5), MAKE_VAL_COMPLEX(-87, 75), MAKE_VAL_COMPLEX(-19, -79), MAKE_VAL_COMPLEX(-90, -76), MAKE_VAL_COMPLEX(0, 211), MAKE_VAL_COMPLEX(-81, 59), MAKE_VAL_COMPLEX(-120, -94) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_101, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_101", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_101() */
/* END TEST 101 */

/*
 *  BEGIN TEST 102
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_102( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_102, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_102", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_102() */
/* END TEST 102 */

/*
 *  BEGIN TEST 103
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_103( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-183, 380), MAKE_VAL_COMPLEX(27, 40), MAKE_VAL_COMPLEX(-21, -407), MAKE_VAL_COMPLEX(-274, -274), MAKE_VAL_COMPLEX(-191, -511), MAKE_VAL_COMPLEX(-287, 1), MAKE_VAL_COMPLEX(39, 49), MAKE_VAL_COMPLEX(54, -71), MAKE_VAL_COMPLEX(-2, -6), MAKE_VAL_COMPLEX(-12, -62), MAKE_VAL_COMPLEX(-3, 19), MAKE_VAL_COMPLEX(57, 21), MAKE_VAL_COMPLEX(-501, 480), MAKE_VAL_COMPLEX(-341, 369), MAKE_VAL_COMPLEX(-50, -616) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-3, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_103, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_103", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_103() */
/* END TEST 103 */

/*
 *  BEGIN TEST 104
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_104( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(104, -33), MAKE_VAL_COMPLEX(24, 11), MAKE_VAL_COMPLEX(-36, 98), MAKE_VAL_COMPLEX(45, -117), MAKE_VAL_COMPLEX(94, -124), MAKE_VAL_COMPLEX(-43, -63), MAKE_VAL_COMPLEX(-22, 10), MAKE_VAL_COMPLEX(20, 17), MAKE_VAL_COMPLEX(-3, -1), MAKE_VAL_COMPLEX(31, 7), MAKE_VAL_COMPLEX(-7, 9), MAKE_VAL_COMPLEX(-8, 24), MAKE_VAL_COMPLEX(106, -207), MAKE_VAL_COMPLEX(111, -41), MAKE_VAL_COMPLEX(68, 210) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_104, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_104", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_104() */
/* END TEST 104 */

/*
 *  BEGIN TEST 105
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_105( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-167, 72), MAKE_VAL_COMPLEX(-5, 76), MAKE_VAL_COMPLEX(121, -17), MAKE_VAL_COMPLEX(60, 52), MAKE_VAL_COMPLEX(19, 147), MAKE_VAL_COMPLEX(121, -55), MAKE_VAL_COMPLEX(-87, 11), MAKE_VAL_COMPLEX(100, 83), MAKE_VAL_COMPLEX(0, -32), MAKE_VAL_COMPLEX(66, 84), MAKE_VAL_COMPLEX(-91, 11), MAKE_VAL_COMPLEX(-81, 95), MAKE_VAL_COMPLEX(-77, 160), MAKE_VAL_COMPLEX(-205, 105), MAKE_VAL_COMPLEX(-22, -192) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_105, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_105", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_105() */
/* END TEST 105 */

/*
 *  BEGIN TEST 106
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_106( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-66, -27), MAKE_VAL_COMPLEX(-57, 64), MAKE_VAL_COMPLEX(-15, 112), MAKE_VAL_COMPLEX(110, 45), MAKE_VAL_COMPLEX(46, 101), MAKE_VAL_COMPLEX(136, -6), MAKE_VAL_COMPLEX(-78, -73), MAKE_VAL_COMPLEX(-10, 150), MAKE_VAL_COMPLEX(24, -21), MAKE_VAL_COMPLEX(-16, 127), MAKE_VAL_COMPLEX(-71, -65), MAKE_VAL_COMPLEX(-143, -2), MAKE_VAL_COMPLEX(53, -4), MAKE_VAL_COMPLEX(-103, -37), MAKE_VAL_COMPLEX(12, 46) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_106, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_106", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_106() */
/* END TEST 106 */

/*
 *  BEGIN TEST 107
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_107( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(693, -383), MAKE_VAL_COMPLEX(104, -106), MAKE_VAL_COMPLEX(-283, 500), MAKE_VAL_COMPLEX(-17, -612), MAKE_VAL_COMPLEX(238, -855), MAKE_VAL_COMPLEX(-433, -113), MAKE_VAL_COMPLEX(64, 23), MAKE_VAL_COMPLEX(-93, -68), MAKE_VAL_COMPLEX(-7, 54), MAKE_VAL_COMPLEX(5, -128), MAKE_VAL_COMPLEX(134, 6), MAKE_VAL_COMPLEX(107, -74), MAKE_VAL_COMPLEX(388, -1260), MAKE_VAL_COMPLEX(776, -488), MAKE_VAL_COMPLEX(534, 1170) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_107, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_107", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_107() */
/* END TEST 107 */

/*
 *  BEGIN TEST 108
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_108( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	oski_value_t* y_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(38, 103), MAKE_VAL_COMPLEX(32, 98), MAKE_VAL_COMPLEX(-80, 15), MAKE_VAL_COMPLEX(112, -12), MAKE_VAL_COMPLEX(53, -110), MAKE_VAL_COMPLEX(31, 135), MAKE_VAL_COMPLEX(-14, 20), MAKE_VAL_COMPLEX(53, 216), MAKE_VAL_COMPLEX(-15, 32), MAKE_VAL_COMPLEX(161, 0), MAKE_VAL_COMPLEX(28, -71), MAKE_VAL_COMPLEX(133, -20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_108, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_108", y_true, y_val, y96_LEN*y96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_108() */
/* END TEST 108 */

/*
 *  BEGIN TEST 109
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_109( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	oski_value_t* y_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-59, 4), MAKE_VAL_COMPLEX(-124, -9), MAKE_VAL_COMPLEX(3, -106), MAKE_VAL_COMPLEX(-30, -102), MAKE_VAL_COMPLEX(-39, 87), MAKE_VAL_COMPLEX(-29, -14), MAKE_VAL_COMPLEX(-20, -47), MAKE_VAL_COMPLEX(-180, -142), MAKE_VAL_COMPLEX(10, 28), MAKE_VAL_COMPLEX(-86, -30), MAKE_VAL_COMPLEX(94, 57), MAKE_VAL_COMPLEX(-128, -40) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_109, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_109", y_true, y_val, y96_LEN*y96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_109() */
/* END TEST 109 */

/*
 *  BEGIN TEST 110
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_110( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	oski_value_t* y_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-74, 10), MAKE_VAL_COMPLEX(-78, -40), MAKE_VAL_COMPLEX(42, -70), MAKE_VAL_COMPLEX(92, 28), MAKE_VAL_COMPLEX(18, 4), MAKE_VAL_COMPLEX(-78, 28), MAKE_VAL_COMPLEX(10, -28), MAKE_VAL_COMPLEX(-76, -66), MAKE_VAL_COMPLEX(-40, 10), MAKE_VAL_COMPLEX(36, 42), MAKE_VAL_COMPLEX(36, 76), MAKE_VAL_COMPLEX(40, -10) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_110, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_110", y_true, y_val, y96_LEN*y96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_110() */
/* END TEST 110 */

/*
 *  BEGIN TEST 111
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_111( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	oski_value_t* y_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-496, 350), MAKE_VAL_COMPLEX(-112, 360), MAKE_VAL_COMPLEX(-180, -74), MAKE_VAL_COMPLEX(344, 1440), MAKE_VAL_COMPLEX(1126, -112), MAKE_VAL_COMPLEX(-918, 414), MAKE_VAL_COMPLEX(-60, 148), MAKE_VAL_COMPLEX(-742, 1312), MAKE_VAL_COMPLEX(-326, -248), MAKE_VAL_COMPLEX(570, 1456), MAKE_VAL_COMPLEX(88, -114), MAKE_VAL_COMPLEX(938, 1264) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_111, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_111", y_true, y_val, y96_LEN*y96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_111() */
/* END TEST 111 */

/*
 *  BEGIN TEST 112
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_112( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(14, 153), MAKE_VAL_COMPLEX(116, 60), MAKE_VAL_COMPLEX(70, -116), MAKE_VAL_COMPLEX(97, -72), MAKE_VAL_COMPLEX(-161, -93), MAKE_VAL_COMPLEX(-20, -76), MAKE_VAL_COMPLEX(19, 51), MAKE_VAL_COMPLEX(37, -82), MAKE_VAL_COMPLEX(-8, 4), MAKE_VAL_COMPLEX(44, -76), MAKE_VAL_COMPLEX(14, 59), MAKE_VAL_COMPLEX(78, 29), MAKE_VAL_COMPLEX(-160, -40), MAKE_VAL_COMPLEX(40, 172), MAKE_VAL_COMPLEX(40, -96) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_112, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_112", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_112() */
/* END TEST 112 */

/*
 *  BEGIN TEST 113
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_113( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(52, -111), MAKE_VAL_COMPLEX(-42, -118), MAKE_VAL_COMPLEX(-32, -2), MAKE_VAL_COMPLEX(-207, 2), MAKE_VAL_COMPLEX(131, -25), MAKE_VAL_COMPLEX(-128, 56), MAKE_VAL_COMPLEX(51, 43), MAKE_VAL_COMPLEX(3, -78), MAKE_VAL_COMPLEX(-22, 14), MAKE_VAL_COMPLEX(-2, -58), MAKE_VAL_COMPLEX(50, 25), MAKE_VAL_COMPLEX(76, 1), MAKE_VAL_COMPLEX(102, 34), MAKE_VAL_COMPLEX(64, -112), MAKE_VAL_COMPLEX(-44, 44) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_113, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_113", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_113() */
/* END TEST 113 */

/*
 *  BEGIN TEST 114
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_114( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(21, -18), MAKE_VAL_COMPLEX(-12, -31), MAKE_VAL_COMPLEX(-36, -25), MAKE_VAL_COMPLEX(-35, 30), MAKE_VAL_COMPLEX(-43, -2), MAKE_VAL_COMPLEX(-22, 48), MAKE_VAL_COMPLEX(39, -14), MAKE_VAL_COMPLEX(-50, -30), MAKE_VAL_COMPLEX(3, 12), MAKE_VAL_COMPLEX(-41, -28), MAKE_VAL_COMPLEX(35, -13), MAKE_VAL_COMPLEX(26, -49), MAKE_VAL_COMPLEX(-8, 19), MAKE_VAL_COMPLEX(31, -29), MAKE_VAL_COMPLEX(-18, -4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_114, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_114", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_114() */
/* END TEST 114 */

/*
 *  BEGIN TEST 115
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_115( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1247, -441), MAKE_VAL_COMPLEX(446, -1096), MAKE_VAL_COMPLEX(-680, -138), MAKE_VAL_COMPLEX(-1033, -1263), MAKE_VAL_COMPLEX(516, 1468), MAKE_VAL_COMPLEX(-840, -268), MAKE_VAL_COMPLEX(40, 172), MAKE_VAL_COMPLEX(17, -103), MAKE_VAL_COMPLEX(-92, -62), MAKE_VAL_COMPLEX(-96, -122), MAKE_VAL_COMPLEX(175, 121), MAKE_VAL_COMPLEX(89, 37), MAKE_VAL_COMPLEX(330, 1320), MAKE_VAL_COMPLEX(1286, -526), MAKE_VAL_COMPLEX(-800, -28) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, -9);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_115, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_115", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_115() */
/* END TEST 115 */

/*
 *  BEGIN TEST 116
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_116( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-76, 6), MAKE_VAL_COMPLEX(-73, 40), MAKE_VAL_COMPLEX(7, -28), MAKE_VAL_COMPLEX(-113, -87), MAKE_VAL_COMPLEX(-95, 172), MAKE_VAL_COMPLEX(-80, 38), MAKE_VAL_COMPLEX(61, -34), MAKE_VAL_COMPLEX(-93, -52), MAKE_VAL_COMPLEX(16, 19), MAKE_VAL_COMPLEX(-68, -57), MAKE_VAL_COMPLEX(51, -18), MAKE_VAL_COMPLEX(43, -93), MAKE_VAL_COMPLEX(-3, 44), MAKE_VAL_COMPLEX(61, 59), MAKE_VAL_COMPLEX(52, -98) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_116, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_116", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_116() */
/* END TEST 116 */

/*
 *  BEGIN TEST 117
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_117( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(160, -18), MAKE_VAL_COMPLEX(93, -140), MAKE_VAL_COMPLEX(-59, -96), MAKE_VAL_COMPLEX(-27, 107), MAKE_VAL_COMPLEX(-13, -248), MAKE_VAL_COMPLEX(-56, 50), MAKE_VAL_COMPLEX(63, 62), MAKE_VAL_COMPLEX(13, -108), MAKE_VAL_COMPLEX(-28, 17), MAKE_VAL_COMPLEX(8, -83), MAKE_VAL_COMPLEX(61, 42), MAKE_VAL_COMPLEX(105, 9), MAKE_VAL_COMPLEX(-49, -8), MAKE_VAL_COMPLEX(67, -83), MAKE_VAL_COMPLEX(-92, 58) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_117, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_117", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_117() */
/* END TEST 117 */

/*
 *  BEGIN TEST 118
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_118( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(30, -30), MAKE_VAL_COMPLEX(-22, -46), MAKE_VAL_COMPLEX(-58, -34), MAKE_VAL_COMPLEX(-50, 50), MAKE_VAL_COMPLEX(-66, 2), MAKE_VAL_COMPLEX(-28, 76), MAKE_VAL_COMPLEX(58, -26), MAKE_VAL_COMPLEX(-80, -40), MAKE_VAL_COMPLEX(6, 18), MAKE_VAL_COMPLEX(-66, -38), MAKE_VAL_COMPLEX(52, -24), MAKE_VAL_COMPLEX(34, -78), MAKE_VAL_COMPLEX(-10, 30), MAKE_VAL_COMPLEX(44, -48), MAKE_VAL_COMPLEX(-28, -4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_118, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_118", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_118() */
/* END TEST 118 */

/*
 *  BEGIN TEST 119
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_ROWMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_119( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_ROWMAJ, x96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-43, 275), MAKE_VAL_COMPLEX(148, 209), MAKE_VAL_COMPLEX(122, -121), MAKE_VAL_COMPLEX(-322, -56), MAKE_VAL_COMPLEX(350, 207), MAKE_VAL_COMPLEX(-134, 12), MAKE_VAL_COMPLEX(-40, 19), MAKE_VAL_COMPLEX(23, 14), MAKE_VAL_COMPLEX(3, -28), MAKE_VAL_COMPLEX(11, -12), MAKE_VAL_COMPLEX(-13, 38), MAKE_VAL_COMPLEX(-18, 24), MAKE_VAL_COMPLEX(32, -21), MAKE_VAL_COMPLEX(219, 115), MAKE_VAL_COMPLEX(-90, -260) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_119, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_119", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_119() */
/* END TEST 119 */

/*
 *  BEGIN TEST 120
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_120( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	oski_value_t* y_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-163, 56), MAKE_VAL_COMPLEX(-179, -56), MAKE_VAL_COMPLEX(-54, -44), MAKE_VAL_COMPLEX(117, 115), MAKE_VAL_COMPLEX(75, -33), MAKE_VAL_COMPLEX(-85, 135), MAKE_VAL_COMPLEX(6, -10), MAKE_VAL_COMPLEX(-48, -9), MAKE_VAL_COMPLEX(-17, 11), MAKE_VAL_COMPLEX(-31, 92), MAKE_VAL_COMPLEX(-62, -60), MAKE_VAL_COMPLEX(-170, -10) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_120, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_120", y_true, y_val, y96_LEN*y96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_120() */
/* END TEST 120 */

/*
 *  BEGIN TEST 121
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_121( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	oski_value_t* y_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(181, -54), MAKE_VAL_COMPLEX(195, 70), MAKE_VAL_COMPLEX(40, 58), MAKE_VAL_COMPLEX(-137, -127), MAKE_VAL_COMPLEX(-79, 31), MAKE_VAL_COMPLEX(105, -137), MAKE_VAL_COMPLEX(-10, 16), MAKE_VAL_COMPLEX(62, 29), MAKE_VAL_COMPLEX(27, -11), MAKE_VAL_COMPLEX(25, -104), MAKE_VAL_COMPLEX(58, 40), MAKE_VAL_COMPLEX(160, 10) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_121, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_121", y_true, y_val, y96_LEN*y96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_121() */
/* END TEST 121 */

/*
 *  BEGIN TEST 122
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_122( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	oski_value_t* y_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(56, 70), MAKE_VAL_COMPLEX(7, 105), MAKE_VAL_COMPLEX(-98, 0), MAKE_VAL_COMPLEX(-28, -112), MAKE_VAL_COMPLEX(-7, -21), MAKE_VAL_COMPLEX(77, 63), MAKE_VAL_COMPLEX(-35, 7), MAKE_VAL_COMPLEX(-21, 119), MAKE_VAL_COMPLEX(35, 35), MAKE_VAL_COMPLEX(21, -63), MAKE_VAL_COMPLEX(56, -84), MAKE_VAL_COMPLEX(-35, -35) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_122, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_122", y_true, y_val, y96_LEN*y96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_122() */
/* END TEST 122 */

/*
 *  BEGIN TEST 123
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_123( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	oski_value_t* y_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-451, -901), MAKE_VAL_COMPLEX(206, -1047), MAKE_VAL_COMPLEX(-225, -443), MAKE_VAL_COMPLEX(-908, 652), MAKE_VAL_COMPLEX(364, 439), MAKE_VAL_COMPLEX(-832, -169), MAKE_VAL_COMPLEX(-70, 89), MAKE_VAL_COMPLEX(-31, 184), MAKE_VAL_COMPLEX(73, 26), MAKE_VAL_COMPLEX(-747, -382), MAKE_VAL_COMPLEX(558, -1062), MAKE_VAL_COMPLEX(-455, -1520) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_123, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_123", y_true, y_val, y96_LEN*y96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_123() */
/* END TEST 123 */

/*
 *  BEGIN TEST 124
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_124( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(2, -69), MAKE_VAL_COMPLEX(148, 32), MAKE_VAL_COMPLEX(67, -35), MAKE_VAL_COMPLEX(-9, -83), MAKE_VAL_COMPLEX(77, -248), MAKE_VAL_COMPLEX(-64, -92), MAKE_VAL_COMPLEX(98, 152), MAKE_VAL_COMPLEX(80, -43), MAKE_VAL_COMPLEX(-11, -8), MAKE_VAL_COMPLEX(115, -178), MAKE_VAL_COMPLEX(51, 142), MAKE_VAL_COMPLEX(103, 48), MAKE_VAL_COMPLEX(-22, -24), MAKE_VAL_COMPLEX(17, 68), MAKE_VAL_COMPLEX(124, 106) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_124, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_124", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_124() */
/* END TEST 124 */

/*
 *  BEGIN TEST 125
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_125( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-62, 99), MAKE_VAL_COMPLEX(-138, 48), MAKE_VAL_COMPLEX(3, 115), MAKE_VAL_COMPLEX(109, 33), MAKE_VAL_COMPLEX(23, 278), MAKE_VAL_COMPLEX(144, -8), MAKE_VAL_COMPLEX(-198, -142), MAKE_VAL_COMPLEX(20, 143), MAKE_VAL_COMPLEX(11, -22), MAKE_VAL_COMPLEX(-35, 268), MAKE_VAL_COMPLEX(-141, -132), MAKE_VAL_COMPLEX(-193, 52), MAKE_VAL_COMPLEX(52, -16), MAKE_VAL_COMPLEX(-107, -18), MAKE_VAL_COMPLEX(-84, -86) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_125, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_125", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_125() */
/* END TEST 125 */

/*
 *  BEGIN TEST 126
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_126( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(42, -36), MAKE_VAL_COMPLEX(-24, -62), MAKE_VAL_COMPLEX(-72, -50), MAKE_VAL_COMPLEX(-70, 60), MAKE_VAL_COMPLEX(-86, -4), MAKE_VAL_COMPLEX(-44, 96), MAKE_VAL_COMPLEX(78, -28), MAKE_VAL_COMPLEX(-100, -60), MAKE_VAL_COMPLEX(6, 24), MAKE_VAL_COMPLEX(-82, -56), MAKE_VAL_COMPLEX(70, -26), MAKE_VAL_COMPLEX(52, -98), MAKE_VAL_COMPLEX(-16, 38), MAKE_VAL_COMPLEX(62, -58), MAKE_VAL_COMPLEX(-36, -8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_126, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_126", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_126() */
/* END TEST 126 */

/*
 *  BEGIN TEST 127
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_127( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-954, -259), MAKE_VAL_COMPLEX(496, -640), MAKE_VAL_COMPLEX(29, -1), MAKE_VAL_COMPLEX(-11, -117), MAKE_VAL_COMPLEX(-1329, -884), MAKE_VAL_COMPLEX(-192, -28), MAKE_VAL_COMPLEX(958, -596), MAKE_VAL_COMPLEX(96, 71), MAKE_VAL_COMPLEX(-1, -92), MAKE_VAL_COMPLEX(-951, -730), MAKE_VAL_COMPLEX(789, -218), MAKE_VAL_COMPLEX(-95, -448), MAKE_VAL_COMPLEX(26, -8), MAKE_VAL_COMPLEX(-33, 40), MAKE_VAL_COMPLEX(1364, -442) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_127, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_127", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_127() */
/* END TEST 127 */

/*
 *  BEGIN TEST 128
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_128( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(95, 9), MAKE_VAL_COMPLEX(84, -55), MAKE_VAL_COMPLEX(87, 4), MAKE_VAL_COMPLEX(36, -93), MAKE_VAL_COMPLEX(182, -238), MAKE_VAL_COMPLEX(2, -112), MAKE_VAL_COMPLEX(17, -54), MAKE_VAL_COMPLEX(110, 7), MAKE_VAL_COMPLEX(-8, -20), MAKE_VAL_COMPLEX(226, -54), MAKE_VAL_COMPLEX(-106, 149), MAKE_VAL_COMPLEX(-43, 79), MAKE_VAL_COMPLEX(-6, -37), MAKE_VAL_COMPLEX(-24, 79), MAKE_VAL_COMPLEX(94, 198) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_128, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_128", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_128() */
/* END TEST 128 */

/*
 *  BEGIN TEST 129
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_129( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-65, 81), MAKE_VAL_COMPLEX(30, 57), MAKE_VAL_COMPLEX(39, -86), MAKE_VAL_COMPLEX(-86, -57), MAKE_VAL_COMPLEX(-120, 104), MAKE_VAL_COMPLEX(-126, -20), MAKE_VAL_COMPLEX(-23, 196), MAKE_VAL_COMPLEX(50, -127), MAKE_VAL_COMPLEX(-34, 14), MAKE_VAL_COMPLEX(-84, -40), MAKE_VAL_COMPLEX(102, -21), MAKE_VAL_COMPLEX(165, 67), MAKE_VAL_COMPLEX(-44, -13), MAKE_VAL_COMPLEX(76, 57), MAKE_VAL_COMPLEX(-58, -250) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_129, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_129", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_129() */
/* END TEST 129 */

/*
 *  BEGIN TEST 130
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_130( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(36, -18), MAKE_VAL_COMPLEX(-6, -48), MAKE_VAL_COMPLEX(-42, -48), MAKE_VAL_COMPLEX(-60, 30), MAKE_VAL_COMPLEX(-60, -18), MAKE_VAL_COMPLEX(-48, 60), MAKE_VAL_COMPLEX(60, -6), MAKE_VAL_COMPLEX(-60, -60), MAKE_VAL_COMPLEX(0, 18), MAKE_VAL_COMPLEX(-48, -54), MAKE_VAL_COMPLEX(54, -6), MAKE_VAL_COMPLEX(54, -60), MAKE_VAL_COMPLEX(-18, 24), MAKE_VAL_COMPLEX(54, -30), MAKE_VAL_COMPLEX(-24, -12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_130, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_130", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_130() */
/* END TEST 130 */

/*
 *  BEGIN TEST 131
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_131( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_ROWMAJ, y96_VW );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(831, 105), MAKE_VAL_COMPLEX(519, -498), MAKE_VAL_COMPLEX(36, -63), MAKE_VAL_COMPLEX(15, -60), MAKE_VAL_COMPLEX(1230, -924), MAKE_VAL_COMPLEX(36, -30), MAKE_VAL_COMPLEX(792, -447), MAKE_VAL_COMPLEX(-21, -48), MAKE_VAL_COMPLEX(48, 18), MAKE_VAL_COMPLEX(978, -36), MAKE_VAL_COMPLEX(-609, 414), MAKE_VAL_COMPLEX(-129, -303), MAKE_VAL_COMPLEX(-39, 12), MAKE_VAL_COMPLEX(45, 24), MAKE_VAL_COMPLEX(-270, 1410) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_131, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_131", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_131() */
/* END TEST 131 */

/*
 *  BEGIN TEST 132
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_132( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	oski_value_t* y_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-188, 35), MAKE_VAL_COMPLEX(-4, 80), MAKE_VAL_COMPLEX(49, -48), MAKE_VAL_COMPLEX(15, 109), MAKE_VAL_COMPLEX(-111, -47), MAKE_VAL_COMPLEX(-5, -14), MAKE_VAL_COMPLEX(21, -13), MAKE_VAL_COMPLEX(-159, -157), MAKE_VAL_COMPLEX(-96, 10), MAKE_VAL_COMPLEX(-21, 140), MAKE_VAL_COMPLEX(27, 64), MAKE_VAL_COMPLEX(-155, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_132, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_132", y_true, y_val, y96_LEN*y96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_132() */
/* END TEST 132 */

/*
 *  BEGIN TEST 133
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_133( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	oski_value_t* y_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(170, -37), MAKE_VAL_COMPLEX(-12, -94), MAKE_VAL_COMPLEX(-35, 34), MAKE_VAL_COMPLEX(5, -97), MAKE_VAL_COMPLEX(115, 49), MAKE_VAL_COMPLEX(-15, 16), MAKE_VAL_COMPLEX(-17, 7), MAKE_VAL_COMPLEX(145, 137), MAKE_VAL_COMPLEX(86, -10), MAKE_VAL_COMPLEX(27, -128), MAKE_VAL_COMPLEX(-23, -44), MAKE_VAL_COMPLEX(165, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_133, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_133", y_true, y_val, y96_LEN*y96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_133() */
/* END TEST 133 */

/*
 *  BEGIN TEST 134
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_134( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	oski_value_t* y_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(46, -4), MAKE_VAL_COMPLEX(47, 27), MAKE_VAL_COMPLEX(-28, 42), MAKE_VAL_COMPLEX(-56, -20), MAKE_VAL_COMPLEX(-11, -3), MAKE_VAL_COMPLEX(49, -15), MAKE_VAL_COMPLEX(-7, 17), MAKE_VAL_COMPLEX(45, 43), MAKE_VAL_COMPLEX(25, -5), MAKE_VAL_COMPLEX(-21, -27), MAKE_VAL_COMPLEX(-20, -48), MAKE_VAL_COMPLEX(-25, 5) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_134, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_134", y_true, y_val, y96_LEN*y96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_134() */
/* END TEST 134 */

/*
 *  BEGIN TEST 135
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_135( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	oski_value_t* y_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-602, 779), MAKE_VAL_COMPLEX(963, 378), MAKE_VAL_COMPLEX(131, 20), MAKE_VAL_COMPLEX(25, 825), MAKE_VAL_COMPLEX(-1014, 329), MAKE_VAL_COMPLEX(195, -543), MAKE_VAL_COMPLEX(128, 9), MAKE_VAL_COMPLEX(-1051, -122), MAKE_VAL_COMPLEX(-356, 315), MAKE_VAL_COMPLEX(202, 876), MAKE_VAL_COMPLEX(-36, -1), MAKE_VAL_COMPLEX(-1100, 1000) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_135, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_135", y_true, y_val, y96_LEN*y96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_135() */
/* END TEST 135 */

/*
 *  BEGIN TEST 136
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_136( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(29, 138), MAKE_VAL_COMPLEX(182, -67), MAKE_VAL_COMPLEX(-2, -82), MAKE_VAL_COMPLEX(-69, -35), MAKE_VAL_COMPLEX(-165, -95), MAKE_VAL_COMPLEX(-20, 92), MAKE_VAL_COMPLEX(-104, 18), MAKE_VAL_COMPLEX(-3, -102), MAKE_VAL_COMPLEX(-25, 22), MAKE_VAL_COMPLEX(-3, 43), MAKE_VAL_COMPLEX(86, -18), MAKE_VAL_COMPLEX(151, -53), MAKE_VAL_COMPLEX(-25, 3), MAKE_VAL_COMPLEX(70, 15), MAKE_VAL_COMPLEX(26, -98) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_136, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_136", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_136() */
/* END TEST 136 */

/*
 *  BEGIN TEST 137
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_137( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(55, -90), MAKE_VAL_COMPLEX(-94, -9), MAKE_VAL_COMPLEX(42, -66), MAKE_VAL_COMPLEX(-71, -45), MAKE_VAL_COMPLEX(121, -49), MAKE_VAL_COMPLEX(-164, -108), MAKE_VAL_COMPLEX(196, 94), MAKE_VAL_COMPLEX(43, -98), MAKE_VAL_COMPLEX(-11, 2), MAKE_VAL_COMPLEX(47, -211), MAKE_VAL_COMPLEX(-2, 118), MAKE_VAL_COMPLEX(41, 81), MAKE_VAL_COMPLEX(-47, -7), MAKE_VAL_COMPLEX(62, 53), MAKE_VAL_COMPLEX(-34, 34) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_137, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_137", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_137() */
/* END TEST 137 */

/*
 *  BEGIN TEST 138
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_138( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(39, 48), MAKE_VAL_COMPLEX(70, -25), MAKE_VAL_COMPLEX(58, -79), MAKE_VAL_COMPLEX(-65, -80), MAKE_VAL_COMPLEX(7, -96), MAKE_VAL_COMPLEX(-106, -52), MAKE_VAL_COMPLEX(29, 88), MAKE_VAL_COMPLEX(70, -110), MAKE_VAL_COMPLEX(-27, 6), MAKE_VAL_COMPLEX(65, -90), MAKE_VAL_COMPLEX(27, 79), MAKE_VAL_COMPLEX(108, 61), MAKE_VAL_COMPLEX(-42, -19), MAKE_VAL_COMPLEX(63, 71), MAKE_VAL_COMPLEX(10, -40) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_138, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_138", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_138() */
/* END TEST 138 */

/*
 *  BEGIN TEST 139
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_139( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-952, 716), MAKE_VAL_COMPLEX(1154, 838), MAKE_VAL_COMPLEX(-59, -117), MAKE_VAL_COMPLEX(77, 27), MAKE_VAL_COMPLEX(-759, -1105), MAKE_VAL_COMPLEX(-94, 1138), MAKE_VAL_COMPLEX(-877, -1337), MAKE_VAL_COMPLEX(-87, -55), MAKE_VAL_COMPLEX(-110, -6), MAKE_VAL_COMPLEX(-1019, 819), MAKE_VAL_COMPLEX(700, -186), MAKE_VAL_COMPLEX(743, -21), MAKE_VAL_COMPLEX(81, 85), MAKE_VAL_COMPLEX(65, -87), MAKE_VAL_COMPLEX(702, -222) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_139, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_139", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_139() */
/* END TEST 139 */

/*
 *  BEGIN TEST 140
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_140( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-100, 63), MAKE_VAL_COMPLEX(-21, -146), MAKE_VAL_COMPLEX(25, -35), MAKE_VAL_COMPLEX(-14, -50), MAKE_VAL_COMPLEX(22, -30), MAKE_VAL_COMPLEX(-90, 62), MAKE_VAL_COMPLEX(-15, 213), MAKE_VAL_COMPLEX(37, -42), MAKE_VAL_COMPLEX(-22, 7), MAKE_VAL_COMPLEX(44, 80), MAKE_VAL_COMPLEX(76, 50), MAKE_VAL_COMPLEX(-4, -12), MAKE_VAL_COMPLEX(-6, -14), MAKE_VAL_COMPLEX(20, 31), MAKE_VAL_COMPLEX(92, -108) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_140, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_140", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_140() */
/* END TEST 140 */

/*
 *  BEGIN TEST 141
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_141( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(88, -87), MAKE_VAL_COMPLEX(-11, 150), MAKE_VAL_COMPLEX(-57, 63), MAKE_VAL_COMPLEX(34, 90), MAKE_VAL_COMPLEX(-34, 70), MAKE_VAL_COMPLEX(130, -30), MAKE_VAL_COMPLEX(11, -253), MAKE_VAL_COMPLEX(-77, 82), MAKE_VAL_COMPLEX(34, -7), MAKE_VAL_COMPLEX(-80, -48), MAKE_VAL_COMPLEX(-80, -86), MAKE_VAL_COMPLEX(-36, -24), MAKE_VAL_COMPLEX(22, 26), MAKE_VAL_COMPLEX(-40, -67), MAKE_VAL_COMPLEX(-100, 124) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_141, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_141", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_141() */
/* END TEST 141 */

/*
 *  BEGIN TEST 142
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_142( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(54, 3), MAKE_VAL_COMPLEX(25, -60), MAKE_VAL_COMPLEX(-17, -84), MAKE_VAL_COMPLEX(-90, -5), MAKE_VAL_COMPLEX(-58, -61), MAKE_VAL_COMPLEX(-96, 38), MAKE_VAL_COMPLEX(74, 33), MAKE_VAL_COMPLEX(-30, -110), MAKE_VAL_COMPLEX(-12, 21), MAKE_VAL_COMPLEX(-20, -95), MAKE_VAL_COMPLEX(67, 29), MAKE_VAL_COMPLEX(103, -34), MAKE_VAL_COMPLEX(-37, 16), MAKE_VAL_COMPLEX(83, 1), MAKE_VAL_COMPLEX(-20, -30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_142, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_142", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_142() */
/* END TEST 142 */

/*
 *  BEGIN TEST 143
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x96', LAYOUT_COLMAJ>
 *  y : <'y96', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_143( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A96_ptr, A96_ind, A96_val,
		A96_M, A96_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y96_val,
		y96_LEN * y96_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y96_LEN, y96_VW, LAYOUT_COLMAJ, y96_LEN );
	oski_value_t* y_val = array_Duplicate( x96_val,
		x96_LEN * x96_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x96_LEN, x96_VW, LAYOUT_COLMAJ, x96_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(104, 536), MAKE_VAL_COMPLEX(-826, 74), MAKE_VAL_COMPLEX(-147, 21), MAKE_VAL_COMPLEX(-79, 65), MAKE_VAL_COMPLEX(-31, 9), MAKE_VAL_COMPLEX(434, 470), MAKE_VAL_COMPLEX(909, 191), MAKE_VAL_COMPLEX(-153, 13), MAKE_VAL_COMPLEX(46, 78), MAKE_VAL_COMPLEX(449, 65), MAKE_VAL_COMPLEX(198, -416), MAKE_VAL_COMPLEX(-291, 39), MAKE_VAL_COMPLEX(7, -3), MAKE_VAL_COMPLEX(3, -103), MAKE_VAL_COMPLEX(-418, -498) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_143, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_143", y_true, y_val, x96_LEN*x96_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_143() */
/* END TEST 143 */

#define A144_M 4    /* # of rows */
#define A144_N 5    /* # of cols */
#define A144_NNZ 4  /* # of stored non-zeros */
static oski_index_t A144_ptr[A144_M+1] = {0, 1, 1, 2, 4}; /* end row pointers for A144 */
static oski_index_t A144_ind[A144_NNZ] = {1, 3, 1, 4}; /* end column indices pointers for A144 */
static oski_value_t A144_val[A144_NNZ] = {MAKE_VAL_COMPLEX(4, -10), MAKE_VAL_COMPLEX(6, 0), MAKE_VAL_COMPLEX(3, 7), MAKE_VAL_COMPLEX(-10, 10)}; /* end stored non-zero values for A144 */

#define x144_LEN 5   /* vector length */
#define x144_VW  3   /* vector width */
static oski_value_t x144_val[x144_LEN*x144_VW] = {MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(4, 8), MAKE_VAL_COMPLEX(2, 1), MAKE_VAL_COMPLEX(4, -3), MAKE_VAL_COMPLEX(-10, 6), MAKE_VAL_COMPLEX(-5, -1), MAKE_VAL_COMPLEX(-7, 10), MAKE_VAL_COMPLEX(-2, 2), MAKE_VAL_COMPLEX(10, 0), MAKE_VAL_COMPLEX(-6, 7), MAKE_VAL_COMPLEX(-8, 4), MAKE_VAL_COMPLEX(-2, -10), MAKE_VAL_COMPLEX(-7, 1), MAKE_VAL_COMPLEX(-8, 2), MAKE_VAL_COMPLEX(-3, 6)}; /* data */

#define y144_LEN 4   /* vector length */
#define y144_VW  3   /* vector width */
static oski_value_t y144_val[y144_LEN*y144_VW] = {MAKE_VAL_COMPLEX(1, 1), MAKE_VAL_COMPLEX(-2, -6), MAKE_VAL_COMPLEX(10, -4), MAKE_VAL_COMPLEX(-8, 7), MAKE_VAL_COMPLEX(-10, 10), MAKE_VAL_COMPLEX(-1, -6), MAKE_VAL_COMPLEX(-3, -10), MAKE_VAL_COMPLEX(10, -7), MAKE_VAL_COMPLEX(10, -4), MAKE_VAL_COMPLEX(-3, 5), MAKE_VAL_COMPLEX(6, -3), MAKE_VAL_COMPLEX(2, 0)}; /* data */

/*
 *  BEGIN TEST 144
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_144( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	oski_value_t* y_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-10, -38), MAKE_VAL_COMPLEX(36, 68), MAKE_VAL_COMPLEX(82, 61), MAKE_VAL_COMPLEX(-103, 20), MAKE_VAL_COMPLEX(-150, 46), MAKE_VAL_COMPLEX(16, -60), MAKE_VAL_COMPLEX(-20, -53), MAKE_VAL_COMPLEX(75, 13), MAKE_VAL_COMPLEX(108, -46), MAKE_VAL_COMPLEX(35, -24), MAKE_VAL_COMPLEX(49, -145), MAKE_VAL_COMPLEX(-22, -128) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_144, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_144", y_true, y_val, y144_LEN*y144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_144() */
/* END TEST 144 */

/*
 *  BEGIN TEST 145
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_145( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	oski_value_t* y_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(21, 43), MAKE_VAL_COMPLEX(-70, -110), MAKE_VAL_COMPLEX(-14, -123), MAKE_VAL_COMPLEX(60, 60), MAKE_VAL_COMPLEX(100, 64), MAKE_VAL_COMPLEX(-42, 15), MAKE_VAL_COMPLEX(-34, -18), MAKE_VAL_COMPLEX(-16, -99), MAKE_VAL_COMPLEX(-40, -16), MAKE_VAL_COMPLEX(-44, 73), MAKE_VAL_COMPLEX(-10, 103), MAKE_VAL_COMPLEX(38, 122) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_145, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_145", y_true, y_val, y144_LEN*y144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_145() */
/* END TEST 145 */

/*
 *  BEGIN TEST 146
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_146( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	oski_value_t* y_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 6), MAKE_VAL_COMPLEX(12, -24), MAKE_VAL_COMPLEX(42, 18), MAKE_VAL_COMPLEX(-45, -3), MAKE_VAL_COMPLEX(-60, 0), MAKE_VAL_COMPLEX(15, -21), MAKE_VAL_COMPLEX(21, -39), MAKE_VAL_COMPLEX(51, 9), MAKE_VAL_COMPLEX(42, 18), MAKE_VAL_COMPLEX(-24, 6), MAKE_VAL_COMPLEX(27, 9), MAKE_VAL_COMPLEX(6, 6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_146, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_146", y_true, y_val, y144_LEN*y144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_146() */
/* END TEST 146 */

/*
 *  BEGIN TEST 147
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_147( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	oski_value_t* y_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-290, 410), MAKE_VAL_COMPLEX(820, -940), MAKE_VAL_COMPLEX(467, -16), MAKE_VAL_COMPLEX(-40, -85), MAKE_VAL_COMPLEX(102, -66), MAKE_VAL_COMPLEX(58, 66), MAKE_VAL_COMPLEX(687, 24), MAKE_VAL_COMPLEX(467, 279), MAKE_VAL_COMPLEX(-438, 444), MAKE_VAL_COMPLEX(-910, -325), MAKE_VAL_COMPLEX(-931, 1073), MAKE_VAL_COMPLEX(-736, 1118) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_147, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_147", y_true, y_val, y144_LEN*y144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_147() */
/* END TEST 147 */

/*
 *  BEGIN TEST 148
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_148( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1, 1), MAKE_VAL_COMPLEX(6, -10), MAKE_VAL_COMPLEX(11, -6), MAKE_VAL_COMPLEX(-41, -9), MAKE_VAL_COMPLEX(-33, 49), MAKE_VAL_COMPLEX(4, -103), MAKE_VAL_COMPLEX(7, -3), MAKE_VAL_COMPLEX(12, -5), MAKE_VAL_COMPLEX(10, -14), MAKE_VAL_COMPLEX(-14, -49), MAKE_VAL_COMPLEX(70, -37), MAKE_VAL_COMPLEX(52, -22), MAKE_VAL_COMPLEX(-19, -73), MAKE_VAL_COMPLEX(-28, 98), MAKE_VAL_COMPLEX(-14, 23) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_148, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_148", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_148() */
/* END TEST 148 */

/*
 *  BEGIN TEST 149
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_149( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1, -1), MAKE_VAL_COMPLEX(54, -10), MAKE_VAL_COMPLEX(-2, -7), MAKE_VAL_COMPLEX(24, -22), MAKE_VAL_COMPLEX(65, 27), MAKE_VAL_COMPLEX(-16, 137), MAKE_VAL_COMPLEX(56, 62), MAKE_VAL_COMPLEX(0, 21), MAKE_VAL_COMPLEX(0, -56), MAKE_VAL_COMPLEX(57, 98), MAKE_VAL_COMPLEX(-50, 97), MAKE_VAL_COMPLEX(-124, 26), MAKE_VAL_COMPLEX(19, 123), MAKE_VAL_COMPLEX(34, -40), MAKE_VAL_COMPLEX(53, 4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_149, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_149", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_149() */
/* END TEST 149 */

/*
 *  BEGIN TEST 150
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_150( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(-4, 32), MAKE_VAL_COMPLEX(4, 7), MAKE_VAL_COMPLEX(18, -1), MAKE_VAL_COMPLEX(-42, -2), MAKE_VAL_COMPLEX(-13, -13), MAKE_VAL_COMPLEX(-41, 16), MAKE_VAL_COMPLEX(-10, 2), MAKE_VAL_COMPLEX(30, 20), MAKE_VAL_COMPLEX(-32, 9), MAKE_VAL_COMPLEX(-32, -4), MAKE_VAL_COMPLEX(14, -34), MAKE_VAL_COMPLEX(-23, -11), MAKE_VAL_COMPLEX(-28, -10), MAKE_VAL_COMPLEX(-21, 12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_150, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_150", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_150() */
/* END TEST 150 */

/*
 *  BEGIN TEST 151
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_151( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-12, 0), MAKE_VAL_COMPLEX(60, 88), MAKE_VAL_COMPLEX(-24, 95), MAKE_VAL_COMPLEX(292, -211), MAKE_VAL_COMPLEX(-82, -446), MAKE_VAL_COMPLEX(585, 661), MAKE_VAL_COMPLEX(9, 98), MAKE_VAL_COMPLEX(-36, 112), MAKE_VAL_COMPLEX(34, 104), MAKE_VAL_COMPLEX(400, 241), MAKE_VAL_COMPLEX(-190, 678), MAKE_VAL_COMPLEX(-222, 442), MAKE_VAL_COMPLEX(549, 353), MAKE_VAL_COMPLEX(-420, -722), MAKE_VAL_COMPLEX(-33, -204) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_151, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_151", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_151() */
/* END TEST 151 */

/*
 *  BEGIN TEST 152
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_152( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1, 1), MAKE_VAL_COMPLEX(54, 6), MAKE_VAL_COMPLEX(23, -10), MAKE_VAL_COMPLEX(13, 25), MAKE_VAL_COMPLEX(29, -11), MAKE_VAL_COMPLEX(60, 85), MAKE_VAL_COMPLEX(19, 65), MAKE_VAL_COMPLEX(12, 11), MAKE_VAL_COMPLEX(50, -54), MAKE_VAL_COMPLEX(-10, 3), MAKE_VAL_COMPLEX(54, 11), MAKE_VAL_COMPLEX(4, -54), MAKE_VAL_COMPLEX(57, 19), MAKE_VAL_COMPLEX(-112, 18), MAKE_VAL_COMPLEX(-2, 19) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_152, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_152", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_152() */
/* END TEST 152 */

/*
 *  BEGIN TEST 153
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_153( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1, -1), MAKE_VAL_COMPLEX(-30, -58), MAKE_VAL_COMPLEX(-26, -4), MAKE_VAL_COMPLEX(-44, -33), MAKE_VAL_COMPLEX(41, 37), MAKE_VAL_COMPLEX(-45, -56), MAKE_VAL_COMPLEX(59, -70), MAKE_VAL_COMPLEX(6, -9), MAKE_VAL_COMPLEX(-90, 4), MAKE_VAL_COMPLEX(69, -1), MAKE_VAL_COMPLEX(-2, 13), MAKE_VAL_COMPLEX(-46, 104), MAKE_VAL_COMPLEX(-24, 12), MAKE_VAL_COMPLEX(154, 14), MAKE_VAL_COMPLEX(44, -28) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_153, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_153", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_153() */
/* END TEST 153 */

/*
 *  BEGIN TEST 154
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_154( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(-40, 0), MAKE_VAL_COMPLEX(-8, 6), MAKE_VAL_COMPLEX(4, 22), MAKE_VAL_COMPLEX(-4, -52), MAKE_VAL_COMPLEX(14, -18), MAKE_VAL_COMPLEX(-26, -48), MAKE_VAL_COMPLEX(-4, -12), MAKE_VAL_COMPLEX(-20, 40), MAKE_VAL_COMPLEX(-16, -38), MAKE_VAL_COMPLEX(0, -40), MAKE_VAL_COMPLEX(44, 12), MAKE_VAL_COMPLEX(10, -30), MAKE_VAL_COMPLEX(8, -36), MAKE_VAL_COMPLEX(-18, -24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_154, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_154", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_154() */
/* END TEST 154 */

/*
 *  BEGIN TEST 155
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_155( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(8, 4), MAKE_VAL_COMPLEX(-96, -36), MAKE_VAL_COMPLEX(37, -34), MAKE_VAL_COMPLEX(191, 358), MAKE_VAL_COMPLEX(62, -682), MAKE_VAL_COMPLEX(665, 183), MAKE_VAL_COMPLEX(-80, -143), MAKE_VAL_COMPLEX(40, -84), MAKE_VAL_COMPLEX(12, 26), MAKE_VAL_COMPLEX(-261, -358), MAKE_VAL_COMPLEX(310, -474), MAKE_VAL_COMPLEX(402, -242), MAKE_VAL_COMPLEX(461, -333), MAKE_VAL_COMPLEX(-582, -64), MAKE_VAL_COMPLEX(-190, -125) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_155, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_155", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_155() */
/* END TEST 155 */

/*
 *  BEGIN TEST 156
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_156( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	oski_value_t* y_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-13, -63), MAKE_VAL_COMPLEX(-22, 39), MAKE_VAL_COMPLEX(-117, 12), MAKE_VAL_COMPLEX(169, -41), MAKE_VAL_COMPLEX(134, 142), MAKE_VAL_COMPLEX(-41, 42), MAKE_VAL_COMPLEX(-95, 94), MAKE_VAL_COMPLEX(-112, -173), MAKE_VAL_COMPLEX(-102, 7), MAKE_VAL_COMPLEX(40, -8), MAKE_VAL_COMPLEX(-50, -81), MAKE_VAL_COMPLEX(-50, -150) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_156, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_156", y_true, y_val, y144_LEN*y144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_156() */
/* END TEST 156 */

/*
 *  BEGIN TEST 157
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_157( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	oski_value_t* y_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(11, 65), MAKE_VAL_COMPLEX(34, -43), MAKE_VAL_COMPLEX(125, 8), MAKE_VAL_COMPLEX(-183, 25), MAKE_VAL_COMPLEX(-154, -162), MAKE_VAL_COMPLEX(53, -44), MAKE_VAL_COMPLEX(115, -100), MAKE_VAL_COMPLEX(126, 193), MAKE_VAL_COMPLEX(110, 13), MAKE_VAL_COMPLEX(-50, 2), MAKE_VAL_COMPLEX(56, 93), MAKE_VAL_COMPLEX(50, 154) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_157, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_157", y_true, y_val, y144_LEN*y144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_157() */
/* END TEST 157 */

/*
 *  BEGIN TEST 158
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_158( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	oski_value_t* y_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(5, 7), MAKE_VAL_COMPLEX(-6, -38), MAKE_VAL_COMPLEX(64, -14), MAKE_VAL_COMPLEX(-55, 34), MAKE_VAL_COMPLEX(-70, 50), MAKE_VAL_COMPLEX(0, -37), MAKE_VAL_COMPLEX(-8, -63), MAKE_VAL_COMPLEX(67, -32), MAKE_VAL_COMPLEX(64, -14), MAKE_VAL_COMPLEX(-23, 27), MAKE_VAL_COMPLEX(39, -12), MAKE_VAL_COMPLEX(12, 2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_158, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_158", y_true, y_val, y144_LEN*y144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_158() */
/* END TEST 158 */

/*
 *  BEGIN TEST 159
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_159( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	oski_value_t* y_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-23, -59), MAKE_VAL_COMPLEX(26, 43), MAKE_VAL_COMPLEX(-119, 94), MAKE_VAL_COMPLEX(144, -118), MAKE_VAL_COMPLEX(94, 42), MAKE_VAL_COMPLEX(4, 53), MAKE_VAL_COMPLEX(-16, 103), MAKE_VAL_COMPLEX(-93, -82), MAKE_VAL_COMPLEX(-104, 89), MAKE_VAL_COMPLEX(14, -44), MAKE_VAL_COMPLEX(-47, -30), MAKE_VAL_COMPLEX(-56, -136) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_159, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_159", y_true, y_val, y144_LEN*y144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_159() */
/* END TEST 159 */

/*
 *  BEGIN TEST 160
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_160( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1, 1), MAKE_VAL_COMPLEX(-6, 58), MAKE_VAL_COMPLEX(18, 5), MAKE_VAL_COMPLEX(-35, -54), MAKE_VAL_COMPLEX(76, 193), MAKE_VAL_COMPLEX(-20, -112), MAKE_VAL_COMPLEX(-45, 32), MAKE_VAL_COMPLEX(-17, -4), MAKE_VAL_COMPLEX(56, 17), MAKE_VAL_COMPLEX(8, 6), MAKE_VAL_COMPLEX(-56, -63), MAKE_VAL_COMPLEX(48, -72), MAKE_VAL_COMPLEX(-27, -159), MAKE_VAL_COMPLEX(-74, 164), MAKE_VAL_COMPLEX(-47, 44) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_160, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_160", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_160() */
/* END TEST 160 */

/*
 *  BEGIN TEST 161
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_161( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1, -1), MAKE_VAL_COMPLEX(86, 22), MAKE_VAL_COMPLEX(10, -1), MAKE_VAL_COMPLEX(71, 2), MAKE_VAL_COMPLEX(-172, -81), MAKE_VAL_COMPLEX(-44, 120), MAKE_VAL_COMPLEX(1, 116), MAKE_VAL_COMPLEX(1, 36), MAKE_VAL_COMPLEX(64, -57), MAKE_VAL_COMPLEX(-52, 102), MAKE_VAL_COMPLEX(-24, 143), MAKE_VAL_COMPLEX(-112, -40), MAKE_VAL_COMPLEX(-53, 199), MAKE_VAL_COMPLEX(-14, -108), MAKE_VAL_COMPLEX(35, 40) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_161, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_161", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_161() */
/* END TEST 161 */

/*
 *  BEGIN TEST 162
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_162( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(-64, -48), MAKE_VAL_COMPLEX(-20, 0), MAKE_VAL_COMPLEX(-20, 40), MAKE_VAL_COMPLEX(56, -88), MAKE_VAL_COMPLEX(44, -12), MAKE_VAL_COMPLEX(16, -108), MAKE_VAL_COMPLEX(8, -24), MAKE_VAL_COMPLEX(-80, 40), MAKE_VAL_COMPLEX(20, -80), MAKE_VAL_COMPLEX(48, -64), MAKE_VAL_COMPLEX(56, 72), MAKE_VAL_COMPLEX(52, -36), MAKE_VAL_COMPLEX(56, -48), MAKE_VAL_COMPLEX(0, -60) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_162, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_162", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_162() */
/* END TEST 162 */

/*
 *  BEGIN TEST 163
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_163( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(2, -2), MAKE_VAL_COMPLEX(60, -80), MAKE_VAL_COMPLEX(-15, -46), MAKE_VAL_COMPLEX(-153, 110), MAKE_VAL_COMPLEX(500, -234), MAKE_VAL_COMPLEX(-163, 47), MAKE_VAL_COMPLEX(138, -37), MAKE_VAL_COMPLEX(14, 8), MAKE_VAL_COMPLEX(-86, -102), MAKE_VAL_COMPLEX(77, -106), MAKE_VAL_COMPLEX(-34, 56), MAKE_VAL_COMPLEX(-110, 22), MAKE_VAL_COMPLEX(-235, 35), MAKE_VAL_COMPLEX(422, 116), MAKE_VAL_COMPLEX(118, 19) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_163, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_163", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_163() */
/* END TEST 163 */

/*
 *  BEGIN TEST 164
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_164( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1, 1), MAKE_VAL_COMPLEX(22, -86), MAKE_VAL_COMPLEX(2, -28), MAKE_VAL_COMPLEX(-39, 77), MAKE_VAL_COMPLEX(-32, -125), MAKE_VAL_COMPLEX(115, 123), MAKE_VAL_COMPLEX(146, -28), MAKE_VAL_COMPLEX(29, -10), MAKE_VAL_COMPLEX(-74, -83), MAKE_VAL_COMPLEX(156, -25), MAKE_VAL_COMPLEX(88, -35), MAKE_VAL_COMPLEX(-26, 78), MAKE_VAL_COMPLEX(214, 58), MAKE_VAL_COMPLEX(-90, 18), MAKE_VAL_COMPLEX(52, -44) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_164, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_164", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_164() */
/* END TEST 164 */

/*
 *  BEGIN TEST 165
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_165( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1, -1), MAKE_VAL_COMPLEX(-90, -30), MAKE_VAL_COMPLEX(-33, 15), MAKE_VAL_COMPLEX(-18, -28), MAKE_VAL_COMPLEX(176, 25), MAKE_VAL_COMPLEX(-39, -113), MAKE_VAL_COMPLEX(-51, -129), MAKE_VAL_COMPLEX(-1, -22), MAKE_VAL_COMPLEX(-76, 93), MAKE_VAL_COMPLEX(-73, -86), MAKE_VAL_COMPLEX(28, -33), MAKE_VAL_COMPLEX(66, 70), MAKE_VAL_COMPLEX(-110, -80), MAKE_VAL_COMPLEX(208, -56), MAKE_VAL_COMPLEX(-13, -49) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_165, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_165", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_165() */
/* END TEST 165 */

/*
 *  BEGIN TEST 166
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_166( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(32, -36), MAKE_VAL_COMPLEX(1, -12), MAKE_VAL_COMPLEX(-23, -14), MAKE_VAL_COMPLEX(50, 38), MAKE_VAL_COMPLEX(5, 27), MAKE_VAL_COMPLEX(64, 15), MAKE_VAL_COMPLEX(14, 6), MAKE_VAL_COMPLEX(-20, -50), MAKE_VAL_COMPLEX(47, 16), MAKE_VAL_COMPLEX(36, 32), MAKE_VAL_COMPLEX(-46, 30), MAKE_VAL_COMPLEX(19, 33), MAKE_VAL_COMPLEX(26, 36), MAKE_VAL_COMPLEX(36, 3) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_166, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_166", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_166() */
/* END TEST 166 */

/*
 *  BEGIN TEST 167
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_ROWMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_167( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_ROWMAJ, x144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-7, -11), MAKE_VAL_COMPLEX(58, -54), MAKE_VAL_COMPLEX(-106, 27), MAKE_VAL_COMPLEX(31, -772), MAKE_VAL_COMPLEX(1100, 1667), MAKE_VAL_COMPLEX(-586, -870), MAKE_VAL_COMPLEX(-151, -36), MAKE_VAL_COMPLEX(-3, 82), MAKE_VAL_COMPLEX(-70, 75), MAKE_VAL_COMPLEX(-538, 6), MAKE_VAL_COMPLEX(-78, 567), MAKE_VAL_COMPLEX(-316, 84), MAKE_VAL_COMPLEX(-1329, -433), MAKE_VAL_COMPLEX(1466, 560), MAKE_VAL_COMPLEX(107, 196) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-9, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_167, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_167", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_167() */
/* END TEST 167 */

/*
 *  BEGIN TEST 168
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_168( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	oski_value_t* y_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(107, -3), MAKE_VAL_COMPLEX(33, 67), MAKE_VAL_COMPLEX(-48, -78), MAKE_VAL_COMPLEX(-39, 88), MAKE_VAL_COMPLEX(-57, 120), MAKE_VAL_COMPLEX(-28, -55), MAKE_VAL_COMPLEX(-28, -88), MAKE_VAL_COMPLEX(117, -84), MAKE_VAL_COMPLEX(13, -49), MAKE_VAL_COMPLEX(-9, -62), MAKE_VAL_COMPLEX(-52, -191), MAKE_VAL_COMPLEX(42, -138) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_168, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_168", y_true, y_val, y144_LEN*y144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_168() */
/* END TEST 168 */

/*
 *  BEGIN TEST 169
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_169( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	oski_value_t* y_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-92, 6), MAKE_VAL_COMPLEX(-87, -109), MAKE_VAL_COMPLEX(114, -18), MAKE_VAL_COMPLEX(9, 23), MAKE_VAL_COMPLEX(27, 30), MAKE_VAL_COMPLEX(-17, 7), MAKE_VAL_COMPLEX(-59, 16), MAKE_VAL_COMPLEX(-69, -39), MAKE_VAL_COMPLEX(53, -47), MAKE_VAL_COMPLEX(12, 125), MAKE_VAL_COMPLEX(88, 128), MAKE_VAL_COMPLEX(-24, 126) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_169, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_169", y_true, y_val, y144_LEN*y144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_169() */
/* END TEST 169 */

/*
 *  BEGIN TEST 170
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_170( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	oski_value_t* y_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(4, 10), MAKE_VAL_COMPLEX(4, -48), MAKE_VAL_COMPLEX(82, 2), MAKE_VAL_COMPLEX(-77, 25), MAKE_VAL_COMPLEX(-100, 40), MAKE_VAL_COMPLEX(11, -45), MAKE_VAL_COMPLEX(9, -79), MAKE_VAL_COMPLEX(91, -19), MAKE_VAL_COMPLEX(82, 2), MAKE_VAL_COMPLEX(-36, 26), MAKE_VAL_COMPLEX(51, -3), MAKE_VAL_COMPLEX(14, 6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_170, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_170", y_true, y_val, y144_LEN*y144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_170() */
/* END TEST 170 */

/*
 *  BEGIN TEST 171
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_171( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	oski_value_t* y_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(81, -211), MAKE_VAL_COMPLEX(259, 17), MAKE_VAL_COMPLEX(-222, 176), MAKE_VAL_COMPLEX(102, 13), MAKE_VAL_COMPLEX(123, 34), MAKE_VAL_COMPLEX(-53, 30), MAKE_VAL_COMPLEX(-53, -1), MAKE_VAL_COMPLEX(-30, -139), MAKE_VAL_COMPLEX(-103, 83), MAKE_VAL_COMPLEX(-177, -118), MAKE_VAL_COMPLEX(-437, 12), MAKE_VAL_COMPLEX(-248, -196) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_171, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_171", y_true, y_val, y144_LEN*y144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_171() */
/* END TEST 171 */

/*
 *  BEGIN TEST 172
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_172( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1, 1), MAKE_VAL_COMPLEX(54, 19), MAKE_VAL_COMPLEX(19, -19), MAKE_VAL_COMPLEX(-17, -108), MAKE_VAL_COMPLEX(-42, 42), MAKE_VAL_COMPLEX(-45, 27), MAKE_VAL_COMPLEX(-8, 165), MAKE_VAL_COMPLEX(12, 23), MAKE_VAL_COMPLEX(136, -125), MAKE_VAL_COMPLEX(-16, 187), MAKE_VAL_COMPLEX(-14, 88), MAKE_VAL_COMPLEX(-89, -162), MAKE_VAL_COMPLEX(-31, 59), MAKE_VAL_COMPLEX(22, 54), MAKE_VAL_COMPLEX(7, 86) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_172, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_172", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_172() */
/* END TEST 172 */

/*
 *  BEGIN TEST 173
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_173( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1, -1), MAKE_VAL_COMPLEX(2, -27), MAKE_VAL_COMPLEX(-9, 9), MAKE_VAL_COMPLEX(7, 78), MAKE_VAL_COMPLEX(58, 30), MAKE_VAL_COMPLEX(29, 1), MAKE_VAL_COMPLEX(54, -103), MAKE_VAL_COMPLEX(-4, -7), MAKE_VAL_COMPLEX(-116, 65), MAKE_VAL_COMPLEX(46, -137), MAKE_VAL_COMPLEX(22, -32), MAKE_VAL_COMPLEX(25, 154), MAKE_VAL_COMPLEX(23, -15), MAKE_VAL_COMPLEX(-26, -2), MAKE_VAL_COMPLEX(23, -56) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_173, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_173", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_173() */
/* END TEST 173 */

/*
 *  BEGIN TEST 174
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_174( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(-28, -36), MAKE_VAL_COMPLEX(-11, -3), MAKE_VAL_COMPLEX(-17, 19), MAKE_VAL_COMPLEX(44, -40), MAKE_VAL_COMPLEX(26, 0), MAKE_VAL_COMPLEX(25, -57), MAKE_VAL_COMPLEX(8, -12), MAKE_VAL_COMPLEX(-50, 10), MAKE_VAL_COMPLEX(23, -41), MAKE_VAL_COMPLEX(36, -28), MAKE_VAL_COMPLEX(20, 48), MAKE_VAL_COMPLEX(34, -12), MAKE_VAL_COMPLEX(38, -18), MAKE_VAL_COMPLEX(9, -33) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_174, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_174", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_174() */
/* END TEST 174 */

/*
 *  BEGIN TEST 175
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_175( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-5, -7), MAKE_VAL_COMPLEX(183, 28), MAKE_VAL_COMPLEX(-6, 61), MAKE_VAL_COMPLEX(53, 377), MAKE_VAL_COMPLEX(88, 444), MAKE_VAL_COMPLEX(38, 34), MAKE_VAL_COMPLEX(295, -269), MAKE_VAL_COMPLEX(-59, 16), MAKE_VAL_COMPLEX(-501, 224), MAKE_VAL_COMPLEX(292, -564), MAKE_VAL_COMPLEX(-24, -26), MAKE_VAL_COMPLEX(-106, 699), MAKE_VAL_COMPLEX(-24, -6), MAKE_VAL_COMPLEX(-352, 54), MAKE_VAL_COMPLEX(146, -142) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_175, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_175", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_175() */
/* END TEST 175 */

/*
 *  BEGIN TEST 176
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_176( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1, 1), MAKE_VAL_COMPLEX(0, 113), MAKE_VAL_COMPLEX(3, 3), MAKE_VAL_COMPLEX(11, -54), MAKE_VAL_COMPLEX(6, -30), MAKE_VAL_COMPLEX(-23, -31), MAKE_VAL_COMPLEX(-36, -63), MAKE_VAL_COMPLEX(-8, -5), MAKE_VAL_COMPLEX(116, -5), MAKE_VAL_COMPLEX(-148, -19), MAKE_VAL_COMPLEX(-46, -16), MAKE_VAL_COMPLEX(115, 6), MAKE_VAL_COMPLEX(-29, -27), MAKE_VAL_COMPLEX(14, -46), MAKE_VAL_COMPLEX(-59, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_176, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_176", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_176() */
/* END TEST 176 */

/*
 *  BEGIN TEST 177
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_177( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1, -1), MAKE_VAL_COMPLEX(-4, -141), MAKE_VAL_COMPLEX(-8, -8), MAKE_VAL_COMPLEX(-26, 59), MAKE_VAL_COMPLEX(30, 22), MAKE_VAL_COMPLEX(37, 39), MAKE_VAL_COMPLEX(67, 40), MAKE_VAL_COMPLEX(16, 1), MAKE_VAL_COMPLEX(-146, -5), MAKE_VAL_COMPLEX(173, 4), MAKE_VAL_COMPLEX(74, 12), MAKE_VAL_COMPLEX(-119, 26), MAKE_VAL_COMPLEX(51, 31), MAKE_VAL_COMPLEX(12, 48), MAKE_VAL_COMPLEX(74, -13) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_177, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_177", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_177() */
/* END TEST 177 */

/*
 *  BEGIN TEST 178
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_178( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(-44, 32), MAKE_VAL_COMPLEX(-4, 13), MAKE_VAL_COMPLEX(22, 21), MAKE_VAL_COMPLEX(-46, -54), MAKE_VAL_COMPLEX(1, -31), MAKE_VAL_COMPLEX(-67, -32), MAKE_VAL_COMPLEX(-14, -10), MAKE_VAL_COMPLEX(10, 60), MAKE_VAL_COMPLEX(-48, -29), MAKE_VAL_COMPLEX(-32, -44), MAKE_VAL_COMPLEX(58, -22), MAKE_VAL_COMPLEX(-13, -41), MAKE_VAL_COMPLEX(-20, -46), MAKE_VAL_COMPLEX(-39, -12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_178, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_178", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_178() */
/* END TEST 178 */

/*
 *  BEGIN TEST 179
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_179( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_ROWMAJ, y144_VW );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-4, 2), MAKE_VAL_COMPLEX(-183, 59), MAKE_VAL_COMPLEX(45, 17), MAKE_VAL_COMPLEX(230, -16), MAKE_VAL_COMPLEX(-124, 268), MAKE_VAL_COMPLEX(-16, -28), MAKE_VAL_COMPLEX(120, 254), MAKE_VAL_COMPLEX(-13, 45), MAKE_VAL_COMPLEX(149, 283), MAKE_VAL_COMPLEX(104, -208), MAKE_VAL_COMPLEX(-78, 34), MAKE_VAL_COMPLEX(-253, 103), MAKE_VAL_COMPLEX(-58, 14), MAKE_VAL_COMPLEX(-62, 194), MAKE_VAL_COMPLEX(32, -4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_179, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_179", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_179() */
/* END TEST 179 */

/*
 *  BEGIN TEST 180
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_180( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	oski_value_t* y_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(106, -4), MAKE_VAL_COMPLEX(-28, -28), MAKE_VAL_COMPLEX(84, -75), MAKE_VAL_COMPLEX(-35, -38), MAKE_VAL_COMPLEX(27, 209), MAKE_VAL_COMPLEX(-32, -29), MAKE_VAL_COMPLEX(7, -59), MAKE_VAL_COMPLEX(-42, -228), MAKE_VAL_COMPLEX(-58, -74), MAKE_VAL_COMPLEX(-8, 34), MAKE_VAL_COMPLEX(-22, -26), MAKE_VAL_COMPLEX(40, -138) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_180, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_180", y_true, y_val, y144_LEN*y144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_180() */
/* END TEST 180 */

/*
 *  BEGIN TEST 181
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_181( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	oski_value_t* y_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-103, 17), MAKE_VAL_COMPLEX(42, -30), MAKE_VAL_COMPLEX(16, 93), MAKE_VAL_COMPLEX(-64, 54), MAKE_VAL_COMPLEX(-157, -179), MAKE_VAL_COMPLEX(54, -24), MAKE_VAL_COMPLEX(19, -36), MAKE_VAL_COMPLEX(157, 222), MAKE_VAL_COMPLEX(158, 92), MAKE_VAL_COMPLEX(-41, -9), MAKE_VAL_COMPLEX(85, 32), MAKE_VAL_COMPLEX(-24, 148) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_181, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_181", y_true, y_val, y144_LEN*y144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_181() */
/* END TEST 181 */

/*
 *  BEGIN TEST 182
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_182( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	oski_value_t* y_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(6, 10), MAKE_VAL_COMPLEX(-4, -52), MAKE_VAL_COMPLEX(88, -12), MAKE_VAL_COMPLEX(-78, 40), MAKE_VAL_COMPLEX(-100, 60), MAKE_VAL_COMPLEX(4, -50), MAKE_VAL_COMPLEX(-4, -86), MAKE_VAL_COMPLEX(94, -36), MAKE_VAL_COMPLEX(88, -12), MAKE_VAL_COMPLEX(-34, 34), MAKE_VAL_COMPLEX(54, -12), MAKE_VAL_COMPLEX(16, 4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_182, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_182", y_true, y_val, y144_LEN*y144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_182() */
/* END TEST 182 */

/*
 *  BEGIN TEST 183
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_183( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	oski_value_t* y_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(215, -887), MAKE_VAL_COMPLEX(90, -14), MAKE_VAL_COMPLEX(-71, -275), MAKE_VAL_COMPLEX(-1006, -341), MAKE_VAL_COMPLEX(1172, -286), MAKE_VAL_COMPLEX(75, 92), MAKE_VAL_COMPLEX(202, -519), MAKE_VAL_COMPLEX(-1607, 382), MAKE_VAL_COMPLEX(-488, 1006), MAKE_VAL_COMPLEX(-101, -15), MAKE_VAL_COMPLEX(-45, 540), MAKE_VAL_COMPLEX(-1110, -628) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, -9);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_183, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_183", y_true, y_val, y144_LEN*y144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_183() */
/* END TEST 183 */

/*
 *  BEGIN TEST 184
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_184( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1, 1), MAKE_VAL_COMPLEX(-13, -31), MAKE_VAL_COMPLEX(22, -8), MAKE_VAL_COMPLEX(56, -45), MAKE_VAL_COMPLEX(-6, -86), MAKE_VAL_COMPLEX(-34, 26), MAKE_VAL_COMPLEX(150, 251), MAKE_VAL_COMPLEX(-3, 6), MAKE_VAL_COMPLEX(32, -107), MAKE_VAL_COMPLEX(-26, 222), MAKE_VAL_COMPLEX(-6, 44), MAKE_VAL_COMPLEX(-45, -129), MAKE_VAL_COMPLEX(-18, 29), MAKE_VAL_COMPLEX(14, 22), MAKE_VAL_COMPLEX(-8, 56) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_184, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_184", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_184() */
/* END TEST 184 */

/*
 *  BEGIN TEST 185
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_185( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1, -1), MAKE_VAL_COMPLEX(53, 31), MAKE_VAL_COMPLEX(-14, 2), MAKE_VAL_COMPLEX(-60, 23), MAKE_VAL_COMPLEX(10, 138), MAKE_VAL_COMPLEX(20, -8), MAKE_VAL_COMPLEX(-124, -203), MAKE_VAL_COMPLEX(7, 6), MAKE_VAL_COMPLEX(-12, 67), MAKE_VAL_COMPLEX(42, -184), MAKE_VAL_COMPLEX(6, -4), MAKE_VAL_COMPLEX(1, 117), MAKE_VAL_COMPLEX(8, 1), MAKE_VAL_COMPLEX(-22, 14), MAKE_VAL_COMPLEX(26, -32) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_185, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_185", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_185() */
/* END TEST 185 */

/*
 *  BEGIN TEST 186
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_186( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(28, -24), MAKE_VAL_COMPLEX(2, -9), MAKE_VAL_COMPLEX(-16, -13), MAKE_VAL_COMPLEX(34, 34), MAKE_VAL_COMPLEX(1, 21), MAKE_VAL_COMPLEX(47, 18), MAKE_VAL_COMPLEX(10, 6), MAKE_VAL_COMPLEX(-10, -40), MAKE_VAL_COMPLEX(34, 17), MAKE_VAL_COMPLEX(24, 28), MAKE_VAL_COMPLEX(-38, 18), MAKE_VAL_COMPLEX(11, 27), MAKE_VAL_COMPLEX(16, 30), MAKE_VAL_COMPLEX(27, 6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_186, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_186", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_186() */
/* END TEST 186 */

/*
 *  BEGIN TEST 187
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_187( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(17, -1), MAKE_VAL_COMPLEX(-963, 169), MAKE_VAL_COMPLEX(33, -115), MAKE_VAL_COMPLEX(266, -575), MAKE_VAL_COMPLEX(-1270, -1358), MAKE_VAL_COMPLEX(30, 148), MAKE_VAL_COMPLEX(2722, 157), MAKE_VAL_COMPLEX(-118, -69), MAKE_VAL_COMPLEX(-697, -414), MAKE_VAL_COMPLEX(1273, 1579), MAKE_VAL_COMPLEX(48, -174), MAKE_VAL_COMPLEX(-793, -783), MAKE_VAL_COMPLEX(37, -116), MAKE_VAL_COMPLEX(156, -532), MAKE_VAL_COMPLEX(-1, 307) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, -9);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_187, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_187", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_187() */
/* END TEST 187 */

/*
 *  BEGIN TEST 188
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_188( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1, 1), MAKE_VAL_COMPLEX(61, 153), MAKE_VAL_COMPLEX(29, 3), MAKE_VAL_COMPLEX(85, -48), MAKE_VAL_COMPLEX(66, 74), MAKE_VAL_COMPLEX(-56, 6), MAKE_VAL_COMPLEX(-213, -60), MAKE_VAL_COMPLEX(-19, 10), MAKE_VAL_COMPLEX(82, -77), MAKE_VAL_COMPLEX(-217, 39), MAKE_VAL_COMPLEX(-58, 40), MAKE_VAL_COMPLEX(55, -13), MAKE_VAL_COMPLEX(-56, 13), MAKE_VAL_COMPLEX(-32, 8), MAKE_VAL_COMPLEX(-41, 37) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_188, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_188", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_188() */
/* END TEST 188 */

/*
 *  BEGIN TEST 189
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_189( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1, -1), MAKE_VAL_COMPLEX(27, -177), MAKE_VAL_COMPLEX(-15, -21), MAKE_VAL_COMPLEX(-107, 2), MAKE_VAL_COMPLEX(-26, 38), MAKE_VAL_COMPLEX(36, 42), MAKE_VAL_COMPLEX(299, 150), MAKE_VAL_COMPLEX(35, 14), MAKE_VAL_COMPLEX(-62, -23), MAKE_VAL_COMPLEX(275, 35), MAKE_VAL_COMPLEX(82, 48), MAKE_VAL_COMPLEX(-159, 13), MAKE_VAL_COMPLEX(52, 59), MAKE_VAL_COMPLEX(36, 76), MAKE_VAL_COMPLEX(95, 5) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_189, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_189", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_189() */
/* END TEST 189 */

/*
 *  BEGIN TEST 190
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_190( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(36, 52), MAKE_VAL_COMPLEX(15, 5), MAKE_VAL_COMPLEX(25, -25), MAKE_VAL_COMPLEX(-64, 52), MAKE_VAL_COMPLEX(-36, -2), MAKE_VAL_COMPLEX(-39, 77), MAKE_VAL_COMPLEX(-12, 16), MAKE_VAL_COMPLEX(70, -10), MAKE_VAL_COMPLEX(-35, 55), MAKE_VAL_COMPLEX(-52, 36), MAKE_VAL_COMPLEX(-24, -68), MAKE_VAL_COMPLEX(-48, 14), MAKE_VAL_COMPLEX(-54, 22), MAKE_VAL_COMPLEX(-15, 45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_190, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_190", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_190() */
/* END TEST 190 */

/*
 *  BEGIN TEST 191
 *  Value type: complex
 *  A : <MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x144', LAYOUT_COLMAJ>
 *  y : <'y144', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_191( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A144_ptr, A144_ind, A144_val,
		A144_M, A144_N, SHARE_INPUTMAT,
		5, MAT_GENERAL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y144_val,
		y144_LEN * y144_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y144_LEN, y144_VW, LAYOUT_COLMAJ, y144_LEN );
	oski_value_t* y_val = array_Duplicate( x144_val,
		x144_LEN * x144_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x144_LEN, x144_VW, LAYOUT_COLMAJ, x144_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1, -1), MAKE_VAL_COMPLEX(49, -29), MAKE_VAL_COMPLEX(-13, -7), MAKE_VAL_COMPLEX(-20, -31), MAKE_VAL_COMPLEX(22, -198), MAKE_VAL_COMPLEX(28, -2), MAKE_VAL_COMPLEX(-166, 109), MAKE_VAL_COMPLEX(-10, -9), MAKE_VAL_COMPLEX(-97, 38), MAKE_VAL_COMPLEX(-33, 131), MAKE_VAL_COMPLEX(8, -46), MAKE_VAL_COMPLEX(111, -59), MAKE_VAL_COMPLEX(15, -30), MAKE_VAL_COMPLEX(0, -68), MAKE_VAL_COMPLEX(-29, -7) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_191, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_191", y_true, y_val, x144_LEN*x144_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_191() */
/* END TEST 191 */

#define A192_M 4    /* # of rows */
#define A192_N 5    /* # of cols */
#define A192_NNZ 4  /* # of stored non-zeros */
static oski_index_t A192_ptr[A192_M+1] = {1, 2, 4, 5, 5}; /* end row pointers for A192 */
static oski_index_t A192_ind[A192_NNZ] = {3, 2, 5, 5}; /* end column indices pointers for A192 */
static oski_value_t A192_val[A192_NNZ] = {MAKE_VAL_COMPLEX(-10, 4), MAKE_VAL_COMPLEX(0, 10), MAKE_VAL_COMPLEX(10, -10), MAKE_VAL_COMPLEX(2, 7)}; /* end stored non-zero values for A192 */

#define x192_LEN 5   /* vector length */
#define x192_VW  3   /* vector width */
static oski_value_t x192_val[x192_LEN*x192_VW] = {MAKE_VAL_COMPLEX(-8, 0), MAKE_VAL_COMPLEX(10, -10), MAKE_VAL_COMPLEX(3, -4), MAKE_VAL_COMPLEX(-4, 2), MAKE_VAL_COMPLEX(5, 6), MAKE_VAL_COMPLEX(-7, -3), MAKE_VAL_COMPLEX(0, 4), MAKE_VAL_COMPLEX(0, -10), MAKE_VAL_COMPLEX(-5, -2), MAKE_VAL_COMPLEX(-10, 10), MAKE_VAL_COMPLEX(0, 1), MAKE_VAL_COMPLEX(-8, 10), MAKE_VAL_COMPLEX(-3, 5), MAKE_VAL_COMPLEX(9, 4), MAKE_VAL_COMPLEX(-9, 5)}; /* data */

#define y192_LEN 4   /* vector length */
#define y192_VW  3   /* vector width */
static oski_value_t y192_val[y192_LEN*y192_VW] = {MAKE_VAL_COMPLEX(-10, 3), MAKE_VAL_COMPLEX(8, -10), MAKE_VAL_COMPLEX(-4, -4), MAKE_VAL_COMPLEX(10, -1), MAKE_VAL_COMPLEX(-7, -6), MAKE_VAL_COMPLEX(8, 10), MAKE_VAL_COMPLEX(-7, 10), MAKE_VAL_COMPLEX(1, 7), MAKE_VAL_COMPLEX(7, -1), MAKE_VAL_COMPLEX(2, 0), MAKE_VAL_COMPLEX(-10, -9), MAKE_VAL_COMPLEX(2, -9)}; /* data */

/*
 *  BEGIN TEST 192
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_192( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	oski_value_t* y_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-60, -114), MAKE_VAL_COMPLEX(136, 144), MAKE_VAL_COMPLEX(82, -40), MAKE_VAL_COMPLEX(28, 118), MAKE_VAL_COMPLEX(104, -68), MAKE_VAL_COMPLEX(-74, 154), MAKE_VAL_COMPLEX(-135, -47), MAKE_VAL_COMPLEX(-64, 93), MAKE_VAL_COMPLEX(-31, 1), MAKE_VAL_COMPLEX(4, 16), MAKE_VAL_COMPLEX(52, -98), MAKE_VAL_COMPLEX(76, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_192, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_192", y_true, y_val, y192_LEN*y192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_192() */
/* END TEST 192 */

/*
 *  BEGIN TEST 193
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_193( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	oski_value_t* y_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(84, -24), MAKE_VAL_COMPLEX(-64, 12), MAKE_VAL_COMPLEX(-10, 16), MAKE_VAL_COMPLEX(-76, 8), MAKE_VAL_COMPLEX(10, 20), MAKE_VAL_COMPLEX(-94, -118), MAKE_VAL_COMPLEX(57, -97), MAKE_VAL_COMPLEX(-26, -123), MAKE_VAL_COMPLEX(1, 89), MAKE_VAL_COMPLEX(-16, 8), MAKE_VAL_COMPLEX(116, 32), MAKE_VAL_COMPLEX(20, 80) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_193, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_193", y_true, y_val, y192_LEN*y192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_193() */
/* END TEST 193 */

/*
 *  BEGIN TEST 194
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_194( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	oski_value_t* y_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(10, -3), MAKE_VAL_COMPLEX(-8, 10), MAKE_VAL_COMPLEX(4, 4), MAKE_VAL_COMPLEX(-10, 1), MAKE_VAL_COMPLEX(7, 6), MAKE_VAL_COMPLEX(-8, -10), MAKE_VAL_COMPLEX(7, -10), MAKE_VAL_COMPLEX(-1, -7), MAKE_VAL_COMPLEX(-7, 1), MAKE_VAL_COMPLEX(-2, 0), MAKE_VAL_COMPLEX(10, 9), MAKE_VAL_COMPLEX(-2, 9) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_194, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_194", y_true, y_val, y192_LEN*y192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_194() */
/* END TEST 194 */

/*
 *  BEGIN TEST 195
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_195( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	oski_value_t* y_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-242, 216), MAKE_VAL_COMPLEX(668, -496), MAKE_VAL_COMPLEX(-108, -440), MAKE_VAL_COMPLEX(278, -96), MAKE_VAL_COMPLEX(-124, -522), MAKE_VAL_COMPLEX(568, -116), MAKE_VAL_COMPLEX(42, 324), MAKE_VAL_COMPLEX(598, -92), MAKE_VAL_COMPLEX(-348, 520), MAKE_VAL_COMPLEX(-8, -4), MAKE_VAL_COMPLEX(22, 56), MAKE_VAL_COMPLEX(-26, 32) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-2, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_195, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_195", y_true, y_val, y192_LEN*y192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_195() */
/* END TEST 195 */

/*
 *  BEGIN TEST 196
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_196( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(40, 16), MAKE_VAL_COMPLEX(-70, 30), MAKE_VAL_COMPLEX(-23, 14), MAKE_VAL_COMPLEX(34, 98), MAKE_VAL_COMPLEX(47, -110), MAKE_VAL_COMPLEX(-71, 109), MAKE_VAL_COMPLEX(96, -90), MAKE_VAL_COMPLEX(-60, 182), MAKE_VAL_COMPLEX(77, 44), MAKE_VAL_COMPLEX(70, -30), MAKE_VAL_COMPLEX(2, -5), MAKE_VAL_COMPLEX(60, -34), MAKE_VAL_COMPLEX(31, -158), MAKE_VAL_COMPLEX(-214, -7), MAKE_VAL_COMPLEX(256, 60) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_196, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_196", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_196() */
/* END TEST 196 */

/*
 *  BEGIN TEST 197
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_197( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-56, 48), MAKE_VAL_COMPLEX(10, -130), MAKE_VAL_COMPLEX(-3, -46), MAKE_VAL_COMPLEX(-26, -62), MAKE_VAL_COMPLEX(11, 82), MAKE_VAL_COMPLEX(33, -59), MAKE_VAL_COMPLEX(-64, 98), MAKE_VAL_COMPLEX(-20, -202), MAKE_VAL_COMPLEX(-103, -8), MAKE_VAL_COMPLEX(-10, 130), MAKE_VAL_COMPLEX(6, 7), MAKE_VAL_COMPLEX(4, 118), MAKE_VAL_COMPLEX(3, 192), MAKE_VAL_COMPLEX(264, -57), MAKE_VAL_COMPLEX(-234, 22) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_197, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_197", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_197() */
/* END TEST 197 */

/*
 *  BEGIN TEST 198
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_198( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 56), MAKE_VAL_COMPLEX(-70, -70), MAKE_VAL_COMPLEX(-28, -21), MAKE_VAL_COMPLEX(14, 28), MAKE_VAL_COMPLEX(42, -35), MAKE_VAL_COMPLEX(-21, 49), MAKE_VAL_COMPLEX(28, 0), MAKE_VAL_COMPLEX(-70, 0), MAKE_VAL_COMPLEX(-14, 35), MAKE_VAL_COMPLEX(70, 70), MAKE_VAL_COMPLEX(7, 0), MAKE_VAL_COMPLEX(70, 56), MAKE_VAL_COMPLEX(35, 21), MAKE_VAL_COMPLEX(28, -63), MAKE_VAL_COMPLEX(35, 63) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_198, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_198", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_198() */
/* END TEST 198 */

/*
 *  BEGIN TEST 199
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_199( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(32, 16), MAKE_VAL_COMPLEX(-60, 20), MAKE_VAL_COMPLEX(-20, 10), MAKE_VAL_COMPLEX(-860, -720), MAKE_VAL_COMPLEX(72, 1006), MAKE_VAL_COMPLEX(182, -1414), MAKE_VAL_COMPLEX(-136, 1248), MAKE_VAL_COMPLEX(-756, -1336), MAKE_VAL_COMPLEX(-624, 274), MAKE_VAL_COMPLEX(60, -20), MAKE_VAL_COMPLEX(2, -4), MAKE_VAL_COMPLEX(52, -24), MAKE_VAL_COMPLEX(1086, 1146), MAKE_VAL_COMPLEX(1140, -1698), MAKE_VAL_COMPLEX(-2098, 1070) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_199, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_199", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_199() */
/* END TEST 199 */

/*
 *  BEGIN TEST 200
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_200( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(48, -16), MAKE_VAL_COMPLEX(-40, 80), MAKE_VAL_COMPLEX(-10, 30), MAKE_VAL_COMPLEX(10, -120), MAKE_VAL_COMPLEX(-102, 44), MAKE_VAL_COMPLEX(148, -76), MAKE_VAL_COMPLEX(104, -14), MAKE_VAL_COMPLEX(-100, 128), MAKE_VAL_COMPLEX(58, 58), MAKE_VAL_COMPLEX(40, -80), MAKE_VAL_COMPLEX(-2, -6), MAKE_VAL_COMPLEX(28, -76), MAKE_VAL_COMPLEX(174, 123), MAKE_VAL_COMPLEX(-21, -129), MAKE_VAL_COMPLEX(31, 81) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_200, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_200", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_200() */
/* END TEST 200 */

/*
 *  BEGIN TEST 201
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_201( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-72, 40), MAKE_VAL_COMPLEX(40, -140), MAKE_VAL_COMPLEX(7, -51), MAKE_VAL_COMPLEX(-16, 138), MAKE_VAL_COMPLEX(135, -41), MAKE_VAL_COMPLEX(-178, 88), MAKE_VAL_COMPLEX(-92, 26), MAKE_VAL_COMPLEX(70, -158), MAKE_VAL_COMPLEX(-79, -49), MAKE_VAL_COMPLEX(-40, 140), MAKE_VAL_COMPLEX(5, 9), MAKE_VAL_COMPLEX(-22, 130), MAKE_VAL_COMPLEX(-168, -99), MAKE_VAL_COMPLEX(60, 114), MAKE_VAL_COMPLEX(-43, -39) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_201, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_201", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_201() */
/* END TEST 201 */

/*
 *  BEGIN TEST 202
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_202( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-8, 8), MAKE_VAL_COMPLEX(0, -20), MAKE_VAL_COMPLEX(-1, -7), MAKE_VAL_COMPLEX(-2, 6), MAKE_VAL_COMPLEX(11, 1), MAKE_VAL_COMPLEX(-10, 4), MAKE_VAL_COMPLEX(4, 4), MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(-7, 3), MAKE_VAL_COMPLEX(0, 20), MAKE_VAL_COMPLEX(1, 1), MAKE_VAL_COMPLEX(2, 18), MAKE_VAL_COMPLEX(2, 8), MAKE_VAL_COMPLEX(13, -5), MAKE_VAL_COMPLEX(-4, 14) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_202, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_202", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_202() */
/* END TEST 202 */

/*
 *  BEGIN TEST 203
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_203( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(72, -40), MAKE_VAL_COMPLEX(-40, 140), MAKE_VAL_COMPLEX(-7, 51), MAKE_VAL_COMPLEX(-114, 582), MAKE_VAL_COMPLEX(425, -329), MAKE_VAL_COMPLEX(-682, 272), MAKE_VAL_COMPLEX(-672, -320), MAKE_VAL_COMPLEX(906, -78), MAKE_VAL_COMPLEX(23, -391), MAKE_VAL_COMPLEX(40, -140), MAKE_VAL_COMPLEX(-5, -9), MAKE_VAL_COMPLEX(22, -130), MAKE_VAL_COMPLEX(-676, -1346), MAKE_VAL_COMPLEX(-593, 665), MAKE_VAL_COMPLEX(392, -838) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_203, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_203", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_203() */
/* END TEST 203 */

/*
 *  BEGIN TEST 204
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_204( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	oski_value_t* y_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(35, 21), MAKE_VAL_COMPLEX(-94, 14), MAKE_VAL_COMPLEX(-57, 29), MAKE_VAL_COMPLEX(-37, -67), MAKE_VAL_COMPLEX(19, 167), MAKE_VAL_COMPLEX(116, -86), MAKE_VAL_COMPLEX(81, 90), MAKE_VAL_COMPLEX(46, -28), MAKE_VAL_COMPLEX(30, -46), MAKE_VAL_COMPLEX(-16, 56), MAKE_VAL_COMPLEX(-86, 44), MAKE_VAL_COMPLEX(-69, 13) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_204, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_204", y_true, y_val, y192_LEN*y192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_204() */
/* END TEST 204 */

/*
 *  BEGIN TEST 205
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_205( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	oski_value_t* y_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(17, 7), MAKE_VAL_COMPLEX(22, -6), MAKE_VAL_COMPLEX(57, 3), MAKE_VAL_COMPLEX(-7, 31), MAKE_VAL_COMPLEX(-15, -115), MAKE_VAL_COMPLEX(-108, 14), MAKE_VAL_COMPLEX(-13, -102), MAKE_VAL_COMPLEX(-22, -4), MAKE_VAL_COMPLEX(-62, 22), MAKE_VAL_COMPLEX(8, -64), MAKE_VAL_COMPLEX(90, 32), MAKE_VAL_COMPLEX(25, 15) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_205, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_205", y_true, y_val, y192_LEN*y192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_205() */
/* END TEST 205 */

/*
 *  BEGIN TEST 206
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_206( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	oski_value_t* y_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-65, 74), MAKE_VAL_COMPLEX(14, -120), MAKE_VAL_COMPLEX(-52, -12), MAKE_VAL_COMPLEX(75, -58), MAKE_VAL_COMPLEX(-86, -13), MAKE_VAL_COMPLEX(114, 40), MAKE_VAL_COMPLEX(-6, 115), MAKE_VAL_COMPLEX(43, 51), MAKE_VAL_COMPLEX(51, -43), MAKE_VAL_COMPLEX(16, -10), MAKE_VAL_COMPLEX(-125, -22), MAKE_VAL_COMPLEX(-29, -82) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_206, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_206", y_true, y_val, y192_LEN*y192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_206() */
/* END TEST 206 */

/*
 *  BEGIN TEST 207
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_207( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	oski_value_t* y_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-376, -63), MAKE_VAL_COMPLEX(356, 98), MAKE_VAL_COMPLEX(-193, 161), MAKE_VAL_COMPLEX(76, 53), MAKE_VAL_COMPLEX(707, -24), MAKE_VAL_COMPLEX(206, -302), MAKE_VAL_COMPLEX(287, 301), MAKE_VAL_COMPLEX(-35, 55), MAKE_VAL_COMPLEX(229, -313), MAKE_VAL_COMPLEX(404, 282), MAKE_VAL_COMPLEX(-493, 36), MAKE_VAL_COMPLEX(68, -51) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_207, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_207", y_true, y_val, y192_LEN*y192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_207() */
/* END TEST 207 */

/*
 *  BEGIN TEST 208
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_208( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-48, 40), MAKE_VAL_COMPLEX(10, -110), MAKE_VAL_COMPLEX(-2, -39), MAKE_VAL_COMPLEX(86, 112), MAKE_VAL_COMPLEX(-40, 91), MAKE_VAL_COMPLEX(-57, 37), MAKE_VAL_COMPLEX(108, -46), MAKE_VAL_COMPLEX(44, -28), MAKE_VAL_COMPLEX(-106, 51), MAKE_VAL_COMPLEX(-10, 110), MAKE_VAL_COMPLEX(5, 6), MAKE_VAL_COMPLEX(2, 100), MAKE_VAL_COMPLEX(7, -171), MAKE_VAL_COMPLEX(170, -30), MAKE_VAL_COMPLEX(34, -33) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_208, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_208", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_208() */
/* END TEST 208 */

/*
 *  BEGIN TEST 209
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_209( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(72, -8), MAKE_VAL_COMPLEX(-80, 100), MAKE_VAL_COMPLEX(-23, 39), MAKE_VAL_COMPLEX(-66, -102), MAKE_VAL_COMPLEX(49, -129), MAKE_VAL_COMPLEX(66, 0), MAKE_VAL_COMPLEX(-92, 34), MAKE_VAL_COMPLEX(-84, 58), MAKE_VAL_COMPLEX(113, -25), MAKE_VAL_COMPLEX(80, -100), MAKE_VAL_COMPLEX(-1, -9), MAKE_VAL_COMPLEX(62, -98), MAKE_VAL_COMPLEX(22, 168), MAKE_VAL_COMPLEX(-181, -18), MAKE_VAL_COMPLEX(13, 54) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_209, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_209", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_209() */
/* END TEST 209 */

/*
 *  BEGIN TEST 210
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_210( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(8, 32), MAKE_VAL_COMPLEX(-50, -30), MAKE_VAL_COMPLEX(-19, -8), MAKE_VAL_COMPLEX(12, 14), MAKE_VAL_COMPLEX(19, -26), MAKE_VAL_COMPLEX(-5, 31), MAKE_VAL_COMPLEX(16, -4), MAKE_VAL_COMPLEX(-40, 10), MAKE_VAL_COMPLEX(-3, 22), MAKE_VAL_COMPLEX(50, 30), MAKE_VAL_COMPLEX(4, -1), MAKE_VAL_COMPLEX(48, 22), MAKE_VAL_COMPLEX(23, 7), MAKE_VAL_COMPLEX(7, -40), MAKE_VAL_COMPLEX(29, 31) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_210, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_210", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_210() */
/* END TEST 210 */

/*
 *  BEGIN TEST 211
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_211( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-56, 64), MAKE_VAL_COMPLEX(-10, -150), MAKE_VAL_COMPLEX(-11, -52), MAKE_VAL_COMPLEX(68, -54), MAKE_VAL_COMPLEX(163, 102), MAKE_VAL_COMPLEX(-53, 35), MAKE_VAL_COMPLEX(-38, -60), MAKE_VAL_COMPLEX(-48, -164), MAKE_VAL_COMPLEX(-13, 92), MAKE_VAL_COMPLEX(10, 150), MAKE_VAL_COMPLEX(8, 7), MAKE_VAL_COMPLEX(24, 134), MAKE_VAL_COMPLEX(-197, 59), MAKE_VAL_COMPLEX(86, -140), MAKE_VAL_COMPLEX(-131, 44) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_211, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_211", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_211() */
/* END TEST 211 */

/*
 *  BEGIN TEST 212
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_212( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(16, -56), MAKE_VAL_COMPLEX(50, 90), MAKE_VAL_COMPLEX(22, 29), MAKE_VAL_COMPLEX(-106, -112), MAKE_VAL_COMPLEX(48, -57), MAKE_VAL_COMPLEX(35, -63), MAKE_VAL_COMPLEX(84, 2), MAKE_VAL_COMPLEX(116, 108), MAKE_VAL_COMPLEX(-50, -49), MAKE_VAL_COMPLEX(-50, -90), MAKE_VAL_COMPLEX(-7, -2), MAKE_VAL_COMPLEX(-54, -76), MAKE_VAL_COMPLEX(115, -31), MAKE_VAL_COMPLEX(-10, 304), MAKE_VAL_COMPLEX(-80, -1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_212, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_212", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_212() */
/* END TEST 212 */

/*
 *  BEGIN TEST 213
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_213( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, -16), MAKE_VAL_COMPLEX(20, 20), MAKE_VAL_COMPLEX(8, 6), MAKE_VAL_COMPLEX(96, 72), MAKE_VAL_COMPLEX(-112, 90), MAKE_VAL_COMPLEX(6, 6), MAKE_VAL_COMPLEX(-120, -10), MAKE_VAL_COMPLEX(-26, -88), MAKE_VAL_COMPLEX(78, 8), MAKE_VAL_COMPLEX(-20, -20), MAKE_VAL_COMPLEX(-2, 0), MAKE_VAL_COMPLEX(-20, -16), MAKE_VAL_COMPLEX(-154, -6), MAKE_VAL_COMPLEX(-44, -231), MAKE_VAL_COMPLEX(53, -90) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_213, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_213", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_213() */
/* END TEST 213 */

/*
 *  BEGIN TEST 214
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_214( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(64, -72), MAKE_VAL_COMPLEX(10, 170), MAKE_VAL_COMPLEX(12, 59), MAKE_VAL_COMPLEX(14, -52), MAKE_VAL_COMPLEX(-94, -3), MAKE_VAL_COMPLEX(83, -39), MAKE_VAL_COMPLEX(-36, -32), MAKE_VAL_COMPLEX(90, 80), MAKE_VAL_COMPLEX(58, -29), MAKE_VAL_COMPLEX(-10, -170), MAKE_VAL_COMPLEX(-9, -8), MAKE_VAL_COMPLEX(-26, -152), MAKE_VAL_COMPLEX(-21, -67), MAKE_VAL_COMPLEX(-108, 49), MAKE_VAL_COMPLEX(27, -121) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_214, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_214", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_214() */
/* END TEST 214 */

/*
 *  BEGIN TEST 215
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_ROWMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_215( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_ROWMAJ, x192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-8, -48), MAKE_VAL_COMPLEX(70, 50), MAKE_VAL_COMPLEX(27, 14), MAKE_VAL_COMPLEX(544, 918), MAKE_VAL_COMPLEX(-1071, 376), MAKE_VAL_COMPLEX(-49, 115), MAKE_VAL_COMPLEX(-890, -412), MAKE_VAL_COMPLEX(-44, -852), MAKE_VAL_COMPLEX(545, 334), MAKE_VAL_COMPLEX(-70, -50), MAKE_VAL_COMPLEX(-6, 1), MAKE_VAL_COMPLEX(-68, -38), MAKE_VAL_COMPLEX(-1185, -445), MAKE_VAL_COMPLEX(444, -2042), MAKE_VAL_COMPLEX(681, -436) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, -3);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_215, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_215", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_215() */
/* END TEST 215 */

/*
 *  BEGIN TEST 216
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_216( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	oski_value_t* y_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-28, -31), MAKE_VAL_COMPLEX(112, 174), MAKE_VAL_COMPLEX(46, -90), MAKE_VAL_COMPLEX(208, 191), MAKE_VAL_COMPLEX(15, 150), MAKE_VAL_COMPLEX(-228, 114), MAKE_VAL_COMPLEX(-105, -19), MAKE_VAL_COMPLEX(-147, -49), MAKE_VAL_COMPLEX(-52, 4), MAKE_VAL_COMPLEX(-2, 16), MAKE_VAL_COMPLEX(82, -71), MAKE_VAL_COMPLEX(70, 25) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_216, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_216", y_true, y_val, y192_LEN*y192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_216() */
/* END TEST 216 */

/*
 *  BEGIN TEST 217
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_217( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	oski_value_t* y_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(6, -115), MAKE_VAL_COMPLEX(12, -42), MAKE_VAL_COMPLEX(18, 42), MAKE_VAL_COMPLEX(-214, -49), MAKE_VAL_COMPLEX(83, -236), MAKE_VAL_COMPLEX(72, -22), MAKE_VAL_COMPLEX(-21, -99), MAKE_VAL_COMPLEX(47, 49), MAKE_VAL_COMPLEX(52, 96), MAKE_VAL_COMPLEX(-2, 12), MAKE_VAL_COMPLEX(64, -51), MAKE_VAL_COMPLEX(52, 21) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_217, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_217", y_true, y_val, y192_LEN*y192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_217() */
/* END TEST 217 */

/*
 *  BEGIN TEST 218
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_218( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	oski_value_t* y_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(69, -97), MAKE_VAL_COMPLEX(-2, 146), MAKE_VAL_COMPLEX(64, 8), MAKE_VAL_COMPLEX(-83, 79), MAKE_VAL_COMPLEX(105, 5), MAKE_VAL_COMPLEX(-142, -34), MAKE_VAL_COMPLEX(-7, -139), MAKE_VAL_COMPLEX(-58, -56), MAKE_VAL_COMPLEX(-56, 58), MAKE_VAL_COMPLEX(-18, 14), MAKE_VAL_COMPLEX(153, 11), MAKE_VAL_COMPLEX(45, 95) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_218, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_218", y_true, y_val, y192_LEN*y192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_218() */
/* END TEST 218 */

/*
 *  BEGIN TEST 219
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_219( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	oski_value_t* y_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(463, 447), MAKE_VAL_COMPLEX(1054, 286), MAKE_VAL_COMPLEX(-482, -422), MAKE_VAL_COMPLEX(2177, -1253), MAKE_VAL_COMPLEX(1591, 1623), MAKE_VAL_COMPLEX(-326, 1526), MAKE_VAL_COMPLEX(310, 521), MAKE_VAL_COMPLEX(-976, 458), MAKE_VAL_COMPLEX(-847, 145), MAKE_VAL_COMPLEX(-14, -6), MAKE_VAL_COMPLEX(43, 93), MAKE_VAL_COMPLEX(-41, 57) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, -9);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_219, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_219", y_true, y_val, y192_LEN*y192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_219() */
/* END TEST 219 */

/*
 *  BEGIN TEST 220
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_220( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(72, -16), MAKE_VAL_COMPLEX(-60, 210), MAKE_VAL_COMPLEX(69, -28), MAKE_VAL_COMPLEX(32, -26), MAKE_VAL_COMPLEX(-51, -183), MAKE_VAL_COMPLEX(69, 13), MAKE_VAL_COMPLEX(52, -106), MAKE_VAL_COMPLEX(-20, 222), MAKE_VAL_COMPLEX(49, 8), MAKE_VAL_COMPLEX(-107, -79), MAKE_VAL_COMPLEX(-2, -9), MAKE_VAL_COMPLEX(-48, -26), MAKE_VAL_COMPLEX(73, -27), MAKE_VAL_COMPLEX(-89, -18), MAKE_VAL_COMPLEX(272, 4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_220, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_220", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_220() */
/* END TEST 220 */

/*
 *  BEGIN TEST 221
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_221( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-24, -32), MAKE_VAL_COMPLEX(60, -90), MAKE_VAL_COMPLEX(-63, 70), MAKE_VAL_COMPLEX(-20, -10), MAKE_VAL_COMPLEX(-15, 177), MAKE_VAL_COMPLEX(-9, -37), MAKE_VAL_COMPLEX(-76, 82), MAKE_VAL_COMPLEX(80, -162), MAKE_VAL_COMPLEX(-7, -26), MAKE_VAL_COMPLEX(107, -41), MAKE_VAL_COMPLEX(-4, 3), MAKE_VAL_COMPLEX(36, -82), MAKE_VAL_COMPLEX(-85, -21), MAKE_VAL_COMPLEX(11, 48), MAKE_VAL_COMPLEX(-248, -88) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_221, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_221", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_221() */
/* END TEST 221 */

/*
 *  BEGIN TEST 222
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_222( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-24, 32), MAKE_VAL_COMPLEX(-10, -70), MAKE_VAL_COMPLEX(-7, -24), MAKE_VAL_COMPLEX(-4, 22), MAKE_VAL_COMPLEX(39, -2), MAKE_VAL_COMPLEX(-33, 19), MAKE_VAL_COMPLEX(16, 12), MAKE_VAL_COMPLEX(-40, -30), MAKE_VAL_COMPLEX(-23, 14), MAKE_VAL_COMPLEX(10, 70), MAKE_VAL_COMPLEX(4, 3), MAKE_VAL_COMPLEX(16, 62), MAKE_VAL_COMPLEX(11, 27), MAKE_VAL_COMPLEX(43, -24), MAKE_VAL_COMPLEX(-7, 51) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_222, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_222", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_222() */
/* END TEST 222 */

/*
 *  BEGIN TEST 223
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_223( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(64, 48), MAKE_VAL_COMPLEX(-120, 220), MAKE_VAL_COMPLEX(128, -126), MAKE_VAL_COMPLEX(44, 8), MAKE_VAL_COMPLEX(8, -356), MAKE_VAL_COMPLEX(38, 66), MAKE_VAL_COMPLEX(144, -172), MAKE_VAL_COMPLEX(-140, 344), MAKE_VAL_COMPLEX(28, 46), MAKE_VAL_COMPLEX(-214, 42), MAKE_VAL_COMPLEX(6, -8), MAKE_VAL_COMPLEX(-76, 128), MAKE_VAL_COMPLEX(166, 26), MAKE_VAL_COMPLEX(-48, -86), MAKE_VAL_COMPLEX(504, 148) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_223, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_223", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_223() */
/* END TEST 223 */

/*
 *  BEGIN TEST 224
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_224( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-40, 0), MAKE_VAL_COMPLEX(40, -150), MAKE_VAL_COMPLEX(127, -10), MAKE_VAL_COMPLEX(-20, 10), MAKE_VAL_COMPLEX(191, 189), MAKE_VAL_COMPLEX(-35, -15), MAKE_VAL_COMPLEX(-60, 90), MAKE_VAL_COMPLEX(-120, 18), MAKE_VAL_COMPLEX(-25, -10), MAKE_VAL_COMPLEX(-9, -73), MAKE_VAL_COMPLEX(0, 5), MAKE_VAL_COMPLEX(60, -30), MAKE_VAL_COMPLEX(9, 81), MAKE_VAL_COMPLEX(45, 20), MAKE_VAL_COMPLEX(-58, 154) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_224, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_224", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_224() */
/* END TEST 224 */

/*
 *  BEGIN TEST 225
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_225( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-56, 32), MAKE_VAL_COMPLEX(40, -10), MAKE_VAL_COMPLEX(-107, -50), MAKE_VAL_COMPLEX(-20, 30), MAKE_VAL_COMPLEX(-107, -137), MAKE_VAL_COMPLEX(-61, 7), MAKE_VAL_COMPLEX(76, -42), MAKE_VAL_COMPLEX(80, -138), MAKE_VAL_COMPLEX(-43, 6), MAKE_VAL_COMPLEX(-71, 233), MAKE_VAL_COMPLEX(4, 7), MAKE_VAL_COMPLEX(-116, 182), MAKE_VAL_COMPLEX(-25, -9), MAKE_VAL_COMPLEX(79, -8), MAKE_VAL_COMPLEX(-30, -58) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_225, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_225", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_225() */
/* END TEST 225 */

/*
 *  BEGIN TEST 226
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_226( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-56, -64), MAKE_VAL_COMPLEX(150, 10), MAKE_VAL_COMPLEX(53, -4), MAKE_VAL_COMPLEX(-44, -18), MAKE_VAL_COMPLEX(-13, 82), MAKE_VAL_COMPLEX(-25, -77), MAKE_VAL_COMPLEX(-32, 28), MAKE_VAL_COMPLEX(80, -70), MAKE_VAL_COMPLEX(-19, -54), MAKE_VAL_COMPLEX(-150, -10), MAKE_VAL_COMPLEX(-8, 7), MAKE_VAL_COMPLEX(-136, 6), MAKE_VAL_COMPLEX(-61, 11), MAKE_VAL_COMPLEX(31, 100), MAKE_VAL_COMPLEX(-103, -37) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_226, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_226", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_226() */
/* END TEST 226 */

/*
 *  BEGIN TEST 227
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_227( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_ROWMAJ, y192_VW );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(56, -32), MAKE_VAL_COMPLEX(140, 800), MAKE_VAL_COMPLEX(-799, 82), MAKE_VAL_COMPLEX(20, -30), MAKE_VAL_COMPLEX(-1380, -969), MAKE_VAL_COMPLEX(61, -7), MAKE_VAL_COMPLEX(334, -578), MAKE_VAL_COMPLEX(812, -526), MAKE_VAL_COMPLEX(43, -6), MAKE_VAL_COMPLEX(-134, 792), MAKE_VAL_COMPLEX(-4, -7), MAKE_VAL_COMPLEX(-604, 558), MAKE_VAL_COMPLEX(-223, -415), MAKE_VAL_COMPLEX(-79, 8), MAKE_VAL_COMPLEX(5, -987) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_227, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_227", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_227() */
/* END TEST 227 */

/*
 *  BEGIN TEST 228
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_228( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	oski_value_t* y_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-19, -1), MAKE_VAL_COMPLEX(252, 160), MAKE_VAL_COMPLEX(-8, 31), MAKE_VAL_COMPLEX(-5, 51), MAKE_VAL_COMPLEX(77, 71), MAKE_VAL_COMPLEX(-98, 230), MAKE_VAL_COMPLEX(-133, -95), MAKE_VAL_COMPLEX(-36, -2), MAKE_VAL_COMPLEX(8, -26), MAKE_VAL_COMPLEX(-142, 70), MAKE_VAL_COMPLEX(2, -94), MAKE_VAL_COMPLEX(43, 19) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_228, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_228", y_true, y_val, y192_LEN*y192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_228() */
/* END TEST 228 */

/*
 *  BEGIN TEST 229
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_229( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	oski_value_t* y_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-68, -71), MAKE_VAL_COMPLEX(-114, -148), MAKE_VAL_COMPLEX(20, -91), MAKE_VAL_COMPLEX(74, 33), MAKE_VAL_COMPLEX(-65, -170), MAKE_VAL_COMPLEX(56, -98), MAKE_VAL_COMPLEX(1, 92), MAKE_VAL_COMPLEX(-21, 53), MAKE_VAL_COMPLEX(43, 83), MAKE_VAL_COMPLEX(154, -52), MAKE_VAL_COMPLEX(19, -50), MAKE_VAL_COMPLEX(50, -55) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_229, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_229", y_true, y_val, y192_LEN*y192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_229() */
/* END TEST 229 */

/*
 *  BEGIN TEST 230
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_230( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	oski_value_t* y_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-34, 32), MAKE_VAL_COMPLEX(12, -56), MAKE_VAL_COMPLEX(-24, -8), MAKE_VAL_COMPLEX(38, -24), MAKE_VAL_COMPLEX(-40, -10), MAKE_VAL_COMPLEX(52, 24), MAKE_VAL_COMPLEX(-8, 54), MAKE_VAL_COMPLEX(18, 26), MAKE_VAL_COMPLEX(26, -18), MAKE_VAL_COMPLEX(8, -4), MAKE_VAL_COMPLEX(-58, -16), MAKE_VAL_COMPLEX(-10, -40) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_230, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_230", y_true, y_val, y192_LEN*y192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_230() */
/* END TEST 230 */

/*
 *  BEGIN TEST 231
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_231( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	oski_value_t* y_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-456, 246), MAKE_VAL_COMPLEX(648, 1860), MAKE_VAL_COMPLEX(-498, 66), MAKE_VAL_COMPLEX(60, -6), MAKE_VAL_COMPLEX(-402, 804), MAKE_VAL_COMPLEX(-1392, 1020), MAKE_VAL_COMPLEX(-282, -780), MAKE_VAL_COMPLEX(6, 42), MAKE_VAL_COMPLEX(474, -318), MAKE_VAL_COMPLEX(-1188, -480), MAKE_VAL_COMPLEX(-60, -690), MAKE_VAL_COMPLEX(12, -54) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_231, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_231", y_true, y_val, y192_LEN*y192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_231() */
/* END TEST 231 */

/*
 *  BEGIN TEST 232
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_232( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(16, -16), MAKE_VAL_COMPLEX(100, 120), MAKE_VAL_COMPLEX(90, -56), MAKE_VAL_COMPLEX(4, -12), MAKE_VAL_COMPLEX(-22, -218), MAKE_VAL_COMPLEX(20, -8), MAKE_VAL_COMPLEX(-108, 72), MAKE_VAL_COMPLEX(114, 52), MAKE_VAL_COMPLEX(14, -6), MAKE_VAL_COMPLEX(96, -49), MAKE_VAL_COMPLEX(-2, -2), MAKE_VAL_COMPLEX(-4, -16), MAKE_VAL_COMPLEX(-70, 22), MAKE_VAL_COMPLEX(-26, 10), MAKE_VAL_COMPLEX(71, -136) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_232, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_232", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_232() */
/* END TEST 232 */

/*
 *  BEGIN TEST 233
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_233( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(8, -32), MAKE_VAL_COMPLEX(-70, -30), MAKE_VAL_COMPLEX(-75, 86), MAKE_VAL_COMPLEX(-4, -18), MAKE_VAL_COMPLEX(-29, 230), MAKE_VAL_COMPLEX(19, -25), MAKE_VAL_COMPLEX(84, -84), MAKE_VAL_COMPLEX(-54, -22), MAKE_VAL_COMPLEX(13, -18), MAKE_VAL_COMPLEX(-126, -41), MAKE_VAL_COMPLEX(-4, -1), MAKE_VAL_COMPLEX(-32, -62), MAKE_VAL_COMPLEX(49, -55), MAKE_VAL_COMPLEX(-25, 32), MAKE_VAL_COMPLEX(-74, 67) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_233, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_233", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_233() */
/* END TEST 233 */

/*
 *  BEGIN TEST 234
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_234( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-32, 32), MAKE_VAL_COMPLEX(0, -80), MAKE_VAL_COMPLEX(-4, -28), MAKE_VAL_COMPLEX(-8, 24), MAKE_VAL_COMPLEX(44, 4), MAKE_VAL_COMPLEX(-40, 16), MAKE_VAL_COMPLEX(16, 16), MAKE_VAL_COMPLEX(-40, -40), MAKE_VAL_COMPLEX(-28, 12), MAKE_VAL_COMPLEX(0, 80), MAKE_VAL_COMPLEX(4, 4), MAKE_VAL_COMPLEX(8, 72), MAKE_VAL_COMPLEX(8, 32), MAKE_VAL_COMPLEX(52, -20), MAKE_VAL_COMPLEX(-16, 56) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_234, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_234", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_234() */
/* END TEST 234 */

/*
 *  BEGIN TEST 235
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_235( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-24, 8), MAKE_VAL_COMPLEX(500, 1000), MAKE_VAL_COMPLEX(989, -223), MAKE_VAL_COMPLEX(-10, 10), MAKE_VAL_COMPLEX(885, -1715), MAKE_VAL_COMPLEX(-24, -2), MAKE_VAL_COMPLEX(-1116, 252), MAKE_VAL_COMPLEX(614, 602), MAKE_VAL_COMPLEX(-17, -1), MAKE_VAL_COMPLEX(784, 352), MAKE_VAL_COMPLEX(1, 3), MAKE_VAL_COMPLEX(-94, 198), MAKE_VAL_COMPLEX(-684, 58), MAKE_VAL_COMPLEX(31, 3), MAKE_VAL_COMPLEX(914, -588) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_235, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_235", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_235() */
/* END TEST 235 */

/*
 *  BEGIN TEST 236
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_236( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(16, -48), MAKE_VAL_COMPLEX(-60, 0), MAKE_VAL_COMPLEX(130, 36), MAKE_VAL_COMPLEX(-4, -28), MAKE_VAL_COMPLEX(98, 18), MAKE_VAL_COMPLEX(32, -36), MAKE_VAL_COMPLEX(76, -88), MAKE_VAL_COMPLEX(106, 108), MAKE_VAL_COMPLEX(22, -26), MAKE_VAL_COMPLEX(-4, 169), MAKE_VAL_COMPLEX(-6, -2), MAKE_VAL_COMPLEX(-44, -88), MAKE_VAL_COMPLEX(-98, -46), MAKE_VAL_COMPLEX(-42, 46), MAKE_VAL_COMPLEX(-75, 8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_236, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_236", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_236() */
/* END TEST 236 */

/*
 *  BEGIN TEST 237
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_237( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(64, 56), MAKE_VAL_COMPLEX(-50, 90), MAKE_VAL_COMPLEX(-164, 1), MAKE_VAL_COMPLEX(46, 12), MAKE_VAL_COMPLEX(-142, -83), MAKE_VAL_COMPLEX(35, 73), MAKE_VAL_COMPLEX(-72, 48), MAKE_VAL_COMPLEX(-116, -8), MAKE_VAL_COMPLEX(26, 51), MAKE_VAL_COMPLEX(114, -259), MAKE_VAL_COMPLEX(7, -8), MAKE_VAL_COMPLEX(134, -4), MAKE_VAL_COMPLEX(133, -1), MAKE_VAL_COMPLEX(-44, -95), MAKE_VAL_COMPLEX(170, -49) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_237, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_237", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_237() */
/* END TEST 237 */

/*
 *  BEGIN TEST 238
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_238( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-56, -24), MAKE_VAL_COMPLEX(100, -40), MAKE_VAL_COMPLEX(33, -19), MAKE_VAL_COMPLEX(-34, 2), MAKE_VAL_COMPLEX(17, 57), MAKE_VAL_COMPLEX(-40, -42), MAKE_VAL_COMPLEX(-12, 28), MAKE_VAL_COMPLEX(30, -70), MAKE_VAL_COMPLEX(-29, -29), MAKE_VAL_COMPLEX(-100, 40), MAKE_VAL_COMPLEX(-3, 7), MAKE_VAL_COMPLEX(-86, 46), MAKE_VAL_COMPLEX(-36, 26), MAKE_VAL_COMPLEX(51, 55), MAKE_VAL_COMPLEX(-78, 8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_238, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_238", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_238() */
/* END TEST 238 */

/*
 *  BEGIN TEST 239
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x192', LAYOUT_COLMAJ>
 *  y : <'y192', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_239( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A192_ptr, A192_ind, A192_val,
		A192_M, A192_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y192_val,
		y192_LEN * y192_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y192_LEN, y192_VW, LAYOUT_COLMAJ, y192_LEN );
	oski_value_t* y_val = array_Duplicate( x192_val,
		x192_LEN * x192_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x192_LEN, x192_VW, LAYOUT_COLMAJ, x192_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-56, 40), MAKE_VAL_COMPLEX(-100, 440), MAKE_VAL_COMPLEX(-183, -511), MAKE_VAL_COMPLEX(-18, 34), MAKE_VAL_COMPLEX(-223, -559), MAKE_VAL_COMPLEX(-64, 14), MAKE_VAL_COMPLEX(-500, -212), MAKE_VAL_COMPLEX(210, -430), MAKE_VAL_COMPLEX(-45, 11), MAKE_VAL_COMPLEX(904, -522), MAKE_VAL_COMPLEX(5, 7), MAKE_VAL_COMPLEX(-86, 150), MAKE_VAL_COMPLEX(80, 382), MAKE_VAL_COMPLEX(83, -17), MAKE_VAL_COMPLEX(376, 188) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-2, -4);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_239, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_239", y_true, y_val, x192_LEN*x192_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_239() */
/* END TEST 239 */

#define A240_M 4    /* # of rows */
#define A240_N 5    /* # of cols */
#define A240_NNZ 7  /* # of stored non-zeros */
static oski_index_t A240_ptr[A240_M+1] = {0, 2, 5, 7, 7}; /* end row pointers for A240 */
static oski_index_t A240_ind[A240_NNZ] = {0, 2, 1, 2, 4, 2, 3}; /* end column indices pointers for A240 */
static oski_value_t A240_val[A240_NNZ] = {MAKE_VAL_COMPLEX(-10, -1), MAKE_VAL_COMPLEX(0, -10), MAKE_VAL_COMPLEX(4, 10), MAKE_VAL_COMPLEX(5, 1), MAKE_VAL_COMPLEX(10, 3), MAKE_VAL_COMPLEX(0, 1), MAKE_VAL_COMPLEX(-1, 1)}; /* end stored non-zero values for A240 */

#define x240_LEN 5   /* vector length */
#define x240_VW  3   /* vector width */
static oski_value_t x240_val[x240_LEN*x240_VW] = {MAKE_VAL_COMPLEX(-10, 7), MAKE_VAL_COMPLEX(-9, -10), MAKE_VAL_COMPLEX(-3, 6), MAKE_VAL_COMPLEX(-2, -9), MAKE_VAL_COMPLEX(-5, -8), MAKE_VAL_COMPLEX(3, -2), MAKE_VAL_COMPLEX(10, -6), MAKE_VAL_COMPLEX(-6, -9), MAKE_VAL_COMPLEX(-1, -4), MAKE_VAL_COMPLEX(-9, 6), MAKE_VAL_COMPLEX(3, 2), MAKE_VAL_COMPLEX(2, -6), MAKE_VAL_COMPLEX(8, 10), MAKE_VAL_COMPLEX(9, 1), MAKE_VAL_COMPLEX(7, 8)}; /* data */

#define y240_LEN 4   /* vector length */
#define y240_VW  3   /* vector width */
static oski_value_t y240_val[y240_LEN*y240_VW] = {MAKE_VAL_COMPLEX(1, -8), MAKE_VAL_COMPLEX(6, 6), MAKE_VAL_COMPLEX(3, 7), MAKE_VAL_COMPLEX(-7, -10), MAKE_VAL_COMPLEX(-7, 10), MAKE_VAL_COMPLEX(-2, 7), MAKE_VAL_COMPLEX(-0, -7), MAKE_VAL_COMPLEX(-10, 1), MAKE_VAL_COMPLEX(6, -8), MAKE_VAL_COMPLEX(-2, -9), MAKE_VAL_COMPLEX(10, 1), MAKE_VAL_COMPLEX(6, -5)}; /* data */

/*
 *  BEGIN TEST 240
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_240( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	oski_value_t* y_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(49, -111), MAKE_VAL_COMPLEX(-52, 139), MAKE_VAL_COMPLEX(-29, -86), MAKE_VAL_COMPLEX(240, 101), MAKE_VAL_COMPLEX(158, -163), MAKE_VAL_COMPLEX(82, 58), MAKE_VAL_COMPLEX(16, 37), MAKE_VAL_COMPLEX(63, -21), MAKE_VAL_COMPLEX(-20, 61), MAKE_VAL_COMPLEX(21, 52), MAKE_VAL_COMPLEX(-61, 4), MAKE_VAL_COMPLEX(-31, 36) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_240, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_240", y_true, y_val, y240_LEN*y240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_240() */
/* END TEST 240 */

/*
 *  BEGIN TEST 241
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_241( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	oski_value_t* y_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-14, 91), MAKE_VAL_COMPLEX(46, -97), MAKE_VAL_COMPLEX(10, 119), MAKE_VAL_COMPLEX(-221, -159), MAKE_VAL_COMPLEX(-219, 165), MAKE_VAL_COMPLEX(-116, -45), MAKE_VAL_COMPLEX(12, -58), MAKE_VAL_COMPLEX(-97, -16), MAKE_VAL_COMPLEX(70, -61), MAKE_VAL_COMPLEX(9, -87), MAKE_VAL_COMPLEX(87, 39), MAKE_VAL_COMPLEX(69, -27) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_241, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_241", y_true, y_val, y240_LEN*y240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_241() */
/* END TEST 241 */

/*
 *  BEGIN TEST 242
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_242( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	oski_value_t* y_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(48, 71), MAKE_VAL_COMPLEX(-90, -6), MAKE_VAL_COMPLEX(-73, -35), MAKE_VAL_COMPLEX(126, 31), MAKE_VAL_COMPLEX(-14, -129), MAKE_VAL_COMPLEX(-33, -70), MAKE_VAL_COMPLEX(49, 56), MAKE_VAL_COMPLEX(73, -78), MAKE_VAL_COMPLEX(8, 106), MAKE_VAL_COMPLEX(79, 58), MAKE_VAL_COMPLEX(-87, 62), MAKE_VAL_COMPLEX(-13, 82) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_242, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_242", y_true, y_val, y240_LEN*y240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_242() */
/* END TEST 242 */

/*
 *  BEGIN TEST 243
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_243( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	oski_value_t* y_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(525, 1139), MAKE_VAL_COMPLEX(-743, -1022), MAKE_VAL_COMPLEX(280, 311), MAKE_VAL_COMPLEX(-1417, 582), MAKE_VAL_COMPLEX(-325, 1276), MAKE_VAL_COMPLEX(-986, -178), MAKE_VAL_COMPLEX(-29, 26), MAKE_VAL_COMPLEX(-69, 57), MAKE_VAL_COMPLEX(-41, -58), MAKE_VAL_COMPLEX(-14, -63), MAKE_VAL_COMPLEX(70, 7), MAKE_VAL_COMPLEX(42, -35) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_243, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_243", y_true, y_val, y240_LEN*y240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_243() */
/* END TEST 243 */

/*
 *  BEGIN TEST 244
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_244( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-54, -15), MAKE_VAL_COMPLEX(44, -118), MAKE_VAL_COMPLEX(-65, -109), MAKE_VAL_COMPLEX(148, -108), MAKE_VAL_COMPLEX(-54, -54), MAKE_VAL_COMPLEX(-68, 36), MAKE_VAL_COMPLEX(-70, 25), MAKE_VAL_COMPLEX(98, -57), MAKE_VAL_COMPLEX(95, 9), MAKE_VAL_COMPLEX(-23, -77), MAKE_VAL_COMPLEX(-13, 9), MAKE_VAL_COMPLEX(46, 42), MAKE_VAL_COMPLEX(-136, -77), MAKE_VAL_COMPLEX(-126, 149), MAKE_VAL_COMPLEX(-119, 104) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_244, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_244", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_244() */
/* END TEST 244 */

/*
 *  BEGIN TEST 245
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_245( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(53, -178), MAKE_VAL_COMPLEX(167, 91), MAKE_VAL_COMPLEX(14, 16), MAKE_VAL_COMPLEX(-13, 163), MAKE_VAL_COMPLEX(203, 61), MAKE_VAL_COMPLEX(67, 21), MAKE_VAL_COMPLEX(58, 159), MAKE_VAL_COMPLEX(73, 60), MAKE_VAL_COMPLEX(-34, 14), MAKE_VAL_COMPLEX(26, -94), MAKE_VAL_COMPLEX(-40, 12), MAKE_VAL_COMPLEX(14, 38), MAKE_VAL_COMPLEX(-66, 91), MAKE_VAL_COMPLEX(32, -41), MAKE_VAL_COMPLEX(-48, -85) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_245, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_245", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_245() */
/* END TEST 245 */

/*
 *  BEGIN TEST 246
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_246( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-106, -45), MAKE_VAL_COMPLEX(35, -122), MAKE_VAL_COMPLEX(-63, 6), MAKE_VAL_COMPLEX(62, -61), MAKE_VAL_COMPLEX(39, -80), MAKE_VAL_COMPLEX(31, 14), MAKE_VAL_COMPLEX(98, 50), MAKE_VAL_COMPLEX(42, -93), MAKE_VAL_COMPLEX(27, -28), MAKE_VAL_COMPLEX(-93, -42), MAKE_VAL_COMPLEX(-1, 34), MAKE_VAL_COMPLEX(58, -14), MAKE_VAL_COMPLEX(-40, 114), MAKE_VAL_COMPLEX(37, 77), MAKE_VAL_COMPLEX(-29, 96) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_246, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_246", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_246() */
/* END TEST 246 */

/*
 *  BEGIN TEST 247
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_247( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(20, -590), MAKE_VAL_COMPLEX(496, 319), MAKE_VAL_COMPLEX(181, 370), MAKE_VAL_COMPLEX(-251, 754), MAKE_VAL_COMPLEX(874, 49), MAKE_VAL_COMPLEX(462, -97), MAKE_VAL_COMPLEX(657, 398), MAKE_VAL_COMPLEX(30, 170), MAKE_VAL_COMPLEX(-343, 16), MAKE_VAL_COMPLEX(-55, -122), MAKE_VAL_COMPLEX(-69, 88), MAKE_VAL_COMPLEX(8, -44), MAKE_VAL_COMPLEX(259, 702), MAKE_VAL_COMPLEX(478, -515), MAKE_VAL_COMPLEX(98, -408) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_247, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_247", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_247() */
/* END TEST 247 */

/*
 *  BEGIN TEST 248
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_248( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-38, 136), MAKE_VAL_COMPLEX(-131, -86), MAKE_VAL_COMPLEX(-40, -31), MAKE_VAL_COMPLEX(-156, -11), MAKE_VAL_COMPLEX(31, 80), MAKE_VAL_COMPLEX(73, 32), MAKE_VAL_COMPLEX(66, -83), MAKE_VAL_COMPLEX(-132, 94), MAKE_VAL_COMPLEX(-94, 43), MAKE_VAL_COMPLEX(-40, 55), MAKE_VAL_COMPLEX(30, 13), MAKE_VAL_COMPLEX(-16, -32), MAKE_VAL_COMPLEX(-40, -45), MAKE_VAL_COMPLEX(7, 108), MAKE_VAL_COMPLEX(52, 102) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_248, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_248", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_248() */
/* END TEST 248 */

/*
 *  BEGIN TEST 249
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_249( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(79, -120), MAKE_VAL_COMPLEX(119, 133), MAKE_VAL_COMPLEX(64, 28), MAKE_VAL_COMPLEX(133, 35), MAKE_VAL_COMPLEX(-45, -49), MAKE_VAL_COMPLEX(-85, -37), MAKE_VAL_COMPLEX(-104, 65), MAKE_VAL_COMPLEX(117, -58), MAKE_VAL_COMPLEX(84, -32), MAKE_VAL_COMPLEX(76, -40), MAKE_VAL_COMPLEX(-30, -26), MAKE_VAL_COMPLEX(-6, 38), MAKE_VAL_COMPLEX(54, 1), MAKE_VAL_COMPLEX(-22, -137), MAKE_VAL_COMPLEX(-42, -139) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_249, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_249", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_249() */
/* END TEST 249 */

/*
 *  BEGIN TEST 250
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_250( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-18, -47), MAKE_VAL_COMPLEX(49, -26), MAKE_VAL_COMPLEX(-21, -18), MAKE_VAL_COMPLEX(38, 1), MAKE_VAL_COMPLEX(37, -12), MAKE_VAL_COMPLEX(5, 14), MAKE_VAL_COMPLEX(14, 46), MAKE_VAL_COMPLEX(42, -15), MAKE_VAL_COMPLEX(17, 0), MAKE_VAL_COMPLEX(-15, -42), MAKE_VAL_COMPLEX(-11, 10), MAKE_VAL_COMPLEX(22, 14), MAKE_VAL_COMPLEX(-48, 22), MAKE_VAL_COMPLEX(-13, 35), MAKE_VAL_COMPLEX(-39, 20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_250, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_250", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_250() */
/* END TEST 250 */

/*
 *  BEGIN TEST 251
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_251( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-455, 234), MAKE_VAL_COMPLEX(65, -540), MAKE_VAL_COMPLEX(203, -374), MAKE_VAL_COMPLEX(-522, -581), MAKE_VAL_COMPLEX(-333, 656), MAKE_VAL_COMPLEX(-41, 454), MAKE_VAL_COMPLEX(291, 43), MAKE_VAL_COMPLEX(-887, -78), MAKE_VAL_COMPLEX(-543, -236), MAKE_VAL_COMPLEX(-95, -14), MAKE_VAL_COMPLEX(-7, 94), MAKE_VAL_COMPLEX(-34, -78), MAKE_VAL_COMPLEX(99, -691), MAKE_VAL_COMPLEX(-700, 184), MAKE_VAL_COMPLEX(-372, 271) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_251, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_251", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_251() */
/* END TEST 251 */

/*
 *  BEGIN TEST 252
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_252( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	oski_value_t* y_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(51, -127), MAKE_VAL_COMPLEX(158, 30), MAKE_VAL_COMPLEX(-10, -30), MAKE_VAL_COMPLEX(38, 33), MAKE_VAL_COMPLEX(8, 122), MAKE_VAL_COMPLEX(127, -126), MAKE_VAL_COMPLEX(11, 23), MAKE_VAL_COMPLEX(39, -14), MAKE_VAL_COMPLEX(-20, -9), MAKE_VAL_COMPLEX(94, 136), MAKE_VAL_COMPLEX(-33, 13), MAKE_VAL_COMPLEX(-19, 26) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_252, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_252", y_true, y_val, y240_LEN*y240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_252() */
/* END TEST 252 */

/*
 *  BEGIN TEST 253
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_253( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	oski_value_t* y_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-14, 91), MAKE_VAL_COMPLEX(-152, 24), MAKE_VAL_COMPLEX(-3, 77), MAKE_VAL_COMPLEX(-33, -111), MAKE_VAL_COMPLEX(-83, -100), MAKE_VAL_COMPLEX(-165, 153), MAKE_VAL_COMPLEX(17, -58), MAKE_VAL_COMPLEX(-93, -21), MAKE_VAL_COMPLEX(82, -7), MAKE_VAL_COMPLEX(-68, -189), MAKE_VAL_COMPLEX(79, 32), MAKE_VAL_COMPLEX(69, -27) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_253, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_253", y_true, y_val, y240_LEN*y240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_253() */
/* END TEST 253 */

/*
 *  BEGIN TEST 254
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_254( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	oski_value_t* y_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(27, 44), MAKE_VAL_COMPLEX(-54, -6), MAKE_VAL_COMPLEX(-43, -23), MAKE_VAL_COMPLEX(75, 22), MAKE_VAL_COMPLEX(-5, -78), MAKE_VAL_COMPLEX(-18, -43), MAKE_VAL_COMPLEX(28, 35), MAKE_VAL_COMPLEX(46, -45), MAKE_VAL_COMPLEX(2, 64), MAKE_VAL_COMPLEX(46, 37), MAKE_VAL_COMPLEX(-54, 35), MAKE_VAL_COMPLEX(-10, 49) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_254, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_254", y_true, y_val, y240_LEN*y240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_254() */
/* END TEST 254 */

/*
 *  BEGIN TEST 255
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_255( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	oski_value_t* y_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-57, 1556), MAKE_VAL_COMPLEX(-1836, -32), MAKE_VAL_COMPLEX(-119, 67), MAKE_VAL_COMPLEX(74, -22), MAKE_VAL_COMPLEX(-294, -1603), MAKE_VAL_COMPLEX(-980, 1090), MAKE_VAL_COMPLEX(16, 67), MAKE_VAL_COMPLEX(14, -62), MAKE_VAL_COMPLEX(166, 467), MAKE_VAL_COMPLEX(-839, -758), MAKE_VAL_COMPLEX(-112, 11), MAKE_VAL_COMPLEX(18, 46) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-9, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_255, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_255", y_true, y_val, y240_LEN*y240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_255() */
/* END TEST 255 */

/*
 *  BEGIN TEST 256
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_256( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(77, 87), MAKE_VAL_COMPLEX(84, 12), MAKE_VAL_COMPLEX(-20, 53), MAKE_VAL_COMPLEX(-69, 148), MAKE_VAL_COMPLEX(-88, 81), MAKE_VAL_COMPLEX(54, -59), MAKE_VAL_COMPLEX(-153, 15), MAKE_VAL_COMPLEX(81, 187), MAKE_VAL_COMPLEX(-96, -68), MAKE_VAL_COMPLEX(74, 5), MAKE_VAL_COMPLEX(-1, -20), MAKE_VAL_COMPLEX(-53, 35), MAKE_VAL_COMPLEX(44, -22), MAKE_VAL_COMPLEX(-90, 13), MAKE_VAL_COMPLEX(5, -179) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_256, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_256", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_256() */
/* END TEST 256 */

/*
 *  BEGIN TEST 257
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_257( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-61, -128), MAKE_VAL_COMPLEX(-37, 0), MAKE_VAL_COMPLEX(17, -77), MAKE_VAL_COMPLEX(93, -125), MAKE_VAL_COMPLEX(119, -67), MAKE_VAL_COMPLEX(-59, 71), MAKE_VAL_COMPLEX(135, 23), MAKE_VAL_COMPLEX(-45, -172), MAKE_VAL_COMPLEX(107, 78), MAKE_VAL_COMPLEX(-59, -41), MAKE_VAL_COMPLEX(-12, 20), MAKE_VAL_COMPLEX(59, -13), MAKE_VAL_COMPLEX(-88, 8), MAKE_VAL_COMPLEX(61, 2), MAKE_VAL_COMPLEX(-42, 169) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_257, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_257", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_257() */
/* END TEST 257 */

/*
 *  BEGIN TEST 258
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_258( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1, 44), MAKE_VAL_COMPLEX(-48, 7), MAKE_VAL_COMPLEX(12, 21), MAKE_VAL_COMPLEX(-31, -12), MAKE_VAL_COMPLEX(-34, -1), MAKE_VAL_COMPLEX(0, -13), MAKE_VAL_COMPLEX(2, -42), MAKE_VAL_COMPLEX(-39, 0), MAKE_VAL_COMPLEX(-14, -5), MAKE_VAL_COMPLEX(0, 39), MAKE_VAL_COMPLEX(12, -5), MAKE_VAL_COMPLEX(-14, -18), MAKE_VAL_COMPLEX(46, -4), MAKE_VAL_COMPLEX(21, -25), MAKE_VAL_COMPLEX(38, -5) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_258, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_258", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_258() */
/* END TEST 258 */

/*
 *  BEGIN TEST 259
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_259( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(122, 415), MAKE_VAL_COMPLEX(-14, -653), MAKE_VAL_COMPLEX(-59, 533), MAKE_VAL_COMPLEX(89, 401), MAKE_VAL_COMPLEX(-321, 231), MAKE_VAL_COMPLEX(173, -483), MAKE_VAL_COMPLEX(-121, 265), MAKE_VAL_COMPLEX(630, 103), MAKE_VAL_COMPLEX(-628, -161), MAKE_VAL_COMPLEX(-91, 53), MAKE_VAL_COMPLEX(66, 14), MAKE_VAL_COMPLEX(-13, 37), MAKE_VAL_COMPLEX(452, 228), MAKE_VAL_COMPLEX(74, 375), MAKE_VAL_COMPLEX(-217, -372) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, -3);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_259, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_259", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_259() */
/* END TEST 259 */

/*
 *  BEGIN TEST 260
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_260( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(110, 92), MAKE_VAL_COMPLEX(63, 17), MAKE_VAL_COMPLEX(5, 62), MAKE_VAL_COMPLEX(44, 39), MAKE_VAL_COMPLEX(49, 134), MAKE_VAL_COMPLEX(-131, -20), MAKE_VAL_COMPLEX(17, 13), MAKE_VAL_COMPLEX(-122, 66), MAKE_VAL_COMPLEX(45, 41), MAKE_VAL_COMPLEX(103, 2), MAKE_VAL_COMPLEX(-16, -25), MAKE_VAL_COMPLEX(-59, 19), MAKE_VAL_COMPLEX(82, -76), MAKE_VAL_COMPLEX(-56, 15), MAKE_VAL_COMPLEX(-48, -182) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_260, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_260", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_260() */
/* END TEST 260 */

/*
 *  BEGIN TEST 261
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_261( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(10, -176), MAKE_VAL_COMPLEX(45, 103), MAKE_VAL_COMPLEX(31, -134), MAKE_VAL_COMPLEX(-20, 69), MAKE_VAL_COMPLEX(11, -38), MAKE_VAL_COMPLEX(95, 44), MAKE_VAL_COMPLEX(-137, 59), MAKE_VAL_COMPLEX(194, 42), MAKE_VAL_COMPLEX(-33, 7), MAKE_VAL_COMPLEX(5, -74), MAKE_VAL_COMPLEX(-20, 1), MAKE_VAL_COMPLEX(35, 53), MAKE_VAL_COMPLEX(-178, -44), MAKE_VAL_COMPLEX(-52, -27), MAKE_VAL_COMPLEX(-36, 86) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_261, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_261", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_261() */
/* END TEST 261 */

/*
 *  BEGIN TEST 262
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_262( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-32, 82), MAKE_VAL_COMPLEX(-94, -24), MAKE_VAL_COMPLEX(6, 48), MAKE_VAL_COMPLEX(-48, -46), MAKE_VAL_COMPLEX(-62, -28), MAKE_VAL_COMPLEX(10, -24), MAKE_VAL_COMPLEX(36, -76), MAKE_VAL_COMPLEX(-72, -30), MAKE_VAL_COMPLEX(-22, -20), MAKE_VAL_COMPLEX(-30, 72), MAKE_VAL_COMPLEX(26, 0), MAKE_VAL_COMPLEX(-12, -44), MAKE_VAL_COMPLEX(88, 28), MAKE_VAL_COMPLEX(58, -30), MAKE_VAL_COMPLEX(74, 20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_262, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_262", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_262() */
/* END TEST 262 */

/*
 *  BEGIN TEST 263
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_ROWMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_263( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_ROWMAJ, x240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-642, 362), MAKE_VAL_COMPLEX(1072, -190), MAKE_VAL_COMPLEX(-901, 69), MAKE_VAL_COMPLEX(711, 371), MAKE_VAL_COMPLEX(-1, 625), MAKE_VAL_COMPLEX(-359, -757), MAKE_VAL_COMPLEX(458, 1048), MAKE_VAL_COMPLEX(-292, -992), MAKE_VAL_COMPLEX(278, 452), MAKE_VAL_COMPLEX(33, -49), MAKE_VAL_COMPLEX(-85, 7), MAKE_VAL_COMPLEX(68, -126), MAKE_VAL_COMPLEX(70, 826), MAKE_VAL_COMPLEX(-505, 435), MAKE_VAL_COMPLEX(334, -690) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_263, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_263", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_263() */
/* END TEST 263 */

/*
 *  BEGIN TEST 264
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_264( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	oski_value_t* y_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(131, -67), MAKE_VAL_COMPLEX(-68, 71), MAKE_VAL_COMPLEX(119, -90), MAKE_VAL_COMPLEX(-61, -203), MAKE_VAL_COMPLEX(-7, 133), MAKE_VAL_COMPLEX(171, 193), MAKE_VAL_COMPLEX(-30, -24), MAKE_VAL_COMPLEX(-21, 51), MAKE_VAL_COMPLEX(-36, -46), MAKE_VAL_COMPLEX(-53, -26), MAKE_VAL_COMPLEX(45, -46), MAKE_VAL_COMPLEX(-1, -50) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_264, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_264", y_true, y_val, y240_LEN*y240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_264() */
/* END TEST 264 */

/*
 *  BEGIN TEST 265
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_265( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	oski_value_t* y_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-154, -9), MAKE_VAL_COMPLEX(146, -41), MAKE_VAL_COMPLEX(-64, 141), MAKE_VAL_COMPLEX(-42, 141), MAKE_VAL_COMPLEX(-16, -15), MAKE_VAL_COMPLEX(-161, -122), MAKE_VAL_COMPLEX(2, -39), MAKE_VAL_COMPLEX(-65, -2), MAKE_VAL_COMPLEX(58, -50), MAKE_VAL_COMPLEX(-1, -47), MAKE_VAL_COMPLEX(49, 15), MAKE_VAL_COMPLEX(35, -19) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_265, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_265", y_true, y_val, y240_LEN*y240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_265() */
/* END TEST 265 */

/*
 *  BEGIN TEST 266
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_266( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	oski_value_t* y_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(69, -32), MAKE_VAL_COMPLEX(-18, 78), MAKE_VAL_COMPLEX(-41, 59), MAKE_VAL_COMPLEX(45, -106), MAKE_VAL_COMPLEX(-115, -6), MAKE_VAL_COMPLEX(-66, 19), MAKE_VAL_COMPLEX(56, -35), MAKE_VAL_COMPLEX(-58, -75), MAKE_VAL_COMPLEX(94, 8), MAKE_VAL_COMPLEX(62, -61), MAKE_VAL_COMPLEX(42, 85), MAKE_VAL_COMPLEX(70, 23) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_266, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_266", y_true, y_val, y240_LEN*y240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_266() */
/* END TEST 266 */

/*
 *  BEGIN TEST 267
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_267( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	oski_value_t* y_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1828, 1183), MAKE_VAL_COMPLEX(-1785, -327), MAKE_VAL_COMPLEX(1554, -212), MAKE_VAL_COMPLEX(1946, -1741), MAKE_VAL_COMPLEX(-892, 289), MAKE_VAL_COMPLEX(-155, 2728), MAKE_VAL_COMPLEX(51, 32), MAKE_VAL_COMPLEX(77, 46), MAKE_VAL_COMPLEX(-234, -56), MAKE_VAL_COMPLEX(40, -75), MAKE_VAL_COMPLEX(64, 67), MAKE_VAL_COMPLEX(72, 1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(9, 9);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_267, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_267", y_true, y_val, y240_LEN*y240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_267() */
/* END TEST 267 */

/*
 *  BEGIN TEST 268
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_268( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-107, 37), MAKE_VAL_COMPLEX(106, -213), MAKE_VAL_COMPLEX(-152, -64), MAKE_VAL_COMPLEX(62, -43), MAKE_VAL_COMPLEX(-4, -188), MAKE_VAL_COMPLEX(-28, -53), MAKE_VAL_COMPLEX(-46, 16), MAKE_VAL_COMPLEX(53, -105), MAKE_VAL_COMPLEX(33, -34), MAKE_VAL_COMPLEX(-178, 40), MAKE_VAL_COMPLEX(-25, -44), MAKE_VAL_COMPLEX(-28, -2), MAKE_VAL_COMPLEX(23, 105), MAKE_VAL_COMPLEX(31, 81), MAKE_VAL_COMPLEX(-69, 145) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_268, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_268", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_268() */
/* END TEST 268 */

/*
 *  BEGIN TEST 269
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_269( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-21, -156), MAKE_VAL_COMPLEX(7, 57), MAKE_VAL_COMPLEX(59, 40), MAKE_VAL_COMPLEX(58, -12), MAKE_VAL_COMPLEX(101, 94), MAKE_VAL_COMPLEX(65, 89), MAKE_VAL_COMPLEX(160, 106), MAKE_VAL_COMPLEX(55, -6), MAKE_VAL_COMPLEX(20, 8), MAKE_VAL_COMPLEX(67, -148), MAKE_VAL_COMPLEX(6, 92), MAKE_VAL_COMPLEX(118, 12), MAKE_VAL_COMPLEX(-139, 37), MAKE_VAL_COMPLEX(-18, 48), MAKE_VAL_COMPLEX(-22, -23) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_269, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_269", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_269() */
/* END TEST 269 */

/*
 *  BEGIN TEST 270
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_270( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(66, -76), MAKE_VAL_COMPLEX(92, 62), MAKE_VAL_COMPLEX(12, -54), MAKE_VAL_COMPLEX(34, 68), MAKE_VAL_COMPLEX(56, 54), MAKE_VAL_COMPLEX(-20, 22), MAKE_VAL_COMPLEX(-68, 68), MAKE_VAL_COMPLEX(66, 60), MAKE_VAL_COMPLEX(16, 30), MAKE_VAL_COMPLEX(60, -66), MAKE_VAL_COMPLEX(-28, -10), MAKE_VAL_COMPLEX(-4, 52), MAKE_VAL_COMPLEX(-84, -64), MAKE_VAL_COMPLEX(-74, 10), MAKE_VAL_COMPLEX(-72, -50) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_270, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_270", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_270() */
/* END TEST 270 */

/*
 *  BEGIN TEST 271
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_271( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-441, -146), MAKE_VAL_COMPLEX(706, 463), MAKE_VAL_COMPLEX(408, -615), MAKE_VAL_COMPLEX(-25, 76), MAKE_VAL_COMPLEX(754, -213), MAKE_VAL_COMPLEX(386, -313), MAKE_VAL_COMPLEX(146, -734), MAKE_VAL_COMPLEX(195, 114), MAKE_VAL_COMPLEX(74, 69), MAKE_VAL_COMPLEX(-444, -633), MAKE_VAL_COMPLEX(424, -143), MAKE_VAL_COMPLEX(-50, -442), MAKE_VAL_COMPLEX(-96, 334), MAKE_VAL_COMPLEX(-121, 17), MAKE_VAL_COMPLEX(-420, -271) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_271, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_271", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_271() */
/* END TEST 271 */

/*
 *  BEGIN TEST 272
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_272( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-37, 31), MAKE_VAL_COMPLEX(-78, -15), MAKE_VAL_COMPLEX(-2, -48), MAKE_VAL_COMPLEX(38, -3), MAKE_VAL_COMPLEX(-60, -104), MAKE_VAL_COMPLEX(-56, -39), MAKE_VAL_COMPLEX(102, 160), MAKE_VAL_COMPLEX(-39, 97), MAKE_VAL_COMPLEX(31, 4), MAKE_VAL_COMPLEX(-70, 76), MAKE_VAL_COMPLEX(-47, -52), MAKE_VAL_COMPLEX(92, 58), MAKE_VAL_COMPLEX(-131, 101), MAKE_VAL_COMPLEX(-19, 47), MAKE_VAL_COMPLEX(-39, 111) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_272, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_272", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_272() */
/* END TEST 272 */

/*
 *  BEGIN TEST 273
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_273( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(92, -144), MAKE_VAL_COMPLEX(209, 60), MAKE_VAL_COMPLEX(-1, -21), MAKE_VAL_COMPLEX(25, 74), MAKE_VAL_COMPLEX(145, 151), MAKE_VAL_COMPLEX(39, 72), MAKE_VAL_COMPLEX(-162, -56), MAKE_VAL_COMPLEX(138, -46), MAKE_VAL_COMPLEX(-2, 27), MAKE_VAL_COMPLEX(121, -175), MAKE_VAL_COMPLEX(10, 49), MAKE_VAL_COMPLEX(-80, 6), MAKE_VAL_COMPLEX(9, -151), MAKE_VAL_COMPLEX(-67, -11), MAKE_VAL_COMPLEX(-64, -148) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_273, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_273", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_273() */
/* END TEST 273 */

/*
 *  BEGIN TEST 274
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_274( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(26, -48), MAKE_VAL_COMPLEX(56, 22), MAKE_VAL_COMPLEX(0, -30), MAKE_VAL_COMPLEX(26, 32), MAKE_VAL_COMPLEX(36, 22), MAKE_VAL_COMPLEX(-8, 14), MAKE_VAL_COMPLEX(-28, 44), MAKE_VAL_COMPLEX(42, 24), MAKE_VAL_COMPLEX(12, 14), MAKE_VAL_COMPLEX(24, -42), MAKE_VAL_COMPLEX(-16, -2), MAKE_VAL_COMPLEX(4, 28), MAKE_VAL_COMPLEX(-52, -24), MAKE_VAL_COMPLEX(-38, 14), MAKE_VAL_COMPLEX(-44, -18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_274, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_274", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_274() */
/* END TEST 274 */

/*
 *  BEGIN TEST 275
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_275( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_ROWMAJ, y240_VW );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(590, 564), MAKE_VAL_COMPLEX(-602, 1162), MAKE_VAL_COMPLEX(-102, -380), MAKE_VAL_COMPLEX(-16, 48), MAKE_VAL_COMPLEX(-1278, 288), MAKE_VAL_COMPLEX(-814, 186), MAKE_VAL_COMPLEX(1360, 28), MAKE_VAL_COMPLEX(458, 1392), MAKE_VAL_COMPLEX(124, -56), MAKE_VAL_COMPLEX(686, 1100), MAKE_VAL_COMPLEX(-736, -100), MAKE_VAL_COMPLEX(756, -248), MAKE_VAL_COMPLEX(70, 1058), MAKE_VAL_COMPLEX(-12, 112), MAKE_VAL_COMPLEX(672, 482) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_275, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_275", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_275() */
/* END TEST 275 */

/*
 *  BEGIN TEST 276
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_276( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	oski_value_t* y_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(192, -35), MAKE_VAL_COMPLEX(5, -174), MAKE_VAL_COMPLEX(-13, 20), MAKE_VAL_COMPLEX(23, -31), MAKE_VAL_COMPLEX(-159, 66), MAKE_VAL_COMPLEX(-52, 59), MAKE_VAL_COMPLEX(35, -10), MAKE_VAL_COMPLEX(-13, -29), MAKE_VAL_COMPLEX(102, -93), MAKE_VAL_COMPLEX(169, 140), MAKE_VAL_COMPLEX(-13, 47), MAKE_VAL_COMPLEX(21, 13) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_276, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_276", y_true, y_val, y240_LEN*y240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_276() */
/* END TEST 276 */

/*
 *  BEGIN TEST 277
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_277( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	oski_value_t* y_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-191, 27), MAKE_VAL_COMPLEX(1, 180), MAKE_VAL_COMPLEX(16, -13), MAKE_VAL_COMPLEX(-30, 21), MAKE_VAL_COMPLEX(152, -56), MAKE_VAL_COMPLEX(50, -52), MAKE_VAL_COMPLEX(-35, 3), MAKE_VAL_COMPLEX(3, 30), MAKE_VAL_COMPLEX(-96, 85), MAKE_VAL_COMPLEX(-171, -149), MAKE_VAL_COMPLEX(23, -46), MAKE_VAL_COMPLEX(-15, -18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_277, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_277", y_true, y_val, y240_LEN*y240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_277() */
/* END TEST 277 */

/*
 *  BEGIN TEST 278
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_278( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	oski_value_t* y_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(8, -64), MAKE_VAL_COMPLEX(48, 48), MAKE_VAL_COMPLEX(24, 56), MAKE_VAL_COMPLEX(-56, -80), MAKE_VAL_COMPLEX(-56, 80), MAKE_VAL_COMPLEX(-16, 56), MAKE_VAL_COMPLEX(0, -56), MAKE_VAL_COMPLEX(-80, 8), MAKE_VAL_COMPLEX(48, -64), MAKE_VAL_COMPLEX(-16, -72), MAKE_VAL_COMPLEX(80, 8), MAKE_VAL_COMPLEX(48, -40) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_278, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_278", y_true, y_val, y240_LEN*y240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_278() */
/* END TEST 278 */

/*
 *  BEGIN TEST 279
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_279( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	oski_value_t* y_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-369, 368), MAKE_VAL_COMPLEX(345, 724), MAKE_VAL_COMPLEX(-55, 78), MAKE_VAL_COMPLEX(24, -136), MAKE_VAL_COMPLEX(76, -451), MAKE_VAL_COMPLEX(-101, -192), MAKE_VAL_COMPLEX(20, -19), MAKE_VAL_COMPLEX(-88, -72), MAKE_VAL_COMPLEX(102, 437), MAKE_VAL_COMPLEX(-686, -265), MAKE_VAL_COMPLEX(100, 0), MAKE_VAL_COMPLEX(88, 8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-3, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_279, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_279", y_true, y_val, y240_LEN*y240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_279() */
/* END TEST 279 */

/*
 *  BEGIN TEST 280
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_280( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-11, 89), MAKE_VAL_COMPLEX(-46, 93), MAKE_VAL_COMPLEX(-57, 32), MAKE_VAL_COMPLEX(-19, -2), MAKE_VAL_COMPLEX(34, 83), MAKE_VAL_COMPLEX(78, -96), MAKE_VAL_COMPLEX(-84, -2), MAKE_VAL_COMPLEX(81, 109), MAKE_VAL_COMPLEX(3, 8), MAKE_VAL_COMPLEX(-35, 73), MAKE_VAL_COMPLEX(-66, 71), MAKE_VAL_COMPLEX(76, -58), MAKE_VAL_COMPLEX(-72, -105), MAKE_VAL_COMPLEX(-10, 0), MAKE_VAL_COMPLEX(15, -103) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_280, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_280", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_280() */
/* END TEST 280 */

/*
 *  BEGIN TEST 281
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_281( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(64, 8), MAKE_VAL_COMPLEX(-53, -22), MAKE_VAL_COMPLEX(108, 1), MAKE_VAL_COMPLEX(-64, 11), MAKE_VAL_COMPLEX(-111, -46), MAKE_VAL_COMPLEX(-93, 67), MAKE_VAL_COMPLEX(40, -94), MAKE_VAL_COMPLEX(-168, -64), MAKE_VAL_COMPLEX(-40, -3), MAKE_VAL_COMPLEX(80, 14), MAKE_VAL_COMPLEX(87, -96), MAKE_VAL_COMPLEX(-128, 34), MAKE_VAL_COMPLEX(170, 43), MAKE_VAL_COMPLEX(28, -80), MAKE_VAL_COMPLEX(64, 48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_281, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_281", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_281() */
/* END TEST 281 */

/*
 *  BEGIN TEST 282
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_282( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(42, -89), MAKE_VAL_COMPLEX(103, 34), MAKE_VAL_COMPLEX(-3, -54), MAKE_VAL_COMPLEX(50, 55), MAKE_VAL_COMPLEX(67, 36), MAKE_VAL_COMPLEX(-13, 26), MAKE_VAL_COMPLEX(-46, 82), MAKE_VAL_COMPLEX(78, 39), MAKE_VAL_COMPLEX(23, 24), MAKE_VAL_COMPLEX(39, -78), MAKE_VAL_COMPLEX(-29, -2), MAKE_VAL_COMPLEX(10, 50), MAKE_VAL_COMPLEX(-96, -38), MAKE_VAL_COMPLEX(-67, 29), MAKE_VAL_COMPLEX(-81, -28) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_282, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_282", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_282() */
/* END TEST 282 */

/*
 *  BEGIN TEST 283
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_283( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(540, -205), MAKE_VAL_COMPLEX(737, -179), MAKE_VAL_COMPLEX(410, 340), MAKE_VAL_COMPLEX(47, 31), MAKE_VAL_COMPLEX(393, -671), MAKE_VAL_COMPLEX(-946, -183), MAKE_VAL_COMPLEX(448, 534), MAKE_VAL_COMPLEX(376, -1117), MAKE_VAL_COMPLEX(36, -102), MAKE_VAL_COMPLEX(537, 16), MAKE_VAL_COMPLEX(795, 205), MAKE_VAL_COMPLEX(-680, -370), MAKE_VAL_COMPLEX(-361, 1052), MAKE_VAL_COMPLEX(147, 91), MAKE_VAL_COMPLEX(-705, 410) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_283, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_283", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_283() */
/* END TEST 283 */

/*
 *  BEGIN TEST 284
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_284( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-55, 133), MAKE_VAL_COMPLEX(20, -87), MAKE_VAL_COMPLEX(111, 70), MAKE_VAL_COMPLEX(-17, -62), MAKE_VAL_COMPLEX(40, -1), MAKE_VAL_COMPLEX(76, -122), MAKE_VAL_COMPLEX(116, 2), MAKE_VAL_COMPLEX(-155, -81), MAKE_VAL_COMPLEX(-17, -16), MAKE_VAL_COMPLEX(-47, 121), MAKE_VAL_COMPLEX(-32, 95), MAKE_VAL_COMPLEX(-92, -54), MAKE_VAL_COMPLEX(120, 59), MAKE_VAL_COMPLEX(46, -14), MAKE_VAL_COMPLEX(3, -43) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_284, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_284", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_284() */
/* END TEST 284 */

/*
 *  BEGIN TEST 285
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_285( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(7, -10), MAKE_VAL_COMPLEX(-161, 51), MAKE_VAL_COMPLEX(-102, 2), MAKE_VAL_COMPLEX(-55, -7), MAKE_VAL_COMPLEX(-133, -41), MAKE_VAL_COMPLEX(-61, 86), MAKE_VAL_COMPLEX(-62, -116), MAKE_VAL_COMPLEX(47, 36), MAKE_VAL_COMPLEX(-16, -14), MAKE_VAL_COMPLEX(2, -13), MAKE_VAL_COMPLEX(71, -95), MAKE_VAL_COMPLEX(74, -12), MAKE_VAL_COMPLEX(12, -17), MAKE_VAL_COMPLEX(41, -31), MAKE_VAL_COMPLEX(108, 73) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_285, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_285", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_285() */
/* END TEST 285 */

/*
 *  BEGIN TEST 286
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_286( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(37, -11), MAKE_VAL_COMPLEX(17, 39), MAKE_VAL_COMPLEX(15, -15), MAKE_VAL_COMPLEX(-3, 29), MAKE_VAL_COMPLEX(7, 29), MAKE_VAL_COMPLEX(-11, 3), MAKE_VAL_COMPLEX(-36, 8), MAKE_VAL_COMPLEX(9, 33), MAKE_VAL_COMPLEX(-1, 13), MAKE_VAL_COMPLEX(33, -9), MAKE_VAL_COMPLEX(-7, -9), MAKE_VAL_COMPLEX(-12, 16), MAKE_VAL_COMPLEX(-14, -38), MAKE_VAL_COMPLEX(-26, -12), MAKE_VAL_COMPLEX(-13, -31) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_286, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_286", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_286() */
/* END TEST 286 */

/*
 *  BEGIN TEST 287
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x240', LAYOUT_COLMAJ>
 *  y : <'y240', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_287( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A240_ptr, A240_ind, A240_val,
		A240_M, A240_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y240_val,
		y240_LEN * y240_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y240_LEN, y240_VW, LAYOUT_COLMAJ, y240_LEN );
	oski_value_t* y_val = array_Duplicate( x240_val,
		x240_LEN * x240_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x240_LEN, x240_VW, LAYOUT_COLMAJ, x240_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-445, 187), MAKE_VAL_COMPLEX(132, 358), MAKE_VAL_COMPLEX(-20, 712), MAKE_VAL_COMPLEX(-16, -50), MAKE_VAL_COMPLEX(-220, 398), MAKE_VAL_COMPLEX(607, 159), MAKE_VAL_COMPLEX(-134, 250), MAKE_VAL_COMPLEX(-47, -619), MAKE_VAL_COMPLEX(-74, -54), MAKE_VAL_COMPLEX(-415, 183), MAKE_VAL_COMPLEX(-446, -176), MAKE_VAL_COMPLEX(-38, -566), MAKE_VAL_COMPLEX(151, 349), MAKE_VAL_COMPLEX(124, -102), MAKE_VAL_COMPLEX(477, -297) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_287, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_287", y_true, y_val, x240_LEN*x240_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_287() */
/* END TEST 287 */

#define A288_M 4    /* # of rows */
#define A288_N 5    /* # of cols */
#define A288_NNZ 4  /* # of stored non-zeros */
static oski_index_t A288_ptr[A288_M+1] = {1, 3, 4, 4, 5}; /* end row pointers for A288 */
static oski_index_t A288_ind[A288_NNZ] = {2, 4, 3, 5}; /* end column indices pointers for A288 */
static oski_value_t A288_val[A288_NNZ] = {MAKE_VAL_COMPLEX(-6, -2), MAKE_VAL_COMPLEX(0, -8), MAKE_VAL_COMPLEX(-5, 10), MAKE_VAL_COMPLEX(9, -9)}; /* end stored non-zero values for A288 */

#define x288_LEN 5   /* vector length */
#define x288_VW  3   /* vector width */
static oski_value_t x288_val[x288_LEN*x288_VW] = {MAKE_VAL_COMPLEX(-8, -4), MAKE_VAL_COMPLEX(-6, 8), MAKE_VAL_COMPLEX(10, 4), MAKE_VAL_COMPLEX(1, 4), MAKE_VAL_COMPLEX(6, -9), MAKE_VAL_COMPLEX(10, -6), MAKE_VAL_COMPLEX(-6, 0), MAKE_VAL_COMPLEX(10, 0), MAKE_VAL_COMPLEX(-9, 7), MAKE_VAL_COMPLEX(-7, -10), MAKE_VAL_COMPLEX(6, 2), MAKE_VAL_COMPLEX(8, 0), MAKE_VAL_COMPLEX(8, -3), MAKE_VAL_COMPLEX(8, 10), MAKE_VAL_COMPLEX(-10, -10)}; /* data */

#define y288_LEN 4   /* vector length */
#define y288_VW  3   /* vector width */
static oski_value_t y288_val[y288_LEN*y288_VW] = {MAKE_VAL_COMPLEX(4, 3), MAKE_VAL_COMPLEX(1, -10), MAKE_VAL_COMPLEX(7, 6), MAKE_VAL_COMPLEX(-3, -10), MAKE_VAL_COMPLEX(-7, -5), MAKE_VAL_COMPLEX(-2, 10), MAKE_VAL_COMPLEX(-2, 0), MAKE_VAL_COMPLEX(-3, -4), MAKE_VAL_COMPLEX(7, -10), MAKE_VAL_COMPLEX(-5, 9), MAKE_VAL_COMPLEX(-10, -4), MAKE_VAL_COMPLEX(10, 5)}; /* data */

/*
 *  BEGIN TEST 288
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_288( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	oski_value_t* y_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-113, 12), MAKE_VAL_COMPLEX(-40, 63), MAKE_VAL_COMPLEX(-110, -73), MAKE_VAL_COMPLEX(59, 1), MAKE_VAL_COMPLEX(3, 114), MAKE_VAL_COMPLEX(-13, -193), MAKE_VAL_COMPLEX(6, -2), MAKE_VAL_COMPLEX(32, 21), MAKE_VAL_COMPLEX(-41, 74), MAKE_VAL_COMPLEX(59, -168), MAKE_VAL_COMPLEX(232, 34), MAKE_VAL_COMPLEX(-237, -20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_288, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_288", y_true, y_val, y288_LEN*y288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_288() */
/* END TEST 288 */

/*
 *  BEGIN TEST 289
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_289( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	oski_value_t* y_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(101, 29), MAKE_VAL_COMPLEX(123, -85), MAKE_VAL_COMPLEX(83, 147), MAKE_VAL_COMPLEX(12, -55), MAKE_VAL_COMPLEX(16, -185), MAKE_VAL_COMPLEX(-73, 207), MAKE_VAL_COMPLEX(-12, -14), MAKE_VAL_COMPLEX(-9, -57), MAKE_VAL_COMPLEX(142, -48), MAKE_VAL_COMPLEX(-146, 155), MAKE_VAL_COMPLEX(-230, -126), MAKE_VAL_COMPLEX(227, 115) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_289, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_289", y_true, y_val, y288_LEN*y288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_289() */
/* END TEST 289 */

/*
 *  BEGIN TEST 290
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_290( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	oski_value_t* y_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(50, 0), MAKE_VAL_COMPLEX(-52, -86), MAKE_VAL_COMPLEX(92, 6), MAKE_VAL_COMPLEX(-84, -62), MAKE_VAL_COMPLEX(-86, 2), MAKE_VAL_COMPLEX(44, 92), MAKE_VAL_COMPLEX(-16, 12), MAKE_VAL_COMPLEX(-48, -14), MAKE_VAL_COMPLEX(-4, -122), MAKE_VAL_COMPLEX(14, 102), MAKE_VAL_COMPLEX(-104, 28), MAKE_VAL_COMPLEX(110, -20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_290, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_290", y_true, y_val, y288_LEN*y288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_290() */
/* END TEST 290 */

/*
 *  BEGIN TEST 291
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_291( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	oski_value_t* y_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(341, -784), MAKE_VAL_COMPLEX(198, -305), MAKE_VAL_COMPLEX(696, -217), MAKE_VAL_COMPLEX(162, 588), MAKE_VAL_COMPLEX(-384, -809), MAKE_VAL_COMPLEX(1061, 671), MAKE_VAL_COMPLEX(40, -32), MAKE_VAL_COMPLEX(-74, 93), MAKE_VAL_COMPLEX(-59, -120), MAKE_VAL_COMPLEX(590, 927), MAKE_VAL_COMPLEX(-1148, 1114), MAKE_VAL_COMPLEX(1037, -1264) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_291, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_291", y_true, y_val, y288_LEN*y288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_291() */
/* END TEST 291 */

/*
 *  BEGIN TEST 292
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_292( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-12, -85), MAKE_VAL_COMPLEX(-99, -10), MAKE_VAL_COMPLEX(35, 110), MAKE_VAL_COMPLEX(-47, -4), MAKE_VAL_COMPLEX(75, 47), MAKE_VAL_COMPLEX(76, 4), MAKE_VAL_COMPLEX(77, -28), MAKE_VAL_COMPLEX(142, 31), MAKE_VAL_COMPLEX(-193, -110), MAKE_VAL_COMPLEX(57, -139), MAKE_VAL_COMPLEX(-70, 48), MAKE_VAL_COMPLEX(106, 13), MAKE_VAL_COMPLEX(108, 172), MAKE_VAL_COMPLEX(-158, 178), MAKE_VAL_COMPLEX(155, -185) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_292, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_292", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_292() */
/* END TEST 292 */

/*
 *  BEGIN TEST 293
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_293( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(56, 17), MAKE_VAL_COMPLEX(33, -52), MAKE_VAL_COMPLEX(-81, -24), MAKE_VAL_COMPLEX(10, 9), MAKE_VAL_COMPLEX(0, 16), MAKE_VAL_COMPLEX(-32, 92), MAKE_VAL_COMPLEX(-71, -26), MAKE_VAL_COMPLEX(-152, 59), MAKE_VAL_COMPLEX(139, 22), MAKE_VAL_COMPLEX(40, 86), MAKE_VAL_COMPLEX(46, 4), MAKE_VAL_COMPLEX(-114, 59), MAKE_VAL_COMPLEX(-89, -97), MAKE_VAL_COMPLEX(60, -116), MAKE_VAL_COMPLEX(-55, 105) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_293, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_293", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_293() */
/* END TEST 293 */

/*
 *  BEGIN TEST 294
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_294( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(16, 48), MAKE_VAL_COMPLEX(56, -8), MAKE_VAL_COMPLEX(-24, -56), MAKE_VAL_COMPLEX(12, -20), MAKE_VAL_COMPLEX(-60, 12), MAKE_VAL_COMPLEX(-64, -16), MAKE_VAL_COMPLEX(24, 24), MAKE_VAL_COMPLEX(-40, -40), MAKE_VAL_COMPLEX(64, 8), MAKE_VAL_COMPLEX(-12, 68), MAKE_VAL_COMPLEX(-16, -32), MAKE_VAL_COMPLEX(-32, -32), MAKE_VAL_COMPLEX(-44, -20), MAKE_VAL_COMPLEX(8, -72), MAKE_VAL_COMPLEX(0, 80) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_294, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_294", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_294() */
/* END TEST 294 */

/*
 *  BEGIN TEST 295
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_295( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(29, 89), MAKE_VAL_COMPLEX(113, -73), MAKE_VAL_COMPLEX(-5, 21), MAKE_VAL_COMPLEX(45, -377), MAKE_VAL_COMPLEX(-556, 218), MAKE_VAL_COMPLEX(-88, -456), MAKE_VAL_COMPLEX(715, 729), MAKE_VAL_COMPLEX(779, 27), MAKE_VAL_COMPLEX(-117, -967), MAKE_VAL_COMPLEX(236, 2), MAKE_VAL_COMPLEX(-586, -566), MAKE_VAL_COMPLEX(629, -99), MAKE_VAL_COMPLEX(-422, 1042), MAKE_VAL_COMPLEX(-1144, -324), MAKE_VAL_COMPLEX(1170, 440) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_295, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_295", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_295() */
/* END TEST 295 */

/*
 *  BEGIN TEST 296
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_296( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(48, 75), MAKE_VAL_COMPLEX(89, -44), MAKE_VAL_COMPLEX(-53, -76), MAKE_VAL_COMPLEX(-21, -57), MAKE_VAL_COMPLEX(-86, 99), MAKE_VAL_COMPLEX(-166, -14), MAKE_VAL_COMPLEX(-39, 110), MAKE_VAL_COMPLEX(-98, 41), MAKE_VAL_COMPLEX(224, -51), MAKE_VAL_COMPLEX(-23, 156), MAKE_VAL_COMPLEX(32, -42), MAKE_VAL_COMPLEX(-102, 21), MAKE_VAL_COMPLEX(-205, 20), MAKE_VAL_COMPLEX(-68, -246), MAKE_VAL_COMPLEX(75, 265) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_296, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_296", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_296() */
/* END TEST 296 */

/*
 *  BEGIN TEST 297
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_297( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(44, -59), MAKE_VAL_COMPLEX(-53, -54), MAKE_VAL_COMPLEX(-59, 66), MAKE_VAL_COMPLEX(-1, 20), MAKE_VAL_COMPLEX(53, 9), MAKE_VAL_COMPLEX(84, 104), MAKE_VAL_COMPLEX(99, -128), MAKE_VAL_COMPLEX(-2, -11), MAKE_VAL_COMPLEX(-155, -46), MAKE_VAL_COMPLEX(123, -77), MAKE_VAL_COMPLEX(-98, 40), MAKE_VAL_COMPLEX(22, 3), MAKE_VAL_COMPLEX(134, 34), MAKE_VAL_COMPLEX(-42, 170), MAKE_VAL_COMPLEX(55, -195) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_297, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_297", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_297() */
/* END TEST 297 */

/*
 *  BEGIN TEST 298
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_298( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-72, 44), MAKE_VAL_COMPLEX(34, 88), MAKE_VAL_COMPLEX(82, -60), MAKE_VAL_COMPLEX(37, 12), MAKE_VAL_COMPLEX(-42, -93), MAKE_VAL_COMPLEX(2, -110), MAKE_VAL_COMPLEX(-30, 48), MAKE_VAL_COMPLEX(50, -80), MAKE_VAL_COMPLEX(11, 107), MAKE_VAL_COMPLEX(-115, 6), MAKE_VAL_COMPLEX(46, -38), MAKE_VAL_COMPLEX(40, -64), MAKE_VAL_COMPLEX(16, -79), MAKE_VAL_COMPLEX(120, -14), MAKE_VAL_COMPLEX(-130, 30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_298, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_298", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_298() */
/* END TEST 298 */

/*
 *  BEGIN TEST 299
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_299( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(9, -15), MAKE_VAL_COMPLEX(7, 15), MAKE_VAL_COMPLEX(-79, -15), MAKE_VAL_COMPLEX(174, 57), MAKE_VAL_COMPLEX(-92, -239), MAKE_VAL_COMPLEX(274, 42), MAKE_VAL_COMPLEX(373, -499), MAKE_VAL_COMPLEX(-31, -453), MAKE_VAL_COMPLEX(-532, 278), MAKE_VAL_COMPLEX(145, -218), MAKE_VAL_COMPLEX(-376, 56), MAKE_VAL_COMPLEX(105, -327), MAKE_VAL_COMPLEX(576, -281), MAKE_VAL_COMPLEX(352, 562), MAKE_VAL_COMPLEX(-310, -620) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_299, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_299", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_299() */
/* END TEST 299 */

/*
 *  BEGIN TEST 300
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_300( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	oski_value_t* y_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-71, 31), MAKE_VAL_COMPLEX(24, -87), MAKE_VAL_COMPLEX(21, 11), MAKE_VAL_COMPLEX(19, -136), MAKE_VAL_COMPLEX(-70, -6), MAKE_VAL_COMPLEX(-40, 123), MAKE_VAL_COMPLEX(4, 2), MAKE_VAL_COMPLEX(155, 11), MAKE_VAL_COMPLEX(-51, -81), MAKE_VAL_COMPLEX(-21, -99), MAKE_VAL_COMPLEX(-43, 5), MAKE_VAL_COMPLEX(-137, 5) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_300, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_300", y_true, y_val, y288_LEN*y288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_300() */
/* END TEST 300 */

/*
 *  BEGIN TEST 301
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_301( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	oski_value_t* y_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(83, -22), MAKE_VAL_COMPLEX(-21, 57), MAKE_VAL_COMPLEX(0, 7), MAKE_VAL_COMPLEX(-28, 106), MAKE_VAL_COMPLEX(49, -9), MAKE_VAL_COMPLEX(34, -93), MAKE_VAL_COMPLEX(-10, -2), MAKE_VAL_COMPLEX(-164, -23), MAKE_VAL_COMPLEX(72, 51), MAKE_VAL_COMPLEX(6, 126), MAKE_VAL_COMPLEX(13, -17), MAKE_VAL_COMPLEX(167, 10) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_301, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_301", y_true, y_val, y288_LEN*y288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_301() */
/* END TEST 301 */

/*
 *  BEGIN TEST 302
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_302( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	oski_value_t* y_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-35, 5), MAKE_VAL_COMPLEX(45, 55), MAKE_VAL_COMPLEX(-65, 5), MAKE_VAL_COMPLEX(65, 35), MAKE_VAL_COMPLEX(60, -10), MAKE_VAL_COMPLEX(-40, -60), MAKE_VAL_COMPLEX(10, -10), MAKE_VAL_COMPLEX(35, 5), MAKE_VAL_COMPLEX(15, 85), MAKE_VAL_COMPLEX(-20, -70), MAKE_VAL_COMPLEX(70, -30), MAKE_VAL_COMPLEX(-75, 25) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_302, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_302", y_true, y_val, y288_LEN*y288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_302() */
/* END TEST 302 */

/*
 *  BEGIN TEST 303
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_303( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	oski_value_t* y_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-121, 163), MAKE_VAL_COMPLEX(23, -212), MAKE_VAL_COMPLEX(31, 55), MAKE_VAL_COMPLEX(-44, -329), MAKE_VAL_COMPLEX(-130, 6), MAKE_VAL_COMPLEX(-21, 294), MAKE_VAL_COMPLEX(6, -12), MAKE_VAL_COMPLEX(339, -159), MAKE_VAL_COMPLEX(-109, -89), MAKE_VAL_COMPLEX(-205, -189), MAKE_VAL_COMPLEX(-77, -15), MAKE_VAL_COMPLEX(-279, 217) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_303, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_303", y_true, y_val, y288_LEN*y288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_303() */
/* END TEST 303 */

/*
 *  BEGIN TEST 304
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_304( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(20, 51), MAKE_VAL_COMPLEX(49, -13), MAKE_VAL_COMPLEX(-17, -66), MAKE_VAL_COMPLEX(-5, -56), MAKE_VAL_COMPLEX(-30, 66), MAKE_VAL_COMPLEX(-131, 39), MAKE_VAL_COMPLEX(126, 90), MAKE_VAL_COMPLEX(-132, -110), MAKE_VAL_COMPLEX(-11, -91), MAKE_VAL_COMPLEX(9, 26), MAKE_VAL_COMPLEX(-59, 20), MAKE_VAL_COMPLEX(-102, -83), MAKE_VAL_COMPLEX(-161, -83), MAKE_VAL_COMPLEX(-55, -81), MAKE_VAL_COMPLEX(135, 35) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_304, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_304", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_304() */
/* END TEST 304 */

/*
 *  BEGIN TEST 305
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_305( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-76, 41), MAKE_VAL_COMPLEX(41, 93), MAKE_VAL_COMPLEX(75, -50), MAKE_VAL_COMPLEX(54, 48), MAKE_VAL_COMPLEX(-72, -147), MAKE_VAL_COMPLEX(69, -165), MAKE_VAL_COMPLEX(-132, -18), MAKE_VAL_COMPLEX(142, -10), MAKE_VAL_COMPLEX(86, 206), MAKE_VAL_COMPLEX(-136, 48), MAKE_VAL_COMPLEX(89, -90), MAKE_VAL_COMPLEX(110, -13), MAKE_VAL_COMPLEX(133, -16), MAKE_VAL_COMPLEX(183, -5), MAKE_VAL_COMPLEX(-265, 75) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_305, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_305", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_305() */
/* END TEST 305 */

/*
 *  BEGIN TEST 306
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_306( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-36, -28), MAKE_VAL_COMPLEX(-38, 34), MAKE_VAL_COMPLEX(46, 30), MAKE_VAL_COMPLEX(1, 21), MAKE_VAL_COMPLEX(39, -39), MAKE_VAL_COMPLEX(56, -20), MAKE_VAL_COMPLEX(-30, -6), MAKE_VAL_COMPLEX(50, 10), MAKE_VAL_COMPLEX(-52, 26), MAKE_VAL_COMPLEX(-25, -57), MAKE_VAL_COMPLEX(28, 16), MAKE_VAL_COMPLEX(40, 8), MAKE_VAL_COMPLEX(43, -7), MAKE_VAL_COMPLEX(30, 58), MAKE_VAL_COMPLEX(-40, -60) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_306, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_306", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_306() */
/* END TEST 306 */

/*
 *  BEGIN TEST 307
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_307( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-78, -22), MAKE_VAL_COMPLEX(34, 22), MAKE_VAL_COMPLEX(64, 132), MAKE_VAL_COMPLEX(319, 135), MAKE_VAL_COMPLEX(-465, -183), MAKE_VAL_COMPLEX(128, -752), MAKE_VAL_COMPLEX(-1038, 210), MAKE_VAL_COMPLEX(1022, -122), MAKE_VAL_COMPLEX(992, 170), MAKE_VAL_COMPLEX(101, 321), MAKE_VAL_COMPLEX(-26, -554), MAKE_VAL_COMPLEX(766, -106), MAKE_VAL_COMPLEX(1123, -331), MAKE_VAL_COMPLEX(462, -266), MAKE_VAL_COMPLEX(-580, 1020) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_307, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_307", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_307() */
/* END TEST 307 */

/*
 *  BEGIN TEST 308
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_308( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-52, 15), MAKE_VAL_COMPLEX(-5, 59), MAKE_VAL_COMPLEX(73, -30), MAKE_VAL_COMPLEX(-8, -4), MAKE_VAL_COMPLEX(44, -43), MAKE_VAL_COMPLEX(-1, 13), MAKE_VAL_COMPLEX(-128, 70), MAKE_VAL_COMPLEX(158, -70), MAKE_VAL_COMPLEX(88, 72), MAKE_VAL_COMPLEX(-102, 0), MAKE_VAL_COMPLEX(75, -74), MAKE_VAL_COMPLEX(130, 29), MAKE_VAL_COMPLEX(91, -164), MAKE_VAL_COMPLEX(89, -45), MAKE_VAL_COMPLEX(-45, 125) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_308, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_308", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_308() */
/* END TEST 308 */

/*
 *  BEGIN TEST 309
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_309( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-32, -87), MAKE_VAL_COMPLEX(-91, 19), MAKE_VAL_COMPLEX(35, 108), MAKE_VAL_COMPLEX(8, 55), MAKE_VAL_COMPLEX(55, -47), MAKE_VAL_COMPLEX(139, -55), MAKE_VAL_COMPLEX(56, -88), MAKE_VAL_COMPLEX(-38, 100), MAKE_VAL_COMPLEX(-217, -15), MAKE_VAL_COMPLEX(48, -141), MAKE_VAL_COMPLEX(-9, 116), MAKE_VAL_COMPLEX(-34, -5), MAKE_VAL_COMPLEX(14, 152), MAKE_VAL_COMPLEX(-23, 189), MAKE_VAL_COMPLEX(-45, -275) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_309, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_309", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_309() */
/* END TEST 309 */

/*
 *  BEGIN TEST 310
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_310( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(56, -32), MAKE_VAL_COMPLEX(-24, -68), MAKE_VAL_COMPLEX(-64, 44), MAKE_VAL_COMPLEX(-28, -10), MAKE_VAL_COMPLEX(30, 72), MAKE_VAL_COMPLEX(-4, 84), MAKE_VAL_COMPLEX(24, -36), MAKE_VAL_COMPLEX(-40, 60), MAKE_VAL_COMPLEX(-6, -82), MAKE_VAL_COMPLEX(88, -2), MAKE_VAL_COMPLEX(-36, 28), MAKE_VAL_COMPLEX(-32, 48), MAKE_VAL_COMPLEX(-14, 60), MAKE_VAL_COMPLEX(-92, 8), MAKE_VAL_COMPLEX(100, -20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_310, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_310", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_310() */
/* END TEST 310 */

/*
 *  BEGIN TEST 311
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_ROWMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_311( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_ROWMAJ, x288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-38, -79), MAKE_VAL_COMPLEX(-81, 5), MAKE_VAL_COMPLEX(55, 122), MAKE_VAL_COMPLEX(48, -3), MAKE_VAL_COMPLEX(3, 53), MAKE_VAL_COMPLEX(-27, -109), MAKE_VAL_COMPLEX(-36, -284), MAKE_VAL_COMPLEX(22, 316), MAKE_VAL_COMPLEX(-219, 195), MAKE_VAL_COMPLEX(4, -195), MAKE_VAL_COMPLEX(111, 190), MAKE_VAL_COMPLEX(-156, 175), MAKE_VAL_COMPLEX(248, 278), MAKE_VAL_COMPLEX(103, 207), MAKE_VAL_COMPLEX(-315, -185) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_311, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_311", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_311() */
/* END TEST 311 */

/*
 *  BEGIN TEST 312
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_312( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	oski_value_t* y_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(104, -27), MAKE_VAL_COMPLEX(109, 8), MAKE_VAL_COMPLEX(87, -36), MAKE_VAL_COMPLEX(-117, 18), MAKE_VAL_COMPLEX(-105, 65), MAKE_VAL_COMPLEX(-16, 165), MAKE_VAL_COMPLEX(-4, 4), MAKE_VAL_COMPLEX(-11, -28), MAKE_VAL_COMPLEX(57, -73), MAKE_VAL_COMPLEX(-61, -68), MAKE_VAL_COMPLEX(-232, -48), MAKE_VAL_COMPLEX(-102, 45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_312, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_312", y_true, y_val, y288_LEN*y288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_312() */
/* END TEST 312 */

/*
 *  BEGIN TEST 313
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_313( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	oski_value_t* y_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-36, 53), MAKE_VAL_COMPLEX(-135, -152), MAKE_VAL_COMPLEX(35, 92), MAKE_VAL_COMPLEX(35, -146), MAKE_VAL_COMPLEX(-13, -107), MAKE_VAL_COMPLEX(28, -17), MAKE_VAL_COMPLEX(-24, 4), MAKE_VAL_COMPLEX(-47, -16), MAKE_VAL_COMPLEX(1, -95), MAKE_VAL_COMPLEX(27, 214), MAKE_VAL_COMPLEX(76, 32), MAKE_VAL_COMPLEX(262, -15) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_313, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_313", y_true, y_val, y288_LEN*y288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_313() */
/* END TEST 313 */

/*
 *  BEGIN TEST 314
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_314( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	oski_value_t* y_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(5, 35), MAKE_VAL_COMPLEX(55, -45), MAKE_VAL_COMPLEX(5, 65), MAKE_VAL_COMPLEX(35, -65), MAKE_VAL_COMPLEX(-10, -60), MAKE_VAL_COMPLEX(-60, 40), MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(5, -35), MAKE_VAL_COMPLEX(85, -15), MAKE_VAL_COMPLEX(-70, 20), MAKE_VAL_COMPLEX(-30, -70), MAKE_VAL_COMPLEX(25, 75) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_314, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_314", y_true, y_val, y288_LEN*y288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_314() */
/* END TEST 314 */

/*
 *  BEGIN TEST 315
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_315( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	oski_value_t* y_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(133, -684), MAKE_VAL_COMPLEX(956, -305), MAKE_VAL_COMPLEX(-204, -591), MAKE_VAL_COMPLEX(-6, 945), MAKE_VAL_COMPLEX(249, 817), MAKE_VAL_COMPLEX(574, 573), MAKE_VAL_COMPLEX(58, -34), MAKE_VAL_COMPLEX(14, -83), MAKE_VAL_COMPLEX(48, -164), MAKE_VAL_COMPLEX(-929, -412), MAKE_VAL_COMPLEX(-1022, 870), MAKE_VAL_COMPLEX(-705, 1092) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_315, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_315", y_true, y_val, y288_LEN*y288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_315() */
/* END TEST 315 */

/*
 *  BEGIN TEST 316
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_316( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(52, -13), MAKE_VAL_COMPLEX(-29, -92), MAKE_VAL_COMPLEX(57, 44), MAKE_VAL_COMPLEX(-1, -35), MAKE_VAL_COMPLEX(48, 186), MAKE_VAL_COMPLEX(-15, 54), MAKE_VAL_COMPLEX(-9, 29), MAKE_VAL_COMPLEX(42, -9), MAKE_VAL_COMPLEX(-82, -76), MAKE_VAL_COMPLEX(-58, 66), MAKE_VAL_COMPLEX(-25, 22), MAKE_VAL_COMPLEX(-64, -8), MAKE_VAL_COMPLEX(-103, -36), MAKE_VAL_COMPLEX(-14, -59), MAKE_VAL_COMPLEX(215, -45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_316, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_316", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_316() */
/* END TEST 316 */

/*
 *  BEGIN TEST 317
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_317( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-8, -95), MAKE_VAL_COMPLEX(-81, 22), MAKE_VAL_COMPLEX(-99, 90), MAKE_VAL_COMPLEX(-50, 52), MAKE_VAL_COMPLEX(75, -117), MAKE_VAL_COMPLEX(103, 70), MAKE_VAL_COMPLEX(3, -107), MAKE_VAL_COMPLEX(-32, 139), MAKE_VAL_COMPLEX(-18, -34), MAKE_VAL_COMPLEX(181, -167), MAKE_VAL_COMPLEX(5, 58), MAKE_VAL_COMPLEX(72, 112), MAKE_VAL_COMPLEX(150, 137), MAKE_VAL_COMPLEX(-108, 173), MAKE_VAL_COMPLEX(-95, -95) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_317, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_317", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_317() */
/* END TEST 317 */

/*
 *  BEGIN TEST 318
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_318( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-40, -40), MAKE_VAL_COMPLEX(-52, 36), MAKE_VAL_COMPLEX(52, 44), MAKE_VAL_COMPLEX(-2, 26), MAKE_VAL_COMPLEX(54, -42), MAKE_VAL_COMPLEX(72, -16), MAKE_VAL_COMPLEX(-36, -12), MAKE_VAL_COMPLEX(60, 20), MAKE_VAL_COMPLEX(-68, 24), MAKE_VAL_COMPLEX(-22, -74), MAKE_VAL_COMPLEX(32, 24), MAKE_VAL_COMPLEX(48, 16), MAKE_VAL_COMPLEX(54, -2), MAKE_VAL_COMPLEX(28, 76), MAKE_VAL_COMPLEX(-40, -80) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_318, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_318", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_318() */
/* END TEST 318 */

/*
 *  BEGIN TEST 319
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_319( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(17, -9), MAKE_VAL_COMPLEX(-257, 123), MAKE_VAL_COMPLEX(261, -791), MAKE_VAL_COMPLEX(-134, -158), MAKE_VAL_COMPLEX(900, -138), MAKE_VAL_COMPLEX(-81, -37), MAKE_VAL_COMPLEX(338, 296), MAKE_VAL_COMPLEX(-261, -643), MAKE_VAL_COMPLEX(-160, 636), MAKE_VAL_COMPLEX(232, 950), MAKE_VAL_COMPLEX(53, -55), MAKE_VAL_COMPLEX(-312, 168), MAKE_VAL_COMPLEX(-649, 485), MAKE_VAL_COMPLEX(-279, -473), MAKE_VAL_COMPLEX(-200, -970) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_319, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_319", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_319() */
/* END TEST 319 */

/*
 *  BEGIN TEST 320
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_320( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 51), MAKE_VAL_COMPLEX(19, -6), MAKE_VAL_COMPLEX(-87, 22), MAKE_VAL_COMPLEX(-11, 28), MAKE_VAL_COMPLEX(-183, 24), MAKE_VAL_COMPLEX(-49, -48), MAKE_VAL_COMPLEX(19, 87), MAKE_VAL_COMPLEX(-38, 41), MAKE_VAL_COMPLEX(123, 35), MAKE_VAL_COMPLEX(-90, -71), MAKE_VAL_COMPLEX(5, -28), MAKE_VAL_COMPLEX(-72, -52), MAKE_VAL_COMPLEX(86, -74), MAKE_VAL_COMPLEX(-4, 1), MAKE_VAL_COMPLEX(15, 205) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_320, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_320", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_320() */
/* END TEST 320 */

/*
 *  BEGIN TEST 321
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_321( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(56, -63), MAKE_VAL_COMPLEX(-21, -58), MAKE_VAL_COMPLEX(21, -2), MAKE_VAL_COMPLEX(-10, -44), MAKE_VAL_COMPLEX(189, 45), MAKE_VAL_COMPLEX(23, 118), MAKE_VAL_COMPLEX(11, -111), MAKE_VAL_COMPLEX(-12, -1), MAKE_VAL_COMPLEX(-106, -106), MAKE_VAL_COMPLEX(165, 93), MAKE_VAL_COMPLEX(-43, 42), MAKE_VAL_COMPLEX(32, 84), MAKE_VAL_COMPLEX(-114, 121), MAKE_VAL_COMPLEX(-76, -19), MAKE_VAL_COMPLEX(75, -195) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_321, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_321", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_321() */
/* END TEST 321 */

/*
 *  BEGIN TEST 322
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_322( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-96, 12), MAKE_VAL_COMPLEX(-6, 108), MAKE_VAL_COMPLEX(114, -24), MAKE_VAL_COMPLEX(33, 30), MAKE_VAL_COMPLEX(0, -117), MAKE_VAL_COMPLEX(54, -114), MAKE_VAL_COMPLEX(-54, 36), MAKE_VAL_COMPLEX(90, -60), MAKE_VAL_COMPLEX(-39, 117), MAKE_VAL_COMPLEX(-123, -48), MAKE_VAL_COMPLEX(66, -18), MAKE_VAL_COMPLEX(72, -48), MAKE_VAL_COMPLEX(54, -75), MAKE_VAL_COMPLEX(132, 42), MAKE_VAL_COMPLEX(-150, -30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_322, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_322", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_322() */
/* END TEST 322 */

/*
 *  BEGIN TEST 323
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_323( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_ROWMAJ, y288_VW );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-6, -31), MAKE_VAL_COMPLEX(99, -206), MAKE_VAL_COMPLEX(-407, -538), MAKE_VAL_COMPLEX(-240, -205), MAKE_VAL_COMPLEX(-30, -765), MAKE_VAL_COMPLEX(105, 70), MAKE_VAL_COMPLEX(-355, -51), MAKE_VAL_COMPLEX(-518, -139), MAKE_VAL_COMPLEX(-145, 369), MAKE_VAL_COMPLEX(863, -250), MAKE_VAL_COMPLEX(-49, 74), MAKE_VAL_COMPLEX(136, -276), MAKE_VAL_COMPLEX(149, 787), MAKE_VAL_COMPLEX(-380, -231), MAKE_VAL_COMPLEX(-805, 65) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_323, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_323", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_323() */
/* END TEST 323 */

/*
 *  BEGIN TEST 324
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_324( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	oski_value_t* y_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(87, 4), MAKE_VAL_COMPLEX(-18, 15), MAKE_VAL_COMPLEX(24, 101), MAKE_VAL_COMPLEX(20, -232), MAKE_VAL_COMPLEX(81, -11), MAKE_VAL_COMPLEX(-142, 166), MAKE_VAL_COMPLEX(-6, -14), MAKE_VAL_COMPLEX(-158, -73), MAKE_VAL_COMPLEX(164, -109), MAKE_VAL_COMPLEX(-105, 132), MAKE_VAL_COMPLEX(-44, -105), MAKE_VAL_COMPLEX(-127, 120) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_324, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_324", y_true, y_val, y288_LEN*y288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_324() */
/* END TEST 324 */

/*
 *  BEGIN TEST 325
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_325( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	oski_value_t* y_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-83, 24), MAKE_VAL_COMPLEX(62, -51), MAKE_VAL_COMPLEX(-20, -49), MAKE_VAL_COMPLEX(8, 180), MAKE_VAL_COMPLEX(-89, -37), MAKE_VAL_COMPLEX(94, -134), MAKE_VAL_COMPLEX(-2, 6), MAKE_VAL_COMPLEX(162, 45), MAKE_VAL_COMPLEX(-96, 97), MAKE_VAL_COMPLEX(49, -116), MAKE_VAL_COMPLEX(20, 49), MAKE_VAL_COMPLEX(147, -60) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_325, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_325", y_true, y_val, y288_LEN*y288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_325() */
/* END TEST 325 */

/*
 *  BEGIN TEST 326
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_326( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	oski_value_t* y_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(63, -9), MAKE_VAL_COMPLEX(-81, -99), MAKE_VAL_COMPLEX(117, -9), MAKE_VAL_COMPLEX(-117, -63), MAKE_VAL_COMPLEX(-108, 18), MAKE_VAL_COMPLEX(72, 108), MAKE_VAL_COMPLEX(-18, 18), MAKE_VAL_COMPLEX(-63, -9), MAKE_VAL_COMPLEX(-27, -153), MAKE_VAL_COMPLEX(36, 126), MAKE_VAL_COMPLEX(-126, 54), MAKE_VAL_COMPLEX(135, -45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_326, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_326", y_true, y_val, y288_LEN*y288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_326() */
/* END TEST 326 */

/*
 *  BEGIN TEST 327
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_327( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	oski_value_t* y_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-764, -128), MAKE_VAL_COMPLEX(1128, -128), MAKE_VAL_COMPLEX(58, -96), MAKE_VAL_COMPLEX(-708, 1017), MAKE_VAL_COMPLEX(-342, -1142), MAKE_VAL_COMPLEX(1056, -268), MAKE_VAL_COMPLEX(-86, -44), MAKE_VAL_COMPLEX(958, 1104), MAKE_VAL_COMPLEX(-758, 182), MAKE_VAL_COMPLEX(616, -541), MAKE_VAL_COMPLEX(-186, 21), MAKE_VAL_COMPLEX(1384, 922) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_327, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_327", y_true, y_val, y288_LEN*y288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_327() */
/* END TEST 327 */

/*
 *  BEGIN TEST 328
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_328( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-40, -49), MAKE_VAL_COMPLEX(-83, 2), MAKE_VAL_COMPLEX(160, 124), MAKE_VAL_COMPLEX(16, -11), MAKE_VAL_COMPLEX(-48, -108), MAKE_VAL_COMPLEX(81, -17), MAKE_VAL_COMPLEX(-12, 36), MAKE_VAL_COMPLEX(-22, -40), MAKE_VAL_COMPLEX(-127, 74), MAKE_VAL_COMPLEX(-82, -100), MAKE_VAL_COMPLEX(43, 22), MAKE_VAL_COMPLEX(-11, 79), MAKE_VAL_COMPLEX(-10, -96), MAKE_VAL_COMPLEX(-44, 43), MAKE_VAL_COMPLEX(95, -145) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_328, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_328", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_328() */
/* END TEST 328 */

/*
 *  BEGIN TEST 329
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_329( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-20, -51), MAKE_VAL_COMPLEX(-39, 44), MAKE_VAL_COMPLEX(-78, -10), MAKE_VAL_COMPLEX(-33, 62), MAKE_VAL_COMPLEX(177, 51), MAKE_VAL_COMPLEX(71, 21), MAKE_VAL_COMPLEX(-54, -78), MAKE_VAL_COMPLEX(132, 110), MAKE_VAL_COMPLEX(-21, -60), MAKE_VAL_COMPLEX(75, -59), MAKE_VAL_COMPLEX(9, 42), MAKE_VAL_COMPLEX(99, -23), MAKE_VAL_COMPLEX(119, 119), MAKE_VAL_COMPLEX(62, 123), MAKE_VAL_COMPLEX(-135, -35) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_329, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_329", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_329() */
/* END TEST 329 */

/*
 *  BEGIN TEST 330
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_330( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(100, -40), MAKE_VAL_COMPLEX(-24, -118), MAKE_VAL_COMPLEX(-116, 58), MAKE_VAL_COMPLEX(-44, -23), MAKE_VAL_COMPLEX(33, 126), MAKE_VAL_COMPLEX(-26, 138), MAKE_VAL_COMPLEX(48, -54), MAKE_VAL_COMPLEX(-80, 90), MAKE_VAL_COMPLEX(9, -137), MAKE_VAL_COMPLEX(146, 17), MAKE_VAL_COMPLEX(-66, 38), MAKE_VAL_COMPLEX(-64, 72), MAKE_VAL_COMPLEX(-37, 96), MAKE_VAL_COMPLEX(-154, -8), MAKE_VAL_COMPLEX(170, -10) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_330, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_330", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_330() */
/* END TEST 330 */

/*
 *  BEGIN TEST 331
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_331( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(92, -47), MAKE_VAL_COMPLEX(-141, 6), MAKE_VAL_COMPLEX(64, -432), MAKE_VAL_COMPLEX(-224, -63), MAKE_VAL_COMPLEX(-114, 636), MAKE_VAL_COMPLEX(-41, 145), MAKE_VAL_COMPLEX(228, -216), MAKE_VAL_COMPLEX(-258, 508), MAKE_VAL_COMPLEX(265, 8), MAKE_VAL_COMPLEX(146, 282), MAKE_VAL_COMPLEX(-103, 10), MAKE_VAL_COMPLEX(231, 269), MAKE_VAL_COMPLEX(-356, 476), MAKE_VAL_COMPLEX(-260, 317), MAKE_VAL_COMPLEX(-175, -495) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, -4);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_331, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_331", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_331() */
/* END TEST 331 */

/*
 *  BEGIN TEST 332
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_332( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(32, 87), MAKE_VAL_COMPLEX(69, -34), MAKE_VAL_COMPLEX(-140, -52), MAKE_VAL_COMPLEX(-6, -13), MAKE_VAL_COMPLEX(-42, -96), MAKE_VAL_COMPLEX(-119, -33), MAKE_VAL_COMPLEX(92, 68), MAKE_VAL_COMPLEX(38, -100), MAKE_VAL_COMPLEX(149, -46), MAKE_VAL_COMPLEX(-12, 56), MAKE_VAL_COMPLEX(-21, -66), MAKE_VAL_COMPLEX(-83, 27), MAKE_VAL_COMPLEX(28, -34), MAKE_VAL_COMPLEX(104, -65), MAKE_VAL_COMPLEX(45, 275) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_332, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_332", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_332() */
/* END TEST 332 */

/*
 *  BEGIN TEST 333
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_333( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-64, -23), MAKE_VAL_COMPLEX(-5, 82), MAKE_VAL_COMPLEX(172, -28), MAKE_VAL_COMPLEX(38, 5), MAKE_VAL_COMPLEX(-30, 48), MAKE_VAL_COMPLEX(71, -47), MAKE_VAL_COMPLEX(-92, -20), MAKE_VAL_COMPLEX(-38, 20), MAKE_VAL_COMPLEX(-93, 118), MAKE_VAL_COMPLEX(-68, 0), MAKE_VAL_COMPLEX(37, 18), MAKE_VAL_COMPLEX(83, -91), MAKE_VAL_COMPLEX(-52, -30), MAKE_VAL_COMPLEX(-24, 1), MAKE_VAL_COMPLEX(-125, -195) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_333, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_333", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_333() */
/* END TEST 333 */

/*
 *  BEGIN TEST 334
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_334( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(68, -56), MAKE_VAL_COMPLEX(-48, -86), MAKE_VAL_COMPLEX(-76, 74), MAKE_VAL_COMPLEX(-40, -7), MAKE_VAL_COMPLEX(57, 90), MAKE_VAL_COMPLEX(14, 114), MAKE_VAL_COMPLEX(24, -54), MAKE_VAL_COMPLEX(-40, 90), MAKE_VAL_COMPLEX(-27, -109), MAKE_VAL_COMPLEX(118, -23), MAKE_VAL_COMPLEX(-42, 46), MAKE_VAL_COMPLEX(-32, 72), MAKE_VAL_COMPLEX(-5, 84), MAKE_VAL_COMPLEX(-122, 32), MAKE_VAL_COMPLEX(130, -50) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_334, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_334", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_334() */
/* END TEST 334 */

/*
 *  BEGIN TEST 335
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x288', LAYOUT_COLMAJ>
 *  y : <'y288', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_335( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A288_ptr, A288_ind, A288_val,
		A288_M, A288_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y288_val,
		y288_LEN * y288_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y288_LEN, y288_VW, LAYOUT_COLMAJ, y288_LEN );
	oski_value_t* y_val = array_Duplicate( x288_val,
		x288_LEN * x288_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x288_LEN, x288_VW, LAYOUT_COLMAJ, x288_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(8, -33), MAKE_VAL_COMPLEX(197, 98), MAKE_VAL_COMPLEX(644, -304), MAKE_VAL_COMPLEX(174, -163), MAKE_VAL_COMPLEX(-432, 864), MAKE_VAL_COMPLEX(37, 83), MAKE_VAL_COMPLEX(-332, -200), MAKE_VAL_COMPLEX(-786, 240), MAKE_VAL_COMPLEX(-253, 372), MAKE_VAL_COMPLEX(-12, 450), MAKE_VAL_COMPLEX(-73, 82), MAKE_VAL_COMPLEX(165, -557), MAKE_VAL_COMPLEX(-750, 26), MAKE_VAL_COMPLEX(-684, -433), MAKE_VAL_COMPLEX(-255, -945) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_335, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_335", y_true, y_val, x288_LEN*x288_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_335() */
/* END TEST 335 */

#define A336_M 4    /* # of rows */
#define A336_N 5    /* # of cols */
#define A336_NNZ 4  /* # of stored non-zeros */
static oski_index_t A336_ptr[A336_M+1] = {0, 1, 2, 3, 4}; /* end row pointers for A336 */
static oski_index_t A336_ind[A336_NNZ] = {1, 3, 4, 4}; /* end column indices pointers for A336 */
static oski_value_t A336_val[A336_NNZ] = {MAKE_VAL_COMPLEX(7, -10), MAKE_VAL_COMPLEX(10, 5), MAKE_VAL_COMPLEX(8, 0), MAKE_VAL_COMPLEX(5, 10)}; /* end stored non-zero values for A336 */

#define x336_LEN 5   /* vector length */
#define x336_VW  3   /* vector width */
static oski_value_t x336_val[x336_LEN*x336_VW] = {MAKE_VAL_COMPLEX(-3, 10), MAKE_VAL_COMPLEX(1, -7), MAKE_VAL_COMPLEX(-10, 5), MAKE_VAL_COMPLEX(5, 10), MAKE_VAL_COMPLEX(1, -5), MAKE_VAL_COMPLEX(9, 1), MAKE_VAL_COMPLEX(10, 7), MAKE_VAL_COMPLEX(10, -7), MAKE_VAL_COMPLEX(0, -7), MAKE_VAL_COMPLEX(-3, 5), MAKE_VAL_COMPLEX(-9, 9), MAKE_VAL_COMPLEX(10, -10), MAKE_VAL_COMPLEX(-10, 9), MAKE_VAL_COMPLEX(1, 8), MAKE_VAL_COMPLEX(-5, -9)}; /* data */

#define y336_LEN 4   /* vector length */
#define y336_VW  3   /* vector width */
static oski_value_t y336_val[y336_LEN*y336_VW] = {MAKE_VAL_COMPLEX(-3, 10), MAKE_VAL_COMPLEX(4, 9), MAKE_VAL_COMPLEX(-10, -3), MAKE_VAL_COMPLEX(-3, 1), MAKE_VAL_COMPLEX(10, -2), MAKE_VAL_COMPLEX(6, 1), MAKE_VAL_COMPLEX(-1, 3), MAKE_VAL_COMPLEX(-9, -10), MAKE_VAL_COMPLEX(-10, 2), MAKE_VAL_COMPLEX(-2, 3), MAKE_VAL_COMPLEX(4, -6), MAKE_VAL_COMPLEX(10, 4)}; /* data */

/*
 *  BEGIN TEST 336
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_336( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	oski_value_t* y_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(60, 52), MAKE_VAL_COMPLEX(-80, 8), MAKE_VAL_COMPLEX(41, -150), MAKE_VAL_COMPLEX(-68, 31), MAKE_VAL_COMPLEX(-82, 92), MAKE_VAL_COMPLEX(177, -9), MAKE_VAL_COMPLEX(-92, 85), MAKE_VAL_COMPLEX(42, -37), MAKE_VAL_COMPLEX(-92, -131), MAKE_VAL_COMPLEX(-169, -50), MAKE_VAL_COMPLEX(-32, 59), MAKE_VAL_COMPLEX(91, -29) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_336, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_336", y_true, y_val, y336_LEN*y336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_336() */
/* END TEST 336 */

/*
 *  BEGIN TEST 337
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_337( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	oski_value_t* y_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-163, -72), MAKE_VAL_COMPLEX(-6, 41), MAKE_VAL_COMPLEX(-21, 47), MAKE_VAL_COMPLEX(55, -60), MAKE_VAL_COMPLEX(112, 6), MAKE_VAL_COMPLEX(-181, 70), MAKE_VAL_COMPLEX(61, -92), MAKE_VAL_COMPLEX(49, -63), MAKE_VAL_COMPLEX(62, 33), MAKE_VAL_COMPLEX(137, 33), MAKE_VAL_COMPLEX(96, -25), MAKE_VAL_COMPLEX(-121, 133) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_337, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_337", y_true, y_val, y336_LEN*y336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_337() */
/* END TEST 337 */

/*
 *  BEGIN TEST 338
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_338( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	oski_value_t* y_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(52, -28), MAKE_VAL_COMPLEX(20, -52), MAKE_VAL_COMPLEX(28, 52), MAKE_VAL_COMPLEX(16, 8), MAKE_VAL_COMPLEX(-48, -32), MAKE_VAL_COMPLEX(-20, -28), MAKE_VAL_COMPLEX(16, -8), MAKE_VAL_COMPLEX(-4, 76), MAKE_VAL_COMPLEX(48, 32), MAKE_VAL_COMPLEX(20, -4), MAKE_VAL_COMPLEX(-40, 8), MAKE_VAL_COMPLEX(-24, -56) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_338, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_338", y_true, y_val, y336_LEN*y336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_338() */
/* END TEST 338 */

/*
 *  BEGIN TEST 339
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_339( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	oski_value_t* y_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(570, -259), MAKE_VAL_COMPLEX(-193, -48), MAKE_VAL_COMPLEX(-92, -339), MAKE_VAL_COMPLEX(-12, 314), MAKE_VAL_COMPLEX(-280, 428), MAKE_VAL_COMPLEX(351, -676), MAKE_VAL_COMPLEX(52, 462), MAKE_VAL_COMPLEX(117, 178), MAKE_VAL_COMPLEX(-359, -23), MAKE_VAL_COMPLEX(-556, 303), MAKE_VAL_COMPLEX(-121, 381), MAKE_VAL_COMPLEX(-34, -622) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, -3);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_339, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_339", y_true, y_val, y336_LEN*y336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_339() */
/* END TEST 339 */

/*
 *  BEGIN TEST 340
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_340( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(18, -60), MAKE_VAL_COMPLEX(-3, 58), MAKE_VAL_COMPLEX(60, -38), MAKE_VAL_COMPLEX(41, 31), MAKE_VAL_COMPLEX(121, 56), MAKE_VAL_COMPLEX(-157, 73), MAKE_VAL_COMPLEX(-71, -46), MAKE_VAL_COMPLEX(-79, 39), MAKE_VAL_COMPLEX(-10, 51), MAKE_VAL_COMPLEX(-16, -37), MAKE_VAL_COMPLEX(177, -39), MAKE_VAL_COMPLEX(-5, 114), MAKE_VAL_COMPLEX(22, -44), MAKE_VAL_COMPLEX(1, -126), MAKE_VAL_COMPLEX(-35, 199) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_340, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_340", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_340() */
/* END TEST 340 */

/*
 *  BEGIN TEST 341
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_341( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(24, 29), MAKE_VAL_COMPLEX(-22, -33), MAKE_VAL_COMPLEX(-5, 48), MAKE_VAL_COMPLEX(-31, -86), MAKE_VAL_COMPLEX(-140, -39), MAKE_VAL_COMPLEX(124, -104), MAKE_VAL_COMPLEX(52, -12), MAKE_VAL_COMPLEX(18, -41), MAKE_VAL_COMPLEX(-11, -23), MAKE_VAL_COMPLEX(43, 26), MAKE_VAL_COMPLEX(-114, 30), MAKE_VAL_COMPLEX(-65, -104), MAKE_VAL_COMPLEX(45, 38), MAKE_VAL_COMPLEX(19, 91), MAKE_VAL_COMPLEX(28, -148) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_341, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_341", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_341() */
/* END TEST 341 */

/*
 *  BEGIN TEST 342
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_342( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(32, -34), MAKE_VAL_COMPLEX(-18, 26), MAKE_VAL_COMPLEX(50, 0), MAKE_VAL_COMPLEX(0, -50), MAKE_VAL_COMPLEX(-14, 18), MAKE_VAL_COMPLEX(-34, -22), MAKE_VAL_COMPLEX(-26, -48), MAKE_VAL_COMPLEX(-54, 8), MAKE_VAL_COMPLEX(-14, 28), MAKE_VAL_COMPLEX(22, -14), MAKE_VAL_COMPLEX(54, -18), MAKE_VAL_COMPLEX(-60, 20), MAKE_VAL_COMPLEX(58, -16), MAKE_VAL_COMPLEX(12, -34), MAKE_VAL_COMPLEX(2, 46) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_342, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_342", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_342() */
/* END TEST 342 */

/*
 *  BEGIN TEST 343
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_343( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-67, -31), MAKE_VAL_COMPLEX(14, -79), MAKE_VAL_COMPLEX(-17, -27), MAKE_VAL_COMPLEX(-592, -450), MAKE_VAL_COMPLEX(-754, -17), MAKE_VAL_COMPLEX(523, -515), MAKE_VAL_COMPLEX(11, 76), MAKE_VAL_COMPLEX(156, 76), MAKE_VAL_COMPLEX(100, -57), MAKE_VAL_COMPLEX(179, -18), MAKE_VAL_COMPLEX(-807, -39), MAKE_VAL_COMPLEX(-324, -189), MAKE_VAL_COMPLEX(165, -177), MAKE_VAL_COMPLEX(-21, 474), MAKE_VAL_COMPLEX(313, -961) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_343, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_343", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_343() */
/* END TEST 343 */

/*
 *  BEGIN TEST 344
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_344( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(11, 36), MAKE_VAL_COMPLEX(-8, -7), MAKE_VAL_COMPLEX(-20, 27), MAKE_VAL_COMPLEX(-94, 51), MAKE_VAL_COMPLEX(-60, 89), MAKE_VAL_COMPLEX(-14, -136), MAKE_VAL_COMPLEX(33, -3), MAKE_VAL_COMPLEX(-3, -44), MAKE_VAL_COMPLEX(-24, -12), MAKE_VAL_COMPLEX(-23, 44), MAKE_VAL_COMPLEX(94, -40), MAKE_VAL_COMPLEX(75, -56), MAKE_VAL_COMPLEX(10, 97), MAKE_VAL_COMPLEX(-94, -136), MAKE_VAL_COMPLEX(-18, -72) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_344, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_344", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_344() */
/* END TEST 344 */

/*
 *  BEGIN TEST 345
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_345( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(31, 42), MAKE_VAL_COMPLEX(-28, -41), MAKE_VAL_COMPLEX(-10, 63), MAKE_VAL_COMPLEX(184, -21), MAKE_VAL_COMPLEX(36, -125), MAKE_VAL_COMPLEX(74, 88), MAKE_VAL_COMPLEX(69, -15), MAKE_VAL_COMPLEX(21, -58), MAKE_VAL_COMPLEX(-18, -30), MAKE_VAL_COMPLEX(35, 4), MAKE_VAL_COMPLEX(-94, 148), MAKE_VAL_COMPLEX(-75, -64), MAKE_VAL_COMPLEX(-16, 17), MAKE_VAL_COMPLEX(148, 178), MAKE_VAL_COMPLEX(-66, 48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_345, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_345", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_345() */
/* END TEST 345 */

/*
 *  BEGIN TEST 346
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_346( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-27, -19), MAKE_VAL_COMPLEX(20, 10), MAKE_VAL_COMPLEX(-5, -35), MAKE_VAL_COMPLEX(-35, 5), MAKE_VAL_COMPLEX(14, 8), MAKE_VAL_COMPLEX(-12, 26), MAKE_VAL_COMPLEX(-31, 23), MAKE_VAL_COMPLEX(11, 37), MAKE_VAL_COMPLEX(21, 7), MAKE_VAL_COMPLEX(-12, -14), MAKE_VAL_COMPLEX(-18, -36), MAKE_VAL_COMPLEX(20, 40), MAKE_VAL_COMPLEX(-17, -39), MAKE_VAL_COMPLEX(-25, -5), MAKE_VAL_COMPLEX(32, -6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_346, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_346", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_346() */
/* END TEST 346 */

/*
 *  BEGIN TEST 347
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_347( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(62, 84), MAKE_VAL_COMPLEX(78, -74), MAKE_VAL_COMPLEX(-84, 110), MAKE_VAL_COMPLEX(358, 1052), MAKE_VAL_COMPLEX(814, 386), MAKE_VAL_COMPLEX(-906, 278), MAKE_VAL_COMPLEX(84, 50), MAKE_VAL_COMPLEX(-20, 30), MAKE_VAL_COMPLEX(16, 38), MAKE_VAL_COMPLEX(206, 246), MAKE_VAL_COMPLEX(-662, -698), MAKE_VAL_COMPLEX(-68, -660), MAKE_VAL_COMPLEX(412, -42), MAKE_VAL_COMPLEX(-1194, 944), MAKE_VAL_COMPLEX(-542, -134) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_347, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_347", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_347() */
/* END TEST 347 */

/*
 *  BEGIN TEST 348
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_348( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	oski_value_t* y_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(81, -18), MAKE_VAL_COMPLEX(-116, 42), MAKE_VAL_COMPLEX(-22, 28), MAKE_VAL_COMPLEX(-140, -71), MAKE_VAL_COMPLEX(-60, 14), MAKE_VAL_COMPLEX(-158, 73), MAKE_VAL_COMPLEX(3, 42), MAKE_VAL_COMPLEX(3, 35), MAKE_VAL_COMPLEX(81, -144), MAKE_VAL_COMPLEX(147, -70), MAKE_VAL_COMPLEX(-16, -37), MAKE_VAL_COMPLEX(21, -57) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_348, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_348", y_true, y_val, y336_LEN*y336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_348() */
/* END TEST 348 */

/*
 *  BEGIN TEST 349
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_349( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	oski_value_t* y_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-105, -11), MAKE_VAL_COMPLEX(81, -48), MAKE_VAL_COMPLEX(51, -52), MAKE_VAL_COMPLEX(143, 60), MAKE_VAL_COMPLEX(46, 20), MAKE_VAL_COMPLEX(143, -57), MAKE_VAL_COMPLEX(-10, -51), MAKE_VAL_COMPLEX(45, -42), MAKE_VAL_COMPLEX(-67, 110), MAKE_VAL_COMPLEX(-152, 58), MAKE_VAL_COMPLEX(26, 61), MAKE_VAL_COMPLEX(-53, 79) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_349, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_349", y_true, y_val, y336_LEN*y336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_349() */
/* END TEST 349 */

/*
 *  BEGIN TEST 350
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_350( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	oski_value_t* y_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(35, -44), MAKE_VAL_COMPLEX(-2, -53), MAKE_VAL_COMPLEX(44, 35), MAKE_VAL_COMPLEX(17, 1), MAKE_VAL_COMPLEX(-54, -10), MAKE_VAL_COMPLEX(-28, -17), MAKE_VAL_COMPLEX(11, -13), MAKE_VAL_COMPLEX(25, 68), MAKE_VAL_COMPLEX(54, 10), MAKE_VAL_COMPLEX(16, -11), MAKE_VAL_COMPLEX(-32, 22), MAKE_VAL_COMPLEX(-42, -40) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_350, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_350", y_true, y_val, y336_LEN*y336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_350() */
/* END TEST 350 */

/*
 *  BEGIN TEST 351
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_351( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	oski_value_t* y_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1052, 533), MAKE_VAL_COMPLEX(-595, 351), MAKE_VAL_COMPLEX(-787, 495), MAKE_VAL_COMPLEX(-1200, -755), MAKE_VAL_COMPLEX(-240, -486), MAKE_VAL_COMPLEX(-1281, 136), MAKE_VAL_COMPLEX(25, 548), MAKE_VAL_COMPLEX(-822, 280), MAKE_VAL_COMPLEX(689, -642), MAKE_VAL_COMPLEX(1504, -131), MAKE_VAL_COMPLEX(-152, -775), MAKE_VAL_COMPLEX(877, -717) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(9, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_351, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_351", y_true, y_val, y336_LEN*y336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_351() */
/* END TEST 351 */

/*
 *  BEGIN TEST 352
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_352( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(98, -36), MAKE_VAL_COMPLEX(-53, 39), MAKE_VAL_COMPLEX(100, 47), MAKE_VAL_COMPLEX(128, -1), MAKE_VAL_COMPLEX(9, -86), MAKE_VAL_COMPLEX(-107, 38), MAKE_VAL_COMPLEX(-24, -132), MAKE_VAL_COMPLEX(-127, -28), MAKE_VAL_COMPLEX(-52, 43), MAKE_VAL_COMPLEX(53, 100), MAKE_VAL_COMPLEX(181, 39), MAKE_VAL_COMPLEX(-175, 14), MAKE_VAL_COMPLEX(37, -32), MAKE_VAL_COMPLEX(104, -180), MAKE_VAL_COMPLEX(5, 175) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_352, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_352", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_352() */
/* END TEST 352 */

/*
 *  BEGIN TEST 353
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_353( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(51, 48), MAKE_VAL_COMPLEX(-48, -32), MAKE_VAL_COMPLEX(0, 78), MAKE_VAL_COMPLEX(-3, -99), MAKE_VAL_COMPLEX(-82, 87), MAKE_VAL_COMPLEX(94, -167), MAKE_VAL_COMPLEX(92, -29), MAKE_VAL_COMPLEX(-1, -91), MAKE_VAL_COMPLEX(-46, -22), MAKE_VAL_COMPLEX(26, -73), MAKE_VAL_COMPLEX(-28, 60), MAKE_VAL_COMPLEX(5, -124), MAKE_VAL_COMPLEX(119, 145), MAKE_VAL_COMPLEX(5, 142), MAKE_VAL_COMPLEX(-116, -78) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_353, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_353", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_353() */
/* END TEST 353 */

/*
 *  BEGIN TEST 354
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_354( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-44, -35), MAKE_VAL_COMPLEX(33, 19), MAKE_VAL_COMPLEX(-5, -60), MAKE_VAL_COMPLEX(-60, 5), MAKE_VAL_COMPLEX(23, 15), MAKE_VAL_COMPLEX(-23, 43), MAKE_VAL_COMPLEX(-55, 36), MAKE_VAL_COMPLEX(15, 64), MAKE_VAL_COMPLEX(35, 14), MAKE_VAL_COMPLEX(-19, -25), MAKE_VAL_COMPLEX(-27, -63), MAKE_VAL_COMPLEX(30, 70), MAKE_VAL_COMPLEX(-25, -68), MAKE_VAL_COMPLEX(-42, -11), MAKE_VAL_COMPLEX(55, -7) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_354, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_354", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_354() */
/* END TEST 354 */

/*
 *  BEGIN TEST 355
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_355( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(34, 32), MAKE_VAL_COMPLEX(49, -47), MAKE_VAL_COMPLEX(-18, 35), MAKE_VAL_COMPLEX(1034, 27), MAKE_VAL_COMPLEX(-391, -884), MAKE_VAL_COMPLEX(413, 912), MAKE_VAL_COMPLEX(-112, 44), MAKE_VAL_COMPLEX(-3, 62), MAKE_VAL_COMPLEX(-2, -33), MAKE_VAL_COMPLEX(625, 582), MAKE_VAL_COMPLEX(419, -171), MAKE_VAL_COMPLEX(9, 320), MAKE_VAL_COMPLEX(-807, 338), MAKE_VAL_COMPLEX(-480, -884), MAKE_VAL_COMPLEX(675, 125) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_355, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_355", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_355() */
/* END TEST 355 */

/*
 *  BEGIN TEST 356
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_356( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(46, 101), MAKE_VAL_COMPLEX(-32, -58), MAKE_VAL_COMPLEX(-45, 107), MAKE_VAL_COMPLEX(-12, 84), MAKE_VAL_COMPLEX(68, 45), MAKE_VAL_COMPLEX(-22, -139), MAKE_VAL_COMPLEX(109, -24), MAKE_VAL_COMPLEX(20, -116), MAKE_VAL_COMPLEX(-45, -55), MAKE_VAL_COMPLEX(96, 127), MAKE_VAL_COMPLEX(56, 96), MAKE_VAL_COMPLEX(5, -96), MAKE_VAL_COMPLEX(-92, 144), MAKE_VAL_COMPLEX(-90, 113), MAKE_VAL_COMPLEX(24, -156) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_356, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_356", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_356() */
/* END TEST 356 */

/*
 *  BEGIN TEST 357
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_357( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(80, 24), MAKE_VAL_COMPLEX(-65, -13), MAKE_VAL_COMPLEX(40, 83), MAKE_VAL_COMPLEX(202, -79), MAKE_VAL_COMPLEX(-135, -100), MAKE_VAL_COMPLEX(109, 12), MAKE_VAL_COMPLEX(76, -70), MAKE_VAL_COMPLEX(-45, -90), MAKE_VAL_COMPLEX(-60, -1), MAKE_VAL_COMPLEX(-45, -42), MAKE_VAL_COMPLEX(7, 111), MAKE_VAL_COMPLEX(-75, -134), MAKE_VAL_COMPLEX(147, 78), MAKE_VAL_COMPLEX(218, -64), MAKE_VAL_COMPLEX(-199, 159) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_357, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_357", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_357() */
/* END TEST 357 */

/*
 *  BEGIN TEST 358
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_358( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-70, -21), MAKE_VAL_COMPLEX(49, 7), MAKE_VAL_COMPLEX(-35, -70), MAKE_VAL_COMPLEX(-70, 35), MAKE_VAL_COMPLEX(35, 7), MAKE_VAL_COMPLEX(-7, 63), MAKE_VAL_COMPLEX(-49, 70), MAKE_VAL_COMPLEX(49, 70), MAKE_VAL_COMPLEX(49, 0), MAKE_VAL_COMPLEX(-35, -21), MAKE_VAL_COMPLEX(-63, -63), MAKE_VAL_COMPLEX(70, 70), MAKE_VAL_COMPLEX(-63, -70), MAKE_VAL_COMPLEX(-56, 7), MAKE_VAL_COMPLEX(63, -35) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_358, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_358", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_358() */
/* END TEST 358 */

/*
 *  BEGIN TEST 359
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_ROWMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_359( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_ROWMAJ, x336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(18, 49), MAKE_VAL_COMPLEX(-1, 25), MAKE_VAL_COMPLEX(-13, 38), MAKE_VAL_COMPLEX(-594, -307), MAKE_VAL_COMPLEX(1, 725), MAKE_VAL_COMPLEX(-29, -763), MAKE_VAL_COMPLEX(21, -122), MAKE_VAL_COMPLEX(-65, -62), MAKE_VAL_COMPLEX(-9, -8), MAKE_VAL_COMPLEX(79, 633), MAKE_VAL_COMPLEX(407, 167), MAKE_VAL_COMPLEX(-226, 126), MAKE_VAL_COMPLEX(-321, -226), MAKE_VAL_COMPLEX(-812, -363), MAKE_VAL_COMPLEX(937, 11) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_359, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_359", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_359() */
/* END TEST 359 */

/*
 *  BEGIN TEST 360
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_360( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	oski_value_t* y_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-134, -4), MAKE_VAL_COMPLEX(128, 24), MAKE_VAL_COMPLEX(-84, -229), MAKE_VAL_COMPLEX(-22, 109), MAKE_VAL_COMPLEX(115, -25), MAKE_VAL_COMPLEX(11, 111), MAKE_VAL_COMPLEX(-23, -22), MAKE_VAL_COMPLEX(-18, -72), MAKE_VAL_COMPLEX(-120, -101), MAKE_VAL_COMPLEX(33, 3), MAKE_VAL_COMPLEX(-11, -28), MAKE_VAL_COMPLEX(106, -13) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_360, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_360", y_true, y_val, y336_LEN*y336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_360() */
/* END TEST 360 */

/*
 *  BEGIN TEST 361
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_361( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	oski_value_t* y_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-11, 15), MAKE_VAL_COMPLEX(-225, 73), MAKE_VAL_COMPLEX(73, 84), MAKE_VAL_COMPLEX(-6, -143), MAKE_VAL_COMPLEX(-39, 145), MAKE_VAL_COMPLEX(6, -28), MAKE_VAL_COMPLEX(-21, 24), MAKE_VAL_COMPLEX(103, -95), MAKE_VAL_COMPLEX(44, -19), MAKE_VAL_COMPLEX(-82, -14), MAKE_VAL_COMPLEX(109, 50), MAKE_VAL_COMPLEX(-108, 163) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_361, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_361", y_true, y_val, y336_LEN*y336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_361() */
/* END TEST 361 */

/*
 *  BEGIN TEST 362
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_362( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	oski_value_t* y_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(20, 6), MAKE_VAL_COMPLEX(18, -8), MAKE_VAL_COMPLEX(-6, 20), MAKE_VAL_COMPLEX(2, 6), MAKE_VAL_COMPLEX(-4, -20), MAKE_VAL_COMPLEX(2, -12), MAKE_VAL_COMPLEX(6, 2), MAKE_VAL_COMPLEX(-20, 18), MAKE_VAL_COMPLEX(4, 20), MAKE_VAL_COMPLEX(6, 4), MAKE_VAL_COMPLEX(-12, -8), MAKE_VAL_COMPLEX(8, -20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_362, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_362", y_true, y_val, y336_LEN*y336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_362() */
/* END TEST 362 */

/*
 *  BEGIN TEST 363
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_363( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	oski_value_t* y_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(402, 482), MAKE_VAL_COMPLEX(-1056, 260), MAKE_VAL_COMPLEX(39, 1155), MAKE_VAL_COMPLEX(91, -817), MAKE_VAL_COMPLEX(-310, 372), MAKE_VAL_COMPLEX(258, -504), MAKE_VAL_COMPLEX(-25, 269), MAKE_VAL_COMPLEX(58, -278), MAKE_VAL_COMPLEX(219, 515), MAKE_VAL_COMPLEX(-436, -2), MAKE_VAL_COMPLEX(465, 103), MAKE_VAL_COMPLEX(-475, 561) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_363, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_363", y_true, y_val, y336_LEN*y336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_363() */
/* END TEST 363 */

/*
 *  BEGIN TEST 364
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_364( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(53, 114), MAKE_VAL_COMPLEX(28, 37), MAKE_VAL_COMPLEX(-41, 123), MAKE_VAL_COMPLEX(83, 38), MAKE_VAL_COMPLEX(-80, -29), MAKE_VAL_COMPLEX(84, -55), MAKE_VAL_COMPLEX(264, -3), MAKE_VAL_COMPLEX(15, -146), MAKE_VAL_COMPLEX(58, -32), MAKE_VAL_COMPLEX(24, -6), MAKE_VAL_COMPLEX(-10, 141), MAKE_VAL_COMPLEX(-94, -80), MAKE_VAL_COMPLEX(-18, 154), MAKE_VAL_COMPLEX(137, 100), MAKE_VAL_COMPLEX(-182, 104) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_364, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_364", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_364() */
/* END TEST 364 */

/*
 *  BEGIN TEST 365
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_365( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(25, -47), MAKE_VAL_COMPLEX(-87, -74), MAKE_VAL_COMPLEX(46, -13), MAKE_VAL_COMPLEX(27, -43), MAKE_VAL_COMPLEX(39, 0), MAKE_VAL_COMPLEX(-39, -22), MAKE_VAL_COMPLEX(-161, -59), MAKE_VAL_COMPLEX(-38, 28), MAKE_VAL_COMPLEX(-121, 4), MAKE_VAL_COMPLEX(9, 53), MAKE_VAL_COMPLEX(55, -24), MAKE_VAL_COMPLEX(44, -50), MAKE_VAL_COMPLEX(59, -28), MAKE_VAL_COMPLEX(-61, -77), MAKE_VAL_COMPLEX(81, -95) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_365, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_365", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_365() */
/* END TEST 365 */

/*
 *  BEGIN TEST 366
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_366( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(91, -49), MAKE_VAL_COMPLEX(-56, 42), MAKE_VAL_COMPLEX(105, 35), MAKE_VAL_COMPLEX(35, -105), MAKE_VAL_COMPLEX(-42, 28), MAKE_VAL_COMPLEX(-56, -70), MAKE_VAL_COMPLEX(-21, -119), MAKE_VAL_COMPLEX(-119, -21), MAKE_VAL_COMPLEX(-49, 49), MAKE_VAL_COMPLEX(56, -14), MAKE_VAL_COMPLEX(126, 0), MAKE_VAL_COMPLEX(-140, 0), MAKE_VAL_COMPLEX(133, 7), MAKE_VAL_COMPLEX(49, -63), MAKE_VAL_COMPLEX(-28, 98) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_366, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_366", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_366() */
/* END TEST 366 */

/*
 *  BEGIN TEST 367
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_367( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-82, -90), MAKE_VAL_COMPLEX(-476, 427), MAKE_VAL_COMPLEX(20, -85), MAKE_VAL_COMPLEX(-70, -220), MAKE_VAL_COMPLEX(-76, -205), MAKE_VAL_COMPLEX(-104, 59), MAKE_VAL_COMPLEX(-200, 648), MAKE_VAL_COMPLEX(25, 47), MAKE_VAL_COMPLEX(-85, 612), MAKE_VAL_COMPLEX(343, -5), MAKE_VAL_COMPLEX(24, -149), MAKE_VAL_COMPLEX(-410, -360), MAKE_VAL_COMPLEX(5, -154), MAKE_VAL_COMPLEX(-266, 282), MAKE_VAL_COMPLEX(-605, -321) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_367, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_367", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_367() */
/* END TEST 367 */

/*
 *  BEGIN TEST 368
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_368( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-21, 70), MAKE_VAL_COMPLEX(-118, -1), MAKE_VAL_COMPLEX(-61, 33), MAKE_VAL_COMPLEX(3, 88), MAKE_VAL_COMPLEX(18, 29), MAKE_VAL_COMPLEX(58, 15), MAKE_VAL_COMPLEX(8, 143), MAKE_VAL_COMPLEX(51, -52), MAKE_VAL_COMPLEX(94, -118), MAKE_VAL_COMPLEX(-130, -120), MAKE_VAL_COMPLEX(-64, 51), MAKE_VAL_COMPLEX(26, -180), MAKE_VAL_COMPLEX(-70, 56), MAKE_VAL_COMPLEX(81, 32), MAKE_VAL_COMPLEX(-20, -118) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_368, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_368", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_368() */
/* END TEST 368 */

/*
 *  BEGIN TEST 369
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_369( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-12, -69), MAKE_VAL_COMPLEX(140, -3), MAKE_VAL_COMPLEX(36, -58), MAKE_VAL_COMPLEX(-28, -63), MAKE_VAL_COMPLEX(-2, -31), MAKE_VAL_COMPLEX(-52, 13), MAKE_VAL_COMPLEX(-19, -106), MAKE_VAL_COMPLEX(-20, 75), MAKE_VAL_COMPLEX(-73, 111), MAKE_VAL_COMPLEX(112, 116), MAKE_VAL_COMPLEX(28, -69), MAKE_VAL_COMPLEX(14, 200), MAKE_VAL_COMPLEX(33, -77), MAKE_VAL_COMPLEX(-104, -21), MAKE_VAL_COMPLEX(42, 94) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_369, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_369", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_369() */
/* END TEST 369 */

/*
 *  BEGIN TEST 370
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_370( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(5, -53), MAKE_VAL_COMPLEX(2, 36), MAKE_VAL_COMPLEX(45, -35), MAKE_VAL_COMPLEX(-35, -45), MAKE_VAL_COMPLEX(0, 26), MAKE_VAL_COMPLEX(-46, 4), MAKE_VAL_COMPLEX(-57, -25), MAKE_VAL_COMPLEX(-43, 45), MAKE_VAL_COMPLEX(7, 35), MAKE_VAL_COMPLEX(10, -28), MAKE_VAL_COMPLEX(36, -54), MAKE_VAL_COMPLEX(-40, 60), MAKE_VAL_COMPLEX(41, -55), MAKE_VAL_COMPLEX(-13, -39), MAKE_VAL_COMPLEX(34, 40) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_370, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_370", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_370() */
/* END TEST 370 */

/*
 *  BEGIN TEST 371
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_371( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_ROWMAJ, y336_VW );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(67, 31), MAKE_VAL_COMPLEX(759, 967), MAKE_VAL_COMPLEX(-20, 0), MAKE_VAL_COMPLEX(333, 209), MAKE_VAL_COMPLEX(509, -303), MAKE_VAL_COMPLEX(103, -41), MAKE_VAL_COMPLEX(1091, 213), MAKE_VAL_COMPLEX(-9, 103), MAKE_VAL_COMPLEX(-952, -646), MAKE_VAL_COMPLEX(-1036, 1472), MAKE_VAL_COMPLEX(-51, 117), MAKE_VAL_COMPLEX(-918, 646), MAKE_VAL_COMPLEX(-10, 100), MAKE_VAL_COMPLEX(-381, -593), MAKE_VAL_COMPLEX(-608, 56) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-3, -9);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_371, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_371", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_371() */
/* END TEST 371 */

/*
 *  BEGIN TEST 372
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_372( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	oski_value_t* y_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-136, -70), MAKE_VAL_COMPLEX(-62, 146), MAKE_VAL_COMPLEX(19, -105), MAKE_VAL_COMPLEX(53, -26), MAKE_VAL_COMPLEX(163, 20), MAKE_VAL_COMPLEX(38, -21), MAKE_VAL_COMPLEX(-35, 26), MAKE_VAL_COMPLEX(5, -75), MAKE_VAL_COMPLEX(-53, -231), MAKE_VAL_COMPLEX(-41, 61), MAKE_VAL_COMPLEX(-8, -35), MAKE_VAL_COMPLEX(38, -17) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_372, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_372", y_true, y_val, y336_LEN*y336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_372() */
/* END TEST 372 */

/*
 *  BEGIN TEST 373
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_373( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	oski_value_t* y_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-6, 71), MAKE_VAL_COMPLEX(-39, -58), MAKE_VAL_COMPLEX(-20, -37), MAKE_VAL_COMPLEX(-78, -9), MAKE_VAL_COMPLEX(-97, 102), MAKE_VAL_COMPLEX(-27, 103), MAKE_VAL_COMPLEX(-8, -27), MAKE_VAL_COMPLEX(89, -82), MAKE_VAL_COMPLEX(-13, 109), MAKE_VAL_COMPLEX(-6, -75), MAKE_VAL_COMPLEX(102, 63), MAKE_VAL_COMPLEX(-50, 163) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_373, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_373", y_true, y_val, y336_LEN*y336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_373() */
/* END TEST 373 */

/*
 *  BEGIN TEST 374
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_374( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	oski_value_t* y_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-6, 20), MAKE_VAL_COMPLEX(8, 18), MAKE_VAL_COMPLEX(-20, -6), MAKE_VAL_COMPLEX(-6, 2), MAKE_VAL_COMPLEX(20, -4), MAKE_VAL_COMPLEX(12, 2), MAKE_VAL_COMPLEX(-2, 6), MAKE_VAL_COMPLEX(-18, -20), MAKE_VAL_COMPLEX(-20, 4), MAKE_VAL_COMPLEX(-4, 6), MAKE_VAL_COMPLEX(8, -12), MAKE_VAL_COMPLEX(20, 8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_374, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_374", y_true, y_val, y336_LEN*y336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_374() */
/* END TEST 374 */

/*
 *  BEGIN TEST 375
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_375( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	oski_value_t* y_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(775, 16), MAKE_VAL_COMPLEX(-434, -939), MAKE_VAL_COMPLEX(217, 354), MAKE_VAL_COMPLEX(-364, 293), MAKE_VAL_COMPLEX(-951, 874), MAKE_VAL_COMPLEX(-108, 558), MAKE_VAL_COMPLEX(2, -305), MAKE_VAL_COMPLEX(494, -5), MAKE_VAL_COMPLEX(1025, 1043), MAKE_VAL_COMPLEX(-115, -614), MAKE_VAL_COMPLEX(512, 259), MAKE_VAL_COMPLEX(-168, 747) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_375, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_375", y_true, y_val, y336_LEN*y336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_375() */
/* END TEST 375 */

/*
 *  BEGIN TEST 376
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_376( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(85, -29), MAKE_VAL_COMPLEX(28, 144), MAKE_VAL_COMPLEX(85, 37), MAKE_VAL_COMPLEX(32, 16), MAKE_VAL_COMPLEX(-146, -26), MAKE_VAL_COMPLEX(-37, -71), MAKE_VAL_COMPLEX(45, -225), MAKE_VAL_COMPLEX(-110, -25), MAKE_VAL_COMPLEX(-3, 72), MAKE_VAL_COMPLEX(100, -125), MAKE_VAL_COMPLEX(107, 11), MAKE_VAL_COMPLEX(-182, 107), MAKE_VAL_COMPLEX(127, 10), MAKE_VAL_COMPLEX(25, -31), MAKE_VAL_COMPLEX(9, 161) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_376, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_376", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_376() */
/* END TEST 376 */

/*
 *  BEGIN TEST 377
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_377( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(53, 5), MAKE_VAL_COMPLEX(-118, -114), MAKE_VAL_COMPLEX(35, 53), MAKE_VAL_COMPLEX(58, -136), MAKE_VAL_COMPLEX(80, 44), MAKE_VAL_COMPLEX(-5, -43), MAKE_VAL_COMPLEX(-21, 63), MAKE_VAL_COMPLEX(-34, -53), MAKE_VAL_COMPLEX(-81, -30), MAKE_VAL_COMPLEX(-22, 131), MAKE_VAL_COMPLEX(55, 43), MAKE_VAL_COMPLEX(2, -167), MAKE_VAL_COMPLEX(41, 56), MAKE_VAL_COMPLEX(65, -29), MAKE_VAL_COMPLEX(-87, -47) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_377, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_377", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_377() */
/* END TEST 377 */

/*
 *  BEGIN TEST 378
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_378( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(79, -9), MAKE_VAL_COMPLEX(-52, 14), MAKE_VAL_COMPLEX(65, 55), MAKE_VAL_COMPLEX(55, -65), MAKE_VAL_COMPLEX(-38, 8), MAKE_VAL_COMPLEX(-20, -66), MAKE_VAL_COMPLEX(19, -91), MAKE_VAL_COMPLEX(-79, -49), MAKE_VAL_COMPLEX(-49, 21), MAKE_VAL_COMPLEX(44, 6), MAKE_VAL_COMPLEX(90, 36), MAKE_VAL_COMPLEX(-100, -40), MAKE_VAL_COMPLEX(93, 43), MAKE_VAL_COMPLEX(53, -31), MAKE_VAL_COMPLEX(-48, 62) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_378, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_378", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_378() */
/* END TEST 378 */

/*
 *  BEGIN TEST 379
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_379( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(116, 122), MAKE_VAL_COMPLEX(1494, 202), MAKE_VAL_COMPLEX(-74, 116), MAKE_VAL_COMPLEX(884, 922), MAKE_VAL_COMPLEX(-1262, 442), MAKE_VAL_COMPLEX(70, -150), MAKE_VAL_COMPLEX(-414, -1412), MAKE_VAL_COMPLEX(-26, -28), MAKE_VAL_COMPLEX(566, -128), MAKE_VAL_COMPLEX(-522, -1286), MAKE_VAL_COMPLEX(-10, 150), MAKE_VAL_COMPLEX(460, 1292), MAKE_VAL_COMPLEX(38, -20), MAKE_VAL_COMPLEX(40, 386), MAKE_VAL_COMPLEX(858, 270) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_379, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_379", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_379() */
/* END TEST 379 */

/*
 *  BEGIN TEST 380
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_380( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(27, 19), MAKE_VAL_COMPLEX(-138, 46), MAKE_VAL_COMPLEX(5, 27), MAKE_VAL_COMPLEX(112, 56), MAKE_VAL_COMPLEX(-100, 8), MAKE_VAL_COMPLEX(13, -29), MAKE_VAL_COMPLEX(117, 57), MAKE_VAL_COMPLEX(-22, -27), MAKE_VAL_COMPLEX(35, -30), MAKE_VAL_COMPLEX(-138, 73), MAKE_VAL_COMPLEX(17, 29), MAKE_VAL_COMPLEX(-122, -113), MAKE_VAL_COMPLEX(31, 24), MAKE_VAL_COMPLEX(29, 41), MAKE_VAL_COMPLEX(95, -113) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_380, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_380", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_380() */
/* END TEST 380 */

/*
 *  BEGIN TEST 381
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_381( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(27, 19), MAKE_VAL_COMPLEX(98, -66), MAKE_VAL_COMPLEX(5, 43), MAKE_VAL_COMPLEX(-42, -66), MAKE_VAL_COMPLEX(72, -24), MAKE_VAL_COMPLEX(11, -23), MAKE_VAL_COMPLEX(-55, -103), MAKE_VAL_COMPLEX(0, -47), MAKE_VAL_COMPLEX(-77, 16), MAKE_VAL_COMPLEX(162, -45), MAKE_VAL_COMPLEX(19, 43), MAKE_VAL_COMPLEX(82, 33), MAKE_VAL_COMPLEX(3, 54), MAKE_VAL_COMPLEX(21, -31), MAKE_VAL_COMPLEX(-159, 125) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_381, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_381", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_381() */
/* END TEST 381 */

/*
 *  BEGIN TEST 382
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_382( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(60, 18), MAKE_VAL_COMPLEX(-42, -6), MAKE_VAL_COMPLEX(30, 60), MAKE_VAL_COMPLEX(60, -30), MAKE_VAL_COMPLEX(-30, -6), MAKE_VAL_COMPLEX(6, -54), MAKE_VAL_COMPLEX(42, -60), MAKE_VAL_COMPLEX(-42, -60), MAKE_VAL_COMPLEX(-42, 0), MAKE_VAL_COMPLEX(30, 18), MAKE_VAL_COMPLEX(54, 54), MAKE_VAL_COMPLEX(-60, -60), MAKE_VAL_COMPLEX(54, 60), MAKE_VAL_COMPLEX(48, -6), MAKE_VAL_COMPLEX(-54, 30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_382, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_382", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_382() */
/* END TEST 382 */

/*
 *  BEGIN TEST 383
 *  Value type: complex
 *  A : <MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x336', LAYOUT_COLMAJ>
 *  y : <'y336', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_383( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A336_ptr, A336_ind, A336_val,
		A336_M, A336_N, SHARE_INPUTMAT,
		5, MAT_TRI_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y336_val,
		y336_LEN * y336_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y336_LEN, y336_VW, LAYOUT_COLMAJ, y336_LEN );
	oski_value_t* y_val = array_Duplicate( x336_val,
		x336_LEN * x336_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x336_LEN, x336_VW, LAYOUT_COLMAJ, x336_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-44, -144), MAKE_VAL_COMPLEX(12, -1194), MAKE_VAL_COMPLEX(152, -108), MAKE_VAL_COMPLEX(-997, 369), MAKE_VAL_COMPLEX(228, -770), MAKE_VAL_COMPLEX(-93, 81), MAKE_VAL_COMPLEX(-1240, 450), MAKE_VAL_COMPLEX(-110, 25), MAKE_VAL_COMPLEX(39, 680), MAKE_VAL_COMPLEX(65, -1670), MAKE_VAL_COMPLEX(103, -161), MAKE_VAL_COMPLEX(1025, -426), MAKE_VAL_COMPLEX(127, -2), MAKE_VAL_COMPLEX(-416, -196), MAKE_VAL_COMPLEX(695, 1687) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, 9);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_383, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_383", y_true, y_val, x336_LEN*x336_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_383() */
/* END TEST 383 */

#define A384_M 4    /* # of rows */
#define A384_N 5    /* # of cols */
#define A384_NNZ 4  /* # of stored non-zeros */
static oski_index_t A384_ptr[A384_M+1] = {1, 1, 2, 3, 5}; /* end row pointers for A384 */
static oski_index_t A384_ind[A384_NNZ] = {2, 3, 2, 4}; /* end column indices pointers for A384 */
static oski_value_t A384_val[A384_NNZ] = {MAKE_VAL_COMPLEX(7, -10), MAKE_VAL_COMPLEX(-10, 5), MAKE_VAL_COMPLEX(-9, 10), MAKE_VAL_COMPLEX(10, 8)}; /* end stored non-zero values for A384 */

#define x384_LEN 5   /* vector length */
#define x384_VW  3   /* vector width */
static oski_value_t x384_val[x384_LEN*x384_VW] = {MAKE_VAL_COMPLEX(1, 2), MAKE_VAL_COMPLEX(0, 10), MAKE_VAL_COMPLEX(-6, 5), MAKE_VAL_COMPLEX(-10, -7), MAKE_VAL_COMPLEX(0, 3), MAKE_VAL_COMPLEX(-4, 7), MAKE_VAL_COMPLEX(-8, 5), MAKE_VAL_COMPLEX(9, -1), MAKE_VAL_COMPLEX(7, -6), MAKE_VAL_COMPLEX(-6, 5), MAKE_VAL_COMPLEX(2, 9), MAKE_VAL_COMPLEX(-5, 1), MAKE_VAL_COMPLEX(5, -8), MAKE_VAL_COMPLEX(10, -10), MAKE_VAL_COMPLEX(6, 0)}; /* data */

#define y384_LEN 4   /* vector length */
#define y384_VW  3   /* vector width */
static oski_value_t y384_val[y384_LEN*y384_VW] = {MAKE_VAL_COMPLEX(1, 10), MAKE_VAL_COMPLEX(10, -4), MAKE_VAL_COMPLEX(-9, -1), MAKE_VAL_COMPLEX(3, -7), MAKE_VAL_COMPLEX(10, -4), MAKE_VAL_COMPLEX(-10, 7), MAKE_VAL_COMPLEX(-6, -10), MAKE_VAL_COMPLEX(7, -9), MAKE_VAL_COMPLEX(-6, 6), MAKE_VAL_COMPLEX(5, 4), MAKE_VAL_COMPLEX(-2, -1), MAKE_VAL_COMPLEX(0, 10)}; /* data */

/*
 *  BEGIN TEST 384
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_384( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	oski_value_t* y_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-33, 74), MAKE_VAL_COMPLEX(86, 12), MAKE_VAL_COMPLEX(-59, -43), MAKE_VAL_COMPLEX(-91, 14), MAKE_VAL_COMPLEX(116, 33), MAKE_VAL_COMPLEX(-56, 98), MAKE_VAL_COMPLEX(53, -184), MAKE_VAL_COMPLEX(0, 20), MAKE_VAL_COMPLEX(-106, 113), MAKE_VAL_COMPLEX(79, 13), MAKE_VAL_COMPLEX(-92, 64), MAKE_VAL_COMPLEX(-132, -63) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_384, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_384", y_true, y_val, y384_LEN*y384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_384() */
/* END TEST 384 */

/*
 *  BEGIN TEST 385
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_385( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	oski_value_t* y_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-93, -21), MAKE_VAL_COMPLEX(6, 102), MAKE_VAL_COMPLEX(36, -78), MAKE_VAL_COMPLEX(194, -3), MAKE_VAL_COMPLEX(-24, 81), MAKE_VAL_COMPLEX(-75, -200), MAKE_VAL_COMPLEX(53, 66), MAKE_VAL_COMPLEX(145, 35), MAKE_VAL_COMPLEX(4, -167), MAKE_VAL_COMPLEX(-111, 68), MAKE_VAL_COMPLEX(97, -94), MAKE_VAL_COMPLEX(2, 103) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_385, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_385", y_true, y_val, y384_LEN*y384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_385() */
/* END TEST 385 */

/*
 *  BEGIN TEST 386
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_386( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	oski_value_t* y_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(34, -64), MAKE_VAL_COMPLEX(-76, -16), MAKE_VAL_COMPLEX(50, 42), MAKE_VAL_COMPLEX(-46, 30), MAKE_VAL_COMPLEX(-76, -16), MAKE_VAL_COMPLEX(88, -2), MAKE_VAL_COMPLEX(-4, 84), MAKE_VAL_COMPLEX(-78, 26), MAKE_VAL_COMPLEX(60, -12), MAKE_VAL_COMPLEX(-14, -44), MAKE_VAL_COMPLEX(8, 14), MAKE_VAL_COMPLEX(40, -60) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_386, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_386", y_true, y_val, y384_LEN*y384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_386() */
/* END TEST 386 */

/*
 *  BEGIN TEST 387
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_387( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	oski_value_t* y_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-63, 77), MAKE_VAL_COMPLEX(98, 42), MAKE_VAL_COMPLEX(-56, -70), MAKE_VAL_COMPLEX(-706, 889), MAKE_VAL_COMPLEX(392, 69), MAKE_VAL_COMPLEX(531, 434), MAKE_VAL_COMPLEX(53, -962), MAKE_VAL_COMPLEX(-263, 711), MAKE_VAL_COMPLEX(16, 825), MAKE_VAL_COMPLEX(287, -422), MAKE_VAL_COMPLEX(-265, 860), MAKE_VAL_COMPLEX(-1246, -493) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, -4);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_387, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_387", y_true, y_val, y384_LEN*y384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_387() */
/* END TEST 387 */

/*
 *  BEGIN TEST 388
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_388( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(8, -19), MAKE_VAL_COMPLEX(70, -60), MAKE_VAL_COMPLEX(71, 12), MAKE_VAL_COMPLEX(-123, 47), MAKE_VAL_COMPLEX(79, -157), MAKE_VAL_COMPLEX(-27, 45), MAKE_VAL_COMPLEX(193, 96), MAKE_VAL_COMPLEX(-86, 68), MAKE_VAL_COMPLEX(-54, -103), MAKE_VAL_COMPLEX(89, 92), MAKE_VAL_COMPLEX(39, -94), MAKE_VAL_COMPLEX(-43, 129), MAKE_VAL_COMPLEX(-86, 13), MAKE_VAL_COMPLEX(-130, -10), MAKE_VAL_COMPLEX(-36, -42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_388, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_388", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_388() */
/* END TEST 388 */

/*
 *  BEGIN TEST 389
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_389( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-2, -14), MAKE_VAL_COMPLEX(20, -60), MAKE_VAL_COMPLEX(46, -18), MAKE_VAL_COMPLEX(180, 127), MAKE_VAL_COMPLEX(-52, 121), MAKE_VAL_COMPLEX(138, -93), MAKE_VAL_COMPLEX(-52, -84), MAKE_VAL_COMPLEX(-31, -137), MAKE_VAL_COMPLEX(-84, 112), MAKE_VAL_COMPLEX(28, -98), MAKE_VAL_COMPLEX(18, -32), MAKE_VAL_COMPLEX(112, -96), MAKE_VAL_COMPLEX(-46, 38), MAKE_VAL_COMPLEX(-80, 40), MAKE_VAL_COMPLEX(-36, -12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_389, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_389", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_389() */
/* END TEST 389 */

/*
 *  BEGIN TEST 390
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_390( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-14, 2), MAKE_VAL_COMPLEX(-60, -20), MAKE_VAL_COMPLEX(-18, -46), MAKE_VAL_COMPLEX(62, -46), MAKE_VAL_COMPLEX(-18, -6), MAKE_VAL_COMPLEX(-34, -38), MAKE_VAL_COMPLEX(-14, -58), MAKE_VAL_COMPLEX(-12, 56), MAKE_VAL_COMPLEX(22, 54), MAKE_VAL_COMPLEX(-18, -46), MAKE_VAL_COMPLEX(-58, -6), MAKE_VAL_COMPLEX(4, -32), MAKE_VAL_COMPLEX(38, 46), MAKE_VAL_COMPLEX(40, 80), MAKE_VAL_COMPLEX(-12, 36) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_390, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_390", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_390() */
/* END TEST 390 */

/*
 *  BEGIN TEST 391
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_391( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-15, 10), MAKE_VAL_COMPLEX(-80, 10), MAKE_VAL_COMPLEX(-46, -43), MAKE_VAL_COMPLEX(1512, -38), MAKE_VAL_COMPLEX(10, 1486), MAKE_VAL_COMPLEX(604, -956), MAKE_VAL_COMPLEX(-1318, -249), MAKE_VAL_COMPLEX(-258, -1154), MAKE_VAL_COMPLEX(145, 980), MAKE_VAL_COMPLEX(-528, -691), MAKE_VAL_COMPLEX(142, 211), MAKE_VAL_COMPLEX(307, -1259), MAKE_VAL_COMPLEX(69, 32), MAKE_VAL_COMPLEX(90, 70), MAKE_VAL_COMPLEX(6, 48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-9, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_391, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_391", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_391() */
/* END TEST 391 */

/*
 *  BEGIN TEST 392
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_392( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1, 18), MAKE_VAL_COMPLEX(-40, 70), MAKE_VAL_COMPLEX(-62, 11), MAKE_VAL_COMPLEX(44, -194), MAKE_VAL_COMPLEX(106, 122), MAKE_VAL_COMPLEX(-96, -108), MAKE_VAL_COMPLEX(-66, 133), MAKE_VAL_COMPLEX(-48, 84), MAKE_VAL_COMPLEX(163, -44), MAKE_VAL_COMPLEX(20, 11), MAKE_VAL_COMPLEX(-50, 77), MAKE_VAL_COMPLEX(41, 87), MAKE_VAL_COMPLEX(67, -36), MAKE_VAL_COMPLEX(110, -30), MAKE_VAL_COMPLEX(42, 24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_392, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_392", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_392() */
/* END TEST 392 */

/*
 *  BEGIN TEST 393
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_393( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-6, -2), MAKE_VAL_COMPLEX(-20, -20), MAKE_VAL_COMPLEX(2, -22), MAKE_VAL_COMPLEX(-52, 99), MAKE_VAL_COMPLEX(-124, -107), MAKE_VAL_COMPLEX(34, 119), MAKE_VAL_COMPLEX(-4, -156), MAKE_VAL_COMPLEX(99, -35), MAKE_VAL_COMPLEX(-92, 56), MAKE_VAL_COMPLEX(-80, -22), MAKE_VAL_COMPLEX(6, -20), MAKE_VAL_COMPLEX(-72, -112), MAKE_VAL_COMPLEX(6, 26), MAKE_VAL_COMPLEX(0, 40), MAKE_VAL_COMPLEX(-12, 12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_393, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_393", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_393() */
/* END TEST 393 */

/*
 *  BEGIN TEST 394
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_394( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-21, -2), MAKE_VAL_COMPLEX(-80, -50), MAKE_VAL_COMPLEX(-10, -73), MAKE_VAL_COMPLEX(106, -45), MAKE_VAL_COMPLEX(-24, -15), MAKE_VAL_COMPLEX(-36, -67), MAKE_VAL_COMPLEX(0, -89), MAKE_VAL_COMPLEX(-37, 77), MAKE_VAL_COMPLEX(13, 86), MAKE_VAL_COMPLEX(-10, -73), MAKE_VAL_COMPLEX(-82, -29), MAKE_VAL_COMPLEX(17, -45), MAKE_VAL_COMPLEX(39, 80), MAKE_VAL_COMPLEX(30, 130), MAKE_VAL_COMPLEX(-30, 48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_394, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_394", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_394() */
/* END TEST 394 */

/*
 *  BEGIN TEST 395
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_395( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(19, 3), MAKE_VAL_COMPLEX(70, 50), MAKE_VAL_COMPLEX(5, 67), MAKE_VAL_COMPLEX(-89, -834), MAKE_VAL_COMPLEX(1015, 48), MAKE_VAL_COMPLEX(-735, -482), MAKE_VAL_COMPLEX(565, 691), MAKE_VAL_COMPLEX(-317, 667), MAKE_VAL_COMPLEX(323, -589), MAKE_VAL_COMPLEX(415, -261), MAKE_VAL_COMPLEX(-43, 173), MAKE_VAL_COMPLEX(782, 220), MAKE_VAL_COMPLEX(-31, -75), MAKE_VAL_COMPLEX(-20, -120), MAKE_VAL_COMPLEX(30, -42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, -4);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_395, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_395", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_395() */
/* END TEST 395 */

/*
 *  BEGIN TEST 396
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_396( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	oski_value_t* y_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-12, -19), MAKE_VAL_COMPLEX(-156, 69), MAKE_VAL_COMPLEX(74, -97), MAKE_VAL_COMPLEX(61, -18), MAKE_VAL_COMPLEX(-16, 18), MAKE_VAL_COMPLEX(43, -3), MAKE_VAL_COMPLEX(-63, 69), MAKE_VAL_COMPLEX(-87, 104), MAKE_VAL_COMPLEX(6, -18), MAKE_VAL_COMPLEX(28, 86), MAKE_VAL_COMPLEX(-35, 95), MAKE_VAL_COMPLEX(-102, -153) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_396, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_396", y_true, y_val, y384_LEN*y384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_396() */
/* END TEST 396 */

/*
 *  BEGIN TEST 397
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_397( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	oski_value_t* y_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(38, -24), MAKE_VAL_COMPLEX(104, -83), MAKE_VAL_COMPLEX(-41, 128), MAKE_VAL_COMPLEX(-94, 37), MAKE_VAL_COMPLEX(-36, -32), MAKE_VAL_COMPLEX(18, 5), MAKE_VAL_COMPLEX(57, -11), MAKE_VAL_COMPLEX(32, -89), MAKE_VAL_COMPLEX(36, 12), MAKE_VAL_COMPLEX(-36, -117), MAKE_VAL_COMPLEX(40, -85), MAKE_VAL_COMPLEX(132, 113) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_397, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_397", y_true, y_val, y384_LEN*y384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_397() */
/* END TEST 397 */

/*
 *  BEGIN TEST 398
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_398( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	oski_value_t* y_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(62, -87), MAKE_VAL_COMPLEX(-108, -38), MAKE_VAL_COMPLEX(65, 71), MAKE_VAL_COMPLEX(-73, 35), MAKE_VAL_COMPLEX(-108, -38), MAKE_VAL_COMPLEX(129, 14), MAKE_VAL_COMPLEX(-22, 122), MAKE_VAL_COMPLEX(-119, 23), MAKE_VAL_COMPLEX(90, -6), MAKE_VAL_COMPLEX(-12, -67), MAKE_VAL_COMPLEX(9, 22), MAKE_VAL_COMPLEX(70, -80) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_398, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_398", y_true, y_val, y384_LEN*y384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_398() */
/* END TEST 398 */

/*
 *  BEGIN TEST 399
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_399( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	oski_value_t* y_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(58, -26), MAKE_VAL_COMPLEX(1318, -231), MAKE_VAL_COMPLEX(-663, 756), MAKE_VAL_COMPLEX(-658, 191), MAKE_VAL_COMPLEX(-44, -52), MAKE_VAL_COMPLEX(-166, -203), MAKE_VAL_COMPLEX(827, -269), MAKE_VAL_COMPLEX(828, -571), MAKE_VAL_COMPLEX(48, 24), MAKE_VAL_COMPLEX(-186, -923), MAKE_VAL_COMPLEX(548, -761), MAKE_VAL_COMPLEX(622, 1361) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-9, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_399, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_399", y_true, y_val, y384_LEN*y384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_399() */
/* END TEST 399 */

/*
 *  BEGIN TEST 400
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_400( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-5, -15), MAKE_VAL_COMPLEX(10, -70), MAKE_VAL_COMPLEX(47, -29), MAKE_VAL_COMPLEX(136, 24), MAKE_VAL_COMPLEX(30, 279), MAKE_VAL_COMPLEX(10, -157), MAKE_VAL_COMPLEX(156, -62), MAKE_VAL_COMPLEX(46, 68), MAKE_VAL_COMPLEX(-30, 35), MAKE_VAL_COMPLEX(133, -75), MAKE_VAL_COMPLEX(137, -99), MAKE_VAL_COMPLEX(-44, 98), MAKE_VAL_COMPLEX(-43, 51), MAKE_VAL_COMPLEX(-80, 60), MAKE_VAL_COMPLEX(-42, -6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_400, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_400", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_400() */
/* END TEST 400 */

/*
 *  BEGIN TEST 401
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_401( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(13, 6), MAKE_VAL_COMPLEX(40, 50), MAKE_VAL_COMPLEX(-10, 49), MAKE_VAL_COMPLEX(-151, 40), MAKE_VAL_COMPLEX(-15, -285), MAKE_VAL_COMPLEX(33, 163), MAKE_VAL_COMPLEX(-115, 92), MAKE_VAL_COMPLEX(-69, -111), MAKE_VAL_COMPLEX(-14, -58), MAKE_VAL_COMPLEX(-96, 95), MAKE_VAL_COMPLEX(-96, 71), MAKE_VAL_COMPLEX(59, -75), MAKE_VAL_COMPLEX(-7, -60), MAKE_VAL_COMPLEX(10, -90), MAKE_VAL_COMPLEX(30, -24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_401, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_401", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_401() */
/* END TEST 401 */

/*
 *  BEGIN TEST 402
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_402( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(15, 0), MAKE_VAL_COMPLEX(60, 30), MAKE_VAL_COMPLEX(12, 51), MAKE_VAL_COMPLEX(-72, 39), MAKE_VAL_COMPLEX(18, 9), MAKE_VAL_COMPLEX(30, 45), MAKE_VAL_COMPLEX(6, 63), MAKE_VAL_COMPLEX(21, -57), MAKE_VAL_COMPLEX(-15, -60), MAKE_VAL_COMPLEX(12, 51), MAKE_VAL_COMPLEX(60, 15), MAKE_VAL_COMPLEX(-9, 33), MAKE_VAL_COMPLEX(-33, -54), MAKE_VAL_COMPLEX(-30, -90), MAKE_VAL_COMPLEX(18, -36) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_402, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_402", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_402() */
/* END TEST 402 */

/*
 *  BEGIN TEST 403
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_403( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(9, -17), MAKE_VAL_COMPLEX(70, -50), MAKE_VAL_COMPLEX(65, 17), MAKE_VAL_COMPLEX(512, -140), MAKE_VAL_COMPLEX(210, 2085), MAKE_VAL_COMPLEX(-106, -791), MAKE_VAL_COMPLEX(740, -214), MAKE_VAL_COMPLEX(718, 432), MAKE_VAL_COMPLEX(98, -19), MAKE_VAL_COMPLEX(667, -305), MAKE_VAL_COMPLEX(1047, -297), MAKE_VAL_COMPLEX(-528, 730), MAKE_VAL_COMPLEX(-81, 5), MAKE_VAL_COMPLEX(-120, -20), MAKE_VAL_COMPLEX(-30, -42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_403, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_403", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_403() */
/* END TEST 403 */

/*
 *  BEGIN TEST 404
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_404( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(7, -11), MAKE_VAL_COMPLEX(50, -30), MAKE_VAL_COMPLEX(43, 15), MAKE_VAL_COMPLEX(8, 176), MAKE_VAL_COMPLEX(-278, -49), MAKE_VAL_COMPLEX(142, -13), MAKE_VAL_COMPLEX(134, 80), MAKE_VAL_COMPLEX(-22, 88), MAKE_VAL_COMPLEX(-36, 3), MAKE_VAL_COMPLEX(17, -79), MAKE_VAL_COMPLEX(37, -183), MAKE_VAL_COMPLEX(100, 122), MAKE_VAL_COMPLEX(-55, -1), MAKE_VAL_COMPLEX(-80, -20), MAKE_VAL_COMPLEX(-18, -30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_404, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_404", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_404() */
/* END TEST 404 */

/*
 *  BEGIN TEST 405
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_405( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(10, -20), MAKE_VAL_COMPLEX(80, -60), MAKE_VAL_COMPLEX(76, 18), MAKE_VAL_COMPLEX(-9, 17), MAKE_VAL_COMPLEX(317, 22), MAKE_VAL_COMPLEX(-15, 2), MAKE_VAL_COMPLEX(3, -21), MAKE_VAL_COMPLEX(-72, -196), MAKE_VAL_COMPLEX(-105, -40), MAKE_VAL_COMPLEX(102, 112), MAKE_VAL_COMPLEX(62, 76), MAKE_VAL_COMPLEX(-42, -66), MAKE_VAL_COMPLEX(-94, 8), MAKE_VAL_COMPLEX(-140, -20), MAKE_VAL_COMPLEX(-36, -48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_405, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_405", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_405() */
/* END TEST 405 */

/*
 *  BEGIN TEST 406
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_406( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-8, 4), MAKE_VAL_COMPLEX(-40, 0), MAKE_VAL_COMPLEX(-20, -24), MAKE_VAL_COMPLEX(28, -40), MAKE_VAL_COMPLEX(-12, 0), MAKE_VAL_COMPLEX(-28, -16), MAKE_VAL_COMPLEX(-20, -32), MAKE_VAL_COMPLEX(4, 36), MAKE_VAL_COMPLEX(24, 28), MAKE_VAL_COMPLEX(-20, -24), MAKE_VAL_COMPLEX(-36, 8), MAKE_VAL_COMPLEX(-4, -20), MAKE_VAL_COMPLEX(32, 20), MAKE_VAL_COMPLEX(40, 40), MAKE_VAL_COMPLEX(0, 24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_406, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_406", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_406() */
/* END TEST 406 */

/*
 *  BEGIN TEST 407
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_ROWMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_407( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_ROWMAJ, x384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-5, 15), MAKE_VAL_COMPLEX(-50, 50), MAKE_VAL_COMPLEX(-55, -5), MAKE_VAL_COMPLEX(524, -898), MAKE_VAL_COMPLEX(1796, 2053), MAKE_VAL_COMPLEX(-792, -471), MAKE_VAL_COMPLEX(-330, -910), MAKE_VAL_COMPLEX(760, -930), MAKE_VAL_COMPLEX(80, -225), MAKE_VAL_COMPLEX(-437, 809), MAKE_VAL_COMPLEX(-897, 1089), MAKE_VAL_COMPLEX(10, -1200), MAKE_VAL_COMPLEX(65, -15), MAKE_VAL_COMPLEX(100, 0), MAKE_VAL_COMPLEX(30, 30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_407, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_407", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_407() */
/* END TEST 407 */

/*
 *  BEGIN TEST 408
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_408( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	oski_value_t* y_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-42, -16), MAKE_VAL_COMPLEX(-4, 48), MAKE_VAL_COMPLEX(22, -34), MAKE_VAL_COMPLEX(122, 96), MAKE_VAL_COMPLEX(-10, 163), MAKE_VAL_COMPLEX(-33, 3), MAKE_VAL_COMPLEX(87, -84), MAKE_VAL_COMPLEX(-63, 101), MAKE_VAL_COMPLEX(-22, 69), MAKE_VAL_COMPLEX(-170, -228), MAKE_VAL_COMPLEX(148, -135), MAKE_VAL_COMPLEX(175, -99) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_408, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_408", y_true, y_val, y384_LEN*y384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_408() */
/* END TEST 408 */

/*
 *  BEGIN TEST 409
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_409( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	oski_value_t* y_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(15, -52), MAKE_VAL_COMPLEX(-58, 0), MAKE_VAL_COMPLEX(43, 23), MAKE_VAL_COMPLEX(-129, -41), MAKE_VAL_COMPLEX(-52, -115), MAKE_VAL_COMPLEX(89, -72), MAKE_VAL_COMPLEX(-25, 142), MAKE_VAL_COMPLEX(32, -24), MAKE_VAL_COMPLEX(52, -123), MAKE_VAL_COMPLEX(127, 210), MAKE_VAL_COMPLEX(-132, 138), MAKE_VAL_COMPLEX(-195, 29) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_409, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_409", y_true, y_val, y384_LEN*y384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_409() */
/* END TEST 409 */

/*
 *  BEGIN TEST 410
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_410( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	oski_value_t* y_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(33, 27), MAKE_VAL_COMPLEX(18, -42), MAKE_VAL_COMPLEX(-30, 24), MAKE_VAL_COMPLEX(-12, -30), MAKE_VAL_COMPLEX(18, -42), MAKE_VAL_COMPLEX(-9, 51), MAKE_VAL_COMPLEX(-48, -12), MAKE_VAL_COMPLEX(-6, -48), MAKE_VAL_COMPLEX(0, 36), MAKE_VAL_COMPLEX(27, -3), MAKE_VAL_COMPLEX(-9, 3), MAKE_VAL_COMPLEX(30, 30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_410, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_410", y_true, y_val, y384_LEN*y384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_410() */
/* END TEST 410 */

/*
 *  BEGIN TEST 411
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_411( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	oski_value_t* y_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(66, -47), MAKE_VAL_COMPLEX(-68, -54), MAKE_VAL_COMPLEX(29, 67), MAKE_VAL_COMPLEX(-471, -1323), MAKE_VAL_COMPLEX(791, -1047), MAKE_VAL_COMPLEX(713, -296), MAKE_VAL_COMPLEX(-921, 557), MAKE_VAL_COMPLEX(1059, 87), MAKE_VAL_COMPLEX(891, -857), MAKE_VAL_COMPLEX(-376, 3117), MAKE_VAL_COMPLEX(-2162, 199), MAKE_VAL_COMPLEX(-2418, -834) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-9, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_411, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_411", y_true, y_val, y384_LEN*y384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_411() */
/* END TEST 411 */

/*
 *  BEGIN TEST 412
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_412( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(4, -2), MAKE_VAL_COMPLEX(-114, -65), MAKE_VAL_COMPLEX(120, 82), MAKE_VAL_COMPLEX(4, 100), MAKE_VAL_COMPLEX(6, 0), MAKE_VAL_COMPLEX(14, 8), MAKE_VAL_COMPLEX(68, -123), MAKE_VAL_COMPLEX(-27, 107), MAKE_VAL_COMPLEX(-24, -40), MAKE_VAL_COMPLEX(10, 12), MAKE_VAL_COMPLEX(18, -4), MAKE_VAL_COMPLEX(-98, 69), MAKE_VAL_COMPLEX(14, -100), MAKE_VAL_COMPLEX(-100, 80), MAKE_VAL_COMPLEX(0, -12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_412, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_412", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_412() */
/* END TEST 412 */

/*
 *  BEGIN TEST 413
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_413( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(8, -9), MAKE_VAL_COMPLEX(184, 45), MAKE_VAL_COMPLEX(-73, -50), MAKE_VAL_COMPLEX(-33, -16), MAKE_VAL_COMPLEX(15, -6), MAKE_VAL_COMPLEX(43, 6), MAKE_VAL_COMPLEX(-17, 169), MAKE_VAL_COMPLEX(2, -168), MAKE_VAL_COMPLEX(-32, 3), MAKE_VAL_COMPLEX(37, 20), MAKE_VAL_COMPLEX(41, -28), MAKE_VAL_COMPLEX(115, -36), MAKE_VAL_COMPLEX(-80, 81), MAKE_VAL_COMPLEX(10, -130), MAKE_VAL_COMPLEX(-12, -30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_413, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_413", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_413() */
/* END TEST 413 */

/*
 *  BEGIN TEST 414
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_414( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(17, 14), MAKE_VAL_COMPLEX(40, 90), MAKE_VAL_COMPLEX(-34, 69), MAKE_VAL_COMPLEX(-118, -23), MAKE_VAL_COMPLEX(12, 27), MAKE_VAL_COMPLEX(-8, 79), MAKE_VAL_COMPLEX(-52, 77), MAKE_VAL_COMPLEX(77, -45), MAKE_VAL_COMPLEX(39, -82), MAKE_VAL_COMPLEX(-34, 69), MAKE_VAL_COMPLEX(54, 73), MAKE_VAL_COMPLEX(-41, 29), MAKE_VAL_COMPLEX(13, -92), MAKE_VAL_COMPLEX(50, -130), MAKE_VAL_COMPLEX(54, -24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_414, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_414", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_414() */
/* END TEST 414 */

/*
 *  BEGIN TEST 415
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_415( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(4, 8), MAKE_VAL_COMPLEX(-191, 775), MAKE_VAL_COMPLEX(216, -600), MAKE_VAL_COMPLEX(342, -198), MAKE_VAL_COMPLEX(0, 12), MAKE_VAL_COMPLEX(-16, 28), MAKE_VAL_COMPLEX(-785, -131), MAKE_VAL_COMPLEX(686, -4), MAKE_VAL_COMPLEX(-90, 62), MAKE_VAL_COMPLEX(-24, 20), MAKE_VAL_COMPLEX(8, 36), MAKE_VAL_COMPLEX(375, 445), MAKE_VAL_COMPLEX(-460, -92), MAKE_VAL_COMPLEX(620, 260), MAKE_VAL_COMPLEX(24, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_415, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_415", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_415() */
/* END TEST 415 */

/*
 *  BEGIN TEST 416
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_416( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(7, -26), MAKE_VAL_COMPLEX(166, -195), MAKE_VAL_COMPLEX(104, 133), MAKE_VAL_COMPLEX(116, 143), MAKE_VAL_COMPLEX(24, -27), MAKE_VAL_COMPLEX(92, -31), MAKE_VAL_COMPLEX(230, 120), MAKE_VAL_COMPLEX(-204, -8), MAKE_VAL_COMPLEX(-139, 4), MAKE_VAL_COMPLEX(94, 3), MAKE_VAL_COMPLEX(54, -97), MAKE_VAL_COMPLEX(13, -110), MAKE_VAL_COMPLEX(-19, 2), MAKE_VAL_COMPLEX(-90, 110), MAKE_VAL_COMPLEX(-54, -48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_416, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_416", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_416() */
/* END TEST 416 */

/*
 *  BEGIN TEST 417
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_417( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 5), MAKE_VAL_COMPLEX(-96, 125), MAKE_VAL_COMPLEX(-27, -126), MAKE_VAL_COMPLEX(-95, -24), MAKE_VAL_COMPLEX(-3, 6), MAKE_VAL_COMPLEX(-15, 10), MAKE_VAL_COMPLEX(-139, -99), MAKE_VAL_COMPLEX(134, -48), MAKE_VAL_COMPLEX(48, -11), MAKE_VAL_COMPLEX(-17, 4), MAKE_VAL_COMPLEX(-5, 20), MAKE_VAL_COMPLEX(29, 138), MAKE_VAL_COMPLEX(-72, 19), MAKE_VAL_COMPLEX(-50, -110), MAKE_VAL_COMPLEX(12, 6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_417, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_417", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_417() */
/* END TEST 417 */

/*
 *  BEGIN TEST 418
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_418( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-17, 6), MAKE_VAL_COMPLEX(-80, -10), MAKE_VAL_COMPLEX(-34, -53), MAKE_VAL_COMPLEX(66, -73), MAKE_VAL_COMPLEX(-24, -3), MAKE_VAL_COMPLEX(-52, -39), MAKE_VAL_COMPLEX(-32, -69), MAKE_VAL_COMPLEX(-1, 73), MAKE_VAL_COMPLEX(41, 62), MAKE_VAL_COMPLEX(-34, -53), MAKE_VAL_COMPLEX(-74, 7), MAKE_VAL_COMPLEX(-3, -41), MAKE_VAL_COMPLEX(59, 48), MAKE_VAL_COMPLEX(70, 90), MAKE_VAL_COMPLEX(-6, 48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_418, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_418", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_418() */
/* END TEST 418 */

/*
 *  BEGIN TEST 419
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_419( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_ROWMAJ, y384_VW );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-13, 4), MAKE_VAL_COMPLEX(331, 1012), MAKE_VAL_COMPLEX(-974, -491), MAKE_VAL_COMPLEX(-276, 521), MAKE_VAL_COMPLEX(-18, -3), MAKE_VAL_COMPLEX(-38, -31), MAKE_VAL_COMPLEX(-1201, 369), MAKE_VAL_COMPLEX(72, -970), MAKE_VAL_COMPLEX(99, -172), MAKE_VAL_COMPLEX(-24, -41), MAKE_VAL_COMPLEX(-56, 3), MAKE_VAL_COMPLEX(1146, 253), MAKE_VAL_COMPLEX(-107, 788), MAKE_VAL_COMPLEX(-970, 230), MAKE_VAL_COMPLEX(-6, 36) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_419, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_419", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_419() */
/* END TEST 419 */

/*
 *  BEGIN TEST 420
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_420( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	oski_value_t* y_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(34, -64), MAKE_VAL_COMPLEX(24, 54), MAKE_VAL_COMPLEX(85, -38), MAKE_VAL_COMPLEX(-190, -210), MAKE_VAL_COMPLEX(-76, -16), MAKE_VAL_COMPLEX(82, 113), MAKE_VAL_COMPLEX(-89, 139), MAKE_VAL_COMPLEX(62, -103), MAKE_VAL_COMPLEX(60, -12), MAKE_VAL_COMPLEX(-39, 13), MAKE_VAL_COMPLEX(-2, 119), MAKE_VAL_COMPLEX(255, -139) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_420, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_420", y_true, y_val, y384_LEN*y384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_420() */
/* END TEST 420 */

/*
 *  BEGIN TEST 421
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_421( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	oski_value_t* y_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-54, 66), MAKE_VAL_COMPLEX(-16, -34), MAKE_VAL_COMPLEX(-83, 20), MAKE_VAL_COMPLEX(204, 216), MAKE_VAL_COMPLEX(84, 36), MAKE_VAL_COMPLEX(-96, -133), MAKE_VAL_COMPLEX(109, -151), MAKE_VAL_COMPLEX(-44, 117), MAKE_VAL_COMPLEX(-72, 0), MAKE_VAL_COMPLEX(31, -3), MAKE_VAL_COMPLEX(4, -123), MAKE_VAL_COMPLEX(-275, 139) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_421, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_421", y_true, y_val, y384_LEN*y384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_421() */
/* END TEST 421 */

/*
 *  BEGIN TEST 422
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_422( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	oski_value_t* y_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-25, -48), MAKE_VAL_COMPLEX(-42, 40), MAKE_VAL_COMPLEX(47, -13), MAKE_VAL_COMPLEX(-1, 41), MAKE_VAL_COMPLEX(-42, 40), MAKE_VAL_COMPLEX(36, -55), MAKE_VAL_COMPLEX(50, 38), MAKE_VAL_COMPLEX(-17, 59), MAKE_VAL_COMPLEX(18, -42), MAKE_VAL_COMPLEX(-33, -10), MAKE_VAL_COMPLEX(12, 1), MAKE_VAL_COMPLEX(-20, -50) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_422, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_422", y_true, y_val, y384_LEN*y384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_422() */
/* END TEST 422 */

/*
 *  BEGIN TEST 423
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_423( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	oski_value_t* y_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(42, -85), MAKE_VAL_COMPLEX(-60, -288), MAKE_VAL_COMPLEX(-128, 63), MAKE_VAL_COMPLEX(-395, 569), MAKE_VAL_COMPLEX(-100, -18), MAKE_VAL_COMPLEX(351, -109), MAKE_VAL_COMPLEX(193, 225), MAKE_VAL_COMPLEX(-499, -114), MAKE_VAL_COMPLEX(78, -18), MAKE_VAL_COMPLEX(119, -64), MAKE_VAL_COMPLEX(231, -67), MAKE_VAL_COMPLEX(-323, -431) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_423, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_423", y_true, y_val, y384_LEN*y384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_423() */
/* END TEST 423 */

/*
 *  BEGIN TEST 424
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_424( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-11, -2), MAKE_VAL_COMPLEX(33, -65), MAKE_VAL_COMPLEX(93, -74), MAKE_VAL_COMPLEX(144, -65), MAKE_VAL_COMPLEX(-12, -9), MAKE_VAL_COMPLEX(-16, -37), MAKE_VAL_COMPLEX(31, 253), MAKE_VAL_COMPLEX(87, 109), MAKE_VAL_COMPLEX(145, 12), MAKE_VAL_COMPLEX(-2, -39), MAKE_VAL_COMPLEX(-42, -19), MAKE_VAL_COMPLEX(-14, -135), MAKE_VAL_COMPLEX(42, 44), MAKE_VAL_COMPLEX(-70, 170), MAKE_VAL_COMPLEX(-18, 24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_424, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_424", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_424() */
/* END TEST 424 */

/*
 *  BEGIN TEST 425
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_425( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-15, 0), MAKE_VAL_COMPLEX(-133, 5), MAKE_VAL_COMPLEX(-107, -16), MAKE_VAL_COMPLEX(-14, 7), MAKE_VAL_COMPLEX(-18, -9), MAKE_VAL_COMPLEX(-30, -45), MAKE_VAL_COMPLEX(-33, -363), MAKE_VAL_COMPLEX(-131, -13), MAKE_VAL_COMPLEX(-127, 94), MAKE_VAL_COMPLEX(-12, -51), MAKE_VAL_COMPLEX(-60, -15), MAKE_VAL_COMPLEX(34, 79), MAKE_VAL_COMPLEX(8, 54), MAKE_VAL_COMPLEX(110, -10), MAKE_VAL_COMPLEX(-18, 36) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_425, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_425", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_425() */
/* END TEST 425 */

/*
 *  BEGIN TEST 426
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_426( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-14, -8), MAKE_VAL_COMPLEX(-40, -60), MAKE_VAL_COMPLEX(16, -54), MAKE_VAL_COMPLEX(88, 2), MAKE_VAL_COMPLEX(-12, -18), MAKE_VAL_COMPLEX(-4, -58), MAKE_VAL_COMPLEX(28, -62), MAKE_VAL_COMPLEX(-50, 42), MAKE_VAL_COMPLEX(-18, 64), MAKE_VAL_COMPLEX(16, -54), MAKE_VAL_COMPLEX(-48, -46), MAKE_VAL_COMPLEX(26, -26), MAKE_VAL_COMPLEX(2, 68), MAKE_VAL_COMPLEX(-20, 100), MAKE_VAL_COMPLEX(-36, 24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_426, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_426", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_426() */
/* END TEST 426 */

/*
 *  BEGIN TEST 427
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_427( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(18, 1), MAKE_VAL_COMPLEX(193, -827), MAKE_VAL_COMPLEX(266, -1003), MAKE_VAL_COMPLEX(13, -1008), MAKE_VAL_COMPLEX(21, 12), MAKE_VAL_COMPLEX(33, 56), MAKE_VAL_COMPLEX(2865, 1633), MAKE_VAL_COMPLEX(1319, -637), MAKE_VAL_COMPLEX(532, -1555), MAKE_VAL_COMPLEX(11, 62), MAKE_VAL_COMPLEX(71, 22), MAKE_VAL_COMPLEX(-1171, -408), MAKE_VAL_COMPLEX(114, -292), MAKE_VAL_COMPLEX(390, 1210), MAKE_VAL_COMPLEX(24, -42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, -9);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_427, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_427", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_427() */
/* END TEST 427 */

/*
 *  BEGIN TEST 428
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_428( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-7, 21), MAKE_VAL_COMPLEX(-57, 175), MAKE_VAL_COMPLEX(8, 48), MAKE_VAL_COMPLEX(-47, -213), MAKE_VAL_COMPLEX(-21, 21), MAKE_VAL_COMPLEX(-77, 21), MAKE_VAL_COMPLEX(-384, -61), MAKE_VAL_COMPLEX(80, 186), MAKE_VAL_COMPLEX(89, -139), MAKE_VAL_COMPLEX(-77, -7), MAKE_VAL_COMPLEX(-49, 77), MAKE_VAL_COMPLEX(53, -40), MAKE_VAL_COMPLEX(106, -1), MAKE_VAL_COMPLEX(220, 100), MAKE_VAL_COMPLEX(42, 42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_428, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_428", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_428() */
/* END TEST 428 */

/*
 *  BEGIN TEST 429
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_429( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-2, 6), MAKE_VAL_COMPLEX(-33, -85), MAKE_VAL_COMPLEX(-107, -57), MAKE_VAL_COMPLEX(20, 60), MAKE_VAL_COMPLEX(-6, 6), MAKE_VAL_COMPLEX(-22, 6), MAKE_VAL_COMPLEX(267, 34), MAKE_VAL_COMPLEX(10, -114), MAKE_VAL_COMPLEX(28, 148), MAKE_VAL_COMPLEX(-22, -2), MAKE_VAL_COMPLEX(-14, 22), MAKE_VAL_COMPLEX(-107, 4), MAKE_VAL_COMPLEX(11, -26), MAKE_VAL_COMPLEX(-40, -100), MAKE_VAL_COMPLEX(12, 12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_429, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_429", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_429() */
/* END TEST 429 */

/*
 *  BEGIN TEST 430
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_430( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(17, -11), MAKE_VAL_COMPLEX(90, -10), MAKE_VAL_COMPLEX(51, 49), MAKE_VAL_COMPLEX(-53, 97), MAKE_VAL_COMPLEX(27, -3), MAKE_VAL_COMPLEX(67, 29), MAKE_VAL_COMPLEX(53, 67), MAKE_VAL_COMPLEX(-18, -80), MAKE_VAL_COMPLEX(-61, -57), MAKE_VAL_COMPLEX(51, 49), MAKE_VAL_COMPLEX(79, -27), MAKE_VAL_COMPLEX(14, 44), MAKE_VAL_COMPLEX(-77, -37), MAKE_VAL_COMPLEX(-100, -80), MAKE_VAL_COMPLEX(-6, -54) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_430, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_430", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_430() */
/* END TEST 430 */

/*
 *  BEGIN TEST 431
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x384', LAYOUT_COLMAJ>
 *  y : <'y384', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_431( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A384_ptr, A384_ind, A384_val,
		A384_M, A384_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y384_val,
		y384_LEN * y384_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y384_LEN, y384_VW, LAYOUT_COLMAJ, y384_LEN );
	oski_value_t* y_val = array_Duplicate( x384_val,
		x384_LEN * x384_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x384_LEN, x384_VW, LAYOUT_COLMAJ, x384_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-4, -8), MAKE_VAL_COMPLEX(170, 472), MAKE_VAL_COMPLEX(504, 170), MAKE_VAL_COMPLEX(-184, -416), MAKE_VAL_COMPLEX(0, -12), MAKE_VAL_COMPLEX(16, -28), MAKE_VAL_COMPLEX(-1473, 73), MAKE_VAL_COMPLEX(144, 644), MAKE_VAL_COMPLEX(-184, -704), MAKE_VAL_COMPLEX(24, -20), MAKE_VAL_COMPLEX(-8, -36), MAKE_VAL_COMPLEX(483, -159), MAKE_VAL_COMPLEX(75, 117), MAKE_VAL_COMPLEX(460, 460), MAKE_VAL_COMPLEX(-24, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_431, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_431", y_true, y_val, x384_LEN*x384_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_431() */
/* END TEST 431 */

#define A432_M 4    /* # of rows */
#define A432_N 5    /* # of cols */
#define A432_NNZ 4  /* # of stored non-zeros */
static oski_index_t A432_ptr[A432_M+1] = {0, 0, 1, 3, 4}; /* end row pointers for A432 */
static oski_index_t A432_ind[A432_NNZ] = {0, 0, 2, 3}; /* end column indices pointers for A432 */
static oski_value_t A432_val[A432_NNZ] = {MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(10, -4), MAKE_VAL_COMPLEX(-5, -3), MAKE_VAL_COMPLEX(7, 10)}; /* end stored non-zero values for A432 */

#define x432_LEN 5   /* vector length */
#define x432_VW  3   /* vector width */
static oski_value_t x432_val[x432_LEN*x432_VW] = {MAKE_VAL_COMPLEX(-3, 1), MAKE_VAL_COMPLEX(1, -1), MAKE_VAL_COMPLEX(1, -5), MAKE_VAL_COMPLEX(9, 10), MAKE_VAL_COMPLEX(2, -3), MAKE_VAL_COMPLEX(10, 1), MAKE_VAL_COMPLEX(5, -5), MAKE_VAL_COMPLEX(-3, -10), MAKE_VAL_COMPLEX(6, -3), MAKE_VAL_COMPLEX(-10, 5), MAKE_VAL_COMPLEX(-8, 4), MAKE_VAL_COMPLEX(0, 5), MAKE_VAL_COMPLEX(0, 7), MAKE_VAL_COMPLEX(2, 2), MAKE_VAL_COMPLEX(4, 2)}; /* data */

#define y432_LEN 4   /* vector length */
#define y432_VW  3   /* vector width */
static oski_value_t y432_val[y432_LEN*y432_VW] = {MAKE_VAL_COMPLEX(5, -2), MAKE_VAL_COMPLEX(1, -5), MAKE_VAL_COMPLEX(-7, -9), MAKE_VAL_COMPLEX(-4, 0), MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(10, 10), MAKE_VAL_COMPLEX(1, 6), MAKE_VAL_COMPLEX(-3, 1), MAKE_VAL_COMPLEX(-7, -8), MAKE_VAL_COMPLEX(6, 5), MAKE_VAL_COMPLEX(7, -8), MAKE_VAL_COMPLEX(-8, 0)}; /* data */

/*
 *  BEGIN TEST 432
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_432( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	oski_value_t* y_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-20, -21), MAKE_VAL_COMPLEX(-27, 5), MAKE_VAL_COMPLEX(-31, 53), MAKE_VAL_COMPLEX(48, 40), MAKE_VAL_COMPLEX(-50, 70), MAKE_VAL_COMPLEX(-30, -30), MAKE_VAL_COMPLEX(-38, 15), MAKE_VAL_COMPLEX(2, 58), MAKE_VAL_COMPLEX(-75, -6), MAKE_VAL_COMPLEX(-107, -105), MAKE_VAL_COMPLEX(-150, -71), MAKE_VAL_COMPLEX(-34, 75) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_432, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_432", y_true, y_val, y432_LEN*y432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_432() */
/* END TEST 432 */

/*
 *  BEGIN TEST 433
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_433( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	oski_value_t* y_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(47, 16), MAKE_VAL_COMPLEX(37, -29), MAKE_VAL_COMPLEX(5, -105), MAKE_VAL_COMPLEX(-68, -44), MAKE_VAL_COMPLEX(10, -130), MAKE_VAL_COMPLEX(70, 90), MAKE_VAL_COMPLEX(37, 16), MAKE_VAL_COMPLEX(-18, -56), MAKE_VAL_COMPLEX(48, -41), MAKE_VAL_COMPLEX(132, 136), MAKE_VAL_COMPLEX(193, 38), MAKE_VAL_COMPLEX(-6, -83) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_433, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_433", y_true, y_val, y432_LEN*y432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_433() */
/* END TEST 433 */

/*
 *  BEGIN TEST 434
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_434( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	oski_value_t* y_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-14, -6), MAKE_VAL_COMPLEX(-12, 8), MAKE_VAL_COMPLEX(-4, 32), MAKE_VAL_COMPLEX(8, 8), MAKE_VAL_COMPLEX(0, 40), MAKE_VAL_COMPLEX(0, -40), MAKE_VAL_COMPLEX(10, -14), MAKE_VAL_COMPLEX(8, 4), MAKE_VAL_COMPLEX(-2, 30), MAKE_VAL_COMPLEX(-2, -22), MAKE_VAL_COMPLEX(-30, 2), MAKE_VAL_COMPLEX(16, 16) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_434, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_434", y_true, y_val, y432_LEN*y432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_434() */
/* END TEST 434 */

/*
 *  BEGIN TEST 435
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_435( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	oski_value_t* y_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(61, 22), MAKE_VAL_COMPLEX(49, -37), MAKE_VAL_COMPLEX(9, -137), MAKE_VAL_COMPLEX(-176, -352), MAKE_VAL_COMPLEX(110, -70), MAKE_VAL_COMPLEX(570, 230), MAKE_VAL_COMPLEX(517, 200), MAKE_VAL_COMPLEX(244, -240), MAKE_VAL_COMPLEX(10, 459), MAKE_VAL_COMPLEX(409, 1083), MAKE_VAL_COMPLEX(443, 776), MAKE_VAL_COMPLEX(403, -24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_435, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_435", y_true, y_val, y432_LEN*y432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_435() */
/* END TEST 435 */

/*
 *  BEGIN TEST 436
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_436( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(53, 73), MAKE_VAL_COMPLEX(-13, 227), MAKE_VAL_COMPLEX(-53, -263), MAKE_VAL_COMPLEX(-54, 121), MAKE_VAL_COMPLEX(35, 6), MAKE_VAL_COMPLEX(31, 94), MAKE_VAL_COMPLEX(78, -8), MAKE_VAL_COMPLEX(96, -63), MAKE_VAL_COMPLEX(62, 103), MAKE_VAL_COMPLEX(-93, 25), MAKE_VAL_COMPLEX(61, -42), MAKE_VAL_COMPLEX(-101, -60), MAKE_VAL_COMPLEX(-63, 28), MAKE_VAL_COMPLEX(-10, 26), MAKE_VAL_COMPLEX(-2, 44) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_436, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_436", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_436() */
/* END TEST 436 */

/*
 *  BEGIN TEST 437
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_437( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-86, -92), MAKE_VAL_COMPLEX(30, -226), MAKE_VAL_COMPLEX(106, 232), MAKE_VAL_COMPLEX(36, 40), MAKE_VAL_COMPLEX(8, -12), MAKE_VAL_COMPLEX(40, 4), MAKE_VAL_COMPLEX(7, 13), MAKE_VAL_COMPLEX(-30, -44), MAKE_VAL_COMPLEX(13, -73), MAKE_VAL_COMPLEX(-32, -75), MAKE_VAL_COMPLEX(-161, 2), MAKE_VAL_COMPLEX(56, 100), MAKE_VAL_COMPLEX(0, 28), MAKE_VAL_COMPLEX(8, 8), MAKE_VAL_COMPLEX(16, 8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_437, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_437", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_437() */
/* END TEST 437 */

/*
 *  BEGIN TEST 438
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_438( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1, -3), MAKE_VAL_COMPLEX(1, 1), MAKE_VAL_COMPLEX(5, 1), MAKE_VAL_COMPLEX(-10, 9), MAKE_VAL_COMPLEX(3, 2), MAKE_VAL_COMPLEX(-1, 10), MAKE_VAL_COMPLEX(5, 5), MAKE_VAL_COMPLEX(10, -3), MAKE_VAL_COMPLEX(3, 6), MAKE_VAL_COMPLEX(-5, -10), MAKE_VAL_COMPLEX(-4, -8), MAKE_VAL_COMPLEX(-5, 0), MAKE_VAL_COMPLEX(-7, 0), MAKE_VAL_COMPLEX(-2, 2), MAKE_VAL_COMPLEX(-2, 4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_438, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_438", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_438() */
/* END TEST 438 */

/*
 *  BEGIN TEST 439
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_439( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-184, 328), MAKE_VAL_COMPLEX(-706, 146), MAKE_VAL_COMPLEX(616, -524), MAKE_VAL_COMPLEX(-12, -134), MAKE_VAL_COMPLEX(-34, 12), MAKE_VAL_COMPLEX(-74, -68), MAKE_VAL_COMPLEX(42, 16), MAKE_VAL_COMPLEX(-30, 156), MAKE_VAL_COMPLEX(-238, 82), MAKE_VAL_COMPLEX(-183, 91), MAKE_VAL_COMPLEX(175, 417), MAKE_VAL_COMPLEX(214, -288), MAKE_VAL_COMPLEX(42, -56), MAKE_VAL_COMPLEX(-4, -28), MAKE_VAL_COMPLEX(-20, -40) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_439, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_439", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_439() */
/* END TEST 439 */

/*
 *  BEGIN TEST 440
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_440( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(40, -4), MAKE_VAL_COMPLEX(166, 12), MAKE_VAL_COMPLEX(-210, -66), MAKE_VAL_COMPLEX(-133, -7), MAKE_VAL_COMPLEX(7, 35), MAKE_VAL_COMPLEX(-77, 63), MAKE_VAL_COMPLEX(-23, 43), MAKE_VAL_COMPLEX(103, 35), MAKE_VAL_COMPLEX(38, 82), MAKE_VAL_COMPLEX(127, -130), MAKE_VAL_COMPLEX(-3, -210), MAKE_VAL_COMPLEX(-91, 45), MAKE_VAL_COMPLEX(-49, -49), MAKE_VAL_COMPLEX(-28, 0), MAKE_VAL_COMPLEX(-42, 14) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_440, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_440", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_440() */
/* END TEST 440 */

/*
 *  BEGIN TEST 441
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_441( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-53, -45), MAKE_VAL_COMPLEX(-151, 5), MAKE_VAL_COMPLEX(289, 87), MAKE_VAL_COMPLEX(-36, 141), MAKE_VAL_COMPLEX(39, 0), MAKE_VAL_COMPLEX(51, 96), MAKE_VAL_COMPLEX(98, 42), MAKE_VAL_COMPLEX(60, -73), MAKE_VAL_COMPLEX(4, 17), MAKE_VAL_COMPLEX(-197, -35), MAKE_VAL_COMPLEX(-53, 78), MAKE_VAL_COMPLEX(11, -50), MAKE_VAL_COMPLEX(-63, 42), MAKE_VAL_COMPLEX(-6, 30), MAKE_VAL_COMPLEX(6, 48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_441, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_441", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_441() */
/* END TEST 441 */

/*
 *  BEGIN TEST 442
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_442( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-9, 3), MAKE_VAL_COMPLEX(3, -3), MAKE_VAL_COMPLEX(3, -15), MAKE_VAL_COMPLEX(27, 30), MAKE_VAL_COMPLEX(6, -9), MAKE_VAL_COMPLEX(30, 3), MAKE_VAL_COMPLEX(15, -15), MAKE_VAL_COMPLEX(-9, -30), MAKE_VAL_COMPLEX(18, -9), MAKE_VAL_COMPLEX(-30, 15), MAKE_VAL_COMPLEX(-24, 12), MAKE_VAL_COMPLEX(0, 15), MAKE_VAL_COMPLEX(0, 21), MAKE_VAL_COMPLEX(6, 6), MAKE_VAL_COMPLEX(12, 6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_442, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_442", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_442() */
/* END TEST 442 */

/*
 *  BEGIN TEST 443
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_443( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-177, 107), MAKE_VAL_COMPLEX(-325, 823), MAKE_VAL_COMPLEX(985, -1001), MAKE_VAL_COMPLEX(106, -23), MAKE_VAL_COMPLEX(-13, -26), MAKE_VAL_COMPLEX(47, -66), MAKE_VAL_COMPLEX(166, -116), MAKE_VAL_COMPLEX(-36, 69), MAKE_VAL_COMPLEX(-210, 203), MAKE_VAL_COMPLEX(-64, 600), MAKE_VAL_COMPLEX(688, 169), MAKE_VAL_COMPLEX(-253, -420), MAKE_VAL_COMPLEX(49, 28), MAKE_VAL_COMPLEX(22, -6), MAKE_VAL_COMPLEX(30, -20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-2, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_443, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_443", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_443() */
/* END TEST 443 */

/*
 *  BEGIN TEST 444
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_444( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	oski_value_t* y_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(15, -35), MAKE_VAL_COMPLEX(20, -10), MAKE_VAL_COMPLEX(-146, 22), MAKE_VAL_COMPLEX(-140, -45), MAKE_VAL_COMPLEX(-100, 0), MAKE_VAL_COMPLEX(80, 0), MAKE_VAL_COMPLEX(26, 70), MAKE_VAL_COMPLEX(-106, -32), MAKE_VAL_COMPLEX(-75, -5), MAKE_VAL_COMPLEX(-5, 35), MAKE_VAL_COMPLEX(-54, -132), MAKE_VAL_COMPLEX(-90, 75) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_444, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_444", y_true, y_val, y432_LEN*y432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_444() */
/* END TEST 444 */

/*
 *  BEGIN TEST 445
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_445( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	oski_value_t* y_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(20, -37), MAKE_VAL_COMPLEX(-59, -55), MAKE_VAL_COMPLEX(-21, -51), MAKE_VAL_COMPLEX(96, 85), MAKE_VAL_COMPLEX(-110, -10), MAKE_VAL_COMPLEX(130, 10), MAKE_VAL_COMPLEX(45, -14), MAKE_VAL_COMPLEX(83, 73), MAKE_VAL_COMPLEX(-82, -13), MAKE_VAL_COMPLEX(121, -40), MAKE_VAL_COMPLEX(51, -26), MAKE_VAL_COMPLEX(2, 5) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_445, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_445", y_true, y_val, y432_LEN*y432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_445() */
/* END TEST 445 */

/*
 *  BEGIN TEST 446
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_446( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	oski_value_t* y_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(28, -46), MAKE_VAL_COMPLEX(-22, -46), MAKE_VAL_COMPLEX(-110, -30), MAKE_VAL_COMPLEX(-32, 24), MAKE_VAL_COMPLEX(-140, -20), MAKE_VAL_COMPLEX(140, 20), MAKE_VAL_COMPLEX(44, 42), MAKE_VAL_COMPLEX(-18, 26), MAKE_VAL_COMPLEX(-104, -22), MAKE_VAL_COMPLEX(78, 4), MAKE_VAL_COMPLEX(8, -106), MAKE_VAL_COMPLEX(-64, 48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_446, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_446", y_true, y_val, y432_LEN*y432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_446() */
/* END TEST 446 */

/*
 *  BEGIN TEST 447
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_447( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	oski_value_t* y_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(21, 38), MAKE_VAL_COMPLEX(-479, 143), MAKE_VAL_COMPLEX(403, -881), MAKE_VAL_COMPLEX(1596, -407), MAKE_VAL_COMPLEX(70, -90), MAKE_VAL_COMPLEX(110, -70), MAKE_VAL_COMPLEX(-326, -463), MAKE_VAL_COMPLEX(1269, -323), MAKE_VAL_COMPLEX(57, -64), MAKE_VAL_COMPLEX(186, -787), MAKE_VAL_COMPLEX(968, 169), MAKE_VAL_COMPLEX(162, -779) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-9, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_447, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_447", y_true, y_val, y432_LEN*y432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_447() */
/* END TEST 447 */

/*
 *  BEGIN TEST 448
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_448( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-152, -40), MAKE_VAL_COMPLEX(32, -134), MAKE_VAL_COMPLEX(42, -184), MAKE_VAL_COMPLEX(-94, -24), MAKE_VAL_COMPLEX(0, 26), MAKE_VAL_COMPLEX(-64, 34), MAKE_VAL_COMPLEX(-2, 116), MAKE_VAL_COMPLEX(71, 15), MAKE_VAL_COMPLEX(-83, 61), MAKE_VAL_COMPLEX(12, -110), MAKE_VAL_COMPLEX(1, -79), MAKE_VAL_COMPLEX(-76, -110), MAKE_VAL_COMPLEX(-28, -42), MAKE_VAL_COMPLEX(-20, -4), MAKE_VAL_COMPLEX(-32, 4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_448, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_448", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_448() */
/* END TEST 448 */

/*
 *  BEGIN TEST 449
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_449( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(156, 52), MAKE_VAL_COMPLEX(-36, 130), MAKE_VAL_COMPLEX(-62, 180), MAKE_VAL_COMPLEX(134, -12), MAKE_VAL_COMPLEX(-12, -34), MAKE_VAL_COMPLEX(68, -74), MAKE_VAL_COMPLEX(-18, -136), MAKE_VAL_COMPLEX(-111, -3), MAKE_VAL_COMPLEX(71, -85), MAKE_VAL_COMPLEX(8, 150), MAKE_VAL_COMPLEX(15, 111), MAKE_VAL_COMPLEX(96, 110), MAKE_VAL_COMPLEX(56, 42), MAKE_VAL_COMPLEX(28, -4), MAKE_VAL_COMPLEX(40, -20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_449, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_449", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_449() */
/* END TEST 449 */

/*
 *  BEGIN TEST 450
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_450( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(17, -19), MAKE_VAL_COMPLEX(-3, 11), MAKE_VAL_COMPLEX(13, 39), MAKE_VAL_COMPLEX(-103, -34), MAKE_VAL_COMPLEX(-2, 29), MAKE_VAL_COMPLEX(-74, 33), MAKE_VAL_COMPLEX(-15, 55), MAKE_VAL_COMPLEX(61, 58), MAKE_VAL_COMPLEX(-30, 45), MAKE_VAL_COMPLEX(50, -75), MAKE_VAL_COMPLEX(40, -60), MAKE_VAL_COMPLEX(-20, -35), MAKE_VAL_COMPLEX(-28, -49), MAKE_VAL_COMPLEX(-22, -6), MAKE_VAL_COMPLEX(-36, 2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_450, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_450", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_450() */
/* END TEST 450 */

/*
 *  BEGIN TEST 451
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_451( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1005, 311), MAKE_VAL_COMPLEX(-363, 831), MAKE_VAL_COMPLEX(-429, 1313), MAKE_VAL_COMPLEX(-2, -143), MAKE_VAL_COMPLEX(-37, 10), MAKE_VAL_COMPLEX(-73, -78), MAKE_VAL_COMPLEX(-57, -399), MAKE_VAL_COMPLEX(-157, 286), MAKE_VAL_COMPLEX(304, -73), MAKE_VAL_COMPLEX(243, 298), MAKE_VAL_COMPLEX(255, 193), MAKE_VAL_COMPLEX(291, 496), MAKE_VAL_COMPLEX(49, -56), MAKE_VAL_COMPLEX(-2, -30), MAKE_VAL_COMPLEX(-18, -44) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_451, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_451", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_451() */
/* END TEST 451 */

/*
 *  BEGIN TEST 452
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_452( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-2, -82), MAKE_VAL_COMPLEX(-206, 72), MAKE_VAL_COMPLEX(32, -34), MAKE_VAL_COMPLEX(-80, 72), MAKE_VAL_COMPLEX(24, 16), MAKE_VAL_COMPLEX(-8, 80), MAKE_VAL_COMPLEX(102, 64), MAKE_VAL_COMPLEX(57, -51), MAKE_VAL_COMPLEX(13, 109), MAKE_VAL_COMPLEX(-68, -40), MAKE_VAL_COMPLEX(-43, -27), MAKE_VAL_COMPLEX(-96, 80), MAKE_VAL_COMPLEX(-56, 0), MAKE_VAL_COMPLEX(-16, 16), MAKE_VAL_COMPLEX(-16, 32) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_452, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_452", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_452() */
/* END TEST 452 */

/*
 *  BEGIN TEST 453
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_453( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-17, 45), MAKE_VAL_COMPLEX(221, -61), MAKE_VAL_COMPLEX(35, 37), MAKE_VAL_COMPLEX(-32, 65), MAKE_VAL_COMPLEX(19, 4), MAKE_VAL_COMPLEX(15, 52), MAKE_VAL_COMPLEX(-27, -9), MAKE_VAL_COMPLEX(67, -8), MAKE_VAL_COMPLEX(38, -37), MAKE_VAL_COMPLEX(-17, -80), MAKE_VAL_COMPLEX(-25, -69), MAKE_VAL_COMPLEX(31, -70), MAKE_VAL_COMPLEX(-35, 14), MAKE_VAL_COMPLEX(-6, 14), MAKE_VAL_COMPLEX(-2, 24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_453, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_453", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_453() */
/* END TEST 453 */

/*
 *  BEGIN TEST 454
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_454( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(12, -34), MAKE_VAL_COMPLEX(2, 16), MAKE_VAL_COMPLEX(38, 44), MAKE_VAL_COMPLEX(-153, 11), MAKE_VAL_COMPLEX(13, 39), MAKE_VAL_COMPLEX(-79, 83), MAKE_VAL_COMPLEX(10, 80), MAKE_VAL_COMPLEX(111, 43), MAKE_VAL_COMPLEX(-15, 75), MAKE_VAL_COMPLEX(25, -125), MAKE_VAL_COMPLEX(20, -100), MAKE_VAL_COMPLEX(-45, -35), MAKE_VAL_COMPLEX(-63, -49), MAKE_VAL_COMPLEX(-32, 4), MAKE_VAL_COMPLEX(-46, 22) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_454, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_454", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_454() */
/* END TEST 454 */

/*
 *  BEGIN TEST 455
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_ROWMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_455( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_ROWMAJ, x432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-60, -558), MAKE_VAL_COMPLEX(-1794, 1014), MAKE_VAL_COMPLEX(-124, -340), MAKE_VAL_COMPLEX(-97, 33), MAKE_VAL_COMPLEX(15, 23), MAKE_VAL_COMPLEX(-37, 67), MAKE_VAL_COMPLEX(626, 142), MAKE_VAL_COMPLEX(-182, -188), MAKE_VAL_COMPLEX(26, 622), MAKE_VAL_COMPLEX(-177, 331), MAKE_VAL_COMPLEX(-29, 287), MAKE_VAL_COMPLEX(-379, 817), MAKE_VAL_COMPLEX(-49, -21), MAKE_VAL_COMPLEX(-20, 8), MAKE_VAL_COMPLEX(-26, 22) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(9, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_455, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_455", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_455() */
/* END TEST 455 */

/*
 *  BEGIN TEST 456
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_456( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	oski_value_t* y_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(43, 35), MAKE_VAL_COMPLEX(50, -16), MAKE_VAL_COMPLEX(46, -108), MAKE_VAL_COMPLEX(20, -16), MAKE_VAL_COMPLEX(-50, -250), MAKE_VAL_COMPLEX(80, 180), MAKE_VAL_COMPLEX(-95, 83), MAKE_VAL_COMPLEX(65, 7), MAKE_VAL_COMPLEX(-6, -66), MAKE_VAL_COMPLEX(-52, 239), MAKE_VAL_COMPLEX(179, 62), MAKE_VAL_COMPLEX(-46, -38) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_456, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_456", y_true, y_val, y432_LEN*y432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_456() */
/* END TEST 456 */

/*
 *  BEGIN TEST 457
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_457( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	oski_value_t* y_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-38, 21), MAKE_VAL_COMPLEX(-3, 41), MAKE_VAL_COMPLEX(65, 65), MAKE_VAL_COMPLEX(-8, -24), MAKE_VAL_COMPLEX(180, 180), MAKE_VAL_COMPLEX(-210, -110), MAKE_VAL_COMPLEX(32, -91), MAKE_VAL_COMPLEX(-66, -40), MAKE_VAL_COMPLEX(107, 20), MAKE_VAL_COMPLEX(-16, -194), MAKE_VAL_COMPLEX(-120, 32), MAKE_VAL_COMPLEX(70, -42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_457, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_457", y_true, y_val, y432_LEN*y432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_457() */
/* END TEST 457 */

/*
 *  BEGIN TEST 458
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_458( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	oski_value_t* y_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(25, -39), MAKE_VAL_COMPLEX(-18, -40), MAKE_VAL_COMPLEX(-94, -28), MAKE_VAL_COMPLEX(-28, 20), MAKE_VAL_COMPLEX(-120, -20), MAKE_VAL_COMPLEX(120, 20), MAKE_VAL_COMPLEX(37, 37), MAKE_VAL_COMPLEX(-16, 22), MAKE_VAL_COMPLEX(-89, -21), MAKE_VAL_COMPLEX(67, 5), MAKE_VAL_COMPLEX(9, -91), MAKE_VAL_COMPLEX(-56, 40) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_458, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_458", y_true, y_val, y432_LEN*y432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_458() */
/* END TEST 458 */

/*
 *  BEGIN TEST 459
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_459( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	oski_value_t* y_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-2, 53), MAKE_VAL_COMPLEX(41, 29), MAKE_VAL_COMPLEX(109, -27), MAKE_VAL_COMPLEX(-184, -236), MAKE_VAL_COMPLEX(450, 790), MAKE_VAL_COMPLEX(-770, -430), MAKE_VAL_COMPLEX(306, -187), MAKE_VAL_COMPLEX(-473, -383), MAKE_VAL_COMPLEX(432, -167), MAKE_VAL_COMPLEX(473, -852), MAKE_VAL_COMPLEX(-310, -283), MAKE_VAL_COMPLEX(136, -264) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_459, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_459", y_true, y_val, y432_LEN*y432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_459() */
/* END TEST 459 */

/*
 *  BEGIN TEST 460
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_460( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(85, 119), MAKE_VAL_COMPLEX(-9, -7), MAKE_VAL_COMPLEX(-28, -36), MAKE_VAL_COMPLEX(63, 13), MAKE_VAL_COMPLEX(-26, -13), MAKE_VAL_COMPLEX(-28, 141), MAKE_VAL_COMPLEX(-45, -35), MAKE_VAL_COMPLEX(-59, 38), MAKE_VAL_COMPLEX(99, -31), MAKE_VAL_COMPLEX(50, 75), MAKE_VAL_COMPLEX(-62, -192), MAKE_VAL_COMPLEX(40, -5), MAKE_VAL_COMPLEX(67, 54), MAKE_VAL_COMPLEX(-42, -98), MAKE_VAL_COMPLEX(12, -34) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_460, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_460", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_460() */
/* END TEST 460 */

/*
 *  BEGIN TEST 461
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_461( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-78, -78), MAKE_VAL_COMPLEX(-2, -8), MAKE_VAL_COMPLEX(-35, 13), MAKE_VAL_COMPLEX(85, -110), MAKE_VAL_COMPLEX(-9, -19), MAKE_VAL_COMPLEX(61, -269), MAKE_VAL_COMPLEX(-10, -40), MAKE_VAL_COMPLEX(-77, -19), MAKE_VAL_COMPLEX(-126, -53), MAKE_VAL_COMPLEX(-5, 65), MAKE_VAL_COMPLEX(98, 304), MAKE_VAL_COMPLEX(25, 15), MAKE_VAL_COMPLEX(24, -40), MAKE_VAL_COMPLEX(72, 76), MAKE_VAL_COMPLEX(22, -14) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_461, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_461", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_461() */
/* END TEST 461 */

/*
 *  BEGIN TEST 462
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_462( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(5, -25), MAKE_VAL_COMPLEX(3, 11), MAKE_VAL_COMPLEX(31, 27), MAKE_VAL_COMPLEX(-106, 23), MAKE_VAL_COMPLEX(13, 26), MAKE_VAL_COMPLEX(-47, 66), MAKE_VAL_COMPLEX(15, 55), MAKE_VAL_COMPLEX(82, 19), MAKE_VAL_COMPLEX(-3, 54), MAKE_VAL_COMPLEX(5, -90), MAKE_VAL_COMPLEX(4, -72), MAKE_VAL_COMPLEX(-35, -20), MAKE_VAL_COMPLEX(-49, -28), MAKE_VAL_COMPLEX(-22, 6), MAKE_VAL_COMPLEX(-30, 20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_462, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_462", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_462() */
/* END TEST 462 */

/*
 *  BEGIN TEST 463
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_463( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-906, -192), MAKE_VAL_COMPLEX(-10, -2), MAKE_VAL_COMPLEX(53, 277), MAKE_VAL_COMPLEX(-403, -704), MAKE_VAL_COMPLEX(-26, 0), MAKE_VAL_COMPLEX(-988, -1526), MAKE_VAL_COMPLEX(-50, -10), MAKE_VAL_COMPLEX(-176, 124), MAKE_VAL_COMPLEX(-886, 537), MAKE_VAL_COMPLEX(70, 40), MAKE_VAL_COMPLEX(1928, 1034), MAKE_VAL_COMPLEX(30, -20), MAKE_VAL_COMPLEX(-329, -339), MAKE_VAL_COMPLEX(740, 180), MAKE_VAL_COMPLEX(-4, -32) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_463, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_463", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_463() */
/* END TEST 463 */

/*
 *  BEGIN TEST 464
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_464( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(32, 22), MAKE_VAL_COMPLEX(-2, 2), MAKE_VAL_COMPLEX(-25, -17), MAKE_VAL_COMPLEX(74, -45), MAKE_VAL_COMPLEX(-4, 6), MAKE_VAL_COMPLEX(146, -4), MAKE_VAL_COMPLEX(-10, 10), MAKE_VAL_COMPLEX(18, 6), MAKE_VAL_COMPLEX(-43, -120), MAKE_VAL_COMPLEX(20, -10), MAKE_VAL_COMPLEX(-222, -116), MAKE_VAL_COMPLEX(0, -10), MAKE_VAL_COMPLEX(59, 5), MAKE_VAL_COMPLEX(-60, 76), MAKE_VAL_COMPLEX(-8, -4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_464, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_464", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_464() */
/* END TEST 464 */

/*
 *  BEGIN TEST 465
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_465( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-13, -55), MAKE_VAL_COMPLEX(1, 15), MAKE_VAL_COMPLEX(56, 70), MAKE_VAL_COMPLEX(-235, 27), MAKE_VAL_COMPLEX(10, 37), MAKE_VAL_COMPLEX(-244, 75), MAKE_VAL_COMPLEX(5, 75), MAKE_VAL_COMPLEX(89, 60), MAKE_VAL_COMPLEX(13, 195), MAKE_VAL_COMPLEX(30, -115), MAKE_VAL_COMPLEX(262, 16), MAKE_VAL_COMPLEX(-40, -35), MAKE_VAL_COMPLEX(-115, -68), MAKE_VAL_COMPLEX(26, -78), MAKE_VAL_COMPLEX(-44, 18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_465, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_465", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_465() */
/* END TEST 465 */

/*
 *  BEGIN TEST 466
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_466( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(17, 1), MAKE_VAL_COMPLEX(-7, 3), MAKE_VAL_COMPLEX(-15, 23), MAKE_VAL_COMPLEX(-25, -68), MAKE_VAL_COMPLEX(-16, 11), MAKE_VAL_COMPLEX(-48, -25), MAKE_VAL_COMPLEX(-35, 15), MAKE_VAL_COMPLEX(-5, 56), MAKE_VAL_COMPLEX(-36, 3), MAKE_VAL_COMPLEX(60, -5), MAKE_VAL_COMPLEX(48, -4), MAKE_VAL_COMPLEX(10, -25), MAKE_VAL_COMPLEX(14, -35), MAKE_VAL_COMPLEX(-6, -14), MAKE_VAL_COMPLEX(-16, -18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_466, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_466", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_466() */
/* END TEST 466 */

/*
 *  BEGIN TEST 467
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_467( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_ROWMAJ, y432_VW );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(10, -98), MAKE_VAL_COMPLEX(-6, 2), MAKE_VAL_COMPLEX(-22, 118), MAKE_VAL_COMPLEX(-250, -192), MAKE_VAL_COMPLEX(-14, 8), MAKE_VAL_COMPLEX(-374, -352), MAKE_VAL_COMPLEX(-30, 10), MAKE_VAL_COMPLEX(-60, 50), MAKE_VAL_COMPLEX(-220, 314), MAKE_VAL_COMPLEX(50, 0), MAKE_VAL_COMPLEX(300, 692), MAKE_VAL_COMPLEX(10, -20), MAKE_VAL_COMPLEX(-66, -184), MAKE_VAL_COMPLEX(268, -60), MAKE_VAL_COMPLEX(-12, -16) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-2, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_467, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_467", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_467() */
/* END TEST 467 */

/*
 *  BEGIN TEST 468
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_468( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	oski_value_t* y_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(20, -8), MAKE_VAL_COMPLEX(44, 0), MAKE_VAL_COMPLEX(-74, 8), MAKE_VAL_COMPLEX(-53, 160), MAKE_VAL_COMPLEX(-40, -40), MAKE_VAL_COMPLEX(-50, -70), MAKE_VAL_COMPLEX(93, 53), MAKE_VAL_COMPLEX(60, 43), MAKE_VAL_COMPLEX(-28, -32), MAKE_VAL_COMPLEX(144, 60), MAKE_VAL_COMPLEX(-15, 5), MAKE_VAL_COMPLEX(-38, 34) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_468, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_468", y_true, y_val, y432_LEN*y432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_468() */
/* END TEST 468 */

/*
 *  BEGIN TEST 469
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_469( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	oski_value_t* y_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(29, 0), MAKE_VAL_COMPLEX(-25, -43), MAKE_VAL_COMPLEX(29, -103), MAKE_VAL_COMPLEX(17, -168), MAKE_VAL_COMPLEX(-30, -70), MAKE_VAL_COMPLEX(120, 180), MAKE_VAL_COMPLEX(-96, 3), MAKE_VAL_COMPLEX(-89, -40), MAKE_VAL_COMPLEX(-19, -54), MAKE_VAL_COMPLEX(-100, -3), MAKE_VAL_COMPLEX(94, -63), MAKE_VAL_COMPLEX(-34, -50) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_469, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_469", y_true, y_val, y432_LEN*y432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_469() */
/* END TEST 469 */

/*
 *  BEGIN TEST 470
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_470( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	oski_value_t* y_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-32, -22), MAKE_VAL_COMPLEX(-34, 14), MAKE_VAL_COMPLEX(-26, 78), MAKE_VAL_COMPLEX(16, 24), MAKE_VAL_COMPLEX(-20, 100), MAKE_VAL_COMPLEX(20, -100), MAKE_VAL_COMPLEX(32, -30), MAKE_VAL_COMPLEX(18, 14), MAKE_VAL_COMPLEX(-20, 74), MAKE_VAL_COMPLEX(6, -56), MAKE_VAL_COMPLEX(-76, -10), MAKE_VAL_COMPLEX(32, 48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_470, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_470", y_true, y_val, y432_LEN*y432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_470() */
/* END TEST 470 */

/*
 *  BEGIN TEST 471
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_471( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	oski_value_t* y_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(10, 25), MAKE_VAL_COMPLEX(85, 285), MAKE_VAL_COMPLEX(-359, -89), MAKE_VAL_COMPLEX(-948, 435), MAKE_VAL_COMPLEX(50, -50), MAKE_VAL_COMPLEX(140, -840), MAKE_VAL_COMPLEX(181, 566), MAKE_VAL_COMPLEX(88, 501), MAKE_VAL_COMPLEX(40, -35), MAKE_VAL_COMPLEX(255, 790), MAKE_VAL_COMPLEX(-317, -32), MAKE_VAL_COMPLEX(-194, 66) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_471, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_471", y_true, y_val, y432_LEN*y432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_471() */
/* END TEST 471 */

/*
 *  BEGIN TEST 472
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_472( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-143, -23), MAKE_VAL_COMPLEX(-9, 5), MAKE_VAL_COMPLEX(-9, 99), MAKE_VAL_COMPLEX(-71, -128), MAKE_VAL_COMPLEX(-20, 17), MAKE_VAL_COMPLEX(-34, -171), MAKE_VAL_COMPLEX(-45, 25), MAKE_VAL_COMPLEX(14, 43), MAKE_VAL_COMPLEX(-79, -14), MAKE_VAL_COMPLEX(80, -15), MAKE_VAL_COMPLEX(92, -230), MAKE_VAL_COMPLEX(10, -35), MAKE_VAL_COMPLEX(-45, -30), MAKE_VAL_COMPLEX(-66, -98), MAKE_VAL_COMPLEX(-24, -22) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_472, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_472", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_472() */
/* END TEST 472 */

/*
 *  BEGIN TEST 473
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_473( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(189, 21), MAKE_VAL_COMPLEX(-9, 5), MAKE_VAL_COMPLEX(-25, -33), MAKE_VAL_COMPLEX(-15, -48), MAKE_VAL_COMPLEX(-20, 17), MAKE_VAL_COMPLEX(-102, 117), MAKE_VAL_COMPLEX(-45, 25), MAKE_VAL_COMPLEX(-12, 109), MAKE_VAL_COMPLEX(-17, 32), MAKE_VAL_COMPLEX(80, -15), MAKE_VAL_COMPLEX(36, 206), MAKE_VAL_COMPLEX(10, -35), MAKE_VAL_COMPLEX(73, -68), MAKE_VAL_COMPLEX(46, 62), MAKE_VAL_COMPLEX(-24, -22) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_473, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_473", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_473() */
/* END TEST 473 */

/*
 *  BEGIN TEST 474
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_474( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(24, 12), MAKE_VAL_COMPLEX(-12, 0), MAKE_VAL_COMPLEX(-36, 24), MAKE_VAL_COMPLEX(6, -114), MAKE_VAL_COMPLEX(-30, 6), MAKE_VAL_COMPLEX(-54, -66), MAKE_VAL_COMPLEX(-60, 0), MAKE_VAL_COMPLEX(-42, 78), MAKE_VAL_COMPLEX(-54, -18), MAKE_VAL_COMPLEX(90, 30), MAKE_VAL_COMPLEX(72, 24), MAKE_VAL_COMPLEX(30, -30), MAKE_VAL_COMPLEX(42, -42), MAKE_VAL_COMPLEX(0, -24), MAKE_VAL_COMPLEX(-12, -36) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_474, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_474", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_474() */
/* END TEST 474 */

/*
 *  BEGIN TEST 475
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_475( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1152, -1196), MAKE_VAL_COMPLEX(8, -4), MAKE_VAL_COMPLEX(-560, -360), MAKE_VAL_COMPLEX(522, 94), MAKE_VAL_COMPLEX(18, -14), MAKE_VAL_COMPLEX(1006, 1162), MAKE_VAL_COMPLEX(40, -20), MAKE_VAL_COMPLEX(188, 236), MAKE_VAL_COMPLEX(412, -116), MAKE_VAL_COMPLEX(-70, 10), MAKE_VAL_COMPLEX(1520, 1540), MAKE_VAL_COMPLEX(-10, 30), MAKE_VAL_COMPLEX(188, -544), MAKE_VAL_COMPLEX(984, 48), MAKE_VAL_COMPLEX(20, 20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_475, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_475", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_475() */
/* END TEST 475 */

/*
 *  BEGIN TEST 476
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_476( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(4, -84), MAKE_VAL_COMPLEX(6, 10), MAKE_VAL_COMPLEX(100, 42), MAKE_VAL_COMPLEX(-126, 92), MAKE_VAL_COMPLEX(20, 22), MAKE_VAL_COMPLEX(-242, 142), MAKE_VAL_COMPLEX(30, 50), MAKE_VAL_COMPLEX(63, -31), MAKE_VAL_COMPLEX(1, 91), MAKE_VAL_COMPLEX(-20, -90), MAKE_VAL_COMPLEX(-24, -114), MAKE_VAL_COMPLEX(-40, -10), MAKE_VAL_COMPLEX(-67, 47), MAKE_VAL_COMPLEX(-76, 92), MAKE_VAL_COMPLEX(-24, 28) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_476, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_476", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_476() */
/* END TEST 476 */

/*
 *  BEGIN TEST 477
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_477( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-14, 84), MAKE_VAL_COMPLEX(-2, -12), MAKE_VAL_COMPLEX(-92, -56), MAKE_VAL_COMPLEX(143, -53), MAKE_VAL_COMPLEX(-11, -29), MAKE_VAL_COMPLEX(271, -129), MAKE_VAL_COMPLEX(-10, -60), MAKE_VAL_COMPLEX(-62, -2), MAKE_VAL_COMPLEX(20, -94), MAKE_VAL_COMPLEX(-15, 95), MAKE_VAL_COMPLEX(-4, 118), MAKE_VAL_COMPLEX(35, 25), MAKE_VAL_COMPLEX(60, -26), MAKE_VAL_COMPLEX(80, -84), MAKE_VAL_COMPLEX(34, -18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_477, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_477", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_477() */
/* END TEST 477 */

/*
 *  BEGIN TEST 478
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_478( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-8, -14), MAKE_VAL_COMPLEX(6, 4), MAKE_VAL_COMPLEX(26, 0), MAKE_VAL_COMPLEX(-41, 55), MAKE_VAL_COMPLEX(17, 7), MAKE_VAL_COMPLEX(5, 51), MAKE_VAL_COMPLEX(30, 20), MAKE_VAL_COMPLEX(47, -25), MAKE_VAL_COMPLEX(21, 27), MAKE_VAL_COMPLEX(-35, -45), MAKE_VAL_COMPLEX(-28, -36), MAKE_VAL_COMPLEX(-25, 5), MAKE_VAL_COMPLEX(-35, 7), MAKE_VAL_COMPLEX(-8, 12), MAKE_VAL_COMPLEX(-6, 22) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_478, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_478", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_478() */
/* END TEST 478 */

/*
 *  BEGIN TEST 479
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x432', LAYOUT_COLMAJ>
 *  y : <'y432', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_479( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A432_ptr, A432_ind, A432_val,
		A432_M, A432_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y432_val,
		y432_LEN * y432_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y432_LEN, y432_VW, LAYOUT_COLMAJ, y432_LEN );
	oski_value_t* y_val = array_Duplicate( x432_val,
		x432_LEN * x432_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x432_LEN, x432_VW, LAYOUT_COLMAJ, x432_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(117, -21), MAKE_VAL_COMPLEX(-3, -11), MAKE_VAL_COMPLEX(-17, 121), MAKE_VAL_COMPLEX(-2, -39), MAKE_VAL_COMPLEX(-13, -26), MAKE_VAL_COMPLEX(-295, -430), MAKE_VAL_COMPLEX(-15, -55), MAKE_VAL_COMPLEX(-51, -92), MAKE_VAL_COMPLEX(-82, -39), MAKE_VAL_COMPLEX(-5, 90), MAKE_VAL_COMPLEX(72, 14), MAKE_VAL_COMPLEX(35, 20), MAKE_VAL_COMPLEX(-84, 67), MAKE_VAL_COMPLEX(-194, -38), MAKE_VAL_COMPLEX(30, -20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_479, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_479", y_true, y_val, x432_LEN*x432_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_479() */
/* END TEST 479 */

#define A480_M 4    /* # of rows */
#define A480_N 5    /* # of cols */
#define A480_NNZ 0  /* # of stored non-zeros */
static oski_index_t A480_ptr[A480_M+1] = {1, 1, 1, 1, 1}; /* end row pointers for A480 */
static oski_index_t* A480_ind = NULL;
static oski_value_t* A480_val = NULL;

#define x480_LEN 5   /* vector length */
#define x480_VW  3   /* vector width */
static oski_value_t x480_val[x480_LEN*x480_VW] = {MAKE_VAL_COMPLEX(6, 10), MAKE_VAL_COMPLEX(8, -8), MAKE_VAL_COMPLEX(0, 10), MAKE_VAL_COMPLEX(8, 7), MAKE_VAL_COMPLEX(-5, 9), MAKE_VAL_COMPLEX(10, 4), MAKE_VAL_COMPLEX(-9, 2), MAKE_VAL_COMPLEX(-1, -5), MAKE_VAL_COMPLEX(-8, 3), MAKE_VAL_COMPLEX(6, 3), MAKE_VAL_COMPLEX(-6, 1), MAKE_VAL_COMPLEX(-10, 4), MAKE_VAL_COMPLEX(2, -3), MAKE_VAL_COMPLEX(-5, -10), MAKE_VAL_COMPLEX(-1, -10)}; /* data */

#define y480_LEN 4   /* vector length */
#define y480_VW  3   /* vector width */
static oski_value_t y480_val[y480_LEN*y480_VW] = {MAKE_VAL_COMPLEX(-2, -4), MAKE_VAL_COMPLEX(3, -10), MAKE_VAL_COMPLEX(-8, -5), MAKE_VAL_COMPLEX(8, 4), MAKE_VAL_COMPLEX(-10, 10), MAKE_VAL_COMPLEX(-6, -3), MAKE_VAL_COMPLEX(-3, -8), MAKE_VAL_COMPLEX(5, -10), MAKE_VAL_COMPLEX(10, 2), MAKE_VAL_COMPLEX(-3, 10), MAKE_VAL_COMPLEX(7, -5), MAKE_VAL_COMPLEX(-9, -10)}; /* data */

/*
 *  BEGIN TEST 480
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_480( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	oski_value_t* y_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-4, 30), MAKE_VAL_COMPLEX(-41, 10), MAKE_VAL_COMPLEX(4, 57), MAKE_VAL_COMPLEX(0, -37), MAKE_VAL_COMPLEX(65, 19), MAKE_VAL_COMPLEX(16, 37), MAKE_VAL_COMPLEX(-32, 38), MAKE_VAL_COMPLEX(-56, 5), MAKE_VAL_COMPLEX(-30, -43), MAKE_VAL_COMPLEX(55, -15), MAKE_VAL_COMPLEX(-47, -12), MAKE_VAL_COMPLEX(-23, 70) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_480, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_480", y_true, y_val, y480_LEN*y480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_480() */
/* END TEST 480 */

/*
 *  BEGIN TEST 481
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_481( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	oski_value_t* y_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-38, 6), MAKE_VAL_COMPLEX(-88, -16), MAKE_VAL_COMPLEX(-40, 54), MAKE_VAL_COMPLEX(24, -71), MAKE_VAL_COMPLEX(85, 71), MAKE_VAL_COMPLEX(-34, 44), MAKE_VAL_COMPLEX(-55, 22), MAKE_VAL_COMPLEX(-79, -35), MAKE_VAL_COMPLEX(24, -83), MAKE_VAL_COMPLEX(74, 21), MAKE_VAL_COMPLEX(-34, -57), MAKE_VAL_COMPLEX(-70, 68) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_481, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_481", y_true, y_val, y480_LEN*y480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_481() */
/* END TEST 481 */

/*
 *  BEGIN TEST 482
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_482( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	oski_value_t* y_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(52, 14), MAKE_VAL_COMPLEX(66, 107), MAKE_VAL_COMPLEX(109, -32), MAKE_VAL_COMPLEX(-100, 40), MAKE_VAL_COMPLEX(-10, -170), MAKE_VAL_COMPLEX(75, -30), MAKE_VAL_COMPLEX(96, 37), MAKE_VAL_COMPLEX(50, 125), MAKE_VAL_COMPLEX(-98, 74), MAKE_VAL_COMPLEX(-66, -107), MAKE_VAL_COMPLEX(-11, 103), MAKE_VAL_COMPLEX(162, -1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_482, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_482", y_true, y_val, y480_LEN*y480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_482() */
/* END TEST 482 */

/*
 *  BEGIN TEST 483
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_483( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	oski_value_t* y_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(14, 60), MAKE_VAL_COMPLEX(142, -111), MAKE_VAL_COMPLEX(-79, 16), MAKE_VAL_COMPLEX(95, 136), MAKE_VAL_COMPLEX(-177, 107), MAKE_VAL_COMPLEX(29, 20), MAKE_VAL_COMPLEX(-78, -84), MAKE_VAL_COMPLEX(77, -108), MAKE_VAL_COMPLEX(1, 46), MAKE_VAL_COMPLEX(-15, 113), MAKE_VAL_COMPLEX(20, -29), MAKE_VAL_COMPLEX(-134, -105) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_483, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_483", y_true, y_val, y480_LEN*y480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_483() */
/* END TEST 483 */

/*
 *  BEGIN TEST 484
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_484( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(72, 6), MAKE_VAL_COMPLEX(-5, -82), MAKE_VAL_COMPLEX(42, 35), MAKE_VAL_COMPLEX(75, -8), MAKE_VAL_COMPLEX(15, 71), MAKE_VAL_COMPLEX(54, -37), MAKE_VAL_COMPLEX(-29, 45), MAKE_VAL_COMPLEX(-24, -25), MAKE_VAL_COMPLEX(-7, 54), MAKE_VAL_COMPLEX(36, -8), MAKE_VAL_COMPLEX(-12, 29), MAKE_VAL_COMPLEX(-29, 56), MAKE_VAL_COMPLEX(-7, -22), MAKE_VAL_COMPLEX(-70, -15), MAKE_VAL_COMPLEX(-54, -35) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_484, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_484", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_484() */
/* END TEST 484 */

/*
 *  BEGIN TEST 485
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_485( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(12, -2), MAKE_VAL_COMPLEX(-11, 2), MAKE_VAL_COMPLEX(18, 5), MAKE_VAL_COMPLEX(-1, -12), MAKE_VAL_COMPLEX(19, -5), MAKE_VAL_COMPLEX(10, -7), MAKE_VAL_COMPLEX(5, 17), MAKE_VAL_COMPLEX(-10, 11), MAKE_VAL_COMPLEX(-7, 6), MAKE_VAL_COMPLEX(6, -16), MAKE_VAL_COMPLEX(-6, 11), MAKE_VAL_COMPLEX(13, 20), MAKE_VAL_COMPLEX(-3, -2), MAKE_VAL_COMPLEX(-10, 5), MAKE_VAL_COMPLEX(-10, 1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_485, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_485", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_485() */
/* END TEST 485 */

/*
 *  BEGIN TEST 486
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_486( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(20, 56), MAKE_VAL_COMPLEX(48, -32), MAKE_VAL_COMPLEX(-10, 50), MAKE_VAL_COMPLEX(33, 43), MAKE_VAL_COMPLEX(-34, 40), MAKE_VAL_COMPLEX(46, 30), MAKE_VAL_COMPLEX(-47, 1), MAKE_VAL_COMPLEX(0, -26), MAKE_VAL_COMPLEX(-43, 7), MAKE_VAL_COMPLEX(27, 21), MAKE_VAL_COMPLEX(-31, -1), MAKE_VAL_COMPLEX(-54, 10), MAKE_VAL_COMPLEX(13, -13), MAKE_VAL_COMPLEX(-15, -55), MAKE_VAL_COMPLEX(5, -51) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_486, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_486", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_486() */
/* END TEST 486 */

/*
 *  BEGIN TEST 487
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_487( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(2, -86), MAKE_VAL_COMPLEX(-18, -46), MAKE_VAL_COMPLEX(-8, -86), MAKE_VAL_COMPLEX(29, -12), MAKE_VAL_COMPLEX(-33, 19), MAKE_VAL_COMPLEX(-58, -76), MAKE_VAL_COMPLEX(40, -35), MAKE_VAL_COMPLEX(39, -27), MAKE_VAL_COMPLEX(97, 44), MAKE_VAL_COMPLEX(-53, 24), MAKE_VAL_COMPLEX(79, -2), MAKE_VAL_COMPLEX(18, -64), MAKE_VAL_COMPLEX(-17, 6), MAKE_VAL_COMPLEX(-10, 55), MAKE_VAL_COMPLEX(-26, 43) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_487, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_487", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_487() */
/* END TEST 487 */

/*
 *  BEGIN TEST 488
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_488( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(18, -16), MAKE_VAL_COMPLEX(-13, -26), MAKE_VAL_COMPLEX(12, -5), MAKE_VAL_COMPLEX(22, -12), MAKE_VAL_COMPLEX(8, 20), MAKE_VAL_COMPLEX(2, -23), MAKE_VAL_COMPLEX(1, 10), MAKE_VAL_COMPLEX(-5, -8), MAKE_VAL_COMPLEX(16, 18), MAKE_VAL_COMPLEX(3, -2), MAKE_VAL_COMPLEX(9, 7), MAKE_VAL_COMPLEX(-1, 10), MAKE_VAL_COMPLEX(-6, -4), MAKE_VAL_COMPLEX(-20, 10), MAKE_VAL_COMPLEX(-20, 2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_488, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_488", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_488() */
/* END TEST 488 */

/*
 *  BEGIN TEST 489
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_489( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(22, -76), MAKE_VAL_COMPLEX(-83, 10), MAKE_VAL_COMPLEX(58, -45), MAKE_VAL_COMPLEX(-13, -79), MAKE_VAL_COMPLEX(80, -30), MAKE_VAL_COMPLEX(-24, -67), MAKE_VAL_COMPLEX(58, 43), MAKE_VAL_COMPLEX(-25, 40), MAKE_VAL_COMPLEX(45, 23), MAKE_VAL_COMPLEX(-12, -55), MAKE_VAL_COMPLEX(28, 30), MAKE_VAL_COMPLEX(79, 40), MAKE_VAL_COMPLEX(-25, 5), MAKE_VAL_COMPLEX(-25, 75), MAKE_VAL_COMPLEX(-45, 55) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_489, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_489", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_489() */
/* END TEST 489 */

/*
 *  BEGIN TEST 490
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_490( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-58, -6), MAKE_VAL_COMPLEX(8, 56), MAKE_VAL_COMPLEX(-40, -30), MAKE_VAL_COMPLEX(-52, 11), MAKE_VAL_COMPLEX(-21, -47), MAKE_VAL_COMPLEX(-46, 28), MAKE_VAL_COMPLEX(19, -42), MAKE_VAL_COMPLEX(23, 11), MAKE_VAL_COMPLEX(12, -41), MAKE_VAL_COMPLEX(-30, 15), MAKE_VAL_COMPLEX(14, -27), MAKE_VAL_COMPLEX(14, -52), MAKE_VAL_COMPLEX(6, 17), MAKE_VAL_COMPLEX(55, 10), MAKE_VAL_COMPLEX(43, 26) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_490, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_490", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_490() */
/* END TEST 490 */

/*
 *  BEGIN TEST 491
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_491( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-88, 10), MAKE_VAL_COMPLEX(-110, 139), MAKE_VAL_COMPLEX(-11, 36), MAKE_VAL_COMPLEX(-112, -98), MAKE_VAL_COMPLEX(144, -84), MAKE_VAL_COMPLEX(-49, 82), MAKE_VAL_COMPLEX(14, 37), MAKE_VAL_COMPLEX(-84, 71), MAKE_VAL_COMPLEX(-30, -136), MAKE_VAL_COMPLEX(46, -53), MAKE_VAL_COMPLEX(-59, -39), MAKE_VAL_COMPLEX(46, 79), MAKE_VAL_COMPLEX(0, 26), MAKE_VAL_COMPLEX(70, 40), MAKE_VAL_COMPLEX(46, 56) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_491, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_491", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_491() */
/* END TEST 491 */

/*
 *  BEGIN TEST 492
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_492( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	oski_value_t* y_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(2, 12), MAKE_VAL_COMPLEX(-2, 4), MAKE_VAL_COMPLEX(-14, 10), MAKE_VAL_COMPLEX(10, -5), MAKE_VAL_COMPLEX(18, 2), MAKE_VAL_COMPLEX(-8, 15), MAKE_VAL_COMPLEX(-9, -2), MAKE_VAL_COMPLEX(-16, -4), MAKE_VAL_COMPLEX(2, 0), MAKE_VAL_COMPLEX(20, 7), MAKE_VAL_COMPLEX(-13, -4), MAKE_VAL_COMPLEX(-20, 13) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_492, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_492", y_true, y_val, y480_LEN*y480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_492() */
/* END TEST 492 */

/*
 *  BEGIN TEST 493
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_493( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	oski_value_t* y_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(36, 4), MAKE_VAL_COMPLEX(41, 84), MAKE_VAL_COMPLEX(100, -23), MAKE_VAL_COMPLEX(-90, 25), MAKE_VAL_COMPLEX(-8, -132), MAKE_VAL_COMPLEX(68, -30), MAKE_VAL_COMPLEX(78, 40), MAKE_VAL_COMPLEX(41, 104), MAKE_VAL_COMPLEX(-84, 46), MAKE_VAL_COMPLEX(-59, -95), MAKE_VAL_COMPLEX(-6, 81), MAKE_VAL_COMPLEX(143, 3) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_493, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_493", y_true, y_val, y480_LEN*y480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_493() */
/* END TEST 493 */

/*
 *  BEGIN TEST 494
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_494( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	oski_value_t* y_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-36, 18), MAKE_VAL_COMPLEX(-90, -27), MAKE_VAL_COMPLEX(-45, 72), MAKE_VAL_COMPLEX(36, -72), MAKE_VAL_COMPLEX(90, 90), MAKE_VAL_COMPLEX(-27, 54), MAKE_VAL_COMPLEX(-72, 27), MAKE_VAL_COMPLEX(-90, -45), MAKE_VAL_COMPLEX(18, -90), MAKE_VAL_COMPLEX(90, 27), MAKE_VAL_COMPLEX(-45, -63), MAKE_VAL_COMPLEX(-90, 81) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_494, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_494", y_true, y_val, y480_LEN*y480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_494() */
/* END TEST 494 */

/*
 *  BEGIN TEST 495
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_495( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	oski_value_t* y_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-98, 32), MAKE_VAL_COMPLEX(-61, 8), MAKE_VAL_COMPLEX(-66, -81), MAKE_VAL_COMPLEX(24, 52), MAKE_VAL_COMPLEX(34, 134), MAKE_VAL_COMPLEX(-108, -43), MAKE_VAL_COMPLEX(9, -42), MAKE_VAL_COMPLEX(-3, -108), MAKE_VAL_COMPLEX(-26, -10), MAKE_VAL_COMPLEX(-27, 136), MAKE_VAL_COMPLEX(1, -103), MAKE_VAL_COMPLEX(-97, -112) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_495, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_495", y_true, y_val, y480_LEN*y480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_495() */
/* END TEST 495 */

/*
 *  BEGIN TEST 496
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_496( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-42, -48), MAKE_VAL_COMPLEX(-42, 58), MAKE_VAL_COMPLEX(0, -48), MAKE_VAL_COMPLEX(-44, -37), MAKE_VAL_COMPLEX(10, -53), MAKE_VAL_COMPLEX(-57, 0), MAKE_VAL_COMPLEX(35, -24), MAKE_VAL_COMPLEX(7, 16), MAKE_VAL_COMPLEX(44, -28), MAKE_VAL_COMPLEX(-25, -5), MAKE_VAL_COMPLEX(34, -21), MAKE_VAL_COMPLEX(37, -40), MAKE_VAL_COMPLEX(-7, 17), MAKE_VAL_COMPLEX(35, 45), MAKE_VAL_COMPLEX(15, 49) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_496, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_496", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_496() */
/* END TEST 496 */

/*
 *  BEGIN TEST 497
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_497( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-32, -30), MAKE_VAL_COMPLEX(-14, 30), MAKE_VAL_COMPLEX(-20, -42), MAKE_VAL_COMPLEX(-42, -10), MAKE_VAL_COMPLEX(17, -38), MAKE_VAL_COMPLEX(-41, -16), MAKE_VAL_COMPLEX(42, -12), MAKE_VAL_COMPLEX(12, 27), MAKE_VAL_COMPLEX(22, -15), MAKE_VAL_COMPLEX(-35, -10), MAKE_VAL_COMPLEX(18, 0), MAKE_VAL_COMPLEX(45, -16), MAKE_VAL_COMPLEX(-5, 14), MAKE_VAL_COMPLEX(30, 35), MAKE_VAL_COMPLEX(14, 39) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_497, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_497", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_497() */
/* END TEST 497 */

/*
 *  BEGIN TEST 498
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_498( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(2, 26), MAKE_VAL_COMPLEX(24, -8), MAKE_VAL_COMPLEX(-10, 20), MAKE_VAL_COMPLEX(9, 22), MAKE_VAL_COMPLEX(-19, 13), MAKE_VAL_COMPLEX(16, 18), MAKE_VAL_COMPLEX(-20, -5), MAKE_VAL_COMPLEX(3, -11), MAKE_VAL_COMPLEX(-19, -2), MAKE_VAL_COMPLEX(9, 12), MAKE_VAL_COMPLEX(-13, -4), MAKE_VAL_COMPLEX(-24, -2), MAKE_VAL_COMPLEX(7, -4), MAKE_VAL_COMPLEX(0, -25), MAKE_VAL_COMPLEX(8, -21) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_498, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_498", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_498() */
/* END TEST 498 */

/*
 *  BEGIN TEST 499
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_499( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-46, -94), MAKE_VAL_COMPLEX(-160, 96), MAKE_VAL_COMPLEX(62, -4), MAKE_VAL_COMPLEX(9, -102), MAKE_VAL_COMPLEX(-15, -107), MAKE_VAL_COMPLEX(-128, 54), MAKE_VAL_COMPLEX(8, -93), MAKE_VAL_COMPLEX(19, -47), MAKE_VAL_COMPLEX(121, -38), MAKE_VAL_COMPLEX(9, 52), MAKE_VAL_COMPLEX(115, -72), MAKE_VAL_COMPLEX(24, -150), MAKE_VAL_COMPLEX(-9, 20), MAKE_VAL_COMPLEX(40, 55), MAKE_VAL_COMPLEX(16, 59) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_499, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_499", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_499() */
/* END TEST 499 */

/*
 *  BEGIN TEST 500
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_500( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(20, 10), MAKE_VAL_COMPLEX(-2, -14), MAKE_VAL_COMPLEX(20, 22), MAKE_VAL_COMPLEX(26, -4), MAKE_VAL_COMPLEX(-7, 20), MAKE_VAL_COMPLEX(21, 8), MAKE_VAL_COMPLEX(-24, 8), MAKE_VAL_COMPLEX(-10, -17), MAKE_VAL_COMPLEX(-6, 9), MAKE_VAL_COMPLEX(23, 4), MAKE_VAL_COMPLEX(-6, -2), MAKE_VAL_COMPLEX(-25, 8), MAKE_VAL_COMPLEX(1, -8), MAKE_VAL_COMPLEX(-20, -15), MAKE_VAL_COMPLEX(-12, -19) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_500, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_500", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_500() */
/* END TEST 500 */

/*
 *  BEGIN TEST 501
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_501( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(90, 60), MAKE_VAL_COMPLEX(42, -106), MAKE_VAL_COMPLEX(30, 78), MAKE_VAL_COMPLEX(89, 34), MAKE_VAL_COMPLEX(2, 95), MAKE_VAL_COMPLEX(99, -18), MAKE_VAL_COMPLEX(-56, 57), MAKE_VAL_COMPLEX(-25, -28), MAKE_VAL_COMPLEX(-59, 61), MAKE_VAL_COMPLEX(52, -4), MAKE_VAL_COMPLEX(-49, 42), MAKE_VAL_COMPLEX(-55, 82), MAKE_VAL_COMPLEX(4, -32), MAKE_VAL_COMPLEX(-80, -60), MAKE_VAL_COMPLEX(-48, -76) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_501, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_501", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_501() */
/* END TEST 501 */

/*
 *  BEGIN TEST 502
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_502( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-52, 4), MAKE_VAL_COMPLEX(16, 48), MAKE_VAL_COMPLEX(-40, -20), MAKE_VAL_COMPLEX(-44, 18), MAKE_VAL_COMPLEX(-26, -38), MAKE_VAL_COMPLEX(-36, 32), MAKE_VAL_COMPLEX(10, -40), MAKE_VAL_COMPLEX(22, 6), MAKE_VAL_COMPLEX(4, -38), MAKE_VAL_COMPLEX(-24, 18), MAKE_VAL_COMPLEX(8, -26), MAKE_VAL_COMPLEX(4, -48), MAKE_VAL_COMPLEX(8, 14), MAKE_VAL_COMPLEX(50, 0), MAKE_VAL_COMPLEX(42, 16) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_502, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_502", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_502() */
/* END TEST 502 */

/*
 *  BEGIN TEST 503
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_ROWMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_503( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_ROWMAJ, x480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-32, 86), MAKE_VAL_COMPLEX(178, -14), MAKE_VAL_COMPLEX(-110, -46), MAKE_VAL_COMPLEX(-86, 111), MAKE_VAL_COMPLEX(-13, 52), MAKE_VAL_COMPLEX(78, -17), MAKE_VAL_COMPLEX(13, 48), MAKE_VAL_COMPLEX(2, 65), MAKE_VAL_COMPLEX(-109, -21), MAKE_VAL_COMPLEX(-44, -42), MAKE_VAL_COMPLEX(-106, 33), MAKE_VAL_COMPLEX(-14, 93), MAKE_VAL_COMPLEX(16, 2), MAKE_VAL_COMPLEX(30, -40), MAKE_VAL_COMPLEX(38, -24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_503, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_503", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_503() */
/* END TEST 503 */

/*
 *  BEGIN TEST 504
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_504( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	oski_value_t* y_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(16, -30), MAKE_VAL_COMPLEX(91, -48), MAKE_VAL_COMPLEX(-32, -82), MAKE_VAL_COMPLEX(40, 68), MAKE_VAL_COMPLEX(-139, 12), MAKE_VAL_COMPLEX(-34, -53), MAKE_VAL_COMPLEX(27, -64), MAKE_VAL_COMPLEX(94, -45), MAKE_VAL_COMPLEX(60, 71), MAKE_VAL_COMPLEX(-73, 59), MAKE_VAL_COMPLEX(71, 10), MAKE_VAL_COMPLEX(-8, -134) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_504, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_504", y_true, y_val, y480_LEN*y480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_504() */
/* END TEST 504 */

/*
 *  BEGIN TEST 505
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_505( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	oski_value_t* y_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(10, -8), MAKE_VAL_COMPLEX(14, 25), MAKE_VAL_COMPLEX(37, -15), MAKE_VAL_COMPLEX(-36, 24), MAKE_VAL_COMPLEX(-1, -52), MAKE_VAL_COMPLEX(31, -16), MAKE_VAL_COMPLEX(30, -3), MAKE_VAL_COMPLEX(21, 40), MAKE_VAL_COMPLEX(-28, 29), MAKE_VAL_COMPLEX(-32, -36), MAKE_VAL_COMPLEX(9, 28), MAKE_VAL_COMPLEX(53, 3) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_505, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_505", y_true, y_val, y480_LEN*y480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_505() */
/* END TEST 505 */

/*
 *  BEGIN TEST 506
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_506( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	oski_value_t* y_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-40, -10), MAKE_VAL_COMPLEX(-52, -81), MAKE_VAL_COMPLEX(-83, 26), MAKE_VAL_COMPLEX(76, -32), MAKE_VAL_COMPLEX(10, 130), MAKE_VAL_COMPLEX(-57, 24), MAKE_VAL_COMPLEX(-74, -27), MAKE_VAL_COMPLEX(-40, -95), MAKE_VAL_COMPLEX(74, -58), MAKE_VAL_COMPLEX(52, 81), MAKE_VAL_COMPLEX(7, -79), MAKE_VAL_COMPLEX(-124, 3) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_506, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_506", y_true, y_val, y480_LEN*y480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_506() */
/* END TEST 506 */

/*
 *  BEGIN TEST 507
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_507( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	oski_value_t* y_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-20, 146), MAKE_VAL_COMPLEX(78, 155), MAKE_VAL_COMPLEX(-32, -59), MAKE_VAL_COMPLEX(116, 16), MAKE_VAL_COMPLEX(-109, -113), MAKE_VAL_COMPLEX(-105, -66), MAKE_VAL_COMPLEX(-60, 97), MAKE_VAL_COMPLEX(56, -19), MAKE_VAL_COMPLEX(19, 17), MAKE_VAL_COMPLEX(-15, 106), MAKE_VAL_COMPLEX(-98, -14), MAKE_VAL_COMPLEX(93, -142) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(9, 9);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_507, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_507", y_true, y_val, y480_LEN*y480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_507() */
/* END TEST 507 */

/*
 *  BEGIN TEST 508
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_508( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(60, -82), MAKE_VAL_COMPLEX(-80, -36), MAKE_VAL_COMPLEX(77, -38), MAKE_VAL_COMPLEX(29, -75), MAKE_VAL_COMPLEX(87, 13), MAKE_VAL_COMPLEX(5, -102), MAKE_VAL_COMPLEX(33, 76), MAKE_VAL_COMPLEX(-32, 13), MAKE_VAL_COMPLEX(55, 50), MAKE_VAL_COMPLEX(6, -57), MAKE_VAL_COMPLEX(18, 40), MAKE_VAL_COMPLEX(56, 65), MAKE_VAL_COMPLEX(-20, -5), MAKE_VAL_COMPLEX(-74, 60), MAKE_VAL_COMPLEX(-77, 38) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_508, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_508", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_508() */
/* END TEST 508 */

/*
 *  BEGIN TEST 509
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_509( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-84, 42), MAKE_VAL_COMPLEX(48, 68), MAKE_VAL_COMPLEX(-77, -2), MAKE_VAL_COMPLEX(-61, 47), MAKE_VAL_COMPLEX(-67, -49), MAKE_VAL_COMPLEX(-45, 86), MAKE_VAL_COMPLEX(3, -84), MAKE_VAL_COMPLEX(36, 7), MAKE_VAL_COMPLEX(-23, -62), MAKE_VAL_COMPLEX(-30, 45), MAKE_VAL_COMPLEX(6, -44), MAKE_VAL_COMPLEX(-16, -81), MAKE_VAL_COMPLEX(12, 17), MAKE_VAL_COMPLEX(94, -20), MAKE_VAL_COMPLEX(81, 2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_509, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_509", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_509() */
/* END TEST 509 */

/*
 *  BEGIN TEST 510
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_510( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(54, -46), MAKE_VAL_COMPLEX(-56, -40), MAKE_VAL_COMPLEX(60, -10), MAKE_VAL_COMPLEX(34, -55), MAKE_VAL_COMPLEX(59, 21), MAKE_VAL_COMPLEX(14, -64), MAKE_VAL_COMPLEX(21, 52), MAKE_VAL_COMPLEX(-29, 11), MAKE_VAL_COMPLEX(26, 45), MAKE_VAL_COMPLEX(12, -39), MAKE_VAL_COMPLEX(12, 35), MAKE_VAL_COMPLEX(34, 56), MAKE_VAL_COMPLEX(-20, -9), MAKE_VAL_COMPLEX(-55, 40), MAKE_VAL_COMPLEX(-59, 16) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_510, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_510", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_510() */
/* END TEST 510 */

/*
 *  BEGIN TEST 511
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_511( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-48, -96), MAKE_VAL_COMPLEX(-44, 60), MAKE_VAL_COMPLEX(-7, -101), MAKE_VAL_COMPLEX(-56, -31), MAKE_VAL_COMPLEX(49, -67), MAKE_VAL_COMPLEX(-77, -75), MAKE_VAL_COMPLEX(54, 33), MAKE_VAL_COMPLEX(8, 6), MAKE_VAL_COMPLEX(83, -38), MAKE_VAL_COMPLEX(-45, -30), MAKE_VAL_COMPLEX(20, -9), MAKE_VAL_COMPLEX(63, -25), MAKE_VAL_COMPLEX(13, 18), MAKE_VAL_COMPLEX(-7, 64), MAKE_VAL_COMPLEX(-2, 81) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_511, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_511", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_511() */
/* END TEST 511 */

/*
 *  BEGIN TEST 512
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_512( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(46, 8), MAKE_VAL_COMPLEX(8, -44), MAKE_VAL_COMPLEX(27, 22), MAKE_VAL_COMPLEX(42, 7), MAKE_VAL_COMPLEX(12, 42), MAKE_VAL_COMPLEX(45, -28), MAKE_VAL_COMPLEX(-31, 43), MAKE_VAL_COMPLEX(-13, -22), MAKE_VAL_COMPLEX(-8, 28), MAKE_VAL_COMPLEX(27, -9), MAKE_VAL_COMPLEX(-23, 16), MAKE_VAL_COMPLEX(-24, 39), MAKE_VAL_COMPLEX(7, -13), MAKE_VAL_COMPLEX(-54, -25), MAKE_VAL_COMPLEX(-33, -27) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_512, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_512", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_512() */
/* END TEST 512 */

/*
 *  BEGIN TEST 513
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_513( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(4, -38), MAKE_VAL_COMPLEX(-48, 4), MAKE_VAL_COMPLEX(23, -22), MAKE_VAL_COMPLEX(-7, -47), MAKE_VAL_COMPLEX(33, -17), MAKE_VAL_COMPLEX(-25, -22), MAKE_VAL_COMPLEX(41, 2), MAKE_VAL_COMPLEX(-12, 27), MAKE_VAL_COMPLEX(23, 12), MAKE_VAL_COMPLEX(-12, -21), MAKE_VAL_COMPLEX(28, 14), MAKE_VAL_COMPLEX(44, 11), MAKE_VAL_COMPLEX(-22, 3), MAKE_VAL_COMPLEX(4, 50), MAKE_VAL_COMPLEX(-17, 32) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_513, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_513", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_513() */
/* END TEST 513 */

/*
 *  BEGIN TEST 514
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_514( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-42, 66), MAKE_VAL_COMPLEX(72, 24), MAKE_VAL_COMPLEX(-60, 30), MAKE_VAL_COMPLEX(-18, 69), MAKE_VAL_COMPLEX(-69, -3), MAKE_VAL_COMPLEX(6, 72), MAKE_VAL_COMPLEX(-39, -48), MAKE_VAL_COMPLEX(27, -21), MAKE_VAL_COMPLEX(-42, -39), MAKE_VAL_COMPLEX(0, 45), MAKE_VAL_COMPLEX(-24, -33), MAKE_VAL_COMPLEX(-54, -48), MAKE_VAL_COMPLEX(24, 3), MAKE_VAL_COMPLEX(45, -60), MAKE_VAL_COMPLEX(57, -36) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_514, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_514", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_514() */
/* END TEST 514 */

/*
 *  BEGIN TEST 515
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_515( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_ROWMAJ, y480_VW );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(36, 4), MAKE_VAL_COMPLEX(28, -20), MAKE_VAL_COMPLEX(19, 3), MAKE_VAL_COMPLEX(19, 32), MAKE_VAL_COMPLEX(3, 37), MAKE_VAL_COMPLEX(57, -35), MAKE_VAL_COMPLEX(-63, 44), MAKE_VAL_COMPLEX(12, -38), MAKE_VAL_COMPLEX(8, 17), MAKE_VAL_COMPLEX(24, -3), MAKE_VAL_COMPLEX(-35, -8), MAKE_VAL_COMPLEX(-37, 17), MAKE_VAL_COMPLEX(28, 3), MAKE_VAL_COMPLEX(-52, -59), MAKE_VAL_COMPLEX(-23, -28) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_515, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_515", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_515() */
/* END TEST 515 */

/*
 *  BEGIN TEST 516
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_516( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	oski_value_t* y_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(14, 16), MAKE_VAL_COMPLEX(12, 15), MAKE_VAL_COMPLEX(21, 12), MAKE_VAL_COMPLEX(-12, 7), MAKE_VAL_COMPLEX(20, -26), MAKE_VAL_COMPLEX(6, 2), MAKE_VAL_COMPLEX(13, 8), MAKE_VAL_COMPLEX(-8, 28), MAKE_VAL_COMPLEX(-28, 7), MAKE_VAL_COMPLEX(-14, -19), MAKE_VAL_COMPLEX(-7, 14), MAKE_VAL_COMPLEX(23, 1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_516, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_516", y_true, y_val, y480_LEN*y480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_516() */
/* END TEST 516 */

/*
 *  BEGIN TEST 517
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_517( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	oski_value_t* y_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-52, -22), MAKE_VAL_COMPLEX(-67, -86), MAKE_VAL_COMPLEX(-96, 19), MAKE_VAL_COMPLEX(80, -43), MAKE_VAL_COMPLEX(0, 146), MAKE_VAL_COMPLEX(-57, 25), MAKE_VAL_COMPLEX(-84, -27), MAKE_VAL_COMPLEX(-37, -113), MAKE_VAL_COMPLEX(92, -67), MAKE_VAL_COMPLEX(69, 90), MAKE_VAL_COMPLEX(7, -88), MAKE_VAL_COMPLEX(-138, 12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_517, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_517", y_true, y_val, y480_LEN*y480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_517() */
/* END TEST 517 */

/*
 *  BEGIN TEST 518
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_518( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	oski_value_t* y_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-24, -28), MAKE_VAL_COMPLEX(4, -86), MAKE_VAL_COMPLEX(-74, -24), MAKE_VAL_COMPLEX(72, 16), MAKE_VAL_COMPLEX(-60, 100), MAKE_VAL_COMPLEX(-54, -12), MAKE_VAL_COMPLEX(-40, -58), MAKE_VAL_COMPLEX(20, -90), MAKE_VAL_COMPLEX(84, -4), MAKE_VAL_COMPLEX(-4, 86), MAKE_VAL_COMPLEX(46, -54), MAKE_VAL_COMPLEX(-92, -62) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_518, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_518", y_true, y_val, y480_LEN*y480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_518() */
/* END TEST 518 */

/*
 *  BEGIN TEST 519
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_519( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	oski_value_t* y_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-42, 116), MAKE_VAL_COMPLEX(68, 19), MAKE_VAL_COMPLEX(-69, 94), MAKE_VAL_COMPLEX(-5, 66), MAKE_VAL_COMPLEX(82, 104), MAKE_VAL_COMPLEX(-65, -27), MAKE_VAL_COMPLEX(11, -12), MAKE_VAL_COMPLEX(-109, -33), MAKE_VAL_COMPLEX(-57, -70), MAKE_VAL_COMPLEX(-52, -57), MAKE_VAL_COMPLEX(4, -15), MAKE_VAL_COMPLEX(28, -48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_519, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_519", y_true, y_val, y480_LEN*y480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_519() */
/* END TEST 519 */

/*
 *  BEGIN TEST 520
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_520( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(114, 8), MAKE_VAL_COMPLEX(-13, -122), MAKE_VAL_COMPLEX(72, 55), MAKE_VAL_COMPLEX(112, -18), MAKE_VAL_COMPLEX(42, 94), MAKE_VAL_COMPLEX(82, -46), MAKE_VAL_COMPLEX(-44, 81), MAKE_VAL_COMPLEX(-49, -30), MAKE_VAL_COMPLEX(-19, 72), MAKE_VAL_COMPLEX(60, -30), MAKE_VAL_COMPLEX(-18, 56), MAKE_VAL_COMPLEX(-31, 114), MAKE_VAL_COMPLEX(-5, -39), MAKE_VAL_COMPLEX(-119, -30), MAKE_VAL_COMPLEX(-86, -52) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_520, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_520", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_520() */
/* END TEST 520 */

/*
 *  BEGIN TEST 521
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_521( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(118, 16), MAKE_VAL_COMPLEX(-19, -102), MAKE_VAL_COMPLEX(88, 65), MAKE_VAL_COMPLEX(96, -26), MAKE_VAL_COMPLEX(42, 94), MAKE_VAL_COMPLEX(102, -66), MAKE_VAL_COMPLEX(-32, 87), MAKE_VAL_COMPLEX(-43, -14), MAKE_VAL_COMPLEX(-29, 92), MAKE_VAL_COMPLEX(60, -30), MAKE_VAL_COMPLEX(-38, 52), MAKE_VAL_COMPLEX(-25, 94), MAKE_VAL_COMPLEX(-19, -29), MAKE_VAL_COMPLEX(-101, -10), MAKE_VAL_COMPLEX(-86, -52) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_521, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_521", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_521() */
/* END TEST 521 */

/*
 *  BEGIN TEST 522
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_522( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(62, 58), MAKE_VAL_COMPLEX(40, -72), MAKE_VAL_COMPLEX(20, 70), MAKE_VAL_COMPLEX(70, 33), MAKE_VAL_COMPLEX(-17, 73), MAKE_VAL_COMPLEX(78, 8), MAKE_VAL_COMPLEX(-59, 32), MAKE_VAL_COMPLEX(-17, -33), MAKE_VAL_COMPLEX(-50, 37), MAKE_VAL_COMPLEX(48, 9), MAKE_VAL_COMPLEX(-40, 19), MAKE_VAL_COMPLEX(-62, 48), MAKE_VAL_COMPLEX(8, -25), MAKE_VAL_COMPLEX(-55, -60), MAKE_VAL_COMPLEX(-27, -68) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_522, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_522", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_522() */
/* END TEST 522 */

/*
 *  BEGIN TEST 523
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_523( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(82, 74), MAKE_VAL_COMPLEX(83, 1), MAKE_VAL_COMPLEX(75, 55), MAKE_VAL_COMPLEX(3, 61), MAKE_VAL_COMPLEX(-26, 68), MAKE_VAL_COMPLEX(74, -82), MAKE_VAL_COMPLEX(-16, 8), MAKE_VAL_COMPLEX(43, -9), MAKE_VAL_COMPLEX(-28, 104), MAKE_VAL_COMPLEX(45, 15), MAKE_VAL_COMPLEX(-101, 53), MAKE_VAL_COMPLEX(-101, -27), MAKE_VAL_COMPLEX(1, 37), MAKE_VAL_COMPLEX(50, -60), MAKE_VAL_COMPLEX(-17, -69) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_523, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_523", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_523() */
/* END TEST 523 */

/*
 *  BEGIN TEST 524
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_524( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-24, -86), MAKE_VAL_COMPLEX(-69, 30), MAKE_VAL_COMPLEX(12, -75), MAKE_VAL_COMPLEX(-34, -61), MAKE_VAL_COMPLEX(53, -53), MAKE_VAL_COMPLEX(-72, -38), MAKE_VAL_COMPLEX(61, 1), MAKE_VAL_COMPLEX(-6, 29), MAKE_VAL_COMPLEX(67, -15), MAKE_VAL_COMPLEX(-36, -33), MAKE_VAL_COMPLEX(54, 7), MAKE_VAL_COMPLEX(75, 2), MAKE_VAL_COMPLEX(-13, 12), MAKE_VAL_COMPLEX(6, 70), MAKE_VAL_COMPLEX(-13, 72) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_524, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_524", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_524() */
/* END TEST 524 */

/*
 *  BEGIN TEST 525
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_525( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-10, -84), MAKE_VAL_COMPLEX(-83, 42), MAKE_VAL_COMPLEX(38, -65), MAKE_VAL_COMPLEX(-43, -77), MAKE_VAL_COMPLEX(62, -48), MAKE_VAL_COMPLEX(-48, -68), MAKE_VAL_COMPLEX(75, 16), MAKE_VAL_COMPLEX(-5, 46), MAKE_VAL_COMPLEX(60, 13), MAKE_VAL_COMPLEX(-33, -39), MAKE_VAL_COMPLEX(35, 9), MAKE_VAL_COMPLEX(85, -8), MAKE_VAL_COMPLEX(-30, 20), MAKE_VAL_COMPLEX(14, 95), MAKE_VAL_COMPLEX(-23, 73) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_525, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_525", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_525() */
/* END TEST 525 */

/*
 *  BEGIN TEST 526
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_526( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(22, -54), MAKE_VAL_COMPLEX(-56, -8), MAKE_VAL_COMPLEX(40, -30), MAKE_VAL_COMPLEX(4, -53), MAKE_VAL_COMPLEX(51, -7), MAKE_VAL_COMPLEX(-14, -52), MAKE_VAL_COMPLEX(35, 30), MAKE_VAL_COMPLEX(-17, 19), MAKE_VAL_COMPLEX(36, 23), MAKE_VAL_COMPLEX(-6, -33), MAKE_VAL_COMPLEX(22, 21), MAKE_VAL_COMPLEX(46, 28), MAKE_VAL_COMPLEX(-18, 1), MAKE_VAL_COMPLEX(-25, 50), MAKE_VAL_COMPLEX(-37, 34) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_526, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_526", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_526() */
/* END TEST 526 */

/*
 *  BEGIN TEST 527
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x480', LAYOUT_COLMAJ>
 *  y : <'y480', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_527( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A480_ptr, A480_ind, A480_val,
		A480_M, A480_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y480_val,
		y480_LEN * y480_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y480_LEN, y480_VW, LAYOUT_COLMAJ, y480_LEN );
	oski_value_t* y_val = array_Duplicate( x480_val,
		x480_LEN * x480_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x480_LEN, x480_VW, LAYOUT_COLMAJ, x480_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(112, -46), MAKE_VAL_COMPLEX(1, -44), MAKE_VAL_COMPLEX(118, -33), MAKE_VAL_COMPLEX(32, -13), MAKE_VAL_COMPLEX(67, 49), MAKE_VAL_COMPLEX(-8, -146), MAKE_VAL_COMPLEX(31, 41), MAKE_VAL_COMPLEX(10, -7), MAKE_VAL_COMPLEX(71, 107), MAKE_VAL_COMPLEX(30, -45), MAKE_VAL_COMPLEX(-20, 107), MAKE_VAL_COMPLEX(-35, 56), MAKE_VAL_COMPLEX(1, 28), MAKE_VAL_COMPLEX(-16, -14), MAKE_VAL_COMPLEX(-81, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_527, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_527", y_true, y_val, x480_LEN*x480_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_527() */
/* END TEST 527 */

#define A528_M 4    /* # of rows */
#define A528_N 5    /* # of cols */
#define A528_NNZ 1  /* # of stored non-zeros */
static oski_index_t A528_ptr[A528_M+1] = {0, 0, 1, 1, 1}; /* end row pointers for A528 */
static oski_index_t A528_ind[A528_NNZ] = {0}; /* end column indices pointers for A528 */
static oski_value_t A528_val[A528_NNZ] = {MAKE_VAL_COMPLEX(6, -10)}; /* end stored non-zero values for A528 */

#define x528_LEN 5   /* vector length */
#define x528_VW  3   /* vector width */
static oski_value_t x528_val[x528_LEN*x528_VW] = {MAKE_VAL_COMPLEX(2, 0), MAKE_VAL_COMPLEX(-8, 4), MAKE_VAL_COMPLEX(0, -10), MAKE_VAL_COMPLEX(10, -3), MAKE_VAL_COMPLEX(-2, 1), MAKE_VAL_COMPLEX(7, 10), MAKE_VAL_COMPLEX(-5, -7), MAKE_VAL_COMPLEX(10, 4), MAKE_VAL_COMPLEX(-10, 5), MAKE_VAL_COMPLEX(-9, 4), MAKE_VAL_COMPLEX(-3, -10), MAKE_VAL_COMPLEX(7, 3), MAKE_VAL_COMPLEX(6, 10), MAKE_VAL_COMPLEX(9, 1), MAKE_VAL_COMPLEX(-6, 5)}; /* data */

#define y528_LEN 4   /* vector length */
#define y528_VW  3   /* vector width */
static oski_value_t y528_val[y528_LEN*y528_VW] = {MAKE_VAL_COMPLEX(-10, 4), MAKE_VAL_COMPLEX(7, -5), MAKE_VAL_COMPLEX(-8, 2), MAKE_VAL_COMPLEX(-10, 7), MAKE_VAL_COMPLEX(-3, -10), MAKE_VAL_COMPLEX(5, -2), MAKE_VAL_COMPLEX(4, 5), MAKE_VAL_COMPLEX(-6, -9), MAKE_VAL_COMPLEX(-8, -9), MAKE_VAL_COMPLEX(-1, 9), MAKE_VAL_COMPLEX(10, 10), MAKE_VAL_COMPLEX(7, 8)}; /* data */

/*
 *  BEGIN TEST 528
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_528( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	oski_value_t* y_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(2, 58), MAKE_VAL_COMPLEX(-19, -41), MAKE_VAL_COMPLEX(-6, 34), MAKE_VAL_COMPLEX(37, 41), MAKE_VAL_COMPLEX(-66, 100), MAKE_VAL_COMPLEX(-93, -79), MAKE_VAL_COMPLEX(28, -17), MAKE_VAL_COMPLEX(-47, 16), MAKE_VAL_COMPLEX(-71, 27), MAKE_VAL_COMPLEX(34, 27), MAKE_VAL_COMPLEX(67, -40), MAKE_VAL_COMPLEX(61, -16) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_528, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_528", y_true, y_val, y528_LEN*y528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_528() */
/* END TEST 528 */

/*
 *  BEGIN TEST 529
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_529( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	oski_value_t* y_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-22, 8), MAKE_VAL_COMPLEX(22, -14), MAKE_VAL_COMPLEX(-16, 14), MAKE_VAL_COMPLEX(-42, 37), MAKE_VAL_COMPLEX(4, -125), MAKE_VAL_COMPLEX(103, 46), MAKE_VAL_COMPLEX(13, 17), MAKE_VAL_COMPLEX(-22, -22), MAKE_VAL_COMPLEX(-6, -23), MAKE_VAL_COMPLEX(7, 14), MAKE_VAL_COMPLEX(23, 30), MAKE_VAL_COMPLEX(7, 13) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_529, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_529", y_true, y_val, y528_LEN*y528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_529() */
/* END TEST 529 */

/*
 *  BEGIN TEST 530
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_530( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	oski_value_t* y_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(66, -38), MAKE_VAL_COMPLEX(-44, 42), MAKE_VAL_COMPLEX(54, -22), MAKE_VAL_COMPLEX(63, -59), MAKE_VAL_COMPLEX(31, 67), MAKE_VAL_COMPLEX(-33, 19), MAKE_VAL_COMPLEX(-33, -31), MAKE_VAL_COMPLEX(51, 57), MAKE_VAL_COMPLEX(65, 55), MAKE_VAL_COMPLEX(-2, -64), MAKE_VAL_COMPLEX(-80, -60), MAKE_VAL_COMPLEX(-57, -49) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_530, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_530", y_true, y_val, y528_LEN*y528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_530() */
/* END TEST 530 */

/*
 *  BEGIN TEST 531
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_531( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	oski_value_t* y_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-86, -64), MAKE_VAL_COMPLEX(130, 64), MAKE_VAL_COMPLEX(-124, -6), MAKE_VAL_COMPLEX(-348, -97), MAKE_VAL_COMPLEX(830, -421), MAKE_VAL_COMPLEX(61, 876), MAKE_VAL_COMPLEX(-44, 116), MAKE_VAL_COMPLEX(21, -173), MAKE_VAL_COMPLEX(98, -51), MAKE_VAL_COMPLEX(-4, 85), MAKE_VAL_COMPLEX(-78, 181), MAKE_VAL_COMPLEX(-28, 28) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_531, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_531", y_true, y_val, y528_LEN*y528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_531() */
/* END TEST 531 */

/*
 *  BEGIN TEST 532
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_532( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-16, 140), MAKE_VAL_COMPLEX(-35, -43), MAKE_VAL_COMPLEX(-28, 20), MAKE_VAL_COMPLEX(-99, 1), MAKE_VAL_COMPLEX(16, -12), MAKE_VAL_COMPLEX(-21, -103), MAKE_VAL_COMPLEX(23, 76), MAKE_VAL_COMPLEX(-74, -71), MAKE_VAL_COMPLEX(87, -19), MAKE_VAL_COMPLEX(83, 4), MAKE_VAL_COMPLEX(4, 99), MAKE_VAL_COMPLEX(-40, -37), MAKE_VAL_COMPLEX(-18, -98), MAKE_VAL_COMPLEX(-69, -35), MAKE_VAL_COMPLEX(63, -22) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_532, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_532", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_532() */
/* END TEST 532 */

/*
 *  BEGIN TEST 533
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_533( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, -156), MAKE_VAL_COMPLEX(131, 75), MAKE_VAL_COMPLEX(-52, 60), MAKE_VAL_COMPLEX(-5, -57), MAKE_VAL_COMPLEX(8, 20), MAKE_VAL_COMPLEX(45, -33), MAKE_VAL_COMPLEX(-39, 20), MAKE_VAL_COMPLEX(26, -41), MAKE_VAL_COMPLEX(33, 59), MAKE_VAL_COMPLEX(21, 36), MAKE_VAL_COMPLEX(-60, 5), MAKE_VAL_COMPLEX(8, -43), MAKE_VAL_COMPLEX(50, -30), MAKE_VAL_COMPLEX(5, -45), MAKE_VAL_COMPLEX(25, 30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_533, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_533", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_533() */
/* END TEST 533 */

/*
 *  BEGIN TEST 534
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_534( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-16, 6), MAKE_VAL_COMPLEX(52, -56), MAKE_VAL_COMPLEX(30, 80), MAKE_VAL_COMPLEX(-71, 54), MAKE_VAL_COMPLEX(13, -14), MAKE_VAL_COMPLEX(-86, -59), MAKE_VAL_COMPLEX(61, 41), MAKE_VAL_COMPLEX(-92, -2), MAKE_VAL_COMPLEX(65, -70), MAKE_VAL_COMPLEX(60, -59), MAKE_VAL_COMPLEX(54, 71), MAKE_VAL_COMPLEX(-65, -3), MAKE_VAL_COMPLEX(-78, -62), MAKE_VAL_COMPLEX(-75, 19), MAKE_VAL_COMPLEX(33, -58) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_534, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_534", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_534() */
/* END TEST 534 */

/*
 *  BEGIN TEST 535
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_535( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(578, -1156), MAKE_VAL_COMPLEX(748, 664), MAKE_VAL_COMPLEX(-196, 502), MAKE_VAL_COMPLEX(96, 53), MAKE_VAL_COMPLEX(-16, 77), MAKE_VAL_COMPLEX(-129, 8), MAKE_VAL_COMPLEX(45, -65), MAKE_VAL_COMPLEX(-42, 144), MAKE_VAL_COMPLEX(28, 29), MAKE_VAL_COMPLEX(47, -134), MAKE_VAL_COMPLEX(29, -108), MAKE_VAL_COMPLEX(-63, -59), MAKE_VAL_COMPLEX(-78, 6), MAKE_VAL_COMPLEX(-33, 51), MAKE_VAL_COMPLEX(-12, -51) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, -4);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_535, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_535", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_535() */
/* END TEST 535 */

/*
 *  BEGIN TEST 536
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_536( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-152, -44), MAKE_VAL_COMPLEX(117, -159), MAKE_VAL_COMPLEX(92, 100), MAKE_VAL_COMPLEX(-55, 75), MAKE_VAL_COMPLEX(4, -26), MAKE_VAL_COMPLEX(-87, -27), MAKE_VAL_COMPLEX(69, 22), MAKE_VAL_COMPLEX(-86, 17), MAKE_VAL_COMPLEX(27, -89), MAKE_VAL_COMPLEX(33, -60), MAKE_VAL_COMPLEX(78, 55), MAKE_VAL_COMPLEX(-50, 25), MAKE_VAL_COMPLEX(-86, -30), MAKE_VAL_COMPLEX(-59, 39), MAKE_VAL_COMPLEX(11, -60) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_536, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_536", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_536() */
/* END TEST 536 */

/*
 *  BEGIN TEST 537
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_537( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(142, 50), MAKE_VAL_COMPLEX(-89, 115), MAKE_VAL_COMPLEX(-62, -50), MAKE_VAL_COMPLEX(14, -30), MAKE_VAL_COMPLEX(3, 15), MAKE_VAL_COMPLEX(22, -2), MAKE_VAL_COMPLEX(-23, -2), MAKE_VAL_COMPLEX(24, -7), MAKE_VAL_COMPLEX(8, 34), MAKE_VAL_COMPLEX(0, 13), MAKE_VAL_COMPLEX(-33, -14), MAKE_VAL_COMPLEX(6, -19), MAKE_VAL_COMPLEX(26, -2), MAKE_VAL_COMPLEX(11, -17), MAKE_VAL_COMPLEX(4, 17) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_537, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_537", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_537() */
/* END TEST 537 */

/*
 *  BEGIN TEST 538
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_538( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-6, 6), MAKE_VAL_COMPLEX(12, -36), MAKE_VAL_COMPLEX(30, 30), MAKE_VAL_COMPLEX(-21, 39), MAKE_VAL_COMPLEX(3, -9), MAKE_VAL_COMPLEX(-51, -9), MAKE_VAL_COMPLEX(36, 6), MAKE_VAL_COMPLEX(-42, 18), MAKE_VAL_COMPLEX(15, -45), MAKE_VAL_COMPLEX(15, -39), MAKE_VAL_COMPLEX(39, 21), MAKE_VAL_COMPLEX(-30, 12), MAKE_VAL_COMPLEX(-48, -12), MAKE_VAL_COMPLEX(-30, 24), MAKE_VAL_COMPLEX(3, -33) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_538, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_538", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_538() */
/* END TEST 538 */

/*
 *  BEGIN TEST 539
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_539( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(182, 1196), MAKE_VAL_COMPLEX(-1021, -243), MAKE_VAL_COMPLEX(112, -454), MAKE_VAL_COMPLEX(89, 42), MAKE_VAL_COMPLEX(-58, 61), MAKE_VAL_COMPLEX(-34, -27), MAKE_VAL_COMPLEX(19, -48), MAKE_VAL_COMPLEX(-39, 78), MAKE_VAL_COMPLEX(-31, 92), MAKE_VAL_COMPLEX(67, -29), MAKE_VAL_COMPLEX(30, -110), MAKE_VAL_COMPLEX(28, -81), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_539, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_539", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_539() */
/* END TEST 539 */

/*
 *  BEGIN TEST 540
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_540( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	oski_value_t* y_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-22, -60), MAKE_VAL_COMPLEX(52, 19), MAKE_VAL_COMPLEX(-17, -55), MAKE_VAL_COMPLEX(-51, -56), MAKE_VAL_COMPLEX(52, -14), MAKE_VAL_COMPLEX(2, 135), MAKE_VAL_COMPLEX(-20, 28), MAKE_VAL_COMPLEX(51, -46), MAKE_VAL_COMPLEX(54, -58), MAKE_VAL_COMPLEX(-147, -56), MAKE_VAL_COMPLEX(-70, 65), MAKE_VAL_COMPLEX(-41, 45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_540, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_540", y_true, y_val, y528_LEN*y528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_540() */
/* END TEST 540 */

/*
 *  BEGIN TEST 541
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_541( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	oski_value_t* y_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-84, -60), MAKE_VAL_COMPLEX(53, 54), MAKE_VAL_COMPLEX(-51, -47), MAKE_VAL_COMPLEX(-97, -49), MAKE_VAL_COMPLEX(73, -78), MAKE_VAL_COMPLEX(51, -75), MAKE_VAL_COMPLEX(-30, 53), MAKE_VAL_COMPLEX(45, -83), MAKE_VAL_COMPLEX(32, -99), MAKE_VAL_COMPLEX(16, 87), MAKE_VAL_COMPLEX(-20, 125), MAKE_VAL_COMPLEX(-36, 93) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_541, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_541", y_true, y_val, y528_LEN*y528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_541() */
/* END TEST 541 */

/*
 *  BEGIN TEST 542
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_542( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	oski_value_t* y_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-86, 46), MAKE_VAL_COMPLEX(58, -52), MAKE_VAL_COMPLEX(-70, 26), MAKE_VAL_COMPLEX(-83, 73), MAKE_VAL_COMPLEX(-37, -87), MAKE_VAL_COMPLEX(43, -23), MAKE_VAL_COMPLEX(41, 41), MAKE_VAL_COMPLEX(-63, -75), MAKE_VAL_COMPLEX(-81, -73), MAKE_VAL_COMPLEX(0, 82), MAKE_VAL_COMPLEX(100, 80), MAKE_VAL_COMPLEX(71, 65) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_542, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_542", y_true, y_val, y528_LEN*y528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_542() */
/* END TEST 542 */

/*
 *  BEGIN TEST 543
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_543( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	oski_value_t* y_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-32, 102), MAKE_VAL_COMPLEX(-158, -297), MAKE_VAL_COMPLEX(-99, 119), MAKE_VAL_COMPLEX(16, 215), MAKE_VAL_COMPLEX(-87, 25), MAKE_VAL_COMPLEX(847, 126), MAKE_VAL_COMPLEX(115, -77), MAKE_VAL_COMPLEX(-206, 5), MAKE_VAL_COMPLEX(-217, -1), MAKE_VAL_COMPLEX(-419, 766), MAKE_VAL_COMPLEX(190, 65), MAKE_VAL_COMPLEX(152, -60) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_543, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_543", y_true, y_val, y528_LEN*y528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_543() */
/* END TEST 543 */

/*
 *  BEGIN TEST 544
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_544( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-28, -94), MAKE_VAL_COMPLEX(43, -100), MAKE_VAL_COMPLEX(86, 105), MAKE_VAL_COMPLEX(-40, 20), MAKE_VAL_COMPLEX(14, -9), MAKE_VAL_COMPLEX(-46, -34), MAKE_VAL_COMPLEX(24, 32), MAKE_VAL_COMPLEX(-50, -5), MAKE_VAL_COMPLEX(55, -25), MAKE_VAL_COMPLEX(31, -22), MAKE_VAL_COMPLEX(19, 38), MAKE_VAL_COMPLEX(-31, 0), MAKE_VAL_COMPLEX(-40, -44), MAKE_VAL_COMPLEX(-46, 4), MAKE_VAL_COMPLEX(25, -31) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_544, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_544", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_544() */
/* END TEST 544 */

/*
 *  BEGIN TEST 545
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_545( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(18, 104), MAKE_VAL_COMPLEX(-23, 40), MAKE_VAL_COMPLEX(-36, -55), MAKE_VAL_COMPLEX(5, 45), MAKE_VAL_COMPLEX(-9, -6), MAKE_VAL_COMPLEX(-39, 19), MAKE_VAL_COMPLEX(36, -22), MAKE_VAL_COMPLEX(-20, 35), MAKE_VAL_COMPLEX(-30, -50), MAKE_VAL_COMPLEX(-6, -43), MAKE_VAL_COMPLEX(46, -3), MAKE_VAL_COMPLEX(-19, 20), MAKE_VAL_COMPLEX(-40, 24), MAKE_VAL_COMPLEX(-4, 36), MAKE_VAL_COMPLEX(-20, -24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_545, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_545", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_545() */
/* END TEST 545 */

/*
 *  BEGIN TEST 546
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_546( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-16, -14), MAKE_VAL_COMPLEX(92, 24), MAKE_VAL_COMPLEX(-70, 80), MAKE_VAL_COMPLEX(-101, -46), MAKE_VAL_COMPLEX(23, 6), MAKE_VAL_COMPLEX(14, -129), MAKE_VAL_COMPLEX(-9, 91), MAKE_VAL_COMPLEX(-52, -102), MAKE_VAL_COMPLEX(115, 30), MAKE_VAL_COMPLEX(100, 31), MAKE_VAL_COMPLEX(-46, 101), MAKE_VAL_COMPLEX(-35, -73), MAKE_VAL_COMPLEX(22, -122), MAKE_VAL_COMPLEX(-65, -71), MAKE_VAL_COMPLEX(83, 2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_546, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_546", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_546() */
/* END TEST 546 */

/*
 *  BEGIN TEST 547
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_547( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-50, 300), MAKE_VAL_COMPLEX(-49, 217), MAKE_VAL_COMPLEX(-203, -201), MAKE_VAL_COMPLEX(-75, -10), MAKE_VAL_COMPLEX(-6, 3), MAKE_VAL_COMPLEX(14, -87), MAKE_VAL_COMPLEX(25, 45), MAKE_VAL_COMPLEX(-35, -65), MAKE_VAL_COMPLEX(35, -30), MAKE_VAL_COMPLEX(85, 0), MAKE_VAL_COMPLEX(-9, 82), MAKE_VAL_COMPLEX(-32, -64), MAKE_VAL_COMPLEX(6, -58), MAKE_VAL_COMPLEX(-33, -31), MAKE_VAL_COMPLEX(39, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-3, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_547, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_547", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_547() */
/* END TEST 547 */

/*
 *  BEGIN TEST 548
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_548( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(80, 60), MAKE_VAL_COMPLEX(23, -40), MAKE_VAL_COMPLEX(-24, 45), MAKE_VAL_COMPLEX(21, 78), MAKE_VAL_COMPLEX(-1, -19), MAKE_VAL_COMPLEX(-88, 55), MAKE_VAL_COMPLEX(53, -31), MAKE_VAL_COMPLEX(-38, 81), MAKE_VAL_COMPLEX(-20, -75), MAKE_VAL_COMPLEX(-33, -69), MAKE_VAL_COMPLEX(77, -23), MAKE_VAL_COMPLEX(-24, 61), MAKE_VAL_COMPLEX(-86, 38), MAKE_VAL_COMPLEX(-17, 71), MAKE_VAL_COMPLEX(-34, -53) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_548, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_548", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_548() */
/* END TEST 548 */

/*
 *  BEGIN TEST 549
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_549( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-74, -62), MAKE_VAL_COMPLEX(-43, 60), MAKE_VAL_COMPLEX(14, -75), MAKE_VAL_COMPLEX(6, -97), MAKE_VAL_COMPLEX(-4, 24), MAKE_VAL_COMPLEX(119, -32), MAKE_VAL_COMPLEX(-75, 15), MAKE_VAL_COMPLEX(72, -79), MAKE_VAL_COMPLEX(-5, 100), MAKE_VAL_COMPLEX(10, 90), MAKE_VAL_COMPLEX(-96, -4), MAKE_VAL_COMPLEX(48, -59), MAKE_VAL_COMPLEX(114, -14), MAKE_VAL_COMPLEX(45, -77), MAKE_VAL_COMPLEX(21, 74) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_549, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_549", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_549() */
/* END TEST 549 */

/*
 *  BEGIN TEST 550
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_550( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-2, 12), MAKE_VAL_COMPLEX(-16, -52), MAKE_VAL_COMPLEX(60, 10), MAKE_VAL_COMPLEX(8, 63), MAKE_VAL_COMPLEX(-4, -13), MAKE_VAL_COMPLEX(-67, 32), MAKE_VAL_COMPLEX(47, -23), MAKE_VAL_COMPLEX(-34, 56), MAKE_VAL_COMPLEX(-20, -65), MAKE_VAL_COMPLEX(-15, -58), MAKE_VAL_COMPLEX(63, -8), MAKE_VAL_COMPLEX(-25, 39), MAKE_VAL_COMPLEX(-66, 26), MAKE_VAL_COMPLEX(-15, 53), MAKE_VAL_COMPLEX(-24, -41) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_550, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_550", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_550() */
/* END TEST 550 */

/*
 *  BEGIN TEST 551
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_ROWMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_551( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_ROWMAJ, x528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1150, -354), MAKE_VAL_COMPLEX(635, -91), MAKE_VAL_COMPLEX(-681, 1171), MAKE_VAL_COMPLEX(80, -192), MAKE_VAL_COMPLEX(17, -43), MAKE_VAL_COMPLEX(188, 128), MAKE_VAL_COMPLEX(-136, 64), MAKE_VAL_COMPLEX(185, -19), MAKE_VAL_COMPLEX(130, 100), MAKE_VAL_COMPLEX(-75, 239), MAKE_VAL_COMPLEX(-219, -89), MAKE_VAL_COMPLEX(209, -9), MAKE_VAL_COMPLEX(108, 44), MAKE_VAL_COMPLEX(78, -46), MAKE_VAL_COMPLEX(-18, 76) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(9, -9);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_551, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_551", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_551() */
/* END TEST 551 */

/*
 *  BEGIN TEST 552
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_552( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	oski_value_t* y_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(46, 52), MAKE_VAL_COMPLEX(-37, -22), MAKE_VAL_COMPLEX(25, 34), MAKE_VAL_COMPLEX(66, 30), MAKE_VAL_COMPLEX(83, 21), MAKE_VAL_COMPLEX(-133, -53), MAKE_VAL_COMPLEX(22, -44), MAKE_VAL_COMPLEX(-32, 58), MAKE_VAL_COMPLEX(-32, 76), MAKE_VAL_COMPLEX(66, -15), MAKE_VAL_COMPLEX(30, -75), MAKE_VAL_COMPLEX(43, -57) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_552, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_552", y_true, y_val, y528_LEN*y528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_552() */
/* END TEST 552 */

/*
 *  BEGIN TEST 553
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_553( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	oski_value_t* y_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-4, -34), MAKE_VAL_COMPLEX(1, 16), MAKE_VAL_COMPLEX(5, -16), MAKE_VAL_COMPLEX(-15, -21), MAKE_VAL_COMPLEX(-104, 18), MAKE_VAL_COMPLEX(112, 44), MAKE_VAL_COMPLEX(-19, 17), MAKE_VAL_COMPLEX(23, -13), MAKE_VAL_COMPLEX(29, -25), MAKE_VAL_COMPLEX(-36, -9), MAKE_VAL_COMPLEX(-30, 15), MAKE_VAL_COMPLEX(-40, 12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_553, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_553", y_true, y_val, y528_LEN*y528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_553() */
/* END TEST 553 */

/*
 *  BEGIN TEST 554
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_554( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	oski_value_t* y_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(34, -60), MAKE_VAL_COMPLEX(-15, 53), MAKE_VAL_COMPLEX(32, -42), MAKE_VAL_COMPLEX(22, -75), MAKE_VAL_COMPLEX(55, 38), MAKE_VAL_COMPLEX(-17, 30), MAKE_VAL_COMPLEX(-40, -9), MAKE_VAL_COMPLEX(66, 21), MAKE_VAL_COMPLEX(76, 13), MAKE_VAL_COMPLEX(-31, -49), MAKE_VAL_COMPLEX(-90, -10), MAKE_VAL_COMPLEX(-67, -12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_554, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_554", y_true, y_val, y528_LEN*y528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_554() */
/* END TEST 554 */

/*
 *  BEGIN TEST 555
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_555( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	oski_value_t* y_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(4, -108), MAKE_VAL_COMPLEX(89, 107), MAKE_VAL_COMPLEX(-39, -119), MAKE_VAL_COMPLEX(-61, -203), MAKE_VAL_COMPLEX(733, -493), MAKE_VAL_COMPLEX(-633, 249), MAKE_VAL_COMPLEX(-87, -29), MAKE_VAL_COMPLEX(161, -37), MAKE_VAL_COMPLEX(165, -13), MAKE_VAL_COMPLEX(-37, -81), MAKE_VAL_COMPLEX(-155, 115), MAKE_VAL_COMPLEX(-45, 17) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, -3);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_555, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_555", y_true, y_val, y528_LEN*y528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_555() */
/* END TEST 555 */

/*
 *  BEGIN TEST 556
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_556( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(6, 130), MAKE_VAL_COMPLEX(-2, 83), MAKE_VAL_COMPLEX(-76, -25), MAKE_VAL_COMPLEX(5, -80), MAKE_VAL_COMPLEX(2, 19), MAKE_VAL_COMPLEX(-10, -61), MAKE_VAL_COMPLEX(-74, 9), MAKE_VAL_COMPLEX(56, -77), MAKE_VAL_COMPLEX(20, 105), MAKE_VAL_COMPLEX(5, 84), MAKE_VAL_COMPLEX(-87, -66), MAKE_VAL_COMPLEX(50, -49), MAKE_VAL_COMPLEX(90, -27), MAKE_VAL_COMPLEX(42, -61), MAKE_VAL_COMPLEX(22, 63) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_556, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_556", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_556() */
/* END TEST 556 */

/*
 *  BEGIN TEST 557
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_557( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(16, -140), MAKE_VAL_COMPLEX(-66, 1), MAKE_VAL_COMPLEX(26, -85), MAKE_VAL_COMPLEX(90, -3), MAKE_VAL_COMPLEX(-19, 2), MAKE_VAL_COMPLEX(137, 136), MAKE_VAL_COMPLEX(-16, -61), MAKE_VAL_COMPLEX(74, 71), MAKE_VAL_COMPLEX(-105, 0), MAKE_VAL_COMPLEX(-84, 5), MAKE_VAL_COMPLEX(4, -29), MAKE_VAL_COMPLEX(42, 47), MAKE_VAL_COMPLEX(26, 107), MAKE_VAL_COMPLEX(62, 27), MAKE_VAL_COMPLEX(-63, 22) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_557, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_557", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_557() */
/* END TEST 557 */

/*
 *  BEGIN TEST 558
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_558( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-18, 2), MAKE_VAL_COMPLEX(68, -44), MAKE_VAL_COMPLEX(10, 90), MAKE_VAL_COMPLEX(-87, 37), MAKE_VAL_COMPLEX(17, -11), MAKE_VAL_COMPLEX(-73, -83), MAKE_VAL_COMPLEX(52, 58), MAKE_VAL_COMPLEX(-94, -26), MAKE_VAL_COMPLEX(85, -55), MAKE_VAL_COMPLEX(77, -45), MAKE_VAL_COMPLEX(37, 87), MAKE_VAL_COMPLEX(-66, -20), MAKE_VAL_COMPLEX(-64, -84), MAKE_VAL_COMPLEX(-82, 0), MAKE_VAL_COMPLEX(49, -51) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_558, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_558", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_558() */
/* END TEST 558 */

/*
 *  BEGIN TEST 559
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_559( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(716, -422), MAKE_VAL_COMPLEX(89, -63), MAKE_VAL_COMPLEX(93, 35), MAKE_VAL_COMPLEX(2, 79), MAKE_VAL_COMPLEX(6, -23), MAKE_VAL_COMPLEX(29, 646), MAKE_VAL_COMPLEX(50, 54), MAKE_VAL_COMPLEX(-129, 109), MAKE_VAL_COMPLEX(50, -195), MAKE_VAL_COMPLEX(31, -100), MAKE_VAL_COMPLEX(-205, 216), MAKE_VAL_COMPLEX(-98, 16), MAKE_VAL_COMPLEX(-143, 45), MAKE_VAL_COMPLEX(-52, 6), MAKE_VAL_COMPLEX(2, -83) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-3, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_559, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_559", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_559() */
/* END TEST 559 */

/*
 *  BEGIN TEST 560
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_560( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-138, -42), MAKE_VAL_COMPLEX(-42, -37), MAKE_VAL_COMPLEX(64, -5), MAKE_VAL_COMPLEX(27, 66), MAKE_VAL_COMPLEX(-8, -11), MAKE_VAL_COMPLEX(36, -43), MAKE_VAL_COMPLEX(34, -47), MAKE_VAL_COMPLEX(-20, 55), MAKE_VAL_COMPLEX(-30, -45), MAKE_VAL_COMPLEX(-33, -50), MAKE_VAL_COMPLEX(99, 12), MAKE_VAL_COMPLEX(-6, 43), MAKE_VAL_COMPLEX(-62, 37), MAKE_VAL_COMPLEX(10, 63), MAKE_VAL_COMPLEX(-36, -31) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_560, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_560", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_560() */
/* END TEST 560 */

/*
 *  BEGIN TEST 561
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_561( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(134, 52), MAKE_VAL_COMPLEX(38, -11), MAKE_VAL_COMPLEX(-14, 25), MAKE_VAL_COMPLEX(-32, -10), MAKE_VAL_COMPLEX(7, -1), MAKE_VAL_COMPLEX(-100, 58), MAKE_VAL_COMPLEX(11, 36), MAKE_VAL_COMPLEX(-20, -13), MAKE_VAL_COMPLEX(25, -15), MAKE_VAL_COMPLEX(31, -3), MAKE_VAL_COMPLEX(-43, -7), MAKE_VAL_COMPLEX(-23, -14), MAKE_VAL_COMPLEX(0, -27), MAKE_VAL_COMPLEX(-33, -20), MAKE_VAL_COMPLEX(23, -9) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_561, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_561", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_561() */
/* END TEST 561 */

/*
 *  BEGIN TEST 562
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_562( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(12, 14), MAKE_VAL_COMPLEX(-76, -32), MAKE_VAL_COMPLEX(70, -60), MAKE_VAL_COMPLEX(81, 52), MAKE_VAL_COMPLEX(-19, -8), MAKE_VAL_COMPLEX(-28, 109), MAKE_VAL_COMPLEX(19, -77), MAKE_VAL_COMPLEX(32, 94), MAKE_VAL_COMPLEX(-95, -40), MAKE_VAL_COMPLEX(-82, -39), MAKE_VAL_COMPLEX(52, -81), MAKE_VAL_COMPLEX(21, 67), MAKE_VAL_COMPLEX(-34, 102), MAKE_VAL_COMPLEX(47, 69), MAKE_VAL_COMPLEX(-71, -12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_562, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_562", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_562() */
/* END TEST 562 */

/*
 *  BEGIN TEST 563
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_563( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_ROWMAJ, y528_VW );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(662, 394), MAKE_VAL_COMPLEX(25, 71), MAKE_VAL_COMPLEX(-95, -109), MAKE_VAL_COMPLEX(70, -148), MAKE_VAL_COMPLEX(-8, 24), MAKE_VAL_COMPLEX(-404, 410), MAKE_VAL_COMPLEX(-91, 37), MAKE_VAL_COMPLEX(133, 3), MAKE_VAL_COMPLEX(-80, 60), MAKE_VAL_COMPLEX(-40, 104), MAKE_VAL_COMPLEX(-274, -298), MAKE_VAL_COMPLEX(53, -27), MAKE_VAL_COMPLEX(159, 85), MAKE_VAL_COMPLEX(53, -111), MAKE_VAL_COMPLEX(-8, 88) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_563, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_563", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_563() */
/* END TEST 563 */

/*
 *  BEGIN TEST 564
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_564( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	oski_value_t* y_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-58, 24), MAKE_VAL_COMPLEX(46, -46), MAKE_VAL_COMPLEX(-48, 2), MAKE_VAL_COMPLEX(-50, 39), MAKE_VAL_COMPLEX(-11, -50), MAKE_VAL_COMPLEX(167, -29), MAKE_VAL_COMPLEX(34, 34), MAKE_VAL_COMPLEX(-46, -49), MAKE_VAL_COMPLEX(-51, -64), MAKE_VAL_COMPLEX(-117, 27), MAKE_VAL_COMPLEX(66, 70), MAKE_VAL_COMPLEX(51, 49) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_564, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_564", y_true, y_val, y528_LEN*y528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_564() */
/* END TEST 564 */

/*
 *  BEGIN TEST 565
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_565( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	oski_value_t* y_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-62, -34), MAKE_VAL_COMPLEX(49, 31), MAKE_VAL_COMPLEX(-42, -22), MAKE_VAL_COMPLEX(-85, -19), MAKE_VAL_COMPLEX(31, -65), MAKE_VAL_COMPLEX(-107, 34), MAKE_VAL_COMPLEX(-19, 36), MAKE_VAL_COMPLEX(31, -71), MAKE_VAL_COMPLEX(16, -66), MAKE_VAL_COMPLEX(62, 58), MAKE_VAL_COMPLEX(-16, 80), MAKE_VAL_COMPLEX(-21, 66) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_565, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_565", y_true, y_val, y528_LEN*y528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_565() */
/* END TEST 565 */

/*
 *  BEGIN TEST 566
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_566( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	oski_value_t* y_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-30, 70), MAKE_VAL_COMPLEX(10, -60), MAKE_VAL_COMPLEX(-30, 50), MAKE_VAL_COMPLEX(-15, 85), MAKE_VAL_COMPLEX(-65, -35), MAKE_VAL_COMPLEX(15, -35), MAKE_VAL_COMPLEX(45, 5), MAKE_VAL_COMPLEX(-75, -15), MAKE_VAL_COMPLEX(-85, -5), MAKE_VAL_COMPLEX(40, 50), MAKE_VAL_COMPLEX(100, 0), MAKE_VAL_COMPLEX(75, 5) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_566, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_566", y_true, y_val, y528_LEN*y528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_566() */
/* END TEST 566 */

/*
 *  BEGIN TEST 567
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_567( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	oski_value_t* y_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-102, -74), MAKE_VAL_COMPLEX(6, -28), MAKE_VAL_COMPLEX(-134, -78), MAKE_VAL_COMPLEX(-131, -107), MAKE_VAL_COMPLEX(143, -119), MAKE_VAL_COMPLEX(225, -825), MAKE_VAL_COMPLEX(27, 19), MAKE_VAL_COMPLEX(49, -47), MAKE_VAL_COMPLEX(-41, -137), MAKE_VAL_COMPLEX(-472, 666), MAKE_VAL_COMPLEX(52, 144), MAKE_VAL_COMPLEX(1, 67) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_567, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_567", y_true, y_val, y528_LEN*y528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_567() */
/* END TEST 567 */

/*
 *  BEGIN TEST 568
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_568( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-32, -106), MAKE_VAL_COMPLEX(83, 7), MAKE_VAL_COMPLEX(-58, 72), MAKE_VAL_COMPLEX(-95, -22), MAKE_VAL_COMPLEX(19, 3), MAKE_VAL_COMPLEX(8, -177), MAKE_VAL_COMPLEX(5, 72), MAKE_VAL_COMPLEX(-46, -73), MAKE_VAL_COMPLEX(89, 6), MAKE_VAL_COMPLEX(83, 17), MAKE_VAL_COMPLEX(47, 140), MAKE_VAL_COMPLEX(-35, -47), MAKE_VAL_COMPLEX(18, -90), MAKE_VAL_COMPLEX(-51, -44), MAKE_VAL_COMPLEX(67, -5) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_568, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_568", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_568() */
/* END TEST 568 */

/*
 *  BEGIN TEST 569
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_569( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(20, 90), MAKE_VAL_COMPLEX(-3, 33), MAKE_VAL_COMPLEX(-22, -12), MAKE_VAL_COMPLEX(11, -40), MAKE_VAL_COMPLEX(1, 7), MAKE_VAL_COMPLEX(30, 61), MAKE_VAL_COMPLEX(-31, 10), MAKE_VAL_COMPLEX(18, -31), MAKE_VAL_COMPLEX(11, 44), MAKE_VAL_COMPLEX(3, 31), MAKE_VAL_COMPLEX(-109, -56), MAKE_VAL_COMPLEX(17, -27), MAKE_VAL_COMPLEX(26, -18), MAKE_VAL_COMPLEX(5, -34), MAKE_VAL_COMPLEX(9, 23) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_569, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_569", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_569() */
/* END TEST 569 */

/*
 *  BEGIN TEST 570
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_570( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-6, 0), MAKE_VAL_COMPLEX(24, -12), MAKE_VAL_COMPLEX(0, 30), MAKE_VAL_COMPLEX(-30, 9), MAKE_VAL_COMPLEX(6, -3), MAKE_VAL_COMPLEX(-21, -30), MAKE_VAL_COMPLEX(15, 21), MAKE_VAL_COMPLEX(-30, -12), MAKE_VAL_COMPLEX(30, -15), MAKE_VAL_COMPLEX(27, -12), MAKE_VAL_COMPLEX(9, 30), MAKE_VAL_COMPLEX(-21, -9), MAKE_VAL_COMPLEX(-18, -30), MAKE_VAL_COMPLEX(-27, -3), MAKE_VAL_COMPLEX(18, -15) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_570, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_570", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_570() */
/* END TEST 570 */

/*
 *  BEGIN TEST 571
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_571( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-78, -374), MAKE_VAL_COMPLEX(32, -72), MAKE_VAL_COMPLEX(18, 38), MAKE_VAL_COMPLEX(-55, 87), MAKE_VAL_COMPLEX(1, -13), MAKE_VAL_COMPLEX(-43, -283), MAKE_VAL_COMPLEX(70, -12), MAKE_VAL_COMPLEX(-34, 58), MAKE_VAL_COMPLEX(-19, -101), MAKE_VAL_COMPLEX(7, -57), MAKE_VAL_COMPLEX(363, 235), MAKE_VAL_COMPLEX(-40, 62), MAKE_VAL_COMPLEX(-28, 40), MAKE_VAL_COMPLEX(-4, 74), MAKE_VAL_COMPLEX(-7, -45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_571, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_571", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_571() */
/* END TEST 571 */

/*
 *  BEGIN TEST 572
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_572( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(72, 32), MAKE_VAL_COMPLEX(71, 23), MAKE_VAL_COMPLEX(-68, 52), MAKE_VAL_COMPLEX(-78, -38), MAKE_VAL_COMPLEX(16, 7), MAKE_VAL_COMPLEX(72, -64), MAKE_VAL_COMPLEX(-12, 63), MAKE_VAL_COMPLEX(-22, -75), MAKE_VAL_COMPLEX(74, 26), MAKE_VAL_COMPLEX(69, 34), MAKE_VAL_COMPLEX(-149, 103), MAKE_VAL_COMPLEX(-18, -48), MAKE_VAL_COMPLEX(40, -76), MAKE_VAL_COMPLEX(-32, -51), MAKE_VAL_COMPLEX(60, 11) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_572, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_572", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_572() */
/* END TEST 572 */

/*
 *  BEGIN TEST 573
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_573( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-92, -40), MAKE_VAL_COMPLEX(25, -31), MAKE_VAL_COMPLEX(28, 48), MAKE_VAL_COMPLEX(-34, 28), MAKE_VAL_COMPLEX(8, -9), MAKE_VAL_COMPLEX(-102, -64), MAKE_VAL_COMPLEX(34, 27), MAKE_VAL_COMPLEX(-62, -5), MAKE_VAL_COMPLEX(46, -36), MAKE_VAL_COMPLEX(37, -38), MAKE_VAL_COMPLEX(139, 9), MAKE_VAL_COMPLEX(-40, -10), MAKE_VAL_COMPLEX(-60, -48), MAKE_VAL_COMPLEX(-54, 5), MAKE_VAL_COMPLEX(20, -37) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_573, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_573", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_573() */
/* END TEST 573 */

/*
 *  BEGIN TEST 574
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_574( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(16, 12), MAKE_VAL_COMPLEX(-88, -16), MAKE_VAL_COMPLEX(60, -80), MAKE_VAL_COMPLEX(98, 36), MAKE_VAL_COMPLEX(-22, -4), MAKE_VAL_COMPLEX(-4, 122), MAKE_VAL_COMPLEX(2, -86), MAKE_VAL_COMPLEX(56, 92), MAKE_VAL_COMPLEX(-110, -20), MAKE_VAL_COMPLEX(-96, -22), MAKE_VAL_COMPLEX(36, -98), MAKE_VAL_COMPLEX(38, 66), MAKE_VAL_COMPLEX(-12, 116), MAKE_VAL_COMPLEX(66, 62), MAKE_VAL_COMPLEX(-78, 4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_574, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_574", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_574() */
/* END TEST 574 */

/*
 *  BEGIN TEST 575
 *  Value type: complex
 *  A : <MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x528', LAYOUT_COLMAJ>
 *  y : <'y528', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_575( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A528_ptr, A528_ind, A528_val,
		A528_M, A528_N, SHARE_INPUTMAT,
		5, MAT_TRI_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y528_val,
		y528_LEN * y528_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y528_LEN, y528_VW, LAYOUT_COLMAJ, y528_LEN );
	oski_value_t* y_val = array_Duplicate( x528_val,
		x528_LEN * x528_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x528_LEN, x528_VW, LAYOUT_COLMAJ, x528_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-784, 318), MAKE_VAL_COMPLEX(-87, 79), MAKE_VAL_COMPLEX(74, -148), MAKE_VAL_COMPLEX(103, -96), MAKE_VAL_COMPLEX(-20, 0), MAKE_VAL_COMPLEX(-462, 269), MAKE_VAL_COMPLEX(-28, -29), MAKE_VAL_COMPLEX(5, 70), MAKE_VAL_COMPLEX(-1, 12), MAKE_VAL_COMPLEX(-88, -4), MAKE_VAL_COMPLEX(395, -1030), MAKE_VAL_COMPLEX(-13, -9), MAKE_VAL_COMPLEX(-122, 114), MAKE_VAL_COMPLEX(-30, 45), MAKE_VAL_COMPLEX(-68, 16) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_575, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_575", y_true, y_val, x528_LEN*x528_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_575() */
/* END TEST 575 */

#define A576_M 5    /* # of rows */
#define A576_N 5    /* # of cols */
#define A576_NNZ 12  /* # of stored non-zeros */
static oski_index_t A576_ptr[A576_M+1] = {1, 4, 5, 8, 11, 13}; /* end row pointers for A576 */
static oski_index_t A576_ind[A576_NNZ] = {2, 3, 4, 1, 1, 4, 5, 1, 3, 5, 3, 4}; /* end column indices pointers for A576 */
static oski_value_t A576_val[A576_NNZ] = {MAKE_VAL_COMPLEX(6, 10), MAKE_VAL_COMPLEX(-5, -6), MAKE_VAL_COMPLEX(10, 3), MAKE_VAL_COMPLEX(6, 10), MAKE_VAL_COMPLEX(-5, -6), MAKE_VAL_COMPLEX(6, -7), MAKE_VAL_COMPLEX(0, -6), MAKE_VAL_COMPLEX(10, 3), MAKE_VAL_COMPLEX(6, -7), MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(0, -6), MAKE_VAL_COMPLEX(-10, -10)}; /* end stored non-zero values for A576 */

#define x576_LEN 5   /* vector length */
#define x576_VW  3   /* vector width */
static oski_value_t x576_val[x576_LEN*x576_VW] = {MAKE_VAL_COMPLEX(3, -9), MAKE_VAL_COMPLEX(-6, 3), MAKE_VAL_COMPLEX(-7, -6), MAKE_VAL_COMPLEX(-10, -6), MAKE_VAL_COMPLEX(8, -10), MAKE_VAL_COMPLEX(-3, 3), MAKE_VAL_COMPLEX(1, 7), MAKE_VAL_COMPLEX(9, -1), MAKE_VAL_COMPLEX(-7, 10), MAKE_VAL_COMPLEX(-1, 9), MAKE_VAL_COMPLEX(-3, 0), MAKE_VAL_COMPLEX(5, -4), MAKE_VAL_COMPLEX(10, 5), MAKE_VAL_COMPLEX(-2, 7), MAKE_VAL_COMPLEX(9, 1)}; /* data */

#define y576_LEN 5   /* vector length */
#define y576_VW  3   /* vector width */
static oski_value_t y576_val[y576_LEN*y576_VW] = {MAKE_VAL_COMPLEX(10, 7), MAKE_VAL_COMPLEX(3, 8), MAKE_VAL_COMPLEX(-2, -2), MAKE_VAL_COMPLEX(-7, -10), MAKE_VAL_COMPLEX(-6, 2), MAKE_VAL_COMPLEX(-2, 9), MAKE_VAL_COMPLEX(-2, 10), MAKE_VAL_COMPLEX(9, 10), MAKE_VAL_COMPLEX(2, 3), MAKE_VAL_COMPLEX(7, -2), MAKE_VAL_COMPLEX(2, -7), MAKE_VAL_COMPLEX(4, 9), MAKE_VAL_COMPLEX(0, 4), MAKE_VAL_COMPLEX(1, 5), MAKE_VAL_COMPLEX(-10, 6)}; /* data */

/*
 *  BEGIN TEST 576
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_576( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	oski_value_t* y_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(25, -147), MAKE_VAL_COMPLEX(104, -61), MAKE_VAL_COMPLEX(101, -33), MAKE_VAL_COMPLEX(65, 21), MAKE_VAL_COMPLEX(-50, -14), MAKE_VAL_COMPLEX(65, -105), MAKE_VAL_COMPLEX(70, 28), MAKE_VAL_COMPLEX(113, -1), MAKE_VAL_COMPLEX(20, -54), MAKE_VAL_COMPLEX(45, -229), MAKE_VAL_COMPLEX(31, -110), MAKE_VAL_COMPLEX(-63, -101), MAKE_VAL_COMPLEX(162, -90), MAKE_VAL_COMPLEX(48, -34), MAKE_VAL_COMPLEX(10, 76) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_576, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_576", y_true, y_val, y576_LEN*y576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_576() */
/* END TEST 576 */

/*
 *  BEGIN TEST 577
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_577( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	oski_value_t* y_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-98, 81), MAKE_VAL_COMPLEX(-120, -6), MAKE_VAL_COMPLEX(-87, 51), MAKE_VAL_COMPLEX(-19, 66), MAKE_VAL_COMPLEX(100, 4), MAKE_VAL_COMPLEX(-40, 35), MAKE_VAL_COMPLEX(-44, -106), MAKE_VAL_COMPLEX(-175, -88), MAKE_VAL_COMPLEX(-33, 28), MAKE_VAL_COMPLEX(-103, 238), MAKE_VAL_COMPLEX(-54, 164), MAKE_VAL_COMPLEX(40, 25), MAKE_VAL_COMPLEX(-158, 58), MAKE_VAL_COMPLEX(-51, -7), MAKE_VAL_COMPLEX(76, -114) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_577, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_577", y_true, y_val, y576_LEN*y576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_577() */
/* END TEST 577 */

/*
 *  BEGIN TEST 578
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_578( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	oski_value_t* y_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(25, 92), MAKE_VAL_COMPLEX(-22, 63), MAKE_VAL_COMPLEX(-2, -22), MAKE_VAL_COMPLEX(8, -95), MAKE_VAL_COMPLEX(-46, -18), MAKE_VAL_COMPLEX(-57, 44), MAKE_VAL_COMPLEX(-62, 50), MAKE_VAL_COMPLEX(4, 105), MAKE_VAL_COMPLEX(-3, 28), MAKE_VAL_COMPLEX(52, 23), MAKE_VAL_COMPLEX(47, -32), MAKE_VAL_COMPLEX(-21, 74), MAKE_VAL_COMPLEX(-20, 24), MAKE_VAL_COMPLEX(-19, 35), MAKE_VAL_COMPLEX(-90, -14) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_578, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_578", y_true, y_val, y576_LEN*y576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_578() */
/* END TEST 578 */

/*
 *  BEGIN TEST 579
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_579( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	oski_value_t* y_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-309, -9), MAKE_VAL_COMPLEX(-119, -253), MAKE_VAL_COMPLEX(-192, -436), MAKE_VAL_COMPLEX(-147, -441), MAKE_VAL_COMPLEX(-180, 288), MAKE_VAL_COMPLEX(-403, -39), MAKE_VAL_COMPLEX(136, -36), MAKE_VAL_COMPLEX(273, -285), MAKE_VAL_COMPLEX(-149, -25), MAKE_VAL_COMPLEX(-769, -275), MAKE_VAL_COMPLEX(-443, -299), MAKE_VAL_COMPLEX(-249, 431), MAKE_VAL_COMPLEX(-332, -556), MAKE_VAL_COMPLEX(-78, -84), MAKE_VAL_COMPLEX(116, 168) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, -4);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_579, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_579", y_true, y_val, y576_LEN*y576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_579() */
/* END TEST 579 */

/*
 *  BEGIN TEST 580
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_580( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(243, -224), MAKE_VAL_COMPLEX(-48, -207), MAKE_VAL_COMPLEX(-118, 53), MAKE_VAL_COMPLEX(-68, 80), MAKE_VAL_COMPLEX(14, 24), MAKE_VAL_COMPLEX(-19, -17), MAKE_VAL_COMPLEX(37, -105), MAKE_VAL_COMPLEX(91, -109), MAKE_VAL_COMPLEX(52, 164), MAKE_VAL_COMPLEX(152, 195), MAKE_VAL_COMPLEX(149, 20), MAKE_VAL_COMPLEX(222, 0), MAKE_VAL_COMPLEX(30, 17), MAKE_VAL_COMPLEX(-58, 41), MAKE_VAL_COMPLEX(129, -117) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_580, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_580", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_580() */
/* END TEST 580 */

/*
 *  BEGIN TEST 581
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_581( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-252, 71), MAKE_VAL_COMPLEX(-24, 288), MAKE_VAL_COMPLEX(-23, -101), MAKE_VAL_COMPLEX(-118, -110), MAKE_VAL_COMPLEX(46, -222), MAKE_VAL_COMPLEX(-8, 80), MAKE_VAL_COMPLEX(20, 204), MAKE_VAL_COMPLEX(38, 40), MAKE_VAL_COMPLEX(-97, 28), MAKE_VAL_COMPLEX(-113, -54), MAKE_VAL_COMPLEX(-194, -2), MAKE_VAL_COMPLEX(-171, -90), MAKE_VAL_COMPLEX(150, -2), MAKE_VAL_COMPLEX(70, 76), MAKE_VAL_COMPLEX(12, 78) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_581, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_581", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_581() */
/* END TEST 581 */

/*
 *  BEGIN TEST 582
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_582( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(18, 36), MAKE_VAL_COMPLEX(9, -27), MAKE_VAL_COMPLEX(39, -3), MAKE_VAL_COMPLEX(48, -12), MAKE_VAL_COMPLEX(6, 54), MAKE_VAL_COMPLEX(0, -18), MAKE_VAL_COMPLEX(-24, -18), MAKE_VAL_COMPLEX(-24, 30), MAKE_VAL_COMPLEX(-9, -51), MAKE_VAL_COMPLEX(-24, -30), MAKE_VAL_COMPLEX(9, -9), MAKE_VAL_COMPLEX(-3, 27), MAKE_VAL_COMPLEX(-45, 15), MAKE_VAL_COMPLEX(-15, -27), MAKE_VAL_COMPLEX(-30, 24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_582, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_582", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_582() */
/* END TEST 582 */

/*
 *  BEGIN TEST 583
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_583( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(101, 1392), MAKE_VAL_COMPLEX(879, 588), MAKE_VAL_COMPLEX(-109, -664), MAKE_VAL_COMPLEX(-436, -500), MAKE_VAL_COMPLEX(-104, -366), MAKE_VAL_COMPLEX(101, 89), MAKE_VAL_COMPLEX(437, 609), MAKE_VAL_COMPLEX(355, 533), MAKE_VAL_COMPLEX(-823, 51), MAKE_VAL_COMPLEX(-1124, 245), MAKE_VAL_COMPLEX(-596, 581), MAKE_VAL_COMPLEX(-611, 721), MAKE_VAL_COMPLEX(147, 34), MAKE_VAL_COMPLEX(69, -164), MAKE_VAL_COMPLEX(303, 755) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-3, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_583, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_583", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_583() */
/* END TEST 583 */

/*
 *  BEGIN TEST 584
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_584( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-23, -138), MAKE_VAL_COMPLEX(-197, -30), MAKE_VAL_COMPLEX(115, 38), MAKE_VAL_COMPLEX(104, -196), MAKE_VAL_COMPLEX(252, 10), MAKE_VAL_COMPLEX(-83, 5), MAKE_VAL_COMPLEX(-115, 127), MAKE_VAL_COMPLEX(49, 29), MAKE_VAL_COMPLEX(-199, 37), MAKE_VAL_COMPLEX(-90, 109), MAKE_VAL_COMPLEX(-46, 127), MAKE_VAL_COMPLEX(81, -129), MAKE_VAL_COMPLEX(-75, 208), MAKE_VAL_COMPLEX(-89, 182), MAKE_VAL_COMPLEX(-85, 51) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_584, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_584", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_584() */
/* END TEST 584 */

/*
 *  BEGIN TEST 585
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_585( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(188, 153), MAKE_VAL_COMPLEX(122, -60), MAKE_VAL_COMPLEX(-41, -181), MAKE_VAL_COMPLEX(-42, 2), MAKE_VAL_COMPLEX(-50, 86), MAKE_VAL_COMPLEX(20, -44), MAKE_VAL_COMPLEX(0, -82), MAKE_VAL_COMPLEX(4, 120), MAKE_VAL_COMPLEX(1, -116), MAKE_VAL_COMPLEX(-67, -90), MAKE_VAL_COMPLEX(34, -178), MAKE_VAL_COMPLEX(7, 198), MAKE_VAL_COMPLEX(30, -18), MAKE_VAL_COMPLEX(-38, -188), MAKE_VAL_COMPLEX(104, 106) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_585, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_585", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_585() */
/* END TEST 585 */

/*
 *  BEGIN TEST 586
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_586( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-57, -39), MAKE_VAL_COMPLEX(9, 48), MAKE_VAL_COMPLEX(-56, 37), MAKE_VAL_COMPLEX(-62, 58), MAKE_VAL_COMPLEX(-54, -76), MAKE_VAL_COMPLEX(15, 27), MAKE_VAL_COMPLEX(51, 7), MAKE_VAL_COMPLEX(11, -65), MAKE_VAL_COMPLEX(56, 69), MAKE_VAL_COMPLEX(61, 25), MAKE_VAL_COMPLEX(-6, 21), MAKE_VAL_COMPLEX(-18, -43), MAKE_VAL_COMPLEX(55, -60), MAKE_VAL_COMPLEX(45, 28), MAKE_VAL_COMPLEX(25, -61) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_586, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_586", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_586() */
/* END TEST 586 */

/*
 *  BEGIN TEST 587
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_587( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-768, 1066), MAKE_VAL_COMPLEX(33, 967), MAKE_VAL_COMPLEX(1221, -899), MAKE_VAL_COMPLEX(-420, -1000), MAKE_VAL_COMPLEX(94, -742), MAKE_VAL_COMPLEX(82, 244), MAKE_VAL_COMPLEX(498, 444), MAKE_VAL_COMPLEX(-398, 252), MAKE_VAL_COMPLEX(205, 381), MAKE_VAL_COMPLEX(382, -36), MAKE_VAL_COMPLEX(1247, 179), MAKE_VAL_COMPLEX(-1165, -25), MAKE_VAL_COMPLEX(579, 845), MAKE_VAL_COMPLEX(1169, 47), MAKE_VAL_COMPLEX(-348, 1170) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_587, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_587", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_587() */
/* END TEST 587 */

/*
 *  BEGIN TEST 588
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_588( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	oski_value_t* y_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(4, 32), MAKE_VAL_COMPLEX(62, 48), MAKE_VAL_COMPLEX(22, 0), MAKE_VAL_COMPLEX(100, -312), MAKE_VAL_COMPLEX(90, -122), MAKE_VAL_COMPLEX(-17, 0), MAKE_VAL_COMPLEX(-158, 2), MAKE_VAL_COMPLEX(46, 186), MAKE_VAL_COMPLEX(56, -73), MAKE_VAL_COMPLEX(82, 20), MAKE_VAL_COMPLEX(177, -71), MAKE_VAL_COMPLEX(-30, -20), MAKE_VAL_COMPLEX(-25, -17), MAKE_VAL_COMPLEX(-138, -34), MAKE_VAL_COMPLEX(-138, -12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_588, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_588", y_true, y_val, y576_LEN*y576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_588() */
/* END TEST 588 */

/*
 *  BEGIN TEST 589
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_589( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	oski_value_t* y_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(38, 57), MAKE_VAL_COMPLEX(-73, 20), MAKE_VAL_COMPLEX(-28, -22), MAKE_VAL_COMPLEX(-109, 214), MAKE_VAL_COMPLEX(-140, 112), MAKE_VAL_COMPLEX(-33, 55), MAKE_VAL_COMPLEX(104, 60), MAKE_VAL_COMPLEX(-23, -80), MAKE_VAL_COMPLEX(-54, 102), MAKE_VAL_COMPLEX(-25, -6), MAKE_VAL_COMPLEX(-135, 30), MAKE_VAL_COMPLEX(22, 99), MAKE_VAL_COMPLEX(9, 45), MAKE_VAL_COMPLEX(125, 73), MAKE_VAL_COMPLEX(44, 14) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_589, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_589", y_true, y_val, y576_LEN*y576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_589() */
/* END TEST 589 */

/*
 *  BEGIN TEST 590
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_590( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	oski_value_t* y_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(115, 6), MAKE_VAL_COMPLEX(64, 49), MAKE_VAL_COMPLEX(-26, -6), MAKE_VAL_COMPLEX(-106, -45), MAKE_VAL_COMPLEX(-38, 46), MAKE_VAL_COMPLEX(29, 82), MAKE_VAL_COMPLEX(34, 90), MAKE_VAL_COMPLEX(122, 35), MAKE_VAL_COMPLEX(31, 14), MAKE_VAL_COMPLEX(46, -51), MAKE_VAL_COMPLEX(-19, -66), MAKE_VAL_COMPLEX(77, 52), MAKE_VAL_COMPLEX(20, 32), MAKE_VAL_COMPLEX(33, 35), MAKE_VAL_COMPLEX(-50, 98) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_590, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_590", y_true, y_val, y576_LEN*y576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_590() */
/* END TEST 590 */

/*
 *  BEGIN TEST 591
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_591( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	oski_value_t* y_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-479, 624), MAKE_VAL_COMPLEX(-818, -359), MAKE_VAL_COMPLEX(42, -276), MAKE_VAL_COMPLEX(-1296, 797), MAKE_VAL_COMPLEX(-1308, -232), MAKE_VAL_COMPLEX(-659, -103), MAKE_VAL_COMPLEX(112, 588), MAKE_VAL_COMPLEX(-214, -601), MAKE_VAL_COMPLEX(-960, 322), MAKE_VAL_COMPLEX(-236, 69), MAKE_VAL_COMPLEX(-826, -275), MAKE_VAL_COMPLEX(-693, 592), MAKE_VAL_COMPLEX(-275, 219), MAKE_VAL_COMPLEX(231, 969), MAKE_VAL_COMPLEX(278, -100) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_591, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_591", y_true, y_val, y576_LEN*y576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_591() */
/* END TEST 591 */

/*
 *  BEGIN TEST 592
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_592( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-71, -60), MAKE_VAL_COMPLEX(-122, -25), MAKE_VAL_COMPLEX(-70, 83), MAKE_VAL_COMPLEX(-48, 92), MAKE_VAL_COMPLEX(-42, 0), MAKE_VAL_COMPLEX(61, -13), MAKE_VAL_COMPLEX(-117, -33), MAKE_VAL_COMPLEX(132, -58), MAKE_VAL_COMPLEX(-30, 142), MAKE_VAL_COMPLEX(110, 185), MAKE_VAL_COMPLEX(-28, 25), MAKE_VAL_COMPLEX(262, -10), MAKE_VAL_COMPLEX(-2, 227), MAKE_VAL_COMPLEX(46, -73), MAKE_VAL_COMPLEX(107, -37) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_592, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_592", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_592() */
/* END TEST 592 */

/*
 *  BEGIN TEST 593
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_593( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(77, 42), MAKE_VAL_COMPLEX(110, 31), MAKE_VAL_COMPLEX(56, -95), MAKE_VAL_COMPLEX(28, -104), MAKE_VAL_COMPLEX(58, -20), MAKE_VAL_COMPLEX(-67, 19), MAKE_VAL_COMPLEX(119, 47), MAKE_VAL_COMPLEX(-114, 56), MAKE_VAL_COMPLEX(16, -122), MAKE_VAL_COMPLEX(-112, -167), MAKE_VAL_COMPLEX(22, -25), MAKE_VAL_COMPLEX(-252, 2), MAKE_VAL_COMPLEX(22, -217), MAKE_VAL_COMPLEX(-50, 87), MAKE_VAL_COMPLEX(-89, 39) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_593, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_593", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_593() */
/* END TEST 593 */

/*
 *  BEGIN TEST 594
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_594( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-51, -87), MAKE_VAL_COMPLEX(-18, 69), MAKE_VAL_COMPLEX(-97, 14), MAKE_VAL_COMPLEX(-118, 38), MAKE_VAL_COMPLEX(-24, -134), MAKE_VAL_COMPLEX(3, 45), MAKE_VAL_COMPLEX(63, 41), MAKE_VAL_COMPLEX(55, -79), MAKE_VAL_COMPLEX(31, 126), MAKE_VAL_COMPLEX(65, 71), MAKE_VAL_COMPLEX(-21, 24), MAKE_VAL_COMPLEX(3, -68), MAKE_VAL_COMPLEX(110, -45), MAKE_VAL_COMPLEX(42, 65), MAKE_VAL_COMPLEX(71, -65) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_594, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_594", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_594() */
/* END TEST 594 */

/*
 *  BEGIN TEST 595
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_595( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(803, -212), MAKE_VAL_COMPLEX(747, -172), MAKE_VAL_COMPLEX(-131, -1048), MAKE_VAL_COMPLEX(-432, -1008), MAKE_VAL_COMPLEX(540, -578), MAKE_VAL_COMPLEX(-471, 461), MAKE_VAL_COMPLEX(1023, 17), MAKE_VAL_COMPLEX(-364, 852), MAKE_VAL_COMPLEX(-567, -709), MAKE_VAL_COMPLEX(-1502, -517), MAKE_VAL_COMPLEX(-15, -272), MAKE_VAL_COMPLEX(-1629, 945), MAKE_VAL_COMPLEX(-499, -1462), MAKE_VAL_COMPLEX(-69, 964), MAKE_VAL_COMPLEX(-263, 679) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_595, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_595", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_595() */
/* END TEST 595 */

/*
 *  BEGIN TEST 596
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_596( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-28, -159), MAKE_VAL_COMPLEX(-12, 180), MAKE_VAL_COMPLEX(11, 49), MAKE_VAL_COMPLEX(2, -26), MAKE_VAL_COMPLEX(62, -70), MAKE_VAL_COMPLEX(-58, -14), MAKE_VAL_COMPLEX(-12, -72), MAKE_VAL_COMPLEX(23, -63), MAKE_VAL_COMPLEX(-9, 160), MAKE_VAL_COMPLEX(227, 14), MAKE_VAL_COMPLEX(-83, 333), MAKE_VAL_COMPLEX(19, -284), MAKE_VAL_COMPLEX(302, -22), MAKE_VAL_COMPLEX(-70, 116), MAKE_VAL_COMPLEX(-4, -104) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_596, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_596", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_596() */
/* END TEST 596 */

/*
 *  BEGIN TEST 597
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_597( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-71, 126), MAKE_VAL_COMPLEX(45, -114), MAKE_VAL_COMPLEX(-77, 28), MAKE_VAL_COMPLEX(-68, 136), MAKE_VAL_COMPLEX(-172, -18), MAKE_VAL_COMPLEX(91, 47), MAKE_VAL_COMPLEX(89, 61), MAKE_VAL_COMPLEX(-34, -36), MAKE_VAL_COMPLEX(119, -83), MAKE_VAL_COMPLEX(-128, -3), MAKE_VAL_COMPLEX(83, -300), MAKE_VAL_COMPLEX(-63, 229), MAKE_VAL_COMPLEX(-247, -88), MAKE_VAL_COMPLEX(147, -94), MAKE_VAL_COMPLEX(15, 5) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_597, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_597", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_597() */
/* END TEST 597 */

/*
 *  BEGIN TEST 598
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_598( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-15, 75), MAKE_VAL_COMPLEX(45, -30), MAKE_VAL_COMPLEX(62, 41), MAKE_VAL_COMPLEX(86, 38), MAKE_VAL_COMPLEX(-54, 88), MAKE_VAL_COMPLEX(21, -27), MAKE_VAL_COMPLEX(-15, -55), MAKE_VAL_COMPLEX(-71, 17), MAKE_VAL_COMPLEX(46, -87), MAKE_VAL_COMPLEX(-1, -73), MAKE_VAL_COMPLEX(24, -3), MAKE_VAL_COMPLEX(-36, 37), MAKE_VAL_COMPLEX(-85, -30), MAKE_VAL_COMPLEX(9, -58), MAKE_VAL_COMPLEX(-73, 1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_598, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_598", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_598() */
/* END TEST 598 */

/*
 *  BEGIN TEST 599
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_ROWMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_599( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_ROWMAJ, x576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(595, 121), MAKE_VAL_COMPLEX(-822, 243), MAKE_VAL_COMPLEX(-1, 988), MAKE_VAL_COMPLEX(842, 886), MAKE_VAL_COMPLEX(-412, 906), MAKE_VAL_COMPLEX(377, -633), MAKE_VAL_COMPLEX(745, -889), MAKE_VAL_COMPLEX(-273, -215), MAKE_VAL_COMPLEX(-289, -342), MAKE_VAL_COMPLEX(789, 1101), MAKE_VAL_COMPLEX(-2569, 122), MAKE_VAL_COMPLEX(1715, -268), MAKE_VAL_COMPLEX(110, 1537), MAKE_VAL_COMPLEX(-614, -857), MAKE_VAL_COMPLEX(81, -687) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_599, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_599", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_599() */
/* END TEST 599 */

/*
 *  BEGIN TEST 600
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_600( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	oski_value_t* y_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-155, 55), MAKE_VAL_COMPLEX(-264, 146), MAKE_VAL_COMPLEX(15, -21), MAKE_VAL_COMPLEX(153, -130), MAKE_VAL_COMPLEX(-94, -50), MAKE_VAL_COMPLEX(-100, -1), MAKE_VAL_COMPLEX(-321, 47), MAKE_VAL_COMPLEX(80, 240), MAKE_VAL_COMPLEX(44, 51), MAKE_VAL_COMPLEX(-156, -2), MAKE_VAL_COMPLEX(174, -147), MAKE_VAL_COMPLEX(-67, -72), MAKE_VAL_COMPLEX(-28, 222), MAKE_VAL_COMPLEX(129, -51), MAKE_VAL_COMPLEX(22, -160) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_600, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_600", y_true, y_val, y576_LEN*y576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_600() */
/* END TEST 600 */

/*
 *  BEGIN TEST 601
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_601( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	oski_value_t* y_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(101, 86), MAKE_VAL_COMPLEX(177, -86), MAKE_VAL_COMPLEX(3, -9), MAKE_VAL_COMPLEX(-54, 16), MAKE_VAL_COMPLEX(52, -16), MAKE_VAL_COMPLEX(-14, 4), MAKE_VAL_COMPLEX(195, -41), MAKE_VAL_COMPLEX(-173, -102), MAKE_VAL_COMPLEX(-74, -18), MAKE_VAL_COMPLEX(201, 80), MAKE_VAL_COMPLEX(-84, 150), MAKE_VAL_COMPLEX(-29, 147), MAKE_VAL_COMPLEX(-20, -210), MAKE_VAL_COMPLEX(-186, 78), MAKE_VAL_COMPLEX(-124, 58) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_601, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_601", y_true, y_val, y576_LEN*y576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_601() */
/* END TEST 601 */

/*
 *  BEGIN TEST 602
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_602( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	oski_value_t* y_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-24, -136), MAKE_VAL_COMPLEX(40, -88), MAKE_VAL_COMPLEX(0, 32), MAKE_VAL_COMPLEX(-24, 136), MAKE_VAL_COMPLEX(64, 32), MAKE_VAL_COMPLEX(88, -56), MAKE_VAL_COMPLEX(96, -64), MAKE_VAL_COMPLEX(8, -152), MAKE_VAL_COMPLEX(8, -40), MAKE_VAL_COMPLEX(-72, -40), MAKE_VAL_COMPLEX(-72, 40), MAKE_VAL_COMPLEX(40, -104), MAKE_VAL_COMPLEX(32, -32), MAKE_VAL_COMPLEX(32, -48), MAKE_VAL_COMPLEX(128, 32) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_602, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_602", y_true, y_val, y576_LEN*y576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_602() */
/* END TEST 602 */

/*
 *  BEGIN TEST 603
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_603( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	oski_value_t* y_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-317, 1232), MAKE_VAL_COMPLEX(887, 1632), MAKE_VAL_COMPLEX(27, -73), MAKE_VAL_COMPLEX(-320, -826), MAKE_VAL_COMPLEX(-44, 408), MAKE_VAL_COMPLEX(-204, 178), MAKE_VAL_COMPLEX(355, 1839), MAKE_VAL_COMPLEX(849, -1056), MAKE_VAL_COMPLEX(108, -488), MAKE_VAL_COMPLEX(-181, 1690), MAKE_VAL_COMPLEX(-1146, -740), MAKE_VAL_COMPLEX(-1159, 261), MAKE_VAL_COMPLEX(1620, -234), MAKE_VAL_COMPLEX(-826, -1230), MAKE_VAL_COMPLEX(-988, -830) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_603, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_603", y_true, y_val, y576_LEN*y576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_603() */
/* END TEST 603 */

/*
 *  BEGIN TEST 604
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_604( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(309, -212), MAKE_VAL_COMPLEX(-85, 112), MAKE_VAL_COMPLEX(42, -267), MAKE_VAL_COMPLEX(151, -4), MAKE_VAL_COMPLEX(124, -46), MAKE_VAL_COMPLEX(-51, -219), MAKE_VAL_COMPLEX(-117, 143), MAKE_VAL_COMPLEX(107, -47), MAKE_VAL_COMPLEX(24, 43), MAKE_VAL_COMPLEX(-119, 59), MAKE_VAL_COMPLEX(-105, 82), MAKE_VAL_COMPLEX(84, -19), MAKE_VAL_COMPLEX(156, 238), MAKE_VAL_COMPLEX(100, 56), MAKE_VAL_COMPLEX(131, -53) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_604, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_604", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_604() */
/* END TEST 604 */

/*
 *  BEGIN TEST 605
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_605( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-195, 200), MAKE_VAL_COMPLEX(22, -163), MAKE_VAL_COMPLEX(-11, 160), MAKE_VAL_COMPLEX(-135, -136), MAKE_VAL_COMPLEX(26, 84), MAKE_VAL_COMPLEX(3, 201), MAKE_VAL_COMPLEX(45, -97), MAKE_VAL_COMPLEX(-51, 141), MAKE_VAL_COMPLEX(-169, -70), MAKE_VAL_COMPLEX(15, -25), MAKE_VAL_COMPLEX(90, -115), MAKE_VAL_COMPLEX(-15, 54), MAKE_VAL_COMPLEX(-161, -103), MAKE_VAL_COMPLEX(-187, -43), MAKE_VAL_COMPLEX(-97, 157) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_605, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_605", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_605() */
/* END TEST 605 */

/*
 *  BEGIN TEST 606
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_606( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(33, -39), MAKE_VAL_COMPLEX(-36, 3), MAKE_VAL_COMPLEX(-23, -44), MAKE_VAL_COMPLEX(-38, -50), MAKE_VAL_COMPLEX(60, -34), MAKE_VAL_COMPLEX(-21, 9), MAKE_VAL_COMPLEX(-9, 37), MAKE_VAL_COMPLEX(47, 13), MAKE_VAL_COMPLEX(-55, 36), MAKE_VAL_COMPLEX(-23, 43), MAKE_VAL_COMPLEX(-15, -6), MAKE_VAL_COMPLEX(33, -10), MAKE_VAL_COMPLEX(40, 45), MAKE_VAL_COMPLEX(-24, 31), MAKE_VAL_COMPLEX(43, 23) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_606, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_606", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_606() */
/* END TEST 606 */

/*
 *  BEGIN TEST 607
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_607( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1467, 1794), MAKE_VAL_COMPLEX(-1281, -51), MAKE_VAL_COMPLEX(1353, 413), MAKE_VAL_COMPLEX(-1266, 1625), MAKE_VAL_COMPLEX(316, -340), MAKE_VAL_COMPLEX(1950, 24), MAKE_VAL_COMPLEX(-664, -542), MAKE_VAL_COMPLEX(1102, 186), MAKE_VAL_COMPLEX(-205, 1595), MAKE_VAL_COMPLEX(78, -238), MAKE_VAL_COMPLEX(-990, -714), MAKE_VAL_COMPLEX(283, 35), MAKE_VAL_COMPLEX(-917, 1054), MAKE_VAL_COMPLEX(-133, 1642), MAKE_VAL_COMPLEX(1310, 570) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 9);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_607, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_607", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_607() */
/* END TEST 607 */

/*
 *  BEGIN TEST 608
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_608( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-176, -129), MAKE_VAL_COMPLEX(136, -16), MAKE_VAL_COMPLEX(-110, 92), MAKE_VAL_COMPLEX(-57, 94), MAKE_VAL_COMPLEX(-154, 10), MAKE_VAL_COMPLEX(-110, 24), MAKE_VAL_COMPLEX(142, 26), MAKE_VAL_COMPLEX(-20, -100), MAKE_VAL_COMPLEX(24, 216), MAKE_VAL_COMPLEX(42, 168), MAKE_VAL_COMPLEX(111, 167), MAKE_VAL_COMPLEX(-46, -30), MAKE_VAL_COMPLEX(-3, -30), MAKE_VAL_COMPLEX(43, -116), MAKE_VAL_COMPLEX(-124, -90) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_608, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_608", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_608() */
/* END TEST 608 */

/*
 *  BEGIN TEST 609
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_609( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(104, 15), MAKE_VAL_COMPLEX(-157, 109), MAKE_VAL_COMPLEX(-19, -69), MAKE_VAL_COMPLEX(-99, -38), MAKE_VAL_COMPLEX(116, -188), MAKE_VAL_COMPLEX(116, 36), MAKE_VAL_COMPLEX(-56, 26), MAKE_VAL_COMPLEX(90, -8), MAKE_VAL_COMPLEX(23, -49), MAKE_VAL_COMPLEX(48, -76), MAKE_VAL_COMPLEX(-138, -134), MAKE_VAL_COMPLEX(47, -61), MAKE_VAL_COMPLEX(148, -35), MAKE_VAL_COMPLEX(16, 201), MAKE_VAL_COMPLEX(216, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_609, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_609", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_609() */
/* END TEST 609 */

/*
 *  BEGIN TEST 610
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_610( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-54, -18), MAKE_VAL_COMPLEX(18, 36), MAKE_VAL_COMPLEX(-36, 42), MAKE_VAL_COMPLEX(-36, 60), MAKE_VAL_COMPLEX(-60, -48), MAKE_VAL_COMPLEX(18, 18), MAKE_VAL_COMPLEX(42, -6), MAKE_VAL_COMPLEX(-6, -54), MAKE_VAL_COMPLEX(60, 42), MAKE_VAL_COMPLEX(54, 6), MAKE_VAL_COMPLEX(0, 18), MAKE_VAL_COMPLEX(-24, -30), MAKE_VAL_COMPLEX(30, -60), MAKE_VAL_COMPLEX(42, 12), MAKE_VAL_COMPLEX(6, -54) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_610, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_610", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_610() */
/* END TEST 610 */

/*
 *  BEGIN TEST 611
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_611( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_ROWMAJ, y576_VW );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(289, -739), MAKE_VAL_COMPLEX(441, 619), MAKE_VAL_COMPLEX(-393, -197), MAKE_VAL_COMPLEX(-251, 97), MAKE_VAL_COMPLEX(-558, -502), MAKE_VAL_COMPLEX(-104, -598), MAKE_VAL_COMPLEX(42, 496), MAKE_VAL_COMPLEX(174, -248), MAKE_VAL_COMPLEX(-735, 69), MAKE_VAL_COMPLEX(-684, 90), MAKE_VAL_COMPLEX(-625, 749), MAKE_VAL_COMPLEX(-97, -173), MAKE_VAL_COMPLEX(-138, -300), MAKE_VAL_COMPLEX(752, -114), MAKE_VAL_COMPLEX(38, -824) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_611, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_611", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_611() */
/* END TEST 611 */

/*
 *  BEGIN TEST 612
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_612( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	oski_value_t* y_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-94, -96), MAKE_VAL_COMPLEX(154, -23), MAKE_VAL_COMPLEX(-245, 19), MAKE_VAL_COMPLEX(-271, -33), MAKE_VAL_COMPLEX(2, 236), MAKE_VAL_COMPLEX(-174, 110), MAKE_VAL_COMPLEX(-2, 18), MAKE_VAL_COMPLEX(183, 93), MAKE_VAL_COMPLEX(127, -132), MAKE_VAL_COMPLEX(168, -123), MAKE_VAL_COMPLEX(-22, -19), MAKE_VAL_COMPLEX(35, -32), MAKE_VAL_COMPLEX(78, 28), MAKE_VAL_COMPLEX(12, -144), MAKE_VAL_COMPLEX(130, -48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_612, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_612", y_true, y_val, y576_LEN*y576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_612() */
/* END TEST 612 */

/*
 *  BEGIN TEST 613
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_613( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	oski_value_t* y_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(65, 31), MAKE_VAL_COMPLEX(-145, -26), MAKE_VAL_COMPLEX(249, -3), MAKE_VAL_COMPLEX(276, 104), MAKE_VAL_COMPLEX(34, -228), MAKE_VAL_COMPLEX(211, -149), MAKE_VAL_COMPLEX(42, -62), MAKE_VAL_COMPLEX(-198, -170), MAKE_VAL_COMPLEX(-128, 111), MAKE_VAL_COMPLEX(-209, 112), MAKE_VAL_COMPLEX(-9, 48), MAKE_VAL_COMPLEX(-28, -25), MAKE_VAL_COMPLEX(-66, -48), MAKE_VAL_COMPLEX(-2, 116), MAKE_VAL_COMPLEX(-62, 48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_613, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_613", y_true, y_val, y576_LEN*y576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_613() */
/* END TEST 613 */

/*
 *  BEGIN TEST 614
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_614( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	oski_value_t* y_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(35, -50), MAKE_VAL_COMPLEX(40, -15), MAKE_VAL_COMPLEX(-10, 10), MAKE_VAL_COMPLEX(-50, 35), MAKE_VAL_COMPLEX(10, 30), MAKE_VAL_COMPLEX(45, 10), MAKE_VAL_COMPLEX(50, 10), MAKE_VAL_COMPLEX(50, -45), MAKE_VAL_COMPLEX(15, -10), MAKE_VAL_COMPLEX(-10, -35), MAKE_VAL_COMPLEX(-35, -10), MAKE_VAL_COMPLEX(45, -20), MAKE_VAL_COMPLEX(20, 0), MAKE_VAL_COMPLEX(25, -5), MAKE_VAL_COMPLEX(30, 50) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_614, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_614", y_true, y_val, y576_LEN*y576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_614() */
/* END TEST 614 */

/*
 *  BEGIN TEST 615
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_615( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	oski_value_t* y_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1023, -215), MAKE_VAL_COMPLEX(756, -212), MAKE_VAL_COMPLEX(-1620, 306), MAKE_VAL_COMPLEX(-1553, -209), MAKE_VAL_COMPLEX(182, 1426), MAKE_VAL_COMPLEX(-1439, 861), MAKE_VAL_COMPLEX(-364, 44), MAKE_VAL_COMPLEX(995, 791), MAKE_VAL_COMPLEX(644, -980), MAKE_VAL_COMPLEX(1120, -768), MAKE_VAL_COMPLEX(84, -30), MAKE_VAL_COMPLEX(-124, -120), MAKE_VAL_COMPLEX(426, 114), MAKE_VAL_COMPLEX(-246, -988), MAKE_VAL_COMPLEX(650, -842) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_615, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_615", y_true, y_val, y576_LEN*y576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_615() */
/* END TEST 615 */

/*
 *  BEGIN TEST 616
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_616( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-101, 60), MAKE_VAL_COMPLEX(29, 100), MAKE_VAL_COMPLEX(-34, -43), MAKE_VAL_COMPLEX(231, 160), MAKE_VAL_COMPLEX(-76, 286), MAKE_VAL_COMPLEX(-71, -61), MAKE_VAL_COMPLEX(-131, -19), MAKE_VAL_COMPLEX(16, -36), MAKE_VAL_COMPLEX(13, -70), MAKE_VAL_COMPLEX(51, -179), MAKE_VAL_COMPLEX(-23, 118), MAKE_VAL_COMPLEX(54, 25), MAKE_VAL_COMPLEX(-70, 96), MAKE_VAL_COMPLEX(224, -62), MAKE_VAL_COMPLEX(-11, -41) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_616, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_616", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_616() */
/* END TEST 616 */

/*
 *  BEGIN TEST 617
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_617( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(152, 27), MAKE_VAL_COMPLEX(-11, -169), MAKE_VAL_COMPLEX(131, 29), MAKE_VAL_COMPLEX(-113, -198), MAKE_VAL_COMPLEX(100, -152), MAKE_VAL_COMPLEX(68, 16), MAKE_VAL_COMPLEX(68, -22), MAKE_VAL_COMPLEX(-71, 115), MAKE_VAL_COMPLEX(-44, -56), MAKE_VAL_COMPLEX(-116, 108), MAKE_VAL_COMPLEX(44, -142), MAKE_VAL_COMPLEX(-57, 43), MAKE_VAL_COMPLEX(-40, -51), MAKE_VAL_COMPLEX(-266, -3), MAKE_VAL_COMPLEX(-60, 106) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_617, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_617", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_617() */
/* END TEST 617 */

/*
 *  BEGIN TEST 618
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_618( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(51, 57), MAKE_VAL_COMPLEX(3, -54), MAKE_VAL_COMPLEX(70, -25), MAKE_VAL_COMPLEX(82, -46), MAKE_VAL_COMPLEX(38, 96), MAKE_VAL_COMPLEX(-9, -33), MAKE_VAL_COMPLEX(-53, -21), MAKE_VAL_COMPLEX(-29, 67), MAKE_VAL_COMPLEX(-42, -89), MAKE_VAL_COMPLEX(-59, -43), MAKE_VAL_COMPLEX(12, -21), MAKE_VAL_COMPLEX(8, 51), MAKE_VAL_COMPLEX(-75, 50), MAKE_VAL_COMPLEX(-41, -42), MAKE_VAL_COMPLEX(-43, 59) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_618, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_618", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_618() */
/* END TEST 618 */

/*
 *  BEGIN TEST 619
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_619( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(532, -73), MAKE_VAL_COMPLEX(-204, -730), MAKE_VAL_COMPLEX(696, 172), MAKE_VAL_COMPLEX(-921, -400), MAKE_VAL_COMPLEX(-206, -970), MAKE_VAL_COMPLEX(504, -38), MAKE_VAL_COMPLEX(452, -404), MAKE_VAL_COMPLEX(-321, 511), MAKE_VAL_COMPLEX(51, -207), MAKE_VAL_COMPLEX(-120, 626), MAKE_VAL_COMPLEX(-7, -723), MAKE_VAL_COMPLEX(-394, 280), MAKE_VAL_COMPLEX(-367, -520), MAKE_VAL_COMPLEX(-1123, 434), MAKE_VAL_COMPLEX(-234, 406) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_619, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_619", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_619() */
/* END TEST 619 */

/*
 *  BEGIN TEST 620
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_620( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-28, -9), MAKE_VAL_COMPLEX(181, -61), MAKE_VAL_COMPLEX(-45, -57), MAKE_VAL_COMPLEX(215, 6), MAKE_VAL_COMPLEX(96, 64), MAKE_VAL_COMPLEX(42, 96), MAKE_VAL_COMPLEX(92, 22), MAKE_VAL_COMPLEX(-107, -3), MAKE_VAL_COMPLEX(20, 260), MAKE_VAL_COMPLEX(-76, -16), MAKE_VAL_COMPLEX(136, 50), MAKE_VAL_COMPLEX(-105, -49), MAKE_VAL_COMPLEX(-88, -41), MAKE_VAL_COMPLEX(46, -255), MAKE_VAL_COMPLEX(-144, -74) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_620, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_620", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_620() */
/* END TEST 620 */

/*
 *  BEGIN TEST 621
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_621( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-35, 108), MAKE_VAL_COMPLEX(-100, 43), MAKE_VAL_COMPLEX(111, 150), MAKE_VAL_COMPLEX(-113, 96), MAKE_VAL_COMPLEX(-222, 32), MAKE_VAL_COMPLEX(3, -123), MAKE_VAL_COMPLEX(-83, -109), MAKE_VAL_COMPLEX(-4, -12), MAKE_VAL_COMPLEX(94, -359), MAKE_VAL_COMPLEX(115, -89), MAKE_VAL_COMPLEX(-100, -41), MAKE_VAL_COMPLEX(33, 82), MAKE_VAL_COMPLEX(-17, -49), MAKE_VAL_COMPLEX(-1, 177), MAKE_VAL_COMPLEX(39, 35) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_621, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_621", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_621() */
/* END TEST 621 */

/*
 *  BEGIN TEST 622
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_622( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-33, -81), MAKE_VAL_COMPLEX(-24, 57), MAKE_VAL_COMPLEX(-85, 0), MAKE_VAL_COMPLEX(-106, 18), MAKE_VAL_COMPLEX(-4, -118), MAKE_VAL_COMPLEX(-3, 39), MAKE_VAL_COMPLEX(49, 43), MAKE_VAL_COMPLEX(57, -61), MAKE_VAL_COMPLEX(11, 112), MAKE_VAL_COMPLEX(47, 69), MAKE_VAL_COMPLEX(-21, 18), MAKE_VAL_COMPLEX(11, -58), MAKE_VAL_COMPLEX(100, -25), MAKE_VAL_COMPLEX(28, 61), MAKE_VAL_COMPLEX(69, -47) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_622, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_622", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_622() */
/* END TEST 622 */

/*
 *  BEGIN TEST 623
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x576', LAYOUT_COLMAJ>
 *  y : <'y576', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_623( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A576_ptr, A576_ind, A576_val,
		A576_M, A576_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y576_val,
		y576_LEN * y576_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y576_LEN, y576_VW, LAYOUT_COLMAJ, y576_LEN );
	oski_value_t* y_val = array_Duplicate( x576_val,
		x576_LEN * x576_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x576_LEN, x576_VW, LAYOUT_COLMAJ, x576_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(408, 466), MAKE_VAL_COMPLEX(-364, 1412), MAKE_VAL_COMPLEX(1360, 64), MAKE_VAL_COMPLEX(-530, 1652), MAKE_VAL_COMPLEX(-1172, 1268), MAKE_VAL_COMPLEX(-960, -528), MAKE_VAL_COMPLEX(-1052, 236), MAKE_VAL_COMPLEX(182, -438), MAKE_VAL_COMPLEX(-2174, -2246), MAKE_VAL_COMPLEX(300, -1072), MAKE_VAL_COMPLEX(-1042, 674), MAKE_VAL_COMPLEX(884, -124), MAKE_VAL_COMPLEX(86, -368), MAKE_VAL_COMPLEX(1614, 1416), MAKE_VAL_COMPLEX(896, -424) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_623, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_623", y_true, y_val, x576_LEN*x576_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_623() */
/* END TEST 623 */

#define A624_M 5    /* # of rows */
#define A624_N 5    /* # of cols */
#define A624_NNZ 13  /* # of stored non-zeros */
static oski_index_t A624_ptr[A624_M+1] = {0, 2, 4, 6, 10, 13}; /* end row pointers for A624 */
static oski_index_t A624_ind[A624_NNZ] = {3, 4, 3, 4, 2, 3, 0, 1, 2, 4, 0, 1, 3}; /* end column indices pointers for A624 */
static oski_value_t A624_val[A624_NNZ] = {MAKE_VAL_COMPLEX(-6, 4), MAKE_VAL_COMPLEX(-10, 5), MAKE_VAL_COMPLEX(-2, -9), MAKE_VAL_COMPLEX(-6, -10), MAKE_VAL_COMPLEX(10, 10), MAKE_VAL_COMPLEX(0, 5), MAKE_VAL_COMPLEX(-6, 4), MAKE_VAL_COMPLEX(-2, -9), MAKE_VAL_COMPLEX(0, 5), MAKE_VAL_COMPLEX(-8, 9), MAKE_VAL_COMPLEX(-10, 5), MAKE_VAL_COMPLEX(-6, -10), MAKE_VAL_COMPLEX(-8, 9)}; /* end stored non-zero values for A624 */

#define x624_LEN 5   /* vector length */
#define x624_VW  3   /* vector width */
static oski_value_t x624_val[x624_LEN*x624_VW] = {MAKE_VAL_COMPLEX(-8, 10), MAKE_VAL_COMPLEX(2, 3), MAKE_VAL_COMPLEX(5, 8), MAKE_VAL_COMPLEX(2, -4), MAKE_VAL_COMPLEX(10, -10), MAKE_VAL_COMPLEX(-5, -10), MAKE_VAL_COMPLEX(-10, 5), MAKE_VAL_COMPLEX(-3, -9), MAKE_VAL_COMPLEX(2, -4), MAKE_VAL_COMPLEX(-4, 6), MAKE_VAL_COMPLEX(-7, 9), MAKE_VAL_COMPLEX(-4, -10), MAKE_VAL_COMPLEX(-5, -1), MAKE_VAL_COMPLEX(7, -6), MAKE_VAL_COMPLEX(2, 9)}; /* data */

#define y624_LEN 5   /* vector length */
#define y624_VW  3   /* vector width */
static oski_value_t y624_val[y624_LEN*y624_VW] = {MAKE_VAL_COMPLEX(8, 9), MAKE_VAL_COMPLEX(-6, 4), MAKE_VAL_COMPLEX(0, -5), MAKE_VAL_COMPLEX(-6, 10), MAKE_VAL_COMPLEX(-7, -2), MAKE_VAL_COMPLEX(8, -1), MAKE_VAL_COMPLEX(10, -6), MAKE_VAL_COMPLEX(-3, -4), MAKE_VAL_COMPLEX(-4, -5), MAKE_VAL_COMPLEX(-4, 9), MAKE_VAL_COMPLEX(-1, 5), MAKE_VAL_COMPLEX(-2, 8), MAKE_VAL_COMPLEX(-3, 7), MAKE_VAL_COMPLEX(-1, -10), MAKE_VAL_COMPLEX(-10, 1)}; /* data */

/*
 *  BEGIN TEST 624
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_624( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	oski_value_t* y_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(75, -117), MAKE_VAL_COMPLEX(-6, 29), MAKE_VAL_COMPLEX(-21, -26), MAKE_VAL_COMPLEX(134, 84), MAKE_VAL_COMPLEX(-1, 43), MAKE_VAL_COMPLEX(-24, -48), MAKE_VAL_COMPLEX(-224, -98), MAKE_VAL_COMPLEX(5, -135), MAKE_VAL_COMPLEX(98, -14), MAKE_VAL_COMPLEX(36, -191), MAKE_VAL_COMPLEX(-69, 10), MAKE_VAL_COMPLEX(-183, -15), MAKE_VAL_COMPLEX(-10, -222), MAKE_VAL_COMPLEX(-258, -171), MAKE_VAL_COMPLEX(-14, 137) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_624, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_624", y_true, y_val, y624_LEN*y624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_624() */
/* END TEST 624 */

/*
 *  BEGIN TEST 625
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_625( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	oski_value_t* y_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-80, 202), MAKE_VAL_COMPLEX(-44, -39), MAKE_VAL_COMPLEX(46, 1), MAKE_VAL_COMPLEX(-214, -64), MAKE_VAL_COMPLEX(-24, -88), MAKE_VAL_COMPLEX(69, 83), MAKE_VAL_COMPLEX(304, 118), MAKE_VAL_COMPLEX(0, 100), MAKE_VAL_COMPLEX(-93, -31), MAKE_VAL_COMPLEX(-101, 216), MAKE_VAL_COMPLEX(39, 10), MAKE_VAL_COMPLEX(133, 45), MAKE_VAL_COMPLEX(-40, 242), MAKE_VAL_COMPLEX(303, 116), MAKE_VAL_COMPLEX(-41, -182) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_625, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_625", y_true, y_val, y624_LEN*y624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_625() */
/* END TEST 625 */

/*
 *  BEGIN TEST 626
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_626( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	oski_value_t* y_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(30, -75), MAKE_VAL_COMPLEX(42, 24), MAKE_VAL_COMPLEX(-30, 15), MAKE_VAL_COMPLEX(78, 6), MAKE_VAL_COMPLEX(9, 48), MAKE_VAL_COMPLEX(-30, -45), MAKE_VAL_COMPLEX(-66, -42), MAKE_VAL_COMPLEX(-15, 30), MAKE_VAL_COMPLEX(-18, 39), MAKE_VAL_COMPLEX(66, -3), MAKE_VAL_COMPLEX(33, -9), MAKE_VAL_COMPLEX(54, -12), MAKE_VAL_COMPLEX(51, -3), MAKE_VAL_COMPLEX(-57, 36), MAKE_VAL_COMPLEX(36, 57) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_626, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_626", y_true, y_val, y624_LEN*y624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_626() */
/* END TEST 626 */

/*
 *  BEGIN TEST 627
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_627( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	oski_value_t* y_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(266, 254), MAKE_VAL_COMPLEX(-72, -140), MAKE_VAL_COMPLEX(154, -14), MAKE_VAL_COMPLEX(-352, 336), MAKE_VAL_COMPLEX(-54, -46), MAKE_VAL_COMPLEX(90, -2), MAKE_VAL_COMPLEX(312, -712), MAKE_VAL_COMPLEX(622, 46), MAKE_VAL_COMPLEX(162, 422), MAKE_VAL_COMPLEX(730, -22), MAKE_VAL_COMPLEX(-76, -356), MAKE_VAL_COMPLEX(8, -864), MAKE_VAL_COMPLEX(860, -168), MAKE_VAL_COMPLEX(798, -902), MAKE_VAL_COMPLEX(-418, -170) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_627, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_627", y_true, y_val, y624_LEN*y624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_627() */
/* END TEST 627 */

/*
 *  BEGIN TEST 628
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_628( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-21, -273), MAKE_VAL_COMPLEX(14, 52), MAKE_VAL_COMPLEX(-8, -142), MAKE_VAL_COMPLEX(187, 46), MAKE_VAL_COMPLEX(-57, 199), MAKE_VAL_COMPLEX(241, 136), MAKE_VAL_COMPLEX(155, -75), MAKE_VAL_COMPLEX(60, -30), MAKE_VAL_COMPLEX(-20, -60), MAKE_VAL_COMPLEX(1, -87), MAKE_VAL_COMPLEX(129, -30), MAKE_VAL_COMPLEX(179, -112), MAKE_VAL_COMPLEX(3, -181), MAKE_VAL_COMPLEX(12, 47), MAKE_VAL_COMPLEX(-157, -157) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_628, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_628", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_628() */
/* END TEST 628 */

/*
 *  BEGIN TEST 629
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_629( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-31, 133), MAKE_VAL_COMPLEX(-56, -50), MAKE_VAL_COMPLEX(-102, 144), MAKE_VAL_COMPLEX(-159, -2), MAKE_VAL_COMPLEX(97, -39), MAKE_VAL_COMPLEX(-111, -126), MAKE_VAL_COMPLEX(-145, -55), MAKE_VAL_COMPLEX(48, 54), MAKE_VAL_COMPLEX(48, 104), MAKE_VAL_COMPLEX(-37, 11), MAKE_VAL_COMPLEX(-177, -94), MAKE_VAL_COMPLEX(-55, 132), MAKE_VAL_COMPLEX(37, 137), MAKE_VAL_COMPLEX(6, 59), MAKE_VAL_COMPLEX(55, 123) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_629, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_629", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_629() */
/* END TEST 629 */

/*
 *  BEGIN TEST 630
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_630( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(132, -42), MAKE_VAL_COMPLEX(0, -39), MAKE_VAL_COMPLEX(3, -102), MAKE_VAL_COMPLEX(-42, 24), MAKE_VAL_COMPLEX(-150, 30), MAKE_VAL_COMPLEX(-15, 120), MAKE_VAL_COMPLEX(120, 15), MAKE_VAL_COMPLEX(-27, 99), MAKE_VAL_COMPLEX(-42, 24), MAKE_VAL_COMPLEX(72, -30), MAKE_VAL_COMPLEX(117, -39), MAKE_VAL_COMPLEX(-24, 114), MAKE_VAL_COMPLEX(39, 39), MAKE_VAL_COMPLEX(-99, 12), MAKE_VAL_COMPLEX(36, -93) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_630, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_630", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_630() */
/* END TEST 630 */

/*
 *  BEGIN TEST 631
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_631( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-696, 1220), MAKE_VAL_COMPLEX(-7, -642), MAKE_VAL_COMPLEX(-1080, 477), MAKE_VAL_COMPLEX(-1199, -947), MAKE_VAL_COMPLEX(724, -298), MAKE_VAL_COMPLEX(-567, -1452), MAKE_VAL_COMPLEX(-755, -690), MAKE_VAL_COMPLEX(-285, 555), MAKE_VAL_COMPLEX(-280, 830), MAKE_VAL_COMPLEX(-188, 132), MAKE_VAL_COMPLEX(-598, -1150), MAKE_VAL_COMPLEX(-1447, 601), MAKE_VAL_COMPLEX(-549, 1291), MAKE_VAL_COMPLEX(-325, 104), MAKE_VAL_COMPLEX(103, 1232) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_631, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_631", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_631() */
/* END TEST 631 */

/*
 *  BEGIN TEST 632
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_632( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(107, -91), MAKE_VAL_COMPLEX(-15, 84), MAKE_VAL_COMPLEX(146, 13), MAKE_VAL_COMPLEX(-119, -128), MAKE_VAL_COMPLEX(83, 31), MAKE_VAL_COMPLEX(-13, -155), MAKE_VAL_COMPLEX(70, -145), MAKE_VAL_COMPLEX(-39, -17), MAKE_VAL_COMPLEX(-44, -2), MAKE_VAL_COMPLEX(-43, -237), MAKE_VAL_COMPLEX(-34, 47), MAKE_VAL_COMPLEX(43, 192), MAKE_VAL_COMPLEX(10, -292), MAKE_VAL_COMPLEX(208, -98), MAKE_VAL_COMPLEX(18, 101) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_632, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_632", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_632() */
/* END TEST 632 */

/*
 *  BEGIN TEST 633
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_633( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-229, 59), MAKE_VAL_COMPLEX(-4, -54), MAKE_VAL_COMPLEX(-198, 64), MAKE_VAL_COMPLEX(163, 130), MAKE_VAL_COMPLEX(47, 19), MAKE_VAL_COMPLEX(83, 70), MAKE_VAL_COMPLEX(-155, 75), MAKE_VAL_COMPLEX(108, -46), MAKE_VAL_COMPLEX(88, 4), MAKE_VAL_COMPLEX(-27, 225), MAKE_VAL_COMPLEX(-75, -74), MAKE_VAL_COMPLEX(31, -268), MAKE_VAL_COMPLEX(-21, 243), MAKE_VAL_COMPLEX(-126, 137), MAKE_VAL_COMPLEX(-91, -47) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_633, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_633", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_633() */
/* END TEST 633 */

/*
 *  BEGIN TEST 634
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_634( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-48, 60), MAKE_VAL_COMPLEX(12, 18), MAKE_VAL_COMPLEX(30, 48), MAKE_VAL_COMPLEX(12, -24), MAKE_VAL_COMPLEX(60, -60), MAKE_VAL_COMPLEX(-30, -60), MAKE_VAL_COMPLEX(-60, 30), MAKE_VAL_COMPLEX(-18, -54), MAKE_VAL_COMPLEX(12, -24), MAKE_VAL_COMPLEX(-24, 36), MAKE_VAL_COMPLEX(-42, 54), MAKE_VAL_COMPLEX(-24, -60), MAKE_VAL_COMPLEX(-30, -6), MAKE_VAL_COMPLEX(42, -36), MAKE_VAL_COMPLEX(12, 54) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_634, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_634", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_634() */
/* END TEST 634 */

/*
 *  BEGIN TEST 635
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_635( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1146, -648), MAKE_VAL_COMPLEX(691, -232), MAKE_VAL_COMPLEX(-586, -1265), MAKE_VAL_COMPLEX(-540, 1518), MAKE_VAL_COMPLEX(-114, -618), MAKE_VAL_COMPLEX(-1068, 789), MAKE_VAL_COMPLEX(-1375, -100), MAKE_VAL_COMPLEX(113, 449), MAKE_VAL_COMPLEX(168, 414), MAKE_VAL_COMPLEX(-1726, 1204), MAKE_VAL_COMPLEX(519, -55), MAKE_VAL_COMPLEX(1474, -1040), MAKE_VAL_COMPLEX(-2319, 1063), MAKE_VAL_COMPLEX(-1644, -1163), MAKE_VAL_COMPLEX(653, -624) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_635, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_635", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_635() */
/* END TEST 635 */

/*
 *  BEGIN TEST 636
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_636( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	oski_value_t* y_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-24, -29), MAKE_VAL_COMPLEX(66, 30), MAKE_VAL_COMPLEX(-145, -60), MAKE_VAL_COMPLEX(-66, -251), MAKE_VAL_COMPLEX(-16, -265), MAKE_VAL_COMPLEX(-43, 71), MAKE_VAL_COMPLEX(15, 93), MAKE_VAL_COMPLEX(49, -168), MAKE_VAL_COMPLEX(-48, -2), MAKE_VAL_COMPLEX(-275, -241), MAKE_VAL_COMPLEX(-34, -53), MAKE_VAL_COMPLEX(-56, -48), MAKE_VAL_COMPLEX(67, -75), MAKE_VAL_COMPLEX(-147, 6), MAKE_VAL_COMPLEX(-25, 27) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_636, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_636", y_true, y_val, y624_LEN*y624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_636() */
/* END TEST 636 */

/*
 *  BEGIN TEST 637
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_637( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	oski_value_t* y_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-173, 43), MAKE_VAL_COMPLEX(-58, -148), MAKE_VAL_COMPLEX(210, 110), MAKE_VAL_COMPLEX(-4, 73), MAKE_VAL_COMPLEX(112, 194), MAKE_VAL_COMPLEX(-24, 43), MAKE_VAL_COMPLEX(-37, 97), MAKE_VAL_COMPLEX(33, 169), MAKE_VAL_COMPLEX(153, 0), MAKE_VAL_COMPLEX(198, 99), MAKE_VAL_COMPLEX(-21, -10), MAKE_VAL_COMPLEX(-28, -58), MAKE_VAL_COMPLEX(-128, -34), MAKE_VAL_COMPLEX(287, 81), MAKE_VAL_COMPLEX(112, -167) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_637, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_637", y_true, y_val, y624_LEN*y624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_637() */
/* END TEST 637 */

/*
 *  BEGIN TEST 638
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_638( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	oski_value_t* y_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(3, 94), MAKE_VAL_COMPLEX(-56, -6), MAKE_VAL_COMPLEX(25, -30), MAKE_VAL_COMPLEX(-86, 30), MAKE_VAL_COMPLEX(-32, -47), MAKE_VAL_COMPLEX(53, 34), MAKE_VAL_COMPLEX(90, 14), MAKE_VAL_COMPLEX(2, -39), MAKE_VAL_COMPLEX(1, -50), MAKE_VAL_COMPLEX(-69, 34), MAKE_VAL_COMPLEX(-31, 25), MAKE_VAL_COMPLEX(-52, 38), MAKE_VAL_COMPLEX(-53, 27), MAKE_VAL_COMPLEX(44, -65), MAKE_VAL_COMPLEX(-65, -44) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_638, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_638", y_true, y_val, y624_LEN*y624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_638() */
/* END TEST 638 */

/*
 *  BEGIN TEST 639
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_639( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	oski_value_t* y_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-892, 154), MAKE_VAL_COMPLEX(-146, -1246), MAKE_VAL_COMPLEX(1090, 1890), MAKE_VAL_COMPLEX(-1171, 1805), MAKE_VAL_COMPLEX(-1160, 2344), MAKE_VAL_COMPLEX(389, 57), MAKE_VAL_COMPLEX(92, -130), MAKE_VAL_COMPLEX(-1252, 1314), MAKE_VAL_COMPLEX(891, 525), MAKE_VAL_COMPLEX(687, 3327), MAKE_VAL_COMPLEX(-203, 339), MAKE_VAL_COMPLEX(-26, 206), MAKE_VAL_COMPLEX(-1214, -386), MAKE_VAL_COMPLEX(1842, 1604), MAKE_VAL_COMPLEX(1043, -545) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-9, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_639, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_639", y_true, y_val, y624_LEN*y624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_639() */
/* END TEST 639 */

/*
 *  BEGIN TEST 640
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_640( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(74, -35), MAKE_VAL_COMPLEX(56, -90), MAKE_VAL_COMPLEX(185, 13), MAKE_VAL_COMPLEX(120, 94), MAKE_VAL_COMPLEX(87, -38), MAKE_VAL_COMPLEX(-68, 98), MAKE_VAL_COMPLEX(-25, -30), MAKE_VAL_COMPLEX(-4, -117), MAKE_VAL_COMPLEX(-54, 13), MAKE_VAL_COMPLEX(65, 13), MAKE_VAL_COMPLEX(-148, -106), MAKE_VAL_COMPLEX(52, -173), MAKE_VAL_COMPLEX(-114, -137), MAKE_VAL_COMPLEX(-108, -55), MAKE_VAL_COMPLEX(210, 18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_640, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_640", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_640() */
/* END TEST 640 */

/*
 *  BEGIN TEST 641
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_641( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-72, 217), MAKE_VAL_COMPLEX(-7, 105), MAKE_VAL_COMPLEX(-58, 30), MAKE_VAL_COMPLEX(-134, -156), MAKE_VAL_COMPLEX(-67, -162), MAKE_VAL_COMPLEX(-77, -163), MAKE_VAL_COMPLEX(-40, 175), MAKE_VAL_COMPLEX(-110, 45), MAKE_VAL_COMPLEX(40, -75), MAKE_VAL_COMPLEX(-55, 89), MAKE_VAL_COMPLEX(152, 268), MAKE_VAL_COMPLEX(-186, 99), MAKE_VAL_COMPLEX(50, 171), MAKE_VAL_COMPLEX(131, -74), MAKE_VAL_COMPLEX(-107, 63) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_641, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_641", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_641() */
/* END TEST 641 */

/*
 *  BEGIN TEST 642
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_642( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-14, 120), MAKE_VAL_COMPLEX(31, 14), MAKE_VAL_COMPLEX(80, 39), MAKE_VAL_COMPLEX(-4, -42), MAKE_VAL_COMPLEX(30, -130), MAKE_VAL_COMPLEX(-90, -55), MAKE_VAL_COMPLEX(-55, 90), MAKE_VAL_COMPLEX(-69, -57), MAKE_VAL_COMPLEX(-4, -42), MAKE_VAL_COMPLEX(-2, 68), MAKE_VAL_COMPLEX(-11, 107), MAKE_VAL_COMPLEX(-82, -60), MAKE_VAL_COMPLEX(-45, 17), MAKE_VAL_COMPLEX(26, -83), MAKE_VAL_COMPLEX(61, 62) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_642, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_642", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_642() */
/* END TEST 642 */

/*
 *  BEGIN TEST 643
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_643( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(236, 1112), MAKE_VAL_COMPLEX(367, 789), MAKE_VAL_COMPLEX(-755, 817), MAKE_VAL_COMPLEX(-1478, 34), MAKE_VAL_COMPLEX(-754, 190), MAKE_VAL_COMPLEX(-715, -791), MAKE_VAL_COMPLEX(535, 565), MAKE_VAL_COMPLEX(252, 786), MAKE_VAL_COMPLEX(52, -524), MAKE_VAL_COMPLEX(-182, 546), MAKE_VAL_COMPLEX(1948, 150), MAKE_VAL_COMPLEX(196, 1504), MAKE_VAL_COMPLEX(1430, 388), MAKE_VAL_COMPLEX(707, -705), MAKE_VAL_COMPLEX(-899, 1095) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_643, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_643", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_643() */
/* END TEST 643 */

/*
 *  BEGIN TEST 644
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_644( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(110, 113), MAKE_VAL_COMPLEX(112, -9), MAKE_VAL_COMPLEX(130, 145), MAKE_VAL_COMPLEX(-14, -166), MAKE_VAL_COMPLEX(27, -220), MAKE_VAL_COMPLEX(77, -150), MAKE_VAL_COMPLEX(-55, 45), MAKE_VAL_COMPLEX(-143, -44), MAKE_VAL_COMPLEX(-8, 71), MAKE_VAL_COMPLEX(-33, -15), MAKE_VAL_COMPLEX(53, 139), MAKE_VAL_COMPLEX(24, -21), MAKE_VAL_COMPLEX(61, -232), MAKE_VAL_COMPLEX(-67, 119), MAKE_VAL_COMPLEX(-74, 33) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_644, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_644", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_644() */
/* END TEST 644 */

/*
 *  BEGIN TEST 645
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_645( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-268, 23), MAKE_VAL_COMPLEX(-89, 63), MAKE_VAL_COMPLEX(-74, -2), MAKE_VAL_COMPLEX(58, 108), MAKE_VAL_COMPLEX(163, 90), MAKE_VAL_COMPLEX(-127, -25), MAKE_VAL_COMPLEX(-120, 5), MAKE_VAL_COMPLEX(122, -109), MAKE_VAL_COMPLEX(52, -129), MAKE_VAL_COMPLEX(-49, 99), MAKE_VAL_COMPLEX(-192, -16), MAKE_VAL_COMPLEX(-58, -151), MAKE_VAL_COMPLEX(-138, 201), MAKE_VAL_COMPLEX(197, -194), MAKE_VAL_COMPLEX(79, 117) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_645, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_645", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_645() */
/* END TEST 645 */

/*
 *  BEGIN TEST 646
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_646( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(10, 90), MAKE_VAL_COMPLEX(25, 5), MAKE_VAL_COMPLEX(65, 15), MAKE_VAL_COMPLEX(-10, -30), MAKE_VAL_COMPLEX(0, -100), MAKE_VAL_COMPLEX(-75, -25), MAKE_VAL_COMPLEX(-25, 75), MAKE_VAL_COMPLEX(-60, -30), MAKE_VAL_COMPLEX(-10, -30), MAKE_VAL_COMPLEX(10, 50), MAKE_VAL_COMPLEX(10, 80), MAKE_VAL_COMPLEX(-70, -30), MAKE_VAL_COMPLEX(-30, 20), MAKE_VAL_COMPLEX(5, -65), MAKE_VAL_COMPLEX(55, 35) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_646, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_646", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_646() */
/* END TEST 646 */

/*
 *  BEGIN TEST 647
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_ROWMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_647( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_ROWMAJ, x624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-27, 661), MAKE_VAL_COMPLEX(193, 436), MAKE_VAL_COMPLEX(-492, 497), MAKE_VAL_COMPLEX(670, -204), MAKE_VAL_COMPLEX(657, -135), MAKE_VAL_COMPLEX(672, 575), MAKE_VAL_COMPLEX(60, -75), MAKE_VAL_COMPLEX(-94, -492), MAKE_VAL_COMPLEX(-509, 63), MAKE_VAL_COMPLEX(282, -202), MAKE_VAL_COMPLEX(-261, 397), MAKE_VAL_COMPLEX(-47, 413), MAKE_VAL_COMPLEX(1300, 224), MAKE_VAL_COMPLEX(-965, -272), MAKE_VAL_COMPLEX(-47, -578) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_647, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_647", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_647() */
/* END TEST 647 */

/*
 *  BEGIN TEST 648
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_648( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	oski_value_t* y_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-120, 135), MAKE_VAL_COMPLEX(48, -88), MAKE_VAL_COMPLEX(-73, 19), MAKE_VAL_COMPLEX(-178, -132), MAKE_VAL_COMPLEX(97, -6), MAKE_VAL_COMPLEX(-44, -102), MAKE_VAL_COMPLEX(-68, 202), MAKE_VAL_COMPLEX(109, -88), MAKE_VAL_COMPLEX(28, 2), MAKE_VAL_COMPLEX(11, 8), MAKE_VAL_COMPLEX(155, -16), MAKE_VAL_COMPLEX(-170, -165), MAKE_VAL_COMPLEX(75, -183), MAKE_VAL_COMPLEX(257, 263), MAKE_VAL_COMPLEX(15, 59) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_648, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_648", y_true, y_val, y624_LEN*y624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_648() */
/* END TEST 648 */

/*
 *  BEGIN TEST 649
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_649( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	oski_value_t* y_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(50, -250), MAKE_VAL_COMPLEX(26, 56), MAKE_VAL_COMPLEX(63, 36), MAKE_VAL_COMPLEX(264, 34), MAKE_VAL_COMPLEX(-24, 42), MAKE_VAL_COMPLEX(-46, 97), MAKE_VAL_COMPLEX(-54, -156), MAKE_VAL_COMPLEX(-84, 138), MAKE_VAL_COMPLEX(6, 61), MAKE_VAL_COMPLEX(51, -99), MAKE_VAL_COMPLEX(-134, -37), MAKE_VAL_COMPLEX(208, 81), MAKE_VAL_COMPLEX(-28, 112), MAKE_VAL_COMPLEX(-266, -151), MAKE_VAL_COMPLEX(97, -50) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_649, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_649", y_true, y_val, y624_LEN*y624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_649() */
/* END TEST 649 */

/*
 *  BEGIN TEST 650
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_650( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	oski_value_t* y_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(47, -74), MAKE_VAL_COMPLEX(40, 34), MAKE_VAL_COMPLEX(-35, 10), MAKE_VAL_COMPLEX(82, 22), MAKE_VAL_COMPLEX(0, 53), MAKE_VAL_COMPLEX(-23, -54), MAKE_VAL_COMPLEX(-62, -58), MAKE_VAL_COMPLEX(-22, 29), MAKE_VAL_COMPLEX(-27, 38), MAKE_VAL_COMPLEX(71, 10), MAKE_VAL_COMPLEX(37, -3), MAKE_VAL_COMPLEX(60, -2), MAKE_VAL_COMPLEX(55, 7), MAKE_VAL_COMPLEX(-68, 27), MAKE_VAL_COMPLEX(27, 68) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_650, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_650", y_true, y_val, y624_LEN*y624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_650() */
/* END TEST 650 */

/*
 *  BEGIN TEST 651
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_651( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	oski_value_t* y_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1366, 1397), MAKE_VAL_COMPLEX(376, -386), MAKE_VAL_COMPLEX(-657, -544), MAKE_VAL_COMPLEX(-1552, -1492), MAKE_VAL_COMPLEX(451, 158), MAKE_VAL_COMPLEX(693, -1056), MAKE_VAL_COMPLEX(-808, 1248), MAKE_VAL_COMPLEX(1287, -584), MAKE_VAL_COMPLEX(57, -268), MAKE_VAL_COMPLEX(-392, 681), MAKE_VAL_COMPLEX(1310, 962), MAKE_VAL_COMPLEX(-997, -1813), MAKE_VAL_COMPLEX(1247, -839), MAKE_VAL_COMPLEX(1118, 2933), MAKE_VAL_COMPLEX(-879, 486) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(9, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_651, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_651", y_true, y_val, y624_LEN*y624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_651() */
/* END TEST 651 */

/*
 *  BEGIN TEST 652
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_652( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(9, -167), MAKE_VAL_COMPLEX(176, -2), MAKE_VAL_COMPLEX(113, -1), MAKE_VAL_COMPLEX(1, -15), MAKE_VAL_COMPLEX(-68, -148), MAKE_VAL_COMPLEX(46, 86), MAKE_VAL_COMPLEX(-22, 69), MAKE_VAL_COMPLEX(-18, -54), MAKE_VAL_COMPLEX(126, 81), MAKE_VAL_COMPLEX(39, -45), MAKE_VAL_COMPLEX(98, -127), MAKE_VAL_COMPLEX(144, 120), MAKE_VAL_COMPLEX(-21, -93), MAKE_VAL_COMPLEX(71, -153), MAKE_VAL_COMPLEX(-84, -126) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_652, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_652", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_652() */
/* END TEST 652 */

/*
 *  BEGIN TEST 653
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_653( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-33, 115), MAKE_VAL_COMPLEX(-192, 4), MAKE_VAL_COMPLEX(-155, 5), MAKE_VAL_COMPLEX(11, 31), MAKE_VAL_COMPLEX(88, 208), MAKE_VAL_COMPLEX(4, -86), MAKE_VAL_COMPLEX(22, -119), MAKE_VAL_COMPLEX(60, 60), MAKE_VAL_COMPLEX(-114, -65), MAKE_VAL_COMPLEX(-55, 17), MAKE_VAL_COMPLEX(-120, 81), MAKE_VAL_COMPLEX(-96, -116), MAKE_VAL_COMPLEX(35, 75), MAKE_VAL_COMPLEX(-61, 193), MAKE_VAL_COMPLEX(44, 116) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_653, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_653", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_653() */
/* END TEST 653 */

/*
 *  BEGIN TEST 654
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_654( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(28, -76), MAKE_VAL_COMPLEX(-18, -14), MAKE_VAL_COMPLEX(-46, -38), MAKE_VAL_COMPLEX(-4, 28), MAKE_VAL_COMPLEX(-40, 80), MAKE_VAL_COMPLEX(50, 50), MAKE_VAL_COMPLEX(50, -50), MAKE_VAL_COMPLEX(36, 48), MAKE_VAL_COMPLEX(-4, 28), MAKE_VAL_COMPLEX(12, -44), MAKE_VAL_COMPLEX(24, -68), MAKE_VAL_COMPLEX(44, 52), MAKE_VAL_COMPLEX(32, -4), MAKE_VAL_COMPLEX(-30, 50), MAKE_VAL_COMPLEX(-30, -50) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_654, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_654", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_654() */
/* END TEST 654 */

/*
 *  BEGIN TEST 655
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_655( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(491, 785), MAKE_VAL_COMPLEX(-877, 302), MAKE_VAL_COMPLEX(-561, 74), MAKE_VAL_COMPLEX(-39, 115), MAKE_VAL_COMPLEX(406, 654), MAKE_VAL_COMPLEX(-422, -153), MAKE_VAL_COMPLEX(157, -369), MAKE_VAL_COMPLEX(159, 387), MAKE_VAL_COMPLEX(-856, -115), MAKE_VAL_COMPLEX(5, 255), MAKE_VAL_COMPLEX(-57, 768), MAKE_VAL_COMPLEX(-994, -136), MAKE_VAL_COMPLEX(352, 482), MAKE_VAL_COMPLEX(-201, 928), MAKE_VAL_COMPLEX(725, 318) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_655, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_655", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_655() */
/* END TEST 655 */

/*
 *  BEGIN TEST 656
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_656( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(97, -99), MAKE_VAL_COMPLEX(-129, -119), MAKE_VAL_COMPLEX(74, -122), MAKE_VAL_COMPLEX(-23, -239), MAKE_VAL_COMPLEX(44, -276), MAKE_VAL_COMPLEX(1, 59), MAKE_VAL_COMPLEX(43, 16), MAKE_VAL_COMPLEX(-30, -20), MAKE_VAL_COMPLEX(-8, 45), MAKE_VAL_COMPLEX(179, -101), MAKE_VAL_COMPLEX(124, -5), MAKE_VAL_COMPLEX(-2, -158), MAKE_VAL_COMPLEX(-53, -11), MAKE_VAL_COMPLEX(56, 214), MAKE_VAL_COMPLEX(9, 103) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_656, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_656", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_656() */
/* END TEST 656 */

/*
 *  BEGIN TEST 657
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_657( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-167, 43), MAKE_VAL_COMPLEX(108, 133), MAKE_VAL_COMPLEX(-130, 157), MAKE_VAL_COMPLEX(51, 253), MAKE_VAL_COMPLEX(26, 346), MAKE_VAL_COMPLEX(69, -94), MAKE_VAL_COMPLEX(-78, -86), MAKE_VAL_COMPLEX(93, -1), MAKE_VAL_COMPLEX(36, -31), MAKE_VAL_COMPLEX(-221, 73), MAKE_VAL_COMPLEX(-187, -44), MAKE_VAL_COMPLEX(72, 130), MAKE_VAL_COMPLEX(60, -24), MAKE_VAL_COMPLEX(-14, -165), MAKE_VAL_COMPLEX(-72, -89) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_657, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_657", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_657() */
/* END TEST 657 */

/*
 *  BEGIN TEST 658
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_658( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(80, -18), MAKE_VAL_COMPLEX(2, -23), MAKE_VAL_COMPLEX(7, -60), MAKE_VAL_COMPLEX(-26, 12), MAKE_VAL_COMPLEX(-90, 10), MAKE_VAL_COMPLEX(-15, 70), MAKE_VAL_COMPLEX(70, 15), MAKE_VAL_COMPLEX(-21, 57), MAKE_VAL_COMPLEX(-26, 12), MAKE_VAL_COMPLEX(44, -14), MAKE_VAL_COMPLEX(71, -17), MAKE_VAL_COMPLEX(-20, 66), MAKE_VAL_COMPLEX(21, 25), MAKE_VAL_COMPLEX(-59, 2), MAKE_VAL_COMPLEX(26, -53) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_658, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_658", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_658() */
/* END TEST 658 */

/*
 *  BEGIN TEST 659
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_659( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_ROWMAJ, y624_VW );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-800, -670), MAKE_VAL_COMPLEX(-729, 1017), MAKE_VAL_COMPLEX(-993, -411), MAKE_VAL_COMPLEX(-1640, 430), MAKE_VAL_COMPLEX(-1976, 68), MAKE_VAL_COMPLEX(487, -41), MAKE_VAL_COMPLEX(94, -392), MAKE_VAL_COMPLEX(-50, 260), MAKE_VAL_COMPLEX(333, 41), MAKE_VAL_COMPLEX(-896, -1202), MAKE_VAL_COMPLEX(-169, -943), MAKE_VAL_COMPLEX(-1034, 202), MAKE_VAL_COMPLEX(6, 362), MAKE_VAL_COMPLEX(1437, -541), MAKE_VAL_COMPLEX(657, -201) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_659, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_659", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_659() */
/* END TEST 659 */

/*
 *  BEGIN TEST 660
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_660( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	oski_value_t* y_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(58, 154), MAKE_VAL_COMPLEX(-192, 14), MAKE_VAL_COMPLEX(-50, 120), MAKE_VAL_COMPLEX(57, 167), MAKE_VAL_COMPLEX(24, -80), MAKE_VAL_COMPLEX(38, -116), MAKE_VAL_COMPLEX(36, -110), MAKE_VAL_COMPLEX(36, -102), MAKE_VAL_COMPLEX(102, 33), MAKE_VAL_COMPLEX(286, 263), MAKE_VAL_COMPLEX(-47, 12), MAKE_VAL_COMPLEX(66, -77), MAKE_VAL_COMPLEX(34, 27), MAKE_VAL_COMPLEX(-252, -137), MAKE_VAL_COMPLEX(-85, 170) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_660, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_660", y_true, y_val, y624_LEN*y624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_660() */
/* END TEST 660 */

/*
 *  BEGIN TEST 661
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_661( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	oski_value_t* y_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(64, -198), MAKE_VAL_COMPLEX(208, 62), MAKE_VAL_COMPLEX(0, -140), MAKE_VAL_COMPLEX(19, -67), MAKE_VAL_COMPLEX(-72, 142), MAKE_VAL_COMPLEX(-16, 32), MAKE_VAL_COMPLEX(-56, -14), MAKE_VAL_COMPLEX(-88, 116), MAKE_VAL_COMPLEX(-168, -13), MAKE_VAL_COMPLEX(-212, -187), MAKE_VAL_COMPLEX(93, 18), MAKE_VAL_COMPLEX(6, 129), MAKE_VAL_COMPLEX(24, 31), MAKE_VAL_COMPLEX(148, 107), MAKE_VAL_COMPLEX(55, -66) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_661, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_661", y_true, y_val, y624_LEN*y624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_661() */
/* END TEST 661 */

/*
 *  BEGIN TEST 662
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_662( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	oski_value_t* y_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-16, -18), MAKE_VAL_COMPLEX(12, -8), MAKE_VAL_COMPLEX(0, 10), MAKE_VAL_COMPLEX(12, -20), MAKE_VAL_COMPLEX(14, 4), MAKE_VAL_COMPLEX(-16, 2), MAKE_VAL_COMPLEX(-20, 12), MAKE_VAL_COMPLEX(6, 8), MAKE_VAL_COMPLEX(8, 10), MAKE_VAL_COMPLEX(8, -18), MAKE_VAL_COMPLEX(2, -10), MAKE_VAL_COMPLEX(4, -16), MAKE_VAL_COMPLEX(6, -14), MAKE_VAL_COMPLEX(2, 20), MAKE_VAL_COMPLEX(20, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_662, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_662", y_true, y_val, y624_LEN*y624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_662() */
/* END TEST 662 */

/*
 *  BEGIN TEST 663
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_663( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	oski_value_t* y_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1146, 167), MAKE_VAL_COMPLEX(-150, -1284), MAKE_VAL_COMPLEX(-850, 185), MAKE_VAL_COMPLEX(-534, 222), MAKE_VAL_COMPLEX(859, 124), MAKE_VAL_COMPLEX(374, -3), MAKE_VAL_COMPLEX(206, 230), MAKE_VAL_COMPLEX(807, 226), MAKE_VAL_COMPLEX(172, 947), MAKE_VAL_COMPLEX(-756, 1825), MAKE_VAL_COMPLEX(-87, -497), MAKE_VAL_COMPLEX(740, -138), MAKE_VAL_COMPLEX(95, -67), MAKE_VAL_COMPLEX(307, -1290), MAKE_VAL_COMPLEX(-694, -159) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_663, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_663", y_true, y_val, y624_LEN*y624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_663() */
/* END TEST 663 */

/*
 *  BEGIN TEST 664
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_664( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(84, -27), MAKE_VAL_COMPLEX(144, 120), MAKE_VAL_COMPLEX(52, -68), MAKE_VAL_COMPLEX(55, -47), MAKE_VAL_COMPLEX(-91, -228), MAKE_VAL_COMPLEX(-21, -116), MAKE_VAL_COMPLEX(57, 92), MAKE_VAL_COMPLEX(-13, -114), MAKE_VAL_COMPLEX(-155, -187), MAKE_VAL_COMPLEX(-110, 30), MAKE_VAL_COMPLEX(149, 60), MAKE_VAL_COMPLEX(-74, 99), MAKE_VAL_COMPLEX(-74, 51), MAKE_VAL_COMPLEX(102, -197), MAKE_VAL_COMPLEX(219, 16) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_664, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_664", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_664() */
/* END TEST 664 */

/*
 *  BEGIN TEST 665
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_665( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-66, 189), MAKE_VAL_COMPLEX(-99, -111), MAKE_VAL_COMPLEX(65, 95), MAKE_VAL_COMPLEX(-73, -7), MAKE_VAL_COMPLEX(91, 48), MAKE_VAL_COMPLEX(-114, 71), MAKE_VAL_COMPLEX(-102, 43), MAKE_VAL_COMPLEX(-95, 60), MAKE_VAL_COMPLEX(137, 133), MAKE_VAL_COMPLEX(128, 60), MAKE_VAL_COMPLEX(-131, 84), MAKE_VAL_COMPLEX(-52, -153), MAKE_VAL_COMPLEX(20, -15), MAKE_VAL_COMPLEX(-93, 80), MAKE_VAL_COMPLEX(-120, 47) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_665, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_665", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_665() */
/* END TEST 665 */

/*
 *  BEGIN TEST 666
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_666( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-64, -84), MAKE_VAL_COMPLEX(-28, 10), MAKE_VAL_COMPLEX(-74, 24), MAKE_VAL_COMPLEX(28, 24), MAKE_VAL_COMPLEX(60, 100), MAKE_VAL_COMPLEX(90, -20), MAKE_VAL_COMPLEX(-20, -90), MAKE_VAL_COMPLEX(78, -6), MAKE_VAL_COMPLEX(28, 24), MAKE_VAL_COMPLEX(-40, -44), MAKE_VAL_COMPLEX(-58, -74), MAKE_VAL_COMPLEX(88, -12), MAKE_VAL_COMPLEX(18, -38), MAKE_VAL_COMPLEX(34, 68), MAKE_VAL_COMPLEX(-76, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_666, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_666", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_666() */
/* END TEST 666 */

/*
 *  BEGIN TEST 667
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_667( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1056, -389), MAKE_VAL_COMPLEX(532, -1337), MAKE_VAL_COMPLEX(-703, 246), MAKE_VAL_COMPLEX(-355, -411), MAKE_VAL_COMPLEX(-881, 1262), MAKE_VAL_COMPLEX(-810, -24), MAKE_VAL_COMPLEX(25, -787), MAKE_VAL_COMPLEX(-762, -1), MAKE_VAL_COMPLEX(-845, 1689), MAKE_VAL_COMPLEX(250, 932), MAKE_VAL_COMPLEX(-488, -1185), MAKE_VAL_COMPLEX(1110, -219), MAKE_VAL_COMPLEX(451, 268), MAKE_VAL_COMPLEX(-1439, -289), MAKE_VAL_COMPLEX(-681, -1379) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-3, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_667, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_667", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_667() */
/* END TEST 667 */

/*
 *  BEGIN TEST 668
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_668( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(92, 33), MAKE_VAL_COMPLEX(-16, -119), MAKE_VAL_COMPLEX(-1, 14), MAKE_VAL_COMPLEX(-9, -77), MAKE_VAL_COMPLEX(149, -250), MAKE_VAL_COMPLEX(94, -64), MAKE_VAL_COMPLEX(-53, -125), MAKE_VAL_COMPLEX(-86, -23), MAKE_VAL_COMPLEX(89, 47), MAKE_VAL_COMPLEX(-122, 192), MAKE_VAL_COMPLEX(32, 117), MAKE_VAL_COMPLEX(150, -133), MAKE_VAL_COMPLEX(-23, 92), MAKE_VAL_COMPLEX(115, 33), MAKE_VAL_COMPLEX(-127, 7) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_668, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_668", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_668() */
/* END TEST 668 */

/*
 *  BEGIN TEST 669
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_669( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-228, -27), MAKE_VAL_COMPLEX(6, 156), MAKE_VAL_COMPLEX(-28, 82), MAKE_VAL_COMPLEX(55, 65), MAKE_VAL_COMPLEX(1, 260), MAKE_VAL_COMPLEX(-49, -46), MAKE_VAL_COMPLEX(-57, 80), MAKE_VAL_COMPLEX(137, -64), MAKE_VAL_COMPLEX(-43, -59), MAKE_VAL_COMPLEX(46, -182), MAKE_VAL_COMPLEX(-153, -110), MAKE_VAL_COMPLEX(-98, 31), MAKE_VAL_COMPLEX(-4, -139), MAKE_VAL_COMPLEX(-18, -19), MAKE_VAL_COMPLEX(69, 72) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_669, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_669", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_669() */
/* END TEST 669 */

/*
 *  BEGIN TEST 670
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_670( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_670, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_670", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_670() */
/* END TEST 670 */

/*
 *  BEGIN TEST 671
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x624', LAYOUT_COLMAJ>
 *  y : <'y624', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_671( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A624_ptr, A624_ind, A624_val,
		A624_M, A624_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y624_val,
		y624_LEN * y624_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y624_LEN, y624_VW, LAYOUT_COLMAJ, y624_LEN );
	oski_value_t* y_val = array_Duplicate( x624_val,
		x624_LEN * x624_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x624_LEN, x624_VW, LAYOUT_COLMAJ, x624_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(390, 1228), MAKE_VAL_COMPLEX(1009, -650), MAKE_VAL_COMPLEX(204, -63), MAKE_VAL_COMPLEX(456, -482), MAKE_VAL_COMPLEX(2326, -238), MAKE_VAL_COMPLEX(498, 591), MAKE_VAL_COMPLEX(883, -534), MAKE_VAL_COMPLEX(-499, -747), MAKE_VAL_COMPLEX(-144, 798), MAKE_VAL_COMPLEX(-1846, -20), MAKE_VAL_COMPLEX(-549, 1041), MAKE_VAL_COMPLEX(1282, 728), MAKE_VAL_COMPLEX(-903, 351), MAKE_VAL_COMPLEX(42, 759), MAKE_VAL_COMPLEX(-233, -986) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_671, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_671", y_true, y_val, x624_LEN*x624_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_671() */
/* END TEST 671 */

#define A672_M 5    /* # of rows */
#define A672_N 5    /* # of cols */
#define A672_NNZ 10  /* # of stored non-zeros */
static oski_index_t A672_ptr[A672_M+1] = {1, 3, 6, 8, 11, 11}; /* end row pointers for A672 */
static oski_index_t A672_ind[A672_NNZ] = {2, 4, 1, 3, 4, 2, 4, 1, 2, 3}; /* end column indices pointers for A672 */
static oski_value_t A672_val[A672_NNZ] = {MAKE_VAL_COMPLEX(-8, 2), MAKE_VAL_COMPLEX(-5, -7), MAKE_VAL_COMPLEX(-8, 2), MAKE_VAL_COMPLEX(0, 10), MAKE_VAL_COMPLEX(-6, 0), MAKE_VAL_COMPLEX(0, 10), MAKE_VAL_COMPLEX(-10, 2), MAKE_VAL_COMPLEX(-5, -7), MAKE_VAL_COMPLEX(-6, 0), MAKE_VAL_COMPLEX(-10, 2)}; /* end stored non-zero values for A672 */

#define x672_LEN 5   /* vector length */
#define x672_VW  3   /* vector width */
static oski_value_t x672_val[x672_LEN*x672_VW] = {MAKE_VAL_COMPLEX(-1, 9), MAKE_VAL_COMPLEX(-9, -3), MAKE_VAL_COMPLEX(-3, -2), MAKE_VAL_COMPLEX(-10, -4), MAKE_VAL_COMPLEX(-4, -10), MAKE_VAL_COMPLEX(10, -5), MAKE_VAL_COMPLEX(-3, -7), MAKE_VAL_COMPLEX(5, 1), MAKE_VAL_COMPLEX(-10, -2), MAKE_VAL_COMPLEX(-8, 10), MAKE_VAL_COMPLEX(-10, 9), MAKE_VAL_COMPLEX(-1, 6), MAKE_VAL_COMPLEX(1, -8), MAKE_VAL_COMPLEX(9, 3), MAKE_VAL_COMPLEX(1, -2)}; /* data */

#define y672_LEN 5   /* vector length */
#define y672_VW  3   /* vector width */
static oski_value_t y672_val[y672_LEN*y672_VW] = {MAKE_VAL_COMPLEX(-6, 5), MAKE_VAL_COMPLEX(-7, 1), MAKE_VAL_COMPLEX(-6, 3), MAKE_VAL_COMPLEX(6, 7), MAKE_VAL_COMPLEX(2, 4), MAKE_VAL_COMPLEX(10, 4), MAKE_VAL_COMPLEX(-8, 9), MAKE_VAL_COMPLEX(-6, -7), MAKE_VAL_COMPLEX(-7, 3), MAKE_VAL_COMPLEX(3, 3), MAKE_VAL_COMPLEX(-9, -10), MAKE_VAL_COMPLEX(-10, -2), MAKE_VAL_COMPLEX(1, 4), MAKE_VAL_COMPLEX(8, 3), MAKE_VAL_COMPLEX(2, 10)}; /* data */

/*
 *  BEGIN TEST 672
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_672( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	oski_value_t* y_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(251, 43), MAKE_VAL_COMPLEX(190, 132), MAKE_VAL_COMPLEX(16, 59), MAKE_VAL_COMPLEX(116, -232), MAKE_VAL_COMPLEX(140, -36), MAKE_VAL_COMPLEX(48, -207), MAKE_VAL_COMPLEX(183, -211), MAKE_VAL_COMPLEX(169, -85), MAKE_VAL_COMPLEX(84, 66), MAKE_VAL_COMPLEX(170, 30), MAKE_VAL_COMPLEX(-38, 241), MAKE_VAL_COMPLEX(72, 135), MAKE_VAL_COMPLEX(21, -30), MAKE_VAL_COMPLEX(-5, -57), MAKE_VAL_COMPLEX(53, -54) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_672, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_672", y_true, y_val, y672_LEN*y672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_672() */
/* END TEST 672 */

/*
 *  BEGIN TEST 673
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_673( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	oski_value_t* y_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-179, -103), MAKE_VAL_COMPLEX(-106, -144), MAKE_VAL_COMPLEX(56, -95), MAKE_VAL_COMPLEX(-188, 148), MAKE_VAL_COMPLEX(-164, -12), MAKE_VAL_COMPLEX(-168, 159), MAKE_VAL_COMPLEX(-87, 103), MAKE_VAL_COMPLEX(-97, 169), MAKE_VAL_COMPLEX(0, -102), MAKE_VAL_COMPLEX(-206, -66), MAKE_VAL_COMPLEX(146, -121), MAKE_VAL_COMPLEX(48, -111), MAKE_VAL_COMPLEX(-33, -18), MAKE_VAL_COMPLEX(-91, 21), MAKE_VAL_COMPLEX(-77, -66) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_673, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_673", y_true, y_val, y672_LEN*y672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_673() */
/* END TEST 673 */

/*
 *  BEGIN TEST 674
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_674( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	oski_value_t* y_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(10, 12), MAKE_VAL_COMPLEX(2, 14), MAKE_VAL_COMPLEX(6, 12), MAKE_VAL_COMPLEX(14, -12), MAKE_VAL_COMPLEX(8, -4), MAKE_VAL_COMPLEX(8, -20), MAKE_VAL_COMPLEX(18, 16), MAKE_VAL_COMPLEX(-14, 12), MAKE_VAL_COMPLEX(6, 14), MAKE_VAL_COMPLEX(6, -6), MAKE_VAL_COMPLEX(-20, 18), MAKE_VAL_COMPLEX(-4, 20), MAKE_VAL_COMPLEX(8, -2), MAKE_VAL_COMPLEX(6, -16), MAKE_VAL_COMPLEX(20, -4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_674, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_674", y_true, y_val, y672_LEN*y672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_674() */
/* END TEST 674 */

/*
 *  BEGIN TEST 675
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_675( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	oski_value_t* y_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-202, 1461), MAKE_VAL_COMPLEX(-707, 1149), MAKE_VAL_COMPLEX(-272, -116), MAKE_VAL_COMPLEX(1273, 700), MAKE_VAL_COMPLEX(100, 886), MAKE_VAL_COMPLEX(1025, 410), MAKE_VAL_COMPLEX(1560, 807), MAKE_VAL_COMPLEX(946, 1295), MAKE_VAL_COMPLEX(-287, 339), MAKE_VAL_COMPLEX(-375, 1151), MAKE_VAL_COMPLEX(-1171, -115), MAKE_VAL_COMPLEX(-563, 362), MAKE_VAL_COMPLEX(92, 32), MAKE_VAL_COMPLEX(64, 31), MAKE_VAL_COMPLEX(100, 73) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_675, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_675", y_true, y_val, y672_LEN*y672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_675() */
/* END TEST 675 */

/*
 *  BEGIN TEST 676
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_676( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-81, -150), MAKE_VAL_COMPLEX(25, 83), MAKE_VAL_COMPLEX(-28, 78), MAKE_VAL_COMPLEX(28, -141), MAKE_VAL_COMPLEX(242, 50), MAKE_VAL_COMPLEX(17, -20), MAKE_VAL_COMPLEX(-69, 92), MAKE_VAL_COMPLEX(21, 102), MAKE_VAL_COMPLEX(143, 89), MAKE_VAL_COMPLEX(128, -232), MAKE_VAL_COMPLEX(148, -34), MAKE_VAL_COMPLEX(35, -94), MAKE_VAL_COMPLEX(17, 71), MAKE_VAL_COMPLEX(-73, 6), MAKE_VAL_COMPLEX(0, 29) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_676, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_676", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_676() */
/* END TEST 676 */

/*
 *  BEGIN TEST 677
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_677( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(121, 118), MAKE_VAL_COMPLEX(-1, -35), MAKE_VAL_COMPLEX(32, -58), MAKE_VAL_COMPLEX(-4, 197), MAKE_VAL_COMPLEX(-266, 6), MAKE_VAL_COMPLEX(-77, 0), MAKE_VAL_COMPLEX(53, -52), MAKE_VAL_COMPLEX(-37, -126), MAKE_VAL_COMPLEX(-111, -41), MAKE_VAL_COMPLEX(-56, 224), MAKE_VAL_COMPLEX(-72, 38), MAKE_VAL_COMPLEX(-7, 74), MAKE_VAL_COMPLEX(-53, -43), MAKE_VAL_COMPLEX(49, -54), MAKE_VAL_COMPLEX(-12, -25) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_677, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_677", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_677() */
/* END TEST 677 */

/*
 *  BEGIN TEST 678
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_678( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(54, 88), MAKE_VAL_COMPLEX(-102, 36), MAKE_VAL_COMPLEX(-41, 3), MAKE_VAL_COMPLEX(-118, 34), MAKE_VAL_COMPLEX(-106, -62), MAKE_VAL_COMPLEX(55, -115), MAKE_VAL_COMPLEX(-76, -42), MAKE_VAL_COMPLEX(52, -26), MAKE_VAL_COMPLEX(-104, 52), MAKE_VAL_COMPLEX(-2, 146), MAKE_VAL_COMPLEX(-27, 151), MAKE_VAL_COMPLEX(33, 61), MAKE_VAL_COMPLEX(-47, -79), MAKE_VAL_COMPLEX(102, -36), MAKE_VAL_COMPLEX(-5, -25) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_678, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_678", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_678() */
/* END TEST 678 */

/*
 *  BEGIN TEST 679
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_679( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(314, 33), MAKE_VAL_COMPLEX(-209, -391), MAKE_VAL_COMPLEX(-115, -341), MAKE_VAL_COMPLEX(489, 32), MAKE_VAL_COMPLEX(-308, 484), MAKE_VAL_COMPLEX(201, 416), MAKE_VAL_COMPLEX(90, -497), MAKE_VAL_COMPLEX(-376, 25), MAKE_VAL_COMPLEX(-497, -81), MAKE_VAL_COMPLEX(102, 594), MAKE_VAL_COMPLEX(-501, 191), MAKE_VAL_COMPLEX(13, 266), MAKE_VAL_COMPLEX(18, -66), MAKE_VAL_COMPLEX(38, 69), MAKE_VAL_COMPLEX(-20, -27) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-2, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_679, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_679", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_679() */
/* END TEST 679 */

/*
 *  BEGIN TEST 680
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_680( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-114, -43), MAKE_VAL_COMPLEX(94, -90), MAKE_VAL_COMPLEX(-12, -125), MAKE_VAL_COMPLEX(132, -7), MAKE_VAL_COMPLEX(76, 94), MAKE_VAL_COMPLEX(194, 104), MAKE_VAL_COMPLEX(60, -113), MAKE_VAL_COMPLEX(110, 113), MAKE_VAL_COMPLEX(117, -101), MAKE_VAL_COMPLEX(4, -192), MAKE_VAL_COMPLEX(-3, -28), MAKE_VAL_COMPLEX(-11, -91), MAKE_VAL_COMPLEX(35, -8), MAKE_VAL_COMPLEX(14, 45), MAKE_VAL_COMPLEX(12, 10) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_680, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_680", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_680() */
/* END TEST 680 */

/*
 *  BEGIN TEST 681
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_681( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(99, 14), MAKE_VAL_COMPLEX(-61, 81), MAKE_VAL_COMPLEX(25, 125), MAKE_VAL_COMPLEX(-94, -1), MAKE_VAL_COMPLEX(-44, -72), MAKE_VAL_COMPLEX(-214, -69), MAKE_VAL_COMPLEX(-37, 128), MAKE_VAL_COMPLEX(-127, -106), MAKE_VAL_COMPLEX(-83, 87), MAKE_VAL_COMPLEX(0, 146), MAKE_VAL_COMPLEX(15, -19), MAKE_VAL_COMPLEX(2, 71), MAKE_VAL_COMPLEX(-22, 34), MAKE_VAL_COMPLEX(-47, -36), MAKE_VAL_COMPLEX(-11, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_681, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_681", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_681() */
/* END TEST 681 */

/*
 *  BEGIN TEST 682
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_682( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-55, -79), MAKE_VAL_COMPLEX(93, -39), MAKE_VAL_COMPLEX(38, -5), MAKE_VAL_COMPLEX(108, -38), MAKE_VAL_COMPLEX(102, 52), MAKE_VAL_COMPLEX(-45, 110), MAKE_VAL_COMPLEX(73, 35), MAKE_VAL_COMPLEX(-47, 27), MAKE_VAL_COMPLEX(94, -54), MAKE_VAL_COMPLEX(-6, -136), MAKE_VAL_COMPLEX(17, -142), MAKE_VAL_COMPLEX(-34, -55), MAKE_VAL_COMPLEX(48, 71), MAKE_VAL_COMPLEX(-93, 39), MAKE_VAL_COMPLEX(6, 23) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_682, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_682", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_682() */
/* END TEST 682 */

/*
 *  BEGIN TEST 683
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_683( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(907, -81), MAKE_VAL_COMPLEX(-412, 1000), MAKE_VAL_COMPLEX(847, 647), MAKE_VAL_COMPLEX(-1291, 621), MAKE_VAL_COMPLEX(-1222, -670), MAKE_VAL_COMPLEX(-1526, 550), MAKE_VAL_COMPLEX(339, 827), MAKE_VAL_COMPLEX(-1337, 107), MAKE_VAL_COMPLEX(-560, 1270), MAKE_VAL_COMPLEX(772, 1724), MAKE_VAL_COMPLEX(-407, 447), MAKE_VAL_COMPLEX(570, 832), MAKE_VAL_COMPLEX(-13, -67), MAKE_VAL_COMPLEX(-29, 71), MAKE_VAL_COMPLEX(-74, -66) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_683, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_683", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_683() */
/* END TEST 683 */

/*
 *  BEGIN TEST 684
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_684( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	oski_value_t* y_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(172, 58), MAKE_VAL_COMPLEX(64, -156), MAKE_VAL_COMPLEX(70, -202), MAKE_VAL_COMPLEX(201, 89), MAKE_VAL_COMPLEX(15, 10), MAKE_VAL_COMPLEX(210, 104), MAKE_VAL_COMPLEX(93, 45), MAKE_VAL_COMPLEX(150, -178), MAKE_VAL_COMPLEX(-46, 169), MAKE_VAL_COMPLEX(27, 15), MAKE_VAL_COMPLEX(-81, -6), MAKE_VAL_COMPLEX(12, -131), MAKE_VAL_COMPLEX(47, 55), MAKE_VAL_COMPLEX(87, 74), MAKE_VAL_COMPLEX(21, 46) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_684, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_684", y_true, y_val, y672_LEN*y672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_684() */
/* END TEST 684 */

/*
 *  BEGIN TEST 685
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_685( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	oski_value_t* y_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-201, -105), MAKE_VAL_COMPLEX(-64, 106), MAKE_VAL_COMPLEX(-85, 157), MAKE_VAL_COMPLEX(-256, -54), MAKE_VAL_COMPLEX(-45, 0), MAKE_VAL_COMPLEX(-248, -38), MAKE_VAL_COMPLEX(-148, -110), MAKE_VAL_COMPLEX(-95, 143), MAKE_VAL_COMPLEX(32, -221), MAKE_VAL_COMPLEX(-51, 3), MAKE_VAL_COMPLEX(160, -47), MAKE_VAL_COMPLEX(12, 63), MAKE_VAL_COMPLEX(-76, -52), MAKE_VAL_COMPLEX(-116, -21), MAKE_VAL_COMPLEX(-93, -42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_685, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_685", y_true, y_val, y672_LEN*y672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_685() */
/* END TEST 685 */

/*
 *  BEGIN TEST 686
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_686( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	oski_value_t* y_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-19, 26), MAKE_VAL_COMPLEX(-27, 11), MAKE_VAL_COMPLEX(-21, 18), MAKE_VAL_COMPLEX(31, 22), MAKE_VAL_COMPLEX(12, 14), MAKE_VAL_COMPLEX(44, 6), MAKE_VAL_COMPLEX(-23, 44), MAKE_VAL_COMPLEX(-31, -22), MAKE_VAL_COMPLEX(-25, 19), MAKE_VAL_COMPLEX(15, 9), MAKE_VAL_COMPLEX(-46, -31), MAKE_VAL_COMPLEX(-42, 2), MAKE_VAL_COMPLEX(8, 15), MAKE_VAL_COMPLEX(35, 4), MAKE_VAL_COMPLEX(18, 38) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_686, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_686", y_true, y_val, y672_LEN*y672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_686() */
/* END TEST 686 */

/*
 *  BEGIN TEST 687
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_687( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	oski_value_t* y_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1273, -1570), MAKE_VAL_COMPLEX(-1925, 449), MAKE_VAL_COMPLEX(-2309, 864), MAKE_VAL_COMPLEX(-729, -1470), MAKE_VAL_COMPLEX(-75, 95), MAKE_VAL_COMPLEX(-382, -1648), MAKE_VAL_COMPLEX(-1051, -796), MAKE_VAL_COMPLEX(-2351, -364), MAKE_VAL_COMPLEX(1050, -956), MAKE_VAL_COMPLEX(-39, -39), MAKE_VAL_COMPLEX(425, -206), MAKE_VAL_COMPLEX(-1431, 383), MAKE_VAL_COMPLEX(-34, -479), MAKE_VAL_COMPLEX(204, -697), MAKE_VAL_COMPLEX(-75, 101) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_687, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_687", y_true, y_val, y672_LEN*y672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_687() */
/* END TEST 687 */

/*
 *  BEGIN TEST 688
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_688( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(137, -68), MAKE_VAL_COMPLEX(70, 16), MAKE_VAL_COMPLEX(34, -65), MAKE_VAL_COMPLEX(-87, -81), MAKE_VAL_COMPLEX(-72, -69), MAKE_VAL_COMPLEX(-26, -38), MAKE_VAL_COMPLEX(-152, -115), MAKE_VAL_COMPLEX(-14, -169), MAKE_VAL_COMPLEX(-101, -34), MAKE_VAL_COMPLEX(231, 60), MAKE_VAL_COMPLEX(145, 15), MAKE_VAL_COMPLEX(71, 110), MAKE_VAL_COMPLEX(-60, -20), MAKE_VAL_COMPLEX(45, -63), MAKE_VAL_COMPLEX(-12, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_688, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_688", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_688() */
/* END TEST 688 */

/*
 *  BEGIN TEST 689
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_689( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-86, 19), MAKE_VAL_COMPLEX(-31, 47), MAKE_VAL_COMPLEX(-26, 92), MAKE_VAL_COMPLEX(127, 155), MAKE_VAL_COMPLEX(46, 149), MAKE_VAL_COMPLEX(-59, 18), MAKE_VAL_COMPLEX(135, 172), MAKE_VAL_COMPLEX(-11, 138), MAKE_VAL_COMPLEX(151, 96), MAKE_VAL_COMPLEX(-133, -80), MAKE_VAL_COMPLEX(-40, -19), MAKE_VAL_COMPLEX(-35, -141), MAKE_VAL_COMPLEX(14, 63), MAKE_VAL_COMPLEX(-84, 0), MAKE_VAL_COMPLEX(-4, 9) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_689, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_689", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_689() */
/* END TEST 689 */

/*
 *  BEGIN TEST 690
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_690( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-58, -52), MAKE_VAL_COMPLEX(66, -48), MAKE_VAL_COMPLEX(29, -11), MAKE_VAL_COMPLEX(78, -50), MAKE_VAL_COMPLEX(90, 22), MAKE_VAL_COMPLEX(-15, 95), MAKE_VAL_COMPLEX(64, 14), MAKE_VAL_COMPLEX(-32, 30), MAKE_VAL_COMPLEX(64, -60), MAKE_VAL_COMPLEX(-30, -106), MAKE_VAL_COMPLEX(-13, -115), MAKE_VAL_COMPLEX(-37, -37), MAKE_VAL_COMPLEX(51, 47), MAKE_VAL_COMPLEX(-66, 48), MAKE_VAL_COMPLEX(9, 17) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_690, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_690", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_690() */
/* END TEST 690 */

/*
 *  BEGIN TEST 691
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_691( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(122, 998), MAKE_VAL_COMPLEX(-396, 924), MAKE_VAL_COMPLEX(170, 826), MAKE_VAL_COMPLEX(1104, 652), MAKE_VAL_COMPLEX(414, 526), MAKE_VAL_COMPLEX(-216, -328), MAKE_VAL_COMPLEX(1306, 170), MAKE_VAL_COMPLEX(994, 618), MAKE_VAL_COMPLEX(1018, 222), MAKE_VAL_COMPLEX(-930, 1154), MAKE_VAL_COMPLEX(-136, 1052), MAKE_VAL_COMPLEX(-718, -370), MAKE_VAL_COMPLEX(0, -40), MAKE_VAL_COMPLEX(-12, 48), MAKE_VAL_COMPLEX(-60, -52) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_691, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_691", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_691() */
/* END TEST 691 */

/*
 *  BEGIN TEST 692
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_692( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(19, 96), MAKE_VAL_COMPLEX(16, -86), MAKE_VAL_COMPLEX(-30, 69), MAKE_VAL_COMPLEX(-59, 19), MAKE_VAL_COMPLEX(-200, -57), MAKE_VAL_COMPLEX(84, -32), MAKE_VAL_COMPLEX(-108, -47), MAKE_VAL_COMPLEX(206, 35), MAKE_VAL_COMPLEX(-185, 102), MAKE_VAL_COMPLEX(105, 44), MAKE_VAL_COMPLEX(-17, 213), MAKE_VAL_COMPLEX(209, 14), MAKE_VAL_COMPLEX(-38, -66), MAKE_VAL_COMPLEX(93, -27), MAKE_VAL_COMPLEX(-2, -12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_692, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_692", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_692() */
/* END TEST 692 */

/*
 *  BEGIN TEST 693
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_693( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1, -12), MAKE_VAL_COMPLEX(-106, 86), MAKE_VAL_COMPLEX(-3, -78), MAKE_VAL_COMPLEX(-43, -25), MAKE_VAL_COMPLEX(134, -21), MAKE_VAL_COMPLEX(-9, -43), MAKE_VAL_COMPLEX(60, -7), MAKE_VAL_COMPLEX(-158, -41), MAKE_VAL_COMPLEX(89, -90), MAKE_VAL_COMPLEX(-147, 70), MAKE_VAL_COMPLEX(-46, -102), MAKE_VAL_COMPLEX(-200, 43), MAKE_VAL_COMPLEX(23, -9), MAKE_VAL_COMPLEX(-3, 27), MAKE_VAL_COMPLEX(5, -9) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_693, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_693", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_693() */
/* END TEST 693 */

/*
 *  BEGIN TEST 694
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_694( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(72, 8), MAKE_VAL_COMPLEX(-24, 72), MAKE_VAL_COMPLEX(-16, 24), MAKE_VAL_COMPLEX(-32, 80), MAKE_VAL_COMPLEX(-80, 32), MAKE_VAL_COMPLEX(-40, -80), MAKE_VAL_COMPLEX(-56, 24), MAKE_VAL_COMPLEX(8, -40), MAKE_VAL_COMPLEX(-16, 80), MAKE_VAL_COMPLEX(80, 64), MAKE_VAL_COMPLEX(72, 80), MAKE_VAL_COMPLEX(48, 8), MAKE_VAL_COMPLEX(-64, -8), MAKE_VAL_COMPLEX(24, -72), MAKE_VAL_COMPLEX(-16, -8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_694, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_694", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_694() */
/* END TEST 694 */

/*
 *  BEGIN TEST 695
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_ROWMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_695( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_ROWMAJ, x672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-26, -231), MAKE_VAL_COMPLEX(122, 1360), MAKE_VAL_COMPLEX(-517, -394), MAKE_VAL_COMPLEX(-263, 277), MAKE_VAL_COMPLEX(683, -854), MAKE_VAL_COMPLEX(-570, -150), MAKE_VAL_COMPLEX(333, -314), MAKE_VAL_COMPLEX(-1325, 816), MAKE_VAL_COMPLEX(84, -1075), MAKE_VAL_COMPLEX(-190, 1299), MAKE_VAL_COMPLEX(-755, -411), MAKE_VAL_COMPLEX(-850, 1549), MAKE_VAL_COMPLEX(16, -68), MAKE_VAL_COMPLEX(15, 81), MAKE_VAL_COMPLEX(-62, -56) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_695, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_695", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_695() */
/* END TEST 695 */

/*
 *  BEGIN TEST 696
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_696( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	oski_value_t* y_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(91, 71), MAKE_VAL_COMPLEX(94, 95), MAKE_VAL_COMPLEX(-38, -149), MAKE_VAL_COMPLEX(21, -73), MAKE_VAL_COMPLEX(-43, 115), MAKE_VAL_COMPLEX(51, -62), MAKE_VAL_COMPLEX(115, -122), MAKE_VAL_COMPLEX(219, -39), MAKE_VAL_COMPLEX(-153, -64), MAKE_VAL_COMPLEX(128, -4), MAKE_VAL_COMPLEX(-71, -21), MAKE_VAL_COMPLEX(162, 38), MAKE_VAL_COMPLEX(-22, -14), MAKE_VAL_COMPLEX(-36, 36), MAKE_VAL_COMPLEX(-43, -14) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_696, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_696", y_true, y_val, y672_LEN*y672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_696() */
/* END TEST 696 */

/*
 *  BEGIN TEST 697
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_697( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	oski_value_t* y_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-140, -81), MAKE_VAL_COMPLEX(-127, -126), MAKE_VAL_COMPLEX(-1, 131), MAKE_VAL_COMPLEX(-32, 131), MAKE_VAL_COMPLEX(31, -89), MAKE_VAL_COMPLEX(-31, 128), MAKE_VAL_COMPLEX(-192, 118), MAKE_VAL_COMPLEX(-208, -19), MAKE_VAL_COMPLEX(110, 41), MAKE_VAL_COMPLEX(-131, 31), MAKE_VAL_COMPLEX(85, -64), MAKE_VAL_COMPLEX(-192, -96), MAKE_VAL_COMPLEX(6, 35), MAKE_VAL_COMPLEX(53, 16), MAKE_VAL_COMPLEX(1, 64) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_697, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_697", y_true, y_val, y672_LEN*y672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_697() */
/* END TEST 697 */

/*
 *  BEGIN TEST 698
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_698( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	oski_value_t* y_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(70, 23), MAKE_VAL_COMPLEX(43, 51), MAKE_VAL_COMPLEX(54, 33), MAKE_VAL_COMPLEX(26, -83), MAKE_VAL_COMPLEX(22, -36), MAKE_VAL_COMPLEX(-18, -100), MAKE_VAL_COMPLEX(112, 19), MAKE_VAL_COMPLEX(-26, 83), MAKE_VAL_COMPLEX(59, 41), MAKE_VAL_COMPLEX(9, -39), MAKE_VAL_COMPLEX(-35, 122), MAKE_VAL_COMPLEX(34, 90), MAKE_VAL_COMPLEX(27, -28), MAKE_VAL_COMPLEX(-16, -79), MAKE_VAL_COMPLEX(70, -66) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_698, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_698", y_true, y_val, y672_LEN*y672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_698() */
/* END TEST 698 */

/*
 *  BEGIN TEST 699
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_699( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	oski_value_t* y_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-291, -519), MAKE_VAL_COMPLEX(-211, -584), MAKE_VAL_COMPLEX(33, 514), MAKE_VAL_COMPLEX(-327, 271), MAKE_VAL_COMPLEX(207, -437), MAKE_VAL_COMPLEX(-442, 289), MAKE_VAL_COMPLEX(-612, 153), MAKE_VAL_COMPLEX(-745, -63), MAKE_VAL_COMPLEX(590, 275), MAKE_VAL_COMPLEX(-594, -106), MAKE_VAL_COMPLEX(511, 149), MAKE_VAL_COMPLEX(-462, -430), MAKE_VAL_COMPLEX(6, 44), MAKE_VAL_COMPLEX(42, -32), MAKE_VAL_COMPLEX(-6, 7) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_699, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_699", y_true, y_val, y672_LEN*y672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_699() */
/* END TEST 699 */

/*
 *  BEGIN TEST 700
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_700( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1, -50), MAKE_VAL_COMPLEX(-103, -86), MAKE_VAL_COMPLEX(-134, 62), MAKE_VAL_COMPLEX(46, -66), MAKE_VAL_COMPLEX(-77, 12), MAKE_VAL_COMPLEX(-71, 6), MAKE_VAL_COMPLEX(125, -11), MAKE_VAL_COMPLEX(81, 62), MAKE_VAL_COMPLEX(61, 134), MAKE_VAL_COMPLEX(62, 79), MAKE_VAL_COMPLEX(-15, 159), MAKE_VAL_COMPLEX(122, -71), MAKE_VAL_COMPLEX(3, 80), MAKE_VAL_COMPLEX(84, -100), MAKE_VAL_COMPLEX(-10, -1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_700, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_700", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_700() */
/* END TEST 700 */

/*
 *  BEGIN TEST 701
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_701( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(142, 11), MAKE_VAL_COMPLEX(112, 239), MAKE_VAL_COMPLEX(122, -5), MAKE_VAL_COMPLEX(-46, 240), MAKE_VAL_COMPLEX(-49, 108), MAKE_VAL_COMPLEX(-64, -126), MAKE_VAL_COMPLEX(-212, 98), MAKE_VAL_COMPLEX(-96, -143), MAKE_VAL_COMPLEX(-31, 28), MAKE_VAL_COMPLEX(136, -19), MAKE_VAL_COMPLEX(210, -63), MAKE_VAL_COMPLEX(-26, 50), MAKE_VAL_COMPLEX(-129, -47), MAKE_VAL_COMPLEX(-93, -53), MAKE_VAL_COMPLEX(-26, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_701, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_701", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_701() */
/* END TEST 701 */

/*
 *  BEGIN TEST 702
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_702( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-58, -52), MAKE_VAL_COMPLEX(66, -48), MAKE_VAL_COMPLEX(29, -11), MAKE_VAL_COMPLEX(78, -50), MAKE_VAL_COMPLEX(90, 22), MAKE_VAL_COMPLEX(-15, 95), MAKE_VAL_COMPLEX(64, 14), MAKE_VAL_COMPLEX(-32, 30), MAKE_VAL_COMPLEX(64, -60), MAKE_VAL_COMPLEX(-30, -106), MAKE_VAL_COMPLEX(-13, -115), MAKE_VAL_COMPLEX(-37, -37), MAKE_VAL_COMPLEX(51, 47), MAKE_VAL_COMPLEX(-66, 48), MAKE_VAL_COMPLEX(9, 17) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_702, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_702", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_702() */
/* END TEST 702 */

/*
 *  BEGIN TEST 703
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_703( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-96, 1165), MAKE_VAL_COMPLEX(-550, 1745), MAKE_VAL_COMPLEX(1398, 489), MAKE_VAL_COMPLEX(-1846, 340), MAKE_VAL_COMPLEX(83, -68), MAKE_VAL_COMPLEX(1222, -266), MAKE_VAL_COMPLEX(-1722, -1296), MAKE_VAL_COMPLEX(178, -1337), MAKE_VAL_COMPLEX(-299, -1432), MAKE_VAL_COMPLEX(-108, -139), MAKE_VAL_COMPLEX(1036, -235), MAKE_VAL_COMPLEX(-1494, 148), MAKE_VAL_COMPLEX(359, -1377), MAKE_VAL_COMPLEX(-767, 81), MAKE_VAL_COMPLEX(74, -100) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-9, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_703, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_703", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_703() */
/* END TEST 703 */

/*
 *  BEGIN TEST 704
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_704( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-38, -71), MAKE_VAL_COMPLEX(142, 83), MAKE_VAL_COMPLEX(36, -71), MAKE_VAL_COMPLEX(64, -132), MAKE_VAL_COMPLEX(-31, 40), MAKE_VAL_COMPLEX(60, -78), MAKE_VAL_COMPLEX(22, 156), MAKE_VAL_COMPLEX(98, 69), MAKE_VAL_COMPLEX(65, 38), MAKE_VAL_COMPLEX(64, 15), MAKE_VAL_COMPLEX(42, -87), MAKE_VAL_COMPLEX(180, 66), MAKE_VAL_COMPLEX(95, -45), MAKE_VAL_COMPLEX(9, -141), MAKE_VAL_COMPLEX(-8, 10) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_704, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_704", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_704() */
/* END TEST 704 */

/*
 *  BEGIN TEST 705
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_705( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(25, 24), MAKE_VAL_COMPLEX(-91, -86), MAKE_VAL_COMPLEX(-17, 75), MAKE_VAL_COMPLEX(-6, 132), MAKE_VAL_COMPLEX(71, 2), MAKE_VAL_COMPLEX(-100, 123), MAKE_VAL_COMPLEX(7, -127), MAKE_VAL_COMPLEX(-125, -64), MAKE_VAL_COMPLEX(-11, -48), MAKE_VAL_COMPLEX(-44, -81), MAKE_VAL_COMPLEX(-10, 22), MAKE_VAL_COMPLEX(-187, -98), MAKE_VAL_COMPLEX(-84, 87), MAKE_VAL_COMPLEX(-60, 144), MAKE_VAL_COMPLEX(7, 2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_705, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_705", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_705() */
/* END TEST 705 */

/*
 *  BEGIN TEST 706
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_706( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(69, -47), MAKE_VAL_COMPLEX(33, 81), MAKE_VAL_COMPLEX(4, 33), MAKE_VAL_COMPLEX(32, 94), MAKE_VAL_COMPLEX(-46, 88), MAKE_VAL_COMPLEX(-95, -40), MAKE_VAL_COMPLEX(-31, 63), MAKE_VAL_COMPLEX(-23, -41), MAKE_VAL_COMPLEX(46, 82), MAKE_VAL_COMPLEX(118, -4), MAKE_VAL_COMPLEX(123, 16), MAKE_VAL_COMPLEX(48, -29), MAKE_VAL_COMPLEX(-62, 41), MAKE_VAL_COMPLEX(-33, -81), MAKE_VAL_COMPLEX(-20, 5) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_706, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_706", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_706() */
/* END TEST 706 */

/*
 *  BEGIN TEST 707
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_707( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_ROWMAJ, y672_VW );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(200, 792), MAKE_VAL_COMPLEX(-750, -1018), MAKE_VAL_COMPLEX(-697, 400), MAKE_VAL_COMPLEX(-1320, 902), MAKE_VAL_COMPLEX(55, -73), MAKE_VAL_COMPLEX(-895, -124), MAKE_VAL_COMPLEX(371, -1159), MAKE_VAL_COMPLEX(-268, -1130), MAKE_VAL_COMPLEX(-411, -277), MAKE_VAL_COMPLEX(-107, -81), MAKE_VAL_COMPLEX(-512, 801), MAKE_VAL_COMPLEX(-740, -1281), MAKE_VAL_COMPLEX(-1146, -257), MAKE_VAL_COMPLEX(-597, 669), MAKE_VAL_COMPLEX(48, -79) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_707, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_707", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_707() */
/* END TEST 707 */

/*
 *  BEGIN TEST 708
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_708( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	oski_value_t* y_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(96, 138), MAKE_VAL_COMPLEX(43, -59), MAKE_VAL_COMPLEX(126, -45), MAKE_VAL_COMPLEX(185, -7), MAKE_VAL_COMPLEX(14, -4), MAKE_VAL_COMPLEX(126, 107), MAKE_VAL_COMPLEX(-20, 166), MAKE_VAL_COMPLEX(140, -32), MAKE_VAL_COMPLEX(-141, 25), MAKE_VAL_COMPLEX(10, 10), MAKE_VAL_COMPLEX(-95, -122), MAKE_VAL_COMPLEX(51, -70), MAKE_VAL_COMPLEX(-140, -21), MAKE_VAL_COMPLEX(167, 59), MAKE_VAL_COMPLEX(37, 22) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_708, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_708", y_true, y_val, y672_LEN*y672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_708() */
/* END TEST 708 */

/*
 *  BEGIN TEST 709
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_709( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	oski_value_t* y_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-103, -122), MAKE_VAL_COMPLEX(-56, 68), MAKE_VAL_COMPLEX(-135, 57), MAKE_VAL_COMPLEX(-166, 15), MAKE_VAL_COMPLEX(-6, 10), MAKE_VAL_COMPLEX(-102, -109), MAKE_VAL_COMPLEX(13, -140), MAKE_VAL_COMPLEX(-159, 24), MAKE_VAL_COMPLEX(130, -12), MAKE_VAL_COMPLEX(-1, -7), MAKE_VAL_COMPLEX(67, 111), MAKE_VAL_COMPLEX(-73, 76), MAKE_VAL_COMPLEX(146, 28), MAKE_VAL_COMPLEX(-148, -61), MAKE_VAL_COMPLEX(-23, -4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_709, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_709", y_true, y_val, y672_LEN*y672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_709() */
/* END TEST 709 */

/*
 *  BEGIN TEST 710
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_710( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	oski_value_t* y_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(77, 7), MAKE_VAL_COMPLEX(56, 42), MAKE_VAL_COMPLEX(63, 21), MAKE_VAL_COMPLEX(7, -91), MAKE_VAL_COMPLEX(14, -42), MAKE_VAL_COMPLEX(-42, -98), MAKE_VAL_COMPLEX(119, -7), MAKE_VAL_COMPLEX(-7, 91), MAKE_VAL_COMPLEX(70, 28), MAKE_VAL_COMPLEX(0, -42), MAKE_VAL_COMPLEX(-7, 133), MAKE_VAL_COMPLEX(56, 84), MAKE_VAL_COMPLEX(21, -35), MAKE_VAL_COMPLEX(-35, -77), MAKE_VAL_COMPLEX(56, -84) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_710, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_710", y_true, y_val, y672_LEN*y672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_710() */
/* END TEST 710 */

/*
 *  BEGIN TEST 711
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_711( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	oski_value_t* y_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-601, -17), MAKE_VAL_COMPLEX(74, 438), MAKE_VAL_COMPLEX(-180, 624), MAKE_VAL_COMPLEX(-407, 455), MAKE_VAL_COMPLEX(44, 12), MAKE_VAL_COMPLEX(-633, -137), MAKE_VAL_COMPLEX(-259, -415), MAKE_VAL_COMPLEX(-451, 637), MAKE_VAL_COMPLEX(412, -368), MAKE_VAL_COMPLEX(-6, -60), MAKE_VAL_COMPLEX(470, 262), MAKE_VAL_COMPLEX(29, 555), MAKE_VAL_COMPLEX(558, -380), MAKE_VAL_COMPLEX(-629, 169), MAKE_VAL_COMPLEX(11, -3) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-3, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_711, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_711", y_true, y_val, y672_LEN*y672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_711() */
/* END TEST 711 */

/*
 *  BEGIN TEST 712
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_712( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(75, -84), MAKE_VAL_COMPLEX(-47, -147), MAKE_VAL_COMPLEX(-95, -124), MAKE_VAL_COMPLEX(153, -18), MAKE_VAL_COMPLEX(-12, -2), MAKE_VAL_COMPLEX(117, -65), MAKE_VAL_COMPLEX(6, -85), MAKE_VAL_COMPLEX(-26, -135), MAKE_VAL_COMPLEX(81, -75), MAKE_VAL_COMPLEX(5, 21), MAKE_VAL_COMPLEX(55, -66), MAKE_VAL_COMPLEX(-1, 59), MAKE_VAL_COMPLEX(-72, -119), MAKE_VAL_COMPLEX(37, 84), MAKE_VAL_COMPLEX(1, 7) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_712, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_712", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_712() */
/* END TEST 712 */

/*
 *  BEGIN TEST 713
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_713( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-110, 71), MAKE_VAL_COMPLEX(68, 114), MAKE_VAL_COMPLEX(106, 114), MAKE_VAL_COMPLEX(-127, -18), MAKE_VAL_COMPLEX(56, -4), MAKE_VAL_COMPLEX(-107, 110), MAKE_VAL_COMPLEX(25, 80), MAKE_VAL_COMPLEX(17, 154), MAKE_VAL_COMPLEX(-63, 37), MAKE_VAL_COMPLEX(-37, -63), MAKE_VAL_COMPLEX(-81, 17), MAKE_VAL_COMPLEX(-22, -69), MAKE_VAL_COMPLEX(103, 131), MAKE_VAL_COMPLEX(-58, -51), MAKE_VAL_COMPLEX(6, -1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_713, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_713", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_713() */
/* END TEST 713 */

/*
 *  BEGIN TEST 714
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_714( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(65, 71), MAKE_VAL_COMPLEX(-87, 51), MAKE_VAL_COMPLEX(-37, 10), MAKE_VAL_COMPLEX(-102, 52), MAKE_VAL_COMPLEX(-108, -38), MAKE_VAL_COMPLEX(30, -115), MAKE_VAL_COMPLEX(-77, -25), MAKE_VAL_COMPLEX(43, -33), MAKE_VAL_COMPLEX(-86, 66), MAKE_VAL_COMPLEX(24, 134), MAKE_VAL_COMPLEX(2, 143), MAKE_VAL_COMPLEX(41, 50), MAKE_VAL_COMPLEX(-57, -64), MAKE_VAL_COMPLEX(87, -51), MAKE_VAL_COMPLEX(-9, -22) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_714, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_714", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_714() */
/* END TEST 714 */

/*
 *  BEGIN TEST 715
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_715( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-108, 423), MAKE_VAL_COMPLEX(299, 649), MAKE_VAL_COMPLEX(480, 437), MAKE_VAL_COMPLEX(-642, 345), MAKE_VAL_COMPLEX(-70, 44), MAKE_VAL_COMPLEX(-463, 257), MAKE_VAL_COMPLEX(-23, 382), MAKE_VAL_COMPLEX(251, 454), MAKE_VAL_COMPLEX(-273, 501), MAKE_VAL_COMPLEX(79, 17), MAKE_VAL_COMPLEX(-46, 439), MAKE_VAL_COMPLEX(17, -225), MAKE_VAL_COMPLEX(311, 392), MAKE_VAL_COMPLEX(-196, -407), MAKE_VAL_COMPLEX(-35, -39) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_715, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_715", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_715() */
/* END TEST 715 */

/*
 *  BEGIN TEST 716
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_716( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-21, 46), MAKE_VAL_COMPLEX(15, -9), MAKE_VAL_COMPLEX(-53, -12), MAKE_VAL_COMPLEX(75, -86), MAKE_VAL_COMPLEX(-20, -22), MAKE_VAL_COMPLEX(131, -141), MAKE_VAL_COMPLEX(-124, -19), MAKE_VAL_COMPLEX(176, 55), MAKE_VAL_COMPLEX(-23, 85), MAKE_VAL_COMPLEX(-11, 41), MAKE_VAL_COMPLEX(-15, 104), MAKE_VAL_COMPLEX(37, 87), MAKE_VAL_COMPLEX(-98, 33), MAKE_VAL_COMPLEX(211, -40), MAKE_VAL_COMPLEX(3, 3) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_716, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_716", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_716() */
/* END TEST 716 */

/*
 *  BEGIN TEST 717
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_717( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-44, -35), MAKE_VAL_COMPLEX(-12, -60), MAKE_VAL_COMPLEX(61, -13), MAKE_VAL_COMPLEX(-67, 8), MAKE_VAL_COMPLEX(82, -26), MAKE_VAL_COMPLEX(-76, 201), MAKE_VAL_COMPLEX(167, -16), MAKE_VAL_COMPLEX(-173, -18), MAKE_VAL_COMPLEX(17, -159), MAKE_VAL_COMPLEX(-75, -77), MAKE_VAL_COMPLEX(-68, -156), MAKE_VAL_COMPLEX(-81, -82), MAKE_VAL_COMPLEX(156, -42), MAKE_VAL_COMPLEX(-214, 109), MAKE_VAL_COMPLEX(13, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_717, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_717", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_717() */
/* END TEST 717 */

/*
 *  BEGIN TEST 718
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_718( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(70, -56), MAKE_VAL_COMPLEX(42, 84), MAKE_VAL_COMPLEX(7, 35), MAKE_VAL_COMPLEX(42, 98), MAKE_VAL_COMPLEX(-42, 98), MAKE_VAL_COMPLEX(-105, -35), MAKE_VAL_COMPLEX(-28, 70), MAKE_VAL_COMPLEX(-28, -42), MAKE_VAL_COMPLEX(56, 84), MAKE_VAL_COMPLEX(126, -14), MAKE_VAL_COMPLEX(133, 7), MAKE_VAL_COMPLEX(49, -35), MAKE_VAL_COMPLEX(-63, 49), MAKE_VAL_COMPLEX(-42, -84), MAKE_VAL_COMPLEX(-21, 7) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_718, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_718", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_718() */
/* END TEST 718 */

/*
 *  BEGIN TEST 719
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x672', LAYOUT_COLMAJ>
 *  y : <'y672', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_719( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A672_ptr, A672_ind, A672_val,
		A672_M, A672_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y672_val,
		y672_LEN * y672_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y672_LEN, y672_VW, LAYOUT_COLMAJ, y672_LEN );
	oski_value_t* y_val = array_Duplicate( x672_val,
		x672_LEN * x672_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x672_LEN, x672_VW, LAYOUT_COLMAJ, x672_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(21, -223), MAKE_VAL_COMPLEX(-222, 360), MAKE_VAL_COMPLEX(251, -255), MAKE_VAL_COMPLEX(-25, 1179), MAKE_VAL_COMPLEX(-82, -58), MAKE_VAL_COMPLEX(332, 1272), MAKE_VAL_COMPLEX(515, -787), MAKE_VAL_COMPLEX(-1167, 831), MAKE_VAL_COMPLEX(-676, -334), MAKE_VAL_COMPLEX(-64, 82), MAKE_VAL_COMPLEX(-541, -219), MAKE_VAL_COMPLEX(-640, -64), MAKE_VAL_COMPLEX(49, -991), MAKE_VAL_COMPLEX(-565, 1467), MAKE_VAL_COMPLEX(-78, -50) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_719, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_719", y_true, y_val, x672_LEN*x672_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_719() */
/* END TEST 719 */

#define A720_M 5    /* # of rows */
#define A720_N 5    /* # of cols */
#define A720_NNZ 8  /* # of stored non-zeros */
static oski_index_t A720_ptr[A720_M+1] = {0, 1, 3, 4, 5, 8}; /* end row pointers for A720 */
static oski_index_t A720_ind[A720_NNZ] = {4, 2, 4, 1, 4, 0, 1, 3}; /* end column indices pointers for A720 */
static oski_value_t A720_val[A720_NNZ] = {MAKE_VAL_COMPLEX(-7, 1), MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(-1, -2), MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(-9, -5), MAKE_VAL_COMPLEX(-7, 1), MAKE_VAL_COMPLEX(-1, -2), MAKE_VAL_COMPLEX(-9, -5)}; /* end stored non-zero values for A720 */

#define x720_LEN 5   /* vector length */
#define x720_VW  3   /* vector width */
static oski_value_t x720_val[x720_LEN*x720_VW] = {MAKE_VAL_COMPLEX(-10, 0), MAKE_VAL_COMPLEX(2, 8), MAKE_VAL_COMPLEX(1, 2), MAKE_VAL_COMPLEX(-8, -10), MAKE_VAL_COMPLEX(10, 4), MAKE_VAL_COMPLEX(-9, -4), MAKE_VAL_COMPLEX(7, -9), MAKE_VAL_COMPLEX(6, 4), MAKE_VAL_COMPLEX(1, -6), MAKE_VAL_COMPLEX(-5, 7), MAKE_VAL_COMPLEX(-6, 10), MAKE_VAL_COMPLEX(6, -6), MAKE_VAL_COMPLEX(-0, -6), MAKE_VAL_COMPLEX(1, 6), MAKE_VAL_COMPLEX(-5, 1)}; /* data */

#define y720_LEN 5   /* vector length */
#define y720_VW  3   /* vector width */
static oski_value_t y720_val[y720_LEN*y720_VW] = {MAKE_VAL_COMPLEX(-5, -6), MAKE_VAL_COMPLEX(-10, -1), MAKE_VAL_COMPLEX(-6, -10), MAKE_VAL_COMPLEX(9, 9), MAKE_VAL_COMPLEX(-3, -3), MAKE_VAL_COMPLEX(8, -7), MAKE_VAL_COMPLEX(9, 0), MAKE_VAL_COMPLEX(10, -10), MAKE_VAL_COMPLEX(2, 8), MAKE_VAL_COMPLEX(3, 10), MAKE_VAL_COMPLEX(9, 5), MAKE_VAL_COMPLEX(-9, 2), MAKE_VAL_COMPLEX(1, 8), MAKE_VAL_COMPLEX(-5, 6), MAKE_VAL_COMPLEX(-8, 0)}; /* data */

/*
 *  BEGIN TEST 720
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_720( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	oski_value_t* y_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(30, 95), MAKE_VAL_COMPLEX(68, -15), MAKE_VAL_COMPLEX(73, 76), MAKE_VAL_COMPLEX(-243, -65), MAKE_VAL_COMPLEX(22, -77), MAKE_VAL_COMPLEX(-143, 103), MAKE_VAL_COMPLEX(-85, 162), MAKE_VAL_COMPLEX(-144, -66), MAKE_VAL_COMPLEX(43, 58), MAKE_VAL_COMPLEX(-49, -22), MAKE_VAL_COMPLEX(-52, -98), MAKE_VAL_COMPLEX(130, 3), MAKE_VAL_COMPLEX(138, -93), MAKE_VAL_COMPLEX(127, -175), MAKE_VAL_COMPLEX(-33, 42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_720, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_720", y_true, y_val, y720_LEN*y720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_720() */
/* END TEST 720 */

/*
 *  BEGIN TEST 721
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_721( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	oski_value_t* y_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(69, -25), MAKE_VAL_COMPLEX(86, -10), MAKE_VAL_COMPLEX(57, 50), MAKE_VAL_COMPLEX(72, -34), MAKE_VAL_COMPLEX(35, 110), MAKE_VAL_COMPLEX(51, 34), MAKE_VAL_COMPLEX(-50, -126), MAKE_VAL_COMPLEX(34, 256), MAKE_VAL_COMPLEX(-105, -170), MAKE_VAL_COMPLEX(-36, -116), MAKE_VAL_COMPLEX(-103, 59), MAKE_VAL_COMPLEX(-3, -69), MAKE_VAL_COMPLEX(-185, -23), MAKE_VAL_COMPLEX(-76, 65), MAKE_VAL_COMPLEX(153, -74) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_721, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_721", y_true, y_val, y720_LEN*y720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_721() */
/* END TEST 721 */

/*
 *  BEGIN TEST 722
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_722( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	oski_value_t* y_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-55, -5), MAKE_VAL_COMPLEX(-55, 45), MAKE_VAL_COMPLEX(-80, -20), MAKE_VAL_COMPLEX(90, 0), MAKE_VAL_COMPLEX(-30, 0), MAKE_VAL_COMPLEX(5, -75), MAKE_VAL_COMPLEX(45, -45), MAKE_VAL_COMPLEX(0, -100), MAKE_VAL_COMPLEX(50, 30), MAKE_VAL_COMPLEX(65, 35), MAKE_VAL_COMPLEX(70, -20), MAKE_VAL_COMPLEX(-35, 55), MAKE_VAL_COMPLEX(45, 35), MAKE_VAL_COMPLEX(5, 55), MAKE_VAL_COMPLEX(-40, 40) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_722, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_722", y_true, y_val, y720_LEN*y720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_722() */
/* END TEST 722 */

/*
 *  BEGIN TEST 723
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_723( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	oski_value_t* y_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(173, 77), MAKE_VAL_COMPLEX(-208, -130), MAKE_VAL_COMPLEX(105, -262), MAKE_VAL_COMPLEX(-494, 876), MAKE_VAL_COMPLEX(-407, -352), MAKE_VAL_COMPLEX(93, 474), MAKE_VAL_COMPLEX(690, 628), MAKE_VAL_COMPLEX(-586, -172), MAKE_VAL_COMPLEX(603, 222), MAKE_VAL_COMPLEX(84, 394), MAKE_VAL_COMPLEX(-141, -119), MAKE_VAL_COMPLEX(149, -247), MAKE_VAL_COMPLEX(251, -589), MAKE_VAL_COMPLEX(-352, -725), MAKE_VAL_COMPLEX(-195, 434) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, -4);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_723, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_723", y_true, y_val, y720_LEN*y720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_723() */
/* END TEST 723 */

/*
 *  BEGIN TEST 724
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_724( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(60, -131), MAKE_VAL_COMPLEX(-53, -98), MAKE_VAL_COMPLEX(28, -27), MAKE_VAL_COMPLEX(68, -67), MAKE_VAL_COMPLEX(-294, 39), MAKE_VAL_COMPLEX(176, -122), MAKE_VAL_COMPLEX(16, -59), MAKE_VAL_COMPLEX(-66, 60), MAKE_VAL_COMPLEX(-114, 53), MAKE_VAL_COMPLEX(25, -158), MAKE_VAL_COMPLEX(62, -146), MAKE_VAL_COMPLEX(57, 132), MAKE_VAL_COMPLEX(116, -39), MAKE_VAL_COMPLEX(-43, -119), MAKE_VAL_COMPLEX(146, 39) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_724, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_724", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_724() */
/* END TEST 724 */

/*
 *  BEGIN TEST 725
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_725( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-40, 11), MAKE_VAL_COMPLEX(-47, 106), MAKE_VAL_COMPLEX(-54, 35), MAKE_VAL_COMPLEX(68, -9), MAKE_VAL_COMPLEX(226, 73), MAKE_VAL_COMPLEX(-110, 22), MAKE_VAL_COMPLEX(78, 161), MAKE_VAL_COMPLEX(6, 4), MAKE_VAL_COMPLEX(184, -29), MAKE_VAL_COMPLEX(-99, 84), MAKE_VAL_COMPLEX(-170, 54), MAKE_VAL_COMPLEX(3, -48), MAKE_VAL_COMPLEX(-44, 51), MAKE_VAL_COMPLEX(-31, 119), MAKE_VAL_COMPLEX(-148, -101) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_725, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_725", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_725() */
/* END TEST 725 */

/*
 *  BEGIN TEST 726
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_726( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, -90), MAKE_VAL_COMPLEX(-72, 18), MAKE_VAL_COMPLEX(-18, 9), MAKE_VAL_COMPLEX(90, -72), MAKE_VAL_COMPLEX(-36, 90), MAKE_VAL_COMPLEX(36, -81), MAKE_VAL_COMPLEX(81, 63), MAKE_VAL_COMPLEX(-36, 54), MAKE_VAL_COMPLEX(54, 9), MAKE_VAL_COMPLEX(-63, -45), MAKE_VAL_COMPLEX(-90, -54), MAKE_VAL_COMPLEX(54, 54), MAKE_VAL_COMPLEX(54, 0), MAKE_VAL_COMPLEX(-54, 9), MAKE_VAL_COMPLEX(-9, -45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_726, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_726", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_726() */
/* END TEST 726 */

/*
 *  BEGIN TEST 727
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_727( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-425, -397), MAKE_VAL_COMPLEX(-167, -433), MAKE_VAL_COMPLEX(244, -373), MAKE_VAL_COMPLEX(-831, -343), MAKE_VAL_COMPLEX(-1345, 999), MAKE_VAL_COMPLEX(123, -1072), MAKE_VAL_COMPLEX(-788, -1238), MAKE_VAL_COMPLEX(280, 334), MAKE_VAL_COMPLEX(-1006, 745), MAKE_VAL_COMPLEX(-136, -688), MAKE_VAL_COMPLEX(410, -650), MAKE_VAL_COMPLEX(681, 33), MAKE_VAL_COMPLEX(125, -967), MAKE_VAL_COMPLEX(-385, -586), MAKE_VAL_COMPLEX(1204, -28) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_727, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_727", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_727() */
/* END TEST 727 */

/*
 *  BEGIN TEST 728
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_728( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-14, -3), MAKE_VAL_COMPLEX(81, -42), MAKE_VAL_COMPLEX(63, -6), MAKE_VAL_COMPLEX(-166, 131), MAKE_VAL_COMPLEX(24, 125), MAKE_VAL_COMPLEX(-117, -33), MAKE_VAL_COMPLEX(-218, -51), MAKE_VAL_COMPLEX(100, -42), MAKE_VAL_COMPLEX(-43, 146), MAKE_VAL_COMPLEX(-9, -20), MAKE_VAL_COMPLEX(78, -28), MAKE_VAL_COMPLEX(33, -80), MAKE_VAL_COMPLEX(-110, -17), MAKE_VAL_COMPLEX(4, 20), MAKE_VAL_COMPLEX(102, 67) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_728, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_728", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_728() */
/* END TEST 728 */

/*
 *  BEGIN TEST 729
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_729( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-76, 153), MAKE_VAL_COMPLEX(57, 84), MAKE_VAL_COMPLEX(-24, 9), MAKE_VAL_COMPLEX(-56, -101), MAKE_VAL_COMPLEX(126, -239), MAKE_VAL_COMPLEX(-24, 132), MAKE_VAL_COMPLEX(146, -135), MAKE_VAL_COMPLEX(14, -12), MAKE_VAL_COMPLEX(-38, -215), MAKE_VAL_COMPLEX(69, 158), MAKE_VAL_COMPLEX(18, 208), MAKE_VAL_COMPLEX(-69, -64), MAKE_VAL_COMPLEX(20, -37), MAKE_VAL_COMPLEX(95, 19), MAKE_VAL_COMPLEX(-132, 17) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_729, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_729", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_729() */
/* END TEST 729 */

/*
 *  BEGIN TEST 730
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_730( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(60, 50), MAKE_VAL_COMPLEX(28, -58), MAKE_VAL_COMPLEX(4, -17), MAKE_VAL_COMPLEX(-2, 100), MAKE_VAL_COMPLEX(-40, -74), MAKE_VAL_COMPLEX(34, 69), MAKE_VAL_COMPLEX(-87, 19), MAKE_VAL_COMPLEX(-16, -54), MAKE_VAL_COMPLEX(-36, 31), MAKE_VAL_COMPLEX(65, -17), MAKE_VAL_COMPLEX(86, -30), MAKE_VAL_COMPLEX(-66, 6), MAKE_VAL_COMPLEX(-30, 36), MAKE_VAL_COMPLEX(24, -41), MAKE_VAL_COMPLEX(35, 19) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_730, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_730", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_730() */
/* END TEST 730 */

/*
 *  BEGIN TEST 731
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_731( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-102, 101), MAKE_VAL_COMPLEX(-87, -46), MAKE_VAL_COMPLEX(-50, -105), MAKE_VAL_COMPLEX(270, 147), MAKE_VAL_COMPLEX(364, -199), MAKE_VAL_COMPLEX(-76, 286), MAKE_VAL_COMPLEX(130, 339), MAKE_VAL_COMPLEX(-90, -156), MAKE_VAL_COMPLEX(304, -133), MAKE_VAL_COMPLEX(-37, 150), MAKE_VAL_COMPLEX(-130, 24), MAKE_VAL_COMPLEX(-169, -94), MAKE_VAL_COMPLEX(34, 171), MAKE_VAL_COMPLEX(89, 31), MAKE_VAL_COMPLEX(-16, -225) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_731, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_731", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_731() */
/* END TEST 731 */

/*
 *  BEGIN TEST 732
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_732( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	oski_value_t* y_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(5, 126), MAKE_VAL_COMPLEX(-96, 85), MAKE_VAL_COMPLEX(-19, 297), MAKE_VAL_COMPLEX(-62, -74), MAKE_VAL_COMPLEX(147, 17), MAKE_VAL_COMPLEX(-125, -18), MAKE_VAL_COMPLEX(-80, -158), MAKE_VAL_COMPLEX(-204, -106), MAKE_VAL_COMPLEX(45, -133), MAKE_VAL_COMPLEX(114, -240), MAKE_VAL_COMPLEX(-16, -109), MAKE_VAL_COMPLEX(21, 91), MAKE_VAL_COMPLEX(90, 46), MAKE_VAL_COMPLEX(137, -14), MAKE_VAL_COMPLEX(-25, 82) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_732, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_732", y_true, y_val, y720_LEN*y720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_732() */
/* END TEST 732 */

/*
 *  BEGIN TEST 733
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_733( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	oski_value_t* y_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-23, 11), MAKE_VAL_COMPLEX(203, 57), MAKE_VAL_COMPLEX(-39, -99), MAKE_VAL_COMPLEX(71, -151), MAKE_VAL_COMPLEX(-150, 58), MAKE_VAL_COMPLEX(-62, -2), MAKE_VAL_COMPLEX(-28, 41), MAKE_VAL_COMPLEX(-46, 96), MAKE_VAL_COMPLEX(35, 11), MAKE_VAL_COMPLEX(-20, 81), MAKE_VAL_COMPLEX(-27, -68), MAKE_VAL_COMPLEX(113, 2), MAKE_VAL_COMPLEX(2, -155), MAKE_VAL_COMPLEX(1, 7), MAKE_VAL_COMPLEX(121, 22) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_733, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_733", y_true, y_val, y720_LEN*y720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_733() */
/* END TEST 733 */

/*
 *  BEGIN TEST 734
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_734( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	oski_value_t* y_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-27, 53), MAKE_VAL_COMPLEX(23, 73), MAKE_VAL_COMPLEX(-52, 72), MAKE_VAL_COMPLEX(36, -90), MAKE_VAL_COMPLEX(-12, 30), MAKE_VAL_COMPLEX(-73, -35), MAKE_VAL_COMPLEX(-27, -63), MAKE_VAL_COMPLEX(-100, -40), MAKE_VAL_COMPLEX(50, -38), MAKE_VAL_COMPLEX(61, -51), MAKE_VAL_COMPLEX(8, -78), MAKE_VAL_COMPLEX(41, 57), MAKE_VAL_COMPLEX(53, -31), MAKE_VAL_COMPLEX(57, 17), MAKE_VAL_COMPLEX(24, 56) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_734, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_734", y_true, y_val, y720_LEN*y720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_734() */
/* END TEST 734 */

/*
 *  BEGIN TEST 735
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_735( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	oski_value_t* y_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-349, -52), MAKE_VAL_COMPLEX(-346, -1687), MAKE_VAL_COMPLEX(-1490, -24), MAKE_VAL_COMPLEX(-629, -155), MAKE_VAL_COMPLEX(405, 1181), MAKE_VAL_COMPLEX(366, -89), MAKE_VAL_COMPLEX(964, -23), MAKE_VAL_COMPLEX(1280, -572), MAKE_VAL_COMPLEX(398, 126), MAKE_VAL_COMPLEX(1198, 651), MAKE_VAL_COMPLEX(112, 392), MAKE_VAL_COMPLEX(-610, -659), MAKE_VAL_COMPLEX(-1126, 615), MAKE_VAL_COMPLEX(-97, 492), MAKE_VAL_COMPLEX(-427, -903) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 9);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_735, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_735", y_true, y_val, y720_LEN*y720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_735() */
/* END TEST 735 */

/*
 *  BEGIN TEST 736
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_736( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-51, -18), MAKE_VAL_COMPLEX(-33, -12), MAKE_VAL_COMPLEX(66, 14), MAKE_VAL_COMPLEX(-79, 74), MAKE_VAL_COMPLEX(-116, 42), MAKE_VAL_COMPLEX(18, -127), MAKE_VAL_COMPLEX(160, 58), MAKE_VAL_COMPLEX(-50, -54), MAKE_VAL_COMPLEX(136, 39), MAKE_VAL_COMPLEX(-35, 85), MAKE_VAL_COMPLEX(-47, -45), MAKE_VAL_COMPLEX(127, 22), MAKE_VAL_COMPLEX(28, -113), MAKE_VAL_COMPLEX(-44, 12), MAKE_VAL_COMPLEX(-26, -47) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_736, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_736", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_736() */
/* END TEST 736 */

/*
 *  BEGIN TEST 737
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_737( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-59, 48), MAKE_VAL_COMPLEX(79, 94), MAKE_VAL_COMPLEX(-49, 5), MAKE_VAL_COMPLEX(-39, -160), MAKE_VAL_COMPLEX(238, -28), MAKE_VAL_COMPLEX(-129, 110), MAKE_VAL_COMPLEX(-110, -178), MAKE_VAL_COMPLEX(128, 80), MAKE_VAL_COMPLEX(-143, -108), MAKE_VAL_COMPLEX(1, 7), MAKE_VAL_COMPLEX(11, 173), MAKE_VAL_COMPLEX(-79, -106), MAKE_VAL_COMPLEX(-46, 47), MAKE_VAL_COMPLEX(73, 51), MAKE_VAL_COMPLEX(-26, 73) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_737, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_737", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_737() */
/* END TEST 737 */

/*
 *  BEGIN TEST 738
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_738( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-40, 60), MAKE_VAL_COMPLEX(56, 20), MAKE_VAL_COMPLEX(16, 2), MAKE_VAL_COMPLEX(-92, 8), MAKE_VAL_COMPLEX(64, -44), MAKE_VAL_COMPLEX(-60, 38), MAKE_VAL_COMPLEX(-26, -78), MAKE_VAL_COMPLEX(48, -20), MAKE_VAL_COMPLEX(-32, -30), MAKE_VAL_COMPLEX(22, 58), MAKE_VAL_COMPLEX(36, 76), MAKE_VAL_COMPLEX(-12, -60), MAKE_VAL_COMPLEX(-36, -24), MAKE_VAL_COMPLEX(40, 18), MAKE_VAL_COMPLEX(-14, 34) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_738, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_738", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_738() */
/* END TEST 738 */

/*
 *  BEGIN TEST 739
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_739( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(256, 23), MAKE_VAL_COMPLEX(-658, -459), MAKE_VAL_COMPLEX(509, -489), MAKE_VAL_COMPLEX(708, 1783), MAKE_VAL_COMPLEX(-1500, 1022), MAKE_VAL_COMPLEX(42, -997), MAKE_VAL_COMPLEX(1304, 188), MAKE_VAL_COMPLEX(-1328, -284), MAKE_VAL_COMPLEX(1396, -147), MAKE_VAL_COMPLEX(577, 277), MAKE_VAL_COMPLEX(-407, -935), MAKE_VAL_COMPLEX(810, -125), MAKE_VAL_COMPLEX(-453, -626), MAKE_VAL_COMPLEX(-461, -51), MAKE_VAL_COMPLEX(-161, -368) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_739, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_739", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_739() */
/* END TEST 739 */

/*
 *  BEGIN TEST 740
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_740( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-77, 18), MAKE_VAL_COMPLEX(15, -8), MAKE_VAL_COMPLEX(74, 31), MAKE_VAL_COMPLEX(87, -54), MAKE_VAL_COMPLEX(76, 232), MAKE_VAL_COMPLEX(-172, -120), MAKE_VAL_COMPLEX(167, -181), MAKE_VAL_COMPLEX(-26, 116), MAKE_VAL_COMPLEX(80, -156), MAKE_VAL_COMPLEX(6, 84), MAKE_VAL_COMPLEX(-129, 23), MAKE_VAL_COMPLEX(121, -88), MAKE_VAL_COMPLEX(-88, -65), MAKE_VAL_COMPLEX(-118, 61), MAKE_VAL_COMPLEX(-91, -134) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_740, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_740", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_740() */
/* END TEST 740 */

/*
 *  BEGIN TEST 741
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_741( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-93, -38), MAKE_VAL_COMPLEX(3, 148), MAKE_VAL_COMPLEX(-61, 5), MAKE_VAL_COMPLEX(-203, -132), MAKE_VAL_COMPLEX(86, -144), MAKE_VAL_COMPLEX(27, 34), MAKE_VAL_COMPLEX(-30, 42), MAKE_VAL_COMPLEX(120, -36), MAKE_VAL_COMPLEX(-51, 56), MAKE_VAL_COMPLEX(-105, 25), MAKE_VAL_COMPLEX(7, 135), MAKE_VAL_COMPLEX(-7, -2), MAKE_VAL_COMPLEX(100, -37), MAKE_VAL_COMPLEX(123, 43), MAKE_VAL_COMPLEX(4, 141) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_741, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_741", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_741() */
/* END TEST 741 */

/*
 *  BEGIN TEST 742
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_742( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(70, -10), MAKE_VAL_COMPLEX(-22, -54), MAKE_VAL_COMPLEX(-9, -13), MAKE_VAL_COMPLEX(66, 62), MAKE_VAL_COMPLEX(-74, -18), MAKE_VAL_COMPLEX(67, 19), MAKE_VAL_COMPLEX(-40, 70), MAKE_VAL_COMPLEX(-46, -22), MAKE_VAL_COMPLEX(-1, 43), MAKE_VAL_COMPLEX(28, -54), MAKE_VAL_COMPLEX(32, -76), MAKE_VAL_COMPLEX(-36, 48), MAKE_VAL_COMPLEX(6, 42), MAKE_VAL_COMPLEX(-13, -41), MAKE_VAL_COMPLEX(34, -12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_742, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_742", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_742() */
/* END TEST 742 */

/*
 *  BEGIN TEST 743
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_ROWMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_743( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_ROWMAJ, x720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-16, -69), MAKE_VAL_COMPLEX(-250, 71), MAKE_VAL_COMPLEX(40, -178), MAKE_VAL_COMPLEX(82, -571), MAKE_VAL_COMPLEX(646, 100), MAKE_VAL_COMPLEX(-303, 218), MAKE_VAL_COMPLEX(-224, -354), MAKE_VAL_COMPLEX(270, 286), MAKE_VAL_COMPLEX(-281, -252), MAKE_VAL_COMPLEX(7, -119), MAKE_VAL_COMPLEX(-273, 277), MAKE_VAL_COMPLEX(-42, -225), MAKE_VAL_COMPLEX(-15, 222), MAKE_VAL_COMPLEX(10, 426), MAKE_VAL_COMPLEX(-467, 130) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, -3);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_743, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_743", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_743() */
/* END TEST 743 */

/*
 *  BEGIN TEST 744
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_744( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	oski_value_t* y_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-177, -32), MAKE_VAL_COMPLEX(-79, 13), MAKE_VAL_COMPLEX(-106, -44), MAKE_VAL_COMPLEX(163, -37), MAKE_VAL_COMPLEX(-45, -109), MAKE_VAL_COMPLEX(-31, -64), MAKE_VAL_COMPLEX(142, -170), MAKE_VAL_COMPLEX(-144, -146), MAKE_VAL_COMPLEX(-38, 50), MAKE_VAL_COMPLEX(29, -30), MAKE_VAL_COMPLEX(202, -71), MAKE_VAL_COMPLEX(-14, 112), MAKE_VAL_COMPLEX(189, 176), MAKE_VAL_COMPLEX(1, 164), MAKE_VAL_COMPLEX(-42, -76) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_744, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_744", y_true, y_val, y720_LEN*y720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_744() */
/* END TEST 744 */

/*
 *  BEGIN TEST 745
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_745( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	oski_value_t* y_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(166, 31), MAKE_VAL_COMPLEX(68, -4), MAKE_VAL_COMPLEX(90, 40), MAKE_VAL_COMPLEX(-145, 37), MAKE_VAL_COMPLEX(39, 109), MAKE_VAL_COMPLEX(32, 49), MAKE_VAL_COMPLEX(-133, 161), MAKE_VAL_COMPLEX(144, 126), MAKE_VAL_COMPLEX(48, -44), MAKE_VAL_COMPLEX(-16, 37), MAKE_VAL_COMPLEX(-188, 67), MAKE_VAL_COMPLEX(7, -101), MAKE_VAL_COMPLEX(-180, -169), MAKE_VAL_COMPLEX(0, -153), MAKE_VAL_COMPLEX(34, 84) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_745, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_745", y_true, y_val, y720_LEN*y720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_745() */
/* END TEST 745 */

/*
 *  BEGIN TEST 746
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_746( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	oski_value_t* y_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(19, 35), MAKE_VAL_COMPLEX(49, 15), MAKE_VAL_COMPLEX(20, 56), MAKE_VAL_COMPLEX(-36, -54), MAKE_VAL_COMPLEX(12, 18), MAKE_VAL_COMPLEX(-47, 27), MAKE_VAL_COMPLEX(-45, -9), MAKE_VAL_COMPLEX(-60, 40), MAKE_VAL_COMPLEX(-2, -42), MAKE_VAL_COMPLEX(-5, -53), MAKE_VAL_COMPLEX(-40, -34), MAKE_VAL_COMPLEX(47, -1), MAKE_VAL_COMPLEX(3, -41), MAKE_VAL_COMPLEX(31, -25), MAKE_VAL_COMPLEX(40, 8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_746, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_746", y_true, y_val, y720_LEN*y720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_746() */
/* END TEST 746 */

/*
 *  BEGIN TEST 747
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_747( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	oski_value_t* y_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-145, -608), MAKE_VAL_COMPLEX(359, -105), MAKE_VAL_COMPLEX(106, 88), MAKE_VAL_COMPLEX(315, 3), MAKE_VAL_COMPLEX(651, -309), MAKE_VAL_COMPLEX(-451, -96), MAKE_VAL_COMPLEX(816, 108), MAKE_VAL_COMPLEX(-516, -862), MAKE_VAL_COMPLEX(-346, -690), MAKE_VAL_COMPLEX(317, -694), MAKE_VAL_COMPLEX(532, 415), MAKE_VAL_COMPLEX(-32, 346), MAKE_VAL_COMPLEX(-351, 1076), MAKE_VAL_COMPLEX(-475, 208), MAKE_VAL_COMPLEX(890, -272) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_747, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_747", y_true, y_val, y720_LEN*y720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_747() */
/* END TEST 747 */

/*
 *  BEGIN TEST 748
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_748( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-10, -131), MAKE_VAL_COMPLEX(-124, -85), MAKE_VAL_COMPLEX(-6, -175), MAKE_VAL_COMPLEX(112, -113), MAKE_VAL_COMPLEX(36, -21), MAKE_VAL_COMPLEX(56, -107), MAKE_VAL_COMPLEX(-130, 59), MAKE_VAL_COMPLEX(-24, 88), MAKE_VAL_COMPLEX(125, -11), MAKE_VAL_COMPLEX(-37, -120), MAKE_VAL_COMPLEX(-14, -70), MAKE_VAL_COMPLEX(112, -43), MAKE_VAL_COMPLEX(-106, 4), MAKE_VAL_COMPLEX(20, 43), MAKE_VAL_COMPLEX(111, 46) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_748, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_748", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_748() */
/* END TEST 748 */

/*
 *  BEGIN TEST 749
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_749( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(40, 121), MAKE_VAL_COMPLEX(110, 63), MAKE_VAL_COMPLEX(1, 170), MAKE_VAL_COMPLEX(-78, 135), MAKE_VAL_COMPLEX(-70, 19), MAKE_VAL_COMPLEX(-25, 110), MAKE_VAL_COMPLEX(118, -25), MAKE_VAL_COMPLEX(2, -94), MAKE_VAL_COMPLEX(-122, 30), MAKE_VAL_COMPLEX(45, 94), MAKE_VAL_COMPLEX(22, 34), MAKE_VAL_COMPLEX(-124, 67), MAKE_VAL_COMPLEX(112, 14), MAKE_VAL_COMPLEX(-29, -60), MAKE_VAL_COMPLEX(-97, -54) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_749, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_749", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_749() */
/* END TEST 749 */

/*
 *  BEGIN TEST 750
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_750( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(70, -10), MAKE_VAL_COMPLEX(-22, -54), MAKE_VAL_COMPLEX(-9, -13), MAKE_VAL_COMPLEX(66, 62), MAKE_VAL_COMPLEX(-74, -18), MAKE_VAL_COMPLEX(67, 19), MAKE_VAL_COMPLEX(-40, 70), MAKE_VAL_COMPLEX(-46, -22), MAKE_VAL_COMPLEX(-1, 43), MAKE_VAL_COMPLEX(28, -54), MAKE_VAL_COMPLEX(32, -76), MAKE_VAL_COMPLEX(-36, 48), MAKE_VAL_COMPLEX(6, 42), MAKE_VAL_COMPLEX(-13, -41), MAKE_VAL_COMPLEX(34, -12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_750, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_750", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_750() */
/* END TEST 750 */

/*
 *  BEGIN TEST 751
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_751( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(241, 449), MAKE_VAL_COMPLEX(621, 769), MAKE_VAL_COMPLEX(83, 1637), MAKE_VAL_COMPLEX(-159, 573), MAKE_VAL_COMPLEX(-611, 937), MAKE_VAL_COMPLEX(-91, 379), MAKE_VAL_COMPLEX(1745, -139), MAKE_VAL_COMPLEX(-172, -392), MAKE_VAL_COMPLEX(-684, 308), MAKE_VAL_COMPLEX(-41, 669), MAKE_VAL_COMPLEX(-512, 164), MAKE_VAL_COMPLEX(-545, 943), MAKE_VAL_COMPLEX(1382, -130), MAKE_VAL_COMPLEX(-657, -307), MAKE_VAL_COMPLEX(-1107, -665) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-9, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_751, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_751", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_751() */
/* END TEST 751 */

/*
 *  BEGIN TEST 752
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_752( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(56, -103), MAKE_VAL_COMPLEX(-142, 53), MAKE_VAL_COMPLEX(-185, -8), MAKE_VAL_COMPLEX(42, -29), MAKE_VAL_COMPLEX(-150, 5), MAKE_VAL_COMPLEX(101, -50), MAKE_VAL_COMPLEX(-16, 263), MAKE_VAL_COMPLEX(18, -10), MAKE_VAL_COMPLEX(42, -34), MAKE_VAL_COMPLEX(-31, -42), MAKE_VAL_COMPLEX(46, -86), MAKE_VAL_COMPLEX(-96, -23), MAKE_VAL_COMPLEX(16, 194), MAKE_VAL_COMPLEX(33, -70), MAKE_VAL_COMPLEX(127, 10) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_752, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_752", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_752() */
/* END TEST 752 */

/*
 *  BEGIN TEST 753
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_753( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-26, 23), MAKE_VAL_COMPLEX(72, -61), MAKE_VAL_COMPLEX(166, 10), MAKE_VAL_COMPLEX(62, -5), MAKE_VAL_COMPLEX(88, 63), MAKE_VAL_COMPLEX(-42, -10), MAKE_VAL_COMPLEX(67, -180), MAKE_VAL_COMPLEX(-68, 46), MAKE_VAL_COMPLEX(3, 60), MAKE_VAL_COMPLEX(-10, -19), MAKE_VAL_COMPLEX(-108, 8), MAKE_VAL_COMPLEX(126, 89), MAKE_VAL_COMPLEX(32, -176), MAKE_VAL_COMPLEX(-84, 60), MAKE_VAL_COMPLEX(-120, -53) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_753, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_753", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_753() */
/* END TEST 753 */

/*
 *  BEGIN TEST 754
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_754( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 50), MAKE_VAL_COMPLEX(40, -10), MAKE_VAL_COMPLEX(10, -5), MAKE_VAL_COMPLEX(-50, 40), MAKE_VAL_COMPLEX(20, -50), MAKE_VAL_COMPLEX(-20, 45), MAKE_VAL_COMPLEX(-45, -35), MAKE_VAL_COMPLEX(20, -30), MAKE_VAL_COMPLEX(-30, -5), MAKE_VAL_COMPLEX(35, 25), MAKE_VAL_COMPLEX(50, 30), MAKE_VAL_COMPLEX(-30, -30), MAKE_VAL_COMPLEX(-30, 0), MAKE_VAL_COMPLEX(30, -5), MAKE_VAL_COMPLEX(5, 25) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_754, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_754", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_754() */
/* END TEST 754 */

/*
 *  BEGIN TEST 755
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_755( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_ROWMAJ, y720_VW );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-498, -376), MAKE_VAL_COMPLEX(126, 1324), MAKE_VAL_COMPLEX(-1035, 1365), MAKE_VAL_COMPLEX(-678, 32), MAKE_VAL_COMPLEX(-574, 544), MAKE_VAL_COMPLEX(-79, -491), MAKE_VAL_COMPLEX(1394, 1214), MAKE_VAL_COMPLEX(310, -614), MAKE_VAL_COMPLEX(-433, -615), MAKE_VAL_COMPLEX(-44, 348), MAKE_VAL_COMPLEX(272, -460), MAKE_VAL_COMPLEX(-1168, 210), MAKE_VAL_COMPLEX(1234, 1030), MAKE_VAL_COMPLEX(53, -747), MAKE_VAL_COMPLEX(906, -610) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_755, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_755", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_755() */
/* END TEST 755 */

/*
 *  BEGIN TEST 756
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_756( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	oski_value_t* y_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-2, -5), MAKE_VAL_COMPLEX(97, -108), MAKE_VAL_COMPLEX(179, -60), MAKE_VAL_COMPLEX(-213, -105), MAKE_VAL_COMPLEX(161, 115), MAKE_VAL_COMPLEX(4, 54), MAKE_VAL_COMPLEX(-66, -43), MAKE_VAL_COMPLEX(-164, 174), MAKE_VAL_COMPLEX(9, -94), MAKE_VAL_COMPLEX(-96, 11), MAKE_VAL_COMPLEX(-79, 21), MAKE_VAL_COMPLEX(11, -16), MAKE_VAL_COMPLEX(-184, -63), MAKE_VAL_COMPLEX(49, -61), MAKE_VAL_COMPLEX(94, -196) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_756, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_756", y_true, y_val, y720_LEN*y720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_756() */
/* END TEST 756 */

/*
 *  BEGIN TEST 757
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_757( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	oski_value_t* y_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(124, 66), MAKE_VAL_COMPLEX(70, 54), MAKE_VAL_COMPLEX(-13, 178), MAKE_VAL_COMPLEX(6, 24), MAKE_VAL_COMPLEX(-92, -88), MAKE_VAL_COMPLEX(-83, 114), MAKE_VAL_COMPLEX(-78, 106), MAKE_VAL_COMPLEX(74, 56), MAKE_VAL_COMPLEX(-97, -20), MAKE_VAL_COMPLEX(-22, -150), MAKE_VAL_COMPLEX(-100, -38), MAKE_VAL_COMPLEX(119, -79), MAKE_VAL_COMPLEX(112, -58), MAKE_VAL_COMPLEX(-11, -70), MAKE_VAL_COMPLEX(34, 140) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_757, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_757", y_true, y_val, y720_LEN*y720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_757() */
/* END TEST 757 */

/*
 *  BEGIN TEST 758
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_758( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	oski_value_t* y_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-16, -68), MAKE_VAL_COMPLEX(-76, -48), MAKE_VAL_COMPLEX(-8, -104), MAKE_VAL_COMPLEX(36, 108), MAKE_VAL_COMPLEX(-12, -36), MAKE_VAL_COMPLEX(92, -24), MAKE_VAL_COMPLEX(72, 36), MAKE_VAL_COMPLEX(120, -40), MAKE_VAL_COMPLEX(-16, 72), MAKE_VAL_COMPLEX(-16, 92), MAKE_VAL_COMPLEX(52, 76), MAKE_VAL_COMPLEX(-80, -20), MAKE_VAL_COMPLEX(-24, 68), MAKE_VAL_COMPLEX(-64, 28), MAKE_VAL_COMPLEX(-64, -32) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_758, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_758", y_true, y_val, y720_LEN*y720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_758() */
/* END TEST 758 */

/*
 *  BEGIN TEST 759
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_759( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	oski_value_t* y_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-366, 593), MAKE_VAL_COMPLEX(-333, -268), MAKE_VAL_COMPLEX(-551, -812), MAKE_VAL_COMPLEX(-1047, 381), MAKE_VAL_COMPLEX(1259, -607), MAKE_VAL_COMPLEX(-372, -286), MAKE_VAL_COMPLEX(-830, -321), MAKE_VAL_COMPLEX(-220, 1354), MAKE_VAL_COMPLEX(-149, -770), MAKE_VAL_COMPLEX(504, 241), MAKE_VAL_COMPLEX(43, -185), MAKE_VAL_COMPLEX(353, 520), MAKE_VAL_COMPLEX(-448, 947), MAKE_VAL_COMPLEX(299, -367), MAKE_VAL_COMPLEX(-1030, -700) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_759, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_759", y_true, y_val, y720_LEN*y720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_759() */
/* END TEST 759 */

/*
 *  BEGIN TEST 760
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_760( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1, -28), MAKE_VAL_COMPLEX(-81, 192), MAKE_VAL_COMPLEX(78, 108), MAKE_VAL_COMPLEX(45, -1), MAKE_VAL_COMPLEX(14, -23), MAKE_VAL_COMPLEX(-25, -118), MAKE_VAL_COMPLEX(-124, -6), MAKE_VAL_COMPLEX(-84, -68), MAKE_VAL_COMPLEX(51, -105), MAKE_VAL_COMPLEX(-71, -39), MAKE_VAL_COMPLEX(13, -7), MAKE_VAL_COMPLEX(105, -60), MAKE_VAL_COMPLEX(135, 66), MAKE_VAL_COMPLEX(45, 62), MAKE_VAL_COMPLEX(-2, -57) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_760, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_760", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_760() */
/* END TEST 760 */

/*
 *  BEGIN TEST 761
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_761( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(21, -52), MAKE_VAL_COMPLEX(13, -192), MAKE_VAL_COMPLEX(-96, -104), MAKE_VAL_COMPLEX(51, -43), MAKE_VAL_COMPLEX(-66, 95), MAKE_VAL_COMPLEX(75, 54), MAKE_VAL_COMPLEX(182, 80), MAKE_VAL_COMPLEX(40, 108), MAKE_VAL_COMPLEX(-5, 125), MAKE_VAL_COMPLEX(25, -15), MAKE_VAL_COMPLEX(-81, -61), MAKE_VAL_COMPLEX(-69, 120), MAKE_VAL_COMPLEX(-87, -54), MAKE_VAL_COMPLEX(-95, -66), MAKE_VAL_COMPLEX(4, 15) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_761, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_761", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_761() */
/* END TEST 761 */

/*
 *  BEGIN TEST 762
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_762( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(20, 40), MAKE_VAL_COMPLEX(28, -24), MAKE_VAL_COMPLEX(6, -8), MAKE_VAL_COMPLEX(-24, 52), MAKE_VAL_COMPLEX(-4, -48), MAKE_VAL_COMPLEX(2, 44), MAKE_VAL_COMPLEX(-50, -10), MAKE_VAL_COMPLEX(4, -32), MAKE_VAL_COMPLEX(-26, 8), MAKE_VAL_COMPLEX(38, 6), MAKE_VAL_COMPLEX(52, 4), MAKE_VAL_COMPLEX(-36, -12), MAKE_VAL_COMPLEX(-24, 12), MAKE_VAL_COMPLEX(22, -16), MAKE_VAL_COMPLEX(14, 18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_762, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_762", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_762() */
/* END TEST 762 */

/*
 *  BEGIN TEST 763
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_763( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-187, 89), MAKE_VAL_COMPLEX(-1503, -583), MAKE_VAL_COMPLEX(-992, 675), MAKE_VAL_COMPLEX(-458, 22), MAKE_VAL_COMPLEX(661, 255), MAKE_VAL_COMPLEX(663, -220), MAKE_VAL_COMPLEX(423, -1555), MAKE_VAL_COMPLEX(988, -554), MAKE_VAL_COMPLEX(896, 293), MAKE_VAL_COMPLEX(251, -257), MAKE_VAL_COMPLEX(-144, 606), MAKE_VAL_COMPLEX(657, 699), MAKE_VAL_COMPLEX(-771, 885), MAKE_VAL_COMPLEX(-517, 600), MAKE_VAL_COMPLEX(302, 118) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 9);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_763, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_763", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_763() */
/* END TEST 763 */

/*
 *  BEGIN TEST 764
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_764( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-47, -72), MAKE_VAL_COMPLEX(99, 102), MAKE_VAL_COMPLEX(92, -79), MAKE_VAL_COMPLEX(93, -111), MAKE_VAL_COMPLEX(-64, 103), MAKE_VAL_COMPLEX(-21, -185), MAKE_VAL_COMPLEX(109, 205), MAKE_VAL_COMPLEX(-80, 158), MAKE_VAL_COMPLEX(-15, -94), MAKE_VAL_COMPLEX(-220, 4), MAKE_VAL_COMPLEX(-61, 19), MAKE_VAL_COMPLEX(-1, -66), MAKE_VAL_COMPLEX(125, -138), MAKE_VAL_COMPLEX(19, 11), MAKE_VAL_COMPLEX(-85, -182) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_764, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_764", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_764() */
/* END TEST 764 */

/*
 *  BEGIN TEST 765
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_765( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-13, 22), MAKE_VAL_COMPLEX(-127, -44), MAKE_VAL_COMPLEX(-96, 96), MAKE_VAL_COMPLEX(-91, 11), MAKE_VAL_COMPLEX(104, -29), MAKE_VAL_COMPLEX(-13, 116), MAKE_VAL_COMPLEX(-22, -224), MAKE_VAL_COMPLEX(96, -104), MAKE_VAL_COMPLEX(51, 63), MAKE_VAL_COMPLEX(155, 13), MAKE_VAL_COMPLEX(-25, 11), MAKE_VAL_COMPLEX(67, 60), MAKE_VAL_COMPLEX(-95, 102), MAKE_VAL_COMPLEX(-43, 30), MAKE_VAL_COMPLEX(50, 163) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_765, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_765", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_765() */
/* END TEST 765 */

/*
 *  BEGIN TEST 766
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_766( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-90, 0), MAKE_VAL_COMPLEX(18, 72), MAKE_VAL_COMPLEX(9, 18), MAKE_VAL_COMPLEX(-72, -90), MAKE_VAL_COMPLEX(90, 36), MAKE_VAL_COMPLEX(-81, -36), MAKE_VAL_COMPLEX(63, -81), MAKE_VAL_COMPLEX(54, 36), MAKE_VAL_COMPLEX(9, -54), MAKE_VAL_COMPLEX(-45, 63), MAKE_VAL_COMPLEX(-54, 90), MAKE_VAL_COMPLEX(54, -54), MAKE_VAL_COMPLEX(0, -54), MAKE_VAL_COMPLEX(9, 54), MAKE_VAL_COMPLEX(-45, 9) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_766, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_766", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_766() */
/* END TEST 766 */

/*
 *  BEGIN TEST 767
 *  Value type: complex
 *  A : <MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x720', LAYOUT_COLMAJ>
 *  y : <'y720', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_767( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A720_ptr, A720_ind, A720_val,
		A720_M, A720_N, SHARE_INPUTMAT,
		5, MAT_SYMM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y720_val,
		y720_LEN * y720_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y720_LEN, y720_VW, LAYOUT_COLMAJ, y720_LEN );
	oski_value_t* y_val = array_Duplicate( x720_val,
		x720_LEN * x720_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x720_LEN, x720_VW, LAYOUT_COLMAJ, x720_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(50, 9), MAKE_VAL_COMPLEX(636, 11), MAKE_VAL_COMPLEX(308, -495), MAKE_VAL_COMPLEX(259, -27), MAKE_VAL_COMPLEX(-363, 102), MAKE_VAL_COMPLEX(-90, -370), MAKE_VAL_COMPLEX(199, 815), MAKE_VAL_COMPLEX(-248, 438), MAKE_VAL_COMPLEX(-335, -200), MAKE_VAL_COMPLEX(-546, 144), MAKE_VAL_COMPLEX(211, -23), MAKE_VAL_COMPLEX(-406, -227), MAKE_VAL_COMPLEX(212, -491), MAKE_VAL_COMPLEX(206, -186), MAKE_VAL_COMPLEX(-342, -549) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_767, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_767", y_true, y_val, x720_LEN*x720_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_767() */
/* END TEST 767 */

#define A768_M 5    /* # of rows */
#define A768_N 5    /* # of cols */
#define A768_NNZ 6  /* # of stored non-zeros */
static oski_index_t A768_ptr[A768_M+1] = {1, 3, 4, 5, 7, 7}; /* end row pointers for A768 */
static oski_index_t A768_ind[A768_NNZ] = {3, 4, 3, 4, 4, 5}; /* end column indices pointers for A768 */
static oski_value_t A768_val[A768_NNZ] = {MAKE_VAL_COMPLEX(5, 10), MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(10, 2), MAKE_VAL_COMPLEX(-5, -10), MAKE_VAL_COMPLEX(4, -9), MAKE_VAL_COMPLEX(-1, -1)}; /* end stored non-zero values for A768 */

#define x768_LEN 5   /* vector length */
#define x768_VW  3   /* vector width */
static oski_value_t x768_val[x768_LEN*x768_VW] = {MAKE_VAL_COMPLEX(1, -4), MAKE_VAL_COMPLEX(10, 10), MAKE_VAL_COMPLEX(2, -7), MAKE_VAL_COMPLEX(5, 3), MAKE_VAL_COMPLEX(-4, 1), MAKE_VAL_COMPLEX(1, -3), MAKE_VAL_COMPLEX(0, -10), MAKE_VAL_COMPLEX(-7, 3), MAKE_VAL_COMPLEX(4, -9), MAKE_VAL_COMPLEX(-8, -4), MAKE_VAL_COMPLEX(-5, 6), MAKE_VAL_COMPLEX(6, 9), MAKE_VAL_COMPLEX(3, 6), MAKE_VAL_COMPLEX(6, 10), MAKE_VAL_COMPLEX(-10, -4)}; /* data */

#define y768_LEN 5   /* vector length */
#define y768_VW  3   /* vector width */
static oski_value_t y768_val[y768_LEN*y768_VW] = {MAKE_VAL_COMPLEX(4, 4), MAKE_VAL_COMPLEX(-3, -2), MAKE_VAL_COMPLEX(-6, 7), MAKE_VAL_COMPLEX(9, -10), MAKE_VAL_COMPLEX(6, 9), MAKE_VAL_COMPLEX(2, 2), MAKE_VAL_COMPLEX(-5, -7), MAKE_VAL_COMPLEX(-0, 0), MAKE_VAL_COMPLEX(9, -2), MAKE_VAL_COMPLEX(-10, 6), MAKE_VAL_COMPLEX(10, 5), MAKE_VAL_COMPLEX(-5, 8), MAKE_VAL_COMPLEX(-7, -4), MAKE_VAL_COMPLEX(3, 10), MAKE_VAL_COMPLEX(-2, -4)}; /* data */

/*
 *  BEGIN TEST 768
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_768( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	oski_value_t* y_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(160, 18), MAKE_VAL_COMPLEX(39, -30), MAKE_VAL_COMPLEX(227, -129), MAKE_VAL_COMPLEX(-106, -141), MAKE_VAL_COMPLEX(-19, -74), MAKE_VAL_COMPLEX(68, -108), MAKE_VAL_COMPLEX(46, 203), MAKE_VAL_COMPLEX(-7, 172), MAKE_VAL_COMPLEX(102, -221), MAKE_VAL_COMPLEX(-121, 193), MAKE_VAL_COMPLEX(108, -202), MAKE_VAL_COMPLEX(3, 64), MAKE_VAL_COMPLEX(-4, 91), MAKE_VAL_COMPLEX(89, -68), MAKE_VAL_COMPLEX(-25, 19) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_768, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_768", y_true, y_val, y768_LEN*y768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_768() */
/* END TEST 768 */

/*
 *  BEGIN TEST 769
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_769( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	oski_value_t* y_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-196, -70), MAKE_VAL_COMPLEX(-10, 58), MAKE_VAL_COMPLEX(-147, 64), MAKE_VAL_COMPLEX(-13, 233), MAKE_VAL_COMPLEX(-29, -37), MAKE_VAL_COMPLEX(-86, 82), MAKE_VAL_COMPLEX(-5, -116), MAKE_VAL_COMPLEX(7, -172), MAKE_VAL_COMPLEX(-205, 225), MAKE_VAL_COMPLEX(243, -239), MAKE_VAL_COMPLEX(-208, 127), MAKE_VAL_COMPLEX(68, -142), MAKE_VAL_COMPLEX(73, -33), MAKE_VAL_COMPLEX(-102, -48), MAKE_VAL_COMPLEX(39, 29) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_769, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_769", y_true, y_val, y768_LEN*y768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_769() */
/* END TEST 769 */

/*
 *  BEGIN TEST 770
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_770( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	oski_value_t* y_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(16, 40), MAKE_VAL_COMPLEX(-15, -23), MAKE_VAL_COMPLEX(-63, 31), MAKE_VAL_COMPLEX(93, -43), MAKE_VAL_COMPLEX(15, 81), MAKE_VAL_COMPLEX(8, 20), MAKE_VAL_COMPLEX(-14, -64), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(69, 13), MAKE_VAL_COMPLEX(-88, 12), MAKE_VAL_COMPLEX(55, 65), MAKE_VAL_COMPLEX(-59, 41), MAKE_VAL_COMPLEX(-37, -49), MAKE_VAL_COMPLEX(-9, 79), MAKE_VAL_COMPLEX(-2, -34) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_770, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_770", y_true, y_val, y768_LEN*y768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_770() */
/* END TEST 770 */

/*
 *  BEGIN TEST 771
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_771( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	oski_value_t* y_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-332, 1272), MAKE_VAL_COMPLEX(641, 216), MAKE_VAL_COMPLEX(1494, 727), MAKE_VAL_COMPLEX(875, 82), MAKE_VAL_COMPLEX(-382, -573), MAKE_VAL_COMPLEX(746, 306), MAKE_VAL_COMPLEX(-781, 967), MAKE_VAL_COMPLEX(-1390, 288), MAKE_VAL_COMPLEX(1467, 1034), MAKE_VAL_COMPLEX(-1444, -1576), MAKE_VAL_COMPLEX(852, 695), MAKE_VAL_COMPLEX(-625, -690), MAKE_VAL_COMPLEX(-21, 20), MAKE_VAL_COMPLEX(-65, 60), MAKE_VAL_COMPLEX(168, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_771, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_771", y_true, y_val, y768_LEN*y768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_771() */
/* END TEST 771 */

/*
 *  BEGIN TEST 772
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_772( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(199, -72), MAKE_VAL_COMPLEX(40, -120), MAKE_VAL_COMPLEX(186, 2), MAKE_VAL_COMPLEX(3, -77), MAKE_VAL_COMPLEX(-21, 18), MAKE_VAL_COMPLEX(91, -23), MAKE_VAL_COMPLEX(170, -12), MAKE_VAL_COMPLEX(14, -24), MAKE_VAL_COMPLEX(18, -57), MAKE_VAL_COMPLEX(-88, 130), MAKE_VAL_COMPLEX(90, 18), MAKE_VAL_COMPLEX(178, 29), MAKE_VAL_COMPLEX(52, 31), MAKE_VAL_COMPLEX(61, 27), MAKE_VAL_COMPLEX(-59, 3) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_772, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_772", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_772() */
/* END TEST 772 */

/*
 *  BEGIN TEST 773
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_773( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-222, 45), MAKE_VAL_COMPLEX(80, 100), MAKE_VAL_COMPLEX(-225, -51), MAKE_VAL_COMPLEX(43, 57), MAKE_VAL_COMPLEX(8, 15), MAKE_VAL_COMPLEX(-107, 1), MAKE_VAL_COMPLEX(-240, -38), MAKE_VAL_COMPLEX(-28, 88), MAKE_VAL_COMPLEX(-61, -16), MAKE_VAL_COMPLEX(20, -94), MAKE_VAL_COMPLEX(-73, 47), MAKE_VAL_COMPLEX(-85, -26), MAKE_VAL_COMPLEX(5, -22), MAKE_VAL_COMPLEX(39, -19), MAKE_VAL_COMPLEX(-19, 47) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_773, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_773", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_773() */
/* END TEST 773 */

/*
 *  BEGIN TEST 774
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_774( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(19, -8), MAKE_VAL_COMPLEX(-10, 70), MAKE_VAL_COMPLEX(34, -13), MAKE_VAL_COMPLEX(3, 29), MAKE_VAL_COMPLEX(-16, -13), MAKE_VAL_COMPLEX(15, -5), MAKE_VAL_COMPLEX(40, -30), MAKE_VAL_COMPLEX(-33, -19), MAKE_VAL_COMPLEX(48, -11), MAKE_VAL_COMPLEX(-8, -44), MAKE_VAL_COMPLEX(-39, -2), MAKE_VAL_COMPLEX(-18, 51), MAKE_VAL_COMPLEX(-15, 30), MAKE_VAL_COMPLEX(-22, 54), MAKE_VAL_COMPLEX(-14, -52) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_774, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_774", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_774() */
/* END TEST 774 */

/*
 *  BEGIN TEST 775
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_775( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-515, 573), MAKE_VAL_COMPLEX(350, 330), MAKE_VAL_COMPLEX(-669, 290), MAKE_VAL_COMPLEX(226, 170), MAKE_VAL_COMPLEX(20, -22), MAKE_VAL_COMPLEX(-272, 216), MAKE_VAL_COMPLEX(-656, 316), MAKE_VAL_COMPLEX(15, 237), MAKE_VAL_COMPLEX(-69, 85), MAKE_VAL_COMPLEX(-112, -454), MAKE_VAL_COMPLEX(-234, 247), MAKE_VAL_COMPLEX(-403, 221), MAKE_VAL_COMPLEX(-98, -4), MAKE_VAL_COMPLEX(-31, -1), MAKE_VAL_COMPLEX(43, 19) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-3, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_775, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_775", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_775() */
/* END TEST 775 */

/*
 *  BEGIN TEST 776
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_776( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-18, -157), MAKE_VAL_COMPLEX(-180, 180), MAKE_VAL_COMPLEX(61, -249), MAKE_VAL_COMPLEX(-63, -5), MAKE_VAL_COMPLEX(-28, -27), MAKE_VAL_COMPLEX(115, -45), MAKE_VAL_COMPLEX(200, -318), MAKE_VAL_COMPLEX(-116, 132), MAKE_VAL_COMPLEX(101, 8), MAKE_VAL_COMPLEX(-76, -168), MAKE_VAL_COMPLEX(-41, 83), MAKE_VAL_COMPLEX(-163, 52), MAKE_VAL_COMPLEX(-29, 38), MAKE_VAL_COMPLEX(-65, 103), MAKE_VAL_COMPLEX(-21, -113) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_776, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_776", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_776() */
/* END TEST 776 */

/*
 *  BEGIN TEST 777
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_777( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(37, 149), MAKE_VAL_COMPLEX(170, -110), MAKE_VAL_COMPLEX(-27, 236), MAKE_VAL_COMPLEX(66, 34), MAKE_VAL_COMPLEX(12, 14), MAKE_VAL_COMPLEX(-100, 40), MAKE_VAL_COMPLEX(-160, 288), MAKE_VAL_COMPLEX(83, -151), MAKE_VAL_COMPLEX(-53, -19), MAKE_VAL_COMPLEX(68, 124), MAKE_VAL_COMPLEX(2, -85), MAKE_VAL_COMPLEX(145, -1), MAKE_VAL_COMPLEX(14, -8), MAKE_VAL_COMPLEX(43, -49), MAKE_VAL_COMPLEX(7, 61) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_777, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_777", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_777() */
/* END TEST 777 */

/*
 *  BEGIN TEST 778
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_778( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(27, -6), MAKE_VAL_COMPLEX(-30, 90), MAKE_VAL_COMPLEX(48, -9), MAKE_VAL_COMPLEX(-3, 39), MAKE_VAL_COMPLEX(-18, -21), MAKE_VAL_COMPLEX(21, -3), MAKE_VAL_COMPLEX(60, -30), MAKE_VAL_COMPLEX(-39, -33), MAKE_VAL_COMPLEX(66, -3), MAKE_VAL_COMPLEX(0, -60), MAKE_VAL_COMPLEX(-51, -12), MAKE_VAL_COMPLEX(-36, 63), MAKE_VAL_COMPLEX(-27, 36), MAKE_VAL_COMPLEX(-42, 66), MAKE_VAL_COMPLEX(-6, -72) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_778, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_778", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_778() */
/* END TEST 778 */

/*
 *  BEGIN TEST 779
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_779( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(884, -352), MAKE_VAL_COMPLEX(-260, -820), MAKE_VAL_COMPLEX(1406, -68), MAKE_VAL_COMPLEX(384, -356), MAKE_VAL_COMPLEX(-26, -2), MAKE_VAL_COMPLEX(240, 500), MAKE_VAL_COMPLEX(1628, 660), MAKE_VAL_COMPLEX(-1086, -338), MAKE_VAL_COMPLEX(-84, 8), MAKE_VAL_COMPLEX(464, -448), MAKE_VAL_COMPLEX(-600, 218), MAKE_VAL_COMPLEX(264, -660), MAKE_VAL_COMPLEX(102, 66), MAKE_VAL_COMPLEX(-14, -18), MAKE_VAL_COMPLEX(26, -62) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_779, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_779", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_779() */
/* END TEST 779 */

/*
 *  BEGIN TEST 780
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_780( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	oski_value_t* y_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(148, 70), MAKE_VAL_COMPLEX(15, -99), MAKE_VAL_COMPLEX(90, 143), MAKE_VAL_COMPLEX(-216, 108), MAKE_VAL_COMPLEX(19, 15), MAKE_VAL_COMPLEX(49, -65), MAKE_VAL_COMPLEX(-88, 14), MAKE_VAL_COMPLEX(-7, 172), MAKE_VAL_COMPLEX(110, -103), MAKE_VAL_COMPLEX(7, 15), MAKE_VAL_COMPLEX(155, -160), MAKE_VAL_COMPLEX(61, -69), MAKE_VAL_COMPLEX(145, -145), MAKE_VAL_COMPLEX(-76, 58), MAKE_VAL_COMPLEX(-3, -17) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_780, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_780", y_true, y_val, y768_LEN*y768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_780() */
/* END TEST 780 */

/*
 *  BEGIN TEST 781
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_781( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	oski_value_t* y_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-160, -74), MAKE_VAL_COMPLEX(-7, 100), MAKE_VAL_COMPLEX(-85, -163), MAKE_VAL_COMPLEX(208, -79), MAKE_VAL_COMPLEX(-40, -27), MAKE_VAL_COMPLEX(-55, 63), MAKE_VAL_COMPLEX(105, -5), MAKE_VAL_COMPLEX(7, -172), MAKE_VAL_COMPLEX(-126, 116), MAKE_VAL_COMPLEX(7, -37), MAKE_VAL_COMPLEX(-180, 160), MAKE_VAL_COMPLEX(-59, 48), MAKE_VAL_COMPLEX(-127, 146), MAKE_VAL_COMPLEX(60, -75), MAKE_VAL_COMPLEX(11, 23) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_781, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_781", y_true, y_val, y768_LEN*y768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_781() */
/* END TEST 781 */

/*
 *  BEGIN TEST 782
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_782( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	oski_value_t* y_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(40, -16), MAKE_VAL_COMPLEX(-23, 15), MAKE_VAL_COMPLEX(31, 63), MAKE_VAL_COMPLEX(-43, -93), MAKE_VAL_COMPLEX(81, -15), MAKE_VAL_COMPLEX(20, -8), MAKE_VAL_COMPLEX(-64, 14), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(13, -69), MAKE_VAL_COMPLEX(12, 88), MAKE_VAL_COMPLEX(65, -55), MAKE_VAL_COMPLEX(41, 59), MAKE_VAL_COMPLEX(-49, 37), MAKE_VAL_COMPLEX(79, 9), MAKE_VAL_COMPLEX(-34, 2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_782, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_782", y_true, y_val, y768_LEN*y768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_782() */
/* END TEST 782 */

/*
 *  BEGIN TEST 783
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_783( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	oski_value_t* y_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(766, 978), MAKE_VAL_COMPLEX(434, -683), MAKE_VAL_COMPLEX(122, 1179), MAKE_VAL_COMPLEX(-1724, 245), MAKE_VAL_COMPLEX(-41, 222), MAKE_VAL_COMPLEX(508, -286), MAKE_VAL_COMPLEX(-557, -217), MAKE_VAL_COMPLEX(-565, 1183), MAKE_VAL_COMPLEX(1087, -270), MAKE_VAL_COMPLEX(-54, -16), MAKE_VAL_COMPLEX(1480, -535), MAKE_VAL_COMPLEX(540, -381), MAKE_VAL_COMPLEX(1516, -663), MAKE_VAL_COMPLEX(-842, 197), MAKE_VAL_COMPLEX(86, -146) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_783, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_783", y_true, y_val, y768_LEN*y768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_783() */
/* END TEST 783 */

/*
 *  BEGIN TEST 784
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_784( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-247, -34), MAKE_VAL_COMPLEX(-130, 90), MAKE_VAL_COMPLEX(152, -251), MAKE_VAL_COMPLEX(-66, 124), MAKE_VAL_COMPLEX(-37, -29), MAKE_VAL_COMPLEX(-28, -66), MAKE_VAL_COMPLEX(-101, -76), MAKE_VAL_COMPLEX(-187, -172), MAKE_VAL_COMPLEX(128, 88), MAKE_VAL_COMPLEX(29, -291), MAKE_VAL_COMPLEX(-55, -128), MAKE_VAL_COMPLEX(6, 76), MAKE_VAL_COMPLEX(-52, 70), MAKE_VAL_COMPLEX(-59, 117), MAKE_VAL_COMPLEX(-27, -131) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_784, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_784", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_784() */
/* END TEST 784 */

/*
 *  BEGIN TEST 785
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_785( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(264, 17), MAKE_VAL_COMPLEX(150, -10), MAKE_VAL_COMPLEX(-121, 222), MAKE_VAL_COMPLEX(82, -94), MAKE_VAL_COMPLEX(14, 22), MAKE_VAL_COMPLEX(42, 54), MAKE_VAL_COMPLEX(131, 26), MAKE_VAL_COMPLEX(143, 166), MAKE_VAL_COMPLEX(-81, -121), MAKE_VAL_COMPLEX(-57, 247), MAKE_VAL_COMPLEX(12, 143), MAKE_VAL_COMPLEX(-3, -13), MAKE_VAL_COMPLEX(49, -31), MAKE_VAL_COMPLEX(59, -49), MAKE_VAL_COMPLEX(-11, 81) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_785, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_785", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_785() */
/* END TEST 785 */

/*
 *  BEGIN TEST 786
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_786( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-33, -21), MAKE_VAL_COMPLEX(120, -60), MAKE_VAL_COMPLEX(-57, -39), MAKE_VAL_COMPLEX(42, -36), MAKE_VAL_COMPLEX(-3, 39), MAKE_VAL_COMPLEX(-24, -18), MAKE_VAL_COMPLEX(-90, -30), MAKE_VAL_COMPLEX(6, 72), MAKE_VAL_COMPLEX(-69, -63), MAKE_VAL_COMPLEX(-60, 60), MAKE_VAL_COMPLEX(39, 63), MAKE_VAL_COMPLEX(99, -27), MAKE_VAL_COMPLEX(63, -9), MAKE_VAL_COMPLEX(108, -24), MAKE_VAL_COMPLEX(-66, 78) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_786, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_786", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_786() */
/* END TEST 786 */

/*
 *  BEGIN TEST 787
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_787( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1007, -2241), MAKE_VAL_COMPLEX(760, -680), MAKE_VAL_COMPLEX(1563, 1321), MAKE_VAL_COMPLEX(-295, -757), MAKE_VAL_COMPLEX(22, -31), MAKE_VAL_COMPLEX(629, -307), MAKE_VAL_COMPLEX(681, -1440), MAKE_VAL_COMPLEX(1404, -561), MAKE_VAL_COMPLEX(-951, -106), MAKE_VAL_COMPLEX(1461, 945), MAKE_VAL_COMPLEX(897, 575), MAKE_VAL_COMPLEX(97, 456), MAKE_VAL_COMPLEX(-8, -179), MAKE_VAL_COMPLEX(-13, -101), MAKE_VAL_COMPLEX(177, 63) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-3, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_787, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_787", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_787() */
/* END TEST 787 */

/*
 *  BEGIN TEST 788
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_788( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(77, 279), MAKE_VAL_COMPLEX(-100, 200), MAKE_VAL_COMPLEX(-157, -29), MAKE_VAL_COMPLEX(-49, 121), MAKE_VAL_COMPLEX(-18, -21), MAKE_VAL_COMPLEX(-57, -29), MAKE_VAL_COMPLEX(141, 76), MAKE_VAL_COMPLEX(-98, -3), MAKE_VAL_COMPLEX(17, -8), MAKE_VAL_COMPLEX(-9, -117), MAKE_VAL_COMPLEX(-33, 45), MAKE_VAL_COMPLEX(-183, 132), MAKE_VAL_COMPLEX(-26, 55), MAKE_VAL_COMPLEX(-49, 77), MAKE_VAL_COMPLEX(-19, -79) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_788, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_788", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_788() */
/* END TEST 788 */

/*
 *  BEGIN TEST 789
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_789( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-36, -290), MAKE_VAL_COMPLEX(60, -60), MAKE_VAL_COMPLEX(230, 12), MAKE_VAL_COMPLEX(47, -61), MAKE_VAL_COMPLEX(-11, -10), MAKE_VAL_COMPLEX(89, 23), MAKE_VAL_COMPLEX(-51, -126), MAKE_VAL_COMPLEX(36, -45), MAKE_VAL_COMPLEX(84, -1), MAKE_VAL_COMPLEX(5, 25), MAKE_VAL_COMPLEX(-46, -60), MAKE_VAL_COMPLEX(132, -33), MAKE_VAL_COMPLEX(-13, 2), MAKE_VAL_COMPLEX(-11, 27), MAKE_VAL_COMPLEX(5, -31) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_789, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_789", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_789() */
/* END TEST 789 */

/*
 *  BEGIN TEST 790
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_790( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(38, -16), MAKE_VAL_COMPLEX(-20, 140), MAKE_VAL_COMPLEX(68, -26), MAKE_VAL_COMPLEX(6, 58), MAKE_VAL_COMPLEX(-32, -26), MAKE_VAL_COMPLEX(30, -10), MAKE_VAL_COMPLEX(80, -60), MAKE_VAL_COMPLEX(-66, -38), MAKE_VAL_COMPLEX(96, -22), MAKE_VAL_COMPLEX(-16, -88), MAKE_VAL_COMPLEX(-78, -4), MAKE_VAL_COMPLEX(-36, 102), MAKE_VAL_COMPLEX(-30, 60), MAKE_VAL_COMPLEX(-44, 108), MAKE_VAL_COMPLEX(-28, -104) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_790, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_790", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_790() */
/* END TEST 790 */

/*
 *  BEGIN TEST 791
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_ROWMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_791( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_ROWMAJ, x768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1535, 2909), MAKE_VAL_COMPLEX(-1400, 540), MAKE_VAL_COMPLEX(-1687, -1625), MAKE_VAL_COMPLEX(-984, 432), MAKE_VAL_COMPLEX(-10, -6), MAKE_VAL_COMPLEX(-512, -784), MAKE_VAL_COMPLEX(7, 1501), MAKE_VAL_COMPLEX(-761, -151), MAKE_VAL_COMPLEX(-380, -398), MAKE_VAL_COMPLEX(310, -600), MAKE_VAL_COMPLEX(-259, 641), MAKE_VAL_COMPLEX(-1812, -378), MAKE_VAL_COMPLEX(-130, 196), MAKE_VAL_COMPLEX(-148, 82), MAKE_VAL_COMPLEX(-80, -182) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(9, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_791, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_791", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_791() */
/* END TEST 791 */

/*
 *  BEGIN TEST 792
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_792( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	oski_value_t* y_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(88, -107), MAKE_VAL_COMPLEX(-211, 6), MAKE_VAL_COMPLEX(18, -56), MAKE_VAL_COMPLEX(2, -131), MAKE_VAL_COMPLEX(-19, 4), MAKE_VAL_COMPLEX(32, 60), MAKE_VAL_COMPLEX(85, 56), MAKE_VAL_COMPLEX(-55, -100), MAKE_VAL_COMPLEX(35, -77), MAKE_VAL_COMPLEX(-68, 77), MAKE_VAL_COMPLEX(9, -25), MAKE_VAL_COMPLEX(305, -29), MAKE_VAL_COMPLEX(-36, 19), MAKE_VAL_COMPLEX(43, 10), MAKE_VAL_COMPLEX(-20, -14) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_792, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_792", y_true, y_val, y768_LEN*y768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_792() */
/* END TEST 792 */

/*
 *  BEGIN TEST 793
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_793( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	oski_value_t* y_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-56, 83), MAKE_VAL_COMPLEX(194, 13), MAKE_VAL_COMPLEX(25, 105), MAKE_VAL_COMPLEX(-63, 58), MAKE_VAL_COMPLEX(88, -37), MAKE_VAL_COMPLEX(-16, -72), MAKE_VAL_COMPLEX(-139, -28), MAKE_VAL_COMPLEX(55, 100), MAKE_VAL_COMPLEX(-40, 12), MAKE_VAL_COMPLEX(100, -1), MAKE_VAL_COMPLEX(36, -40), MAKE_VAL_COMPLEX(-254, 72), MAKE_VAL_COMPLEX(1, 26), MAKE_VAL_COMPLEX(30, -21), MAKE_VAL_COMPLEX(-10, 24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_793, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_793", y_true, y_val, y768_LEN*y768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_793() */
/* END TEST 793 */

/*
 *  BEGIN TEST 794
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_794( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	oski_value_t* y_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(40, -8), MAKE_VAL_COMPLEX(-24, 10), MAKE_VAL_COMPLEX(18, 64), MAKE_VAL_COMPLEX(-24, -94), MAKE_VAL_COMPLEX(78, 0), MAKE_VAL_COMPLEX(20, -4), MAKE_VAL_COMPLEX(-62, 2), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(24, -62), MAKE_VAL_COMPLEX(-4, 84), MAKE_VAL_COMPLEX(70, -40), MAKE_VAL_COMPLEX(28, 62), MAKE_VAL_COMPLEX(-52, 26), MAKE_VAL_COMPLEX(72, 22), MAKE_VAL_COMPLEX(-32, -4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_794, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_794", y_true, y_val, y768_LEN*y768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_794() */
/* END TEST 794 */

/*
 *  BEGIN TEST 795
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_795( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	oski_value_t* y_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(258, -904), MAKE_VAL_COMPLEX(-1547, 361), MAKE_VAL_COMPLEX(-205, -845), MAKE_VAL_COMPLEX(87, -517), MAKE_VAL_COMPLEX(-627, 223), MAKE_VAL_COMPLEX(260, 480), MAKE_VAL_COMPLEX(1172, 144), MAKE_VAL_COMPLEX(-640, -690), MAKE_VAL_COMPLEX(99, -255), MAKE_VAL_COMPLEX(-530, 224), MAKE_VAL_COMPLEX(-333, 167), MAKE_VAL_COMPLEX(2087, -1171), MAKE_VAL_COMPLEX(21, -39), MAKE_VAL_COMPLEX(-125, -1), MAKE_VAL_COMPLEX(18, -110) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_795, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_795", y_true, y_val, y768_LEN*y768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_795() */
/* END TEST 795 */

/*
 *  BEGIN TEST 796
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_796( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(183, -42), MAKE_VAL_COMPLEX(-6, -150), MAKE_VAL_COMPLEX(161, 52), MAKE_VAL_COMPLEX(-23, 99), MAKE_VAL_COMPLEX(29, 22), MAKE_VAL_COMPLEX(-67, -149), MAKE_VAL_COMPLEX(-50, 20), MAKE_VAL_COMPLEX(76, -34), MAKE_VAL_COMPLEX(49, -35), MAKE_VAL_COMPLEX(-9, 33), MAKE_VAL_COMPLEX(235, 63), MAKE_VAL_COMPLEX(127, -50), MAKE_VAL_COMPLEX(45, -18), MAKE_VAL_COMPLEX(153, -57), MAKE_VAL_COMPLEX(13, 55) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_796, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_796", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_796() */
/* END TEST 796 */

/*
 *  BEGIN TEST 797
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_797( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-217, 8), MAKE_VAL_COMPLEX(166, 110), MAKE_VAL_COMPLEX(-219, -114), MAKE_VAL_COMPLEX(83, -131), MAKE_VAL_COMPLEX(-43, 24), MAKE_VAL_COMPLEX(43, 121), MAKE_VAL_COMPLEX(-50, -80), MAKE_VAL_COMPLEX(-88, 122), MAKE_VAL_COMPLEX(-115, -59), MAKE_VAL_COMPLEX(-79, 23), MAKE_VAL_COMPLEX(-205, 23), MAKE_VAL_COMPLEX(-1, 44), MAKE_VAL_COMPLEX(33, 24), MAKE_VAL_COMPLEX(-17, 57), MAKE_VAL_COMPLEX(-113, 21) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_797, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_797", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_797() */
/* END TEST 797 */

/*
 *  BEGIN TEST 798
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_798( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(41, -28), MAKE_VAL_COMPLEX(10, 170), MAKE_VAL_COMPLEX(74, -47), MAKE_VAL_COMPLEX(21, 67), MAKE_VAL_COMPLEX(-44, -23), MAKE_VAL_COMPLEX(33, -19), MAKE_VAL_COMPLEX(80, -90), MAKE_VAL_COMPLEX(-87, -29), MAKE_VAL_COMPLEX(108, -49), MAKE_VAL_COMPLEX(-40, -100), MAKE_VAL_COMPLEX(-93, 14), MAKE_VAL_COMPLEX(-18, 129), MAKE_VAL_COMPLEX(-21, 78), MAKE_VAL_COMPLEX(-26, 138), MAKE_VAL_COMPLEX(-58, -116) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_798, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_798", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_798() */
/* END TEST 798 */

/*
 *  BEGIN TEST 799
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_799( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1011, 452), MAKE_VAL_COMPLEX(250, 234), MAKE_VAL_COMPLEX(-1104, -3), MAKE_VAL_COMPLEX(-7, -729), MAKE_VAL_COMPLEX(-50, 13), MAKE_VAL_COMPLEX(375, 715), MAKE_VAL_COMPLEX(-60, 70), MAKE_VAL_COMPLEX(-105, 383), MAKE_VAL_COMPLEX(-559, 306), MAKE_VAL_COMPLEX(72, 146), MAKE_VAL_COMPLEX(-954, -67), MAKE_VAL_COMPLEX(-456, 5), MAKE_VAL_COMPLEX(-99, -84), MAKE_VAL_COMPLEX(-550, 44), MAKE_VAL_COMPLEX(-16, 116) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_799, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_799", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_799() */
/* END TEST 799 */

/*
 *  BEGIN TEST 800
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_800( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-59, -129), MAKE_VAL_COMPLEX(-104, -100), MAKE_VAL_COMPLEX(112, -240), MAKE_VAL_COMPLEX(-88, -96), MAKE_VAL_COMPLEX(20, -20), MAKE_VAL_COMPLEX(-154, 62), MAKE_VAL_COMPLEX(0, 40), MAKE_VAL_COMPLEX(-29, 161), MAKE_VAL_COMPLEX(16, 129), MAKE_VAL_COMPLEX(17, 21), MAKE_VAL_COMPLEX(15, -254), MAKE_VAL_COMPLEX(62, -74), MAKE_VAL_COMPLEX(-3, -3), MAKE_VAL_COMPLEX(-145, -81), MAKE_VAL_COMPLEX(37, 3) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_800, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_800", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_800() */
/* END TEST 800 */

/*
 *  BEGIN TEST 801
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_801( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(73, 175), MAKE_VAL_COMPLEX(-56, 60), MAKE_VAL_COMPLEX(-90, 322), MAKE_VAL_COMPLEX(20, 96), MAKE_VAL_COMPLEX(14, -14), MAKE_VAL_COMPLEX(162, -26), MAKE_VAL_COMPLEX(60, 60), MAKE_VAL_COMPLEX(81, -233), MAKE_VAL_COMPLEX(-2, -15), MAKE_VAL_COMPLEX(87, -29), MAKE_VAL_COMPLEX(-1, 164), MAKE_VAL_COMPLEX(-176, 20), MAKE_VAL_COMPLEX(-63, -39), MAKE_VAL_COMPLEX(25, 17), MAKE_VAL_COMPLEX(87, -23) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_801, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_801", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_801() */
/* END TEST 801 */

/*
 *  BEGIN TEST 802
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_802( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(12, 20), MAKE_VAL_COMPLEX(-80, 0), MAKE_VAL_COMPLEX(20, 36), MAKE_VAL_COMPLEX(-32, 8), MAKE_VAL_COMPLEX(12, -20), MAKE_VAL_COMPLEX(8, 16), MAKE_VAL_COMPLEX(40, 40), MAKE_VAL_COMPLEX(16, -40), MAKE_VAL_COMPLEX(20, 52), MAKE_VAL_COMPLEX(48, -16), MAKE_VAL_COMPLEX(-4, -44), MAKE_VAL_COMPLEX(-60, -12), MAKE_VAL_COMPLEX(-36, -12), MAKE_VAL_COMPLEX(-64, -16), MAKE_VAL_COMPLEX(56, -24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_802, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_802", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_802() */
/* END TEST 802 */

/*
 *  BEGIN TEST 803
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_803( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_ROWMAJ, y768_VW );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(313, -1318), MAKE_VAL_COMPLEX(-18, -770), MAKE_VAL_COMPLEX(2134, -1327), MAKE_VAL_COMPLEX(-7, -953), MAKE_VAL_COMPLEX(96, -55), MAKE_VAL_COMPLEX(-1319, -423), MAKE_VAL_COMPLEX(-80, -50), MAKE_VAL_COMPLEX(-1275, 997), MAKE_VAL_COMPLEX(-293, 734), MAKE_VAL_COMPLEX(-202, 4), MAKE_VAL_COMPLEX(1138, -1565), MAKE_VAL_COMPLEX(894, 161), MAKE_VAL_COMPLEX(21, 198), MAKE_VAL_COMPLEX(-532, -890), MAKE_VAL_COMPLEX(-38, -46) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_803, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_803", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_803() */
/* END TEST 803 */

/*
 *  BEGIN TEST 804
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_804( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	oski_value_t* y_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(52, -103), MAKE_VAL_COMPLEX(40, -62), MAKE_VAL_COMPLEX(142, 31), MAKE_VAL_COMPLEX(-96, 35), MAKE_VAL_COMPLEX(-14, -26), MAKE_VAL_COMPLEX(-199, -9), MAKE_VAL_COMPLEX(-66, 30), MAKE_VAL_COMPLEX(-55, -100), MAKE_VAL_COMPLEX(-54, 19), MAKE_VAL_COMPLEX(7, -7), MAKE_VAL_COMPLEX(-25, -110), MAKE_VAL_COMPLEX(28, 50), MAKE_VAL_COMPLEX(41, -20), MAKE_VAL_COMPLEX(269, -90), MAKE_VAL_COMPLEX(8, -8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_804, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_804", y_true, y_val, y768_LEN*y768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_804() */
/* END TEST 804 */

/*
 *  BEGIN TEST 805
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_805( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	oski_value_t* y_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-44, 95), MAKE_VAL_COMPLEX(-44, 68), MAKE_VAL_COMPLEX(-128, -19), MAKE_VAL_COMPLEX(76, -53), MAKE_VAL_COMPLEX(32, 14), MAKE_VAL_COMPLEX(203, 5), MAKE_VAL_COMPLEX(52, -20), MAKE_VAL_COMPLEX(55, 100), MAKE_VAL_COMPLEX(50, -37), MAKE_VAL_COMPLEX(5, 27), MAKE_VAL_COMPLEX(35, 90), MAKE_VAL_COMPLEX(-12, -40), MAKE_VAL_COMPLEX(-49, 34), MAKE_VAL_COMPLEX(-249, 84), MAKE_VAL_COMPLEX(-16, 12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_805, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_805", y_true, y_val, y768_LEN*y768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_805() */
/* END TEST 805 */

/*
 *  BEGIN TEST 806
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_806( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	oski_value_t* y_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-48, -24), MAKE_VAL_COMPLEX(33, 9), MAKE_VAL_COMPLEX(33, -81), MAKE_VAL_COMPLEX(-51, 117), MAKE_VAL_COMPLEX(-81, -63), MAKE_VAL_COMPLEX(-24, -12), MAKE_VAL_COMPLEX(66, 48), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(-75, 45), MAKE_VAL_COMPLEX(72, -84), MAKE_VAL_COMPLEX(-105, -15), MAKE_VAL_COMPLEX(21, -87), MAKE_VAL_COMPLEX(75, 15), MAKE_VAL_COMPLEX(-57, -81), MAKE_VAL_COMPLEX(30, 30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_806, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_806", y_true, y_val, y768_LEN*y768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_806() */
/* END TEST 806 */

/*
 *  BEGIN TEST 807
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_807( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	oski_value_t* y_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(762, 71), MAKE_VAL_COMPLEX(491, 10), MAKE_VAL_COMPLEX(128, 934), MAKE_VAL_COMPLEX(-335, -451), MAKE_VAL_COMPLEX(66, -39), MAKE_VAL_COMPLEX(-549, -1187), MAKE_VAL_COMPLEX(-343, -405), MAKE_VAL_COMPLEX(435, -630), MAKE_VAL_COMPLEX(-193, -191), MAKE_VAL_COMPLEX(-67, -37), MAKE_VAL_COMPLEX(605, -345), MAKE_VAL_COMPLEX(-331, 324), MAKE_VAL_COMPLEX(234, 88), MAKE_VAL_COMPLEX(1268, 1444), MAKE_VAL_COMPLEX(98, -24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_807, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_807", y_true, y_val, y768_LEN*y768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_807() */
/* END TEST 807 */

/*
 *  BEGIN TEST 808
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_808( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-323, -19), MAKE_VAL_COMPLEX(-4, -32), MAKE_VAL_COMPLEX(-249, -15), MAKE_VAL_COMPLEX(68, -234), MAKE_VAL_COMPLEX(-7, 32), MAKE_VAL_COMPLEX(-135, -75), MAKE_VAL_COMPLEX(-80, 10), MAKE_VAL_COMPLEX(-80, -77), MAKE_VAL_COMPLEX(-42, -148), MAKE_VAL_COMPLEX(-35, 61), MAKE_VAL_COMPLEX(128, -186), MAKE_VAL_COMPLEX(4, -111), MAKE_VAL_COMPLEX(64, 85), MAKE_VAL_COMPLEX(119, -99), MAKE_VAL_COMPLEX(-15, 71) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_808, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_808", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_808() */
/* END TEST 808 */

/*
 *  BEGIN TEST 809
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_809( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(283, 43), MAKE_VAL_COMPLEX(4, -128), MAKE_VAL_COMPLEX(177, 55), MAKE_VAL_COMPLEX(-84, 170), MAKE_VAL_COMPLEX(47, -8), MAKE_VAL_COMPLEX(103, 91), MAKE_VAL_COMPLEX(0, 70), MAKE_VAL_COMPLEX(160, 109), MAKE_VAL_COMPLEX(-62, 188), MAKE_VAL_COMPLEX(67, 35), MAKE_VAL_COMPLEX(-40, 178), MAKE_VAL_COMPLEX(20, -9), MAKE_VAL_COMPLEX(-40, -157), MAKE_VAL_COMPLEX(-87, -29), MAKE_VAL_COMPLEX(63, 41) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_809, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_809", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_809() */
/* END TEST 809 */

/*
 *  BEGIN TEST 810
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_810( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-13, -33), MAKE_VAL_COMPLEX(120, 20), MAKE_VAL_COMPLEX(-21, -59), MAKE_VAL_COMPLEX(50, -4), MAKE_VAL_COMPLEX(-23, 27), MAKE_VAL_COMPLEX(-8, -26), MAKE_VAL_COMPLEX(-50, -70), MAKE_VAL_COMPLEX(-34, 56), MAKE_VAL_COMPLEX(-17, -83), MAKE_VAL_COMPLEX(-76, 12), MAKE_VAL_COMPLEX(-5, 67), MAKE_VAL_COMPLEX(87, 33), MAKE_VAL_COMPLEX(51, 27), MAKE_VAL_COMPLEX(92, 40), MAKE_VAL_COMPLEX(-90, 22) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_810, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_810", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_810() */
/* END TEST 810 */

/*
 *  BEGIN TEST 811
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_811( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-2010, -696), MAKE_VAL_COMPLEX(-584, 248), MAKE_VAL_COMPLEX(-1342, -444), MAKE_VAL_COMPLEX(744, -1248), MAKE_VAL_COMPLEX(-140, -17), MAKE_VAL_COMPLEX(-637, -719), MAKE_VAL_COMPLEX(-30, -20), MAKE_VAL_COMPLEX(-522, -1105), MAKE_VAL_COMPLEX(469, -837), MAKE_VAL_COMPLEX(-91, -55), MAKE_VAL_COMPLEX(973, -1363), MAKE_VAL_COMPLEX(-287, -502), MAKE_VAL_COMPLEX(-73, 846), MAKE_VAL_COMPLEX(439, -195), MAKE_VAL_COMPLEX(43, -55) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_811, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_811", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_811() */
/* END TEST 811 */

/*
 *  BEGIN TEST 812
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_812( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(30, 297), MAKE_VAL_COMPLEX(-46, 2), MAKE_VAL_COMPLEX(45, 126), MAKE_VAL_COMPLEX(-17, -89), MAKE_VAL_COMPLEX(21, 31), MAKE_VAL_COMPLEX(-86, 118), MAKE_VAL_COMPLEX(-40, 40), MAKE_VAL_COMPLEX(-19, 46), MAKE_VAL_COMPLEX(-34, 77), MAKE_VAL_COMPLEX(9, 59), MAKE_VAL_COMPLEX(-161, -24), MAKE_VAL_COMPLEX(-66, -86), MAKE_VAL_COMPLEX(-37, -41), MAKE_VAL_COMPLEX(-131, 5), MAKE_VAL_COMPLEX(11, 49) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_812, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_812", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_812() */
/* END TEST 812 */

/*
 *  BEGIN TEST 813
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_813( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-42, -317), MAKE_VAL_COMPLEX(126, -2), MAKE_VAL_COMPLEX(-65, -162), MAKE_VAL_COMPLEX(49, 81), MAKE_VAL_COMPLEX(-33, -11), MAKE_VAL_COMPLEX(78, -134), MAKE_VAL_COMPLEX(0, -80), MAKE_VAL_COMPLEX(3, -6), MAKE_VAL_COMPLEX(14, -129), MAKE_VAL_COMPLEX(-57, -43), MAKE_VAL_COMPLEX(165, 68), MAKE_VAL_COMPLEX(126, 98), MAKE_VAL_COMPLEX(73, 53), MAKE_VAL_COMPLEX(195, 11), MAKE_VAL_COMPLEX(-67, -25) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_813, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_813", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_813() */
/* END TEST 813 */

/*
 *  BEGIN TEST 814
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_814( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-5, -14), MAKE_VAL_COMPLEX(50, 10), MAKE_VAL_COMPLEX(-8, -25), MAKE_VAL_COMPLEX(21, -1), MAKE_VAL_COMPLEX(-10, 11), MAKE_VAL_COMPLEX(-3, -11), MAKE_VAL_COMPLEX(-20, -30), MAKE_VAL_COMPLEX(-15, 23), MAKE_VAL_COMPLEX(-6, -35), MAKE_VAL_COMPLEX(-32, 4), MAKE_VAL_COMPLEX(-3, 28), MAKE_VAL_COMPLEX(36, 15), MAKE_VAL_COMPLEX(21, 12), MAKE_VAL_COMPLEX(38, 18), MAKE_VAL_COMPLEX(-38, 8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_814, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_814", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_814() */
/* END TEST 814 */

/*
 *  BEGIN TEST 815
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x768', LAYOUT_COLMAJ>
 *  y : <'y768', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_815( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A768_ptr, A768_ind, A768_val,
		A768_M, A768_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y768_val,
		y768_LEN * y768_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y768_LEN, y768_VW, LAYOUT_COLMAJ, y768_LEN );
	oski_value_t* y_val = array_Duplicate( x768_val,
		x768_LEN * x768_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x768_LEN, x768_VW, LAYOUT_COLMAJ, x768_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-95, -2308), MAKE_VAL_COMPLEX(350, -590), MAKE_VAL_COMPLEX(-508, -889), MAKE_VAL_COMPLEX(-23, 483), MAKE_VAL_COMPLEX(21, -181), MAKE_VAL_COMPLEX(684, -792), MAKE_VAL_COMPLEX(60, 40), MAKE_VAL_COMPLEX(512, -235), MAKE_VAL_COMPLEX(-49, -414), MAKE_VAL_COMPLEX(123, -113), MAKE_VAL_COMPLEX(1604, 311), MAKE_VAL_COMPLEX(520, 286), MAKE_VAL_COMPLEX(339, 83), MAKE_VAL_COMPLEX(1161, -409), MAKE_VAL_COMPLEX(161, 25) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_815, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_815", y_true, y_val, x768_LEN*x768_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_815() */
/* END TEST 815 */

#define A816_M 5    /* # of rows */
#define A816_N 5    /* # of cols */
#define A816_NNZ 9  /* # of stored non-zeros */
static oski_index_t A816_ptr[A816_M+1] = {0, 1, 4, 7, 8, 9}; /* end row pointers for A816 */
static oski_index_t A816_ind[A816_NNZ] = {4, 2, 3, 4, 2, 3, 4, 4, 4}; /* end column indices pointers for A816 */
static oski_value_t A816_val[A816_NNZ] = {MAKE_VAL_COMPLEX(-6, -5), MAKE_VAL_COMPLEX(-3, -10), MAKE_VAL_COMPLEX(-4, -6), MAKE_VAL_COMPLEX(-1, -7), MAKE_VAL_COMPLEX(-4, -8), MAKE_VAL_COMPLEX(-10, -3), MAKE_VAL_COMPLEX(5, 1), MAKE_VAL_COMPLEX(-8, -10), MAKE_VAL_COMPLEX(10, 10)}; /* end stored non-zero values for A816 */

#define x816_LEN 5   /* vector length */
#define x816_VW  3   /* vector width */
static oski_value_t x816_val[x816_LEN*x816_VW] = {MAKE_VAL_COMPLEX(-7, 10), MAKE_VAL_COMPLEX(-10, 4), MAKE_VAL_COMPLEX(-7, 9), MAKE_VAL_COMPLEX(1, -2), MAKE_VAL_COMPLEX(-2, 10), MAKE_VAL_COMPLEX(2, 2), MAKE_VAL_COMPLEX(10, 4), MAKE_VAL_COMPLEX(-6, -2), MAKE_VAL_COMPLEX(-7, -10), MAKE_VAL_COMPLEX(-9, 3), MAKE_VAL_COMPLEX(-8, -5), MAKE_VAL_COMPLEX(-7, -10), MAKE_VAL_COMPLEX(-10, -3), MAKE_VAL_COMPLEX(6, -2), MAKE_VAL_COMPLEX(-6, 8)}; /* data */

#define y816_LEN 5   /* vector length */
#define y816_VW  3   /* vector width */
static oski_value_t y816_val[y816_LEN*y816_VW] = {MAKE_VAL_COMPLEX(2, -7), MAKE_VAL_COMPLEX(10, 6), MAKE_VAL_COMPLEX(-5, -8), MAKE_VAL_COMPLEX(-10, -6), MAKE_VAL_COMPLEX(-6, 2), MAKE_VAL_COMPLEX(3, -7), MAKE_VAL_COMPLEX(-5, -7), MAKE_VAL_COMPLEX(-3, 10), MAKE_VAL_COMPLEX(3, -10), MAKE_VAL_COMPLEX(-5, 3), MAKE_VAL_COMPLEX(8, 0), MAKE_VAL_COMPLEX(-1, -9), MAKE_VAL_COMPLEX(1, -2), MAKE_VAL_COMPLEX(-2, 8), MAKE_VAL_COMPLEX(9, 0)}; /* data */

/*
 *  BEGIN TEST 816
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_816( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	oski_value_t* y_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(52, 70), MAKE_VAL_COMPLEX(-52, -8), MAKE_VAL_COMPLEX(84, -23), MAKE_VAL_COMPLEX(59, -7), MAKE_VAL_COMPLEX(-22, 88), MAKE_VAL_COMPLEX(-42, 219), MAKE_VAL_COMPLEX(28, -133), MAKE_VAL_COMPLEX(201, 113), MAKE_VAL_COMPLEX(-26, 228), MAKE_VAL_COMPLEX(-57, 51), MAKE_VAL_COMPLEX(54, -26), MAKE_VAL_COMPLEX(181, 96), MAKE_VAL_COMPLEX(157, -63), MAKE_VAL_COMPLEX(210, 172), MAKE_VAL_COMPLEX(-110, 87) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_816, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_816", y_true, y_val, y816_LEN*y816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_816() */
/* END TEST 816 */

/*
 *  BEGIN TEST 817
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_817( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	oski_value_t* y_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-120, -44), MAKE_VAL_COMPLEX(40, -108), MAKE_VAL_COMPLEX(-118, 111), MAKE_VAL_COMPLEX(-47, 123), MAKE_VAL_COMPLEX(74, -52), MAKE_VAL_COMPLEX(-32, -201), MAKE_VAL_COMPLEX(-54, 215), MAKE_VAL_COMPLEX(-103, -149), MAKE_VAL_COMPLEX(-72, -192), MAKE_VAL_COMPLEX(111, -29), MAKE_VAL_COMPLEX(-102, -38), MAKE_VAL_COMPLEX(-247, -34), MAKE_VAL_COMPLEX(-179, 67), MAKE_VAL_COMPLEX(-134, -204), MAKE_VAL_COMPLEX(56, -159) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_817, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_817", y_true, y_val, y816_LEN*y816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_817() */
/* END TEST 817 */

/*
 *  BEGIN TEST 818
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_818( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	oski_value_t* y_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(13, 34), MAKE_VAL_COMPLEX(-58, 6), MAKE_VAL_COMPLEX(44, 17), MAKE_VAL_COMPLEX(58, -6), MAKE_VAL_COMPLEX(18, -26), MAKE_VAL_COMPLEX(9, 37), MAKE_VAL_COMPLEX(41, 13), MAKE_VAL_COMPLEX(-18, -49), MAKE_VAL_COMPLEX(18, 49), MAKE_VAL_COMPLEX(11, -27), MAKE_VAL_COMPLEX(-32, 24), MAKE_VAL_COMPLEX(31, 33), MAKE_VAL_COMPLEX(2, 11), MAKE_VAL_COMPLEX(-16, -38), MAKE_VAL_COMPLEX(-36, 27) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_818, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_818", y_true, y_val, y816_LEN*y816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_818() */
/* END TEST 818 */

/*
 *  BEGIN TEST 819
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_819( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	oski_value_t* y_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(207, 105), MAKE_VAL_COMPLEX(-152, 80), MAKE_VAL_COMPLEX(190, -147), MAKE_VAL_COMPLEX(151, -117), MAKE_VAL_COMPLEX(40, 166), MAKE_VAL_COMPLEX(167, 502), MAKE_VAL_COMPLEX(-37, -312), MAKE_VAL_COMPLEX(468, 0), MAKE_VAL_COMPLEX(223, 507), MAKE_VAL_COMPLEX(-73, 131), MAKE_VAL_COMPLEX(74, -66), MAKE_VAL_COMPLEX(504, 15), MAKE_VAL_COMPLEX(260, -276), MAKE_VAL_COMPLEX(554, 116), MAKE_VAL_COMPLEX(-142, 329) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_819, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_819", y_true, y_val, y816_LEN*y816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_819() */
/* END TEST 819 */

/*
 *  BEGIN TEST 820
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_820( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-80, -8), MAKE_VAL_COMPLEX(12, -80), MAKE_VAL_COMPLEX(-113, -62), MAKE_VAL_COMPLEX(-20, 85), MAKE_VAL_COMPLEX(81, -32), MAKE_VAL_COMPLEX(-174, -7), MAKE_VAL_COMPLEX(0, 220), MAKE_VAL_COMPLEX(30, 18), MAKE_VAL_COMPLEX(-107, 54), MAKE_VAL_COMPLEX(-28, 136), MAKE_VAL_COMPLEX(201, -157), MAKE_VAL_COMPLEX(-150, -44), MAKE_VAL_COMPLEX(-2, 28), MAKE_VAL_COMPLEX(-177, 7), MAKE_VAL_COMPLEX(-81, 170) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_820, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_820", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_820() */
/* END TEST 820 */

/*
 *  BEGIN TEST 821
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_821( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(82, -80), MAKE_VAL_COMPLEX(32, 16), MAKE_VAL_COMPLEX(119, -20), MAKE_VAL_COMPLEX(22, -69), MAKE_VAL_COMPLEX(-109, -36), MAKE_VAL_COMPLEX(154, 3), MAKE_VAL_COMPLEX(-76, -204), MAKE_VAL_COMPLEX(14, -30), MAKE_VAL_COMPLEX(189, -22), MAKE_VAL_COMPLEX(70, -190), MAKE_VAL_COMPLEX(-133, 155), MAKE_VAL_COMPLEX(232, 76), MAKE_VAL_COMPLEX(74, -50), MAKE_VAL_COMPLEX(149, 29), MAKE_VAL_COMPLEX(85, -242) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_821, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_821", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_821() */
/* END TEST 821 */

/*
 *  BEGIN TEST 822
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_822( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-105, 1), MAKE_VAL_COMPLEX(-78, -50), MAKE_VAL_COMPLEX(-98, -4), MAKE_VAL_COMPLEX(19, -3), MAKE_VAL_COMPLEX(-80, 36), MAKE_VAL_COMPLEX(-4, 24), MAKE_VAL_COMPLEX(22, 90), MAKE_VAL_COMPLEX(-16, -52), MAKE_VAL_COMPLEX(35, -99), MAKE_VAL_COMPLEX(-66, -48), MAKE_VAL_COMPLEX(-5, -81), MAKE_VAL_COMPLEX(35, -99), MAKE_VAL_COMPLEX(-29, -85), MAKE_VAL_COMPLEX(44, 32), MAKE_VAL_COMPLEX(-86, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_822, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_822", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_822() */
/* END TEST 822 */

/*
 *  BEGIN TEST 823
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_823( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-52, -5), MAKE_VAL_COMPLEX(300, -206), MAKE_VAL_COMPLEX(-242, -261), MAKE_VAL_COMPLEX(-164, 428), MAKE_VAL_COMPLEX(683, -250), MAKE_VAL_COMPLEX(-848, -113), MAKE_VAL_COMPLEX(-40, 794), MAKE_VAL_COMPLEX(184, 268), MAKE_VAL_COMPLEX(-687, 585), MAKE_VAL_COMPLEX(61, 863), MAKE_VAL_COMPLEX(992, -515), MAKE_VAL_COMPLEX(-902, 95), MAKE_VAL_COMPLEX(55, 432), MAKE_VAL_COMPLEX(-1019, -87), MAKE_VAL_COMPLEX(-121, 888) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_823, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_823", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_823() */
/* END TEST 823 */

/*
 *  BEGIN TEST 824
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_824( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-27, -109), MAKE_VAL_COMPLEX(10, -166), MAKE_VAL_COMPLEX(-77, -74), MAKE_VAL_COMPLEX(109, -40), MAKE_VAL_COMPLEX(-243, -120), MAKE_VAL_COMPLEX(110, 155), MAKE_VAL_COMPLEX(108, -98), MAKE_VAL_COMPLEX(-94, -98), MAKE_VAL_COMPLEX(340, 207), MAKE_VAL_COMPLEX(198, -48), MAKE_VAL_COMPLEX(44, -266), MAKE_VAL_COMPLEX(87, 259), MAKE_VAL_COMPLEX(137, -205), MAKE_VAL_COMPLEX(-133, 265), MAKE_VAL_COMPLEX(287, -134) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_824, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_824", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_824() */
/* END TEST 824 */

/*
 *  BEGIN TEST 825
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_825( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1, -149), MAKE_VAL_COMPLEX(94, -38), MAKE_VAL_COMPLEX(63, -168), MAKE_VAL_COMPLEX(-97, 86), MAKE_VAL_COMPLEX(135, -68), MAKE_VAL_COMPLEX(-170, -159), MAKE_VAL_COMPLEX(-324, 174), MAKE_VAL_COMPLEX(218, 46), MAKE_VAL_COMPLEX(-88, -145), MAKE_VAL_COMPLEX(-96, -126), MAKE_VAL_COMPLEX(154, 234), MAKE_VAL_COMPLEX(165, -197), MAKE_VAL_COMPLEX(65, 113), MAKE_VAL_COMPLEX(65, -149), MAKE_VAL_COMPLEX(-303, -78) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_825, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_825", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_825() */
/* END TEST 825 */

/*
 *  BEGIN TEST 826
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_826( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-31, 23), MAKE_VAL_COMPLEX(-34, 2), MAKE_VAL_COMPLEX(-30, 20), MAKE_VAL_COMPLEX(5, -5), MAKE_VAL_COMPLEX(-16, 28), MAKE_VAL_COMPLEX(4, 8), MAKE_VAL_COMPLEX(26, 22), MAKE_VAL_COMPLEX(-16, -12), MAKE_VAL_COMPLEX(-11, -37), MAKE_VAL_COMPLEX(-30, 0), MAKE_VAL_COMPLEX(-19, -23), MAKE_VAL_COMPLEX(-11, -37), MAKE_VAL_COMPLEX(-27, -19), MAKE_VAL_COMPLEX(20, 0), MAKE_VAL_COMPLEX(-26, 18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_826, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_826", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_826() */
/* END TEST 826 */

/*
 *  BEGIN TEST 827
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_827( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(173, 183), MAKE_VAL_COMPLEX(-650, -202), MAKE_VAL_COMPLEX(-100, 809), MAKE_VAL_COMPLEX(360, -1209), MAKE_VAL_COMPLEX(-1612, 1017), MAKE_VAL_COMPLEX(2209, 242), MAKE_VAL_COMPLEX(690, -2704), MAKE_VAL_COMPLEX(-1698, 574), MAKE_VAL_COMPLEX(2961, 62), MAKE_VAL_COMPLEX(1566, -696), MAKE_VAL_COMPLEX(-2132, -1512), MAKE_VAL_COMPLEX(1301, 2249), MAKE_VAL_COMPLEX(-733, -1431), MAKE_VAL_COMPLEX(579, 2335), MAKE_VAL_COMPLEX(2278, -2341) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_827, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_827", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_827() */
/* END TEST 827 */

/*
 *  BEGIN TEST 828
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_828( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	oski_value_t* y_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(100, 114), MAKE_VAL_COMPLEX(-41, 69), MAKE_VAL_COMPLEX(113, -141), MAKE_VAL_COMPLEX(40, -10), MAKE_VAL_COMPLEX(161, -126), MAKE_VAL_COMPLEX(5, 37), MAKE_VAL_COMPLEX(63, 77), MAKE_VAL_COMPLEX(133, 49), MAKE_VAL_COMPLEX(132, 33), MAKE_VAL_COMPLEX(211, 117), MAKE_VAL_COMPLEX(44, 54), MAKE_VAL_COMPLEX(36, 243), MAKE_VAL_COMPLEX(-22, 242), MAKE_VAL_COMPLEX(108, 47), MAKE_VAL_COMPLEX(-146, 159) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_828, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_828", y_true, y_val, y816_LEN*y816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_828() */
/* END TEST 828 */

/*
 *  BEGIN TEST 829
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_829( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	oski_value_t* y_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-3, -109), MAKE_VAL_COMPLEX(-7, 79), MAKE_VAL_COMPLEX(-24, 52), MAKE_VAL_COMPLEX(8, -138), MAKE_VAL_COMPLEX(-205, 54), MAKE_VAL_COMPLEX(95, -19), MAKE_VAL_COMPLEX(13, -163), MAKE_VAL_COMPLEX(-272, -58), MAKE_VAL_COMPLEX(7, -24), MAKE_VAL_COMPLEX(-265, -173), MAKE_VAL_COMPLEX(-20, 50), MAKE_VAL_COMPLEX(78, -283), MAKE_VAL_COMPLEX(51, -235), MAKE_VAL_COMPLEX(-218, -49), MAKE_VAL_COMPLEX(173, -42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_829, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_829", y_true, y_val, y816_LEN*y816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_829() */
/* END TEST 829 */

/*
 *  BEGIN TEST 830
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_830( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	oski_value_t* y_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-8, 28), MAKE_VAL_COMPLEX(-40, -24), MAKE_VAL_COMPLEX(20, 32), MAKE_VAL_COMPLEX(40, 24), MAKE_VAL_COMPLEX(24, -8), MAKE_VAL_COMPLEX(-12, 28), MAKE_VAL_COMPLEX(20, 28), MAKE_VAL_COMPLEX(12, -40), MAKE_VAL_COMPLEX(-12, 40), MAKE_VAL_COMPLEX(20, -12), MAKE_VAL_COMPLEX(-32, 0), MAKE_VAL_COMPLEX(4, 36), MAKE_VAL_COMPLEX(-4, 8), MAKE_VAL_COMPLEX(8, -32), MAKE_VAL_COMPLEX(-36, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_830, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_830", y_true, y_val, y816_LEN*y816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_830() */
/* END TEST 830 */

/*
 *  BEGIN TEST 831
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_831( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	oski_value_t* y_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-243, -66), MAKE_VAL_COMPLEX(-74, 82), MAKE_VAL_COMPLEX(180, 297), MAKE_VAL_COMPLEX(-42, -202), MAKE_VAL_COMPLEX(-188, 460), MAKE_VAL_COMPLEX(113, -79), MAKE_VAL_COMPLEX(-179, -227), MAKE_VAL_COMPLEX(-630, 219), MAKE_VAL_COMPLEX(-64, 147), MAKE_VAL_COMPLEX(-785, 109), MAKE_VAL_COMPLEX(-100, 164), MAKE_VAL_COMPLEX(-363, -545), MAKE_VAL_COMPLEX(-382, -529), MAKE_VAL_COMPLEX(-518, 172), MAKE_VAL_COMPLEX(82, -403) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-2, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_831, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_831", y_true, y_val, y816_LEN*y816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_831() */
/* END TEST 831 */

/*
 *  BEGIN TEST 832
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_832( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(41, 89), MAKE_VAL_COMPLEX(7, 57), MAKE_VAL_COMPLEX(-62, 21), MAKE_VAL_COMPLEX(-42, 185), MAKE_VAL_COMPLEX(83, 110), MAKE_VAL_COMPLEX(40, -83), MAKE_VAL_COMPLEX(98, 38), MAKE_VAL_COMPLEX(-87, 164), MAKE_VAL_COMPLEX(-83, -57), MAKE_VAL_COMPLEX(54, 97), MAKE_VAL_COMPLEX(53, -8), MAKE_VAL_COMPLEX(-203, -60), MAKE_VAL_COMPLEX(-151, 34), MAKE_VAL_COMPLEX(-302, 118), MAKE_VAL_COMPLEX(77, 71) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_832, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_832", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_832() */
/* END TEST 832 */

/*
 *  BEGIN TEST 833
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_833( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-81, 32), MAKE_VAL_COMPLEX(-95, 13), MAKE_VAL_COMPLEX(19, 90), MAKE_VAL_COMPLEX(46, -208), MAKE_VAL_COMPLEX(-73, -4), MAKE_VAL_COMPLEX(-14, 97), MAKE_VAL_COMPLEX(14, -28), MAKE_VAL_COMPLEX(21, -166), MAKE_VAL_COMPLEX(-17, -22), MAKE_VAL_COMPLEX(-135, -40), MAKE_VAL_COMPLEX(-148, -18), MAKE_VAL_COMPLEX(103, -19), MAKE_VAL_COMPLEX(42, -34), MAKE_VAL_COMPLEX(356, -156), MAKE_VAL_COMPLEX(-113, 27) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_833, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_833", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_833() */
/* END TEST 833 */

/*
 *  BEGIN TEST 834
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_834( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(51, -9), MAKE_VAL_COMPLEX(42, 18), MAKE_VAL_COMPLEX(48, -6), MAKE_VAL_COMPLEX(-9, 3), MAKE_VAL_COMPLEX(36, -24), MAKE_VAL_COMPLEX(0, -12), MAKE_VAL_COMPLEX(-18, -42), MAKE_VAL_COMPLEX(12, 24), MAKE_VAL_COMPLEX(-9, 51), MAKE_VAL_COMPLEX(36, 18), MAKE_VAL_COMPLEX(9, 39), MAKE_VAL_COMPLEX(-9, 51), MAKE_VAL_COMPLEX(21, 39), MAKE_VAL_COMPLEX(-24, -12), MAKE_VAL_COMPLEX(42, -6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_834, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_834", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_834() */
/* END TEST 834 */

/*
 *  BEGIN TEST 835
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_835( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(316, 138), MAKE_VAL_COMPLEX(228, 44), MAKE_VAL_COMPLEX(-416, -48), MAKE_VAL_COMPLEX(146, 1248), MAKE_VAL_COMPLEX(530, 290), MAKE_VAL_COMPLEX(-2, -566), MAKE_VAL_COMPLEX(416, 188), MAKE_VAL_COMPLEX(-54, 1058), MAKE_VAL_COMPLEX(-280, -170), MAKE_VAL_COMPLEX(596, 228), MAKE_VAL_COMPLEX(540, -250), MAKE_VAL_COMPLEX(-1006, 52), MAKE_VAL_COMPLEX(-612, 334), MAKE_VAL_COMPLEX(-1628, 1476), MAKE_VAL_COMPLEX(524, 12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_835, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_835", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_835() */
/* END TEST 835 */

/*
 *  BEGIN TEST 836
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_836( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-70, 10), MAKE_VAL_COMPLEX(-81, -51), MAKE_VAL_COMPLEX(-146, 89), MAKE_VAL_COMPLEX(179, -118), MAKE_VAL_COMPLEX(-115, 32), MAKE_VAL_COMPLEX(-24, 59), MAKE_VAL_COMPLEX(148, 192), MAKE_VAL_COMPLEX(-45, -4), MAKE_VAL_COMPLEX(130, -170), MAKE_VAL_COMPLEX(-58, 13), MAKE_VAL_COMPLEX(28, -257), MAKE_VAL_COMPLEX(-34, 35), MAKE_VAL_COMPLEX(-30, 45), MAKE_VAL_COMPLEX(178, 280), MAKE_VAL_COMPLEX(-35, -103) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_836, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_836", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_836() */
/* END TEST 836 */

/*
 *  BEGIN TEST 837
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_837( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-18, 137), MAKE_VAL_COMPLEX(-55, 117), MAKE_VAL_COMPLEX(57, 44), MAKE_VAL_COMPLEX(-167, 89), MAKE_VAL_COMPLEX(97, 110), MAKE_VAL_COMPLEX(54, -33), MAKE_VAL_COMPLEX(-4, -146), MAKE_VAL_COMPLEX(-41, -18), MAKE_VAL_COMPLEX(-238, 37), MAKE_VAL_COMPLEX(-65, 38), MAKE_VAL_COMPLEX(-145, 195), MAKE_VAL_COMPLEX(-74, -168), MAKE_VAL_COMPLEX(-113, -77), MAKE_VAL_COMPLEX(-96, -314), MAKE_VAL_COMPLEX(-41, 221) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_837, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_837", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_837() */
/* END TEST 837 */

/*
 *  BEGIN TEST 838
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_838( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-88, -2), MAKE_VAL_COMPLEX(-64, -44), MAKE_VAL_COMPLEX(-82, -6), MAKE_VAL_COMPLEX(16, -2), MAKE_VAL_COMPLEX(-68, 28), MAKE_VAL_COMPLEX(-4, 20), MAKE_VAL_COMPLEX(16, 76), MAKE_VAL_COMPLEX(-12, -44), MAKE_VAL_COMPLEX(32, -82), MAKE_VAL_COMPLEX(-54, -42), MAKE_VAL_COMPLEX(-2, -68), MAKE_VAL_COMPLEX(32, -82), MAKE_VAL_COMPLEX(-22, -72), MAKE_VAL_COMPLEX(36, 28), MAKE_VAL_COMPLEX(-72, -4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_838, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_838", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_838() */
/* END TEST 838 */

/*
 *  BEGIN TEST 839
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_ROWMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_839( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_ROWMAJ, x816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-285, -221), MAKE_VAL_COMPLEX(-300, -154), MAKE_VAL_COMPLEX(245, 425), MAKE_VAL_COMPLEX(-576, -1248), MAKE_VAL_COMPLEX(-351, 381), MAKE_VAL_COMPLEX(213, 263), MAKE_VAL_COMPLEX(930, -454), MAKE_VAL_COMPLEX(243, 63), MAKE_VAL_COMPLEX(-285, -1101), MAKE_VAL_COMPLEX(189, -163), MAKE_VAL_COMPLEX(-1226, -702), MAKE_VAL_COMPLEX(986, 252), MAKE_VAL_COMPLEX(794, -170), MAKE_VAL_COMPLEX(2034, -578), MAKE_VAL_COMPLEX(-970, -456) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_839, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_839", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_839() */
/* END TEST 839 */

/*
 *  BEGIN TEST 840
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_840( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	oski_value_t* y_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(60, -96), MAKE_VAL_COMPLEX(141, 43), MAKE_VAL_COMPLEX(30, -56), MAKE_VAL_COMPLEX(95, 33), MAKE_VAL_COMPLEX(-36, 232), MAKE_VAL_COMPLEX(30, 67), MAKE_VAL_COMPLEX(90, 141), MAKE_VAL_COMPLEX(12, 137), MAKE_VAL_COMPLEX(-169, 162), MAKE_VAL_COMPLEX(253, -57), MAKE_VAL_COMPLEX(188, 12), MAKE_VAL_COMPLEX(163, 86), MAKE_VAL_COMPLEX(-60, 151), MAKE_VAL_COMPLEX(-172, 30), MAKE_VAL_COMPLEX(-241, 62) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_840, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_840", y_true, y_val, y816_LEN*y816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_840() */
/* END TEST 840 */

/*
 *  BEGIN TEST 841
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_841( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	oski_value_t* y_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-24, 23), MAKE_VAL_COMPLEX(-43, 43), MAKE_VAL_COMPLEX(-69, -42), MAKE_VAL_COMPLEX(-193, -119), MAKE_VAL_COMPLEX(-34, -222), MAKE_VAL_COMPLEX(17, -138), MAKE_VAL_COMPLEX(-131, -228), MAKE_VAL_COMPLEX(-65, -33), MAKE_VAL_COMPLEX(222, -266), MAKE_VAL_COMPLEX(-314, 80), MAKE_VAL_COMPLEX(-100, 4), MAKE_VAL_COMPLEX(-156, -187), MAKE_VAL_COMPLEX(75, -171), MAKE_VAL_COMPLEX(134, 54), MAKE_VAL_COMPLEX(340, -44) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_841, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_841", y_true, y_val, y816_LEN*y816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_841() */
/* END TEST 841 */

/*
 *  BEGIN TEST 842
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_842( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	oski_value_t* y_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(48, -9), MAKE_VAL_COMPLEX(-6, 78), MAKE_VAL_COMPLEX(33, -54), MAKE_VAL_COMPLEX(6, -78), MAKE_VAL_COMPLEX(-30, -30), MAKE_VAL_COMPLEX(51, -3), MAKE_VAL_COMPLEX(27, -51), MAKE_VAL_COMPLEX(-69, 12), MAKE_VAL_COMPLEX(69, -12), MAKE_VAL_COMPLEX(-33, -21), MAKE_VAL_COMPLEX(24, 48), MAKE_VAL_COMPLEX(51, -33), MAKE_VAL_COMPLEX(15, 0), MAKE_VAL_COMPLEX(-54, 12), MAKE_VAL_COMPLEX(27, 54) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_842, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_842", y_true, y_val, y816_LEN*y816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_842() */
/* END TEST 842 */

/*
 *  BEGIN TEST 843
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_843( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	oski_value_t* y_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-369, 0), MAKE_VAL_COMPLEX(-168, -354), MAKE_VAL_COMPLEX(-276, -111), MAKE_VAL_COMPLEX(-312, -582), MAKE_VAL_COMPLEX(678, -606), MAKE_VAL_COMPLEX(228, -390), MAKE_VAL_COMPLEX(126, -864), MAKE_VAL_COMPLEX(231, -294), MAKE_VAL_COMPLEX(1137, -132), MAKE_VAL_COMPLEX(-1047, -579), MAKE_VAL_COMPLEX(-384, -528), MAKE_VAL_COMPLEX(-168, -918), MAKE_VAL_COMPLEX(669, -300), MAKE_VAL_COMPLEX(498, 552), MAKE_VAL_COMPLEX(1071, 618) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-3, -3);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_843, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_843", y_true, y_val, y816_LEN*y816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_843() */
/* END TEST 843 */

/*
 *  BEGIN TEST 844
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_844( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-95, 56), MAKE_VAL_COMPLEX(-114, 82), MAKE_VAL_COMPLEX(-76, 204), MAKE_VAL_COMPLEX(18, 164), MAKE_VAL_COMPLEX(-41, 148), MAKE_VAL_COMPLEX(60, -18), MAKE_VAL_COMPLEX(193, 16), MAKE_VAL_COMPLEX(-4, 20), MAKE_VAL_COMPLEX(173, -198), MAKE_VAL_COMPLEX(-271, -25), MAKE_VAL_COMPLEX(-95, -104), MAKE_VAL_COMPLEX(-187, -112), MAKE_VAL_COMPLEX(-204, 58), MAKE_VAL_COMPLEX(-138, 15), MAKE_VAL_COMPLEX(-95, 222) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_844, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_844", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_844() */
/* END TEST 844 */

/*
 *  BEGIN TEST 845
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_845( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(82, -80), MAKE_VAL_COMPLEX(116, -106), MAKE_VAL_COMPLEX(65, -227), MAKE_VAL_COMPLEX(-15, -160), MAKE_VAL_COMPLEX(23, -162), MAKE_VAL_COMPLEX(-66, 20), MAKE_VAL_COMPLEX(-211, 0), MAKE_VAL_COMPLEX(14, -30), MAKE_VAL_COMPLEX(-146, 194), MAKE_VAL_COMPLEX(274, 4), MAKE_VAL_COMPLEX(113, 93), MAKE_VAL_COMPLEX(214, 108), MAKE_VAL_COMPLEX(220, -75), MAKE_VAL_COMPLEX(136, -1), MAKE_VAL_COMPLEX(85, -242) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_845, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_845", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_845() */
/* END TEST 845 */

/*
 *  BEGIN TEST 846
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_846( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-7, 10), MAKE_VAL_COMPLEX(-10, 4), MAKE_VAL_COMPLEX(-7, 9), MAKE_VAL_COMPLEX(1, -2), MAKE_VAL_COMPLEX(-2, 10), MAKE_VAL_COMPLEX(2, 2), MAKE_VAL_COMPLEX(10, 4), MAKE_VAL_COMPLEX(-6, -2), MAKE_VAL_COMPLEX(-7, -10), MAKE_VAL_COMPLEX(-9, 3), MAKE_VAL_COMPLEX(-8, -5), MAKE_VAL_COMPLEX(-7, -10), MAKE_VAL_COMPLEX(-10, -3), MAKE_VAL_COMPLEX(6, -2), MAKE_VAL_COMPLEX(-6, 8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_846, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_846", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_846() */
/* END TEST 846 */

/*
 *  BEGIN TEST 847
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_847( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-24, -51), MAKE_VAL_COMPLEX(-736, 76), MAKE_VAL_COMPLEX(-1224, 542), MAKE_VAL_COMPLEX(-1401, 557), MAKE_VAL_COMPLEX(-583, 292), MAKE_VAL_COMPLEX(476, 286), MAKE_VAL_COMPLEX(773, 874), MAKE_VAL_COMPLEX(-336, 460), MAKE_VAL_COMPLEX(1352, 1271), MAKE_VAL_COMPLEX(-421, -1577), MAKE_VAL_COMPLEX(158, -503), MAKE_VAL_COMPLEX(-416, -1351), MAKE_VAL_COMPLEX(-1325, -737), MAKE_VAL_COMPLEX(-662, -1503), MAKE_VAL_COMPLEX(-1495, 368) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_847, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_847", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_847() */
/* END TEST 847 */

/*
 *  BEGIN TEST 848
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_848( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-36, -11), MAKE_VAL_COMPLEX(84, -102), MAKE_VAL_COMPLEX(174, -178), MAKE_VAL_COMPLEX(167, 49), MAKE_VAL_COMPLEX(3, -154), MAKE_VAL_COMPLEX(-36, -50), MAKE_VAL_COMPLEX(-193, 6), MAKE_VAL_COMPLEX(-144, -88), MAKE_VAL_COMPLEX(-12, -265), MAKE_VAL_COMPLEX(-119, 167), MAKE_VAL_COMPLEX(-34, 13), MAKE_VAL_COMPLEX(180, 125), MAKE_VAL_COMPLEX(223, 153), MAKE_VAL_COMPLEX(-34, 269), MAKE_VAL_COMPLEX(277, -54) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_848, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_848", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_848() */
/* END TEST 848 */

/*
 *  BEGIN TEST 849
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_849( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(93, 36), MAKE_VAL_COMPLEX(-54, 148), MAKE_VAL_COMPLEX(-122, 204), MAKE_VAL_COMPLEX(-178, -52), MAKE_VAL_COMPLEX(49, 154), MAKE_VAL_COMPLEX(44, 38), MAKE_VAL_COMPLEX(203, -60), MAKE_VAL_COMPLEX(140, 120), MAKE_VAL_COMPLEX(-31, 310), MAKE_VAL_COMPLEX(143, -125), MAKE_VAL_COMPLEX(17, 32), MAKE_VAL_COMPLEX(-223, -80), MAKE_VAL_COMPLEX(-228, -100), MAKE_VAL_COMPLEX(18, -297), MAKE_VAL_COMPLEX(-231, 76) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_849, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_849", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_849() */
/* END TEST 849 */

/*
 *  BEGIN TEST 850
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_850( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(28, -40), MAKE_VAL_COMPLEX(40, -16), MAKE_VAL_COMPLEX(28, -36), MAKE_VAL_COMPLEX(-4, 8), MAKE_VAL_COMPLEX(8, -40), MAKE_VAL_COMPLEX(-8, -8), MAKE_VAL_COMPLEX(-40, -16), MAKE_VAL_COMPLEX(24, 8), MAKE_VAL_COMPLEX(28, 40), MAKE_VAL_COMPLEX(36, -12), MAKE_VAL_COMPLEX(32, 20), MAKE_VAL_COMPLEX(28, 40), MAKE_VAL_COMPLEX(40, 12), MAKE_VAL_COMPLEX(-24, 8), MAKE_VAL_COMPLEX(24, -32) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_850, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_850", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_850() */
/* END TEST 850 */

/*
 *  BEGIN TEST 851
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_851( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_ROWMAJ, y816_VW );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(107, 22), MAKE_VAL_COMPLEX(394, -812), MAKE_VAL_COMPLEX(677, -1799), MAKE_VAL_COMPLEX(1133, -354), MAKE_VAL_COMPLEX(-336, -1102), MAKE_VAL_COMPLEX(-414, -242), MAKE_VAL_COMPLEX(-1256, 504), MAKE_VAL_COMPLEX(-1134, 222), MAKE_VAL_COMPLEX(-1205, -1178), MAKE_VAL_COMPLEX(209, 1613), MAKE_VAL_COMPLEX(-94, 495), MAKE_VAL_COMPLEX(1507, 394), MAKE_VAL_COMPLEX(2094, 309), MAKE_VAL_COMPLEX(702, 1660), MAKE_VAL_COMPLEX(1748, -1468) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, -4);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_851, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_851", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_851() */
/* END TEST 851 */

/*
 *  BEGIN TEST 852
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_852( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	oski_value_t* y_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(41, -56), MAKE_VAL_COMPLEX(185, 19), MAKE_VAL_COMPLEX(110, 188), MAKE_VAL_COMPLEX(259, -55), MAKE_VAL_COMPLEX(-56, 183), MAKE_VAL_COMPLEX(48, 18), MAKE_VAL_COMPLEX(-25, 223), MAKE_VAL_COMPLEX(40, 80), MAKE_VAL_COMPLEX(110, 19), MAKE_VAL_COMPLEX(-167, -7), MAKE_VAL_COMPLEX(76, -42), MAKE_VAL_COMPLEX(-1, 118), MAKE_VAL_COMPLEX(-173, 225), MAKE_VAL_COMPLEX(211, 144), MAKE_VAL_COMPLEX(-295, 53) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_852, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_852", y_true, y_val, y816_LEN*y816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_852() */
/* END TEST 852 */

/*
 *  BEGIN TEST 853
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_853( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	oski_value_t* y_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-54, 75), MAKE_VAL_COMPLEX(-209, -47), MAKE_VAL_COMPLEX(-103, -159), MAKE_VAL_COMPLEX(-235, 83), MAKE_VAL_COMPLEX(76, -183), MAKE_VAL_COMPLEX(-64, 0), MAKE_VAL_COMPLEX(33, -197), MAKE_VAL_COMPLEX(-21, -107), MAKE_VAL_COMPLEX(-129, 8), MAKE_VAL_COMPLEX(185, 3), MAKE_VAL_COMPLEX(-100, 34), MAKE_VAL_COMPLEX(-5, -90), MAKE_VAL_COMPLEX(168, -220), MAKE_VAL_COMPLEX(-197, -166), MAKE_VAL_COMPLEX(268, -62) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_853, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_853", y_true, y_val, y816_LEN*y816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_853() */
/* END TEST 853 */

/*
 *  BEGIN TEST 854
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_854( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	oski_value_t* y_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-45, -28), MAKE_VAL_COMPLEX(62, -58), MAKE_VAL_COMPLEX(-66, 19), MAKE_VAL_COMPLEX(-62, 58), MAKE_VAL_COMPLEX(2, 46), MAKE_VAL_COMPLEX(-43, -35), MAKE_VAL_COMPLEX(-59, 21), MAKE_VAL_COMPLEX(64, 41), MAKE_VAL_COMPLEX(-64, -41), MAKE_VAL_COMPLEX(11, 41), MAKE_VAL_COMPLEX(16, -56), MAKE_VAL_COMPLEX(-65, -11), MAKE_VAL_COMPLEX(-12, -11), MAKE_VAL_COMPLEX(52, 30), MAKE_VAL_COMPLEX(18, -63) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_854, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_854", y_true, y_val, y816_LEN*y816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_854() */
/* END TEST 854 */

/*
 *  BEGIN TEST 855
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_855( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	oski_value_t* y_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(608, 104), MAKE_VAL_COMPLEX(502, 1136), MAKE_VAL_COMPLEX(-111, 1527), MAKE_VAL_COMPLEX(1898, 904), MAKE_VAL_COMPLEX(-1127, 459), MAKE_VAL_COMPLEX(254, 532), MAKE_VAL_COMPLEX(-984, 1008), MAKE_VAL_COMPLEX(-373, 341), MAKE_VAL_COMPLEX(628, 904), MAKE_VAL_COMPLEX(-774, -1042), MAKE_VAL_COMPLEX(438, 354), MAKE_VAL_COMPLEX(-369, 733), MAKE_VAL_COMPLEX(-1963, 321), MAKE_VAL_COMPLEX(189, 1577), MAKE_VAL_COMPLEX(-1911, -1003) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_855, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_855", y_true, y_val, y816_LEN*y816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_855() */
/* END TEST 855 */

/*
 *  BEGIN TEST 856
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_856( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(61, 103), MAKE_VAL_COMPLEX(-71, 268), MAKE_VAL_COMPLEX(46, 128), MAKE_VAL_COMPLEX(85, 40), MAKE_VAL_COMPLEX(-52, 79), MAKE_VAL_COMPLEX(65, 7), MAKE_VAL_COMPLEX(133, 24), MAKE_VAL_COMPLEX(-91, 176), MAKE_VAL_COMPLEX(23, -22), MAKE_VAL_COMPLEX(-356, 206), MAKE_VAL_COMPLEX(-119, -30), MAKE_VAL_COMPLEX(-61, -102), MAKE_VAL_COMPLEX(-83, 7), MAKE_VAL_COMPLEX(-118, -71), MAKE_VAL_COMPLEX(93, 83) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_856, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_856", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_856() */
/* END TEST 856 */

/*
 *  BEGIN TEST 857
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_857( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-134, -20), MAKE_VAL_COMPLEX(-23, -242), MAKE_VAL_COMPLEX(-118, -54), MAKE_VAL_COMPLEX(-74, -57), MAKE_VAL_COMPLEX(24, 9), MAKE_VAL_COMPLEX(-49, 13), MAKE_VAL_COMPLEX(-47, 22), MAKE_VAL_COMPLEX(39, -200), MAKE_VAL_COMPLEX(-76, -75), MAKE_VAL_COMPLEX(272, -188), MAKE_VAL_COMPLEX(52, -23), MAKE_VAL_COMPLEX(8, 5), MAKE_VAL_COMPLEX(-4, -44), MAKE_VAL_COMPLEX(174, 59), MAKE_VAL_COMPLEX(-155, -17) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_857, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_857", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_857() */
/* END TEST 857 */

/*
 *  BEGIN TEST 858
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_858( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(57, 25), MAKE_VAL_COMPLEX(30, 46), MAKE_VAL_COMPLEX(52, 26), MAKE_VAL_COMPLEX(-11, -3), MAKE_VAL_COMPLEX(52, 0), MAKE_VAL_COMPLEX(8, -12), MAKE_VAL_COMPLEX(10, -54), MAKE_VAL_COMPLEX(-4, 32), MAKE_VAL_COMPLEX(-43, 45), MAKE_VAL_COMPLEX(24, 42), MAKE_VAL_COMPLEX(-17, 45), MAKE_VAL_COMPLEX(-43, 45), MAKE_VAL_COMPLEX(-5, 53), MAKE_VAL_COMPLEX(-16, -28), MAKE_VAL_COMPLEX(46, 22) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_858, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_858", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_858() */
/* END TEST 858 */

/*
 *  BEGIN TEST 859
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_859( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(552, -301), MAKE_VAL_COMPLEX(1389, 1712), MAKE_VAL_COMPLEX(719, -3), MAKE_VAL_COMPLEX(1056, -322), MAKE_VAL_COMPLEX(-446, 795), MAKE_VAL_COMPLEX(361, -333), MAKE_VAL_COMPLEX(805, -188), MAKE_VAL_COMPLEX(933, 1534), MAKE_VAL_COMPLEX(726, -812), MAKE_VAL_COMPLEX(-1033, 3633), MAKE_VAL_COMPLEX(-689, 181), MAKE_VAL_COMPLEX(-502, -700), MAKE_VAL_COMPLEX(-279, 2), MAKE_VAL_COMPLEX(-1268, 885), MAKE_VAL_COMPLEX(757, -619) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_859, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_859", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_859() */
/* END TEST 859 */

/*
 *  BEGIN TEST 860
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_860( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(132, -87), MAKE_VAL_COMPLEX(263, -88), MAKE_VAL_COMPLEX(185, 83), MAKE_VAL_COMPLEX(8, 36), MAKE_VAL_COMPLEX(104, 39), MAKE_VAL_COMPLEX(9, -69), MAKE_VAL_COMPLEX(-119, -122), MAKE_VAL_COMPLEX(33, 82), MAKE_VAL_COMPLEX(78, -70), MAKE_VAL_COMPLEX(209, 293), MAKE_VAL_COMPLEX(-15, 125), MAKE_VAL_COMPLEX(-26, 150), MAKE_VAL_COMPLEX(211, 12), MAKE_VAL_COMPLEX(-76, 129), MAKE_VAL_COMPLEX(133, -177) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_860, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_860", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_860() */
/* END TEST 860 */

/*
 *  BEGIN TEST 861
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_861( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-90, 27), MAKE_VAL_COMPLEX(-203, 64), MAKE_VAL_COMPLEX(-143, -137), MAKE_VAL_COMPLEX(-14, -24), MAKE_VAL_COMPLEX(-92, -99), MAKE_VAL_COMPLEX(-21, 57), MAKE_VAL_COMPLEX(59, 98), MAKE_VAL_COMPLEX(3, -70), MAKE_VAL_COMPLEX(-36, 130), MAKE_VAL_COMPLEX(-155, -311), MAKE_VAL_COMPLEX(63, -95), MAKE_VAL_COMPLEX(68, -90), MAKE_VAL_COMPLEX(-151, 6), MAKE_VAL_COMPLEX(40, -117), MAKE_VAL_COMPLEX(-97, 129) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_861, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_861", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_861() */
/* END TEST 861 */

/*
 *  BEGIN TEST 862
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_862( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-42, -89), MAKE_VAL_COMPLEX(12, -86), MAKE_VAL_COMPLEX(-35, -85), MAKE_VAL_COMPLEX(10, 15), MAKE_VAL_COMPLEX(-62, -54), MAKE_VAL_COMPLEX(-22, 6), MAKE_VAL_COMPLEX(-68, 54), MAKE_VAL_COMPLEX(38, -34), MAKE_VAL_COMPLEX(98, -9), MAKE_VAL_COMPLEX(15, -75), MAKE_VAL_COMPLEX(67, -36), MAKE_VAL_COMPLEX(98, -9), MAKE_VAL_COMPLEX(61, -58), MAKE_VAL_COMPLEX(-10, 50), MAKE_VAL_COMPLEX(-32, -74) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_862, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_862", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_862() */
/* END TEST 862 */

/*
 *  BEGIN TEST 863
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x816', LAYOUT_COLMAJ>
 *  y : <'y816', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_863( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A816_ptr, A816_ind, A816_val,
		A816_M, A816_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y816_val,
		y816_LEN * y816_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y816_LEN, y816_VW, LAYOUT_COLMAJ, y816_LEN );
	oski_value_t* y_val = array_Duplicate( x816_val,
		x816_LEN * x816_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x816_LEN, x816_VW, LAYOUT_COLMAJ, x816_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(414, 262), MAKE_VAL_COMPLEX(1003, 1294), MAKE_VAL_COMPLEX(-832, 884), MAKE_VAL_COMPLEX(-186, 217), MAKE_VAL_COMPLEX(-762, 469), MAKE_VAL_COMPLEX(383, 69), MAKE_VAL_COMPLEX(333, -576), MAKE_VAL_COMPLEX(-333, 36), MAKE_VAL_COMPLEX(1444, 407), MAKE_VAL_COMPLEX(-2066, 1332), MAKE_VAL_COMPLEX(-486, -343), MAKE_VAL_COMPLEX(-420, -205), MAKE_VAL_COMPLEX(578, 1174), MAKE_VAL_COMPLEX(-1154, -57), MAKE_VAL_COMPLEX(1229, 297) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_863, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_863", y_true, y_val, x816_LEN*x816_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_863() */
/* END TEST 863 */

#define A864_M 5    /* # of rows */
#define A864_N 5    /* # of cols */
#define A864_NNZ 5  /* # of stored non-zeros */
static oski_index_t A864_ptr[A864_M+1] = {1, 5, 5, 6, 6, 6}; /* end row pointers for A864 */
static oski_index_t A864_ind[A864_NNZ] = {2, 3, 4, 5, 5}; /* end column indices pointers for A864 */
static oski_value_t A864_val[A864_NNZ] = {MAKE_VAL_COMPLEX(1, -6), MAKE_VAL_COMPLEX(-4, -3), MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(10, -2), MAKE_VAL_COMPLEX(-9, 0)}; /* end stored non-zero values for A864 */

#define x864_LEN 5   /* vector length */
#define x864_VW  3   /* vector width */
static oski_value_t x864_val[x864_LEN*x864_VW] = {MAKE_VAL_COMPLEX(-5, 6), MAKE_VAL_COMPLEX(8, 3), MAKE_VAL_COMPLEX(5, 9), MAKE_VAL_COMPLEX(-5, -1), MAKE_VAL_COMPLEX(5, 0), MAKE_VAL_COMPLEX(10, -10), MAKE_VAL_COMPLEX(-8, 10), MAKE_VAL_COMPLEX(-10, -6), MAKE_VAL_COMPLEX(10, 10), MAKE_VAL_COMPLEX(2, -9), MAKE_VAL_COMPLEX(0, 3), MAKE_VAL_COMPLEX(-5, 4), MAKE_VAL_COMPLEX(-6, 5), MAKE_VAL_COMPLEX(-6, -4), MAKE_VAL_COMPLEX(-2, 1)}; /* data */

#define y864_LEN 5   /* vector length */
#define y864_VW  3   /* vector width */
static oski_value_t y864_val[y864_LEN*y864_VW] = {MAKE_VAL_COMPLEX(-3, -10), MAKE_VAL_COMPLEX(5, -8), MAKE_VAL_COMPLEX(0, 2), MAKE_VAL_COMPLEX(-1, 9), MAKE_VAL_COMPLEX(-1, 4), MAKE_VAL_COMPLEX(10, 10), MAKE_VAL_COMPLEX(-10, -1), MAKE_VAL_COMPLEX(-6, 4), MAKE_VAL_COMPLEX(-3, 8), MAKE_VAL_COMPLEX(8, -9), MAKE_VAL_COMPLEX(9, -2), MAKE_VAL_COMPLEX(-9, 2), MAKE_VAL_COMPLEX(-4, -9), MAKE_VAL_COMPLEX(2, -5), MAKE_VAL_COMPLEX(-5, 10)}; /* data */

/*
 *  BEGIN TEST 864
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_864( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	oski_value_t* y_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-15, 154), MAKE_VAL_COMPLEX(54, 38), MAKE_VAL_COMPLEX(-1, -113), MAKE_VAL_COMPLEX(-52, -1), MAKE_VAL_COMPLEX(-2, -66), MAKE_VAL_COMPLEX(-51, 29), MAKE_VAL_COMPLEX(123, -131), MAKE_VAL_COMPLEX(3, -72), MAKE_VAL_COMPLEX(-28, -101), MAKE_VAL_COMPLEX(169, 80), MAKE_VAL_COMPLEX(-59, -20), MAKE_VAL_COMPLEX(44, -223), MAKE_VAL_COMPLEX(121, -24), MAKE_VAL_COMPLEX(209, 97), MAKE_VAL_COMPLEX(-99, -84) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_864, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_864", y_true, y_val, y864_LEN*y864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_864() */
/* END TEST 864 */

/*
 *  BEGIN TEST 865
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_865( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	oski_value_t* y_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(188, -195), MAKE_VAL_COMPLEX(77, 55), MAKE_VAL_COMPLEX(-33, 111), MAKE_VAL_COMPLEX(-100, -25), MAKE_VAL_COMPLEX(-65, 45), MAKE_VAL_COMPLEX(-129, 131), MAKE_VAL_COMPLEX(-96, -38), MAKE_VAL_COMPLEX(-65, -34), MAKE_VAL_COMPLEX(-105, 42), MAKE_VAL_COMPLEX(-24, 65), MAKE_VAL_COMPLEX(84, 175), MAKE_VAL_COMPLEX(-69, 68), MAKE_VAL_COMPLEX(36, -35), MAKE_VAL_COMPLEX(-126, -58), MAKE_VAL_COMPLEX(-66, -11) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_865, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_865", y_true, y_val, y864_LEN*y864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_865() */
/* END TEST 865 */

/*
 *  BEGIN TEST 866
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_866( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	oski_value_t* y_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-6, -20), MAKE_VAL_COMPLEX(10, -16), MAKE_VAL_COMPLEX(0, 4), MAKE_VAL_COMPLEX(-2, 18), MAKE_VAL_COMPLEX(-2, 8), MAKE_VAL_COMPLEX(20, 20), MAKE_VAL_COMPLEX(-20, -2), MAKE_VAL_COMPLEX(-12, 8), MAKE_VAL_COMPLEX(-6, 16), MAKE_VAL_COMPLEX(16, -18), MAKE_VAL_COMPLEX(18, -4), MAKE_VAL_COMPLEX(-18, 4), MAKE_VAL_COMPLEX(-8, -18), MAKE_VAL_COMPLEX(4, -10), MAKE_VAL_COMPLEX(-10, 20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_866, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_866", y_true, y_val, y864_LEN*y864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_866() */
/* END TEST 866 */

/*
 *  BEGIN TEST 867
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_867( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	oski_value_t* y_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1552, -106), MAKE_VAL_COMPLEX(122, -119), MAKE_VAL_COMPLEX(829, -323), MAKE_VAL_COMPLEX(-80, 265), MAKE_VAL_COMPLEX(470, 15), MAKE_VAL_COMPLEX(483, 229), MAKE_VAL_COMPLEX(708, 479), MAKE_VAL_COMPLEX(192, 131), MAKE_VAL_COMPLEX(559, 7), MAKE_VAL_COMPLEX(471, 723), MAKE_VAL_COMPLEX(474, -818), MAKE_VAL_COMPLEX(1167, -259), MAKE_VAL_COMPLEX(191, 223), MAKE_VAL_COMPLEX(188, 1469), MAKE_VAL_COMPLEX(62, -244) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_867, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_867", y_true, y_val, y864_LEN*y864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_867() */
/* END TEST 867 */

/*
 *  BEGIN TEST 868
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_868( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-201, -22), MAKE_VAL_COMPLEX(-3, -62), MAKE_VAL_COMPLEX(171, 188), MAKE_VAL_COMPLEX(-89, -14), MAKE_VAL_COMPLEX(-14, -9), MAKE_VAL_COMPLEX(132, 2), MAKE_VAL_COMPLEX(-90, 149), MAKE_VAL_COMPLEX(-98, -20), MAKE_VAL_COMPLEX(58, 20), MAKE_VAL_COMPLEX(-5, 77), MAKE_VAL_COMPLEX(-136, 46), MAKE_VAL_COMPLEX(-39, -19), MAKE_VAL_COMPLEX(-25, -94), MAKE_VAL_COMPLEX(74, -185), MAKE_VAL_COMPLEX(9, -46) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_868, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_868", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_868() */
/* END TEST 868 */

/*
 *  BEGIN TEST 869
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_869( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(97, 37), MAKE_VAL_COMPLEX(56, 164), MAKE_VAL_COMPLEX(-202, -53), MAKE_VAL_COMPLEX(48, -41), MAKE_VAL_COMPLEX(64, 54), MAKE_VAL_COMPLEX(58, -12), MAKE_VAL_COMPLEX(-80, -121), MAKE_VAL_COMPLEX(52, -130), MAKE_VAL_COMPLEX(-48, 170), MAKE_VAL_COMPLEX(106, -149), MAKE_VAL_COMPLEX(109, -16), MAKE_VAL_COMPLEX(-47, 14), MAKE_VAL_COMPLEX(-80, 90), MAKE_VAL_COMPLEX(-98, 91), MAKE_VAL_COMPLEX(-38, 38) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_869, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_869", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_869() */
/* END TEST 869 */

/*
 *  BEGIN TEST 870
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_870( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(4, -78), MAKE_VAL_COMPLEX(-82, 24), MAKE_VAL_COMPLEX(-94, -42), MAKE_VAL_COMPLEX(46, -22), MAKE_VAL_COMPLEX(-40, 30), MAKE_VAL_COMPLEX(-20, 140), MAKE_VAL_COMPLEX(4, -128), MAKE_VAL_COMPLEX(116, -12), MAKE_VAL_COMPLEX(-140, -20), MAKE_VAL_COMPLEX(38, 84), MAKE_VAL_COMPLEX(-18, -24), MAKE_VAL_COMPLEX(16, -62), MAKE_VAL_COMPLEX(18, -76), MAKE_VAL_COMPLEX(72, -4), MAKE_VAL_COMPLEX(10, -20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_870, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_870", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_870() */
/* END TEST 870 */

/*
 *  BEGIN TEST 871
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_871( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1105, -255), MAKE_VAL_COMPLEX(-287, -987), MAKE_VAL_COMPLEX(1510, 848), MAKE_VAL_COMPLEX(-510, 152), MAKE_VAL_COMPLEX(-357, -287), MAKE_VAL_COMPLEX(136, 76), MAKE_VAL_COMPLEX(102, 1046), MAKE_VAL_COMPLEX(-552, 564), MAKE_VAL_COMPLEX(404, -760), MAKE_VAL_COMPLEX(-525, 971), MAKE_VAL_COMPLEX(-959, 221), MAKE_VAL_COMPLEX(105, -133), MAKE_VAL_COMPLEX(309, -739), MAKE_VAL_COMPLEX(714, -1026), MAKE_VAL_COMPLEX(213, -331) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_871, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_871", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_871() */
/* END TEST 871 */

/*
 *  BEGIN TEST 872
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_872( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-35, -49), MAKE_VAL_COMPLEX(-136, 60), MAKE_VAL_COMPLEX(-174, -21), MAKE_VAL_COMPLEX(104, -51), MAKE_VAL_COMPLEX(12, 66), MAKE_VAL_COMPLEX(-2, 172), MAKE_VAL_COMPLEX(52, -33), MAKE_VAL_COMPLEX(108, 64), MAKE_VAL_COMPLEX(-124, -90), MAKE_VAL_COMPLEX(194, 149), MAKE_VAL_COMPLEX(15, 104), MAKE_VAL_COMPLEX(-21, -90), MAKE_VAL_COMPLEX(84, -194), MAKE_VAL_COMPLEX(202, -127), MAKE_VAL_COMPLEX(26, -66) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_872, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_872", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_872() */
/* END TEST 872 */

/*
 *  BEGIN TEST 873
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_873( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(69, -4), MAKE_VAL_COMPLEX(69, -76), MAKE_VAL_COMPLEX(125, -46), MAKE_VAL_COMPLEX(-63, 54), MAKE_VAL_COMPLEX(-52, -61), MAKE_VAL_COMPLEX(-68, -82), MAKE_VAL_COMPLEX(2, -55), MAKE_VAL_COMPLEX(-22, -26), MAKE_VAL_COMPLEX(34, 20), MAKE_VAL_COMPLEX(-201, -75), MAKE_VAL_COMPLEX(-18, -128), MAKE_VAL_COMPLEX(57, 53), MAKE_VAL_COMPLEX(-41, 148), MAKE_VAL_COMPLEX(-150, 153), MAKE_VAL_COMPLEX(-11, 56) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_873, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_873", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_873() */
/* END TEST 873 */

/*
 *  BEGIN TEST 874
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_874( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-69, -27), MAKE_VAL_COMPLEX(-3, 81), MAKE_VAL_COMPLEX(-66, 72), MAKE_VAL_COMPLEX(-6, -48), MAKE_VAL_COMPLEX(15, 45), MAKE_VAL_COMPLEX(120, 60), MAKE_VAL_COMPLEX(-114, -42), MAKE_VAL_COMPLEX(24, -108), MAKE_VAL_COMPLEX(-60, 120), MAKE_VAL_COMPLEX(87, -9), MAKE_VAL_COMPLEX(-27, 9), MAKE_VAL_COMPLEX(-51, -33), MAKE_VAL_COMPLEX(-63, -39), MAKE_VAL_COMPLEX(18, -66), MAKE_VAL_COMPLEX(-15, -15) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_874, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_874", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_874() */
/* END TEST 874 */

/*
 *  BEGIN TEST 875
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_875( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(185, 147), MAKE_VAL_COMPLEX(116, 246), MAKE_VAL_COMPLEX(65, 328), MAKE_VAL_COMPLEX(-105, -282), MAKE_VAL_COMPLEX(140, -260), MAKE_VAL_COMPLEX(80, -10), MAKE_VAL_COMPLEX(539, -320), MAKE_VAL_COMPLEX(460, 88), MAKE_VAL_COMPLEX(-430, -160), MAKE_VAL_COMPLEX(309, -708), MAKE_VAL_COMPLEX(640, -189), MAKE_VAL_COMPLEX(-100, 153), MAKE_VAL_COMPLEX(-542, -370), MAKE_VAL_COMPLEX(-567, -618), MAKE_VAL_COMPLEX(-214, -88) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_875, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_875", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_875() */
/* END TEST 875 */

/*
 *  BEGIN TEST 876
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_876( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	oski_value_t* y_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-163, 242), MAKE_VAL_COMPLEX(-65, 56), MAKE_VAL_COMPLEX(98, -58), MAKE_VAL_COMPLEX(182, -75), MAKE_VAL_COMPLEX(63, -36), MAKE_VAL_COMPLEX(-3, -171), MAKE_VAL_COMPLEX(94, 32), MAKE_VAL_COMPLEX(91, 8), MAKE_VAL_COMPLEX(27, -142), MAKE_VAL_COMPLEX(51, 71), MAKE_VAL_COMPLEX(-60, -156), MAKE_VAL_COMPLEX(146, 18), MAKE_VAL_COMPLEX(0, 41), MAKE_VAL_COMPLEX(-14, -115), MAKE_VAL_COMPLEX(81, -44) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_876, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_876", y_true, y_val, y864_LEN*y864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_876() */
/* END TEST 876 */

/*
 *  BEGIN TEST 877
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_877( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	oski_value_t* y_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(134, -157), MAKE_VAL_COMPLEX(-10, -25), MAKE_VAL_COMPLEX(-88, 44), MAKE_VAL_COMPLEX(-130, 17), MAKE_VAL_COMPLEX(-36, 13), MAKE_VAL_COMPLEX(-17, 51), MAKE_VAL_COMPLEX(-29, 25), MAKE_VAL_COMPLEX(-29, -6), MAKE_VAL_COMPLEX(34, 101), MAKE_VAL_COMPLEX(-152, -48), MAKE_VAL_COMPLEX(-13, 125), MAKE_VAL_COMPLEX(-73, 13), MAKE_VAL_COMPLEX(-17, 42), MAKE_VAL_COMPLEX(-25, 140), MAKE_VAL_COMPLEX(4, -1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_877, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_877", y_true, y_val, y864_LEN*y864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_877() */
/* END TEST 877 */

/*
 *  BEGIN TEST 878
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_878( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	oski_value_t* y_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(38, -55), MAKE_VAL_COMPLEX(60, -7), MAKE_VAL_COMPLEX(-10, 8), MAKE_VAL_COMPLEX(-49, 31), MAKE_VAL_COMPLEX(-24, 11), MAKE_VAL_COMPLEX(-10, 90), MAKE_VAL_COMPLEX(-35, -54), MAKE_VAL_COMPLEX(-44, -14), MAKE_VAL_COMPLEX(-52, 17), MAKE_VAL_COMPLEX(77, 4), MAKE_VAL_COMPLEX(46, 37), MAKE_VAL_COMPLEX(-46, -37), MAKE_VAL_COMPLEX(29, -56), MAKE_VAL_COMPLEX(33, -10), MAKE_VAL_COMPLEX(-70, 15) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_878, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_878", y_true, y_val, y864_LEN*y864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_878() */
/* END TEST 878 */

/*
 *  BEGIN TEST 879
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_879( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	oski_value_t* y_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1920, -428), MAKE_VAL_COMPLEX(20, -506), MAKE_VAL_COMPLEX(-972, -230), MAKE_VAL_COMPLEX(-996, -601), MAKE_VAL_COMPLEX(-316, -60), MAKE_VAL_COMPLEX(-123, 259), MAKE_VAL_COMPLEX(-597, 180), MAKE_VAL_COMPLEX(-212, -63), MAKE_VAL_COMPLEX(-326, 1242), MAKE_VAL_COMPLEX(-924, -1761), MAKE_VAL_COMPLEX(-866, 695), MAKE_VAL_COMPLEX(-788, -193), MAKE_VAL_COMPLEX(-678, 144), MAKE_VAL_COMPLEX(-1246, 820), MAKE_VAL_COMPLEX(154, 290) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_879, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_879", y_true, y_val, y864_LEN*y864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_879() */
/* END TEST 879 */

/*
 *  BEGIN TEST 880
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_880( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(92, -86), MAKE_VAL_COMPLEX(198, -123), MAKE_VAL_COMPLEX(-73, 204), MAKE_VAL_COMPLEX(-52, 22), MAKE_VAL_COMPLEX(50, -71), MAKE_VAL_COMPLEX(-72, -74), MAKE_VAL_COMPLEX(47, 27), MAKE_VAL_COMPLEX(-92, 67), MAKE_VAL_COMPLEX(19, -178), MAKE_VAL_COMPLEX(-111, 149), MAKE_VAL_COMPLEX(9, -198), MAKE_VAL_COMPLEX(-82, -63), MAKE_VAL_COMPLEX(-19, -94), MAKE_VAL_COMPLEX(178, 67), MAKE_VAL_COMPLEX(125, 59) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_880, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_880", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_880() */
/* END TEST 880 */

/*
 *  BEGIN TEST 881
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_881( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(13, 82), MAKE_VAL_COMPLEX(-240, 16), MAKE_VAL_COMPLEX(118, -335), MAKE_VAL_COMPLEX(87, 37), MAKE_VAL_COMPLEX(-95, 21), MAKE_VAL_COMPLEX(-118, 64), MAKE_VAL_COMPLEX(125, -37), MAKE_VAL_COMPLEX(122, 87), MAKE_VAL_COMPLEX(-9, -12), MAKE_VAL_COMPLEX(3, -88), MAKE_VAL_COMPLEX(21, 171), MAKE_VAL_COMPLEX(167, 77), MAKE_VAL_COMPLEX(123, 109), MAKE_VAL_COMPLEX(-164, 29), MAKE_VAL_COMPLEX(-97, -48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_881, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_881", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_881() */
/* END TEST 881 */

/*
 *  BEGIN TEST 882
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_882( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(36, -31), MAKE_VAL_COMPLEX(-45, -26), MAKE_VAL_COMPLEX(-21, -59), MAKE_VAL_COMPLEX(29, 11), MAKE_VAL_COMPLEX(-30, -5), MAKE_VAL_COMPLEX(-70, 50), MAKE_VAL_COMPLEX(58, -52), MAKE_VAL_COMPLEX(54, 46), MAKE_VAL_COMPLEX(-50, -70), MAKE_VAL_COMPLEX(-21, 52), MAKE_VAL_COMPLEX(3, -18), MAKE_VAL_COMPLEX(34, -19), MAKE_VAL_COMPLEX(41, -24), MAKE_VAL_COMPLEX(32, 30), MAKE_VAL_COMPLEX(13, -4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_882, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_882", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_882() */
/* END TEST 882 */

/*
 *  BEGIN TEST 883
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_883( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(379, -528), MAKE_VAL_COMPLEX(1238, -555), MAKE_VAL_COMPLEX(-535, 1431), MAKE_VAL_COMPLEX(-359, 29), MAKE_VAL_COMPLEX(365, -336), MAKE_VAL_COMPLEX(-122, -394), MAKE_VAL_COMPLEX(-2, 148), MAKE_VAL_COMPLEX(-574, 144), MAKE_VAL_COMPLEX(64, -758), MAKE_VAL_COMPLEX(-478, 813), MAKE_VAL_COMPLEX(0, -1149), MAKE_VAL_COMPLEX(-629, -416), MAKE_VAL_COMPLEX(-282, -607), MAKE_VAL_COMPLEX(1062, 242), MAKE_VAL_COMPLEX(706, 331) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_883, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_883", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_883() */
/* END TEST 883 */

/*
 *  BEGIN TEST 884
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_884( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-85, -33), MAKE_VAL_COMPLEX(103, -246), MAKE_VAL_COMPLEX(7, 180), MAKE_VAL_COMPLEX(38, -46), MAKE_VAL_COMPLEX(-35, 74), MAKE_VAL_COMPLEX(72, 14), MAKE_VAL_COMPLEX(1, 39), MAKE_VAL_COMPLEX(-192, 35), MAKE_VAL_COMPLEX(51, -4), MAKE_VAL_COMPLEX(148, 36), MAKE_VAL_COMPLEX(-206, 23), MAKE_VAL_COMPLEX(-97, 120), MAKE_VAL_COMPLEX(-46, -101), MAKE_VAL_COMPLEX(116, 49), MAKE_VAL_COMPLEX(114, 92) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_884, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_884", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_884() */
/* END TEST 884 */

/*
 *  BEGIN TEST 885
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_885( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(15, -5), MAKE_VAL_COMPLEX(-117, 332), MAKE_VAL_COMPLEX(-87, -112), MAKE_VAL_COMPLEX(-38, -6), MAKE_VAL_COMPLEX(45, -24), MAKE_VAL_COMPLEX(48, 66), MAKE_VAL_COMPLEX(-117, -99), MAKE_VAL_COMPLEX(232, -147), MAKE_VAL_COMPLEX(-131, 124), MAKE_VAL_COMPLEX(-54, -34), MAKE_VAL_COMPLEX(176, -17), MAKE_VAL_COMPLEX(47, -162), MAKE_VAL_COMPLEX(-16, 51), MAKE_VAL_COMPLEX(-88, -117), MAKE_VAL_COMPLEX(-128, -110) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_885, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_885", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_885() */
/* END TEST 885 */

/*
 *  BEGIN TEST 886
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_886( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-46, 43), MAKE_VAL_COMPLEX(61, 32), MAKE_VAL_COMPLEX(31, 77), MAKE_VAL_COMPLEX(-39, -13), MAKE_VAL_COMPLEX(40, 5), MAKE_VAL_COMPLEX(90, -70), MAKE_VAL_COMPLEX(-74, 72), MAKE_VAL_COMPLEX(-74, -58), MAKE_VAL_COMPLEX(70, 90), MAKE_VAL_COMPLEX(25, -70), MAKE_VAL_COMPLEX(-3, 24), MAKE_VAL_COMPLEX(-44, 27), MAKE_VAL_COMPLEX(-53, 34), MAKE_VAL_COMPLEX(-44, -38), MAKE_VAL_COMPLEX(-17, 6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_886, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_886", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_886() */
/* END TEST 886 */

/*
 *  BEGIN TEST 887
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_ROWMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_887( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_ROWMAJ, x864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(242, 369), MAKE_VAL_COMPLEX(-584, 1239), MAKE_VAL_COMPLEX(220, -666), MAKE_VAL_COMPLEX(-351, 143), MAKE_VAL_COMPLEX(369, -310), MAKE_VAL_COMPLEX(-56, -332), MAKE_VAL_COMPLEX(-208, 4), MAKE_VAL_COMPLEX(785, -177), MAKE_VAL_COMPLEX(-69, 259), MAKE_VAL_COMPLEX(-611, -534), MAKE_VAL_COMPLEX(1038, 163), MAKE_VAL_COMPLEX(465, -432), MAKE_VAL_COMPLEX(-40, 641), MAKE_VAL_COMPLEX(-655, -487), MAKE_VAL_COMPLEX(-531, -560) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_887, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_887", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_887() */
/* END TEST 887 */

/*
 *  BEGIN TEST 888
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_888( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	oski_value_t* y_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(140, -3), MAKE_VAL_COMPLEX(74, -155), MAKE_VAL_COMPLEX(58, 141), MAKE_VAL_COMPLEX(34, 2), MAKE_VAL_COMPLEX(-58, -77), MAKE_VAL_COMPLEX(-37, -23), MAKE_VAL_COMPLEX(39, -6), MAKE_VAL_COMPLEX(-78, 63), MAKE_VAL_COMPLEX(25, -51), MAKE_VAL_COMPLEX(82, 33), MAKE_VAL_COMPLEX(-224, 27), MAKE_VAL_COMPLEX(58, -51), MAKE_VAL_COMPLEX(-53, 21), MAKE_VAL_COMPLEX(169, -53), MAKE_VAL_COMPLEX(68, -59) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_888, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_888", y_true, y_val, y864_LEN*y864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_888() */
/* END TEST 888 */

/*
 *  BEGIN TEST 889
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_889( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	oski_value_t* y_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-157, 128), MAKE_VAL_COMPLEX(-169, 218), MAKE_VAL_COMPLEX(-48, -163), MAKE_VAL_COMPLEX(22, -96), MAKE_VAL_COMPLEX(89, 38), MAKE_VAL_COMPLEX(-23, -137), MAKE_VAL_COMPLEX(66, 67), MAKE_VAL_COMPLEX(164, -77), MAKE_VAL_COMPLEX(48, -22), MAKE_VAL_COMPLEX(-215, 26), MAKE_VAL_COMPLEX(115, -50), MAKE_VAL_COMPLEX(51, 74), MAKE_VAL_COMPLEX(52, 98), MAKE_VAL_COMPLEX(-216, 98), MAKE_VAL_COMPLEX(37, -26) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_889, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_889", y_true, y_val, y864_LEN*y864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_889() */
/* END TEST 889 */

/*
 *  BEGIN TEST 890
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_890( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	oski_value_t* y_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(77, 75), MAKE_VAL_COMPLEX(-5, 97), MAKE_VAL_COMPLEX(-10, -18), MAKE_VAL_COMPLEX(-36, -86), MAKE_VAL_COMPLEX(-11, -41), MAKE_VAL_COMPLEX(-140, -40), MAKE_VAL_COMPLEX(95, -41), MAKE_VAL_COMPLEX(34, -66), MAKE_VAL_COMPLEX(-13, -87), MAKE_VAL_COMPLEX(-27, 121), MAKE_VAL_COMPLEX(-71, 63), MAKE_VAL_COMPLEX(71, -63), MAKE_VAL_COMPLEX(81, 61), MAKE_VAL_COMPLEX(7, 55), MAKE_VAL_COMPLEX(-5, -115) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_890, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_890", y_true, y_val, y864_LEN*y864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_890() */
/* END TEST 890 */

/*
 *  BEGIN TEST 891
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_891( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	oski_value_t* y_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(592, -1298), MAKE_VAL_COMPLEX(-844, -2266), MAKE_VAL_COMPLEX(1676, 720), MAKE_VAL_COMPLEX(638, 38), MAKE_VAL_COMPLEX(-940, 2), MAKE_VAL_COMPLEX(220, -28), MAKE_VAL_COMPLEX(-58, 32), MAKE_VAL_COMPLEX(-120, 1492), MAKE_VAL_COMPLEX(44, -258), MAKE_VAL_COMPLEX(766, -980), MAKE_VAL_COMPLEX(-1408, 1574), MAKE_VAL_COMPLEX(-112, -438), MAKE_VAL_COMPLEX(-746, 508), MAKE_VAL_COMPLEX(774, -2000), MAKE_VAL_COMPLEX(352, -526) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_891, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_891", y_true, y_val, y864_LEN*y864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_891() */
/* END TEST 891 */

/*
 *  BEGIN TEST 892
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_892( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-202, -33), MAKE_VAL_COMPLEX(-42, 80), MAKE_VAL_COMPLEX(-21, 204), MAKE_VAL_COMPLEX(-81, 86), MAKE_VAL_COMPLEX(61, -64), MAKE_VAL_COMPLEX(74, -110), MAKE_VAL_COMPLEX(-144, -32), MAKE_VAL_COMPLEX(-82, -24), MAKE_VAL_COMPLEX(-131, 138), MAKE_VAL_COMPLEX(154, -164), MAKE_VAL_COMPLEX(168, 124), MAKE_VAL_COMPLEX(-27, 2), MAKE_VAL_COMPLEX(-12, -101), MAKE_VAL_COMPLEX(5, -74), MAKE_VAL_COMPLEX(10, -49) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_892, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_892", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_892() */
/* END TEST 892 */

/*
 *  BEGIN TEST 893
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_893( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(161, 70), MAKE_VAL_COMPLEX(95, -51), MAKE_VAL_COMPLEX(47, -136), MAKE_VAL_COMPLEX(47, -98), MAKE_VAL_COMPLEX(-26, 69), MAKE_VAL_COMPLEX(6, 50), MAKE_VAL_COMPLEX(78, 94), MAKE_VAL_COMPLEX(18, -28), MAKE_VAL_COMPLEX(191, -58), MAKE_VAL_COMPLEX(-131, 103), MAKE_VAL_COMPLEX(-171, -103), MAKE_VAL_COMPLEX(-12, 21), MAKE_VAL_COMPLEX(-35, 130), MAKE_VAL_COMPLEX(-43, 40), MAKE_VAL_COMPLEX(-25, 54) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_893, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_893", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_893() */
/* END TEST 893 */

/*
 *  BEGIN TEST 894
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_894( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-9, -99), MAKE_VAL_COMPLEX(-99, 45), MAKE_VAL_COMPLEX(-126, -36), MAKE_VAL_COMPLEX(54, -36), MAKE_VAL_COMPLEX(-45, 45), MAKE_VAL_COMPLEX(0, 180), MAKE_VAL_COMPLEX(-18, -162), MAKE_VAL_COMPLEX(144, -36), MAKE_VAL_COMPLEX(-180, 0), MAKE_VAL_COMPLEX(63, 99), MAKE_VAL_COMPLEX(-27, -27), MAKE_VAL_COMPLEX(9, -81), MAKE_VAL_COMPLEX(9, -99), MAKE_VAL_COMPLEX(90, -18), MAKE_VAL_COMPLEX(9, -27) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_894, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_894", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_894() */
/* END TEST 894 */

/*
 *  BEGIN TEST 895
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_895( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1076, -379), MAKE_VAL_COMPLEX(-509, 43), MAKE_VAL_COMPLEX(-24, 1029), MAKE_VAL_COMPLEX(-626, 925), MAKE_VAL_COMPLEX(387, -736), MAKE_VAL_COMPLEX(-198, -1046), MAKE_VAL_COMPLEX(-286, -274), MAKE_VAL_COMPLEX(-644, 494), MAKE_VAL_COMPLEX(-946, 73), MAKE_VAL_COMPLEX(817, -975), MAKE_VAL_COMPLEX(1391, 1061), MAKE_VAL_COMPLEX(175, 142), MAKE_VAL_COMPLEX(488, -643), MAKE_VAL_COMPLEX(84, -113), MAKE_VAL_COMPLEX(252, -301) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_895, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_895", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_895() */
/* END TEST 895 */

/*
 *  BEGIN TEST 896
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_896( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-12, 21), MAKE_VAL_COMPLEX(32, -28), MAKE_VAL_COMPLEX(53, 84), MAKE_VAL_COMPLEX(153, 64), MAKE_VAL_COMPLEX(61, -106), MAKE_VAL_COMPLEX(-78, 50), MAKE_VAL_COMPLEX(76, -4), MAKE_VAL_COMPLEX(10, 114), MAKE_VAL_COMPLEX(9, 98), MAKE_VAL_COMPLEX(116, -84), MAKE_VAL_COMPLEX(-62, 2), MAKE_VAL_COMPLEX(13, 0), MAKE_VAL_COMPLEX(54, -105), MAKE_VAL_COMPLEX(-11, -6), MAKE_VAL_COMPLEX(24, -45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_896, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_896", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_896() */
/* END TEST 896 */

/*
 *  BEGIN TEST 897
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_897( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-61, -31), MAKE_VAL_COMPLEX(-16, 107), MAKE_VAL_COMPLEX(-100, 1), MAKE_VAL_COMPLEX(-170, -109), MAKE_VAL_COMPLEX(-36, 146), MAKE_VAL_COMPLEX(208, -20), MAKE_VAL_COMPLEX(-196, -10), MAKE_VAL_COMPLEX(-12, -224), MAKE_VAL_COMPLEX(-39, 32), MAKE_VAL_COMPLEX(-34, 55), MAKE_VAL_COMPLEX(38, 13), MAKE_VAL_COMPLEX(-70, -20), MAKE_VAL_COMPLEX(-124, 82), MAKE_VAL_COMPLEX(13, -62), MAKE_VAL_COMPLEX(-42, 34) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_897, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_897", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_897() */
/* END TEST 897 */

/*
 *  BEGIN TEST 898
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_898( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(21, 48), MAKE_VAL_COMPLEX(42, -39), MAKE_VAL_COMPLEX(69, -3), MAKE_VAL_COMPLEX(-21, 27), MAKE_VAL_COMPLEX(15, -30), MAKE_VAL_COMPLEX(-30, -90), MAKE_VAL_COMPLEX(36, 78), MAKE_VAL_COMPLEX(-66, 42), MAKE_VAL_COMPLEX(90, -30), MAKE_VAL_COMPLEX(-48, -39), MAKE_VAL_COMPLEX(18, 9), MAKE_VAL_COMPLEX(9, 42), MAKE_VAL_COMPLEX(12, 51), MAKE_VAL_COMPLEX(-42, 24), MAKE_VAL_COMPLEX(0, 15) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_898, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_898", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_898() */
/* END TEST 898 */

/*
 *  BEGIN TEST 899
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_899( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_ROWMAJ, y864_VW );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(171, -146), MAKE_VAL_COMPLEX(-314, 61), MAKE_VAL_COMPLEX(-177, -701), MAKE_VAL_COMPLEX(-747, -671), MAKE_VAL_COMPLEX(-623, 504), MAKE_VAL_COMPLEX(458, -74), MAKE_VAL_COMPLEX(-372, -202), MAKE_VAL_COMPLEX(246, -630), MAKE_VAL_COMPLEX(72, -716), MAKE_VAL_COMPLEX(-900, 349), MAKE_VAL_COMPLEX(382, 85), MAKE_VAL_COMPLEX(-25, -52), MAKE_VAL_COMPLEX(-470, 489), MAKE_VAL_COMPLEX(100, 106), MAKE_VAL_COMPLEX(-214, 217) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_899, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_899", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_899() */
/* END TEST 899 */

/*
 *  BEGIN TEST 900
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_900( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	oski_value_t* y_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(179, -91), MAKE_VAL_COMPLEX(110, 50), MAKE_VAL_COMPLEX(-16, 6), MAKE_VAL_COMPLEX(39, 9), MAKE_VAL_COMPLEX(-109, -6), MAKE_VAL_COMPLEX(46, -92), MAKE_VAL_COMPLEX(-81, -133), MAKE_VAL_COMPLEX(-144, 55), MAKE_VAL_COMPLEX(-255, 13), MAKE_VAL_COMPLEX(259, -46), MAKE_VAL_COMPLEX(101, 206), MAKE_VAL_COMPLEX(-28, -50), MAKE_VAL_COMPLEX(72, -71), MAKE_VAL_COMPLEX(65, -35), MAKE_VAL_COMPLEX(-27, -19) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_900, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_900", y_true, y_val, y864_LEN*y864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_900() */
/* END TEST 900 */

/*
 *  BEGIN TEST 901
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_901( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	oski_value_t* y_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-193, 8), MAKE_VAL_COMPLEX(-62, -109), MAKE_VAL_COMPLEX(14, 10), MAKE_VAL_COMPLEX(-56, 62), MAKE_VAL_COMPLEX(97, 37), MAKE_VAL_COMPLEX(24, 182), MAKE_VAL_COMPLEX(2, 115), MAKE_VAL_COMPLEX(92, -29), MAKE_VAL_COMPLEX(223, 48), MAKE_VAL_COMPLEX(-186, -18), MAKE_VAL_COMPLEX(-27, -213), MAKE_VAL_COMPLEX(-46, 57), MAKE_VAL_COMPLEX(-95, -5), MAKE_VAL_COMPLEX(-44, -3), MAKE_VAL_COMPLEX(-23, 94) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_901, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_901", y_true, y_val, y864_LEN*y864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_901() */
/* END TEST 901 */

/*
 *  BEGIN TEST 902
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_902( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	oski_value_t* y_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(8, -46), MAKE_VAL_COMPLEX(36, -22), MAKE_VAL_COMPLEX(-4, 8), MAKE_VAL_COMPLEX(-22, 34), MAKE_VAL_COMPLEX(-12, 14), MAKE_VAL_COMPLEX(20, 60), MAKE_VAL_COMPLEX(-38, -24), MAKE_VAL_COMPLEX(-32, 4), MAKE_VAL_COMPLEX(-28, 26), MAKE_VAL_COMPLEX(50, -20), MAKE_VAL_COMPLEX(40, 10), MAKE_VAL_COMPLEX(-40, -10), MAKE_VAL_COMPLEX(2, -44), MAKE_VAL_COMPLEX(18, -16), MAKE_VAL_COMPLEX(-40, 30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_902, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_902", y_true, y_val, y864_LEN*y864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_902() */
/* END TEST 902 */

/*
 *  BEGIN TEST 903
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_903( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	oski_value_t* y_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(523, -99), MAKE_VAL_COMPLEX(145, 227), MAKE_VAL_COMPLEX(-14, -14), MAKE_VAL_COMPLEX(400, -110), MAKE_VAL_COMPLEX(-317, -75), MAKE_VAL_COMPLEX(244, -808), MAKE_VAL_COMPLEX(-159, -263), MAKE_VAL_COMPLEX(-362, 294), MAKE_VAL_COMPLEX(-763, -25), MAKE_VAL_COMPLEX(659, -213), MAKE_VAL_COMPLEX(183, 637), MAKE_VAL_COMPLEX(109, -13), MAKE_VAL_COMPLEX(139, -13), MAKE_VAL_COMPLEX(97, -95), MAKE_VAL_COMPLEX(237, -141) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_903, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_903", y_true, y_val, y864_LEN*y864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_903() */
/* END TEST 903 */

/*
 *  BEGIN TEST 904
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_904( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(41, -159), MAKE_VAL_COMPLEX(-119, 41), MAKE_VAL_COMPLEX(-97, 5), MAKE_VAL_COMPLEX(-39, 109), MAKE_VAL_COMPLEX(-76, -73), MAKE_VAL_COMPLEX(222, 35), MAKE_VAL_COMPLEX(30, -157), MAKE_VAL_COMPLEX(4, -25), MAKE_VAL_COMPLEX(-123, -172), MAKE_VAL_COMPLEX(235, 94), MAKE_VAL_COMPLEX(-120, 227), MAKE_VAL_COMPLEX(-15, -109), MAKE_VAL_COMPLEX(-6, -185), MAKE_VAL_COMPLEX(-50, -97), MAKE_VAL_COMPLEX(120, 34) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_904, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_904", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_904() */
/* END TEST 904 */

/*
 *  BEGIN TEST 905
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_905( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-135, 101), MAKE_VAL_COMPLEX(93, 77), MAKE_VAL_COMPLEX(-19, 83), MAKE_VAL_COMPLEX(43, -181), MAKE_VAL_COMPLEX(86, 143), MAKE_VAL_COMPLEX(-62, 85), MAKE_VAL_COMPLEX(-186, 65), MAKE_VAL_COMPLEX(60, -127), MAKE_VAL_COMPLEX(3, 332), MAKE_VAL_COMPLEX(-105, -84), MAKE_VAL_COMPLEX(78, -221), MAKE_VAL_COMPLEX(-51, 47), MAKE_VAL_COMPLEX(-76, 111), MAKE_VAL_COMPLEX(94, 5), MAKE_VAL_COMPLEX(-138, -60) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_905, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_905", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_905() */
/* END TEST 905 */

/*
 *  BEGIN TEST 906
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_906( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(28, -58), MAKE_VAL_COMPLEX(-70, -8), MAKE_VAL_COMPLEX(-58, -62), MAKE_VAL_COMPLEX(42, -2), MAKE_VAL_COMPLEX(-40, 10), MAKE_VAL_COMPLEX(-60, 100), MAKE_VAL_COMPLEX(44, -96), MAKE_VAL_COMPLEX(92, 28), MAKE_VAL_COMPLEX(-100, -60), MAKE_VAL_COMPLEX(2, 76), MAKE_VAL_COMPLEX(-6, -24), MAKE_VAL_COMPLEX(32, -42), MAKE_VAL_COMPLEX(38, -52), MAKE_VAL_COMPLEX(56, 20), MAKE_VAL_COMPLEX(14, -12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_906, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_906", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_906() */
/* END TEST 906 */

/*
 *  BEGIN TEST 907
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_907( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-396, -940), MAKE_VAL_COMPLEX(-346, 458), MAKE_VAL_COMPLEX(-32, 142), MAKE_VAL_COMPLEX(646, 1156), MAKE_VAL_COMPLEX(-1031, -143), MAKE_VAL_COMPLEX(455, -1719), MAKE_VAL_COMPLEX(-105, -779), MAKE_VAL_COMPLEX(-247, 635), MAKE_VAL_COMPLEX(-1479, -899), MAKE_VAL_COMPLEX(1219, -1047), MAKE_VAL_COMPLEX(1175, 1891), MAKE_VAL_COMPLEX(-450, -242), MAKE_VAL_COMPLEX(-847, -651), MAKE_VAL_COMPLEX(-1009, 349), MAKE_VAL_COMPLEX(956, -574) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_907, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_907", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_907() */
/* END TEST 907 */

/*
 *  BEGIN TEST 908
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_908( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-34, -21), MAKE_VAL_COMPLEX(93, -70), MAKE_VAL_COMPLEX(106, -10), MAKE_VAL_COMPLEX(114, 102), MAKE_VAL_COMPLEX(-1, -145), MAKE_VAL_COMPLEX(36, -339), MAKE_VAL_COMPLEX(-26, 129), MAKE_VAL_COMPLEX(-198, 113), MAKE_VAL_COMPLEX(-133, -22), MAKE_VAL_COMPLEX(101, 47), MAKE_VAL_COMPLEX(6, 136), MAKE_VAL_COMPLEX(22, 87), MAKE_VAL_COMPLEX(24, -24), MAKE_VAL_COMPLEX(-100, 127), MAKE_VAL_COMPLEX(126, 101) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_908, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_908", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_908() */
/* END TEST 908 */

/*
 *  BEGIN TEST 909
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_909( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(10, 1), MAKE_VAL_COMPLEX(-105, 102), MAKE_VAL_COMPLEX(-142, 30), MAKE_VAL_COMPLEX(-110, -122), MAKE_VAL_COMPLEX(1, 165), MAKE_VAL_COMPLEX(4, 379), MAKE_VAL_COMPLEX(-14, -161), MAKE_VAL_COMPLEX(222, -153), MAKE_VAL_COMPLEX(93, 62), MAKE_VAL_COMPLEX(-65, -39), MAKE_VAL_COMPLEX(-18, -136), MAKE_VAL_COMPLEX(-38, -107), MAKE_VAL_COMPLEX(-44, 0), MAKE_VAL_COMPLEX(116, -151), MAKE_VAL_COMPLEX(-130, -109) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_909, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_909", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_909() */
/* END TEST 909 */

/*
 *  BEGIN TEST 910
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_910( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-49, -51), MAKE_VAL_COMPLEX(-35, 69), MAKE_VAL_COMPLEX(-86, 36), MAKE_VAL_COMPLEX(14, -44), MAKE_VAL_COMPLEX(-5, 45), MAKE_VAL_COMPLEX(80, 100), MAKE_VAL_COMPLEX(-82, -82), MAKE_VAL_COMPLEX(64, -84), MAKE_VAL_COMPLEX(-100, 80), MAKE_VAL_COMPLEX(79, 27), MAKE_VAL_COMPLEX(-27, -3), MAKE_VAL_COMPLEX(-31, -49), MAKE_VAL_COMPLEX(-39, -59), MAKE_VAL_COMPLEX(42, -50), MAKE_VAL_COMPLEX(-7, -19) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_910, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_910", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_910() */
/* END TEST 910 */

/*
 *  BEGIN TEST 911
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x864', LAYOUT_COLMAJ>
 *  y : <'y864', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_911( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A864_ptr, A864_ind, A864_val,
		A864_M, A864_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y864_val,
		y864_LEN * y864_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y864_LEN, y864_VW, LAYOUT_COLMAJ, y864_LEN );
	oski_value_t* y_val = array_Duplicate( x864_val,
		x864_LEN * x864_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x864_LEN, x864_VW, LAYOUT_COLMAJ, x864_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-270, -372), MAKE_VAL_COMPLEX(338, -48), MAKE_VAL_COMPLEX(196, 94), MAKE_VAL_COMPLEX(564, 482), MAKE_VAL_COMPLEX(75, -571), MAKE_VAL_COMPLEX(699, -1219), MAKE_VAL_COMPLEX(-465, 241), MAKE_VAL_COMPLEX(-787, 135), MAKE_VAL_COMPLEX(-1083, -63), MAKE_VAL_COMPLEX(711, 515), MAKE_VAL_COMPLEX(-199, 647), MAKE_VAL_COMPLEX(-36, 250), MAKE_VAL_COMPLEX(67, -347), MAKE_VAL_COMPLEX(-425, 401), MAKE_VAL_COMPLEX(524, 556) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_911, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_911", y_true, y_val, x864_LEN*x864_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_911() */
/* END TEST 911 */

#define A912_M 5    /* # of rows */
#define A912_N 5    /* # of cols */
#define A912_NNZ 6  /* # of stored non-zeros */
static oski_index_t A912_ptr[A912_M+1] = {0, 2, 4, 5, 6, 6}; /* end row pointers for A912 */
static oski_index_t A912_ind[A912_NNZ] = {1, 3, 3, 4, 3, 4}; /* end column indices pointers for A912 */
static oski_value_t A912_val[A912_NNZ] = {MAKE_VAL_COMPLEX(-6, -10), MAKE_VAL_COMPLEX(-2, -7), MAKE_VAL_COMPLEX(-5, 7), MAKE_VAL_COMPLEX(-3, -3), MAKE_VAL_COMPLEX(-10, 6), MAKE_VAL_COMPLEX(0, 2)}; /* end stored non-zero values for A912 */

#define x912_LEN 5   /* vector length */
#define x912_VW  3   /* vector width */
static oski_value_t x912_val[x912_LEN*x912_VW] = {MAKE_VAL_COMPLEX(-4, 7), MAKE_VAL_COMPLEX(-7, -10), MAKE_VAL_COMPLEX(9, 10), MAKE_VAL_COMPLEX(-4, 5), MAKE_VAL_COMPLEX(-9, -9), MAKE_VAL_COMPLEX(-9, 0), MAKE_VAL_COMPLEX(0, 10), MAKE_VAL_COMPLEX(-10, 6), MAKE_VAL_COMPLEX(1, -1), MAKE_VAL_COMPLEX(-3, -5), MAKE_VAL_COMPLEX(7, 0), MAKE_VAL_COMPLEX(1, 8), MAKE_VAL_COMPLEX(10, -6), MAKE_VAL_COMPLEX(-5, 3), MAKE_VAL_COMPLEX(-8, -6)}; /* data */

#define y912_LEN 5   /* vector length */
#define y912_VW  3   /* vector width */
static oski_value_t y912_val[y912_LEN*y912_VW] = {MAKE_VAL_COMPLEX(-1, 10), MAKE_VAL_COMPLEX(3, 7), MAKE_VAL_COMPLEX(10, -5), MAKE_VAL_COMPLEX(3, 6), MAKE_VAL_COMPLEX(3, 5), MAKE_VAL_COMPLEX(-9, -7), MAKE_VAL_COMPLEX(3, -7), MAKE_VAL_COMPLEX(-1, -10), MAKE_VAL_COMPLEX(-1, 8), MAKE_VAL_COMPLEX(3, 8), MAKE_VAL_COMPLEX(7, -5), MAKE_VAL_COMPLEX(3, -5), MAKE_VAL_COMPLEX(-10, 10), MAKE_VAL_COMPLEX(-4, -7), MAKE_VAL_COMPLEX(-1, -4)}; /* data */

/*
 *  BEGIN TEST 912
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_912( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	oski_value_t* y_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-27, 21), MAKE_VAL_COMPLEX(-112, 92), MAKE_VAL_COMPLEX(132, 157), MAKE_VAL_COMPLEX(44, 4), MAKE_VAL_COMPLEX(-119, 187), MAKE_VAL_COMPLEX(49, -190), MAKE_VAL_COMPLEX(103, 77), MAKE_VAL_COMPLEX(-8, 61), MAKE_VAL_COMPLEX(-111, -98), MAKE_VAL_COMPLEX(-71, -119), MAKE_VAL_COMPLEX(138, -20), MAKE_VAL_COMPLEX(135, -107), MAKE_VAL_COMPLEX(-3, -105), MAKE_VAL_COMPLEX(52, 57), MAKE_VAL_COMPLEX(33, 24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_912, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_912", y_true, y_val, y912_LEN*y912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_912() */
/* END TEST 912 */

/*
 *  BEGIN TEST 913
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_913( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	oski_value_t* y_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-42, -38), MAKE_VAL_COMPLEX(60, -78), MAKE_VAL_COMPLEX(-107, -82), MAKE_VAL_COMPLEX(-89, 11), MAKE_VAL_COMPLEX(81, -171), MAKE_VAL_COMPLEX(9, 134), MAKE_VAL_COMPLEX(-57, -49), MAKE_VAL_COMPLEX(79, -58), MAKE_VAL_COMPLEX(56, 83), MAKE_VAL_COMPLEX(12, 132), MAKE_VAL_COMPLEX(-110, 74), MAKE_VAL_COMPLEX(-103, 133), MAKE_VAL_COMPLEX(-57, 25), MAKE_VAL_COMPLEX(1, -78), MAKE_VAL_COMPLEX(-4, -27) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_913, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_913", y_true, y_val, y912_LEN*y912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_913() */
/* END TEST 913 */

/*
 *  BEGIN TEST 914
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_914( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	oski_value_t* y_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(26, -58), MAKE_VAL_COMPLEX(-4, -48), MAKE_VAL_COMPLEX(-70, 10), MAKE_VAL_COMPLEX(-6, -42), MAKE_VAL_COMPLEX(-8, -36), MAKE_VAL_COMPLEX(40, 60), MAKE_VAL_COMPLEX(-32, 36), MAKE_VAL_COMPLEX(-14, 62), MAKE_VAL_COMPLEX(22, -46), MAKE_VAL_COMPLEX(-2, -54), MAKE_VAL_COMPLEX(-52, 16), MAKE_VAL_COMPLEX(-28, 24), MAKE_VAL_COMPLEX(80, -40), MAKE_VAL_COMPLEX(10, 50), MAKE_VAL_COMPLEX(-2, 26) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_914, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_914", y_true, y_val, y912_LEN*y912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_914() */
/* END TEST 914 */

/*
 *  BEGIN TEST 915
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_915( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	oski_value_t* y_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-27, -353), MAKE_VAL_COMPLEX(668, -328), MAKE_VAL_COMPLEX(-461, -918), MAKE_VAL_COMPLEX(-504, -240), MAKE_VAL_COMPLEX(1051, -821), MAKE_VAL_COMPLEX(-416, 937), MAKE_VAL_COMPLEX(-275, -491), MAKE_VAL_COMPLEX(539, -90), MAKE_VAL_COMPLEX(176, 669), MAKE_VAL_COMPLEX(-239, 787), MAKE_VAL_COMPLEX(-944, 42), MAKE_VAL_COMPLEX(-1075, 461), MAKE_VAL_COMPLEX(-297, 79), MAKE_VAL_COMPLEX(167, -414), MAKE_VAL_COMPLEX(14, -159) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, -3);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_915, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_915", y_true, y_val, y912_LEN*y912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_915() */
/* END TEST 915 */

/*
 *  BEGIN TEST 916
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_916( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(39, -67), MAKE_VAL_COMPLEX(-16, -180), MAKE_VAL_COMPLEX(-33, 212), MAKE_VAL_COMPLEX(54, -49), MAKE_VAL_COMPLEX(28, -50), MAKE_VAL_COMPLEX(-162, -52), MAKE_VAL_COMPLEX(-115, -9), MAKE_VAL_COMPLEX(-125, 78), MAKE_VAL_COMPLEX(9, 74), MAKE_VAL_COMPLEX(12, 12), MAKE_VAL_COMPLEX(126, 70), MAKE_VAL_COMPLEX(-14, -129), MAKE_VAL_COMPLEX(67, -7), MAKE_VAL_COMPLEX(-30, -19), MAKE_VAL_COMPLEX(-9, -18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_916, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_916", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_916() */
/* END TEST 916 */

/*
 *  BEGIN TEST 917
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_917( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-96, 53), MAKE_VAL_COMPLEX(72, 111), MAKE_VAL_COMPLEX(-19, -129), MAKE_VAL_COMPLEX(-97, 31), MAKE_VAL_COMPLEX(17, -31), MAKE_VAL_COMPLEX(144, -11), MAKE_VAL_COMPLEX(45, 29), MAKE_VAL_COMPLEX(63, -136), MAKE_VAL_COMPLEX(0, -69), MAKE_VAL_COMPLEX(17, -43), MAKE_VAL_COMPLEX(-112, -21), MAKE_VAL_COMPLEX(-40, 152), MAKE_VAL_COMPLEX(-5, 65), MAKE_VAL_COMPLEX(-1, -10), MAKE_VAL_COMPLEX(35, -50) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_917, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_917", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_917() */
/* END TEST 917 */

/*
 *  BEGIN TEST 918
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_918( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(35, 20), MAKE_VAL_COMPLEX(-50, 35), MAKE_VAL_COMPLEX(50, -45), MAKE_VAL_COMPLEX(25, 20), MAKE_VAL_COMPLEX(-45, 45), MAKE_VAL_COMPLEX(0, 45), MAKE_VAL_COMPLEX(50, 0), MAKE_VAL_COMPLEX(30, 50), MAKE_VAL_COMPLEX(-5, -5), MAKE_VAL_COMPLEX(-25, 15), MAKE_VAL_COMPLEX(0, -35), MAKE_VAL_COMPLEX(40, -5), MAKE_VAL_COMPLEX(-30, -50), MAKE_VAL_COMPLEX(15, 25), MAKE_VAL_COMPLEX(-30, 40) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_918, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_918", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_918() */
/* END TEST 918 */

/*
 *  BEGIN TEST 919
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_919( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(16, -1088), MAKE_VAL_COMPLEX(-676, -440), MAKE_VAL_COMPLEX(454, 942), MAKE_VAL_COMPLEX(230, -950), MAKE_VAL_COMPLEX(480, 0), MAKE_VAL_COMPLEX(-744, 588), MAKE_VAL_COMPLEX(-824, 36), MAKE_VAL_COMPLEX(270, 778), MAKE_VAL_COMPLEX(446, 458), MAKE_VAL_COMPLEX(364, 276), MAKE_VAL_COMPLEX(756, -280), MAKE_VAL_COMPLEX(-982, -1162), MAKE_VAL_COMPLEX(-192, -4), MAKE_VAL_COMPLEX(-18, -154), MAKE_VAL_COMPLEX(366, 242) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_919, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_919", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_919() */
/* END TEST 919 */

/*
 *  BEGIN TEST 920
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_920( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-211, 34), MAKE_VAL_COMPLEX(-33, -46), MAKE_VAL_COMPLEX(166, 102), MAKE_VAL_COMPLEX(-108, -174), MAKE_VAL_COMPLEX(-131, -139), MAKE_VAL_COMPLEX(-117, 82), MAKE_VAL_COMPLEX(-39, -35), MAKE_VAL_COMPLEX(-207, -20), MAKE_VAL_COMPLEX(-48, 39), MAKE_VAL_COMPLEX(-81, -51), MAKE_VAL_COMPLEX(-43, 112), MAKE_VAL_COMPLEX(23, 163), MAKE_VAL_COMPLEX(85, 13), MAKE_VAL_COMPLEX(-91, -36), MAKE_VAL_COMPLEX(17, -106) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_920, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_920", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_920() */
/* END TEST 920 */

/*
 *  BEGIN TEST 921
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_921( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(160, -26), MAKE_VAL_COMPLEX(55, -29), MAKE_VAL_COMPLEX(-180, -17), MAKE_VAL_COMPLEX(67, 174), MAKE_VAL_COMPLEX(140, 58), MAKE_VAL_COMPLEX(81, -127), MAKE_VAL_COMPLEX(-11, 75), MAKE_VAL_COMPLEX(137, -6), MAKE_VAL_COMPLEX(57, -38), MAKE_VAL_COMPLEX(94, 16), MAKE_VAL_COMPLEX(71, -77), MAKE_VAL_COMPLEX(-59, -126), MAKE_VAL_COMPLEX(-15, 13), MAKE_VAL_COMPLEX(56, 23), MAKE_VAL_COMPLEX(-19, 42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_921, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_921", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_921() */
/* END TEST 921 */

/*
 *  BEGIN TEST 922
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_922( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-84, 17), MAKE_VAL_COMPLEX(31, -126), MAKE_VAL_COMPLEX(-17, 142), MAKE_VAL_COMPLEX(-68, 3), MAKE_VAL_COMPLEX(9, -135), MAKE_VAL_COMPLEX(-63, -72), MAKE_VAL_COMPLEX(-80, 70), MAKE_VAL_COMPLEX(-118, -38), MAKE_VAL_COMPLEX(15, 1), MAKE_VAL_COMPLEX(19, -59), MAKE_VAL_COMPLEX(49, 56), MAKE_VAL_COMPLEX(-57, 64), MAKE_VAL_COMPLEX(118, 38), MAKE_VAL_COMPLEX(-59, -19), MAKE_VAL_COMPLEX(-8, -106) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_922, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_922", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_922() */
/* END TEST 922 */

/*
 *  BEGIN TEST 923
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_923( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-395, -1152), MAKE_VAL_COMPLEX(-532, -249), MAKE_VAL_COMPLEX(414, 1303), MAKE_VAL_COMPLEX(1353, -808), MAKE_VAL_COMPLEX(13, -1065), MAKE_VAL_COMPLEX(-1178, -223), MAKE_VAL_COMPLEX(812, -62), MAKE_VAL_COMPLEX(-208, -894), MAKE_VAL_COMPLEX(-437, -399), MAKE_VAL_COMPLEX(-155, -727), MAKE_VAL_COMPLEX(-758, -547), MAKE_VAL_COMPLEX(-738, 705), MAKE_VAL_COMPLEX(20, -96), MAKE_VAL_COMPLEX(129, -399), MAKE_VAL_COMPLEX(260, 220) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_923, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_923", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_923() */
/* END TEST 923 */

/*
 *  BEGIN TEST 924
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_924( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	oski_value_t* y_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-23, -19), MAKE_VAL_COMPLEX(25, -26), MAKE_VAL_COMPLEX(35, 142), MAKE_VAL_COMPLEX(-69, -139), MAKE_VAL_COMPLEX(-6, -24), MAKE_VAL_COMPLEX(46, 64), MAKE_VAL_COMPLEX(-47, 239), MAKE_VAL_COMPLEX(-4, 101), MAKE_VAL_COMPLEX(67, -134), MAKE_VAL_COMPLEX(-79, 44), MAKE_VAL_COMPLEX(110, 156), MAKE_VAL_COMPLEX(-1, -90), MAKE_VAL_COMPLEX(-67, -205), MAKE_VAL_COMPLEX(179, -124), MAKE_VAL_COMPLEX(37, 40) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_924, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_924", y_true, y_val, y912_LEN*y912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_924() */
/* END TEST 924 */

/*
 *  BEGIN TEST 925
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_925( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	oski_value_t* y_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-98, 17), MAKE_VAL_COMPLEX(-106, 69), MAKE_VAL_COMPLEX(35, -27), MAKE_VAL_COMPLEX(0, 181), MAKE_VAL_COMPLEX(-51, 65), MAKE_VAL_COMPLEX(29, -179), MAKE_VAL_COMPLEX(134, -210), MAKE_VAL_COMPLEX(123, -123), MAKE_VAL_COMPLEX(-164, 130), MAKE_VAL_COMPLEX(-14, 0), MAKE_VAL_COMPLEX(-43, -77), MAKE_VAL_COMPLEX(64, 121), MAKE_VAL_COMPLEX(-63, 95), MAKE_VAL_COMPLEX(-99, 69), MAKE_VAL_COMPLEX(10, -56) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_925, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_925", y_true, y_val, y912_LEN*y912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_925() */
/* END TEST 925 */

/*
 *  BEGIN TEST 926
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_926( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	oski_value_t* y_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(2, 81), MAKE_VAL_COMPLEX(31, 53), MAKE_VAL_COMPLEX(75, -50), MAKE_VAL_COMPLEX(30, 45), MAKE_VAL_COMPLEX(29, 37), MAKE_VAL_COMPLEX(-79, -47), MAKE_VAL_COMPLEX(17, -59), MAKE_VAL_COMPLEX(-18, -79), MAKE_VAL_COMPLEX(0, 65), MAKE_VAL_COMPLEX(32, 61), MAKE_VAL_COMPLEX(51, -47), MAKE_VAL_COMPLEX(19, -43), MAKE_VAL_COMPLEX(-70, 90), MAKE_VAL_COMPLEX(-39, -52), MAKE_VAL_COMPLEX(-12, -31) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_926, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_926", y_true, y_val, y912_LEN*y912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_926() */
/* END TEST 926 */

/*
 *  BEGIN TEST 927
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_927( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	oski_value_t* y_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-607, -127), MAKE_VAL_COMPLEX(-696, 596), MAKE_VAL_COMPLEX(-727, 9), MAKE_VAL_COMPLEX(825, 941), MAKE_VAL_COMPLEX(-278, 404), MAKE_VAL_COMPLEX(-70, -1020), MAKE_VAL_COMPLEX(-436, -1886), MAKE_VAL_COMPLEX(341, -873), MAKE_VAL_COMPLEX(-455, 1343), MAKE_VAL_COMPLEX(-375, -593), MAKE_VAL_COMPLEX(-1396, 74), MAKE_VAL_COMPLEX(988, 1012), MAKE_VAL_COMPLEX(906, 278), MAKE_VAL_COMPLEX(-31, 1737), MAKE_VAL_COMPLEX(-167, -169) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_927, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_927", y_true, y_val, y912_LEN*y912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_927() */
/* END TEST 927 */

/*
 *  BEGIN TEST 928
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_928( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(94, -26), MAKE_VAL_COMPLEX(-138, -39), MAKE_VAL_COMPLEX(11, 62), MAKE_VAL_COMPLEX(55, -21), MAKE_VAL_COMPLEX(-157, 27), MAKE_VAL_COMPLEX(-92, 22), MAKE_VAL_COMPLEX(-6, 23), MAKE_VAL_COMPLEX(-79, -4), MAKE_VAL_COMPLEX(74, 44), MAKE_VAL_COMPLEX(-115, 75), MAKE_VAL_COMPLEX(105, 206), MAKE_VAL_COMPLEX(62, -111), MAKE_VAL_COMPLEX(43, -111), MAKE_VAL_COMPLEX(-63, 64), MAKE_VAL_COMPLEX(-97, -8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_928, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_928", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_928() */
/* END TEST 928 */

/*
 *  BEGIN TEST 929
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_929( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-111, 7), MAKE_VAL_COMPLEX(175, 28), MAKE_VAL_COMPLEX(-50, -45), MAKE_VAL_COMPLEX(-66, 4), MAKE_VAL_COMPLEX(193, -45), MAKE_VAL_COMPLEX(101, -49), MAKE_VAL_COMPLEX(-24, -33), MAKE_VAL_COMPLEX(71, -32), MAKE_VAL_COMPLEX(-72, -40), MAKE_VAL_COMPLEX(133, -79), MAKE_VAL_COMPLEX(-112, -185), MAKE_VAL_COMPLEX(-87, 106), MAKE_VAL_COMPLEX(-35, 147), MAKE_VAL_COMPLEX(59, -82), MAKE_VAL_COMPLEX(123, -10) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_929, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_929", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_929() */
/* END TEST 929 */

/*
 *  BEGIN TEST 930
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_930( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-92, 31), MAKE_VAL_COMPLEX(17, -146), MAKE_VAL_COMPLEX(1, 162), MAKE_VAL_COMPLEX(-76, 13), MAKE_VAL_COMPLEX(-9, -153), MAKE_VAL_COMPLEX(-81, -72), MAKE_VAL_COMPLEX(-80, 90), MAKE_VAL_COMPLEX(-138, -26), MAKE_VAL_COMPLEX(17, -1), MAKE_VAL_COMPLEX(13, -69), MAKE_VAL_COMPLEX(63, 56), MAKE_VAL_COMPLEX(-55, 80), MAKE_VAL_COMPLEX(138, 26), MAKE_VAL_COMPLEX(-69, -13), MAKE_VAL_COMPLEX(-24, -118) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_930, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_930", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_930() */
/* END TEST 930 */

/*
 *  BEGIN TEST 931
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_931( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(458, -392), MAKE_VAL_COMPLEX(-139, -155), MAKE_VAL_COMPLEX(-591, 195), MAKE_VAL_COMPLEX(310, -334), MAKE_VAL_COMPLEX(-254, 68), MAKE_VAL_COMPLEX(-158, -216), MAKE_VAL_COMPLEX(-313, -251), MAKE_VAL_COMPLEX(-187, -575), MAKE_VAL_COMPLEX(316, 356), MAKE_VAL_COMPLEX(-412, 362), MAKE_VAL_COMPLEX(67, 1317), MAKE_VAL_COMPLEX(177, -755), MAKE_VAL_COMPLEX(122, -36), MAKE_VAL_COMPLEX(-277, 19), MAKE_VAL_COMPLEX(-33, -129) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_931, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_931", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_931() */
/* END TEST 931 */

/*
 *  BEGIN TEST 932
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_932( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-107, -13), MAKE_VAL_COMPLEX(-3, 96), MAKE_VAL_COMPLEX(80, -17), MAKE_VAL_COMPLEX(-62, -132), MAKE_VAL_COMPLEX(173, -49), MAKE_VAL_COMPLEX(33, 185), MAKE_VAL_COMPLEX(36, -123), MAKE_VAL_COMPLEX(109, -88), MAKE_VAL_COMPLEX(-18, 106), MAKE_VAL_COMPLEX(-149, -67), MAKE_VAL_COMPLEX(-60, 59), MAKE_VAL_COMPLEX(131, -16), MAKE_VAL_COMPLEX(-67, -9), MAKE_VAL_COMPLEX(57, 38), MAKE_VAL_COMPLEX(11, 68) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_932, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_932", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_932() */
/* END TEST 932 */

/*
 *  BEGIN TEST 933
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_933( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(67, 18), MAKE_VAL_COMPLEX(22, -154), MAKE_VAL_COMPLEX(-93, 83), MAKE_VAL_COMPLEX(30, 131), MAKE_VAL_COMPLEX(-164, -14), MAKE_VAL_COMPLEX(-60, -221), MAKE_VAL_COMPLEX(-76, 153), MAKE_VAL_COMPLEX(-163, 66), MAKE_VAL_COMPLEX(25, -105), MAKE_VAL_COMPLEX(160, 40), MAKE_VAL_COMPLEX(81, -31), MAKE_VAL_COMPLEX(-160, 44), MAKE_VAL_COMPLEX(121, 31), MAKE_VAL_COMPLEX(-84, -49), MAKE_VAL_COMPLEX(-11, -118) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_933, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_933", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_933() */
/* END TEST 933 */

/*
 *  BEGIN TEST 934
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_934( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(7, 4), MAKE_VAL_COMPLEX(-10, 7), MAKE_VAL_COMPLEX(10, -9), MAKE_VAL_COMPLEX(5, 4), MAKE_VAL_COMPLEX(-9, 9), MAKE_VAL_COMPLEX(0, 9), MAKE_VAL_COMPLEX(10, 0), MAKE_VAL_COMPLEX(6, 10), MAKE_VAL_COMPLEX(-1, -1), MAKE_VAL_COMPLEX(-5, 3), MAKE_VAL_COMPLEX(0, -7), MAKE_VAL_COMPLEX(8, -1), MAKE_VAL_COMPLEX(-6, -10), MAKE_VAL_COMPLEX(3, 5), MAKE_VAL_COMPLEX(-6, 8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_934, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_934", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_934() */
/* END TEST 934 */

/*
 *  BEGIN TEST 935
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_ROWMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_935( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_ROWMAJ, x912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(784, -871), MAKE_VAL_COMPLEX(-179, -308), MAKE_VAL_COMPLEX(-831, 320), MAKE_VAL_COMPLEX(1252, 187), MAKE_VAL_COMPLEX(-303, 1557), MAKE_VAL_COMPLEX(-975, -1020), MAKE_VAL_COMPLEX(402, 584), MAKE_VAL_COMPLEX(172, 840), MAKE_VAL_COMPLEX(-553, -695), MAKE_VAL_COMPLEX(1467, -343), MAKE_VAL_COMPLEX(-253, -630), MAKE_VAL_COMPLEX(-823, 598), MAKE_VAL_COMPLEX(158, -6), MAKE_VAL_COMPLEX(-421, -57), MAKE_VAL_COMPLEX(-106, -216) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_935, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_935", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_935() */
/* END TEST 935 */

/*
 *  BEGIN TEST 936
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_936( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	oski_value_t* y_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-44, 102), MAKE_VAL_COMPLEX(46, -91), MAKE_VAL_COMPLEX(77, 26), MAKE_VAL_COMPLEX(39, -32), MAKE_VAL_COMPLEX(20, 120), MAKE_VAL_COMPLEX(35, -62), MAKE_VAL_COMPLEX(25, -20), MAKE_VAL_COMPLEX(21, 69), MAKE_VAL_COMPLEX(23, -109), MAKE_VAL_COMPLEX(-13, -75), MAKE_VAL_COMPLEX(3, -68), MAKE_VAL_COMPLEX(-132, 59), MAKE_VAL_COMPLEX(-8, -46), MAKE_VAL_COMPLEX(70, -10), MAKE_VAL_COMPLEX(24, -26) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_936, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_936", y_true, y_val, y912_LEN*y912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_936() */
/* END TEST 936 */

/*
 *  BEGIN TEST 937
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_937( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	oski_value_t* y_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-37, -201), MAKE_VAL_COMPLEX(-136, 55), MAKE_VAL_COMPLEX(-122, 109), MAKE_VAL_COMPLEX(-120, 5), MAKE_VAL_COMPLEX(-92, -138), MAKE_VAL_COMPLEX(109, 44), MAKE_VAL_COMPLEX(11, 110), MAKE_VAL_COMPLEX(78, 12), MAKE_VAL_COMPLEX(-86, 28), MAKE_VAL_COMPLEX(-86, 30), MAKE_VAL_COMPLEX(-21, 176), MAKE_VAL_COMPLEX(150, 13), MAKE_VAL_COMPLEX(8, -134), MAKE_VAL_COMPLEX(29, 37), MAKE_VAL_COMPLEX(21, 53) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_937, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_937", y_true, y_val, y912_LEN*y912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_937() */
/* END TEST 937 */

/*
 *  BEGIN TEST 938
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_938( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	oski_value_t* y_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(27, -68), MAKE_VAL_COMPLEX(-7, -55), MAKE_VAL_COMPLEX(-80, 15), MAKE_VAL_COMPLEX(-9, -48), MAKE_VAL_COMPLEX(-11, -41), MAKE_VAL_COMPLEX(49, 67), MAKE_VAL_COMPLEX(-35, 43), MAKE_VAL_COMPLEX(-13, 72), MAKE_VAL_COMPLEX(23, -54), MAKE_VAL_COMPLEX(-5, -62), MAKE_VAL_COMPLEX(-59, 21), MAKE_VAL_COMPLEX(-31, 29), MAKE_VAL_COMPLEX(90, -50), MAKE_VAL_COMPLEX(14, 57), MAKE_VAL_COMPLEX(-1, 30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_938, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_938", y_true, y_val, y912_LEN*y912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_938() */
/* END TEST 938 */

/*
 *  BEGIN TEST 939
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_939( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	oski_value_t* y_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(341, 920), MAKE_VAL_COMPLEX(228, -502), MAKE_VAL_COMPLEX(573, -506), MAKE_VAL_COMPLEX(342, -211), MAKE_VAL_COMPLEX(675, 581), MAKE_VAL_COMPLEX(-432, -338), MAKE_VAL_COMPLEX(-56, -434), MAKE_VAL_COMPLEX(7, 60), MAKE_VAL_COMPLEX(-13, -386), MAKE_VAL_COMPLEX(9, -220), MAKE_VAL_COMPLEX(-154, -670), MAKE_VAL_COMPLEX(-548, 482), MAKE_VAL_COMPLEX(-148, 324), MAKE_VAL_COMPLEX(24, -323), MAKE_VAL_COMPLEX(-95, -274) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, -3);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_939, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_939", y_true, y_val, y912_LEN*y912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_939() */
/* END TEST 939 */

/*
 *  BEGIN TEST 940
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_940( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(58, -149), MAKE_VAL_COMPLEX(196, -72), MAKE_VAL_COMPLEX(-181, -46), MAKE_VAL_COMPLEX(-9, 6), MAKE_VAL_COMPLEX(82, -38), MAKE_VAL_COMPLEX(22, -155), MAKE_VAL_COMPLEX(-24, 0), MAKE_VAL_COMPLEX(-43, -12), MAKE_VAL_COMPLEX(87, 53), MAKE_VAL_COMPLEX(59, -18), MAKE_VAL_COMPLEX(-75, 165), MAKE_VAL_COMPLEX(-168, -41), MAKE_VAL_COMPLEX(1, 170), MAKE_VAL_COMPLEX(11, -228), MAKE_VAL_COMPLEX(89, 18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_940, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_940", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_940() */
/* END TEST 940 */

/*
 *  BEGIN TEST 941
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_941( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-76, 148), MAKE_VAL_COMPLEX(-183, 48), MAKE_VAL_COMPLEX(170, 74), MAKE_VAL_COMPLEX(-5, -9), MAKE_VAL_COMPLEX(-73, 11), MAKE_VAL_COMPLEX(-31, 137), MAKE_VAL_COMPLEX(4, 10), MAKE_VAL_COMPLEX(21, -2), MAKE_VAL_COMPLEX(-84, -52), MAKE_VAL_COMPLEX(-52, 7), MAKE_VAL_COMPLEX(82, -151), MAKE_VAL_COMPLEX(153, 51), MAKE_VAL_COMPLEX(21, -156), MAKE_VAL_COMPLEX(-22, 221), MAKE_VAL_COMPLEX(-85, -40) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_941, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_941", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_941() */
/* END TEST 941 */

/*
 *  BEGIN TEST 942
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_942( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-7, -4), MAKE_VAL_COMPLEX(10, -7), MAKE_VAL_COMPLEX(-10, 9), MAKE_VAL_COMPLEX(-5, -4), MAKE_VAL_COMPLEX(9, -9), MAKE_VAL_COMPLEX(0, -9), MAKE_VAL_COMPLEX(-10, 0), MAKE_VAL_COMPLEX(-6, -10), MAKE_VAL_COMPLEX(1, 1), MAKE_VAL_COMPLEX(5, -3), MAKE_VAL_COMPLEX(0, 7), MAKE_VAL_COMPLEX(-8, 1), MAKE_VAL_COMPLEX(6, 10), MAKE_VAL_COMPLEX(-3, -5), MAKE_VAL_COMPLEX(6, -8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_942, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_942", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_942() */
/* END TEST 942 */

/*
 *  BEGIN TEST 943
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_943( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-516, -822), MAKE_VAL_COMPLEX(-216, -877), MAKE_VAL_COMPLEX(-656, 462), MAKE_VAL_COMPLEX(361, 33), MAKE_VAL_COMPLEX(-93, 25), MAKE_VAL_COMPLEX(-699, -131), MAKE_VAL_COMPLEX(322, -212), MAKE_VAL_COMPLEX(432, 419), MAKE_VAL_COMPLEX(470, -502), MAKE_VAL_COMPLEX(-79, -148), MAKE_VAL_COMPLEX(739, 596), MAKE_VAL_COMPLEX(-365, 753), MAKE_VAL_COMPLEX(590, 191), MAKE_VAL_COMPLEX(-1273, -462), MAKE_VAL_COMPLEX(386, -63) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_943, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_943", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_943() */
/* END TEST 943 */

/*
 *  BEGIN TEST 944
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_944( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-70, 31), MAKE_VAL_COMPLEX(-126, -102), MAKE_VAL_COMPLEX(93, -206), MAKE_VAL_COMPLEX(-37, 28), MAKE_VAL_COMPLEX(-84, 94), MAKE_VAL_COMPLEX(-26, 147), MAKE_VAL_COMPLEX(-32, -42), MAKE_VAL_COMPLEX(-27, 76), MAKE_VAL_COMPLEX(-103, 63), MAKE_VAL_COMPLEX(-77, 10), MAKE_VAL_COMPLEX(149, -85), MAKE_VAL_COMPLEX(16, 111), MAKE_VAL_COMPLEX(-135, -38), MAKE_VAL_COMPLEX(101, 140), MAKE_VAL_COMPLEX(-1, 68) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_944, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_944", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_944() */
/* END TEST 944 */

/*
 *  BEGIN TEST 945
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_945( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(170, -11), MAKE_VAL_COMPLEX(34, 226), MAKE_VAL_COMPLEX(-9, 58), MAKE_VAL_COMPLEX(113, 0), MAKE_VAL_COMPLEX(12, 50), MAKE_VAL_COMPLEX(62, -39), MAKE_VAL_COMPLEX(152, 2), MAKE_VAL_COMPLEX(139, 20), MAKE_VAL_COMPLEX(87, -71), MAKE_VAL_COMPLEX(29, 46), MAKE_VAL_COMPLEX(-177, 1), MAKE_VAL_COMPLEX(76, -155), MAKE_VAL_COMPLEX(23, -58), MAKE_VAL_COMPLEX(-45, -92), MAKE_VAL_COMPLEX(-39, 52) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_945, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_945", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_945() */
/* END TEST 945 */

/*
 *  BEGIN TEST 946
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_946( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(39, 13), MAKE_VAL_COMPLEX(-43, 45), MAKE_VAL_COMPLEX(41, -55), MAKE_VAL_COMPLEX(29, 15), MAKE_VAL_COMPLEX(-36, 54), MAKE_VAL_COMPLEX(9, 45), MAKE_VAL_COMPLEX(50, -10), MAKE_VAL_COMPLEX(40, 44), MAKE_VAL_COMPLEX(-6, -4), MAKE_VAL_COMPLEX(-22, 20), MAKE_VAL_COMPLEX(-7, -35), MAKE_VAL_COMPLEX(39, -13), MAKE_VAL_COMPLEX(-40, -44), MAKE_VAL_COMPLEX(20, 22), MAKE_VAL_COMPLEX(-22, 46) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_946, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_946", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_946() */
/* END TEST 946 */

/*
 *  BEGIN TEST 947
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_947( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_ROWMAJ, y912_VW );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(631, -1375), MAKE_VAL_COMPLEX(1994, 524), MAKE_VAL_COMPLEX(747, 671), MAKE_VAL_COMPLEX(445, -867), MAKE_VAL_COMPLEX(240, -128), MAKE_VAL_COMPLEX(-311, -753), MAKE_VAL_COMPLEX(778, -1058), MAKE_VAL_COMPLEX(575, -971), MAKE_VAL_COMPLEX(-174, -1168), MAKE_VAL_COMPLEX(469, -181), MAKE_VAL_COMPLEX(-666, 1598), MAKE_VAL_COMPLEX(-985, -1219), MAKE_VAL_COMPLEX(-107, -677), MAKE_VAL_COMPLEX(-1203, 35), MAKE_VAL_COMPLEX(33, 431) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 9);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_947, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_947", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_947() */
/* END TEST 947 */

/*
 *  BEGIN TEST 948
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_948( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	oski_value_t* y_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(68, 93), MAKE_VAL_COMPLEX(107, -84), MAKE_VAL_COMPLEX(-91, -109), MAKE_VAL_COMPLEX(53, -110), MAKE_VAL_COMPLEX(-9, -25), MAKE_VAL_COMPLEX(89, 56), MAKE_VAL_COMPLEX(-27, 161), MAKE_VAL_COMPLEX(-87, 100), MAKE_VAL_COMPLEX(94, -162), MAKE_VAL_COMPLEX(72, -113), MAKE_VAL_COMPLEX(23, -50), MAKE_VAL_COMPLEX(-92, -59), MAKE_VAL_COMPLEX(192, -56), MAKE_VAL_COMPLEX(-160, 106), MAKE_VAL_COMPLEX(-18, -7) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_948, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_948", y_true, y_val, y912_LEN*y912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_948() */
/* END TEST 948 */

/*
 *  BEGIN TEST 949
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_949( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	oski_value_t* y_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-19, -78), MAKE_VAL_COMPLEX(-69, 76), MAKE_VAL_COMPLEX(76, 54), MAKE_VAL_COMPLEX(-20, 101), MAKE_VAL_COMPLEX(37, 15), MAKE_VAL_COMPLEX(-133, -18), MAKE_VAL_COMPLEX(-5, -183), MAKE_VAL_COMPLEX(36, -105), MAKE_VAL_COMPLEX(-55, 175), MAKE_VAL_COMPLEX(-29, 106), MAKE_VAL_COMPLEX(-41, 10), MAKE_VAL_COMPLEX(70, 39), MAKE_VAL_COMPLEX(-152, 116), MAKE_VAL_COMPLEX(121, -93), MAKE_VAL_COMPLEX(-3, 8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_949, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_949", y_true, y_val, y912_LEN*y912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_949() */
/* END TEST 949 */

/*
 *  BEGIN TEST 950
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_950( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	oski_value_t* y_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(12, 82), MAKE_VAL_COMPLEX(38, 50), MAKE_VAL_COMPLEX(70, -60), MAKE_VAL_COMPLEX(36, 42), MAKE_VAL_COMPLEX(34, 34), MAKE_VAL_COMPLEX(-86, -38), MAKE_VAL_COMPLEX(10, -62), MAKE_VAL_COMPLEX(-28, -78), MAKE_VAL_COMPLEX(8, 66), MAKE_VAL_COMPLEX(40, 58), MAKE_VAL_COMPLEX(46, -54), MAKE_VAL_COMPLEX(14, -46), MAKE_VAL_COMPLEX(-60, 100), MAKE_VAL_COMPLEX(-46, -48), MAKE_VAL_COMPLEX(-16, -30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_950, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_950", y_true, y_val, y912_LEN*y912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_950() */
/* END TEST 950 */

/*
 *  BEGIN TEST 951
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_951( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	oski_value_t* y_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(955, -1056), MAKE_VAL_COMPLEX(-552, -344), MAKE_VAL_COMPLEX(-543, 348), MAKE_VAL_COMPLEX(-399, 124), MAKE_VAL_COMPLEX(365, -145), MAKE_VAL_COMPLEX(-864, 130), MAKE_VAL_COMPLEX(281, -1165), MAKE_VAL_COMPLEX(175, -10), MAKE_VAL_COMPLEX(-692, 631), MAKE_VAL_COMPLEX(-357, 18), MAKE_VAL_COMPLEX(-998, -350), MAKE_VAL_COMPLEX(-174, 668), MAKE_VAL_COMPLEX(-514, 222), MAKE_VAL_COMPLEX(1049, 550), MAKE_VAL_COMPLEX(-273, 298) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_951, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_951", y_true, y_val, y912_LEN*y912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_951() */
/* END TEST 951 */

/*
 *  BEGIN TEST 952
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_952( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(124, -111), MAKE_VAL_COMPLEX(56, -15), MAKE_VAL_COMPLEX(-62, -114), MAKE_VAL_COMPLEX(-38, 87), MAKE_VAL_COMPLEX(12, 44), MAKE_VAL_COMPLEX(-3, 23), MAKE_VAL_COMPLEX(-19, 5), MAKE_VAL_COMPLEX(19, -90), MAKE_VAL_COMPLEX(49, 242), MAKE_VAL_COMPLEX(-46, 47), MAKE_VAL_COMPLEX(-130, 16), MAKE_VAL_COMPLEX(-9, -58), MAKE_VAL_COMPLEX(14, 50), MAKE_VAL_COMPLEX(44, -159), MAKE_VAL_COMPLEX(3, 42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_952, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_952", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_952() */
/* END TEST 952 */

/*
 *  BEGIN TEST 953
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_953( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-31, 127), MAKE_VAL_COMPLEX(-138, 132), MAKE_VAL_COMPLEX(136, -25), MAKE_VAL_COMPLEX(109, -63), MAKE_VAL_COMPLEX(-75, 91), MAKE_VAL_COMPLEX(39, 76), MAKE_VAL_COMPLEX(129, -45), MAKE_VAL_COMPLEX(87, 176), MAKE_VAL_COMPLEX(-64, -249), MAKE_VAL_COMPLEX(3, 6), MAKE_VAL_COMPLEX(102, -93), MAKE_VAL_COMPLEX(93, 15), MAKE_VAL_COMPLEX(-120, -136), MAKE_VAL_COMPLEX(9, 202), MAKE_VAL_COMPLEX(-37, 70) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_953, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_953", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_953() */
/* END TEST 953 */

/*
 *  BEGIN TEST 954
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_954( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(64, -47), MAKE_VAL_COMPLEX(23, 118), MAKE_VAL_COMPLEX(-41, -126), MAKE_VAL_COMPLEX(56, -29), MAKE_VAL_COMPLEX(45, 117), MAKE_VAL_COMPLEX(81, 36), MAKE_VAL_COMPLEX(40, -90), MAKE_VAL_COMPLEX(114, -14), MAKE_VAL_COMPLEX(-13, 5), MAKE_VAL_COMPLEX(7, 57), MAKE_VAL_COMPLEX(-63, -28), MAKE_VAL_COMPLEX(23, -76), MAKE_VAL_COMPLEX(-114, 14), MAKE_VAL_COMPLEX(57, -7), MAKE_VAL_COMPLEX(48, 86) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_954, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_954", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_954() */
/* END TEST 954 */

/*
 *  BEGIN TEST 955
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_955( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-4, 1011), MAKE_VAL_COMPLEX(-4, 801), MAKE_VAL_COMPLEX(603, -63), MAKE_VAL_COMPLEX(-28, -905), MAKE_VAL_COMPLEX(50, 192), MAKE_VAL_COMPLEX(272, -126), MAKE_VAL_COMPLEX(119, -535), MAKE_VAL_COMPLEX(764, 419), MAKE_VAL_COMPLEX(-1548, -1169), MAKE_VAL_COMPLEX(149, -298), MAKE_VAL_COMPLEX(413, -767), MAKE_VAL_COMPLEX(327, -28), MAKE_VAL_COMPLEX(-762, -14), MAKE_VAL_COMPLEX(745, 1066), MAKE_VAL_COMPLEX(82, 33) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_955, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_955", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_955() */
/* END TEST 955 */

/*
 *  BEGIN TEST 956
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_956( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-114, 48), MAKE_VAL_COMPLEX(-159, -115), MAKE_VAL_COMPLEX(93, -98), MAKE_VAL_COMPLEX(-138, -58), MAKE_VAL_COMPLEX(-87, 5), MAKE_VAL_COMPLEX(-38, 87), MAKE_VAL_COMPLEX(205, -73), MAKE_VAL_COMPLEX(57, -16), MAKE_VAL_COMPLEX(-34, 65), MAKE_VAL_COMPLEX(-3, 40), MAKE_VAL_COMPLEX(117, 6), MAKE_VAL_COMPLEX(40, 186), MAKE_VAL_COMPLEX(-12, 36), MAKE_VAL_COMPLEX(119, 52), MAKE_VAL_COMPLEX(-63, 50) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_956, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_956", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_956() */
/* END TEST 956 */

/*
 *  BEGIN TEST 957
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_957( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(162, -67), MAKE_VAL_COMPLEX(154, 193), MAKE_VAL_COMPLEX(-98, 12), MAKE_VAL_COMPLEX(178, 49), MAKE_VAL_COMPLEX(96, 76), MAKE_VAL_COMPLEX(83, -51), MAKE_VAL_COMPLEX(-165, 23), MAKE_VAL_COMPLEX(17, 26), MAKE_VAL_COMPLEX(25, -64), MAKE_VAL_COMPLEX(-2, -3), MAKE_VAL_COMPLEX(-152, -34), MAKE_VAL_COMPLEX(-13, -230), MAKE_VAL_COMPLEX(-62, -46), MAKE_VAL_COMPLEX(-82, -47), MAKE_VAL_COMPLEX(79, 12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_957, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_957", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_957() */
/* END TEST 957 */

/*
 *  BEGIN TEST 958
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_958( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-28, 49), MAKE_VAL_COMPLEX(-49, -70), MAKE_VAL_COMPLEX(63, 70), MAKE_VAL_COMPLEX(-28, 35), MAKE_VAL_COMPLEX(-63, -63), MAKE_VAL_COMPLEX(-63, 0), MAKE_VAL_COMPLEX(0, 70), MAKE_VAL_COMPLEX(-70, 42), MAKE_VAL_COMPLEX(7, -7), MAKE_VAL_COMPLEX(-21, -35), MAKE_VAL_COMPLEX(49, 0), MAKE_VAL_COMPLEX(7, 56), MAKE_VAL_COMPLEX(70, -42), MAKE_VAL_COMPLEX(-35, 21), MAKE_VAL_COMPLEX(-56, -42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_958, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_958", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_958() */
/* END TEST 958 */

/*
 *  BEGIN TEST 959
 *  Value type: complex
 *  A : <MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x912', LAYOUT_COLMAJ>
 *  y : <'y912', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_959( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A912_ptr, A912_ind, A912_val,
		A912_M, A912_N, SHARE_INPUTMAT,
		5, MAT_SYMM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y912_val,
		y912_LEN * y912_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y912_LEN, y912_VW, LAYOUT_COLMAJ, y912_LEN );
	oski_value_t* y_val = array_Duplicate( x912_val,
		x912_LEN * x912_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x912_LEN, x912_VW, LAYOUT_COLMAJ, x912_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-10, -748), MAKE_VAL_COMPLEX(598, -532), MAKE_VAL_COMPLEX(632, 84), MAKE_VAL_COMPLEX(574, -844), MAKE_VAL_COMPLEX(-48, -72), MAKE_VAL_COMPLEX(-306, 6), MAKE_VAL_COMPLEX(698, 990), MAKE_VAL_COMPLEX(492, 458), MAKE_VAL_COMPLEX(-440, -206), MAKE_VAL_COMPLEX(-298, 180), MAKE_VAL_COMPLEX(-204, 520), MAKE_VAL_COMPLEX(-932, 22), MAKE_VAL_COMPLEX(-612, -188), MAKE_VAL_COMPLEX(-114, 772), MAKE_VAL_COMPLEX(-264, -26) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_959, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_959", y_true, y_val, x912_LEN*x912_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_959() */
/* END TEST 959 */

#define A960_M 5    /* # of rows */
#define A960_N 5    /* # of cols */
#define A960_NNZ 9  /* # of stored non-zeros */
static oski_index_t A960_ptr[A960_M+1] = {1, 2, 3, 5, 7, 10}; /* end row pointers for A960 */
static oski_index_t A960_ind[A960_NNZ] = {1, 1, 1, 2, 2, 4, 1, 4, 5}; /* end column indices pointers for A960 */
static oski_value_t A960_val[A960_NNZ] = {MAKE_VAL_COMPLEX(1, 5), MAKE_VAL_COMPLEX(2, -10), MAKE_VAL_COMPLEX(-1, -4), MAKE_VAL_COMPLEX(-2, -1), MAKE_VAL_COMPLEX(5, -5), MAKE_VAL_COMPLEX(10, 10), MAKE_VAL_COMPLEX(-4, -6), MAKE_VAL_COMPLEX(-10, -4), MAKE_VAL_COMPLEX(-10, -8)}; /* end stored non-zero values for A960 */

#define x960_LEN 5   /* vector length */
#define x960_VW  3   /* vector width */
static oski_value_t x960_val[x960_LEN*x960_VW] = {MAKE_VAL_COMPLEX(-2, -10), MAKE_VAL_COMPLEX(6, 4), MAKE_VAL_COMPLEX(4, 2), MAKE_VAL_COMPLEX(2, 5), MAKE_VAL_COMPLEX(-1, -8), MAKE_VAL_COMPLEX(-9, -3), MAKE_VAL_COMPLEX(-7, -1), MAKE_VAL_COMPLEX(-5, -2), MAKE_VAL_COMPLEX(-10, -2), MAKE_VAL_COMPLEX(3, 7), MAKE_VAL_COMPLEX(9, 4), MAKE_VAL_COMPLEX(-3, 5), MAKE_VAL_COMPLEX(-8, 10), MAKE_VAL_COMPLEX(7, -8), MAKE_VAL_COMPLEX(10, 9)}; /* data */

#define y960_LEN 5   /* vector length */
#define y960_VW  3   /* vector width */
static oski_value_t y960_val[y960_LEN*y960_VW] = {MAKE_VAL_COMPLEX(-0, -10), MAKE_VAL_COMPLEX(10, -9), MAKE_VAL_COMPLEX(-2, 4), MAKE_VAL_COMPLEX(-7, -5), MAKE_VAL_COMPLEX(8, 8), MAKE_VAL_COMPLEX(6, -6), MAKE_VAL_COMPLEX(-6, 1), MAKE_VAL_COMPLEX(-0, -2), MAKE_VAL_COMPLEX(6, 9), MAKE_VAL_COMPLEX(-1, 1), MAKE_VAL_COMPLEX(6, 1), MAKE_VAL_COMPLEX(-4, 8), MAKE_VAL_COMPLEX(-4, -5), MAKE_VAL_COMPLEX(-6, 10), MAKE_VAL_COMPLEX(-10, 7)}; /* data */

/*
 *  BEGIN TEST 960
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_960( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	oski_value_t* y_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(197, 67), MAKE_VAL_COMPLEX(-235, 94), MAKE_VAL_COMPLEX(-26, 28), MAKE_VAL_COMPLEX(1, 59), MAKE_VAL_COMPLEX(77, -116), MAKE_VAL_COMPLEX(20, 54), MAKE_VAL_COMPLEX(-1, 0), MAKE_VAL_COMPLEX(4, 1), MAKE_VAL_COMPLEX(-17, -57), MAKE_VAL_COMPLEX(121, 41), MAKE_VAL_COMPLEX(-133, 141), MAKE_VAL_COMPLEX(-180, -128), MAKE_VAL_COMPLEX(130, -36), MAKE_VAL_COMPLEX(-172, -164), MAKE_VAL_COMPLEX(78, -282) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_960, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_960", y_true, y_val, y960_LEN*y960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_960() */
/* END TEST 960 */

/*
 *  BEGIN TEST 961
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_961( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	oski_value_t* y_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-187, 43), MAKE_VAL_COMPLEX(134, 15), MAKE_VAL_COMPLEX(44, -74), MAKE_VAL_COMPLEX(81, -11), MAKE_VAL_COMPLEX(-173, 36), MAKE_VAL_COMPLEX(-80, 18), MAKE_VAL_COMPLEX(66, -17), MAKE_VAL_COMPLEX(-2, 21), MAKE_VAL_COMPLEX(-58, -36), MAKE_VAL_COMPLEX(-111, -53), MAKE_VAL_COMPLEX(66, -146), MAKE_VAL_COMPLEX(216, 36), MAKE_VAL_COMPLEX(-81, 87), MAKE_VAL_COMPLEX(228, 48), MAKE_VAL_COMPLEX(25, 195) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_961, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_961", y_true, y_val, y960_LEN*y960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_961() */
/* END TEST 961 */

/*
 *  BEGIN TEST 962
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_962( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	oski_value_t* y_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-30, 60), MAKE_VAL_COMPLEX(-87, 24), MAKE_VAL_COMPLEX(24, -18), MAKE_VAL_COMPLEX(27, 51), MAKE_VAL_COMPLEX(-24, -72), MAKE_VAL_COMPLEX(-54, 18), MAKE_VAL_COMPLEX(39, 12), MAKE_VAL_COMPLEX(-6, 12), MAKE_VAL_COMPLEX(-9, -72), MAKE_VAL_COMPLEX(9, -3), MAKE_VAL_COMPLEX(-33, -24), MAKE_VAL_COMPLEX(48, -36), MAKE_VAL_COMPLEX(9, 42), MAKE_VAL_COMPLEX(66, -42), MAKE_VAL_COMPLEX(81, -12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_962, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_962", y_true, y_val, y960_LEN*y960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_962() */
/* END TEST 962 */

/*
 *  BEGIN TEST 963
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_963( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	oski_value_t* y_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1080, -930), MAKE_VAL_COMPLEX(-641, 1153), MAKE_VAL_COMPLEX(50, 430), MAKE_VAL_COMPLEX(-16, 318), MAKE_VAL_COMPLEX(221, -933), MAKE_VAL_COMPLEX(394, -142), MAKE_VAL_COMPLEX(-149, 177), MAKE_VAL_COMPLEX(-23, -71), MAKE_VAL_COMPLEX(14, -92), MAKE_VAL_COMPLEX(806, -348), MAKE_VAL_COMPLEX(232, 1254), MAKE_VAL_COMPLEX(-1460, 580), MAKE_VAL_COMPLEX(238, -874), MAKE_VAL_COMPLEX(-1608, 464), MAKE_VAL_COMPLEX(-1132, -1364) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_963, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_963", y_true, y_val, y960_LEN*y960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_963() */
/* END TEST 963 */

/*
 *  BEGIN TEST 964
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_964( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-66, 137), MAKE_VAL_COMPLEX(241, -59), MAKE_VAL_COMPLEX(48, -101), MAKE_VAL_COMPLEX(-64, -21), MAKE_VAL_COMPLEX(-76, -126), MAKE_VAL_COMPLEX(47, 112), MAKE_VAL_COMPLEX(-29, 63), MAKE_VAL_COMPLEX(-59, -28), MAKE_VAL_COMPLEX(0, 62), MAKE_VAL_COMPLEX(-28, 54), MAKE_VAL_COMPLEX(241, -55), MAKE_VAL_COMPLEX(36, -40), MAKE_VAL_COMPLEX(12, 146), MAKE_VAL_COMPLEX(-57, -137), MAKE_VAL_COMPLEX(295, -117) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_964, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_964", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_964() */
/* END TEST 964 */

/*
 *  BEGIN TEST 965
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_965( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(52, -207), MAKE_VAL_COMPLEX(-199, 87), MAKE_VAL_COMPLEX(-20, 115), MAKE_VAL_COMPLEX(78, 56), MAKE_VAL_COMPLEX(69, 70), MAKE_VAL_COMPLEX(-110, -133), MAKE_VAL_COMPLEX(-20, -70), MAKE_VAL_COMPLEX(24, 14), MAKE_VAL_COMPLEX(-70, -76), MAKE_VAL_COMPLEX(49, -5), MAKE_VAL_COMPLEX(-178, 83), MAKE_VAL_COMPLEX(-57, 75), MAKE_VAL_COMPLEX(-68, -76), MAKE_VAL_COMPLEX(106, 81), MAKE_VAL_COMPLEX(-225, 180) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_965, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_965", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_965() */
/* END TEST 965 */

/*
 *  BEGIN TEST 966
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_966( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-74, 46), MAKE_VAL_COMPLEX(14, -60), MAKE_VAL_COMPLEX(4, -38), MAKE_VAL_COMPLEX(34, -31), MAKE_VAL_COMPLEX(-61, 32), MAKE_VAL_COMPLEX(3, 81), MAKE_VAL_COMPLEX(13, 59), MAKE_VAL_COMPLEX(-1, 46), MAKE_VAL_COMPLEX(14, 86), MAKE_VAL_COMPLEX(47, -45), MAKE_VAL_COMPLEX(5, -84), MAKE_VAL_COMPLEX(49, 9), MAKE_VAL_COMPLEX(104, 34), MAKE_VAL_COMPLEX(-85, -32), MAKE_VAL_COMPLEX(42, -107) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_966, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_966", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_966() */
/* END TEST 966 */

/*
 *  BEGIN TEST 967
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_967( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(172, -539), MAKE_VAL_COMPLEX(-1199, 143), MAKE_VAL_COMPLEX(-248, 411), MAKE_VAL_COMPLEX(388, 14), MAKE_VAL_COMPLEX(247, 736), MAKE_VAL_COMPLEX(-190, -365), MAKE_VAL_COMPLEX(204, -178), MAKE_VAL_COMPLEX(314, 252), MAKE_VAL_COMPLEX(74, -108), MAKE_VAL_COMPLEX(233, -401), MAKE_VAL_COMPLEX(-1232, 69), MAKE_VAL_COMPLEX(-57, 199), MAKE_VAL_COMPLEX(208, -696), MAKE_VAL_COMPLEX(64, 647), MAKE_VAL_COMPLEX(-1423, 306) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_967, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_967", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_967() */
/* END TEST 967 */

/*
 *  BEGIN TEST 968
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_968( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(52, -133), MAKE_VAL_COMPLEX(-129, -21), MAKE_VAL_COMPLEX(46, -1), MAKE_VAL_COMPLEX(93, -19), MAKE_VAL_COMPLEX(152, 111), MAKE_VAL_COMPLEX(-128, -25), MAKE_VAL_COMPLEX(54, -2), MAKE_VAL_COMPLEX(1, 29), MAKE_VAL_COMPLEX(-26, -16), MAKE_VAL_COMPLEX(39, 7), MAKE_VAL_COMPLEX(91, -72), MAKE_VAL_COMPLEX(159, 9), MAKE_VAL_COMPLEX(118, 38), MAKE_VAL_COMPLEX(-47, -32), MAKE_VAL_COMPLEX(28, -245) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_968, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_968", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_968() */
/* END TEST 968 */

/*
 *  BEGIN TEST 969
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_969( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(60, 121), MAKE_VAL_COMPLEX(79, 83), MAKE_VAL_COMPLEX(-72, 43), MAKE_VAL_COMPLEX(-150, 36), MAKE_VAL_COMPLEX(-63, -114), MAKE_VAL_COMPLEX(170, -71), MAKE_VAL_COMPLEX(-36, -74), MAKE_VAL_COMPLEX(26, -82), MAKE_VAL_COMPLEX(58, -92), MAKE_VAL_COMPLEX(-119, 19), MAKE_VAL_COMPLEX(-144, 167), MAKE_VAL_COMPLEX(-211, -47), MAKE_VAL_COMPLEX(-220, -136), MAKE_VAL_COMPLEX(128, 117), MAKE_VAL_COMPLEX(-137, 346) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_969, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_969", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_969() */
/* END TEST 969 */

/*
 *  BEGIN TEST 970
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_970( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(6, -74), MAKE_VAL_COMPLEX(34, 40), MAKE_VAL_COMPLEX(24, 22), MAKE_VAL_COMPLEX(4, 39), MAKE_VAL_COMPLEX(9, -58), MAKE_VAL_COMPLEX(-57, -39), MAKE_VAL_COMPLEX(-47, -21), MAKE_VAL_COMPLEX(-31, -24), MAKE_VAL_COMPLEX(-66, -34), MAKE_VAL_COMPLEX(7, 55), MAKE_VAL_COMPLEX(55, 46), MAKE_VAL_COMPLEX(-31, 29), MAKE_VAL_COMPLEX(-76, 54), MAKE_VAL_COMPLEX(65, -42), MAKE_VAL_COMPLEX(52, 83) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_970, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_970", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_970() */
/* END TEST 970 */

/*
 *  BEGIN TEST 971
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_971( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(744, -732), MAKE_VAL_COMPLEX(-908, -848), MAKE_VAL_COMPLEX(384, 72), MAKE_VAL_COMPLEX(914, 136), MAKE_VAL_COMPLEX(604, 1582), MAKE_VAL_COMPLEX(-918, -490), MAKE_VAL_COMPLEX(474, 362), MAKE_VAL_COMPLEX(-112, 362), MAKE_VAL_COMPLEX(-108, 4), MAKE_VAL_COMPLEX(414, 90), MAKE_VAL_COMPLEX(1032, -442), MAKE_VAL_COMPLEX(1370, 734), MAKE_VAL_COMPLEX(1076, 872), MAKE_VAL_COMPLEX(-480, -534), MAKE_VAL_COMPLEX(1322, -2140) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_971, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_971", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_971() */
/* END TEST 971 */

/*
 *  BEGIN TEST 972
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_972( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	oski_value_t* y_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(167, -23), MAKE_VAL_COMPLEX(-38, -28), MAKE_VAL_COMPLEX(-31, 24), MAKE_VAL_COMPLEX(79, 53), MAKE_VAL_COMPLEX(154, -66), MAKE_VAL_COMPLEX(-175, 4), MAKE_VAL_COMPLEX(110, -47), MAKE_VAL_COMPLEX(-2, -17), MAKE_VAL_COMPLEX(-52, 156), MAKE_VAL_COMPLEX(-208, -98), MAKE_VAL_COMPLEX(-17, 37), MAKE_VAL_COMPLEX(68, 54), MAKE_VAL_COMPLEX(-8, -6), MAKE_VAL_COMPLEX(-192, -32), MAKE_VAL_COMPLEX(9, -189) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_972, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_972", y_true, y_val, y960_LEN*y960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_972() */
/* END TEST 972 */

/*
 *  BEGIN TEST 973
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_973( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	oski_value_t* y_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-197, -97), MAKE_VAL_COMPLEX(131, -110), MAKE_VAL_COMPLEX(19, 30), MAKE_VAL_COMPLEX(-178, -92), MAKE_VAL_COMPLEX(-34, 138), MAKE_VAL_COMPLEX(229, -94), MAKE_VAL_COMPLEX(-179, 77), MAKE_VAL_COMPLEX(-4, -7), MAKE_VAL_COMPLEX(151, -66), MAKE_VAL_COMPLEX(199, 113), MAKE_VAL_COMPLEX(92, -43), MAKE_VAL_COMPLEX(-92, 54), MAKE_VAL_COMPLEX(-55, -42), MAKE_VAL_COMPLEX(150, 170), MAKE_VAL_COMPLEX(-108, 303) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_973, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_973", y_true, y_val, y960_LEN*y960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_973() */
/* END TEST 973 */

/*
 *  BEGIN TEST 974
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_974( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	oski_value_t* y_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-10, 70), MAKE_VAL_COMPLEX(-79, 53), MAKE_VAL_COMPLEX(18, -26), MAKE_VAL_COMPLEX(44, 42), MAKE_VAL_COMPLEX(-48, -64), MAKE_VAL_COMPLEX(-48, 36), MAKE_VAL_COMPLEX(43, -1), MAKE_VAL_COMPLEX(-2, 14), MAKE_VAL_COMPLEX(-33, -69), MAKE_VAL_COMPLEX(8, -6), MAKE_VAL_COMPLEX(-41, -13), MAKE_VAL_COMPLEX(36, -52), MAKE_VAL_COMPLEX(23, 39), MAKE_VAL_COMPLEX(52, -64), MAKE_VAL_COMPLEX(77, -39) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_974, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_974", y_true, y_val, y960_LEN*y960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_974() */
/* END TEST 974 */

/*
 *  BEGIN TEST 975
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_975( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	oski_value_t* y_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-32, 1222), MAKE_VAL_COMPLEX(-205, -200), MAKE_VAL_COMPLEX(-32, -240), MAKE_VAL_COMPLEX(-249, 703), MAKE_VAL_COMPLEX(356, 612), MAKE_VAL_COMPLEX(-258, -1020), MAKE_VAL_COMPLEX(431, 740), MAKE_VAL_COMPLEX(68, 32), MAKE_VAL_COMPLEX(-915, -612), MAKE_VAL_COMPLEX(627, -1253), MAKE_VAL_COMPLEX(-337, -226), MAKE_VAL_COMPLEX(-100, 300), MAKE_VAL_COMPLEX(39, 130), MAKE_VAL_COMPLEX(494, -1270), MAKE_VAL_COMPLEX(1473, 70) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_975, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_975", y_true, y_val, y960_LEN*y960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_975() */
/* END TEST 975 */

/*
 *  BEGIN TEST 976
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_976( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(66, -256), MAKE_VAL_COMPLEX(36, 142), MAKE_VAL_COMPLEX(143, 172), MAKE_VAL_COMPLEX(-174, 16), MAKE_VAL_COMPLEX(69, -91), MAKE_VAL_COMPLEX(27, -30), MAKE_VAL_COMPLEX(-91, -28), MAKE_VAL_COMPLEX(-25, -52), MAKE_VAL_COMPLEX(-14, -105), MAKE_VAL_COMPLEX(-93, -281), MAKE_VAL_COMPLEX(-29, 249), MAKE_VAL_COMPLEX(-54, 72), MAKE_VAL_COMPLEX(-118, -34), MAKE_VAL_COMPLEX(10, -118), MAKE_VAL_COMPLEX(224, -10) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_976, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_976", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_976() */
/* END TEST 976 */

/*
 *  BEGIN TEST 977
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_977( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-94, 220), MAKE_VAL_COMPLEX(-4, -138), MAKE_VAL_COMPLEX(-123, -172), MAKE_VAL_COMPLEX(192, 0), MAKE_VAL_COMPLEX(-89, 61), MAKE_VAL_COMPLEX(-69, 36), MAKE_VAL_COMPLEX(61, 38), MAKE_VAL_COMPLEX(1, 54), MAKE_VAL_COMPLEX(-30, 117), MAKE_VAL_COMPLEX(119, 303), MAKE_VAL_COMPLEX(73, -251), MAKE_VAL_COMPLEX(52, -46), MAKE_VAL_COMPLEX(106, 90), MAKE_VAL_COMPLEX(2, 72), MAKE_VAL_COMPLEX(-166, 26) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_977, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_977", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_977() */
/* END TEST 977 */

/*
 *  BEGIN TEST 978
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_978( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(26, -78), MAKE_VAL_COMPLEX(26, 52), MAKE_VAL_COMPLEX(20, 30), MAKE_VAL_COMPLEX(-6, 43), MAKE_VAL_COMPLEX(25, -60), MAKE_VAL_COMPLEX(-51, -57), MAKE_VAL_COMPLEX(-45, -35), MAKE_VAL_COMPLEX(-27, -34), MAKE_VAL_COMPLEX(-62, -54), MAKE_VAL_COMPLEX(-7, 61), MAKE_VAL_COMPLEX(47, 64), MAKE_VAL_COMPLEX(-41, 23), MAKE_VAL_COMPLEX(-96, 38), MAKE_VAL_COMPLEX(81, -28), MAKE_VAL_COMPLEX(34, 103) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_978, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_978", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_978() */
/* END TEST 978 */

/*
 *  BEGIN TEST 979
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_979( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-346, -762), MAKE_VAL_COMPLEX(280, 250), MAKE_VAL_COMPLEX(806, 262), MAKE_VAL_COMPLEX(-651, 198), MAKE_VAL_COMPLEX(-6, -197), MAKE_VAL_COMPLEX(321, 87), MAKE_VAL_COMPLEX(-185, 225), MAKE_VAL_COMPLEX(-58, -1), MAKE_VAL_COMPLEX(60, -150), MAKE_VAL_COMPLEX(-923, -1241), MAKE_VAL_COMPLEX(126, 757), MAKE_VAL_COMPLEX(121, 267), MAKE_VAL_COMPLEX(-82, -124), MAKE_VAL_COMPLEX(-584, -323), MAKE_VAL_COMPLEX(669, -982) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_979, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_979", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_979() */
/* END TEST 979 */

/*
 *  BEGIN TEST 980
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_980( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-106, 184), MAKE_VAL_COMPLEX(-58, -192), MAKE_VAL_COMPLEX(-73, -206), MAKE_VAL_COMPLEX(117, -153), MAKE_VAL_COMPLEX(-4, 208), MAKE_VAL_COMPLEX(-11, 196), MAKE_VAL_COMPLEX(83, 110), MAKE_VAL_COMPLEX(56, 85), MAKE_VAL_COMPLEX(62, 111), MAKE_VAL_COMPLEX(-101, -113), MAKE_VAL_COMPLEX(76, -126), MAKE_VAL_COMPLEX(124, 52), MAKE_VAL_COMPLEX(168, 28), MAKE_VAL_COMPLEX(-217, -15), MAKE_VAL_COMPLEX(25, -413) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_980, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_980", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_980() */
/* END TEST 980 */

/*
 *  BEGIN TEST 981
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_981( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(68, -10), MAKE_VAL_COMPLEX(-10, 86), MAKE_VAL_COMPLEX(23, 146), MAKE_VAL_COMPLEX(-114, 59), MAKE_VAL_COMPLEX(-36, -73), MAKE_VAL_COMPLEX(134, -85), MAKE_VAL_COMPLEX(22, -45), MAKE_VAL_COMPLEX(10, -18), MAKE_VAL_COMPLEX(84, -9), MAKE_VAL_COMPLEX(102, -20), MAKE_VAL_COMPLEX(-192, -1), MAKE_VAL_COMPLEX(-41, -111), MAKE_VAL_COMPLEX(30, -132), MAKE_VAL_COMPLEX(49, 94), MAKE_VAL_COMPLEX(-122, 199) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_981, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_981", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_981() */
/* END TEST 981 */

/*
 *  BEGIN TEST 982
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_982( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-36, 80), MAKE_VAL_COMPLEX(-22, -58), MAKE_VAL_COMPLEX(-18, -34), MAKE_VAL_COMPLEX(11, -45), MAKE_VAL_COMPLEX(-33, 61), MAKE_VAL_COMPLEX(48, 66), MAKE_VAL_COMPLEX(44, 42), MAKE_VAL_COMPLEX(25, 39), MAKE_VAL_COMPLEX(60, 64), MAKE_VAL_COMPLEX(14, -64), MAKE_VAL_COMPLEX(-43, -73), MAKE_VAL_COMPLEX(46, -20), MAKE_VAL_COMPLEX(106, -30), MAKE_VAL_COMPLEX(-89, 21), MAKE_VAL_COMPLEX(-25, -113) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_982, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_982", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_982() */
/* END TEST 982 */

/*
 *  BEGIN TEST 983
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_ROWMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_983( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_ROWMAJ, x960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(338, 8), MAKE_VAL_COMPLEX(-342, 288), MAKE_VAL_COMPLEX(-535, 386), MAKE_VAL_COMPLEX(-445, -243), MAKE_VAL_COMPLEX(452, -404), MAKE_VAL_COMPLEX(375, 148), MAKE_VAL_COMPLEX(89, -166), MAKE_VAL_COMPLEX(36, -153), MAKE_VAL_COMPLEX(-26, 17), MAKE_VAL_COMPLEX(53, 609), MAKE_VAL_COMPLEX(-116, -446), MAKE_VAL_COMPLEX(208, -256), MAKE_VAL_COMPLEX(128, -28), MAKE_VAL_COMPLEX(29, 319), MAKE_VAL_COMPLEX(-961, 161) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, -4);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_983, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_983", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_983() */
/* END TEST 983 */

/*
 *  BEGIN TEST 984
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_984( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	oski_value_t* y_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(50, -122), MAKE_VAL_COMPLEX(70, -77), MAKE_VAL_COMPLEX(85, 45), MAKE_VAL_COMPLEX(-129, -21), MAKE_VAL_COMPLEX(-36, 181), MAKE_VAL_COMPLEX(115, -217), MAKE_VAL_COMPLEX(-87, 17), MAKE_VAL_COMPLEX(8, 34), MAKE_VAL_COMPLEX(69, 10), MAKE_VAL_COMPLEX(-8, 152), MAKE_VAL_COMPLEX(-79, -171), MAKE_VAL_COMPLEX(76, -40), MAKE_VAL_COMPLEX(-139, 51), MAKE_VAL_COMPLEX(104, 108), MAKE_VAL_COMPLEX(-205, -129) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_984, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_984", y_true, y_val, y960_LEN*y960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_984() */
/* END TEST 984 */

/*
 *  BEGIN TEST 985
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_985( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	oski_value_t* y_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 102), MAKE_VAL_COMPLEX(-5, 109), MAKE_VAL_COMPLEX(-109, -47), MAKE_VAL_COMPLEX(140, -24), MAKE_VAL_COMPLEX(12, -125), MAKE_VAL_COMPLEX(-73, 235), MAKE_VAL_COMPLEX(70, -45), MAKE_VAL_COMPLEX(2, -38), MAKE_VAL_COMPLEX(-102, 38), MAKE_VAL_COMPLEX(1, -155), MAKE_VAL_COMPLEX(86, 203), MAKE_VAL_COMPLEX(-124, 36), MAKE_VAL_COMPLEX(156, -81), MAKE_VAL_COMPLEX(-166, -118), MAKE_VAL_COMPLEX(150, 93) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_985, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_985", y_true, y_val, y960_LEN*y960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_985() */
/* END TEST 985 */

/*
 *  BEGIN TEST 986
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_986( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	oski_value_t* y_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(20, -50), MAKE_VAL_COMPLEX(68, -25), MAKE_VAL_COMPLEX(-18, 16), MAKE_VAL_COMPLEX(-25, -39), MAKE_VAL_COMPLEX(24, 56), MAKE_VAL_COMPLEX(42, -18), MAKE_VAL_COMPLEX(-32, -7), MAKE_VAL_COMPLEX(4, -10), MAKE_VAL_COMPLEX(12, 57), MAKE_VAL_COMPLEX(-7, 3), MAKE_VAL_COMPLEX(28, 17), MAKE_VAL_COMPLEX(-36, 32), MAKE_VAL_COMPLEX(-10, -33), MAKE_VAL_COMPLEX(-50, 38), MAKE_VAL_COMPLEX(-64, 15) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_986, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_986", y_true, y_val, y960_LEN*y960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_986() */
/* END TEST 986 */

/*
 *  BEGIN TEST 987
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_987( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	oski_value_t* y_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(506, 466), MAKE_VAL_COMPLEX(-67, 115), MAKE_VAL_COMPLEX(100, 760), MAKE_VAL_COMPLEX(-158, -520), MAKE_VAL_COMPLEX(-1312, -630), MAKE_VAL_COMPLEX(1438, 330), MAKE_VAL_COMPLEX(-67, -351), MAKE_VAL_COMPLEX(-370, 194), MAKE_VAL_COMPLEX(385, -49), MAKE_VAL_COMPLEX(-1144, 266), MAKE_VAL_COMPLEX(1081, -1347), MAKE_VAL_COMPLEX(1052, 508), MAKE_VAL_COMPLEX(-847, -627), MAKE_VAL_COMPLEX(100, 1080), MAKE_VAL_COMPLEX(1331, -1545) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_987, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_987", y_true, y_val, y960_LEN*y960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_987() */
/* END TEST 987 */

/*
 *  BEGIN TEST 988
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_988( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-28, 119), MAKE_VAL_COMPLEX(-83, -12), MAKE_VAL_COMPLEX(-29, 23), MAKE_VAL_COMPLEX(-55, 74), MAKE_VAL_COMPLEX(-54, 117), MAKE_VAL_COMPLEX(224, -16), MAKE_VAL_COMPLEX(-38, -132), MAKE_VAL_COMPLEX(-56, -50), MAKE_VAL_COMPLEX(228, 4), MAKE_VAL_COMPLEX(-3, -113), MAKE_VAL_COMPLEX(46, -88), MAKE_VAL_COMPLEX(58, 67), MAKE_VAL_COMPLEX(10, 18), MAKE_VAL_COMPLEX(0, -57), MAKE_VAL_COMPLEX(269, -68) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_988, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_988", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_988() */
/* END TEST 988 */

/*
 *  BEGIN TEST 989
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_989( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-62, -153), MAKE_VAL_COMPLEX(145, -16), MAKE_VAL_COMPLEX(65, -45), MAKE_VAL_COMPLEX(105, -65), MAKE_VAL_COMPLEX(-15, -149), MAKE_VAL_COMPLEX(-293, 73), MAKE_VAL_COMPLEX(-5, 183), MAKE_VAL_COMPLEX(15, 80), MAKE_VAL_COMPLEX(-294, 66), MAKE_VAL_COMPLEX(74, 124), MAKE_VAL_COMPLEX(31, 36), MAKE_VAL_COMPLEX(-33, -18), MAKE_VAL_COMPLEX(30, 96), MAKE_VAL_COMPLEX(-29, -39), MAKE_VAL_COMPLEX(-147, 33) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_989, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_989", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_989() */
/* END TEST 989 */

/*
 *  BEGIN TEST 990
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_990( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(72, -4), MAKE_VAL_COMPLEX(-34, 38), MAKE_VAL_COMPLEX(-18, 26), MAKE_VAL_COMPLEX(-37, 9), MAKE_VAL_COMPLEX(57, 1), MAKE_VAL_COMPLEX(30, -60), MAKE_VAL_COMPLEX(14, -48), MAKE_VAL_COMPLEX(19, -33), MAKE_VAL_COMPLEX(24, -68), MAKE_VAL_COMPLEX(-52, 14), MAKE_VAL_COMPLEX(-37, 59), MAKE_VAL_COMPLEX(-32, -26), MAKE_VAL_COMPLEX(-62, -66), MAKE_VAL_COMPLEX(49, 57), MAKE_VAL_COMPLEX(-73, 61) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_990, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_990", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_990() */
/* END TEST 990 */

/*
 *  BEGIN TEST 991
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_991( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-897, 673), MAKE_VAL_COMPLEX(-581, -617), MAKE_VAL_COMPLEX(-412, -6), MAKE_VAL_COMPLEX(-975, 127), MAKE_VAL_COMPLEX(-1095, 477), MAKE_VAL_COMPLEX(1770, 1366), MAKE_VAL_COMPLEX(712, -1268), MAKE_VAL_COMPLEX(12, -790), MAKE_VAL_COMPLEX(1652, 1516), MAKE_VAL_COMPLEX(668, -818), MAKE_VAL_COMPLEX(836, -210), MAKE_VAL_COMPLEX(-105, 809), MAKE_VAL_COMPLEX(-128, 40), MAKE_VAL_COMPLEX(453, -267), MAKE_VAL_COMPLEX(2191, 1473) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_991, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_991", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_991() */
/* END TEST 991 */

/*
 *  BEGIN TEST 992
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_992( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(112, -93), MAKE_VAL_COMPLEX(49, -2), MAKE_VAL_COMPLEX(29, 33), MAKE_VAL_COMPLEX(7, -12), MAKE_VAL_COMPLEX(206, 69), MAKE_VAL_COMPLEX(-70, -88), MAKE_VAL_COMPLEX(172, 76), MAKE_VAL_COMPLEX(36, 14), MAKE_VAL_COMPLEX(144, -156), MAKE_VAL_COMPLEX(-131, -45), MAKE_VAL_COMPLEX(-18, 36), MAKE_VAL_COMPLEX(-148, -45), MAKE_VAL_COMPLEX(-58, -90), MAKE_VAL_COMPLEX(200, 91), MAKE_VAL_COMPLEX(-67, -240) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_992, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_992", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_992() */
/* END TEST 992 */

/*
 *  BEGIN TEST 993
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_993( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-54, 123), MAKE_VAL_COMPLEX(-93, 16), MAKE_VAL_COMPLEX(-55, -21), MAKE_VAL_COMPLEX(-40, 2), MAKE_VAL_COMPLEX(-162, -42), MAKE_VAL_COMPLEX(121, 55), MAKE_VAL_COMPLEX(-139, -107), MAKE_VAL_COMPLEX(-6, -31), MAKE_VAL_COMPLEX(-94, 114), MAKE_VAL_COMPLEX(84, 32), MAKE_VAL_COMPLEX(-38, -7), MAKE_VAL_COMPLEX(135, 10), MAKE_VAL_COMPLEX(40, 10), MAKE_VAL_COMPLEX(-188, -24), MAKE_VAL_COMPLEX(-18, 254) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_993, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_993", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_993() */
/* END TEST 993 */

/*
 *  BEGIN TEST 994
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_994( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-46, -22), MAKE_VAL_COMPLEX(34, -12), MAKE_VAL_COMPLEX(20, -10), MAKE_VAL_COMPLEX(26, 7), MAKE_VAL_COMPLEX(-35, -20), MAKE_VAL_COMPLEX(-39, 27), MAKE_VAL_COMPLEX(-25, 25), MAKE_VAL_COMPLEX(-23, 14), MAKE_VAL_COMPLEX(-38, 34), MAKE_VAL_COMPLEX(37, 9), MAKE_VAL_COMPLEX(43, -24), MAKE_VAL_COMPLEX(11, 27), MAKE_VAL_COMPLEX(16, 62), MAKE_VAL_COMPLEX(-11, -52), MAKE_VAL_COMPLEX(66, -13) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_994, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_994", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_994() */
/* END TEST 994 */

/*
 *  BEGIN TEST 995
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_995( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_ROWMAJ, y960_VW );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(21, 573), MAKE_VAL_COMPLEX(-485, 305), MAKE_VAL_COMPLEX(-308, 34), MAKE_VAL_COMPLEX(-276, 116), MAKE_VAL_COMPLEX(-714, -114), MAKE_VAL_COMPLEX(660, -26), MAKE_VAL_COMPLEX(-826, -528), MAKE_VAL_COMPLEX(-55, -251), MAKE_VAL_COMPLEX(-380, 472), MAKE_VAL_COMPLEX(344, 172), MAKE_VAL_COMPLEX(-215, 189), MAKE_VAL_COMPLEX(577, -109), MAKE_VAL_COMPLEX(-18, -74), MAKE_VAL_COMPLEX(-778, 146), MAKE_VAL_COMPLEX(62, 1522) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_995, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_995", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_995() */
/* END TEST 995 */

/*
 *  BEGIN TEST 996
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_996( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	oski_value_t* y_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, -72), MAKE_VAL_COMPLEX(-109, -71), MAKE_VAL_COMPLEX(-26, 24), MAKE_VAL_COMPLEX(-46, 176), MAKE_VAL_COMPLEX(-42, 50), MAKE_VAL_COMPLEX(-15, -52), MAKE_VAL_COMPLEX(-106, 171), MAKE_VAL_COMPLEX(-2, 44), MAKE_VAL_COMPLEX(-56, -190), MAKE_VAL_COMPLEX(140, 40), MAKE_VAL_COMPLEX(113, -19), MAKE_VAL_COMPLEX(119, -129), MAKE_VAL_COMPLEX(-20, -33), MAKE_VAL_COMPLEX(144, -44), MAKE_VAL_COMPLEX(-120, -114) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_996, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_996", y_true, y_val, y960_LEN*y960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_996() */
/* END TEST 996 */

/*
 *  BEGIN TEST 997
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_997( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	oski_value_t* y_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-130, 42), MAKE_VAL_COMPLEX(22, -86), MAKE_VAL_COMPLEX(72, 14), MAKE_VAL_COMPLEX(-40, -100), MAKE_VAL_COMPLEX(170, -130), MAKE_VAL_COMPLEX(-45, -44), MAKE_VAL_COMPLEX(101, -90), MAKE_VAL_COMPLEX(-24, -50), MAKE_VAL_COMPLEX(191, 139), MAKE_VAL_COMPLEX(-130, -24), MAKE_VAL_COMPLEX(-82, -56), MAKE_VAL_COMPLEX(-27, 205), MAKE_VAL_COMPLEX(-57, 70), MAKE_VAL_COMPLEX(-32, 152), MAKE_VAL_COMPLEX(181, 265) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_997, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_997", y_true, y_val, y960_LEN*y960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_997() */
/* END TEST 997 */

/*
 *  BEGIN TEST 998
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_998( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	oski_value_t* y_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(50, 40), MAKE_VAL_COMPLEX(5, 86), MAKE_VAL_COMPLEX(-12, -26), MAKE_VAL_COMPLEX(53, -15), MAKE_VAL_COMPLEX(-72, 8), MAKE_VAL_COMPLEX(6, 54), MAKE_VAL_COMPLEX(19, -34), MAKE_VAL_COMPLEX(10, 8), MAKE_VAL_COMPLEX(-69, -6), MAKE_VAL_COMPLEX(-1, -9), MAKE_VAL_COMPLEX(-29, 26), MAKE_VAL_COMPLEX(-24, -52), MAKE_VAL_COMPLEX(41, 0), MAKE_VAL_COMPLEX(-26, -70), MAKE_VAL_COMPLEX(5, -78) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_998, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_998", y_true, y_val, y960_LEN*y960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_998() */
/* END TEST 998 */

/*
 *  BEGIN TEST 999
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_999( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	oski_value_t* y_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(636, 300), MAKE_VAL_COMPLEX(-583, -573), MAKE_VAL_COMPLEX(-214, -364), MAKE_VAL_COMPLEX(-1146, 800), MAKE_VAL_COMPLEX(-1186, -566), MAKE_VAL_COMPLEX(-19, 52), MAKE_VAL_COMPLEX(-1508, -95), MAKE_VAL_COMPLEX(-350, 336), MAKE_VAL_COMPLEX(790, -1956), MAKE_VAL_COMPLEX(440, 1248), MAKE_VAL_COMPLEX(315, 779), MAKE_VAL_COMPLEX(1843, -245), MAKE_VAL_COMPLEX(458, -19), MAKE_VAL_COMPLEX(1408, 236), MAKE_VAL_COMPLEX(930, -2052) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_999, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_999", y_true, y_val, y960_LEN*y960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_999() */
/* END TEST 999 */

/*
 *  BEGIN TEST 1000
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1000( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(88, -146), MAKE_VAL_COMPLEX(-218, -8), MAKE_VAL_COMPLEX(-109, 28), MAKE_VAL_COMPLEX(-107, -350), MAKE_VAL_COMPLEX(29, 24), MAKE_VAL_COMPLEX(117, 57), MAKE_VAL_COMPLEX(80, -88), MAKE_VAL_COMPLEX(30, -30), MAKE_VAL_COMPLEX(41, 133), MAKE_VAL_COMPLEX(-129, -159), MAKE_VAL_COMPLEX(52, 166), MAKE_VAL_COMPLEX(36, -17), MAKE_VAL_COMPLEX(10, -155), MAKE_VAL_COMPLEX(-13, 168), MAKE_VAL_COMPLEX(114, -109) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1000, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1000", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1000() */
/* END TEST 1000 */

/*
 *  BEGIN TEST 1001
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1001( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-8, 234), MAKE_VAL_COMPLEX(134, 4), MAKE_VAL_COMPLEX(57, -24), MAKE_VAL_COMPLEX(57, 312), MAKE_VAL_COMPLEX(29, 50), MAKE_VAL_COMPLEX(-9, -81), MAKE_VAL_COMPLEX(-4, 56), MAKE_VAL_COMPLEX(32, 20), MAKE_VAL_COMPLEX(71, -173), MAKE_VAL_COMPLEX(57, 107), MAKE_VAL_COMPLEX(-166, -152), MAKE_VAL_COMPLEX(-36, -51), MAKE_VAL_COMPLEX(10, 7), MAKE_VAL_COMPLEX(-9, -46), MAKE_VAL_COMPLEX(-268, 79) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1001, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1001", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1001() */
/* END TEST 1001 */

/*
 *  BEGIN TEST 1002
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1002( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-16, 24), MAKE_VAL_COMPLEX(-4, -20), MAKE_VAL_COMPLEX(-4, -12), MAKE_VAL_COMPLEX(6, -14), MAKE_VAL_COMPLEX(-14, 18), MAKE_VAL_COMPLEX(12, 24), MAKE_VAL_COMPLEX(12, 16), MAKE_VAL_COMPLEX(6, 14), MAKE_VAL_COMPLEX(16, 24), MAKE_VAL_COMPLEX(8, -20), MAKE_VAL_COMPLEX(-10, -26), MAKE_VAL_COMPLEX(16, -4), MAKE_VAL_COMPLEX(36, -4), MAKE_VAL_COMPLEX(-30, 2), MAKE_VAL_COMPLEX(-2, -38) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1002, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1002", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1002() */
/* END TEST 1002 */

/*
 *  BEGIN TEST 1003
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1003( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1420, -844), MAKE_VAL_COMPLEX(-648, -1222), MAKE_VAL_COMPLEX(-477, -441), MAKE_VAL_COMPLEX(2007, -2124), MAKE_VAL_COMPLEX(-14, -255), MAKE_VAL_COMPLEX(-423, 801), MAKE_VAL_COMPLEX(529, -21), MAKE_VAL_COMPLEX(25, -132), MAKE_VAL_COMPLEX(-1416, 710), MAKE_VAL_COMPLEX(611, -1171), MAKE_VAL_COMPLEX(-303, 1568), MAKE_VAL_COMPLEX(36, 492), MAKE_VAL_COMPLEX(467, -75), MAKE_VAL_COMPLEX(-664, 251), MAKE_VAL_COMPLEX(1953, 992) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1003, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1003", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1003() */
/* END TEST 1003 */

/*
 *  BEGIN TEST 1004
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1004( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-36, 118), MAKE_VAL_COMPLEX(70, -112), MAKE_VAL_COMPLEX(15, 26), MAKE_VAL_COMPLEX(-140, -45), MAKE_VAL_COMPLEX(2, 79), MAKE_VAL_COMPLEX(-7, -81), MAKE_VAL_COMPLEX(86, 138), MAKE_VAL_COMPLEX(47, 35), MAKE_VAL_COMPLEX(159, 9), MAKE_VAL_COMPLEX(-87, -55), MAKE_VAL_COMPLEX(-87, -177), MAKE_VAL_COMPLEX(-48, 71), MAKE_VAL_COMPLEX(32, -29), MAKE_VAL_COMPLEX(16, 95), MAKE_VAL_COMPLEX(3, -288) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1004, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1004", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1004() */
/* END TEST 1004 */

/*
 *  BEGIN TEST 1005
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1005( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-22, -44), MAKE_VAL_COMPLEX(-78, 46), MAKE_VAL_COMPLEX(-25, -66), MAKE_VAL_COMPLEX(163, 1), MAKE_VAL_COMPLEX(-52, -24), MAKE_VAL_COMPLEX(40, 162), MAKE_VAL_COMPLEX(-51, -83), MAKE_VAL_COMPLEX(-31, 12), MAKE_VAL_COMPLEX(-113, 73), MAKE_VAL_COMPLEX(118, -8), MAKE_VAL_COMPLEX(61, 90), MAKE_VAL_COMPLEX(101, -80), MAKE_VAL_COMPLEX(86, 25), MAKE_VAL_COMPLEX(-114, -96), MAKE_VAL_COMPLEX(0, 164) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1005, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1005", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1005() */
/* END TEST 1005 */

/*
 *  BEGIN TEST 1006
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1006( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(82, -6), MAKE_VAL_COMPLEX(-38, 44), MAKE_VAL_COMPLEX(-20, 30), MAKE_VAL_COMPLEX(-42, 11), MAKE_VAL_COMPLEX(65, 0), MAKE_VAL_COMPLEX(33, -69), MAKE_VAL_COMPLEX(15, -55), MAKE_VAL_COMPLEX(21, -38), MAKE_VAL_COMPLEX(26, -78), MAKE_VAL_COMPLEX(-59, 17), MAKE_VAL_COMPLEX(-41, 68), MAKE_VAL_COMPLEX(-37, -29), MAKE_VAL_COMPLEX(-72, -74), MAKE_VAL_COMPLEX(57, 64), MAKE_VAL_COMPLEX(-82, 71) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1006, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1006", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1006() */
/* END TEST 1006 */

/*
 *  BEGIN TEST 1007
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x960', LAYOUT_COLMAJ>
 *  y : <'y960', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1007( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A960_ptr, A960_ind, A960_val,
		A960_M, A960_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y960_val,
		y960_LEN * y960_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y960_LEN, y960_VW, LAYOUT_COLMAJ, y960_LEN );
	oski_value_t* y_val = array_Duplicate( x960_val,
		x960_LEN * x960_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x960_LEN, x960_VW, LAYOUT_COLMAJ, x960_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(326, 430), MAKE_VAL_COMPLEX(-220, -712), MAKE_VAL_COMPLEX(253, -23), MAKE_VAL_COMPLEX(-554, 604), MAKE_VAL_COMPLEX(284, 136), MAKE_VAL_COMPLEX(-582, -130), MAKE_VAL_COMPLEX(844, 62), MAKE_VAL_COMPLEX(225, -91), MAKE_VAL_COMPLEX(366, -664), MAKE_VAL_COMPLEX(-406, 322), MAKE_VAL_COMPLEX(-977, -161), MAKE_VAL_COMPLEX(237, 557), MAKE_VAL_COMPLEX(-23, -13), MAKE_VAL_COMPLEX(510, 10), MAKE_VAL_COMPLEX(-1184, -892) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1007, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1007", y_true, y_val, x960_LEN*x960_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1007() */
/* END TEST 1007 */

#define A1008_M 5    /* # of rows */
#define A1008_N 5    /* # of cols */
#define A1008_NNZ 9  /* # of stored non-zeros */
static oski_index_t A1008_ptr[A1008_M+1] = {0, 0, 1, 3, 5, 9}; /* end row pointers for A1008 */
static oski_index_t A1008_ind[A1008_NNZ] = {0, 0, 1, 1, 2, 0, 1, 2, 4}; /* end column indices pointers for A1008 */
static oski_value_t A1008_val[A1008_NNZ] = {MAKE_VAL_COMPLEX(-9, 0), MAKE_VAL_COMPLEX(-9, -10), MAKE_VAL_COMPLEX(2, 8), MAKE_VAL_COMPLEX(-10, -2), MAKE_VAL_COMPLEX(0, -2), MAKE_VAL_COMPLEX(-10, -9), MAKE_VAL_COMPLEX(-10, -3), MAKE_VAL_COMPLEX(-4, -4), MAKE_VAL_COMPLEX(10, 10)}; /* end stored non-zero values for A1008 */

#define x1008_LEN 5   /* vector length */
#define x1008_VW  3   /* vector width */
static oski_value_t x1008_val[x1008_LEN*x1008_VW] = {MAKE_VAL_COMPLEX(7, 1), MAKE_VAL_COMPLEX(5, 1), MAKE_VAL_COMPLEX(-1, -3), MAKE_VAL_COMPLEX(5, -10), MAKE_VAL_COMPLEX(-6, 2), MAKE_VAL_COMPLEX(10, -7), MAKE_VAL_COMPLEX(-9, 10), MAKE_VAL_COMPLEX(5, -5), MAKE_VAL_COMPLEX(-1, -7), MAKE_VAL_COMPLEX(-5, 8), MAKE_VAL_COMPLEX(8, -2), MAKE_VAL_COMPLEX(0, 4), MAKE_VAL_COMPLEX(-10, 1), MAKE_VAL_COMPLEX(1, -7), MAKE_VAL_COMPLEX(9, 7)}; /* data */

#define y1008_LEN 5   /* vector length */
#define y1008_VW  3   /* vector width */
static oski_value_t y1008_val[y1008_LEN*y1008_VW] = {MAKE_VAL_COMPLEX(-6, -9), MAKE_VAL_COMPLEX(-3, 2), MAKE_VAL_COMPLEX(-5, 10), MAKE_VAL_COMPLEX(1, -10), MAKE_VAL_COMPLEX(-9, 0), MAKE_VAL_COMPLEX(8, 3), MAKE_VAL_COMPLEX(4, 7), MAKE_VAL_COMPLEX(-5, -7), MAKE_VAL_COMPLEX(9, 3), MAKE_VAL_COMPLEX(0, 8), MAKE_VAL_COMPLEX(-10, -8), MAKE_VAL_COMPLEX(-5, 4), MAKE_VAL_COMPLEX(-1, -10), MAKE_VAL_COMPLEX(10, 9), MAKE_VAL_COMPLEX(8, 8)}; /* data */

/*
 *  BEGIN TEST 1008
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1008( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	oski_value_t* y_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(323, 170), MAKE_VAL_COMPLEX(-114, 12), MAKE_VAL_COMPLEX(-218, -85), MAKE_VAL_COMPLEX(64, -65), MAKE_VAL_COMPLEX(-74, 38), MAKE_VAL_COMPLEX(-48, -96), MAKE_VAL_COMPLEX(39, -17), MAKE_VAL_COMPLEX(-37, -97), MAKE_VAL_COMPLEX(1, 81), MAKE_VAL_COMPLEX(-98, 76), MAKE_VAL_COMPLEX(142, -46), MAKE_VAL_COMPLEX(-132, 6), MAKE_VAL_COMPLEX(-111, -48), MAKE_VAL_COMPLEX(-29, -93), MAKE_VAL_COMPLEX(-222, 287) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1008, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1008", y_true, y_val, y1008_LEN*y1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1008() */
/* END TEST 1008 */

/*
 *  BEGIN TEST 1009
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1009( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	oski_value_t* y_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-356, -161), MAKE_VAL_COMPLEX(117, -1), MAKE_VAL_COMPLEX(243, 110), MAKE_VAL_COMPLEX(-93, 52), MAKE_VAL_COMPLEX(65, -11), MAKE_VAL_COMPLEX(65, 75), MAKE_VAL_COMPLEX(-14, 12), MAKE_VAL_COMPLEX(11, 105), MAKE_VAL_COMPLEX(17, -105), MAKE_VAL_COMPLEX(122, -68), MAKE_VAL_COMPLEX(-176, 68), MAKE_VAL_COMPLEX(139, 13), MAKE_VAL_COMPLEX(80, 41), MAKE_VAL_COMPLEX(66, 72), MAKE_VAL_COMPLEX(254, -303) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1009, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1009", y_true, y_val, y1008_LEN*y1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1009() */
/* END TEST 1009 */

/*
 *  BEGIN TEST 1010
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1010( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	oski_value_t* y_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-48, -33), MAKE_VAL_COMPLEX(-11, 16), MAKE_VAL_COMPLEX(-5, 60), MAKE_VAL_COMPLEX(-15, -52), MAKE_VAL_COMPLEX(-45, 18), MAKE_VAL_COMPLEX(46, -1), MAKE_VAL_COMPLEX(34, 27), MAKE_VAL_COMPLEX(-39, -25), MAKE_VAL_COMPLEX(51, -3), MAKE_VAL_COMPLEX(16, 40), MAKE_VAL_COMPLEX(-66, -20), MAKE_VAL_COMPLEX(-17, 30), MAKE_VAL_COMPLEX(-25, -48), MAKE_VAL_COMPLEX(68, 25), MAKE_VAL_COMPLEX(56, 24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1010, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1010", y_true, y_val, y1008_LEN*y1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1010() */
/* END TEST 1010 */

/*
 *  BEGIN TEST 1011
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1011( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	oski_value_t* y_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1549, -652), MAKE_VAL_COMPLEX(-131, 696), MAKE_VAL_COMPLEX(-534, 845), MAKE_VAL_COMPLEX(-594, -349), MAKE_VAL_COMPLEX(157, 880), MAKE_VAL_COMPLEX(-660, -463), MAKE_VAL_COMPLEX(256, -569), MAKE_VAL_COMPLEX(-799, 261), MAKE_VAL_COMPLEX(351, -221), MAKE_VAL_COMPLEX(438, 550), MAKE_VAL_COMPLEX(54, -240), MAKE_VAL_COMPLEX(-85, 814), MAKE_VAL_COMPLEX(-992, 665), MAKE_VAL_COMPLEX(-366, -763), MAKE_VAL_COMPLEX(953, 1451) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1011, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1011", y_true, y_val, y1008_LEN*y1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1011() */
/* END TEST 1011 */

/*
 *  BEGIN TEST 1012
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1012( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-98, 47), MAKE_VAL_COMPLEX(8, -104), MAKE_VAL_COMPLEX(-142, -269), MAKE_VAL_COMPLEX(-93, 190), MAKE_VAL_COMPLEX(138, -70), MAKE_VAL_COMPLEX(-71, -157), MAKE_VAL_COMPLEX(149, 157), MAKE_VAL_COMPLEX(-56, -111), MAKE_VAL_COMPLEX(110, 31), MAKE_VAL_COMPLEX(67, 64), MAKE_VAL_COMPLEX(8, -6), MAKE_VAL_COMPLEX(-44, -92), MAKE_VAL_COMPLEX(117, 140), MAKE_VAL_COMPLEX(91, 315), MAKE_VAL_COMPLEX(24, -100) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1012, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1012", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1012() */
/* END TEST 1012 */

/*
 *  BEGIN TEST 1013
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1013( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(66, -73), MAKE_VAL_COMPLEX(-30, 84), MAKE_VAL_COMPLEX(138, 287), MAKE_VAL_COMPLEX(38, -155), MAKE_VAL_COMPLEX(-102, 78), MAKE_VAL_COMPLEX(0, 162), MAKE_VAL_COMPLEX(-74, -180), MAKE_VAL_COMPLEX(16, 121), MAKE_VAL_COMPLEX(-126, 7), MAKE_VAL_COMPLEX(-18, -89), MAKE_VAL_COMPLEX(-54, -8), MAKE_VAL_COMPLEX(56, 72), MAKE_VAL_COMPLEX(-64, -115), MAKE_VAL_COMPLEX(-117, -283), MAKE_VAL_COMPLEX(-48, 38) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1013, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1013", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1013() */
/* END TEST 1013 */

/*
 *  BEGIN TEST 1014
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1014( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-21, 47), MAKE_VAL_COMPLEX(-17, 33), MAKE_VAL_COMPLEX(23, -1), MAKE_VAL_COMPLEX(60, 55), MAKE_VAL_COMPLEX(-2, -46), MAKE_VAL_COMPLEX(29, 84), MAKE_VAL_COMPLEX(-52, -83), MAKE_VAL_COMPLEX(25, 45), MAKE_VAL_COMPLEX(51, 7), MAKE_VAL_COMPLEX(-46, -51), MAKE_VAL_COMPLEX(-2, 60), MAKE_VAL_COMPLEX(-28, -8), MAKE_VAL_COMPLEX(13, -72), MAKE_VAL_COMPLEX(47, 21), MAKE_VAL_COMPLEX(-67, 49) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1014, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1014", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1014() */
/* END TEST 1014 */

/*
 *  BEGIN TEST 1015
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1015( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-361, 930), MAKE_VAL_COMPLEX(159, -711), MAKE_VAL_COMPLEX(-1777, -2370), MAKE_VAL_COMPLEX(223, 1411), MAKE_VAL_COMPLEX(636, -984), MAKE_VAL_COMPLEX(-48, -1383), MAKE_VAL_COMPLEX(711, 1460), MAKE_VAL_COMPLEX(-221, -1042), MAKE_VAL_COMPLEX(1231, -438), MAKE_VAL_COMPLEX(121, 795), MAKE_VAL_COMPLEX(658, 78), MAKE_VAL_COMPLEX(-720, -476), MAKE_VAL_COMPLEX(638, 742), MAKE_VAL_COMPLEX(1760, 2226), MAKE_VAL_COMPLEX(365, -171) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(9, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1015, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1015", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1015() */
/* END TEST 1015 */

/*
 *  BEGIN TEST 1016
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1016( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1, 196), MAKE_VAL_COMPLEX(25, 41), MAKE_VAL_COMPLEX(-323, 14), MAKE_VAL_COMPLEX(207, 75), MAKE_VAL_COMPLEX(-78, -16), MAKE_VAL_COMPLEX(90, -233), MAKE_VAL_COMPLEX(17, 14), MAKE_VAL_COMPLEX(-31, 18), MAKE_VAL_COMPLEX(-55, -234), MAKE_VAL_COMPLEX(-59, 109), MAKE_VAL_COMPLEX(138, 6), MAKE_VAL_COMPLEX(-112, 16), MAKE_VAL_COMPLEX(-28, -10), MAKE_VAL_COMPLEX(378, -92), MAKE_VAL_COMPLEX(-25, -61) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1016, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1016", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1016() */
/* END TEST 1016 */

/*
 *  BEGIN TEST 1017
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1017( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, -203), MAKE_VAL_COMPLEX(-24, -46), MAKE_VAL_COMPLEX(320, -13), MAKE_VAL_COMPLEX(-217, -80), MAKE_VAL_COMPLEX(80, 22), MAKE_VAL_COMPLEX(-97, 223), MAKE_VAL_COMPLEX(-7, -5), MAKE_VAL_COMPLEX(26, -23), MAKE_VAL_COMPLEX(48, 235), MAKE_VAL_COMPLEX(67, -104), MAKE_VAL_COMPLEX(-140, -14), MAKE_VAL_COMPLEX(116, -16), MAKE_VAL_COMPLEX(29, 20), MAKE_VAL_COMPLEX(-385, 91), MAKE_VAL_COMPLEX(32, 52) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1017, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1017", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1017() */
/* END TEST 1017 */

/*
 *  BEGIN TEST 1018
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1018( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(24, 32), MAKE_VAL_COMPLEX(16, 24), MAKE_VAL_COMPLEX(8, -16), MAKE_VAL_COMPLEX(60, -20), MAKE_VAL_COMPLEX(-32, -16), MAKE_VAL_COMPLEX(68, 12), MAKE_VAL_COMPLEX(-76, 4), MAKE_VAL_COMPLEX(40, 0), MAKE_VAL_COMPLEX(24, -32), MAKE_VAL_COMPLEX(-52, 12), MAKE_VAL_COMPLEX(40, 24), MAKE_VAL_COMPLEX(-16, 16), MAKE_VAL_COMPLEX(-44, -36), MAKE_VAL_COMPLEX(32, -24), MAKE_VAL_COMPLEX(8, 64) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1018, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1018", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1018() */
/* END TEST 1018 */

/*
 *  BEGIN TEST 1019
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1019( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1178, 1193), MAKE_VAL_COMPLEX(224, -52), MAKE_VAL_COMPLEX(-2138, 2875), MAKE_VAL_COMPLEX(1594, -621), MAKE_VAL_COMPLEX(-288, 490), MAKE_VAL_COMPLEX(-1936, -2098), MAKE_VAL_COMPLEX(922, -484), MAKE_VAL_COMPLEX(-448, 589), MAKE_VAL_COMPLEX(-2310, -769), MAKE_VAL_COMPLEX(882, 843), MAKE_VAL_COMPLEX(510, -1072), MAKE_VAL_COMPLEX(-592, 772), MAKE_VAL_COMPLEX(330, 255), MAKE_VAL_COMPLEX(1739, -3271), MAKE_VAL_COMPLEX(-1082, -784) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1019, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1019", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1019() */
/* END TEST 1019 */

/*
 *  BEGIN TEST 1020
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1020( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	oski_value_t* y_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(263, 119), MAKE_VAL_COMPLEX(-9, -117), MAKE_VAL_COMPLEX(42, -3), MAKE_VAL_COMPLEX(-7, 82), MAKE_VAL_COMPLEX(-202, -118), MAKE_VAL_COMPLEX(-102, 79), MAKE_VAL_COMPLEX(-126, 129), MAKE_VAL_COMPLEX(-86, -136), MAKE_VAL_COMPLEX(69, 27), MAKE_VAL_COMPLEX(33, -93), MAKE_VAL_COMPLEX(-176, -79), MAKE_VAL_COMPLEX(-29, -140), MAKE_VAL_COMPLEX(92, 5), MAKE_VAL_COMPLEX(-134, 119), MAKE_VAL_COMPLEX(-150, 327) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1020, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1020", y_true, y_val, y1008_LEN*y1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1020() */
/* END TEST 1020 */

/*
 *  BEGIN TEST 1021
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1021( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	oski_value_t* y_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-323, -170), MAKE_VAL_COMPLEX(-8, 137), MAKE_VAL_COMPLEX(-57, 83), MAKE_VAL_COMPLEX(-6, -154), MAKE_VAL_COMPLEX(139, 136), MAKE_VAL_COMPLEX(164, -74), MAKE_VAL_COMPLEX(168, -88), MAKE_VAL_COMPLEX(37, 97), MAKE_VAL_COMPLEX(0, -24), MAKE_VAL_COMPLEX(-17, 149), MAKE_VAL_COMPLEX(90, 43), MAKE_VAL_COMPLEX(2, 178), MAKE_VAL_COMPLEX(-119, -73), MAKE_VAL_COMPLEX(222, -76), MAKE_VAL_COMPLEX(222, -287) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1021, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1021", y_true, y_val, y1008_LEN*y1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1021() */
/* END TEST 1021 */

/*
 *  BEGIN TEST 1022
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1022( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	oski_value_t* y_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-6, -9), MAKE_VAL_COMPLEX(-3, 2), MAKE_VAL_COMPLEX(-5, 10), MAKE_VAL_COMPLEX(1, -10), MAKE_VAL_COMPLEX(-9, 0), MAKE_VAL_COMPLEX(8, 3), MAKE_VAL_COMPLEX(4, 7), MAKE_VAL_COMPLEX(-5, -7), MAKE_VAL_COMPLEX(9, 3), MAKE_VAL_COMPLEX(0, 8), MAKE_VAL_COMPLEX(-10, -8), MAKE_VAL_COMPLEX(-5, 4), MAKE_VAL_COMPLEX(-1, -10), MAKE_VAL_COMPLEX(10, 9), MAKE_VAL_COMPLEX(8, 8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1022, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1022", y_true, y_val, y1008_LEN*y1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1022() */
/* END TEST 1022 */

/*
 *  BEGIN TEST 1023
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1023( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	oski_value_t* y_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(2730, 398), MAKE_VAL_COMPLEX(-374, -950), MAKE_VAL_COMPLEX(694, -582), MAKE_VAL_COMPLEX(86, 1130), MAKE_VAL_COMPLEX(-1674, -10), MAKE_VAL_COMPLEX(-826, 790), MAKE_VAL_COMPLEX(-578, 1158), MAKE_VAL_COMPLEX(-1100, -352), MAKE_VAL_COMPLEX(288, -324), MAKE_VAL_COMPLEX(4, -1244), MAKE_VAL_COMPLEX(-1376, 580), MAKE_VAL_COMPLEX(-506, -1118), MAKE_VAL_COMPLEX(662, 146), MAKE_VAL_COMPLEX(-930, 934), MAKE_VAL_COMPLEX(-148, 2736) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, -4);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1023, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1023", y_true, y_val, y1008_LEN*y1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1023() */
/* END TEST 1023 */

/*
 *  BEGIN TEST 1024
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1024( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(263, -34), MAKE_VAL_COMPLEX(14, -71), MAKE_VAL_COMPLEX(-77, -77), MAKE_VAL_COMPLEX(-61, 156), MAKE_VAL_COMPLEX(-64, -163), MAKE_VAL_COMPLEX(-36, -243), MAKE_VAL_COMPLEX(47, 217), MAKE_VAL_COMPLEX(-97, -146), MAKE_VAL_COMPLEX(-69, 71), MAKE_VAL_COMPLEX(123, 28), MAKE_VAL_COMPLEX(-64, -130), MAKE_VAL_COMPLEX(70, -32), MAKE_VAL_COMPLEX(3, 102), MAKE_VAL_COMPLEX(-217, -57), MAKE_VAL_COMPLEX(101, 270) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1024, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1024", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1024() */
/* END TEST 1024 */

/*
 *  BEGIN TEST 1025
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1025( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-263, -66), MAKE_VAL_COMPLEX(-10, -1), MAKE_VAL_COMPLEX(37, 97), MAKE_VAL_COMPLEX(-89, -206), MAKE_VAL_COMPLEX(104, 243), MAKE_VAL_COMPLEX(-82, 117), MAKE_VAL_COMPLEX(111, -111), MAKE_VAL_COMPLEX(17, 86), MAKE_VAL_COMPLEX(-27, -43), MAKE_VAL_COMPLEX(-1, 26), MAKE_VAL_COMPLEX(20, 22), MAKE_VAL_COMPLEX(-14, 24), MAKE_VAL_COMPLEX(31, 36), MAKE_VAL_COMPLEX(117, 57), MAKE_VAL_COMPLEX(-21, -410) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1025, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1025", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1025() */
/* END TEST 1025 */

/*
 *  BEGIN TEST 1026
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1026( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(3, -21), MAKE_VAL_COMPLEX(3, -15), MAKE_VAL_COMPLEX(-9, 3), MAKE_VAL_COMPLEX(-30, -15), MAKE_VAL_COMPLEX(6, 18), MAKE_VAL_COMPLEX(-21, -30), MAKE_VAL_COMPLEX(30, 27), MAKE_VAL_COMPLEX(-15, -15), MAKE_VAL_COMPLEX(-21, 3), MAKE_VAL_COMPLEX(24, 15), MAKE_VAL_COMPLEX(-6, -24), MAKE_VAL_COMPLEX(12, 0), MAKE_VAL_COMPLEX(3, 30), MAKE_VAL_COMPLEX(-21, -3), MAKE_VAL_COMPLEX(21, -27) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1026, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1026", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1026() */
/* END TEST 1026 */

/*
 *  BEGIN TEST 1027
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1027( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-2312, 1635), MAKE_VAL_COMPLEX(83, 304), MAKE_VAL_COMPLEX(1053, 351), MAKE_VAL_COMPLEX(-1544, -1245), MAKE_VAL_COMPLEX(2201, 1060), MAKE_VAL_COMPLEX(863, 1616), MAKE_VAL_COMPLEX(-667, -1605), MAKE_VAL_COMPLEX(1148, 559), MAKE_VAL_COMPLEX(-279, -489), MAKE_VAL_COMPLEX(-356, 351), MAKE_VAL_COMPLEX(730, 272), MAKE_VAL_COMPLEX(-104, 458), MAKE_VAL_COMPLEX(40, -287), MAKE_VAL_COMPLEX(1657, -617), MAKE_VAL_COMPLEX(-2940, -2479) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1027, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1027", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1027() */
/* END TEST 1027 */

/*
 *  BEGIN TEST 1028
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1028( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(106, -197), MAKE_VAL_COMPLEX(37, -98), MAKE_VAL_COMPLEX(10, 10), MAKE_VAL_COMPLEX(309, 171), MAKE_VAL_COMPLEX(-310, -109), MAKE_VAL_COMPLEX(-109, -65), MAKE_VAL_COMPLEX(97, 40), MAKE_VAL_COMPLEX(-16, 11), MAKE_VAL_COMPLEX(138, -14), MAKE_VAL_COMPLEX(-69, -5), MAKE_VAL_COMPLEX(26, -46), MAKE_VAL_COMPLEX(42, -24), MAKE_VAL_COMPLEX(-20, -6), MAKE_VAL_COMPLEX(2, 28), MAKE_VAL_COMPLEX(442, 65) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1028, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1028", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1028() */
/* END TEST 1028 */

/*
 *  BEGIN TEST 1029
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1029( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-15, 260), MAKE_VAL_COMPLEX(26, 147), MAKE_VAL_COMPLEX(-3, -59), MAKE_VAL_COMPLEX(-169, -276), MAKE_VAL_COMPLEX(212, 95), MAKE_VAL_COMPLEX(298, 37), MAKE_VAL_COMPLEX(-293, 37), MAKE_VAL_COMPLEX(121, -46), MAKE_VAL_COMPLEX(-103, -91), MAKE_VAL_COMPLEX(-57, 82), MAKE_VAL_COMPLEX(100, 74), MAKE_VAL_COMPLEX(-70, 80), MAKE_VAL_COMPLEX(-127, -50), MAKE_VAL_COMPLEX(61, -119), MAKE_VAL_COMPLEX(-365, 96) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1029, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1029", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1029() */
/* END TEST 1029 */

/*
 *  BEGIN TEST 1030
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1030( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-33, 31), MAKE_VAL_COMPLEX(-25, 21), MAKE_VAL_COMPLEX(19, 7), MAKE_VAL_COMPLEX(30, 65), MAKE_VAL_COMPLEX(14, -38), MAKE_VAL_COMPLEX(-5, 78), MAKE_VAL_COMPLEX(-14, -85), MAKE_VAL_COMPLEX(5, 45), MAKE_VAL_COMPLEX(39, 23), MAKE_VAL_COMPLEX(-20, -57), MAKE_VAL_COMPLEX(-22, 48), MAKE_VAL_COMPLEX(-20, -16), MAKE_VAL_COMPLEX(35, -54), MAKE_VAL_COMPLEX(31, 33), MAKE_VAL_COMPLEX(-71, 17) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1030, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1030", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1030() */
/* END TEST 1030 */

/*
 *  BEGIN TEST 1031
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_ROWMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1031( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_ROWMAJ, x1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1347, 1474), MAKE_VAL_COMPLEX(-540, 919), MAKE_VAL_COMPLEX(141, -287), MAKE_VAL_COMPLEX(-609, -2448), MAKE_VAL_COMPLEX(1380, 1661), MAKE_VAL_COMPLEX(1346, 1329), MAKE_VAL_COMPLEX(-1533, -929), MAKE_VAL_COMPLEX(681, 152), MAKE_VAL_COMPLEX(-575, -775), MAKE_VAL_COMPLEX(-269, 254), MAKE_VAL_COMPLEX(60, 698), MAKE_VAL_COMPLEX(-666, 144), MAKE_VAL_COMPLEX(-313, -532), MAKE_VAL_COMPLEX(617, -387), MAKE_VAL_COMPLEX(-2993, -1350) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, -4);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1031, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1031", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1031() */
/* END TEST 1031 */

/*
 *  BEGIN TEST 1032
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1032( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	oski_value_t* y_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-45, 152), MAKE_VAL_COMPLEX(138, -111), MAKE_VAL_COMPLEX(183, -91), MAKE_VAL_COMPLEX(-139, 96), MAKE_VAL_COMPLEX(66, 181), MAKE_VAL_COMPLEX(-198, -173), MAKE_VAL_COMPLEX(8, -95), MAKE_VAL_COMPLEX(-263, -26), MAKE_VAL_COMPLEX(-155, -213), MAKE_VAL_COMPLEX(18, -50), MAKE_VAL_COMPLEX(68, 30), MAKE_VAL_COMPLEX(66, 9), MAKE_VAL_COMPLEX(-282, -73), MAKE_VAL_COMPLEX(-172, -189), MAKE_VAL_COMPLEX(18, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1032, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1032", y_true, y_val, y1008_LEN*y1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1032() */
/* END TEST 1032 */

/*
 *  BEGIN TEST 1033
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1033( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	oski_value_t* y_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(60, -71), MAKE_VAL_COMPLEX(-111, 106), MAKE_VAL_COMPLEX(-118, 36), MAKE_VAL_COMPLEX(102, -29), MAKE_VAL_COMPLEX(-3, -154), MAKE_VAL_COMPLEX(151, 128), MAKE_VAL_COMPLEX(-15, 34), MAKE_VAL_COMPLEX(277, 90), MAKE_VAL_COMPLEX(101, 165), MAKE_VAL_COMPLEX(6, -6), MAKE_VAL_COMPLEX(-22, 56), MAKE_VAL_COMPLEX(-19, -22), MAKE_VAL_COMPLEX(259, 146), MAKE_VAL_COMPLEX(129, 96), MAKE_VAL_COMPLEX(-50, -80) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1033, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1033", y_true, y_val, y1008_LEN*y1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1033() */
/* END TEST 1033 */

/*
 *  BEGIN TEST 1034
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1034( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	oski_value_t* y_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-3, -24), MAKE_VAL_COMPLEX(-8, 1), MAKE_VAL_COMPLEX(-20, 15), MAKE_VAL_COMPLEX(12, -19), MAKE_VAL_COMPLEX(-18, -9), MAKE_VAL_COMPLEX(13, 14), MAKE_VAL_COMPLEX(1, 18), MAKE_VAL_COMPLEX(-3, -19), MAKE_VAL_COMPLEX(15, 15), MAKE_VAL_COMPLEX(-8, 16), MAKE_VAL_COMPLEX(-12, -26), MAKE_VAL_COMPLEX(-14, 3), MAKE_VAL_COMPLEX(8, -21), MAKE_VAL_COMPLEX(11, 28), MAKE_VAL_COMPLEX(8, 24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1034, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1034", y_true, y_val, y1008_LEN*y1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1034() */
/* END TEST 1034 */

/*
 *  BEGIN TEST 1035
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1035( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	oski_value_t* y_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-494, 238), MAKE_VAL_COMPLEX(1136, 524), MAKE_VAL_COMPLEX(888, 379), MAKE_VAL_COMPLEX(-623, -213), MAKE_VAL_COMPLEX(-658, 482), MAKE_VAL_COMPLEX(277, -1587), MAKE_VAL_COMPLEX(203, -353), MAKE_VAL_COMPLEX(187, -1712), MAKE_VAL_COMPLEX(590, -1328), MAKE_VAL_COMPLEX(82, -398), MAKE_VAL_COMPLEX(760, 580), MAKE_VAL_COMPLEX(182, 78), MAKE_VAL_COMPLEX(380, -1628), MAKE_VAL_COMPLEX(107, -1679), MAKE_VAL_COMPLEX(-692, 6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1035, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1035", y_true, y_val, y1008_LEN*y1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1035() */
/* END TEST 1035 */

/*
 *  BEGIN TEST 1036
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1036( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-103, 132), MAKE_VAL_COMPLEX(-34, 174), MAKE_VAL_COMPLEX(50, 185), MAKE_VAL_COMPLEX(14, 180), MAKE_VAL_COMPLEX(65, 39), MAKE_VAL_COMPLEX(19, 35), MAKE_VAL_COMPLEX(78, -206), MAKE_VAL_COMPLEX(9, -56), MAKE_VAL_COMPLEX(124, 64), MAKE_VAL_COMPLEX(122, 194), MAKE_VAL_COMPLEX(-167, -236), MAKE_VAL_COMPLEX(17, -170), MAKE_VAL_COMPLEX(199, -90), MAKE_VAL_COMPLEX(-32, -16), MAKE_VAL_COMPLEX(-51, 15) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1036, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1036", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1036() */
/* END TEST 1036 */

/*
 *  BEGIN TEST 1037
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1037( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-5, -126), MAKE_VAL_COMPLEX(-44, -174), MAKE_VAL_COMPLEX(-26, -143), MAKE_VAL_COMPLEX(-59, -15), MAKE_VAL_COMPLEX(19, -87), MAKE_VAL_COMPLEX(-148, 100), MAKE_VAL_COMPLEX(27, 29), MAKE_VAL_COMPLEX(-69, 146), MAKE_VAL_COMPLEX(-88, 38), MAKE_VAL_COMPLEX(-71, -329), MAKE_VAL_COMPLEX(53, 290), MAKE_VAL_COMPLEX(-29, 110), MAKE_VAL_COMPLEX(-52, 45), MAKE_VAL_COMPLEX(38, 124), MAKE_VAL_COMPLEX(-105, -93) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1037, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1037", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1037() */
/* END TEST 1037 */

/*
 *  BEGIN TEST 1038
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1038( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(28, 4), MAKE_VAL_COMPLEX(20, 4), MAKE_VAL_COMPLEX(-4, -12), MAKE_VAL_COMPLEX(20, -40), MAKE_VAL_COMPLEX(-24, 8), MAKE_VAL_COMPLEX(40, -28), MAKE_VAL_COMPLEX(-36, 40), MAKE_VAL_COMPLEX(20, -20), MAKE_VAL_COMPLEX(-4, -28), MAKE_VAL_COMPLEX(-20, 32), MAKE_VAL_COMPLEX(32, -8), MAKE_VAL_COMPLEX(0, 16), MAKE_VAL_COMPLEX(-40, 4), MAKE_VAL_COMPLEX(4, -28), MAKE_VAL_COMPLEX(36, 28) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1038, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1038", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1038() */
/* END TEST 1038 */

/*
 *  BEGIN TEST 1039
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1039( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-110, 142), MAKE_VAL_COMPLEX(6, 264), MAKE_VAL_COMPLEX(32, 356), MAKE_VAL_COMPLEX(-107, 155), MAKE_VAL_COMPLEX(102, 214), MAKE_VAL_COMPLEX(15, -197), MAKE_VAL_COMPLEX(247, -131), MAKE_VAL_COMPLEX(-22, -262), MAKE_VAL_COMPLEX(104, 70), MAKE_VAL_COMPLEX(341, 571), MAKE_VAL_COMPLEX(-284, -646), MAKE_VAL_COMPLEX(110, -296), MAKE_VAL_COMPLEX(307, 21), MAKE_VAL_COMPLEX(-186, -128), MAKE_VAL_COMPLEX(130, -64) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1039, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1039", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1039() */
/* END TEST 1039 */

/*
 *  BEGIN TEST 1040
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1040( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(5, 118), MAKE_VAL_COMPLEX(158, 52), MAKE_VAL_COMPLEX(74, 29), MAKE_VAL_COMPLEX(-54, 60), MAKE_VAL_COMPLEX(7, 77), MAKE_VAL_COMPLEX(-7, -61), MAKE_VAL_COMPLEX(-8, 82), MAKE_VAL_COMPLEX(-91, -32), MAKE_VAL_COMPLEX(60, -36), MAKE_VAL_COMPLEX(378, -30), MAKE_VAL_COMPLEX(-331, -24), MAKE_VAL_COMPLEX(49, -254), MAKE_VAL_COMPLEX(-101, -146), MAKE_VAL_COMPLEX(-132, -16), MAKE_VAL_COMPLEX(43, -167) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1040, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1040", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1040() */
/* END TEST 1040 */

/*
 *  BEGIN TEST 1041
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1041( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-26, -171), MAKE_VAL_COMPLEX(-171, -91), MAKE_VAL_COMPLEX(-91, -10), MAKE_VAL_COMPLEX(-36, -55), MAKE_VAL_COMPLEX(31, -43), MAKE_VAL_COMPLEX(-82, 19), MAKE_VAL_COMPLEX(114, -59), MAKE_VAL_COMPLEX(36, 17), MAKE_VAL_COMPLEX(-105, 71), MAKE_VAL_COMPLEX(-302, 33), MAKE_VAL_COMPLEX(285, -24), MAKE_VAL_COMPLEX(-21, 238), MAKE_VAL_COMPLEX(148, 212), MAKE_VAL_COMPLEX(79, 37), MAKE_VAL_COMPLEX(-30, 76) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1041, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1041", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1041() */
/* END TEST 1041 */

/*
 *  BEGIN TEST 1042
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1042( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(4, -28), MAKE_VAL_COMPLEX(4, -20), MAKE_VAL_COMPLEX(-12, 4), MAKE_VAL_COMPLEX(-40, -20), MAKE_VAL_COMPLEX(8, 24), MAKE_VAL_COMPLEX(-28, -40), MAKE_VAL_COMPLEX(40, 36), MAKE_VAL_COMPLEX(-20, -20), MAKE_VAL_COMPLEX(-28, 4), MAKE_VAL_COMPLEX(32, 20), MAKE_VAL_COMPLEX(-8, -32), MAKE_VAL_COMPLEX(16, 0), MAKE_VAL_COMPLEX(4, 40), MAKE_VAL_COMPLEX(-28, -4), MAKE_VAL_COMPLEX(28, -36) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1042, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1042", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1042() */
/* END TEST 1042 */

/*
 *  BEGIN TEST 1043
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1043( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_ROWMAJ, y1008_VW );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1350, 138), MAKE_VAL_COMPLEX(-394, 1252), MAKE_VAL_COMPLEX(-56, 844), MAKE_VAL_COMPLEX(-1003, 243), MAKE_VAL_COMPLEX(-214, 154), MAKE_VAL_COMPLEX(-213, 139), MAKE_VAL_COMPLEX(-13, -411), MAKE_VAL_COMPLEX(-286, -542), MAKE_VAL_COMPLEX(392, 846), MAKE_VAL_COMPLEX(1389, 2531), MAKE_VAL_COMPLEX(-980, -2662), MAKE_VAL_COMPLEX(2174, -360), MAKE_VAL_COMPLEX(1587, -1051), MAKE_VAL_COMPLEX(-274, -672), MAKE_VAL_COMPLEX(950, -516) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1043, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1043", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1043() */
/* END TEST 1043 */

/*
 *  BEGIN TEST 1044
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1044( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	oski_value_t* y_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(66, 26), MAKE_VAL_COMPLEX(-57, 47), MAKE_VAL_COMPLEX(-99, -61), MAKE_VAL_COMPLEX(6, -12), MAKE_VAL_COMPLEX(-196, -176), MAKE_VAL_COMPLEX(90, -82), MAKE_VAL_COMPLEX(-12, 124), MAKE_VAL_COMPLEX(-178, -129), MAKE_VAL_COMPLEX(82, -38), MAKE_VAL_COMPLEX(-261, -63), MAKE_VAL_COMPLEX(121, -156), MAKE_VAL_COMPLEX(-217, -119), MAKE_VAL_COMPLEX(-86, -126), MAKE_VAL_COMPLEX(-44, 40), MAKE_VAL_COMPLEX(-70, 152) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1044, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1044", y_true, y_val, y1008_LEN*y1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1044() */
/* END TEST 1044 */

/*
 *  BEGIN TEST 1045
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1045( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	oski_value_t* y_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-105, -104), MAKE_VAL_COMPLEX(31, -34), MAKE_VAL_COMPLEX(49, 136), MAKE_VAL_COMPLEX(12, -67), MAKE_VAL_COMPLEX(124, 167), MAKE_VAL_COMPLEX(-29, 114), MAKE_VAL_COMPLEX(37, -64), MAKE_VAL_COMPLEX(145, 68), MAKE_VAL_COMPLEX(-13, 71), MAKE_VAL_COMPLEX(253, 127), MAKE_VAL_COMPLEX(-193, 82), MAKE_VAL_COMPLEX(173, 146), MAKE_VAL_COMPLEX(88, 45), MAKE_VAL_COMPLEX(115, 42), MAKE_VAL_COMPLEX(126, -80) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1045, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1045", y_true, y_val, y1008_LEN*y1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1045() */
/* END TEST 1045 */

/*
 *  BEGIN TEST 1046
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1046( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	oski_value_t* y_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(87, 33), MAKE_VAL_COMPLEX(11, -29), MAKE_VAL_COMPLEX(-15, -95), MAKE_VAL_COMPLEX(43, 75), MAKE_VAL_COMPLEX(63, -45), MAKE_VAL_COMPLEX(-71, 19), MAKE_VAL_COMPLEX(-63, -29), MAKE_VAL_COMPLEX(70, 24), MAKE_VAL_COMPLEX(-78, 24), MAKE_VAL_COMPLEX(-40, -56), MAKE_VAL_COMPLEX(110, 6), MAKE_VAL_COMPLEX(15, -53), MAKE_VAL_COMPLEX(57, 65), MAKE_VAL_COMPLEX(-115, -13), MAKE_VAL_COMPLEX(-96, -16) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1046, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1046", y_true, y_val, y1008_LEN*y1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1046() */
/* END TEST 1046 */

/*
 *  BEGIN TEST 1047
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1047( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	oski_value_t* y_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-130, 479), MAKE_VAL_COMPLEX(-438, 395), MAKE_VAL_COMPLEX(-146, -225), MAKE_VAL_COMPLEX(-419, -272), MAKE_VAL_COMPLEX(-1155, -1174), MAKE_VAL_COMPLEX(1064, -749), MAKE_VAL_COMPLEX(78, 749), MAKE_VAL_COMPLEX(-1413, -1114), MAKE_VAL_COMPLEX(935, -507), MAKE_VAL_COMPLEX(-1301, -843), MAKE_VAL_COMPLEX(609, -470), MAKE_VAL_COMPLEX(-1156, -957), MAKE_VAL_COMPLEX(-865, -1154), MAKE_VAL_COMPLEX(212, -173), MAKE_VAL_COMPLEX(-274, 644) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1047, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1047", y_true, y_val, y1008_LEN*y1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1047() */
/* END TEST 1047 */

/*
 *  BEGIN TEST 1048
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1048( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(261, -20), MAKE_VAL_COMPLEX(25, 155), MAKE_VAL_COMPLEX(-59, 164), MAKE_VAL_COMPLEX(-11, -24), MAKE_VAL_COMPLEX(3, 57), MAKE_VAL_COMPLEX(-41, -83), MAKE_VAL_COMPLEX(-17, -165), MAKE_VAL_COMPLEX(-87, -136), MAKE_VAL_COMPLEX(-81, -55), MAKE_VAL_COMPLEX(-107, -34), MAKE_VAL_COMPLEX(-74, -134), MAKE_VAL_COMPLEX(54, -174), MAKE_VAL_COMPLEX(2, 115), MAKE_VAL_COMPLEX(-5, -27), MAKE_VAL_COMPLEX(87, 288) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1048, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1048", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1048() */
/* END TEST 1048 */

/*
 *  BEGIN TEST 1049
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1049( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-275, -82), MAKE_VAL_COMPLEX(-31, -229), MAKE_VAL_COMPLEX(21, -138), MAKE_VAL_COMPLEX(-149, -6), MAKE_VAL_COMPLEX(49, 19), MAKE_VAL_COMPLEX(-97, -29), MAKE_VAL_COMPLEX(193, 251), MAKE_VAL_COMPLEX(-3, 86), MAKE_VAL_COMPLEX(-13, 97), MAKE_VAL_COMPLEX(239, 72), MAKE_VAL_COMPLEX(14, 30), MAKE_VAL_COMPLEX(2, 158), MAKE_VAL_COMPLEX(52, 21), MAKE_VAL_COMPLEX(-97, 41), MAKE_VAL_COMPLEX(-25, -442) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1049, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1049", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1049() */
/* END TEST 1049 */

/*
 *  BEGIN TEST 1050
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1050( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(5, 15), MAKE_VAL_COMPLEX(3, 11), MAKE_VAL_COMPLEX(5, -5), MAKE_VAL_COMPLEX(25, 0), MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(24, 13), MAKE_VAL_COMPLEX(-29, -8), MAKE_VAL_COMPLEX(15, 5), MAKE_VAL_COMPLEX(13, -9), MAKE_VAL_COMPLEX(-21, -2), MAKE_VAL_COMPLEX(12, 14), MAKE_VAL_COMPLEX(-8, 4), MAKE_VAL_COMPLEX(-12, -19), MAKE_VAL_COMPLEX(15, -5), MAKE_VAL_COMPLEX(-5, 25) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1050, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1050", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1050() */
/* END TEST 1050 */

/*
 *  BEGIN TEST 1051
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1051( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(72, 1910), MAKE_VAL_COMPLEX(-1301, 391), MAKE_VAL_COMPLEX(-1100, -138), MAKE_VAL_COMPLEX(147, 444), MAKE_VAL_COMPLEX(-174, -136), MAKE_VAL_COMPLEX(247, 148), MAKE_VAL_COMPLEX(1324, -913), MAKE_VAL_COMPLEX(750, -420), MAKE_VAL_COMPLEX(495, -335), MAKE_VAL_COMPLEX(183, -1240), MAKE_VAL_COMPLEX(554, -396), MAKE_VAL_COMPLEX(1188, 28), MAKE_VAL_COMPLEX(-384, -125), MAKE_VAL_COMPLEX(287, 267), MAKE_VAL_COMPLEX(-2472, 778) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1051, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1051", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1051() */
/* END TEST 1051 */

/*
 *  BEGIN TEST 1052
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1052( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(20, -195), MAKE_VAL_COMPLEX(192, 246), MAKE_VAL_COMPLEX(236, 23), MAKE_VAL_COMPLEX(51, 49), MAKE_VAL_COMPLEX(71, -15), MAKE_VAL_COMPLEX(6, -25), MAKE_VAL_COMPLEX(-283, -206), MAKE_VAL_COMPLEX(-66, 81), MAKE_VAL_COMPLEX(14, -44), MAKE_VAL_COMPLEX(-71, -3), MAKE_VAL_COMPLEX(-24, 102), MAKE_VAL_COMPLEX(-242, -86), MAKE_VAL_COMPLEX(153, -35), MAKE_VAL_COMPLEX(106, -10), MAKE_VAL_COMPLEX(320, -1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1052, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1052", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1052() */
/* END TEST 1052 */

/*
 *  BEGIN TEST 1053
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1053( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-123, 216), MAKE_VAL_COMPLEX(-267, -235), MAKE_VAL_COMPLEX(-207, 14), MAKE_VAL_COMPLEX(-71, 116), MAKE_VAL_COMPLEX(3, -43), MAKE_VAL_COMPLEX(-111, 173), MAKE_VAL_COMPLEX(359, 21), MAKE_VAL_COMPLEX(21, 14), MAKE_VAL_COMPLEX(35, 137), MAKE_VAL_COMPLEX(101, -134), MAKE_VAL_COMPLEX(-78, -34), MAKE_VAL_COMPLEX(222, 30), MAKE_VAL_COMPLEX(-18, -29), MAKE_VAL_COMPLEX(-85, 113), MAKE_VAL_COMPLEX(-481, -52) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1053, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1053", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1053() */
/* END TEST 1053 */

/*
 *  BEGIN TEST 1054
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1054( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(8, 44), MAKE_VAL_COMPLEX(4, 32), MAKE_VAL_COMPLEX(16, -12), MAKE_VAL_COMPLEX(70, 10), MAKE_VAL_COMPLEX(-24, -32), MAKE_VAL_COMPLEX(62, 46), MAKE_VAL_COMPLEX(-78, -34), MAKE_VAL_COMPLEX(40, 20), MAKE_VAL_COMPLEX(40, -20), MAKE_VAL_COMPLEX(-58, -14), MAKE_VAL_COMPLEX(28, 44), MAKE_VAL_COMPLEX(-24, 8), MAKE_VAL_COMPLEX(-26, -58), MAKE_VAL_COMPLEX(44, -8), MAKE_VAL_COMPLEX(-24, 68) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1054, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1054", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1054() */
/* END TEST 1054 */

/*
 *  BEGIN TEST 1055
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1008', LAYOUT_COLMAJ>
 *  y : <'y1008', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1055( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1008_ptr, A1008_ind, A1008_val,
		A1008_M, A1008_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1008_val,
		y1008_LEN * y1008_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1008_LEN, y1008_VW, LAYOUT_COLMAJ, y1008_LEN );
	oski_value_t* y_val = array_Duplicate( x1008_val,
		x1008_LEN * x1008_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1008_LEN, x1008_VW, LAYOUT_COLMAJ, x1008_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1729, 768), MAKE_VAL_COMPLEX(189, -2463), MAKE_VAL_COMPLEX(-985, -1330), MAKE_VAL_COMPLEX(-321, 94), MAKE_VAL_COMPLEX(-43, -465), MAKE_VAL_COMPLEX(-861, 521), MAKE_VAL_COMPLEX(803, 2095), MAKE_VAL_COMPLEX(471, 306), MAKE_VAL_COMPLEX(-289, 601), MAKE_VAL_COMPLEX(691, -72), MAKE_VAL_COMPLEX(198, -258), MAKE_VAL_COMPLEX(706, 1618), MAKE_VAL_COMPLEX(-312, -779), MAKE_VAL_COMPLEX(-677, -127), MAKE_VAL_COMPLEX(-2179, -2404) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1055, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1055", y_true, y_val, x1008_LEN*x1008_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1055() */
/* END TEST 1055 */

#define A1056_M 5    /* # of rows */
#define A1056_N 5    /* # of cols */
#define A1056_NNZ 6  /* # of stored non-zeros */
static oski_index_t A1056_ptr[A1056_M+1] = {1, 1, 2, 4, 6, 7}; /* end row pointers for A1056 */
static oski_index_t A1056_ind[A1056_NNZ] = {1, 1, 2, 1, 2, 2}; /* end column indices pointers for A1056 */
static oski_value_t A1056_val[A1056_NNZ] = {MAKE_VAL_COMPLEX(-6, -7), MAKE_VAL_COMPLEX(-5, 0), MAKE_VAL_COMPLEX(-6, -2), MAKE_VAL_COMPLEX(-3, -4), MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(-9, -7)}; /* end stored non-zero values for A1056 */

#define x1056_LEN 5   /* vector length */
#define x1056_VW  3   /* vector width */
static oski_value_t x1056_val[x1056_LEN*x1056_VW] = {MAKE_VAL_COMPLEX(0, 5), MAKE_VAL_COMPLEX(-9, 3), MAKE_VAL_COMPLEX(-6, -7), MAKE_VAL_COMPLEX(-3, -2), MAKE_VAL_COMPLEX(-3, -7), MAKE_VAL_COMPLEX(-0, -6), MAKE_VAL_COMPLEX(-10, -6), MAKE_VAL_COMPLEX(-7, -10), MAKE_VAL_COMPLEX(8, -1), MAKE_VAL_COMPLEX(10, -7), MAKE_VAL_COMPLEX(7, 2), MAKE_VAL_COMPLEX(-7, -3), MAKE_VAL_COMPLEX(1, 10), MAKE_VAL_COMPLEX(3, -6), MAKE_VAL_COMPLEX(10, -1)}; /* data */

#define y1056_LEN 5   /* vector length */
#define y1056_VW  3   /* vector width */
static oski_value_t y1056_val[y1056_LEN*y1056_VW] = {MAKE_VAL_COMPLEX(-2, -8), MAKE_VAL_COMPLEX(-9, -10), MAKE_VAL_COMPLEX(10, 5), MAKE_VAL_COMPLEX(-10, -9), MAKE_VAL_COMPLEX(-10, 4), MAKE_VAL_COMPLEX(6, 7), MAKE_VAL_COMPLEX(1, 1), MAKE_VAL_COMPLEX(0, -3), MAKE_VAL_COMPLEX(-6, 10), MAKE_VAL_COMPLEX(-5, 3), MAKE_VAL_COMPLEX(8, -4), MAKE_VAL_COMPLEX(-10, -8), MAKE_VAL_COMPLEX(9, -6), MAKE_VAL_COMPLEX(9, 2), MAKE_VAL_COMPLEX(0, 4)}; /* data */

/*
 *  BEGIN TEST 1056
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1056( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	oski_value_t* y_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(6, 21), MAKE_VAL_COMPLEX(-25, 34), MAKE_VAL_COMPLEX(-59, 106), MAKE_VAL_COMPLEX(-41, -150), MAKE_VAL_COMPLEX(-63, 47), MAKE_VAL_COMPLEX(-116, 140), MAKE_VAL_COMPLEX(5, -8), MAKE_VAL_COMPLEX(48, 14), MAKE_VAL_COMPLEX(-12, 88), MAKE_VAL_COMPLEX(19, 27), MAKE_VAL_COMPLEX(38, 133), MAKE_VAL_COMPLEX(-91, 58), MAKE_VAL_COMPLEX(53, 49), MAKE_VAL_COMPLEX(4, 102), MAKE_VAL_COMPLEX(-40, 65) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1056, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1056", y_true, y_val, y1056_LEN*y1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1056() */
/* END TEST 1056 */

/*
 *  BEGIN TEST 1057
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1057( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	oski_value_t* y_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(66, -5), MAKE_VAL_COMPLEX(141, -66), MAKE_VAL_COMPLEX(-21, -46), MAKE_VAL_COMPLEX(153, 106), MAKE_VAL_COMPLEX(71, -143), MAKE_VAL_COMPLEX(36, -120), MAKE_VAL_COMPLEX(-17, 12), MAKE_VAL_COMPLEX(-24, -2), MAKE_VAL_COMPLEX(-44, -176), MAKE_VAL_COMPLEX(-23, -79), MAKE_VAL_COMPLEX(-38, -53), MAKE_VAL_COMPLEX(195, -106), MAKE_VAL_COMPLEX(-41, 47), MAKE_VAL_COMPLEX(-56, -38), MAKE_VAL_COMPLEX(8, -81) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1057, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1057", y_true, y_val, y1056_LEN*y1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1057() */
/* END TEST 1057 */

/*
 *  BEGIN TEST 1058
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1058( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	oski_value_t* y_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-12, 54), MAKE_VAL_COMPLEX(24, 87), MAKE_VAL_COMPLEX(-45, -60), MAKE_VAL_COMPLEX(33, 84), MAKE_VAL_COMPLEX(72, 6), MAKE_VAL_COMPLEX(-15, -60), MAKE_VAL_COMPLEX(-3, -9), MAKE_VAL_COMPLEX(-9, 18), MAKE_VAL_COMPLEX(66, -42), MAKE_VAL_COMPLEX(39, -3), MAKE_VAL_COMPLEX(-60, 0), MAKE_VAL_COMPLEX(36, 78), MAKE_VAL_COMPLEX(-72, 9), MAKE_VAL_COMPLEX(-48, -39), MAKE_VAL_COMPLEX(12, -24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1058, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1058", y_true, y_val, y1056_LEN*y1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1058() */
/* END TEST 1058 */

/*
 *  BEGIN TEST 1059
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1059( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	oski_value_t* y_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(191, 39), MAKE_VAL_COMPLEX(297, 167), MAKE_VAL_COMPLEX(351, 441), MAKE_VAL_COMPLEX(-639, 47), MAKE_VAL_COMPLEX(220, 250), MAKE_VAL_COMPLEX(480, 712), MAKE_VAL_COMPLEX(-51, -33), MAKE_VAL_COMPLEX(142, -202), MAKE_VAL_COMPLEX(396, 20), MAKE_VAL_COMPLEX(170, -132), MAKE_VAL_COMPLEX(671, 39), MAKE_VAL_COMPLEX(343, 497), MAKE_VAL_COMPLEX(274, -96), MAKE_VAL_COMPLEX(426, 138), MAKE_VAL_COMPLEX(253, 233) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1059, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1059", y_true, y_val, y1056_LEN*y1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1059() */
/* END TEST 1059 */

/*
 *  BEGIN TEST 1060
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1060( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(37, 112), MAKE_VAL_COMPLEX(69, 61), MAKE_VAL_COMPLEX(35, -27), MAKE_VAL_COMPLEX(-103, 72), MAKE_VAL_COMPLEX(-241, 50), MAKE_VAL_COMPLEX(61, 15), MAKE_VAL_COMPLEX(49, 167), MAKE_VAL_COMPLEX(87, 91), MAKE_VAL_COMPLEX(-98, -99), MAKE_VAL_COMPLEX(-69, 199), MAKE_VAL_COMPLEX(129, 90), MAKE_VAL_COMPLEX(-8, -159), MAKE_VAL_COMPLEX(74, 121), MAKE_VAL_COMPLEX(97, 36), MAKE_VAL_COMPLEX(-29, -139) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1060, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1060", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1060() */
/* END TEST 1060 */

/*
 *  BEGIN TEST 1061
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1061( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-52, -77), MAKE_VAL_COMPLEX(-141, -67), MAKE_VAL_COMPLEX(-56, -40), MAKE_VAL_COMPLEX(88, -95), MAKE_VAL_COMPLEX(241, -108), MAKE_VAL_COMPLEX(-43, -57), MAKE_VAL_COMPLEX(-101, -239), MAKE_VAL_COMPLEX(-106, -182), MAKE_VAL_COMPLEX(157, 116), MAKE_VAL_COMPLEX(160, -218), MAKE_VAL_COMPLEX(-86, -55), MAKE_VAL_COMPLEX(-32, 117), MAKE_VAL_COMPLEX(-97, -48), MAKE_VAL_COMPLEX(-58, -69), MAKE_VAL_COMPLEX(102, 162) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1061, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1061", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1061() */
/* END TEST 1061 */

/*
 *  BEGIN TEST 1062
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1062( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-35, 40), MAKE_VAL_COMPLEX(-93, -39), MAKE_VAL_COMPLEX(1, -98), MAKE_VAL_COMPLEX(-10, -37), MAKE_VAL_COMPLEX(25, -77), MAKE_VAL_COMPLEX(42, -48), MAKE_VAL_COMPLEX(-38, -118), MAKE_VAL_COMPLEX(14, -129), MAKE_VAL_COMPLEX(71, 48), MAKE_VAL_COMPLEX(129, 14), MAKE_VAL_COMPLEX(42, 65), MAKE_VAL_COMPLEX(-35, -73), MAKE_VAL_COMPLEX(-62, 87), MAKE_VAL_COMPLEX(66, -27), MAKE_VAL_COMPLEX(87, 62) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1062, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1062", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1062() */
/* END TEST 1062 */

/*
 *  BEGIN TEST 1063
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1063( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-675, -485), MAKE_VAL_COMPLEX(-410, 100), MAKE_VAL_COMPLEX(-6, 548), MAKE_VAL_COMPLEX(193, -789), MAKE_VAL_COMPLEX(923, -1259), MAKE_VAL_COMPLEX(-464, 362), MAKE_VAL_COMPLEX(-944, -318), MAKE_VAL_COMPLEX(-876, 298), MAKE_VAL_COMPLEX(795, -85), MAKE_VAL_COMPLEX(-968, -1326), MAKE_VAL_COMPLEX(-1221, 53), MAKE_VAL_COMPLEX(947, 919), MAKE_VAL_COMPLEX(-857, -469), MAKE_VAL_COMPLEX(-815, 395), MAKE_VAL_COMPLEX(606, 432) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1063, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1063", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1063() */
/* END TEST 1063 */

/*
 *  BEGIN TEST 1064
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1064( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(149, -18), MAKE_VAL_COMPLEX(-39, 33), MAKE_VAL_COMPLEX(-73, -81), MAKE_VAL_COMPLEX(-5, 60), MAKE_VAL_COMPLEX(-81, 146), MAKE_VAL_COMPLEX(43, -129), MAKE_VAL_COMPLEX(-27, 87), MAKE_VAL_COMPLEX(-19, -35), MAKE_VAL_COMPLEX(-48, -101), MAKE_VAL_COMPLEX(261, -107), MAKE_VAL_COMPLEX(203, -176), MAKE_VAL_COMPLEX(-264, 25), MAKE_VAL_COMPLEX(230, 79), MAKE_VAL_COMPLEX(59, -170), MAKE_VAL_COMPLEX(-29, -85) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1064, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1064", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1064() */
/* END TEST 1064 */

/*
 *  BEGIN TEST 1065
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1065( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-134, 53), MAKE_VAL_COMPLEX(-15, 15), MAKE_VAL_COMPLEX(10, 50), MAKE_VAL_COMPLEX(-22, -65), MAKE_VAL_COMPLEX(39, -186), MAKE_VAL_COMPLEX(-61, 87), MAKE_VAL_COMPLEX(-61, -99), MAKE_VAL_COMPLEX(-60, -14), MAKE_VAL_COMPLEX(101, 70), MAKE_VAL_COMPLEX(-212, 28), MAKE_VAL_COMPLEX(-148, 169), MAKE_VAL_COMPLEX(206, -25), MAKE_VAL_COMPLEX(-193, -12), MAKE_VAL_COMPLEX(-56, 119), MAKE_VAL_COMPLEX(96, 48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1065, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1065", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1065() */
/* END TEST 1065 */

/*
 *  BEGIN TEST 1066
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1066( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-10, 15), MAKE_VAL_COMPLEX(-33, -9), MAKE_VAL_COMPLEX(-4, -33), MAKE_VAL_COMPLEX(-5, -12), MAKE_VAL_COMPLEX(5, -27), MAKE_VAL_COMPLEX(12, -18), MAKE_VAL_COMPLEX(-18, -38), MAKE_VAL_COMPLEX(-1, -44), MAKE_VAL_COMPLEX(26, 13), MAKE_VAL_COMPLEX(44, -1), MAKE_VAL_COMPLEX(17, 20), MAKE_VAL_COMPLEX(-15, -23), MAKE_VAL_COMPLEX(-17, 32), MAKE_VAL_COMPLEX(21, -12), MAKE_VAL_COMPLEX(32, 17) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1066, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1066", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1066() */
/* END TEST 1066 */

/*
 *  BEGIN TEST 1067
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1067( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-136, 1237), MAKE_VAL_COMPLEX(291, 333), MAKE_VAL_COMPLEX(416, 442), MAKE_VAL_COMPLEX(-614, -43), MAKE_VAL_COMPLEX(-1387, -1036), MAKE_VAL_COMPLEX(259, 1043), MAKE_VAL_COMPLEX(-1029, 333), MAKE_VAL_COMPLEX(-492, 764), MAKE_VAL_COMPLEX(883, -614), MAKE_VAL_COMPLEX(-1190, 1756), MAKE_VAL_COMPLEX(516, 1835), MAKE_VAL_COMPLEX(904, -1559), MAKE_VAL_COMPLEX(-861, 1262), MAKE_VAL_COMPLEX(480, 1097), MAKE_VAL_COMPLEX(642, -728) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1067, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1067", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1067() */
/* END TEST 1067 */

/*
 *  BEGIN TEST 1068
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1068( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	oski_value_t* y_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(14, 53), MAKE_VAL_COMPLEX(0, -111), MAKE_VAL_COMPLEX(-16, 2), MAKE_VAL_COMPLEX(68, 17), MAKE_VAL_COMPLEX(16, 25), MAKE_VAL_COMPLEX(-38, 87), MAKE_VAL_COMPLEX(-28, 56), MAKE_VAL_COMPLEX(48, 26), MAKE_VAL_COMPLEX(-8, 107), MAKE_VAL_COMPLEX(-20, 65), MAKE_VAL_COMPLEX(-79, 91), MAKE_VAL_COMPLEX(-94, 95), MAKE_VAL_COMPLEX(29, 94), MAKE_VAL_COMPLEX(-90, 118), MAKE_VAL_COMPLEX(-40, 49) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1068, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1068", y_true, y_val, y1056_LEN*y1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1068() */
/* END TEST 1068 */

/*
 *  BEGIN TEST 1069
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1069( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	oski_value_t* y_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-36, -5), MAKE_VAL_COMPLEX(5, 197), MAKE_VAL_COMPLEX(-14, -67), MAKE_VAL_COMPLEX(-54, 68), MAKE_VAL_COMPLEX(50, -5), MAKE_VAL_COMPLEX(36, -146), MAKE_VAL_COMPLEX(27, -65), MAKE_VAL_COMPLEX(-60, -11), MAKE_VAL_COMPLEX(78, -133), MAKE_VAL_COMPLEX(57, -60), MAKE_VAL_COMPLEX(23, -103), MAKE_VAL_COMPLEX(112, -15), MAKE_VAL_COMPLEX(-98, -100), MAKE_VAL_COMPLEX(53, -164), MAKE_VAL_COMPLEX(56, -69) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1069, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1069", y_true, y_val, y1056_LEN*y1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1069() */
/* END TEST 1069 */

/*
 *  BEGIN TEST 1070
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1070( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	oski_value_t* y_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-82, -56), MAKE_VAL_COMPLEX(-161, -18), MAKE_VAL_COMPLEX(130, -35), MAKE_VAL_COMPLEX(-162, -1), MAKE_VAL_COMPLEX(-58, 116), MAKE_VAL_COMPLEX(110, 15), MAKE_VAL_COMPLEX(17, 1), MAKE_VAL_COMPLEX(-24, -27), MAKE_VAL_COMPLEX(26, 138), MAKE_VAL_COMPLEX(-21, 67), MAKE_VAL_COMPLEX(40, -100), MAKE_VAL_COMPLEX(-154, 8), MAKE_VAL_COMPLEX(33, -126), MAKE_VAL_COMPLEX(97, -54), MAKE_VAL_COMPLEX(32, 36) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1070, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1070", y_true, y_val, y1056_LEN*y1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1070() */
/* END TEST 1070 */

/*
 *  BEGIN TEST 1071
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1071( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	oski_value_t* y_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(122, -403), MAKE_VAL_COMPLEX(414, 916), MAKE_VAL_COMPLEX(-151, 172), MAKE_VAL_COMPLEX(-209, -279), MAKE_VAL_COMPLEX(-92, -559), MAKE_VAL_COMPLEX(57, -733), MAKE_VAL_COMPLEX(209, -493), MAKE_VAL_COMPLEX(-351, -186), MAKE_VAL_COMPLEX(-102, -1285), MAKE_VAL_COMPLEX(179, -768), MAKE_VAL_COMPLEX(691, -539), MAKE_VAL_COMPLEX(1222, -997), MAKE_VAL_COMPLEX(-243, -507), MAKE_VAL_COMPLEX(612, -851), MAKE_VAL_COMPLEX(252, -505) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-9, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1071, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1071", y_true, y_val, y1056_LEN*y1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1071() */
/* END TEST 1071 */

/*
 *  BEGIN TEST 1072
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1072( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-64, 162), MAKE_VAL_COMPLEX(62, 24), MAKE_VAL_COMPLEX(-72, 107), MAKE_VAL_COMPLEX(18, 230), MAKE_VAL_COMPLEX(217, -98), MAKE_VAL_COMPLEX(-206, -174), MAKE_VAL_COMPLEX(32, 137), MAKE_VAL_COMPLEX(-61, -42), MAKE_VAL_COMPLEX(19, 65), MAKE_VAL_COMPLEX(-50, 186), MAKE_VAL_COMPLEX(15, -67), MAKE_VAL_COMPLEX(-24, 173), MAKE_VAL_COMPLEX(22, 165), MAKE_VAL_COMPLEX(-16, -25), MAKE_VAL_COMPLEX(42, 125) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1072, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1072", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1072() */
/* END TEST 1072 */

/*
 *  BEGIN TEST 1073
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1073( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(39, -202), MAKE_VAL_COMPLEX(-5, -93), MAKE_VAL_COMPLEX(155, -81), MAKE_VAL_COMPLEX(16, -229), MAKE_VAL_COMPLEX(-158, 139), MAKE_VAL_COMPLEX(236, 222), MAKE_VAL_COMPLEX(78, -139), MAKE_VAL_COMPLEX(167, 87), MAKE_VAL_COMPLEX(-78, -17), MAKE_VAL_COMPLEX(5, -80), MAKE_VAL_COMPLEX(-81, 86), MAKE_VAL_COMPLEX(95, -184), MAKE_VAL_COMPLEX(-80, -240), MAKE_VAL_COMPLEX(22, 88), MAKE_VAL_COMPLEX(-117, -67) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1073, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1073", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1073() */
/* END TEST 1073 */

/*
 *  BEGIN TEST 1074
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1074( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(25, 35), MAKE_VAL_COMPLEX(-48, 66), MAKE_VAL_COMPLEX(-77, -19), MAKE_VAL_COMPLEX(-31, 1), MAKE_VAL_COMPLEX(-56, -34), MAKE_VAL_COMPLEX(-30, -42), MAKE_VAL_COMPLEX(-100, 8), MAKE_VAL_COMPLEX(-99, -35), MAKE_VAL_COMPLEX(51, -47), MAKE_VAL_COMPLEX(35, -99), MAKE_VAL_COMPLEX(59, -21), MAKE_VAL_COMPLEX(-64, 14), MAKE_VAL_COMPLEX(57, 65), MAKE_VAL_COMPLEX(-9, -57), MAKE_VAL_COMPLEX(65, -57) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1074, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1074", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1074() */
/* END TEST 1074 */

/*
 *  BEGIN TEST 1075
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1075( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1005, -1260), MAKE_VAL_COMPLEX(-239, 403), MAKE_VAL_COMPLEX(-497, -822), MAKE_VAL_COMPLEX(-1866, -715), MAKE_VAL_COMPLEX(-73, 2289), MAKE_VAL_COMPLEX(2174, -832), MAKE_VAL_COMPLEX(-1066, -102), MAKE_VAL_COMPLEX(650, -71), MAKE_VAL_COMPLEX(-763, -144), MAKE_VAL_COMPLEX(-1537, -1074), MAKE_VAL_COMPLEX(350, 299), MAKE_VAL_COMPLEX(-1169, -775), MAKE_VAL_COMPLEX(-1358, -691), MAKE_VAL_COMPLEX(162, 71), MAKE_VAL_COMPLEX(-1375, -214) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1075, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1075", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1075() */
/* END TEST 1075 */

/*
 *  BEGIN TEST 1076
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1076( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(178, -49), MAKE_VAL_COMPLEX(-5, 41), MAKE_VAL_COMPLEX(-12, 82), MAKE_VAL_COMPLEX(225, -78), MAKE_VAL_COMPLEX(-150, -179), MAKE_VAL_COMPLEX(-258, 160), MAKE_VAL_COMPLEX(46, 167), MAKE_VAL_COMPLEX(-118, 14), MAKE_VAL_COMPLEX(37, -22), MAKE_VAL_COMPLEX(162, -63), MAKE_VAL_COMPLEX(-56, -43), MAKE_VAL_COMPLEX(157, 82), MAKE_VAL_COMPLEX(221, 23), MAKE_VAL_COMPLEX(-69, -23), MAKE_VAL_COMPLEX(138, -74) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1076, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1076", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1076() */
/* END TEST 1076 */

/*
 *  BEGIN TEST 1077
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1077( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-178, 24), MAKE_VAL_COMPLEX(50, -56), MAKE_VAL_COMPLEX(42, -47), MAKE_VAL_COMPLEX(-210, 88), MAKE_VAL_COMPLEX(165, 214), MAKE_VAL_COMPLEX(258, -130), MAKE_VAL_COMPLEX(4, -137), MAKE_VAL_COMPLEX(153, 36), MAKE_VAL_COMPLEX(-77, 27), MAKE_VAL_COMPLEX(-212, 98), MAKE_VAL_COMPLEX(21, 33), MAKE_VAL_COMPLEX(-122, -67), MAKE_VAL_COMPLEX(-226, -73), MAKE_VAL_COMPLEX(54, 53), MAKE_VAL_COMPLEX(-188, 79) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1077, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1077", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1077() */
/* END TEST 1077 */

/*
 *  BEGIN TEST 1078
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1078( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(25, 30), MAKE_VAL_COMPLEX(-39, 63), MAKE_VAL_COMPLEX(-71, -12), MAKE_VAL_COMPLEX(-28, 3), MAKE_VAL_COMPLEX(-53, -27), MAKE_VAL_COMPLEX(-30, -36), MAKE_VAL_COMPLEX(-90, 14), MAKE_VAL_COMPLEX(-92, -25), MAKE_VAL_COMPLEX(43, -46), MAKE_VAL_COMPLEX(25, -92), MAKE_VAL_COMPLEX(52, -23), MAKE_VAL_COMPLEX(-57, 17), MAKE_VAL_COMPLEX(56, 55), MAKE_VAL_COMPLEX(-12, -51), MAKE_VAL_COMPLEX(55, -56) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1078, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1078", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1078() */
/* END TEST 1078 */

/*
 *  BEGIN TEST 1079
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_ROWMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1079( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_ROWMAJ, x1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-439, 137), MAKE_VAL_COMPLEX(90, 42), MAKE_VAL_COMPLEX(-85, -118), MAKE_VAL_COMPLEX(-707, 295), MAKE_VAL_COMPLEX(323, 608), MAKE_VAL_COMPLEX(660, -468), MAKE_VAL_COMPLEX(-232, -299), MAKE_VAL_COMPLEX(178, 111), MAKE_VAL_COMPLEX(-146, -84), MAKE_VAL_COMPLEX(-639, 13), MAKE_VAL_COMPLEX(192, -8), MAKE_VAL_COMPLEX(-514, -107), MAKE_VAL_COMPLEX(-475, -108), MAKE_VAL_COMPLEX(87, 24), MAKE_VAL_COMPLEX(-453, 34) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-3, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1079, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1079", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1079() */
/* END TEST 1079 */

/*
 *  BEGIN TEST 1080
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1080( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	oski_value_t* y_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(52, 57), MAKE_VAL_COMPLEX(-88, 96), MAKE_VAL_COMPLEX(85, 10), MAKE_VAL_COMPLEX(-79, 148), MAKE_VAL_COMPLEX(-309, 105), MAKE_VAL_COMPLEX(-131, -138), MAKE_VAL_COMPLEX(66, -30), MAKE_VAL_COMPLEX(26, 55), MAKE_VAL_COMPLEX(10, 132), MAKE_VAL_COMPLEX(117, 89), MAKE_VAL_COMPLEX(60, 109), MAKE_VAL_COMPLEX(-80, 54), MAKE_VAL_COMPLEX(132, -58), MAKE_VAL_COMPLEX(131, 86), MAKE_VAL_COMPLEX(72, 103) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1080, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1080", y_true, y_val, y1056_LEN*y1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1080() */
/* END TEST 1080 */

/*
 *  BEGIN TEST 1081
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1081( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	oski_value_t* y_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-106, -69), MAKE_VAL_COMPLEX(1, -72), MAKE_VAL_COMPLEX(-25, -55), MAKE_VAL_COMPLEX(-5, -115), MAKE_VAL_COMPLEX(303, -33), MAKE_VAL_COMPLEX(191, 123), MAKE_VAL_COMPLEX(-57, 27), MAKE_VAL_COMPLEX(-44, -64), MAKE_VAL_COMPLEX(32, -66), MAKE_VAL_COMPLEX(-114, -50), MAKE_VAL_COMPLEX(-60, -169), MAKE_VAL_COMPLEX(2, -18), MAKE_VAL_COMPLEX(-141, -14), MAKE_VAL_COMPLEX(-92, -134), MAKE_VAL_COMPLEX(-48, -91) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1081, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1081", y_true, y_val, y1056_LEN*y1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1081() */
/* END TEST 1081 */

/*
 *  BEGIN TEST 1082
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1082( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	oski_value_t* y_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-14, -22), MAKE_VAL_COMPLEX(-37, -21), MAKE_VAL_COMPLEX(35, 5), MAKE_VAL_COMPLEX(-39, -17), MAKE_VAL_COMPLEX(-26, 22), MAKE_VAL_COMPLEX(25, 15), MAKE_VAL_COMPLEX(4, 2), MAKE_VAL_COMPLEX(-3, -9), MAKE_VAL_COMPLEX(-8, 36), MAKE_VAL_COMPLEX(-12, 14), MAKE_VAL_COMPLEX(20, -20), MAKE_VAL_COMPLEX(-38, -14), MAKE_VAL_COMPLEX(21, -27), MAKE_VAL_COMPLEX(29, -3), MAKE_VAL_COMPLEX(4, 12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1082, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1082", y_true, y_val, y1056_LEN*y1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1082() */
/* END TEST 1082 */

/*
 *  BEGIN TEST 1083
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1083( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	oski_value_t* y_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1327, -74), MAKE_VAL_COMPLEX(992, 646), MAKE_VAL_COMPLEX(90, 145), MAKE_VAL_COMPLEX(1348, 809), MAKE_VAL_COMPLEX(-1295, 2005), MAKE_VAL_COMPLEX(-2342, 449), MAKE_VAL_COMPLEX(100, -580), MAKE_VAL_COMPLEX(763, 184), MAKE_VAL_COMPLEX(316, 158), MAKE_VAL_COMPLEX(1163, -691), MAKE_VAL_COMPLEX(1323, 874), MAKE_VAL_COMPLEX(610, 240), MAKE_VAL_COMPLEX(722, -534), MAKE_VAL_COMPLEX(1132, 241), MAKE_VAL_COMPLEX(857, 66) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1083, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1083", y_true, y_val, y1056_LEN*y1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1083() */
/* END TEST 1083 */

/*
 *  BEGIN TEST 1084
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1084( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-18, 97), MAKE_VAL_COMPLEX(-77, -22), MAKE_VAL_COMPLEX(132, 108), MAKE_VAL_COMPLEX(8, 214), MAKE_VAL_COMPLEX(100, 159), MAKE_VAL_COMPLEX(81, 61), MAKE_VAL_COMPLEX(-127, -16), MAKE_VAL_COMPLEX(218, 44), MAKE_VAL_COMPLEX(102, 183), MAKE_VAL_COMPLEX(126, 141), MAKE_VAL_COMPLEX(2, -26), MAKE_VAL_COMPLEX(141, -31), MAKE_VAL_COMPLEX(-153, -112), MAKE_VAL_COMPLEX(17, -142), MAKE_VAL_COMPLEX(-48, -26) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1084, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1084", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1084() */
/* END TEST 1084 */

/*
 *  BEGIN TEST 1085
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1085( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-37, -87), MAKE_VAL_COMPLEX(26, -71), MAKE_VAL_COMPLEX(-67, -188), MAKE_VAL_COMPLEX(8, -251), MAKE_VAL_COMPLEX(-29, -206), MAKE_VAL_COMPLEX(-15, -73), MAKE_VAL_COMPLEX(173, -106), MAKE_VAL_COMPLEX(-122, -141), MAKE_VAL_COMPLEX(-75, -97), MAKE_VAL_COMPLEX(-29, -45), MAKE_VAL_COMPLEX(-10, 107), MAKE_VAL_COMPLEX(-122, -52), MAKE_VAL_COMPLEX(45, 143), MAKE_VAL_COMPLEX(55, 163), MAKE_VAL_COMPLEX(79, 134) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1085, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1085", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1085() */
/* END TEST 1085 */

/*
 *  BEGIN TEST 1086
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1086( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, -40), MAKE_VAL_COMPLEX(72, -24), MAKE_VAL_COMPLEX(48, 56), MAKE_VAL_COMPLEX(24, 16), MAKE_VAL_COMPLEX(24, 56), MAKE_VAL_COMPLEX(0, 48), MAKE_VAL_COMPLEX(80, 48), MAKE_VAL_COMPLEX(56, 80), MAKE_VAL_COMPLEX(-64, 8), MAKE_VAL_COMPLEX(-80, 56), MAKE_VAL_COMPLEX(-56, -16), MAKE_VAL_COMPLEX(56, 24), MAKE_VAL_COMPLEX(-8, -80), MAKE_VAL_COMPLEX(-24, 48), MAKE_VAL_COMPLEX(-80, 8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1086, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1086", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1086() */
/* END TEST 1086 */

/*
 *  BEGIN TEST 1087
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1087( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-954, -501), MAKE_VAL_COMPLEX(122, -1020), MAKE_VAL_COMPLEX(-1037, -208), MAKE_VAL_COMPLEX(-1458, -1277), MAKE_VAL_COMPLEX(-1168, -468), MAKE_VAL_COMPLEX(-394, 130), MAKE_VAL_COMPLEX(965, -1671), MAKE_VAL_COMPLEX(-822, 575), MAKE_VAL_COMPLEX(-1542, 361), MAKE_VAL_COMPLEX(-886, 753), MAKE_VAL_COMPLEX(180, 699), MAKE_VAL_COMPLEX(-423, 565), MAKE_VAL_COMPLEX(841, -218), MAKE_VAL_COMPLEX(1413, 916), MAKE_VAL_COMPLEX(715, 502) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1087, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1087", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1087() */
/* END TEST 1087 */

/*
 *  BEGIN TEST 1088
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1088( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(99, -28), MAKE_VAL_COMPLEX(-35, 40), MAKE_VAL_COMPLEX(81, 9), MAKE_VAL_COMPLEX(213, -15), MAKE_VAL_COMPLEX(172, -49), MAKE_VAL_COMPLEX(39, -81), MAKE_VAL_COMPLEX(-51, 108), MAKE_VAL_COMPLEX(95, -85), MAKE_VAL_COMPLEX(187, -124), MAKE_VAL_COMPLEX(159, -106), MAKE_VAL_COMPLEX(51, -21), MAKE_VAL_COMPLEX(49, -127), MAKE_VAL_COMPLEX(-140, 19), MAKE_VAL_COMPLEX(-148, -17), MAKE_VAL_COMPLEX(-39, 17) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1088, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1088", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1088() */
/* END TEST 1088 */

/*
 *  BEGIN TEST 1089
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1089( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-104, 43), MAKE_VAL_COMPLEX(5, -40), MAKE_VAL_COMPLEX(-92, -36), MAKE_VAL_COMPLEX(-220, 6), MAKE_VAL_COMPLEX(-174, 25), MAKE_VAL_COMPLEX(-33, 63), MAKE_VAL_COMPLEX(27, -136), MAKE_VAL_COMPLEX(-106, 48), MAKE_VAL_COMPLEX(-162, 129), MAKE_VAL_COMPLEX(-122, 95), MAKE_VAL_COMPLEX(-32, 34), MAKE_VAL_COMPLEX(-67, 111), MAKE_VAL_COMPLEX(133, 12), MAKE_VAL_COMPLEX(163, 2), MAKE_VAL_COMPLEX(70, -10) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1089, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1089", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1089() */
/* END TEST 1089 */

/*
 *  BEGIN TEST 1090
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1090( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(25, 10), MAKE_VAL_COMPLEX(-3, 51), MAKE_VAL_COMPLEX(-47, 16), MAKE_VAL_COMPLEX(-16, 11), MAKE_VAL_COMPLEX(-41, 1), MAKE_VAL_COMPLEX(-30, -12), MAKE_VAL_COMPLEX(-50, 38), MAKE_VAL_COMPLEX(-64, 15), MAKE_VAL_COMPLEX(11, -42), MAKE_VAL_COMPLEX(-15, -64), MAKE_VAL_COMPLEX(24, -31), MAKE_VAL_COMPLEX(-29, 29), MAKE_VAL_COMPLEX(52, 15), MAKE_VAL_COMPLEX(-24, -27), MAKE_VAL_COMPLEX(15, -52) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1090, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1090", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1090() */
/* END TEST 1090 */

/*
 *  BEGIN TEST 1091
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1091( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_ROWMAJ, y1056_VW );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(21, 1343), MAKE_VAL_COMPLEX(-733, 26), MAKE_VAL_COMPLEX(-1044, 415), MAKE_VAL_COMPLEX(-987, 1940), MAKE_VAL_COMPLEX(-682, 1367), MAKE_VAL_COMPLEX(375, 267), MAKE_VAL_COMPLEX(-1646, -969), MAKE_VAL_COMPLEX(-421, 746), MAKE_VAL_COMPLEX(879, 1847), MAKE_VAL_COMPLEX(767, 1049), MAKE_VAL_COMPLEX(527, 448), MAKE_VAL_COMPLEX(372, 976), MAKE_VAL_COMPLEX(787, -689), MAKE_VAL_COMPLEX(751, -1771), MAKE_VAL_COMPLEX(650, -817) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, 9);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1091, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1091", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1091() */
/* END TEST 1091 */

/*
 *  BEGIN TEST 1092
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1092( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	oski_value_t* y_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(114, 33), MAKE_VAL_COMPLEX(-6, 54), MAKE_VAL_COMPLEX(119, 38), MAKE_VAL_COMPLEX(84, -59), MAKE_VAL_COMPLEX(7, 31), MAKE_VAL_COMPLEX(52, 195), MAKE_VAL_COMPLEX(-254, 38), MAKE_VAL_COMPLEX(50, 52), MAKE_VAL_COMPLEX(-54, 239), MAKE_VAL_COMPLEX(9, 126), MAKE_VAL_COMPLEX(66, 17), MAKE_VAL_COMPLEX(-264, -251), MAKE_VAL_COMPLEX(92, 11), MAKE_VAL_COMPLEX(96, 103), MAKE_VAL_COMPLEX(40, 107) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1092, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1092", y_true, y_val, y1056_LEN*y1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1092() */
/* END TEST 1092 */

/*
 *  BEGIN TEST 1093
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1093( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	oski_value_t* y_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-154, -91), MAKE_VAL_COMPLEX(-96, -107), MAKE_VAL_COMPLEX(-24, -28), MAKE_VAL_COMPLEX(-191, 17), MAKE_VAL_COMPLEX(-75, 31), MAKE_VAL_COMPLEX(17, -157), MAKE_VAL_COMPLEX(265, -33), MAKE_VAL_COMPLEX(-59, -76), MAKE_VAL_COMPLEX(36, -141), MAKE_VAL_COMPLEX(-40, -87), MAKE_VAL_COMPLEX(-14, -73), MAKE_VAL_COMPLEX(160, 217), MAKE_VAL_COMPLEX(-38, -86), MAKE_VAL_COMPLEX(-18, -114), MAKE_VAL_COMPLEX(-28, -75) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1093, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1093", y_true, y_val, y1056_LEN*y1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1093() */
/* END TEST 1093 */

/*
 *  BEGIN TEST 1094
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1094( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	oski_value_t* y_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-62, -10), MAKE_VAL_COMPLEX(-97, 33), MAKE_VAL_COMPLEX(65, -55), MAKE_VAL_COMPLEX(-93, 43), MAKE_VAL_COMPLEX(-2, 82), MAKE_VAL_COMPLEX(67, -21), MAKE_VAL_COMPLEX(10, -4), MAKE_VAL_COMPLEX(-21, -9), MAKE_VAL_COMPLEX(52, 72), MAKE_VAL_COMPLEX(6, 44), MAKE_VAL_COMPLEX(-4, -68), MAKE_VAL_COMPLEX(-86, 46), MAKE_VAL_COMPLEX(-15, -81), MAKE_VAL_COMPLEX(41, -57), MAKE_VAL_COMPLEX(28, 12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1094, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1094", y_true, y_val, y1056_LEN*y1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1094() */
/* END TEST 1094 */

/*
 *  BEGIN TEST 1095
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1095( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	oski_value_t* y_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-762, -1093), MAKE_VAL_COMPLEX(13, -1444), MAKE_VAL_COMPLEX(-585, 125), MAKE_VAL_COMPLEX(-1124, -586), MAKE_VAL_COMPLEX(-771, -449), MAKE_VAL_COMPLEX(8, -1192), MAKE_VAL_COMPLEX(2383, 267), MAKE_VAL_COMPLEX(-226, -751), MAKE_VAL_COMPLEX(198, -1673), MAKE_VAL_COMPLEX(-254, -1169), MAKE_VAL_COMPLEX(88, -209), MAKE_VAL_COMPLEX(1584, 1899), MAKE_VAL_COMPLEX(-17, -289), MAKE_VAL_COMPLEX(-189, -637), MAKE_VAL_COMPLEX(-306, -799) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-9, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1095, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1095", y_true, y_val, y1056_LEN*y1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1095() */
/* END TEST 1095 */

/*
 *  BEGIN TEST 1096
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1096( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-39, 157), MAKE_VAL_COMPLEX(46, 289), MAKE_VAL_COMPLEX(5, 165), MAKE_VAL_COMPLEX(-60, 234), MAKE_VAL_COMPLEX(-48, 178), MAKE_VAL_COMPLEX(23, 3), MAKE_VAL_COMPLEX(192, -27), MAKE_VAL_COMPLEX(-104, 3), MAKE_VAL_COMPLEX(-3, -111), MAKE_VAL_COMPLEX(-56, -83), MAKE_VAL_COMPLEX(-38, 53), MAKE_VAL_COMPLEX(-215, -119), MAKE_VAL_COMPLEX(83, 75), MAKE_VAL_COMPLEX(-53, 141), MAKE_VAL_COMPLEX(27, 76) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1096, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1096", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1096() */
/* END TEST 1096 */

/*
 *  BEGIN TEST 1097
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1097( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(89, -132), MAKE_VAL_COMPLEX(-61, -184), MAKE_VAL_COMPLEX(-105, -140), MAKE_VAL_COMPLEX(25, -214), MAKE_VAL_COMPLEX(-37, -183), MAKE_VAL_COMPLEX(-83, -33), MAKE_VAL_COMPLEX(-302, 97), MAKE_VAL_COMPLEX(-31, 17), MAKE_VAL_COMPLEX(33, 26), MAKE_VAL_COMPLEX(36, -52), MAKE_VAL_COMPLEX(93, -113), MAKE_VAL_COMPLEX(150, 174), MAKE_VAL_COMPLEX(22, -35), MAKE_VAL_COMPLEX(8, -201), MAKE_VAL_COMPLEX(13, -181) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1097, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1097", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1097() */
/* END TEST 1097 */

/*
 *  BEGIN TEST 1098
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1098( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-35, 15), MAKE_VAL_COMPLEX(-48, -54), MAKE_VAL_COMPLEX(31, -63), MAKE_VAL_COMPLEX(5, -27), MAKE_VAL_COMPLEX(40, -42), MAKE_VAL_COMPLEX(42, -18), MAKE_VAL_COMPLEX(12, -88), MAKE_VAL_COMPLEX(49, -79), MAKE_VAL_COMPLEX(31, 53), MAKE_VAL_COMPLEX(79, 49), MAKE_VAL_COMPLEX(7, 55), MAKE_VAL_COMPLEX(0, -58), MAKE_VAL_COMPLEX(-67, 37), MAKE_VAL_COMPLEX(51, 3), MAKE_VAL_COMPLEX(37, 67) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1098, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1098", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1098() */
/* END TEST 1098 */

/*
 *  BEGIN TEST 1099
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1099( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-131, -462), MAKE_VAL_COMPLEX(-481, -339), MAKE_VAL_COMPLEX(-403, -96), MAKE_VAL_COMPLEX(-348, -497), MAKE_VAL_COMPLEX(-365, -291), MAKE_VAL_COMPLEX(-178, 124), MAKE_VAL_COMPLEX(-316, 732), MAKE_VAL_COMPLEX(90, 73), MAKE_VAL_COMPLEX(95, 62), MAKE_VAL_COMPLEX(-9, -58), MAKE_VAL_COMPLEX(-86, -357), MAKE_VAL_COMPLEX(703, -3), MAKE_VAL_COMPLEX(-120, -145), MAKE_VAL_COMPLEX(-344, -367), MAKE_VAL_COMPLEX(-367, -294) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-2, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1099, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1099", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1099() */
/* END TEST 1099 */

/*
 *  BEGIN TEST 1100
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1100( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(173, -9), MAKE_VAL_COMPLEX(190, -15), MAKE_VAL_COMPLEX(-3, 73), MAKE_VAL_COMPLEX(180, 22), MAKE_VAL_COMPLEX(68, -4), MAKE_VAL_COMPLEX(-71, -79), MAKE_VAL_COMPLEX(-216, -181), MAKE_VAL_COMPLEX(-164, -73), MAKE_VAL_COMPLEX(-15, -51), MAKE_VAL_COMPLEX(10, -125), MAKE_VAL_COMPLEX(114, 1), MAKE_VAL_COMPLEX(-287, 185), MAKE_VAL_COMPLEX(123, 115), MAKE_VAL_COMPLEX(163, -43), MAKE_VAL_COMPLEX(219, -72) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1100, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1100", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1100() */
/* END TEST 1100 */

/*
 *  BEGIN TEST 1101
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1101( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-178, 89), MAKE_VAL_COMPLEX(-337, 54), MAKE_VAL_COMPLEX(-86, -191), MAKE_VAL_COMPLEX(-226, -57), MAKE_VAL_COMPLEX(-109, -111), MAKE_VAL_COMPLEX(77, -17), MAKE_VAL_COMPLEX(62, 75), MAKE_VAL_COMPLEX(62, -94), MAKE_VAL_COMPLEX(144, 43), MAKE_VAL_COMPLEX(157, 23), MAKE_VAL_COMPLEX(-4, 38), MAKE_VAL_COMPLEX(178, -240), MAKE_VAL_COMPLEX(-117, 46), MAKE_VAL_COMPLEX(-109, -50), MAKE_VAL_COMPLEX(-58, 66) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1101, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1101", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1101() */
/* END TEST 1101 */

/*
 *  BEGIN TEST 1102
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1102( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, -15), MAKE_VAL_COMPLEX(27, -9), MAKE_VAL_COMPLEX(18, 21), MAKE_VAL_COMPLEX(9, 6), MAKE_VAL_COMPLEX(9, 21), MAKE_VAL_COMPLEX(0, 18), MAKE_VAL_COMPLEX(30, 18), MAKE_VAL_COMPLEX(21, 30), MAKE_VAL_COMPLEX(-24, 3), MAKE_VAL_COMPLEX(-30, 21), MAKE_VAL_COMPLEX(-21, -6), MAKE_VAL_COMPLEX(21, 9), MAKE_VAL_COMPLEX(-3, -30), MAKE_VAL_COMPLEX(-9, 18), MAKE_VAL_COMPLEX(-30, 3) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1102, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1102", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1102() */
/* END TEST 1102 */

/*
 *  BEGIN TEST 1103
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1056', LAYOUT_COLMAJ>
 *  y : <'y1056', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1103( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1056_ptr, A1056_ind, A1056_val,
		A1056_M, A1056_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1056_val,
		y1056_LEN * y1056_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1056_LEN, y1056_VW, LAYOUT_COLMAJ, y1056_LEN );
	oski_value_t* y_val = array_Duplicate( x1056_val,
		x1056_LEN * x1056_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1056_LEN, x1056_VW, LAYOUT_COLMAJ, x1056_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(984, 893), MAKE_VAL_COMPLEX(1831, 1583), MAKE_VAL_COMPLEX(-358, 1169), MAKE_VAL_COMPLEX(729, 1842), MAKE_VAL_COMPLEX(269, 1297), MAKE_VAL_COMPLEX(84, -338), MAKE_VAL_COMPLEX(1230, -1466), MAKE_VAL_COMPLEX(125, -386), MAKE_VAL_COMPLEX(-424, -585), MAKE_VAL_COMPLEX(-178, -223), MAKE_VAL_COMPLEX(-101, 426), MAKE_VAL_COMPLEX(-2123, -975), MAKE_VAL_COMPLEX(-79, 490), MAKE_VAL_COMPLEX(459, 1546), MAKE_VAL_COMPLEX(498, 1115) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1103, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1103", y_true, y_val, x1056_LEN*x1056_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1103() */
/* END TEST 1103 */

#define A1104_M 5    /* # of rows */
#define A1104_N 5    /* # of cols */
#define A1104_NNZ 4  /* # of stored non-zeros */
static oski_index_t A1104_ptr[A1104_M+1] = {0, 0, 0, 0, 2, 4}; /* end row pointers for A1104 */
static oski_index_t A1104_ind[A1104_NNZ] = {0, 2, 0, 3}; /* end column indices pointers for A1104 */
static oski_value_t A1104_val[A1104_NNZ] = {MAKE_VAL_COMPLEX(-10, -7), MAKE_VAL_COMPLEX(-5, -6), MAKE_VAL_COMPLEX(-2, 4), MAKE_VAL_COMPLEX(-7, -3)}; /* end stored non-zero values for A1104 */

#define x1104_LEN 5   /* vector length */
#define x1104_VW  3   /* vector width */
static oski_value_t x1104_val[x1104_LEN*x1104_VW] = {MAKE_VAL_COMPLEX(-9, -7), MAKE_VAL_COMPLEX(2, -2), MAKE_VAL_COMPLEX(8, -10), MAKE_VAL_COMPLEX(-10, 8), MAKE_VAL_COMPLEX(1, -6), MAKE_VAL_COMPLEX(-4, 10), MAKE_VAL_COMPLEX(0, -4), MAKE_VAL_COMPLEX(0, -4), MAKE_VAL_COMPLEX(-8, 10), MAKE_VAL_COMPLEX(8, 1), MAKE_VAL_COMPLEX(-10, 6), MAKE_VAL_COMPLEX(10, 10), MAKE_VAL_COMPLEX(7, -6), MAKE_VAL_COMPLEX(7, 7), MAKE_VAL_COMPLEX(-3, 3)}; /* data */

#define y1104_LEN 5   /* vector length */
#define y1104_VW  3   /* vector width */
static oski_value_t y1104_val[y1104_LEN*y1104_VW] = {MAKE_VAL_COMPLEX(10, 8), MAKE_VAL_COMPLEX(-3, -5), MAKE_VAL_COMPLEX(-7, 0), MAKE_VAL_COMPLEX(9, -6), MAKE_VAL_COMPLEX(7, -6), MAKE_VAL_COMPLEX(-8, 9), MAKE_VAL_COMPLEX(7, 8), MAKE_VAL_COMPLEX(-6, 10), MAKE_VAL_COMPLEX(5, -8), MAKE_VAL_COMPLEX(1, -7), MAKE_VAL_COMPLEX(4, 0), MAKE_VAL_COMPLEX(7, 5), MAKE_VAL_COMPLEX(4, -8), MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(-6, -3)}; /* data */

/*
 *  BEGIN TEST 1104
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1104( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	oski_value_t* y_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(16, 21), MAKE_VAL_COMPLEX(73, -15), MAKE_VAL_COMPLEX(-84, -191), MAKE_VAL_COMPLEX(56, -49), MAKE_VAL_COMPLEX(51, -61), MAKE_VAL_COMPLEX(-59, 90), MAKE_VAL_COMPLEX(30, 0), MAKE_VAL_COMPLEX(48, 112), MAKE_VAL_COMPLEX(34, -169), MAKE_VAL_COMPLEX(-41, 118), MAKE_VAL_COMPLEX(-64, -42), MAKE_VAL_COMPLEX(51, 73), MAKE_VAL_COMPLEX(24, -127), MAKE_VAL_COMPLEX(9, -63), MAKE_VAL_COMPLEX(-70, -63) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1104, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1104", y_true, y_val, y1104_LEN*y1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1104() */
/* END TEST 1104 */

/*
 *  BEGIN TEST 1105
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1105( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	oski_value_t* y_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(104, 75), MAKE_VAL_COMPLEX(-109, -45), MAKE_VAL_COMPLEX(0, 191), MAKE_VAL_COMPLEX(52, -23), MAKE_VAL_COMPLEX(33, -11), MAKE_VAL_COMPLEX(-37, 18), MAKE_VAL_COMPLEX(54, 96), MAKE_VAL_COMPLEX(-120, 8), MAKE_VAL_COMPLEX(26, 73), MAKE_VAL_COMPLEX(53, -202), MAKE_VAL_COMPLEX(112, 42), MAKE_VAL_COMPLEX(33, -13), MAKE_VAL_COMPLEX(24, 31), MAKE_VAL_COMPLEX(-129, -57), MAKE_VAL_COMPLEX(-2, 27) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1105, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1105", y_true, y_val, y1104_LEN*y1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1105() */
/* END TEST 1105 */

/*
 *  BEGIN TEST 1106
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1106( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	oski_value_t* y_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(88, -28), MAKE_VAL_COMPLEX(-42, -2), MAKE_VAL_COMPLEX(-28, 42), MAKE_VAL_COMPLEX(0, -78), MAKE_VAL_COMPLEX(-8, -66), MAKE_VAL_COMPLEX(22, 84), MAKE_VAL_COMPLEX(76, -10), MAKE_VAL_COMPLEX(36, 76), MAKE_VAL_COMPLEX(-28, -62), MAKE_VAL_COMPLEX(-38, -34), MAKE_VAL_COMPLEX(16, -24), MAKE_VAL_COMPLEX(58, -22), MAKE_VAL_COMPLEX(-32, -56), MAKE_VAL_COMPLEX(-100, 20), MAKE_VAL_COMPLEX(-42, 24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1106, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1106", y_true, y_val, y1104_LEN*y1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1106() */
/* END TEST 1106 */

/*
 *  BEGIN TEST 1107
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1107( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	oski_value_t* y_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-5, -4), MAKE_VAL_COMPLEX(5, 79), MAKE_VAL_COMPLEX(191, -70), MAKE_VAL_COMPLEX(37, 38), MAKE_VAL_COMPLEX(49, 37), MAKE_VAL_COMPLEX(-72, -43), MAKE_VAL_COMPLEX(16, 16), MAKE_VAL_COMPLEX(-92, 60), MAKE_VAL_COMPLEX(153, 24), MAKE_VAL_COMPLEX(-132, -43), MAKE_VAL_COMPLEX(42, -72), MAKE_VAL_COMPLEX(-63, 37), MAKE_VAL_COMPLEX(111, 16), MAKE_VAL_COMPLEX(43, 29), MAKE_VAL_COMPLEX(57, -58) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1107, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1107", y_true, y_val, y1104_LEN*y1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1107() */
/* END TEST 1107 */

/*
 *  BEGIN TEST 1108
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1108( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-33, 39), MAKE_VAL_COMPLEX(33, -53), MAKE_VAL_COMPLEX(54, -125), MAKE_VAL_COMPLEX(-63, -14), MAKE_VAL_COMPLEX(35, -26), MAKE_VAL_COMPLEX(-64, 33), MAKE_VAL_COMPLEX(-24, 21), MAKE_VAL_COMPLEX(-10, -30), MAKE_VAL_COMPLEX(-72, -67), MAKE_VAL_COMPLEX(-54, -159), MAKE_VAL_COMPLEX(65, 141), MAKE_VAL_COMPLEX(37, 183), MAKE_VAL_COMPLEX(-24, 66), MAKE_VAL_COMPLEX(-12, 32), MAKE_VAL_COMPLEX(-50, -87) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1108, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1108", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1108() */
/* END TEST 1108 */

/*
 *  BEGIN TEST 1109
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1109( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-73, -107), MAKE_VAL_COMPLEX(-13, 29), MAKE_VAL_COMPLEX(24, 7), MAKE_VAL_COMPLEX(-39, 112), MAKE_VAL_COMPLEX(-30, -41), MAKE_VAL_COMPLEX(30, 81), MAKE_VAL_COMPLEX(20, -65), MAKE_VAL_COMPLEX(6, -14), MAKE_VAL_COMPLEX(-6, 185), MAKE_VAL_COMPLEX(143, 162), MAKE_VAL_COMPLEX(-169, -65), MAKE_VAL_COMPLEX(83, -83), MAKE_VAL_COMPLEX(95, -139), MAKE_VAL_COMPLEX(96, 38), MAKE_VAL_COMPLEX(20, 123) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1109, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1109", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1109() */
/* END TEST 1109 */

/*
 *  BEGIN TEST 1110
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1110( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(18, 14), MAKE_VAL_COMPLEX(-4, 4), MAKE_VAL_COMPLEX(-16, 20), MAKE_VAL_COMPLEX(20, -16), MAKE_VAL_COMPLEX(-2, 12), MAKE_VAL_COMPLEX(8, -20), MAKE_VAL_COMPLEX(0, 8), MAKE_VAL_COMPLEX(0, 8), MAKE_VAL_COMPLEX(16, -20), MAKE_VAL_COMPLEX(-16, -2), MAKE_VAL_COMPLEX(20, -12), MAKE_VAL_COMPLEX(-20, -20), MAKE_VAL_COMPLEX(-14, 12), MAKE_VAL_COMPLEX(-14, -14), MAKE_VAL_COMPLEX(6, -6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1110, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1110", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1110() */
/* END TEST 1110 */

/*
 *  BEGIN TEST 1111
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1111( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(12, 336), MAKE_VAL_COMPLEX(-2, -180), MAKE_VAL_COMPLEX(-173, -351), MAKE_VAL_COMPLEX(19, -9), MAKE_VAL_COMPLEX(10, -32), MAKE_VAL_COMPLEX(-9, 49), MAKE_VAL_COMPLEX(-87, 147), MAKE_VAL_COMPLEX(-96, -20), MAKE_VAL_COMPLEX(-73, -207), MAKE_VAL_COMPLEX(-432, -510), MAKE_VAL_COMPLEX(540, 358), MAKE_VAL_COMPLEX(234, 272), MAKE_VAL_COMPLEX(-165, 249), MAKE_VAL_COMPLEX(-46, -60), MAKE_VAL_COMPLEX(-165, -229) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1111, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1111", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1111() */
/* END TEST 1111 */

/*
 *  BEGIN TEST 1112
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1112( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-56, 20), MAKE_VAL_COMPLEX(-45, 69), MAKE_VAL_COMPLEX(-38, -43), MAKE_VAL_COMPLEX(-79, 48), MAKE_VAL_COMPLEX(21, -53), MAKE_VAL_COMPLEX(-50, 85), MAKE_VAL_COMPLEX(48, 17), MAKE_VAL_COMPLEX(-22, 2), MAKE_VAL_COMPLEX(-134, 81), MAKE_VAL_COMPLEX(-179, 69), MAKE_VAL_COMPLEX(53, 21), MAKE_VAL_COMPLEX(221, 119), MAKE_VAL_COMPLEX(92, -53), MAKE_VAL_COMPLEX(-3, 87), MAKE_VAL_COMPLEX(-83, 32) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1112, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1112", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1112() */
/* END TEST 1112 */

/*
 *  BEGIN TEST 1113
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1113( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-67, -159), MAKE_VAL_COMPLEX(83, -95), MAKE_VAL_COMPLEX(196, -93), MAKE_VAL_COMPLEX(-105, 50), MAKE_VAL_COMPLEX(13, -40), MAKE_VAL_COMPLEX(-44, 63), MAKE_VAL_COMPLEX(-36, -81), MAKE_VAL_COMPLEX(34, -66), MAKE_VAL_COMPLEX(-24, 55), MAKE_VAL_COMPLEX(304, -29), MAKE_VAL_COMPLEX(-231, 45), MAKE_VAL_COMPLEX(-91, 71), MAKE_VAL_COMPLEX(38, -22), MAKE_VAL_COMPLEX(94, 46), MAKE_VAL_COMPLEX(26, 7) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1113, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1113", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1113() */
/* END TEST 1113 */

/*
 *  BEGIN TEST 1114
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1114( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-42, 54), MAKE_VAL_COMPLEX(-12, -12), MAKE_VAL_COMPLEX(-60, -48), MAKE_VAL_COMPLEX(48, 60), MAKE_VAL_COMPLEX(-36, -6), MAKE_VAL_COMPLEX(60, 24), MAKE_VAL_COMPLEX(-24, 0), MAKE_VAL_COMPLEX(-24, 0), MAKE_VAL_COMPLEX(60, 48), MAKE_VAL_COMPLEX(6, -48), MAKE_VAL_COMPLEX(36, 60), MAKE_VAL_COMPLEX(60, -60), MAKE_VAL_COMPLEX(-36, -42), MAKE_VAL_COMPLEX(42, -42), MAKE_VAL_COMPLEX(18, 18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1114, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1114", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1114() */
/* END TEST 1114 */

/*
 *  BEGIN TEST 1115
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1115( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-521, 395), MAKE_VAL_COMPLEX(-874, -16), MAKE_VAL_COMPLEX(-659, -605), MAKE_VAL_COMPLEX(-7, -23), MAKE_VAL_COMPLEX(128, 9), MAKE_VAL_COMPLEX(-195, -13), MAKE_VAL_COMPLEX(-91, 541), MAKE_VAL_COMPLEX(-336, -24), MAKE_VAL_COMPLEX(-467, -409), MAKE_VAL_COMPLEX(-1565, -1362), MAKE_VAL_COMPLEX(736, 826), MAKE_VAL_COMPLEX(502, 1312), MAKE_VAL_COMPLEX(303, 188), MAKE_VAL_COMPLEX(-463, -167), MAKE_VAL_COMPLEX(-390, -352) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1115, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1115", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1115() */
/* END TEST 1115 */

/*
 *  BEGIN TEST 1116
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1116( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	oski_value_t* y_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-218, -35), MAKE_VAL_COMPLEX(52, 32), MAKE_VAL_COMPLEX(29, -106), MAKE_VAL_COMPLEX(-81, 292), MAKE_VAL_COMPLEX(-21, 44), MAKE_VAL_COMPLEX(111, -115), MAKE_VAL_COMPLEX(-118, -29), MAKE_VAL_COMPLEX(70, -106), MAKE_VAL_COMPLEX(-85, 69), MAKE_VAL_COMPLEX(139, 77), MAKE_VAL_COMPLEX(-64, -170), MAKE_VAL_COMPLEX(-102, 14), MAKE_VAL_COMPLEX(22, 0), MAKE_VAL_COMPLEX(150, 60), MAKE_VAL_COMPLEX(56, -60) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1116, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1116", y_true, y_val, y1104_LEN*y1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1116() */
/* END TEST 1116 */

/*
 *  BEGIN TEST 1117
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1117( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	oski_value_t* y_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(186, -7), MAKE_VAL_COMPLEX(-45, -9), MAKE_VAL_COMPLEX(-1, 113), MAKE_VAL_COMPLEX(39, -277), MAKE_VAL_COMPLEX(-13, -27), MAKE_VAL_COMPLEX(-70, 87), MAKE_VAL_COMPLEX(98, -10), MAKE_VAL_COMPLEX(-36, 72), MAKE_VAL_COMPLEX(57, -42), MAKE_VAL_COMPLEX(-150, -50), MAKE_VAL_COMPLEX(48, 166), MAKE_VAL_COMPLEX(79, -41), MAKE_VAL_COMPLEX(-46, 28), MAKE_VAL_COMPLEX(-120, -10), MAKE_VAL_COMPLEX(-35, 78) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1117, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1117", y_true, y_val, y1104_LEN*y1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1117() */
/* END TEST 1117 */

/*
 *  BEGIN TEST 1118
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1118( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	oski_value_t* y_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(52, 58), MAKE_VAL_COMPLEX(-13, -33), MAKE_VAL_COMPLEX(-42, -7), MAKE_VAL_COMPLEX(60, -27), MAKE_VAL_COMPLEX(48, -29), MAKE_VAL_COMPLEX(-57, 46), MAKE_VAL_COMPLEX(34, 55), MAKE_VAL_COMPLEX(-46, 54), MAKE_VAL_COMPLEX(38, -43), MAKE_VAL_COMPLEX(13, -41), MAKE_VAL_COMPLEX(24, 4), MAKE_VAL_COMPLEX(37, 37), MAKE_VAL_COMPLEX(32, -44), MAKE_VAL_COMPLEX(-50, -70), MAKE_VAL_COMPLEX(-33, -24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1118, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1118", y_true, y_val, y1104_LEN*y1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1118() */
/* END TEST 1118 */

/*
 *  BEGIN TEST 1119
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1119( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	oski_value_t* y_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-89, -104), MAKE_VAL_COMPLEX(-32, -12), MAKE_VAL_COMPLEX(-32, -155), MAKE_VAL_COMPLEX(-226, 299), MAKE_VAL_COMPLEX(86, -129), MAKE_VAL_COMPLEX(149, 165), MAKE_VAL_COMPLEX(21, 20), MAKE_VAL_COMPLEX(118, 162), MAKE_VAL_COMPLEX(-131, -191), MAKE_VAL_COMPLEX(201, 93), MAKE_VAL_COMPLEX(154, -420), MAKE_VAL_COMPLEX(-2, 38), MAKE_VAL_COMPLEX(124, -218), MAKE_VAL_COMPLEX(-80, 30), MAKE_VAL_COMPLEX(-8, -124) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1119, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1119", y_true, y_val, y1104_LEN*y1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1119() */
/* END TEST 1119 */

/*
 *  BEGIN TEST 1120
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1120( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-132, 15), MAKE_VAL_COMPLEX(-80, 68), MAKE_VAL_COMPLEX(92, 130), MAKE_VAL_COMPLEX(-41, 9), MAKE_VAL_COMPLEX(16, -9), MAKE_VAL_COMPLEX(-15, 31), MAKE_VAL_COMPLEX(-84, -36), MAKE_VAL_COMPLEX(-75, 8), MAKE_VAL_COMPLEX(-40, 124), MAKE_VAL_COMPLEX(-44, -82), MAKE_VAL_COMPLEX(174, -2), MAKE_VAL_COMPLEX(-65, 57), MAKE_VAL_COMPLEX(-99, 22), MAKE_VAL_COMPLEX(-64, 12), MAKE_VAL_COMPLEX(14, 119) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1120, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1120", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1120() */
/* END TEST 1120 */

/*
 *  BEGIN TEST 1121
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1121( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(179, 36), MAKE_VAL_COMPLEX(66, -58), MAKE_VAL_COMPLEX(-150, -78), MAKE_VAL_COMPLEX(109, -47), MAKE_VAL_COMPLEX(-28, 44), MAKE_VAL_COMPLEX(49, -87), MAKE_VAL_COMPLEX(80, 60), MAKE_VAL_COMPLEX(71, 16), MAKE_VAL_COMPLEX(98, -176), MAKE_VAL_COMPLEX(-3, 68), MAKE_VAL_COMPLEX(-108, -24), MAKE_VAL_COMPLEX(15, -127), MAKE_VAL_COMPLEX(51, 7), MAKE_VAL_COMPLEX(29, -61), MAKE_VAL_COMPLEX(7, -134) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1121, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1121", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1121() */
/* END TEST 1121 */

/*
 *  BEGIN TEST 1122
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1122( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-43, -19), MAKE_VAL_COMPLEX(6, -10), MAKE_VAL_COMPLEX(22, -48), MAKE_VAL_COMPLEX(-32, 42), MAKE_VAL_COMPLEX(-2, -25), MAKE_VAL_COMPLEX(-6, 44), MAKE_VAL_COMPLEX(-4, -16), MAKE_VAL_COMPLEX(-4, -16), MAKE_VAL_COMPLEX(-22, 48), MAKE_VAL_COMPLEX(33, -4), MAKE_VAL_COMPLEX(-34, 34), MAKE_VAL_COMPLEX(50, 30), MAKE_VAL_COMPLEX(22, -31), MAKE_VAL_COMPLEX(35, 21), MAKE_VAL_COMPLEX(-9, 15) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1122, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1122", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1122() */
/* END TEST 1122 */

/*
 *  BEGIN TEST 1123
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1123( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1139, -368), MAKE_VAL_COMPLEX(-1224, -24), MAKE_VAL_COMPLEX(-664, 1702), MAKE_VAL_COMPLEX(91, -125), MAKE_VAL_COMPLEX(-8, 161), MAKE_VAL_COMPLEX(53, 9), MAKE_VAL_COMPLEX(-536, -776), MAKE_VAL_COMPLEX(-772, -361), MAKE_VAL_COMPLEX(-698, 694), MAKE_VAL_COMPLEX(51, -1221), MAKE_VAL_COMPLEX(1830, 1342), MAKE_VAL_COMPLEX(-879, -539), MAKE_VAL_COMPLEX(-1295, -570), MAKE_VAL_COMPLEX(-568, -730), MAKE_VAL_COMPLEX(-559, 1062) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1123, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1123", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1123() */
/* END TEST 1123 */

/*
 *  BEGIN TEST 1124
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1124( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-4, 41), MAKE_VAL_COMPLEX(-16, 154), MAKE_VAL_COMPLEX(256, 142), MAKE_VAL_COMPLEX(-65, -103), MAKE_VAL_COMPLEX(60, 23), MAKE_VAL_COMPLEX(-79, -41), MAKE_VAL_COMPLEX(20, 88), MAKE_VAL_COMPLEX(53, 84), MAKE_VAL_COMPLEX(32, -100), MAKE_VAL_COMPLEX(-160, 76), MAKE_VAL_COMPLEX(-38, -284), MAKE_VAL_COMPLEX(-71, 165), MAKE_VAL_COMPLEX(21, 76), MAKE_VAL_COMPLEX(-28, 134), MAKE_VAL_COMPLEX(62, -9) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1124, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1124", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1124() */
/* END TEST 1124 */

/*
 *  BEGIN TEST 1125
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1125( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(204, -131), MAKE_VAL_COMPLEX(32, -102), MAKE_VAL_COMPLEX(-158, 84), MAKE_VAL_COMPLEX(19, -139), MAKE_VAL_COMPLEX(33, 48), MAKE_VAL_COMPLEX(-55, -117), MAKE_VAL_COMPLEX(48, -52), MAKE_VAL_COMPLEX(15, -48), MAKE_VAL_COMPLEX(-130, -126), MAKE_VAL_COMPLEX(71, 51), MAKE_VAL_COMPLEX(26, 60), MAKE_VAL_COMPLEX(-189, -85), MAKE_VAL_COMPLEX(18, 97), MAKE_VAL_COMPLEX(-154, -78), MAKE_VAL_COMPLEX(-86, -69) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1125, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1125", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1125() */
/* END TEST 1125 */

/*
 *  BEGIN TEST 1126
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1126( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(101, -37), MAKE_VAL_COMPLEX(6, 26), MAKE_VAL_COMPLEX(40, 114), MAKE_VAL_COMPLEX(-14, -120), MAKE_VAL_COMPLEX(43, 38), MAKE_VAL_COMPLEX(-60, -82), MAKE_VAL_COMPLEX(32, 20), MAKE_VAL_COMPLEX(32, 20), MAKE_VAL_COMPLEX(-40, -114), MAKE_VAL_COMPLEX(-48, 59), MAKE_VAL_COMPLEX(2, -110), MAKE_VAL_COMPLEX(-130, 30), MAKE_VAL_COMPLEX(13, 86), MAKE_VAL_COMPLEX(-91, 21), MAKE_VAL_COMPLEX(-9, -39) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1126, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1126", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1126() */
/* END TEST 1126 */

/*
 *  BEGIN TEST 1127
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_ROWMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1127( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_ROWMAJ, x1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-864, 4), MAKE_VAL_COMPLEX(-936, 376), MAKE_VAL_COMPLEX(334, 1430), MAKE_VAL_COMPLEX(52, -180), MAKE_VAL_COMPLEX(13, 135), MAKE_VAL_COMPLEX(-36, -56), MAKE_VAL_COMPLEX(-540, 276), MAKE_VAL_COMPLEX(-384, 458), MAKE_VAL_COMPLEX(566, 466), MAKE_VAL_COMPLEX(-681, -791), MAKE_VAL_COMPLEX(1200, -840), MAKE_VAL_COMPLEX(-554, 494), MAKE_VAL_COMPLEX(-167, -25), MAKE_VAL_COMPLEX(-412, 550), MAKE_VAL_COMPLEX(224, 552) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1127, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1127", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1127() */
/* END TEST 1127 */

/*
 *  BEGIN TEST 1128
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1128( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	oski_value_t* y_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(181, -57), MAKE_VAL_COMPLEX(112, 18), MAKE_VAL_COMPLEX(-23, -103), MAKE_VAL_COMPLEX(-40, -26), MAKE_VAL_COMPLEX(-38, -20), MAKE_VAL_COMPLEX(62, 24), MAKE_VAL_COMPLEX(124, -34), MAKE_VAL_COMPLEX(152, -2), MAKE_VAL_COMPLEX(-28, -87), MAKE_VAL_COMPLEX(-124, 192), MAKE_VAL_COMPLEX(17, -89), MAKE_VAL_COMPLEX(114, -45), MAKE_VAL_COMPLEX(101, -54), MAKE_VAL_COMPLEX(42, -21), MAKE_VAL_COMPLEX(-35, -89) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1128, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1128", y_true, y_val, y1104_LEN*y1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1128() */
/* END TEST 1128 */

/*
 *  BEGIN TEST 1129
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1129( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	oski_value_t* y_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-291, -31), MAKE_VAL_COMPLEX(-79, 37), MAKE_VAL_COMPLEX(100, 103), MAKE_VAL_COMPLEX(-59, 92), MAKE_VAL_COMPLEX(-39, 86), MAKE_VAL_COMPLEX(26, -123), MAKE_VAL_COMPLEX(-201, -54), MAKE_VAL_COMPLEX(-86, -108), MAKE_VAL_COMPLEX(-27, 175), MAKE_VAL_COMPLEX(113, -115), MAKE_VAL_COMPLEX(-61, 89), MAKE_VAL_COMPLEX(-191, -10), MAKE_VAL_COMPLEX(-145, 142), MAKE_VAL_COMPLEX(68, 131), MAKE_VAL_COMPLEX(101, 122) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1129, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1129", y_true, y_val, y1104_LEN*y1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1129() */
/* END TEST 1129 */

/*
 *  BEGIN TEST 1130
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1130( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	oski_value_t* y_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-4, -118), MAKE_VAL_COMPLEX(-17, 51), MAKE_VAL_COMPLEX(42, 49), MAKE_VAL_COMPLEX(-96, -27), MAKE_VAL_COMPLEX(-84, -13), MAKE_VAL_COMPLEX(111, 2), MAKE_VAL_COMPLEX(14, -97), MAKE_VAL_COMPLEX(106, -18), MAKE_VAL_COMPLEX(-86, 13), MAKE_VAL_COMPLEX(-55, 35), MAKE_VAL_COMPLEX(-24, -28), MAKE_VAL_COMPLEX(-7, -79), MAKE_VAL_COMPLEX(-80, 20), MAKE_VAL_COMPLEX(-10, 130), MAKE_VAL_COMPLEX(15, 60) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1130, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1130", y_true, y_val, y1104_LEN*y1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1130() */
/* END TEST 1130 */

/*
 *  BEGIN TEST 1131
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1131( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	oski_value_t* y_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-826, -1272), MAKE_VAL_COMPLEX(-684, -826), MAKE_VAL_COMPLEX(-711, 963), MAKE_VAL_COMPLEX(-93, -49), MAKE_VAL_COMPLEX(-91, -11), MAKE_VAL_COMPLEX(107, -91), MAKE_VAL_COMPLEX(-425, -865), MAKE_VAL_COMPLEX(-454, -658), MAKE_VAL_COMPLEX(-722, 306), MAKE_VAL_COMPLEX(1692, -238), MAKE_VAL_COMPLEX(-648, 162), MAKE_VAL_COMPLEX(-575, -785), MAKE_VAL_COMPLEX(-1151, -721), MAKE_VAL_COMPLEX(-917, 71), MAKE_VAL_COMPLEX(-661, 891) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1131, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1131", y_true, y_val, y1104_LEN*y1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1131() */
/* END TEST 1131 */

/*
 *  BEGIN TEST 1132
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1132( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-21, 135), MAKE_VAL_COMPLEX(1, -6), MAKE_VAL_COMPLEX(-76, 41), MAKE_VAL_COMPLEX(-46, -191), MAKE_VAL_COMPLEX(-90, 72), MAKE_VAL_COMPLEX(45, -65), MAKE_VAL_COMPLEX(-1, 2), MAKE_VAL_COMPLEX(-34, -6), MAKE_VAL_COMPLEX(165, 153), MAKE_VAL_COMPLEX(-26, -42), MAKE_VAL_COMPLEX(14, -109), MAKE_VAL_COMPLEX(-8, -31), MAKE_VAL_COMPLEX(-26, -77), MAKE_VAL_COMPLEX(37, 75), MAKE_VAL_COMPLEX(-14, -87) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1132, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1132", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1132() */
/* END TEST 1132 */

/*
 *  BEGIN TEST 1133
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1133( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(73, -109), MAKE_VAL_COMPLEX(-9, 18), MAKE_VAL_COMPLEX(46, 17), MAKE_VAL_COMPLEX(88, 141), MAKE_VAL_COMPLEX(91, -41), MAKE_VAL_COMPLEX(-35, 11), MAKE_VAL_COMPLEX(5, 18), MAKE_VAL_COMPLEX(38, 26), MAKE_VAL_COMPLEX(-135, -211), MAKE_VAL_COMPLEX(-15, 45), MAKE_VAL_COMPLEX(30, 69), MAKE_VAL_COMPLEX(-52, -9), MAKE_VAL_COMPLEX(-3, 114), MAKE_VAL_COMPLEX(-79, -103), MAKE_VAL_COMPLEX(26, 69) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1133, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1133", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1133() */
/* END TEST 1133 */

/*
 *  BEGIN TEST 1134
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1134( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-63, -49), MAKE_VAL_COMPLEX(14, -14), MAKE_VAL_COMPLEX(56, -70), MAKE_VAL_COMPLEX(-70, 56), MAKE_VAL_COMPLEX(7, -42), MAKE_VAL_COMPLEX(-28, 70), MAKE_VAL_COMPLEX(0, -28), MAKE_VAL_COMPLEX(0, -28), MAKE_VAL_COMPLEX(-56, 70), MAKE_VAL_COMPLEX(56, 7), MAKE_VAL_COMPLEX(-70, 42), MAKE_VAL_COMPLEX(70, 70), MAKE_VAL_COMPLEX(49, -42), MAKE_VAL_COMPLEX(49, 49), MAKE_VAL_COMPLEX(-21, 21) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1134, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1134", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1134() */
/* END TEST 1134 */

/*
 *  BEGIN TEST 1135
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1135( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1009, 679), MAKE_VAL_COMPLEX(106, 26), MAKE_VAL_COMPLEX(-678, -8), MAKE_VAL_COMPLEX(968, -2218), MAKE_VAL_COMPLEX(-1126, -91), MAKE_VAL_COMPLEX(606, -316), MAKE_VAL_COMPLEX(92, 24), MAKE_VAL_COMPLEX(-108, -304), MAKE_VAL_COMPLEX(-162, 2272), MAKE_VAL_COMPLEX(67, -316), MAKE_VAL_COMPLEX(850, -1074), MAKE_VAL_COMPLEX(-146, -62), MAKE_VAL_COMPLEX(554, -597), MAKE_VAL_COMPLEX(-535, 1071), MAKE_VAL_COMPLEX(509, -907) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1135, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1135", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1135() */
/* END TEST 1135 */

/*
 *  BEGIN TEST 1136
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1136( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(30, 58), MAKE_VAL_COMPLEX(15, 0), MAKE_VAL_COMPLEX(74, 73), MAKE_VAL_COMPLEX(-266, 23), MAKE_VAL_COMPLEX(48, -9), MAKE_VAL_COMPLEX(-93, 71), MAKE_VAL_COMPLEX(19, -6), MAKE_VAL_COMPLEX(-14, 34), MAKE_VAL_COMPLEX(109, -41), MAKE_VAL_COMPLEX(-55, 48), MAKE_VAL_COMPLEX(-130, -1), MAKE_VAL_COMPLEX(-38, 39), MAKE_VAL_COMPLEX(-42, 30), MAKE_VAL_COMPLEX(130, 50), MAKE_VAL_COMPLEX(-65, 2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1136, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1136", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1136() */
/* END TEST 1136 */

/*
 *  BEGIN TEST 1137
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1137( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-15, -133), MAKE_VAL_COMPLEX(3, 6), MAKE_VAL_COMPLEX(10, -55), MAKE_VAL_COMPLEX(188, -59), MAKE_VAL_COMPLEX(-9, -3), MAKE_VAL_COMPLEX(21, -65), MAKE_VAL_COMPLEX(5, -6), MAKE_VAL_COMPLEX(38, -46), MAKE_VAL_COMPLEX(-193, 23), MAKE_VAL_COMPLEX(73, 3), MAKE_VAL_COMPLEX(64, -41), MAKE_VAL_COMPLEX(8, 51), MAKE_VAL_COMPLEX(99, -6), MAKE_VAL_COMPLEX(-151, 13), MAKE_VAL_COMPLEX(38, -11) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1137, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1137", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1137() */
/* END TEST 1137 */

/*
 *  BEGIN TEST 1138
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1138( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(74, -58), MAKE_VAL_COMPLEX(12, 20), MAKE_VAL_COMPLEX(64, 84), MAKE_VAL_COMPLEX(-44, -96), MAKE_VAL_COMPLEX(46, 20), MAKE_VAL_COMPLEX(-72, -52), MAKE_VAL_COMPLEX(32, 8), MAKE_VAL_COMPLEX(32, 8), MAKE_VAL_COMPLEX(-64, -84), MAKE_VAL_COMPLEX(-24, 62), MAKE_VAL_COMPLEX(-28, -92), MAKE_VAL_COMPLEX(-100, 60), MAKE_VAL_COMPLEX(34, 68), MAKE_VAL_COMPLEX(-70, 42), MAKE_VAL_COMPLEX(-18, -30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1138, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1138", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1138() */
/* END TEST 1138 */

/*
 *  BEGIN TEST 1139
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1139( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_ROWMAJ, y1104_VW );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(645, -339), MAKE_VAL_COMPLEX(-102, -50), MAKE_VAL_COMPLEX(134, -548), MAKE_VAL_COMPLEX(1472, 1742), MAKE_VAL_COMPLEX(-250, -173), MAKE_VAL_COMPLEX(982, 152), MAKE_VAL_COMPLEX(-96, -16), MAKE_VAL_COMPLEX(356, 88), MAKE_VAL_COMPLEX(-610, -1044), MAKE_VAL_COMPLEX(373, 276), MAKE_VAL_COMPLEX(750, 806), MAKE_VAL_COMPLEX(114, -62), MAKE_VAL_COMPLEX(254, 433), MAKE_VAL_COMPLEX(-365, -1387), MAKE_VAL_COMPLEX(339, 407) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1139, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1139", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1139() */
/* END TEST 1139 */

/*
 *  BEGIN TEST 1140
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1140( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	oski_value_t* y_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(175, 53), MAKE_VAL_COMPLEX(8, -26), MAKE_VAL_COMPLEX(85, -11), MAKE_VAL_COMPLEX(-49, 191), MAKE_VAL_COMPLEX(180, -51), MAKE_VAL_COMPLEX(75, -1), MAKE_VAL_COMPLEX(-3, 41), MAKE_VAL_COMPLEX(52, 6), MAKE_VAL_COMPLEX(64, -82), MAKE_VAL_COMPLEX(86, -99), MAKE_VAL_COMPLEX(-25, -119), MAKE_VAL_COMPLEX(16, 46), MAKE_VAL_COMPLEX(50, -95), MAKE_VAL_COMPLEX(108, -67), MAKE_VAL_COMPLEX(-44, -146) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1140, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1140", y_true, y_val, y1104_LEN*y1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1140() */
/* END TEST 1140 */

/*
 *  BEGIN TEST 1141
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1141( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	oski_value_t* y_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-247, 37), MAKE_VAL_COMPLEX(37, -1), MAKE_VAL_COMPLEX(-85, -52), MAKE_VAL_COMPLEX(103, -110), MAKE_VAL_COMPLEX(-126, 114), MAKE_VAL_COMPLEX(-156, -71), MAKE_VAL_COMPLEX(-69, 22), MAKE_VAL_COMPLEX(-142, -60), MAKE_VAL_COMPLEX(8, 127), MAKE_VAL_COMPLEX(-23, 108), MAKE_VAL_COMPLEX(25, 155), MAKE_VAL_COMPLEX(-61, 17), MAKE_VAL_COMPLEX(22, 131), MAKE_VAL_COMPLEX(-18, -23), MAKE_VAL_COMPLEX(71, 92) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1141, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1141", y_true, y_val, y1104_LEN*y1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1141() */
/* END TEST 1141 */

/*
 *  BEGIN TEST 1142
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1142( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	oski_value_t* y_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-22, -116), MAKE_VAL_COMPLEX(-9, 53), MAKE_VAL_COMPLEX(49, 42), MAKE_VAL_COMPLEX(-99, -12), MAKE_VAL_COMPLEX(-85, 0), MAKE_VAL_COMPLEX(110, -15), MAKE_VAL_COMPLEX(-1, -98), MAKE_VAL_COMPLEX(102, -34), MAKE_VAL_COMPLEX(-83, 26), MAKE_VAL_COMPLEX(-49, 43), MAKE_VAL_COMPLEX(-28, -24), MAKE_VAL_COMPLEX(-19, -77), MAKE_VAL_COMPLEX(-76, 32), MAKE_VAL_COMPLEX(10, 130), MAKE_VAL_COMPLEX(24, 57) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1142, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1142", y_true, y_val, y1104_LEN*y1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1142() */
/* END TEST 1142 */

/*
 *  BEGIN TEST 1143
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1143( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	oski_value_t* y_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(922, -1056), MAKE_VAL_COMPLEX(31, 5), MAKE_VAL_COMPLEX(745, -590), MAKE_VAL_COMPLEX(477, 1716), MAKE_VAL_COMPLEX(485, -1114), MAKE_VAL_COMPLEX(770, -859), MAKE_VAL_COMPLEX(-89, -66), MAKE_VAL_COMPLEX(586, -718), MAKE_VAL_COMPLEX(-307, -522), MAKE_VAL_COMPLEX(-107, -807), MAKE_VAL_COMPLEX(-1036, -556), MAKE_VAL_COMPLEX(61, -21), MAKE_VAL_COMPLEX(-426, -518), MAKE_VAL_COMPLEX(696, -640), MAKE_VAL_COMPLEX(-876, -495) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1143, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1143", y_true, y_val, y1104_LEN*y1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1143() */
/* END TEST 1143 */

/*
 *  BEGIN TEST 1144
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1144( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-125, 136), MAKE_VAL_COMPLEX(-25, -11), MAKE_VAL_COMPLEX(-190, -40), MAKE_VAL_COMPLEX(29, -55), MAKE_VAL_COMPLEX(-172, 50), MAKE_VAL_COMPLEX(-2, 60), MAKE_VAL_COMPLEX(-21, 24), MAKE_VAL_COMPLEX(-107, 36), MAKE_VAL_COMPLEX(312, 6), MAKE_VAL_COMPLEX(-103, -76), MAKE_VAL_COMPLEX(140, 198), MAKE_VAL_COMPLEX(37, -105), MAKE_VAL_COMPLEX(-76, 77), MAKE_VAL_COMPLEX(-64, -60), MAKE_VAL_COMPLEX(59, 122) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1144, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1144", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1144() */
/* END TEST 1144 */

/*
 *  BEGIN TEST 1145
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1145( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(104, -109), MAKE_VAL_COMPLEX(19, 5), MAKE_VAL_COMPLEX(160, 16), MAKE_VAL_COMPLEX(-5, 85), MAKE_VAL_COMPLEX(154, -53), MAKE_VAL_COMPLEX(32, -48), MAKE_VAL_COMPLEX(9, -24), MAKE_VAL_COMPLEX(95, -36), MAKE_VAL_COMPLEX(-282, 18), MAKE_VAL_COMPLEX(106, 52), MAKE_VAL_COMPLEX(-122, -168), MAKE_VAL_COMPLEX(-7, 75), MAKE_VAL_COMPLEX(58, -98), MAKE_VAL_COMPLEX(85, 39), MAKE_VAL_COMPLEX(-50, -113) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1145, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1145", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1145() */
/* END TEST 1145 */

/*
 *  BEGIN TEST 1146
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1146( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(117, -39), MAKE_VAL_COMPLEX(6, 30), MAKE_VAL_COMPLEX(42, 132), MAKE_VAL_COMPLEX(-12, -138), MAKE_VAL_COMPLEX(48, 45), MAKE_VAL_COMPLEX(-66, -96), MAKE_VAL_COMPLEX(36, 24), MAKE_VAL_COMPLEX(36, 24), MAKE_VAL_COMPLEX(-42, -132), MAKE_VAL_COMPLEX(-57, 66), MAKE_VAL_COMPLEX(6, -126), MAKE_VAL_COMPLEX(-150, 30), MAKE_VAL_COMPLEX(12, 99), MAKE_VAL_COMPLEX(-105, 21), MAKE_VAL_COMPLEX(-9, -45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1146, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1146", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1146() */
/* END TEST 1146 */

/*
 *  BEGIN TEST 1147
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1147( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(791, -139), MAKE_VAL_COMPLEX(30, 46), MAKE_VAL_COMPLEX(584, 354), MAKE_VAL_COMPLEX(122, 636), MAKE_VAL_COMPLEX(873, 44), MAKE_VAL_COMPLEX(580, -258), MAKE_VAL_COMPLEX(6, -74), MAKE_VAL_COMPLEX(546, 26), MAKE_VAL_COMPLEX(-1384, -394), MAKE_VAL_COMPLEX(452, 319), MAKE_VAL_COMPLEX(-122, -1090), MAKE_VAL_COMPLEX(-82, 66), MAKE_VAL_COMPLEX(309, -562), MAKE_VAL_COMPLEX(509, 145), MAKE_VAL_COMPLEX(37, -745) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1147, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1147", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1147() */
/* END TEST 1147 */

/*
 *  BEGIN TEST 1148
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1148( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-44, 241), MAKE_VAL_COMPLEX(-35, -1), MAKE_VAL_COMPLEX(-158, 118), MAKE_VAL_COMPLEX(3, 3), MAKE_VAL_COMPLEX(-77, 54), MAKE_VAL_COMPLEX(74, 72), MAKE_VAL_COMPLEX(-21, 44), MAKE_VAL_COMPLEX(-11, 116), MAKE_VAL_COMPLEX(148, -222), MAKE_VAL_COMPLEX(-23, 13), MAKE_VAL_COMPLEX(306, 76), MAKE_VAL_COMPLEX(-13, -155), MAKE_VAL_COMPLEX(9, -13), MAKE_VAL_COMPLEX(15, -27), MAKE_VAL_COMPLEX(134, 15) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1148, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1148", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1148() */
/* END TEST 1148 */

/*
 *  BEGIN TEST 1149
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1149( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(20, -173), MAKE_VAL_COMPLEX(19, -7), MAKE_VAL_COMPLEX(82, -146), MAKE_VAL_COMPLEX(65, 41), MAKE_VAL_COMPLEX(39, -48), MAKE_VAL_COMPLEX(-6, -68), MAKE_VAL_COMPLEX(-3, -36), MAKE_VAL_COMPLEX(-13, -108), MAKE_VAL_COMPLEX(-72, 250), MAKE_VAL_COMPLEX(13, -63), MAKE_VAL_COMPLEX(-250, -28), MAKE_VAL_COMPLEX(53, 75), MAKE_VAL_COMPLEX(-59, -17), MAKE_VAL_COMPLEX(13, -29), MAKE_VAL_COMPLEX(-110, -3) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1149, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1149", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1149() */
/* END TEST 1149 */

/*
 *  BEGIN TEST 1150
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1150( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-10, -80), MAKE_VAL_COMPLEX(20, 0), MAKE_VAL_COMPLEX(90, -10), MAKE_VAL_COMPLEX(-90, -10), MAKE_VAL_COMPLEX(35, -25), MAKE_VAL_COMPLEX(-70, 30), MAKE_VAL_COMPLEX(20, -20), MAKE_VAL_COMPLEX(20, -20), MAKE_VAL_COMPLEX(-90, 10), MAKE_VAL_COMPLEX(35, 45), MAKE_VAL_COMPLEX(-80, -20), MAKE_VAL_COMPLEX(0, 100), MAKE_VAL_COMPLEX(65, 5), MAKE_VAL_COMPLEX(0, 70), MAKE_VAL_COMPLEX(-30, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1150, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1150", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1150() */
/* END TEST 1150 */

/*
 *  BEGIN TEST 1151
 *  Value type: complex
 *  A : <MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1104', LAYOUT_COLMAJ>
 *  y : <'y1104', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1151( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1104_ptr, A1104_ind, A1104_val,
		A1104_M, A1104_N, SHARE_INPUTMAT,
		5, MAT_SYMM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1104_val,
		y1104_LEN * y1104_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1104_LEN, y1104_VW, LAYOUT_COLMAJ, y1104_LEN );
	oski_value_t* y_val = array_Duplicate( x1104_val,
		x1104_LEN * x1104_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1104_LEN, x1104_VW, LAYOUT_COLMAJ, x1104_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(242, 1283), MAKE_VAL_COMPLEX(-59, 3), MAKE_VAL_COMPLEX(366, 778), MAKE_VAL_COMPLEX(-925, 807), MAKE_VAL_COMPLEX(-129, 250), MAKE_VAL_COMPLEX(578, 1048), MAKE_VAL_COMPLEX(109, 42), MAKE_VAL_COMPLEX(611, 486), MAKE_VAL_COMPLEX(-1060, -1446), MAKE_VAL_COMPLEX(703, 311), MAKE_VAL_COMPLEX(1630, -696), MAKE_VAL_COMPLEX(-21, -47), MAKE_VAL_COMPLEX(659, -747), MAKE_VAL_COMPLEX(643, 393), MAKE_VAL_COMPLEX(740, -399) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1151, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1151", y_true, y_val, x1104_LEN*x1104_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1151() */
/* END TEST 1151 */

#define A1152_M 5    /* # of rows */
#define A1152_N 5    /* # of cols */
#define A1152_NNZ 13  /* # of stored non-zeros */
static oski_index_t A1152_ptr[A1152_M+1] = {1, 4, 6, 7, 11, 14}; /* end row pointers for A1152 */
static oski_index_t A1152_ind[A1152_NNZ] = {2, 4, 5, 1, 5, 4, 1, 3, 4, 5, 1, 2, 4}; /* end column indices pointers for A1152 */
static oski_value_t A1152_val[A1152_NNZ] = {MAKE_VAL_COMPLEX(-7, 2), MAKE_VAL_COMPLEX(-10, 7), MAKE_VAL_COMPLEX(-4, 10), MAKE_VAL_COMPLEX(-7, -2), MAKE_VAL_COMPLEX(0, -9), MAKE_VAL_COMPLEX(-9, 8), MAKE_VAL_COMPLEX(-10, -7), MAKE_VAL_COMPLEX(-9, -8), MAKE_VAL_COMPLEX(10, 0), MAKE_VAL_COMPLEX(-5, -10), MAKE_VAL_COMPLEX(-4, -10), MAKE_VAL_COMPLEX(0, 9), MAKE_VAL_COMPLEX(-5, 10)}; /* end stored non-zero values for A1152 */

#define x1152_LEN 5   /* vector length */
#define x1152_VW  3   /* vector width */
static oski_value_t x1152_val[x1152_LEN*x1152_VW] = {MAKE_VAL_COMPLEX(4, -1), MAKE_VAL_COMPLEX(-5, 0), MAKE_VAL_COMPLEX(9, 9), MAKE_VAL_COMPLEX(-10, 1), MAKE_VAL_COMPLEX(-2, 10), MAKE_VAL_COMPLEX(10, -4), MAKE_VAL_COMPLEX(-2, -9), MAKE_VAL_COMPLEX(6, -8), MAKE_VAL_COMPLEX(9, -4), MAKE_VAL_COMPLEX(1, -1), MAKE_VAL_COMPLEX(10, 6), MAKE_VAL_COMPLEX(2, 9), MAKE_VAL_COMPLEX(5, 0), MAKE_VAL_COMPLEX(-8, 1), MAKE_VAL_COMPLEX(8, -10)}; /* data */

#define y1152_LEN 5   /* vector length */
#define y1152_VW  3   /* vector width */
static oski_value_t y1152_val[y1152_LEN*y1152_VW] = {MAKE_VAL_COMPLEX(7, 6), MAKE_VAL_COMPLEX(-1, 2), MAKE_VAL_COMPLEX(-5, -7), MAKE_VAL_COMPLEX(-5, -10), MAKE_VAL_COMPLEX(-10, 0), MAKE_VAL_COMPLEX(-4, -1), MAKE_VAL_COMPLEX(1, 3), MAKE_VAL_COMPLEX(3, 7), MAKE_VAL_COMPLEX(7, 9), MAKE_VAL_COMPLEX(-10, 6), MAKE_VAL_COMPLEX(10, 5), MAKE_VAL_COMPLEX(2, 7), MAKE_VAL_COMPLEX(-4, 3), MAKE_VAL_COMPLEX(-3, 10), MAKE_VAL_COMPLEX(-4, 9)}; /* data */

/*
 *  BEGIN TEST 1152
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1152( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	oski_value_t* y_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(41, -12), MAKE_VAL_COMPLEX(-114, -152), MAKE_VAL_COMPLEX(-85, 140), MAKE_VAL_COMPLEX(-50, 14), MAKE_VAL_COMPLEX(84, 122), MAKE_VAL_COMPLEX(-123, -133), MAKE_VAL_COMPLEX(7, 1), MAKE_VAL_COMPLEX(-122, -14), MAKE_VAL_COMPLEX(-82, -129), MAKE_VAL_COMPLEX(-52, 35), MAKE_VAL_COMPLEX(62, 134), MAKE_VAL_COMPLEX(-240, -165), MAKE_VAL_COMPLEX(-2, -107), MAKE_VAL_COMPLEX(-128, 74), MAKE_VAL_COMPLEX(42, -81) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1152, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1152", y_true, y_val, y1152_LEN*y1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1152() */
/* END TEST 1152 */

/*
 *  BEGIN TEST 1153
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1153( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	oski_value_t* y_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-67, -71), MAKE_VAL_COMPLEX(132, 141), MAKE_VAL_COMPLEX(90, -59), MAKE_VAL_COMPLEX(40, 91), MAKE_VAL_COMPLEX(-4, -72), MAKE_VAL_COMPLEX(150, 161), MAKE_VAL_COMPLEX(0, -30), MAKE_VAL_COMPLEX(133, -57), MAKE_VAL_COMPLEX(71, 22), MAKE_VAL_COMPLEX(162, -33), MAKE_VAL_COMPLEX(-117, -224), MAKE_VAL_COMPLEX(259, 99), MAKE_VAL_COMPLEX(49, 103), MAKE_VAL_COMPLEX(202, -139), MAKE_VAL_COMPLEX(35, 29) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1153, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1153", y_true, y_val, y1152_LEN*y1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1153() */
/* END TEST 1153 */

/*
 *  BEGIN TEST 1154
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1154( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	oski_value_t* y_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(85, 0), MAKE_VAL_COMPLEX(5, 20), MAKE_VAL_COMPLEX(-77, -19), MAKE_VAL_COMPLEX(-95, -40), MAKE_VAL_COMPLEX(-70, 60), MAKE_VAL_COMPLEX(-34, 17), MAKE_VAL_COMPLEX(25, 15), MAKE_VAL_COMPLEX(63, 31), MAKE_VAL_COMPLEX(103, 21), MAKE_VAL_COMPLEX(-34, 102), MAKE_VAL_COMPLEX(100, -25), MAKE_VAL_COMPLEX(56, 37), MAKE_VAL_COMPLEX(-10, 45), MAKE_VAL_COMPLEX(39, 88), MAKE_VAL_COMPLEX(26, 87) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1154, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1154", y_true, y_val, y1152_LEN*y1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1154() */
/* END TEST 1154 */

/*
 *  BEGIN TEST 1155
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1155( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	oski_value_t* y_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(354, -288), MAKE_VAL_COMPLEX(-1038, 901), MAKE_VAL_COMPLEX(583, 498), MAKE_VAL_COMPLEX(-392, 145), MAKE_VAL_COMPLEX(494, -278), MAKE_VAL_COMPLEX(-1106, 949), MAKE_VAL_COMPLEX(136, 28), MAKE_VAL_COMPLEX(227, 1013), MAKE_VAL_COMPLEX(-372, 681), MAKE_VAL_COMPLEX(71, 890), MAKE_VAL_COMPLEX(1453, -564), MAKE_VAL_COMPLEX(-866, 1870), MAKE_VAL_COMPLEX(-800, 246), MAKE_VAL_COMPLEX(720, 1349), MAKE_VAL_COMPLEX(-432, 154) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1155, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1155", y_true, y_val, y1152_LEN*y1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1155() */
/* END TEST 1155 */

/*
 *  BEGIN TEST 1156
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1156( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(216, 153), MAKE_VAL_COMPLEX(112, -155), MAKE_VAL_COMPLEX(89, 25), MAKE_VAL_COMPLEX(-107, -153), MAKE_VAL_COMPLEX(-179, -51), MAKE_VAL_COMPLEX(14, 89), MAKE_VAL_COMPLEX(217, -1), MAKE_VAL_COMPLEX(28, -79), MAKE_VAL_COMPLEX(83, -2), MAKE_VAL_COMPLEX(-245, -17), MAKE_VAL_COMPLEX(-116, 0), MAKE_VAL_COMPLEX(-165, 22), MAKE_VAL_COMPLEX(-63, 206), MAKE_VAL_COMPLEX(-33, -124), MAKE_VAL_COMPLEX(239, 21) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1156, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1156", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1156() */
/* END TEST 1156 */

/*
 *  BEGIN TEST 1157
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1157( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-222, -109), MAKE_VAL_COMPLEX(-92, 105), MAKE_VAL_COMPLEX(-215, 29), MAKE_VAL_COMPLEX(137, 49), MAKE_VAL_COMPLEX(87, -9), MAKE_VAL_COMPLEX(-14, 27), MAKE_VAL_COMPLEX(-119, 17), MAKE_VAL_COMPLEX(28, 171), MAKE_VAL_COMPLEX(-79, 108), MAKE_VAL_COMPLEX(251, 31), MAKE_VAL_COMPLEX(16, 76), MAKE_VAL_COMPLEX(67, -38), MAKE_VAL_COMPLEX(43, -156), MAKE_VAL_COMPLEX(55, 40), MAKE_VAL_COMPLEX(-171, 99) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1157, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1157", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1157() */
/* END TEST 1157 */

/*
 *  BEGIN TEST 1158
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1158( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-27, -6), MAKE_VAL_COMPLEX(30, 15), MAKE_VAL_COMPLEX(-27, -81), MAKE_VAL_COMPLEX(63, 24), MAKE_VAL_COMPLEX(42, -54), MAKE_VAL_COMPLEX(-72, -6), MAKE_VAL_COMPLEX(-15, 60), MAKE_VAL_COMPLEX(-60, 30), MAKE_VAL_COMPLEX(-66, -3), MAKE_VAL_COMPLEX(-9, 3), MAKE_VAL_COMPLEX(-42, -66), MAKE_VAL_COMPLEX(15, -60), MAKE_VAL_COMPLEX(-30, -15), MAKE_VAL_COMPLEX(51, 18), MAKE_VAL_COMPLEX(-78, 36) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1158, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1158", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1158() */
/* END TEST 1158 */

/*
 *  BEGIN TEST 1159
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1159( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(9, -518), MAKE_VAL_COMPLEX(-307, -124), MAKE_VAL_COMPLEX(-345, -311), MAKE_VAL_COMPLEX(20, 234), MAKE_VAL_COMPLEX(13, 157), MAKE_VAL_COMPLEX(-22, 85), MAKE_VAL_COMPLEX(-74, -248), MAKE_VAL_COMPLEX(-236, 273), MAKE_VAL_COMPLEX(-250, 27), MAKE_VAL_COMPLEX(199, 541), MAKE_VAL_COMPLEX(-180, 204), MAKE_VAL_COMPLEX(64, 123), MAKE_VAL_COMPLEX(360, -25), MAKE_VAL_COMPLEX(-42, 79), MAKE_VAL_COMPLEX(-271, -181) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1159, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1159", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1159() */
/* END TEST 1159 */

/*
 *  BEGIN TEST 1160
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1160( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(90, -107), MAKE_VAL_COMPLEX(-138, -85), MAKE_VAL_COMPLEX(-167, -133), MAKE_VAL_COMPLEX(17, -53), MAKE_VAL_COMPLEX(77, -21), MAKE_VAL_COMPLEX(84, 137), MAKE_VAL_COMPLEX(75, -113), MAKE_VAL_COMPLEX(-124, 77), MAKE_VAL_COMPLEX(-89, -8), MAKE_VAL_COMPLEX(-63, -53), MAKE_VAL_COMPLEX(220, -58), MAKE_VAL_COMPLEX(107, 12), MAKE_VAL_COMPLEX(97, -254), MAKE_VAL_COMPLEX(-55, -40), MAKE_VAL_COMPLEX(-115, 81) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1160, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1160", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1160() */
/* END TEST 1160 */

/*
 *  BEGIN TEST 1161
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1161( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-79, 100), MAKE_VAL_COMPLEX(123, 90), MAKE_VAL_COMPLEX(203, 151), MAKE_VAL_COMPLEX(-46, 66), MAKE_VAL_COMPLEX(-73, 53), MAKE_VAL_COMPLEX(-58, -159), MAKE_VAL_COMPLEX(-90, 88), MAKE_VAL_COMPLEX(134, -107), MAKE_VAL_COMPLEX(112, -13), MAKE_VAL_COMPLEX(65, 49), MAKE_VAL_COMPLEX(-184, 66), MAKE_VAL_COMPLEX(-92, 13), MAKE_VAL_COMPLEX(-82, 249), MAKE_VAL_COMPLEX(32, 51), MAKE_VAL_COMPLEX(129, -119) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1161, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1161", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1161() */
/* END TEST 1161 */

/*
 *  BEGIN TEST 1162
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1162( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-31, -22), MAKE_VAL_COMPLEX(30, 35), MAKE_VAL_COMPLEX(9, -117), MAKE_VAL_COMPLEX(67, 64), MAKE_VAL_COMPLEX(82, -46), MAKE_VAL_COMPLEX(-88, -46), MAKE_VAL_COMPLEX(-51, 68), MAKE_VAL_COMPLEX(-92, 6), MAKE_VAL_COMPLEX(-82, -39), MAKE_VAL_COMPLEX(-13, -1), MAKE_VAL_COMPLEX(-18, -106), MAKE_VAL_COMPLEX(51, -68), MAKE_VAL_COMPLEX(-30, -35), MAKE_VAL_COMPLEX(55, 50), MAKE_VAL_COMPLEX(-118, 4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1162, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1162", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1162() */
/* END TEST 1162 */

/*
 *  BEGIN TEST 1163
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1163( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(90, 645), MAKE_VAL_COMPLEX(639, -234), MAKE_VAL_COMPLEX(819, 87), MAKE_VAL_COMPLEX(33, 66), MAKE_VAL_COMPLEX(-330, 324), MAKE_VAL_COMPLEX(-543, -33), MAKE_VAL_COMPLEX(237, 480), MAKE_VAL_COMPLEX(297, -561), MAKE_VAL_COMPLEX(405, -132), MAKE_VAL_COMPLEX(369, -21), MAKE_VAL_COMPLEX(-516, 1020), MAKE_VAL_COMPLEX(-480, 369), MAKE_VAL_COMPLEX(501, 1128), MAKE_VAL_COMPLEX(222, -159), MAKE_VAL_COMPLEX(300, -528) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-3, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1163, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1163", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1163() */
/* END TEST 1163 */

/*
 *  BEGIN TEST 1164
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1164( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	oski_value_t* y_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(149, 32), MAKE_VAL_COMPLEX(-22, -22), MAKE_VAL_COMPLEX(-97, 1), MAKE_VAL_COMPLEX(-236, -21), MAKE_VAL_COMPLEX(-110, -31), MAKE_VAL_COMPLEX(-166, -124), MAKE_VAL_COMPLEX(76, 98), MAKE_VAL_COMPLEX(-58, 58), MAKE_VAL_COMPLEX(210, 210), MAKE_VAL_COMPLEX(-212, 230), MAKE_VAL_COMPLEX(43, 52), MAKE_VAL_COMPLEX(-63, -113), MAKE_VAL_COMPLEX(-98, -9), MAKE_VAL_COMPLEX(-204, -25), MAKE_VAL_COMPLEX(30, 43) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1164, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1164", y_true, y_val, y1152_LEN*y1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1164() */
/* END TEST 1164 */

/*
 *  BEGIN TEST 1165
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1165( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	oski_value_t* y_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-8, -69), MAKE_VAL_COMPLEX(39, 53), MAKE_VAL_COMPLEX(-39, 1), MAKE_VAL_COMPLEX(61, -4), MAKE_VAL_COMPLEX(20, 161), MAKE_VAL_COMPLEX(117, 167), MAKE_VAL_COMPLEX(-28, -84), MAKE_VAL_COMPLEX(176, -34), MAKE_VAL_COMPLEX(-30, -220), MAKE_VAL_COMPLEX(200, -46), MAKE_VAL_COMPLEX(112, -137), MAKE_VAL_COMPLEX(172, 150), MAKE_VAL_COMPLEX(101, 88), MAKE_VAL_COMPLEX(307, 154), MAKE_VAL_COMPLEX(51, 90) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1165, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1165", y_true, y_val, y1152_LEN*y1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1165() */
/* END TEST 1165 */

/*
 *  BEGIN TEST 1166
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1166( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	oski_value_t* y_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-7, -6), MAKE_VAL_COMPLEX(1, -2), MAKE_VAL_COMPLEX(5, 7), MAKE_VAL_COMPLEX(5, 10), MAKE_VAL_COMPLEX(10, 0), MAKE_VAL_COMPLEX(4, 1), MAKE_VAL_COMPLEX(-1, -3), MAKE_VAL_COMPLEX(-3, -7), MAKE_VAL_COMPLEX(-7, -9), MAKE_VAL_COMPLEX(10, -6), MAKE_VAL_COMPLEX(-10, -5), MAKE_VAL_COMPLEX(-2, -7), MAKE_VAL_COMPLEX(4, -3), MAKE_VAL_COMPLEX(3, -10), MAKE_VAL_COMPLEX(4, -9) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1166, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1166", y_true, y_val, y1152_LEN*y1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1166() */
/* END TEST 1166 */

/*
 *  BEGIN TEST 1167
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1167( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	oski_value_t* y_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(520, -200), MAKE_VAL_COMPLEX(-569, 25), MAKE_VAL_COMPLEX(208, 104), MAKE_VAL_COMPLEX(-334, 1164), MAKE_VAL_COMPLEX(-1099, -335), MAKE_VAL_COMPLEX(-1937, 379), MAKE_VAL_COMPLEX(938, 4), MAKE_VAL_COMPLEX(-506, 1352), MAKE_VAL_COMPLEX(2076, 222), MAKE_VAL_COMPLEX(38, 2050), MAKE_VAL_COMPLEX(368, 1178), MAKE_VAL_COMPLEX(-2097, 425), MAKE_VAL_COMPLEX(-1030, 450), MAKE_VAL_COMPLEX(-2496, 1630), MAKE_VAL_COMPLEX(-624, -280) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, -9);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1167, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1167", y_true, y_val, y1152_LEN*y1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1167() */
/* END TEST 1167 */

/*
 *  BEGIN TEST 1168
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1168( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(58, 229), MAKE_VAL_COMPLEX(62, -101), MAKE_VAL_COMPLEX(233, -47), MAKE_VAL_COMPLEX(-124, -142), MAKE_VAL_COMPLEX(-66, -37), MAKE_VAL_COMPLEX(-89, -45), MAKE_VAL_COMPLEX(-20, 70), MAKE_VAL_COMPLEX(69, -167), MAKE_VAL_COMPLEX(173, -63), MAKE_VAL_COMPLEX(-2, -191), MAKE_VAL_COMPLEX(66, -31), MAKE_VAL_COMPLEX(-238, 36), MAKE_VAL_COMPLEX(-115, 170), MAKE_VAL_COMPLEX(57, -178), MAKE_VAL_COMPLEX(166, 6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1168, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1168", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1168() */
/* END TEST 1168 */

/*
 *  BEGIN TEST 1169
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1169( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-37, -196), MAKE_VAL_COMPLEX(-77, 56), MAKE_VAL_COMPLEX(-287, 155), MAKE_VAL_COMPLEX(85, 55), MAKE_VAL_COMPLEX(-30, 49), MAKE_VAL_COMPLEX(155, 123), MAKE_VAL_COMPLEX(95, -115), MAKE_VAL_COMPLEX(21, 197), MAKE_VAL_COMPLEX(-110, 132), MAKE_VAL_COMPLEX(14, 197), MAKE_VAL_COMPLEX(-90, 139), MAKE_VAL_COMPLEX(163, 9), MAKE_VAL_COMPLEX(130, -125), MAKE_VAL_COMPLEX(-90, 109), MAKE_VAL_COMPLEX(-52, 36) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1169, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1169", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1169() */
/* END TEST 1169 */

/*
 *  BEGIN TEST 1170
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1170( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(12, -3), MAKE_VAL_COMPLEX(-15, 0), MAKE_VAL_COMPLEX(27, 27), MAKE_VAL_COMPLEX(-30, 3), MAKE_VAL_COMPLEX(-6, 30), MAKE_VAL_COMPLEX(30, -12), MAKE_VAL_COMPLEX(-6, -27), MAKE_VAL_COMPLEX(18, -24), MAKE_VAL_COMPLEX(27, -12), MAKE_VAL_COMPLEX(3, -3), MAKE_VAL_COMPLEX(30, 18), MAKE_VAL_COMPLEX(6, 27), MAKE_VAL_COMPLEX(15, 0), MAKE_VAL_COMPLEX(-24, 3), MAKE_VAL_COMPLEX(24, -30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1170, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1170", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1170() */
/* END TEST 1170 */

/*
 *  BEGIN TEST 1171
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1171( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1286, -1174), MAKE_VAL_COMPLEX(-137, 961), MAKE_VAL_COMPLEX(-596, 1852), MAKE_VAL_COMPLEX(923, 376), MAKE_VAL_COMPLEX(563, 450), MAKE_VAL_COMPLEX(1263, -481), MAKE_VAL_COMPLEX(-419, -988), MAKE_VAL_COMPLEX(673, 861), MAKE_VAL_COMPLEX(-273, 946), MAKE_VAL_COMPLEX(1012, 1073), MAKE_VAL_COMPLEX(353, 750), MAKE_VAL_COMPLEX(1437, -938), MAKE_VAL_COMPLEX(110, -1640), MAKE_VAL_COMPLEX(125, 1479), MAKE_VAL_COMPLEX(-684, 182) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1171, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1171", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1171() */
/* END TEST 1171 */

/*
 *  BEGIN TEST 1172
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1172( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(190, -45), MAKE_VAL_COMPLEX(-196, -199), MAKE_VAL_COMPLEX(-115, -161), MAKE_VAL_COMPLEX(-100, 10), MAKE_VAL_COMPLEX(38, 159), MAKE_VAL_COMPLEX(93, -13), MAKE_VAL_COMPLEX(140, -10), MAKE_VAL_COMPLEX(-75, -55), MAKE_VAL_COMPLEX(13, -111), MAKE_VAL_COMPLEX(-30, -9), MAKE_VAL_COMPLEX(284, 177), MAKE_VAL_COMPLEX(60, 40), MAKE_VAL_COMPLEX(169, -88), MAKE_VAL_COMPLEX(-197, 60), MAKE_VAL_COMPLEX(-60, -218) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1172, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1172", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1172() */
/* END TEST 1172 */

/*
 *  BEGIN TEST 1173
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1173( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-130, 81), MAKE_VAL_COMPLEX(136, 139), MAKE_VAL_COMPLEX(115, 377), MAKE_VAL_COMPLEX(-32, -118), MAKE_VAL_COMPLEX(-182, -63), MAKE_VAL_COMPLEX(75, 85), MAKE_VAL_COMPLEX(-56, -122), MAKE_VAL_COMPLEX(243, 31), MAKE_VAL_COMPLEX(143, 171), MAKE_VAL_COMPLEX(54, 9), MAKE_VAL_COMPLEX(-236, 15), MAKE_VAL_COMPLEX(-144, 92), MAKE_VAL_COMPLEX(-109, 148), MAKE_VAL_COMPLEX(89, -144), MAKE_VAL_COMPLEX(276, 194) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1173, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1173", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1173() */
/* END TEST 1173 */

/*
 *  BEGIN TEST 1174
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1174( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-17, 34), MAKE_VAL_COMPLEX(30, -35), MAKE_VAL_COMPLEX(-117, 9), MAKE_VAL_COMPLEX(53, -76), MAKE_VAL_COMPLEX(-58, -74), MAKE_VAL_COMPLEX(-32, 94), MAKE_VAL_COMPLEX(75, 40), MAKE_VAL_COMPLEX(20, 90), MAKE_VAL_COMPLEX(-26, 87), MAKE_VAL_COMPLEX(1, 13), MAKE_VAL_COMPLEX(-102, 34), MAKE_VAL_COMPLEX(-75, -40), MAKE_VAL_COMPLEX(-30, 35), MAKE_VAL_COMPLEX(41, -62), MAKE_VAL_COMPLEX(22, 116) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1174, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1174", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1174() */
/* END TEST 1174 */

/*
 *  BEGIN TEST 1175
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_ROWMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1175( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_ROWMAJ, x1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-388, -1146), MAKE_VAL_COMPLEX(-1253, 1177), MAKE_VAL_COMPLEX(-1730, 904), MAKE_VAL_COMPLEX(311, 282), MAKE_VAL_COMPLEX(779, -624), MAKE_VAL_COMPLEX(-215, -149), MAKE_VAL_COMPLEX(337, -806), MAKE_VAL_COMPLEX(-241, 983), MAKE_VAL_COMPLEX(-873, 372), MAKE_VAL_COMPLEX(-52, 277), MAKE_VAL_COMPLEX(725, -1766), MAKE_VAL_COMPLEX(-127, -594), MAKE_VAL_COMPLEX(-756, -988), MAKE_VAL_COMPLEX(605, 1039), MAKE_VAL_COMPLEX(-1360, 1012) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1175, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1175", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1175() */
/* END TEST 1175 */

/*
 *  BEGIN TEST 1176
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1176( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	oski_value_t* y_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(81, -160), MAKE_VAL_COMPLEX(-19, 186), MAKE_VAL_COMPLEX(66, -6), MAKE_VAL_COMPLEX(5, 7), MAKE_VAL_COMPLEX(-117, 39), MAKE_VAL_COMPLEX(-164, -121), MAKE_VAL_COMPLEX(97, -84), MAKE_VAL_COMPLEX(-12, 117), MAKE_VAL_COMPLEX(121, -74), MAKE_VAL_COMPLEX(-52, -133), MAKE_VAL_COMPLEX(-121, -76), MAKE_VAL_COMPLEX(-289, -177), MAKE_VAL_COMPLEX(14, -161), MAKE_VAL_COMPLEX(27, 50), MAKE_VAL_COMPLEX(-7, -148) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1176, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1176", y_true, y_val, y1152_LEN*y1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1176() */
/* END TEST 1176 */

/*
 *  BEGIN TEST 1177
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1177( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	oski_value_t* y_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(14, 205), MAKE_VAL_COMPLEX(14, -161), MAKE_VAL_COMPLEX(-142, -56), MAKE_VAL_COMPLEX(-90, -102), MAKE_VAL_COMPLEX(7, -9), MAKE_VAL_COMPLEX(117, 122), MAKE_VAL_COMPLEX(-77, 114), MAKE_VAL_COMPLEX(66, -49), MAKE_VAL_COMPLEX(-17, 152), MAKE_VAL_COMPLEX(-40, 229), MAKE_VAL_COMPLEX(246, 101), MAKE_VAL_COMPLEX(332, 248), MAKE_VAL_COMPLEX(-49, 206), MAKE_VAL_COMPLEX(-30, 69), MAKE_VAL_COMPLEX(-10, 259) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1177, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1177", y_true, y_val, y1152_LEN*y1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1177() */
/* END TEST 1177 */

/*
 *  BEGIN TEST 1178
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1178( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	oski_value_t* y_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(98, -1), MAKE_VAL_COMPLEX(6, 23), MAKE_VAL_COMPLEX(-89, -21), MAKE_VAL_COMPLEX(-110, -45), MAKE_VAL_COMPLEX(-80, 70), MAKE_VAL_COMPLEX(-39, 20), MAKE_VAL_COMPLEX(29, 17), MAKE_VAL_COMPLEX(73, 35), MAKE_VAL_COMPLEX(119, 23), MAKE_VAL_COMPLEX(-38, 118), MAKE_VAL_COMPLEX(115, -30), MAKE_VAL_COMPLEX(65, 42), MAKE_VAL_COMPLEX(-11, 52), MAKE_VAL_COMPLEX(46, 101), MAKE_VAL_COMPLEX(31, 100) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1178, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1178", y_true, y_val, y1152_LEN*y1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1178() */
/* END TEST 1178 */

/*
 *  BEGIN TEST 1179
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1179( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	oski_value_t* y_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-957, -642), MAKE_VAL_COMPLEX(1145, 696), MAKE_VAL_COMPLEX(319, -829), MAKE_VAL_COMPLEX(344, -429), MAKE_VAL_COMPLEX(-298, 546), MAKE_VAL_COMPLEX(-1388, 607), MAKE_VAL_COMPLEX(-392, -826), MAKE_VAL_COMPLEX(576, 706), MAKE_VAL_COMPLEX(-352, -598), MAKE_VAL_COMPLEX(-1541, -293), MAKE_VAL_COMPLEX(-870, 1119), MAKE_VAL_COMPLEX(-2335, 1724), MAKE_VAL_COMPLEX(-1290, -671), MAKE_VAL_COMPLEX(-25, 64), MAKE_VAL_COMPLEX(-1496, -353) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1179, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1179", y_true, y_val, y1152_LEN*y1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1179() */
/* END TEST 1179 */

/*
 *  BEGIN TEST 1180
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1180( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(172, 113), MAKE_VAL_COMPLEX(-53, -49), MAKE_VAL_COMPLEX(102, -64), MAKE_VAL_COMPLEX(-182, -2), MAKE_VAL_COMPLEX(-24, 97), MAKE_VAL_COMPLEX(35, -112), MAKE_VAL_COMPLEX(-100, 26), MAKE_VAL_COMPLEX(-116, -87), MAKE_VAL_COMPLEX(-147, -95), MAKE_VAL_COMPLEX(-26, -49), MAKE_VAL_COMPLEX(109, -137), MAKE_VAL_COMPLEX(-19, -66), MAKE_VAL_COMPLEX(3, -94), MAKE_VAL_COMPLEX(-27, 12), MAKE_VAL_COMPLEX(55, 5) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1180, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1180", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1180() */
/* END TEST 1180 */

/*
 *  BEGIN TEST 1181
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1181( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-193, -146), MAKE_VAL_COMPLEX(68, 94), MAKE_VAL_COMPLEX(-48, -44), MAKE_VAL_COMPLEX(221, 89), MAKE_VAL_COMPLEX(120, -109), MAKE_VAL_COMPLEX(-101, 34), MAKE_VAL_COMPLEX(25, 19), MAKE_VAL_COMPLEX(26, 57), MAKE_VAL_COMPLEX(84, 26), MAKE_VAL_COMPLEX(14, 43), MAKE_VAL_COMPLEX(-85, 29), MAKE_VAL_COMPLEX(94, 21), MAKE_VAL_COMPLEX(-18, 49), MAKE_VAL_COMPLEX(60, 57), MAKE_VAL_COMPLEX(-169, -47) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1181, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1181", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1181() */
/* END TEST 1181 */

/*
 *  BEGIN TEST 1182
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1182( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-21, 18), MAKE_VAL_COMPLEX(30, -15), MAKE_VAL_COMPLEX(-81, -27), MAKE_VAL_COMPLEX(57, -36), MAKE_VAL_COMPLEX(-18, -66), MAKE_VAL_COMPLEX(-48, 54), MAKE_VAL_COMPLEX(39, 48), MAKE_VAL_COMPLEX(-12, 66), MAKE_VAL_COMPLEX(-42, 51), MAKE_VAL_COMPLEX(-3, 9), MAKE_VAL_COMPLEX(-78, -6), MAKE_VAL_COMPLEX(-39, -48), MAKE_VAL_COMPLEX(-30, 15), MAKE_VAL_COMPLEX(45, -30), MAKE_VAL_COMPLEX(-18, 84) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1182, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1182", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1182() */
/* END TEST 1182 */

/*
 *  BEGIN TEST 1183
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1183( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1072, -1421), MAKE_VAL_COMPLEX(434, 697), MAKE_VAL_COMPLEX(-933, -587), MAKE_VAL_COMPLEX(1727, 908), MAKE_VAL_COMPLEX(933, -949), MAKE_VAL_COMPLEX(-1157, 457), MAKE_VAL_COMPLEX(511, 574), MAKE_VAL_COMPLEX(-13, 1059), MAKE_VAL_COMPLEX(210, 923), MAKE_VAL_COMPLEX(-46, 424), MAKE_VAL_COMPLEX(-1360, -10), MAKE_VAL_COMPLEX(202, 63), MAKE_VAL_COMPLEX(-513, 454), MAKE_VAL_COMPLEX(600, 267), MAKE_VAL_COMPLEX(-1096, -92) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, -3);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1183, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1183", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1183() */
/* END TEST 1183 */

/*
 *  BEGIN TEST 1184
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1184( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(78, -138), MAKE_VAL_COMPLEX(10, 5), MAKE_VAL_COMPLEX(51, -215), MAKE_VAL_COMPLEX(-18, -13), MAKE_VAL_COMPLEX(170, -299), MAKE_VAL_COMPLEX(-213, -104), MAKE_VAL_COMPLEX(64, 61), MAKE_VAL_COMPLEX(-194, 37), MAKE_VAL_COMPLEX(212, -99), MAKE_VAL_COMPLEX(-85, -14), MAKE_VAL_COMPLEX(-123, -207), MAKE_VAL_COMPLEX(139, 49), MAKE_VAL_COMPLEX(-94, -72), MAKE_VAL_COMPLEX(177, 69), MAKE_VAL_COMPLEX(-203, 27) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1184, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1184", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1184() */
/* END TEST 1184 */

/*
 *  BEGIN TEST 1185
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1185( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-62, 151), MAKE_VAL_COMPLEX(-25, -25), MAKE_VAL_COMPLEX(-60, 278), MAKE_VAL_COMPLEX(-16, -24), MAKE_VAL_COMPLEX(-216, 321), MAKE_VAL_COMPLEX(259, 132), MAKE_VAL_COMPLEX(-34, -96), MAKE_VAL_COMPLEX(244, -37), MAKE_VAL_COMPLEX(-169, 123), MAKE_VAL_COMPLEX(92, 15), MAKE_VAL_COMPLEX(129, 265), MAKE_VAL_COMPLEX(-169, -14), MAKE_VAL_COMPLEX(109, 92), MAKE_VAL_COMPLEX(-205, -98), MAKE_VAL_COMPLEX(267, -25) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1185, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1185", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1185() */
/* END TEST 1185 */

/*
 *  BEGIN TEST 1186
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1186( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-7, 6), MAKE_VAL_COMPLEX(10, -5), MAKE_VAL_COMPLEX(-27, -9), MAKE_VAL_COMPLEX(19, -12), MAKE_VAL_COMPLEX(-6, -22), MAKE_VAL_COMPLEX(-16, 18), MAKE_VAL_COMPLEX(13, 16), MAKE_VAL_COMPLEX(-4, 22), MAKE_VAL_COMPLEX(-14, 17), MAKE_VAL_COMPLEX(-1, 3), MAKE_VAL_COMPLEX(-26, -2), MAKE_VAL_COMPLEX(-13, -16), MAKE_VAL_COMPLEX(-10, 5), MAKE_VAL_COMPLEX(15, -10), MAKE_VAL_COMPLEX(-6, 28) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1186, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1186", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1186() */
/* END TEST 1186 */

/*
 *  BEGIN TEST 1187
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1187( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_ROWMAJ, y1152_VW );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1497, 432), MAKE_VAL_COMPLEX(145, -205), MAKE_VAL_COMPLEX(1302, -104), MAKE_VAL_COMPLEX(282, -874), MAKE_VAL_COMPLEX(2801, -92), MAKE_VAL_COMPLEX(36, -1583), MAKE_VAL_COMPLEX(334, 964), MAKE_VAL_COMPLEX(-785, -980), MAKE_VAL_COMPLEX(1721, 2199), MAKE_VAL_COMPLEX(-181, -728), MAKE_VAL_COMPLEX(693, -1485), MAKE_VAL_COMPLEX(-512, 1303), MAKE_VAL_COMPLEX(122, -819), MAKE_VAL_COMPLEX(264, 1335), MAKE_VAL_COMPLEX(-665, -915) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, 9);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1187, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1187", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1187() */
/* END TEST 1187 */

/*
 *  BEGIN TEST 1188
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1188( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	oski_value_t* y_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(42, -242), MAKE_VAL_COMPLEX(82, 13), MAKE_VAL_COMPLEX(56, -7), MAKE_VAL_COMPLEX(-96, -91), MAKE_VAL_COMPLEX(74, -106), MAKE_VAL_COMPLEX(-8, 214), MAKE_VAL_COMPLEX(-69, -27), MAKE_VAL_COMPLEX(-11, 42), MAKE_VAL_COMPLEX(-141, -161), MAKE_VAL_COMPLEX(104, 52), MAKE_VAL_COMPLEX(89, -115), MAKE_VAL_COMPLEX(-104, -192), MAKE_VAL_COMPLEX(112, -59), MAKE_VAL_COMPLEX(-225, -226), MAKE_VAL_COMPLEX(65, -213) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1188, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1188", y_true, y_val, y1152_LEN*y1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1188() */
/* END TEST 1188 */

/*
 *  BEGIN TEST 1189
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1189( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	oski_value_t* y_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(30, 158), MAKE_VAL_COMPLEX(-58, -1), MAKE_VAL_COMPLEX(-140, 67), MAKE_VAL_COMPLEX(-24, 151), MAKE_VAL_COMPLEX(-74, 226), MAKE_VAL_COMPLEX(-4, -166), MAKE_VAL_COMPLEX(105, 15), MAKE_VAL_COMPLEX(95, -78), MAKE_VAL_COMPLEX(249, 77), MAKE_VAL_COMPLEX(-32, 68), MAKE_VAL_COMPLEX(-29, -5), MAKE_VAL_COMPLEX(188, 168), MAKE_VAL_COMPLEX(-76, 107), MAKE_VAL_COMPLEX(345, 262), MAKE_VAL_COMPLEX(43, 261) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1189, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1189", y_true, y_val, y1152_LEN*y1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1189() */
/* END TEST 1189 */

/*
 *  BEGIN TEST 1190
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1190( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	oski_value_t* y_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-71, 12), MAKE_VAL_COMPLEX(-7, -16), MAKE_VAL_COMPLEX(67, 5), MAKE_VAL_COMPLEX(85, 20), MAKE_VAL_COMPLEX(50, -60), MAKE_VAL_COMPLEX(26, -19), MAKE_VAL_COMPLEX(-23, -9), MAKE_VAL_COMPLEX(-57, -17), MAKE_VAL_COMPLEX(-89, -3), MAKE_VAL_COMPLEX(14, -90), MAKE_VAL_COMPLEX(-80, 35), MAKE_VAL_COMPLEX(-52, -23), MAKE_VAL_COMPLEX(2, -39), MAKE_VAL_COMPLEX(-45, -68), MAKE_VAL_COMPLEX(-34, -69) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1190, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1190", y_true, y_val, y1152_LEN*y1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1190() */
/* END TEST 1190 */

/*
 *  BEGIN TEST 1191
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1191( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	oski_value_t* y_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-199, 457), MAKE_VAL_COMPLEX(-187, -72), MAKE_VAL_COMPLEX(-162, 281), MAKE_VAL_COMPLEX(243, 608), MAKE_VAL_COMPLEX(28, 488), MAKE_VAL_COMPLEX(107, -549), MAKE_VAL_COMPLEX(233, -11), MAKE_VAL_COMPLEX(77, -352), MAKE_VAL_COMPLEX(401, 139), MAKE_VAL_COMPLEX(40, -136), MAKE_VAL_COMPLEX(-432, 50), MAKE_VAL_COMPLEX(381, 367), MAKE_VAL_COMPLEX(-185, 170), MAKE_VAL_COMPLEX(920, 479), MAKE_VAL_COMPLEX(58, 482) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-3, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1191, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1191", y_true, y_val, y1152_LEN*y1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1191() */
/* END TEST 1191 */

/*
 *  BEGIN TEST 1192
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1192( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(31, 206), MAKE_VAL_COMPLEX(-66, -98), MAKE_VAL_COMPLEX(10, 103), MAKE_VAL_COMPLEX(-17, -147), MAKE_VAL_COMPLEX(-107, 173), MAKE_VAL_COMPLEX(86, -130), MAKE_VAL_COMPLEX(-62, -92), MAKE_VAL_COMPLEX(-17, -169), MAKE_VAL_COMPLEX(17, -137), MAKE_VAL_COMPLEX(105, -165), MAKE_VAL_COMPLEX(240, -162), MAKE_VAL_COMPLEX(-123, -50), MAKE_VAL_COMPLEX(112, -86), MAKE_VAL_COMPLEX(-227, 9), MAKE_VAL_COMPLEX(56, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1192, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1192", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1192() */
/* END TEST 1192 */

/*
 *  BEGIN TEST 1193
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1193( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-41, -229), MAKE_VAL_COMPLEX(71, 128), MAKE_VAL_COMPLEX(35, -166), MAKE_VAL_COMPLEX(33, 206), MAKE_VAL_COMPLEX(169, -171), MAKE_VAL_COMPLEX(-120, 74), MAKE_VAL_COMPLEX(10, 113), MAKE_VAL_COMPLEX(-37, 141), MAKE_VAL_COMPLEX(-50, 87), MAKE_VAL_COMPLEX(-112, 160), MAKE_VAL_COMPLEX(-214, 96), MAKE_VAL_COMPLEX(175, 29), MAKE_VAL_COMPLEX(-117, 56), MAKE_VAL_COMPLEX(241, 38), MAKE_VAL_COMPLEX(-124, -38) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1193, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1193", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1193() */
/* END TEST 1193 */

/*
 *  BEGIN TEST 1194
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1194( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-16, -13), MAKE_VAL_COMPLEX(15, 20), MAKE_VAL_COMPLEX(9, -63), MAKE_VAL_COMPLEX(34, 37), MAKE_VAL_COMPLEX(46, -22), MAKE_VAL_COMPLEX(-46, -28), MAKE_VAL_COMPLEX(-30, 35), MAKE_VAL_COMPLEX(-50, 0), MAKE_VAL_COMPLEX(-43, -24), MAKE_VAL_COMPLEX(-7, -1), MAKE_VAL_COMPLEX(-6, -58), MAKE_VAL_COMPLEX(30, -35), MAKE_VAL_COMPLEX(-15, -20), MAKE_VAL_COMPLEX(28, 29), MAKE_VAL_COMPLEX(-64, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1194, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1194", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1194() */
/* END TEST 1194 */

/*
 *  BEGIN TEST 1195
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1195( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1095, 368), MAKE_VAL_COMPLEX(539, -408), MAKE_VAL_COMPLEX(-676, -90), MAKE_VAL_COMPLEX(952, -215), MAKE_VAL_COMPLEX(-886, -584), MAKE_VAL_COMPLEX(490, 352), MAKE_VAL_COMPLEX(408, -187), MAKE_VAL_COMPLEX(658, -94), MAKE_VAL_COMPLEX(479, 4), MAKE_VAL_COMPLEX(903, 379), MAKE_VAL_COMPLEX(844, 860), MAKE_VAL_COMPLEX(117, -880), MAKE_VAL_COMPLEX(427, 454), MAKE_VAL_COMPLEX(-84, -1117), MAKE_VAL_COMPLEX(-168, 478) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1195, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1195", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1195() */
/* END TEST 1195 */

/*
 *  BEGIN TEST 1196
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1196( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(203, -27), MAKE_VAL_COMPLEX(-77, -6), MAKE_VAL_COMPLEX(125, 194), MAKE_VAL_COMPLEX(-127, -78), MAKE_VAL_COMPLEX(43, -39), MAKE_VAL_COMPLEX(-54, -136), MAKE_VAL_COMPLEX(136, 17), MAKE_VAL_COMPLEX(-23, -41), MAKE_VAL_COMPLEX(346, 151), MAKE_VAL_COMPLEX(-130, 78), MAKE_VAL_COMPLEX(-110, -114), MAKE_VAL_COMPLEX(-35, 69), MAKE_VAL_COMPLEX(-13, -74), MAKE_VAL_COMPLEX(3, -76), MAKE_VAL_COMPLEX(6, -198) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1196, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1196", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1196() */
/* END TEST 1196 */

/*
 *  BEGIN TEST 1197
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1197( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-119, 74), MAKE_VAL_COMPLEX(-8, -74), MAKE_VAL_COMPLEX(-116, 103), MAKE_VAL_COMPLEX(-59, -65), MAKE_VAL_COMPLEX(-237, 177), MAKE_VAL_COMPLEX(288, 228), MAKE_VAL_COMPLEX(-26, -202), MAKE_VAL_COMPLEX(253, 1), MAKE_VAL_COMPLEX(-129, -75), MAKE_VAL_COMPLEX(163, -79), MAKE_VAL_COMPLEX(184, 376), MAKE_VAL_COMPLEX(-75, 116), MAKE_VAL_COMPLEX(98, 154), MAKE_VAL_COMPLEX(-155, -35), MAKE_VAL_COMPLEX(290, 156) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1197, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1197", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1197() */
/* END TEST 1197 */

/*
 *  BEGIN TEST 1198
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1198( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-28, -10), MAKE_VAL_COMPLEX(30, 20), MAKE_VAL_COMPLEX(-18, -90), MAKE_VAL_COMPLEX(64, 34), MAKE_VAL_COMPLEX(52, -52), MAKE_VAL_COMPLEX(-76, -16), MAKE_VAL_COMPLEX(-24, 62), MAKE_VAL_COMPLEX(-68, 24), MAKE_VAL_COMPLEX(-70, -12), MAKE_VAL_COMPLEX(-10, 2), MAKE_VAL_COMPLEX(-36, -76), MAKE_VAL_COMPLEX(24, -62), MAKE_VAL_COMPLEX(-30, -20), MAKE_VAL_COMPLEX(52, 26), MAKE_VAL_COMPLEX(-88, 28) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1198, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1198", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1198() */
/* END TEST 1198 */

/*
 *  BEGIN TEST 1199
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1152', LAYOUT_COLMAJ>
 *  y : <'y1152', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1199( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1152_ptr, A1152_ind, A1152_val,
		A1152_M, A1152_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1152_val,
		y1152_LEN * y1152_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1152_LEN, y1152_VW, LAYOUT_COLMAJ, y1152_LEN );
	oski_value_t* y_val = array_Duplicate( x1152_val,
		x1152_LEN * x1152_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1152_LEN, x1152_VW, LAYOUT_COLMAJ, x1152_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-812, 1854), MAKE_VAL_COMPLEX(-55, -590), MAKE_VAL_COMPLEX(-1342, 779), MAKE_VAL_COMPLEX(279, -264), MAKE_VAL_COMPLEX(-173, 2171), MAKE_VAL_COMPLEX(3062, -88), MAKE_VAL_COMPLEX(-1630, -195), MAKE_VAL_COMPLEX(1335, -1105), MAKE_VAL_COMPLEX(-2866, 1227), MAKE_VAL_COMPLEX(472, -1950), MAKE_VAL_COMPLEX(3422, 682), MAKE_VAL_COMPLEX(48, 416), MAKE_VAL_COMPLEX(1495, 420), MAKE_VAL_COMPLEX(-489, 868), MAKE_VAL_COMPLEX(2784, 100) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, 9);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1199, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1199", y_true, y_val, x1152_LEN*x1152_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1199() */
/* END TEST 1199 */

#define A1200_M 5    /* # of rows */
#define A1200_N 5    /* # of cols */
#define A1200_NNZ 13  /* # of stored non-zeros */
static oski_index_t A1200_ptr[A1200_M+1] = {0, 2, 5, 9, 11, 13}; /* end row pointers for A1200 */
static oski_index_t A1200_ind[A1200_NNZ] = {1, 2, 0, 2, 4, 0, 1, 2, 3, 2, 3, 1, 4}; /* end column indices pointers for A1200 */
static oski_value_t A1200_val[A1200_NNZ] = {MAKE_VAL_COMPLEX(-4, 8), MAKE_VAL_COMPLEX(-2, -7), MAKE_VAL_COMPLEX(-4, -8), MAKE_VAL_COMPLEX(-10, -9), MAKE_VAL_COMPLEX(-4, 4), MAKE_VAL_COMPLEX(-2, 7), MAKE_VAL_COMPLEX(-10, 9), MAKE_VAL_COMPLEX(2, 0), MAKE_VAL_COMPLEX(-2, -10), MAKE_VAL_COMPLEX(-2, 10), MAKE_VAL_COMPLEX(10, 0), MAKE_VAL_COMPLEX(-4, -4), MAKE_VAL_COMPLEX(-1, 0)}; /* end stored non-zero values for A1200 */

#define x1200_LEN 5   /* vector length */
#define x1200_VW  3   /* vector width */
static oski_value_t x1200_val[x1200_LEN*x1200_VW] = {MAKE_VAL_COMPLEX(7, 0), MAKE_VAL_COMPLEX(-2, -1), MAKE_VAL_COMPLEX(-1, -8), MAKE_VAL_COMPLEX(-10, 4), MAKE_VAL_COMPLEX(-6, 8), MAKE_VAL_COMPLEX(8, -5), MAKE_VAL_COMPLEX(4, 7), MAKE_VAL_COMPLEX(10, 9), MAKE_VAL_COMPLEX(-7, -10), MAKE_VAL_COMPLEX(-5, -4), MAKE_VAL_COMPLEX(-3, 10), MAKE_VAL_COMPLEX(4, -5), MAKE_VAL_COMPLEX(0, 1), MAKE_VAL_COMPLEX(1, -4), MAKE_VAL_COMPLEX(5, -1)}; /* data */

#define y1200_LEN 5   /* vector length */
#define y1200_VW  3   /* vector width */
static oski_value_t y1200_val[y1200_LEN*y1200_VW] = {MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(6, -9), MAKE_VAL_COMPLEX(8, -6), MAKE_VAL_COMPLEX(-8, 9), MAKE_VAL_COMPLEX(-8, -2), MAKE_VAL_COMPLEX(7, 2), MAKE_VAL_COMPLEX(-2, 3), MAKE_VAL_COMPLEX(6, -6), MAKE_VAL_COMPLEX(6, 10), MAKE_VAL_COMPLEX(2, -8), MAKE_VAL_COMPLEX(-8, 3), MAKE_VAL_COMPLEX(9, 7), MAKE_VAL_COMPLEX(7, 5), MAKE_VAL_COMPLEX(10, -3), MAKE_VAL_COMPLEX(8, -10)}; /* data */

/*
 *  BEGIN TEST 1200
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1200( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	oski_value_t* y_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-61, -128), MAKE_VAL_COMPLEX(-21, -249), MAKE_VAL_COMPLEX(-44, 75), MAKE_VAL_COMPLEX(5, -73), MAKE_VAL_COMPLEX(-59, -102), MAKE_VAL_COMPLEX(-49, 195), MAKE_VAL_COMPLEX(36, 18), MAKE_VAL_COMPLEX(119, -184), MAKE_VAL_COMPLEX(41, 95), MAKE_VAL_COMPLEX(-166, -66), MAKE_VAL_COMPLEX(-162, 245), MAKE_VAL_COMPLEX(241, -119), MAKE_VAL_COMPLEX(121, 6), MAKE_VAL_COMPLEX(87, -79), MAKE_VAL_COMPLEX(-77, -109) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1200, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1200", y_true, y_val, y1200_LEN*y1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1200() */
/* END TEST 1200 */

/*
 *  BEGIN TEST 1201
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1201( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	oski_value_t* y_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-29, 38), MAKE_VAL_COMPLEX(75, 168), MAKE_VAL_COMPLEX(116, -129), MAKE_VAL_COMPLEX(-77, 154), MAKE_VAL_COMPLEX(-13, 84), MAKE_VAL_COMPLEX(112, -177), MAKE_VAL_COMPLEX(-54, 9), MAKE_VAL_COMPLEX(-65, 130), MAKE_VAL_COMPLEX(13, -5), MAKE_VAL_COMPLEX(184, -6), MAKE_VAL_COMPLEX(90, -218), MAKE_VAL_COMPLEX(-160, 182), MAKE_VAL_COMPLEX(-58, 39), MAKE_VAL_COMPLEX(3, 52), MAKE_VAL_COMPLEX(149, 19) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1201, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1201", y_true, y_val, y1200_LEN*y1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1201() */
/* END TEST 1201 */

/*
 *  BEGIN TEST 1202
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1202( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	oski_value_t* y_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(20, -80), MAKE_VAL_COMPLEX(63, 3), MAKE_VAL_COMPLEX(54, 22), MAKE_VAL_COMPLEX(-69, -13), MAKE_VAL_COMPLEX(-14, -46), MAKE_VAL_COMPLEX(11, 41), MAKE_VAL_COMPLEX(-21, -1), MAKE_VAL_COMPLEX(48, 12), MAKE_VAL_COMPLEX(-32, 60), MAKE_VAL_COMPLEX(46, -14), MAKE_VAL_COMPLEX(-39, -31), MAKE_VAL_COMPLEX(-8, 66), MAKE_VAL_COMPLEX(-4, 50), MAKE_VAL_COMPLEX(45, 41), MAKE_VAL_COMPLEX(74, 10) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1202, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1202", y_true, y_val, y1200_LEN*y1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1202() */
/* END TEST 1202 */

/*
 *  BEGIN TEST 1203
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1203( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	oski_value_t* y_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(148, -286), MAKE_VAL_COMPLEX(21, -300), MAKE_VAL_COMPLEX(-94, 342), MAKE_VAL_COMPLEX(-29, -374), MAKE_VAL_COMPLEX(8, -300), MAKE_VAL_COMPLEX(-212, 471), MAKE_VAL_COMPLEX(51, -30), MAKE_VAL_COMPLEX(256, -206), MAKE_VAL_COMPLEX(-140, 160), MAKE_VAL_COMPLEX(-236, -6), MAKE_VAL_COMPLEX(-273, 334), MAKE_VAL_COMPLEX(269, -187), MAKE_VAL_COMPLEX(83, 57), MAKE_VAL_COMPLEX(99, 28), MAKE_VAL_COMPLEX(-100, 22) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1203, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1203", y_true, y_val, y1200_LEN*y1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1203() */
/* END TEST 1203 */

/*
 *  BEGIN TEST 1204
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1204( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(94, 50), MAKE_VAL_COMPLEX(50, 113), MAKE_VAL_COMPLEX(-47, -56), MAKE_VAL_COMPLEX(71, -192), MAKE_VAL_COMPLEX(-64, 142), MAKE_VAL_COMPLEX(-168, 93), MAKE_VAL_COMPLEX(145, 145), MAKE_VAL_COMPLEX(-59, 39), MAKE_VAL_COMPLEX(-133, -83), MAKE_VAL_COMPLEX(73, -100), MAKE_VAL_COMPLEX(-215, -26), MAKE_VAL_COMPLEX(212, 9), MAKE_VAL_COMPLEX(-17, -72), MAKE_VAL_COMPLEX(55, -19), MAKE_VAL_COMPLEX(-33, 59) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1204, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1204", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1204() */
/* END TEST 1204 */

/*
 *  BEGIN TEST 1205
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1205( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-66, 27), MAKE_VAL_COMPLEX(-47, -139), MAKE_VAL_COMPLEX(131, 13), MAKE_VAL_COMPLEX(-155, 98), MAKE_VAL_COMPLEX(-48, -176), MAKE_VAL_COMPLEX(255, -25), MAKE_VAL_COMPLEX(-206, -73), MAKE_VAL_COMPLEX(0, 107), MAKE_VAL_COMPLEX(215, -34), MAKE_VAL_COMPLEX(-49, 29), MAKE_VAL_COMPLEX(93, 33), MAKE_VAL_COMPLEX(-141, 15), MAKE_VAL_COMPLEX(6, 76), MAKE_VAL_COMPLEX(-7, 14), MAKE_VAL_COMPLEX(64, -8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1205, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1205", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1205() */
/* END TEST 1205 */

/*
 *  BEGIN TEST 1206
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1206( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-21, 14), MAKE_VAL_COMPLEX(8, -1), MAKE_VAL_COMPLEX(19, 22), MAKE_VAL_COMPLEX(22, -32), MAKE_VAL_COMPLEX(2, -36), MAKE_VAL_COMPLEX(-14, 31), MAKE_VAL_COMPLEX(-26, -13), MAKE_VAL_COMPLEX(-48, -7), MAKE_VAL_COMPLEX(41, 16), MAKE_VAL_COMPLEX(23, 2), MAKE_VAL_COMPLEX(-11, -36), MAKE_VAL_COMPLEX(-2, 23), MAKE_VAL_COMPLEX(-2, -3), MAKE_VAL_COMPLEX(5, 14), MAKE_VAL_COMPLEX(-13, 13) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1206, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1206", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1206() */
/* END TEST 1206 */

/*
 *  BEGIN TEST 1207
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1207( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-292, -60), MAKE_VAL_COMPLEX(-204, -504), MAKE_VAL_COMPLEX(336, 108), MAKE_VAL_COMPLEX(-484, 616), MAKE_VAL_COMPLEX(24, -592), MAKE_VAL_COMPLEX(868, -272), MAKE_VAL_COMPLEX(-672, -416), MAKE_VAL_COMPLEX(176, 152), MAKE_VAL_COMPLEX(648, 72), MAKE_VAL_COMPLEX(-272, 252), MAKE_VAL_COMPLEX(624, 164), MAKE_VAL_COMPLEX(-700, -16), MAKE_VAL_COMPLEX(48, 300), MAKE_VAL_COMPLEX(-128, 48), MAKE_VAL_COMPLEX(212, -148) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1207, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1207", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1207() */
/* END TEST 1207 */

/*
 *  BEGIN TEST 1208
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1208( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-78, -43), MAKE_VAL_COMPLEX(19, -91), MAKE_VAL_COMPLEX(79, 51), MAKE_VAL_COMPLEX(21, 10), MAKE_VAL_COMPLEX(-240, -68), MAKE_VAL_COMPLEX(-79, -21), MAKE_VAL_COMPLEX(-84, -245), MAKE_VAL_COMPLEX(54, 59), MAKE_VAL_COMPLEX(135, 68), MAKE_VAL_COMPLEX(67, -105), MAKE_VAL_COMPLEX(-75, -9), MAKE_VAL_COMPLEX(-23, 183), MAKE_VAL_COMPLEX(54, -18), MAKE_VAL_COMPLEX(33, 86), MAKE_VAL_COMPLEX(-66, 18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1208, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1208", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1208() */
/* END TEST 1208 */

/*
 *  BEGIN TEST 1209
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1209( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(57, 64), MAKE_VAL_COMPLEX(-10, 88), MAKE_VAL_COMPLEX(-52, -30), MAKE_VAL_COMPLEX(-3, -52), MAKE_VAL_COMPLEX(234, 26), MAKE_VAL_COMPLEX(70, 60), MAKE_VAL_COMPLEX(51, 236), MAKE_VAL_COMPLEX(-111, -56), MAKE_VAL_COMPLEX(-84, -59), MAKE_VAL_COMPLEX(-40, 102), MAKE_VAL_COMPLEX(54, -30), MAKE_VAL_COMPLEX(26, -156), MAKE_VAL_COMPLEX(-57, 15), MAKE_VAL_COMPLEX(-24, -71), MAKE_VAL_COMPLEX(54, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1209, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1209", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1209() */
/* END TEST 1209 */

/*
 *  BEGIN TEST 1210
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1210( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(49, 35), MAKE_VAL_COMPLEX(-9, -17), MAKE_VAL_COMPLEX(33, -61), MAKE_VAL_COMPLEX(-90, -22), MAKE_VAL_COMPLEX(-82, 26), MAKE_VAL_COMPLEX(81, 5), MAKE_VAL_COMPLEX(-7, 69), MAKE_VAL_COMPLEX(25, 113), MAKE_VAL_COMPLEX(1, -105), MAKE_VAL_COMPLEX(-15, -53), MAKE_VAL_COMPLEX(-71, 55), MAKE_VAL_COMPLEX(53, -15), MAKE_VAL_COMPLEX(-5, 7), MAKE_VAL_COMPLEX(27, -23), MAKE_VAL_COMPLEX(40, 18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1210, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1210", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1210() */
/* END TEST 1210 */

/*
 *  BEGIN TEST 1211
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1211( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(450, -437), MAKE_VAL_COMPLEX(507, -380), MAKE_VAL_COMPLEX(170, 73), MAKE_VAL_COMPLEX(-812, 164), MAKE_VAL_COMPLEX(-1216, -1310), MAKE_VAL_COMPLEX(531, -686), MAKE_VAL_COMPLEX(1219, -864), MAKE_VAL_COMPLEX(321, 1498), MAKE_VAL_COMPLEX(-82, 159), MAKE_VAL_COMPLEX(654, -451), MAKE_VAL_COMPLEX(-752, 147), MAKE_VAL_COMPLEX(-757, 350), MAKE_VAL_COMPLEX(295, 324), MAKE_VAL_COMPLEX(-196, 283), MAKE_VAL_COMPLEX(17, -251) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1211, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1211", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1211() */
/* END TEST 1211 */

/*
 *  BEGIN TEST 1212
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1212( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	oski_value_t* y_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(9, -58), MAKE_VAL_COMPLEX(-75, -184), MAKE_VAL_COMPLEX(-24, -45), MAKE_VAL_COMPLEX(-58, 16), MAKE_VAL_COMPLEX(60, 75), MAKE_VAL_COMPLEX(1, -214), MAKE_VAL_COMPLEX(15, -134), MAKE_VAL_COMPLEX(77, -142), MAKE_VAL_COMPLEX(-92, 126), MAKE_VAL_COMPLEX(3, 0), MAKE_VAL_COMPLEX(-14, 195), MAKE_VAL_COMPLEX(-72, 159), MAKE_VAL_COMPLEX(-33, 29), MAKE_VAL_COMPLEX(116, -154), MAKE_VAL_COMPLEX(-133, -39) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1212, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1212", y_true, y_val, y1200_LEN*y1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1212() */
/* END TEST 1212 */

/*
 *  BEGIN TEST 1213
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1213( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	oski_value_t* y_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-139, 228), MAKE_VAL_COMPLEX(-72, 112), MAKE_VAL_COMPLEX(-82, -63), MAKE_VAL_COMPLEX(209, 86), MAKE_VAL_COMPLEX(-74, 49), MAKE_VAL_COMPLEX(15, 105), MAKE_VAL_COMPLEX(34, 158), MAKE_VAL_COMPLEX(-179, 64), MAKE_VAL_COMPLEX(230, -236), MAKE_VAL_COMPLEX(-127, -14), MAKE_VAL_COMPLEX(75, -81), MAKE_VAL_COMPLEX(159, -308), MAKE_VAL_COMPLEX(94, -144), MAKE_VAL_COMPLEX(-181, 10), MAKE_VAL_COMPLEX(-33, -61) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1213, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1213", y_true, y_val, y1200_LEN*y1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1213() */
/* END TEST 1213 */

/*
 *  BEGIN TEST 1214
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1214( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	oski_value_t* y_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-80, 60), MAKE_VAL_COMPLEX(-57, -51), MAKE_VAL_COMPLEX(-34, -62), MAKE_VAL_COMPLEX(55, 65), MAKE_VAL_COMPLEX(-22, 54), MAKE_VAL_COMPLEX(21, -47), MAKE_VAL_COMPLEX(19, 17), MAKE_VAL_COMPLEX(-36, -48), MAKE_VAL_COMPLEX(76, -32), MAKE_VAL_COMPLEX(-54, -22), MAKE_VAL_COMPLEX(13, 59), MAKE_VAL_COMPLEX(58, -56), MAKE_VAL_COMPLEX(42, -44), MAKE_VAL_COMPLEX(-11, -73), MAKE_VAL_COMPLEX(-62, -66) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1214, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1214", y_true, y_val, y1200_LEN*y1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1214() */
/* END TEST 1214 */

/*
 *  BEGIN TEST 1215
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1215( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	oski_value_t* y_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1418, 9), MAKE_VAL_COMPLEX(1365, -742), MAKE_VAL_COMPLEX(111, 190), MAKE_VAL_COMPLEX(-283, -1197), MAKE_VAL_COMPLEX(43, 662), MAKE_VAL_COMPLEX(1503, -704), MAKE_VAL_COMPLEX(1263, -624), MAKE_VAL_COMPLEX(1490, 641), MAKE_VAL_COMPLEX(-2158, -624), MAKE_VAL_COMPLEX(190, 505), MAKE_VAL_COMPLEX(-1508, 221), MAKE_VAL_COMPLEX(-2423, -54), MAKE_VAL_COMPLEX(-925, -191), MAKE_VAL_COMPLEX(1445, 931), MAKE_VAL_COMPLEX(-239, -563) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, 9);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1215, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1215", y_true, y_val, y1200_LEN*y1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1215() */
/* END TEST 1215 */

/*
 *  BEGIN TEST 1216
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1216( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-70, 91), MAKE_VAL_COMPLEX(67, 48), MAKE_VAL_COMPLEX(11, -66), MAKE_VAL_COMPLEX(98, 82), MAKE_VAL_COMPLEX(-130, 156), MAKE_VAL_COMPLEX(-154, -15), MAKE_VAL_COMPLEX(-284, 36), MAKE_VAL_COMPLEX(-98, 13), MAKE_VAL_COMPLEX(84, -20), MAKE_VAL_COMPLEX(-136, -3), MAKE_VAL_COMPLEX(-62, 37), MAKE_VAL_COMPLEX(161, -90), MAKE_VAL_COMPLEX(15, 62), MAKE_VAL_COMPLEX(14, -7), MAKE_VAL_COMPLEX(-67, 43) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1216, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1216", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1216() */
/* END TEST 1216 */

/*
 *  BEGIN TEST 1217
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1217( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(84, -21), MAKE_VAL_COMPLEX(-61, -70), MAKE_VAL_COMPLEX(67, 40), MAKE_VAL_COMPLEX(-158, -174), MAKE_VAL_COMPLEX(38, -200), MAKE_VAL_COMPLEX(220, 85), MAKE_VAL_COMPLEX(222, 18), MAKE_VAL_COMPLEX(28, 105), MAKE_VAL_COMPLEX(2, -70), MAKE_VAL_COMPLEX(166, -55), MAKE_VAL_COMPLEX(-44, -47), MAKE_VAL_COMPLEX(-103, 120), MAKE_VAL_COMPLEX(-25, -60), MAKE_VAL_COMPLEX(28, 9), MAKE_VAL_COMPLEX(87, 5) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1217, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1217", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1217() */
/* END TEST 1217 */

/*
 *  BEGIN TEST 1218
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1218( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(35, -35), MAKE_VAL_COMPLEX(-15, 5), MAKE_VAL_COMPLEX(-45, -35), MAKE_VAL_COMPLEX(-30, 70), MAKE_VAL_COMPLEX(10, 70), MAKE_VAL_COMPLEX(15, -65), MAKE_VAL_COMPLEX(55, 15), MAKE_VAL_COMPLEX(95, -5), MAKE_VAL_COMPLEX(-85, -15), MAKE_VAL_COMPLEX(-45, 5), MAKE_VAL_COMPLEX(35, 65), MAKE_VAL_COMPLEX(-5, -45), MAKE_VAL_COMPLEX(5, 5), MAKE_VAL_COMPLEX(-15, -25), MAKE_VAL_COMPLEX(20, -30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1218, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1218", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1218() */
/* END TEST 1218 */

/*
 *  BEGIN TEST 1219
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1219( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(840, 0), MAKE_VAL_COMPLEX(-94, -772), MAKE_VAL_COMPLEX(-197, 589), MAKE_VAL_COMPLEX(-38, -1584), MAKE_VAL_COMPLEX(1766, -722), MAKE_VAL_COMPLEX(919, 1417), MAKE_VAL_COMPLEX(1875, 1380), MAKE_VAL_COMPLEX(205, 555), MAKE_VAL_COMPLEX(-202, -301), MAKE_VAL_COMPLEX(1198, 794), MAKE_VAL_COMPLEX(430, -260), MAKE_VAL_COMPLEX(-1624, -72), MAKE_VAL_COMPLEX(236, -552), MAKE_VAL_COMPLEX(-44, 108), MAKE_VAL_COMPLEX(618, 284) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1219, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1219", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1219() */
/* END TEST 1219 */

/*
 *  BEGIN TEST 1220
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1220( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-17, 89), MAKE_VAL_COMPLEX(-61, -71), MAKE_VAL_COMPLEX(-14, -14), MAKE_VAL_COMPLEX(-152, 10), MAKE_VAL_COMPLEX(-152, -68), MAKE_VAL_COMPLEX(66, 72), MAKE_VAL_COMPLEX(180, 165), MAKE_VAL_COMPLEX(-8, -34), MAKE_VAL_COMPLEX(-117, -174), MAKE_VAL_COMPLEX(-3, 151), MAKE_VAL_COMPLEX(41, 141), MAKE_VAL_COMPLEX(67, 63), MAKE_VAL_COMPLEX(-59, 13), MAKE_VAL_COMPLEX(45, 15), MAKE_VAL_COMPLEX(-14, -18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1220, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1220", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1220() */
/* END TEST 1220 */

/*
 *  BEGIN TEST 1221
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1221( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(17, -40), MAKE_VAL_COMPLEX(68, 57), MAKE_VAL_COMPLEX(70, 7), MAKE_VAL_COMPLEX(124, -80), MAKE_VAL_COMPLEX(96, 26), MAKE_VAL_COMPLEX(-31, -16), MAKE_VAL_COMPLEX(-229, -137), MAKE_VAL_COMPLEX(-55, 104), MAKE_VAL_COMPLEX(187, 125), MAKE_VAL_COMPLEX(31, -186), MAKE_VAL_COMPLEX(-111, -162), MAKE_VAL_COMPLEX(-32, -35), MAKE_VAL_COMPLEX(52, -13), MAKE_VAL_COMPLEX(-17, -8), MAKE_VAL_COMPLEX(21, 53) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1221, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1221", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1221() */
/* END TEST 1221 */

/*
 *  BEGIN TEST 1222
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1222( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(56, -14), MAKE_VAL_COMPLEX(-18, -4), MAKE_VAL_COMPLEX(-24, -62), MAKE_VAL_COMPLEX(-72, 52), MAKE_VAL_COMPLEX(-32, 76), MAKE_VAL_COMPLEX(54, -56), MAKE_VAL_COMPLEX(46, 48), MAKE_VAL_COMPLEX(98, 52), MAKE_VAL_COMPLEX(-76, -66), MAKE_VAL_COMPLEX(-48, -22), MAKE_VAL_COMPLEX(-4, 86), MAKE_VAL_COMPLEX(22, -48), MAKE_VAL_COMPLEX(2, 8), MAKE_VAL_COMPLEX(0, -34), MAKE_VAL_COMPLEX(38, -18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1222, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1222", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1222() */
/* END TEST 1222 */

/*
 *  BEGIN TEST 1223
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_ROWMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1223( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_ROWMAJ, x1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(111, -172), MAKE_VAL_COMPLEX(114, 461), MAKE_VAL_COMPLEX(198, 335), MAKE_VAL_COMPLEX(882, 98), MAKE_VAL_COMPLEX(444, 358), MAKE_VAL_COMPLEX(-209, -174), MAKE_VAL_COMPLEX(-486, -1348), MAKE_VAL_COMPLEX(-561, 62), MAKE_VAL_COMPLEX(360, 1234), MAKE_VAL_COMPLEX(701, -562), MAKE_VAL_COMPLEX(165, -1064), MAKE_VAL_COMPLEX(-112, -217), MAKE_VAL_COMPLEX(256, 93), MAKE_VAL_COMPLEX(-91, -48), MAKE_VAL_COMPLEX(-139, 241) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, -3);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1223, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1223", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1223() */
/* END TEST 1223 */

/*
 *  BEGIN TEST 1224
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1224( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	oski_value_t* y_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(42, -69), MAKE_VAL_COMPLEX(43, -36), MAKE_VAL_COMPLEX(79, 114), MAKE_VAL_COMPLEX(-170, -87), MAKE_VAL_COMPLEX(-39, -292), MAKE_VAL_COMPLEX(69, 54), MAKE_VAL_COMPLEX(49, 101), MAKE_VAL_COMPLEX(-102, 188), MAKE_VAL_COMPLEX(-181, 93), MAKE_VAL_COMPLEX(46, 62), MAKE_VAL_COMPLEX(-204, -82), MAKE_VAL_COMPLEX(-56, 30), MAKE_VAL_COMPLEX(-30, 60), MAKE_VAL_COMPLEX(41, 40), MAKE_VAL_COMPLEX(39, 69) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1224, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1224", y_true, y_val, y1200_LEN*y1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1224() */
/* END TEST 1224 */

/*
 *  BEGIN TEST 1225
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1225( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	oski_value_t* y_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(88, 19), MAKE_VAL_COMPLEX(14, 126), MAKE_VAL_COMPLEX(-57, -18), MAKE_VAL_COMPLEX(121, -21), MAKE_VAL_COMPLEX(89, 228), MAKE_VAL_COMPLEX(-115, 1), MAKE_VAL_COMPLEX(-68, -131), MAKE_VAL_COMPLEX(132, -110), MAKE_VAL_COMPLEX(67, -79), MAKE_VAL_COMPLEX(18, -12), MAKE_VAL_COMPLEX(209, -2), MAKE_VAL_COMPLEX(-43, 23), MAKE_VAL_COMPLEX(-43, -17), MAKE_VAL_COMPLEX(-54, 62), MAKE_VAL_COMPLEX(19, 43) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1225, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1225", y_true, y_val, y1200_LEN*y1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1225() */
/* END TEST 1225 */

/*
 *  BEGIN TEST 1226
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1226( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	oski_value_t* y_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(130, 30), MAKE_VAL_COMPLEX(-3, 102), MAKE_VAL_COMPLEX(-34, 88), MAKE_VAL_COMPLEX(19, -112), MAKE_VAL_COMPLEX(74, -24), MAKE_VAL_COMPLEX(-66, 19), MAKE_VAL_COMPLEX(1, -34), MAKE_VAL_COMPLEX(-18, 78), MAKE_VAL_COMPLEX(-98, -50), MAKE_VAL_COMPLEX(24, 74), MAKE_VAL_COMPLEX(49, -64), MAKE_VAL_COMPLEX(-107, -11), MAKE_VAL_COMPLEX(-81, -5), MAKE_VAL_COMPLEX(-65, 74), MAKE_VAL_COMPLEX(-14, 120) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1226, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1226", y_true, y_val, y1200_LEN*y1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1226() */
/* END TEST 1226 */

/*
 *  BEGIN TEST 1227
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1227( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	oski_value_t* y_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(79, -373), MAKE_VAL_COMPLEX(821, -75), MAKE_VAL_COMPLEX(-481, 329), MAKE_VAL_COMPLEX(281, -977), MAKE_VAL_COMPLEX(2171, -319), MAKE_VAL_COMPLEX(-125, 811), MAKE_VAL_COMPLEX(-1138, 506), MAKE_VAL_COMPLEX(-1098, -1406), MAKE_VAL_COMPLEX(-420, -966), MAKE_VAL_COMPLEX(-344, -144), MAKE_VAL_COMPLEX(366, -1684), MAKE_VAL_COMPLEX(268, 72), MAKE_VAL_COMPLEX(-128, 112), MAKE_VAL_COMPLEX(307, 291), MAKE_VAL_COMPLEX(28, -250) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 9);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1227, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1227", y_true, y_val, y1200_LEN*y1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1227() */
/* END TEST 1227 */

/*
 *  BEGIN TEST 1228
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1228( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(31, 50), MAKE_VAL_COMPLEX(127, -140), MAKE_VAL_COMPLEX(239, 172), MAKE_VAL_COMPLEX(110, -158), MAKE_VAL_COMPLEX(-11, -173), MAKE_VAL_COMPLEX(12, 214), MAKE_VAL_COMPLEX(-84, 138), MAKE_VAL_COMPLEX(-149, -42), MAKE_VAL_COMPLEX(-36, 20), MAKE_VAL_COMPLEX(94, -19), MAKE_VAL_COMPLEX(-130, -140), MAKE_VAL_COMPLEX(-208, 114), MAKE_VAL_COMPLEX(-192, -39), MAKE_VAL_COMPLEX(194, 28), MAKE_VAL_COMPLEX(-78, 68) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1228, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1228", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1228() */
/* END TEST 1228 */

/*
 *  BEGIN TEST 1229
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1229( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-66, 20), MAKE_VAL_COMPLEX(-107, 125), MAKE_VAL_COMPLEX(-154, -142), MAKE_VAL_COMPLEX(-100, 38), MAKE_VAL_COMPLEX(-39, 73), MAKE_VAL_COMPLEX(-2, -109), MAKE_VAL_COMPLEX(-6, -133), MAKE_VAL_COMPLEX(9, 97), MAKE_VAL_COMPLEX(171, -40), MAKE_VAL_COMPLEX(-29, -11), MAKE_VAL_COMPLEX(45, 60), MAKE_VAL_COMPLEX(238, -49), MAKE_VAL_COMPLEX(182, 34), MAKE_VAL_COMPLEX(-159, 2), MAKE_VAL_COMPLEX(63, -13) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1229, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1229", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1229() */
/* END TEST 1229 */

/*
 *  BEGIN TEST 1230
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1230( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-42, 49), MAKE_VAL_COMPLEX(19, -8), MAKE_VAL_COMPLEX(62, 41), MAKE_VAL_COMPLEX(32, -94), MAKE_VAL_COMPLEX(-20, -90), MAKE_VAL_COMPLEX(-13, 86), MAKE_VAL_COMPLEX(-73, -14), MAKE_VAL_COMPLEX(-123, 16), MAKE_VAL_COMPLEX(112, 11), MAKE_VAL_COMPLEX(58, -11), MAKE_VAL_COMPLEX(-52, -81), MAKE_VAL_COMPLEX(11, 58), MAKE_VAL_COMPLEX(-7, -6), MAKE_VAL_COMPLEX(22, 31), MAKE_VAL_COMPLEX(-23, 41) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1230, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1230", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1230() */
/* END TEST 1230 */

/*
 *  BEGIN TEST 1231
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1231( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-571, -303), MAKE_VAL_COMPLEX(-1145, 631), MAKE_VAL_COMPLEX(-925, -1365), MAKE_VAL_COMPLEX(-604, 288), MAKE_VAL_COMPLEX(-170, 548), MAKE_VAL_COMPLEX(82, -1014), MAKE_VAL_COMPLEX(574, -1138), MAKE_VAL_COMPLEX(17, 293), MAKE_VAL_COMPLEX(1016, 548), MAKE_VAL_COMPLEX(-275, 39), MAKE_VAL_COMPLEX(506, 590), MAKE_VAL_COMPLEX(1610, 266), MAKE_VAL_COMPLEX(1207, 777), MAKE_VAL_COMPLEX(-1266, -470), MAKE_VAL_COMPLEX(410, -70) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, -3);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1231, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1231", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1231() */
/* END TEST 1231 */

/*
 *  BEGIN TEST 1232
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1232( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-29, -106), MAKE_VAL_COMPLEX(-39, 122), MAKE_VAL_COMPLEX(-13, -192), MAKE_VAL_COMPLEX(22, -94), MAKE_VAL_COMPLEX(89, -13), MAKE_VAL_COMPLEX(-32, -92), MAKE_VAL_COMPLEX(-232, 24), MAKE_VAL_COMPLEX(205, 32), MAKE_VAL_COMPLEX(-38, 136), MAKE_VAL_COMPLEX(16, 61), MAKE_VAL_COMPLEX(40, -28), MAKE_VAL_COMPLEX(-60, -120), MAKE_VAL_COMPLEX(4, 25), MAKE_VAL_COMPLEX(-32, 116), MAKE_VAL_COMPLEX(-40, -34) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1232, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1232", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1232() */
/* END TEST 1232 */

/*
 *  BEGIN TEST 1233
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1233( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-20, 92), MAKE_VAL_COMPLEX(51, -111), MAKE_VAL_COMPLEX(4, 250), MAKE_VAL_COMPLEX(56, 86), MAKE_VAL_COMPLEX(-31, -31), MAKE_VAL_COMPLEX(-34, 111), MAKE_VAL_COMPLEX(218, -81), MAKE_VAL_COMPLEX(-257, -115), MAKE_VAL_COMPLEX(67, -52), MAKE_VAL_COMPLEX(11, -23), MAKE_VAL_COMPLEX(1, -36), MAKE_VAL_COMPLEX(22, 147), MAKE_VAL_COMPLEX(-2, -32), MAKE_VAL_COMPLEX(17, -90), MAKE_VAL_COMPLEX(3, 31) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1233, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1233", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1233() */
/* END TEST 1233 */

/*
 *  BEGIN TEST 1234
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1234( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-35, 21), MAKE_VAL_COMPLEX(13, -1), MAKE_VAL_COMPLEX(29, 37), MAKE_VAL_COMPLEX(38, -50), MAKE_VAL_COMPLEX(6, -58), MAKE_VAL_COMPLEX(-25, 49), MAKE_VAL_COMPLEX(-41, -23), MAKE_VAL_COMPLEX(-77, -15), MAKE_VAL_COMPLEX(65, 29), MAKE_VAL_COMPLEX(37, 5), MAKE_VAL_COMPLEX(-15, -59), MAKE_VAL_COMPLEX(-5, 37), MAKE_VAL_COMPLEX(-3, -5), MAKE_VAL_COMPLEX(7, 23), MAKE_VAL_COMPLEX(-22, 20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1234, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1234", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1234() */
/* END TEST 1234 */

/*
 *  BEGIN TEST 1235
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1235( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_ROWMAJ, y1200_VW );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-762, -368), MAKE_VAL_COMPLEX(705, 909), MAKE_VAL_COMPLEX(-1714, -1110), MAKE_VAL_COMPLEX(-932, -414), MAKE_VAL_COMPLEX(223, -453), MAKE_VAL_COMPLEX(-682, -449), MAKE_VAL_COMPLEX(-766, 2167), MAKE_VAL_COMPLEX(1701, -1283), MAKE_VAL_COMPLEX(567, 724), MAKE_VAL_COMPLEX(363, 95), MAKE_VAL_COMPLEX(-23, -100), MAKE_VAL_COMPLEX(-1178, -325), MAKE_VAL_COMPLEX(230, 126), MAKE_VAL_COMPLEX(761, 694), MAKE_VAL_COMPLEX(-317, 67) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1235, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1235", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1235() */
/* END TEST 1235 */

/*
 *  BEGIN TEST 1236
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1236( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	oski_value_t* y_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-88, -39), MAKE_VAL_COMPLEX(-68, -68), MAKE_VAL_COMPLEX(113, 87), MAKE_VAL_COMPLEX(-58, 91), MAKE_VAL_COMPLEX(-30, -6), MAKE_VAL_COMPLEX(6, -74), MAKE_VAL_COMPLEX(-65, -213), MAKE_VAL_COMPLEX(-120, 110), MAKE_VAL_COMPLEX(-150, 32), MAKE_VAL_COMPLEX(27, -80), MAKE_VAL_COMPLEX(-9, 65), MAKE_VAL_COMPLEX(130, 33), MAKE_VAL_COMPLEX(-66, 70), MAKE_VAL_COMPLEX(50, -57), MAKE_VAL_COMPLEX(-1, -45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1236, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1236", y_true, y_val, y1200_LEN*y1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1236() */
/* END TEST 1236 */

/*
 *  BEGIN TEST 1237
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1237( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	oski_value_t* y_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-92, 19), MAKE_VAL_COMPLEX(56, -70), MAKE_VAL_COMPLEX(-81, -211), MAKE_VAL_COMPLEX(50, 63), MAKE_VAL_COMPLEX(-66, 50), MAKE_VAL_COMPLEX(80, 38), MAKE_VAL_COMPLEX(69, 259), MAKE_VAL_COMPLEX(132, -218), MAKE_VAL_COMPLEX(290, 20), MAKE_VAL_COMPLEX(-71, -16), MAKE_VAL_COMPLEX(-47, 29), MAKE_VAL_COMPLEX(16, -35), MAKE_VAL_COMPLEX(176, -76), MAKE_VAL_COMPLEX(26, -53), MAKE_VAL_COMPLEX(1, -119) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1237, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1237", y_true, y_val, y1200_LEN*y1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1237() */
/* END TEST 1237 */

/*
 *  BEGIN TEST 1238
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1238( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	oski_value_t* y_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-90, -70), MAKE_VAL_COMPLEX(39, -78), MAKE_VAL_COMPLEX(58, -56), MAKE_VAL_COMPLEX(-55, 80), MAKE_VAL_COMPLEX(-66, -8), MAKE_VAL_COMPLEX(58, 9), MAKE_VAL_COMPLEX(-13, 26), MAKE_VAL_COMPLEX(42, -54), MAKE_VAL_COMPLEX(58, 74), MAKE_VAL_COMPLEX(8, -66), MAKE_VAL_COMPLEX(-61, 32), MAKE_VAL_COMPLEX(79, 47), MAKE_VAL_COMPLEX(61, 33), MAKE_VAL_COMPLEX(77, -34), MAKE_VAL_COMPLEX(54, -88) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1238, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1238", y_true, y_val, y1200_LEN*y1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1238() */
/* END TEST 1238 */

/*
 *  BEGIN TEST 1239
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1239( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	oski_value_t* y_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-280, 345), MAKE_VAL_COMPLEX(-866, 373), MAKE_VAL_COMPLEX(1193, 321), MAKE_VAL_COMPLEX(188, 496), MAKE_VAL_COMPLEX(66, 12), MAKE_VAL_COMPLEX(-681, -452), MAKE_VAL_COMPLEX(-1739, -1248), MAKE_VAL_COMPLEX(-234, 1832), MAKE_VAL_COMPLEX(-1296, 938), MAKE_VAL_COMPLEX(-165, -410), MAKE_VAL_COMPLEX(519, 218), MAKE_VAL_COMPLEX(643, -564), MAKE_VAL_COMPLEX(-393, 889), MAKE_VAL_COMPLEX(-250, -360), MAKE_VAL_COMPLEX(-301, 213) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1239, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1239", y_true, y_val, y1200_LEN*y1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1239() */
/* END TEST 1239 */

/*
 *  BEGIN TEST 1240
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1240( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-63, 56), MAKE_VAL_COMPLEX(116, 131), MAKE_VAL_COMPLEX(-250, 8), MAKE_VAL_COMPLEX(-166, 26), MAKE_VAL_COMPLEX(14, 70), MAKE_VAL_COMPLEX(70, 53), MAKE_VAL_COMPLEX(-86, 193), MAKE_VAL_COMPLEX(-43, -28), MAKE_VAL_COMPLEX(-19, 42), MAKE_VAL_COMPLEX(-11, -16), MAKE_VAL_COMPLEX(-32, -51), MAKE_VAL_COMPLEX(-175, -60), MAKE_VAL_COMPLEX(34, 16), MAKE_VAL_COMPLEX(137, -114), MAKE_VAL_COMPLEX(-67, 17) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1240, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1240", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1240() */
/* END TEST 1240 */

/*
 *  BEGIN TEST 1241
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1241( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(28, -112), MAKE_VAL_COMPLEX(-114, -110), MAKE_VAL_COMPLEX(191, 40), MAKE_VAL_COMPLEX(248, 34), MAKE_VAL_COMPLEX(80, -62), MAKE_VAL_COMPLEX(-150, -92), MAKE_VAL_COMPLEX(122, -260), MAKE_VAL_COMPLEX(65, -97), MAKE_VAL_COMPLEX(-26, 64), MAKE_VAL_COMPLEX(4, 76), MAKE_VAL_COMPLEX(127, 25), MAKE_VAL_COMPLEX(115, 53), MAKE_VAL_COMPLEX(-26, -21), MAKE_VAL_COMPLEX(-174, 126), MAKE_VAL_COMPLEX(34, -52) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1241, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1241", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1241() */
/* END TEST 1241 */

/*
 *  BEGIN TEST 1242
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1242( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(14, -56), MAKE_VAL_COMPLEX(-12, 14), MAKE_VAL_COMPLEX(-66, -8), MAKE_VAL_COMPLEX(12, 88), MAKE_VAL_COMPLEX(52, 64), MAKE_VAL_COMPLEX(-24, -74), MAKE_VAL_COMPLEX(64, -18), MAKE_VAL_COMPLEX(92, -62), MAKE_VAL_COMPLEX(-94, 36), MAKE_VAL_COMPLEX(-42, 32), MAKE_VAL_COMPLEX(74, 44), MAKE_VAL_COMPLEX(-32, -42), MAKE_VAL_COMPLEX(8, 2), MAKE_VAL_COMPLEX(-30, -16), MAKE_VAL_COMPLEX(2, -42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1242, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1242", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1242() */
/* END TEST 1242 */

/*
 *  BEGIN TEST 1243
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1243( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-7, 259), MAKE_VAL_COMPLEX(495, 23), MAKE_VAL_COMPLEX(-461, 505), MAKE_VAL_COMPLEX(-302, 358), MAKE_VAL_COMPLEX(138, 102), MAKE_VAL_COMPLEX(269, -15), MAKE_VAL_COMPLEX(197, 577), MAKE_VAL_COMPLEX(-159, 69), MAKE_VAL_COMPLEX(69, 91), MAKE_VAL_COMPLEX(-47, -29), MAKE_VAL_COMPLEX(-199, -37), MAKE_VAL_COMPLEX(-451, 237), MAKE_VAL_COMPLEX(97, -35), MAKE_VAL_COMPLEX(59, -503), MAKE_VAL_COMPLEX(-92, 182) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1243, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1243", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1243() */
/* END TEST 1243 */

/*
 *  BEGIN TEST 1244
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1244( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-66, 89), MAKE_VAL_COMPLEX(-111, 78), MAKE_VAL_COMPLEX(297, 201), MAKE_VAL_COMPLEX(16, 80), MAKE_VAL_COMPLEX(-60, -92), MAKE_VAL_COMPLEX(-99, 38), MAKE_VAL_COMPLEX(-183, -46), MAKE_VAL_COMPLEX(-78, -97), MAKE_VAL_COMPLEX(234, 203), MAKE_VAL_COMPLEX(86, 1), MAKE_VAL_COMPLEX(-117, -116), MAKE_VAL_COMPLEX(42, 79), MAKE_VAL_COMPLEX(-201, -143), MAKE_VAL_COMPLEX(56, 69), MAKE_VAL_COMPLEX(-49, -11) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1244, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1244", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1244() */
/* END TEST 1244 */

/*
 *  BEGIN TEST 1245
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1245( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(45, 16), MAKE_VAL_COMPLEX(132, -105), MAKE_VAL_COMPLEX(-174, -192), MAKE_VAL_COMPLEX(-46, -242), MAKE_VAL_COMPLEX(-42, -22), MAKE_VAL_COMPLEX(150, 97), MAKE_VAL_COMPLEX(66, 85), MAKE_VAL_COMPLEX(-87, 220), MAKE_VAL_COMPLEX(-63, -278), MAKE_VAL_COMPLEX(-11, -64), MAKE_VAL_COMPLEX(-24, 41), MAKE_VAL_COMPLEX(21, -4), MAKE_VAL_COMPLEX(186, 140), MAKE_VAL_COMPLEX(1, -42), MAKE_VAL_COMPLEX(49, 89) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1245, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1245", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1245() */
/* END TEST 1245 */

/*
 *  BEGIN TEST 1246
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1246( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-56, 21), MAKE_VAL_COMPLEX(19, 2), MAKE_VAL_COMPLEX(32, 61), MAKE_VAL_COMPLEX(68, -62), MAKE_VAL_COMPLEX(24, -82), MAKE_VAL_COMPLEX(-49, 64), MAKE_VAL_COMPLEX(-53, -44), MAKE_VAL_COMPLEX(-107, -42), MAKE_VAL_COMPLEX(86, 59), MAKE_VAL_COMPLEX(52, 17), MAKE_VAL_COMPLEX(-6, -89), MAKE_VAL_COMPLEX(-17, 52), MAKE_VAL_COMPLEX(-3, -8), MAKE_VAL_COMPLEX(4, 35), MAKE_VAL_COMPLEX(-37, 23) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1246, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1246", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1246() */
/* END TEST 1246 */

/*
 *  BEGIN TEST 1247
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1200', LAYOUT_COLMAJ>
 *  y : <'y1200', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1247( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1200_ptr, A1200_ind, A1200_val,
		A1200_M, A1200_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1200_val,
		y1200_LEN * y1200_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1200_LEN, y1200_VW, LAYOUT_COLMAJ, y1200_LEN );
	oski_value_t* y_val = array_Duplicate( x1200_val,
		x1200_LEN * x1200_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1200_LEN, x1200_VW, LAYOUT_COLMAJ, x1200_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-409, -90), MAKE_VAL_COMPLEX(-742, -1199), MAKE_VAL_COMPLEX(-1289, 2153), MAKE_VAL_COMPLEX(-1568, -352), MAKE_VAL_COMPLEX(-84, -524), MAKE_VAL_COMPLEX(466, -595), MAKE_VAL_COMPLEX(134, -967), MAKE_VAL_COMPLEX(785, 522), MAKE_VAL_COMPLEX(-1499, 1042), MAKE_VAL_COMPLEX(-1, 190), MAKE_VAL_COMPLEX(156, -709), MAKE_VAL_COMPLEX(-127, 386), MAKE_VAL_COMPLEX(1215, -1753), MAKE_VAL_COMPLEX(-277, 352), MAKE_VAL_COMPLEX(451, -137) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 9);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1247, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1247", y_true, y_val, x1200_LEN*x1200_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1247() */
/* END TEST 1247 */

#define A1248_M 5    /* # of rows */
#define A1248_N 5    /* # of cols */
#define A1248_NNZ 10  /* # of stored non-zeros */
static oski_index_t A1248_ptr[A1248_M+1] = {1, 3, 6, 8, 9, 11}; /* end row pointers for A1248 */
static oski_index_t A1248_ind[A1248_NNZ] = {2, 3, 1, 3, 5, 1, 2, 5, 2, 4}; /* end column indices pointers for A1248 */
static oski_value_t A1248_val[A1248_NNZ] = {MAKE_VAL_COMPLEX(-3, -5), MAKE_VAL_COMPLEX(0, 1), MAKE_VAL_COMPLEX(-3, 5), MAKE_VAL_COMPLEX(-2, 4), MAKE_VAL_COMPLEX(-10, 10), MAKE_VAL_COMPLEX(0, -1), MAKE_VAL_COMPLEX(-2, -4), MAKE_VAL_COMPLEX(-7, 10), MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(-7, -10)}; /* end stored non-zero values for A1248 */

#define x1248_LEN 5   /* vector length */
#define x1248_VW  3   /* vector width */
static oski_value_t x1248_val[x1248_LEN*x1248_VW] = {MAKE_VAL_COMPLEX(0, -10), MAKE_VAL_COMPLEX(-10, -4), MAKE_VAL_COMPLEX(5, 1), MAKE_VAL_COMPLEX(1, -3), MAKE_VAL_COMPLEX(-10, -1), MAKE_VAL_COMPLEX(2, -1), MAKE_VAL_COMPLEX(4, -10), MAKE_VAL_COMPLEX(-6, 10), MAKE_VAL_COMPLEX(0, 10), MAKE_VAL_COMPLEX(10, 7), MAKE_VAL_COMPLEX(-6, -10), MAKE_VAL_COMPLEX(-5, 8), MAKE_VAL_COMPLEX(3, 9), MAKE_VAL_COMPLEX(2, 10), MAKE_VAL_COMPLEX(-0, -5)}; /* data */

#define y1248_LEN 5   /* vector length */
#define y1248_VW  3   /* vector width */
static oski_value_t y1248_val[y1248_LEN*y1248_VW] = {MAKE_VAL_COMPLEX(10, 7), MAKE_VAL_COMPLEX(-1, -10), MAKE_VAL_COMPLEX(-7, 0), MAKE_VAL_COMPLEX(10, 3), MAKE_VAL_COMPLEX(-2, 1), MAKE_VAL_COMPLEX(8, -10), MAKE_VAL_COMPLEX(2, -1), MAKE_VAL_COMPLEX(-6, 10), MAKE_VAL_COMPLEX(6, 6), MAKE_VAL_COMPLEX(-10, 10), MAKE_VAL_COMPLEX(0, -2), MAKE_VAL_COMPLEX(-9, 9), MAKE_VAL_COMPLEX(6, -6), MAKE_VAL_COMPLEX(-4, 5), MAKE_VAL_COMPLEX(10, -6)}; /* data */

/*
 *  BEGIN TEST 1248
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1248( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	oski_value_t* y_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(72, 54), MAKE_VAL_COMPLEX(-3, -37), MAKE_VAL_COMPLEX(-72, -6), MAKE_VAL_COMPLEX(43, 27), MAKE_VAL_COMPLEX(-124, -155), MAKE_VAL_COMPLEX(56, -29), MAKE_VAL_COMPLEX(-4, -16), MAKE_VAL_COMPLEX(-42, 142), MAKE_VAL_COMPLEX(41, 47), MAKE_VAL_COMPLEX(-181, 54), MAKE_VAL_COMPLEX(-120, -76), MAKE_VAL_COMPLEX(-27, 115), MAKE_VAL_COMPLEX(11, -168), MAKE_VAL_COMPLEX(2, 290), MAKE_VAL_COMPLEX(165, -69) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1248, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1248", y_true, y_val, y1248_LEN*y1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1248() */
/* END TEST 1248 */

/*
 *  BEGIN TEST 1249
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1249( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	oski_value_t* y_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(130, -2), MAKE_VAL_COMPLEX(-73, -117), MAKE_VAL_COMPLEX(-40, 48), MAKE_VAL_COMPLEX(135, -39), MAKE_VAL_COMPLEX(98, 183), MAKE_VAL_COMPLEX(12, -179), MAKE_VAL_COMPLEX(30, -12), MAKE_VAL_COMPLEX(6, 54), MAKE_VAL_COMPLEX(91, 13), MAKE_VAL_COMPLEX(81, 166), MAKE_VAL_COMPLEX(108, 44), MAKE_VAL_COMPLEX(-63, 83), MAKE_VAL_COMPLEX(49, 36), MAKE_VAL_COMPLEX(-36, -186), MAKE_VAL_COMPLEX(-41, -87) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1249, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1249", y_true, y_val, y1248_LEN*y1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1249() */
/* END TEST 1249 */

/*
 *  BEGIN TEST 1250
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1250( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	oski_value_t* y_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-52, 53), MAKE_VAL_COMPLEX(61, 4), MAKE_VAL_COMPLEX(7, -42), MAKE_VAL_COMPLEX(-28, 57), MAKE_VAL_COMPLEX(-4, -13), MAKE_VAL_COMPLEX(52, 58), MAKE_VAL_COMPLEX(4, 13), MAKE_VAL_COMPLEX(-54, -46), MAKE_VAL_COMPLEX(-42, 30), MAKE_VAL_COMPLEX(-50, -70), MAKE_VAL_COMPLEX(12, 2), MAKE_VAL_COMPLEX(-45, -63), MAKE_VAL_COMPLEX(30, 42), MAKE_VAL_COMPLEX(-26, -29), MAKE_VAL_COMPLEX(26, 66) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1250, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1250", y_true, y_val, y1248_LEN*y1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1250() */
/* END TEST 1250 */

/*
 *  BEGIN TEST 1251
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1251( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	oski_value_t* y_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-144, -23), MAKE_VAL_COMPLEX(376, 419), MAKE_VAL_COMPLEX(-143, 88), MAKE_VAL_COMPLEX(-373, 282), MAKE_VAL_COMPLEX(-2260, -606), MAKE_VAL_COMPLEX(260, 577), MAKE_VAL_COMPLEX(-260, 91), MAKE_VAL_COMPLEX(612, 452), MAKE_VAL_COMPLEX(-101, -7), MAKE_VAL_COMPLEX(-1027, 524), MAKE_VAL_COMPLEX(-1564, 434), MAKE_VAL_COMPLEX(830, -18), MAKE_VAL_COMPLEX(-1347, -754), MAKE_VAL_COMPLEX(2344, 1951), MAKE_VAL_COMPLEX(515, -847) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(9, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1251, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1251", y_true, y_val, y1248_LEN*y1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1251() */
/* END TEST 1251 */

/*
 *  BEGIN TEST 1252
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1252( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-26, 6), MAKE_VAL_COMPLEX(-26, -43), MAKE_VAL_COMPLEX(44, 73), MAKE_VAL_COMPLEX(-106, -85), MAKE_VAL_COMPLEX(54, 16), MAKE_VAL_COMPLEX(-110, -53), MAKE_VAL_COMPLEX(-11, 7), MAKE_VAL_COMPLEX(-30, 33), MAKE_VAL_COMPLEX(20, 91), MAKE_VAL_COMPLEX(-79, 30), MAKE_VAL_COMPLEX(64, -43), MAKE_VAL_COMPLEX(-167, -22), MAKE_VAL_COMPLEX(-151, -67), MAKE_VAL_COMPLEX(24, 31), MAKE_VAL_COMPLEX(8, 1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1252, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1252", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1252() */
/* END TEST 1252 */

/*
 *  BEGIN TEST 1253
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1253( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(106, 34), MAKE_VAL_COMPLEX(98, -21), MAKE_VAL_COMPLEX(-72, -37), MAKE_VAL_COMPLEX(126, 105), MAKE_VAL_COMPLEX(-6, -92), MAKE_VAL_COMPLEX(110, 73), MAKE_VAL_COMPLEX(75, 65), MAKE_VAL_COMPLEX(-26, -121), MAKE_VAL_COMPLEX(-100, -131), MAKE_VAL_COMPLEX(-17, 22), MAKE_VAL_COMPLEX(40, 35), MAKE_VAL_COMPLEX(123, -50), MAKE_VAL_COMPLEX(67, 55), MAKE_VAL_COMPLEX(-112, -55), MAKE_VAL_COMPLEX(32, 19) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1253, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1253", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1253() */
/* END TEST 1253 */

/*
 *  BEGIN TEST 1254
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1254( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(80, -40), MAKE_VAL_COMPLEX(-8, -96), MAKE_VAL_COMPLEX(12, 44), MAKE_VAL_COMPLEX(28, -4), MAKE_VAL_COMPLEX(-32, -84), MAKE_VAL_COMPLEX(16, 12), MAKE_VAL_COMPLEX(96, -8), MAKE_VAL_COMPLEX(-104, -8), MAKE_VAL_COMPLEX(-80, 40), MAKE_VAL_COMPLEX(-16, 108), MAKE_VAL_COMPLEX(56, -88), MAKE_VAL_COMPLEX(-84, -8), MAKE_VAL_COMPLEX(-60, 60), MAKE_VAL_COMPLEX(-72, 56), MAKE_VAL_COMPLEX(40, -20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1254, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1254", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1254() */
/* END TEST 1254 */

/*
 *  BEGIN TEST 1255
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1255( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-396, 342), MAKE_VAL_COMPLEX(132, -153), MAKE_VAL_COMPLEX(95, 636), MAKE_VAL_COMPLEX(-863, -890), MAKE_VAL_COMPLEX(779, 436), MAKE_VAL_COMPLEX(-967, -693), MAKE_VAL_COMPLEX(-399, 321), MAKE_VAL_COMPLEX(18, -13), MAKE_VAL_COMPLEX(148, 519), MAKE_VAL_COMPLEX(-1006, -276), MAKE_VAL_COMPLEX(716, 171), MAKE_VAL_COMPLEX(-1169, -729), MAKE_VAL_COMPLEX(-1192, -1256), MAKE_VAL_COMPLEX(236, -43), MAKE_VAL_COMPLEX(-5, 195) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(9, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1255, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1255", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1255() */
/* END TEST 1255 */

/*
 *  BEGIN TEST 1256
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1256( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(46, -90), MAKE_VAL_COMPLEX(-20, -75), MAKE_VAL_COMPLEX(-72, 25), MAKE_VAL_COMPLEX(-36, 155), MAKE_VAL_COMPLEX(-22, -162), MAKE_VAL_COMPLEX(-34, 133), MAKE_VAL_COMPLEX(67, -77), MAKE_VAL_COMPLEX(-82, 27), MAKE_VAL_COMPLEX(-100, 41), MAKE_VAL_COMPLEX(13, 190), MAKE_VAL_COMPLEX(4, -147), MAKE_VAL_COMPLEX(-79, 158), MAKE_VAL_COMPLEX(73, -55), MAKE_VAL_COMPLEX(-36, 79), MAKE_VAL_COMPLEX(8, 21) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1256, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1256", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1256() */
/* END TEST 1256 */

/*
 *  BEGIN TEST 1257
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1257( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(14, -20), MAKE_VAL_COMPLEX(-66, -29), MAKE_VAL_COMPLEX(121, 16), MAKE_VAL_COMPLEX(65, -182), MAKE_VAL_COMPLEX(-82, 91), MAKE_VAL_COMPLEX(62, -132), MAKE_VAL_COMPLEX(37, -9), MAKE_VAL_COMPLEX(-44, 47), MAKE_VAL_COMPLEX(40, 69), MAKE_VAL_COMPLEX(55, -53), MAKE_VAL_COMPLEX(-10, 1), MAKE_VAL_COMPLEX(-24, -100), MAKE_VAL_COMPLEX(-94, 172), MAKE_VAL_COMPLEX(-2, 43), MAKE_VAL_COMPLEX(22, -76) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1257, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1257", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1257() */
/* END TEST 1257 */

/*
 *  BEGIN TEST 1258
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1258( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(10, -20), MAKE_VAL_COMPLEX(-16, -18), MAKE_VAL_COMPLEX(9, 7), MAKE_VAL_COMPLEX(5, -5), MAKE_VAL_COMPLEX(-19, -12), MAKE_VAL_COMPLEX(5, 0), MAKE_VAL_COMPLEX(18, -16), MAKE_VAL_COMPLEX(-22, 14), MAKE_VAL_COMPLEX(-10, 20), MAKE_VAL_COMPLEX(13, 24), MAKE_VAL_COMPLEX(-2, -26), MAKE_VAL_COMPLEX(-18, 11), MAKE_VAL_COMPLEX(-3, 21), MAKE_VAL_COMPLEX(-6, 22), MAKE_VAL_COMPLEX(5, -10) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1258, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1258", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1258() */
/* END TEST 1258 */

/*
 *  BEGIN TEST 1259
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1259( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-154, 272), MAKE_VAL_COMPLEX(25, 80), MAKE_VAL_COMPLEX(308, 261), MAKE_VAL_COMPLEX(694, -467), MAKE_VAL_COMPLEX(-318, 419), MAKE_VAL_COMPLEX(555, -365), MAKE_VAL_COMPLEX(-219, 277), MAKE_VAL_COMPLEX(139, -92), MAKE_VAL_COMPLEX(223, 86), MAKE_VAL_COMPLEX(258, -534), MAKE_VAL_COMPLEX(-127, 446), MAKE_VAL_COMPLEX(575, -585), MAKE_VAL_COMPLEX(-742, 46), MAKE_VAL_COMPLEX(71, -198), MAKE_VAL_COMPLEX(181, -83) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, -3);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1259, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1259", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1259() */
/* END TEST 1259 */

/*
 *  BEGIN TEST 1260
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1260( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	oski_value_t* y_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(45, 50), MAKE_VAL_COMPLEX(-33, -58), MAKE_VAL_COMPLEX(-62, -15), MAKE_VAL_COMPLEX(-44, 2), MAKE_VAL_COMPLEX(-50, -116), MAKE_VAL_COMPLEX(63, -9), MAKE_VAL_COMPLEX(-95, -167), MAKE_VAL_COMPLEX(-40, 116), MAKE_VAL_COMPLEX(-90, -18), MAKE_VAL_COMPLEX(-36, 300), MAKE_VAL_COMPLEX(-14, -18), MAKE_VAL_COMPLEX(-71, 96), MAKE_VAL_COMPLEX(35, -31), MAKE_VAL_COMPLEX(16, 69), MAKE_VAL_COMPLEX(151, -47) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1260, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1260", y_true, y_val, y1248_LEN*y1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1260() */
/* END TEST 1260 */

/*
 *  BEGIN TEST 1261
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1261( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	oski_value_t* y_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-32, -26), MAKE_VAL_COMPLEX(41, 37), MAKE_VAL_COMPLEX(48, 8), MAKE_VAL_COMPLEX(61, 14), MAKE_VAL_COMPLEX(45, 116), MAKE_VAL_COMPLEX(-37, -3), MAKE_VAL_COMPLEX(100, 167), MAKE_VAL_COMPLEX(18, -102), MAKE_VAL_COMPLEX(96, 36), MAKE_VAL_COMPLEX(6, -290), MAKE_VAL_COMPLEX(16, 14), MAKE_VAL_COMPLEX(44, -87), MAKE_VAL_COMPLEX(-17, 25), MAKE_VAL_COMPLEX(-29, -63), MAKE_VAL_COMPLEX(-125, 45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1261, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1261", y_true, y_val, y1248_LEN*y1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1261() */
/* END TEST 1261 */

/*
 *  BEGIN TEST 1262
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1262( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	oski_value_t* y_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-3, -17), MAKE_VAL_COMPLEX(-9, 11), MAKE_VAL_COMPLEX(7, 7), MAKE_VAL_COMPLEX(-7, -13), MAKE_VAL_COMPLEX(3, 1), MAKE_VAL_COMPLEX(-18, 2), MAKE_VAL_COMPLEX(-3, -1), MAKE_VAL_COMPLEX(16, -4), MAKE_VAL_COMPLEX(0, -12), MAKE_VAL_COMPLEX(20, 0), MAKE_VAL_COMPLEX(-2, 2), MAKE_VAL_COMPLEX(18, 0), MAKE_VAL_COMPLEX(-12, 0), MAKE_VAL_COMPLEX(9, -1), MAKE_VAL_COMPLEX(-16, -4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1262, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1262", y_true, y_val, y1248_LEN*y1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1262() */
/* END TEST 1262 */

/*
 *  BEGIN TEST 1263
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1263( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	oski_value_t* y_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(73, -93), MAKE_VAL_COMPLEX(-162, -202), MAKE_VAL_COMPLEX(-21, -67), MAKE_VAL_COMPLEX(-21, -618), MAKE_VAL_COMPLEX(527, -408), MAKE_VAL_COMPLEX(-251, 25), MAKE_VAL_COMPLEX(598, -883), MAKE_VAL_COMPLEX(-246, 226), MAKE_VAL_COMPLEX(120, -744), MAKE_VAL_COMPLEX(-1142, 770), MAKE_VAL_COMPLEX(-16, -98), MAKE_VAL_COMPLEX(-235, 152), MAKE_VAL_COMPLEX(-33, -97), MAKE_VAL_COMPLEX(-102, 354), MAKE_VAL_COMPLEX(263, 295) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1263, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1263", y_true, y_val, y1248_LEN*y1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1263() */
/* END TEST 1263 */

/*
 *  BEGIN TEST 1264
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1264( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(103, 89), MAKE_VAL_COMPLEX(83, -11), MAKE_VAL_COMPLEX(-53, -65), MAKE_VAL_COMPLEX(55, -24), MAKE_VAL_COMPLEX(234, -42), MAKE_VAL_COMPLEX(-221, -24), MAKE_VAL_COMPLEX(48, 92), MAKE_VAL_COMPLEX(-6, -46), MAKE_VAL_COMPLEX(-50, -110), MAKE_VAL_COMPLEX(-44, 21), MAKE_VAL_COMPLEX(246, 62), MAKE_VAL_COMPLEX(-141, -113), MAKE_VAL_COMPLEX(-43, 137), MAKE_VAL_COMPLEX(-172, 16), MAKE_VAL_COMPLEX(8, -236) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1264, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1264", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1264() */
/* END TEST 1264 */

/*
 *  BEGIN TEST 1265
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1265( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-123, 51), MAKE_VAL_COMPLEX(49, 87), MAKE_VAL_COMPLEX(-15, 41), MAKE_VAL_COMPLEX(-75, 64), MAKE_VAL_COMPLEX(-96, 76), MAKE_VAL_COMPLEX(191, 34), MAKE_VAL_COMPLEX(-124, 40), MAKE_VAL_COMPLEX(110, -82), MAKE_VAL_COMPLEX(70, -30), MAKE_VAL_COMPLEX(-82, -139), MAKE_VAL_COMPLEX(-182, 90), MAKE_VAL_COMPLEX(227, 11), MAKE_VAL_COMPLEX(19, -269), MAKE_VAL_COMPLEX(164, -160), MAKE_VAL_COMPLEX(-18, 306) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1265, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1265", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1265() */
/* END TEST 1265 */

/*
 *  BEGIN TEST 1266
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1266( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-90, 30), MAKE_VAL_COMPLEX(-6, 102), MAKE_VAL_COMPLEX(-6, -48), MAKE_VAL_COMPLEX(-30, 0), MAKE_VAL_COMPLEX(21, 93), MAKE_VAL_COMPLEX(-15, -15), MAKE_VAL_COMPLEX(-102, -6), MAKE_VAL_COMPLEX(108, 24), MAKE_VAL_COMPLEX(90, -30), MAKE_VAL_COMPLEX(33, -111), MAKE_VAL_COMPLEX(-72, 84), MAKE_VAL_COMPLEX(87, 21), MAKE_VAL_COMPLEX(72, -54), MAKE_VAL_COMPLEX(84, -48), MAKE_VAL_COMPLEX(-45, 15) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1266, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1266", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1266() */
/* END TEST 1266 */

/*
 *  BEGIN TEST 1267
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1267( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-222, -226), MAKE_VAL_COMPLEX(-126, -94), MAKE_VAL_COMPLEX(128, 342), MAKE_VAL_COMPLEX(-176, 154), MAKE_VAL_COMPLEX(-787, -9), MAKE_VAL_COMPLEX(877, 147), MAKE_VAL_COMPLEX(-54, -162), MAKE_VAL_COMPLEX(-88, -60), MAKE_VAL_COMPLEX(10, 310), MAKE_VAL_COMPLEX(-87, 15), MAKE_VAL_COMPLEX(-716, -232), MAKE_VAL_COMPLEX(477, 253), MAKE_VAL_COMPLEX(-38, -608), MAKE_VAL_COMPLEX(472, -156), MAKE_VAL_COMPLEX(63, 1009) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1267, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1267", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1267() */
/* END TEST 1267 */

/*
 *  BEGIN TEST 1268
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1268( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(33, -5), MAKE_VAL_COMPLEX(-25, -109), MAKE_VAL_COMPLEX(91, 61), MAKE_VAL_COMPLEX(-17, -44), MAKE_VAL_COMPLEX(-33, -249), MAKE_VAL_COMPLEX(-12, 221), MAKE_VAL_COMPLEX(48, 2), MAKE_VAL_COMPLEX(-118, -6), MAKE_VAL_COMPLEX(-12, 52), MAKE_VAL_COMPLEX(5, 74), MAKE_VAL_COMPLEX(22, -246), MAKE_VAL_COMPLEX(-90, 144), MAKE_VAL_COMPLEX(-183, 47), MAKE_VAL_COMPLEX(-84, -48), MAKE_VAL_COMPLEX(303, -21) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1268, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1268", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1268() */
/* END TEST 1268 */

/*
 *  BEGIN TEST 1269
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1269( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(57, -105), MAKE_VAL_COMPLEX(-49, -25), MAKE_VAL_COMPLEX(-45, -5), MAKE_VAL_COMPLEX(55, 20), MAKE_VAL_COMPLEX(-68, 148), MAKE_VAL_COMPLEX(43, -214), MAKE_VAL_COMPLEX(86, -76), MAKE_VAL_COMPLEX(-38, 62), MAKE_VAL_COMPLEX(-78, 58), MAKE_VAL_COMPLEX(42, 93), MAKE_VAL_COMPLEX(2, 82), MAKE_VAL_COMPLEX(-37, -101), MAKE_VAL_COMPLEX(135, 79), MAKE_VAL_COMPLEX(16, 176), MAKE_VAL_COMPLEX(-258, -34) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1269, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1269", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1269() */
/* END TEST 1269 */

/*
 *  BEGIN TEST 1270
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1270( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-90, -10), MAKE_VAL_COMPLEX(-46, 86), MAKE_VAL_COMPLEX(14, -44), MAKE_VAL_COMPLEX(-26, -12), MAKE_VAL_COMPLEX(-19, 89), MAKE_VAL_COMPLEX(-7, -19), MAKE_VAL_COMPLEX(-86, -46), MAKE_VAL_COMPLEX(84, 64), MAKE_VAL_COMPLEX(90, 10), MAKE_VAL_COMPLEX(73, -83), MAKE_VAL_COMPLEX(-96, 44), MAKE_VAL_COMPLEX(67, 53), MAKE_VAL_COMPLEX(84, -18), MAKE_VAL_COMPLEX(92, -8), MAKE_VAL_COMPLEX(-45, -5) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1270, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1270", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1270() */
/* END TEST 1270 */

/*
 *  BEGIN TEST 1271
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_ROWMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1271( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_ROWMAJ, x1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-27, 405), MAKE_VAL_COMPLEX(-175, -89), MAKE_VAL_COMPLEX(571, -249), MAKE_VAL_COMPLEX(-442, -19), MAKE_VAL_COMPLEX(-863, -1069), MAKE_VAL_COMPLEX(893, 1406), MAKE_VAL_COMPLEX(-142, 292), MAKE_VAL_COMPLEX(-178, 74), MAKE_VAL_COMPLEX(388, -228), MAKE_VAL_COMPLEX(-60, -201), MAKE_VAL_COMPLEX(-938, -866), MAKE_VAL_COMPLEX(640, 934), MAKE_VAL_COMPLEX(-863, 597), MAKE_VAL_COMPLEX(-664, -508), MAKE_VAL_COMPLEX(1613, -1341) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1271, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1271", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1271() */
/* END TEST 1271 */

/*
 *  BEGIN TEST 1272
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1272( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	oski_value_t* y_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-55, 42), MAKE_VAL_COMPLEX(-45, 51), MAKE_VAL_COMPLEX(75, -20), MAKE_VAL_COMPLEX(80, -41), MAKE_VAL_COMPLEX(-187, -20), MAKE_VAL_COMPLEX(51, 118), MAKE_VAL_COMPLEX(-9, 58), MAKE_VAL_COMPLEX(-45, -50), MAKE_VAL_COMPLEX(-7, 1), MAKE_VAL_COMPLEX(111, -166), MAKE_VAL_COMPLEX(-136, 71), MAKE_VAL_COMPLEX(79, -18), MAKE_VAL_COMPLEX(-5, 192), MAKE_VAL_COMPLEX(-20, -36), MAKE_VAL_COMPLEX(178, -75) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1272, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1272", y_true, y_val, y1248_LEN*y1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1272() */
/* END TEST 1272 */

/*
 *  BEGIN TEST 1273
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1273( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	oski_value_t* y_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(41, -22), MAKE_VAL_COMPLEX(65, -53), MAKE_VAL_COMPLEX(-75, 6), MAKE_VAL_COMPLEX(-86, 61), MAKE_VAL_COMPLEX(185, 16), MAKE_VAL_COMPLEX(-31, -102), MAKE_VAL_COMPLEX(11, -54), MAKE_VAL_COMPLEX(25, 38), MAKE_VAL_COMPLEX(-5, 11), MAKE_VAL_COMPLEX(-131, 146), MAKE_VAL_COMPLEX(140, -71), MAKE_VAL_COMPLEX(-97, 0), MAKE_VAL_COMPLEX(17, -180), MAKE_VAL_COMPLEX(10, 28), MAKE_VAL_COMPLEX(-166, 95) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1273, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1273", y_true, y_val, y1248_LEN*y1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1273() */
/* END TEST 1273 */

/*
 *  BEGIN TEST 1274
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1274( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	oski_value_t* y_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-16, -41), MAKE_VAL_COMPLEX(-17, 32), MAKE_VAL_COMPLEX(21, 14), MAKE_VAL_COMPLEX(-24, -29), MAKE_VAL_COMPLEX(8, 1), MAKE_VAL_COMPLEX(-44, 14), MAKE_VAL_COMPLEX(-8, -1), MAKE_VAL_COMPLEX(38, -18), MAKE_VAL_COMPLEX(-6, -30), MAKE_VAL_COMPLEX(50, -10), MAKE_VAL_COMPLEX(-4, 6), MAKE_VAL_COMPLEX(45, -9), MAKE_VAL_COMPLEX(-30, 6), MAKE_VAL_COMPLEX(22, -7), MAKE_VAL_COMPLEX(-42, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1274, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1274", y_true, y_val, y1248_LEN*y1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1274() */
/* END TEST 1274 */

/*
 *  BEGIN TEST 1275
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1275( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	oski_value_t* y_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-473, 35), MAKE_VAL_COMPLEX(102, -468), MAKE_VAL_COMPLEX(23, 251), MAKE_VAL_COMPLEX(141, 1039), MAKE_VAL_COMPLEX(273, -1367), MAKE_VAL_COMPLEX(-409, 515), MAKE_VAL_COMPLEX(-343, -43), MAKE_VAL_COMPLEX(-61, -457), MAKE_VAL_COMPLEX(-216, 238), MAKE_VAL_COMPLEX(571, 583), MAKE_VAL_COMPLEX(-277, -1035), MAKE_VAL_COMPLEX(-385, 247), MAKE_VAL_COMPLEX(-1051, -11), MAKE_VAL_COMPLEX(38, -242), MAKE_VAL_COMPLEX(659, 1705) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1275, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1275", y_true, y_val, y1248_LEN*y1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1275() */
/* END TEST 1275 */

/*
 *  BEGIN TEST 1276
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1276( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(4, -24), MAKE_VAL_COMPLEX(-167, -142), MAKE_VAL_COMPLEX(-6, 70), MAKE_VAL_COMPLEX(-93, -25), MAKE_VAL_COMPLEX(-220, -153), MAKE_VAL_COMPLEX(28, -16), MAKE_VAL_COMPLEX(161, -24), MAKE_VAL_COMPLEX(-78, 45), MAKE_VAL_COMPLEX(38, 73), MAKE_VAL_COMPLEX(68, 78), MAKE_VAL_COMPLEX(23, -30), MAKE_VAL_COMPLEX(-186, -15), MAKE_VAL_COMPLEX(15, 126), MAKE_VAL_COMPLEX(-165, 29), MAKE_VAL_COMPLEX(23, -14) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1276, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1276", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1276() */
/* END TEST 1276 */

/*
 *  BEGIN TEST 1277
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1277( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-14, -126), MAKE_VAL_COMPLEX(13, 92), MAKE_VAL_COMPLEX(82, -60), MAKE_VAL_COMPLEX(105, -21), MAKE_VAL_COMPLEX(69, 148), MAKE_VAL_COMPLEX(1, -1), MAKE_VAL_COMPLEX(-111, -130), MAKE_VAL_COMPLEX(-2, 111), MAKE_VAL_COMPLEX(-28, 77), MAKE_VAL_COMPLEX(89, 17), MAKE_VAL_COMPLEX(-123, -114), MAKE_VAL_COMPLEX(119, 140), MAKE_VAL_COMPLEX(39, 6), MAKE_VAL_COMPLEX(205, 119), MAKE_VAL_COMPLEX(-28, -61) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1277, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1277", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1277() */
/* END TEST 1277 */

/*
 *  BEGIN TEST 1278
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1278( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-40, -10), MAKE_VAL_COMPLEX(-26, 36), MAKE_VAL_COMPLEX(9, -19), MAKE_VAL_COMPLEX(-11, -7), MAKE_VAL_COMPLEX(-14, 39), MAKE_VAL_COMPLEX(-2, -9), MAKE_VAL_COMPLEX(-36, -26), MAKE_VAL_COMPLEX(34, 34), MAKE_VAL_COMPLEX(40, 10), MAKE_VAL_COMPLEX(38, -33), MAKE_VAL_COMPLEX(-46, 14), MAKE_VAL_COMPLEX(27, 28), MAKE_VAL_COMPLEX(39, -3), MAKE_VAL_COMPLEX(42, 2), MAKE_VAL_COMPLEX(-20, -5) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1278, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1278", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1278() */
/* END TEST 1278 */

/*
 *  BEGIN TEST 1279
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1279( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-110, -460), MAKE_VAL_COMPLEX(861, -155), MAKE_VAL_COMPLEX(1, -425), MAKE_VAL_COMPLEX(587, -541), MAKE_VAL_COMPLEX(1244, -185), MAKE_VAL_COMPLEX(39, 118), MAKE_VAL_COMPLEX(-655, 241), MAKE_VAL_COMPLEX(15, 111), MAKE_VAL_COMPLEX(-345, 425), MAKE_VAL_COMPLEX(17, 160), MAKE_VAL_COMPLEX(-535, -209), MAKE_VAL_COMPLEX(873, -270), MAKE_VAL_COMPLEX(-336, -78), MAKE_VAL_COMPLEX(1010, -412), MAKE_VAL_COMPLEX(-150, -115) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1279, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1279", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1279() */
/* END TEST 1279 */

/*
 *  BEGIN TEST 1280
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1280( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(26, -20), MAKE_VAL_COMPLEX(-13, 144), MAKE_VAL_COMPLEX(-17, -45), MAKE_VAL_COMPLEX(14, 124), MAKE_VAL_COMPLEX(139, -133), MAKE_VAL_COMPLEX(-3, 0), MAKE_VAL_COMPLEX(31, -66), MAKE_VAL_COMPLEX(-20, -31), MAKE_VAL_COMPLEX(-52, -107), MAKE_VAL_COMPLEX(-45, 38), MAKE_VAL_COMPLEX(-39, 8), MAKE_VAL_COMPLEX(-56, 88), MAKE_VAL_COMPLEX(-86, -17), MAKE_VAL_COMPLEX(-55, 127), MAKE_VAL_COMPLEX(-2, 56) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1280, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1280", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1280() */
/* END TEST 1280 */

/*
 *  BEGIN TEST 1281
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1281( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-56, 130), MAKE_VAL_COMPLEX(111, -70), MAKE_VAL_COMPLEX(-35, 19), MAKE_VAL_COMPLEX(-34, -94), MAKE_VAL_COMPLEX(-32, 174), MAKE_VAL_COMPLEX(-22, 5), MAKE_VAL_COMPLEX(-105, 164), MAKE_VAL_COMPLEX(116, -61), MAKE_VAL_COMPLEX(82, -3), MAKE_VAL_COMPLEX(-44, -145), MAKE_VAL_COMPLEX(75, 120), MAKE_VAL_COMPLEX(135, -161), MAKE_VAL_COMPLEX(80, -91), MAKE_VAL_COMPLEX(63, -243), MAKE_VAL_COMPLEX(-13, -1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1281, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1281", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1281() */
/* END TEST 1281 */

/*
 *  BEGIN TEST 1282
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1282( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(50, 0), MAKE_VAL_COMPLEX(20, -50), MAKE_VAL_COMPLEX(-5, 25), MAKE_VAL_COMPLEX(15, 5), MAKE_VAL_COMPLEX(5, -50), MAKE_VAL_COMPLEX(5, 10), MAKE_VAL_COMPLEX(50, 20), MAKE_VAL_COMPLEX(-50, -30), MAKE_VAL_COMPLEX(-50, 0), MAKE_VAL_COMPLEX(-35, 50), MAKE_VAL_COMPLEX(50, -30), MAKE_VAL_COMPLEX(-40, -25), MAKE_VAL_COMPLEX(-45, 15), MAKE_VAL_COMPLEX(-50, 10), MAKE_VAL_COMPLEX(25, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1282, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1282", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1282() */
/* END TEST 1282 */

/*
 *  BEGIN TEST 1283
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1283( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_ROWMAJ, y1248_VW );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(452, 408), MAKE_VAL_COMPLEX(-808, -1740), MAKE_VAL_COMPLEX(426, 470), MAKE_VAL_COMPLEX(-958, -818), MAKE_VAL_COMPLEX(42, 1680), MAKE_VAL_COMPLEX(66, 80), MAKE_VAL_COMPLEX(764, 1016), MAKE_VAL_COMPLEX(-68, -252), MAKE_VAL_COMPLEX(772, 400), MAKE_VAL_COMPLEX(-334, -192), MAKE_VAL_COMPLEX(772, -636), MAKE_VAL_COMPLEX(-636, -1434), MAKE_VAL_COMPLEX(362, -438), MAKE_VAL_COMPLEX(-1084, -1396), MAKE_VAL_COMPLEX(-182, -444) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1283, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1283", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1283() */
/* END TEST 1283 */

/*
 *  BEGIN TEST 1284
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1284( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	oski_value_t* y_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(118, 29), MAKE_VAL_COMPLEX(60, -99), MAKE_VAL_COMPLEX(-43, 98), MAKE_VAL_COMPLEX(162, -148), MAKE_VAL_COMPLEX(8, 170), MAKE_VAL_COMPLEX(-92, -113), MAKE_VAL_COMPLEX(-190, -31), MAKE_VAL_COMPLEX(-21, 114), MAKE_VAL_COMPLEX(-62, 55), MAKE_VAL_COMPLEX(-20, 127), MAKE_VAL_COMPLEX(26, -18), MAKE_VAL_COMPLEX(80, 169), MAKE_VAL_COMPLEX(29, -59), MAKE_VAL_COMPLEX(63, 103), MAKE_VAL_COMPLEX(234, -231) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1284, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1284", y_true, y_val, y1248_LEN*y1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1284() */
/* END TEST 1284 */

/*
 *  BEGIN TEST 1285
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1285( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	oski_value_t* y_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(12, -87), MAKE_VAL_COMPLEX(-166, 49), MAKE_VAL_COMPLEX(1, -28), MAKE_VAL_COMPLEX(-72, 66), MAKE_VAL_COMPLEX(-10, -144), MAKE_VAL_COMPLEX(40, -27), MAKE_VAL_COMPLEX(192, 5), MAKE_VAL_COMPLEX(85, 6), MAKE_VAL_COMPLEX(158, -79), MAKE_VAL_COMPLEX(60, 33), MAKE_VAL_COMPLEX(-46, 6), MAKE_VAL_COMPLEX(-44, -25), MAKE_VAL_COMPLEX(-53, -37), MAKE_VAL_COMPLEX(-37, -33), MAKE_VAL_COMPLEX(-234, 95) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1285, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1285", y_true, y_val, y1248_LEN*y1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1285() */
/* END TEST 1285 */

/*
 *  BEGIN TEST 1286
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1286( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	oski_value_t* y_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(23, 31), MAKE_VAL_COMPLEX(7, -31), MAKE_VAL_COMPLEX(-21, -7), MAKE_VAL_COMPLEX(27, 19), MAKE_VAL_COMPLEX(-7, 1), MAKE_VAL_COMPLEX(34, -22), MAKE_VAL_COMPLEX(7, -1), MAKE_VAL_COMPLEX(-28, 24), MAKE_VAL_COMPLEX(12, 24), MAKE_VAL_COMPLEX(-40, 20), MAKE_VAL_COMPLEX(2, -6), MAKE_VAL_COMPLEX(-36, 18), MAKE_VAL_COMPLEX(24, -12), MAKE_VAL_COMPLEX(-17, 11), MAKE_VAL_COMPLEX(36, -8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1286, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1286", y_true, y_val, y1248_LEN*y1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1286() */
/* END TEST 1286 */

/*
 *  BEGIN TEST 1287
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1287( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	oski_value_t* y_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-289, -491), MAKE_VAL_COMPLEX(-964, 838), MAKE_VAL_COMPLEX(-97, -381), MAKE_VAL_COMPLEX(-414, 973), MAKE_VAL_COMPLEX(-540, -1163), MAKE_VAL_COMPLEX(483, -190), MAKE_VAL_COMPLEX(1579, -495), MAKE_VAL_COMPLEX(460, -309), MAKE_VAL_COMPLEX(913, -956), MAKE_VAL_COMPLEX(329, -106), MAKE_VAL_COMPLEX(-306, 180), MAKE_VAL_COMPLEX(-652, -239), MAKE_VAL_COMPLEX(-385, -23), MAKE_VAL_COMPLEX(-517, -226), MAKE_VAL_COMPLEX(-1425, 1664) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1287, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1287", y_true, y_val, y1248_LEN*y1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1287() */
/* END TEST 1287 */

/*
 *  BEGIN TEST 1288
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1288( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(13, -21), MAKE_VAL_COMPLEX(-32, -17), MAKE_VAL_COMPLEX(63, 7), MAKE_VAL_COMPLEX(25, -7), MAKE_VAL_COMPLEX(-57, 214), MAKE_VAL_COMPLEX(24, -7), MAKE_VAL_COMPLEX(154, -87), MAKE_VAL_COMPLEX(18, 118), MAKE_VAL_COMPLEX(226, 96), MAKE_VAL_COMPLEX(-27, 50), MAKE_VAL_COMPLEX(-110, -110), MAKE_VAL_COMPLEX(-205, 36), MAKE_VAL_COMPLEX(53, -21), MAKE_VAL_COMPLEX(-72, -3), MAKE_VAL_COMPLEX(-37, -291) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1288, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1288", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1288() */
/* END TEST 1288 */

/*
 *  BEGIN TEST 1289
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1289( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-153, -59), MAKE_VAL_COMPLEX(-104, 125), MAKE_VAL_COMPLEX(-9, -69), MAKE_VAL_COMPLEX(-59, -31), MAKE_VAL_COMPLEX(-37, -82), MAKE_VAL_COMPLEX(-22, -29), MAKE_VAL_COMPLEX(-262, -49), MAKE_VAL_COMPLEX(74, 46), MAKE_VAL_COMPLEX(-86, -16), MAKE_VAL_COMPLEX(205, -134), MAKE_VAL_COMPLEX(-78, 114), MAKE_VAL_COMPLEX(277, 98), MAKE_VAL_COMPLEX(97, 51), MAKE_VAL_COMPLEX(228, 55), MAKE_VAL_COMPLEX(-33, 251) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1289, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1289", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1289() */
/* END TEST 1289 */

/*
 *  BEGIN TEST 1290
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1290( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-10, -90), MAKE_VAL_COMPLEX(-94, -26), MAKE_VAL_COMPLEX(46, 4), MAKE_VAL_COMPLEX(6, -28), MAKE_VAL_COMPLEX(-91, 1), MAKE_VAL_COMPLEX(17, -11), MAKE_VAL_COMPLEX(26, -94), MAKE_VAL_COMPLEX(-44, 96), MAKE_VAL_COMPLEX(10, 90), MAKE_VAL_COMPLEX(97, 53), MAKE_VAL_COMPLEX(-64, -84), MAKE_VAL_COMPLEX(-37, 77), MAKE_VAL_COMPLEX(36, 78), MAKE_VAL_COMPLEX(28, 88), MAKE_VAL_COMPLEX(-5, -45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1290, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1290", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1290() */
/* END TEST 1290 */

/*
 *  BEGIN TEST 1291
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1291( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-59, -677), MAKE_VAL_COMPLEX(-605, -116), MAKE_VAL_COMPLEX(118, -332), MAKE_VAL_COMPLEX(-40, -334), MAKE_VAL_COMPLEX(1070, -664), MAKE_VAL_COMPLEX(3, -179), MAKE_VAL_COMPLEX(-817, -1344), MAKE_VAL_COMPLEX(212, -12), MAKE_VAL_COMPLEX(-372, -1296), MAKE_VAL_COMPLEX(1030, 598), MAKE_VAL_COMPLEX(-592, 456), MAKE_VAL_COMPLEX(625, 1757), MAKE_VAL_COMPLEX(-284, 358), MAKE_VAL_COMPLEX(261, 1214), MAKE_VAL_COMPLEX(-1819, 1088) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1291, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1291", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1291() */
/* END TEST 1291 */

/*
 *  BEGIN TEST 1292
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1292( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-17, 105), MAKE_VAL_COMPLEX(36, -31), MAKE_VAL_COMPLEX(-75, 17), MAKE_VAL_COMPLEX(13, -1), MAKE_VAL_COMPLEX(-60, -23), MAKE_VAL_COMPLEX(-25, -12), MAKE_VAL_COMPLEX(-8, -97), MAKE_VAL_COMPLEX(-2, -86), MAKE_VAL_COMPLEX(-44, -234), MAKE_VAL_COMPLEX(-106, -131), MAKE_VAL_COMPLEX(140, 80), MAKE_VAL_COMPLEX(-8, 145), MAKE_VAL_COMPLEX(19, -45), MAKE_VAL_COMPLEX(-48, 81), MAKE_VAL_COMPLEX(278, 34) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1292, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1292", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1292() */
/* END TEST 1292 */

/*
 *  BEGIN TEST 1293
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1293( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-53, -25), MAKE_VAL_COMPLEX(16, 133), MAKE_VAL_COMPLEX(42, -60), MAKE_VAL_COMPLEX(-42, 18), MAKE_VAL_COMPLEX(133, 101), MAKE_VAL_COMPLEX(2, 6), MAKE_VAL_COMPLEX(-94, 149), MAKE_VAL_COMPLEX(120, 48), MAKE_VAL_COMPLEX(114, 154), MAKE_VAL_COMPLEX(75, 5), MAKE_VAL_COMPLEX(-162, 42), MAKE_VAL_COMPLEX(104, -174), MAKE_VAL_COMPLEX(20, -48), MAKE_VAL_COMPLEX(102, -175), MAKE_VAL_COMPLEX(-313, 6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1293, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1293", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1293() */
/* END TEST 1293 */

/*
 *  BEGIN TEST 1294
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1294( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-10, -70), MAKE_VAL_COMPLEX(-74, -18), MAKE_VAL_COMPLEX(36, 2), MAKE_VAL_COMPLEX(4, -22), MAKE_VAL_COMPLEX(-71, 3), MAKE_VAL_COMPLEX(13, -9), MAKE_VAL_COMPLEX(18, -74), MAKE_VAL_COMPLEX(-32, 76), MAKE_VAL_COMPLEX(10, 70), MAKE_VAL_COMPLEX(77, 39), MAKE_VAL_COMPLEX(-52, -64), MAKE_VAL_COMPLEX(-27, 61), MAKE_VAL_COMPLEX(30, 60), MAKE_VAL_COMPLEX(24, 68), MAKE_VAL_COMPLEX(-5, -35) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1294, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1294", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1294() */
/* END TEST 1294 */

/*
 *  BEGIN TEST 1295
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1248', LAYOUT_COLMAJ>
 *  y : <'y1248', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1295( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1248_ptr, A1248_ind, A1248_val,
		A1248_M, A1248_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1248_val,
		y1248_LEN * y1248_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1248_LEN, y1248_VW, LAYOUT_COLMAJ, y1248_LEN );
	oski_value_t* y_val = array_Duplicate( x1248_val,
		x1248_LEN * x1248_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1248_LEN, x1248_VW, LAYOUT_COLMAJ, x1248_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(100, 414), MAKE_VAL_COMPLEX(-487, 51), MAKE_VAL_COMPLEX(-86, 358), MAKE_VAL_COMPLEX(-84, -220), MAKE_VAL_COMPLEX(-746, 832), MAKE_VAL_COMPLEX(-218, -18), MAKE_VAL_COMPLEX(-1177, -879), MAKE_VAL_COMPLEX(-200, 144), MAKE_VAL_COMPLEX(-1308, -632), MAKE_VAL_COMPLEX(-832, -302), MAKE_VAL_COMPLEX(572, -480), MAKE_VAL_COMPLEX(1300, 1240), MAKE_VAL_COMPLEX(344, -316), MAKE_VAL_COMPLEX(923, 861), MAKE_VAL_COMPLEX(1353, -1891) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1295, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1295", y_true, y_val, x1248_LEN*x1248_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1295() */
/* END TEST 1295 */

#define A1296_M 5    /* # of rows */
#define A1296_N 5    /* # of cols */
#define A1296_NNZ 14  /* # of stored non-zeros */
static oski_index_t A1296_ptr[A1296_M+1] = {0, 2, 4, 7, 11, 14}; /* end row pointers for A1296 */
static oski_index_t A1296_ind[A1296_NNZ] = {2, 3, 3, 4, 0, 3, 4, 0, 1, 2, 4, 1, 2, 3}; /* end column indices pointers for A1296 */
static oski_value_t A1296_val[A1296_NNZ] = {MAKE_VAL_COMPLEX(-2, -9), MAKE_VAL_COMPLEX(6, 5), MAKE_VAL_COMPLEX(-8, -1), MAKE_VAL_COMPLEX(4, -7), MAKE_VAL_COMPLEX(-2, 9), MAKE_VAL_COMPLEX(5, -9), MAKE_VAL_COMPLEX(7, -4), MAKE_VAL_COMPLEX(6, -5), MAKE_VAL_COMPLEX(-8, 1), MAKE_VAL_COMPLEX(5, 9), MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(4, 7), MAKE_VAL_COMPLEX(7, 4), MAKE_VAL_COMPLEX(-10, 10)}; /* end stored non-zero values for A1296 */

#define x1296_LEN 5   /* vector length */
#define x1296_VW  3   /* vector width */
static oski_value_t x1296_val[x1296_LEN*x1296_VW] = {MAKE_VAL_COMPLEX(3, 10), MAKE_VAL_COMPLEX(-8, -2), MAKE_VAL_COMPLEX(-2, 10), MAKE_VAL_COMPLEX(10, 3), MAKE_VAL_COMPLEX(5, 3), MAKE_VAL_COMPLEX(-5, 5), MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(1, -2), MAKE_VAL_COMPLEX(4, -4), MAKE_VAL_COMPLEX(6, 9), MAKE_VAL_COMPLEX(1, 8), MAKE_VAL_COMPLEX(6, 6), MAKE_VAL_COMPLEX(-10, -5), MAKE_VAL_COMPLEX(9, -5), MAKE_VAL_COMPLEX(8, -5)}; /* data */

#define y1296_LEN 5   /* vector length */
#define y1296_VW  3   /* vector width */
static oski_value_t y1296_val[y1296_LEN*y1296_VW] = {MAKE_VAL_COMPLEX(-3, 10), MAKE_VAL_COMPLEX(3, -8), MAKE_VAL_COMPLEX(-10, 2), MAKE_VAL_COMPLEX(4, 10), MAKE_VAL_COMPLEX(-7, -8), MAKE_VAL_COMPLEX(10, 9), MAKE_VAL_COMPLEX(-7, 3), MAKE_VAL_COMPLEX(5, -3), MAKE_VAL_COMPLEX(3, -7), MAKE_VAL_COMPLEX(-9, -7), MAKE_VAL_COMPLEX(-6, 5), MAKE_VAL_COMPLEX(2, -10), MAKE_VAL_COMPLEX(8, 1), MAKE_VAL_COMPLEX(-7, 1), MAKE_VAL_COMPLEX(-3, -10)}; /* data */

/*
 *  BEGIN TEST 1296
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1296( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	oski_value_t* y_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-25, 252), MAKE_VAL_COMPLEX(-101, 4), MAKE_VAL_COMPLEX(-58, 114), MAKE_VAL_COMPLEX(-32, -19), MAKE_VAL_COMPLEX(-63, -127), MAKE_VAL_COMPLEX(34, -158), MAKE_VAL_COMPLEX(-88, 44), MAKE_VAL_COMPLEX(152, -149), MAKE_VAL_COMPLEX(5, -172), MAKE_VAL_COMPLEX(12, 83), MAKE_VAL_COMPLEX(-205, 27), MAKE_VAL_COMPLEX(-49, -25), MAKE_VAL_COMPLEX(-141, -108), MAKE_VAL_COMPLEX(-82, 7), MAKE_VAL_COMPLEX(-192, -44) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1296, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1296", y_true, y_val, y1296_LEN*y1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1296() */
/* END TEST 1296 */

/*
 *  BEGIN TEST 1297
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1297( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	oski_value_t* y_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(109, -96), MAKE_VAL_COMPLEX(41, -136), MAKE_VAL_COMPLEX(-38, 30), MAKE_VAL_COMPLEX(200, 91), MAKE_VAL_COMPLEX(-117, 115), MAKE_VAL_COMPLEX(194, 146), MAKE_VAL_COMPLEX(40, 76), MAKE_VAL_COMPLEX(-128, 53), MAKE_VAL_COMPLEX(-53, 52), MAKE_VAL_COMPLEX(-204, -59), MAKE_VAL_COMPLEX(193, 105), MAKE_VAL_COMPLEX(-47, -119), MAKE_VAL_COMPLEX(249, 24), MAKE_VAL_COMPLEX(10, 89), MAKE_VAL_COMPLEX(36, -40) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1297, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1297", y_true, y_val, y1296_LEN*y1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1297() */
/* END TEST 1297 */

/*
 *  BEGIN TEST 1298
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1298( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	oski_value_t* y_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(29, -24), MAKE_VAL_COMPLEX(-25, 18), MAKE_VAL_COMPLEX(34, 14), MAKE_VAL_COMPLEX(8, -38), MAKE_VAL_COMPLEX(5, 38), MAKE_VAL_COMPLEX(-12, -47), MAKE_VAL_COMPLEX(27, 5), MAKE_VAL_COMPLEX(-21, -1), MAKE_VAL_COMPLEX(-23, 15), MAKE_VAL_COMPLEX(13, 39), MAKE_VAL_COMPLEX(28, -3), MAKE_VAL_COMPLEX(-26, 26), MAKE_VAL_COMPLEX(-22, -19), MAKE_VAL_COMPLEX(23, 11), MAKE_VAL_COMPLEX(-11, 36) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1298, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1298", y_true, y_val, y1296_LEN*y1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1298() */
/* END TEST 1298 */

/*
 *  BEGIN TEST 1299
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1299( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	oski_value_t* y_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(117, 353), MAKE_VAL_COMPLEX(-3, 49), MAKE_VAL_COMPLEX(-60, 112), MAKE_VAL_COMPLEX(-91, 145), MAKE_VAL_COMPLEX(-196, -200), MAKE_VAL_COMPLEX(-96, -22), MAKE_VAL_COMPLEX(-138, 106), MAKE_VAL_COMPLEX(77, -291), MAKE_VAL_COMPLEX(-81, -223), MAKE_VAL_COMPLEX(61, -71), MAKE_VAL_COMPLEX(-278, 234), MAKE_VAL_COMPLEX(26, -60), MAKE_VAL_COMPLEX(-177, 107), MAKE_VAL_COMPLEX(-153, 43), MAKE_VAL_COMPLEX(-186, 24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1299, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1299", y_true, y_val, y1296_LEN*y1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1299() */
/* END TEST 1299 */

/*
 *  BEGIN TEST 1300
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1300( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-94, 17), MAKE_VAL_COMPLEX(-45, 81), MAKE_VAL_COMPLEX(-15, 41), MAKE_VAL_COMPLEX(175, 148), MAKE_VAL_COMPLEX(33, -73), MAKE_VAL_COMPLEX(22, 60), MAKE_VAL_COMPLEX(99, -147), MAKE_VAL_COMPLEX(-192, -77), MAKE_VAL_COMPLEX(192, -59), MAKE_VAL_COMPLEX(-164, 171), MAKE_VAL_COMPLEX(157, -40), MAKE_VAL_COMPLEX(-21, -74), MAKE_VAL_COMPLEX(12, 177), MAKE_VAL_COMPLEX(110, -39), MAKE_VAL_COMPLEX(34, -52) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1300, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1300", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1300() */
/* END TEST 1300 */

/*
 *  BEGIN TEST 1301
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1301( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(60, 15), MAKE_VAL_COMPLEX(37, -117), MAKE_VAL_COMPLEX(-29, -29), MAKE_VAL_COMPLEX(-167, -102), MAKE_VAL_COMPLEX(-35, 99), MAKE_VAL_COMPLEX(-52, -70), MAKE_VAL_COMPLEX(-79, 87), MAKE_VAL_COMPLEX(202, 77), MAKE_VAL_COMPLEX(-168, 67), MAKE_VAL_COMPLEX(140, -129), MAKE_VAL_COMPLEX(-187, 60), MAKE_VAL_COMPLEX(9, 110), MAKE_VAL_COMPLEX(-12, -227), MAKE_VAL_COMPLEX(-72, 65), MAKE_VAL_COMPLEX(2, 74) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1301, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1301", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1301() */
/* END TEST 1301 */

/*
 *  BEGIN TEST 1302
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1302( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-89, -6), MAKE_VAL_COMPLEX(40, -58), MAKE_VAL_COMPLEX(-74, -46), MAKE_VAL_COMPLEX(-54, 71), MAKE_VAL_COMPLEX(-39, 31), MAKE_VAL_COMPLEX(-25, -55), MAKE_VAL_COMPLEX(110, -50), MAKE_VAL_COMPLEX(13, 14), MAKE_VAL_COMPLEX(20, 44), MAKE_VAL_COMPLEX(-90, 21), MAKE_VAL_COMPLEX(-67, -16), MAKE_VAL_COMPLEX(-66, 30), MAKE_VAL_COMPLEX(70, -65), MAKE_VAL_COMPLEX(13, 87), MAKE_VAL_COMPLEX(16, 79) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1302, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1302", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1302() */
/* END TEST 1302 */

/*
 *  BEGIN TEST 1303
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1303( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-103, -661), MAKE_VAL_COMPLEX(-402, 518), MAKE_VAL_COMPLEX(232, -44), MAKE_VAL_COMPLEX(1, 813), MAKE_VAL_COMPLEX(432, -434), MAKE_VAL_COMPLEX(168, 418), MAKE_VAL_COMPLEX(804, 448), MAKE_VAL_COMPLEX(-565, -1071), MAKE_VAL_COMPLEX(748, 460), MAKE_VAL_COMPLEX(-1109, -425), MAKE_VAL_COMPLEX(1093, 243), MAKE_VAL_COMPLEX(320, -764), MAKE_VAL_COMPLEX(-635, 1281), MAKE_VAL_COMPLEX(184, 30), MAKE_VAL_COMPLEX(-45, -285) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1303, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1303", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1303() */
/* END TEST 1303 */

/*
 *  BEGIN TEST 1304
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1304( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(75, -124), MAKE_VAL_COMPLEX(-47, 33), MAKE_VAL_COMPLEX(79, -125), MAKE_VAL_COMPLEX(52, -81), MAKE_VAL_COMPLEX(9, -53), MAKE_VAL_COMPLEX(-18, 68), MAKE_VAL_COMPLEX(-139, 137), MAKE_VAL_COMPLEX(17, 162), MAKE_VAL_COMPLEX(-200, -227), MAKE_VAL_COMPLEX(-127, -266), MAKE_VAL_COMPLEX(224, 17), MAKE_VAL_COMPLEX(-129, 16), MAKE_VAL_COMPLEX(93, 162), MAKE_VAL_COMPLEX(-34, -223), MAKE_VAL_COMPLEX(-1, 155) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1304, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1304", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1304() */
/* END TEST 1304 */

/*
 *  BEGIN TEST 1305
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1305( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-133, -33), MAKE_VAL_COMPLEX(177, -9), MAKE_VAL_COMPLEX(-57, -37), MAKE_VAL_COMPLEX(-215, 43), MAKE_VAL_COMPLEX(-92, 10), MAKE_VAL_COMPLEX(93, -153), MAKE_VAL_COMPLEX(309, 13), MAKE_VAL_COMPLEX(-31, -129), MAKE_VAL_COMPLEX(140, 295), MAKE_VAL_COMPLEX(22, 128), MAKE_VAL_COMPLEX(-248, -144), MAKE_VAL_COMPLEX(27, -106), MAKE_VAL_COMPLEX(72, -92), MAKE_VAL_COMPLEX(-105, 312), MAKE_VAL_COMPLEX(-122, -67) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1305, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1305", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1305() */
/* END TEST 1305 */

/*
 *  BEGIN TEST 1306
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1306( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(2, 43), MAKE_VAL_COMPLEX(-30, -16), MAKE_VAL_COMPLEX(-18, 38), MAKE_VAL_COMPLEX(37, 22), MAKE_VAL_COMPLEX(17, 17), MAKE_VAL_COMPLEX(-25, 15), MAKE_VAL_COMPLEX(-30, -50), MAKE_VAL_COMPLEX(6, -7), MAKE_VAL_COMPLEX(20, -12), MAKE_VAL_COMPLEX(15, 42), MAKE_VAL_COMPLEX(-4, 33), MAKE_VAL_COMPLEX(18, 30), MAKE_VAL_COMPLEX(-35, -30), MAKE_VAL_COMPLEX(41, -11), MAKE_VAL_COMPLEX(37, -12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1306, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1306", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1306() */
/* END TEST 1306 */

/*
 *  BEGIN TEST 1307
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1307( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-58, -130), MAKE_VAL_COMPLEX(-327, 11), MAKE_VAL_COMPLEX(-223, -211), MAKE_VAL_COMPLEX(302, -102), MAKE_VAL_COMPLEX(54, -2), MAKE_VAL_COMPLEX(-163, 345), MAKE_VAL_COMPLEX(-353, 199), MAKE_VAL_COMPLEX(438, 396), MAKE_VAL_COMPLEX(-858, -365), MAKE_VAL_COMPLEX(-829, -145), MAKE_VAL_COMPLEX(629, -103), MAKE_VAL_COMPLEX(-229, 600), MAKE_VAL_COMPLEX(323, -15), MAKE_VAL_COMPLEX(-163, -650), MAKE_VAL_COMPLEX(657, 404) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1307, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1307", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1307() */
/* END TEST 1307 */

/*
 *  BEGIN TEST 1308
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1308( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	oski_value_t* y_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-105, 119), MAKE_VAL_COMPLEX(-85, 46), MAKE_VAL_COMPLEX(-25, -71), MAKE_VAL_COMPLEX(3, 0), MAKE_VAL_COMPLEX(-82, -42), MAKE_VAL_COMPLEX(-177, 33), MAKE_VAL_COMPLEX(40, -201), MAKE_VAL_COMPLEX(135, -64), MAKE_VAL_COMPLEX(-203, 40), MAKE_VAL_COMPLEX(31, -34), MAKE_VAL_COMPLEX(-23, -17), MAKE_VAL_COMPLEX(-14, -45), MAKE_VAL_COMPLEX(-23, -100), MAKE_VAL_COMPLEX(49, -25), MAKE_VAL_COMPLEX(-52, 23) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1308, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1308", y_true, y_val, y1296_LEN*y1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1308() */
/* END TEST 1308 */

/*
 *  BEGIN TEST 1309
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1309( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	oski_value_t* y_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(61, -154), MAKE_VAL_COMPLEX(119, -15), MAKE_VAL_COMPLEX(35, 17), MAKE_VAL_COMPLEX(-61, 0), MAKE_VAL_COMPLEX(136, 23), MAKE_VAL_COMPLEX(112, -1), MAKE_VAL_COMPLEX(-41, 160), MAKE_VAL_COMPLEX(-130, 95), MAKE_VAL_COMPLEX(232, -11), MAKE_VAL_COMPLEX(22, 3), MAKE_VAL_COMPLEX(10, -23), MAKE_VAL_COMPLEX(60, 75), MAKE_VAL_COMPLEX(2, 138), MAKE_VAL_COMPLEX(-40, -12), MAKE_VAL_COMPLEX(108, -18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1309, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1309", y_true, y_val, y1296_LEN*y1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1309() */
/* END TEST 1309 */

/*
 *  BEGIN TEST 1310
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1310( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	oski_value_t* y_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-17, -16), MAKE_VAL_COMPLEX(13, 14), MAKE_VAL_COMPLEX(6, -22), MAKE_VAL_COMPLEX(-24, -2), MAKE_VAL_COMPLEX(23, -6), MAKE_VAL_COMPLEX(-28, 11), MAKE_VAL_COMPLEX(1, -17), MAKE_VAL_COMPLEX(1, 13), MAKE_VAL_COMPLEX(11, 13), MAKE_VAL_COMPLEX(23, -11), MAKE_VAL_COMPLEX(-4, -17), MAKE_VAL_COMPLEX(18, 14), MAKE_VAL_COMPLEX(-10, 15), MAKE_VAL_COMPLEX(5, -15), MAKE_VAL_COMPLEX(23, 4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1310, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1310", y_true, y_val, y1296_LEN*y1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1310() */
/* END TEST 1310 */

/*
 *  BEGIN TEST 1311
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1311( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	oski_value_t* y_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1260, 1085), MAKE_VAL_COMPLEX(-702, -564), MAKE_VAL_COMPLEX(-493, -331), MAKE_VAL_COMPLEX(413, 586), MAKE_VAL_COMPLEX(-959, -1028), MAKE_VAL_COMPLEX(-615, -50), MAKE_VAL_COMPLEX(685, -960), MAKE_VAL_COMPLEX(1474, -167), MAKE_VAL_COMPLEX(-1490, -1015), MAKE_VAL_COMPLEX(-312, -429), MAKE_VAL_COMPLEX(-401, 188), MAKE_VAL_COMPLEX(68, -1029), MAKE_VAL_COMPLEX(757, -841), MAKE_VAL_COMPLEX(16, 182), MAKE_VAL_COMPLEX(-785, -635) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1311, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1311", y_true, y_val, y1296_LEN*y1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1311() */
/* END TEST 1311 */

/*
 *  BEGIN TEST 1312
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1312( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(89, -136), MAKE_VAL_COMPLEX(66, 51), MAKE_VAL_COMPLEX(-12, 44), MAKE_VAL_COMPLEX(-79, -229), MAKE_VAL_COMPLEX(-39, -73), MAKE_VAL_COMPLEX(175, -106), MAKE_VAL_COMPLEX(39, 131), MAKE_VAL_COMPLEX(93, -192), MAKE_VAL_COMPLEX(-8, -79), MAKE_VAL_COMPLEX(10, 130), MAKE_VAL_COMPLEX(259, -68), MAKE_VAL_COMPLEX(61, 10), MAKE_VAL_COMPLEX(7, -67), MAKE_VAL_COMPLEX(-185, 57), MAKE_VAL_COMPLEX(-9, -21) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1312, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1312", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1312() */
/* END TEST 1312 */

/*
 *  BEGIN TEST 1313
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1313( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1, 66), MAKE_VAL_COMPLEX(-54, 37), MAKE_VAL_COMPLEX(120, -64), MAKE_VAL_COMPLEX(69, 117), MAKE_VAL_COMPLEX(49, 11), MAKE_VAL_COMPLEX(-105, 136), MAKE_VAL_COMPLEX(-99, 9), MAKE_VAL_COMPLEX(-117, 190), MAKE_VAL_COMPLEX(-48, 55), MAKE_VAL_COMPLEX(56, -226), MAKE_VAL_COMPLEX(-183, 26), MAKE_VAL_COMPLEX(-25, -94), MAKE_VAL_COMPLEX(-17, 187), MAKE_VAL_COMPLEX(99, -127), MAKE_VAL_COMPLEX(-73, -39) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1313, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1313", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1313() */
/* END TEST 1313 */

/*
 *  BEGIN TEST 1314
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1314( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(29, -85), MAKE_VAL_COMPLEX(46, 54), MAKE_VAL_COMPLEX(64, -60), MAKE_VAL_COMPLEX(-55, -71), MAKE_VAL_COMPLEX(-20, -46), MAKE_VAL_COMPLEX(60, -10), MAKE_VAL_COMPLEX(20, 120), MAKE_VAL_COMPLEX(-17, 9), MAKE_VAL_COMPLEX(-48, 8), MAKE_VAL_COMPLEX(3, -93), MAKE_VAL_COMPLEX(33, -61), MAKE_VAL_COMPLEX(-12, -72), MAKE_VAL_COMPLEX(45, 85), MAKE_VAL_COMPLEX(-88, -10), MAKE_VAL_COMPLEX(-81, -5) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1314, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1314", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1314() */
/* END TEST 1314 */

/*
 *  BEGIN TEST 1315
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1315( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-303, -177), MAKE_VAL_COMPLEX(-34, -44), MAKE_VAL_COMPLEX(604, 72), MAKE_VAL_COMPLEX(-605, 373), MAKE_VAL_COMPLEX(-76, 106), MAKE_VAL_COMPLEX(-584, -268), MAKE_VAL_COMPLEX(6, -38), MAKE_VAL_COMPLEX(-1033, -35), MAKE_VAL_COMPLEX(-452, 14), MAKE_VAL_COMPLEX(887, -519), MAKE_VAL_COMPLEX(-459, -917), MAKE_VAL_COMPLEX(182, -492), MAKE_VAL_COMPLEX(-517, 501), MAKE_VAL_COMPLEX(420, 242), MAKE_VAL_COMPLEX(-247, -265) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-2, -4);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1315, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1315", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1315() */
/* END TEST 1315 */

/*
 *  BEGIN TEST 1316
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1316( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-15, 96), MAKE_VAL_COMPLEX(90, -41), MAKE_VAL_COMPLEX(-44, -178), MAKE_VAL_COMPLEX(-183, -99), MAKE_VAL_COMPLEX(-163, 67), MAKE_VAL_COMPLEX(17, -70), MAKE_VAL_COMPLEX(15, 21), MAKE_VAL_COMPLEX(-243, 10), MAKE_VAL_COMPLEX(-144, -21), MAKE_VAL_COMPLEX(-106, 150), MAKE_VAL_COMPLEX(225, 92), MAKE_VAL_COMPLEX(-111, 302), MAKE_VAL_COMPLEX(-77, -65), MAKE_VAL_COMPLEX(-43, 95), MAKE_VAL_COMPLEX(123, -37) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1316, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1316", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1316() */
/* END TEST 1316 */

/*
 *  BEGIN TEST 1317
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1317( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-95, -172), MAKE_VAL_COMPLEX(6, -3), MAKE_VAL_COMPLEX(-16, 62), MAKE_VAL_COMPLEX(59, 149), MAKE_VAL_COMPLEX(89, -57), MAKE_VAL_COMPLEX(-7, -20), MAKE_VAL_COMPLEX(165, -1), MAKE_VAL_COMPLEX(249, 18), MAKE_VAL_COMPLEX(136, 93), MAKE_VAL_COMPLEX(-26, -192), MAKE_VAL_COMPLEX(-299, -164), MAKE_VAL_COMPLEX(3, -314), MAKE_VAL_COMPLEX(217, 35), MAKE_VAL_COMPLEX(-7, 27), MAKE_VAL_COMPLEX(-163, 151) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1317, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1317", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1317() */
/* END TEST 1317 */

/*
 *  BEGIN TEST 1318
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1318( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(53, -5), MAKE_VAL_COMPLEX(-18, 38), MAKE_VAL_COMPLEX(48, 20), MAKE_VAL_COMPLEX(25, -47), MAKE_VAL_COMPLEX(20, -22), MAKE_VAL_COMPLEX(20, 30), MAKE_VAL_COMPLEX(-60, 40), MAKE_VAL_COMPLEX(-9, -7), MAKE_VAL_COMPLEX(-16, -24), MAKE_VAL_COMPLEX(51, -21), MAKE_VAL_COMPLEX(41, 3), MAKE_VAL_COMPLEX(36, -24), MAKE_VAL_COMPLEX(-35, 45), MAKE_VAL_COMPLEX(-16, -50), MAKE_VAL_COMPLEX(-17, -45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1318, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1318", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1318() */
/* END TEST 1318 */

/*
 *  BEGIN TEST 1319
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_ROWMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1319( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_ROWMAJ, x1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-231, 751), MAKE_VAL_COMPLEX(164, -130), MAKE_VAL_COMPLEX(16, -448), MAKE_VAL_COMPLEX(-191, -561), MAKE_VAL_COMPLEX(-628, 98), MAKE_VAL_COMPLEX(-22, -106), MAKE_VAL_COMPLEX(-262, -346), MAKE_VAL_COMPLEX(-937, -511), MAKE_VAL_COMPLEX(-350, -484), MAKE_VAL_COMPLEX(-583, 775), MAKE_VAL_COMPLEX(681, 1123), MAKE_VAL_COMPLEX(-880, 1262), MAKE_VAL_COMPLEX(-503, -689), MAKE_VAL_COMPLEX(16, 190), MAKE_VAL_COMPLEX(907, -15) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1319, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1319", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1319() */
/* END TEST 1319 */

/*
 *  BEGIN TEST 1320
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1320( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	oski_value_t* y_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(55, 148), MAKE_VAL_COMPLEX(94, -58), MAKE_VAL_COMPLEX(-47, 81), MAKE_VAL_COMPLEX(-68, 55), MAKE_VAL_COMPLEX(26, -102), MAKE_VAL_COMPLEX(-38, 102), MAKE_VAL_COMPLEX(-55, -72), MAKE_VAL_COMPLEX(91, -71), MAKE_VAL_COMPLEX(21, -230), MAKE_VAL_COMPLEX(-15, -109), MAKE_VAL_COMPLEX(58, -21), MAKE_VAL_COMPLEX(-56, -227), MAKE_VAL_COMPLEX(-131, 128), MAKE_VAL_COMPLEX(-18, -64), MAKE_VAL_COMPLEX(-67, 18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1320, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1320", y_true, y_val, y1296_LEN*y1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1320() */
/* END TEST 1320 */

/*
 *  BEGIN TEST 1321
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1321( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	oski_value_t* y_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-116, -90), MAKE_VAL_COMPLEX(-41, 14), MAKE_VAL_COMPLEX(-31, -107), MAKE_VAL_COMPLEX(56, 31), MAKE_VAL_COMPLEX(-43, 18), MAKE_VAL_COMPLEX(72, 1), MAKE_VAL_COMPLEX(-6, 65), MAKE_VAL_COMPLEX(-44, 70), MAKE_VAL_COMPLEX(28, 193), MAKE_VAL_COMPLEX(-20, 24), MAKE_VAL_COMPLEX(-120, 32), MAKE_VAL_COMPLEX(110, 165), MAKE_VAL_COMPLEX(183, -89), MAKE_VAL_COMPLEX(-35, 43), MAKE_VAL_COMPLEX(86, -100) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1321, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1321", y_true, y_val, y1296_LEN*y1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1321() */
/* END TEST 1321 */

/*
 *  BEGIN TEST 1322
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1322( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	oski_value_t* y_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-104, 56), MAKE_VAL_COMPLEX(88, -40), MAKE_VAL_COMPLEX(-96, -64), MAKE_VAL_COMPLEX(-48, 112), MAKE_VAL_COMPLEX(8, -120), MAKE_VAL_COMPLEX(8, 152), MAKE_VAL_COMPLEX(-80, -32), MAKE_VAL_COMPLEX(64, 16), MAKE_VAL_COMPLEX(80, -32), MAKE_VAL_COMPLEX(-16, -128), MAKE_VAL_COMPLEX(-88, -8), MAKE_VAL_COMPLEX(96, -64), MAKE_VAL_COMPLEX(56, 72), MAKE_VAL_COMPLEX(-64, -48), MAKE_VAL_COMPLEX(56, -104) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1322, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1322", y_true, y_val, y1296_LEN*y1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1322() */
/* END TEST 1322 */

/*
 *  BEGIN TEST 1323
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1323( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	oski_value_t* y_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-81, 558), MAKE_VAL_COMPLEX(30, 7), MAKE_VAL_COMPLEX(-360, 318), MAKE_VAL_COMPLEX(173, -149), MAKE_VAL_COMPLEX(-46, 109), MAKE_VAL_COMPLEX(111, -236), MAKE_VAL_COMPLEX(168, 50), MAKE_VAL_COMPLEX(269, -15), MAKE_VAL_COMPLEX(508, -370), MAKE_VAL_COMPLEX(-59, 63), MAKE_VAL_COMPLEX(212, 433), MAKE_VAL_COMPLEX(303, -605), MAKE_VAL_COMPLEX(-368, -531), MAKE_VAL_COMPLEX(111, 141), MAKE_VAL_COMPLEX(-513, -218) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1323, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1323", y_true, y_val, y1296_LEN*y1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1323() */
/* END TEST 1323 */

/*
 *  BEGIN TEST 1324
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1324( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-98, 40), MAKE_VAL_COMPLEX(40, 83), MAKE_VAL_COMPLEX(121, 19), MAKE_VAL_COMPLEX(-113, 149), MAKE_VAL_COMPLEX(116, 259), MAKE_VAL_COMPLEX(-46, 138), MAKE_VAL_COMPLEX(-69, -209), MAKE_VAL_COMPLEX(-188, -80), MAKE_VAL_COMPLEX(202, -125), MAKE_VAL_COMPLEX(78, 80), MAKE_VAL_COMPLEX(2, 47), MAKE_VAL_COMPLEX(104, 96), MAKE_VAL_COMPLEX(80, -100), MAKE_VAL_COMPLEX(34, -149), MAKE_VAL_COMPLEX(55, -54) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1324, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1324", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1324() */
/* END TEST 1324 */

/*
 *  BEGIN TEST 1325
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1325( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(49, 160), MAKE_VAL_COMPLEX(-156, -197), MAKE_VAL_COMPLEX(-255, 131), MAKE_VAL_COMPLEX(253, 2), MAKE_VAL_COMPLEX(-61, -158), MAKE_VAL_COMPLEX(-89, -103), MAKE_VAL_COMPLEX(-1, -61), MAKE_VAL_COMPLEX(225, 56), MAKE_VAL_COMPLEX(-94, 97), MAKE_VAL_COMPLEX(-66, 133), MAKE_VAL_COMPLEX(-65, 99), MAKE_VAL_COMPLEX(-62, 66), MAKE_VAL_COMPLEX(-200, -85), MAKE_VAL_COMPLEX(169, 154), MAKE_VAL_COMPLEX(131, 49) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1325, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1325", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1325() */
/* END TEST 1325 */

/*
 *  BEGIN TEST 1326
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1326( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(117, 63), MAKE_VAL_COMPLEX(-90, 54), MAKE_VAL_COMPLEX(72, 108), MAKE_VAL_COMPLEX(117, -63), MAKE_VAL_COMPLEX(72, -18), MAKE_VAL_COMPLEX(0, 90), MAKE_VAL_COMPLEX(-180, 0), MAKE_VAL_COMPLEX(-9, -27), MAKE_VAL_COMPLEX(0, -72), MAKE_VAL_COMPLEX(135, 27), MAKE_VAL_COMPLEX(81, 63), MAKE_VAL_COMPLEX(108, 0), MAKE_VAL_COMPLEX(-135, 45), MAKE_VAL_COMPLEX(36, -126), MAKE_VAL_COMPLEX(27, -117) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1326, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1326", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1326() */
/* END TEST 1326 */

/*
 *  BEGIN TEST 1327
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1327( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(428, -977), MAKE_VAL_COMPLEX(-973, 1099), MAKE_VAL_COMPLEX(776, 1398), MAKE_VAL_COMPLEX(-1079, -1682), MAKE_VAL_COMPLEX(-1900, 914), MAKE_VAL_COMPLEX(-923, 209), MAKE_VAL_COMPLEX(842, -100), MAKE_VAL_COMPLEX(374, -1881), MAKE_VAL_COMPLEX(1035, 1305), MAKE_VAL_COMPLEX(195, 380), MAKE_VAL_COMPLEX(271, 76), MAKE_VAL_COMPLEX(-178, 594), MAKE_VAL_COMPLEX(435, 1410), MAKE_VAL_COMPLEX(1058, -668), MAKE_VAL_COMPLEX(212, -299) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 9);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1327, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1327", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1327() */
/* END TEST 1327 */

/*
 *  BEGIN TEST 1328
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1328( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-8, -110), MAKE_VAL_COMPLEX(180, 41), MAKE_VAL_COMPLEX(-121, -113), MAKE_VAL_COMPLEX(-241, -173), MAKE_VAL_COMPLEX(8, 15), MAKE_VAL_COMPLEX(-50, -92), MAKE_VAL_COMPLEX(115, 101), MAKE_VAL_COMPLEX(32, 172), MAKE_VAL_COMPLEX(132, 125), MAKE_VAL_COMPLEX(24, -272), MAKE_VAL_COMPLEX(-26, -133), MAKE_VAL_COMPLEX(-152, 14), MAKE_VAL_COMPLEX(-46, -182), MAKE_VAL_COMPLEX(-210, 157), MAKE_VAL_COMPLEX(31, 224) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1328, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1328", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1328() */
/* END TEST 1328 */

/*
 *  BEGIN TEST 1329
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1329( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(5, -9), MAKE_VAL_COMPLEX(-98, 5), MAKE_VAL_COMPLEX(173, 9), MAKE_VAL_COMPLEX(140, 110), MAKE_VAL_COMPLEX(-54, -63), MAKE_VAL_COMPLEX(120, 52), MAKE_VAL_COMPLEX(-35, 39), MAKE_VAL_COMPLEX(-49, -153), MAKE_VAL_COMPLEX(-188, -93), MAKE_VAL_COMPLEX(-63, 155), MAKE_VAL_COMPLEX(39, 42), MAKE_VAL_COMPLEX(104, -98), MAKE_VAL_COMPLEX(141, 267), MAKE_VAL_COMPLEX(96, -129), MAKE_VAL_COMPLEX(-134, -193) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1329, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1329", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1329() */
/* END TEST 1329 */

/*
 *  BEGIN TEST 1330
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1330( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(17, -16), MAKE_VAL_COMPLEX(4, 18), MAKE_VAL_COMPLEX(22, -6), MAKE_VAL_COMPLEX(-4, -23), MAKE_VAL_COMPLEX(1, -13), MAKE_VAL_COMPLEX(15, 5), MAKE_VAL_COMPLEX(-10, 30), MAKE_VAL_COMPLEX(-5, 0), MAKE_VAL_COMPLEX(-12, -4), MAKE_VAL_COMPLEX(12, -21), MAKE_VAL_COMPLEX(15, -10), MAKE_VAL_COMPLEX(6, -18), MAKE_VAL_COMPLEX(0, 25), MAKE_VAL_COMPLEX(-19, -13), MAKE_VAL_COMPLEX(-18, -11) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1330, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1330", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1330() */
/* END TEST 1330 */

/*
 *  BEGIN TEST 1331
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1331( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_ROWMAJ, y1296_VW );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(85, 362), MAKE_VAL_COMPLEX(-1111, 610), MAKE_VAL_COMPLEX(1325, -685), MAKE_VAL_COMPLEX(2341, 68), MAKE_VAL_COMPLEX(-652, 27), MAKE_VAL_COMPLEX(1079, -209), MAKE_VAL_COMPLEX(-437, 67), MAKE_VAL_COMPLEX(-1411, -970), MAKE_VAL_COMPLEX(-1959, 400), MAKE_VAL_COMPLEX(830, 2113), MAKE_VAL_COMPLEX(628, 434), MAKE_VAL_COMPLEX(404, -1218), MAKE_VAL_COMPLEX(2552, 889), MAKE_VAL_COMPLEX(294, -1907), MAKE_VAL_COMPLEX(-2040, -812) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1331, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1331", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1331() */
/* END TEST 1331 */

/*
 *  BEGIN TEST 1332
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1332( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	oski_value_t* y_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(223, 24), MAKE_VAL_COMPLEX(-113, -21), MAKE_VAL_COMPLEX(108, -11), MAKE_VAL_COMPLEX(56, -86), MAKE_VAL_COMPLEX(-196, 169), MAKE_VAL_COMPLEX(3, -127), MAKE_VAL_COMPLEX(108, 33), MAKE_VAL_COMPLEX(-25, -83), MAKE_VAL_COMPLEX(79, 1), MAKE_VAL_COMPLEX(72, 72), MAKE_VAL_COMPLEX(127, 124), MAKE_VAL_COMPLEX(-148, 19), MAKE_VAL_COMPLEX(-98, -240), MAKE_VAL_COMPLEX(-79, -114), MAKE_VAL_COMPLEX(-139, 214) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1332, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1332", y_true, y_val, y1296_LEN*y1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1332() */
/* END TEST 1332 */

/*
 *  BEGIN TEST 1333
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1333( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	oski_value_t* y_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-186, -111), MAKE_VAL_COMPLEX(78, 90), MAKE_VAL_COMPLEX(-16, 3), MAKE_VAL_COMPLEX(-82, -8), MAKE_VAL_COMPLEX(251, -90), MAKE_VAL_COMPLEX(-84, 36), MAKE_VAL_COMPLEX(-42, -53), MAKE_VAL_COMPLEX(-23, 105), MAKE_VAL_COMPLEX(-113, 59), MAKE_VAL_COMPLEX(2, 0), MAKE_VAL_COMPLEX(-68, -163), MAKE_VAL_COMPLEX(120, 69), MAKE_VAL_COMPLEX(27, 223), MAKE_VAL_COMPLEX(143, 112), MAKE_VAL_COMPLEX(156, -121) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1333, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1333", y_true, y_val, y1296_LEN*y1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1333() */
/* END TEST 1333 */

/*
 *  BEGIN TEST 1334
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1334( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	oski_value_t* y_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(59, -15), MAKE_VAL_COMPLEX(-49, 9), MAKE_VAL_COMPLEX(40, 44), MAKE_VAL_COMPLEX(38, -50), MAKE_VAL_COMPLEX(-19, 59), MAKE_VAL_COMPLEX(15, -77), MAKE_VAL_COMPLEX(36, 26), MAKE_VAL_COMPLEX(-30, -16), MAKE_VAL_COMPLEX(-44, 6), MAKE_VAL_COMPLEX(-8, 66), MAKE_VAL_COMPLEX(43, 15), MAKE_VAL_COMPLEX(-56, 20), MAKE_VAL_COMPLEX(-19, -43), MAKE_VAL_COMPLEX(26, 32), MAKE_VAL_COMPLEX(-41, 45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1334, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1334", y_true, y_val, y1296_LEN*y1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1334() */
/* END TEST 1334 */

/*
 *  BEGIN TEST 1335
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1335( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	oski_value_t* y_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-268, 337), MAKE_VAL_COMPLEX(137, -68), MAKE_VAL_COMPLEX(203, 217), MAKE_VAL_COMPLEX(26, -12), MAKE_VAL_COMPLEX(-60, -567), MAKE_VAL_COMPLEX(34, -65), MAKE_VAL_COMPLEX(-22, 162), MAKE_VAL_COMPLEX(147, 125), MAKE_VAL_COMPLEX(-127, 457), MAKE_VAL_COMPLEX(15, 293), MAKE_VAL_COMPLEX(-355, 76), MAKE_VAL_COMPLEX(93, -161), MAKE_VAL_COMPLEX(580, -35), MAKE_VAL_COMPLEX(618, -194), MAKE_VAL_COMPLEX(-356, -359) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1335, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1335", y_true, y_val, y1296_LEN*y1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1335() */
/* END TEST 1335 */

/*
 *  BEGIN TEST 1336
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1336( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(139, -42), MAKE_VAL_COMPLEX(-39, -121), MAKE_VAL_COMPLEX(55, 43), MAKE_VAL_COMPLEX(60, 172), MAKE_VAL_COMPLEX(-25, -171), MAKE_VAL_COMPLEX(30, 43), MAKE_VAL_COMPLEX(-91, 11), MAKE_VAL_COMPLEX(99, -214), MAKE_VAL_COMPLEX(219, -32), MAKE_VAL_COMPLEX(-27, 35), MAKE_VAL_COMPLEX(-18, 126), MAKE_VAL_COMPLEX(163, -110), MAKE_VAL_COMPLEX(-10, -45), MAKE_VAL_COMPLEX(73, 18), MAKE_VAL_COMPLEX(61, -87) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1336, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1336", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1336() */
/* END TEST 1336 */

/*
 *  BEGIN TEST 1337
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1337( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-112, 132), MAKE_VAL_COMPLEX(-33, 103), MAKE_VAL_COMPLEX(-73, 47), MAKE_VAL_COMPLEX(30, -145), MAKE_VAL_COMPLEX(70, 198), MAKE_VAL_COMPLEX(-75, 2), MAKE_VAL_COMPLEX(1, -101), MAKE_VAL_COMPLEX(-90, 196), MAKE_VAL_COMPLEX(-183, -4), MAKE_VAL_COMPLEX(81, 46), MAKE_VAL_COMPLEX(27, -54), MAKE_VAL_COMPLEX(-109, 164), MAKE_VAL_COMPLEX(-80, 0), MAKE_VAL_COMPLEX(8, -63), MAKE_VAL_COMPLEX(11, 42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1337, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1337", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1337() */
/* END TEST 1337 */

/*
 *  BEGIN TEST 1338
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1338( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(90, -27), MAKE_VAL_COMPLEX(-18, 72), MAKE_VAL_COMPLEX(90, 18), MAKE_VAL_COMPLEX(27, -90), MAKE_VAL_COMPLEX(27, -45), MAKE_VAL_COMPLEX(45, 45), MAKE_VAL_COMPLEX(-90, 90), MAKE_VAL_COMPLEX(-18, -9), MAKE_VAL_COMPLEX(-36, -36), MAKE_VAL_COMPLEX(81, -54), MAKE_VAL_COMPLEX(72, -9), MAKE_VAL_COMPLEX(54, -54), MAKE_VAL_COMPLEX(-45, 90), MAKE_VAL_COMPLEX(-45, -81), MAKE_VAL_COMPLEX(-45, -72) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1338, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1338", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1338() */
/* END TEST 1338 */

/*
 *  BEGIN TEST 1339
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1339( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-303, -599), MAKE_VAL_COMPLEX(-1341, -105), MAKE_VAL_COMPLEX(53, 33), MAKE_VAL_COMPLEX(1841, 86), MAKE_VAL_COMPLEX(-1228, 298), MAKE_VAL_COMPLEX(9, -82), MAKE_VAL_COMPLEX(-233, 9), MAKE_VAL_COMPLEX(-1515, -1077), MAKE_VAL_COMPLEX(255, -1876), MAKE_VAL_COMPLEX(316, 833), MAKE_VAL_COMPLEX(838, 679), MAKE_VAL_COMPLEX(-579, -976), MAKE_VAL_COMPLEX(-725, -455), MAKE_VAL_COMPLEX(774, -581), MAKE_VAL_COMPLEX(-126, -615) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1339, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1339", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1339() */
/* END TEST 1339 */

/*
 *  BEGIN TEST 1340
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1340( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(45, 78), MAKE_VAL_COMPLEX(-51, -11), MAKE_VAL_COMPLEX(11, -127), MAKE_VAL_COMPLEX(-120, 138), MAKE_VAL_COMPLEX(-179, -159), MAKE_VAL_COMPLEX(96, -67), MAKE_VAL_COMPLEX(-103, 231), MAKE_VAL_COMPLEX(-255, 4), MAKE_VAL_COMPLEX(177, 164), MAKE_VAL_COMPLEX(35, -53), MAKE_VAL_COMPLEX(-20, -168), MAKE_VAL_COMPLEX(-35, -114), MAKE_VAL_COMPLEX(-62, 47), MAKE_VAL_COMPLEX(-165, 336), MAKE_VAL_COMPLEX(93, -85) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1340, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1340", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1340() */
/* END TEST 1340 */

/*
 *  BEGIN TEST 1341
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1341( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-53, -250), MAKE_VAL_COMPLEX(171, 75), MAKE_VAL_COMPLEX(61, -25), MAKE_VAL_COMPLEX(-28, -226), MAKE_VAL_COMPLEX(111, 91), MAKE_VAL_COMPLEX(4, 7), MAKE_VAL_COMPLEX(223, -31), MAKE_VAL_COMPLEX(231, 24), MAKE_VAL_COMPLEX(-257, -116), MAKE_VAL_COMPLEX(-95, -115), MAKE_VAL_COMPLEX(36, 36), MAKE_VAL_COMPLEX(-37, -6), MAKE_VAL_COMPLEX(202, 73), MAKE_VAL_COMPLEX(1, -292), MAKE_VAL_COMPLEX(-241, 133) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1341, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1341", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1341() */
/* END TEST 1341 */

/*
 *  BEGIN TEST 1342
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1342( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(9, 30), MAKE_VAL_COMPLEX(-24, -6), MAKE_VAL_COMPLEX(-6, 30), MAKE_VAL_COMPLEX(30, 9), MAKE_VAL_COMPLEX(15, 9), MAKE_VAL_COMPLEX(-15, 15), MAKE_VAL_COMPLEX(-30, -30), MAKE_VAL_COMPLEX(3, -6), MAKE_VAL_COMPLEX(12, -12), MAKE_VAL_COMPLEX(18, 27), MAKE_VAL_COMPLEX(3, 24), MAKE_VAL_COMPLEX(18, 18), MAKE_VAL_COMPLEX(-30, -15), MAKE_VAL_COMPLEX(27, -15), MAKE_VAL_COMPLEX(24, -15) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1342, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1342", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1342() */
/* END TEST 1342 */

/*
 *  BEGIN TEST 1343
 *  Value type: complex
 *  A : <MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1296', LAYOUT_COLMAJ>
 *  y : <'y1296', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1343( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1296_ptr, A1296_ind, A1296_val,
		A1296_M, A1296_N, SHARE_INPUTMAT,
		5, MAT_HERM_FULL, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1296_val,
		y1296_LEN * y1296_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1296_LEN, y1296_VW, LAYOUT_COLMAJ, y1296_LEN );
	oski_value_t* y_val = array_Duplicate( x1296_val,
		x1296_LEN * x1296_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1296_LEN, x1296_VW, LAYOUT_COLMAJ, x1296_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(685, 1177), MAKE_VAL_COMPLEX(-975, -161), MAKE_VAL_COMPLEX(-753, -207), MAKE_VAL_COMPLEX(421, 1844), MAKE_VAL_COMPLEX(-1560, -84), MAKE_VAL_COMPLEX(-61, -558), MAKE_VAL_COMPLEX(-387, 1099), MAKE_VAL_COMPLEX(-1659, 913), MAKE_VAL_COMPLEX(2267, 204), MAKE_VAL_COMPLEX(336, 275), MAKE_VAL_COMPLEX(-872, -471), MAKE_VAL_COMPLEX(-347, -100), MAKE_VAL_COMPLEX(-921, 27), MAKE_VAL_COMPLEX(958, 2785), MAKE_VAL_COMPLEX(1004, -1211) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, -4);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1343, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1343", y_true, y_val, x1296_LEN*x1296_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1343() */
/* END TEST 1343 */

#define A1344_M 5    /* # of rows */
#define A1344_N 5    /* # of cols */
#define A1344_NNZ 7  /* # of stored non-zeros */
static oski_index_t A1344_ptr[A1344_M+1] = {1, 4, 5, 6, 7, 8}; /* end row pointers for A1344 */
static oski_index_t A1344_ind[A1344_NNZ] = {1, 3, 5, 4, 4, 5, 5}; /* end column indices pointers for A1344 */
static oski_value_t A1344_val[A1344_NNZ] = {MAKE_VAL_COMPLEX(-4, 0), MAKE_VAL_COMPLEX(2, -6), MAKE_VAL_COMPLEX(-3, -2), MAKE_VAL_COMPLEX(-10, 1), MAKE_VAL_COMPLEX(-6, 10), MAKE_VAL_COMPLEX(-4, -6), MAKE_VAL_COMPLEX(10, 0)}; /* end stored non-zero values for A1344 */

#define x1344_LEN 5   /* vector length */
#define x1344_VW  3   /* vector width */
static oski_value_t x1344_val[x1344_LEN*x1344_VW] = {MAKE_VAL_COMPLEX(-10, 3), MAKE_VAL_COMPLEX(0, -9), MAKE_VAL_COMPLEX(-9, 0), MAKE_VAL_COMPLEX(10, -8), MAKE_VAL_COMPLEX(6, -4), MAKE_VAL_COMPLEX(7, 10), MAKE_VAL_COMPLEX(-3, -10), MAKE_VAL_COMPLEX(-9, 0), MAKE_VAL_COMPLEX(2, -8), MAKE_VAL_COMPLEX(-2, 2), MAKE_VAL_COMPLEX(3, 6), MAKE_VAL_COMPLEX(3, -5), MAKE_VAL_COMPLEX(-5, 7), MAKE_VAL_COMPLEX(-1, 4), MAKE_VAL_COMPLEX(-8, 3)}; /* data */

#define y1344_LEN 5   /* vector length */
#define y1344_VW  3   /* vector width */
static oski_value_t y1344_val[y1344_LEN*y1344_VW] = {MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(-6, 9), MAKE_VAL_COMPLEX(2, 3), MAKE_VAL_COMPLEX(7, -7), MAKE_VAL_COMPLEX(-6, 6), MAKE_VAL_COMPLEX(2, -1), MAKE_VAL_COMPLEX(-3, 3), MAKE_VAL_COMPLEX(4, -6), MAKE_VAL_COMPLEX(-8, -1), MAKE_VAL_COMPLEX(8, -3), MAKE_VAL_COMPLEX(-3, 6), MAKE_VAL_COMPLEX(2, -4), MAKE_VAL_COMPLEX(4, -3), MAKE_VAL_COMPLEX(10, 10), MAKE_VAL_COMPLEX(3, -1)}; /* data */

/*
 *  BEGIN TEST 1344
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1344( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	oski_value_t* y_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-127, -75), MAKE_VAL_COMPLEX(-25, 185), MAKE_VAL_COMPLEX(52, -2), MAKE_VAL_COMPLEX(53, -113), MAKE_VAL_COMPLEX(-66, 21), MAKE_VAL_COMPLEX(-11, 36), MAKE_VAL_COMPLEX(-61, -47), MAKE_VAL_COMPLEX(-12, -94), MAKE_VAL_COMPLEX(-62, 29), MAKE_VAL_COMPLEX(-68, 103), MAKE_VAL_COMPLEX(15, 180), MAKE_VAL_COMPLEX(-100, -87), MAKE_VAL_COMPLEX(-6, -22), MAKE_VAL_COMPLEX(90, 111), MAKE_VAL_COMPLEX(-12, 29) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1344, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1344", y_true, y_val, y1344_LEN*y1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1344() */
/* END TEST 1344 */

/*
 *  BEGIN TEST 1345
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1345( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	oski_value_t* y_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-93, -5), MAKE_VAL_COMPLEX(-2, -8), MAKE_VAL_COMPLEX(-1, 33), MAKE_VAL_COMPLEX(3, -41), MAKE_VAL_COMPLEX(18, 111), MAKE_VAL_COMPLEX(34, -65), MAKE_VAL_COMPLEX(37, 113), MAKE_VAL_COMPLEX(30, -24), MAKE_VAL_COMPLEX(-65, 12), MAKE_VAL_COMPLEX(167, -204), MAKE_VAL_COMPLEX(-18, -69), MAKE_VAL_COMPLEX(102, 13), MAKE_VAL_COMPLEX(45, -51), MAKE_VAL_COMPLEX(130, -31), MAKE_VAL_COMPLEX(50, -65) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1345, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1345", y_true, y_val, y1344_LEN*y1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1345() */
/* END TEST 1345 */

/*
 *  BEGIN TEST 1346
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1346( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	oski_value_t* y_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(60, 60), MAKE_VAL_COMPLEX(36, -54), MAKE_VAL_COMPLEX(-12, -18), MAKE_VAL_COMPLEX(-42, 42), MAKE_VAL_COMPLEX(36, -36), MAKE_VAL_COMPLEX(-12, 6), MAKE_VAL_COMPLEX(18, -18), MAKE_VAL_COMPLEX(-24, 36), MAKE_VAL_COMPLEX(48, 6), MAKE_VAL_COMPLEX(-48, 18), MAKE_VAL_COMPLEX(18, -36), MAKE_VAL_COMPLEX(-12, 24), MAKE_VAL_COMPLEX(-24, 18), MAKE_VAL_COMPLEX(-60, -60), MAKE_VAL_COMPLEX(-18, 6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1346, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1346", y_true, y_val, y1344_LEN*y1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1346() */
/* END TEST 1346 */

/*
 *  BEGIN TEST 1347
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1347( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	oski_value_t* y_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(211, -193), MAKE_VAL_COMPLEX(-405, 469), MAKE_VAL_COMPLEX(214, -56), MAKE_VAL_COMPLEX(235, -19), MAKE_VAL_COMPLEX(-42, -597), MAKE_VAL_COMPLEX(-387, 286), MAKE_VAL_COMPLEX(13, -813), MAKE_VAL_COMPLEX(-48, -222), MAKE_VAL_COMPLEX(104, 53), MAKE_VAL_COMPLEX(-1550, 679), MAKE_VAL_COMPLEX(-357, 834), MAKE_VAL_COMPLEX(-524, -685), MAKE_VAL_COMPLEX(-288, 60), MAKE_VAL_COMPLEX(-614, 297), MAKE_VAL_COMPLEX(-448, 231) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1347, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1347", y_true, y_val, y1344_LEN*y1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1347() */
/* END TEST 1347 */

/*
 *  BEGIN TEST 1348
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1348( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(64, 7), MAKE_VAL_COMPLEX(36, 20), MAKE_VAL_COMPLEX(29, -71), MAKE_VAL_COMPLEX(-127, 90), MAKE_VAL_COMPLEX(8, -21), MAKE_VAL_COMPLEX(-86, -8), MAKE_VAL_COMPLEX(-120, 32), MAKE_VAL_COMPLEX(174, 30), MAKE_VAL_COMPLEX(-26, 50), MAKE_VAL_COMPLEX(-65, 49), MAKE_VAL_COMPLEX(-40, 0), MAKE_VAL_COMPLEX(25, -4), MAKE_VAL_COMPLEX(16, -68), MAKE_VAL_COMPLEX(182, 53), MAKE_VAL_COMPLEX(40, -53) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1348, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1348", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1348() */
/* END TEST 1348 */

/*
 *  BEGIN TEST 1349
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1349( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(16, -31), MAKE_VAL_COMPLEX(-36, 52), MAKE_VAL_COMPLEX(43, 71), MAKE_VAL_COMPLEX(47, -26), MAKE_VAL_COMPLEX(-56, 53), MAKE_VAL_COMPLEX(30, -72), MAKE_VAL_COMPLEX(144, 48), MAKE_VAL_COMPLEX(-102, -30), MAKE_VAL_COMPLEX(10, 14), MAKE_VAL_COMPLEX(81, -65), MAKE_VAL_COMPLEX(16, -48), MAKE_VAL_COMPLEX(-49, 44), MAKE_VAL_COMPLEX(24, 12), MAKE_VAL_COMPLEX(-174, -85), MAKE_VAL_COMPLEX(24, 29) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1349, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1349", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1349() */
/* END TEST 1349 */

/*
 *  BEGIN TEST 1350
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1350( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-72, 87), MAKE_VAL_COMPLEX(-54, -81), MAKE_VAL_COMPLEX(-81, 54), MAKE_VAL_COMPLEX(42, -132), MAKE_VAL_COMPLEX(30, -72), MAKE_VAL_COMPLEX(123, 48), MAKE_VAL_COMPLEX(-87, -72), MAKE_VAL_COMPLEX(-81, 54), MAKE_VAL_COMPLEX(-30, -84), MAKE_VAL_COMPLEX(-6, 30), MAKE_VAL_COMPLEX(63, 36), MAKE_VAL_COMPLEX(-3, -63), MAKE_VAL_COMPLEX(-3, 93), MAKE_VAL_COMPLEX(15, 42), MAKE_VAL_COMPLEX(-54, 75) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1350, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1350", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1350() */
/* END TEST 1350 */

/*
 *  BEGIN TEST 1351
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1351( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-150, 31), MAKE_VAL_COMPLEX(136, 99), MAKE_VAL_COMPLEX(239, -100), MAKE_VAL_COMPLEX(-118, -308), MAKE_VAL_COMPLEX(210, 156), MAKE_VAL_COMPLEX(-173, -126), MAKE_VAL_COMPLEX(97, -602), MAKE_VAL_COMPLEX(-165, 480), MAKE_VAL_COMPLEX(2, -96), MAKE_VAL_COMPLEX(-254, -298), MAKE_VAL_COMPLEX(-129, -58), MAKE_VAL_COMPLEX(151, 147), MAKE_VAL_COMPLEX(79, -21), MAKE_VAL_COMPLEX(-313, 724), MAKE_VAL_COMPLEX(100, -17) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1351, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1351", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1351() */
/* END TEST 1351 */

/*
 *  BEGIN TEST 1352
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1352( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-27, 116), MAKE_VAL_COMPLEX(-41, -185), MAKE_VAL_COMPLEX(-104, 58), MAKE_VAL_COMPLEX(-31, -48), MAKE_VAL_COMPLEX(54, -109), MAKE_VAL_COMPLEX(63, 91), MAKE_VAL_COMPLEX(-29, -43), MAKE_VAL_COMPLEX(-171, -57), MAKE_VAL_COMPLEX(4, 0), MAKE_VAL_COMPLEX(-71, 83), MAKE_VAL_COMPLEX(41, -125), MAKE_VAL_COMPLEX(5, 36), MAKE_VAL_COMPLEX(62, 104), MAKE_VAL_COMPLEX(81, 50), MAKE_VAL_COMPLEX(-13, 58) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1352, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1352", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1352() */
/* END TEST 1352 */

/*
 *  BEGIN TEST 1353
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1353( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(46, -89), MAKE_VAL_COMPLEX(14, 194), MAKE_VAL_COMPLEX(113, -31), MAKE_VAL_COMPLEX(-3, 26), MAKE_VAL_COMPLEX(-72, 95), MAKE_VAL_COMPLEX(-40, -122), MAKE_VAL_COMPLEX(2, 62), MAKE_VAL_COMPLEX(180, 84), MAKE_VAL_COMPLEX(-30, 2), MAKE_VAL_COMPLEX(79, -79), MAKE_VAL_COMPLEX(-26, 110), MAKE_VAL_COMPLEX(-23, -40), MAKE_VAL_COMPLEX(-36, -96), MAKE_VAL_COMPLEX(-68, -51), MAKE_VAL_COMPLEX(30, -37) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1353, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1353", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1353() */
/* END TEST 1353 */

/*
 *  BEGIN TEST 1354
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1354( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(28, 57), MAKE_VAL_COMPLEX(-54, 9), MAKE_VAL_COMPLEX(9, 54), MAKE_VAL_COMPLEX(-58, -52), MAKE_VAL_COMPLEX(-30, -32), MAKE_VAL_COMPLEX(53, -52), MAKE_VAL_COMPLEX(-57, 28), MAKE_VAL_COMPLEX(9, 54), MAKE_VAL_COMPLEX(-50, -4), MAKE_VAL_COMPLEX(14, 10), MAKE_VAL_COMPLEX(33, -24), MAKE_VAL_COMPLEX(-33, -13), MAKE_VAL_COMPLEX(47, 23), MAKE_VAL_COMPLEX(25, 2), MAKE_VAL_COMPLEX(26, 45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1354, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1354", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1354() */
/* END TEST 1354 */

/*
 *  BEGIN TEST 1355
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1355( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(54, -363), MAKE_VAL_COMPLEX(-490, 612), MAKE_VAL_COMPLEX(275, 81), MAKE_VAL_COMPLEX(533, -10), MAKE_VAL_COMPLEX(-368, 147), MAKE_VAL_COMPLEX(370, -142), MAKE_VAL_COMPLEX(-136, -214), MAKE_VAL_COMPLEX(150, 924), MAKE_VAL_COMPLEX(126, -430), MAKE_VAL_COMPLEX(571, -35), MAKE_VAL_COMPLEX(-576, 794), MAKE_VAL_COMPLEX(303, -450), MAKE_VAL_COMPLEX(-194, -422), MAKE_VAL_COMPLEX(-278, -367), MAKE_VAL_COMPLEX(-142, -119) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, -4);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1355, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1355", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1355() */
/* END TEST 1355 */

/*
 *  BEGIN TEST 1356
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1356( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	oski_value_t* y_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-17, -85), MAKE_VAL_COMPLEX(-24, 2), MAKE_VAL_COMPLEX(-44, -70), MAKE_VAL_COMPLEX(-86, 148), MAKE_VAL_COMPLEX(-66, 33), MAKE_VAL_COMPLEX(3, 80), MAKE_VAL_COMPLEX(-54, -51), MAKE_VAL_COMPLEX(4, -40), MAKE_VAL_COMPLEX(-12, 94), MAKE_VAL_COMPLEX(-2, 65), MAKE_VAL_COMPLEX(-2, -3), MAKE_VAL_COMPLEX(-9, 41), MAKE_VAL_COMPLEX(36, 2), MAKE_VAL_COMPLEX(-82, 17), MAKE_VAL_COMPLEX(-21, 52) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1356, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1356", y_true, y_val, y1344_LEN*y1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1356() */
/* END TEST 1356 */

/*
 *  BEGIN TEST 1357
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1357( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	oski_value_t* y_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-13, 95), MAKE_VAL_COMPLEX(36, 19), MAKE_VAL_COMPLEX(52, 69), MAKE_VAL_COMPLEX(79, -169), MAKE_VAL_COMPLEX(72, -15), MAKE_VAL_COMPLEX(-3, -85), MAKE_VAL_COMPLEX(57, 60), MAKE_VAL_COMPLEX(-12, 26), MAKE_VAL_COMPLEX(2, -79), MAKE_VAL_COMPLEX(4, -84), MAKE_VAL_COMPLEX(11, 15), MAKE_VAL_COMPLEX(3, -49), MAKE_VAL_COMPLEX(-38, -13), MAKE_VAL_COMPLEX(112, -27), MAKE_VAL_COMPLEX(22, -59) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1357, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1357", y_true, y_val, y1344_LEN*y1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1357() */
/* END TEST 1357 */

/*
 *  BEGIN TEST 1358
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1358( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	oski_value_t* y_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-90, 10), MAKE_VAL_COMPLEX(21, 66), MAKE_VAL_COMPLEX(23, 2), MAKE_VAL_COMPLEX(-7, -63), MAKE_VAL_COMPLEX(6, 54), MAKE_VAL_COMPLEX(3, -14), MAKE_VAL_COMPLEX(3, 27), MAKE_VAL_COMPLEX(-14, -44), MAKE_VAL_COMPLEX(-37, 36), MAKE_VAL_COMPLEX(17, -52), MAKE_VAL_COMPLEX(18, 39), MAKE_VAL_COMPLEX(-12, -26), MAKE_VAL_COMPLEX(1, -32), MAKE_VAL_COMPLEX(90, -10), MAKE_VAL_COMPLEX(7, -19) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1358, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1358", y_true, y_val, y1344_LEN*y1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1358() */
/* END TEST 1358 */

/*
 *  BEGIN TEST 1359
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1359( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	oski_value_t* y_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-67, -69), MAKE_VAL_COMPLEX(-261, 43), MAKE_VAL_COMPLEX(-87, 521), MAKE_VAL_COMPLEX(1110, -250), MAKE_VAL_COMPLEX(87, -6), MAKE_VAL_COMPLEX(291, -288), MAKE_VAL_COMPLEX(-75, 330), MAKE_VAL_COMPLEX(106, 162), MAKE_VAL_COMPLEX(223, -589), MAKE_VAL_COMPLEX(426, -73), MAKE_VAL_COMPLEX(-226, 33), MAKE_VAL_COMPLEX(309, -147), MAKE_VAL_COMPLEX(17, -51), MAKE_VAL_COMPLEX(259, 638), MAKE_VAL_COMPLEX(320, -75) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, -3);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1359, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1359", y_true, y_val, y1344_LEN*y1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1359() */
/* END TEST 1359 */

/*
 *  BEGIN TEST 1360
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1360( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(59, 9), MAKE_VAL_COMPLEX(27, 68), MAKE_VAL_COMPLEX(58, -6), MAKE_VAL_COMPLEX(-99, 97), MAKE_VAL_COMPLEX(65, 36), MAKE_VAL_COMPLEX(-121, -133), MAKE_VAL_COMPLEX(-173, 39), MAKE_VAL_COMPLEX(63, 63), MAKE_VAL_COMPLEX(72, -104), MAKE_VAL_COMPLEX(1, -162), MAKE_VAL_COMPLEX(34, 88), MAKE_VAL_COMPLEX(-20, 140), MAKE_VAL_COMPLEX(-112, 70), MAKE_VAL_COMPLEX(97, 8), MAKE_VAL_COMPLEX(92, -139) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1360, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1360", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1360() */
/* END TEST 1360 */

/*
 *  BEGIN TEST 1361
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1361( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-120, -67), MAKE_VAL_COMPLEX(36, -104), MAKE_VAL_COMPLEX(-94, -57), MAKE_VAL_COMPLEX(195, -59), MAKE_VAL_COMPLEX(-13, -10), MAKE_VAL_COMPLEX(79, 222), MAKE_VAL_COMPLEX(231, -100), MAKE_VAL_COMPLEX(-99, -126), MAKE_VAL_COMPLEX(-8, 86), MAKE_VAL_COMPLEX(-23, 156), MAKE_VAL_COMPLEX(-64, -43), MAKE_VAL_COMPLEX(67, -139), MAKE_VAL_COMPLEX(43, -77), MAKE_VAL_COMPLEX(-129, 1), MAKE_VAL_COMPLEX(-145, 95) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1361, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1361", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1361() */
/* END TEST 1361 */

/*
 *  BEGIN TEST 1362
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1362( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(11, -36), MAKE_VAL_COMPLEX(27, 18), MAKE_VAL_COMPLEX(18, -27), MAKE_VAL_COMPLEX(4, 46), MAKE_VAL_COMPLEX(0, 26), MAKE_VAL_COMPLEX(-44, 1), MAKE_VAL_COMPLEX(36, 11), MAKE_VAL_COMPLEX(18, -27), MAKE_VAL_COMPLEX(20, 22), MAKE_VAL_COMPLEX(-2, -10), MAKE_VAL_COMPLEX(-24, -3), MAKE_VAL_COMPLEX(9, 19), MAKE_VAL_COMPLEX(-11, -29), MAKE_VAL_COMPLEX(-10, -11), MAKE_VAL_COMPLEX(7, -30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1362, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1362", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1362() */
/* END TEST 1362 */

/*
 *  BEGIN TEST 1363
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1363( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(57, 35), MAKE_VAL_COMPLEX(-37, 158), MAKE_VAL_COMPLEX(106, 38), MAKE_VAL_COMPLEX(-269, 11), MAKE_VAL_COMPLEX(1, 230), MAKE_VAL_COMPLEX(37, -359), MAKE_VAL_COMPLEX(-159, -291), MAKE_VAL_COMPLEX(-27, 117), MAKE_VAL_COMPLEX(336, 88), MAKE_VAL_COMPLEX(317, -184), MAKE_VAL_COMPLEX(-202, 156), MAKE_VAL_COMPLEX(-272, 144), MAKE_VAL_COMPLEX(-288, -222), MAKE_VAL_COMPLEX(53, 178), MAKE_VAL_COMPLEX(378, -31) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1363, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1363", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1363() */
/* END TEST 1363 */

/*
 *  BEGIN TEST 1364
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1364( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(15, 62), MAKE_VAL_COMPLEX(-75, -111), MAKE_VAL_COMPLEX(-81, -48), MAKE_VAL_COMPLEX(9, 3), MAKE_VAL_COMPLEX(125, -36), MAKE_VAL_COMPLEX(-44, -17), MAKE_VAL_COMPLEX(34, -45), MAKE_VAL_COMPLEX(-4, -55), MAKE_VAL_COMPLEX(-194, -32), MAKE_VAL_COMPLEX(133, -92), MAKE_VAL_COMPLEX(-71, -22), MAKE_VAL_COMPLEX(-77, -41), MAKE_VAL_COMPLEX(-29, 201), MAKE_VAL_COMPLEX(110, -34), MAKE_VAL_COMPLEX(-134, 18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1364, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1364", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1364() */
/* END TEST 1364 */

/*
 *  BEGIN TEST 1365
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1365( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-117, 34), MAKE_VAL_COMPLEX(21, 3), MAKE_VAL_COMPLEX(-27, 102), MAKE_VAL_COMPLEX(63, -159), MAKE_VAL_COMPLEX(-77, -48), MAKE_VAL_COMPLEX(188, 95), MAKE_VAL_COMPLEX(-130, -57), MAKE_VAL_COMPLEX(-104, 109), MAKE_VAL_COMPLEX(170, -76), MAKE_VAL_COMPLEX(-145, 128), MAKE_VAL_COMPLEX(143, 76), MAKE_VAL_COMPLEX(83, -37), MAKE_VAL_COMPLEX(11, -87), MAKE_VAL_COMPLEX(-98, 88), MAKE_VAL_COMPLEX(56, 66) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1365, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1365", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1365() */
/* END TEST 1365 */

/*
 *  BEGIN TEST 1366
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1366( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(70, -21), MAKE_VAL_COMPLEX(0, 63), MAKE_VAL_COMPLEX(63, 0), MAKE_VAL_COMPLEX(-70, 56), MAKE_VAL_COMPLEX(-42, 28), MAKE_VAL_COMPLEX(-49, -70), MAKE_VAL_COMPLEX(21, 70), MAKE_VAL_COMPLEX(63, 0), MAKE_VAL_COMPLEX(-14, 56), MAKE_VAL_COMPLEX(14, -14), MAKE_VAL_COMPLEX(-21, -42), MAKE_VAL_COMPLEX(-21, 35), MAKE_VAL_COMPLEX(35, -49), MAKE_VAL_COMPLEX(7, -28), MAKE_VAL_COMPLEX(56, -21) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1366, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1366", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1366() */
/* END TEST 1366 */

/*
 *  BEGIN TEST 1367
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_ROWMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1367( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_ROWMAJ, x1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-589, -198), MAKE_VAL_COMPLEX(342, 339), MAKE_VAL_COMPLEX(6, 489), MAKE_VAL_COMPLEX(446, -566), MAKE_VAL_COMPLEX(-601, -149), MAKE_VAL_COMPLEX(770, 677), MAKE_VAL_COMPLEX(-534, -265), MAKE_VAL_COMPLEX(-540, 461), MAKE_VAL_COMPLEX(1376, -54), MAKE_VAL_COMPLEX(-1121, 641), MAKE_VAL_COMPLEX(694, 543), MAKE_VAL_COMPLEX(629, 55), MAKE_VAL_COMPLEX(175, -939), MAKE_VAL_COMPLEX(-829, 364), MAKE_VAL_COMPLEX(524, 243) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1367, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1367", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1367() */
/* END TEST 1367 */

/*
 *  BEGIN TEST 1368
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1368( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	oski_value_t* y_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(106, 52), MAKE_VAL_COMPLEX(-45, -72), MAKE_VAL_COMPLEX(23, 19), MAKE_VAL_COMPLEX(-99, 167), MAKE_VAL_COMPLEX(-6, 16), MAKE_VAL_COMPLEX(-1, -25), MAKE_VAL_COMPLEX(-15, 61), MAKE_VAL_COMPLEX(28, 186), MAKE_VAL_COMPLEX(-11, -38), MAKE_VAL_COMPLEX(-36, 218), MAKE_VAL_COMPLEX(82, 146), MAKE_VAL_COMPLEX(123, 125), MAKE_VAL_COMPLEX(83, 61), MAKE_VAL_COMPLEX(-131, 38), MAKE_VAL_COMPLEX(-134, 17) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1368, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1368", y_true, y_val, y1344_LEN*y1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1368() */
/* END TEST 1368 */

/*
 *  BEGIN TEST 1369
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1369( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	oski_value_t* y_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-56, 78), MAKE_VAL_COMPLEX(135, 15), MAKE_VAL_COMPLEX(-29, -54), MAKE_VAL_COMPLEX(8, -132), MAKE_VAL_COMPLEX(84, -46), MAKE_VAL_COMPLEX(-21, 26), MAKE_VAL_COMPLEX(54, -76), MAKE_VAL_COMPLEX(-88, -148), MAKE_VAL_COMPLEX(79, 79), MAKE_VAL_COMPLEX(-48, -223), MAKE_VAL_COMPLEX(-31, -188), MAKE_VAL_COMPLEX(-157, -97), MAKE_VAL_COMPLEX(-131, -50), MAKE_VAL_COMPLEX(81, -168), MAKE_VAL_COMPLEX(103, -20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1369, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1369", y_true, y_val, y1344_LEN*y1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1369() */
/* END TEST 1369 */

/*
 *  BEGIN TEST 1370
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1370( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	oski_value_t* y_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-110, -10), MAKE_VAL_COMPLEX(9, 84), MAKE_VAL_COMPLEX(27, 8), MAKE_VAL_COMPLEX(7, -77), MAKE_VAL_COMPLEX(-6, 66), MAKE_VAL_COMPLEX(7, -16), MAKE_VAL_COMPLEX(-3, 33), MAKE_VAL_COMPLEX(-6, -56), MAKE_VAL_COMPLEX(-53, 34), MAKE_VAL_COMPLEX(33, -58), MAKE_VAL_COMPLEX(12, 51), MAKE_VAL_COMPLEX(-8, -34), MAKE_VAL_COMPLEX(9, -38), MAKE_VAL_COMPLEX(110, 10), MAKE_VAL_COMPLEX(13, -21) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1370, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1370", y_true, y_val, y1344_LEN*y1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1370() */
/* END TEST 1370 */

/*
 *  BEGIN TEST 1371
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1371( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	oski_value_t* y_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(344, -32), MAKE_VAL_COMPLEX(69, 72), MAKE_VAL_COMPLEX(52, -281), MAKE_VAL_COMPLEX(549, 461), MAKE_VAL_COMPLEX(416, -100), MAKE_VAL_COMPLEX(-216, 31), MAKE_VAL_COMPLEX(485, -43), MAKE_VAL_COMPLEX(646, -176), MAKE_VAL_COMPLEX(99, 274), MAKE_VAL_COMPLEX(776, -71), MAKE_VAL_COMPLEX(867, -724), MAKE_VAL_COMPLEX(356, -628), MAKE_VAL_COMPLEX(20, -437), MAKE_VAL_COMPLEX(131, 67), MAKE_VAL_COMPLEX(90, 636) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1371, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1371", y_true, y_val, y1344_LEN*y1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1371() */
/* END TEST 1371 */

/*
 *  BEGIN TEST 1372
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1372( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-56, 152), MAKE_VAL_COMPLEX(-155, -59), MAKE_VAL_COMPLEX(-239, 50), MAKE_VAL_COMPLEX(-47, -87), MAKE_VAL_COMPLEX(22, -100), MAKE_VAL_COMPLEX(161, 0), MAKE_VAL_COMPLEX(-71, -123), MAKE_VAL_COMPLEX(39, 120), MAKE_VAL_COMPLEX(-56, -58), MAKE_VAL_COMPLEX(182, 113), MAKE_VAL_COMPLEX(50, -23), MAKE_VAL_COMPLEX(-37, -31), MAKE_VAL_COMPLEX(-19, 101), MAKE_VAL_COMPLEX(56, 4), MAKE_VAL_COMPLEX(-50, 72) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1372, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1372", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1372() */
/* END TEST 1372 */

/*
 *  BEGIN TEST 1373
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1373( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(89, -42), MAKE_VAL_COMPLEX(56, 59), MAKE_VAL_COMPLEX(239, 49), MAKE_VAL_COMPLEX(-41, -23), MAKE_VAL_COMPLEX(-66, 34), MAKE_VAL_COMPLEX(-51, -77), MAKE_VAL_COMPLEX(-39, 156), MAKE_VAL_COMPLEX(-39, -21), MAKE_VAL_COMPLEX(-32, 36), MAKE_VAL_COMPLEX(-160, -91), MAKE_VAL_COMPLEX(16, -10), MAKE_VAL_COMPLEX(-18, -2), MAKE_VAL_COMPLEX(96, -46), MAKE_VAL_COMPLEX(-12, 7), MAKE_VAL_COMPLEX(83, 16) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1373, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1373", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1373() */
/* END TEST 1373 */

/*
 *  BEGIN TEST 1374
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1374( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-60, 18), MAKE_VAL_COMPLEX(0, -54), MAKE_VAL_COMPLEX(-54, 0), MAKE_VAL_COMPLEX(60, -48), MAKE_VAL_COMPLEX(36, -24), MAKE_VAL_COMPLEX(42, 60), MAKE_VAL_COMPLEX(-18, -60), MAKE_VAL_COMPLEX(-54, 0), MAKE_VAL_COMPLEX(12, -48), MAKE_VAL_COMPLEX(-12, 12), MAKE_VAL_COMPLEX(18, 36), MAKE_VAL_COMPLEX(18, -30), MAKE_VAL_COMPLEX(-30, 42), MAKE_VAL_COMPLEX(-6, 24), MAKE_VAL_COMPLEX(-48, 18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1374, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1374", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1374() */
/* END TEST 1374 */

/*
 *  BEGIN TEST 1375
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1375( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-170, -247), MAKE_VAL_COMPLEX(438, -247), MAKE_VAL_COMPLEX(758, -93), MAKE_VAL_COMPLEX(400, -396), MAKE_VAL_COMPLEX(72, 86), MAKE_VAL_COMPLEX(-64, 283), MAKE_VAL_COMPLEX(-91, 246), MAKE_VAL_COMPLEX(-702, -165), MAKE_VAL_COMPLEX(72, -198), MAKE_VAL_COMPLEX(-1021, -209), MAKE_VAL_COMPLEX(166, 291), MAKE_VAL_COMPLEX(125, -229), MAKE_VAL_COMPLEX(87, -3), MAKE_VAL_COMPLEX(-151, 252), MAKE_VAL_COMPLEX(-34, 71) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1375, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1375", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1375() */
/* END TEST 1375 */

/*
 *  BEGIN TEST 1376
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1376( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(135, 111), MAKE_VAL_COMPLEX(-140, 110), MAKE_VAL_COMPLEX(94, 81), MAKE_VAL_COMPLEX(-199, 57), MAKE_VAL_COMPLEX(0, 30), MAKE_VAL_COMPLEX(0, -251), MAKE_VAL_COMPLEX(-22, 38), MAKE_VAL_COMPLEX(-36, -21), MAKE_VAL_COMPLEX(-70, -108), MAKE_VAL_COMPLEX(106, 17), MAKE_VAL_COMPLEX(-23, -38), MAKE_VAL_COMPLEX(-75, 61), MAKE_VAL_COMPLEX(103, 41), MAKE_VAL_COMPLEX(35, 55), MAKE_VAL_COMPLEX(119, 45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1376, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1376", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1376() */
/* END TEST 1376 */

/*
 *  BEGIN TEST 1377
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1377( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-115, -117), MAKE_VAL_COMPLEX(140, -92), MAKE_VAL_COMPLEX(-76, -81), MAKE_VAL_COMPLEX(179, -41), MAKE_VAL_COMPLEX(-12, -22), MAKE_VAL_COMPLEX(-14, 231), MAKE_VAL_COMPLEX(28, -18), MAKE_VAL_COMPLEX(54, 21), MAKE_VAL_COMPLEX(66, 124), MAKE_VAL_COMPLEX(-102, -21), MAKE_VAL_COMPLEX(17, 26), MAKE_VAL_COMPLEX(69, -51), MAKE_VAL_COMPLEX(-93, -55), MAKE_VAL_COMPLEX(-33, -63), MAKE_VAL_COMPLEX(-103, -51) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1377, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1377", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1377() */
/* END TEST 1377 */

/*
 *  BEGIN TEST 1378
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1378( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-28, 52), MAKE_VAL_COMPLEX(-36, -36), MAKE_VAL_COMPLEX(-36, 36), MAKE_VAL_COMPLEX(8, -72), MAKE_VAL_COMPLEX(8, -40), MAKE_VAL_COMPLEX(68, 12), MAKE_VAL_COMPLEX(-52, -28), MAKE_VAL_COMPLEX(-36, 36), MAKE_VAL_COMPLEX(-24, -40), MAKE_VAL_COMPLEX(0, 16), MAKE_VAL_COMPLEX(36, 12), MAKE_VAL_COMPLEX(-8, -32), MAKE_VAL_COMPLEX(8, 48), MAKE_VAL_COMPLEX(12, 20), MAKE_VAL_COMPLEX(-20, 44) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1378, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1378", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1378() */
/* END TEST 1378 */

/*
 *  BEGIN TEST 1379
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1379( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_ROWMAJ, y1344_VW );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(382, -4), MAKE_VAL_COMPLEX(9, 443), MAKE_VAL_COMPLEX(116, -47), MAKE_VAL_COMPLEX(185, 395), MAKE_VAL_COMPLEX(352, -326), MAKE_VAL_COMPLEX(-594, -175), MAKE_VAL_COMPLEX(-303, -257), MAKE_VAL_COMPLEX(-654, 399), MAKE_VAL_COMPLEX(-806, -348), MAKE_VAL_COMPLEX(249, -332), MAKE_VAL_COMPLEX(97, 270), MAKE_VAL_COMPLEX(169, 145), MAKE_VAL_COMPLEX(349, 83), MAKE_VAL_COMPLEX(408, 176), MAKE_VAL_COMPLEX(126, -114) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1379, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1379", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1379() */
/* END TEST 1379 */

/*
 *  BEGIN TEST 1380
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1380( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	oski_value_t* y_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-44, 42), MAKE_VAL_COMPLEX(-86, 120), MAKE_VAL_COMPLEX(-8, 96), MAKE_VAL_COMPLEX(-3, 132), MAKE_VAL_COMPLEX(92, 47), MAKE_VAL_COMPLEX(-34, 6), MAKE_VAL_COMPLEX(-12, 94), MAKE_VAL_COMPLEX(18, 110), MAKE_VAL_COMPLEX(76, 211), MAKE_VAL_COMPLEX(-11, 26), MAKE_VAL_COMPLEX(56, 45), MAKE_VAL_COMPLEX(2, -53), MAKE_VAL_COMPLEX(-62, -18), MAKE_VAL_COMPLEX(155, 91), MAKE_VAL_COMPLEX(-117, -12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1380, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1380", y_true, y_val, y1344_LEN*y1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1380() */
/* END TEST 1380 */

/*
 *  BEGIN TEST 1381
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1381( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	oski_value_t* y_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-106, -92), MAKE_VAL_COMPLEX(71, 0), MAKE_VAL_COMPLEX(43, -76), MAKE_VAL_COMPLEX(38, -237), MAKE_VAL_COMPLEX(-122, 43), MAKE_VAL_COMPLEX(49, -26), MAKE_VAL_COMPLEX(-3, -49), MAKE_VAL_COMPLEX(-8, -190), MAKE_VAL_COMPLEX(-161, -181), MAKE_VAL_COMPLEX(76, -96), MAKE_VAL_COMPLEX(-56, 30), MAKE_VAL_COMPLEX(-2, 3), MAKE_VAL_COMPLEX(87, -32), MAKE_VAL_COMPLEX(-5, -41), MAKE_VAL_COMPLEX(142, -13) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1381, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1381", y_true, y_val, y1344_LEN*y1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1381() */
/* END TEST 1381 */

/*
 *  BEGIN TEST 1382
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1382( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	oski_value_t* y_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-20, -60), MAKE_VAL_COMPLEX(-42, 24), MAKE_VAL_COMPLEX(2, 16), MAKE_VAL_COMPLEX(42, -14), MAKE_VAL_COMPLEX(-36, 12), MAKE_VAL_COMPLEX(10, 0), MAKE_VAL_COMPLEX(-18, 6), MAKE_VAL_COMPLEX(28, -16), MAKE_VAL_COMPLEX(-30, -20), MAKE_VAL_COMPLEX(38, 4), MAKE_VAL_COMPLEX(-24, 18), MAKE_VAL_COMPLEX(16, -12), MAKE_VAL_COMPLEX(22, -4), MAKE_VAL_COMPLEX(20, 60), MAKE_VAL_COMPLEX(14, 2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1382, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1382", y_true, y_val, y1344_LEN*y1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1382() */
/* END TEST 1382 */

/*
 *  BEGIN TEST 1383
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1383( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	oski_value_t* y_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(236, -238), MAKE_VAL_COMPLEX(1153, 385), MAKE_VAL_COMPLEX(869, -305), MAKE_VAL_COMPLEX(1337, -860), MAKE_VAL_COMPLEX(-312, -779), MAKE_VAL_COMPLEX(291, 213), MAKE_VAL_COMPLEX(698, -278), MAKE_VAL_COMPLEX(948, -892), MAKE_VAL_COMPLEX(1031, -1722), MAKE_VAL_COMPLEX(552, -123), MAKE_VAL_COMPLEX(-43, -472), MAKE_VAL_COMPLEX(-352, 115), MAKE_VAL_COMPLEX(315, 493), MAKE_VAL_COMPLEX(273, -1315), MAKE_VAL_COMPLEX(597, 970) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1383, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1383", y_true, y_val, y1344_LEN*y1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1383() */
/* END TEST 1383 */

/*
 *  BEGIN TEST 1384
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1384( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(29, 127), MAKE_VAL_COMPLEX(-158, 36), MAKE_VAL_COMPLEX(-219, 93), MAKE_VAL_COMPLEX(-45, -268), MAKE_VAL_COMPLEX(-138, 30), MAKE_VAL_COMPLEX(129, 8), MAKE_VAL_COMPLEX(-20, 15), MAKE_VAL_COMPLEX(9, 153), MAKE_VAL_COMPLEX(-17, 61), MAKE_VAL_COMPLEX(110, 45), MAKE_VAL_COMPLEX(94, -6), MAKE_VAL_COMPLEX(-126, -152), MAKE_VAL_COMPLEX(118, -64), MAKE_VAL_COMPLEX(17, 143), MAKE_VAL_COMPLEX(74, -41) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1384, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1384", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1384() */
/* END TEST 1384 */

/*
 *  BEGIN TEST 1385
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1385( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-23, -107), MAKE_VAL_COMPLEX(140, -36), MAKE_VAL_COMPLEX(219, -75), MAKE_VAL_COMPLEX(29, 248), MAKE_VAL_COMPLEX(130, -42), MAKE_VAL_COMPLEX(-109, -22), MAKE_VAL_COMPLEX(0, -9), MAKE_VAL_COMPLEX(-9, -135), MAKE_VAL_COMPLEX(1, -65), MAKE_VAL_COMPLEX(-106, -41), MAKE_VAL_COMPLEX(-82, 0), MAKE_VAL_COMPLEX(116, 146), MAKE_VAL_COMPLEX(-104, 74), MAKE_VAL_COMPLEX(-9, -141), MAKE_VAL_COMPLEX(-68, 57) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1385, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1385", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1385() */
/* END TEST 1385 */

/*
 *  BEGIN TEST 1386
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1386( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-36, 98), MAKE_VAL_COMPLEX(-72, -54), MAKE_VAL_COMPLEX(-54, 72), MAKE_VAL_COMPLEX(-4, -128), MAKE_VAL_COMPLEX(4, -72), MAKE_VAL_COMPLEX(122, 4), MAKE_VAL_COMPLEX(-98, -36), MAKE_VAL_COMPLEX(-54, 72), MAKE_VAL_COMPLEX(-52, -64), MAKE_VAL_COMPLEX(4, 28), MAKE_VAL_COMPLEX(66, 12), MAKE_VAL_COMPLEX(-22, -54), MAKE_VAL_COMPLEX(26, 82), MAKE_VAL_COMPLEX(26, 32), MAKE_VAL_COMPLEX(-24, 82) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1386, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1386", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1386() */
/* END TEST 1386 */

/*
 *  BEGIN TEST 1387
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1387( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(340, 37), MAKE_VAL_COMPLEX(-214, 627), MAKE_VAL_COMPLEX(-843, 654), MAKE_VAL_COMPLEX(-583, -709), MAKE_VAL_COMPLEX(-374, 864), MAKE_VAL_COMPLEX(170, 47), MAKE_VAL_COMPLEX(456, -51), MAKE_VAL_COMPLEX(477, 270), MAKE_VAL_COMPLEX(520, 436), MAKE_VAL_COMPLEX(575, -203), MAKE_VAL_COMPLEX(149, -138), MAKE_VAL_COMPLEX(-817, -241), MAKE_VAL_COMPLEX(19, -913), MAKE_VAL_COMPLEX(303, 624), MAKE_VAL_COMPLEX(67, -834) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, -3);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1387, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1387", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1387() */
/* END TEST 1387 */

/*
 *  BEGIN TEST 1388
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1388( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(54, -26), MAKE_VAL_COMPLEX(-9, 59), MAKE_VAL_COMPLEX(50, -22), MAKE_VAL_COMPLEX(215, -66), MAKE_VAL_COMPLEX(40, 168), MAKE_VAL_COMPLEX(-112, 23), MAKE_VAL_COMPLEX(135, -36), MAKE_VAL_COMPLEX(50, -118), MAKE_VAL_COMPLEX(-49, -71), MAKE_VAL_COMPLEX(98, -75), MAKE_VAL_COMPLEX(-39, -27), MAKE_VAL_COMPLEX(-74, -82), MAKE_VAL_COMPLEX(-254, 18), MAKE_VAL_COMPLEX(-122, 4), MAKE_VAL_COMPLEX(-107, -56) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1388, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1388", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1388() */
/* END TEST 1388 */

/*
 *  BEGIN TEST 1389
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1389( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-126, -105), MAKE_VAL_COMPLEX(135, -86), MAKE_VAL_COMPLEX(-77, -104), MAKE_VAL_COMPLEX(-73, 182), MAKE_VAL_COMPLEX(34, -96), MAKE_VAL_COMPLEX(-7, 105), MAKE_VAL_COMPLEX(-4, -36), MAKE_VAL_COMPLEX(-77, -8), MAKE_VAL_COMPLEX(167, 75), MAKE_VAL_COMPLEX(-132, 53), MAKE_VAL_COMPLEX(-36, 87), MAKE_VAL_COMPLEX(153, 109), MAKE_VAL_COMPLEX(141, -67), MAKE_VAL_COMPLEX(63, -6), MAKE_VAL_COMPLEX(41, -47) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1389, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1389", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1389() */
/* END TEST 1389 */

/*
 *  BEGIN TEST 1390
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1390( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-49, 91), MAKE_VAL_COMPLEX(-63, -63), MAKE_VAL_COMPLEX(-63, 63), MAKE_VAL_COMPLEX(14, -126), MAKE_VAL_COMPLEX(14, -70), MAKE_VAL_COMPLEX(119, 21), MAKE_VAL_COMPLEX(-91, -49), MAKE_VAL_COMPLEX(-63, 63), MAKE_VAL_COMPLEX(-42, -70), MAKE_VAL_COMPLEX(0, 28), MAKE_VAL_COMPLEX(63, 21), MAKE_VAL_COMPLEX(-14, -56), MAKE_VAL_COMPLEX(14, 84), MAKE_VAL_COMPLEX(21, 35), MAKE_VAL_COMPLEX(-35, 77) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1390, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1390", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1390() */
/* END TEST 1390 */

/*
 *  BEGIN TEST 1391
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1344', LAYOUT_COLMAJ>
 *  y : <'y1344', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1391( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1344_ptr, A1344_ind, A1344_val,
		A1344_M, A1344_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1344_val,
		y1344_LEN * y1344_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1344_LEN, y1344_VW, LAYOUT_COLMAJ, y1344_LEN );
	oski_value_t* y_val = array_Duplicate( x1344_val,
		x1344_LEN * x1344_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1344_LEN, x1344_VW, LAYOUT_COLMAJ, x1344_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(282, 265), MAKE_VAL_COMPLEX(-365, 182), MAKE_VAL_COMPLEX(177, 232), MAKE_VAL_COMPLEX(736, -389), MAKE_VAL_COMPLEX(-98, 512), MAKE_VAL_COMPLEX(-136, -180), MAKE_VAL_COMPLEX(261, 31), MAKE_VAL_COMPLEX(273, -152), MAKE_VAL_COMPLEX(-355, -433), MAKE_VAL_COMPLEX(517, -132), MAKE_VAL_COMPLEX(66, -207), MAKE_VAL_COMPLEX(-349, -517), MAKE_VAL_COMPLEX(-849, 3), MAKE_VAL_COMPLEX(-377, -56), MAKE_VAL_COMPLEX(-322, -76) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1391, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1391", y_true, y_val, x1344_LEN*x1344_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1391() */
/* END TEST 1391 */

#define A1392_M 5    /* # of rows */
#define A1392_N 5    /* # of cols */
#define A1392_NNZ 7  /* # of stored non-zeros */
static oski_index_t A1392_ptr[A1392_M+1] = {0, 2, 3, 5, 7, 7}; /* end row pointers for A1392 */
static oski_index_t A1392_ind[A1392_NNZ] = {1, 4, 1, 3, 4, 3, 4}; /* end column indices pointers for A1392 */
static oski_value_t A1392_val[A1392_NNZ] = {MAKE_VAL_COMPLEX(-10, 7), MAKE_VAL_COMPLEX(-10, 4), MAKE_VAL_COMPLEX(10, 0), MAKE_VAL_COMPLEX(-10, 9), MAKE_VAL_COMPLEX(-8, 10), MAKE_VAL_COMPLEX(-4, 0), MAKE_VAL_COMPLEX(-6, -6)}; /* end stored non-zero values for A1392 */

#define x1392_LEN 5   /* vector length */
#define x1392_VW  3   /* vector width */
static oski_value_t x1392_val[x1392_LEN*x1392_VW] = {MAKE_VAL_COMPLEX(-9, -4), MAKE_VAL_COMPLEX(-0, -10), MAKE_VAL_COMPLEX(10, 10), MAKE_VAL_COMPLEX(-2, -4), MAKE_VAL_COMPLEX(9, -10), MAKE_VAL_COMPLEX(5, -5), MAKE_VAL_COMPLEX(5, -7), MAKE_VAL_COMPLEX(-1, -8), MAKE_VAL_COMPLEX(3, 4), MAKE_VAL_COMPLEX(-8, 9), MAKE_VAL_COMPLEX(1, 8), MAKE_VAL_COMPLEX(5, 1), MAKE_VAL_COMPLEX(-10, -6), MAKE_VAL_COMPLEX(6, -9), MAKE_VAL_COMPLEX(1, -9)}; /* data */

#define y1392_LEN 5   /* vector length */
#define y1392_VW  3   /* vector width */
static oski_value_t y1392_val[y1392_LEN*y1392_VW] = {MAKE_VAL_COMPLEX(-10, -1), MAKE_VAL_COMPLEX(4, -10), MAKE_VAL_COMPLEX(1, -6), MAKE_VAL_COMPLEX(5, 4), MAKE_VAL_COMPLEX(-10, -10), MAKE_VAL_COMPLEX(4, -10), MAKE_VAL_COMPLEX(-5, -2), MAKE_VAL_COMPLEX(5, 4), MAKE_VAL_COMPLEX(8, 2), MAKE_VAL_COMPLEX(2, 10), MAKE_VAL_COMPLEX(-1, 8), MAKE_VAL_COMPLEX(6, -6), MAKE_VAL_COMPLEX(10, -8), MAKE_VAL_COMPLEX(-6, -6), MAKE_VAL_COMPLEX(8, 7)}; /* data */

/*
 *  BEGIN TEST 1392
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1392( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	oski_value_t* y_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(123, 132), MAKE_VAL_COMPLEX(-118, 201), MAKE_VAL_COMPLEX(-39, 146), MAKE_VAL_COMPLEX(98, 34), MAKE_VAL_COMPLEX(-110, 50), MAKE_VAL_COMPLEX(-54, -296), MAKE_VAL_COMPLEX(101, -177), MAKE_VAL_COMPLEX(16, 32), MAKE_VAL_COMPLEX(73, 53), MAKE_VAL_COMPLEX(41, 107), MAKE_VAL_COMPLEX(-88, 116), MAKE_VAL_COMPLEX(-104, -101), MAKE_VAL_COMPLEX(-74, -142), MAKE_VAL_COMPLEX(-244, 162), MAKE_VAL_COMPLEX(15, -222) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1392, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1392", y_true, y_val, y1392_LEN*y1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1392() */
/* END TEST 1392 */

/*
 *  BEGIN TEST 1393
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1393( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	oski_value_t* y_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-175, -16), MAKE_VAL_COMPLEX(14, -289), MAKE_VAL_COMPLEX(-29, -182), MAKE_VAL_COMPLEX(-30, -78), MAKE_VAL_COMPLEX(-50, 30), MAKE_VAL_COMPLEX(-50, 208), MAKE_VAL_COMPLEX(-145, 229), MAKE_VAL_COMPLEX(52, -76), MAKE_VAL_COMPLEX(-17, -141), MAKE_VAL_COMPLEX(87, -91), MAKE_VAL_COMPLEX(180, -72), MAKE_VAL_COMPLEX(56, 5), MAKE_VAL_COMPLEX(18, -10), MAKE_VAL_COMPLEX(148, -114), MAKE_VAL_COMPLEX(101, 154) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1393, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1393", y_true, y_val, y1392_LEN*y1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1393() */
/* END TEST 1393 */

/*
 *  BEGIN TEST 1394
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1394( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	oski_value_t* y_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-40, -4), MAKE_VAL_COMPLEX(16, -40), MAKE_VAL_COMPLEX(4, -24), MAKE_VAL_COMPLEX(20, 16), MAKE_VAL_COMPLEX(-40, -40), MAKE_VAL_COMPLEX(16, -40), MAKE_VAL_COMPLEX(-20, -8), MAKE_VAL_COMPLEX(20, 16), MAKE_VAL_COMPLEX(32, 8), MAKE_VAL_COMPLEX(8, 40), MAKE_VAL_COMPLEX(-4, 32), MAKE_VAL_COMPLEX(24, -24), MAKE_VAL_COMPLEX(40, -32), MAKE_VAL_COMPLEX(-24, -24), MAKE_VAL_COMPLEX(32, 28) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1394, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1394", y_true, y_val, y1392_LEN*y1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1394() */
/* END TEST 1394 */

/*
 *  BEGIN TEST 1395
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1395( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	oski_value_t* y_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(878, -1062), MAKE_VAL_COMPLEX(1791, 1179), MAKE_VAL_COMPLEX(1282, 484), MAKE_VAL_COMPLEX(547, -121), MAKE_VAL_COMPLEX(100, -100), MAKE_VAL_COMPLEX(-1496, -760), MAKE_VAL_COMPLEX(-1061, -1607), MAKE_VAL_COMPLEX(287, 447), MAKE_VAL_COMPLEX(856, 182), MAKE_VAL_COMPLEX(416, 614), MAKE_VAL_COMPLEX(61, 1285), MAKE_VAL_COMPLEX(-407, 473), MAKE_VAL_COMPLEX(-306, 266), MAKE_VAL_COMPLEX(450, 1582), MAKE_VAL_COMPLEX(-1518, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1395, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1395", y_true, y_val, y1392_LEN*y1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1395() */
/* END TEST 1395 */

/*
 *  BEGIN TEST 1396
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1396( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-204, 29), MAKE_VAL_COMPLEX(-14, 234), MAKE_VAL_COMPLEX(-62, -90), MAKE_VAL_COMPLEX(121, -12), MAKE_VAL_COMPLEX(-132, -64), MAKE_VAL_COMPLEX(42, -83), MAKE_VAL_COMPLEX(-136, -208), MAKE_VAL_COMPLEX(4, 29), MAKE_VAL_COMPLEX(-70, -146), MAKE_VAL_COMPLEX(104, 125), MAKE_VAL_COMPLEX(56, -19), MAKE_VAL_COMPLEX(-194, 44), MAKE_VAL_COMPLEX(144, -68), MAKE_VAL_COMPLEX(-86, 26), MAKE_VAL_COMPLEX(-212, 102) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1396, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1396", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1396() */
/* END TEST 1396 */

/*
 *  BEGIN TEST 1397
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1397( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(140, 18), MAKE_VAL_COMPLEX(-56, -274), MAKE_VAL_COMPLEX(172, 60), MAKE_VAL_COMPLEX(-157, 10), MAKE_VAL_COMPLEX(98, -39), MAKE_VAL_COMPLEX(-57, 28), MAKE_VAL_COMPLEX(107, 145), MAKE_VAL_COMPLEX(-64, -54), MAKE_VAL_COMPLEX(110, 141), MAKE_VAL_COMPLEX(-73, -33), MAKE_VAL_COMPLEX(4, 44), MAKE_VAL_COMPLEX(221, -75), MAKE_VAL_COMPLEX(-226, 114), MAKE_VAL_COMPLEX(47, -104), MAKE_VAL_COMPLEX(153, -145) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1397, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1397", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1397() */
/* END TEST 1397 */

/*
 *  BEGIN TEST 1398
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1398( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-18, -8), MAKE_VAL_COMPLEX(0, -20), MAKE_VAL_COMPLEX(20, 20), MAKE_VAL_COMPLEX(-4, -8), MAKE_VAL_COMPLEX(18, -20), MAKE_VAL_COMPLEX(10, -10), MAKE_VAL_COMPLEX(10, -14), MAKE_VAL_COMPLEX(-2, -16), MAKE_VAL_COMPLEX(6, 8), MAKE_VAL_COMPLEX(-16, 18), MAKE_VAL_COMPLEX(2, 16), MAKE_VAL_COMPLEX(10, 2), MAKE_VAL_COMPLEX(-20, -12), MAKE_VAL_COMPLEX(12, -18), MAKE_VAL_COMPLEX(2, -18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1398, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1398", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1398() */
/* END TEST 1398 */

/*
 *  BEGIN TEST 1399
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1399( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1063, 16), MAKE_VAL_COMPLEX(-570, -1378), MAKE_VAL_COMPLEX(842, 18), MAKE_VAL_COMPLEX(-874, 293), MAKE_VAL_COMPLEX(400, -86), MAKE_VAL_COMPLEX(-399, 350), MAKE_VAL_COMPLEX(710, 930), MAKE_VAL_COMPLEX(-385, -96), MAKE_VAL_COMPLEX(722, 664), MAKE_VAL_COMPLEX(-339, -346), MAKE_VAL_COMPLEX(15, 96), MAKE_VAL_COMPLEX(1142, -672), MAKE_VAL_COMPLEX(-1008, 996), MAKE_VAL_COMPLEX(88, -458), MAKE_VAL_COMPLEX(788, -830) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1399, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1399", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1399() */
/* END TEST 1399 */

/*
 *  BEGIN TEST 1400
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1400( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-212, 118), MAKE_VAL_COMPLEX(224, 6), MAKE_VAL_COMPLEX(12, 120), MAKE_VAL_COMPLEX(119, 102), MAKE_VAL_COMPLEX(-186, -115), MAKE_VAL_COMPLEX(3, -92), MAKE_VAL_COMPLEX(-101, 25), MAKE_VAL_COMPLEX(16, -146), MAKE_VAL_COMPLEX(-110, 153), MAKE_VAL_COMPLEX(-105, 91), MAKE_VAL_COMPLEX(20, 0), MAKE_VAL_COMPLEX(-59, -167), MAKE_VAL_COMPLEX(-34, 62), MAKE_VAL_COMPLEX(-113, -124), MAKE_VAL_COMPLEX(-99, -25) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1400, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1400", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1400() */
/* END TEST 1400 */

/*
 *  BEGIN TEST 1401
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1401( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(121, -83), MAKE_VAL_COMPLEX(-294, -76), MAKE_VAL_COMPLEX(128, -120), MAKE_VAL_COMPLEX(-161, -116), MAKE_VAL_COMPLEX(179, -18), MAKE_VAL_COMPLEX(-3, 22), MAKE_VAL_COMPLEX(87, -109), MAKE_VAL_COMPLEX(-79, 97), MAKE_VAL_COMPLEX(159, -146), MAKE_VAL_COMPLEX(112, 28), MAKE_VAL_COMPLEX(43, 49), MAKE_VAL_COMPLEX(101, 139), MAKE_VAL_COMPLEX(-78, -34), MAKE_VAL_COMPLEX(92, 19), MAKE_VAL_COMPLEX(43, -45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1401, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1401", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1401() */
/* END TEST 1401 */

/*
 *  BEGIN TEST 1402
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1402( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(21, -23), MAKE_VAL_COMPLEX(30, 10), MAKE_VAL_COMPLEX(-40, 20), MAKE_VAL_COMPLEX(14, -2), MAKE_VAL_COMPLEX(21, 37), MAKE_VAL_COMPLEX(10, 20), MAKE_VAL_COMPLEX(16, 22), MAKE_VAL_COMPLEX(25, 5), MAKE_VAL_COMPLEX(-15, 5), MAKE_VAL_COMPLEX(-19, -33), MAKE_VAL_COMPLEX(-25, -5), MAKE_VAL_COMPLEX(-8, 14), MAKE_VAL_COMPLEX(28, -24), MAKE_VAL_COMPLEX(21, 27), MAKE_VAL_COMPLEX(26, 12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1402, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1402", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1402() */
/* END TEST 1402 */

/*
 *  BEGIN TEST 1403
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1403( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1291, -763), MAKE_VAL_COMPLEX(360, 1896), MAKE_VAL_COMPLEX(-844, -136), MAKE_VAL_COMPLEX(-403, 1317), MAKE_VAL_COMPLEX(-302, -1580), MAKE_VAL_COMPLEX(363, -235), MAKE_VAL_COMPLEX(-822, -550), MAKE_VAL_COMPLEX(885, -51), MAKE_VAL_COMPLEX(-1386, -516), MAKE_VAL_COMPLEX(-461, -525), MAKE_VAL_COMPLEX(245, -135), MAKE_VAL_COMPLEX(864, -1080), MAKE_VAL_COMPLEX(-388, 404), MAKE_VAL_COMPLEX(100, -1046), MAKE_VAL_COMPLEX(-396, -516) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1403, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1403", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1403() */
/* END TEST 1403 */

/*
 *  BEGIN TEST 1404
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1404( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	oski_value_t* y_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(221, -40), MAKE_VAL_COMPLEX(116, 139), MAKE_VAL_COMPLEX(189, -181), MAKE_VAL_COMPLEX(-113, 114), MAKE_VAL_COMPLEX(88, -70), MAKE_VAL_COMPLEX(30, 353), MAKE_VAL_COMPLEX(58, -37), MAKE_VAL_COMPLEX(-96, 90), MAKE_VAL_COMPLEX(-206, 139), MAKE_VAL_COMPLEX(-264, 110), MAKE_VAL_COMPLEX(-57, 138), MAKE_VAL_COMPLEX(50, -142), MAKE_VAL_COMPLEX(55, 239), MAKE_VAL_COMPLEX(4, -53), MAKE_VAL_COMPLEX(-175, -134) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1404, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1404", y_true, y_val, y1392_LEN*y1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1404() */
/* END TEST 1404 */

/*
 *  BEGIN TEST 1405
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1405( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	oski_value_t* y_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-160, 36), MAKE_VAL_COMPLEX(-130, -75), MAKE_VAL_COMPLEX(-189, 218), MAKE_VAL_COMPLEX(79, -133), MAKE_VAL_COMPLEX(-18, 120), MAKE_VAL_COMPLEX(-44, -289), MAKE_VAL_COMPLEX(-26, 44), MAKE_VAL_COMPLEX(62, -109), MAKE_VAL_COMPLEX(156, -143), MAKE_VAL_COMPLEX(242, -168), MAKE_VAL_COMPLEX(55, -187), MAKE_VAL_COMPLEX(-80, 184), MAKE_VAL_COMPLEX(-107, -181), MAKE_VAL_COMPLEX(38, 83), MAKE_VAL_COMPLEX(120, 100) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1405, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1405", y_true, y_val, y1392_LEN*y1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1405() */
/* END TEST 1405 */

/*
 *  BEGIN TEST 1406
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1406( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	oski_value_t* y_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-46, 56), MAKE_VAL_COMPLEX(-44, -64), MAKE_VAL_COMPLEX(-32, -30), MAKE_VAL_COMPLEX(44, -14), MAKE_VAL_COMPLEX(-100, 20), MAKE_VAL_COMPLEX(-44, -64), MAKE_VAL_COMPLEX(-32, 22), MAKE_VAL_COMPLEX(44, -14), MAKE_VAL_COMPLEX(44, -40), MAKE_VAL_COMPLEX(68, 28), MAKE_VAL_COMPLEX(44, 38), MAKE_VAL_COMPLEX(-12, -60), MAKE_VAL_COMPLEX(-8, -92), MAKE_VAL_COMPLEX(-60, 12), MAKE_VAL_COMPLEX(74, -20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1406, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1406", y_true, y_val, y1392_LEN*y1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1406() */
/* END TEST 1406 */

/*
 *  BEGIN TEST 1407
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1407( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	oski_value_t* y_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-173, 795), MAKE_VAL_COMPLEX(-265, 288), MAKE_VAL_COMPLEX(1107, 732), MAKE_VAL_COMPLEX(-478, -270), MAKE_VAL_COMPLEX(220, -230), MAKE_VAL_COMPLEX(-1335, -142), MAKE_VAL_COMPLEX(39, 75), MAKE_VAL_COMPLEX(-358, -185), MAKE_VAL_COMPLEX(-429, -734), MAKE_VAL_COMPLEX(-740, -866), MAKE_VAL_COMPLEX(-946, 8), MAKE_VAL_COMPLEX(1112, 172), MAKE_VAL_COMPLEX(-579, 225), MAKE_VAL_COMPLEX(187, -382), MAKE_VAL_COMPLEX(801, -379) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1407, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1407", y_true, y_val, y1392_LEN*y1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1407() */
/* END TEST 1407 */

/*
 *  BEGIN TEST 1408
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1408( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-99, 104), MAKE_VAL_COMPLEX(136, -143), MAKE_VAL_COMPLEX(-144, 86), MAKE_VAL_COMPLEX(161, -212), MAKE_VAL_COMPLEX(141, 90), MAKE_VAL_COMPLEX(99, -152), MAKE_VAL_COMPLEX(67, 72), MAKE_VAL_COMPLEX(77, -272), MAKE_VAL_COMPLEX(7, 38), MAKE_VAL_COMPLEX(0, 70), MAKE_VAL_COMPLEX(-245, 29), MAKE_VAL_COMPLEX(-57, 249), MAKE_VAL_COMPLEX(108, -160), MAKE_VAL_COMPLEX(10, 41), MAKE_VAL_COMPLEX(59, 79) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1408, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1408", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1408() */
/* END TEST 1408 */

/*
 *  BEGIN TEST 1409
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1409( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(95, -192), MAKE_VAL_COMPLEX(-56, 103), MAKE_VAL_COMPLEX(104, 34), MAKE_VAL_COMPLEX(-137, 180), MAKE_VAL_COMPLEX(-25, -58), MAKE_VAL_COMPLEX(-39, 172), MAKE_VAL_COMPLEX(9, -60), MAKE_VAL_COMPLEX(-17, 232), MAKE_VAL_COMPLEX(-27, 2), MAKE_VAL_COMPLEX(-104, -98), MAKE_VAL_COMPLEX(185, 11), MAKE_VAL_COMPLEX(69, -205), MAKE_VAL_COMPLEX(-100, 56), MAKE_VAL_COMPLEX(86, -29), MAKE_VAL_COMPLEX(17, -107) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1409, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1409", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1409() */
/* END TEST 1409 */

/*
 *  BEGIN TEST 1410
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1410( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(10, -71), MAKE_VAL_COMPLEX(70, -20), MAKE_VAL_COMPLEX(-50, 90), MAKE_VAL_COMPLEX(24, -22), MAKE_VAL_COMPLEX(88, 43), MAKE_VAL_COMPLEX(45, 25), MAKE_VAL_COMPLEX(59, 21), MAKE_VAL_COMPLEX(54, -23), MAKE_VAL_COMPLEX(-22, 29), MAKE_VAL_COMPLEX(-79, -38), MAKE_VAL_COMPLEX(-54, 23), MAKE_VAL_COMPLEX(3, 37), MAKE_VAL_COMPLEX(22, -82), MAKE_VAL_COMPLEX(75, 24), MAKE_VAL_COMPLEX(65, -11) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1410, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1410", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1410() */
/* END TEST 1410 */

/*
 *  BEGIN TEST 1411
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1411( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-426, -292), MAKE_VAL_COMPLEX(10, 125), MAKE_VAL_COMPLEX(402, -224), MAKE_VAL_COMPLEX(149, 446), MAKE_VAL_COMPLEX(-200, -224), MAKE_VAL_COMPLEX(280, 135), MAKE_VAL_COMPLEX(-164, -211), MAKE_VAL_COMPLEX(326, 208), MAKE_VAL_COMPLEX(60, 50), MAKE_VAL_COMPLEX(-246, 303), MAKE_VAL_COMPLEX(328, -301), MAKE_VAL_COMPLEX(-282, -404), MAKE_VAL_COMPLEX(-162, 342), MAKE_VAL_COMPLEX(-44, -366), MAKE_VAL_COMPLEX(-314, -236) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1411, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1411", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1411() */
/* END TEST 1411 */

/*
 *  BEGIN TEST 1412
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1412( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(262, 175), MAKE_VAL_COMPLEX(54, -27), MAKE_VAL_COMPLEX(-256, 34), MAKE_VAL_COMPLEX(169, 2), MAKE_VAL_COMPLEX(-182, 177), MAKE_VAL_COMPLEX(111, -68), MAKE_VAL_COMPLEX(89, 66), MAKE_VAL_COMPLEX(-166, 51), MAKE_VAL_COMPLEX(-64, 13), MAKE_VAL_COMPLEX(-65, 43), MAKE_VAL_COMPLEX(-46, -224), MAKE_VAL_COMPLEX(-199, -59), MAKE_VAL_COMPLEX(84, 92), MAKE_VAL_COMPLEX(-143, 140), MAKE_VAL_COMPLEX(-9, -35) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1412, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1412", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1412() */
/* END TEST 1412 */

/*
 *  BEGIN TEST 1413
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1413( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-201, -191), MAKE_VAL_COMPLEX(-14, 77), MAKE_VAL_COMPLEX(166, -44), MAKE_VAL_COMPLEX(-143, 10), MAKE_VAL_COMPLEX(177, -91), MAKE_VAL_COMPLEX(-116, 113), MAKE_VAL_COMPLEX(-86, -11), MAKE_VAL_COMPLEX(203, -15), MAKE_VAL_COMPLEX(33, -21), MAKE_VAL_COMPLEX(69, -120), MAKE_VAL_COMPLEX(9, 188), MAKE_VAL_COMPLEX(170, 74), MAKE_VAL_COMPLEX(-10, -102), MAKE_VAL_COMPLEX(149, -71), MAKE_VAL_COMPLEX(40, 84) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1413, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1413", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1413() */
/* END TEST 1413 */

/*
 *  BEGIN TEST 1414
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1414( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-93, -9), MAKE_VAL_COMPLEX(-30, -90), MAKE_VAL_COMPLEX(120, 60), MAKE_VAL_COMPLEX(-30, -30), MAKE_VAL_COMPLEX(51, -117), MAKE_VAL_COMPLEX(30, -60), MAKE_VAL_COMPLEX(24, -78), MAKE_VAL_COMPLEX(-33, -69), MAKE_VAL_COMPLEX(39, 27), MAKE_VAL_COMPLEX(-45, 105), MAKE_VAL_COMPLEX(33, 69), MAKE_VAL_COMPLEX(48, -6), MAKE_VAL_COMPLEX(-108, -24), MAKE_VAL_COMPLEX(27, -99), MAKE_VAL_COMPLEX(-18, -84) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1414, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1414", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1414() */
/* END TEST 1414 */

/*
 *  BEGIN TEST 1415
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_ROWMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1415( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_ROWMAJ, x1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-559, -941), MAKE_VAL_COMPLEX(-133, 354), MAKE_VAL_COMPLEX(648, -70), MAKE_VAL_COMPLEX(-570, -79), MAKE_VAL_COMPLEX(745, -127), MAKE_VAL_COMPLEX(-607, 366), MAKE_VAL_COMPLEX(-363, -88), MAKE_VAL_COMPLEX(833, 191), MAKE_VAL_COMPLEX(135, -75), MAKE_VAL_COMPLEX(444, -465), MAKE_VAL_COMPLEX(-158, 713), MAKE_VAL_COMPLEX(576, 460), MAKE_VAL_COMPLEX(122, -382), MAKE_VAL_COMPLEX(631, -81), MAKE_VAL_COMPLEX(70, 430) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1415, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1415", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1415() */
/* END TEST 1415 */

/*
 *  BEGIN TEST 1416
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1416( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	oski_value_t* y_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(84, 202), MAKE_VAL_COMPLEX(59, 59), MAKE_VAL_COMPLEX(-13, 159), MAKE_VAL_COMPLEX(16, -1), MAKE_VAL_COMPLEX(65, -35), MAKE_VAL_COMPLEX(112, -1), MAKE_VAL_COMPLEX(122, 184), MAKE_VAL_COMPLEX(-138, -169), MAKE_VAL_COMPLEX(47, 246), MAKE_VAL_COMPLEX(-168, -220), MAKE_VAL_COMPLEX(2, 15), MAKE_VAL_COMPLEX(-50, 294), MAKE_VAL_COMPLEX(102, -4), MAKE_VAL_COMPLEX(-124, 110), MAKE_VAL_COMPLEX(-16, 144) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1416, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1416", y_true, y_val, y1392_LEN*y1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1416() */
/* END TEST 1416 */

/*
 *  BEGIN TEST 1417
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1417( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	oski_value_t* y_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-47, -269), MAKE_VAL_COMPLEX(-1, -1), MAKE_VAL_COMPLEX(52, -134), MAKE_VAL_COMPLEX(-59, 24), MAKE_VAL_COMPLEX(35, -5), MAKE_VAL_COMPLEX(-54, 59), MAKE_VAL_COMPLEX(-93, -213), MAKE_VAL_COMPLEX(95, 192), MAKE_VAL_COMPLEX(-85, -196), MAKE_VAL_COMPLEX(92, 204), MAKE_VAL_COMPLEX(-55, -46), MAKE_VAL_COMPLEX(74, -234), MAKE_VAL_COMPLEX(-76, 98), MAKE_VAL_COMPLEX(184, -134), MAKE_VAL_COMPLEX(-57, -109) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1417, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1417", y_true, y_val, y1392_LEN*y1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1417() */
/* END TEST 1417 */

/*
 *  BEGIN TEST 1418
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1418( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	oski_value_t* y_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-18, -22), MAKE_VAL_COMPLEX(28, -12), MAKE_VAL_COMPLEX(14, -10), MAKE_VAL_COMPLEX(2, 18), MAKE_VAL_COMPLEX(0, -40), MAKE_VAL_COMPLEX(28, -12), MAKE_VAL_COMPLEX(-6, -14), MAKE_VAL_COMPLEX(2, 18), MAKE_VAL_COMPLEX(12, 20), MAKE_VAL_COMPLEX(-16, 24), MAKE_VAL_COMPLEX(-18, 14), MAKE_VAL_COMPLEX(24, 0), MAKE_VAL_COMPLEX(36, 4), MAKE_VAL_COMPLEX(0, -24), MAKE_VAL_COMPLEX(2, 30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1418, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1418", y_true, y_val, y1392_LEN*y1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1418() */
/* END TEST 1418 */

/*
 *  BEGIN TEST 1419
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1419( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	oski_value_t* y_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1186, 2131), MAKE_VAL_COMPLEX(211, -287), MAKE_VAL_COMPLEX(142, 1229), MAKE_VAL_COMPLEX(567, -289), MAKE_VAL_COMPLEX(-545, -205), MAKE_VAL_COMPLEX(448, -1039), MAKE_VAL_COMPLEX(1543, 1446), MAKE_VAL_COMPLEX(-1491, -1185), MAKE_VAL_COMPLEX(1791, 1544), MAKE_VAL_COMPLEX(-1650, -1134), MAKE_VAL_COMPLEX(591, 443), MAKE_VAL_COMPLEX(504, 2252), MAKE_VAL_COMPLEX(668, -1372), MAKE_VAL_COMPLEX(-1270, 1708), MAKE_VAL_COMPLEX(1156, 1033) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(9, -4);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1419, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1419", y_true, y_val, y1392_LEN*y1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1419() */
/* END TEST 1419 */

/*
 *  BEGIN TEST 1420
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1420( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-168, -52), MAKE_VAL_COMPLEX(167, -40), MAKE_VAL_COMPLEX(-80, -124), MAKE_VAL_COMPLEX(48, 33), MAKE_VAL_COMPLEX(240, -147), MAKE_VAL_COMPLEX(81, 249), MAKE_VAL_COMPLEX(-53, 19), MAKE_VAL_COMPLEX(76, 20), MAKE_VAL_COMPLEX(-8, -16), MAKE_VAL_COMPLEX(-51, 102), MAKE_VAL_COMPLEX(-168, -13), MAKE_VAL_COMPLEX(81, -26), MAKE_VAL_COMPLEX(-122, -152), MAKE_VAL_COMPLEX(-191, 70), MAKE_VAL_COMPLEX(-131, 111) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1420, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1420", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1420() */
/* END TEST 1420 */

/*
 *  BEGIN TEST 1421
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1421( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(128, -52), MAKE_VAL_COMPLEX(-87, -40), MAKE_VAL_COMPLEX(80, 284), MAKE_VAL_COMPLEX(-32, -81), MAKE_VAL_COMPLEX(-88, 139), MAKE_VAL_COMPLEX(-1, -249), MAKE_VAL_COMPLEX(149, -35), MAKE_VAL_COMPLEX(-20, -92), MAKE_VAL_COMPLEX(0, 72), MAKE_VAL_COMPLEX(-85, -94), MAKE_VAL_COMPLEX(112, 85), MAKE_VAL_COMPLEX(-49, 74), MAKE_VAL_COMPLEX(90, 24), MAKE_VAL_COMPLEX(311, -94), MAKE_VAL_COMPLEX(211, -175) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1421, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1421", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1421() */
/* END TEST 1421 */

/*
 *  BEGIN TEST 1422
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1422( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-47, -64), MAKE_VAL_COMPLEX(40, -70), MAKE_VAL_COMPLEX(30, 110), MAKE_VAL_COMPLEX(2, -36), MAKE_VAL_COMPLEX(103, -34), MAKE_VAL_COMPLEX(55, -15), MAKE_VAL_COMPLEX(63, -29), MAKE_VAL_COMPLEX(25, -60), MAKE_VAL_COMPLEX(5, 40), MAKE_VAL_COMPLEX(-92, 31), MAKE_VAL_COMPLEX(-25, 60), MAKE_VAL_COMPLEX(31, 27), MAKE_VAL_COMPLEX(-46, -82), MAKE_VAL_COMPLEX(78, -39), MAKE_VAL_COMPLEX(43, -59) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1422, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1422", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1422() */
/* END TEST 1422 */

/*
 *  BEGIN TEST 1423
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1423( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-6, 599), MAKE_VAL_COMPLEX(-197, -648), MAKE_VAL_COMPLEX(-526, 594), MAKE_VAL_COMPLEX(132, -259), MAKE_VAL_COMPLEX(-680, -716), MAKE_VAL_COMPLEX(990, -518), MAKE_VAL_COMPLEX(230, 244), MAKE_VAL_COMPLEX(106, -353), MAKE_VAL_COMPLEX(-102, 95), MAKE_VAL_COMPLEX(326, 16), MAKE_VAL_COMPLEX(14, 714), MAKE_VAL_COMPLEX(-188, -231), MAKE_VAL_COMPLEX(-428, 498), MAKE_VAL_COMPLEX(600, 754), MAKE_VAL_COMPLEX(694, 408) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, -4);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1423, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1423", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1423() */
/* END TEST 1423 */

/*
 *  BEGIN TEST 1424
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1424( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-86, 77), MAKE_VAL_COMPLEX(203, 160), MAKE_VAL_COMPLEX(-210, 102), MAKE_VAL_COMPLEX(-52, 17), MAKE_VAL_COMPLEX(68, 162), MAKE_VAL_COMPLEX(264, 116), MAKE_VAL_COMPLEX(-188, 30), MAKE_VAL_COMPLEX(98, -75), MAKE_VAL_COMPLEX(-46, -43), MAKE_VAL_COMPLEX(-144, -136), MAKE_VAL_COMPLEX(-130, 64), MAKE_VAL_COMPLEX(-38, -21), MAKE_VAL_COMPLEX(-64, 102), MAKE_VAL_COMPLEX(-62, -86), MAKE_VAL_COMPLEX(-28, 74) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1424, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1424", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1424() */
/* END TEST 1424 */

/*
 *  BEGIN TEST 1425
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1425( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(209, -87), MAKE_VAL_COMPLEX(-143, -50), MAKE_VAL_COMPLEX(40, -152), MAKE_VAL_COMPLEX(98, 15), MAKE_VAL_COMPLEX(-107, 2), MAKE_VAL_COMPLEX(-289, -31), MAKE_VAL_COMPLEX(175, 77), MAKE_VAL_COMPLEX(-39, 157), MAKE_VAL_COMPLEX(-11, 17), MAKE_VAL_COMPLEX(178, -11), MAKE_VAL_COMPLEX(71, -146), MAKE_VAL_COMPLEX(-23, 40), MAKE_VAL_COMPLEX(210, -96), MAKE_VAL_COMPLEX(50, 221), MAKE_VAL_COMPLEX(71, 31) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1425, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1425", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1425() */
/* END TEST 1425 */

/*
 *  BEGIN TEST 1426
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1426( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-37, -38), MAKE_VAL_COMPLEX(20, -50), MAKE_VAL_COMPLEX(30, 70), MAKE_VAL_COMPLEX(-2, -24), MAKE_VAL_COMPLEX(65, -32), MAKE_VAL_COMPLEX(35, -15), MAKE_VAL_COMPLEX(39, -25), MAKE_VAL_COMPLEX(11, -42), MAKE_VAL_COMPLEX(7, 26), MAKE_VAL_COMPLEX(-58, 29), MAKE_VAL_COMPLEX(-11, 42), MAKE_VAL_COMPLEX(23, 15), MAKE_VAL_COMPLEX(-38, -50), MAKE_VAL_COMPLEX(48, -33), MAKE_VAL_COMPLEX(23, -43) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1426, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1426", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1426() */
/* END TEST 1426 */

/*
 *  BEGIN TEST 1427
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1427( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_ROWMAJ, y1392_VW );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(40, -618), MAKE_VAL_COMPLEX(-446, -14), MAKE_VAL_COMPLEX(36, -244), MAKE_VAL_COMPLEX(162, -234), MAKE_VAL_COMPLEX(-90, -36), MAKE_VAL_COMPLEX(-570, 386), MAKE_VAL_COMPLEX(562, -366), MAKE_VAL_COMPLEX(168, 238), MAKE_VAL_COMPLEX(96, 118), MAKE_VAL_COMPLEX(228, -150), MAKE_VAL_COMPLEX(-82, -280), MAKE_VAL_COMPLEX(140, 116), MAKE_VAL_COMPLEX(-8, -672), MAKE_VAL_COMPLEX(608, 126), MAKE_VAL_COMPLEX(170, -266) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-2, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1427, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1427", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1427() */
/* END TEST 1427 */

/*
 *  BEGIN TEST 1428
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1428( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	oski_value_t* y_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(86, 182), MAKE_VAL_COMPLEX(98, 87), MAKE_VAL_COMPLEX(114, 234), MAKE_VAL_COMPLEX(-170, -162), MAKE_VAL_COMPLEX(250, -92), MAKE_VAL_COMPLEX(79, 67), MAKE_VAL_COMPLEX(7, -73), MAKE_VAL_COMPLEX(-146, -159), MAKE_VAL_COMPLEX(-32, 103), MAKE_VAL_COMPLEX(-256, 50), MAKE_VAL_COMPLEX(-73, 65), MAKE_VAL_COMPLEX(96, -5), MAKE_VAL_COMPLEX(91, 334), MAKE_VAL_COMPLEX(34, 234), MAKE_VAL_COMPLEX(-30, 160) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1428, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1428", y_true, y_val, y1392_LEN*y1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1428() */
/* END TEST 1428 */

/*
 *  BEGIN TEST 1429
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1429( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	oski_value_t* y_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(30, -231), MAKE_VAL_COMPLEX(-82, 47), MAKE_VAL_COMPLEX(-89, -162), MAKE_VAL_COMPLEX(91, 148), MAKE_VAL_COMPLEX(-80, 142), MAKE_VAL_COMPLEX(-63, 67), MAKE_VAL_COMPLEX(60, 65), MAKE_VAL_COMPLEX(67, 145), MAKE_VAL_COMPLEX(-68, -77), MAKE_VAL_COMPLEX(174, -148), MAKE_VAL_COMPLEX(36, -159), MAKE_VAL_COMPLEX(-126, 107), MAKE_VAL_COMPLEX(-153, -186), MAKE_VAL_COMPLEX(68, -204), MAKE_VAL_COMPLEX(-100, -189) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1429, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1429", y_true, y_val, y1392_LEN*y1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1429() */
/* END TEST 1429 */

/*
 *  BEGIN TEST 1430
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1430( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	oski_value_t* y_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-7, 70), MAKE_VAL_COMPLEX(-70, -28), MAKE_VAL_COMPLEX(-42, -7), MAKE_VAL_COMPLEX(28, -35), MAKE_VAL_COMPLEX(-70, 70), MAKE_VAL_COMPLEX(-70, -28), MAKE_VAL_COMPLEX(-14, 35), MAKE_VAL_COMPLEX(28, -35), MAKE_VAL_COMPLEX(14, -56), MAKE_VAL_COMPLEX(70, -14), MAKE_VAL_COMPLEX(56, 7), MAKE_VAL_COMPLEX(-42, -42), MAKE_VAL_COMPLEX(-56, -70), MAKE_VAL_COMPLEX(-42, 42), MAKE_VAL_COMPLEX(49, -56) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1430, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1430", y_true, y_val, y1392_LEN*y1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1430() */
/* END TEST 1430 */

/*
 *  BEGIN TEST 1431
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1431( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	oski_value_t* y_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1215, -2337), MAKE_VAL_COMPLEX(-438, -393), MAKE_VAL_COMPLEX(453, -2241), MAKE_VAL_COMPLEX(15, 2265), MAKE_VAL_COMPLEX(-1662, -72), MAKE_VAL_COMPLEX(-387, -99), MAKE_VAL_COMPLEX(-12, 654), MAKE_VAL_COMPLEX(-183, 2094), MAKE_VAL_COMPLEX(156, -693), MAKE_VAL_COMPLEX(2160, 270), MAKE_VAL_COMPLEX(918, -870), MAKE_VAL_COMPLEX(-1254, 153), MAKE_VAL_COMPLEX(531, -2586), MAKE_VAL_COMPLEX(1782, -1950), MAKE_VAL_COMPLEX(345, -1653) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-9, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1431, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1431", y_true, y_val, y1392_LEN*y1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1431() */
/* END TEST 1431 */

/*
 *  BEGIN TEST 1432
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1432( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-154, 252), MAKE_VAL_COMPLEX(67, -240), MAKE_VAL_COMPLEX(126, 95), MAKE_VAL_COMPLEX(96, 37), MAKE_VAL_COMPLEX(142, -178), MAKE_VAL_COMPLEX(56, -133), MAKE_VAL_COMPLEX(-36, 12), MAKE_VAL_COMPLEX(-50, -248), MAKE_VAL_COMPLEX(-134, -43), MAKE_VAL_COMPLEX(-108, 210), MAKE_VAL_COMPLEX(-82, -28), MAKE_VAL_COMPLEX(62, -179), MAKE_VAL_COMPLEX(-116, 10), MAKE_VAL_COMPLEX(-118, 80), MAKE_VAL_COMPLEX(-86, 72) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1432, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1432", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1432() */
/* END TEST 1432 */

/*
 *  BEGIN TEST 1433
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1433( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(22, -149), MAKE_VAL_COMPLEX(-217, 160), MAKE_VAL_COMPLEX(104, -165), MAKE_VAL_COMPLEX(-172, -39), MAKE_VAL_COMPLEX(-220, -37), MAKE_VAL_COMPLEX(-91, 18), MAKE_VAL_COMPLEX(-29, -143), MAKE_VAL_COMPLEX(-78, 199), MAKE_VAL_COMPLEX(218, 30), MAKE_VAL_COMPLEX(179, -18), MAKE_VAL_COMPLEX(210, 77), MAKE_VAL_COMPLEX(-7, 112), MAKE_VAL_COMPLEX(-54, 92), MAKE_VAL_COMPLEX(31, -242), MAKE_VAL_COMPLEX(-41, -159) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1433, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1433", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1433() */
/* END TEST 1433 */

/*
 *  BEGIN TEST 1434
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1434( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(53, 2), MAKE_VAL_COMPLEX(20, 50), MAKE_VAL_COMPLEX(-70, -30), MAKE_VAL_COMPLEX(18, 16), MAKE_VAL_COMPLEX(-25, 68), MAKE_VAL_COMPLEX(-15, 35), MAKE_VAL_COMPLEX(-11, 45), MAKE_VAL_COMPLEX(21, 38), MAKE_VAL_COMPLEX(-23, -14), MAKE_VAL_COMPLEX(22, -61), MAKE_VAL_COMPLEX(-21, -38), MAKE_VAL_COMPLEX(-27, 5), MAKE_VAL_COMPLEX(62, 10), MAKE_VAL_COMPLEX(-12, 57), MAKE_VAL_COMPLEX(13, 47) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1434, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1434", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1434() */
/* END TEST 1434 */

/*
 *  BEGIN TEST 1435
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1435( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-617, -1438), MAKE_VAL_COMPLEX(-302, 1498), MAKE_VAL_COMPLEX(-66, -716), MAKE_VAL_COMPLEX(-1110, 250), MAKE_VAL_COMPLEX(-811, 652), MAKE_VAL_COMPLEX(-129, 487), MAKE_VAL_COMPLEX(-329, -793), MAKE_VAL_COMPLEX(583, 1206), MAKE_VAL_COMPLEX(1383, -452), MAKE_VAL_COMPLEX(414, -815), MAKE_VAL_COMPLEX(1313, -78), MAKE_VAL_COMPLEX(535, 913), MAKE_VAL_COMPLEX(-70, 210), MAKE_VAL_COMPLEX(-260, -1657), MAKE_VAL_COMPLEX(-525, -1051) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1435, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1435", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1435() */
/* END TEST 1435 */

/*
 *  BEGIN TEST 1436
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1436( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(95, 187), MAKE_VAL_COMPLEX(103, -90), MAKE_VAL_COMPLEX(194, 25), MAKE_VAL_COMPLEX(-110, 133), MAKE_VAL_COMPLEX(7, -3), MAKE_VAL_COMPLEX(24, -157), MAKE_VAL_COMPLEX(-146, -12), MAKE_VAL_COMPLEX(-245, -61), MAKE_VAL_COMPLEX(35, -146), MAKE_VAL_COMPLEX(-169, 125), MAKE_VAL_COMPLEX(-95, 117), MAKE_VAL_COMPLEX(164, -141), MAKE_VAL_COMPLEX(-108, 18), MAKE_VAL_COMPLEX(-139, -157), MAKE_VAL_COMPLEX(-66, -178) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1436, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1436", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1436() */
/* END TEST 1436 */

/*
 *  BEGIN TEST 1437
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1437( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-116, -164), MAKE_VAL_COMPLEX(-133, 80), MAKE_VAL_COMPLEX(-154, -45), MAKE_VAL_COMPLEX(96, -131), MAKE_VAL_COMPLEX(-28, -34), MAKE_VAL_COMPLEX(-34, 137), MAKE_VAL_COMPLEX(130, -10), MAKE_VAL_COMPLEX(220, 56), MAKE_VAL_COMPLEX(-20, 141), MAKE_VAL_COMPLEX(188, -92), MAKE_VAL_COMPLEX(120, -112), MAKE_VAL_COMPLEX(-156, 127), MAKE_VAL_COMPLEX(80, 6), MAKE_VAL_COMPLEX(118, 130), MAKE_VAL_COMPLEX(40, 166) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1437, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1437", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1437() */
/* END TEST 1437 */

/*
 *  BEGIN TEST 1438
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1438( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-61, -81), MAKE_VAL_COMPLEX(50, -90), MAKE_VAL_COMPLEX(40, 140), MAKE_VAL_COMPLEX(2, -46), MAKE_VAL_COMPLEX(131, -45), MAKE_VAL_COMPLEX(70, -20), MAKE_VAL_COMPLEX(80, -38), MAKE_VAL_COMPLEX(31, -77), MAKE_VAL_COMPLEX(7, 51), MAKE_VAL_COMPLEX(-117, 41), MAKE_VAL_COMPLEX(-31, 77), MAKE_VAL_COMPLEX(40, 34), MAKE_VAL_COMPLEX(-60, -104), MAKE_VAL_COMPLEX(99, -51), MAKE_VAL_COMPLEX(54, -76) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1438, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1438", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1438() */
/* END TEST 1438 */

/*
 *  BEGIN TEST 1439
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1392', LAYOUT_COLMAJ>
 *  y : <'y1392', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1439( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1392_ptr, A1392_ind, A1392_val,
		A1392_M, A1392_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1392_val,
		y1392_LEN * y1392_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1392_LEN, y1392_VW, LAYOUT_COLMAJ, y1392_LEN );
	oski_value_t* y_val = array_Duplicate( x1392_val,
		x1392_LEN * x1392_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1392_LEN, x1392_VW, LAYOUT_COLMAJ, x1392_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-2281, 974), MAKE_VAL_COMPLEX(-259, 1187), MAKE_VAL_COMPLEX(-37, 921), MAKE_VAL_COMPLEX(-1173, -1241), MAKE_VAL_COMPLEX(-745, -622), MAKE_VAL_COMPLEX(966, 302), MAKE_VAL_COMPLEX(19, -1665), MAKE_VAL_COMPLEX(675, -1954), MAKE_VAL_COMPLEX(1516, 529), MAKE_VAL_COMPLEX(58, -1279), MAKE_VAL_COMPLEX(-159, -1274), MAKE_VAL_COMPLEX(858, 1520), MAKE_VAL_COMPLEX(-304, -272), MAKE_VAL_COMPLEX(1152, -1245), MAKE_VAL_COMPLEX(1117, -145) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-3, 9);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1439, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1439", y_true, y_val, x1392_LEN*x1392_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1439() */
/* END TEST 1439 */

#define A1440_M 5    /* # of rows */
#define A1440_N 5    /* # of cols */
#define A1440_NNZ 5  /* # of stored non-zeros */
static oski_index_t A1440_ptr[A1440_M+1] = {1, 2, 3, 5, 6, 6}; /* end row pointers for A1440 */
static oski_index_t A1440_ind[A1440_NNZ] = {4, 4, 4, 5, 5}; /* end column indices pointers for A1440 */
static oski_value_t A1440_val[A1440_NNZ] = {MAKE_VAL_COMPLEX(3, 6), MAKE_VAL_COMPLEX(-10, 9), MAKE_VAL_COMPLEX(-8, 5), MAKE_VAL_COMPLEX(2, -10), MAKE_VAL_COMPLEX(-1, -5)}; /* end stored non-zero values for A1440 */

#define x1440_LEN 5   /* vector length */
#define x1440_VW  3   /* vector width */
static oski_value_t x1440_val[x1440_LEN*x1440_VW] = {MAKE_VAL_COMPLEX(8, 4), MAKE_VAL_COMPLEX(10, -1), MAKE_VAL_COMPLEX(3, 4), MAKE_VAL_COMPLEX(-2, 4), MAKE_VAL_COMPLEX(1, 0), MAKE_VAL_COMPLEX(-3, 10), MAKE_VAL_COMPLEX(-10, 3), MAKE_VAL_COMPLEX(3, -5), MAKE_VAL_COMPLEX(1, 10), MAKE_VAL_COMPLEX(-3, -1), MAKE_VAL_COMPLEX(-5, 10), MAKE_VAL_COMPLEX(5, 10), MAKE_VAL_COMPLEX(0, -10), MAKE_VAL_COMPLEX(-8, -3), MAKE_VAL_COMPLEX(6, -9)}; /* data */

#define y1440_LEN 5   /* vector length */
#define y1440_VW  3   /* vector width */
static oski_value_t y1440_val[y1440_LEN*y1440_VW] = {MAKE_VAL_COMPLEX(5, 8), MAKE_VAL_COMPLEX(-8, -6), MAKE_VAL_COMPLEX(-5, -2), MAKE_VAL_COMPLEX(-10, 8), MAKE_VAL_COMPLEX(-9, -7), MAKE_VAL_COMPLEX(4, -5), MAKE_VAL_COMPLEX(-8, 10), MAKE_VAL_COMPLEX(7, 2), MAKE_VAL_COMPLEX(10, 8), MAKE_VAL_COMPLEX(-9, 7), MAKE_VAL_COMPLEX(9, -9), MAKE_VAL_COMPLEX(1, -10), MAKE_VAL_COMPLEX(8, 0), MAKE_VAL_COMPLEX(3, 0), MAKE_VAL_COMPLEX(-1, 0)}; /* data */

/*
 *  BEGIN TEST 1440
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1440( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	oski_value_t* y_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(39, -105), MAKE_VAL_COMPLEX(-65, 99), MAKE_VAL_COMPLEX(-28, 116), MAKE_VAL_COMPLEX(161, 19), MAKE_VAL_COMPLEX(-41, -31), MAKE_VAL_COMPLEX(-207, -47), MAKE_VAL_COMPLEX(47, -20), MAKE_VAL_COMPLEX(-79, -104), MAKE_VAL_COMPLEX(-163, -251), MAKE_VAL_COMPLEX(256, 7), MAKE_VAL_COMPLEX(-173, -12), MAKE_VAL_COMPLEX(63, -123), MAKE_VAL_COMPLEX(-90, -182), MAKE_VAL_COMPLEX(-15, -42), MAKE_VAL_COMPLEX(-141, 44) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1440, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1440", y_true, y_val, y1440_LEN*y1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1440() */
/* END TEST 1440 */

/*
 *  BEGIN TEST 1441
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1441( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	oski_value_t* y_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-82, 54), MAKE_VAL_COMPLEX(127, -65), MAKE_VAL_COMPLEX(65, -107), MAKE_VAL_COMPLEX(-99, -85), MAKE_VAL_COMPLEX(111, 71), MAKE_VAL_COMPLEX(184, 86), MAKE_VAL_COMPLEX(-1, -58), MAKE_VAL_COMPLEX(28, 97), MAKE_VAL_COMPLEX(85, 205), MAKE_VAL_COMPLEX(-200, -65), MAKE_VAL_COMPLEX(119, 84), MAKE_VAL_COMPLEX(-60, 194), MAKE_VAL_COMPLEX(34, 190), MAKE_VAL_COMPLEX(-6, 45), MAKE_VAL_COMPLEX(148, -45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1441, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1441", y_true, y_val, y1440_LEN*y1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1441() */
/* END TEST 1441 */

/*
 *  BEGIN TEST 1442
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1442( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	oski_value_t* y_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(82, -29), MAKE_VAL_COMPLEX(-70, 60), MAKE_VAL_COMPLEX(-28, 41), MAKE_VAL_COMPLEX(52, 106), MAKE_VAL_COMPLEX(-81, 67), MAKE_VAL_COMPLEX(-37, -46), MAKE_VAL_COMPLEX(74, 92), MAKE_VAL_COMPLEX(32, -59), MAKE_VAL_COMPLEX(92, -74), MAKE_VAL_COMPLEX(45, 95), MAKE_VAL_COMPLEX(-63, -99), MAKE_VAL_COMPLEX(-88, -29), MAKE_VAL_COMPLEX(16, -72), MAKE_VAL_COMPLEX(6, -27), MAKE_VAL_COMPLEX(-2, 9) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1442, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1442", y_true, y_val, y1440_LEN*y1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1442() */
/* END TEST 1442 */

/*
 *  BEGIN TEST 1443
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1443( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	oski_value_t* y_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-110, -6), MAKE_VAL_COMPLEX(185, 203), MAKE_VAL_COMPLEX(307, -7), MAKE_VAL_COMPLEX(-87, -155), MAKE_VAL_COMPLEX(-291, 415), MAKE_VAL_COMPLEX(146, 556), MAKE_VAL_COMPLEX(100, 217), MAKE_VAL_COMPLEX(-43, 242), MAKE_VAL_COMPLEX(-109, 737), MAKE_VAL_COMPLEX(-337, -454), MAKE_VAL_COMPLEX(94, 201), MAKE_VAL_COMPLEX(-798, -44), MAKE_VAL_COMPLEX(-310, 386), MAKE_VAL_COMPLEX(-75, 36), MAKE_VAL_COMPLEX(407, 366) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-2, -3);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1443, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1443", y_true, y_val, y1440_LEN*y1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1443() */
/* END TEST 1443 */

/*
 *  BEGIN TEST 1444
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1444( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(48, 187), MAKE_VAL_COMPLEX(54, -5), MAKE_VAL_COMPLEX(-74, 21), MAKE_VAL_COMPLEX(91, 33), MAKE_VAL_COMPLEX(-172, 11), MAKE_VAL_COMPLEX(-210, 139), MAKE_VAL_COMPLEX(8, 13), MAKE_VAL_COMPLEX(-35, 46), MAKE_VAL_COMPLEX(-132, 162), MAKE_VAL_COMPLEX(-23, -224), MAKE_VAL_COMPLEX(-25, -17), MAKE_VAL_COMPLEX(-166, 146), MAKE_VAL_COMPLEX(226, 58), MAKE_VAL_COMPLEX(-54, -198), MAKE_VAL_COMPLEX(177, -97) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1444, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1444", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1444() */
/* END TEST 1444 */

/*
 *  BEGIN TEST 1445
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1445( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-24, -55), MAKE_VAL_COMPLEX(48, 116), MAKE_VAL_COMPLEX(53, 51), MAKE_VAL_COMPLEX(-157, -21), MAKE_VAL_COMPLEX(181, 1), MAKE_VAL_COMPLEX(63, -85), MAKE_VAL_COMPLEX(-134, -106), MAKE_VAL_COMPLEX(122, -55), MAKE_VAL_COMPLEX(21, -60), MAKE_VAL_COMPLEX(8, 179), MAKE_VAL_COMPLEX(-140, 47), MAKE_VAL_COMPLEX(91, 4), MAKE_VAL_COMPLEX(-106, -148), MAKE_VAL_COMPLEX(18, 75), MAKE_VAL_COMPLEX(-15, 88) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1445, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1445", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1445() */
/* END TEST 1445 */

/*
 *  BEGIN TEST 1446
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1446( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(24, -68), MAKE_VAL_COMPLEX(-18, -79), MAKE_VAL_COMPLEX(29, -28), MAKE_VAL_COMPLEX(34, 12), MAKE_VAL_COMPLEX(-1, -8), MAKE_VAL_COMPLEX(83, 14), MAKE_VAL_COMPLEX(34, 77), MAKE_VAL_COMPLEX(-43, -19), MAKE_VAL_COMPLEX(79, -18), MAKE_VAL_COMPLEX(-5, 25), MAKE_VAL_COMPLEX(85, 30), MAKE_VAL_COMPLEX(75, -50), MAKE_VAL_COMPLEX(-80, 10), MAKE_VAL_COMPLEX(-16, 67), MAKE_VAL_COMPLEX(-78, -39) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1446, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1446", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1446() */
/* END TEST 1446 */

/*
 *  BEGIN TEST 1447
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1447( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-524, 578), MAKE_VAL_COMPLEX(457, -890), MAKE_VAL_COMPLEX(-38, -759), MAKE_VAL_COMPLEX(746, 1268), MAKE_VAL_COMPLEX(-1098, -1437), MAKE_VAL_COMPLEX(-1168, -245), MAKE_VAL_COMPLEX(105, 1478), MAKE_VAL_COMPLEX(-1147, -495), MAKE_VAL_COMPLEX(-796, 9), MAKE_VAL_COMPLEX(1461, -1169), MAKE_VAL_COMPLEX(1146, 553), MAKE_VAL_COMPLEX(-784, -867), MAKE_VAL_COMPLEX(-378, 1936), MAKE_VAL_COMPLEX(703, -670), MAKE_VAL_COMPLEX(827, -126) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1447, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1447", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1447() */
/* END TEST 1447 */

/*
 *  BEGIN TEST 1448
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1448( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(8, -49), MAKE_VAL_COMPLEX(127, -45), MAKE_VAL_COMPLEX(100, -20), MAKE_VAL_COMPLEX(29, -107), MAKE_VAL_COMPLEX(-12, 158), MAKE_VAL_COMPLEX(126, 182), MAKE_VAL_COMPLEX(3, -93), MAKE_VAL_COMPLEX(-26, 41), MAKE_VAL_COMPLEX(116, 157), MAKE_VAL_COMPLEX(308, -57), MAKE_VAL_COMPLEX(-43, 196), MAKE_VAL_COMPLEX(-60, -51), MAKE_VAL_COMPLEX(-194, -172), MAKE_VAL_COMPLEX(-33, 158), MAKE_VAL_COMPLEX(-30, 41) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1448, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1448", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1448() */
/* END TEST 1448 */

/*
 *  BEGIN TEST 1449
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1449( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(148, 17), MAKE_VAL_COMPLEX(2, -79), MAKE_VAL_COMPLEX(-14, 43), MAKE_VAL_COMPLEX(-13, 185), MAKE_VAL_COMPLEX(26, -169), MAKE_VAL_COMPLEX(-58, -9), MAKE_VAL_COMPLEX(-110, 245), MAKE_VAL_COMPLEX(13, -144), MAKE_VAL_COMPLEX(8, -28), MAKE_VAL_COMPLEX(-361, 76), MAKE_VAL_COMPLEX(83, -1), MAKE_VAL_COMPLEX(240, 136), MAKE_VAL_COMPLEX(84, 32), MAKE_VAL_COMPLEX(-112, -112), MAKE_VAL_COMPLEX(15, -233) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1449, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1449", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1449() */
/* END TEST 1449 */

/*
 *  BEGIN TEST 1450
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1450( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-48, -4), MAKE_VAL_COMPLEX(-48, 25), MAKE_VAL_COMPLEX(-23, -14), MAKE_VAL_COMPLEX(2, -24), MAKE_VAL_COMPLEX(-5, 2), MAKE_VAL_COMPLEX(-5, -56), MAKE_VAL_COMPLEX(44, -35), MAKE_VAL_COMPLEX(-5, 31), MAKE_VAL_COMPLEX(-25, -48), MAKE_VAL_COMPLEX(17, -1), MAKE_VAL_COMPLEX(5, -60), MAKE_VAL_COMPLEX(-45, -40), MAKE_VAL_COMPLEX(20, 50), MAKE_VAL_COMPLEX(46, -1), MAKE_VAL_COMPLEX(-12, 57) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1450, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1450", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1450() */
/* END TEST 1450 */

/*
 *  BEGIN TEST 1451
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1451( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-354, -500), MAKE_VAL_COMPLEX(516, 681), MAKE_VAL_COMPLEX(603, 184), MAKE_VAL_COMPLEX(952, -1038), MAKE_VAL_COMPLEX(-1123, 1212), MAKE_VAL_COMPLEX(-47, 1362), MAKE_VAL_COMPLEX(1312, -1163), MAKE_VAL_COMPLEX(-591, 627), MAKE_VAL_COMPLEX(-293, 1182), MAKE_VAL_COMPLEX(3063, 1411), MAKE_VAL_COMPLEX(-1325, 420), MAKE_VAL_COMPLEX(-769, -1458), MAKE_VAL_COMPLEX(-320, -1750), MAKE_VAL_COMPLEX(-520, 1143), MAKE_VAL_COMPLEX(-780, 979) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1451, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1451", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1451() */
/* END TEST 1451 */

/*
 *  BEGIN TEST 1452
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1452( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	oski_value_t* y_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(12, 12), MAKE_VAL_COMPLEX(19, -39), MAKE_VAL_COMPLEX(-94, -35), MAKE_VAL_COMPLEX(108, -9), MAKE_VAL_COMPLEX(-62, -148), MAKE_VAL_COMPLEX(-48, -12), MAKE_VAL_COMPLEX(-73, -123), MAKE_VAL_COMPLEX(-34, -23), MAKE_VAL_COMPLEX(-25, 40), MAKE_VAL_COMPLEX(-31, -6), MAKE_VAL_COMPLEX(-6, 46), MAKE_VAL_COMPLEX(-130, -74), MAKE_VAL_COMPLEX(-143, -115), MAKE_VAL_COMPLEX(158, -172), MAKE_VAL_COMPLEX(-150, 35) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1452, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1452", y_true, y_val, y1440_LEN*y1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1452() */
/* END TEST 1452 */

/*
 *  BEGIN TEST 1453
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1453( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	oski_value_t* y_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-22, -28), MAKE_VAL_COMPLEX(-3, 51), MAKE_VAL_COMPLEX(104, 39), MAKE_VAL_COMPLEX(-88, -7), MAKE_VAL_COMPLEX(80, 162), MAKE_VAL_COMPLEX(40, 22), MAKE_VAL_COMPLEX(89, 103), MAKE_VAL_COMPLEX(20, 19), MAKE_VAL_COMPLEX(5, -56), MAKE_VAL_COMPLEX(49, -8), MAKE_VAL_COMPLEX(-12, -28), MAKE_VAL_COMPLEX(128, 94), MAKE_VAL_COMPLEX(127, 115), MAKE_VAL_COMPLEX(-164, 172), MAKE_VAL_COMPLEX(152, -35) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1453, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1453", y_true, y_val, y1440_LEN*y1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1453() */
/* END TEST 1453 */

/*
 *  BEGIN TEST 1454
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1454( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	oski_value_t* y_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-3, 102), MAKE_VAL_COMPLEX(-36, -102), MAKE_VAL_COMPLEX(-33, -48), MAKE_VAL_COMPLEX(-138, 12), MAKE_VAL_COMPLEX(-39, -117), MAKE_VAL_COMPLEX(66, -21), MAKE_VAL_COMPLEX(-132, 42), MAKE_VAL_COMPLEX(51, 60), MAKE_VAL_COMPLEX(42, 132), MAKE_VAL_COMPLEX(-123, 9), MAKE_VAL_COMPLEX(135, -27), MAKE_VAL_COMPLEX(69, -84), MAKE_VAL_COMPLEX(72, 48), MAKE_VAL_COMPLEX(27, 18), MAKE_VAL_COMPLEX(-9, -6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1454, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1454", y_true, y_val, y1440_LEN*y1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1454() */
/* END TEST 1454 */

/*
 *  BEGIN TEST 1455
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1455( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	oski_value_t* y_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(178, -74), MAKE_VAL_COMPLEX(287, 199), MAKE_VAL_COMPLEX(-329, -698), MAKE_VAL_COMPLEX(1051, 922), MAKE_VAL_COMPLEX(544, -982), MAKE_VAL_COMPLEX(-394, -478), MAKE_VAL_COMPLEX(803, -1109), MAKE_VAL_COMPLEX(-53, -606), MAKE_VAL_COMPLEX(-290, -315), MAKE_VAL_COMPLEX(156, -53), MAKE_VAL_COMPLEX(-718, 54), MAKE_VAL_COMPLEX(-572, -1284), MAKE_VAL_COMPLEX(-133, -1931), MAKE_VAL_COMPLEX(2122, -16), MAKE_VAL_COMPLEX(-1135, -810) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1455, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1455", y_true, y_val, y1440_LEN*y1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1455() */
/* END TEST 1455 */

/*
 *  BEGIN TEST 1456
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1456( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-21, 120), MAKE_VAL_COMPLEX(57, 12), MAKE_VAL_COMPLEX(-11, -24), MAKE_VAL_COMPLEX(150, -18), MAKE_VAL_COMPLEX(-39, -155), MAKE_VAL_COMPLEX(-70, -82), MAKE_VAL_COMPLEX(182, -179), MAKE_VAL_COMPLEX(-105, -158), MAKE_VAL_COMPLEX(-71, -50), MAKE_VAL_COMPLEX(199, -9), MAKE_VAL_COMPLEX(-85, -243), MAKE_VAL_COMPLEX(36, 166), MAKE_VAL_COMPLEX(61, 111), MAKE_VAL_COMPLEX(94, -148), MAKE_VAL_COMPLEX(39, -38) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1456, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1456", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1456() */
/* END TEST 1456 */

/*
 *  BEGIN TEST 1457
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1457( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-75, -88), MAKE_VAL_COMPLEX(-129, 76), MAKE_VAL_COMPLEX(-45, 16), MAKE_VAL_COMPLEX(-166, -30), MAKE_VAL_COMPLEX(31, 163), MAKE_VAL_COMPLEX(14, -22), MAKE_VAL_COMPLEX(-126, 75), MAKE_VAL_COMPLEX(121, 222), MAKE_VAL_COMPLEX(-17, -22), MAKE_VAL_COMPLEX(-167, -7), MAKE_VAL_COMPLEX(45, 123), MAKE_VAL_COMPLEX(-156, -206), MAKE_VAL_COMPLEX(19, -31), MAKE_VAL_COMPLEX(-6, 108), MAKE_VAL_COMPLEX(-15, 158) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1457, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1457", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1457() */
/* END TEST 1457 */

/*
 *  BEGIN TEST 1458
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1458( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-76, 52), MAKE_VAL_COMPLEX(-41, 95), MAKE_VAL_COMPLEX(-51, 7), MAKE_VAL_COMPLEX(-26, -38), MAKE_VAL_COMPLEX(-5, 9), MAKE_VAL_COMPLEX(-75, -77), MAKE_VAL_COMPLEX(23, -105), MAKE_VAL_COMPLEX(30, 52), MAKE_VAL_COMPLEX(-95, -41), MAKE_VAL_COMPLEX(24, -22), MAKE_VAL_COMPLEX(-65, -95), MAKE_VAL_COMPLEX(-115, -5), MAKE_VAL_COMPLEX(90, 50), MAKE_VAL_COMPLEX(67, -57), MAKE_VAL_COMPLEX(51, 99) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1458, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1458", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1458() */
/* END TEST 1458 */

/*
 *  BEGIN TEST 1459
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1459( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(384, -555), MAKE_VAL_COMPLEX(-561, -324), MAKE_VAL_COMPLEX(-181, -14), MAKE_VAL_COMPLEX(-600, -828), MAKE_VAL_COMPLEX(-658, 812), MAKE_VAL_COMPLEX(17, 297), MAKE_VAL_COMPLEX(-1224, -281), MAKE_VAL_COMPLEX(-502, 1343), MAKE_VAL_COMPLEX(25, 162), MAKE_VAL_COMPLEX(-727, -911), MAKE_VAL_COMPLEX(-650, 1022), MAKE_VAL_COMPLEX(521, -1249), MAKE_VAL_COMPLEX(281, -359), MAKE_VAL_COMPLEX(-813, 263), MAKE_VAL_COMPLEX(-607, 290) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1459, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1459", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1459() */
/* END TEST 1459 */

/*
 *  BEGIN TEST 1460
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1460( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-45, 36), MAKE_VAL_COMPLEX(51, 123), MAKE_VAL_COMPLEX(16, 48), MAKE_VAL_COMPLEX(-12, -162), MAKE_VAL_COMPLEX(-174, 25), MAKE_VAL_COMPLEX(-97, 62), MAKE_VAL_COMPLEX(-128, -72), MAKE_VAL_COMPLEX(-18, 77), MAKE_VAL_COMPLEX(-62, 90), MAKE_VAL_COMPLEX(70, 206), MAKE_VAL_COMPLEX(80, -37), MAKE_VAL_COMPLEX(-207, 60), MAKE_VAL_COMPLEX(21, -179), MAKE_VAL_COMPLEX(-98, 65), MAKE_VAL_COMPLEX(93, 71) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1460, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1460", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1460() */
/* END TEST 1460 */

/*
 *  BEGIN TEST 1461
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1461( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(13, -32), MAKE_VAL_COMPLEX(-79, -100), MAKE_VAL_COMPLEX(-33, -54), MAKE_VAL_COMPLEX(10, 146), MAKE_VAL_COMPLEX(171, -23), MAKE_VAL_COMPLEX(86, -98), MAKE_VAL_COMPLEX(152, 43), MAKE_VAL_COMPLEX(19, -56), MAKE_VAL_COMPLEX(39, -118), MAKE_VAL_COMPLEX(-59, -209), MAKE_VAL_COMPLEX(-85, -3), MAKE_VAL_COMPLEX(172, -80), MAKE_VAL_COMPLEX(-1, 209), MAKE_VAL_COMPLEX(128, -72), MAKE_VAL_COMPLEX(-93, -32) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1461, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1461", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1461() */
/* END TEST 1461 */

/*
 *  BEGIN TEST 1462
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1462( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(52, -44), MAKE_VAL_COMPLEX(23, -73), MAKE_VAL_COMPLEX(37, -9), MAKE_VAL_COMPLEX(22, 26), MAKE_VAL_COMPLEX(3, -7), MAKE_VAL_COMPLEX(61, 51), MAKE_VAL_COMPLEX(-9, 79), MAKE_VAL_COMPLEX(-26, -36), MAKE_VAL_COMPLEX(73, 23), MAKE_VAL_COMPLEX(-16, 18), MAKE_VAL_COMPLEX(55, 65), MAKE_VAL_COMPLEX(85, -5), MAKE_VAL_COMPLEX(-70, -30), MAKE_VAL_COMPLEX(-45, 47), MAKE_VAL_COMPLEX(-45, -69) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1462, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1462", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1462() */
/* END TEST 1462 */

/*
 *  BEGIN TEST 1463
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_ROWMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1463( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_ROWMAJ, x1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(247, 14), MAKE_VAL_COMPLEX(-97, -176), MAKE_VAL_COMPLEX(-105, 40), MAKE_VAL_COMPLEX(254, 554), MAKE_VAL_COMPLEX(501, -411), MAKE_VAL_COMPLEX(-40, -126), MAKE_VAL_COMPLEX(202, -73), MAKE_VAL_COMPLEX(47, -376), MAKE_VAL_COMPLEX(-85, -92), MAKE_VAL_COMPLEX(-697, -543), MAKE_VAL_COMPLEX(-431, 501), MAKE_VAL_COMPLEX(526, -244), MAKE_VAL_COMPLEX(415, 289), MAKE_VAL_COMPLEX(-32, -574), MAKE_VAL_COMPLEX(-139, -216) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-3, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1463, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1463", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1463() */
/* END TEST 1463 */

/*
 *  BEGIN TEST 1464
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1464( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	oski_value_t* y_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-101, 20), MAKE_VAL_COMPLEX(12, 0), MAKE_VAL_COMPLEX(20, -81), MAKE_VAL_COMPLEX(-40, -163), MAKE_VAL_COMPLEX(-27, -139), MAKE_VAL_COMPLEX(140, 15), MAKE_VAL_COMPLEX(-55, -142), MAKE_VAL_COMPLEX(-108, -2), MAKE_VAL_COMPLEX(-93, -48), MAKE_VAL_COMPLEX(-97, -257), MAKE_VAL_COMPLEX(173, 258), MAKE_VAL_COMPLEX(53, 9), MAKE_VAL_COMPLEX(-75, 88), MAKE_VAL_COMPLEX(-7, 38), MAKE_VAL_COMPLEX(132, -74) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1464, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1464", y_true, y_val, y1440_LEN*y1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1464() */
/* END TEST 1464 */

/*
 *  BEGIN TEST 1465
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1465( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	oski_value_t* y_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(78, -39), MAKE_VAL_COMPLEX(18, 10), MAKE_VAL_COMPLEX(-3, 82), MAKE_VAL_COMPLEX(62, 129), MAKE_VAL_COMPLEX(61, 151), MAKE_VAL_COMPLEX(-147, 4), MAKE_VAL_COMPLEX(69, 104), MAKE_VAL_COMPLEX(85, 3), MAKE_VAL_COMPLEX(55, 34), MAKE_VAL_COMPLEX(117, 227), MAKE_VAL_COMPLEX(-191, -222), MAKE_VAL_COMPLEX(-46, 22), MAKE_VAL_COMPLEX(51, -80), MAKE_VAL_COMPLEX(-2, -35), MAKE_VAL_COMPLEX(-129, 73) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1465, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1465", y_true, y_val, y1440_LEN*y1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1465() */
/* END TEST 1465 */

/*
 *  BEGIN TEST 1466
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1466( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	oski_value_t* y_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(8, 84), MAKE_VAL_COMPLEX(-40, -80), MAKE_VAL_COMPLEX(-32, -36), MAKE_VAL_COMPLEX(-112, 24), MAKE_VAL_COMPLEX(-44, -92), MAKE_VAL_COMPLEX(52, -24), MAKE_VAL_COMPLEX(-104, 48), MAKE_VAL_COMPLEX(48, 44), MAKE_VAL_COMPLEX(48, 104), MAKE_VAL_COMPLEX(-100, 20), MAKE_VAL_COMPLEX(108, -36), MAKE_VAL_COMPLEX(48, -76), MAKE_VAL_COMPLEX(64, 32), MAKE_VAL_COMPLEX(24, 12), MAKE_VAL_COMPLEX(-8, -4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1466, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1466", y_true, y_val, y1440_LEN*y1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1466() */
/* END TEST 1466 */

/*
 *  BEGIN TEST 1467
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1467( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	oski_value_t* y_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-240, -34), MAKE_VAL_COMPLEX(-428, 190), MAKE_VAL_COMPLEX(59, -369), MAKE_VAL_COMPLEX(88, -533), MAKE_VAL_COMPLEX(-498, -848), MAKE_VAL_COMPLEX(854, 54), MAKE_VAL_COMPLEX(91, -442), MAKE_VAL_COMPLEX(-447, -476), MAKE_VAL_COMPLEX(107, -714), MAKE_VAL_COMPLEX(-136, -1380), MAKE_VAL_COMPLEX(578, 1477), MAKE_VAL_COMPLEX(-56, -185), MAKE_VAL_COMPLEX(-453, 114), MAKE_VAL_COMPLEX(-32, 120), MAKE_VAL_COMPLEX(1041, -210) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1467, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1467", y_true, y_val, y1440_LEN*y1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1467() */
/* END TEST 1467 */

/*
 *  BEGIN TEST 1468
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1468( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-32, 47), MAKE_VAL_COMPLEX(72, 16), MAKE_VAL_COMPLEX(98, 48), MAKE_VAL_COMPLEX(10, -215), MAKE_VAL_COMPLEX(129, 137), MAKE_VAL_COMPLEX(-4, -154), MAKE_VAL_COMPLEX(-107, -9), MAKE_VAL_COMPLEX(-130, 91), MAKE_VAL_COMPLEX(108, -123), MAKE_VAL_COMPLEX(3, -92), MAKE_VAL_COMPLEX(-17, -103), MAKE_VAL_COMPLEX(-121, 11), MAKE_VAL_COMPLEX(-60, 143), MAKE_VAL_COMPLEX(-63, 50), MAKE_VAL_COMPLEX(-3, -22) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1468, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1468", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1468() */
/* END TEST 1468 */

/*
 *  BEGIN TEST 1469
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1469( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-20, -143), MAKE_VAL_COMPLEX(-179, -76), MAKE_VAL_COMPLEX(-100, -109), MAKE_VAL_COMPLEX(38, 189), MAKE_VAL_COMPLEX(-139, -144), MAKE_VAL_COMPLEX(104, 75), MAKE_VAL_COMPLEX(228, 49), MAKE_VAL_COMPLEX(65, -62), MAKE_VAL_COMPLEX(-48, 16), MAKE_VAL_COMPLEX(20, 123), MAKE_VAL_COMPLEX(137, 38), MAKE_VAL_COMPLEX(141, -146), MAKE_VAL_COMPLEX(-10, -43), MAKE_VAL_COMPLEX(122, 36), MAKE_VAL_COMPLEX(-120, 70) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1469, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1469", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1469() */
/* END TEST 1469 */

/*
 *  BEGIN TEST 1470
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1470( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-44, -92), MAKE_VAL_COMPLEX(-97, -61), MAKE_VAL_COMPLEX(1, -57), MAKE_VAL_COMPLEX(46, -22), MAKE_VAL_COMPLEX(-9, -7), MAKE_VAL_COMPLEX(97, -69), MAKE_VAL_COMPLEX(111, 43), MAKE_VAL_COMPLEX(-62, 24), MAKE_VAL_COMPLEX(61, -97), MAKE_VAL_COMPLEX(20, 30), MAKE_VAL_COMPLEX(115, -55), MAKE_VAL_COMPLEX(25, -125), MAKE_VAL_COMPLEX(-70, 90), MAKE_VAL_COMPLEX(51, 83), MAKE_VAL_COMPLEX(-117, 39) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1470, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1470", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1470() */
/* END TEST 1470 */

/*
 *  BEGIN TEST 1471
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1471( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-380, -359), MAKE_VAL_COMPLEX(-756, 423), MAKE_VAL_COMPLEX(-859, 66), MAKE_VAL_COMPLEX(808, 939), MAKE_VAL_COMPLEX(-1228, -151), MAKE_VAL_COMPLEX(561, 350), MAKE_VAL_COMPLEX(867, -668), MAKE_VAL_COMPLEX(271, -746), MAKE_VAL_COMPLEX(-263, 715), MAKE_VAL_COMPLEX(476, 453), MAKE_VAL_COMPLEX(492, 7), MAKE_VAL_COMPLEX(206, -799), MAKE_VAL_COMPLEX(-92, -605), MAKE_VAL_COMPLEX(449, -541), MAKE_VAL_COMPLEX(55, 521) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1471, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1471", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1471() */
/* END TEST 1471 */

/*
 *  BEGIN TEST 1472
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1472( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-48, -97), MAKE_VAL_COMPLEX(-11, -221), MAKE_VAL_COMPLEX(71, -203), MAKE_VAL_COMPLEX(368, -61), MAKE_VAL_COMPLEX(-136, -120), MAKE_VAL_COMPLEX(159, 25), MAKE_VAL_COMPLEX(26, 238), MAKE_VAL_COMPLEX(-60, 75), MAKE_VAL_COMPLEX(5, 78), MAKE_VAL_COMPLEX(31, 154), MAKE_VAL_COMPLEX(148, -6), MAKE_VAL_COMPLEX(154, 44), MAKE_VAL_COMPLEX(-30, 123), MAKE_VAL_COMPLEX(-158, -11), MAKE_VAL_COMPLEX(-96, 101) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1472, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1472", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1472() */
/* END TEST 1472 */

/*
 *  BEGIN TEST 1473
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1473( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(40, -27), MAKE_VAL_COMPLEX(-71, 108), MAKE_VAL_COMPLEX(-44, 139), MAKE_VAL_COMPLEX(-306, 57), MAKE_VAL_COMPLEX(129, 108), MAKE_VAL_COMPLEX(-18, -59), MAKE_VAL_COMPLEX(80, -139), MAKE_VAL_COMPLEX(-21, -76), MAKE_VAL_COMPLEX(108, -160), MAKE_VAL_COMPLEX(-22, -111), MAKE_VAL_COMPLEX(7, -4), MAKE_VAL_COMPLEX(-69, -174), MAKE_VAL_COMPLEX(-90, -53), MAKE_VAL_COMPLEX(178, 128), MAKE_VAL_COMPLEX(-54, -110) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1473, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1473", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1473() */
/* END TEST 1473 */

/*
 *  BEGIN TEST 1474
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1474( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(40, 0), MAKE_VAL_COMPLEX(38, -24), MAKE_VAL_COMPLEX(20, 10), MAKE_VAL_COMPLEX(0, 20), MAKE_VAL_COMPLEX(4, -2), MAKE_VAL_COMPLEX(8, 46), MAKE_VAL_COMPLEX(-34, 32), MAKE_VAL_COMPLEX(2, -26), MAKE_VAL_COMPLEX(24, 38), MAKE_VAL_COMPLEX(-14, 2), MAKE_VAL_COMPLEX(0, 50), MAKE_VAL_COMPLEX(40, 30), MAKE_VAL_COMPLEX(-20, -40), MAKE_VAL_COMPLEX(-38, 4), MAKE_VAL_COMPLEX(6, -48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1474, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1474", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1474() */
/* END TEST 1474 */

/*
 *  BEGIN TEST 1475
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1475( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_ROWMAJ, y1440_VW );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(52, 62), MAKE_VAL_COMPLEX(-99, 343), MAKE_VAL_COMPLEX(-131, 329), MAKE_VAL_COMPLEX(-670, 100), MAKE_VAL_COMPLEX(261, 229), MAKE_VAL_COMPLEX(-175, -127), MAKE_VAL_COMPLEX(91, -399), MAKE_VAL_COMPLEX(32, -128), MAKE_VAL_COMPLEX(89, -277), MAKE_VAL_COMPLEX(-40, -264), MAKE_VAL_COMPLEX(-131, -43), MAKE_VAL_COMPLEX(-253, -253), MAKE_VAL_COMPLEX(-50, -136), MAKE_VAL_COMPLEX(371, 143), MAKE_VAL_COMPLEX(27, -169) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-2, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1475, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1475", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1475() */
/* END TEST 1475 */

/*
 *  BEGIN TEST 1476
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1476( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	oski_value_t* y_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-6, -95), MAKE_VAL_COMPLEX(16, 45), MAKE_VAL_COMPLEX(27, 3), MAKE_VAL_COMPLEX(68, -158), MAKE_VAL_COMPLEX(-28, 143), MAKE_VAL_COMPLEX(-127, 58), MAKE_VAL_COMPLEX(24, -112), MAKE_VAL_COMPLEX(-113, -117), MAKE_VAL_COMPLEX(104, 25), MAKE_VAL_COMPLEX(123, 21), MAKE_VAL_COMPLEX(-146, -38), MAKE_VAL_COMPLEX(34, 41), MAKE_VAL_COMPLEX(-63, -160), MAKE_VAL_COMPLEX(-48, -50), MAKE_VAL_COMPLEX(137, -59) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1476, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1476", y_true, y_val, y1440_LEN*y1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1476() */
/* END TEST 1476 */

/*
 *  BEGIN TEST 1477
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1477( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	oski_value_t* y_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(40, -82), MAKE_VAL_COMPLEX(18, 143), MAKE_VAL_COMPLEX(17, 90), MAKE_VAL_COMPLEX(176, 176), MAKE_VAL_COMPLEX(63, 72), MAKE_VAL_COMPLEX(6, -40), MAKE_VAL_COMPLEX(218, 76), MAKE_VAL_COMPLEX(41, -2), MAKE_VAL_COMPLEX(-140, -267), MAKE_VAL_COMPLEX(94, -2), MAKE_VAL_COMPLEX(-97, 47), MAKE_VAL_COMPLEX(-178, 86), MAKE_VAL_COMPLEX(-49, 56), MAKE_VAL_COMPLEX(6, 11), MAKE_VAL_COMPLEX(-123, 72) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1477, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1477", y_true, y_val, y1440_LEN*y1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1477() */
/* END TEST 1477 */

/*
 *  BEGIN TEST 1478
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1478( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	oski_value_t* y_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-59, 48), MAKE_VAL_COMPLEX(40, -70), MAKE_VAL_COMPLEX(11, -42), MAKE_VAL_COMPLEX(-74, -72), MAKE_VAL_COMPLEX(47, -79), MAKE_VAL_COMPLEX(44, 27), MAKE_VAL_COMPLEX(-88, -54), MAKE_VAL_COMPLEX(-9, 58), MAKE_VAL_COMPLEX(-54, 88), MAKE_VAL_COMPLEX(-65, -65), MAKE_VAL_COMPLEX(81, 63), MAKE_VAL_COMPLEX(81, -2), MAKE_VAL_COMPLEX(8, 64), MAKE_VAL_COMPLEX(3, 24), MAKE_VAL_COMPLEX(-1, -8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1478, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1478", y_true, y_val, y1440_LEN*y1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1478() */
/* END TEST 1478 */

/*
 *  BEGIN TEST 1479
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1479( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	oski_value_t* y_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-100, -123), MAKE_VAL_COMPLEX(315, -380), MAKE_VAL_COMPLEX(312, -264), MAKE_VAL_COMPLEX(740, -1378), MAKE_VAL_COMPLEX(-545, -196), MAKE_VAL_COMPLEX(-691, -184), MAKE_VAL_COMPLEX(-26, -1218), MAKE_VAL_COMPLEX(-24, -836), MAKE_VAL_COMPLEX(-265, 1832), MAKE_VAL_COMPLEX(-87, 171), MAKE_VAL_COMPLEX(254, -440), MAKE_VAL_COMPLEX(850, 547), MAKE_VAL_COMPLEX(766, -657), MAKE_VAL_COMPLEX(71, -357), MAKE_VAL_COMPLEX(1232, 512) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1479, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1479", y_true, y_val, y1440_LEN*y1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1479() */
/* END TEST 1479 */

/*
 *  BEGIN TEST 1480
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1480( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(11, 96), MAKE_VAL_COMPLEX(155, 15), MAKE_VAL_COMPLEX(160, -121), MAKE_VAL_COMPLEX(183, -3), MAKE_VAL_COMPLEX(10, 82), MAKE_VAL_COMPLEX(75, -54), MAKE_VAL_COMPLEX(-29, -173), MAKE_VAL_COMPLEX(-119, -180), MAKE_VAL_COMPLEX(-61, -197), MAKE_VAL_COMPLEX(59, -119), MAKE_VAL_COMPLEX(13, -42), MAKE_VAL_COMPLEX(-44, -42), MAKE_VAL_COMPLEX(-8, -15), MAKE_VAL_COMPLEX(112, 166), MAKE_VAL_COMPLEX(15, -80) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1480, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1480", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1480() */
/* END TEST 1480 */

/*
 *  BEGIN TEST 1481
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1481( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-79, -100), MAKE_VAL_COMPLEX(-222, 22), MAKE_VAL_COMPLEX(-193, 102), MAKE_VAL_COMPLEX(-181, -31), MAKE_VAL_COMPLEX(-17, -79), MAKE_VAL_COMPLEX(-84, -25), MAKE_VAL_COMPLEX(90, 122), MAKE_VAL_COMPLEX(113, 224), MAKE_VAL_COMPLEX(24, 130), MAKE_VAL_COMPLEX(-35, 117), MAKE_VAL_COMPLEX(-8, -43), MAKE_VAL_COMPLEX(-21, -13), MAKE_VAL_COMPLEX(38, 85), MAKE_VAL_COMPLEX(-47, -169), MAKE_VAL_COMPLEX(-30, 161) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1481, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1481", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1481() */
/* END TEST 1481 */

/*
 *  BEGIN TEST 1482
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1482( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(32, 56), MAKE_VAL_COMPLEX(64, 34), MAKE_VAL_COMPLEX(2, 36), MAKE_VAL_COMPLEX(-28, 16), MAKE_VAL_COMPLEX(6, 4), MAKE_VAL_COMPLEX(-58, 48), MAKE_VAL_COMPLEX(-72, -22), MAKE_VAL_COMPLEX(38, -18), MAKE_VAL_COMPLEX(-34, 64), MAKE_VAL_COMPLEX(-14, -18), MAKE_VAL_COMPLEX(-70, 40), MAKE_VAL_COMPLEX(-10, 80), MAKE_VAL_COMPLEX(40, -60), MAKE_VAL_COMPLEX(-36, -50), MAKE_VAL_COMPLEX(72, -30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1482, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1482", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1482() */
/* END TEST 1482 */

/*
 *  BEGIN TEST 1483
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1483( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(666, 422), MAKE_VAL_COMPLEX(990, -938), MAKE_VAL_COMPLEX(268, -1724), MAKE_VAL_COMPLEX(1124, -1104), MAKE_VAL_COMPLEX(550, 422), MAKE_VAL_COMPLEX(232, -764), MAKE_VAL_COMPLEX(-1162, -770), MAKE_VAL_COMPLEX(-1850, -386), MAKE_VAL_COMPLEX(-1450, -846), MAKE_VAL_COMPLEX(-364, -1036), MAKE_VAL_COMPLEX(-64, -300), MAKE_VAL_COMPLEX(-426, -58), MAKE_VAL_COMPLEX(-238, -22), MAKE_VAL_COMPLEX(1654, 410), MAKE_VAL_COMPLEX(-492, -612) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1483, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1483", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1483() */
/* END TEST 1483 */

/*
 *  BEGIN TEST 1484
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1484( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-25, -44), MAKE_VAL_COMPLEX(56, -220), MAKE_VAL_COMPLEX(-25, -36), MAKE_VAL_COMPLEX(91, 251), MAKE_VAL_COMPLEX(-25, -123), MAKE_VAL_COMPLEX(14, 131), MAKE_VAL_COMPLEX(-208, 72), MAKE_VAL_COMPLEX(-69, 60), MAKE_VAL_COMPLEX(204, -36), MAKE_VAL_COMPLEX(-81, 131), MAKE_VAL_COMPLEX(38, 69), MAKE_VAL_COMPLEX(31, 37), MAKE_VAL_COMPLEX(-58, -15), MAKE_VAL_COMPLEX(-231, -5), MAKE_VAL_COMPLEX(0, 35) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1484, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1484", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1484() */
/* END TEST 1484 */

/*
 *  BEGIN TEST 1485
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1485( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(77, 100), MAKE_VAL_COMPLEX(27, 242), MAKE_VAL_COMPLEX(37, 77), MAKE_VAL_COMPLEX(-119, -225), MAKE_VAL_COMPLEX(33, 126), MAKE_VAL_COMPLEX(-68, -60), MAKE_VAL_COMPLEX(119, -78), MAKE_VAL_COMPLEX(108, -91), MAKE_VAL_COMPLEX(-226, 119), MAKE_VAL_COMPLEX(60, -148), MAKE_VAL_COMPLEX(-108, -4), MAKE_VAL_COMPLEX(-21, 58), MAKE_VAL_COMPLEX(88, -65), MAKE_VAL_COMPLEX(176, -43), MAKE_VAL_COMPLEX(75, -89) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1485, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1485", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1485() */
/* END TEST 1485 */

/*
 *  BEGIN TEST 1486
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1486( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(36, 8), MAKE_VAL_COMPLEX(39, -14), MAKE_VAL_COMPLEX(16, 13), MAKE_VAL_COMPLEX(-4, 18), MAKE_VAL_COMPLEX(4, -1), MAKE_VAL_COMPLEX(-2, 43), MAKE_VAL_COMPLEX(-37, 22), MAKE_VAL_COMPLEX(7, -23), MAKE_VAL_COMPLEX(14, 39), MAKE_VAL_COMPLEX(-13, -1), MAKE_VAL_COMPLEX(-10, 45), MAKE_VAL_COMPLEX(30, 35), MAKE_VAL_COMPLEX(-10, -40), MAKE_VAL_COMPLEX(-35, -4), MAKE_VAL_COMPLEX(15, -42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1486, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1486", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1486() */
/* END TEST 1486 */

/*
 *  BEGIN TEST 1487
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1440', LAYOUT_COLMAJ>
 *  y : <'y1440', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1487( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1440_ptr, A1440_ind, A1440_val,
		A1440_M, A1440_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1440_val,
		y1440_LEN * y1440_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1440_LEN, y1440_VW, LAYOUT_COLMAJ, y1440_LEN );
	oski_value_t* y_val = array_Duplicate( x1440_val,
		x1440_LEN * x1440_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1440_LEN, x1440_VW, LAYOUT_COLMAJ, x1440_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-151, -465), MAKE_VAL_COMPLEX(862, -400), MAKE_VAL_COMPLEX(9, -411), MAKE_VAL_COMPLEX(-878, 1060), MAKE_VAL_COMPLEX(500, -500), MAKE_VAL_COMPLEX(-433, 81), MAKE_VAL_COMPLEX(-566, -884), MAKE_VAL_COMPLEX(-657, 17), MAKE_VAL_COMPLEX(812, 506), MAKE_VAL_COMPLEX(-784, 46), MAKE_VAL_COMPLEX(29, 27), MAKE_VAL_COMPLEX(-232, -164), MAKE_VAL_COMPLEX(-159, 65), MAKE_VAL_COMPLEX(-366, -1002), MAKE_VAL_COMPLEX(-469, 429) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1487, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1487", y_true, y_val, x1440_LEN*x1440_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1487() */
/* END TEST 1487 */

#define A1488_M 5    /* # of rows */
#define A1488_N 5    /* # of cols */
#define A1488_NNZ 5  /* # of stored non-zeros */
static oski_index_t A1488_ptr[A1488_M+1] = {0, 3, 4, 5, 5, 5}; /* end row pointers for A1488 */
static oski_index_t A1488_ind[A1488_NNZ] = {1, 2, 4, 2, 3}; /* end column indices pointers for A1488 */
static oski_value_t A1488_val[A1488_NNZ] = {MAKE_VAL_COMPLEX(-10, 5), MAKE_VAL_COMPLEX(10, -5), MAKE_VAL_COMPLEX(6, -2), MAKE_VAL_COMPLEX(-1, 1), MAKE_VAL_COMPLEX(1, -10)}; /* end stored non-zero values for A1488 */

#define x1488_LEN 5   /* vector length */
#define x1488_VW  3   /* vector width */
static oski_value_t x1488_val[x1488_LEN*x1488_VW] = {MAKE_VAL_COMPLEX(-7, -10), MAKE_VAL_COMPLEX(-6, 4), MAKE_VAL_COMPLEX(-10, -7), MAKE_VAL_COMPLEX(-1, 7), MAKE_VAL_COMPLEX(10, -1), MAKE_VAL_COMPLEX(-3, 2), MAKE_VAL_COMPLEX(5, 5), MAKE_VAL_COMPLEX(1, 2), MAKE_VAL_COMPLEX(9, -5), MAKE_VAL_COMPLEX(6, 7), MAKE_VAL_COMPLEX(-5, -8), MAKE_VAL_COMPLEX(3, 7), MAKE_VAL_COMPLEX(-8, 10), MAKE_VAL_COMPLEX(4, 2), MAKE_VAL_COMPLEX(9, 0)}; /* data */

#define y1488_LEN 5   /* vector length */
#define y1488_VW  3   /* vector width */
static oski_value_t y1488_val[y1488_LEN*y1488_VW] = {MAKE_VAL_COMPLEX(-10, -4), MAKE_VAL_COMPLEX(9, -2), MAKE_VAL_COMPLEX(-2, -6), MAKE_VAL_COMPLEX(-1, 8), MAKE_VAL_COMPLEX(1, 0), MAKE_VAL_COMPLEX(-10, -9), MAKE_VAL_COMPLEX(-7, -10), MAKE_VAL_COMPLEX(5, -8), MAKE_VAL_COMPLEX(2, -5), MAKE_VAL_COMPLEX(5, 2), MAKE_VAL_COMPLEX(10, 9), MAKE_VAL_COMPLEX(7, -8), MAKE_VAL_COMPLEX(-5, 10), MAKE_VAL_COMPLEX(6, 5), MAKE_VAL_COMPLEX(-4, 6)}; /* data */

/*
 *  BEGIN TEST 1488
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1488( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	oski_value_t* y_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(39, 130), MAKE_VAL_COMPLEX(-125, 14), MAKE_VAL_COMPLEX(87, -101), MAKE_VAL_COMPLEX(87, 103), MAKE_VAL_COMPLEX(81, -21), MAKE_VAL_COMPLEX(37, 280), MAKE_VAL_COMPLEX(21, -66), MAKE_VAL_COMPLEX(-277, 48), MAKE_VAL_COMPLEX(-35, -135), MAKE_VAL_COMPLEX(-51, 5), MAKE_VAL_COMPLEX(-3, -140), MAKE_VAL_COMPLEX(-52, 77), MAKE_VAL_COMPLEX(90, -79), MAKE_VAL_COMPLEX(-31, -70), MAKE_VAL_COMPLEX(41, -62) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1488, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1488", y_true, y_val, y1488_LEN*y1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1488() */
/* END TEST 1488 */

/*
 *  BEGIN TEST 1489
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1489( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	oski_value_t* y_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-133, 18), MAKE_VAL_COMPLEX(120, -164), MAKE_VAL_COMPLEX(-189, 115), MAKE_VAL_COMPLEX(38, -63), MAKE_VAL_COMPLEX(-78, 5), MAKE_VAL_COMPLEX(-211, -147), MAKE_VAL_COMPLEX(-202, 148), MAKE_VAL_COMPLEX(164, -152), MAKE_VAL_COMPLEX(-39, 88), MAKE_VAL_COMPLEX(98, -79), MAKE_VAL_COMPLEX(177, 7), MAKE_VAL_COMPLEX(-55, -213), MAKE_VAL_COMPLEX(55, 189), MAKE_VAL_COMPLEX(129, -11), MAKE_VAL_COMPLEX(43, 144) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1489, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1489", y_true, y_val, y1488_LEN*y1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1489() */
/* END TEST 1489 */

/*
 *  BEGIN TEST 1490
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1490( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	oski_value_t* y_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(28, -70), MAKE_VAL_COMPLEX(14, 63), MAKE_VAL_COMPLEX(42, -14), MAKE_VAL_COMPLEX(-56, -7), MAKE_VAL_COMPLEX(0, 7), MAKE_VAL_COMPLEX(63, -70), MAKE_VAL_COMPLEX(70, -49), MAKE_VAL_COMPLEX(56, 35), MAKE_VAL_COMPLEX(35, 14), MAKE_VAL_COMPLEX(-14, 35), MAKE_VAL_COMPLEX(-63, 70), MAKE_VAL_COMPLEX(56, 49), MAKE_VAL_COMPLEX(-70, -35), MAKE_VAL_COMPLEX(-35, 42), MAKE_VAL_COMPLEX(-42, -28) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1490, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1490", y_true, y_val, y1488_LEN*y1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1490() */
/* END TEST 1490 */

/*
 *  BEGIN TEST 1491
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1491( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	oski_value_t* y_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(177, -152), MAKE_VAL_COMPLEX(436, 488), MAKE_VAL_COMPLEX(-741, -929), MAKE_VAL_COMPLEX(792, 55), MAKE_VAL_COMPLEX(12, -525), MAKE_VAL_COMPLEX(985, -275), MAKE_VAL_COMPLEX(-954, -636), MAKE_VAL_COMPLEX(152, 1164), MAKE_VAL_COMPLEX(-821, -246), MAKE_VAL_COMPLEX(300, 335), MAKE_VAL_COMPLEX(-111, 211), MAKE_VAL_COMPLEX(665, -193), MAKE_VAL_COMPLEX(-489, 41), MAKE_VAL_COMPLEX(-19, 293), MAKE_VAL_COMPLEX(-451, 106) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1491, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1491", y_true, y_val, y1488_LEN*y1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1491() */
/* END TEST 1491 */

/*
 *  BEGIN TEST 1492
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1492( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(59, -122), MAKE_VAL_COMPLEX(141, -72), MAKE_VAL_COMPLEX(160, 131), MAKE_VAL_COMPLEX(95, -38), MAKE_VAL_COMPLEX(-158, 115), MAKE_VAL_COMPLEX(46, 18), MAKE_VAL_COMPLEX(-204, 28), MAKE_VAL_COMPLEX(-11, 27), MAKE_VAL_COMPLEX(15, 77), MAKE_VAL_COMPLEX(-172, 37), MAKE_VAL_COMPLEX(2, -13), MAKE_VAL_COMPLEX(-90, -70), MAKE_VAL_COMPLEX(-57, -96), MAKE_VAL_COMPLEX(20, -23), MAKE_VAL_COMPLEX(-91, 10) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1492, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1492", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1492() */
/* END TEST 1492 */

/*
 *  BEGIN TEST 1493
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1493( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-5, 114), MAKE_VAL_COMPLEX(-145, 40), MAKE_VAL_COMPLEX(-112, -157), MAKE_VAL_COMPLEX(-121, 20), MAKE_VAL_COMPLEX(142, -73), MAKE_VAL_COMPLEX(-48, -34), MAKE_VAL_COMPLEX(174, -18), MAKE_VAL_COMPLEX(1, -27), MAKE_VAL_COMPLEX(-13, -31), MAKE_VAL_COMPLEX(132, -27), MAKE_VAL_COMPLEX(40, 9), MAKE_VAL_COMPLEX(56, 68), MAKE_VAL_COMPLEX(33, 44), MAKE_VAL_COMPLEX(-36, 35), MAKE_VAL_COMPLEX(73, 26) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1493, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1493", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1493() */
/* END TEST 1493 */

/*
 *  BEGIN TEST 1494
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1494( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(35, 50), MAKE_VAL_COMPLEX(30, -20), MAKE_VAL_COMPLEX(50, 35), MAKE_VAL_COMPLEX(5, -35), MAKE_VAL_COMPLEX(-50, 5), MAKE_VAL_COMPLEX(15, -10), MAKE_VAL_COMPLEX(-25, -25), MAKE_VAL_COMPLEX(-5, -10), MAKE_VAL_COMPLEX(-45, 25), MAKE_VAL_COMPLEX(-30, -35), MAKE_VAL_COMPLEX(25, 40), MAKE_VAL_COMPLEX(-15, -35), MAKE_VAL_COMPLEX(40, -50), MAKE_VAL_COMPLEX(-20, -10), MAKE_VAL_COMPLEX(-45, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1494, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1494", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1494() */
/* END TEST 1494 */

/*
 *  BEGIN TEST 1495
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1495( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(98, 1137), MAKE_VAL_COMPLEX(-813, 206), MAKE_VAL_COMPLEX(-543, -826), MAKE_VAL_COMPLEX(-769, -56), MAKE_VAL_COMPLEX(697, -552), MAKE_VAL_COMPLEX(-235, -275), MAKE_VAL_COMPLEX(1108, -306), MAKE_VAL_COMPLEX(-8, -254), MAKE_VAL_COMPLEX(-387, -135), MAKE_VAL_COMPLEX(799, -434), MAKE_VAL_COMPLEX(369, 330), MAKE_VAL_COMPLEX(354, 252), MAKE_VAL_COMPLEX(533, 74), MAKE_VAL_COMPLEX(-354, 159), MAKE_VAL_COMPLEX(250, 119) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1495, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1495", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1495() */
/* END TEST 1495 */

/*
 *  BEGIN TEST 1496
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1496( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-285, -78), MAKE_VAL_COMPLEX(57, -16), MAKE_VAL_COMPLEX(20, 63), MAKE_VAL_COMPLEX(147, 144), MAKE_VAL_COMPLEX(-54, -97), MAKE_VAL_COMPLEX(-16, 102), MAKE_VAL_COMPLEX(12, -170), MAKE_VAL_COMPLEX(225, -73), MAKE_VAL_COMPLEX(-55, -231), MAKE_VAL_COMPLEX(184, -91), MAKE_VAL_COMPLEX(6, 51), MAKE_VAL_COMPLEX(130, 18), MAKE_VAL_COMPLEX(-17, 80), MAKE_VAL_COMPLEX(100, -11), MAKE_VAL_COMPLEX(41, -106) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1496, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1496", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1496() */
/* END TEST 1496 */

/*
 *  BEGIN TEST 1497
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1497( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(238, 96), MAKE_VAL_COMPLEX(-47, 44), MAKE_VAL_COMPLEX(-58, -30), MAKE_VAL_COMPLEX(-120, -133), MAKE_VAL_COMPLEX(60, 56), MAKE_VAL_COMPLEX(21, -88), MAKE_VAL_COMPLEX(13, 155), MAKE_VAL_COMPLEX(-216, 71), MAKE_VAL_COMPLEX(44, 190), MAKE_VAL_COMPLEX(-150, 74), MAKE_VAL_COMPLEX(-43, -39), MAKE_VAL_COMPLEX(-99, -23), MAKE_VAL_COMPLEX(49, -38), MAKE_VAL_COMPLEX(-88, -3), MAKE_VAL_COMPLEX(-32, 70) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1497, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1497", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1497() */
/* END TEST 1497 */

/*
 *  BEGIN TEST 1498
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1498( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-6, -115), MAKE_VAL_COMPLEX(-68, 2), MAKE_VAL_COMPLEX(-45, -106), MAKE_VAL_COMPLEX(-43, 51), MAKE_VAL_COMPLEX(85, 42), MAKE_VAL_COMPLEX(-34, 1), MAKE_VAL_COMPLEX(15, 65), MAKE_VAL_COMPLEX(-2, 21), MAKE_VAL_COMPLEX(97, 5), MAKE_VAL_COMPLEX(13, 86), MAKE_VAL_COMPLEX(0, -89), MAKE_VAL_COMPLEX(-11, 71), MAKE_VAL_COMPLEX(-114, 40), MAKE_VAL_COMPLEX(22, 36), MAKE_VAL_COMPLEX(72, 45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1498, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1498", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1498() */
/* END TEST 1498 */

/*
 *  BEGIN TEST 1499
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1499( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-516, -1039), MAKE_VAL_COMPLEX(647, -198), MAKE_VAL_COMPLEX(577, 652), MAKE_VAL_COMPLEX(101, 864), MAKE_VAL_COMPLEX(-475, -470), MAKE_VAL_COMPLEX(-339, 273), MAKE_VAL_COMPLEX(350, -1022), MAKE_VAL_COMPLEX(1322, 430), MAKE_VAL_COMPLEX(185, -917), MAKE_VAL_COMPLEX(799, -44), MAKE_VAL_COMPLEX(267, 682), MAKE_VAL_COMPLEX(276, 234), MAKE_VAL_COMPLEX(-87, -414), MAKE_VAL_COMPLEX(266, 291), MAKE_VAL_COMPLEX(80, -213) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1499, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1499", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1499() */
/* END TEST 1499 */

/*
 *  BEGIN TEST 1500
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1500( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	oski_value_t* y_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(91, -58), MAKE_VAL_COMPLEX(-9, 231), MAKE_VAL_COMPLEX(131, -183), MAKE_VAL_COMPLEX(-107, 21), MAKE_VAL_COMPLEX(-34, -55), MAKE_VAL_COMPLEX(68, 29), MAKE_VAL_COMPLEX(205, -35), MAKE_VAL_COMPLEX(-123, 122), MAKE_VAL_COMPLEX(13, 42), MAKE_VAL_COMPLEX(-78, 51), MAKE_VAL_COMPLEX(8, -101), MAKE_VAL_COMPLEX(102, 231), MAKE_VAL_COMPLEX(-48, -232), MAKE_VAL_COMPLEX(-7, 126), MAKE_VAL_COMPLEX(-75, -122) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1500, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1500", y_true, y_val, y1488_LEN*y1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1500() */
/* END TEST 1500 */

/*
 *  BEGIN TEST 1501
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1501( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	oski_value_t* y_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(23, -24), MAKE_VAL_COMPLEX(-32, -118), MAKE_VAL_COMPLEX(-51, 203), MAKE_VAL_COMPLEX(26, -88), MAKE_VAL_COMPLEX(27, 66), MAKE_VAL_COMPLEX(101, -76), MAKE_VAL_COMPLEX(-46, 28), MAKE_VAL_COMPLEX(176, -11), MAKE_VAL_COMPLEX(28, 15), MAKE_VAL_COMPLEX(21, -10), MAKE_VAL_COMPLEX(-177, 148), MAKE_VAL_COMPLEX(-63, -98), MAKE_VAL_COMPLEX(-27, 107), MAKE_VAL_COMPLEX(-90, -95), MAKE_VAL_COMPLEX(37, 36) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1501, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1501", y_true, y_val, y1488_LEN*y1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1501() */
/* END TEST 1501 */

/*
 *  BEGIN TEST 1502
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1502( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	oski_value_t* y_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(48, 112), MAKE_VAL_COMPLEX(-88, -56), MAKE_VAL_COMPLEX(-32, 64), MAKE_VAL_COMPLEX(72, -56), MAKE_VAL_COMPLEX(-8, -8), MAKE_VAL_COMPLEX(8, 152), MAKE_VAL_COMPLEX(-24, 136), MAKE_VAL_COMPLEX(-104, 24), MAKE_VAL_COMPLEX(-56, 24), MAKE_VAL_COMPLEX(-24, -56), MAKE_VAL_COMPLEX(-8, -152), MAKE_VAL_COMPLEX(-120, 8), MAKE_VAL_COMPLEX(120, -40), MAKE_VAL_COMPLEX(-8, -88), MAKE_VAL_COMPLEX(80, -16) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1502, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1502", y_true, y_val, y1488_LEN*y1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1502() */
/* END TEST 1502 */

/*
 *  BEGIN TEST 1503
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1503( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	oski_value_t* y_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(122, -86), MAKE_VAL_COMPLEX(941, 540), MAKE_VAL_COMPLEX(-844, -1208), MAKE_VAL_COMPLEX(179, 518), MAKE_VAL_COMPLEX(-499, -72), MAKE_VAL_COMPLEX(272, 565), MAKE_VAL_COMPLEX(281, -648), MAKE_VAL_COMPLEX(-373, 1210), MAKE_VAL_COMPLEX(-42, 143), MAKE_VAL_COMPLEX(-59, 326), MAKE_VAL_COMPLEX(-400, -1309), MAKE_VAL_COMPLEX(1115, 184), MAKE_VAL_COMPLEX(-859, -690), MAKE_VAL_COMPLEX(810, 45), MAKE_VAL_COMPLEX(-564, -12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1503, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1503", y_true, y_val, y1488_LEN*y1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1503() */
/* END TEST 1503 */

/*
 *  BEGIN TEST 1504
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1504( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-142, -123), MAKE_VAL_COMPLEX(110, 121), MAKE_VAL_COMPLEX(-290, 149), MAKE_VAL_COMPLEX(135, 26), MAKE_VAL_COMPLEX(163, 9), MAKE_VAL_COMPLEX(-131, -39), MAKE_VAL_COMPLEX(-180, 23), MAKE_VAL_COMPLEX(-63, -24), MAKE_VAL_COMPLEX(123, 92), MAKE_VAL_COMPLEX(-25, 38), MAKE_VAL_COMPLEX(-109, -85), MAKE_VAL_COMPLEX(127, 87), MAKE_VAL_COMPLEX(-79, 52), MAKE_VAL_COMPLEX(-53, -32), MAKE_VAL_COMPLEX(110, 22) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1504, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1504", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1504() */
/* END TEST 1504 */

/*
 *  BEGIN TEST 1505
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1505( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(21, 14), MAKE_VAL_COMPLEX(-176, -51), MAKE_VAL_COMPLEX(139, -210), MAKE_VAL_COMPLEX(-127, 68), MAKE_VAL_COMPLEX(-36, -52), MAKE_VAL_COMPLEX(98, 74), MAKE_VAL_COMPLEX(260, 27), MAKE_VAL_COMPLEX(82, 47), MAKE_VAL_COMPLEX(-21, -184), MAKE_VAL_COMPLEX(124, 35), MAKE_VAL_COMPLEX(20, -4), MAKE_VAL_COMPLEX(-67, -5), MAKE_VAL_COMPLEX(5, 102), MAKE_VAL_COMPLEX(111, 46), MAKE_VAL_COMPLEX(7, -49) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1505, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1505", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1505() */
/* END TEST 1505 */

/*
 *  BEGIN TEST 1506
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1506( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-102, -18), MAKE_VAL_COMPLEX(-12, 60), MAKE_VAL_COMPLEX(-102, 18), MAKE_VAL_COMPLEX(36, 48), MAKE_VAL_COMPLEX(54, -66), MAKE_VAL_COMPLEX(-6, 30), MAKE_VAL_COMPLEX(60, 0), MAKE_VAL_COMPLEX(18, 6), MAKE_VAL_COMPLEX(24, -84), MAKE_VAL_COMPLEX(78, 6), MAKE_VAL_COMPLEX(-78, -18), MAKE_VAL_COMPLEX(60, 24), MAKE_VAL_COMPLEX(12, 108), MAKE_VAL_COMPLEX(36, -12), MAKE_VAL_COMPLEX(54, -54) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1506, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1506", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1506() */
/* END TEST 1506 */

/*
 *  BEGIN TEST 1507
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1507( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-60, 1757), MAKE_VAL_COMPLEX(-275, -1951), MAKE_VAL_COMPLEX(3519, 758), MAKE_VAL_COMPLEX(-1132, -1092), MAKE_VAL_COMPLEX(-842, -1461), MAKE_VAL_COMPLEX(641, 1594), MAKE_VAL_COMPLEX(2022, 1748), MAKE_VAL_COMPLEX(376, 903), MAKE_VAL_COMPLEX(268, -2074), MAKE_VAL_COMPLEX(787, 338), MAKE_VAL_COMPLEX(63, 1252), MAKE_VAL_COMPLEX(-299, -1513), MAKE_VAL_COMPLEX(615, 661), MAKE_VAL_COMPLEX(373, 927), MAKE_VAL_COMPLEX(-315, -1062) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-9, -9);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1507, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1507", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1507() */
/* END TEST 1507 */

/*
 *  BEGIN TEST 1508
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1508( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-185, 0), MAKE_VAL_COMPLEX(148, -17), MAKE_VAL_COMPLEX(-83, 302), MAKE_VAL_COMPLEX(115, 116), MAKE_VAL_COMPLEX(78, 110), MAKE_VAL_COMPLEX(-56, -148), MAKE_VAL_COMPLEX(16, -85), MAKE_VAL_COMPLEX(-92, -153), MAKE_VAL_COMPLEX(103, 54), MAKE_VAL_COMPLEX(96, -15), MAKE_VAL_COMPLEX(48, 28), MAKE_VAL_COMPLEX(-69, -23), MAKE_VAL_COMPLEX(-45, 10), MAKE_VAL_COMPLEX(-19, -78), MAKE_VAL_COMPLEX(65, 53) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1508, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1508", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1508() */
/* END TEST 1508 */

/*
 *  BEGIN TEST 1509
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1509( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(25, -37), MAKE_VAL_COMPLEX(-172, 111), MAKE_VAL_COMPLEX(-80, -282), MAKE_VAL_COMPLEX(-62, -37), MAKE_VAL_COMPLEX(13, -210), MAKE_VAL_COMPLEX(44, 195), MAKE_VAL_COMPLEX(79, 90), MAKE_VAL_COMPLEX(120, 164), MAKE_VAL_COMPLEX(-58, -185), MAKE_VAL_COMPLEX(27, 31), MAKE_VAL_COMPLEX(-170, -63), MAKE_VAL_COMPLEX(162, 66), MAKE_VAL_COMPLEX(55, 162), MAKE_VAL_COMPLEX(77, 62), MAKE_VAL_COMPLEX(25, -134) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1509, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1509", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1509() */
/* END TEST 1509 */

/*
 *  BEGIN TEST 1510
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1510( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-75, -22), MAKE_VAL_COMPLEX(-14, 44), MAKE_VAL_COMPLEX(-78, 5), MAKE_VAL_COMPLEX(23, 39), MAKE_VAL_COMPLEX(46, -45), MAKE_VAL_COMPLEX(-7, 22), MAKE_VAL_COMPLEX(45, 5), MAKE_VAL_COMPLEX(13, 6), MAKE_VAL_COMPLEX(25, -61), MAKE_VAL_COMPLEX(58, 11), MAKE_VAL_COMPLEX(-57, -20), MAKE_VAL_COMPLEX(43, 23), MAKE_VAL_COMPLEX(0, 82), MAKE_VAL_COMPLEX(28, -6), MAKE_VAL_COMPLEX(45, -36) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1510, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1510", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1510() */
/* END TEST 1510 */

/*
 *  BEGIN TEST 1511
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_ROWMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1511( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_ROWMAJ, x1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-665, 1279), MAKE_VAL_COMPLEX(513, -1702), MAKE_VAL_COMPLEX(2506, 2044), MAKE_VAL_COMPLEX(1393, -352), MAKE_VAL_COMPLEX(1556, 433), MAKE_VAL_COMPLEX(-1779, -518), MAKE_VAL_COMPLEX(-884, -389), MAKE_VAL_COMPLEX(-2020, -32), MAKE_VAL_COMPLEX(1389, -196), MAKE_VAL_COMPLEX(135, -631), MAKE_VAL_COMPLEX(899, -544), MAKE_VAL_COMPLEX(-942, 662), MAKE_VAL_COMPLEX(-700, 131), MAKE_VAL_COMPLEX(-892, -95), MAKE_VAL_COMPLEX(912, 165) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, -9);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1511, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1511", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1511() */
/* END TEST 1511 */

/*
 *  BEGIN TEST 1512
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1512( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	oski_value_t* y_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-16, -196), MAKE_VAL_COMPLEX(6, 85), MAKE_VAL_COMPLEX(-4, 45), MAKE_VAL_COMPLEX(-25, 129), MAKE_VAL_COMPLEX(42, 6), MAKE_VAL_COMPLEX(74, 24), MAKE_VAL_COMPLEX(119, -172), MAKE_VAL_COMPLEX(-24, -63), MAKE_VAL_COMPLEX(45, -129), MAKE_VAL_COMPLEX(45, -65), MAKE_VAL_COMPLEX(-73, 77), MAKE_VAL_COMPLEX(-48, -19), MAKE_VAL_COMPLEX(-82, -110), MAKE_VAL_COMPLEX(-51, 55), MAKE_VAL_COMPLEX(-47, -86) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1512, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1512", y_true, y_val, y1488_LEN*y1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1512() */
/* END TEST 1512 */

/*
 *  BEGIN TEST 1513
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1513( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	oski_value_t* y_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(48, 232), MAKE_VAL_COMPLEX(-46, -95), MAKE_VAL_COMPLEX(0, -17), MAKE_VAL_COMPLEX(45, -159), MAKE_VAL_COMPLEX(-46, -8), MAKE_VAL_COMPLEX(-52, 32), MAKE_VAL_COMPLEX(-111, 226), MAKE_VAL_COMPLEX(-12, 85), MAKE_VAL_COMPLEX(-63, 145), MAKE_VAL_COMPLEX(-61, 47), MAKE_VAL_COMPLEX(51, -133), MAKE_VAL_COMPLEX(4, 37), MAKE_VAL_COMPLEX(122, 80), MAKE_VAL_COMPLEX(37, -87), MAKE_VAL_COMPLEX(75, 70) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1513, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1513", y_true, y_val, y1488_LEN*y1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1513() */
/* END TEST 1513 */

/*
 *  BEGIN TEST 1514
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1514( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	oski_value_t* y_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(78, -50), MAKE_VAL_COMPLEX(-31, 73), MAKE_VAL_COMPLEX(52, 16), MAKE_VAL_COMPLEX(-51, -47), MAKE_VAL_COMPLEX(-5, 7), MAKE_VAL_COMPLEX(113, -25), MAKE_VAL_COMPLEX(105, 1), MAKE_VAL_COMPLEX(31, 75), MAKE_VAL_COMPLEX(25, 39), MAKE_VAL_COMPLEX(-39, 25), MAKE_VAL_COMPLEX(-113, 25), MAKE_VAL_COMPLEX(21, 89), MAKE_VAL_COMPLEX(-45, -85), MAKE_VAL_COMPLEX(-65, 17), MAKE_VAL_COMPLEX(-22, -58) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1514, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1514", y_true, y_val, y1488_LEN*y1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1514() */
/* END TEST 1514 */

/*
 *  BEGIN TEST 1515
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1515( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	oski_value_t* y_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(156, -80), MAKE_VAL_COMPLEX(20, -68), MAKE_VAL_COMPLEX(-158, -132), MAKE_VAL_COMPLEX(-248, 130), MAKE_VAL_COMPLEX(10, 80), MAKE_VAL_COMPLEX(-304, -10), MAKE_VAL_COMPLEX(150, 46), MAKE_VAL_COMPLEX(204, -244), MAKE_VAL_COMPLEX(282, -28), MAKE_VAL_COMPLEX(248, 114), MAKE_VAL_COMPLEX(102, 12), MAKE_VAL_COMPLEX(160, -300), MAKE_VAL_COMPLEX(150, 76), MAKE_VAL_COMPLEX(42, -16), MAKE_VAL_COMPLEX(108, 54) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1515, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1515", y_true, y_val, y1488_LEN*y1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1515() */
/* END TEST 1515 */

/*
 *  BEGIN TEST 1516
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1516( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-49, -255), MAKE_VAL_COMPLEX(62, 25), MAKE_VAL_COMPLEX(-198, -36), MAKE_VAL_COMPLEX(-130, 108), MAKE_VAL_COMPLEX(0, 29), MAKE_VAL_COMPLEX(88, -15), MAKE_VAL_COMPLEX(-72, 118), MAKE_VAL_COMPLEX(5, 54), MAKE_VAL_COMPLEX(13, -57), MAKE_VAL_COMPLEX(77, 42), MAKE_VAL_COMPLEX(51, 51), MAKE_VAL_COMPLEX(33, 102), MAKE_VAL_COMPLEX(-28, 52), MAKE_VAL_COMPLEX(-19, -7), MAKE_VAL_COMPLEX(35, 1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1516, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1516", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1516() */
/* END TEST 1516 */

/*
 *  BEGIN TEST 1517
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1517( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-2, 246), MAKE_VAL_COMPLEX(-68, 5), MAKE_VAL_COMPLEX(147, 45), MAKE_VAL_COMPLEX(148, -84), MAKE_VAL_COMPLEX(27, -62), MAKE_VAL_COMPLEX(-91, 30), MAKE_VAL_COMPLEX(102, -118), MAKE_VAL_COMPLEX(4, -51), MAKE_VAL_COMPLEX(-1, 15), MAKE_VAL_COMPLEX(-38, -39), MAKE_VAL_COMPLEX(-90, -60), MAKE_VAL_COMPLEX(-3, -90), MAKE_VAL_COMPLEX(34, 2), MAKE_VAL_COMPLEX(37, 1), MAKE_VAL_COMPLEX(-8, -28) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1517, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1517", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1517() */
/* END TEST 1517 */

/*
 *  BEGIN TEST 1518
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1518( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-39, -77), MAKE_VAL_COMPLEX(-46, 22), MAKE_VAL_COMPLEX(-63, -59), MAKE_VAL_COMPLEX(-14, 48), MAKE_VAL_COMPLEX(71, 3), MAKE_VAL_COMPLEX(-23, 11), MAKE_VAL_COMPLEX(30, 40), MAKE_VAL_COMPLEX(5, 15), MAKE_VAL_COMPLEX(68, -26), MAKE_VAL_COMPLEX(35, 55), MAKE_VAL_COMPLEX(-27, -61), MAKE_VAL_COMPLEX(14, 52), MAKE_VAL_COMPLEX(-66, 62), MAKE_VAL_COMPLEX(26, 18), MAKE_VAL_COMPLEX(63, 9) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1518, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1518", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1518() */
/* END TEST 1518 */

/*
 *  BEGIN TEST 1519
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1519( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-989, -1167), MAKE_VAL_COMPLEX(1001, -502), MAKE_VAL_COMPLEX(-912, 1060), MAKE_VAL_COMPLEX(-538, 980), MAKE_VAL_COMPLEX(-579, 466), MAKE_VAL_COMPLEX(819, -756), MAKE_VAL_COMPLEX(-436, 1014), MAKE_VAL_COMPLEX(204, 275), MAKE_VAL_COMPLEX(-767, -7), MAKE_VAL_COMPLEX(270, -471), MAKE_VAL_COMPLEX(1161, 665), MAKE_VAL_COMPLEX(440, 181), MAKE_VAL_COMPLEX(468, -312), MAKE_VAL_COMPLEX(-515, -45), MAKE_VAL_COMPLEX(-381, -23) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, -5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1519, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1519", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1519() */
/* END TEST 1519 */

/*
 *  BEGIN TEST 1520
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1520( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-150, -98), MAKE_VAL_COMPLEX(88, 89), MAKE_VAL_COMPLEX(-39, -175), MAKE_VAL_COMPLEX(84, -80), MAKE_VAL_COMPLEX(-55, -14), MAKE_VAL_COMPLEX(51, -90), MAKE_VAL_COMPLEX(-106, -2), MAKE_VAL_COMPLEX(200, -73), MAKE_VAL_COMPLEX(105, 69), MAKE_VAL_COMPLEX(50, 23), MAKE_VAL_COMPLEX(142, 8), MAKE_VAL_COMPLEX(-31, 74), MAKE_VAL_COMPLEX(-80, -150), MAKE_VAL_COMPLEX(55, 15), MAKE_VAL_COMPLEX(-4, -16) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1520, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1520", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1520() */
/* END TEST 1520 */

/*
 *  BEGIN TEST 1521
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1521( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(172, 23), MAKE_VAL_COMPLEX(-132, -103), MAKE_VAL_COMPLEX(34, 97), MAKE_VAL_COMPLEX(-123, 103), MAKE_VAL_COMPLEX(100, 60), MAKE_VAL_COMPLEX(-73, 83), MAKE_VAL_COMPLEX(101, 47), MAKE_VAL_COMPLEX(-206, 86), MAKE_VAL_COMPLEX(-44, -44), MAKE_VAL_COMPLEX(-61, 35), MAKE_VAL_COMPLEX(-122, -65), MAKE_VAL_COMPLEX(8, -31), MAKE_VAL_COMPLEX(-2, 150), MAKE_VAL_COMPLEX(-49, 13), MAKE_VAL_COMPLEX(40, 61) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1521, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1521", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1521() */
/* END TEST 1521 */

/*
 *  BEGIN TEST 1522
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1522( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(41, -133), MAKE_VAL_COMPLEX(-78, -26), MAKE_VAL_COMPLEX(-7, -139), MAKE_VAL_COMPLEX(-70, 40), MAKE_VAL_COMPLEX(79, 83), MAKE_VAL_COMPLEX(-39, -13), MAKE_VAL_COMPLEX(-10, 80), MAKE_VAL_COMPLEX(-11, 23), MAKE_VAL_COMPLEX(108, 46), MAKE_VAL_COMPLEX(-21, 103), MAKE_VAL_COMPLEX(37, -101), MAKE_VAL_COMPLEX(-42, 76), MAKE_VAL_COMPLEX(-146, -2), MAKE_VAL_COMPLEX(10, 50), MAKE_VAL_COMPLEX(63, 81) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1522, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1522", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1522() */
/* END TEST 1522 */

/*
 *  BEGIN TEST 1523
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1523( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_ROWMAJ, y1488_VW );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-911, -766), MAKE_VAL_COMPLEX(421, 686), MAKE_VAL_COMPLEX(-171, -1078), MAKE_VAL_COMPLEX(619, -341), MAKE_VAL_COMPLEX(-252, -206), MAKE_VAL_COMPLEX(387, -457), MAKE_VAL_COMPLEX(-569, -113), MAKE_VAL_COMPLEX(1292, -230), MAKE_VAL_COMPLEX(594, 430), MAKE_VAL_COMPLEX(361, 201), MAKE_VAL_COMPLEX(761, 164), MAKE_VAL_COMPLEX(-197, 444), MAKE_VAL_COMPLEX(-326, -862), MAKE_VAL_COMPLEX(355, 135), MAKE_VAL_COMPLEX(55, -154) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, 1);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1523, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1523", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1523() */
/* END TEST 1523 */

/*
 *  BEGIN TEST 1524
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1524( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	oski_value_t* y_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-102, -126), MAKE_VAL_COMPLEX(72, 108), MAKE_VAL_COMPLEX(27, -149), MAKE_VAL_COMPLEX(70, -58), MAKE_VAL_COMPLEX(-7, -77), MAKE_VAL_COMPLEX(-76, -3), MAKE_VAL_COMPLEX(-13, -37), MAKE_VAL_COMPLEX(-71, -148), MAKE_VAL_COMPLEX(-10, -22), MAKE_VAL_COMPLEX(13, 13), MAKE_VAL_COMPLEX(22, 84), MAKE_VAL_COMPLEX(30, 40), MAKE_VAL_COMPLEX(5, -83), MAKE_VAL_COMPLEX(-64, -55), MAKE_VAL_COMPLEX(-13, -20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1524, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1524", y_true, y_val, y1488_LEN*y1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1524() */
/* END TEST 1524 */

/*
 *  BEGIN TEST 1525
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1525( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	oski_value_t* y_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-16, 44), MAKE_VAL_COMPLEX(51, -107), MAKE_VAL_COMPLEX(-35, 65), MAKE_VAL_COMPLEX(-107, 159), MAKE_VAL_COMPLEX(20, 80), MAKE_VAL_COMPLEX(-27, -144), MAKE_VAL_COMPLEX(-48, -114), MAKE_VAL_COMPLEX(160, 59), MAKE_VAL_COMPLEX(51, -37), MAKE_VAL_COMPLEX(46, 28), MAKE_VAL_COMPLEX(81, 63), MAKE_VAL_COMPLEX(85, -123), MAKE_VAL_COMPLEX(-100, 198), MAKE_VAL_COMPLEX(127, 138), MAKE_VAL_COMPLEX(-57, 86) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1525, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1525", y_true, y_val, y1488_LEN*y1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1525() */
/* END TEST 1525 */

/*
 *  BEGIN TEST 1526
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1526( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	oski_value_t* y_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-48, 62), MAKE_VAL_COMPLEX(4, -67), MAKE_VAL_COMPLEX(-46, 2), MAKE_VAL_COMPLEX(54, 23), MAKE_VAL_COMPLEX(2, -7), MAKE_VAL_COMPLEX(-83, 52), MAKE_VAL_COMPLEX(-84, 29), MAKE_VAL_COMPLEX(-46, -51), MAKE_VAL_COMPLEX(-31, -24), MAKE_VAL_COMPLEX(24, -31), MAKE_VAL_COMPLEX(83, -52), MAKE_VAL_COMPLEX(-42, -65), MAKE_VAL_COMPLEX(60, 55), MAKE_VAL_COMPLEX(47, -32), MAKE_VAL_COMPLEX(34, 40) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1526, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1526", y_true, y_val, y1488_LEN*y1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1526() */
/* END TEST 1526 */

/*
 *  BEGIN TEST 1527
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1527( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	oski_value_t* y_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1126, -198), MAKE_VAL_COMPLEX(-1253, 67), MAKE_VAL_COMPLEX(1006, 568), MAKE_VAL_COMPLEX(903, 655), MAKE_VAL_COMPLEX(689, -41), MAKE_VAL_COMPLEX(-282, -32), MAKE_VAL_COMPLEX(-127, 415), MAKE_VAL_COMPLEX(908, -678), MAKE_VAL_COMPLEX(-89, -123), MAKE_VAL_COMPLEX(-75, -193), MAKE_VAL_COMPLEX(-393, -535), MAKE_VAL_COMPLEX(-907, -67), MAKE_VAL_COMPLEX(1347, 293), MAKE_VAL_COMPLEX(768, -906), MAKE_VAL_COMPLEX(567, 57) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 9);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1527, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1527", y_true, y_val, y1488_LEN*y1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1527() */
/* END TEST 1527 */

/*
 *  BEGIN TEST 1528
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1528( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-62, -179), MAKE_VAL_COMPLEX(77, -24), MAKE_VAL_COMPLEX(-208, -75), MAKE_VAL_COMPLEX(-109, 44), MAKE_VAL_COMPLEX(-21, 52), MAKE_VAL_COMPLEX(102, 83), MAKE_VAL_COMPLEX(115, 83), MAKE_VAL_COMPLEX(-79, -16), MAKE_VAL_COMPLEX(-7, -29), MAKE_VAL_COMPLEX(-91, 32), MAKE_VAL_COMPLEX(-208, 95), MAKE_VAL_COMPLEX(-153, -7), MAKE_VAL_COMPLEX(5, 122), MAKE_VAL_COMPLEX(105, 97), MAKE_VAL_COMPLEX(110, 94) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1528, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1528", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1528() */
/* END TEST 1528 */

/*
 *  BEGIN TEST 1529
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1529( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(154, 55), MAKE_VAL_COMPLEX(-149, -32), MAKE_VAL_COMPLEX(252, -73), MAKE_VAL_COMPLEX(21, -28), MAKE_VAL_COMPLEX(73, 64), MAKE_VAL_COMPLEX(-138, -111), MAKE_VAL_COMPLEX(-155, -3), MAKE_VAL_COMPLEX(59, 36), MAKE_VAL_COMPLEX(103, 117), MAKE_VAL_COMPLEX(31, 68), MAKE_VAL_COMPLEX(284, -187), MAKE_VAL_COMPLEX(81, 71), MAKE_VAL_COMPLEX(-157, -178), MAKE_VAL_COMPLEX(-113, -41), MAKE_VAL_COMPLEX(-74, 14) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1529, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1529", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1529() */
/* END TEST 1529 */

/*
 *  BEGIN TEST 1530
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1530( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-72, -39), MAKE_VAL_COMPLEX(-24, 42), MAKE_VAL_COMPLEX(-81, -12), MAKE_VAL_COMPLEX(15, 45), MAKE_VAL_COMPLEX(57, -36), MAKE_VAL_COMPLEX(-12, 21), MAKE_VAL_COMPLEX(45, 15), MAKE_VAL_COMPLEX(12, 9), MAKE_VAL_COMPLEX(39, -57), MAKE_VAL_COMPLEX(57, 24), MAKE_VAL_COMPLEX(-54, -33), MAKE_VAL_COMPLEX(39, 33), MAKE_VAL_COMPLEX(-18, 84), MAKE_VAL_COMPLEX(30, 0), MAKE_VAL_COMPLEX(54, -27) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1530, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1530", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1530() */
/* END TEST 1530 */

/*
 *  BEGIN TEST 1531
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1531( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(315, -750), MAKE_VAL_COMPLEX(403, 547), MAKE_VAL_COMPLEX(-586, -1039), MAKE_VAL_COMPLEX(-342, -298), MAKE_VAL_COMPLEX(-234, -271), MAKE_VAL_COMPLEX(-83, 876), MAKE_VAL_COMPLEX(145, 729), MAKE_VAL_COMPLEX(-83, -450), MAKE_VAL_COMPLEX(77, -461), MAKE_VAL_COMPLEX(-144, -461), MAKE_VAL_COMPLEX(-1407, -696), MAKE_VAL_COMPLEX(-171, -795), MAKE_VAL_COMPLEX(-381, 759), MAKE_VAL_COMPLEX(-60, 716), MAKE_VAL_COMPLEX(-32, 553) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, 5);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1531, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1531", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1531() */
/* END TEST 1531 */

/*
 *  BEGIN TEST 1532
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1532( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-16, -14), MAKE_VAL_COMPLEX(85, 22), MAKE_VAL_COMPLEX(89, -147), MAKE_VAL_COMPLEX(-2, -55), MAKE_VAL_COMPLEX(-82, 50), MAKE_VAL_COMPLEX(148, -82), MAKE_VAL_COMPLEX(-14, 168), MAKE_VAL_COMPLEX(-123, -155), MAKE_VAL_COMPLEX(96, 138), MAKE_VAL_COMPLEX(-124, -46), MAKE_VAL_COMPLEX(60, 280), MAKE_VAL_COMPLEX(-128, -164), MAKE_VAL_COMPLEX(33, -16), MAKE_VAL_COMPLEX(-133, -7), MAKE_VAL_COMPLEX(2, 161) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1532, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1532", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1532() */
/* END TEST 1532 */

/*
 *  BEGIN TEST 1533
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1533( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(102, -76), MAKE_VAL_COMPLEX(-137, -74), MAKE_VAL_COMPLEX(-39, 33), MAKE_VAL_COMPLEX(-70, 59), MAKE_VAL_COMPLEX(112, 48), MAKE_VAL_COMPLEX(-174, 56), MAKE_VAL_COMPLEX(-26, -108), MAKE_VAL_COMPLEX(105, 169), MAKE_VAL_COMPLEX(-28, -58), MAKE_VAL_COMPLEX(66, 120), MAKE_VAL_COMPLEX(10, -346), MAKE_VAL_COMPLEX(64, 208), MAKE_VAL_COMPLEX(-149, -44), MAKE_VAL_COMPLEX(121, 51), MAKE_VAL_COMPLEX(16, -71) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1533, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1533", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1533() */
/* END TEST 1533 */

/*
 *  BEGIN TEST 1534
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1534( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(31, -126), MAKE_VAL_COMPLEX(-74, -20), MAKE_VAL_COMPLEX(-14, -129), MAKE_VAL_COMPLEX(-63, 41), MAKE_VAL_COMPLEX(78, 73), MAKE_VAL_COMPLEX(-37, -10), MAKE_VAL_COMPLEX(-5, 75), MAKE_VAL_COMPLEX(-9, 22), MAKE_VAL_COMPLEX(103, 37), MAKE_VAL_COMPLEX(-14, 97), MAKE_VAL_COMPLEX(29, -96), MAKE_VAL_COMPLEX(-35, 73), MAKE_VAL_COMPLEX(-136, 6), MAKE_VAL_COMPLEX(12, 46), MAKE_VAL_COMPLEX(63, 72) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1534, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1534", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1534() */
/* END TEST 1534 */

/*
 *  BEGIN TEST 1535
 *  Value type: complex
 *  A : <MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1488', LAYOUT_COLMAJ>
 *  y : <'y1488', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1535( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1488_ptr, A1488_ind, A1488_val,
		A1488_M, A1488_N, SHARE_INPUTMAT,
		5, MAT_HERM_UPPER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1488_val,
		y1488_LEN * y1488_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1488_LEN, y1488_VW, LAYOUT_COLMAJ, y1488_LEN );
	oski_value_t* y_val = array_Duplicate( x1488_val,
		x1488_LEN * x1488_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1488_LEN, x1488_VW, LAYOUT_COLMAJ, x1488_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-116, 92), MAKE_VAL_COMPLEX(159, -11), MAKE_VAL_COMPLEX(-108, -122), MAKE_VAL_COMPLEX(15, -57), MAKE_VAL_COMPLEX(-62, 34), MAKE_VAL_COMPLEX(92, -204), MAKE_VAL_COMPLEX(194, 122), MAKE_VAL_COMPLEX(-260, -46), MAKE_VAL_COMPLEX(166, -38), MAKE_VAL_COMPLEX(-112, 4), MAKE_VAL_COMPLEX(270, 286), MAKE_VAL_COMPLEX(-228, -80), MAKE_VAL_COMPLEX(133, 11), MAKE_VAL_COMPLEX(-128, 82), MAKE_VAL_COMPLEX(145, 69) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1535, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1535", y_true, y_val, x1488_LEN*x1488_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1535() */
/* END TEST 1535 */

#define A1536_M 5    /* # of rows */
#define A1536_N 5    /* # of cols */
#define A1536_NNZ 8  /* # of stored non-zeros */
static oski_index_t A1536_ptr[A1536_M+1] = {1, 1, 2, 4, 6, 9}; /* end row pointers for A1536 */
static oski_index_t A1536_ind[A1536_NNZ] = {1, 2, 3, 2, 3, 1, 2, 3}; /* end column indices pointers for A1536 */
static oski_value_t A1536_val[A1536_NNZ] = {MAKE_VAL_COMPLEX(-10, -4), MAKE_VAL_COMPLEX(1, 10), MAKE_VAL_COMPLEX(3, 0), MAKE_VAL_COMPLEX(0, -6), MAKE_VAL_COMPLEX(1, 0), MAKE_VAL_COMPLEX(-9, 6), MAKE_VAL_COMPLEX(-1, 10), MAKE_VAL_COMPLEX(10, 3)}; /* end stored non-zero values for A1536 */

#define x1536_LEN 5   /* vector length */
#define x1536_VW  3   /* vector width */
static oski_value_t x1536_val[x1536_LEN*x1536_VW] = {MAKE_VAL_COMPLEX(-1, -8), MAKE_VAL_COMPLEX(-1, 8), MAKE_VAL_COMPLEX(7, 10), MAKE_VAL_COMPLEX(3, -10), MAKE_VAL_COMPLEX(-8, 8), MAKE_VAL_COMPLEX(3, 2), MAKE_VAL_COMPLEX(-5, -2), MAKE_VAL_COMPLEX(7, -6), MAKE_VAL_COMPLEX(5, -3), MAKE_VAL_COMPLEX(-10, -8), MAKE_VAL_COMPLEX(9, -5), MAKE_VAL_COMPLEX(5, 1), MAKE_VAL_COMPLEX(10, 5), MAKE_VAL_COMPLEX(7, -4), MAKE_VAL_COMPLEX(-6, 5)}; /* data */

#define y1536_LEN 5   /* vector length */
#define y1536_VW  3   /* vector width */
static oski_value_t y1536_val[y1536_LEN*y1536_VW] = {MAKE_VAL_COMPLEX(5, 0), MAKE_VAL_COMPLEX(2, 5), MAKE_VAL_COMPLEX(10, 9), MAKE_VAL_COMPLEX(7, 5), MAKE_VAL_COMPLEX(-10, -7), MAKE_VAL_COMPLEX(-7, -10), MAKE_VAL_COMPLEX(-6, 4), MAKE_VAL_COMPLEX(-10, 10), MAKE_VAL_COMPLEX(-7, 1), MAKE_VAL_COMPLEX(1, 10), MAKE_VAL_COMPLEX(7, -9), MAKE_VAL_COMPLEX(-9, -1), MAKE_VAL_COMPLEX(-5, 5), MAKE_VAL_COMPLEX(7, -1), MAKE_VAL_COMPLEX(-1, -9)}; /* data */

/*
 *  BEGIN TEST 1536
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1536( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	oski_value_t* y_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-20, -13), MAKE_VAL_COMPLEX(-7, -96), MAKE_VAL_COMPLEX(142, -3), MAKE_VAL_COMPLEX(103, -31), MAKE_VAL_COMPLEX(-116, -166), MAKE_VAL_COMPLEX(-87, -128), MAKE_VAL_COMPLEX(173, 74), MAKE_VAL_COMPLEX(-20, -56), MAKE_VAL_COMPLEX(-80, 126), MAKE_VAL_COMPLEX(-19, 36), MAKE_VAL_COMPLEX(61, -40), MAKE_VAL_COMPLEX(-41, 9), MAKE_VAL_COMPLEX(100, 121), MAKE_VAL_COMPLEX(15, -239), MAKE_VAL_COMPLEX(-129, -85) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1536, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1536", y_true, y_val, y1536_LEN*y1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1536() */
/* END TEST 1536 */

/*
 *  BEGIN TEST 1537
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1537( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	oski_value_t* y_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(55, 13), MAKE_VAL_COMPLEX(21, 131), MAKE_VAL_COMPLEX(-72, 66), MAKE_VAL_COMPLEX(-54, 66), MAKE_VAL_COMPLEX(46, 117), MAKE_VAL_COMPLEX(38, 58), MAKE_VAL_COMPLEX(-215, -46), MAKE_VAL_COMPLEX(-50, 126), MAKE_VAL_COMPLEX(31, -119), MAKE_VAL_COMPLEX(26, 34), MAKE_VAL_COMPLEX(-12, -23), MAKE_VAL_COMPLEX(-22, -16), MAKE_VAL_COMPLEX(-135, -86), MAKE_VAL_COMPLEX(34, 232), MAKE_VAL_COMPLEX(122, 22) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1537, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1537", y_true, y_val, y1536_LEN*y1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1537() */
/* END TEST 1537 */

/*
 *  BEGIN TEST 1538
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1538( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	oski_value_t* y_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-10, -40), MAKE_VAL_COMPLEX(36, -26), MAKE_VAL_COMPLEX(52, -98), MAKE_VAL_COMPLEX(26, -66), MAKE_VAL_COMPLEX(-36, 94), MAKE_VAL_COMPLEX(-66, 76), MAKE_VAL_COMPLEX(44, 40), MAKE_VAL_COMPLEX(100, 60), MAKE_VAL_COMPLEX(22, 54), MAKE_VAL_COMPLEX(78, -28), MAKE_VAL_COMPLEX(-86, -38), MAKE_VAL_COMPLEX(10, 74), MAKE_VAL_COMPLEX(50, 30), MAKE_VAL_COMPLEX(-22, -54), MAKE_VAL_COMPLEX(-70, 26) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1538, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1538", y_true, y_val, y1536_LEN*y1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1538() */
/* END TEST 1538 */

/*
 *  BEGIN TEST 1539
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1539( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	oski_value_t* y_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(168, -375), MAKE_VAL_COMPLEX(927, 300), MAKE_VAL_COMPLEX(-136, 209), MAKE_VAL_COMPLEX(-23, 315), MAKE_VAL_COMPLEX(1148, 788), MAKE_VAL_COMPLEX(577, 586), MAKE_VAL_COMPLEX(-1041, 1044), MAKE_VAL_COMPLEX(1096, 780), MAKE_VAL_COMPLEX(-278, -664), MAKE_VAL_COMPLEX(517, -378), MAKE_VAL_COMPLEX(-655, 136), MAKE_VAL_COMPLEX(105, 287), MAKE_VAL_COMPLEX(-896, 305), MAKE_VAL_COMPLEX(1281, 839), MAKE_VAL_COMPLEX(581, -267) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -8);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1539, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1539", y_true, y_val, y1536_LEN*y1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1539() */
/* END TEST 1539 */

/*
 *  BEGIN TEST 1540
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1540( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(8, -134), MAKE_VAL_COMPLEX(-22, 132), MAKE_VAL_COMPLEX(22, 208), MAKE_VAL_COMPLEX(-40, -52), MAKE_VAL_COMPLEX(-217, -167), MAKE_VAL_COMPLEX(-87, -57), MAKE_VAL_COMPLEX(0, -27), MAKE_VAL_COMPLEX(-21, 178), MAKE_VAL_COMPLEX(-120, 3), MAKE_VAL_COMPLEX(34, 20), MAKE_VAL_COMPLEX(30, 10), MAKE_VAL_COMPLEX(33, -19), MAKE_VAL_COMPLEX(-105, -12), MAKE_VAL_COMPLEX(-119, 227), MAKE_VAL_COMPLEX(-203, -75) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1540, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1540", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1540() */
/* END TEST 1540 */

/*
 *  BEGIN TEST 1541
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1541( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-29, 161), MAKE_VAL_COMPLEX(49, -153), MAKE_VAL_COMPLEX(-13, -259), MAKE_VAL_COMPLEX(1, 73), MAKE_VAL_COMPLEX(265, 167), MAKE_VAL_COMPLEX(84, 42), MAKE_VAL_COMPLEX(9, 48), MAKE_VAL_COMPLEX(-18, -181), MAKE_VAL_COMPLEX(96, -9), MAKE_VAL_COMPLEX(-28, 34), MAKE_VAL_COMPLEX(-72, -22), MAKE_VAL_COMPLEX(-45, 1), MAKE_VAL_COMPLEX(90, -33), MAKE_VAL_COMPLEX(86, -236), MAKE_VAL_COMPLEX(236, 78) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1541, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1541", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1541() */
/* END TEST 1541 */

/*
 *  BEGIN TEST 1542
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1542( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(39, -13), MAKE_VAL_COMPLEX(-41, 3), MAKE_VAL_COMPLEX(-43, 45), MAKE_VAL_COMPLEX(53, 5), MAKE_VAL_COMPLEX(-48, -32), MAKE_VAL_COMPLEX(-7, 17), MAKE_VAL_COMPLEX(5, -27), MAKE_VAL_COMPLEX(37, 29), MAKE_VAL_COMPLEX(20, 22), MAKE_VAL_COMPLEX(30, -58), MAKE_VAL_COMPLEX(34, 40), MAKE_VAL_COMPLEX(0, 26), MAKE_VAL_COMPLEX(-15, 55), MAKE_VAL_COMPLEX(27, 31), MAKE_VAL_COMPLEX(-31, -25) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1542, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1542", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1542() */
/* END TEST 1542 */

/*
 *  BEGIN TEST 1543
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1543( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1094, 123), MAKE_VAL_COMPLEX(1138, 23), MAKE_VAL_COMPLEX(1476, -481), MAKE_VAL_COMPLEX(-762, 513), MAKE_VAL_COMPLEX(-898, 1264), MAKE_VAL_COMPLEX(-544, 405), MAKE_VAL_COMPLEX(-59, 170), MAKE_VAL_COMPLEX(823, 366), MAKE_VAL_COMPLEX(-354, 826), MAKE_VAL_COMPLEX(334, 320), MAKE_VAL_COMPLEX(-349, -245), MAKE_VAL_COMPLEX(-255, -411), MAKE_VAL_COMPLEX(-424, 293), MAKE_VAL_COMPLEX(1118, 1029), MAKE_VAL_COMPLEX(-387, 1316) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1543, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1543", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1543() */
/* END TEST 1543 */

/*
 *  BEGIN TEST 1544
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1544( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(17, -106), MAKE_VAL_COMPLEX(11, 56), MAKE_VAL_COMPLEX(71, 274), MAKE_VAL_COMPLEX(53, 30), MAKE_VAL_COMPLEX(23, 49), MAKE_VAL_COMPLEX(-130, -63), MAKE_VAL_COMPLEX(-125, 121), MAKE_VAL_COMPLEX(190, -130), MAKE_VAL_COMPLEX(81, -164), MAKE_VAL_COMPLEX(-16, -152), MAKE_VAL_COMPLEX(45, 75), MAKE_VAL_COMPLEX(-32, 76), MAKE_VAL_COMPLEX(-119, 207), MAKE_VAL_COMPLEX(-22, -53), MAKE_VAL_COMPLEX(-183, -82) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1544, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1544", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1544() */
/* END TEST 1544 */

/*
 *  BEGIN TEST 1545
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1545( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(71, -35), MAKE_VAL_COMPLEX(-131, 59), MAKE_VAL_COMPLEX(-89, -23), MAKE_VAL_COMPLEX(125, -151), MAKE_VAL_COMPLEX(-255, -25), MAKE_VAL_COMPLEX(152, 134), MAKE_VAL_COMPLEX(71, -218), MAKE_VAL_COMPLEX(0, 125), MAKE_VAL_COMPLEX(38, 181), MAKE_VAL_COMPLEX(-40, -106), MAKE_VAL_COMPLEX(164, -38), MAKE_VAL_COMPLEX(99, 5), MAKE_VAL_COMPLEX(214, 3), MAKE_VAL_COMPLEX(186, 80), MAKE_VAL_COMPLEX(22, 84) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1545, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1545", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1545() */
/* END TEST 1545 */

/*
 *  BEGIN TEST 1546
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1546( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-4, -32), MAKE_VAL_COMPLEX(-4, 32), MAKE_VAL_COMPLEX(28, 40), MAKE_VAL_COMPLEX(12, -40), MAKE_VAL_COMPLEX(-32, 32), MAKE_VAL_COMPLEX(12, 8), MAKE_VAL_COMPLEX(-20, -8), MAKE_VAL_COMPLEX(28, -24), MAKE_VAL_COMPLEX(20, -12), MAKE_VAL_COMPLEX(-40, -32), MAKE_VAL_COMPLEX(36, -20), MAKE_VAL_COMPLEX(20, 4), MAKE_VAL_COMPLEX(40, 20), MAKE_VAL_COMPLEX(28, -16), MAKE_VAL_COMPLEX(-24, 20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 0);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1546, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1546", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1546() */
/* END TEST 1546 */

/*
 *  BEGIN TEST 1547
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1547( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(174, 225), MAKE_VAL_COMPLEX(-296, 59), MAKE_VAL_COMPLEX(-726, -741), MAKE_VAL_COMPLEX(-92, -421), MAKE_VAL_COMPLEX(-629, 41), MAKE_VAL_COMPLEX(877, 170), MAKE_VAL_COMPLEX(200, -992), MAKE_VAL_COMPLEX(-337, 861), MAKE_VAL_COMPLEX(161, 914), MAKE_VAL_COMPLEX(62, 290), MAKE_VAL_COMPLEX(49, -391), MAKE_VAL_COMPLEX(202, -348), MAKE_VAL_COMPLEX(536, -958), MAKE_VAL_COMPLEX(547, 134), MAKE_VAL_COMPLEX(778, 183) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1547, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1547", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1547() */
/* END TEST 1547 */

/*
 *  BEGIN TEST 1548
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1548( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	oski_value_t* y_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-15, 52), MAKE_VAL_COMPLEX(10, 20), MAKE_VAL_COMPLEX(182, 179), MAKE_VAL_COMPLEX(-61, 78), MAKE_VAL_COMPLEX(103, -68), MAKE_VAL_COMPLEX(2, -251), MAKE_VAL_COMPLEX(-106, -190), MAKE_VAL_COMPLEX(-160, -176), MAKE_VAL_COMPLEX(-3, -14), MAKE_VAL_COMPLEX(-106, -126), MAKE_VAL_COMPLEX(176, -17), MAKE_VAL_COMPLEX(-59, -184), MAKE_VAL_COMPLEX(-122, 82), MAKE_VAL_COMPLEX(75, 35), MAKE_VAL_COMPLEX(-13, -107) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1548, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1548", y_true, y_val, y1536_LEN*y1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1548() */
/* END TEST 1548 */

/*
 *  BEGIN TEST 1549
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1549( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	oski_value_t* y_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(25, -17), MAKE_VAL_COMPLEX(-41, 4), MAKE_VAL_COMPLEX(-225, -91), MAKE_VAL_COMPLEX(40, -19), MAKE_VAL_COMPLEX(-74, -16), MAKE_VAL_COMPLEX(54, 182), MAKE_VAL_COMPLEX(66, 156), MAKE_VAL_COMPLEX(70, 126), MAKE_VAL_COMPLEX(-18, -33), MAKE_VAL_COMPLEX(38, 153), MAKE_VAL_COMPLEX(-99, 48), MAKE_VAL_COMPLEX(48, 119), MAKE_VAL_COMPLEX(77, -107), MAKE_VAL_COMPLEX(-54, 12), MAKE_VAL_COMPLEX(74, 82) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1549, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1549", y_true, y_val, y1536_LEN*y1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1549() */
/* END TEST 1549 */

/*
 *  BEGIN TEST 1550
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1550( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	oski_value_t* y_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(15, 10), MAKE_VAL_COMPLEX(-4, 19), MAKE_VAL_COMPLEX(12, 47), MAKE_VAL_COMPLEX(11, 29), MAKE_VAL_COMPLEX(-16, -41), MAKE_VAL_COMPLEX(-1, -44), MAKE_VAL_COMPLEX(-26, 0), MAKE_VAL_COMPLEX(-50, 10), MAKE_VAL_COMPLEX(-23, -11), MAKE_VAL_COMPLEX(-17, 32), MAKE_VAL_COMPLEX(39, -13), MAKE_VAL_COMPLEX(-25, -21), MAKE_VAL_COMPLEX(-25, 5), MAKE_VAL_COMPLEX(23, 11), MAKE_VAL_COMPLEX(15, -29) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1550, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1550", y_true, y_val, y1536_LEN*y1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1550() */
/* END TEST 1550 */

/*
 *  BEGIN TEST 1551
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1551( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	oski_value_t* y_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(399, 9), MAKE_VAL_COMPLEX(-365, 153), MAKE_VAL_COMPLEX(-1459, -691), MAKE_VAL_COMPLEX(494, 226), MAKE_VAL_COMPLEX(-757, -697), MAKE_VAL_COMPLEX(27, 1101), MAKE_VAL_COMPLEX(-58, 1398), MAKE_VAL_COMPLEX(-212, 1288), MAKE_VAL_COMPLEX(-328, -400), MAKE_VAL_COMPLEX(-159, 1649), MAKE_VAL_COMPLEX(-486, 22), MAKE_VAL_COMPLEX(-132, 844), MAKE_VAL_COMPLEX(570, -632), MAKE_VAL_COMPLEX(-206, 88), MAKE_VAL_COMPLEX(566, 488) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, -2);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1551, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1551", y_true, y_val, y1536_LEN*y1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1551() */
/* END TEST 1551 */

/*
 *  BEGIN TEST 1552
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1552( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(142, -40), MAKE_VAL_COMPLEX(1, -117), MAKE_VAL_COMPLEX(125, 108), MAKE_VAL_COMPLEX(-16, 17), MAKE_VAL_COMPLEX(-87, 0), MAKE_VAL_COMPLEX(-12, 29), MAKE_VAL_COMPLEX(22, -84), MAKE_VAL_COMPLEX(-31, 217), MAKE_VAL_COMPLEX(-17, 21), MAKE_VAL_COMPLEX(8, 27), MAKE_VAL_COMPLEX(-47, -7), MAKE_VAL_COMPLEX(-10, -46), MAKE_VAL_COMPLEX(105, 5), MAKE_VAL_COMPLEX(-31, 333), MAKE_VAL_COMPLEX(-146, 179) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1552, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1552", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1552() */
/* END TEST 1552 */

/*
 *  BEGIN TEST 1553
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1553( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-94, 99), MAKE_VAL_COMPLEX(-33, 48), MAKE_VAL_COMPLEX(-231, -153), MAKE_VAL_COMPLEX(42, 78), MAKE_VAL_COMPLEX(111, -104), MAKE_VAL_COMPLEX(-22, -30), MAKE_VAL_COMPLEX(28, 75), MAKE_VAL_COMPLEX(5, -134), MAKE_VAL_COMPLEX(-8, 28), MAKE_VAL_COMPLEX(112, -13), MAKE_VAL_COMPLEX(0, 92), MAKE_VAL_COMPLEX(-35, 63), MAKE_VAL_COMPLEX(-210, 5), MAKE_VAL_COMPLEX(-5, -266), MAKE_VAL_COMPLEX(169, -249) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1553, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1553", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1553() */
/* END TEST 1553 */

/*
 *  BEGIN TEST 1554
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1554( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-24, -62), MAKE_VAL_COMPLEX(8, 66), MAKE_VAL_COMPLEX(76, 66), MAKE_VAL_COMPLEX(4, -86), MAKE_VAL_COMPLEX(-48, 80), MAKE_VAL_COMPLEX(28, 10), MAKE_VAL_COMPLEX(-44, -6), MAKE_VAL_COMPLEX(44, -62), MAKE_VAL_COMPLEX(34, -34), MAKE_VAL_COMPLEX(-96, -44), MAKE_VAL_COMPLEX(62, -58), MAKE_VAL_COMPLEX(42, -2), MAKE_VAL_COMPLEX(90, 20), MAKE_VAL_COMPLEX(48, -46), MAKE_VAL_COMPLEX(-38, 52) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1554, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1554", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1554() */
/* END TEST 1554 */

/*
 *  BEGIN TEST 1555
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1555( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1272, -568), MAKE_VAL_COMPLEX(-792, 670), MAKE_VAL_COMPLEX(-225, -1825), MAKE_VAL_COMPLEX(157, 167), MAKE_VAL_COMPLEX(769, 481), MAKE_VAL_COMPLEX(237, -161), MAKE_VAL_COMPLEX(-656, 480), MAKE_VAL_COMPLEX(1598, -1208), MAKE_VAL_COMPLEX(170, 16), MAKE_VAL_COMPLEX(297, -81), MAKE_VAL_COMPLEX(108, 450), MAKE_VAL_COMPLEX(-340, 364), MAKE_VAL_COMPLEX(-870, -880), MAKE_VAL_COMPLEX(2414, -2056), MAKE_VAL_COMPLEX(2393, -309) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1555, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1555", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1555() */
/* END TEST 1555 */

/*
 *  BEGIN TEST 1556
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1556( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1, 139), MAKE_VAL_COMPLEX(118, -214), MAKE_VAL_COMPLEX(20, -23), MAKE_VAL_COMPLEX(-81, 102), MAKE_VAL_COMPLEX(285, 136), MAKE_VAL_COMPLEX(-161, 158), MAKE_VAL_COMPLEX(-101, 41), MAKE_VAL_COMPLEX(-104, 100), MAKE_VAL_COMPLEX(-85, -153), MAKE_VAL_COMPLEX(94, 73), MAKE_VAL_COMPLEX(-59, 63), MAKE_VAL_COMPLEX(-44, 42), MAKE_VAL_COMPLEX(-84, 110), MAKE_VAL_COMPLEX(-98, 68), MAKE_VAL_COMPLEX(-3, 46) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1556, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1556", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1556() */
/* END TEST 1556 */

/*
 *  BEGIN TEST 1557
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1557( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, -147), MAKE_VAL_COMPLEX(-119, 222), MAKE_VAL_COMPLEX(-13, 33), MAKE_VAL_COMPLEX(84, -112), MAKE_VAL_COMPLEX(-293, -128), MAKE_VAL_COMPLEX(164, -156), MAKE_VAL_COMPLEX(96, -43), MAKE_VAL_COMPLEX(111, -106), MAKE_VAL_COMPLEX(90, 150), MAKE_VAL_COMPLEX(-104, -81), MAKE_VAL_COMPLEX(68, -68), MAKE_VAL_COMPLEX(49, -41), MAKE_VAL_COMPLEX(94, -105), MAKE_VAL_COMPLEX(105, -72), MAKE_VAL_COMPLEX(-3, -41) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1557, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1557", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1557() */
/* END TEST 1557 */

/*
 *  BEGIN TEST 1558
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1558( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(12, -34), MAKE_VAL_COMPLEX(-20, 30), MAKE_VAL_COMPLEX(8, 54), MAKE_VAL_COMPLEX(32, -34), MAKE_VAL_COMPLEX(-48, 16), MAKE_VAL_COMPLEX(8, 14), MAKE_VAL_COMPLEX(-16, -18), MAKE_VAL_COMPLEX(40, -10), MAKE_VAL_COMPLEX(26, -2), MAKE_VAL_COMPLEX(-24, -52), MAKE_VAL_COMPLEX(46, -2), MAKE_VAL_COMPLEX(18, 14), MAKE_VAL_COMPLEX(30, 40), MAKE_VAL_COMPLEX(36, -2), MAKE_VAL_COMPLEX(-34, 8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1558, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1558", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1558() */
/* END TEST 1558 */

/*
 *  BEGIN TEST 1559
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_ROWMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1559( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_ROWMAJ, x1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(166, 186), MAKE_VAL_COMPLEX(-29, -463), MAKE_VAL_COMPLEX(44, 79), MAKE_VAL_COMPLEX(23, 190), MAKE_VAL_COMPLEX(578, 35), MAKE_VAL_COMPLEX(-141, 514), MAKE_VAL_COMPLEX(-206, 136), MAKE_VAL_COMPLEX(-1, 273), MAKE_VAL_COMPLEX(-253, -229), MAKE_VAL_COMPLEX(191, -86), MAKE_VAL_COMPLEX(69, 175), MAKE_VAL_COMPLEX(4, 164), MAKE_VAL_COMPLEX(27, 409), MAKE_VAL_COMPLEX(-31, 225), MAKE_VAL_COMPLEX(-51, 120) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1559, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1559", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1559() */
/* END TEST 1559 */

/*
 *  BEGIN TEST 1560
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1560( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	oski_value_t* y_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(113, -63), MAKE_VAL_COMPLEX(61, 165), MAKE_VAL_COMPLEX(-21, 118), MAKE_VAL_COMPLEX(209, 192), MAKE_VAL_COMPLEX(-94, -81), MAKE_VAL_COMPLEX(99, -77), MAKE_VAL_COMPLEX(-167, 80), MAKE_VAL_COMPLEX(-203, -183), MAKE_VAL_COMPLEX(-43, 70), MAKE_VAL_COMPLEX(-32, 55), MAKE_VAL_COMPLEX(97, 60), MAKE_VAL_COMPLEX(-2, -109), MAKE_VAL_COMPLEX(-42, 139), MAKE_VAL_COMPLEX(104, -27), MAKE_VAL_COMPLEX(97, 192) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1560, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1560", y_true, y_val, y1536_LEN*y1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1560() */
/* END TEST 1560 */

/*
 *  BEGIN TEST 1561
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1561( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	oski_value_t* y_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-113, 113), MAKE_VAL_COMPLEX(-111, -145), MAKE_VAL_COMPLEX(-69, -18), MAKE_VAL_COMPLEX(-259, -122), MAKE_VAL_COMPLEX(164, -19), MAKE_VAL_COMPLEX(1, 7), MAKE_VAL_COMPLEX(127, -140), MAKE_VAL_COMPLEX(103, 83), MAKE_VAL_COMPLEX(33, -140), MAKE_VAL_COMPLEX(-68, -45), MAKE_VAL_COMPLEX(-7, 10), MAKE_VAL_COMPLEX(12, 19), MAKE_VAL_COMPLEX(-8, -189), MAKE_VAL_COMPLEX(-94, 97), MAKE_VAL_COMPLEX(-7, -202) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1561, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1561", y_true, y_val, y1536_LEN*y1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1561() */
/* END TEST 1561 */

/*
 *  BEGIN TEST 1562
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1562( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	oski_value_t* y_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-25, -30), MAKE_VAL_COMPLEX(20, -37), MAKE_VAL_COMPLEX(4, -105), MAKE_VAL_COMPLEX(-5, -67), MAKE_VAL_COMPLEX(8, 95), MAKE_VAL_COMPLEX(-25, 92), MAKE_VAL_COMPLEX(54, 16), MAKE_VAL_COMPLEX(110, 10), MAKE_VAL_COMPLEX(41, 37), MAKE_VAL_COMPLEX(55, -56), MAKE_VAL_COMPLEX(-89, 3), MAKE_VAL_COMPLEX(39, 59), MAKE_VAL_COMPLEX(55, 5), MAKE_VAL_COMPLEX(-41, -37), MAKE_VAL_COMPLEX(-49, 51) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1562, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1562", y_true, y_val, y1536_LEN*y1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1562() */
/* END TEST 1562 */

/*
 *  BEGIN TEST 1563
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1563( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	oski_value_t* y_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1761, 163), MAKE_VAL_COMPLEX(175, -2013), MAKE_VAL_COMPLEX(-247, -326), MAKE_VAL_COMPLEX(-1181, -2940), MAKE_VAL_COMPLEX(1378, 817), MAKE_VAL_COMPLEX(-171, -319), MAKE_VAL_COMPLEX(2031, -176), MAKE_VAL_COMPLEX(-157, 1791), MAKE_VAL_COMPLEX(1183, -1034), MAKE_VAL_COMPLEX(-320, -619), MAKE_VAL_COMPLEX(171, -184), MAKE_VAL_COMPLEX(-322, 145), MAKE_VAL_COMPLEX(1230, -1655), MAKE_VAL_COMPLEX(-1388, 159), MAKE_VAL_COMPLEX(1611, -2172) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-9, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1563, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1563", y_true, y_val, y1536_LEN*y1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1563() */
/* END TEST 1563 */

/*
 *  BEGIN TEST 1564
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1564( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-50, -143), MAKE_VAL_COMPLEX(-64, -103), MAKE_VAL_COMPLEX(-12, -32), MAKE_VAL_COMPLEX(-59, 50), MAKE_VAL_COMPLEX(-26, -39), MAKE_VAL_COMPLEX(16, 153), MAKE_VAL_COMPLEX(-200, -91), MAKE_VAL_COMPLEX(-49, 117), MAKE_VAL_COMPLEX(21, -57), MAKE_VAL_COMPLEX(-124, 208), MAKE_VAL_COMPLEX(74, 190), MAKE_VAL_COMPLEX(-71, -77), MAKE_VAL_COMPLEX(-120, -56), MAKE_VAL_COMPLEX(38, -51), MAKE_VAL_COMPLEX(-180, -23) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1564, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1564", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1564() */
/* END TEST 1564 */

/*
 *  BEGIN TEST 1565
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1565( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(23, 122), MAKE_VAL_COMPLEX(85, 130), MAKE_VAL_COMPLEX(63, 41), MAKE_VAL_COMPLEX(38, -89), MAKE_VAL_COMPLEX(26, 87), MAKE_VAL_COMPLEX(-1, -156), MAKE_VAL_COMPLEX(179, 100), MAKE_VAL_COMPLEX(52, -156), MAKE_VAL_COMPLEX(-15, 33), MAKE_VAL_COMPLEX(70, -202), MAKE_VAL_COMPLEX(-62, -232), MAKE_VAL_COMPLEX(89, 65), MAKE_VAL_COMPLEX(165, 41), MAKE_VAL_COMPLEX(-29, 18), MAKE_VAL_COMPLEX(177, 56) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1565, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1565", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1565() */
/* END TEST 1565 */

/*
 *  BEGIN TEST 1566
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1566( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 65), MAKE_VAL_COMPLEX(16, -63), MAKE_VAL_COMPLEX(-46, -87), MAKE_VAL_COMPLEX(-34, 77), MAKE_VAL_COMPLEX(72, -56), MAKE_VAL_COMPLEX(-22, -19), MAKE_VAL_COMPLEX(38, 21), MAKE_VAL_COMPLEX(-62, 41), MAKE_VAL_COMPLEX(-43, 19), MAKE_VAL_COMPLEX(72, 74), MAKE_VAL_COMPLEX(-77, 31), MAKE_VAL_COMPLEX(-39, -13), MAKE_VAL_COMPLEX(-75, -50), MAKE_VAL_COMPLEX(-60, 25), MAKE_VAL_COMPLEX(53, -34) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1566, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1566", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1566() */
/* END TEST 1566 */

/*
 *  BEGIN TEST 1567
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1567( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1019, -271), MAKE_VAL_COMPLEX(723, -529), MAKE_VAL_COMPLEX(144, -177), MAKE_VAL_COMPLEX(-370, -310), MAKE_VAL_COMPLEX(345, -270), MAKE_VAL_COMPLEX(-1103, 95), MAKE_VAL_COMPLEX(689, -1385), MAKE_VAL_COMPLEX(-883, -276), MAKE_VAL_COMPLEX(352, 182), MAKE_VAL_COMPLEX(-1348, -798), MAKE_VAL_COMPLEX(-1415, 577), MAKE_VAL_COMPLEX(488, -502), MAKE_VAL_COMPLEX(287, -880), MAKE_VAL_COMPLEX(291, 313), MAKE_VAL_COMPLEX(216, -1316) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1567, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1567", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1567() */
/* END TEST 1567 */

/*
 *  BEGIN TEST 1568
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1568( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(45, -77), MAKE_VAL_COMPLEX(-89, 119), MAKE_VAL_COMPLEX(-147, 258), MAKE_VAL_COMPLEX(116, -54), MAKE_VAL_COMPLEX(-270, 85), MAKE_VAL_COMPLEX(55, 29), MAKE_VAL_COMPLEX(123, -23), MAKE_VAL_COMPLEX(180, -85), MAKE_VAL_COMPLEX(-8, 98), MAKE_VAL_COMPLEX(-74, -168), MAKE_VAL_COMPLEX(141, 211), MAKE_VAL_COMPLEX(-132, -50), MAKE_VAL_COMPLEX(26, -65), MAKE_VAL_COMPLEX(-7, 83), MAKE_VAL_COMPLEX(-174, -120) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1568, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1568", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1568() */
/* END TEST 1568 */

/*
 *  BEGIN TEST 1569
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1569( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(17, -12), MAKE_VAL_COMPLEX(7, -48), MAKE_VAL_COMPLEX(127, -95), MAKE_VAL_COMPLEX(4, -19), MAKE_VAL_COMPLEX(118, -77), MAKE_VAL_COMPLEX(-43, 18), MAKE_VAL_COMPLEX(-155, -42), MAKE_VAL_COMPLEX(-56, 88), MAKE_VAL_COMPLEX(85, -83), MAKE_VAL_COMPLEX(46, -2), MAKE_VAL_COMPLEX(-6, -180), MAKE_VAL_COMPLEX(173, 105), MAKE_VAL_COMPLEX(29, 205), MAKE_VAL_COMPLEX(113, -60), MAKE_VAL_COMPLEX(69, 116) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1569, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1569", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1569() */
/* END TEST 1569 */

/*
 *  BEGIN TEST 1570
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1570( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-69, 33), MAKE_VAL_COMPLEX(75, -15), MAKE_VAL_COMPLEX(69, -93), MAKE_VAL_COMPLEX(-99, 3), MAKE_VAL_COMPLEX(96, 48), MAKE_VAL_COMPLEX(9, -33), MAKE_VAL_COMPLEX(-3, 51), MAKE_VAL_COMPLEX(-75, -45), MAKE_VAL_COMPLEX(-42, -36), MAKE_VAL_COMPLEX(-42, 114), MAKE_VAL_COMPLEX(-72, -66), MAKE_VAL_COMPLEX(-6, -48), MAKE_VAL_COMPLEX(15, -105), MAKE_VAL_COMPLEX(-57, -51), MAKE_VAL_COMPLEX(63, 39) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1570, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1570", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1570() */
/* END TEST 1570 */

/*
 *  BEGIN TEST 1571
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1571( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_ROWMAJ, y1536_VW );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(189, -145), MAKE_VAL_COMPLEX(-607, -91), MAKE_VAL_COMPLEX(-1012, -661), MAKE_VAL_COMPLEX(204, 84), MAKE_VAL_COMPLEX(-803, -1106), MAKE_VAL_COMPLEX(55, 401), MAKE_VAL_COMPLEX(-221, 917), MAKE_VAL_COMPLEX(813, 636), MAKE_VAL_COMPLEX(-484, -422), MAKE_VAL_COMPLEX(268, -654), MAKE_VAL_COMPLEX(-1099, 353), MAKE_VAL_COMPLEX(696, -1042), MAKE_VAL_COMPLEX(1255, 132), MAKE_VAL_COMPLEX(-291, -549), MAKE_VAL_COMPLEX(648, -692) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1571, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1571", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1571() */
/* END TEST 1571 */

/*
 *  BEGIN TEST 1572
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1572( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	oski_value_t* y_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(108, -118), MAKE_VAL_COMPLEX(247, 120), MAKE_VAL_COMPLEX(-75, 120), MAKE_VAL_COMPLEX(79, 12), MAKE_VAL_COMPLEX(-16, 175), MAKE_VAL_COMPLEX(66, 126), MAKE_VAL_COMPLEX(-131, 50), MAKE_VAL_COMPLEX(-83, -83), MAKE_VAL_COMPLEX(-17, 40), MAKE_VAL_COMPLEX(96, -69), MAKE_VAL_COMPLEX(26, -31), MAKE_VAL_COMPLEX(10, 32), MAKE_VAL_COMPLEX(-13, 150), MAKE_VAL_COMPLEX(28, -41), MAKE_VAL_COMPLEX(-1, 212) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1572, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1572", y_true, y_val, y1536_LEN*y1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1572() */
/* END TEST 1572 */

/*
 *  BEGIN TEST 1573
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1573( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	oski_value_t* y_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-103, 108), MAKE_VAL_COMPLEX(-235, -119), MAKE_VAL_COMPLEX(103, -131), MAKE_VAL_COMPLEX(-62, -21), MAKE_VAL_COMPLEX(-8, -162), MAKE_VAL_COMPLEX(-93, -122), MAKE_VAL_COMPLEX(133, -34), MAKE_VAL_COMPLEX(93, 113), MAKE_VAL_COMPLEX(12, -25), MAKE_VAL_COMPLEX(-75, 77), MAKE_VAL_COMPLEX(-37, 8), MAKE_VAL_COMPLEX(-21, -15), MAKE_VAL_COMPLEX(18, -135), MAKE_VAL_COMPLEX(-23, 26), MAKE_VAL_COMPLEX(-18, -219) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1573, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1573", y_true, y_val, y1536_LEN*y1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1573() */
/* END TEST 1573 */

/*
 *  BEGIN TEST 1574
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1574( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	oski_value_t* y_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-20, -30), MAKE_VAL_COMPLEX(22, -32), MAKE_VAL_COMPLEX(14, -96), MAKE_VAL_COMPLEX(2, -62), MAKE_VAL_COMPLEX(-2, 88), MAKE_VAL_COMPLEX(-32, 82), MAKE_VAL_COMPLEX(48, 20), MAKE_VAL_COMPLEX(100, 20), MAKE_VAL_COMPLEX(34, 38), MAKE_VAL_COMPLEX(56, -46), MAKE_VAL_COMPLEX(-82, -6), MAKE_VAL_COMPLEX(30, 58), MAKE_VAL_COMPLEX(50, 10), MAKE_VAL_COMPLEX(-34, -38), MAKE_VAL_COMPLEX(-50, 42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1574, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1574", y_true, y_val, y1536_LEN*y1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1574() */
/* END TEST 1574 */

/*
 *  BEGIN TEST 1575
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1575( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	oski_value_t* y_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-48, 1731), MAKE_VAL_COMPLEX(-2996, 798), MAKE_VAL_COMPLEX(26, -2106), MAKE_VAL_COMPLEX(-640, 231), MAKE_VAL_COMPLEX(-1489, -1285), MAKE_VAL_COMPLEX(-1964, -293), MAKE_VAL_COMPLEX(959, -1280), MAKE_VAL_COMPLEX(1841, 433), MAKE_VAL_COMPLEX(-100, -227), MAKE_VAL_COMPLEX(74, 1310), MAKE_VAL_COMPLEX(-365, 250), MAKE_VAL_COMPLEX(-349, 147), MAKE_VAL_COMPLEX(-868, -1279), MAKE_VAL_COMPLEX(9, 324), MAKE_VAL_COMPLEX(-2034, -1841) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-9, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1575, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1575", y_true, y_val, y1536_LEN*y1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1575() */
/* END TEST 1575 */

/*
 *  BEGIN TEST 1576
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1576( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(92, -115), MAKE_VAL_COMPLEX(4, 62), MAKE_VAL_COMPLEX(106, -31), MAKE_VAL_COMPLEX(-36, -71), MAKE_VAL_COMPLEX(98, 101), MAKE_VAL_COMPLEX(39, -96), MAKE_VAL_COMPLEX(-143, 28), MAKE_VAL_COMPLEX(9, 122), MAKE_VAL_COMPLEX(-6, -70), MAKE_VAL_COMPLEX(-133, 294), MAKE_VAL_COMPLEX(201, 45), MAKE_VAL_COMPLEX(40, 18), MAKE_VAL_COMPLEX(90, 10), MAKE_VAL_COMPLEX(41, -109), MAKE_VAL_COMPLEX(-181, 259) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1576, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1576", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1576() */
/* END TEST 1576 */

/*
 *  BEGIN TEST 1577
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1577( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-82, 0), MAKE_VAL_COMPLEX(-42, 47), MAKE_VAL_COMPLEX(-38, 192), MAKE_VAL_COMPLEX(108, -60), MAKE_VAL_COMPLEX(-234, -13), MAKE_VAL_COMPLEX(-3, 133), MAKE_VAL_COMPLEX(79, -71), MAKE_VAL_COMPLEX(107, -185), MAKE_VAL_COMPLEX(85, 43), MAKE_VAL_COMPLEX(17, -436), MAKE_VAL_COMPLEX(-60, -88), MAKE_VAL_COMPLEX(27, 11), MAKE_VAL_COMPLEX(35, 90), MAKE_VAL_COMPLEX(69, 74), MAKE_VAL_COMPLEX(82, -207) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1577, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1577", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1577() */
/* END TEST 1577 */

/*
 *  BEGIN TEST 1578
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1578( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(28, -36), MAKE_VAL_COMPLEX(-36, 28), MAKE_VAL_COMPLEX(-12, 68), MAKE_VAL_COMPLEX(52, -28), MAKE_VAL_COMPLEX(-64, 0), MAKE_VAL_COMPLEX(4, 20), MAKE_VAL_COMPLEX(-12, -28), MAKE_VAL_COMPLEX(52, 4), MAKE_VAL_COMPLEX(32, 8), MAKE_VAL_COMPLEX(-8, -72), MAKE_VAL_COMPLEX(56, 16), MAKE_VAL_COMPLEX(16, 24), MAKE_VAL_COMPLEX(20, 60), MAKE_VAL_COMPLEX(44, 12), MAKE_VAL_COMPLEX(-44, -4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1578, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1578", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1578() */
/* END TEST 1578 */

/*
 *  BEGIN TEST 1579
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1579( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(234, 1110), MAKE_VAL_COMPLEX(18, -108), MAKE_VAL_COMPLEX(606, 362), MAKE_VAL_COMPLEX(-30, -238), MAKE_VAL_COMPLEX(-412, 1014), MAKE_VAL_COMPLEX(788, 292), MAKE_VAL_COMPLEX(-16, -860), MAKE_VAL_COMPLEX(-1448, -560), MAKE_VAL_COMPLEX(330, -198), MAKE_VAL_COMPLEX(-2498, -932), MAKE_VAL_COMPLEX(-1172, 992), MAKE_VAL_COMPLEX(-210, -46), MAKE_VAL_COMPLEX(-40, 10), MAKE_VAL_COMPLEX(464, 138), MAKE_VAL_COMPLEX(-1330, -1632) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-2, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1579, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1579", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1579() */
/* END TEST 1579 */

/*
 *  BEGIN TEST 1580
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1580( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-26, 134), MAKE_VAL_COMPLEX(6, -5), MAKE_VAL_COMPLEX(-124, -78), MAKE_VAL_COMPLEX(-28, 42), MAKE_VAL_COMPLEX(64, 157), MAKE_VAL_COMPLEX(87, -187), MAKE_VAL_COMPLEX(233, 213), MAKE_VAL_COMPLEX(-115, 73), MAKE_VAL_COMPLEX(-31, 39), MAKE_VAL_COMPLEX(-21, 152), MAKE_VAL_COMPLEX(-30, 46), MAKE_VAL_COMPLEX(-169, 147), MAKE_VAL_COMPLEX(-79, -244), MAKE_VAL_COMPLEX(-73, 48), MAKE_VAL_COMPLEX(6, 69) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1580, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1580", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1580() */
/* END TEST 1580 */

/*
 *  BEGIN TEST 1581
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1581( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(48, -23), MAKE_VAL_COMPLEX(0, -108), MAKE_VAL_COMPLEX(16, -55), MAKE_VAL_COMPLEX(-4, 101), MAKE_VAL_COMPLEX(40, -277), MAKE_VAL_COMPLEX(-131, 162), MAKE_VAL_COMPLEX(-161, -190), MAKE_VAL_COMPLEX(23, 18), MAKE_VAL_COMPLEX(-36, 8), MAKE_VAL_COMPLEX(169, -50), MAKE_VAL_COMPLEX(-91, 33), MAKE_VAL_COMPLEX(98, -156), MAKE_VAL_COMPLEX(-66, 184), MAKE_VAL_COMPLEX(-21, 15), MAKE_VAL_COMPLEX(73, -145) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1581, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1581", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1581() */
/* END TEST 1581 */

/*
 *  BEGIN TEST 1582
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1582( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(66, -57), MAKE_VAL_COMPLEX(-78, 39), MAKE_VAL_COMPLEX(-48, 123), MAKE_VAL_COMPLEX(108, -33), MAKE_VAL_COMPLEX(-120, -24), MAKE_VAL_COMPLEX(0, 39), MAKE_VAL_COMPLEX(-12, -57), MAKE_VAL_COMPLEX(96, 27), MAKE_VAL_COMPLEX(57, 27), MAKE_VAL_COMPLEX(12, -138), MAKE_VAL_COMPLEX(99, 51), MAKE_VAL_COMPLEX(21, 51), MAKE_VAL_COMPLEX(15, 120), MAKE_VAL_COMPLEX(78, 39), MAKE_VAL_COMPLEX(-81, -24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1582, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1582", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1582() */
/* END TEST 1582 */

/*
 *  BEGIN TEST 1583
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED>
 *  x : <'x1536', LAYOUT_COLMAJ>
 *  y : <'y1536', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1583( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1536_ptr, A1536_ind, A1536_val,
		A1536_M, A1536_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1536_val,
		y1536_LEN * y1536_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1536_LEN, y1536_VW, LAYOUT_COLMAJ, y1536_LEN );
	oski_value_t* y_val = array_Duplicate( x1536_val,
		x1536_LEN * x1536_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1536_LEN, x1536_VW, LAYOUT_COLMAJ, x1536_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-3, 533), MAKE_VAL_COMPLEX(-179, 195), MAKE_VAL_COMPLEX(-757, 3), MAKE_VAL_COMPLEX(155, 11), MAKE_VAL_COMPLEX(-48, 1006), MAKE_VAL_COMPLEX(569, -991), MAKE_VAL_COMPLEX(1289, 1101), MAKE_VAL_COMPLEX(-335, 413), MAKE_VAL_COMPLEX(38, 256), MAKE_VAL_COMPLEX(-252, 434), MAKE_VAL_COMPLEX(214, 328), MAKE_VAL_COMPLEX(-882, 1028), MAKE_VAL_COMPLEX(-279, -1079), MAKE_VAL_COMPLEX(-129, 325), MAKE_VAL_COMPLEX(-265, 431) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1583, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1583", y_true, y_val, x1536_LEN*x1536_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1583() */
/* END TEST 1583 */

#define A1584_M 5    /* # of rows */
#define A1584_N 5    /* # of cols */
#define A1584_NNZ 7  /* # of stored non-zeros */
static oski_index_t A1584_ptr[A1584_M+1] = {0, 1, 1, 2, 3, 7}; /* end row pointers for A1584 */
static oski_index_t A1584_ind[A1584_NNZ] = {0, 1, 3, 0, 2, 3, 4}; /* end column indices pointers for A1584 */
static oski_value_t A1584_val[A1584_NNZ] = {MAKE_VAL_COMPLEX(5, 0), MAKE_VAL_COMPLEX(0, 2), MAKE_VAL_COMPLEX(-3, 0), MAKE_VAL_COMPLEX(-10, -8), MAKE_VAL_COMPLEX(-7, 3), MAKE_VAL_COMPLEX(-4, 10), MAKE_VAL_COMPLEX(10, 0)}; /* end stored non-zero values for A1584 */

#define x1584_LEN 5   /* vector length */
#define x1584_VW  3   /* vector width */
static oski_value_t x1584_val[x1584_LEN*x1584_VW] = {MAKE_VAL_COMPLEX(-4, 2), MAKE_VAL_COMPLEX(9, 0), MAKE_VAL_COMPLEX(-4, 6), MAKE_VAL_COMPLEX(-4, -4), MAKE_VAL_COMPLEX(2, 9), MAKE_VAL_COMPLEX(-2, -6), MAKE_VAL_COMPLEX(6, 0), MAKE_VAL_COMPLEX(1, 2), MAKE_VAL_COMPLEX(10, -3), MAKE_VAL_COMPLEX(5, -10), MAKE_VAL_COMPLEX(-6, 10), MAKE_VAL_COMPLEX(3, 0), MAKE_VAL_COMPLEX(-10, -1), MAKE_VAL_COMPLEX(-1, 1), MAKE_VAL_COMPLEX(-6, 5)}; /* data */

#define y1584_LEN 5   /* vector length */
#define y1584_VW  3   /* vector width */
static oski_value_t y1584_val[y1584_LEN*y1584_VW] = {MAKE_VAL_COMPLEX(10, -5), MAKE_VAL_COMPLEX(-0, -3), MAKE_VAL_COMPLEX(-10, -8), MAKE_VAL_COMPLEX(2, -9), MAKE_VAL_COMPLEX(-6, 8), MAKE_VAL_COMPLEX(-6, -2), MAKE_VAL_COMPLEX(-4, -1), MAKE_VAL_COMPLEX(1, -6), MAKE_VAL_COMPLEX(-8, -10), MAKE_VAL_COMPLEX(3, 0), MAKE_VAL_COMPLEX(-9, -8), MAKE_VAL_COMPLEX(-1, -3), MAKE_VAL_COMPLEX(-5, 3), MAKE_VAL_COMPLEX(-8, 10), MAKE_VAL_COMPLEX(-7, 10)}; /* data */

/*
 *  BEGIN TEST 1584
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1584( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	oski_value_t* y_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(48, 10), MAKE_VAL_COMPLEX(59, 0), MAKE_VAL_COMPLEX(92, -60), MAKE_VAL_COMPLEX(24, 50), MAKE_VAL_COMPLEX(8, -74), MAKE_VAL_COMPLEX(38, -32), MAKE_VAL_COMPLEX(103, 19), MAKE_VAL_COMPLEX(10, 40), MAKE_VAL_COMPLEX(157, 7), MAKE_VAL_COMPLEX(-3, 146), MAKE_VAL_COMPLEX(118, -12), MAKE_VAL_COMPLEX(83, 54), MAKE_VAL_COMPLEX(12, 72), MAKE_VAL_COMPLEX(-181, -265), MAKE_VAL_COMPLEX(-43, 15) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1584, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1584", y_true, y_val, y1584_LEN*y1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1584() */
/* END TEST 1584 */

/*
 *  BEGIN TEST 1585
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1585( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	oski_value_t* y_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-63, 135), MAKE_VAL_COMPLEX(-26, 21), MAKE_VAL_COMPLEX(66, 6), MAKE_VAL_COMPLEX(61, 35), MAKE_VAL_COMPLEX(-54, -48), MAKE_VAL_COMPLEX(26, -20), MAKE_VAL_COMPLEX(-64, -56), MAKE_VAL_COMPLEX(49, 13), MAKE_VAL_COMPLEX(9, -25), MAKE_VAL_COMPLEX(-18, -113), MAKE_VAL_COMPLEX(33, -31), MAKE_VAL_COMPLEX(-43, -44), MAKE_VAL_COMPLEX(-10, -148), MAKE_VAL_COMPLEX(127, 107), MAKE_VAL_COMPLEX(-18, -162) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1585, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1585", y_true, y_val, y1584_LEN*y1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1585() */
/* END TEST 1585 */

/*
 *  BEGIN TEST 1586
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1586( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	oski_value_t* y_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(30, 10), MAKE_VAL_COMPLEX(6, -6), MAKE_VAL_COMPLEX(-4, -36), MAKE_VAL_COMPLEX(22, -14), MAKE_VAL_COMPLEX(-28, 4), MAKE_VAL_COMPLEX(-8, -16), MAKE_VAL_COMPLEX(-6, -10), MAKE_VAL_COMPLEX(14, -10), MAKE_VAL_COMPLEX(4, -36), MAKE_VAL_COMPLEX(6, 6), MAKE_VAL_COMPLEX(-2, -34), MAKE_VAL_COMPLEX(4, -8), MAKE_VAL_COMPLEX(-16, -4), MAKE_VAL_COMPLEX(-36, 4), MAKE_VAL_COMPLEX(-34, 6) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1586, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1586", y_true, y_val, y1584_LEN*y1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1586() */
/* END TEST 1586 */

/*
 *  BEGIN TEST 1587
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1587( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	oski_value_t* y_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-365, 225), MAKE_VAL_COMPLEX(-244, 63), MAKE_VAL_COMPLEX(-114, 298), MAKE_VAL_COMPLEX(-9, -27), MAKE_VAL_COMPLEX(-50, 100), MAKE_VAL_COMPLEX(-30, 100), MAKE_VAL_COMPLEX(-414, -142), MAKE_VAL_COMPLEX(31, -57), MAKE_VAL_COMPLEX(-447, 39), MAKE_VAL_COMPLEX(-48, -679), MAKE_VAL_COMPLEX(-265, 75), MAKE_VAL_COMPLEX(-343, -224), MAKE_VAL_COMPLEX(-6, -508), MAKE_VAL_COMPLEX(903, 979), MAKE_VAL_COMPLEX(192, -404) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1587, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1587", y_true, y_val, y1584_LEN*y1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1587() */
/* END TEST 1587 */

/*
 *  BEGIN TEST 1588
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1588( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(136, 9), MAKE_VAL_COMPLEX(160, -105), MAKE_VAL_COMPLEX(136, -60), MAKE_VAL_COMPLEX(-22, 16), MAKE_VAL_COMPLEX(66, -10), MAKE_VAL_COMPLEX(-16, -4), MAKE_VAL_COMPLEX(8, -76), MAKE_VAL_COMPLEX(54, -88), MAKE_VAL_COMPLEX(-3, -139), MAKE_VAL_COMPLEX(-79, -92), MAKE_VAL_COMPLEX(19, -60), MAKE_VAL_COMPLEX(-69, -119), MAKE_VAL_COMPLEX(-103, 209), MAKE_VAL_COMPLEX(-119, 297), MAKE_VAL_COMPLEX(124, 252) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1588, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1588", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1588() */
/* END TEST 1588 */

/*
 *  BEGIN TEST 1589
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1589( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-160, 13), MAKE_VAL_COMPLEX(-97, 87), MAKE_VAL_COMPLEX(-152, 110), MAKE_VAL_COMPLEX(-14, -36), MAKE_VAL_COMPLEX(-34, 69), MAKE_VAL_COMPLEX(-10, -34), MAKE_VAL_COMPLEX(34, 64), MAKE_VAL_COMPLEX(-43, 100), MAKE_VAL_COMPLEX(67, 98), MAKE_VAL_COMPLEX(94, 12), MAKE_VAL_COMPLEX(-41, 142), MAKE_VAL_COMPLEX(90, 113), MAKE_VAL_COMPLEX(31, -196), MAKE_VAL_COMPLEX(114, -288), MAKE_VAL_COMPLEX(-156, -205) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1589, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1589", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1589() */
/* END TEST 1589 */

/*
 *  BEGIN TEST 1590
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1590( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(28, 26), MAKE_VAL_COMPLEX(-27, -72), MAKE_VAL_COMPLEX(60, 14), MAKE_VAL_COMPLEX(-20, 44), MAKE_VAL_COMPLEX(66, -43), MAKE_VAL_COMPLEX(-42, 34), MAKE_VAL_COMPLEX(-18, -48), MAKE_VAL_COMPLEX(13, -14), MAKE_VAL_COMPLEX(-54, -71), MAKE_VAL_COMPLEX(-95, -10), MAKE_VAL_COMPLEX(98, 18), MAKE_VAL_COMPLEX(-9, -24), MAKE_VAL_COMPLEX(22, 83), MAKE_VAL_COMPLEX(11, 5), MAKE_VAL_COMPLEX(58, 33) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1590, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1590", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1590() */
/* END TEST 1590 */

/*
 *  BEGIN TEST 1591
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1591( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-890, 878), MAKE_VAL_COMPLEX(-983, 1332), MAKE_VAL_COMPLEX(-268, 1282), MAKE_VAL_COMPLEX(20, 96), MAKE_VAL_COMPLEX(-74, 39), MAKE_VAL_COMPLEX(-86, 262), MAKE_VAL_COMPLEX(170, 344), MAKE_VAL_COMPLEX(163, 902), MAKE_VAL_COMPLEX(332, 685), MAKE_VAL_COMPLEX(479, 372), MAKE_VAL_COMPLEX(950, 536), MAKE_VAL_COMPLEX(1155, 382), MAKE_VAL_COMPLEX(-112, -1733), MAKE_VAL_COMPLEX(-741, -3075), MAKE_VAL_COMPLEX(-2022, -1145) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1591, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1591", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1591() */
/* END TEST 1591 */

/*
 *  BEGIN TEST 1592
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1592( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(42, -73), MAKE_VAL_COMPLEX(81, -188), MAKE_VAL_COMPLEX(-90, -138), MAKE_VAL_COMPLEX(-42, -24), MAKE_VAL_COMPLEX(15, 77), MAKE_VAL_COMPLEX(-44, -36), MAKE_VAL_COMPLEX(116, -8), MAKE_VAL_COMPLEX(81, -41), MAKE_VAL_COMPLEX(170, -98), MAKE_VAL_COMPLEX(76, -57), MAKE_VAL_COMPLEX(115, 160), MAKE_VAL_COMPLEX(158, 36), MAKE_VAL_COMPLEX(-262, 26), MAKE_VAL_COMPLEX(15, 127), MAKE_VAL_COMPLEX(37, 359) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1592, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1592", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1592() */
/* END TEST 1592 */

/*
 *  BEGIN TEST 1593
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1593( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-98, 101), MAKE_VAL_COMPLEX(45, 188), MAKE_VAL_COMPLEX(34, 222), MAKE_VAL_COMPLEX(-14, -32), MAKE_VAL_COMPLEX(13, 49), MAKE_VAL_COMPLEX(16, -48), MAKE_VAL_COMPLEX(-32, 8), MAKE_VAL_COMPLEX(-67, 69), MAKE_VAL_COMPLEX(-30, 56), MAKE_VAL_COMPLEX(-6, -83), MAKE_VAL_COMPLEX(-199, -20), MAKE_VAL_COMPLEX(-116, -36), MAKE_VAL_COMPLEX(122, -40), MAKE_VAL_COMPLEX(-29, -113), MAKE_VAL_COMPLEX(-121, -289) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1593, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1593", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1593() */
/* END TEST 1593 */

/*
 *  BEGIN TEST 1594
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1594( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-18, -36), MAKE_VAL_COMPLEX(0, 81), MAKE_VAL_COMPLEX(-54, -36), MAKE_VAL_COMPLEX(36, -36), MAKE_VAL_COMPLEX(-81, 18), MAKE_VAL_COMPLEX(54, -18), MAKE_VAL_COMPLEX(0, 54), MAKE_VAL_COMPLEX(-18, 9), MAKE_VAL_COMPLEX(27, 90), MAKE_VAL_COMPLEX(90, 45), MAKE_VAL_COMPLEX(-90, -54), MAKE_VAL_COMPLEX(0, 27), MAKE_VAL_COMPLEX(9, -90), MAKE_VAL_COMPLEX(-9, -9), MAKE_VAL_COMPLEX(-45, -54) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1594, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1594", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1594() */
/* END TEST 1594 */

/*
 *  BEGIN TEST 1595
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1595( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-871, -213), MAKE_VAL_COMPLEX(-1262, 456), MAKE_VAL_COMPLEX(-1134, 1038), MAKE_VAL_COMPLEX(30, 30), MAKE_VAL_COMPLEX(101, 63), MAKE_VAL_COMPLEX(120, 116), MAKE_VAL_COMPLEX(-206, -482), MAKE_VAL_COMPLEX(-599, -327), MAKE_VAL_COMPLEX(-713, -485), MAKE_VAL_COMPLEX(48, -436), MAKE_VAL_COMPLEX(67, -1261), MAKE_VAL_COMPLEX(-123, -1061), MAKE_VAL_COMPLEX(689, 1213), MAKE_VAL_COMPLEX(760, -504), MAKE_VAL_COMPLEX(1949, -1477) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-3, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1595, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1595", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1595() */
/* END TEST 1595 */

/*
 *  BEGIN TEST 1596
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1596( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	oski_value_t* y_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(223, -15), MAKE_VAL_COMPLEX(27, -39), MAKE_VAL_COMPLEX(57, -133), MAKE_VAL_COMPLEX(114, 71), MAKE_VAL_COMPLEX(-132, 128), MAKE_VAL_COMPLEX(11, -90), MAKE_VAL_COMPLEX(-23, -47), MAKE_VAL_COMPLEX(55, -45), MAKE_VAL_COMPLEX(50, -186), MAKE_VAL_COMPLEX(-162, -146), MAKE_VAL_COMPLEX(-9, -221), MAKE_VAL_COMPLEX(12, -56), MAKE_VAL_COMPLEX(-3, -39), MAKE_VAL_COMPLEX(-97, 58), MAKE_VAL_COMPLEX(-198, 130) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1596, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1596", y_true, y_val, y1584_LEN*y1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1596() */
/* END TEST 1596 */

/*
 *  BEGIN TEST 1597
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1597( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	oski_value_t* y_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-163, 85), MAKE_VAL_COMPLEX(-3, 33), MAKE_VAL_COMPLEX(-13, 37), MAKE_VAL_COMPLEX(-38, -73), MAKE_VAL_COMPLEX(56, -160), MAKE_VAL_COMPLEX(-7, 38), MAKE_VAL_COMPLEX(23, 13), MAKE_VAL_COMPLEX(-5, 41), MAKE_VAL_COMPLEX(14, 102), MAKE_VAL_COMPLEX(168, 170), MAKE_VAL_COMPLEX(55, 133), MAKE_VAL_COMPLEX(10, 42), MAKE_VAL_COMPLEX(-31, 5), MAKE_VAL_COMPLEX(1, -102), MAKE_VAL_COMPLEX(104, -166) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1597, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1597", y_true, y_val, y1584_LEN*y1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1597() */
/* END TEST 1597 */

/*
 *  BEGIN TEST 1598
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1598( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	oski_value_t* y_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(115, 55), MAKE_VAL_COMPLEX(27, -21), MAKE_VAL_COMPLEX(2, -146), MAKE_VAL_COMPLEX(95, -45), MAKE_VAL_COMPLEX(-114, 2), MAKE_VAL_COMPLEX(-24, -68), MAKE_VAL_COMPLEX(-19, -43), MAKE_VAL_COMPLEX(61, -33), MAKE_VAL_COMPLEX(34, -142), MAKE_VAL_COMPLEX(21, 27), MAKE_VAL_COMPLEX(9, -137), MAKE_VAL_COMPLEX(20, -30), MAKE_VAL_COMPLEX(-62, -24), MAKE_VAL_COMPLEX(-146, -2), MAKE_VAL_COMPLEX(-139, 7) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1598, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1598", y_true, y_val, y1584_LEN*y1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1598() */
/* END TEST 1598 */

/*
 *  BEGIN TEST 1599
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1599( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	oski_value_t* y_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(690, -71), MAKE_VAL_COMPLEX(36, -81), MAKE_VAL_COMPLEX(218, 314), MAKE_VAL_COMPLEX(-313, 652), MAKE_VAL_COMPLEX(-402, 588), MAKE_VAL_COMPLEX(247, 37), MAKE_VAL_COMPLEX(-2, -5), MAKE_VAL_COMPLEX(-33, -66), MAKE_VAL_COMPLEX(176, -94), MAKE_VAL_COMPLEX(-405, -1432), MAKE_VAL_COMPLEX(141, -396), MAKE_VAL_COMPLEX(23, -139), MAKE_VAL_COMPLEX(373, 123), MAKE_VAL_COMPLEX(149, 465), MAKE_VAL_COMPLEX(-583, 450) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1599, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1599", y_true, y_val, y1584_LEN*y1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1599() */
/* END TEST 1599 */

/*
 *  BEGIN TEST 1600
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1600( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(140, -45), MAKE_VAL_COMPLEX(12, -25), MAKE_VAL_COMPLEX(67, -40), MAKE_VAL_COMPLEX(-12, -56), MAKE_VAL_COMPLEX(19, 76), MAKE_VAL_COMPLEX(-16, -60), MAKE_VAL_COMPLEX(60, -68), MAKE_VAL_COMPLEX(-17, 34), MAKE_VAL_COMPLEX(96, -103), MAKE_VAL_COMPLEX(-12, -140), MAKE_VAL_COMPLEX(-46, 134), MAKE_VAL_COMPLEX(-24, -137), MAKE_VAL_COMPLEX(-251, 294), MAKE_VAL_COMPLEX(4, 138), MAKE_VAL_COMPLEX(207, 176) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1600, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1600", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1600() */
/* END TEST 1600 */

/*
 *  BEGIN TEST 1601
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1601( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-144, 57), MAKE_VAL_COMPLEX(6, 7), MAKE_VAL_COMPLEX(-63, 60), MAKE_VAL_COMPLEX(-4, 56), MAKE_VAL_COMPLEX(3, -62), MAKE_VAL_COMPLEX(0, 52), MAKE_VAL_COMPLEX(-48, 56), MAKE_VAL_COMPLEX(23, -32), MAKE_VAL_COMPLEX(-82, 77), MAKE_VAL_COMPLEX(2, 110), MAKE_VAL_COMPLEX(54, -102), MAKE_VAL_COMPLEX(30, 131), MAKE_VAL_COMPLEX(229, -276), MAKE_VAL_COMPLEX(-4, -134), MAKE_VAL_COMPLEX(-209, -154) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1601, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1601", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1601() */
/* END TEST 1601 */

/*
 *  BEGIN TEST 1602
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1602( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-2, -34), MAKE_VAL_COMPLEX(-27, 63), MAKE_VAL_COMPLEX(-30, -46), MAKE_VAL_COMPLEX(40, -16), MAKE_VAL_COMPLEX(-69, -13), MAKE_VAL_COMPLEX(48, 4), MAKE_VAL_COMPLEX(-18, 42), MAKE_VAL_COMPLEX(-17, 1), MAKE_VAL_COMPLEX(-9, 79), MAKE_VAL_COMPLEX(55, 65), MAKE_VAL_COMPLEX(-52, -72), MAKE_VAL_COMPLEX(-9, 21), MAKE_VAL_COMPLEX(37, -67), MAKE_VAL_COMPLEX(-4, -10), MAKE_VAL_COMPLEX(-17, -57) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1602, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1602", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1602() */
/* END TEST 1602 */

/*
 *  BEGIN TEST 1603
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1603( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-707, 851), MAKE_VAL_COMPLEX(456, -82), MAKE_VAL_COMPLEX(-249, 803), MAKE_VAL_COMPLEX(-76, 156), MAKE_VAL_COMPLEX(18, 64), MAKE_VAL_COMPLEX(0, 12), MAKE_VAL_COMPLEX(198, 358), MAKE_VAL_COMPLEX(86, -150), MAKE_VAL_COMPLEX(238, 386), MAKE_VAL_COMPLEX(455, 39), MAKE_VAL_COMPLEX(-196, -156), MAKE_VAL_COMPLEX(678, 532), MAKE_VAL_COMPLEX(-144, -2002), MAKE_VAL_COMPLEX(-456, -602), MAKE_VAL_COMPLEX(-1722, 148) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1603, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1603", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1603() */
/* END TEST 1603 */

/*
 *  BEGIN TEST 1604
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1604( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(36, -163), MAKE_VAL_COMPLEX(-51, 41), MAKE_VAL_COMPLEX(-77, -202), MAKE_VAL_COMPLEX(-8, 4), MAKE_VAL_COMPLEX(-37, 13), MAKE_VAL_COMPLEX(22, -2), MAKE_VAL_COMPLEX(78, -20), MAKE_VAL_COMPLEX(-24, -12), MAKE_VAL_COMPLEX(104, -24), MAKE_VAL_COMPLEX(133, 60), MAKE_VAL_COMPLEX(-24, -8), MAKE_VAL_COMPLEX(155, 9), MAKE_VAL_COMPLEX(-31, 101), MAKE_VAL_COMPLEX(213, 71), MAKE_VAL_COMPLEX(-107, 83) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1604, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1604", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1604() */
/* END TEST 1604 */

/*
 *  BEGIN TEST 1605
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1605( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-30, 105), MAKE_VAL_COMPLEX(-12, 58), MAKE_VAL_COMPLEX(39, 116), MAKE_VAL_COMPLEX(80, -20), MAKE_VAL_COMPLEX(-76, -54), MAKE_VAL_COMPLEX(58, 22), MAKE_VAL_COMPLEX(-120, 86), MAKE_VAL_COMPLEX(-5, 9), MAKE_VAL_COMPLEX(-141, 155), MAKE_VAL_COMPLEX(-58, 65), MAKE_VAL_COMPLEX(-44, -128), MAKE_VAL_COMPLEX(-176, 24), MAKE_VAL_COMPLEX(112, -204), MAKE_VAL_COMPLEX(-217, -89), MAKE_VAL_COMPLEX(94, -184) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1605, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1605", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1605() */
/* END TEST 1605 */

/*
 *  BEGIN TEST 1606
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1606( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(14, 38), MAKE_VAL_COMPLEX(9, -81), MAKE_VAL_COMPLEX(50, 42), MAKE_VAL_COMPLEX(-40, 32), MAKE_VAL_COMPLEX(83, -9), MAKE_VAL_COMPLEX(-56, 12), MAKE_VAL_COMPLEX(6, -54), MAKE_VAL_COMPLEX(19, -7), MAKE_VAL_COMPLEX(-17, -93), MAKE_VAL_COMPLEX(-85, -55), MAKE_VAL_COMPLEX(84, 64), MAKE_VAL_COMPLEX(3, -27), MAKE_VAL_COMPLEX(-19, 89), MAKE_VAL_COMPLEX(8, 10), MAKE_VAL_COMPLEX(39, 59) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1606, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1606", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1606() */
/* END TEST 1606 */

/*
 *  BEGIN TEST 1607
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_ROWMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1607( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_ROWMAJ, x1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(150, -600), MAKE_VAL_COMPLEX(-168, 65), MAKE_VAL_COMPLEX(-258, -740), MAKE_VAL_COMPLEX(-92, 44), MAKE_VAL_COMPLEX(-41, 66), MAKE_VAL_COMPLEX(14, -10), MAKE_VAL_COMPLEX(336, -146), MAKE_VAL_COMPLEX(-70, -51), MAKE_VAL_COMPLEX(423, -218), MAKE_VAL_COMPLEX(442, 145), MAKE_VAL_COMPLEX(-10, 74), MAKE_VAL_COMPLEX(632, 3), MAKE_VAL_COMPLEX(-175, 510), MAKE_VAL_COMPLEX(859, 299), MAKE_VAL_COMPLEX(-397, 418) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1607, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1607", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1607() */
/* END TEST 1607 */

/*
 *  BEGIN TEST 1608
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1608( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	oski_value_t* y_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-122, -109), MAKE_VAL_COMPLEX(8, 107), MAKE_VAL_COMPLEX(-52, -16), MAKE_VAL_COMPLEX(-22, -9), MAKE_VAL_COMPLEX(30, 30), MAKE_VAL_COMPLEX(-16, 42), MAKE_VAL_COMPLEX(5, -36), MAKE_VAL_COMPLEX(-88, 57), MAKE_VAL_COMPLEX(9, 11), MAKE_VAL_COMPLEX(97, -56), MAKE_VAL_COMPLEX(-191, 27), MAKE_VAL_COMPLEX(64, 38), MAKE_VAL_COMPLEX(149, 47), MAKE_VAL_COMPLEX(31, 119), MAKE_VAL_COMPLEX(180, -1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1608, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1608", y_true, y_val, y1584_LEN*y1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1608() */
/* END TEST 1608 */

/*
 *  BEGIN TEST 1609
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1609( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	oski_value_t* y_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(207, 54), MAKE_VAL_COMPLEX(-11, -134), MAKE_VAL_COMPLEX(-46, -46), MAKE_VAL_COMPLEX(31, -74), MAKE_VAL_COMPLEX(-76, 48), MAKE_VAL_COMPLEX(-40, -54), MAKE_VAL_COMPLEX(-42, 31), MAKE_VAL_COMPLEX(91, -112), MAKE_VAL_COMPLEX(-91, -93), MAKE_VAL_COMPLEX(-70, 53), MAKE_VAL_COMPLEX(102, -90), MAKE_VAL_COMPLEX(-76, -64), MAKE_VAL_COMPLEX(-191, -15), MAKE_VAL_COMPLEX(-93, -21), MAKE_VAL_COMPLEX(-233, 98) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1609, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1609", y_true, y_val, y1584_LEN*y1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1609() */
/* END TEST 1609 */

/*
 *  BEGIN TEST 1610
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1610( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	oski_value_t* y_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(65, 55), MAKE_VAL_COMPLEX(21, -9), MAKE_VAL_COMPLEX(26, -94), MAKE_VAL_COMPLEX(69, -13), MAKE_VAL_COMPLEX(-74, -18), MAKE_VAL_COMPLEX(-4, -48), MAKE_VAL_COMPLEX(-5, -31), MAKE_VAL_COMPLEX(45, -11), MAKE_VAL_COMPLEX(46, -86), MAKE_VAL_COMPLEX(9, 21), MAKE_VAL_COMPLEX(29, -87), MAKE_VAL_COMPLEX(18, -16), MAKE_VAL_COMPLEX(-36, -26), MAKE_VAL_COMPLEX(-94, -26), MAKE_VAL_COMPLEX(-91, -19) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1610, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1610", y_true, y_val, y1584_LEN*y1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1610() */
/* END TEST 1610 */

/*
 *  BEGIN TEST 1611
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1611( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	oski_value_t* y_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-656, 703), MAKE_VAL_COMPLEX(914, 33), MAKE_VAL_COMPLEX(-510, -68), MAKE_VAL_COMPLEX(88, -165), MAKE_VAL_COMPLEX(-46, 48), MAKE_VAL_COMPLEX(98, 50), MAKE_VAL_COMPLEX(-420, -207), MAKE_VAL_COMPLEX(403, 598), MAKE_VAL_COMPLEX(-66, -552), MAKE_VAL_COMPLEX(-137, -846), MAKE_VAL_COMPLEX(-405, 1144), MAKE_VAL_COMPLEX(435, -567), MAKE_VAL_COMPLEX(437, -1051), MAKE_VAL_COMPLEX(562, 268), MAKE_VAL_COMPLEX(-61, -1150) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1611, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1611", y_true, y_val, y1584_LEN*y1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1611() */
/* END TEST 1611 */

/*
 *  BEGIN TEST 1612
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1612( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(152, -9), MAKE_VAL_COMPLEX(-43, -44), MAKE_VAL_COMPLEX(52, -54), MAKE_VAL_COMPLEX(-15, -26), MAKE_VAL_COMPLEX(-71, 96), MAKE_VAL_COMPLEX(146, -13), MAKE_VAL_COMPLEX(-18, -22), MAKE_VAL_COMPLEX(45, -96), MAKE_VAL_COMPLEX(-103, -121), MAKE_VAL_COMPLEX(-190, 321), MAKE_VAL_COMPLEX(170, -110), MAKE_VAL_COMPLEX(5, -28), MAKE_VAL_COMPLEX(61, -34), MAKE_VAL_COMPLEX(-60, -102), MAKE_VAL_COMPLEX(144, 215) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1612, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1612", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1612() */
/* END TEST 1612 */

/*
 *  BEGIN TEST 1613
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1613( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-150, 33), MAKE_VAL_COMPLEX(61, -1), MAKE_VAL_COMPLEX(-30, 86), MAKE_VAL_COMPLEX(-13, 38), MAKE_VAL_COMPLEX(120, -88), MAKE_VAL_COMPLEX(-180, 11), MAKE_VAL_COMPLEX(30, -8), MAKE_VAL_COMPLEX(-33, 95), MAKE_VAL_COMPLEX(108, 65), MAKE_VAL_COMPLEX(150, -366), MAKE_VAL_COMPLEX(-132, 160), MAKE_VAL_COMPLEX(1, 13), MAKE_VAL_COMPLEX(-86, 82), MAKE_VAL_COMPLEX(63, 109), MAKE_VAL_COMPLEX(-131, -175) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1613, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1613", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1613() */
/* END TEST 1613 */

/*
 *  BEGIN TEST 1614
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1614( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(8, 36), MAKE_VAL_COMPLEX(18, -72), MAKE_VAL_COMPLEX(40, 44), MAKE_VAL_COMPLEX(-40, 24), MAKE_VAL_COMPLEX(76, 2), MAKE_VAL_COMPLEX(-52, 4), MAKE_VAL_COMPLEX(12, -48), MAKE_VAL_COMPLEX(18, -4), MAKE_VAL_COMPLEX(-4, -86), MAKE_VAL_COMPLEX(-70, -60), MAKE_VAL_COMPLEX(68, 68), MAKE_VAL_COMPLEX(6, -24), MAKE_VAL_COMPLEX(-28, 78), MAKE_VAL_COMPLEX(6, 10), MAKE_VAL_COMPLEX(28, 58) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1614, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1614", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1614() */
/* END TEST 1614 */

/*
 *  BEGIN TEST 1615
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1615( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-765, -1011), MAKE_VAL_COMPLEX(-37, -23), MAKE_VAL_COMPLEX(-390, 178), MAKE_VAL_COMPLEX(-503, 481), MAKE_VAL_COMPLEX(1881, 163), MAKE_VAL_COMPLEX(-1299, -1205), MAKE_VAL_COMPLEX(-12, -148), MAKE_VAL_COMPLEX(-933, 37), MAKE_VAL_COMPLEX(-624, 784), MAKE_VAL_COMPLEX(3219, -405), MAKE_VAL_COMPLEX(-1236, -400), MAKE_VAL_COMPLEX(-229, -115), MAKE_VAL_COMPLEX(-841, 305), MAKE_VAL_COMPLEX(-564, 1136), MAKE_VAL_COMPLEX(1469, -1871) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-5, -9);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1615, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1615", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1615() */
/* END TEST 1615 */

/*
 *  BEGIN TEST 1616
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1616( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(36, -95), MAKE_VAL_COMPLEX(70, 44), MAKE_VAL_COMPLEX(6, 30), MAKE_VAL_COMPLEX(25, -10), MAKE_VAL_COMPLEX(-191, 105), MAKE_VAL_COMPLEX(8, -235), MAKE_VAL_COMPLEX(36, 22), MAKE_VAL_COMPLEX(70, -38), MAKE_VAL_COMPLEX(251, 80), MAKE_VAL_COMPLEX(103, 57), MAKE_VAL_COMPLEX(-148, -140), MAKE_VAL_COMPLEX(4, 28), MAKE_VAL_COMPLEX(7, -109), MAKE_VAL_COMPLEX(119, 43), MAKE_VAL_COMPLEX(18, 324) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1616, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1616", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1616() */
/* END TEST 1616 */

/*
 *  BEGIN TEST 1617
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1617( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-58, 111), MAKE_VAL_COMPLEX(-16, -53), MAKE_VAL_COMPLEX(-24, 10), MAKE_VAL_COMPLEX(-53, -10), MAKE_VAL_COMPLEX(212, -53), MAKE_VAL_COMPLEX(-26, 201), MAKE_VAL_COMPLEX(0, -28), MAKE_VAL_COMPLEX(-62, 49), MAKE_VAL_COMPLEX(-194, -108), MAKE_VAL_COMPLEX(-83, -122), MAKE_VAL_COMPLEX(122, 206), MAKE_VAL_COMPLEX(14, -31), MAKE_VAL_COMPLEX(-68, 113), MAKE_VAL_COMPLEX(-124, -36), MAKE_VAL_COMPLEX(-49, -288) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1617, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1617", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1617() */
/* END TEST 1617 */

/*
 *  BEGIN TEST 1618
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1618( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-42, 6), MAKE_VAL_COMPLEX(81, 27), MAKE_VAL_COMPLEX(-54, 42), MAKE_VAL_COMPLEX(-24, -48), MAKE_VAL_COMPLEX(-9, 87), MAKE_VAL_COMPLEX(0, -60), MAKE_VAL_COMPLEX(54, 18), MAKE_VAL_COMPLEX(3, 21), MAKE_VAL_COMPLEX(99, 3), MAKE_VAL_COMPLEX(75, -75), MAKE_VAL_COMPLEX(-84, 72), MAKE_VAL_COMPLEX(27, 9), MAKE_VAL_COMPLEX(-87, -39), MAKE_VAL_COMPLEX(-12, 6), MAKE_VAL_COMPLEX(-69, 27) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1618, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1618", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1618() */
/* END TEST 1618 */

/*
 *  BEGIN TEST 1619
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1619( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_ROWMAJ, y1584_VW );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(43, -244), MAKE_VAL_COMPLEX(49, -9), MAKE_VAL_COMPLEX(120, -24), MAKE_VAL_COMPLEX(88, 27), MAKE_VAL_COMPLEX(-280, 260), MAKE_VAL_COMPLEX(-207, -382), MAKE_VAL_COMPLEX(4, -10), MAKE_VAL_COMPLEX(93, -179), MAKE_VAL_COMPLEX(423, -76), MAKE_VAL_COMPLEX(158, 146), MAKE_VAL_COMPLEX(-316, -264), MAKE_VAL_COMPLEX(-9, 43), MAKE_VAL_COMPLEX(52, -180), MAKE_VAL_COMPLEX(299, -45), MAKE_VAL_COMPLEX(465, 573) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1619, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1619", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1619() */
/* END TEST 1619 */

/*
 *  BEGIN TEST 1620
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1620( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	oski_value_t* y_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-67, -99), MAKE_VAL_COMPLEX(9, -7), MAKE_VAL_COMPLEX(-45, -81), MAKE_VAL_COMPLEX(95, -91), MAKE_VAL_COMPLEX(120, 70), MAKE_VAL_COMPLEX(-12, 106), MAKE_VAL_COMPLEX(-17, -3), MAKE_VAL_COMPLEX(-66, 36), MAKE_VAL_COMPLEX(-200, -43), MAKE_VAL_COMPLEX(14, 74), MAKE_VAL_COMPLEX(-63, -79), MAKE_VAL_COMPLEX(-10, 6), MAKE_VAL_COMPLEX(35, 9), MAKE_VAL_COMPLEX(47, 95), MAKE_VAL_COMPLEX(122, 18) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1620, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1620", y_true, y_val, y1584_LEN*y1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1620() */
/* END TEST 1620 */

/*
 *  BEGIN TEST 1621
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1621( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	oski_value_t* y_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(17, 49), MAKE_VAL_COMPLEX(-27, 13), MAKE_VAL_COMPLEX(17, 157), MAKE_VAL_COMPLEX(-153, 97), MAKE_VAL_COMPLEX(-60, -50), MAKE_VAL_COMPLEX(12, -66), MAKE_VAL_COMPLEX(19, 29), MAKE_VAL_COMPLEX(28, -30), MAKE_VAL_COMPLEX(156, 111), MAKE_VAL_COMPLEX(-20, -92), MAKE_VAL_COMPLEX(33, 149), MAKE_VAL_COMPLEX(-6, 6), MAKE_VAL_COMPLEX(-7, 15), MAKE_VAL_COMPLEX(29, -67), MAKE_VAL_COMPLEX(-48, 4) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1621, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1621", y_true, y_val, y1584_LEN*y1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1621() */
/* END TEST 1621 */

/*
 *  BEGIN TEST 1622
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1622( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	oski_value_t* y_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-30, 90), MAKE_VAL_COMPLEX(18, 18), MAKE_VAL_COMPLEX(108, -12), MAKE_VAL_COMPLEX(42, 66), MAKE_VAL_COMPLEX(-12, -84), MAKE_VAL_COMPLEX(48, -24), MAKE_VAL_COMPLEX(30, -18), MAKE_VAL_COMPLEX(30, 42), MAKE_VAL_COMPLEX(108, 12), MAKE_VAL_COMPLEX(-18, 18), MAKE_VAL_COMPLEX(102, -6), MAKE_VAL_COMPLEX(24, 12), MAKE_VAL_COMPLEX(12, -48), MAKE_VAL_COMPLEX(-12, -108), MAKE_VAL_COMPLEX(-18, -102) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1622, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1622", y_true, y_val, y1584_LEN*y1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1622() */
/* END TEST 1622 */

/*
 *  BEGIN TEST 1623
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1623( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	oski_value_t* y_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(889, 23), MAKE_VAL_COMPLEX(-99, -21), MAKE_VAL_COMPLEX(37, 331), MAKE_VAL_COMPLEX(-435, 517), MAKE_VAL_COMPLEX(-934, 312), MAKE_VAL_COMPLEX(-470, -600), MAKE_VAL_COMPLEX(-31, 25), MAKE_VAL_COMPLEX(186, -616), MAKE_VAL_COMPLEX(869, -466), MAKE_VAL_COMPLEX(-216, -468), MAKE_VAL_COMPLEX(169, 243), MAKE_VAL_COMPLEX(-54, -134), MAKE_VAL_COMPLEX(-321, 251), MAKE_VAL_COMPLEX(-587, 47), MAKE_VAL_COMPLEX(-776, 712) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 3);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1623, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1623", y_true, y_val, y1584_LEN*y1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1623() */
/* END TEST 1623 */

/*
 *  BEGIN TEST 1624
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1624( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(158, -19), MAKE_VAL_COMPLEX(79, -74), MAKE_VAL_COMPLEX(20, -8), MAKE_VAL_COMPLEX(-114, -69), MAKE_VAL_COMPLEX(-104, 363), MAKE_VAL_COMPLEX(-110, -64), MAKE_VAL_COMPLEX(54, -34), MAKE_VAL_COMPLEX(-4, 25), MAKE_VAL_COMPLEX(64, -21), MAKE_VAL_COMPLEX(-12, 31), MAKE_VAL_COMPLEX(123, 22), MAKE_VAL_COMPLEX(15, -28), MAKE_VAL_COMPLEX(-63, -36), MAKE_VAL_COMPLEX(-49, -127), MAKE_VAL_COMPLEX(248, 213) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1624, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1624", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1624() */
/* END TEST 1624 */

/*
 *  BEGIN TEST 1625
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1625( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-130, 75), MAKE_VAL_COMPLEX(-79, -52), MAKE_VAL_COMPLEX(64, 64), MAKE_VAL_COMPLEX(58, 125), MAKE_VAL_COMPLEX(230, -391), MAKE_VAL_COMPLEX(26, 92), MAKE_VAL_COMPLEX(-54, -50), MAKE_VAL_COMPLEX(32, -39), MAKE_VAL_COMPLEX(-106, -119), MAKE_VAL_COMPLEX(-128, -101), MAKE_VAL_COMPLEX(17, 62), MAKE_VAL_COMPLEX(-15, -14), MAKE_VAL_COMPLEX(49, 176), MAKE_VAL_COMPLEX(63, 141), MAKE_VAL_COMPLEX(-178, -129) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1625, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1625", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1625() */
/* END TEST 1625 */

/*
 *  BEGIN TEST 1626
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1626( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(6, 32), MAKE_VAL_COMPLEX(18, -63), MAKE_VAL_COMPLEX(34, 40), MAKE_VAL_COMPLEX(-36, 20), MAKE_VAL_COMPLEX(67, 4), MAKE_VAL_COMPLEX(-46, 2), MAKE_VAL_COMPLEX(12, -42), MAKE_VAL_COMPLEX(16, -3), MAKE_VAL_COMPLEX(-1, -76), MAKE_VAL_COMPLEX(-60, -55), MAKE_VAL_COMPLEX(58, 62), MAKE_VAL_COMPLEX(6, -21), MAKE_VAL_COMPLEX(-27, 68), MAKE_VAL_COMPLEX(5, 9), MAKE_VAL_COMPLEX(23, 52) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1626, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1626", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1626() */
/* END TEST 1626 */

/*
 *  BEGIN TEST 1627
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1627( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(990, 1215), MAKE_VAL_COMPLEX(181, 68), MAKE_VAL_COMPLEX(640, -144), MAKE_VAL_COMPLEX(346, -679), MAKE_VAL_COMPLEX(-3018, -467), MAKE_VAL_COMPLEX(98, -564), MAKE_VAL_COMPLEX(2, 102), MAKE_VAL_COMPLEX(-208, -139), MAKE_VAL_COMPLEX(-474, 285), MAKE_VAL_COMPLEX(-1024, 527), MAKE_VAL_COMPLEX(1005, 558), MAKE_VAL_COMPLEX(53, -78), MAKE_VAL_COMPLEX(781, -160), MAKE_VAL_COMPLEX(979, -807), MAKE_VAL_COMPLEX(-338, 2491) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1627, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1627", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1627() */
/* END TEST 1627 */

/*
 *  BEGIN TEST 1628
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1628( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(8, -149), MAKE_VAL_COMPLEX(56, 47), MAKE_VAL_COMPLEX(22, -2), MAKE_VAL_COMPLEX(78, 11), MAKE_VAL_COMPLEX(-35, 210), MAKE_VAL_COMPLEX(-58, -40), MAKE_VAL_COMPLEX(36, 16), MAKE_VAL_COMPLEX(-17, 2), MAKE_VAL_COMPLEX(101, 6), MAKE_VAL_COMPLEX(287, 8), MAKE_VAL_COMPLEX(-133, -134), MAKE_VAL_COMPLEX(30, 19), MAKE_VAL_COMPLEX(8, -89), MAKE_VAL_COMPLEX(141, 5), MAKE_VAL_COMPLEX(-149, 118) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1628, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1628", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1628() */
/* END TEST 1628 */

/*
 *  BEGIN TEST 1629
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1629( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-10, 165), MAKE_VAL_COMPLEX(-38, -74), MAKE_VAL_COMPLEX(-12, 26), MAKE_VAL_COMPLEX(-98, -7), MAKE_VAL_COMPLEX(66, -198), MAKE_VAL_COMPLEX(36, 34), MAKE_VAL_COMPLEX(-24, -34), MAKE_VAL_COMPLEX(25, -1), MAKE_VAL_COMPLEX(-90, -42), MAKE_VAL_COMPLEX(-307, -43), MAKE_VAL_COMPLEX(151, 172), MAKE_VAL_COMPLEX(-24, -28), MAKE_VAL_COMPLEX(-31, 117), MAKE_VAL_COMPLEX(-140, 0), MAKE_VAL_COMPLEX(152, -90) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1629, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1629", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1629() */
/* END TEST 1629 */

/*
 *  BEGIN TEST 1630
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1630( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-14, -38), MAKE_VAL_COMPLEX(-9, 81), MAKE_VAL_COMPLEX(-50, -42), MAKE_VAL_COMPLEX(40, -32), MAKE_VAL_COMPLEX(-83, 9), MAKE_VAL_COMPLEX(56, -12), MAKE_VAL_COMPLEX(-6, 54), MAKE_VAL_COMPLEX(-19, 7), MAKE_VAL_COMPLEX(17, 93), MAKE_VAL_COMPLEX(85, 55), MAKE_VAL_COMPLEX(-84, -64), MAKE_VAL_COMPLEX(-3, 27), MAKE_VAL_COMPLEX(19, -89), MAKE_VAL_COMPLEX(-8, -10), MAKE_VAL_COMPLEX(-39, -59) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1630, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1630", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1630() */
/* END TEST 1630 */

/*
 *  BEGIN TEST 1631
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1584', LAYOUT_COLMAJ>
 *  y : <'y1584', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1631( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1584_ptr, A1584_ind, A1584_val,
		A1584_M, A1584_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_DIAG_EXPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1584_val,
		y1584_LEN * y1584_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1584_LEN, y1584_VW, LAYOUT_COLMAJ, y1584_LEN );
	oski_value_t* y_val = array_Duplicate( x1584_val,
		x1584_LEN * x1584_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1584_LEN, x1584_VW, LAYOUT_COLMAJ, x1584_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1134, -698), MAKE_VAL_COMPLEX(110, 213), MAKE_VAL_COMPLEX(-182, -684), MAKE_VAL_COMPLEX(664, -686), MAKE_VAL_COMPLEX(959, 448), MAKE_VAL_COMPLEX(26, 510), MAKE_VAL_COMPLEX(-52, 122), MAKE_VAL_COMPLEX(-186, 119), MAKE_VAL_COMPLEX(19, -40), MAKE_VAL_COMPLEX(1058, -1545), MAKE_VAL_COMPLEX(-1716, -2), MAKE_VAL_COMPLEX(86, -13), MAKE_VAL_COMPLEX(-213, -830), MAKE_VAL_COMPLEX(301, -1223), MAKE_VAL_COMPLEX(583, 840) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(2, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1631, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1631", y_true, y_val, x1584_LEN*x1584_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1631() */
/* END TEST 1631 */

#define A1632_M 5    /* # of rows */
#define A1632_N 5    /* # of cols */
#define A1632_NNZ 7  /* # of stored non-zeros */
static oski_index_t A1632_ptr[A1632_M+1] = {1, 1, 1, 2, 4, 8}; /* end row pointers for A1632 */
static oski_index_t A1632_ind[A1632_NNZ] = {1, 1, 2, 1, 2, 3, 4}; /* end column indices pointers for A1632 */
static oski_value_t A1632_val[A1632_NNZ] = {MAKE_VAL_COMPLEX(-10, 10), MAKE_VAL_COMPLEX(-2, 2), MAKE_VAL_COMPLEX(-4, 3), MAKE_VAL_COMPLEX(-5, -5), MAKE_VAL_COMPLEX(10, 8), MAKE_VAL_COMPLEX(-3, 6), MAKE_VAL_COMPLEX(-3, -3)}; /* end stored non-zero values for A1632 */

#define x1632_LEN 5   /* vector length */
#define x1632_VW  3   /* vector width */
static oski_value_t x1632_val[x1632_LEN*x1632_VW] = {MAKE_VAL_COMPLEX(9, -5), MAKE_VAL_COMPLEX(8, 7), MAKE_VAL_COMPLEX(-10, -9), MAKE_VAL_COMPLEX(1, 8), MAKE_VAL_COMPLEX(-6, -5), MAKE_VAL_COMPLEX(6, 0), MAKE_VAL_COMPLEX(9, -10), MAKE_VAL_COMPLEX(-4, 10), MAKE_VAL_COMPLEX(-9, 0), MAKE_VAL_COMPLEX(10, 7), MAKE_VAL_COMPLEX(-6, 0), MAKE_VAL_COMPLEX(-8, -2), MAKE_VAL_COMPLEX(-6, -5), MAKE_VAL_COMPLEX(3, 7), MAKE_VAL_COMPLEX(-5, 3)}; /* data */

#define y1632_LEN 5   /* vector length */
#define y1632_VW  3   /* vector width */
static oski_value_t y1632_val[y1632_LEN*y1632_VW] = {MAKE_VAL_COMPLEX(6, -4), MAKE_VAL_COMPLEX(3, 6), MAKE_VAL_COMPLEX(-8, 3), MAKE_VAL_COMPLEX(-1, -10), MAKE_VAL_COMPLEX(10, 1), MAKE_VAL_COMPLEX(1, -3), MAKE_VAL_COMPLEX(-10, -6), MAKE_VAL_COMPLEX(-2, -10), MAKE_VAL_COMPLEX(8, 8), MAKE_VAL_COMPLEX(5, 0), MAKE_VAL_COMPLEX(-4, 9), MAKE_VAL_COMPLEX(5, -9), MAKE_VAL_COMPLEX(-3, 4), MAKE_VAL_COMPLEX(4, -3), MAKE_VAL_COMPLEX(7, 10)}; /* data */

/*
 *  BEGIN TEST 1632
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1632( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	oski_value_t* y_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-150, -22), MAKE_VAL_COMPLEX(101, -79), MAKE_VAL_COMPLEX(126, 52), MAKE_VAL_COMPLEX(-115, -22), MAKE_VAL_COMPLEX(74, 56), MAKE_VAL_COMPLEX(3, 111), MAKE_VAL_COMPLEX(-13, 199), MAKE_VAL_COMPLEX(-115, 11), MAKE_VAL_COMPLEX(190, -13), MAKE_VAL_COMPLEX(-8, 3), MAKE_VAL_COMPLEX(-15, -35), MAKE_VAL_COMPLEX(-3, 17), MAKE_VAL_COMPLEX(-97, 84), MAKE_VAL_COMPLEX(-64, -193), MAKE_VAL_COMPLEX(84, 92) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1632, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1632", y_true, y_val, y1632_LEN*y1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1632() */
/* END TEST 1632 */

/*
 *  BEGIN TEST 1633
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1633( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	oski_value_t* y_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(74, 38), MAKE_VAL_COMPLEX(-107, 7), MAKE_VAL_COMPLEX(-34, -50), MAKE_VAL_COMPLEX(85, 126), MAKE_VAL_COMPLEX(-170, -106), MAKE_VAL_COMPLEX(-25, -85), MAKE_VAL_COMPLEX(89, -99), MAKE_VAL_COMPLEX(95, 97), MAKE_VAL_COMPLEX(-238, -99), MAKE_VAL_COMPLEX(-42, -23), MAKE_VAL_COMPLEX(91, -39), MAKE_VAL_COMPLEX(-83, 53), MAKE_VAL_COMPLEX(143, -112), MAKE_VAL_COMPLEX(12, 207), MAKE_VAL_COMPLEX(-114, -220) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1633, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1633", y_true, y_val, y1632_LEN*y1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1633() */
/* END TEST 1633 */

/*
 *  BEGIN TEST 1634
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1634( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	oski_value_t* y_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0), MAKE_VAL_COMPLEX(0, 0) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1634, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1634", y_true, y_val, y1632_LEN*y1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1634() */
/* END TEST 1634 */

/*
 *  BEGIN TEST 1635
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1635( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	oski_value_t* y_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1074, 824), MAKE_VAL_COMPLEX(199, -1252), MAKE_VAL_COMPLEX(1104, -415), MAKE_VAL_COMPLEX(-1165, 626), MAKE_VAL_COMPLEX(1110, -517), MAKE_VAL_COMPLEX(841, 557), MAKE_VAL_COMPLEX(1156, 1464), MAKE_VAL_COMPLEX(-920, 852), MAKE_VAL_COMPLEX(1412, -1670), MAKE_VAL_COMPLEX(71, -58), MAKE_VAL_COMPLEX(-194, 193), MAKE_VAL_COMPLEX(-39, -185), MAKE_VAL_COMPLEX(145, 1440), MAKE_VAL_COMPLEX(-1936, -815), MAKE_VAL_COMPLEX(1653, -126) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(6, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1635, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1635", y_true, y_val, y1632_LEN*y1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1635() */
/* END TEST 1635 */

/*
 *  BEGIN TEST 1636
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1636( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(185, -71), MAKE_VAL_COMPLEX(107, 38), MAKE_VAL_COMPLEX(-182, -33), MAKE_VAL_COMPLEX(-58, 26), MAKE_VAL_COMPLEX(42, -32), MAKE_VAL_COMPLEX(4, 186), MAKE_VAL_COMPLEX(-146, -93), MAKE_VAL_COMPLEX(60, -45), MAKE_VAL_COMPLEX(28, 97), MAKE_VAL_COMPLEX(11, 13), MAKE_VAL_COMPLEX(-62, -28), MAKE_VAL_COMPLEX(9, -19), MAKE_VAL_COMPLEX(-145, 68), MAKE_VAL_COMPLEX(22, -87), MAKE_VAL_COMPLEX(58, -95) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1636, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1636", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1636() */
/* END TEST 1636 */

/*
 *  BEGIN TEST 1637
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1637( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-205, 35), MAKE_VAL_COMPLEX(-79, -70), MAKE_VAL_COMPLEX(146, 73), MAKE_VAL_COMPLEX(90, -30), MAKE_VAL_COMPLEX(-62, 56), MAKE_VAL_COMPLEX(-4, -210), MAKE_VAL_COMPLEX(106, 57), MAKE_VAL_COMPLEX(-20, 61), MAKE_VAL_COMPLEX(-28, -61), MAKE_VAL_COMPLEX(17, -53), MAKE_VAL_COMPLEX(62, 52), MAKE_VAL_COMPLEX(-17, 51), MAKE_VAL_COMPLEX(125, -44), MAKE_VAL_COMPLEX(6, 75), MAKE_VAL_COMPLEX(-46, 115) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1637, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1637", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1637() */
/* END TEST 1637 */

/*
 *  BEGIN TEST 1638
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1638( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-30, -54), MAKE_VAL_COMPLEX(42, -48), MAKE_VAL_COMPLEX(-54, 60), MAKE_VAL_COMPLEX(48, -6), MAKE_VAL_COMPLEX(-30, 36), MAKE_VAL_COMPLEX(0, -36), MAKE_VAL_COMPLEX(-60, -54), MAKE_VAL_COMPLEX(60, 24), MAKE_VAL_COMPLEX(0, 54), MAKE_VAL_COMPLEX(42, -60), MAKE_VAL_COMPLEX(0, 36), MAKE_VAL_COMPLEX(-12, 48), MAKE_VAL_COMPLEX(-30, 36), MAKE_VAL_COMPLEX(42, -18), MAKE_VAL_COMPLEX(18, 30) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1638, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1638", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1638() */
/* END TEST 1638 */

/*
 *  BEGIN TEST 1639
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1639( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-560, 858), MAKE_VAL_COMPLEX(-455, 117), MAKE_VAL_COMPLEX(699, -397), MAKE_VAL_COMPLEX(279, -363), MAKE_VAL_COMPLEX(-129, 415), MAKE_VAL_COMPLEX(-766, -842), MAKE_VAL_COMPLEX(757, -373), MAKE_VAL_COMPLEX(134, 486), MAKE_VAL_COMPLEX(-491, -105), MAKE_VAL_COMPLEX(27, -205), MAKE_VAL_COMPLEX(366, -22), MAKE_VAL_COMPLEX(10, 266), MAKE_VAL_COMPLEX(219, -733), MAKE_VAL_COMPLEX(390, 372), MAKE_VAL_COMPLEX(210, 704) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-4, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1639, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1639", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1639() */
/* END TEST 1639 */

/*
 *  BEGIN TEST 1640
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1640( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(81, 95), MAKE_VAL_COMPLEX(-23, 194), MAKE_VAL_COMPLEX(-162, -219), MAKE_VAL_COMPLEX(-22, 80), MAKE_VAL_COMPLEX(44, -116), MAKE_VAL_COMPLEX(134, 68), MAKE_VAL_COMPLEX(58, 59), MAKE_VAL_COMPLEX(-152, -9), MAKE_VAL_COMPLEX(52, -183), MAKE_VAL_COMPLEX(75, 81), MAKE_VAL_COMPLEX(-98, 44), MAKE_VAL_COMPLEX(-69, -43), MAKE_VAL_COMPLEX(43, -202), MAKE_VAL_COMPLEX(224, 83), MAKE_VAL_COMPLEX(-46, 59) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1640, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1640", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1640() */
/* END TEST 1640 */

/*
 *  BEGIN TEST 1641
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1641( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-30, -194), MAKE_VAL_COMPLEX(137, -179), MAKE_VAL_COMPLEX(18, 198), MAKE_VAL_COMPLEX(79, -14), MAKE_VAL_COMPLEX(-128, 107), MAKE_VAL_COMPLEX(-80, -104), MAKE_VAL_COMPLEX(-37, -203), MAKE_VAL_COMPLEX(176, 123), MAKE_VAL_COMPLEX(-133, 237), MAKE_VAL_COMPLEX(57, -78), MAKE_VAL_COMPLEX(44, -8), MAKE_VAL_COMPLEX(-15, 73), MAKE_VAL_COMPLEX(-127, 193), MAKE_VAL_COMPLEX(-155, -38), MAKE_VAL_COMPLEX(19, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1641, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1641", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1641() */
/* END TEST 1641 */

/*
 *  BEGIN TEST 1642
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1642( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-43, 71), MAKE_VAL_COMPLEX(-84, -17), MAKE_VAL_COMPLEX(106, 23), MAKE_VAL_COMPLEX(-39, -52), MAKE_VAL_COMPLEX(62, 11), MAKE_VAL_COMPLEX(-42, 24), MAKE_VAL_COMPLEX(-23, 106), MAKE_VAL_COMPLEX(-12, -86), MAKE_VAL_COMPLEX(63, -36), MAKE_VAL_COMPLEX(-98, -9), MAKE_VAL_COMPLEX(42, -24), MAKE_VAL_COMPLEX(64, -18), MAKE_VAL_COMPLEX(62, 11), MAKE_VAL_COMPLEX(-49, -37), MAKE_VAL_COMPLEX(23, -41) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1642, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1642", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1642() */
/* END TEST 1642 */

/*
 *  BEGIN TEST 1643
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1643( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-948, 426), MAKE_VAL_COMPLEX(-1443, -321), MAKE_VAL_COMPLEX(1391, -1293), MAKE_VAL_COMPLEX(-402, -128), MAKE_VAL_COMPLEX(854, 512), MAKE_VAL_COMPLEX(-454, 1022), MAKE_VAL_COMPLEX(-857, 163), MAKE_VAL_COMPLEX(333, -1169), MAKE_VAL_COMPLEX(1663, 663), MAKE_VAL_COMPLEX(-320, 364), MAKE_VAL_COMPLEX(-518, -586), MAKE_VAL_COMPLEX(202, -336), MAKE_VAL_COMPLEX(1627, 417), MAKE_VAL_COMPLEX(-217, 2001), MAKE_VAL_COMPLEX(-471, -133) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 9);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1643, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1643", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1643() */
/* END TEST 1643 */

/*
 *  BEGIN TEST 1644
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1644( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	oski_value_t* y_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-116, -88), MAKE_VAL_COMPLEX(-70, -31), MAKE_VAL_COMPLEX(-76, 239), MAKE_VAL_COMPLEX(-49, -52), MAKE_VAL_COMPLEX(-41, 52), MAKE_VAL_COMPLEX(101, -84), MAKE_VAL_COMPLEX(14, 73), MAKE_VAL_COMPLEX(-183, -69), MAKE_VAL_COMPLEX(61, 0), MAKE_VAL_COMPLEX(-22, -227), MAKE_VAL_COMPLEX(123, 135), MAKE_VAL_COMPLEX(-9, 23), MAKE_VAL_COMPLEX(216, 50), MAKE_VAL_COMPLEX(21, -48), MAKE_VAL_COMPLEX(197, 147) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1644, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1644", y_true, y_val, y1632_LEN*y1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1644() */
/* END TEST 1644 */

/*
 *  BEGIN TEST 1645
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1645( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	oski_value_t* y_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(100, 38), MAKE_VAL_COMPLEX(118, 22), MAKE_VAL_COMPLEX(81, -177), MAKE_VAL_COMPLEX(-23, 39), MAKE_VAL_COMPLEX(68, -120), MAKE_VAL_COMPLEX(-120, 71), MAKE_VAL_COMPLEX(-76, -15), MAKE_VAL_COMPLEX(109, 63), MAKE_VAL_COMPLEX(11, -40), MAKE_VAL_COMPLEX(32, 192), MAKE_VAL_COMPLEX(-68, -89), MAKE_VAL_COMPLEX(-44, -76), MAKE_VAL_COMPLEX(-194, -21), MAKE_VAL_COMPLEX(-34, 14), MAKE_VAL_COMPLEX(-113, -176) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1645, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1645", y_true, y_val, y1632_LEN*y1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1645() */
/* END TEST 1645 */

/*
 *  BEGIN TEST 1646
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1646( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	oski_value_t* y_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-12, 8), MAKE_VAL_COMPLEX(-6, -12), MAKE_VAL_COMPLEX(16, -6), MAKE_VAL_COMPLEX(2, 20), MAKE_VAL_COMPLEX(-20, -2), MAKE_VAL_COMPLEX(-2, 6), MAKE_VAL_COMPLEX(20, 12), MAKE_VAL_COMPLEX(4, 20), MAKE_VAL_COMPLEX(-16, -16), MAKE_VAL_COMPLEX(-10, 0), MAKE_VAL_COMPLEX(8, -18), MAKE_VAL_COMPLEX(-10, 18), MAKE_VAL_COMPLEX(6, -8), MAKE_VAL_COMPLEX(-8, 6), MAKE_VAL_COMPLEX(-14, -20) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1646, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1646", y_true, y_val, y1632_LEN*y1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1646() */
/* END TEST 1646 */

/*
 *  BEGIN TEST 1647
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1647( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	oski_value_t* y_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1094, 434), MAKE_VAL_COMPLEX(856, 597), MAKE_VAL_COMPLEX(450, -1459), MAKE_VAL_COMPLEX(31, -100), MAKE_VAL_COMPLEX(995, -566), MAKE_VAL_COMPLEX(-908, 369), MAKE_VAL_COMPLEX(-779, -702), MAKE_VAL_COMPLEX(1027, 195), MAKE_VAL_COMPLEX(184, 211), MAKE_VAL_COMPLEX(244, 1713), MAKE_VAL_COMPLEX(-860, -572), MAKE_VAL_COMPLEX(165, -831), MAKE_VAL_COMPLEX(-1755, -305), MAKE_VAL_COMPLEX(-55, 86), MAKE_VAL_COMPLEX(-766, -958) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 9);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1647, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1647", y_true, y_val, y1632_LEN*y1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1647() */
/* END TEST 1647 */

/*
 *  BEGIN TEST 1648
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1648( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(99, -223), MAKE_VAL_COMPLEX(157, 91), MAKE_VAL_COMPLEX(-118, -183), MAKE_VAL_COMPLEX(162, 202), MAKE_VAL_COMPLEX(-85, -1), MAKE_VAL_COMPLEX(42, 153), MAKE_VAL_COMPLEX(-93, -77), MAKE_VAL_COMPLEX(-63, 142), MAKE_VAL_COMPLEX(-35, -7), MAKE_VAL_COMPLEX(69, -47), MAKE_VAL_COMPLEX(-47, 69), MAKE_VAL_COMPLEX(-86, -37), MAKE_VAL_COMPLEX(84, 126), MAKE_VAL_COMPLEX(-187, 136), MAKE_VAL_COMPLEX(-46, -116) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1648, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1648", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1648() */
/* END TEST 1648 */

/*
 *  BEGIN TEST 1649
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1649( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-19, 49), MAKE_VAL_COMPLEX(40, -74), MAKE_VAL_COMPLEX(-131, 158), MAKE_VAL_COMPLEX(-59, -93), MAKE_VAL_COMPLEX(-60, -8), MAKE_VAL_COMPLEX(48, -219), MAKE_VAL_COMPLEX(118, -172), MAKE_VAL_COMPLEX(113, 52), MAKE_VAL_COMPLEX(-100, 106), MAKE_VAL_COMPLEX(158, 42), MAKE_VAL_COMPLEX(-43, -3), MAKE_VAL_COMPLEX(-56, 95), MAKE_VAL_COMPLEX(-229, -135), MAKE_VAL_COMPLEX(309, -64), MAKE_VAL_COMPLEX(4, 216) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1649, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1649", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1649() */
/* END TEST 1649 */

/*
 *  BEGIN TEST 1650
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1650( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-3, -69), MAKE_VAL_COMPLEX(66, -27), MAKE_VAL_COMPLEX(-84, 33), MAKE_VAL_COMPLEX(51, 18), MAKE_VAL_COMPLEX(-48, 21), MAKE_VAL_COMPLEX(18, -36), MAKE_VAL_COMPLEX(-33, -84), MAKE_VAL_COMPLEX(48, 54), MAKE_VAL_COMPLEX(-27, 54), MAKE_VAL_COMPLEX(72, -39), MAKE_VAL_COMPLEX(-18, 36), MAKE_VAL_COMPLEX(-36, 42), MAKE_VAL_COMPLEX(-48, 21), MAKE_VAL_COMPLEX(51, 3), MAKE_VAL_COMPLEX(3, 39) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1650, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1650", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1650() */
/* END TEST 1650 */

/*
 *  BEGIN TEST 1651
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1651( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1209, 455), MAKE_VAL_COMPLEX(-507, 507), MAKE_VAL_COMPLEX(1092, 65), MAKE_VAL_COMPLEX(-1232, 912), MAKE_VAL_COMPLEX(-173, -189), MAKE_VAL_COMPLEX(-1350, -237), MAKE_VAL_COMPLEX(-117, -1157), MAKE_VAL_COMPLEX(-317, -506), MAKE_VAL_COMPLEX(217, 357), MAKE_VAL_COMPLEX(657, -199), MAKE_VAL_COMPLEX(-421, -25), MAKE_VAL_COMPLEX(354, -63), MAKE_VAL_COMPLEX(-1358, 1036), MAKE_VAL_COMPLEX(-453, -1940), MAKE_VAL_COMPLEX(1254, 56) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 8);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1651, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1651", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1651() */
/* END TEST 1651 */

/*
 *  BEGIN TEST 1652
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1652( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(48, 16), MAKE_VAL_COMPLEX(-8, 81), MAKE_VAL_COMPLEX(-155, 17), MAKE_VAL_COMPLEX(154, 11), MAKE_VAL_COMPLEX(-38, -79), MAKE_VAL_COMPLEX(160, -13), MAKE_VAL_COMPLEX(-87, -82), MAKE_VAL_COMPLEX(63, 82), MAKE_VAL_COMPLEX(-59, -126), MAKE_VAL_COMPLEX(38, -23), MAKE_VAL_COMPLEX(25, 73), MAKE_VAL_COMPLEX(-106, 67), MAKE_VAL_COMPLEX(-149, 74), MAKE_VAL_COMPLEX(70, -149), MAKE_VAL_COMPLEX(157, -43) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1652, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1652", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1652() */
/* END TEST 1652 */

/*
 *  BEGIN TEST 1653
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1653( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-71, -121), MAKE_VAL_COMPLEX(102, -140), MAKE_VAL_COMPLEX(35, 56), MAKE_VAL_COMPLEX(-71, 3), MAKE_VAL_COMPLEX(-30, 124), MAKE_VAL_COMPLEX(-142, -47), MAKE_VAL_COMPLEX(14, -38), MAKE_VAL_COMPLEX(25, -12), MAKE_VAL_COMPLEX(32, 216), MAKE_VAL_COMPLEX(62, -56), MAKE_VAL_COMPLEX(-43, -13), MAKE_VAL_COMPLEX(62, 7), MAKE_VAL_COMPLEX(81, -29), MAKE_VAL_COMPLEX(9, 140), MAKE_VAL_COMPLEX(-142, 102) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1653, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1653", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1653() */
/* END TEST 1653 */

/*
 *  BEGIN TEST 1654
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1654( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(58, 62), MAKE_VAL_COMPLEX(-40, 78), MAKE_VAL_COMPLEX(52, -98), MAKE_VAL_COMPLEX(-62, 24), MAKE_VAL_COMPLEX(28, -58), MAKE_VAL_COMPLEX(12, 48), MAKE_VAL_COMPLEX(98, 52), MAKE_VAL_COMPLEX(-88, -12), MAKE_VAL_COMPLEX(-18, -72), MAKE_VAL_COMPLEX(-36, 94), MAKE_VAL_COMPLEX(-12, -48), MAKE_VAL_COMPLEX(0, -68), MAKE_VAL_COMPLEX(28, -58), MAKE_VAL_COMPLEX(-50, 38), MAKE_VAL_COMPLEX(-34, -34) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1654, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1654", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1654() */
/* END TEST 1654 */

/*
 *  BEGIN TEST 1655
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_ROWMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1655( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_ROWMAJ, x1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(882, 90), MAKE_VAL_COMPLEX(385, 1068), MAKE_VAL_COMPLEX(-299, 285), MAKE_VAL_COMPLEX(261, -720), MAKE_VAL_COMPLEX(-607, -558), MAKE_VAL_COMPLEX(741, -824), MAKE_VAL_COMPLEX(-54, 483), MAKE_VAL_COMPLEX(118, -25), MAKE_VAL_COMPLEX(-1406, -613), MAKE_VAL_COMPLEX(-95, 446), MAKE_VAL_COMPLEX(419, -231), MAKE_VAL_COMPLEX(-95, 482), MAKE_VAL_COMPLEX(20, 831), MAKE_VAL_COMPLEX(-1073, -688), MAKE_VAL_COMPLEX(-7, -1465) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(4, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -9);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1655, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1655", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1655() */
/* END TEST 1655 */

/*
 *  BEGIN TEST 1656
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1656( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	oski_value_t* y_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(26, 186), MAKE_VAL_COMPLEX(64, -87), MAKE_VAL_COMPLEX(100, 39), MAKE_VAL_COMPLEX(-83, 62), MAKE_VAL_COMPLEX(113, -22), MAKE_VAL_COMPLEX(-40, 56), MAKE_VAL_COMPLEX(16, 256), MAKE_VAL_COMPLEX(-54, 83), MAKE_VAL_COMPLEX(31, -132), MAKE_VAL_COMPLEX(-72, 19), MAKE_VAL_COMPLEX(-24, 15), MAKE_VAL_COMPLEX(-4, 26), MAKE_VAL_COMPLEX(88, 19), MAKE_VAL_COMPLEX(87, -59), MAKE_VAL_COMPLEX(-22, -206) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1656, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1656", y_true, y_val, y1632_LEN*y1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1656() */
/* END TEST 1656 */

/*
 *  BEGIN TEST 1657
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1657( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	oski_value_t* y_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-136, -156), MAKE_VAL_COMPLEX(-79, -18), MAKE_VAL_COMPLEX(35, -44), MAKE_VAL_COMPLEX(48, 93), MAKE_VAL_COMPLEX(-258, -43), MAKE_VAL_COMPLEX(10, -16), MAKE_VAL_COMPLEX(104, -116), MAKE_VAL_COMPLEX(34, 77), MAKE_VAL_COMPLEX(-111, -28), MAKE_VAL_COMPLEX(-3, -44), MAKE_VAL_COMPLEX(129, -130), MAKE_VAL_COMPLEX(-116, 84), MAKE_VAL_COMPLEX(-23, -64), MAKE_VAL_COMPLEX(-162, 84), MAKE_VAL_COMPLEX(-33, 21) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1657, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1657", y_true, y_val, y1632_LEN*y1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1657() */
/* END TEST 1657 */

/*
 *  BEGIN TEST 1658
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1658( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	oski_value_t* y_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(28, 16), MAKE_VAL_COMPLEX(-18, 24), MAKE_VAL_COMPLEX(-28, -26), MAKE_VAL_COMPLEX(38, -24), MAKE_VAL_COMPLEX(16, 42), MAKE_VAL_COMPLEX(14, -2), MAKE_VAL_COMPLEX(4, -52), MAKE_VAL_COMPLEX(36, -28), MAKE_VAL_COMPLEX(-16, 48), MAKE_VAL_COMPLEX(10, 20), MAKE_VAL_COMPLEX(-44, 2), MAKE_VAL_COMPLEX(46, 2), MAKE_VAL_COMPLEX(-22, -4), MAKE_VAL_COMPLEX(20, 10), MAKE_VAL_COMPLEX(-26, 48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1658, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1658", y_true, y_val, y1632_LEN*y1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1658() */
/* END TEST 1658 */

/*
 *  BEGIN TEST 1659
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1659( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	oski_value_t* y_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(592, -1596), MAKE_VAL_COMPLEX(-654, -403), MAKE_VAL_COMPLEX(232, -441), MAKE_VAL_COMPLEX(203, 696), MAKE_VAL_COMPLEX(-1165, -1470), MAKE_VAL_COMPLEX(360, -10), MAKE_VAL_COMPLEX(1644, -632), MAKE_VAL_COMPLEX(217, 444), MAKE_VAL_COMPLEX(-822, -369), MAKE_VAL_COMPLEX(360, 5), MAKE_VAL_COMPLEX(1106, 17), MAKE_VAL_COMPLEX(-692, -144), MAKE_VAL_COMPLEX(36, -663), MAKE_VAL_COMPLEX(-1330, -440), MAKE_VAL_COMPLEX(-827, 441) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1659, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1659", y_true, y_val, y1632_LEN*y1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1659() */
/* END TEST 1659 */

/*
 *  BEGIN TEST 1660
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1660( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(129, -87), MAKE_VAL_COMPLEX(-58, -56), MAKE_VAL_COMPLEX(-158, 41), MAKE_VAL_COMPLEX(33, 5), MAKE_VAL_COMPLEX(-141, 112), MAKE_VAL_COMPLEX(60, 13), MAKE_VAL_COMPLEX(-34, -78), MAKE_VAL_COMPLEX(116, -69), MAKE_VAL_COMPLEX(-29, 17), MAKE_VAL_COMPLEX(17, -176), MAKE_VAL_COMPLEX(-127, -12), MAKE_VAL_COMPLEX(8, 266), MAKE_VAL_COMPLEX(20, 127), MAKE_VAL_COMPLEX(63, -83), MAKE_VAL_COMPLEX(90, -87) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1660, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1660", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1660() */
/* END TEST 1660 */

/*
 *  BEGIN TEST 1661
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1661( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-195, 53), MAKE_VAL_COMPLEX(68, -20), MAKE_VAL_COMPLEX(144, 55), MAKE_VAL_COMPLEX(11, -43), MAKE_VAL_COMPLEX(135, -56), MAKE_VAL_COMPLEX(-84, -49), MAKE_VAL_COMPLEX(-62, 64), MAKE_VAL_COMPLEX(-40, 53), MAKE_VAL_COMPLEX(65, 37), MAKE_VAL_COMPLEX(-15, 88), MAKE_VAL_COMPLEX(151, 48), MAKE_VAL_COMPLEX(12, -210), MAKE_VAL_COMPLEX(-26, -71), MAKE_VAL_COMPLEX(-33, 37), MAKE_VAL_COMPLEX(-52, 105) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1661, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1661", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1661() */
/* END TEST 1661 */

/*
 *  BEGIN TEST 1662
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1662( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-39, 57), MAKE_VAL_COMPLEX(-69, -18), MAKE_VAL_COMPLEX(87, 24), MAKE_VAL_COMPLEX(-30, -45), MAKE_VAL_COMPLEX(51, 12), MAKE_VAL_COMPLEX(-36, 18), MAKE_VAL_COMPLEX(-24, 87), MAKE_VAL_COMPLEX(-6, -72), MAKE_VAL_COMPLEX(54, -27), MAKE_VAL_COMPLEX(-81, -12), MAKE_VAL_COMPLEX(36, -18), MAKE_VAL_COMPLEX(54, -12), MAKE_VAL_COMPLEX(51, 12), MAKE_VAL_COMPLEX(-39, -33), MAKE_VAL_COMPLEX(21, -33) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1662, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1662", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1662() */
/* END TEST 1662 */

/*
 *  BEGIN TEST 1663
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1663( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(957, -1775), MAKE_VAL_COMPLEX(-462, 754), MAKE_VAL_COMPLEX(-1262, 681), MAKE_VAL_COMPLEX(149, 355), MAKE_VAL_COMPLEX(-391, 1418), MAKE_VAL_COMPLEX(944, -245), MAKE_VAL_COMPLEX(8, -758), MAKE_VAL_COMPLEX(-260, -807), MAKE_VAL_COMPLEX(-697, 135), MAKE_VAL_COMPLEX(-757, -638), MAKE_VAL_COMPLEX(-1405, 788), MAKE_VAL_COMPLEX(1672, 1466), MAKE_VAL_COMPLEX(856, 235), MAKE_VAL_COMPLEX(-181, -523), MAKE_VAL_COMPLEX(-488, -1239) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1663, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1663", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1663() */
/* END TEST 1663 */

/*
 *  BEGIN TEST 1664
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1664( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-33, 123), MAKE_VAL_COMPLEX(-53, -19), MAKE_VAL_COMPLEX(45, 174), MAKE_VAL_COMPLEX(42, -14), MAKE_VAL_COMPLEX(94, -129), MAKE_VAL_COMPLEX(-92, 139), MAKE_VAL_COMPLEX(-5, -43), MAKE_VAL_COMPLEX(-78, -107), MAKE_VAL_COMPLEX(-14, 68), MAKE_VAL_COMPLEX(170, -17), MAKE_VAL_COMPLEX(-85, -152), MAKE_VAL_COMPLEX(148, 90), MAKE_VAL_COMPLEX(123, -132), MAKE_VAL_COMPLEX(-35, -73), MAKE_VAL_COMPLEX(18, 41) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1664, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1664", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1664() */
/* END TEST 1664 */

/*
 *  BEGIN TEST 1665
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1665( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(4, -154), MAKE_VAL_COMPLEX(73, -20), MAKE_VAL_COMPLEX(-71, -125), MAKE_VAL_COMPLEX(-11, 2), MAKE_VAL_COMPLEX(-108, 158), MAKE_VAL_COMPLEX(86, -163), MAKE_VAL_COMPLEX(-44, 17), MAKE_VAL_COMPLEX(122, 113), MAKE_VAL_COMPLEX(23, -32), MAKE_VAL_COMPLEX(-152, -30), MAKE_VAL_COMPLEX(91, 176), MAKE_VAL_COMPLEX(-148, -56), MAKE_VAL_COMPLEX(-137, 161), MAKE_VAL_COMPLEX(60, 54), MAKE_VAL_COMPLEX(-1, -24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1665, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1665", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1665() */
/* END TEST 1665 */

/*
 *  BEGIN TEST 1666
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1666( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-48, -44), MAKE_VAL_COMPLEX(26, -62), MAKE_VAL_COMPLEX(-34, 78), MAKE_VAL_COMPLEX(46, -22), MAKE_VAL_COMPLEX(-18, 46), MAKE_VAL_COMPLEX(-12, -36), MAKE_VAL_COMPLEX(-78, -34), MAKE_VAL_COMPLEX(68, 4), MAKE_VAL_COMPLEX(18, 54), MAKE_VAL_COMPLEX(22, -74), MAKE_VAL_COMPLEX(12, 36), MAKE_VAL_COMPLEX(4, 52), MAKE_VAL_COMPLEX(-18, 46), MAKE_VAL_COMPLEX(36, -32), MAKE_VAL_COMPLEX(28, 24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1666, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1666", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1666() */
/* END TEST 1666 */

/*
 *  BEGIN TEST 1667
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1667( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_ROWMAJ, y1632_VW );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-894, -632), MAKE_VAL_COMPLEX(-65, -412), MAKE_VAL_COMPLEX(-667, -651), MAKE_VAL_COMPLEX(-407, -56), MAKE_VAL_COMPLEX(524, 1624), MAKE_VAL_COMPLEX(-508, -1417), MAKE_VAL_COMPLEX(-4, 963), MAKE_VAL_COMPLEX(1220, -321), MAKE_VAL_COMPLEX(167, -722), MAKE_VAL_COMPLEX(-1744, 948), MAKE_VAL_COMPLEX(1825, 446), MAKE_VAL_COMPLEX(-1116, 220), MAKE_VAL_COMPLEX(339, 1819), MAKE_VAL_COMPLEX(398, -16), MAKE_VAL_COMPLEX(-161, -428) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, 6);
	oski_value_t beta = MAKE_VAL_COMPLEX(1, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1667, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1667", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1667() */
/* END TEST 1667 */

/*
 *  BEGIN TEST 1668
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1668( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	oski_value_t* y_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(98, 190), MAKE_VAL_COMPLEX(-99, -24), MAKE_VAL_COMPLEX(-66, 147), MAKE_VAL_COMPLEX(14, 35), MAKE_VAL_COMPLEX(39, 88), MAKE_VAL_COMPLEX(90, -20), MAKE_VAL_COMPLEX(235, -27), MAKE_VAL_COMPLEX(-10, -9), MAKE_VAL_COMPLEX(-118, 112), MAKE_VAL_COMPLEX(124, -58), MAKE_VAL_COMPLEX(-10, 25), MAKE_VAL_COMPLEX(6, 60), MAKE_VAL_COMPLEX(74, -60), MAKE_VAL_COMPLEX(67, -26), MAKE_VAL_COMPLEX(-26, -84) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1668, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1668", y_true, y_val, y1632_LEN*y1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1668() */
/* END TEST 1668 */

/*
 *  BEGIN TEST 1669
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1669( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	oski_value_t* y_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-50, -144), MAKE_VAL_COMPLEX(51, 63), MAKE_VAL_COMPLEX(23, -213), MAKE_VAL_COMPLEX(74, -64), MAKE_VAL_COMPLEX(-28, 4), MAKE_VAL_COMPLEX(-61, 23), MAKE_VAL_COMPLEX(-201, -75), MAKE_VAL_COMPLEX(96, -29), MAKE_VAL_COMPLEX(62, -24), MAKE_VAL_COMPLEX(-114, 103), MAKE_VAL_COMPLEX(-79, -43), MAKE_VAL_COMPLEX(85, -33), MAKE_VAL_COMPLEX(-116, 41), MAKE_VAL_COMPLEX(-32, 56), MAKE_VAL_COMPLEX(-50, 167) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1669, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1669", y_true, y_val, y1632_LEN*y1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1669() */
/* END TEST 1669 */

/*
 *  BEGIN TEST 1670
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1670( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	oski_value_t* y_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(4, 32), MAKE_VAL_COMPLEX(-30, 0), MAKE_VAL_COMPLEX(4, -38), MAKE_VAL_COMPLEX(42, 16), MAKE_VAL_COMPLEX(-24, 38), MAKE_VAL_COMPLEX(10, 10), MAKE_VAL_COMPLEX(44, -28), MAKE_VAL_COMPLEX(44, 12), MAKE_VAL_COMPLEX(-48, 16), MAKE_VAL_COMPLEX(-10, 20), MAKE_VAL_COMPLEX(-28, -34), MAKE_VAL_COMPLEX(26, 38), MAKE_VAL_COMPLEX(-10, -20), MAKE_VAL_COMPLEX(4, 22), MAKE_VAL_COMPLEX(-54, 8) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1670, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1670", y_true, y_val, y1632_LEN*y1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1670() */
/* END TEST 1670 */

/*
 *  BEGIN TEST 1671
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1671( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	oski_value_t* y_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-554, 1800), MAKE_VAL_COMPLEX(-336, -753), MAKE_VAL_COMPLEX(-1653, 792), MAKE_VAL_COMPLEX(-354, 91), MAKE_VAL_COMPLEX(-55, 736), MAKE_VAL_COMPLEX(743, 391), MAKE_VAL_COMPLEX(1456, 1474), MAKE_VAL_COMPLEX(-241, -367), MAKE_VAL_COMPLEX(-1250, 30), MAKE_VAL_COMPLEX(1409, 372), MAKE_VAL_COMPLEX(-191, 420), MAKE_VAL_COMPLEX(-405, 129), MAKE_VAL_COMPLEX(929, 260), MAKE_VAL_COMPLEX(705, 134), MAKE_VAL_COMPLEX(775, -626) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1671, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1671", y_true, y_val, y1632_LEN*y1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1671() */
/* END TEST 1671 */

/*
 *  BEGIN TEST 1672
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1672( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(30, -220), MAKE_VAL_COMPLEX(195, 106), MAKE_VAL_COMPLEX(-228, 73), MAKE_VAL_COMPLEX(9, -62), MAKE_VAL_COMPLEX(105, 174), MAKE_VAL_COMPLEX(82, 16), MAKE_VAL_COMPLEX(-49, -58), MAKE_VAL_COMPLEX(-9, 94), MAKE_VAL_COMPLEX(-20, 105), MAKE_VAL_COMPLEX(-163, 43), MAKE_VAL_COMPLEX(-19, -96), MAKE_VAL_COMPLEX(-48, 213), MAKE_VAL_COMPLEX(-2, -13), MAKE_VAL_COMPLEX(43, -40), MAKE_VAL_COMPLEX(-7, -119) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -6);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1672, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1672", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1672() */
/* END TEST 1672 */

/*
 *  BEGIN TEST 1673
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1673( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-85, 227), MAKE_VAL_COMPLEX(-221, -157), MAKE_VAL_COMPLEX(260, -8), MAKE_VAL_COMPLEX(2, 20), MAKE_VAL_COMPLEX(-85, -137), MAKE_VAL_COMPLEX(-112, -28), MAKE_VAL_COMPLEX(-16, 90), MAKE_VAL_COMPLEX(49, -136), MAKE_VAL_COMPLEX(65, -87), MAKE_VAL_COMPLEX(127, -98), MAKE_VAL_COMPLEX(49, 108), MAKE_VAL_COMPLEX(84, -187), MAKE_VAL_COMPLEX(22, 50), MAKE_VAL_COMPLEX(-44, -1), MAKE_VAL_COMPLEX(38, 114) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1673, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1673", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1673() */
/* END TEST 1673 */

/*
 *  BEGIN TEST 1674
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1674( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(16, 50), MAKE_VAL_COMPLEX(-43, 33), MAKE_VAL_COMPLEX(55, -41), MAKE_VAL_COMPLEX(-41, -3), MAKE_VAL_COMPLEX(31, -25), MAKE_VAL_COMPLEX(-6, 30), MAKE_VAL_COMPLEX(41, 55), MAKE_VAL_COMPLEX(-46, -30), MAKE_VAL_COMPLEX(9, -45), MAKE_VAL_COMPLEX(-45, 43), MAKE_VAL_COMPLEX(6, -30), MAKE_VAL_COMPLEX(18, -38), MAKE_VAL_COMPLEX(31, -25), MAKE_VAL_COMPLEX(-38, 8), MAKE_VAL_COMPLEX(-10, -28) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1674, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1674", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1674() */
/* END TEST 1674 */

/*
 *  BEGIN TEST 1675
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1675( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(76, -788), MAKE_VAL_COMPLEX(646, 567), MAKE_VAL_COMPLEX(-723, 324), MAKE_VAL_COMPLEX(-160, -455), MAKE_VAL_COMPLEX(766, 837), MAKE_VAL_COMPLEX(284, 218), MAKE_VAL_COMPLEX(-204, 127), MAKE_VAL_COMPLEX(-191, 168), MAKE_VAL_COMPLEX(89, 318), MAKE_VAL_COMPLEX(-1186, 298), MAKE_VAL_COMPLEX(31, -618), MAKE_VAL_COMPLEX(-26, 923), MAKE_VAL_COMPLEX(231, -98), MAKE_VAL_COMPLEX(-9, -278), MAKE_VAL_COMPLEX(1, -773) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(5, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -7);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1675, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1675", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1675() */
/* END TEST 1675 */

/*
 *  BEGIN TEST 1676
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1676( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(91, 157), MAKE_VAL_COMPLEX(59, 41), MAKE_VAL_COMPLEX(-18, -38), MAKE_VAL_COMPLEX(-114, 44), MAKE_VAL_COMPLEX(-61, 5), MAKE_VAL_COMPLEX(-92, 146), MAKE_VAL_COMPLEX(110, -68), MAKE_VAL_COMPLEX(-65, -4), MAKE_VAL_COMPLEX(37, -29), MAKE_VAL_COMPLEX(-23, -86), MAKE_VAL_COMPLEX(-45, -54), MAKE_VAL_COMPLEX(126, -17), MAKE_VAL_COMPLEX(4, -244), MAKE_VAL_COMPLEX(-86, 45), MAKE_VAL_COMPLEX(130, -122) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1676, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1676", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1676() */
/* END TEST 1676 */

/*
 *  BEGIN TEST 1677
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1677( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-73, -61), MAKE_VAL_COMPLEX(-146, 10), MAKE_VAL_COMPLEX(129, -25), MAKE_VAL_COMPLEX(39, -59), MAKE_VAL_COMPLEX(124, -44), MAKE_VAL_COMPLEX(74, -92), MAKE_VAL_COMPLEX(-47, 179), MAKE_VAL_COMPLEX(-13, -62), MAKE_VAL_COMPLEX(-10, -52), MAKE_VAL_COMPLEX(-70, 155), MAKE_VAL_COMPLEX(63, 0), MAKE_VAL_COMPLEX(-84, -49), MAKE_VAL_COMPLEX(59, 205), MAKE_VAL_COMPLEX(14, -39), MAKE_VAL_COMPLEX(-142, 68) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 3);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1677, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1677", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1677() */
/* END TEST 1677 */

/*
 *  BEGIN TEST 1678
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1678( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-49, 39), MAKE_VAL_COMPLEX(-55, -34), MAKE_VAL_COMPLEX(69, 44), MAKE_VAL_COMPLEX(-14, -47), MAKE_VAL_COMPLEX(41, 24), MAKE_VAL_COMPLEX(-36, 6), MAKE_VAL_COMPLEX(-44, 69), MAKE_VAL_COMPLEX(14, -64), MAKE_VAL_COMPLEX(54, -9), MAKE_VAL_COMPLEX(-67, -32), MAKE_VAL_COMPLEX(36, -6), MAKE_VAL_COMPLEX(50, 4), MAKE_VAL_COMPLEX(41, 24), MAKE_VAL_COMPLEX(-25, -39), MAKE_VAL_COMPLEX(27, -23) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1678, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1678", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1678() */
/* END TEST 1678 */

/*
 *  BEGIN TEST 1679
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED>
 *  x : <'x1632', LAYOUT_COLMAJ>
 *  y : <'y1632', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1679( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1632_ptr, A1632_ind, A1632_val,
		A1632_M, A1632_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ONE_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1632_val,
		y1632_LEN * y1632_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1632_LEN, y1632_VW, LAYOUT_COLMAJ, y1632_LEN );
	oski_value_t* y_val = array_Duplicate( x1632_val,
		x1632_LEN * x1632_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1632_LEN, x1632_VW, LAYOUT_COLMAJ, x1632_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1231, 593), MAKE_VAL_COMPLEX(474, -853), MAKE_VAL_COMPLEX(-48, 743), MAKE_VAL_COMPLEX(-387, 698), MAKE_VAL_COMPLEX(-196, 1043), MAKE_VAL_COMPLEX(30, 1724), MAKE_VAL_COMPLEX(-567, -1024), MAKE_VAL_COMPLEX(16, -62), MAKE_VAL_COMPLEX(667, -258), MAKE_VAL_COMPLEX(-1332, -1273), MAKE_VAL_COMPLEX(-390, 108), MAKE_VAL_COMPLEX(1348, -770), MAKE_VAL_COMPLEX(-1668, -1469), MAKE_VAL_COMPLEX(-319, 489), MAKE_VAL_COMPLEX(489, -2067) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1679, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1679", y_true, y_val, x1632_LEN*x1632_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1679() */
/* END TEST 1679 */

#define A1680_M 5    /* # of rows */
#define A1680_N 5    /* # of cols */
#define A1680_NNZ 5  /* # of stored non-zeros */
static oski_index_t A1680_ptr[A1680_M+1] = {0, 0, 0, 1, 2, 5}; /* end row pointers for A1680 */
static oski_index_t A1680_ind[A1680_NNZ] = {1, 1, 1, 2, 3}; /* end column indices pointers for A1680 */
static oski_value_t A1680_val[A1680_NNZ] = {MAKE_VAL_COMPLEX(-5, 7), MAKE_VAL_COMPLEX(-10, 10), MAKE_VAL_COMPLEX(-8, 0), MAKE_VAL_COMPLEX(-7, 0), MAKE_VAL_COMPLEX(0, -8)}; /* end stored non-zero values for A1680 */

#define x1680_LEN 5   /* vector length */
#define x1680_VW  3   /* vector width */
static oski_value_t x1680_val[x1680_LEN*x1680_VW] = {MAKE_VAL_COMPLEX(-3, 10), MAKE_VAL_COMPLEX(-10, -3), MAKE_VAL_COMPLEX(8, 1), MAKE_VAL_COMPLEX(8, 0), MAKE_VAL_COMPLEX(-5, 0), MAKE_VAL_COMPLEX(-10, -1), MAKE_VAL_COMPLEX(-9, -10), MAKE_VAL_COMPLEX(10, 4), MAKE_VAL_COMPLEX(2, 9), MAKE_VAL_COMPLEX(4, -5), MAKE_VAL_COMPLEX(-7, -3), MAKE_VAL_COMPLEX(-1, -9), MAKE_VAL_COMPLEX(1, -3), MAKE_VAL_COMPLEX(6, 1), MAKE_VAL_COMPLEX(-7, 4)}; /* data */

#define y1680_LEN 5   /* vector length */
#define y1680_VW  3   /* vector width */
static oski_value_t y1680_val[y1680_LEN*y1680_VW] = {MAKE_VAL_COMPLEX(0, 3), MAKE_VAL_COMPLEX(7, 6), MAKE_VAL_COMPLEX(-10, -6), MAKE_VAL_COMPLEX(-0, -2), MAKE_VAL_COMPLEX(3, -3), MAKE_VAL_COMPLEX(-10, 3), MAKE_VAL_COMPLEX(-1, 2), MAKE_VAL_COMPLEX(7, -10), MAKE_VAL_COMPLEX(5, 10), MAKE_VAL_COMPLEX(10, -2), MAKE_VAL_COMPLEX(1, 4), MAKE_VAL_COMPLEX(5, 6), MAKE_VAL_COMPLEX(-1, -10), MAKE_VAL_COMPLEX(9, 1), MAKE_VAL_COMPLEX(-9, 3)}; /* data */

/*
 *  BEGIN TEST 1680
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1680( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	oski_value_t* y_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-6, -5), MAKE_VAL_COMPLEX(-51, -26), MAKE_VAL_COMPLEX(64, 21), MAKE_VAL_COMPLEX(-113, 157), MAKE_VAL_COMPLEX(-47, 20), MAKE_VAL_COMPLEX(66, -17), MAKE_VAL_COMPLEX(-53, 56), MAKE_VAL_COMPLEX(-32, 19), MAKE_VAL_COMPLEX(73, -129), MAKE_VAL_COMPLEX(-100, 103), MAKE_VAL_COMPLEX(26, -24), MAKE_VAL_COMPLEX(46, -180), MAKE_VAL_COMPLEX(-25, 84), MAKE_VAL_COMPLEX(-94, 33), MAKE_VAL_COMPLEX(29, -67) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1680, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1680", y_true, y_val, y1680_LEN*y1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1680() */
/* END TEST 1680 */

/*
 *  BEGIN TEST 1681
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1681( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	oski_value_t* y_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-9, -16), MAKE_VAL_COMPLEX(-28, 19), MAKE_VAL_COMPLEX(36, -29), MAKE_VAL_COMPLEX(123, -143), MAKE_VAL_COMPLEX(41, 16), MAKE_VAL_COMPLEX(-11, -54), MAKE_VAL_COMPLEX(50, -75), MAKE_VAL_COMPLEX(33, 86), MAKE_VAL_COMPLEX(-158, 84), MAKE_VAL_COMPLEX(40, -39), MAKE_VAL_COMPLEX(-53, 1), MAKE_VAL_COMPLEX(-111, 163), MAKE_VAL_COMPLEX(82, -19), MAKE_VAL_COMPLEX(26, 5), MAKE_VAL_COMPLEX(19, 1) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1681, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1681", y_true, y_val, y1680_LEN*y1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1681() */
/* END TEST 1681 */

/*
 *  BEGIN TEST 1682
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1682( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	oski_value_t* y_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(18, -3), MAKE_VAL_COMPLEX(29, -48), MAKE_VAL_COMPLEX(-26, 66), MAKE_VAL_COMPLEX(-12, 2), MAKE_VAL_COMPLEX(-21, -15), MAKE_VAL_COMPLEX(28, 57), MAKE_VAL_COMPLEX(13, 4), MAKE_VAL_COMPLEX(-67, -32), MAKE_VAL_COMPLEX(55, -40), MAKE_VAL_COMPLEX(-22, -58), MAKE_VAL_COMPLEX(23, -10), MAKE_VAL_COMPLEX(31, -36), MAKE_VAL_COMPLEX(-59, 16), MAKE_VAL_COMPLEX(-3, -55), MAKE_VAL_COMPLEX(27, 51) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1682, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1682", y_true, y_val, y1680_LEN*y1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1682() */
/* END TEST 1682 */

/*
 *  BEGIN TEST 1683
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1683( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	oski_value_t* y_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(1, 55), MAKE_VAL_COMPLEX(-91, -75), MAKE_VAL_COMPLEX(93, 75), MAKE_VAL_COMPLEX(-666, 1156), MAKE_VAL_COMPLEX(-273, 55), MAKE_VAL_COMPLEX(213, 59), MAKE_VAL_COMPLEX(-311, 517), MAKE_VAL_COMPLEX(-169, -227), MAKE_VAL_COMPLEX(682, -776), MAKE_VAL_COMPLEX(-349, 605), MAKE_VAL_COMPLEX(250, -98), MAKE_VAL_COMPLEX(378, -1218), MAKE_VAL_COMPLEX(-279, 349), MAKE_VAL_COMPLEX(-357, 209), MAKE_VAL_COMPLEX(-68, -270) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(7, -1);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1683, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1683", y_true, y_val, y1680_LEN*y1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1683() */
/* END TEST 1683 */

/*
 *  BEGIN TEST 1684
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1684( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(29, 88), MAKE_VAL_COMPLEX(-78, 35), MAKE_VAL_COMPLEX(51, -39), MAKE_VAL_COMPLEX(-25, 141), MAKE_VAL_COMPLEX(-119, 83), MAKE_VAL_COMPLEX(-218, -3), MAKE_VAL_COMPLEX(-121, 57), MAKE_VAL_COMPLEX(-2, -45), MAKE_VAL_COMPLEX(198, 97), MAKE_VAL_COMPLEX(-87, -29), MAKE_VAL_COMPLEX(-115, -54), MAKE_VAL_COMPLEX(107, 90), MAKE_VAL_COMPLEX(14, 46), MAKE_VAL_COMPLEX(-49, 80), MAKE_VAL_COMPLEX(-41, 12) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1684, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1684", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1684() */
/* END TEST 1684 */

/*
 *  BEGIN TEST 1685
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1685( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(41, -67), MAKE_VAL_COMPLEX(57, 35), MAKE_VAL_COMPLEX(-44, -17), MAKE_VAL_COMPLEX(25, -197), MAKE_VAL_COMPLEX(119, -48), MAKE_VAL_COMPLEX(211, 73), MAKE_VAL_COMPLEX(51, 6), MAKE_VAL_COMPLEX(30, -25), MAKE_VAL_COMPLEX(-135, -111), MAKE_VAL_COMPLEX(52, 1), MAKE_VAL_COMPLEX(94, 103), MAKE_VAL_COMPLEX(-170, -83), MAKE_VAL_COMPLEX(-35, -53), MAKE_VAL_COMPLEX(56, -122), MAKE_VAL_COMPLEX(69, 37) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1685, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1685", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1685() */
/* END TEST 1685 */

/*
 *  BEGIN TEST 1686
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1686( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(27, -90), MAKE_VAL_COMPLEX(90, 27), MAKE_VAL_COMPLEX(-72, -9), MAKE_VAL_COMPLEX(-72, 0), MAKE_VAL_COMPLEX(45, 0), MAKE_VAL_COMPLEX(90, 9), MAKE_VAL_COMPLEX(81, 90), MAKE_VAL_COMPLEX(-90, -36), MAKE_VAL_COMPLEX(-18, -81), MAKE_VAL_COMPLEX(-36, 45), MAKE_VAL_COMPLEX(63, 27), MAKE_VAL_COMPLEX(9, 81), MAKE_VAL_COMPLEX(-9, 27), MAKE_VAL_COMPLEX(-54, -9), MAKE_VAL_COMPLEX(63, -36) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-9, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1686, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1686", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1686() */
/* END TEST 1686 */

/*
 *  BEGIN TEST 1687
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1687( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-69, -30), MAKE_VAL_COMPLEX(62, -145), MAKE_VAL_COMPLEX(-41, 148), MAKE_VAL_COMPLEX(1348, 458), MAKE_VAL_COMPLEX(490, 485), MAKE_VAL_COMPLEX(-170, 957), MAKE_VAL_COMPLEX(672, -107), MAKE_VAL_COMPLEX(-105, 757), MAKE_VAL_COMPLEX(88, -999), MAKE_VAL_COMPLEX(317, 640), MAKE_VAL_COMPLEX(-398, 362), MAKE_VAL_COMPLEX(958, -1270), MAKE_VAL_COMPLEX(509, -217), MAKE_VAL_COMPLEX(808, 623), MAKE_VAL_COMPLEX(-381, 72) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(0, 9);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1687, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1687", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1687() */
/* END TEST 1687 */

/*
 *  BEGIN TEST 1688
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1688( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-19, 30), MAKE_VAL_COMPLEX(-20, -13), MAKE_VAL_COMPLEX(13, 5), MAKE_VAL_COMPLEX(-69, 3), MAKE_VAL_COMPLEX(-159, -65), MAKE_VAL_COMPLEX(88, -229), MAKE_VAL_COMPLEX(3, 43), MAKE_VAL_COMPLEX(-24, 41), MAKE_VAL_COMPLEX(94, -67), MAKE_VAL_COMPLEX(127, -1), MAKE_VAL_COMPLEX(-25, 120), MAKE_VAL_COMPLEX(57, -224), MAKE_VAL_COMPLEX(-4, -96), MAKE_VAL_COMPLEX(-15, 96), MAKE_VAL_COMPLEX(59, -126) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1688, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1688", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1688() */
/* END TEST 1688 */

/*
 *  BEGIN TEST 1689
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1689( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-101, 43), MAKE_VAL_COMPLEX(-53, -107), MAKE_VAL_COMPLEX(58, 77), MAKE_VAL_COMPLEX(149, 69), MAKE_VAL_COMPLEX(109, 20), MAKE_VAL_COMPLEX(-179, 129), MAKE_VAL_COMPLEX(-3, -224), MAKE_VAL_COMPLEX(88, 89), MAKE_VAL_COMPLEX(-155, 175), MAKE_VAL_COMPLEX(-42, -13), MAKE_VAL_COMPLEX(-18, -213), MAKE_VAL_COMPLEX(14, 125), MAKE_VAL_COMPLEX(41, 75), MAKE_VAL_COMPLEX(66, -32), MAKE_VAL_COMPLEX(-165, 103) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, 8);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1689, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1689", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1689() */
/* END TEST 1689 */

/*
 *  BEGIN TEST 1690
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1690( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-1, 76), MAKE_VAL_COMPLEX(-76, -1), MAKE_VAL_COMPLEX(58, -9), MAKE_VAL_COMPLEX(56, -16), MAKE_VAL_COMPLEX(-35, 10), MAKE_VAL_COMPLEX(-72, 13), MAKE_VAL_COMPLEX(-83, -52), MAKE_VAL_COMPLEX(78, 8), MAKE_VAL_COMPLEX(32, 59), MAKE_VAL_COMPLEX(18, -43), MAKE_VAL_COMPLEX(-55, -7), MAKE_VAL_COMPLEX(-25, -61), MAKE_VAL_COMPLEX(1, -23), MAKE_VAL_COMPLEX(44, -5), MAKE_VAL_COMPLEX(-41, 42) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(7, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1690, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1690", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1690() */
/* END TEST 1690 */

/*
 *  BEGIN TEST 1691
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1691( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-7, -71), MAKE_VAL_COMPLEX(77, -85), MAKE_VAL_COMPLEX(-50, 105), MAKE_VAL_COMPLEX(209, 605), MAKE_VAL_COMPLEX(97, 1024), MAKE_VAL_COMPLEX(-1593, -89), MAKE_VAL_COMPLEX(517, -352), MAKE_VAL_COMPLEX(198, 263), MAKE_VAL_COMPLEX(-913, -331), MAKE_VAL_COMPLEX(-270, -649), MAKE_VAL_COMPLEX(884, -379), MAKE_VAL_COMPLEX(-1288, 323), MAKE_VAL_COMPLEX(-491, 343), MAKE_VAL_COMPLEX(584, -50), MAKE_VAL_COMPLEX(-1019, -159) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-3, -6);
	oski_value_t beta = MAKE_VAL_COMPLEX(-5, 4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1691, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1691", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1691() */
/* END TEST 1691 */

/*
 *  BEGIN TEST 1692
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1692( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	oski_value_t* y_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-3, 28), MAKE_VAL_COMPLEX(-73, 183), MAKE_VAL_COMPLEX(-116, 31), MAKE_VAL_COMPLEX(-52, 71), MAKE_VAL_COMPLEX(-22, 17), MAKE_VAL_COMPLEX(-70, 15), MAKE_VAL_COMPLEX(-41, 14), MAKE_VAL_COMPLEX(35, -98), MAKE_VAL_COMPLEX(65, 55), MAKE_VAL_COMPLEX(12, 17), MAKE_VAL_COMPLEX(14, 25), MAKE_VAL_COMPLEX(49, 44), MAKE_VAL_COMPLEX(102, -144), MAKE_VAL_COMPLEX(131, -149), MAKE_VAL_COMPLEX(-67, -25) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1692, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1692", y_true, y_val, y1680_LEN*y1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1692() */
/* END TEST 1692 */

/*
 *  BEGIN TEST 1693
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1693( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	oski_value_t* y_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-15, -34), MAKE_VAL_COMPLEX(23, -153), MAKE_VAL_COMPLEX(172, -79), MAKE_VAL_COMPLEX(64, -67), MAKE_VAL_COMPLEX(34, 7), MAKE_VAL_COMPLEX(72, -81), MAKE_VAL_COMPLEX(31, -24), MAKE_VAL_COMPLEX(11, 160), MAKE_VAL_COMPLEX(-135, -45), MAKE_VAL_COMPLEX(-20, 47), MAKE_VAL_COMPLEX(-40, -27), MAKE_VAL_COMPLEX(-95, -26), MAKE_VAL_COMPLEX(-40, 158), MAKE_VAL_COMPLEX(-155, 201), MAKE_VAL_COMPLEX(67, -35) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 6);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1693, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1693", y_true, y_val, y1680_LEN*y1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1693() */
/* END TEST 1693 */

/*
 *  BEGIN TEST 1694
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1694( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	oski_value_t* y_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(0, 15), MAKE_VAL_COMPLEX(35, 30), MAKE_VAL_COMPLEX(-50, -30), MAKE_VAL_COMPLEX(0, -10), MAKE_VAL_COMPLEX(15, -15), MAKE_VAL_COMPLEX(-50, 15), MAKE_VAL_COMPLEX(-5, 10), MAKE_VAL_COMPLEX(35, -50), MAKE_VAL_COMPLEX(25, 50), MAKE_VAL_COMPLEX(50, -10), MAKE_VAL_COMPLEX(5, 20), MAKE_VAL_COMPLEX(25, 30), MAKE_VAL_COMPLEX(-5, -50), MAKE_VAL_COMPLEX(45, 5), MAKE_VAL_COMPLEX(-45, 15) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 0);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1694, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1694", y_true, y_val, y1680_LEN*y1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1694() */
/* END TEST 1694 */

/*
 *  BEGIN TEST 1695
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1695( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	oski_value_t* y_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(6, 93), MAKE_VAL_COMPLEX(-602, 1649), MAKE_VAL_COMPLEX(-289, 793), MAKE_VAL_COMPLEX(-221, 907), MAKE_VAL_COMPLEX(-264, 438), MAKE_VAL_COMPLEX(-76, 7), MAKE_VAL_COMPLEX(-305, 120), MAKE_VAL_COMPLEX(-201, -308), MAKE_VAL_COMPLEX(300, -175), MAKE_VAL_COMPLEX(-367, 399), MAKE_VAL_COMPLEX(77, -24), MAKE_VAL_COMPLEX(191, -2), MAKE_VAL_COMPLEX(712, -1059), MAKE_VAL_COMPLEX(211, -1637), MAKE_VAL_COMPLEX(-225, -345) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(9, -3);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1695, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1695", y_true, y_val, y1680_LEN*y1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1695() */
/* END TEST 1695 */

/*
 *  BEGIN TEST 1696
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1696( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(62, -22), MAKE_VAL_COMPLEX(15, 65), MAKE_VAL_COMPLEX(-26, -40), MAKE_VAL_COMPLEX(63, -30), MAKE_VAL_COMPLEX(-176, 92), MAKE_VAL_COMPLEX(87, 159), MAKE_VAL_COMPLEX(-38, 21), MAKE_VAL_COMPLEX(-64, -65), MAKE_VAL_COMPLEX(116, -142), MAKE_VAL_COMPLEX(-75, -156), MAKE_VAL_COMPLEX(32, -33), MAKE_VAL_COMPLEX(2, 4), MAKE_VAL_COMPLEX(14, -2), MAKE_VAL_COMPLEX(-130, 58), MAKE_VAL_COMPLEX(-2, 116) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1696, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1696", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1696() */
/* END TEST 1696 */

/*
 *  BEGIN TEST 1697
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1697( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(98, -39), MAKE_VAL_COMPLEX(46, 95), MAKE_VAL_COMPLEX(-41, -74), MAKE_VAL_COMPLEX(-143, -74), MAKE_VAL_COMPLEX(226, -27), MAKE_VAL_COMPLEX(0, -19), MAKE_VAL_COMPLEX(-2, 196), MAKE_VAL_COMPLEX(16, -105), MAKE_VAL_COMPLEX(-19, 26), MAKE_VAL_COMPLEX(-30, 154), MAKE_VAL_COMPLEX(-1, 154), MAKE_VAL_COMPLEX(-109, 99), MAKE_VAL_COMPLEX(-63, 19), MAKE_VAL_COMPLEX(83, -146), MAKE_VAL_COMPLEX(124, -65) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1697, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1697", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1697() */
/* END TEST 1697 */

/*
 *  BEGIN TEST 1698
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1698( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(25, 62), MAKE_VAL_COMPLEX(-62, 25), MAKE_VAL_COMPLEX(44, -27), MAKE_VAL_COMPLEX(40, -32), MAKE_VAL_COMPLEX(-25, 20), MAKE_VAL_COMPLEX(-54, 35), MAKE_VAL_COMPLEX(-85, -14), MAKE_VAL_COMPLEX(66, -20), MAKE_VAL_COMPLEX(46, 37), MAKE_VAL_COMPLEX(0, -41), MAKE_VAL_COMPLEX(-47, 13), MAKE_VAL_COMPLEX(-41, -41), MAKE_VAL_COMPLEX(-7, -19), MAKE_VAL_COMPLEX(34, -19), MAKE_VAL_COMPLEX(-19, 48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, -4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1698, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1698", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1698() */
/* END TEST 1698 */

/*
 *  BEGIN TEST 1699
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1699( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(42, -62), MAKE_VAL_COMPLEX(29, 23), MAKE_VAL_COMPLEX(-74, -16), MAKE_VAL_COMPLEX(181, 386), MAKE_VAL_COMPLEX(-816, -568), MAKE_VAL_COMPLEX(-187, 561), MAKE_VAL_COMPLEX(226, -181), MAKE_VAL_COMPLEX(-204, -235), MAKE_VAL_COMPLEX(632, -50), MAKE_VAL_COMPLEX(475, -576), MAKE_VAL_COMPLEX(424, -119), MAKE_VAL_COMPLEX(258, 136), MAKE_VAL_COMPLEX(118, 126), MAKE_VAL_COMPLEX(-746, -194), MAKE_VAL_COMPLEX(-470, 80) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(3, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -3);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1699, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1699", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1699() */
/* END TEST 1699 */

/*
 *  BEGIN TEST 1700
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1700( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(59, -12), MAKE_VAL_COMPLEX(5, 62), MAKE_VAL_COMPLEX(-18, -39), MAKE_VAL_COMPLEX(-53, 110), MAKE_VAL_COMPLEX(-121, -106), MAKE_VAL_COMPLEX(-43, -8), MAKE_VAL_COMPLEX(-131, 109), MAKE_VAL_COMPLEX(-82, -75), MAKE_VAL_COMPLEX(34, -63), MAKE_VAL_COMPLEX(-143, 27), MAKE_VAL_COMPLEX(17, 104), MAKE_VAL_COMPLEX(-167, -49), MAKE_VAL_COMPLEX(-17, -5), MAKE_VAL_COMPLEX(36, -21), MAKE_VAL_COMPLEX(7, -24) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1700, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1700", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1700() */
/* END TEST 1700 */

/*
 *  BEGIN TEST 1701
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1701( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(49, -21), MAKE_VAL_COMPLEX(28, 46), MAKE_VAL_COMPLEX(-21, -39), MAKE_VAL_COMPLEX(5, -182), MAKE_VAL_COMPLEX(151, 151), MAKE_VAL_COMPLEX(94, 104), MAKE_VAL_COMPLEX(95, 32), MAKE_VAL_COMPLEX(58, -39), MAKE_VAL_COMPLEX(35, -9), MAKE_VAL_COMPLEX(74, -33), MAKE_VAL_COMPLEX(-2, -23), MAKE_VAL_COMPLEX(92, 112), MAKE_VAL_COMPLEX(-16, 14), MAKE_VAL_COMPLEX(-63, -39), MAKE_VAL_COMPLEX(71, 63) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1701, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1701", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1701() */
/* END TEST 1701 */

/*
 *  BEGIN TEST 1702
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1702( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(82, -19), MAKE_VAL_COMPLEX(19, 82), MAKE_VAL_COMPLEX(-25, -60), MAKE_VAL_COMPLEX(-32, -56), MAKE_VAL_COMPLEX(20, 35), MAKE_VAL_COMPLEX(33, 74), MAKE_VAL_COMPLEX(-34, 103), MAKE_VAL_COMPLEX(-12, -86), MAKE_VAL_COMPLEX(55, -50), MAKE_VAL_COMPLEX(-51, -8), MAKE_VAL_COMPLEX(7, 61), MAKE_VAL_COMPLEX(-59, 43), MAKE_VAL_COMPLEX(-25, 5), MAKE_VAL_COMPLEX(-17, -46), MAKE_VAL_COMPLEX(56, 33) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1702, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1702", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1702() */
/* END TEST 1702 */

/*
 *  BEGIN TEST 1703
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_ROWMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1703( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_ROWMAJ, x1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(100, 14), MAKE_VAL_COMPLEX(6, 166), MAKE_VAL_COMPLEX(36, -62), MAKE_VAL_COMPLEX(1313, 355), MAKE_VAL_COMPLEX(-1310, 1128), MAKE_VAL_COMPLEX(-614, 622), MAKE_VAL_COMPLEX(137, 1079), MAKE_VAL_COMPLEX(-171, 561), MAKE_VAL_COMPLEX(-178, -4), MAKE_VAL_COMPLEX(143, 963), MAKE_VAL_COMPLEX(537, 13), MAKE_VAL_COMPLEX(-916, 1060), MAKE_VAL_COMPLEX(-102, -22), MAKE_VAL_COMPLEX(158, -472), MAKE_VAL_COMPLEX(-422, 304) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, -9);
	oski_value_t beta = MAKE_VAL_COMPLEX(-1, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1703, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1703", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1703() */
/* END TEST 1703 */

/*
 *  BEGIN TEST 1704
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1704( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	oski_value_t* y_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-24, 22), MAKE_VAL_COMPLEX(-24, 72), MAKE_VAL_COMPLEX(-5, -97), MAKE_VAL_COMPLEX(-69, -152), MAKE_VAL_COMPLEX(40, -161), MAKE_VAL_COMPLEX(-82, -161), MAKE_VAL_COMPLEX(96, -53), MAKE_VAL_COMPLEX(195, 35), MAKE_VAL_COMPLEX(68, 82), MAKE_VAL_COMPLEX(192, -48), MAKE_VAL_COMPLEX(208, 74), MAKE_VAL_COMPLEX(52, 84), MAKE_VAL_COMPLEX(85, -94), MAKE_VAL_COMPLEX(107, 98), MAKE_VAL_COMPLEX(-55, -2) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, 7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1704, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1704", y_true, y_val, y1680_LEN*y1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1704() */
/* END TEST 1704 */

/*
 *  BEGIN TEST 1705
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1705( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	oski_value_t* y_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(24, -34), MAKE_VAL_COMPLEX(-4, -96), MAKE_VAL_COMPLEX(45, 121), MAKE_VAL_COMPLEX(69, 160), MAKE_VAL_COMPLEX(-52, 173), MAKE_VAL_COMPLEX(122, 149), MAKE_VAL_COMPLEX(-92, 45), MAKE_VAL_COMPLEX(-223, 5), MAKE_VAL_COMPLEX(-88, -122), MAKE_VAL_COMPLEX(-232, 56), MAKE_VAL_COMPLEX(-212, -90), MAKE_VAL_COMPLEX(-72, -108), MAKE_VAL_COMPLEX(-81, 134), MAKE_VAL_COMPLEX(-143, -102), MAKE_VAL_COMPLEX(91, -10) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -7);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1705, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1705", y_true, y_val, y1680_LEN*y1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1705() */
/* END TEST 1705 */

/*
 *  BEGIN TEST 1706
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1706( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	oski_value_t* y_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(15, -9), MAKE_VAL_COMPLEX(9, -53), MAKE_VAL_COMPLEX(0, 68), MAKE_VAL_COMPLEX(-10, 6), MAKE_VAL_COMPLEX(-24, -6), MAKE_VAL_COMPLEX(45, 41), MAKE_VAL_COMPLEX(13, -1), MAKE_VAL_COMPLEX(-71, -5), MAKE_VAL_COMPLEX(35, -55), MAKE_VAL_COMPLEX(-40, -44), MAKE_VAL_COMPLEX(17, -17), MAKE_VAL_COMPLEX(15, -43), MAKE_VAL_COMPLEX(-47, 35), MAKE_VAL_COMPLEX(-22, -48), MAKE_VAL_COMPLEX(42, 36) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -5);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1706, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1706", y_true, y_val, y1680_LEN*y1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1706() */
/* END TEST 1706 */

/*
 *  BEGIN TEST 1707
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1707( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	oski_value_t* y_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(89, -40), MAKE_VAL_COMPLEX(52, 127), MAKE_VAL_COMPLEX(29, 25), MAKE_VAL_COMPLEX(-563, 1668), MAKE_VAL_COMPLEX(-1391, 1140), MAKE_VAL_COMPLEX(-709, 855), MAKE_VAL_COMPLEX(-1240, -534), MAKE_VAL_COMPLEX(-417, -950), MAKE_VAL_COMPLEX(-800, -953), MAKE_VAL_COMPLEX(-1818, -298), MAKE_VAL_COMPLEX(-1230, -2201), MAKE_VAL_COMPLEX(-332, -735), MAKE_VAL_COMPLEX(-471, 153), MAKE_VAL_COMPLEX(-284, -803), MAKE_VAL_COMPLEX(348, -389) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-7, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1707, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1707", y_true, y_val, y1680_LEN*y1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1707() */
/* END TEST 1707 */

/*
 *  BEGIN TEST 1708
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1708( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-58, 51), MAKE_VAL_COMPLEX(-129, 123), MAKE_VAL_COMPLEX(36, 120), MAKE_VAL_COMPLEX(-42, 58), MAKE_VAL_COMPLEX(-8, 52), MAKE_VAL_COMPLEX(-49, -40), MAKE_VAL_COMPLEX(-98, -38), MAKE_VAL_COMPLEX(-48, 41), MAKE_VAL_COMPLEX(-75, -6), MAKE_VAL_COMPLEX(-52, 89), MAKE_VAL_COMPLEX(-40, -52), MAKE_VAL_COMPLEX(-113, -104), MAKE_VAL_COMPLEX(157, 30), MAKE_VAL_COMPLEX(191, 178), MAKE_VAL_COMPLEX(-70, -55) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, 4);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1708, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1708", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1708() */
/* END TEST 1708 */

/*
 *  BEGIN TEST 1709
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1709( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-14, 80), MAKE_VAL_COMPLEX(-2, -195), MAKE_VAL_COMPLEX(73, -82), MAKE_VAL_COMPLEX(154, -34), MAKE_VAL_COMPLEX(-62, -67), MAKE_VAL_COMPLEX(-88, -4), MAKE_VAL_COMPLEX(2, -129), MAKE_VAL_COMPLEX(176, 45), MAKE_VAL_COMPLEX(76, 138), MAKE_VAL_COMPLEX(123, -147), MAKE_VAL_COMPLEX(-49, -11), MAKE_VAL_COMPLEX(126, -25), MAKE_VAL_COMPLEX(-134, -69), MAKE_VAL_COMPLEX(-110, -146), MAKE_VAL_COMPLEX(-40, 90) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1709, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1709", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1709() */
/* END TEST 1709 */

/*
 *  BEGIN TEST 1710
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1710( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(86, 4), MAKE_VAL_COMPLEX(-4, 86), MAKE_VAL_COMPLEX(-8, -66), MAKE_VAL_COMPLEX(-16, -64), MAKE_VAL_COMPLEX(10, 40), MAKE_VAL_COMPLEX(12, 82), MAKE_VAL_COMPLEX(-62, 92), MAKE_VAL_COMPLEX(12, -88), MAKE_VAL_COMPLEX(68, -34), MAKE_VAL_COMPLEX(-48, -22), MAKE_VAL_COMPLEX(-10, 62), MAKE_VAL_COMPLEX(-70, 26), MAKE_VAL_COMPLEX(-26, -2), MAKE_VAL_COMPLEX(-4, -50), MAKE_VAL_COMPLEX(46, 48) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-2, -8);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1710, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1710", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1710() */
/* END TEST 1710 */

/*
 *  BEGIN TEST 1711
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1711( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(48, -56), MAKE_VAL_COMPLEX(880, 2120), MAKE_VAL_COMPLEX(528, 712), MAKE_VAL_COMPLEX(-576, 928), MAKE_VAL_COMPLEX(792, 400), MAKE_VAL_COMPLEX(184, 0), MAKE_VAL_COMPLEX(-136, 1216), MAKE_VAL_COMPLEX(-1000, 520), MAKE_VAL_COMPLEX(-968, -208), MAKE_VAL_COMPLEX(24, 1632), MAKE_VAL_COMPLEX(-72, 56), MAKE_VAL_COMPLEX(-1504, 848), MAKE_VAL_COMPLEX(1456, -736), MAKE_VAL_COMPLEX(2408, -96), MAKE_VAL_COMPLEX(-448, -344) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(8, -8);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1711, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1711", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1711() */
/* END TEST 1711 */

/*
 *  BEGIN TEST 1712
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1712( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-75, 35), MAKE_VAL_COMPLEX(-125, -80), MAKE_VAL_COMPLEX(54, 135), MAKE_VAL_COMPLEX(150, 58), MAKE_VAL_COMPLEX(-35, -118), MAKE_VAL_COMPLEX(-37, -59), MAKE_VAL_COMPLEX(-129, -164), MAKE_VAL_COMPLEX(-24, 99), MAKE_VAL_COMPLEX(-51, 193), MAKE_VAL_COMPLEX(18, 86), MAKE_VAL_COMPLEX(-27, -63), MAKE_VAL_COMPLEX(166, -267), MAKE_VAL_COMPLEX(120, -105), MAKE_VAL_COMPLEX(75, -155), MAKE_VAL_COMPLEX(25, -153) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(5, 6);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1712, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1712", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1712() */
/* END TEST 1712 */

/*
 *  BEGIN TEST 1713
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1713( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-19, 24), MAKE_VAL_COMPLEX(66, -14), MAKE_VAL_COMPLEX(3, -71), MAKE_VAL_COMPLEX(-86, -2), MAKE_VAL_COMPLEX(-5, 83), MAKE_VAL_COMPLEX(-36, -19), MAKE_VAL_COMPLEX(127, 21), MAKE_VAL_COMPLEX(76, 3), MAKE_VAL_COMPLEX(4, -107), MAKE_VAL_COMPLEX(49, -98), MAKE_VAL_COMPLEX(-8, -10), MAKE_VAL_COMPLEX(-111, 188), MAKE_VAL_COMPLEX(-91, 88), MAKE_VAL_COMPLEX(-34, 205), MAKE_VAL_COMPLEX(-109, 136) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, 1);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1713, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1713", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1713() */
/* END TEST 1713 */

/*
 *  BEGIN TEST 1714
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1714( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(74, -65), MAKE_VAL_COMPLEX(65, 74), MAKE_VAL_COMPLEX(-59, -48), MAKE_VAL_COMPLEX(-64, -40), MAKE_VAL_COMPLEX(40, 25), MAKE_VAL_COMPLEX(75, 58), MAKE_VAL_COMPLEX(22, 125), MAKE_VAL_COMPLEX(-60, -82), MAKE_VAL_COMPLEX(29, -82), MAKE_VAL_COMPLEX(-57, 20), MAKE_VAL_COMPLEX(41, 59), MAKE_VAL_COMPLEX(-37, 77), MAKE_VAL_COMPLEX(-23, 19), MAKE_VAL_COMPLEX(-43, -38), MAKE_VAL_COMPLEX(76, 3) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-8, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1714, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1714", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1714() */
/* END TEST 1714 */

/*
 *  BEGIN TEST 1715
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_ROWMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1715( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_ROWMAJ, y1680_VW );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(13, -70), MAKE_VAL_COMPLEX(755, 683), MAKE_VAL_COMPLEX(380, -784), MAKE_VAL_COMPLEX(-842, -834), MAKE_VAL_COMPLEX(-516, 764), MAKE_VAL_COMPLEX(28, -113), MAKE_VAL_COMPLEX(788, 1510), MAKE_VAL_COMPLEX(485, 202), MAKE_VAL_COMPLEX(982, -1071), MAKE_VAL_COMPLEX(859, -444), MAKE_VAL_COMPLEX(72, 116), MAKE_VAL_COMPLEX(-2426, 943), MAKE_VAL_COMPLEX(-1446, 103), MAKE_VAL_COMPLEX(-1806, 1219), MAKE_VAL_COMPLEX(-1569, 430) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, -7);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, 2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1715, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1715", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1715() */
/* END TEST 1715 */

/*
 *  BEGIN TEST 1716
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=1, beta=general>
 */
static void
test_MatMult_1716( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	oski_value_t* y_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(3, -2), MAKE_VAL_COMPLEX(-99, -182), MAKE_VAL_COMPLEX(142, -10), MAKE_VAL_COMPLEX(134, -102), MAKE_VAL_COMPLEX(1, -41), MAKE_VAL_COMPLEX(36, 7), MAKE_VAL_COMPLEX(15, -176), MAKE_VAL_COMPLEX(49, 52), MAKE_VAL_COMPLEX(232, 1), MAKE_VAL_COMPLEX(34, 19), MAKE_VAL_COMPLEX(-3, -21), MAKE_VAL_COMPLEX(-29, -137), MAKE_VAL_COMPLEX(102, 49), MAKE_VAL_COMPLEX(40, 3), MAKE_VAL_COMPLEX(44, 55) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-4, -2);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1716, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1716", y_true, y_val, y1680_LEN*y1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1716() */
/* END TEST 1716 */

/*
 *  BEGIN TEST 1717
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=-1, beta=general>
 */
static void
test_MatMult_1717( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	oski_value_t* y_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(15, 8), MAKE_VAL_COMPLEX(149, 152), MAKE_VAL_COMPLEX(-198, 58), MAKE_VAL_COMPLEX(-146, 98), MAKE_VAL_COMPLEX(-13, 17), MAKE_VAL_COMPLEX(-38, 59), MAKE_VAL_COMPLEX(-5, 186), MAKE_VAL_COMPLEX(-95, -114), MAKE_VAL_COMPLEX(-162, -11), MAKE_VAL_COMPLEX(-26, -83), MAKE_VAL_COMPLEX(29, 23), MAKE_VAL_COMPLEX(75, 119), MAKE_VAL_COMPLEX(-164, -63), MAKE_VAL_COMPLEX(-16, -55), MAKE_VAL_COMPLEX(-44, 5) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(6, -4);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1717, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1717", y_true, y_val, y1680_LEN*y1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1717() */
/* END TEST 1717 */

/*
 *  BEGIN TEST 1718
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=0, beta=general>
 */
static void
test_MatMult_1718( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	oski_value_t* y_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(9, 24), MAKE_VAL_COMPLEX(74, 27), MAKE_VAL_COMPLEX(-98, -18), MAKE_VAL_COMPLEX(-6, -16), MAKE_VAL_COMPLEX(15, -33), MAKE_VAL_COMPLEX(-71, 54), MAKE_VAL_COMPLEX(-2, 19), MAKE_VAL_COMPLEX(26, -101), MAKE_VAL_COMPLEX(70, 65), MAKE_VAL_COMPLEX(74, -46), MAKE_VAL_COMPLEX(20, 29), MAKE_VAL_COMPLEX(58, 33), MAKE_VAL_COMPLEX(-38, -77), MAKE_VAL_COMPLEX(75, -19), MAKE_VAL_COMPLEX(-63, 51) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, -3);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1718, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1718", y_true, y_val, y1680_LEN*y1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1718() */
/* END TEST 1718 */

/*
 *  BEGIN TEST 1719
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_NORMAL, alpha=general, beta=general>
 */
static void
test_MatMult_1719( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	oski_value_t* y_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(46, 29), MAKE_VAL_COMPLEX(-424, 523), MAKE_VAL_COMPLEX(-426, -446), MAKE_VAL_COMPLEX(-580, -460), MAKE_VAL_COMPLEX(-183, -59), MAKE_VAL_COMPLEX(-81, 78), MAKE_VAL_COMPLEX(-694, 161), MAKE_VAL_COMPLEX(60, -511), MAKE_VAL_COMPLEX(27, -894), MAKE_VAL_COMPLEX(134, -371), MAKE_VAL_COMPLEX(8, 66), MAKE_VAL_COMPLEX(-340, 236), MAKE_VAL_COMPLEX(-109, -568), MAKE_VAL_COMPLEX(108, -321), MAKE_VAL_COMPLEX(116, -21) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, -4);
	oski_value_t beta = MAKE_VAL_COMPLEX(9, -1);
	int err;

	err = oski_MatMult( A, OP_NORMAL, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1719, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1719", y_true, y_val, y1680_LEN*y1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1719() */
/* END TEST 1719 */

/*
 *  BEGIN TEST 1720
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1720( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(31, -64), MAKE_VAL_COMPLEX(162, 41), MAKE_VAL_COMPLEX(-79, -79), MAKE_VAL_COMPLEX(-90, -164), MAKE_VAL_COMPLEX(68, -4), MAKE_VAL_COMPLEX(59, 20), MAKE_VAL_COMPLEX(-143, 146), MAKE_VAL_COMPLEX(-110, -37), MAKE_VAL_COMPLEX(14, -145), MAKE_VAL_COMPLEX(-144, 123), MAKE_VAL_COMPLEX(47, 32), MAKE_VAL_COMPLEX(50, 169), MAKE_VAL_COMPLEX(69, -76), MAKE_VAL_COMPLEX(2, -50), MAKE_VAL_COMPLEX(3, 76) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -1);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1720, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1720", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1720() */
/* END TEST 1720 */

/*
 *  BEGIN TEST 1721
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1721( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-6, 17), MAKE_VAL_COMPLEX(-115, -16), MAKE_VAL_COMPLEX(40, 66), MAKE_VAL_COMPLEX(50, 156), MAKE_VAL_COMPLEX(-43, 9), MAKE_VAL_COMPLEX(-10, -5), MAKE_VAL_COMPLEX(178, -87), MAKE_VAL_COMPLEX(64, 7), MAKE_VAL_COMPLEX(-15, 98), MAKE_VAL_COMPLEX(119, -102), MAKE_VAL_COMPLEX(-15, -10), MAKE_VAL_COMPLEX(-54, -123), MAKE_VAL_COMPLEX(-77, 90), MAKE_VAL_COMPLEX(-31, 39), MAKE_VAL_COMPLEX(36, -89) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(2, 0);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1721, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1721", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1721() */
/* END TEST 1721 */

/*
 *  BEGIN TEST 1722
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1722( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-32, -75), MAKE_VAL_COMPLEX(75, -32), MAKE_VAL_COMPLEX(-53, 34), MAKE_VAL_COMPLEX(-48, 40), MAKE_VAL_COMPLEX(30, -25), MAKE_VAL_COMPLEX(65, -44), MAKE_VAL_COMPLEX(104, 15), MAKE_VAL_COMPLEX(-80, 26), MAKE_VAL_COMPLEX(-57, -44), MAKE_VAL_COMPLEX(1, 50), MAKE_VAL_COMPLEX(57, -17), MAKE_VAL_COMPLEX(51, 49), MAKE_VAL_COMPLEX(9, 23), MAKE_VAL_COMPLEX(-41, 24), MAKE_VAL_COMPLEX(22, -59) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-6, 5);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1722, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1722", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1722() */
/* END TEST 1722 */

/*
 *  BEGIN TEST 1723
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1723( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(-56, 50), MAKE_VAL_COMPLEX(-874, 256), MAKE_VAL_COMPLEX(510, 440), MAKE_VAL_COMPLEX(960, 1128), MAKE_VAL_COMPLEX(-268, 194), MAKE_VAL_COMPLEX(-10, -92), MAKE_VAL_COMPLEX(1248, -1418), MAKE_VAL_COMPLEX(420, -132), MAKE_VAL_COMPLEX(166, 792), MAKE_VAL_COMPLEX(562, -1212), MAKE_VAL_COMPLEX(-74, -66), MAKE_VAL_COMPLEX(-826, -706), MAKE_VAL_COMPLEX(-234, 1062), MAKE_VAL_COMPLEX(-150, 488), MAKE_VAL_COMPLEX(-52, -958) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-8, 4);
	oski_value_t beta = MAKE_VAL_COMPLEX(8, 2);
	int err;

	err = oski_MatMult( A, OP_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1723, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1723", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1723() */
/* END TEST 1723 */

/*
 *  BEGIN TEST 1724
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=1, beta=general>
 */
static void
test_MatMult_1724( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(29, -21), MAKE_VAL_COMPLEX(-5, 179), MAKE_VAL_COMPLEX(-130, 15), MAKE_VAL_COMPLEX(-130, 16), MAKE_VAL_COMPLEX(16, 1), MAKE_VAL_COMPLEX(18, 26), MAKE_VAL_COMPLEX(-129, -83), MAKE_VAL_COMPLEX(-94, -45), MAKE_VAL_COMPLEX(23, 29), MAKE_VAL_COMPLEX(27, 19), MAKE_VAL_COMPLEX(16, 27), MAKE_VAL_COMPLEX(-83, -32), MAKE_VAL_COMPLEX(-14, -19), MAKE_VAL_COMPLEX(-141, -96), MAKE_VAL_COMPLEX(-5, -45) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-3, -2);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1724, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1724", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1724() */
/* END TEST 1724 */

/*
 *  BEGIN TEST 1725
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=-1, beta=general>
 */
static void
test_MatMult_1725( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(31, 39), MAKE_VAL_COMPLEX(-13, -119), MAKE_VAL_COMPLEX(136, -63), MAKE_VAL_COMPLEX(130, -64), MAKE_VAL_COMPLEX(-16, 29), MAKE_VAL_COMPLEX(-24, 34), MAKE_VAL_COMPLEX(69, 137), MAKE_VAL_COMPLEX(118, -15), MAKE_VAL_COMPLEX(31, -41), MAKE_VAL_COMPLEX(-57, -43), MAKE_VAL_COMPLEX(-34, 15), MAKE_VAL_COMPLEX(29, 38), MAKE_VAL_COMPLEX(-4, 13), MAKE_VAL_COMPLEX(147, 60), MAKE_VAL_COMPLEX(29, 87) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-1, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(3, -4);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1725, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1725", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1725() */
/* END TEST 1725 */

/*
 *  BEGIN TEST 1726
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=0, beta=general>
 */
static void
test_MatMult_1726( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(91, -49), MAKE_VAL_COMPLEX(49, 91), MAKE_VAL_COMPLEX(-49, -63), MAKE_VAL_COMPLEX(-56, -56), MAKE_VAL_COMPLEX(35, 35), MAKE_VAL_COMPLEX(63, 77), MAKE_VAL_COMPLEX(-7, 133), MAKE_VAL_COMPLEX(-42, -98), MAKE_VAL_COMPLEX(49, -77), MAKE_VAL_COMPLEX(-63, 7), MAKE_VAL_COMPLEX(28, 70), MAKE_VAL_COMPLEX(-56, 70), MAKE_VAL_COMPLEX(-28, 14), MAKE_VAL_COMPLEX(-35, -49), MAKE_VAL_COMPLEX(77, 21) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(0, 0);
	oski_value_t beta = MAKE_VAL_COMPLEX(-7, -7);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1726, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1726", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1726() */
/* END TEST 1726 */

/*
 *  BEGIN TEST 1727
 *  Value type: complex
 *  A : <MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED>
 *  x : <'x1680', LAYOUT_COLMAJ>
 *  y : <'y1680', LAYOUT_COLMAJ>
 *  Kernel : <op=OP_CONJ_TRANS, alpha=general, beta=general>
 */
static void
test_MatMult_1727( void )
{
	oski_matrix_t A = oski_CreateMatCSR( A1680_ptr, A1680_ind, A1680_val,
		A1680_M, A1680_N, SHARE_INPUTMAT,
		5, MAT_HERM_LOWER, MAT_UNIT_DIAG_IMPLICIT, INDEX_ZERO_BASED,
		INDEX_SORTED, INDEX_UNIQUE );
	oski_value_t* x_val = array_Duplicate( y1680_val,
		y1680_LEN * y1680_VW );
	oski_vecview_t x = oski_CreateMultiVecView( x_val,
		y1680_LEN, y1680_VW, LAYOUT_COLMAJ, y1680_LEN );
	oski_value_t* y_val = array_Duplicate( x1680_val,
		x1680_LEN * x1680_VW );
	oski_vecview_t y = oski_CreateMultiVecView( y_val,
		x1680_LEN, x1680_VW, LAYOUT_COLMAJ, x1680_LEN );
	const oski_value_t y_true[] = {MAKE_VAL_COMPLEX(17, 37), MAKE_VAL_COMPLEX(-931, -1065), MAKE_VAL_COMPLEX(447, -996), MAKE_VAL_COMPLEX(444, -974), MAKE_VAL_COMPLEX(37, 86), MAKE_VAL_COMPLEX(-6, -42), MAKE_VAL_COMPLEX(1647, -161), MAKE_VAL_COMPLEX(583, -460), MAKE_VAL_COMPLEX(-433, -257), MAKE_VAL_COMPLEX(-387, 231), MAKE_VAL_COMPLEX(-77, 6), MAKE_VAL_COMPLEX(786, -141), MAKE_VAL_COMPLEX(201, 104), MAKE_VAL_COMPLEX(1346, -415), MAKE_VAL_COMPLEX(525, 95) };
	oski_value_t alpha = MAKE_VAL_COMPLEX(-6, 7);
	oski_value_t beta = MAKE_VAL_COMPLEX(4, -5);
	int err;

	err = oski_MatMult( A, OP_CONJ_TRANS, alpha, x, beta, y );
	if( err != 0 ) { OSKI_ERR(test_MatMult_1727, err); exit( 1 ); }

	array_CompareElems( "test_MatMult_1727", y_true, y_val, x1680_LEN*x1680_VW );

	oski_DestroyVecView( y );
	oski_DestroyVecView( x );
	oski_DestroyMat( A );
	array_Destroy( y_val );
	array_Destroy( x_val );
} /* end test_MatMult_1727() */
/* END TEST 1727 */


int
main( int argc, char* argv[] )
{
	oski_Init();
	test_MatMult_0();
	test_MatMult_1();
	test_MatMult_2();
	test_MatMult_3();
	test_MatMult_4();
	test_MatMult_5();
	test_MatMult_6();
	test_MatMult_7();
	test_MatMult_8();
	test_MatMult_9();
	test_MatMult_10();
	test_MatMult_11();
	test_MatMult_12();
	test_MatMult_13();
	test_MatMult_14();
	test_MatMult_15();
	test_MatMult_16();
	test_MatMult_17();
	test_MatMult_18();
	test_MatMult_19();
	test_MatMult_20();
	test_MatMult_21();
	test_MatMult_22();
	test_MatMult_23();
	test_MatMult_24();
	test_MatMult_25();
	test_MatMult_26();
	test_MatMult_27();
	test_MatMult_28();
	test_MatMult_29();
	test_MatMult_30();
	test_MatMult_31();
	test_MatMult_32();
	test_MatMult_33();
	test_MatMult_34();
	test_MatMult_35();
	test_MatMult_36();
	test_MatMult_37();
	test_MatMult_38();
	test_MatMult_39();
	test_MatMult_40();
	test_MatMult_41();
	test_MatMult_42();
	test_MatMult_43();
	test_MatMult_44();
	test_MatMult_45();
	test_MatMult_46();
	test_MatMult_47();
	test_MatMult_48();
	test_MatMult_49();
	test_MatMult_50();
	test_MatMult_51();
	test_MatMult_52();
	test_MatMult_53();
	test_MatMult_54();
	test_MatMult_55();
	test_MatMult_56();
	test_MatMult_57();
	test_MatMult_58();
	test_MatMult_59();
	test_MatMult_60();
	test_MatMult_61();
	test_MatMult_62();
	test_MatMult_63();
	test_MatMult_64();
	test_MatMult_65();
	test_MatMult_66();
	test_MatMult_67();
	test_MatMult_68();
	test_MatMult_69();
	test_MatMult_70();
	test_MatMult_71();
	test_MatMult_72();
	test_MatMult_73();
	test_MatMult_74();
	test_MatMult_75();
	test_MatMult_76();
	test_MatMult_77();
	test_MatMult_78();
	test_MatMult_79();
	test_MatMult_80();
	test_MatMult_81();
	test_MatMult_82();
	test_MatMult_83();
	test_MatMult_84();
	test_MatMult_85();
	test_MatMult_86();
	test_MatMult_87();
	test_MatMult_88();
	test_MatMult_89();
	test_MatMult_90();
	test_MatMult_91();
	test_MatMult_92();
	test_MatMult_93();
	test_MatMult_94();
	test_MatMult_95();
	test_MatMult_96();
	test_MatMult_97();
	test_MatMult_98();
	test_MatMult_99();
	test_MatMult_100();
	test_MatMult_101();
	test_MatMult_102();
	test_MatMult_103();
	test_MatMult_104();
	test_MatMult_105();
	test_MatMult_106();
	test_MatMult_107();
	test_MatMult_108();
	test_MatMult_109();
	test_MatMult_110();
	test_MatMult_111();
	test_MatMult_112();
	test_MatMult_113();
	test_MatMult_114();
	test_MatMult_115();
	test_MatMult_116();
	test_MatMult_117();
	test_MatMult_118();
	test_MatMult_119();
	test_MatMult_120();
	test_MatMult_121();
	test_MatMult_122();
	test_MatMult_123();
	test_MatMult_124();
	test_MatMult_125();
	test_MatMult_126();
	test_MatMult_127();
	test_MatMult_128();
	test_MatMult_129();
	test_MatMult_130();
	test_MatMult_131();
	test_MatMult_132();
	test_MatMult_133();
	test_MatMult_134();
	test_MatMult_135();
	test_MatMult_136();
	test_MatMult_137();
	test_MatMult_138();
	test_MatMult_139();
	test_MatMult_140();
	test_MatMult_141();
	test_MatMult_142();
	test_MatMult_143();
	test_MatMult_144();
	test_MatMult_145();
	test_MatMult_146();
	test_MatMult_147();
	test_MatMult_148();
	test_MatMult_149();
	test_MatMult_150();
	test_MatMult_151();
	test_MatMult_152();
	test_MatMult_153();
	test_MatMult_154();
	test_MatMult_155();
	test_MatMult_156();
	test_MatMult_157();
	test_MatMult_158();
	test_MatMult_159();
	test_MatMult_160();
	test_MatMult_161();
	test_MatMult_162();
	test_MatMult_163();
	test_MatMult_164();
	test_MatMult_165();
	test_MatMult_166();
	test_MatMult_167();
	test_MatMult_168();
	test_MatMult_169();
	test_MatMult_170();
	test_MatMult_171();
	test_MatMult_172();
	test_MatMult_173();
	test_MatMult_174();
	test_MatMult_175();
	test_MatMult_176();
	test_MatMult_177();
	test_MatMult_178();
	test_MatMult_179();
	test_MatMult_180();
	test_MatMult_181();
	test_MatMult_182();
	test_MatMult_183();
	test_MatMult_184();
	test_MatMult_185();
	test_MatMult_186();
	test_MatMult_187();
	test_MatMult_188();
	test_MatMult_189();
	test_MatMult_190();
	test_MatMult_191();
	test_MatMult_192();
	test_MatMult_193();
	test_MatMult_194();
	test_MatMult_195();
	test_MatMult_196();
	test_MatMult_197();
	test_MatMult_198();
	test_MatMult_199();
	test_MatMult_200();
	test_MatMult_201();
	test_MatMult_202();
	test_MatMult_203();
	test_MatMult_204();
	test_MatMult_205();
	test_MatMult_206();
	test_MatMult_207();
	test_MatMult_208();
	test_MatMult_209();
	test_MatMult_210();
	test_MatMult_211();
	test_MatMult_212();
	test_MatMult_213();
	test_MatMult_214();
	test_MatMult_215();
	test_MatMult_216();
	test_MatMult_217();
	test_MatMult_218();
	test_MatMult_219();
	test_MatMult_220();
	test_MatMult_221();
	test_MatMult_222();
	test_MatMult_223();
	test_MatMult_224();
	test_MatMult_225();
	test_MatMult_226();
	test_MatMult_227();
	test_MatMult_228();
	test_MatMult_229();
	test_MatMult_230();
	test_MatMult_231();
	test_MatMult_232();
	test_MatMult_233();
	test_MatMult_234();
	test_MatMult_235();
	test_MatMult_236();
	test_MatMult_237();
	test_MatMult_238();
	test_MatMult_239();
	test_MatMult_240();
	test_MatMult_241();
	test_MatMult_242();
	test_MatMult_243();
	test_MatMult_244();
	test_MatMult_245();
	test_MatMult_246();
	test_MatMult_247();
	test_MatMult_248();
	test_MatMult_249();
	test_MatMult_250();
	test_MatMult_251();
	test_MatMult_252();
	test_MatMult_253();
	test_MatMult_254();
	test_MatMult_255();
	test_MatMult_256();
	test_MatMult_257();
	test_MatMult_258();
	test_MatMult_259();
	test_MatMult_260();
	test_MatMult_261();
	test_MatMult_262();
	test_MatMult_263();
	test_MatMult_264();
	test_MatMult_265();
	test_MatMult_266();
	test_MatMult_267();
	test_MatMult_268();
	test_MatMult_269();
	test_MatMult_270();
	test_MatMult_271();
	test_MatMult_272();
	test_MatMult_273();
	test_MatMult_274();
	test_MatMult_275();
	test_MatMult_276();
	test_MatMult_277();
	test_MatMult_278();
	test_MatMult_279();
	test_MatMult_280();
	test_MatMult_281();
	test_MatMult_282();
	test_MatMult_283();
	test_MatMult_284();
	test_MatMult_285();
	test_MatMult_286();
	test_MatMult_287();
	test_MatMult_288();
	test_MatMult_289();
	test_MatMult_290();
	test_MatMult_291();
	test_MatMult_292();
	test_MatMult_293();
	test_MatMult_294();
	test_MatMult_295();
	test_MatMult_296();
	test_MatMult_297();
	test_MatMult_298();
	test_MatMult_299();
	test_MatMult_300();
	test_MatMult_301();
	test_MatMult_302();
	test_MatMult_303();
	test_MatMult_304();
	test_MatMult_305();
	test_MatMult_306();
	test_MatMult_307();
	test_MatMult_308();
	test_MatMult_309();
	test_MatMult_310();
	test_MatMult_311();
	test_MatMult_312();
	test_MatMult_313();
	test_MatMult_314();
	test_MatMult_315();
	test_MatMult_316();
	test_MatMult_317();
	test_MatMult_318();
	test_MatMult_319();
	test_MatMult_320();
	test_MatMult_321();
	test_MatMult_322();
	test_MatMult_323();
	test_MatMult_324();
	test_MatMult_325();
	test_MatMult_326();
	test_MatMult_327();
	test_MatMult_328();
	test_MatMult_329();
	test_MatMult_330();
	test_MatMult_331();
	test_MatMult_332();
	test_MatMult_333();
	test_MatMult_334();
	test_MatMult_335();
	test_MatMult_336();
	test_MatMult_337();
	test_MatMult_338();
	test_MatMult_339();
	test_MatMult_340();
	test_MatMult_341();
	test_MatMult_342();
	test_MatMult_343();
	test_MatMult_344();
	test_MatMult_345();
	test_MatMult_346();
	test_MatMult_347();
	test_MatMult_348();
	test_MatMult_349();
	test_MatMult_350();
	test_MatMult_351();
	test_MatMult_352();
	test_MatMult_353();
	test_MatMult_354();
	test_MatMult_355();
	test_MatMult_356();
	test_MatMult_357();
	test_MatMult_358();
	test_MatMult_359();
	test_MatMult_360();
	test_MatMult_361();
	test_MatMult_362();
	test_MatMult_363();
	test_MatMult_364();
	test_MatMult_365();
	test_MatMult_366();
	test_MatMult_367();
	test_MatMult_368();
	test_MatMult_369();
	test_MatMult_370();
	test_MatMult_371();
	test_MatMult_372();
	test_MatMult_373();
	test_MatMult_374();
	test_MatMult_375();
	test_MatMult_376();
	test_MatMult_377();
	test_MatMult_378();
	test_MatMult_379();
	test_MatMult_380();
	test_MatMult_381();
	test_MatMult_382();
	test_MatMult_383();
	test_MatMult_384();
	test_MatMult_385();
	test_MatMult_386();
	test_MatMult_387();
	test_MatMult_388();
	test_MatMult_389();
	test_MatMult_390();
	test_MatMult_391();
	test_MatMult_392();
	test_MatMult_393();
	test_MatMult_394();
	test_MatMult_395();
	test_MatMult_396();
	test_MatMult_397();
	test_MatMult_398();
	test_MatMult_399();
	test_MatMult_400();
	test_MatMult_401();
	test_MatMult_402();
	test_MatMult_403();
	test_MatMult_404();
	test_MatMult_405();
	test_MatMult_406();
	test_MatMult_407();
	test_MatMult_408();
	test_MatMult_409();
	test_MatMult_410();
	test_MatMult_411();
	test_MatMult_412();
	test_MatMult_413();
	test_MatMult_414();
	test_MatMult_415();
	test_MatMult_416();
	test_MatMult_417();
	test_MatMult_418();
	test_MatMult_419();
	test_MatMult_420();
	test_MatMult_421();
	test_MatMult_422();
	test_MatMult_423();
	test_MatMult_424();
	test_MatMult_425();
	test_MatMult_426();
	test_MatMult_427();
	test_MatMult_428();
	test_MatMult_429();
	test_MatMult_430();
	test_MatMult_431();
	test_MatMult_432();
	test_MatMult_433();
	test_MatMult_434();
	test_MatMult_435();
	test_MatMult_436();
	test_MatMult_437();
	test_MatMult_438();
	test_MatMult_439();
	test_MatMult_440();
	test_MatMult_441();
	test_MatMult_442();
	test_MatMult_443();
	test_MatMult_444();
	test_MatMult_445();
	test_MatMult_446();
	test_MatMult_447();
	test_MatMult_448();
	test_MatMult_449();
	test_MatMult_450();
	test_MatMult_451();
	test_MatMult_452();
	test_MatMult_453();
	test_MatMult_454();
	test_MatMult_455();
	test_MatMult_456();
	test_MatMult_457();
	test_MatMult_458();
	test_MatMult_459();
	test_MatMult_460();
	test_MatMult_461();
	test_MatMult_462();
	test_MatMult_463();
	test_MatMult_464();
	test_MatMult_465();
	test_MatMult_466();
	test_MatMult_467();
	test_MatMult_468();
	test_MatMult_469();
	test_MatMult_470();
	test_MatMult_471();
	test_MatMult_472();
	test_MatMult_473();
	test_MatMult_474();
	test_MatMult_475();
	test_MatMult_476();
	test_MatMult_477();
	test_MatMult_478();
	test_MatMult_479();
	test_MatMult_480();
	test_MatMult_481();
	test_MatMult_482();
	test_MatMult_483();
	test_MatMult_484();
	test_MatMult_485();
	test_MatMult_486();
	test_MatMult_487();
	test_MatMult_488();
	test_MatMult_489();
	test_MatMult_490();
	test_MatMult_491();
	test_MatMult_492();
	test_MatMult_493();
	test_MatMult_494();
	test_MatMult_495();
	test_MatMult_496();
	test_MatMult_497();
	test_MatMult_498();
	test_MatMult_499();
	test_MatMult_500();
	test_MatMult_501();
	test_MatMult_502();
	test_MatMult_503();
	test_MatMult_504();
	test_MatMult_505();
	test_MatMult_506();
	test_MatMult_507();
	test_MatMult_508();
	test_MatMult_509();
	test_MatMult_510();
	test_MatMult_511();
	test_MatMult_512();
	test_MatMult_513();
	test_MatMult_514();
	test_MatMult_515();
	test_MatMult_516();
	test_MatMult_517();
	test_MatMult_518();
	test_MatMult_519();
	test_MatMult_520();
	test_MatMult_521();
	test_MatMult_522();
	test_MatMult_523();
	test_MatMult_524();
	test_MatMult_525();
	test_MatMult_526();
	test_MatMult_527();
	test_MatMult_528();
	test_MatMult_529();
	test_MatMult_530();
	test_MatMult_531();
	test_MatMult_532();
	test_MatMult_533();
	test_MatMult_534();
	test_MatMult_535();
	test_MatMult_536();
	test_MatMult_537();
	test_MatMult_538();
	test_MatMult_539();
	test_MatMult_540();
	test_MatMult_541();
	test_MatMult_542();
	test_MatMult_543();
	test_MatMult_544();
	test_MatMult_545();
	test_MatMult_546();
	test_MatMult_547();
	test_MatMult_548();
	test_MatMult_549();
	test_MatMult_550();
	test_MatMult_551();
	test_MatMult_552();
	test_MatMult_553();
	test_MatMult_554();
	test_MatMult_555();
	test_MatMult_556();
	test_MatMult_557();
	test_MatMult_558();
	test_MatMult_559();
	test_MatMult_560();
	test_MatMult_561();
	test_MatMult_562();
	test_MatMult_563();
	test_MatMult_564();
	test_MatMult_565();
	test_MatMult_566();
	test_MatMult_567();
	test_MatMult_568();
	test_MatMult_569();
	test_MatMult_570();
	test_MatMult_571();
	test_MatMult_572();
	test_MatMult_573();
	test_MatMult_574();
	test_MatMult_575();
	test_MatMult_576();
	test_MatMult_577();
	test_MatMult_578();
	test_MatMult_579();
	test_MatMult_580();
	test_MatMult_581();
	test_MatMult_582();
	test_MatMult_583();
	test_MatMult_584();
	test_MatMult_585();
	test_MatMult_586();
	test_MatMult_587();
	test_MatMult_588();
	test_MatMult_589();
	test_MatMult_590();
	test_MatMult_591();
	test_MatMult_592();
	test_MatMult_593();
	test_MatMult_594();
	test_MatMult_595();
	test_MatMult_596();
	test_MatMult_597();
	test_MatMult_598();
	test_MatMult_599();
	test_MatMult_600();
	test_MatMult_601();
	test_MatMult_602();
	test_MatMult_603();
	test_MatMult_604();
	test_MatMult_605();
	test_MatMult_606();
	test_MatMult_607();
	test_MatMult_608();
	test_MatMult_609();
	test_MatMult_610();
	test_MatMult_611();
	test_MatMult_612();
	test_MatMult_613();
	test_MatMult_614();
	test_MatMult_615();
	test_MatMult_616();
	test_MatMult_617();
	test_MatMult_618();
	test_MatMult_619();
	test_MatMult_620();
	test_MatMult_621();
	test_MatMult_622();
	test_MatMult_623();
	test_MatMult_624();
	test_MatMult_625();
	test_MatMult_626();
	test_MatMult_627();
	test_MatMult_628();
	test_MatMult_629();
	test_MatMult_630();
	test_MatMult_631();
	test_MatMult_632();
	test_MatMult_633();
	test_MatMult_634();
	test_MatMult_635();
	test_MatMult_636();
	test_MatMult_637();
	test_MatMult_638();
	test_MatMult_639();
	test_MatMult_640();
	test_MatMult_641();
	test_MatMult_642();
	test_MatMult_643();
	test_MatMult_644();
	test_MatMult_645();
	test_MatMult_646();
	test_MatMult_647();
	test_MatMult_648();
	test_MatMult_649();
	test_MatMult_650();
	test_MatMult_651();
	test_MatMult_652();
	test_MatMult_653();
	test_MatMult_654();
	test_MatMult_655();
	test_MatMult_656();
	test_MatMult_657();
	test_MatMult_658();
	test_MatMult_659();
	test_MatMult_660();
	test_MatMult_661();
	test_MatMult_662();
	test_MatMult_663();
	test_MatMult_664();
	test_MatMult_665();
	test_MatMult_666();
	test_MatMult_667();
	test_MatMult_668();
	test_MatMult_669();
	test_MatMult_670();
	test_MatMult_671();
	test_MatMult_672();
	test_MatMult_673();
	test_MatMult_674();
	test_MatMult_675();
	test_MatMult_676();
	test_MatMult_677();
	test_MatMult_678();
	test_MatMult_679();
	test_MatMult_680();
	test_MatMult_681();
	test_MatMult_682();
	test_MatMult_683();
	test_MatMult_684();
	test_MatMult_685();
	test_MatMult_686();
	test_MatMult_687();
	test_MatMult_688();
	test_MatMult_689();
	test_MatMult_690();
	test_MatMult_691();
	test_MatMult_692();
	test_MatMult_693();
	test_MatMult_694();
	test_MatMult_695();
	test_MatMult_696();
	test_MatMult_697();
	test_MatMult_698();
	test_MatMult_699();
	test_MatMult_700();
	test_MatMult_701();
	test_MatMult_702();
	test_MatMult_703();
	test_MatMult_704();
	test_MatMult_705();
	test_MatMult_706();
	test_MatMult_707();
	test_MatMult_708();
	test_MatMult_709();
	test_MatMult_710();
	test_MatMult_711();
	test_MatMult_712();
	test_MatMult_713();
	test_MatMult_714();
	test_MatMult_715();
	test_MatMult_716();
	test_MatMult_717();
	test_MatMult_718();
	test_MatMult_719();
	test_MatMult_720();
	test_MatMult_721();
	test_MatMult_722();
	test_MatMult_723();
	test_MatMult_724();
	test_MatMult_725();
	test_MatMult_726();
	test_MatMult_727();
	test_MatMult_728();
	test_MatMult_729();
	test_MatMult_730();
	test_MatMult_731();
	test_MatMult_732();
	test_MatMult_733();
	test_MatMult_734();
	test_MatMult_735();
	test_MatMult_736();
	test_MatMult_737();
	test_MatMult_738();
	test_MatMult_739();
	test_MatMult_740();
	test_MatMult_741();
	test_MatMult_742();
	test_MatMult_743();
	test_MatMult_744();
	test_MatMult_745();
	test_MatMult_746();
	test_MatMult_747();
	test_MatMult_748();
	test_MatMult_749();
	test_MatMult_750();
	test_MatMult_751();
	test_MatMult_752();
	test_MatMult_753();
	test_MatMult_754();
	test_MatMult_755();
	test_MatMult_756();
	test_MatMult_757();
	test_MatMult_758();
	test_MatMult_759();
	test_MatMult_760();
	test_MatMult_761();
	test_MatMult_762();
	test_MatMult_763();
	test_MatMult_764();
	test_MatMult_765();
	test_MatMult_766();
	test_MatMult_767();
	test_MatMult_768();
	test_MatMult_769();
	test_MatMult_770();
	test_MatMult_771();
	test_MatMult_772();
	test_MatMult_773();
	test_MatMult_774();
	test_MatMult_775();
	test_MatMult_776();
	test_MatMult_777();
	test_MatMult_778();
	test_MatMult_779();
	test_MatMult_780();
	test_MatMult_781();
	test_MatMult_782();
	test_MatMult_783();
	test_MatMult_784();
	test_MatMult_785();
	test_MatMult_786();
	test_MatMult_787();
	test_MatMult_788();
	test_MatMult_789();
	test_MatMult_790();
	test_MatMult_791();
	test_MatMult_792();
	test_MatMult_793();
	test_MatMult_794();
	test_MatMult_795();
	test_MatMult_796();
	test_MatMult_797();
	test_MatMult_798();
	test_MatMult_799();
	test_MatMult_800();
	test_MatMult_801();
	test_MatMult_802();
	test_MatMult_803();
	test_MatMult_804();
	test_MatMult_805();
	test_MatMult_806();
	test_MatMult_807();
	test_MatMult_808();
	test_MatMult_809();
	test_MatMult_810();
	test_MatMult_811();
	test_MatMult_812();
	test_MatMult_813();
	test_MatMult_814();
	test_MatMult_815();
	test_MatMult_816();
	test_MatMult_817();
	test_MatMult_818();
	test_MatMult_819();
	test_MatMult_820();
	test_MatMult_821();
	test_MatMult_822();
	test_MatMult_823();
	test_MatMult_824();
	test_MatMult_825();
	test_MatMult_826();
	test_MatMult_827();
	test_MatMult_828();
	test_MatMult_829();
	test_MatMult_830();
	test_MatMult_831();
	test_MatMult_832();
	test_MatMult_833();
	test_MatMult_834();
	test_MatMult_835();
	test_MatMult_836();
	test_MatMult_837();
	test_MatMult_838();
	test_MatMult_839();
	test_MatMult_840();
	test_MatMult_841();
	test_MatMult_842();
	test_MatMult_843();
	test_MatMult_844();
	test_MatMult_845();
	test_MatMult_846();
	test_MatMult_847();
	test_MatMult_848();
	test_MatMult_849();
	test_MatMult_850();
	test_MatMult_851();
	test_MatMult_852();
	test_MatMult_853();
	test_MatMult_854();
	test_MatMult_855();
	test_MatMult_856();
	test_MatMult_857();
	test_MatMult_858();
	test_MatMult_859();
	test_MatMult_860();
	test_MatMult_861();
	test_MatMult_862();
	test_MatMult_863();
	test_MatMult_864();
	test_MatMult_865();
	test_MatMult_866();
	test_MatMult_867();
	test_MatMult_868();
	test_MatMult_869();
	test_MatMult_870();
	test_MatMult_871();
	test_MatMult_872();
	test_MatMult_873();
	test_MatMult_874();
	test_MatMult_875();
	test_MatMult_876();
	test_MatMult_877();
	test_MatMult_878();
	test_MatMult_879();
	test_MatMult_880();
	test_MatMult_881();
	test_MatMult_882();
	test_MatMult_883();
	test_MatMult_884();
	test_MatMult_885();
	test_MatMult_886();
	test_MatMult_887();
	test_MatMult_888();
	test_MatMult_889();
	test_MatMult_890();
	test_MatMult_891();
	test_MatMult_892();
	test_MatMult_893();
	test_MatMult_894();
	test_MatMult_895();
	test_MatMult_896();
	test_MatMult_897();
	test_MatMult_898();
	test_MatMult_899();
	test_MatMult_900();
	test_MatMult_901();
	test_MatMult_902();
	test_MatMult_903();
	test_MatMult_904();
	test_MatMult_905();
	test_MatMult_906();
	test_MatMult_907();
	test_MatMult_908();
	test_MatMult_909();
	test_MatMult_910();
	test_MatMult_911();
	test_MatMult_912();
	test_MatMult_913();
	test_MatMult_914();
	test_MatMult_915();
	test_MatMult_916();
	test_MatMult_917();
	test_MatMult_918();
	test_MatMult_919();
	test_MatMult_920();
	test_MatMult_921();
	test_MatMult_922();
	test_MatMult_923();
	test_MatMult_924();
	test_MatMult_925();
	test_MatMult_926();
	test_MatMult_927();
	test_MatMult_928();
	test_MatMult_929();
	test_MatMult_930();
	test_MatMult_931();
	test_MatMult_932();
	test_MatMult_933();
	test_MatMult_934();
	test_MatMult_935();
	test_MatMult_936();
	test_MatMult_937();
	test_MatMult_938();
	test_MatMult_939();
	test_MatMult_940();
	test_MatMult_941();
	test_MatMult_942();
	test_MatMult_943();
	test_MatMult_944();
	test_MatMult_945();
	test_MatMult_946();
	test_MatMult_947();
	test_MatMult_948();
	test_MatMult_949();
	test_MatMult_950();
	test_MatMult_951();
	test_MatMult_952();
	test_MatMult_953();
	test_MatMult_954();
	test_MatMult_955();
	test_MatMult_956();
	test_MatMult_957();
	test_MatMult_958();
	test_MatMult_959();
	test_MatMult_960();
	test_MatMult_961();
	test_MatMult_962();
	test_MatMult_963();
	test_MatMult_964();
	test_MatMult_965();
	test_MatMult_966();
	test_MatMult_967();
	test_MatMult_968();
	test_MatMult_969();
	test_MatMult_970();
	test_MatMult_971();
	test_MatMult_972();
	test_MatMult_973();
	test_MatMult_974();
	test_MatMult_975();
	test_MatMult_976();
	test_MatMult_977();
	test_MatMult_978();
	test_MatMult_979();
	test_MatMult_980();
	test_MatMult_981();
	test_MatMult_982();
	test_MatMult_983();
	test_MatMult_984();
	test_MatMult_985();
	test_MatMult_986();
	test_MatMult_987();
	test_MatMult_988();
	test_MatMult_989();
	test_MatMult_990();
	test_MatMult_991();
	test_MatMult_992();
	test_MatMult_993();
	test_MatMult_994();
	test_MatMult_995();
	test_MatMult_996();
	test_MatMult_997();
	test_MatMult_998();
	test_MatMult_999();
	test_MatMult_1000();
	test_MatMult_1001();
	test_MatMult_1002();
	test_MatMult_1003();
	test_MatMult_1004();
	test_MatMult_1005();
	test_MatMult_1006();
	test_MatMult_1007();
	test_MatMult_1008();
	test_MatMult_1009();
	test_MatMult_1010();
	test_MatMult_1011();
	test_MatMult_1012();
	test_MatMult_1013();
	test_MatMult_1014();
	test_MatMult_1015();
	test_MatMult_1016();
	test_MatMult_1017();
	test_MatMult_1018();
	test_MatMult_1019();
	test_MatMult_1020();
	test_MatMult_1021();
	test_MatMult_1022();
	test_MatMult_1023();
	test_MatMult_1024();
	test_MatMult_1025();
	test_MatMult_1026();
	test_MatMult_1027();
	test_MatMult_1028();
	test_MatMult_1029();
	test_MatMult_1030();
	test_MatMult_1031();
	test_MatMult_1032();
	test_MatMult_1033();
	test_MatMult_1034();
	test_MatMult_1035();
	test_MatMult_1036();
	test_MatMult_1037();
	test_MatMult_1038();
	test_MatMult_1039();
	test_MatMult_1040();
	test_MatMult_1041();
	test_MatMult_1042();
	test_MatMult_1043();
	test_MatMult_1044();
	test_MatMult_1045();
	test_MatMult_1046();
	test_MatMult_1047();
	test_MatMult_1048();
	test_MatMult_1049();
	test_MatMult_1050();
	test_MatMult_1051();
	test_MatMult_1052();
	test_MatMult_1053();
	test_MatMult_1054();
	test_MatMult_1055();
	test_MatMult_1056();
	test_MatMult_1057();
	test_MatMult_1058();
	test_MatMult_1059();
	test_MatMult_1060();
	test_MatMult_1061();
	test_MatMult_1062();
	test_MatMult_1063();
	test_MatMult_1064();
	test_MatMult_1065();
	test_MatMult_1066();
	test_MatMult_1067();
	test_MatMult_1068();
	test_MatMult_1069();
	test_MatMult_1070();
	test_MatMult_1071();
	test_MatMult_1072();
	test_MatMult_1073();
	test_MatMult_1074();
	test_MatMult_1075();
	test_MatMult_1076();
	test_MatMult_1077();
	test_MatMult_1078();
	test_MatMult_1079();
	test_MatMult_1080();
	test_MatMult_1081();
	test_MatMult_1082();
	test_MatMult_1083();
	test_MatMult_1084();
	test_MatMult_1085();
	test_MatMult_1086();
	test_MatMult_1087();
	test_MatMult_1088();
	test_MatMult_1089();
	test_MatMult_1090();
	test_MatMult_1091();
	test_MatMult_1092();
	test_MatMult_1093();
	test_MatMult_1094();
	test_MatMult_1095();
	test_MatMult_1096();
	test_MatMult_1097();
	test_MatMult_1098();
	test_MatMult_1099();
	test_MatMult_1100();
	test_MatMult_1101();
	test_MatMult_1102();
	test_MatMult_1103();
	test_MatMult_1104();
	test_MatMult_1105();
	test_MatMult_1106();
	test_MatMult_1107();
	test_MatMult_1108();
	test_MatMult_1109();
	test_MatMult_1110();
	test_MatMult_1111();
	test_MatMult_1112();
	test_MatMult_1113();
	test_MatMult_1114();
	test_MatMult_1115();
	test_MatMult_1116();
	test_MatMult_1117();
	test_MatMult_1118();
	test_MatMult_1119();
	test_MatMult_1120();
	test_MatMult_1121();
	test_MatMult_1122();
	test_MatMult_1123();
	test_MatMult_1124();
	test_MatMult_1125();
	test_MatMult_1126();
	test_MatMult_1127();
	test_MatMult_1128();
	test_MatMult_1129();
	test_MatMult_1130();
	test_MatMult_1131();
	test_MatMult_1132();
	test_MatMult_1133();
	test_MatMult_1134();
	test_MatMult_1135();
	test_MatMult_1136();
	test_MatMult_1137();
	test_MatMult_1138();
	test_MatMult_1139();
	test_MatMult_1140();
	test_MatMult_1141();
	test_MatMult_1142();
	test_MatMult_1143();
	test_MatMult_1144();
	test_MatMult_1145();
	test_MatMult_1146();
	test_MatMult_1147();
	test_MatMult_1148();
	test_MatMult_1149();
	test_MatMult_1150();
	test_MatMult_1151();
	test_MatMult_1152();
	test_MatMult_1153();
	test_MatMult_1154();
	test_MatMult_1155();
	test_MatMult_1156();
	test_MatMult_1157();
	test_MatMult_1158();
	test_MatMult_1159();
	test_MatMult_1160();
	test_MatMult_1161();
	test_MatMult_1162();
	test_MatMult_1163();
	test_MatMult_1164();
	test_MatMult_1165();
	test_MatMult_1166();
	test_MatMult_1167();
	test_MatMult_1168();
	test_MatMult_1169();
	test_MatMult_1170();
	test_MatMult_1171();
	test_MatMult_1172();
	test_MatMult_1173();
	test_MatMult_1174();
	test_MatMult_1175();
	test_MatMult_1176();
	test_MatMult_1177();
	test_MatMult_1178();
	test_MatMult_1179();
	test_MatMult_1180();
	test_MatMult_1181();
	test_MatMult_1182();
	test_MatMult_1183();
	test_MatMult_1184();
	test_MatMult_1185();
	test_MatMult_1186();
	test_MatMult_1187();
	test_MatMult_1188();
	test_MatMult_1189();
	test_MatMult_1190();
	test_MatMult_1191();
	test_MatMult_1192();
	test_MatMult_1193();
	test_MatMult_1194();
	test_MatMult_1195();
	test_MatMult_1196();
	test_MatMult_1197();
	test_MatMult_1198();
	test_MatMult_1199();
	test_MatMult_1200();
	test_MatMult_1201();
	test_MatMult_1202();
	test_MatMult_1203();
	test_MatMult_1204();
	test_MatMult_1205();
	test_MatMult_1206();
	test_MatMult_1207();
	test_MatMult_1208();
	test_MatMult_1209();
	test_MatMult_1210();
	test_MatMult_1211();
	test_MatMult_1212();
	test_MatMult_1213();
	test_MatMult_1214();
	test_MatMult_1215();
	test_MatMult_1216();
	test_MatMult_1217();
	test_MatMult_1218();
	test_MatMult_1219();
	test_MatMult_1220();
	test_MatMult_1221();
	test_MatMult_1222();
	test_MatMult_1223();
	test_MatMult_1224();
	test_MatMult_1225();
	test_MatMult_1226();
	test_MatMult_1227();
	test_MatMult_1228();
	test_MatMult_1229();
	test_MatMult_1230();
	test_MatMult_1231();
	test_MatMult_1232();
	test_MatMult_1233();
	test_MatMult_1234();
	test_MatMult_1235();
	test_MatMult_1236();
	test_MatMult_1237();
	test_MatMult_1238();
	test_MatMult_1239();
	test_MatMult_1240();
	test_MatMult_1241();
	test_MatMult_1242();
	test_MatMult_1243();
	test_MatMult_1244();
	test_MatMult_1245();
	test_MatMult_1246();
	test_MatMult_1247();
	test_MatMult_1248();
	test_MatMult_1249();
	test_MatMult_1250();
	test_MatMult_1251();
	test_MatMult_1252();
	test_MatMult_1253();
	test_MatMult_1254();
	test_MatMult_1255();
	test_MatMult_1256();
	test_MatMult_1257();
	test_MatMult_1258();
	test_MatMult_1259();
	test_MatMult_1260();
	test_MatMult_1261();
	test_MatMult_1262();
	test_MatMult_1263();
	test_MatMult_1264();
	test_MatMult_1265();
	test_MatMult_1266();
	test_MatMult_1267();
	test_MatMult_1268();
	test_MatMult_1269();
	test_MatMult_1270();
	test_MatMult_1271();
	test_MatMult_1272();
	test_MatMult_1273();
	test_MatMult_1274();
	test_MatMult_1275();
	test_MatMult_1276();
	test_MatMult_1277();
	test_MatMult_1278();
	test_MatMult_1279();
	test_MatMult_1280();
	test_MatMult_1281();
	test_MatMult_1282();
	test_MatMult_1283();
	test_MatMult_1284();
	test_MatMult_1285();
	test_MatMult_1286();
	test_MatMult_1287();
	test_MatMult_1288();
	test_MatMult_1289();
	test_MatMult_1290();
	test_MatMult_1291();
	test_MatMult_1292();
	test_MatMult_1293();
	test_MatMult_1294();
	test_MatMult_1295();
	test_MatMult_1296();
	test_MatMult_1297();
	test_MatMult_1298();
	test_MatMult_1299();
	test_MatMult_1300();
	test_MatMult_1301();
	test_MatMult_1302();
	test_MatMult_1303();
	test_MatMult_1304();
	test_MatMult_1305();
	test_MatMult_1306();
	test_MatMult_1307();
	test_MatMult_1308();
	test_MatMult_1309();
	test_MatMult_1310();
	test_MatMult_1311();
	test_MatMult_1312();
	test_MatMult_1313();
	test_MatMult_1314();
	test_MatMult_1315();
	test_MatMult_1316();
	test_MatMult_1317();
	test_MatMult_1318();
	test_MatMult_1319();
	test_MatMult_1320();
	test_MatMult_1321();
	test_MatMult_1322();
	test_MatMult_1323();
	test_MatMult_1324();
	test_MatMult_1325();
	test_MatMult_1326();
	test_MatMult_1327();
	test_MatMult_1328();
	test_MatMult_1329();
	test_MatMult_1330();
	test_MatMult_1331();
	test_MatMult_1332();
	test_MatMult_1333();
	test_MatMult_1334();
	test_MatMult_1335();
	test_MatMult_1336();
	test_MatMult_1337();
	test_MatMult_1338();
	test_MatMult_1339();
	test_MatMult_1340();
	test_MatMult_1341();
	test_MatMult_1342();
	test_MatMult_1343();
	test_MatMult_1344();
	test_MatMult_1345();
	test_MatMult_1346();
	test_MatMult_1347();
	test_MatMult_1348();
	test_MatMult_1349();
	test_MatMult_1350();
	test_MatMult_1351();
	test_MatMult_1352();
	test_MatMult_1353();
	test_MatMult_1354();
	test_MatMult_1355();
	test_MatMult_1356();
	test_MatMult_1357();
	test_MatMult_1358();
	test_MatMult_1359();
	test_MatMult_1360();
	test_MatMult_1361();
	test_MatMult_1362();
	test_MatMult_1363();
	test_MatMult_1364();
	test_MatMult_1365();
	test_MatMult_1366();
	test_MatMult_1367();
	test_MatMult_1368();
	test_MatMult_1369();
	test_MatMult_1370();
	test_MatMult_1371();
	test_MatMult_1372();
	test_MatMult_1373();
	test_MatMult_1374();
	test_MatMult_1375();
	test_MatMult_1376();
	test_MatMult_1377();
	test_MatMult_1378();
	test_MatMult_1379();
	test_MatMult_1380();
	test_MatMult_1381();
	test_MatMult_1382();
	test_MatMult_1383();
	test_MatMult_1384();
	test_MatMult_1385();
	test_MatMult_1386();
	test_MatMult_1387();
	test_MatMult_1388();
	test_MatMult_1389();
	test_MatMult_1390();
	test_MatMult_1391();
	test_MatMult_1392();
	test_MatMult_1393();
	test_MatMult_1394();
	test_MatMult_1395();
	test_MatMult_1396();
	test_MatMult_1397();
	test_MatMult_1398();
	test_MatMult_1399();
	test_MatMult_1400();
	test_MatMult_1401();
	test_MatMult_1402();
	test_MatMult_1403();
	test_MatMult_1404();
	test_MatMult_1405();
	test_MatMult_1406();
	test_MatMult_1407();
	test_MatMult_1408();
	test_MatMult_1409();
	test_MatMult_1410();
	test_MatMult_1411();
	test_MatMult_1412();
	test_MatMult_1413();
	test_MatMult_1414();
	test_MatMult_1415();
	test_MatMult_1416();
	test_MatMult_1417();
	test_MatMult_1418();
	test_MatMult_1419();
	test_MatMult_1420();
	test_MatMult_1421();
	test_MatMult_1422();
	test_MatMult_1423();
	test_MatMult_1424();
	test_MatMult_1425();
	test_MatMult_1426();
	test_MatMult_1427();
	test_MatMult_1428();
	test_MatMult_1429();
	test_MatMult_1430();
	test_MatMult_1431();
	test_MatMult_1432();
	test_MatMult_1433();
	test_MatMult_1434();
	test_MatMult_1435();
	test_MatMult_1436();
	test_MatMult_1437();
	test_MatMult_1438();
	test_MatMult_1439();
	test_MatMult_1440();
	test_MatMult_1441();
	test_MatMult_1442();
	test_MatMult_1443();
	test_MatMult_1444();
	test_MatMult_1445();
	test_MatMult_1446();
	test_MatMult_1447();
	test_MatMult_1448();
	test_MatMult_1449();
	test_MatMult_1450();
	test_MatMult_1451();
	test_MatMult_1452();
	test_MatMult_1453();
	test_MatMult_1454();
	test_MatMult_1455();
	test_MatMult_1456();
	test_MatMult_1457();
	test_MatMult_1458();
	test_MatMult_1459();
	test_MatMult_1460();
	test_MatMult_1461();
	test_MatMult_1462();
	test_MatMult_1463();
	test_MatMult_1464();
	test_MatMult_1465();
	test_MatMult_1466();
	test_MatMult_1467();
	test_MatMult_1468();
	test_MatMult_1469();
	test_MatMult_1470();
	test_MatMult_1471();
	test_MatMult_1472();
	test_MatMult_1473();
	test_MatMult_1474();
	test_MatMult_1475();
	test_MatMult_1476();
	test_MatMult_1477();
	test_MatMult_1478();
	test_MatMult_1479();
	test_MatMult_1480();
	test_MatMult_1481();
	test_MatMult_1482();
	test_MatMult_1483();
	test_MatMult_1484();
	test_MatMult_1485();
	test_MatMult_1486();
	test_MatMult_1487();
	test_MatMult_1488();
	test_MatMult_1489();
	test_MatMult_1490();
	test_MatMult_1491();
	test_MatMult_1492();
	test_MatMult_1493();
	test_MatMult_1494();
	test_MatMult_1495();
	test_MatMult_1496();
	test_MatMult_1497();
	test_MatMult_1498();
	test_MatMult_1499();
	test_MatMult_1500();
	test_MatMult_1501();
	test_MatMult_1502();
	test_MatMult_1503();
	test_MatMult_1504();
	test_MatMult_1505();
	test_MatMult_1506();
	test_MatMult_1507();
	test_MatMult_1508();
	test_MatMult_1509();
	test_MatMult_1510();
	test_MatMult_1511();
	test_MatMult_1512();
	test_MatMult_1513();
	test_MatMult_1514();
	test_MatMult_1515();
	test_MatMult_1516();
	test_MatMult_1517();
	test_MatMult_1518();
	test_MatMult_1519();
	test_MatMult_1520();
	test_MatMult_1521();
	test_MatMult_1522();
	test_MatMult_1523();
	test_MatMult_1524();
	test_MatMult_1525();
	test_MatMult_1526();
	test_MatMult_1527();
	test_MatMult_1528();
	test_MatMult_1529();
	test_MatMult_1530();
	test_MatMult_1531();
	test_MatMult_1532();
	test_MatMult_1533();
	test_MatMult_1534();
	test_MatMult_1535();
	test_MatMult_1536();
	test_MatMult_1537();
	test_MatMult_1538();
	test_MatMult_1539();
	test_MatMult_1540();
	test_MatMult_1541();
	test_MatMult_1542();
	test_MatMult_1543();
	test_MatMult_1544();
	test_MatMult_1545();
	test_MatMult_1546();
	test_MatMult_1547();
	test_MatMult_1548();
	test_MatMult_1549();
	test_MatMult_1550();
	test_MatMult_1551();
	test_MatMult_1552();
	test_MatMult_1553();
	test_MatMult_1554();
	test_MatMult_1555();
	test_MatMult_1556();
	test_MatMult_1557();
	test_MatMult_1558();
	test_MatMult_1559();
	test_MatMult_1560();
	test_MatMult_1561();
	test_MatMult_1562();
	test_MatMult_1563();
	test_MatMult_1564();
	test_MatMult_1565();
	test_MatMult_1566();
	test_MatMult_1567();
	test_MatMult_1568();
	test_MatMult_1569();
	test_MatMult_1570();
	test_MatMult_1571();
	test_MatMult_1572();
	test_MatMult_1573();
	test_MatMult_1574();
	test_MatMult_1575();
	test_MatMult_1576();
	test_MatMult_1577();
	test_MatMult_1578();
	test_MatMult_1579();
	test_MatMult_1580();
	test_MatMult_1581();
	test_MatMult_1582();
	test_MatMult_1583();
	test_MatMult_1584();
	test_MatMult_1585();
	test_MatMult_1586();
	test_MatMult_1587();
	test_MatMult_1588();
	test_MatMult_1589();
	test_MatMult_1590();
	test_MatMult_1591();
	test_MatMult_1592();
	test_MatMult_1593();
	test_MatMult_1594();
	test_MatMult_1595();
	test_MatMult_1596();
	test_MatMult_1597();
	test_MatMult_1598();
	test_MatMult_1599();
	test_MatMult_1600();
	test_MatMult_1601();
	test_MatMult_1602();
	test_MatMult_1603();
	test_MatMult_1604();
	test_MatMult_1605();
	test_MatMult_1606();
	test_MatMult_1607();
	test_MatMult_1608();
	test_MatMult_1609();
	test_MatMult_1610();
	test_MatMult_1611();
	test_MatMult_1612();
	test_MatMult_1613();
	test_MatMult_1614();
	test_MatMult_1615();
	test_MatMult_1616();
	test_MatMult_1617();
	test_MatMult_1618();
	test_MatMult_1619();
	test_MatMult_1620();
	test_MatMult_1621();
	test_MatMult_1622();
	test_MatMult_1623();
	test_MatMult_1624();
	test_MatMult_1625();
	test_MatMult_1626();
	test_MatMult_1627();
	test_MatMult_1628();
	test_MatMult_1629();
	test_MatMult_1630();
	test_MatMult_1631();
	test_MatMult_1632();
	test_MatMult_1633();
	test_MatMult_1634();
	test_MatMult_1635();
	test_MatMult_1636();
	test_MatMult_1637();
	test_MatMult_1638();
	test_MatMult_1639();
	test_MatMult_1640();
	test_MatMult_1641();
	test_MatMult_1642();
	test_MatMult_1643();
	test_MatMult_1644();
	test_MatMult_1645();
	test_MatMult_1646();
	test_MatMult_1647();
	test_MatMult_1648();
	test_MatMult_1649();
	test_MatMult_1650();
	test_MatMult_1651();
	test_MatMult_1652();
	test_MatMult_1653();
	test_MatMult_1654();
	test_MatMult_1655();
	test_MatMult_1656();
	test_MatMult_1657();
	test_MatMult_1658();
	test_MatMult_1659();
	test_MatMult_1660();
	test_MatMult_1661();
	test_MatMult_1662();
	test_MatMult_1663();
	test_MatMult_1664();
	test_MatMult_1665();
	test_MatMult_1666();
	test_MatMult_1667();
	test_MatMult_1668();
	test_MatMult_1669();
	test_MatMult_1670();
	test_MatMult_1671();
	test_MatMult_1672();
	test_MatMult_1673();
	test_MatMult_1674();
	test_MatMult_1675();
	test_MatMult_1676();
	test_MatMult_1677();
	test_MatMult_1678();
	test_MatMult_1679();
	test_MatMult_1680();
	test_MatMult_1681();
	test_MatMult_1682();
	test_MatMult_1683();
	test_MatMult_1684();
	test_MatMult_1685();
	test_MatMult_1686();
	test_MatMult_1687();
	test_MatMult_1688();
	test_MatMult_1689();
	test_MatMult_1690();
	test_MatMult_1691();
	test_MatMult_1692();
	test_MatMult_1693();
	test_MatMult_1694();
	test_MatMult_1695();
	test_MatMult_1696();
	test_MatMult_1697();
	test_MatMult_1698();
	test_MatMult_1699();
	test_MatMult_1700();
	test_MatMult_1701();
	test_MatMult_1702();
	test_MatMult_1703();
	test_MatMult_1704();
	test_MatMult_1705();
	test_MatMult_1706();
	test_MatMult_1707();
	test_MatMult_1708();
	test_MatMult_1709();
	test_MatMult_1710();
	test_MatMult_1711();
	test_MatMult_1712();
	test_MatMult_1713();
	test_MatMult_1714();
	test_MatMult_1715();
	test_MatMult_1716();
	test_MatMult_1717();
	test_MatMult_1718();
	test_MatMult_1719();
	test_MatMult_1720();
	test_MatMult_1721();
	test_MatMult_1722();
	test_MatMult_1723();
	test_MatMult_1724();
	test_MatMult_1725();
	test_MatMult_1726();
	test_MatMult_1727();
	oski_Close();
	return 0; /* exit with no errors */
}

/* eof */
